# -*- coding: latin-1 -*-
# J1939 data decodes (modified from TURP1210)

J1939BitDecodings = {
        1081: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available"
        },
        1082: {
            0: "no coolant load increase",
            1: "coolant load increase possible",
            2: "error",
            3: "not available"
        },
        1107: {
            0: "inactive",
            1: "active",
            2: "error",
            3: "not available"
        },
        1108: {
            0: "inactive",
            1: "active",
            2: "error",
            3: "not available"
        },
        1109: {
            0: "not approaching",
            1: "approaching",
            2: "error",
            3: "not available"
        },
        1110: {
            0: "no",
            1: "yes",
            2: "error",
            3: "not available"
        },
        1111: {
            0: "disabled in calibration",
            1: "enabled in calibration",
            2: "error",
            3: "not available"
        },
        1121: {
            0: "brake pedal is not being pressed",
            1: "brake pedal is being pressed",
            2: "error",
            3: "not available"
        },
        1194: {
            0: "random number is not present",
            1: "random number is present",
            2: "error",
            3: "not available"
        },
        1195: {
            0: "password is not a validated password",
            1: "password is a validated password",
            2: "error",
            3: "not available"
        },
        1196: {
            0: "b -  unlocked = this state indicates that the component can be started without the end user being required to enter a password.",
            1: "b - locked = this state indicates that the component can not be started (i.e., unlocked) without the end user being required to enter a password.",
            2: "b - blocked = this state indicates that a lock or unlock command cannot be executed because some other algorithm or command of higher priority is commanding differently.",
            3: " not defined"
        },
        1197: {
            0: "b - ok = this state indicates that the request was successfully performed.",
            1: "b - full_of_passwords = this state indicates that the component can not store any additional passwords in its memory.",
            2: "b - empty_of_passwords = this state indicates that the component would be empty of passwords (an unacceptable condition) if the password under which the end user is logged in, is deleted.  thus the delete password command is not successfully executed.",
            3: "b -  not_valid = this state indicates that the request is not a valid one."
        },
        1199: {
            0: "b - encryption_seed_request = this state represents a request to the component to provide a random number seed.",
            1: "b - encrypted_code_present = this state is used to indicate that an encrypted password is being provided to the component.",
            2: " not defined",
            3: "b - not_available = this state is used to indicate that a random number is not being requested nor is an encrypted password being provided to the component."
        },
        1200: {
            0: "b -   lock_upon_operator_request-this state is used to indicate that the end user would have to manually enter a password to lock the engine.",
            1: "b -  lock_when_keyoff-this state is used to indicate that the component would automatically transition to the locked state when the end user turns off the engine (i.e. without the end user being required to manually enter the password).",
            2: " not defined",
            3: "b - not_available-this state is indicates that the option is not selectable or changeable by the operator via using current tool."
        },
        1201: {
            0: "b - add_password-this state represents a request to the component to add a password to the list of passwords that the component has stored as valid codes.  this command will not be performed if the component has already stored, the maximum number of passwords that it is capable of storing.  the login command must precede this command.",
            1: "b - delete_password-this state represents a request to the component to delete the password (the same one used when the end-user logged in).  see spn 1197 for limitations.",
            2: "b - change_password-this state represents a request to the component to change the password (the same one that the end-user logged in with) to a different password, which is to be specified by the end user.  the login command must precede this command.",
            3: "b - lock_or_unlock-this state represents a request to the component to change from the locked state to the unlocked state or from the unlocked state to the locked state.",
            4: "b - check_status-this state represents a request to check to see if the component is in the locked or unlocked state.",
            5: "b -  login-this state represents a request to validate the end user, before performing commands such as add_password and change_password.",
            6: " not defined",
            7: " not defined"
        },
        1205: {
            0: "safety wire has not been activated",
            1: "safety wire has been activated",
            2: "error",
            3: "not available"
        },
        1206: {
            0: "not engaged",
            1: "engaged",
            2: "error",
            3: "not available"
        },
        1211: {
            0: " do not reset",
            1: " reset",
            2: " reserved",
            3: " take no action"
        },
        1213: {
            0: "lamp off",
            1: "lamp on",
            2: "short mil for wwh obd",
            3: "not available"
        },
        1230: {
            0: " stop broadcast",
            1: " start broadcast",
            2: " reserved",
            3: " don-t care/take no action (leave as is)"
        },
        1231: {
            0: " stop broadcast",
            1: " start broadcast",
            2: " reserved",
            3: " don-t care/take no action (leave as is)"
        },
        1232: {
            0: " stop broadcast",
            1: " start broadcast",
            2: " reserved",
            3: " don-t care/take no action (leave as is)"
        },
        1233: {
            0: " stop broadcast",
            1: " start broadcast",
            2: " reserved",
            3: " don-t care/take no action (leave as is)"
        },
        1234: {
            0: " stop broadcast",
            1: " start broadcast",
            2: " reserved",
            3: " don-t care/take no action (leave as is)"
        },
        1235: {
            0: " stop broadcast",
            1: " start broadcast",
            2: " reserved",
            3: " don-t care/take no action (leave as is)"
        },
        1236: {
            0: "  all devices",
            1: "  devices whose broadcast state has been modified",
            15: " not avai l abl e",
            2: " to 1110 reserved"
        },
        1237: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available"
        },
        1238: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available or not installed"
        },
        1239: {
            0: "no leakage detected",
            1: "leakage detected",
            2: "error",
            3: "not available"
        },
        1240: {
            0: "no leakage detected",
            1: "leakage detected",
            2: "error",
            3: "not available"
        },
        1243: {
            0: "not fully operational",
            1: "fully operational",
            2: "reserved",
            3: "not available"
        },
        1254: {
            0: "disabled",
            1: "enabled",
            2: "error",
            3: "not available"
        },
        1351: {
            0: "     compressor not active",
            1: "     compressor active",
            2: "     error indicator",
            3: "     not available"
        },
        1377: {
            0: " functionality is disabled (off)",
            1: " functionality is enabled (on)",
            2: " error",
            3: " not available or unused"
        },
        1437: {
            0: "active",
            1: "not active",
            2: "error",
            3: "not available"
        },
        1438: {
            0: " off",
            1: " on",
            2: " reserved",
            3: " take no action"
        },
        1439: {
            0: " off",
            1: " on",
            2: " reserved",
            3: " take no action"
        },
        1466: {
            0: " maintain",
            1: " inflate",
            10: " reserved",
            11: " reserved",
            12: " reserved",
            13: " reserved",
            14: " error condition",
            15: " not available",
            2: " deflate",
            3: " confirm",
            4: " inflate wait - system will inflate when conditions allow",
            5: " deflate wait - system will deflate when conditions allow",
            6: " pressure check",
            7: " reserved",
            8: " reserved",
            9: " reserved"
        },
        1467: {
            0: " maintain",
            1: " inflate",
            10: " reserved",
            11: " reserved",
            12: " reserved",
            13: " reserved",
            14: " error condition",
            15: " not available",
            2: " deflate",
            3: " confirm",
            4: " inflate wait - system will inflate when conditions allow",
            5: " deflate wait - system will deflate when conditions allow",
            6: " pressure check",
            7: " reserved",
            8: " reserved",
            9: " reserved"
        },
        1468: {
            0: " maintain",
            1: " inflate",
            10: " reserved",
            11: " reserved",
            12: " reserved",
            13: " reserved",
            14: " error condition",
            15: " not available",
            2: " deflate",
            3: " confirm",
            4: " inflate wait - system will inflate when conditions allow",
            5: " deflate wait - system will deflate when conditions allow",
            6: " pressure check",
            7: " reserved",
            8: " reserved",
            9: " reserved"
        },
        1469: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available"
        },
        1470: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available"
        },
        1471: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available"
        },
        1472: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available"
        },
        1473: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available"
        },
        1474: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available"
        },
        1475: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available"
        },
        1497: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available or not installed"
        },
        1498: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available or not installed"
        },
        1499: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available or not installed"
        },
        1503: {
            0: "off",
            1: "on",
            2: "reserved",
            3: "don't care/take no action"
        },
        1519: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available or not installed"
        },
        1520: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available or not installed"
        },
        1521: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available or not installed"
        },
        1522: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available or not installed"
        },
        1523: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available or not installed"
        },
        1524: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available or not installed"
        },
        1525: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available or not installed"
        },
        1526: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available or not installed"
        },
        1527: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available or not installed"
        },
        1528: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available or not installed"
        },
        1529: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available or not installed"
        },
        1530: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available or not installed"
        },
        1531: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available or not installed"
        },
        1532: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available or not installed"
        },
        1557: {
            0: " fan off",
            1: " engine system-general",
            10: " transmission retarder",
            11: " a/c system",
            12: " timer",
            13: " engine brake",
            14: " other",
            15: " not available",
            2: " excessive engine air temperature",
            3: " excessive engine oil temperature",
            4: " excessive engine coolant temperature",
            5: " excessive transmission oil temperature",
            6: " excessive hydraulic oil temperature",
            7: " default operation",
            8: " reverse operation",
            9: " manual control"
        },
        1573: {
            128: " low coarse",
            129: " low medium/coarse",
            130: " low medium",
            131: " low medium/fine",
            132: " low fine",
            144: " on grade",
            145: " high fine",
            146: " high medium/fine",
            147: " high medium",
            148: " high medium/coarse",
            149: " high coarse",
            159: " no data (i.e. no laser strikes)",
            16: " low fine led on",
            2: " high coarse led on",
            254: " error",
            255: " not available",
            32: " low coarse led on",
            4: " high fine led on",
            8: " on-grade led on"
        },
        1578: {
            0: " manual mode",
            1: " automatic mode",
            2: " inactive automatic mode"
        },
        1582: {
            1: " on-grade -a- led on",
            128: " low coarse",
            129: " low medium/coarse",
            130: " low medium",
            131: " low medium/fine",
            132: " low fine",
            144: " on grade",
            145: " high fine",
            146: " high medium/fine",
            147: " high medium",
            148: " high medium/coarse",
            149: " high coarse",
            159: " no data (i.e. no laser strikes)",
            16: " down led on",
            2: " on-grade -b- led on",
            254: " error",
            255: " not available",
            32: " left led on",
            4: " on-grade -c- led on",
            64: " right led on",
            8: " up led on"
        },
        1583: {
            1: " laser power is on",
            16: " stringline (1 = yes)",
            2: " laser is ready",
            32: " curb (1 = yes)",
            4: " valid target (1 = yes)",
            8: " previous pass (1 = yes)"
        },
        1589: {
            0: " acc distance mode #1 (largest distance)",
            1: " acc distance mode #2",
            2: " acc distance mode #3",
            3: " acc distance mode #4",
            4: " acc distance mode #5 (shortest distance)",
            5: " conventional cruise control mode",
            6: " error condition",
            7: " not available/not valid"
        },
        1590: {
            0: "b - off-used to indicate the acc is enabled in calibration or configuration and there are no faults that would prevent the system from operating.",
            1: "b - speed control active-used to indicate that acc is on but not currently sending control messages.  in other words, there is no target ahead and regular vehicle cruise control is controlling the vehicle speed to the driver-s set speed.",
            2: "b - distance control active-used to indicate that acc is on and actively sending control messages to maintain the appropriate following interval.",
            3: "b - overtake mode-used to indicate that acc is on but temporarily disabled because the driver is manually overriding cruise control by using either the accelerator pedal or the cruise control -accel- switch.",
            4: "b - hold mode-used to indicate that the acc has lost the previous target vehicle and is in hold mode. in this mode, the acc shall limit the speed to the speed held when the target was lost. for example,  if the driver activates the typical cruise buttons (resume/inc/dec) the hold mode shall be exited and normal cruise functionality resumed. if a new target is detected, the distance control active mode (010b) is again entered, unless existing conditions prohibit this.",
            5: "b - finish mode-used to indicate that acc is on with no target ahead, and acc is currently sending control messages to return to the driver-s set speed. this occurs when the target the acc system was tracking moves out of the way so acc returns the vehicle to the driver-s set speed.",
            6: "b - disabled or error condition-used to indicate that acc is in an error state and can not operate.",
            7: " not available/not valid"
        },
        1611: {
            0: " vehicle motion not detected",
            1: " vehicle motion detected",
            2: "error",
            3: "not available"
        },
        1612: {
            0: " rest - sleeping",
            1: " driver available - short break",
            2: " work - loading, unloading, working in an office",
            3: " drive - behind wheel ",
            4: " reserved",
            5: " reserved",
            6: " error",
            7: " not available"
        },
        1613: {
            0: " rest - sleeping",
            1: " driver available - short break",
            2: " work - loading, unloading, working in an office",
            3: " drive - behind wheel ",
            4: " reserved",
            5: " reserved",
            6: " error",
            7: " not available"
        },
        1614: {
            0: " no overspeed",
            1: " overspeed",
            2: " error",
            3: " not available"
        },
        1615: {
            0: "driver card not present",
            1: "driver card present",
            2: "error",
            3: "not available"
        },
        1616: {
            0: "driver card not present",
            1: "driver card present",
            2: "error",
            3: "not available"
        },
        1617: {
            0: " normal/no limits reached",
            1: " limit #1 - 15 min before 4 - h",
            10: " reserved",
            11: " reserved",
            12: " reserved",
            13: " other",
            14: " error",
            15: " not available",
            2: " limit #2 - 4 - h reached",
            3: " limit #3 - 15 min before 9 h",
            4: " limit #4 - 9 h reached",
            5: " limit #5 - 15 min before 16 h (not having 8h rest during the last 24h)",
            6: " limit #6 - 16 h reached",
            7: " reserved",
            8: " reserved",
            9: " reserved"
        },
        1618: {
            0: " normal/no limits reached",
            1: " limit #1 - 15 min before 4 - h",
            10: " reserved",
            11: " reserved",
            12: " reserved",
            13: " other",
            14: " error",
            15: " not available",
            2: " limit #2 - 4 - h reached",
            3: " limit #3 - 15 min before 9 h",
            4: " limit #4 - 9 h reached",
            5: " limit #5 - 15 min before 16 h (not having 8h rest during the last 24h)",
            6: " limit #6 - 16 h reached",
            7: " reserved",
            8: " reserved",
            9: " reserved"
        },
        1619: {
            0: "forward",
            1: "reverse",
            2: "error",
            3: "not available"
        },
        1620: {
            0: "normal performance",
            1: "performance analysis",
            2: "error",
            3: "not available"
        },
        1621: {
            0: "no handling information",
            1: "handling information",
            2: "error",
            3: "not available"
        },
        1622: {
            0: "no tachograph event",
            1: "tachograph event",
            2: "error",
            3: "not available"
        },
        1632: {
            0: " reserved",
            1: " highest torque rating",
            2: " first torque rating",
            3: " previous torque rating (rating prior to the current rating)",
            4: " current torque rating",
            5: " reserved",
            6: " reserved",
            7: " not available"
        },
        1633: {
            0: "off",
            1: "on",
            2: "error indicator",
            3: "take no action"
        },
        1641: {
            0: "   direct memory addressing",
            1: "   direct spatial addressing"
        },
        1653: {
            0: "switch disabled",
            1: "switched enabled",
            2: "error",
            3: "not available"
        },
        1654: {
            0: "switch in the off state",
            1: "switch in the on state - increase",
            2: "error",
            3: "not available"
        },
        1655: {
            0: "switch in the off state",
            1: "switch in the on state - decrease",
            2: "error",
            3: "not available"
        },
        1656: {
            0: "switch in the off state",
            1: "switch in the on state",
            2: "error",
            3: "not available"
        },
        1665: {
            0: "no oil present at turbocharger",
            1: "oil present at turbocharger",
            2: "error",
            3: "not available"
        },
        1666: {
            0: "automated gear shifting is disabled",
            1: "automated gear shifting is enabled",
            2: "error",
            3: "not available"
        },
        1667: {
            0: "retarder is not requesting that the brake lights are illuminated",
            1: "retarder is requesting that the brake lights are illuminated",
            2: "reserved",
            3: "not available/take no action"
        },
        1675: {
            0: "start not requested",
            1: "starter active, gear not engaged",
            10: "starter inhibited due to active scr inducement",
            11: "reserved",
            12: "starter inhibited - reason unknown",
            13: "error (legacy implementation only, use 1110)",
            14: "error",
            15: "not available",
            2: "starter active, gear engaged",
            3: "start finished; starter not active after having been actively engaged (after 50ms mode goes to 0000)",
            4: "starter inhibited due to engine already running",
            5: "starter inhibited due to engine not ready for start (preheating)",
            6: "starter inhibited due to driveline engaged or other transmission inhibit",
            7: "starter inhibited due to active immobilizer",
            8: "starter inhibited due to starter over-temp",
            9: "starter inhibited due to intake air shutoff valve being active"
        },
        1676: {
            0: " water pump is not running",
            1: " water pump is  running",
            2: " reserved",
            3: " not available"
        },
        1677: {
            0: " heater not active",
            1: " off due to adr per european regulations for transport of hazardous materials",
            10: " not defined",
            11: " not defined",
            12: " not defined",
            13: " not defined",
            14: " error",
            15: " not available",
            2: " economy mode",
            3: " normal mode",
            4: " not defined",
            5: " not defined",
            6: " not defined",
            7: " not defined",
            8: " not defined",
            9: " not defined"
        },
        1678: {
            0: " cab not ventilated",
            1: " cab is ventilated",
            2: " reserved",
            3: " not available"
        },
        1679: {
            0: " engine heating zone off",
            1: " engine heating zone on",
            2: " reserved",
            3: " not available"
        },
        1680: {
            0: " cab heating zone off",
            1: " cab heating zone on",
            2: " reserved",
            3: " not available"
        },
        1681: {
            0: " battery main switch not held",
            1: " battery main switch held",
            2: " reserved",
            3: " don't care/take no action"
        },
        1682: {
            0: " release battery main switch",
            1: " hold battery main switch",
            2: " undefined",
            3: " don't care/take no action"
        },
        1683: {
            0: " de-activate auxiliary heater",
            1: " off due to adr per european regulations for transport of hazardous materials",
            10: " not defined",
            11: " not defined",
            12: " not defined",
            13: " not defined",
            14: " reserved",
            15: " don't care/take no action",
            2: " economy mode",
            3: " normal mode",
            4: " not defined",
            5: " not defined",
            6: " not defined",
            7: " not defined",
            8: " not defined",
            9: " not defined"
        },
        1684: {
            0: " deactivate water pump",
            1: " activate water pump",
            2: " reserved",
            3: " don't care/take no action"
        },
        1685: {
            0: " do not heat engine zone",
            1: " heat engine zone",
            2: " reserved",
            3: " don't care/take no action"
        },
        1686: {
            0: " do not cab engine zone",
            1: " heat cab zone",
            2: " reserved",
            3: " don't care/take no action"
        },
        1696: {
            0: " engine not using the exhaust gas oxygen sensor (open loop)",
            1: " engine using the exhaust gas oxygen sensor for a closed loop operation",
            2: " reserved",
            3: " don't care/take no action"
        },
        1697: {
            0: " ok ( no fault)",
            1: " not defined",
            2: " error",
            3: " not supported"
        },
        1698: {
            0: " ok (no fault)",
            1: " tire leak detected",
            2: " error",
            3: " not supported"
        },
        1699: {
            0: " off / isolated from cti pressure controller",
            1: " on (tire is polled)",
            2: " not defined",
            3: " not supported"
        },
        1700: {
            0: " not imminent",
            1: " imminent",
            2: " reserved",
            3: " not used"
        },
        1701: {
            0: " not imminent",
            1: " imminent",
            2: " reserved",
            3: " not used"
        },
        1702: {
            0: " lane departure indication disabled",
            1: " lane departure indication enabled ",
            2: " reserved",
            3: " not used"
        },
        1706: {
            0: " means convert spns per the version 4 definition below ",
            1: ".  spn assumed to be sent most significant bit first",
            2: ".  spn represented as intel format for most significant 16 bits  with 3 least significant bits of 19 bits in with fmi value.",
            3: ".  spn represented as intel format for all 19 bits (least significant sent first) ",
            4: ".  spn represented as intel format for all 19 bits with the spn conversion method set to 0."
        },
        1710: {
            0: " not tracking left side",
            1: " tracking left side",
            2: " reserved",
            3: " don't care/take no action"
        },
        1711: {
            0: " not tracking right side",
            1: " tracking right side",
            2: " reserved",
            3: " don't care/take no action"
        },
        1713: {
            0: " no restriction",
            1: " restriction exists on oil filter",
            2: " error",
            3: " not available"
        },
        1714: {
            0: " operator seat not facing forward",
            1: " operator seat is facing forward",
            2: " error",
            3: " not available"
        },
        1733: {
            0: " level not specified,(i.e. the nominal level is none of the specified levels, no error condition)",
            1: " \"normal level 1,(i.e. the level prescribed for normal driving, given by design)",
            10: " lowering",
            11: " reserved",
            12: " reserved",
            13: " reserved",
            14: " error",
            15: " not available",
            2: " \"normal level 2,(i.e. a level permitted for driving, for example to lower the vehicle in case of high speed)",
            3: " \"normal level 3,(i.e. a level permitted for driving, for example to lift the vehicle in case of offroad)",
            4: " \"preset level,(i.e. a level to be defined externally via can)",
            5: " \"customer level,(i.e. a level to be defined by customer via parameter setting)",
            6: " \"upper level,(i.e. the highest level to be controlled)",
            7: " \"lower level,(i.e. the lowest level to be controlled)",
            8: " level change stopped",
            9: " raising"
        },
        1734: {
            0: " level not specified,(i.e. the nominal level is none of the specified levels, no error condition)",
            1: " \"normal level 1,(i.e. the level prescribed for normal driving, given by design)",
            10: " lowering",
            11: " reserved",
            12: " reserved",
            13: " reserved",
            14: " error",
            15: " not available",
            2: " \"normal level 2,(i.e. a level permitted for driving, for example to lower the vehicle in case of high speed)",
            3: " \"normal level 3,(i.e. a level permitted for driving, for example to lift the vehicle in case of offroad)",
            4: " \"preset level,(i.e. a level to be defined externally via can)",
            5: " \"customer level,(i.e. a level to be defined by customer via parameter setting)",
            6: " \"upper level,(i.e. the highest level to be controlled)",
            7: " \"lower level,(i.e. the lowest level to be controlled)",
            8: " level change stopped",
            9: " raising"
        },
        1736: {
            0: " not above",
            1: " above",
            2: " error",
            3: " not available"
        },
        1737: {
            0: " not above",
            1: " above",
            2: " error",
            3: " not available"
        },
        1738: {
            0: " not below",
            1: " below",
            2: " error",
            3: " not available"
        },
        1739: {
            0: " lifting not active",
            1: " lifting active",
            2: " error",
            3: " not available"
        },
        1740: {
            0: " lowering not active",
            1: " lowering active",
            2: " error",
            3: " not available"
        },
        1741: {
            0: "    normal operation; i.e. the system performs a \"\"pure\"\" control of the vehicle height",
            1: "    traction help (load transfer); i.e. the driven axle is loaded to a maximum value given by legislation or design",
            10: "    automatic air suspension control prohibited",
            11: "    force to bumper level request is in effect",
            12: "    not defined",
            13: "    not defined",
            14: "    error",
            15: "    not available",
            2: "    load fixing; i.e. the driven axlen is loaded to a value defined by the driver",
            3: "    pressure ratio 1; i.e. the ratio between the pressures at the driven axle and at the third axle is controlled, so that the ratio equals a fixed value 1",
            4: "    pressure ratio 2; i.e. the ratio between the pressures at the driven axle and at the third axle is controlled, so that the ratio equals a fixed value 2",
            5: "    optimum traction 1; i.e. the pressure at the driven axle is controlled at a fixed value 1",
            6: "    optimum traction 2; i.e. the pressure at the driven axle is controlled at a fixed value 2",
            7: "    traction help - load reduce; (i.e. the driven axle load is reduced to normal load condition)",
            8: "    exhausting bellow function; i.e. the bellows are exhausted totally",
            9: "    air suspension control prohibited ( i.e. hold current pressures in all suspension devices.)"
        },
        1742: {
            0: " not active,(i.e. the kneeling function is not active\")",
            1: "lowering active,(i.e. the vehicle is lowered due to a kneeling request)",
            10: " not defined",
            11: " not defined",
            12: " not defined",
            13: " not defined",
            14: " error",
            15: " not available",
            2: " kneeling level reached,(i.e. the vehicle is at the fixed kneeling level)",
            3: " lifting active,(i.e. the vehicle is lifted due to a recover request)",
            4: " kneeling aborted,(i.e. in case of manual actuation the request was dropped before the kneeling level was reached)",
            5: " not defined",
            6: " not defined",
            7: " not defined",
            8: " not defined",
            9: " not defined"
        },
        1743: {
            0: " lift axle position down / tag axle laden",
            1: " lift axle position up / tag axle unladen",
            2: " error",
            3: " not available"
        },
        1744: {
            0: " doors may not be opened",
            1: " doors may be opened",
            2: " error",
            3: " not available"
        },
        1745: {
            0: " vehicle may be moved",
            1: " vehicle motion is inhibited",
            2: " error",
            3: " not available"
        },
        1746: {
            0: " not active",
            1: " active",
            2: " error",
            3: " not available"
        },
        1747: {
            0: " automatically actuated",
            1: " manually actuated",
            2: " reserved",
            3: " don't care/take no action"
        },
        1748: {
            0: " no kneeling request",
            1: " kneeling request",
            2: " reserved",
            3: " don't care/take no action"
        },
        1749: {
            0: " no kneeling request",
            1: " kneeling request",
            2: " reserved",
            3: " don't care/take no action"
        },
        1750: {
            0: " no level request",
            1: " normal level 1,(i.e. the level prescribed for normal driving, given by design)",
            10: "  lower request (continuously lowering front axle, as long as active).",
            11: " not defined",
            12: " not defined",
            13: " not defined",
            14: " reserved",
            15: " don't care/take no action",
            2: " normal level 2,(i.e. a level permitted for driving, for example to lower the vehicle in case of high speed)",
            3: " normal level 3,(i.e. a level permitted for driving, for example to lift the vehicle in case of offroad)",
            4: " preset level,(i.e. a level to be defined externally via can)",
            5: " customer level,(i.e. a level to be defined by customer via parameter setting)",
            6: " upper level,(i.e. the highest level to be controlled)",
            7: " lower level,(i.e. the lowest level to be controlled)",
            8: " stop level change,(i.e. the level change in process shall be stopped immediately)",
            9: "  raise request (continuously raising front axle, as long as active)."
        },
        1751: {
            0: " no level request",
            1: " normal level 1,(i.e. the level prescribed for normal driving, given by design)",
            10: "  lower request (continuously lowering front axle, as long as active).",
            11: " not defined",
            12: " not defined",
            13: " not defined",
            14: " reserved",
            15: " don't care/take no action",
            2: " normal level 2,(i.e. a level permitted for driving, for example to lower the vehicle in case of high speed)",
            3: " normal level 3,(i.e. a level permitted for driving, for example to lift the vehicle in case of offroad)",
            4: " preset level,(i.e. a level to be defined externally via can)",
            5: " customer level,(i.e. a level to be defined by customer via parameter setting)",
            6: " upper level,(i.e. the highest level to be controlled)",
            7: " lower level,(i.e. the lowest level to be controlled)",
            8: " stop level change,(i.e. the level change in process shall be stopped immediately)",
            9: "  raise request (continuously raising front axle, as long as active)."
        },
        1752: {
            0: " lift axle position down / tag axle laden",
            1: " lift axle position up / tag axle unladen",
            2: " reserved",
            3: " don't care/take no action"
        },
        1753: {
            0: "    normal operation; i.e. the system performs a \"\"pure\"\" control of the vehicle height",
            1: "    traction help (load transfer); i.e. the driven axle is loaded to a maximum value given by legislation or design",
            10: "    not defined",
            11: "    not defined",
            12: "    not defined",
            13: "    not defined",
            14: "    reserved",
            15: "    don't care/take no action",
            2: "    load fixing; i.e. the driven axle is loaded to a value defined by the driver",
            3: "    pressure ratio 1; i.e. the ratio between the pressures at the driven axle and at the third axle is controlled, so that the ratio equals a fixed value 1",
            4: "    pressure ratio 2; i.e. the ratio between the pressures at the driven axle and at the third axle is controlled, so that the ratio equals a fixed value 2",
            5: "    optimum traction 1; i.e. the pressure at the driven axle is controlled at a fixed value 1",
            6: "    optimum traction 2; i.e. the pressure at the driven axle is controlled at a fixed value 2",
            7: "    traction help - load reduce; (i.e. reduce axle load of driven axle to normal load condition)",
            8: "    exhaust bellows",
            9: "    force to bumper level"
        },
        1754: {
            0: " not below",
            1: " below",
            2: " error",
            3: " not available"
        },
        1755: {
            0: " lowering not active",
            1: " lowering active",
            2: " error",
            3: " not available"
        },
        1756: {
            0: " lifting not active",
            1: " lifting active",
            2: " error",
            3: " not available"
        },
        1763: {
            0: " disabled",
            1: " enabled",
            2: " error indicator",
            3: " not available"
        },
        1764: {
            0: " pressure mode inactive",
            1: " pressure mode active",
            2: " error indicator",
            3: " not available"
        },
        1792: {
            0: " off",
            1: " on ",
            2: " reserved",
            3: " take no action"
        },
        1793: {
            0: " off",
            1: " on ",
            2: " reserved",
            3: " take no action"
        },
        1796: {
            0: " acc das not active",
            1: " acc das active",
            2: " reserved",
            3: " take no action"
        },
        1797: {
            0: " acc ssow not active",
            1: " acc ssow active",
            2: " reserved",
            3: " take no action"
        },
        1798: {
            0: " no targets detected",
            1: " target detected",
            2: " reserved",
            3: " take no action"
        },
        1799: {
            0: " requested acc distance mode #1 (largest distance)",
            1: " requested acc distance mode #2",
            2: " requested acc distance mode #3",
            3: " requested acc distance mode #4",
            4: " requested acc distance mode #5 (shortest distance)",
            5: " not defined",
            6: " error condition",
            7: " not available"
        },
        1804: {
            0: "start enable off",
            1: "start enable on",
            2: "reserved",
            3: "not available"
        },
        1805: {
            0: "center on-grade display mode 1 ( 5 channel )",
            1: "offset on-grade display mode ",
            10: " reserved",
            11: " reserved",
            12: " reserved",
            13: " reserved",
            14: " reserved",
            15: "not available or not applicable",
            2: "center on-grade display mode 2 ( 7 channel )",
            3: " reserved",
            4: " reserved",
            5: " reserved",
            6: " reserved",
            7: " reserved",
            8: " reserved",
            9: " reserved"
        },
        1806: {
            0: "    +/- 4.5 mm  (0.015 feet, 0.18 inches)",
            1: "    +/- 12 mm   (0.040 feet, 0.45 inches)",
            10: "  reserved",
            11: "  reserved",
            12: "  reserved",
            13: "  reserved",
            14: "  reserved",
            15: "    not available or not applicable",
            2: "    +/- 24 mm   (0.080 feet, 0.96 inches)",
            3: "    +/- 5  mm    (0.017 feet, 0.2 inches)",
            4: "    +/- 1 mm     (0.003 feet, 0.04 inches)",
            5: "  reserved",
            6: "  reserved",
            7: "  reserved",
            8: "  reserved",
            9: "  reserved"
        },
        1812: {
            0: " relative measuring principle",
            1: " absolute measuring principle",
            2: " reserved",
            3: " not available"
        },
        1813: {
            0: " off",
            1: " on",
            2: " reserved",
            3: " don't care/take no action"
        },
        1814: {
            0: " not fully operational",
            1: " fully operational",
            2: " reserved",
            3: " don't care/take no action"
        },
        1815: {
            0: " turn brake light not on",
            1: " turn brake light on",
            2: " reserved",
            3: " don't care/take no action"
        },
        1816: {
            0: " rop engine control passive but installed",
            1: " rop engine control active",
            2: " reserved",
            3: " don't care/take no action"
        },
        1817: {
            0: " yc engine control passive but installed",
            1: " yc engine control active",
            2: " reserved",
            3: " don't care/take no action"
        },
        1818: {
            0: " rop brake control passive but installed",
            1: " rop brake control active",
            2: " reserved",
            3: " don't care/take no action"
        },
        1819: {
            0: " yc brake control passive but installed",
            1: " yc brake control active",
            2: " reserved",
            3: " don't care/take no action"
        },
        1820: {
            0: " inside bus",
            1: " outside bus",
            2: " error",
            3: " not available"
        },
        1821: {
            0: " at least 1 door is open",
            1: " closing last door",
            10: " not defined",
            11: " not defined",
            12: " not defined",
            13: " not defined",
            14: " error",
            15: " not available",
            2: " all doors closed",
            3: " not defined",
            4: " not defined",
            5: " not defined",
            6: " not defined",
            7: " not defined",
            8: " not defined",
            9: " not defined"
        },
        1822: {
            0: " lift axle position down / tag axle laden",
            1: " lift axle position up / tag axle unladen",
            2: " error",
            3: " not available"
        },
        1823: {
            0: " actual level out of bumper range",
            1: " actual level within bumper range",
            2: " error",
            3: " not available"
        },
        1824: {
            0: " actual level out of bumper range",
            1: " actual level within bumper range",
            2: " error",
            3: " not available"
        },
        1825: {
            0: " not active",
            1: " active",
            2: " error",
            3: " not available"
        },
        1826: {
            0: " not active",
            1: " active",
            2: " error",
            3: " not available"
        },
        1827: {
            0: " actual request not refused",
            1: " axle load limit reached  (load transfer)",
            10: "  not defined",
            11: "  not defined",
            12: "  not defined",
            13: "  not defined",
            14: " error",
            15: " not available",
            2: " would exceed axle load limit (tag axle)",
            3: " bogie differential not locked",
            4: " above speed limit",
            5: " below speed limit",
            6: " general reject; i.e. no specified reason applies",
            7: " requested level not available",
            8: "  not defined",
            9: "  not defined"
        },
        1828: {
            0: " lift axle position down / tag axle laden",
            1: " lift axle position up / tag axle unladen",
            2: " reserved",
            3: " don't care/take no action"
        },
        1829: {
            0: " deactivate kneeling",
            1: " activate kneeling",
            2: " reserved",
            3: " don't care/take no action"
        },
        1830: {
            0: " deactivate kneeling",
            1: " activate kneeling",
            2: " reserved",
            3: " don't care/take no action"
        },
        1831: {
            0: " normal operation dampers passive",
            1: " normal operation dampers active",
            2: " error",
            3: " not available"
        },
        1832: {
            0: " normal operation dampers passive",
            1: " normal operation dampers active",
            2: " error",
            3: " not available"
        },
        1833: {
            0: " normal operation dampers passive",
            1: " normal operation dampers active",
            2: " error",
            3: " not available"
        },
        1836: {
            0: " trailer abs status information available but not active",
            1: " trailer abs active",
            2: " reserved",
            3: " trailer abs status information not available or parameter not supported"
        },
        1837: {
            0: " off",
            1: " on",
            2: " reserved",
            3: " not supported"
        },
        1838: {
            0: " off",
            1: " on",
            2: " reserved",
            3: " not supported"
        },
        1839: {
            0: " off",
            1: " on",
            2: " reserved",
            3: " not supported"
        },
        1840: {
            0: " off",
            1: " on",
            2: " reserved",
            3: " not supported"
        },
        1841: {
            0: " off",
            1: " on",
            2: " reserved",
            3: " not supported"
        },
        1842: {
            0: " off",
            1: " on",
            2: " reserved",
            3: " not supported"
        },
        1843: {
            0: " off",
            1: " on",
            2: " reserved",
            3: " not supported"
        },
        1849: {
            0: " inactive; transmission requested range display should not be flashing",
            1: " active; transmission requested range display should be flashing",
            2: " reserved",
            3: " take no action"
        },
        1850: {
            0: " inactive; transmission requested range display should not be blanked",
            1: " active; transmission requested range display should be blanked",
            2: " reserved",
            3: " take no action"
        },
        1851: {
            0: " inactive; shift is not inhibited",
            1: " active (on continuously); shift is inhibited",
            2: " active (flashing)",
            3: " take no action"
        },
        1852: {
            0: " disable",
            1: " enable",
            2: " reserved",
            3: " take no action"
        },
        1853: {
            0: " disable",
            1: " enable",
            2: " reserved",
            3: " take no action"
        },
        1854: {
            0: " disable",
            1: " enable",
            2: " reserved",
            3: " take no action"
        },
        1855: {
            0: " disable",
            1: " enable",
            2: " reserved",
            3: " take no action"
        },
        1856: {
            0: " not buckled",
            1: " ok - seat belt is buckled",
            2: " error - switch state cannot be determined",
            3: " not available"
        },
        1857: {
            0: " not ok- oil pressure is too low",
            1: " ok - oil pressure is above minimum",
            2: " error - switch state cannot be determined",
            3: " not available"
        },
        1861: {
            0: " reverse",
            1: " forward",
            2: " error indication",
            3: " not available"
        },
        1864: {
            0: " reverse",
            1: " forward",
            2: " error indication",
            3: " not available"
        },
        1865: {
            0: " key switch off",
            1: " key switch not off",
            2: " error indication",
            3: " not available"
        },
        1867: {
            0: " no further requirement for ecu_pwr",
            1: " requirement for 2 more seconds of ecu_pwr",
            2: " reserved ",
            3: " don-t care"
        },
        1868: {
            0: " no further requirement for pwr",
            1: " requirement for 2 more seconds of pwr",
            2: " reserved ",
            3: " don-t care"
        },
        1869: {
            0: " implement may not be transported ",
            1: " implement may be transported",
            2: " error indication",
            3: " not available"
        },
        1870: {
            0: " implement may not be disconnected ",
            1: " implement may be disconnected",
            2: " error indication",
            3: " not available"
        },
        1871: {
            0: " implement is not ready for field work ",
            1: " implement is ready for field work",
            2: " error indication",
            3: " not available"
        },
        1876: {
            0: " hitch position is out-of-work ",
            1: " hitch position is in-work",
            2: " error indication",
            3: " not available"
        },
        1877: {
            0: " hitch position is out-of-work ",
            1: " hitch position is in-work",
            2: " error indication",
            3: " not available"
        },
        1888: {
            0: " power take off disengaged",
            1: " power take off engaged",
            2: " error indication",
            3: " not available"
        },
        1889: {
            0: " power take off mode is 540",
            1: " power take off mode is 1000",
            2: " error indication",
            3: " not available"
        },
        1890: {
            0: " power take off mode is 540",
            1: " power take off mode is 1000",
            2: " error indication",
            3: " not available"
        },
        1891: {
            0: " power take off economy mode is disengaged",
            1: " power take off economy mode is engaged",
            2: " error indication",
            3: " not available"
        },
        1892: {
            0: " power take off economy mode is disengaged",
            1: " power take off economy mode is engaged",
            2: " error indication",
            3: " not available"
        },
        1893: {
            0: " disengage power take off",
            1: " engage power take off ",
            2: " reserved",
            3: " don-t care"
        },
        1894: {
            0: " disengage power take off",
            1: " engage power take off ",
            2: " reserved",
            3: " don-t care"
        },
        1895: {
            0: " select 540 rpm mode",
            1: " select 1000 rpm mode ",
            2: " reserved",
            3: " don-t care"
        },
        1896: {
            0: " select 540 rpm mode",
            1: " select 1000 rpm mode ",
            2: " reserved",
            3: " don-t care"
        },
        1897: {
            0: " disengage power take off-s economy mode",
            1: " engage power take off-s economy mode",
            2: " reserved",
            3: " don-t care"
        },
        1898: {
            0: " disengage power take off-s economy mode",
            1: " engage power take off-s economy mode",
            2: " reserved",
            3: " don-t care"
        },
        1903: {
            0: " blocked",
            1: " extend",
            10: " reserved",
            11: " reserved",
            12: " reserved",
            13: " reserved",
            14: " reserved",
            15: " don-t care",
            2: " retract",
            3: " floating",
            4: " reserved",
            5: " reserved",
            6: " reserved",
            7: " reserved",
            8: " reserved",
            9: " reserved"
        },
        1908: {
            0: " block",
            1: " extend",
            10: " reserved",
            11: " reserved",
            12: " reserved",
            13: " reserved",
            14: " reserved",
            15: " don-t care",
            2: " retract",
            3: " float",
            4: " reserved",
            5: " reserved",
            6: " reserved",
            7: " reserved",
            8: " reserved",
            9: " reserved"
        },
        1909: {
            0: " block",
            1: " float",
            2: " reserved",
            3: " don-t care"
        },
        1910: {
            0: " block",
            1: " float",
            2: " reserved   <=== ???? should be error indicator since measured ???? =====",
            3: " don-t care"
        },
        1915: {
            0: " blocked",
            1: " extend",
            10: " reserved",
            11: " reserved",
            12: " reserved",
            13: " reserved",
            14: " reserved",
            15: " don-t care",
            2: " retract",
            3: " floating",
            4: " reserved",
            5: " reserved",
            6: " reserved",
            7: " reserved",
            8: " reserved",
            9: " reserved"
        },
        1920: {
            0: " block",
            1: " extend",
            10: " reserved",
            11: " reserved",
            12: " reserved",
            13: " reserved",
            14: " reserved",
            15: " don-t care",
            2: " retract",
            3: " float",
            4: " reserved",
            5: " reserved",
            6: " reserved",
            7: " reserved",
            8: " reserved",
            9: " reserved"
        },
        1921: {
            0: " block",
            1: " float",
            2: " reserved",
            3: " don-t care"
        },
        1922: {
            0: " block",
            1: " float",
            2: " reserved   <=== ???? should be error indicator since measured ???? =====",
            3: " don-t care"
        },
        1927: {
            0: " blocked",
            1: " extend",
            10: " reserved",
            11: " reserved",
            12: " reserved",
            13: " reserved",
            14: " reserved",
            15: " don-t care",
            2: " retract",
            3: " floating",
            4: " reserved",
            5: " reserved",
            6: " reserved",
            7: " reserved",
            8: " reserved",
            9: " reserved"
        },
        1932: {
            0: " block",
            1: " extend",
            10: " reserved",
            11: " reserved",
            12: " reserved",
            13: " reserved",
            14: " reserved",
            15: " don-t care",
            2: " retract",
            3: " float",
            4: " reserved",
            5: " reserved",
            6: " reserved",
            7: " reserved",
            8: " reserved",
            9: " reserved"
        },
        1933: {
            0: " block",
            1: " float",
            2: " reserved",
            3: " don-t care"
        },
        1934: {
            0: " block",
            1: " float",
            2: " reserved   <=== ???? should be error indicator since measured ???? =====",
            3: " don-t care"
        },
        1939: {
            0: " blocked",
            1: " extend",
            10: " reserved",
            11: " reserved",
            12: " reserved",
            13: " reserved",
            14: " reserved",
            15: " don-t care",
            2: " retract",
            3: " floating",
            4: " reserved",
            5: " reserved",
            6: " reserved",
            7: " reserved",
            8: " reserved",
            9: " reserved"
        },
        1944: {
            0: " block",
            1: " extend",
            10: " reserved",
            11: " reserved",
            12: " reserved",
            13: " reserved",
            14: " reserved",
            15: " don-t care",
            2: " retract",
            3: " float",
            4: " reserved",
            5: " reserved",
            6: " reserved",
            7: " reserved",
            8: " reserved",
            9: " reserved"
        },
        1945: {
            0: " block",
            1: " float",
            2: " reserved",
            3: " don-t care"
        },
        1946: {
            0: " block",
            1: " float",
            2: " reserved   <=== ???? should be error indicator since measured ???? =====",
            3: " don-t care"
        },
        1951: {
            0: " blocked",
            1: " extend",
            10: " reserved",
            11: " reserved",
            12: " reserved",
            13: " reserved",
            14: " reserved",
            15: " don-t care",
            2: " retract",
            3: " floating",
            4: " reserved",
            5: " reserved",
            6: " reserved",
            7: " reserved",
            8: " reserved",
            9: " reserved"
        },
        1956: {
            0: " block",
            1: " extend",
            10: " reserved",
            11: " reserved",
            12: " reserved",
            13: " reserved",
            14: " reserved",
            15: " don-t care",
            2: " retract",
            3: " float",
            4: " reserved",
            5: " reserved",
            6: " reserved",
            7: " reserved",
            8: " reserved",
            9: " reserved"
        },
        1957: {
            0: " block",
            1: " float",
            2: " reserved",
            3: " don-t care"
        },
        1958: {
            0: " block",
            1: " float",
            2: " reserved   <=== ???? should be error indicator since measured ???? =====",
            3: " don-t care"
        },
        1963: {
            0: " blocked",
            1: " extend",
            10: " reserved",
            11: " reserved",
            12: " reserved",
            13: " reserved",
            14: " reserved",
            15: " don-t care",
            2: " retract",
            3: " floating",
            4: " reserved",
            5: " reserved",
            6: " reserved",
            7: " reserved",
            8: " reserved",
            9: " reserved"
        },
        1968: {
            0: " block",
            1: " extend",
            10: " reserved",
            11: " reserved",
            12: " reserved",
            13: " reserved",
            14: " reserved",
            15: " don-t care",
            2: " retract",
            3: " float",
            4: " reserved",
            5: " reserved",
            6: " reserved",
            7: " reserved",
            8: " reserved",
            9: " reserved"
        },
        1969: {
            0: " block",
            1: " float",
            2: " reserved",
            3: " don-t care"
        },
        1970: {
            0: " block",
            1: " float",
            2: " reserved   <=== ???? should be error indicator since measured ???? =====",
            3: " don-t care"
        },
        1975: {
            0: " blocked",
            1: " extend",
            10: " reserved",
            11: " reserved",
            12: " reserved",
            13: " reserved",
            14: " reserved",
            15: " don-t care",
            2: " retract",
            3: " floating",
            4: " reserved",
            5: " reserved",
            6: " reserved",
            7: " reserved",
            8: " reserved",
            9: " reserved"
        },
        1980: {
            0: " block",
            1: " extend",
            10: " reserved",
            11: " reserved",
            12: " reserved",
            13: " reserved",
            14: " reserved",
            15: " don-t care",
            2: " retract",
            3: " float",
            4: " reserved",
            5: " reserved",
            6: " reserved",
            7: " reserved",
            8: " reserved",
            9: " reserved"
        },
        1981: {
            0: " block",
            1: " float",
            2: " reserved",
            3: " don-t care"
        },
        1982: {
            0: " block",
            1: " float",
            2: " reserved   <=== ???? should be error indicator since measured ???? =====",
            3: " don-t care"
        },
        1987: {
            0: " blocked",
            1: " extend",
            10: " reserved",
            11: " reserved",
            12: " reserved",
            13: " reserved",
            14: " reserved",
            15: " don-t care",
            2: " retract",
            3: " floating",
            4: " reserved",
            5: " reserved",
            6: " reserved",
            7: " reserved",
            8: " reserved",
            9: " reserved"
        },
        1992: {
            0: " block",
            1: " extend",
            10: " reserved",
            11: " reserved",
            12: " reserved",
            13: " reserved",
            14: " reserved",
            15: " don-t care",
            2: " retract",
            3: " float",
            4: " reserved",
            5: " reserved",
            6: " reserved",
            7: " reserved",
            8: " reserved",
            9: " reserved"
        },
        1993: {
            0: " block",
            1: " float",
            2: " reserved",
            3: " don-t care"
        },
        1994: {
            0: " block",
            1: " float",
            2: " reserved   <=== ???? should be error indicator since measured ???? =====",
            3: " don-t care"
        },
        1999: {
            0: " blocked",
            1: " extend",
            10: " reserved",
            11: " reserved",
            12: " reserved",
            13: " reserved",
            14: " reserved",
            15: " don-t care",
            2: " retract",
            3: " floating",
            4: " reserved",
            5: " reserved",
            6: " reserved",
            7: " reserved",
            8: " reserved",
            9: " reserved"
        },
        2260: {
            0: " block",
            1: " extend",
            10: " reserved",
            11: " reserved",
            12: " reserved",
            13: " reserved",
            14: " reserved",
            15: " don-t care",
            2: " retract",
            3: " float",
            4: " reserved",
            5: " reserved",
            6: " reserved",
            7: " reserved",
            8: " reserved",
            9: " reserved"
        },
        2261: {
            0: " block",
            1: " float",
            2: " reserved",
            3: " don-t care"
        },
        2262: {
            0: " block",
            1: " float",
            2: " reserved   <=== ???? should be error indicator since measured ???? =====",
            3: " don-t care"
        },
        2267: {
            0: " blocked",
            1: " extend",
            10: " reserved",
            11: " reserved",
            12: " reserved",
            13: " reserved",
            14: " reserved",
            15: " don-t care",
            2: " retract",
            3: " floating",
            4: " reserved",
            5: " reserved",
            6: " reserved",
            7: " reserved",
            8: " reserved",
            9: " reserved"
        },
        2272: {
            0: " block",
            1: " extend",
            10: " reserved",
            11: " reserved",
            12: " reserved",
            13: " reserved",
            14: " reserved",
            15: " don-t care",
            2: " retract",
            3: " float",
            4: " reserved",
            5: " reserved",
            6: " reserved",
            7: " reserved",
            8: " reserved",
            9: " reserved"
        },
        2273: {
            0: " block",
            1: " float",
            2: " reserved",
            3: " don-t care"
        },
        2274: {
            0: " block",
            1: " float",
            2: " reserved   <=== ???? should be error indicator since measured ???? =====",
            3: " don-t care"
        },
        2279: {
            0: " blocked",
            1: " extend",
            10: " reserved",
            11: " reserved",
            12: " reserved",
            13: " reserved",
            14: " reserved",
            15: " don-t care",
            2: " retract",
            3: " floating",
            4: " reserved",
            5: " reserved",
            6: " reserved",
            7: " reserved",
            8: " reserved",
            9: " reserved"
        },
        2284: {
            0: " block",
            1: " extend",
            10: " reserved",
            11: " reserved",
            12: " reserved",
            13: " reserved",
            14: " reserved",
            15: " don-t care",
            2: " retract",
            3: " float",
            4: " reserved",
            5: " reserved",
            6: " reserved",
            7: " reserved",
            8: " reserved",
            9: " reserved"
        },
        2285: {
            0: " block",
            1: " float",
            2: " reserved",
            3: " don-t care"
        },
        2286: {
            0: " block",
            1: " float",
            2: " reserved   <=== ???? should be error indicator since measured ???? =====",
            3: " don-t care"
        },
        2291: {
            0: " blocked",
            1: " extend",
            10: " reserved",
            11: " reserved",
            12: " reserved",
            13: " reserved",
            14: " reserved",
            15: " don-t care",
            2: " retract",
            3: " floating",
            4: " reserved",
            5: " reserved",
            6: " reserved",
            7: " reserved",
            8: " reserved",
            9: " reserved"
        },
        2296: {
            0: " block",
            1: " extend",
            10: " reserved",
            11: " reserved",
            12: " reserved",
            13: " reserved",
            14: " reserved",
            15: " don-t care",
            2: " retract",
            3: " float",
            4: " reserved",
            5: " reserved",
            6: " reserved",
            7: " reserved",
            8: " reserved",
            9: " reserved"
        },
        2297: {
            0: " block",
            1: " float",
            2: " reserved",
            3: " don-t care"
        },
        2298: {
            0: " block",
            1: " float",
            2: " reserved   <=== ???? should be error indicator since measured ???? =====",
            3: " don-t care"
        },
        2303: {
            0: " blocked",
            1: " extend",
            10: " reserved",
            11: " reserved",
            12: " reserved",
            13: " reserved",
            14: " reserved",
            15: " don-t care",
            2: " retract",
            3: " floating",
            4: " reserved",
            5: " reserved",
            6: " reserved",
            7: " reserved",
            8: " reserved",
            9: " reserved"
        },
        2308: {
            0: " block",
            1: " extend",
            10: " reserved",
            11: " reserved",
            12: " reserved",
            13: " reserved",
            14: " reserved",
            15: " don-t care",
            2: " retract",
            3: " float",
            4: " reserved",
            5: " reserved",
            6: " reserved",
            7: " reserved",
            8: " reserved",
            9: " reserved"
        },
        2309: {
            0: " block",
            1: " float",
            2: " reserved",
            3: " don-t care"
        },
        2310: {
            0: " block",
            1: " float",
            2: " reserved   <=== ???? should be error indicator since measured ???? =====",
            3: " don-t care"
        },
        2315: {
            0: " blocked",
            1: " extend",
            10: " reserved",
            11: " reserved",
            12: " reserved",
            13: " reserved",
            14: " reserved",
            15: " don-t care",
            2: " retract",
            3: " floating",
            4: " reserved",
            5: " reserved",
            6: " reserved",
            7: " reserved",
            8: " reserved",
            9: " reserved"
        },
        2320: {
            0: " block",
            1: " extend",
            10: " reserved",
            11: " reserved",
            12: " reserved",
            13: " reserved",
            14: " reserved",
            15: " don-t care",
            2: " retract",
            3: " float",
            4: " reserved",
            5: " reserved",
            6: " reserved",
            7: " reserved",
            8: " reserved",
            9: " reserved"
        },
        2321: {
            0: " block",
            1: " float",
            2: " reserved",
            3: " don-t care"
        },
        2322: {
            0: " block",
            1: " float",
            2: " reserved   <=== ???? should be error indicator since measured ???? =====",
            3: " don-t care"
        },
        2327: {
            0: " blocked",
            1: " extend",
            10: " reserved",
            11: " reserved",
            12: " reserved",
            13: " reserved",
            14: " reserved",
            15: " don-t care",
            2: " retract",
            3: " floating",
            4: " reserved",
            5: " reserved",
            6: " reserved",
            7: " reserved",
            8: " reserved",
            9: " reserved"
        },
        2332: {
            0: " block",
            1: " extend",
            10: " reserved",
            11: " reserved",
            12: " reserved",
            13: " reserved",
            14: " reserved",
            15: " don-t care",
            2: " retract",
            3: " float",
            4: " reserved",
            5: " reserved",
            6: " reserved",
            7: " reserved",
            8: " reserved",
            9: " reserved"
        },
        2333: {
            0: " block",
            1: " float",
            2: " reserved",
            3: " don-t care"
        },
        2334: {
            0: " block",
            1: " float",
            2: " reserved   <=== ???? should be error indicator since measured ???? =====",
            3: " don-t care"
        },
        2339: {
            0: " blocked",
            1: " extend",
            10: " reserved",
            11: " reserved",
            12: " reserved",
            13: " reserved",
            14: " reserved",
            15: " don-t care",
            2: " retract",
            3: " floating",
            4: " reserved",
            5: " reserved",
            6: " reserved",
            7: " reserved",
            8: " reserved",
            9: " reserved"
        },
        2344: {
            0: " block",
            1: " extend",
            10: " reserved",
            11: " reserved",
            12: " reserved",
            13: " reserved",
            14: " reserved",
            15: " don-t care",
            2: " retract",
            3: " float",
            4: " reserved",
            5: " reserved",
            6: " reserved",
            7: " reserved",
            8: " reserved",
            9: " reserved"
        },
        2345: {
            0: " block",
            1: " float",
            2: " reserved",
            3: " don-t care"
        },
        2346: {
            0: " block",
            1: " float",
            2: " error",
            3: " not available"
        },
        2347: {
            0: " de-activate",
            1: " activate",
            2: " reserved",
            3: " don-t care"
        },
        2348: {
            0: " de-activated",
            1: " activated",
            2: " fault detected",
            3: " not available"
        },
        2349: {
            0: " de-activate",
            1: " activate",
            2: " reserved",
            3: " don-t care"
        },
        2350: {
            0: " de-activated",
            1: " activated",
            2: " fault detected",
            3: " not available"
        },
        2351: {
            0: " de-activate",
            1: " activate",
            2: " reserved",
            3: " don-t care"
        },
        2352: {
            0: " de-activated",
            1: " activated",
            2: " fault detected",
            3: " not available"
        },
        2353: {
            0: " de-activate",
            1: " activate",
            2: " reserved",
            3: " don-t care"
        },
        2354: {
            0: " de-activated",
            1: " activated",
            2: " fault detected",
            3: " not available"
        },
        2355: {
            0: " de-activate",
            1: " activate",
            2: " reserved",
            3: " don-t care"
        },
        2356: {
            0: " de-activated",
            1: " activated",
            2: " fault detected",
            3: " not available"
        },
        2357: {
            0: " de-activate",
            1: " activate",
            2: " reserved",
            3: " don-t care"
        },
        2358: {
            0: " de-activated",
            1: " activated",
            2: " fault detected",
            3: " not available"
        },
        2359: {
            0: " de-activate",
            1: " activate",
            2: " reserved",
            3: " don-t care"
        },
        2360: {
            0: " de-activated",
            1: " activated",
            2: " fault detected",
            3: " not available"
        },
        2361: {
            0: " de-activate",
            1: " activate",
            2: " reserved",
            3: " don-t care"
        },
        2362: {
            0: " de-activated",
            1: " activated",
            2: " fault detected",
            3: " not available"
        },
        2363: {
            0: " de-activate",
            1: " activate",
            2: " reserved",
            3: " don-t care"
        },
        2364: {
            0: " de-activated",
            1: " activated",
            2: " fault detected",
            3: " not available"
        },
        2365: {
            0: " de-activate",
            1: " activate",
            2: " reserved",
            3: " don-t care"
        },
        2366: {
            0: " de-activated",
            1: " activated",
            2: " fault detected",
            3: " not available"
        },
        2367: {
            0: " de-activate",
            1: " activate",
            2: " reserved",
            3: " don-t care"
        },
        2368: {
            0: " de-activated",
            1: " activated",
            2: " fault detected",
            3: " not available"
        },
        2369: {
            0: " de-activate",
            1: " activate",
            2: " reserved",
            3: " don-t care"
        },
        2370: {
            0: " de-activated",
            1: " activated",
            2: " fault detected",
            3: " not available"
        },
        2371: {
            0: " de-activate",
            1: " activate",
            2: " reserved",
            3: " don-t care"
        },
        2372: {
            0: " de-activated",
            1: " activated",
            2: " fault detected",
            3: " not available"
        },
        2373: {
            0: " de-activate",
            1: " activate",
            2: " reserved",
            3: " don-t care"
        },
        2374: {
            0: " de-activated",
            1: " activated",
            2: " fault detected",
            3: " not available"
        },
        2375: {
            0: " de-activate",
            1: " activate",
            2: " reserved",
            3: " don-t care"
        },
        2376: {
            0: " de-activated",
            1: " activated",
            2: " fault detected",
            3: " not available"
        },
        2377: {
            0: " de-activate",
            1: " activate",
            2: " reserved",
            3: " don-t care"
        },
        2378: {
            0: " de-activated",
            1: " activated",
            2: " fault detected",
            3: " not available"
        },
        2379: {
            0: " de-activate",
            1: " activate",
            2: " reserved",
            3: " don-t care"
        },
        2380: {
            0: " de-activated",
            1: " activated",
            2: " fault detected",
            3: " not available"
        },
        2381: {
            0: " de-activate",
            1: " activate",
            2: " reserved",
            3: " don-t care"
        },
        2382: {
            0: " de-activated",
            1: " activated",
            2: " fault detected",
            3: " not available"
        },
        2383: {
            0: " de-activate",
            1: " activate",
            2: " reserved",
            3: " don-t care"
        },
        2384: {
            0: " de-activated",
            1: " activated",
            2: " fault detected",
            3: " not available"
        },
        2385: {
            0: " de-activate",
            1: " activate",
            2: " reserved",
            3: " don-t care"
        },
        2386: {
            0: " de-activated",
            1: " activated",
            2: " fault detected",
            3: " not available"
        },
        2387: {
            0: " de-activate",
            1: " activate",
            2: " reserved",
            3: " don-t care"
        },
        2388: {
            0: " de-activated",
            1: " activated",
            2: " fault detected",
            3: " not available"
        },
        2389: {
            0: " de-activate",
            1: " activate",
            2: " reserved",
            3: " don-t care"
        },
        2390: {
            0: " de-activated",
            1: " activated",
            2: " fault detected",
            3: " not available"
        },
        2391: {
            0: " de-activate",
            1: " activate",
            2: " reserved",
            3: " don-t care"
        },
        2392: {
            0: " de-activated",
            1: " activated",
            2: " fault detected",
            3: " not available"
        },
        2393: {
            0: " de-activate",
            1: " activate",
            2: " reserved",
            3: " don-t care"
        },
        2394: {
            0: " de-activated",
            1: " activated",
            2: " fault detected",
            3: " not available"
        },
        2395: {
            0: " de-activate",
            1: " activate",
            2: " reserved",
            3: " don-t care"
        },
        2396: {
            0: " de-activated",
            1: " activated",
            2: " fault detected",
            3: " not available"
        },
        2397: {
            0: " de-activate",
            1: " activate",
            2: " reserved",
            3: " don-t care"
        },
        2398: {
            0: " de-activated",
            1: " activated",
            2: " fault detected",
            3: " not available"
        },
        2399: {
            0: " de-activate",
            1: " activate",
            2: " reserved",
            3: " don-t care"
        },
        2400: {
            0: " de-activated",
            1: " activated",
            2: " fault detected",
            3: " not available"
        },
        2401: {
            0: " de-activate",
            1: " activate",
            2: " reserved",
            3: " don-t care"
        },
        2402: {
            0: " de-activated",
            1: " activated",
            2: " fault detected",
            3: " not available"
        },
        2403: {
            0: " de-activate",
            1: " activate",
            2: " reserved",
            3: " don-t care"
        },
        2404: {
            0: " de-activated",
            1: " activated",
            2: " fault detected",
            3: " not available"
        },
        2405: {
            0: " de-activate",
            1: " activate",
            2: " reserved",
            3: " don-t care"
        },
        2406: {
            0: " de-activate",
            1: " activate",
            2: " reserved",
            3: " don-t care"
        },
        2407: {
            0: " de-activated",
            1: " activated",
            2: " fault detected",
            3: " not available"
        },
        2408: {
            0: " power take off disengaged",
            1: " power take off engaged",
            2: " error indication",
            3: " not available"
        },
        2411: {
            0: " comma is used",
            1: " point is used",
            2: " reserved",
            3: " no action"
        },
        2413: {
            0: " 24 hour",
            1: " 12 hour (am/ pm)",
            2: " reserved",
            3: " no action"
        },
        2414: {
            0: " metric (km, m, -)",
            1: " imperial (miles, feet, ..)",
            2: " reserved",
            3: " no action"
        },
        2415: {
            0: " metric (ha or m2)",
            1: " imperial (acres or feet2)",
            2: " reserved",
            3: " no action"
        },
        2416: {
            0: " metric (litre)",
            1: " imperial (gallon)",
            2: " us (gallon)",
            3: " no action"
        },
        2417: {
            0: " metric (tonne, kg, -)",
            1: " imperial (long tons-uk, pounds, -.)",
            2: " tons/pounds (short tons-us, pounds, -)",
            3: " no action"
        },
        2419: {
            0: " data format is long integer",
            1: " data format is floating",
            2: " data format is unsigned long integer",
            3: " not available or error"
        },
        2420: {
            0: " data is set point",
            1: " data is actual values",
            2: " message is set point request",
            3: " message is actual value request"
        },
        2518: {
            0: " leading",
            1: " lagging",
            2: " error",
            3: " not available or not installed"
        },
        2519: {
            0: " leading",
            1: " lagging",
            2: " error",
            3: " not available or not installed"
        },
        2520: {
            0: " leading",
            1: " lagging",
            2: " error",
            3: " not available or not installed"
        },
        2521: {
            0: " leading",
            1: " lagging",
            2: " error",
            3: " not available or not installed"
        },
        2522: {
            0: " leading",
            1: " lagging",
            2: " error",
            3: " not available or not installed"
        },
        2523: {
            0: " leading",
            1: " lagging",
            2: " error",
            3: " not available or not installed"
        },
        2524: {
            0: " leading",
            1: " lagging",
            2: " error",
            3: " not available or not installed"
        },
        2525: {
            0: " leading",
            1: " lagging",
            2: " error",
            3: " not available or not installed"
        },
        2526: {
            0: " phases do not match",
            1: " phases match",
            2: " error",
            3: " not available or not installed"
        },
        2527: {
            0: " voltages do not match",
            1: " voltages match",
            2: " error",
            3: " not available or not installed"
        },
        2528: {
            0: " frequencies do not match",
            1: " frequencies match",
            2: " error",
            3: " not available or not installed"
        },
        2529: {
            0: " not synchronized",
            1: " synchronized",
            2: " error",
            3: " not available or not installed"
        },
        2530: {
            0: " bus is not dead",
            1: " bus is dead",
            2: " error",
            3: " not available or not installed"
        },
        2531: {
            0: " phases do not match",
            1: " phases match",
            2: " error",
            3: " not available or not installed"
        },
        2532: {
            0: " voltages do not match",
            1: " voltages match",
            2: " error",
            3: " not available or not installed"
        },
        2533: {
            0: " frequencies do not match",
            1: " frequencies match",
            2: " error",
            3: " not available or not installed"
        },
        2534: {
            0: " not synchronized",
            1: " synchronized",
            2: " error",
            3: " not available or not installed"
        },
        2535: {
            0: " bus is not dead",
            1: " bus is dead",
            2: " error",
            3: " not available or not installed"
        },
        2536: {
            0: " transmission mode 1 not active",
            1: " transmission mode 1 active",
            2: " error",
            3: " not available"
        },
        2537: {
            0: " transmission mode 2 not active",
            1: " transmission mode 2 active",
            2: " error",
            3: " not available"
        },
        2538: {
            0: " transmission mode 3 not active",
            1: " transmission mode 3 active",
            2: " error",
            3: " not available"
        },
        2539: {
            0: " transmission mode 4 not active",
            1: " transmission mode 4 active",
            2: " error",
            3: " not available"
        },
        2575: {
            0: " no",
            1: " yes",
            2: " undefined",
            3: " na"
        },
        2577: {
            0: "narrow = +/- 4.5mm",
            1: "standard = +/- 12mm",
            10: " reserved",
            11: " reserved",
            12: " reserved",
            13: " reserved",
            14: " reserved",
            15: " not available",
            2: "wide = +/- 24mm",
            3: " reserved",
            4: " reserved",
            5: " reserved",
            6: " reserved",
            7: " reserved",
            8: " reserved",
            9: " reserved"
        },
        2582: {
            0: " supply is not reliable",
            1: " supply is reliable",
            2: " error indicator",
            3: " not available"
        },
        2583: {
            0: " supply is not reliable",
            1: " supply is reliable",
            2: " error indicator",
            3: " not available"
        },
        2584: {
            0: " pressure level sufficient",
            1: " pressure level below warning level",
            2: " error indicator",
            3: " not available"
        },
        2585: {
            0: " pressure level sufficient",
            1: " pressure level below warning level",
            2: " error indicator",
            3: " not available"
        },
        2587: {
            0: " extreme over pressure - the tire pressure is at a level where the safety of the vehicle may be jeopardised.",
            1: " over pressure - the tire pressure is higher than the pressure defined by the vehicle or tire manufacturer.",
            2: " no warning pressure - the tire pressure is within the thresholds defined by the vehicle or tire manufacturer.",
            3: " under pressure - the tire pressure is lower than the pressure defined by the vehicle or tire manufacturer.",
            4: " extreme under pressure - the tire pressure is at a level where the safety of the vehicle may be jeopardised.",
            5: " not defined",
            6: " error indicator",
            7: " not available"
        },
        2597: {
            0: " de-activate",
            1: " activate",
            2: " reserved",
            3: " don-t care"
        },
        2598: {
            0: " de-activate",
            1: " activate",
            2: " fault detected",
            3: " not available"
        },
        2599: {
            0: " pump not engaged",
            1: " pump engaged",
            2: " error",
            3: " not available or not installed"
        },
        2612: {
            0: " front wheel drive actuator not engaged",
            1: " front wheel drive actuator engaged",
            2: " error",
            3: " not available"
        },
        2615: {
            0: " not synchronized",
            1: " synchronized center",
            10: " reserved",
            11: " reserved",
            12: " reserved",
            13: " reserved",
            14: " reserved",
            15: " take no action",
            2: " synchronized port",
            3: " synchronized starboard",
            4: " synchronized master",
            5: " reserved",
            6: " reserved",
            7: " reserved",
            8: " reserved",
            9: " reserved"
        },
        2616: {
            0: " trolling mode is off.",
            1: " trolling mode is active.",
            2: " reserved",
            3: " take no action"
        },
        2617: {
            0: " slow vessel mode is off.",
            1: " slow vessel mode is active.",
            2: " reserved",
            3: " take no action"
        },
        2618: {
            0: " indefinite suspension of all broadcasts",
            1: " indefinite suspension of some messages",
            10: " sae reserved",
            11: " sae reserved",
            12: " sae reserved",
            13: " sae reserved",
            14: " sae reserved",
            15: " not available",
            2: " temporary suspension of all broadcasts",
            3: " temporary suspension of some messages",
            4: " sae reserved",
            5: " sae reserved",
            6: " sae reserved",
            7: " sae reserved",
            8: " sae reserved",
            9: " sae reserved"
        },
        2665: {
            0: " not on positive side of neutral",
            1: " on positive side of neutral",
            2: " error indicator",
            3: " not available"
        },
        2666: {
            0: " not on positive side of neutral",
            1: " on positive side of neutral",
            2: " error indicator",
            3: " not available"
        },
        2667: {
            0: " not on positive side of neutral",
            1: " on positive side of neutral",
            2: " error indicator",
            3: " not available"
        },
        2668: {
            0: " not on positive side of neutral",
            1: " on positive side of neutral",
            2: " error indicator",
            3: " not available"
        },
        2669: {
            0: " not on positive side of neutral",
            1: " on positive side of neutral",
            2: " error indicator",
            3: " not available"
        },
        2670: {
            0: " not on negative side of neutral",
            1: " on negative side of neutral",
            2: " error indicator",
            3: " not available"
        },
        2671: {
            0: " not on negative side of neutral",
            1: " on negative side of neutral",
            2: " error indicator",
            3: " not available"
        },
        2672: {
            0: " not on negative side of neutral",
            1: " on negative side of neutral",
            2: " error indicator",
            3: " not available"
        },
        2673: {
            0: " not on negative side of neutral",
            1: " on negative side of neutral",
            2: " error indicator",
            3: " not available"
        },
        2674: {
            0: " not on negative side of neutral",
            1: " on negative side of neutral",
            2: " error indicator",
            3: " not available"
        },
        2675: {
            0: " not in neutral position",
            1: " in neutral position",
            2: " error indicator",
            3: " not available"
        },
        2676: {
            0: " not in neutral position",
            1: " in neutral position",
            2: " error indicator",
            3: " not available"
        },
        2677: {
            0: " not in neutral position",
            1: " in neutral position",
            2: " error indicator",
            3: " not available"
        },
        2678: {
            0: " not in neutral position",
            1: " in neutral position",
            2: " error indicator",
            3: " not available"
        },
        2679: {
            0: " not in neutral position",
            1: " in neutral position",
            2: " error indicator",
            3: " not available"
        },
        2680: {
            0: " not in the detent position",
            1: " in the detent position",
            2: " error indicator",
            3: " not available"
        },
        2681: {
            0: " not in the detent position",
            1: " in the detent position",
            2: " error indicator",
            3: " not available"
        },
        2682: {
            0: " not in the detent position",
            1: " in the detent position",
            2: " error indicator",
            3: " not available"
        },
        2683: {
            0: " not in the detent position",
            1: " in the detent position",
            2: " error indicator",
            3: " not available"
        },
        2684: {
            0: " not in the detent position",
            1: " in the detent position",
            2: " error indicator",
            3: " not available"
        },
        2685: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        2686: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        2687: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        2688: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        2689: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        2690: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        2691: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        2692: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        2693: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        2694: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        2695: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        2696: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        2702: {
            0: " not on positive side of neutral",
            1: " on positive side of neutral",
            2: " error indicator",
            3: " not available"
        },
        2703: {
            0: " not on positive side of neutral",
            1: " on positive side of neutral",
            2: " error indicator",
            3: " not available"
        },
        2704: {
            0: " not on positive side of neutral",
            1: " on positive side of neutral",
            2: " error indicator",
            3: " not available"
        },
        2705: {
            0: " not on positive side of neutral",
            1: " on positive side of neutral",
            2: " error indicator",
            3: " not available"
        },
        2706: {
            0: " not on positive side of neutral",
            1: " on positive side of neutral",
            2: " error indicator",
            3: " not available"
        },
        2707: {
            0: " not on negative side of neutral",
            1: " on negative side of neutral",
            2: " error indicator",
            3: " not available"
        },
        2708: {
            0: " not on negative side of neutral",
            1: " on negative side of neutral",
            2: " error indicator",
            3: " not available"
        },
        2709: {
            0: " not on negative side of neutral",
            1: " on negative side of neutral",
            2: " error indicator",
            3: " not available"
        },
        2710: {
            0: " not on negative side of neutral",
            1: " on negative side of neutral",
            2: " error indicator",
            3: " not available"
        },
        2711: {
            0: " not on negative side of neutral",
            1: " on negative side of neutral",
            2: " error indicator",
            3: " not available"
        },
        2712: {
            0: " not in neutral position",
            1: " in neutral position",
            2: " error indicator",
            3: " not available"
        },
        2713: {
            0: " not in neutral position",
            1: " in neutral position",
            2: " error indicator",
            3: " not available"
        },
        2714: {
            0: " not on negative side of neutral",
            1: " on negative side of neutral",
            2: " error indicator",
            3: " not available"
        },
        2715: {
            0: " not in neutral position",
            1: " in neutral position",
            2: " error indicator",
            3: " not available"
        },
        2716: {
            0: " not in neutral position",
            1: " in neutral position",
            2: " error indicator",
            3: " not available"
        },
        2717: {
            0: " not in the detent position",
            1: " in the detent position",
            2: " error indicator",
            3: " not available"
        },
        2718: {
            0: " not in the detent position",
            1: " in the detent position",
            2: " error indicator",
            3: " not available"
        },
        2719: {
            0: " not in the detent position",
            1: " in the detent position",
            2: " error indicator",
            3: " not available"
        },
        2720: {
            0: " not in the detent position",
            1: " in the detent position",
            2: " error indicator",
            3: " not available"
        },
        2721: {
            0: " not in the detent position",
            1: " in the detent position",
            2: " error indicator",
            3: " not available"
        },
        2722: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        2723: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        2724: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        2725: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        2726: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        2727: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        2728: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        2729: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        2730: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        2731: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        2732: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        2733: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        2739: {
            0: " not on positive side of neutral",
            1: " on positive side of neutral",
            2: " error indicator",
            3: " not available"
        },
        2740: {
            0: " not on positive side of neutral",
            1: " on positive side of neutral",
            2: " error indicator",
            3: " not available"
        },
        2741: {
            0: " not on positive side of neutral",
            1: " on positive side of neutral",
            2: " error indicator",
            3: " not available"
        },
        2742: {
            0: " not on positive side of neutral",
            1: " on positive side of neutral",
            2: " error indicator",
            3: " not available"
        },
        2743: {
            0: " not on positive side of neutral",
            1: " on positive side of neutral",
            2: " error indicator",
            3: " not available"
        },
        2744: {
            0: " not on negative side of neutral",
            1: " on negative side of neutral",
            2: " error indicator",
            3: " not available"
        },
        2745: {
            0: " not on negative side of neutral",
            1: " on negative side of neutral",
            2: " error indicator",
            3: " not available"
        },
        2746: {
            0: " not on negative side of neutral",
            1: " on negative side of neutral",
            2: " error indicator",
            3: " not available"
        },
        2747: {
            0: " not on negative side of neutral",
            1: " on negative side of neutral",
            2: " error indicator",
            3: " not available"
        },
        2748: {
            0: " not on negative side of neutral",
            1: " on negative side of neutral",
            2: " error indicator",
            3: " not available"
        },
        2749: {
            0: " not in neutral position",
            1: " in neutral position",
            2: " error indicator",
            3: " not available"
        },
        2750: {
            0: " not in neutral position",
            1: " in neutral position",
            2: " error indicator",
            3: " not available"
        },
        2751: {
            0: " not on negative side of neutral",
            1: " on negative side of neutral",
            2: " error indicator",
            3: " not available"
        },
        2752: {
            0: " not in neutral position",
            1: " in neutral position",
            2: " error indicator",
            3: " not available"
        },
        2753: {
            0: " not in neutral position",
            1: " in neutral position",
            2: " error indicator",
            3: " not available"
        },
        2754: {
            0: " not in the detent position",
            1: " in the detent position",
            2: " error indicator",
            3: " not available"
        },
        2755: {
            0: " not in the detent position",
            1: " in the detent position",
            2: " error indicator",
            3: " not available"
        },
        2756: {
            0: " not in the detent position",
            1: " in the detent position",
            2: " error indicator",
            3: " not available"
        },
        2757: {
            0: " not in the detent position",
            1: " in the detent position",
            2: " error indicator",
            3: " not available"
        },
        2758: {
            0: " not in the detent position",
            1: " in the detent position",
            2: " error indicator",
            3: " not available"
        },
        2759: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        2760: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        2761: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        2762: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        2763: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        2764: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        2765: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        2766: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        2767: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        2768: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        2769: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        2770: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        2792: {
            0: "   vgt air control shutoff valve is off",
            1: "   vgt air control shutoff valve is on",
            2: "   reserved",
            3: "   not available"
        },
        2796: {
            0: "\t2 wheel high (normal or -on highway- range)",
            1: "\t4 wheel high (normal or -on highway- range)",
            2: "\tneutral",
            3: "\t2 wheel low (or -off highway- range)",
            4: "\t4 wheel low (or -off highway- range)",
            5: "\treserved for sae assignment",
            6: "\terror indicator",
            7: "\tnot available"
        },
        2804: {
            0: " diagnostics is not supported",
            1: " diagnostics is supported",
            2: " reserved",
            3: " don't care"
        },
        2805: {
            0: " on request mode is not supported",
            1: " on request mode is supported",
            2: " reserved",
            3: " don't care"
        },
        2807: {
            0: "open (fuel supplied to engine)",
            1: "closed (no fuel supplied to engine)",
            2: "reserved",
            3: "don't care / take no action"
        },
        2812: {
            0: " engine overspeed test not active",
            1: " engine overspeed test active",
            2: " reserved",
            3: " not available"
        },
        2813: {
            0: " air shutoff disabled, not attempting to shutoff engine air supply",
            1: " air shutoff enabled, attempting to shutoff engine air supply",
            2: " reserved",
            3: " not available"
        },
        2814: {
            0: " engine alarm output command not active",
            1: " engine alarm output command active",
            2: " reserved",
            3: " not available"
        },
        2815: {
            0: " engine alarm acknowledge  not active ",
            1: " engine alarm acknowledge active",
            2: " error",
            3: " not available"
        },
        2863: {
            0: " off",
            1: " low",
            10: "   reserved   ",
            11: "   reserved   ",
            12: "   reserved   ",
            13: "   reserved   ",
            14: "   reserved   ",
            15: " not available (do not change)",
            2: " medium",
            3: " high",
            4: " delayed 1 (used for the first delay choice when the wiper switch position controls the delay)",
            5: " delayed 2 (used for the second delay choice when the wiper switch position controls the delay)",
            6: " mist (position where external sensor  controls wiper rate)",
            7: "   reserved   ",
            8: "   reserved   ",
            9: "   reserved   "
        },
        2864: {
            0: " off",
            1: " low",
            10: "   reserved   ",
            11: "   reserved   ",
            12: "   reserved   ",
            13: "   reserved   ",
            14: "   reserved   ",
            15: " not available (do not change)",
            2: " medium",
            3: " high",
            4: " delayed 1 (used for the first delay choice when the wiper switch position controls the delay)",
            5: " delayed 2 (used for the second delay choice when the wiper switch position controls the delay)",
            6: " mist (position where external sensor  controls wiper rate)",
            7: "   reserved   ",
            8: "   reserved   ",
            9: "   reserved   "
        },
        2865: {
            0: " off",
            1: " low",
            10: "   reserved   ",
            11: "   reserved   ",
            12: "   reserved   ",
            13: "   reserved   ",
            14: "   reserved   ",
            15: " not available (do not change)",
            2: " medium",
            3: " high",
            4: " delayed 1 (used for the first delay choice when the wiper switch position controls the delay)",
            5: " delayed 2 (used for the second delay choice when the wiper switch position controls the delay)",
            6: " mist (position where external sensor  controls wiper rate)",
            7: "   reserved   ",
            8: "   reserved   ",
            9: "   reserved   "
        },
        2866: {
            0: " off",
            1: " low",
            2: " medium",
            3: " high",
            4: "   reserved   ",
            5: "   reserved   ",
            6: "   reserved   ",
            7: " not available (do not change)"
        },
        2867: {
            0: " off",
            1: " low",
            2: " medium",
            3: " high",
            4: "   reserved   ",
            5: "   reserved   ",
            6: "   reserved   ",
            7: " not available (do not change)"
        },
        2868: {
            0: " off",
            1: " low",
            2: " medium",
            3: " high",
            4: "   reserved   ",
            5: "   reserved   ",
            6: "   reserved   ",
            7: " not available (do not change)"
        },
        2872: {
            0: " off - the position by which the operator selects that none of the lamps are to be on.",
            1: " park on - the position by which the operator selects that the park lamps are to be on. ",
            10: "  reserved   ",
            11: "  reserved   ",
            12: "  reserved   ",
            13: "  reserved   ",
            14: "  error",
            15: "  not available (do not change)",
            2: " headlight on - the position by which the operator selects that the headlamps are to be on.",
            3: " headlight and park on - the position by which the operator selects that both the headlamps and the park lamps are to be on. ",
            4: "   reserved   ",
            5: "   reserved   ",
            6: "   reserved   ",
            7: "   reserved   ",
            8: "   delayed off - the position by which the operator selects that a certain set of lamps are to come on and then are to be turned off following a delay time (operators desired - delayed lamp off time).",
            9: "  reserved   "
        },
        2873: {
            0: "   off - the position by which the operator selects that none of the work lamps are to be on.",
            1: "   work light combination #1 on - the position by which the operator selects that the lamps in the combination defined as work light combination #1 are to be on.",
            10: "    reserved   ",
            11: "    reserved   ",
            12: "    reserved   ",
            13: "    reserved   ",
            14: "   error",
            15: "   not available (do not change)",
            2: "   work light combination #2 on - the position by which the operator selects that the lamps in the combination defined as work light combination #2 are to be on.",
            3: "   work light combination #3 on - the position by which the operator selects that the lamps in the combination defined as work light combination #3 are to be on.",
            4: "   work light combination #4 on - the position by which the operator selects that the lamps in the combination defined as work light combination #4 are to be on.",
            5: "    reserved   ",
            6: "    reserved   ",
            7: "    reserved   ",
            8: "    reserved   ",
            9: "    reserved   "
        },
        2874: {
            0: "  low beam selected",
            1: "  high beam selected",
            2: "  error",
            3: "  not available (do not change)"
        },
        2875: {
            0: "   hazard lamps to be off",
            1: "   hazard lamps to be flashing",
            2: "   error",
            3: "   not available (do not change)"
        },
        2876: {
            0: "   no turn being signaled",
            1: "   left turn to be flashing",
            10: "   reserved   ",
            11: "   reserved   ",
            12: "   reserved   ",
            13: "   reserved   ",
            14: "   error (to include both left and right selected simultaneously)",
            15: "   not available (do not change)",
            2: "   right turn to be flashing",
            3: "   reserved   ",
            4: "   reserved   ",
            5: "   reserved   ",
            6: "   reserved   ",
            7: "   reserved   ",
            8: "   reserved   ",
            9: "   reserved   "
        },
        2879: {
            0: "normal droop setting is selected",
            1: " thru 1101 - one of alternate droop setting 1 through 13 (in sequential order) is selected",
            14: "error condition",
            15: "not available"
        },
        2880: {
            0: "indicates that the isc functionality is not requested, engine operates normally",
            1: " thru 1101 - indicates that the isc setting 1 thru 13 (in sequential order) is selected",
            14: "error condition",
            15: "not available"
        },
        2881: {
            0: "normal droop setting is selected",
            1: " thru 1101 - one of alternate droop setting 1 through 13 (in sequential order) is selected",
            14: "error condition",
            15: "not available"
        },
        2883: {
            0: " default low idle point is selected",
            1: " alternate low idle point is selected",
            2: " error",
            3: " not available or unused"
        },
        2884: {
            0: "auxiliary governor is disabled",
            1: "auxiliary governor is enabled",
            2: "error condition",
            3: "not available"
        },
        2885: {
            0: "normal droop setting is selected",
            1: " thru 1101 - one of alternate droop setting 1 through 13 (in sequential order) is selected",
            14: "error condition",
            15: "not available"
        },
        2886: {
            0: "normal droop setting is selected",
            1: " thru 1101 - one of alternate droop setting 1 through 13 (in sequential order) is selected",
            14: "error condition",
            15: "not available"
        },
        2888: {
            0: "control state is maximum power fueling",
            1: "control state is alternate power fueling 1 ",
            10: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            100: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            101: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            102: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            103: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            104: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            105: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            106: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            107: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            108: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            109: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            11: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            110: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            111: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            112: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            113: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            114: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            115: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            116: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            117: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            118: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            119: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            12: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            120: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            121: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            122: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            123: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            124: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            125: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            126: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            127: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            128: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            129: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            13: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            130: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            131: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            132: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            133: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            134: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            135: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            136: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            137: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            138: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            139: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            14: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            140: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            141: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            142: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            143: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            144: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            145: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            146: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            147: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            148: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            149: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            15: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            150: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            151: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            152: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            153: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            154: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            155: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            156: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            157: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            158: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            159: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            16: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            160: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            161: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            162: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            163: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            164: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            165: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            166: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            167: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            168: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            169: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            17: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            170: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            171: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            172: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            173: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            174: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            175: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            176: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            177: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            178: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            179: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            18: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            180: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            181: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            182: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            183: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            184: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            185: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            186: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            187: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            188: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            189: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            19: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            190: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            191: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            192: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            193: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            194: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            195: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            196: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            197: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            198: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            199: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            2: "control state is alternate power fueling 2 ",
            20: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            200: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            201: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            202: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            203: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            204: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            205: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            206: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            207: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            208: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            209: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            21: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            210: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            211: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            212: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            213: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            214: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            215: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            216: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            217: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            218: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            219: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            22: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            220: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            221: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            222: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            223: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            224: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            225: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            226: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            227: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            228: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            229: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            23: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            230: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            231: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            232: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            233: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            234: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            235: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            236: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            237: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            238: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            239: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            24: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            240: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            241: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            242: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            243: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            244: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            245: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            246: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            247: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            248: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            249: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            25: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            250: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            251: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            252: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            253: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            254: "sae reserved",
            255: "not available",
            26: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            27: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            28: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            29: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            3: "control state is alternate power fueling 3",
            30: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            31: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            32: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            33: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            34: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            35: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            36: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            37: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            38: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            39: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            4: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            40: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            41: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            42: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            43: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            44: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            45: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            46: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            47: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            48: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            49: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            5: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            50: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            51: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            52: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            53: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            54: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            55: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            56: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            57: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            58: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            59: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            6: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            60: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            61: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            62: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            63: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            64: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            65: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            66: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            67: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            68: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            69: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            7: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            70: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            71: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            72: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            73: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            74: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            75: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            76: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            77: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            78: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            79: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            8: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            80: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            81: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            82: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            83: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            84: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            85: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            86: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            87: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            88: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            89: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            9: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            90: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            91: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            92: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            93: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            94: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            95: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            96: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            97: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            98: "control state is alternate power fueling 4 thru 253 (in sequential order) ",
            99: "control state is alternate power fueling 4 thru 253 (in sequential order) "
        },
        2889: {
            0: "control state is normal droop setting",
            1: " thru 1101 - control state is one of alternate droop setting 1 through 13 (in sequential order) ",
            14: "sae reserved",
            15: "not available"
        },
        2890: {
            0: " control state is functionality disabled (off)",
            1: " control state is functionality enabled (on)",
            2: " sae reserved",
            3: " not available"
        },
        2891: {
            0: "normal idle state",
            1: "alternate idle state",
            2: "sae reserved",
            3: "not available"
        },
        2892: {
            0: "control state is isc functionality is not requested, engine operates normally",
            1: " thru 1101 - control state is  isc setting 1 thru 13 (in sequential order) ",
            1110: "sae reserved",
            1111: "not available",
            13: " preprogrammed intermediate speed control settings are available for the controlling ecm to select.  if no speed setting is requested, the engine operates normally. the intermediate speed control is widely used in the industrial application to control the engine to an intermediate speed setting which can either replace the accelerator position control altogether, or limit the accelerator position control to a selectable speed point minimum or maximum. this parameter indicates which state has been selected by the controlling ecm."
        },
        2893: {
            0: "control's state is normal droop setting",
            1: " thru 1101 - control's state is one of alternate droop setting 1 through 13 (in sequential order)",
            14: "sae reserved",
            15: "not available"
        },
        2894: {
            0: "control state is normal droop setting",
            1: " thru 1101 - control state is one of alternate droop setting 1 through 13 (in sequential order) ",
            14: "sae reserved",
            15: "not available"
        },
        2895: {
            0: "control state is normal droop setting",
            1: " thru 1101 - control state is one of alternate droop setting 1 through 13 (in sequential order)",
            14: "sae reserved",
            15: "not available"
        },
        2896: {
            0: "control state is auxiliary governor disabled",
            1: "control state is auxiliary governor enabled",
            2: "sae reserved",
            3: "not available"
        },
        2897: {
            0: " pto set speed memory 1 selected",
            1: " pto set speed memory 2 selected",
            2: " error",
            3: " not available"
        },
        2898: {
            0: "no start enable device 2 installed",
            1: "glow plugs installed",
            10: "reserved",
            11: "reserved",
            12: "reserved",
            13: "reserved",
            14: "reserved",
            15: "not available",
            2: "fuelled start installed",
            3: "ether injection installed",
            4: "electric intake air heater installed",
            5: "reserved",
            6: "reserved",
            7: "reserved",
            8: "reserved",
            9: "reserved"
        },
        2899: {
            0: "no start enable device 1 installed",
            1: "glow plugs installed",
            10: "reserved",
            11: "reserved",
            12: "reserved",
            13: "reserved",
            14: "reserved",
            15: "not available",
            2: "fuelled start installed",
            3: "ether injection installed",
            4: "electric intake air heater installed",
            5: "reserved",
            6: "reserved",
            7: "reserved",
            8: "reserved",
            9: "reserved"
        },
        2900: {
            0: "cranking disabled; engine cranking is prohibited by the transmission",
            1: "cranking enabled; engine cranking is allowed by the transmission",
            2: "error",
            3: "not available"
        },
        2911: {
            0: "    halt brake switch passive",
            1: "    halt brake switch active",
            2: "    error",
            3: "    not available"
        },
        2912: {
            0: "            inactive",
            1: "            active",
            2: "            active, but will change to inactive in a short time.  (this mode may be used to warn the driver)",
            3: "  reserved",
            4: "  reserved",
            5: "  reserved",
            6: "            hill holder not functional",
            7: "            not available"
        },
        2913: {
            0: "            inactive",
            1: "            active",
            2: "            active, but not functioning properly.  (this mode may be used to warn the driver)",
            3: "  not defined",
            4: "  not defined",
            5: "  not defined",
            6: "            halt brake not functional",
            7: "            not available"
        },
        2914: {
            0: "no endurance brake integration allowed",
            1: "only endurance brakes allowed",
            2: "endurance brake integration allowed",
            3: "not defined"
        },
        2915: {
            0: "highest priority - used for emergency situations, e.g. for future collision avoidance system.  this mode overrides any brake protection measures of the brake system.",
            1: "high priority - not defined",
            2: "medium priority - used for acc-systems. this mode does not override brake protection measures of the brake system.",
            3: "low priority - used in -override disabled- xbr control mode"
        },
        2916: {
            0: "override disabled - disable any existing control commanded by the source of this command.",
            1: "acceleration control with addition mode - add the xbr acceleration demand to the driver-s acceleration demand.",
            2: "acceleration control with maximum mode - execute the xbr acceleration demand if it is higher than the driver-s acceleration demand.",
            3: "not defined"
        },
        2917: {
            0: "any external brake demand will be accepted (brake system fully operational)",
            1: "only external brake demand of highest xbr priority (00) will be accepted  (e.g. because the temperature limit of the brake system is exceeded)",
            2: "no external brake demand will be accepted (e.g. because of fault in brake system)",
            3: "not available"
        },
        2918: {
            0: "                no brake demand being executed (default mode)",
            1: "                driver's brake demand being executed, no external brake demand",
            10: "    reserved for sae assignment",
            11: "    reserved for sae assignment",
            12: "    reserved for sae assignment",
            13: "    reserved for sae assignment",
            14: "    reserved for sae assignment",
            15: "                not available",
            2: "                addition mode of xbr acceleration control being executed",
            3: "                maximum mode of xbr acceleration control being executed",
            4: "    reserved for sae assignment",
            5: "    reserved for sae assignment",
            6: "    reserved for sae assignment",
            7: "    reserved for sae assignment",
            8: "    reserved for sae assignment",
            9: "    reserved for sae assignment"
        },
        2919: {
            0: "   foundation brakes not in use",
            1: "   foundation brakes in use",
            2: "   reserved",
            3: "   not available"
        },
        2922: {
            0: "      lowering allowed",
            1: "      lowering inhibited",
            2: "      reserved",
            3: "      not available"
        },
        2923: {
            0: "               axle steering not active (adhesion steering)",
            1: "               axle steering active",
            10: "     reserved for sae assignment",
            11: "     reserved for sae assignment",
            12: "     reserved for sae assignment",
            13: "     reserved for sae assignment",
            14: "     reserved for sae assignment",
            15: "               not available",
            2: "               axle centered",
            3: "               axle centered, because of an error",
            4: "               axle not active because of an error (adhesion steering)",
            5: "               axle steering in special mode (diagnosis, calibration mode)",
            6: "     reserved for sae assignment",
            7: "     reserved for sae assignment",
            8: "     reserved for sae assignment",
            9: "     reserved for sae assignment"
        },
        2924: {
            0: "    main steering system - the steering equipment of a vehicle which is mainly responsible for determining the direction of travel.",
            1: "    auxiliary steering equipment - a system in which the wheels on axle(s) of vehicles of categories m and n are steered in addition to the wheels of the main steering equipment in the same or opposite direction to those of the main steering equipment and/or the steering angle of the front and/or the rear wheels may be adjusted relative to vehicle behaviour.",
            10: "   reserved for assignment by sae",
            11: "   reserved for assignment by sae",
            12: "   reserved for assignment by sae",
            13: "   reserved for assignment by sae",
            14: "   reserved for assignment by sae",
            15: "             not available",
            2: "   reserved for assignment by sae",
            3: "   reserved for assignment by sae",
            4: "   reserved for assignment by sae",
            5: "   reserved for assignment by sae",
            6: "   reserved for assignment by sae",
            7: "   reserved for assignment by sae",
            8: "   reserved for assignment by sae",
            9: "   reserved for assignment by sae"
        },
        2925: {
            0: "  manual steering equipment - the steering forces result solely from the muscular effort of the driver.",
            1: "  power assisted steering equipment - the steering forces result from both the muscular effort of the driver and the energy supply or supplies.",
            10: "   reserved for sae assignment",
            11: "   reserved for sae assignment",
            12: "   reserved for sae assignment",
            13: "   reserved for sae assignment",
            14: "   reserved for sae assignment",
            15: "             not available",
            2: "  full power steering equipment - the steering forces are provided solely by one or more energy supplies.",
            3: "  self tracking steering equipment - a system designed to create a change of steering angle on one or more wheels only when acted upon by forces and/or moments applied through the tire to road contact.",
            4: "   reserved for sae assignment",
            5: "   reserved for sae assignment",
            6: "   reserved for sae assignment",
            7: "   reserved for sae assignment",
            8: "   reserved for sae assignment",
            9: "   reserved for sae assignment"
        },
        2926: {
            0: "    purely mechanical steering transmission - a steering transmission in which the steering forces are transmitted entirely by mechanical means.",
            1: "    purely hydraulic steering transmission - a steering transmission in which the steering forces, somewhere in the transmission, are transmitted only by hydraulic means.",
            10: "   reserved for sae assignment",
            11: "   reserved for sae assignment",
            12: "   reserved for sae assignment",
            13: "   reserved for sae assignment",
            14: "   reserved for sae assignment",
            15: "             not available",
            2: "    purely electric steering transmission - a steering transmission in which the steering forces, somewhere in the transmission, are transmitted only through electric means.",
            3: "    hybrid steering transmission - a steering transmission in which part of the steering forces is transmitted through one and the other part through another of the above mentioned means.  however, in the case where any mechanical part of the transmission is designed only to give position feedback and is too weak to transmit the total sum of the steering forces, this system shall be considered to be purely hydraulic or purely electric steering transmission.",
            4: "   reserved for sae assignment",
            5: "   reserved for sae assignment",
            6: "   reserved for sae assignment",
            7: "   reserved for sae assignment",
            8: "   reserved for sae assignment",
            9: "   reserved for sae assignment"
        },
        2930: {
            0: "       audible warning off",
            1: "       audible warning on",
            2: "       reserved",
            3: "       don't care"
        },
        2931: {
            0: "      fluid level is not sufficient",
            1: "      fluid level is sufficient",
            2: "      error indicator",
            3: "      not available"
        },
        2932: {
            0: "            blocked",
            1: "            extend",
            10: "  reserved",
            11: "  reserved",
            12: "  reserved",
            13: "  reserved",
            14: "  reserved",
            15: "            don-t care",
            2: "            retract",
            3: "            floating",
            4: "  reserved",
            5: "  reserved",
            6: "  reserved",
            7: "  reserved",
            8: "  reserved",
            9: "  reserved"
        },
        2933: {
            0: "              block",
            1: "              extend",
            10: "    reserved",
            11: "    reserved",
            12: "    reserved",
            13: "    reserved",
            14: "    reserved",
            15: "              don-t care",
            2: "              retract",
            3: "              float",
            4: "    reserved",
            5: "    reserved",
            6: "    reserved",
            7: "    reserved",
            8: "    reserved",
            9: "    reserved"
        },
        2934: {
            0: " block",
            1: " float",
            2: " reserved",
            3: " don-t care"
        },
        2935: {
            0: " block",
            1: " float",
            2: " error",
            3: " not available"
        },
        2945: {
            0: "        primary shift console is active",
            1: "        secondary shift console is active",
            2: "        reserved",
            3: "        not available"
        },
        2970: {
            0: "    accelerator pedal 2 not in low idle condition",
            1: "    accelerator pedal 2 in low idle condition",
            2: "    error",
            3: "    not available"
        },
        2971: {
            0: "    accelerator pedal 3 not in low idle condition",
            1: "    accelerator pedal 3 in low idle condition",
            2: "    error",
            3: "    not available"
        },
        2979: {
            0: "     limit not active",
            1: "     limit active",
            2: "     reserved",
            3: "     not available"
        },
        2984: {
            0: "   disable automatic traction help ",
            1: "   enable automatic traction help ",
            2: "   reserved",
            3: "   don't care"
        },
        2985: {
            0: " off",
            1: " on",
            2: " error",
            3: " not available"
        },
        2992: {
            0: " stop",
            1: " move",
            2: " reserved",
            3: " don't care/take no action"
        },
        2993: {
            0: " stop",
            1: " move",
            2: " reserved",
            3: " don't care/take no action"
        },
        2994: {
            0: " stop",
            1: " raise",
            2: " reserved",
            3: " don't care/take no action"
        },
        2995: {
            0: " stop",
            1: " lower",
            2: " reserved",
            3: " don't care/take no action"
        },
        2996: {
            0: " stop",
            1: " raise",
            2: " reserved",
            3: " don't care/take no action"
        },
        2997: {
            0: " stop",
            1: " lower",
            2: " reserved",
            3: " don't care/take no action"
        },
        2998: {
            0: " stop",
            1: " raise",
            2: " reserved",
            3: " don't care/take no action"
        },
        2999: {
            0: " stop",
            1: " lower",
            2: " reserved",
            3: " don't care/take no action"
        },
        3000: {
            0: " stop",
            1: " tilt",
            2: " reserved",
            3: " don't care/take no action"
        },
        3001: {
            0: " stop",
            1: " tilt",
            2: " reserved",
            3: " don't care/take no action"
        },
        3002: {
            0: " stop ",
            1: " fold",
            2: " reserved",
            3: " don't care/take no action"
        },
        3003: {
            0: " stop",
            1: " unfold",
            2: " reserved",
            3: " don't care/take no action"
        },
        3004: {
            0: " stop ",
            1: " increase",
            2: " reserved",
            3: " don't care/take no action"
        },
        3005: {
            0: " stop",
            1: " decrease",
            2: " reserved",
            3: " don't care/take no action"
        },
        3006: {
            0: " stop ",
            1: " increase",
            2: " reserved",
            3: " don't care/take no action"
        },
        3007: {
            0: " stop",
            1: " decrease",
            2: " reserved",
            3: " don't care/take no action"
        },
        3008: {
            0: " stop ",
            1: " increase",
            2: " reserved",
            3: " don't care/take no action"
        },
        3009: {
            0: " stop",
            1: " decrease",
            2: " reserved",
            3: " don't care/take no action"
        },
        3010: {
            0: " stop ",
            1: " increase",
            2: " reserved",
            3: " don't care/take no action"
        },
        3011: {
            0: " stop",
            1: " decrease",
            2: " reserved",
            3: " don't care/take no action"
        },
        3012: {
            0: " stop ",
            1: " increase",
            2: " reserved",
            3: " don't care/take no action"
        },
        3013: {
            0: " stop",
            1: " decrease",
            2: " reserved",
            3: " don't care/take no action"
        },
        3015: {
            0: " stop",
            1: " move",
            2: " reserved",
            3: " don't care/take no action"
        },
        3016: {
            0: " stop",
            1: " move",
            2: " reserved",
            3: " don't care/take no action"
        },
        3017: {
            0: " stop",
            1: " raise",
            2: " reserved",
            3: " don't care/take no action"
        },
        3018: {
            0: " stop",
            1: " raise",
            2: " reserved",
            3: " don't care/take no action"
        },
        3019: {
            0: " disengage",
            1: " engage",
            2: " reserved",
            3: " don't care/take no action"
        },
        3020: {
            0: " disengage",
            1: " engage",
            2: " reserved",
            3: " don't care/take no action"
        },
        3021: {
            0: " not full",
            1: " full",
            2: " error",
            3: " not available"
        },
        3022: {
            0: " disable",
            1: " enable",
            2: " reserved",
            3: " don't care/take no action"
        },
        3023: {
            0: " not engaged",
            1: " engaged",
            2: " error",
            3: " not available"
        },
        3024: {
            0: " off",
            1: " on",
            2: " error",
            3: " not available"
        },
        3025: {
            0: " off",
            1: " on",
            2: " error",
            3: " not available"
        },
        3026: {
            0: " conditions valid for transmission oil level measurement",
            1: " conditions not valid - settling timer still counting down",
            10: " not defined",
            11: " not defined",
            12: " not defined",
            13: " conditions not valid - other",
            14: " error",
            15: " not available",
            2: " conditions not valid - transmission in gear",
            3: " conditions not valid - transmission fluid temperature too low",
            4: " conditions not valid - transmission fluid temperature too high",
            5: " conditions not valid - vehicle moving; output shaft speed too high",
            6: " conditions not valid - vehicle not level",
            7: " conditions not valid - engine speed too low",
            8: " conditions not valid - engine speed too high",
            9: " conditions not valid - no request for reading"
        },
        3028: {
            0: " less than 1 minute",
            1: " one minute",
            10: " ten minutes",
            11: " eleven minutes",
            12: " twelve minutes",
            13: " thirteen minutes",
            14: " error",
            15: " not available",
            2: " two minutes",
            3: " three minutes",
            4: " four minutes",
            5: " five minutes",
            6: " six minutes",
            7: " seven minutes",
            8: " eight minutes",
            9: " nine minutes"
        },
        3034: {
            0: "                                   reserved for sae assignment",
            1: "                                   request to clear/reset a specific previously active dtc",
            10: "                         reserved for sae assignment",
            100: "                       reserved for sae assignment",
            101: "                       reserved for sae assignment",
            102: "                       reserved for sae assignment",
            103: "                       reserved for sae assignment",
            104: "                       reserved for sae assignment",
            105: "                       reserved for sae assignment",
            106: "                       reserved for sae assignment",
            107: "                       reserved for sae assignment",
            108: "                       reserved for sae assignment",
            109: "                       reserved for sae assignment",
            11: "                         reserved for sae assignment",
            110: "                       reserved for sae assignment",
            111: "                       reserved for sae assignment",
            112: "                       reserved for sae assignment",
            113: "                       reserved for sae assignment",
            114: "                       reserved for sae assignment",
            115: "                       reserved for sae assignment",
            116: "                       reserved for sae assignment",
            117: "                       reserved for sae assignment",
            118: "                       reserved for sae assignment",
            119: "                       reserved for sae assignment",
            12: "                         reserved for sae assignment",
            120: "                       reserved for sae assignment",
            121: "                       reserved for sae assignment",
            122: "                       reserved for sae assignment",
            123: "                       reserved for sae assignment",
            124: "                       reserved for sae assignment",
            125: "                       reserved for sae assignment",
            126: "                       reserved for sae assignment",
            127: "                       reserved for sae assignment",
            128: "                       reserved for sae assignment",
            129: "                       reserved for sae assignment",
            13: "                         reserved for sae assignment",
            130: "                       reserved for sae assignment",
            131: "                       reserved for sae assignment",
            132: "                       reserved for sae assignment",
            133: "                       reserved for sae assignment",
            134: "                       reserved for sae assignment",
            135: "                       reserved for sae assignment",
            136: "                       reserved for sae assignment",
            137: "                       reserved for sae assignment",
            138: "                       reserved for sae assignment",
            139: "                       reserved for sae assignment",
            14: "                         reserved for sae assignment",
            140: "                       reserved for sae assignment",
            141: "                       reserved for sae assignment",
            142: "                       reserved for sae assignment",
            143: "                       reserved for sae assignment",
            144: "                       reserved for sae assignment",
            145: "                       reserved for sae assignment",
            146: "                       reserved for sae assignment",
            147: "                       reserved for sae assignment",
            148: "                       reserved for sae assignment",
            149: "                       reserved for sae assignment",
            15: "                         reserved for sae assignment",
            150: "                       reserved for sae assignment",
            151: "                       reserved for sae assignment",
            152: "                       reserved for sae assignment",
            153: "                       reserved for sae assignment",
            154: "                       reserved for sae assignment",
            155: "                       reserved for sae assignment",
            156: "                       reserved for sae assignment",
            157: "                       reserved for sae assignment",
            158: "                       reserved for sae assignment",
            159: "                       reserved for sae assignment",
            16: "                         reserved for sae assignment",
            160: "                       reserved for sae assignment",
            161: "                       reserved for sae assignment",
            162: "                       reserved for sae assignment",
            163: "                       reserved for sae assignment",
            164: "                       reserved for sae assignment",
            165: "                       reserved for sae assignment",
            166: "                       reserved for sae assignment",
            167: "                       reserved for sae assignment",
            168: "                       reserved for sae assignment",
            169: "                       reserved for sae assignment",
            17: "                                   request to clear/reset a specific active dtc",
            170: "                       reserved for sae assignment",
            171: "                       reserved for sae assignment",
            172: "                       reserved for sae assignment",
            173: "                       reserved for sae assignment",
            174: "                       reserved for sae assignment",
            175: "                       reserved for sae assignment",
            176: "                       reserved for sae assignment",
            177: "                       reserved for sae assignment",
            178: "                       reserved for sae assignment",
            179: "                       reserved for sae assignment",
            18: "                                   positive acknowledge of clear/reset of a specific active dtc",
            180: "                       reserved for sae assignment",
            181: "                       reserved for sae assignment",
            182: "                       reserved for sae assignment",
            183: "                       reserved for sae assignment",
            184: "                       reserved for sae assignment",
            185: "                       reserved for sae assignment",
            186: "                       reserved for sae assignment",
            187: "                       reserved for sae assignment",
            188: "                       reserved for sae assignment",
            189: "                       reserved for sae assignment",
            19: "                                   negative acknowledge of clear/reset of a specific active dtc",
            190: "                       reserved for sae assignment",
            191: "                       reserved for sae assignment",
            192: "                       reserved for sae assignment",
            193: "                       reserved for sae assignment",
            194: "                       reserved for sae assignment",
            195: "                       reserved for sae assignment",
            196: "                       reserved for sae assignment",
            197: "                       reserved for sae assignment",
            198: "                       reserved for sae assignment",
            199: "                       reserved for sae assignment",
            2: "                                   positive acknowledge of clear/reset of a specific previously active dtc",
            20: "                       reserved for sae assignment",
            200: "                       reserved for sae assignment",
            201: "                       reserved for sae assignment",
            202: "                       reserved for sae assignment",
            203: "                       reserved for sae assignment",
            204: "                       reserved for sae assignment",
            205: "                       reserved for sae assignment",
            206: "                       reserved for sae assignment",
            207: "                       reserved for sae assignment",
            208: "                       reserved for sae assignment",
            209: "                       reserved for sae assignment",
            21: "                       reserved for sae assignment",
            210: "                       reserved for sae assignment",
            211: "                       reserved for sae assignment",
            212: "                       reserved for sae assignment",
            213: "                       reserved for sae assignment",
            214: "                       reserved for sae assignment",
            215: "                       reserved for sae assignment",
            216: "                       reserved for sae assignment",
            217: "                       reserved for sae assignment",
            218: "                       reserved for sae assignment",
            219: "                       reserved for sae assignment",
            22: "                       reserved for sae assignment",
            220: "                       reserved for sae assignment",
            221: "                       reserved for sae assignment",
            222: "                       reserved for sae assignment",
            223: "                       reserved for sae assignment",
            224: "                       reserved for sae assignment",
            225: "                       reserved for sae assignment",
            226: "                       reserved for sae assignment",
            227: "                       reserved for sae assignment",
            228: "                       reserved for sae assignment",
            229: "                       reserved for sae assignment",
            23: "                       reserved for sae assignment",
            230: "                       reserved for sae assignment",
            231: "                       reserved for sae assignment",
            232: "                       reserved for sae assignment",
            233: "                       reserved for sae assignment",
            234: "                       reserved for sae assignment",
            235: "                       reserved for sae assignment",
            236: "                       reserved for sae assignment",
            237: "                       reserved for sae assignment",
            238: "                       reserved for sae assignment",
            239: "                       reserved for sae assignment",
            24: "                       reserved for sae assignment",
            240: "                       reserved for sae assignment",
            241: "                       reserved for sae assignment",
            242: "                       reserved for sae assignment",
            243: "                       reserved for sae assignment",
            244: "                       reserved for sae assignment",
            245: "                       reserved for sae assignment",
            246: "                       reserved for sae assignment",
            247: "                       reserved for sae assignment",
            248: "                       reserved for sae assignment",
            249: "                       reserved for sae assignment",
            25: "                       reserved for sae assignment",
            250: "                       reserved for sae assignment",
            251: "                       reserved for sae assignment",
            252: "                       reserved for sae assignment",
            253: "                       reserved for sae assignment",
            254: "                       reserved for sae assignment",
            255: "                                 not available",
            26: "                       reserved for sae assignment",
            27: "                       reserved for sae assignment",
            28: "                       reserved for sae assignment",
            29: "                       reserved for sae assignment",
            3: "                                   negative acknowledge of clear/reset of a specific previously active dtc",
            30: "                       reserved for sae assignment",
            31: "                       reserved for sae assignment",
            32: "                       reserved for sae assignment",
            33: "                       reserved for sae assignment",
            34: "                       reserved for sae assignment",
            35: "                       reserved for sae assignment",
            36: "                       reserved for sae assignment",
            37: "                       reserved for sae assignment",
            38: "                       reserved for sae assignment",
            39: "                       reserved for sae assignment",
            4: "                         reserved for sae assignment",
            40: "                       reserved for sae assignment",
            41: "                       reserved for sae assignment",
            42: "                       reserved for sae assignment",
            43: "                       reserved for sae assignment",
            44: "                       reserved for sae assignment",
            45: "                       reserved for sae assignment",
            46: "                       reserved for sae assignment",
            47: "                       reserved for sae assignment",
            48: "                       reserved for sae assignment",
            49: "                       reserved for sae assignment",
            5: "                         reserved for sae assignment",
            50: "                       reserved for sae assignment",
            51: "                       reserved for sae assignment",
            52: "                       reserved for sae assignment",
            53: "                       reserved for sae assignment",
            54: "                       reserved for sae assignment",
            55: "                       reserved for sae assignment",
            56: "                       reserved for sae assignment",
            57: "                       reserved for sae assignment",
            58: "                       reserved for sae assignment",
            59: "                       reserved for sae assignment",
            6: "                         reserved for sae assignment",
            60: "                       reserved for sae assignment",
            61: "                       reserved for sae assignment",
            62: "                       reserved for sae assignment",
            63: "                       reserved for sae assignment",
            64: "                       reserved for sae assignment",
            65: "                       reserved for sae assignment",
            66: "                       reserved for sae assignment",
            67: "                       reserved for sae assignment",
            68: "                       reserved for sae assignment",
            69: "                       reserved for sae assignment",
            7: "                         reserved for sae assignment",
            70: "                       reserved for sae assignment",
            71: "                       reserved for sae assignment",
            72: "                       reserved for sae assignment",
            73: "                       reserved for sae assignment",
            74: "                       reserved for sae assignment",
            75: "                       reserved for sae assignment",
            76: "                       reserved for sae assignment",
            77: "                       reserved for sae assignment",
            78: "                       reserved for sae assignment",
            79: "                       reserved for sae assignment",
            8: "                         reserved for sae assignment",
            80: "                       reserved for sae assignment",
            81: "                       reserved for sae assignment",
            82: "                       reserved for sae assignment",
            83: "                       reserved for sae assignment",
            84: "                       reserved for sae assignment",
            85: "                       reserved for sae assignment",
            86: "                       reserved for sae assignment",
            87: "                       reserved for sae assignment",
            88: "                       reserved for sae assignment",
            89: "                       reserved for sae assignment",
            9: "                         reserved for sae assignment",
            90: "                       reserved for sae assignment",
            91: "                       reserved for sae assignment",
            92: "                       reserved for sae assignment",
            93: "                       reserved for sae assignment",
            94: "                       reserved for sae assignment",
            95: "                       reserved for sae assignment",
            96: "                       reserved for sae assignment",
            97: "                       reserved for sae assignment",
            98: "                       reserved for sae assignment",
            99: "                       reserved for sae assignment"
        },
        3035: {
            0: "               general negative acknowledge",
            1: "               access denied (security denied access)",
            10: "   reserved for sae assignment",
            100: "   reserved for sae assignment",
            101: "   reserved for sae assignment",
            102: "   reserved for sae assignment",
            103: "   reserved for sae assignment",
            104: "   reserved for sae assignment",
            105: "   reserved for sae assignment",
            106: "   reserved for sae assignment",
            107: "   reserved for sae assignment",
            108: "   reserved for sae assignment",
            109: "   reserved for sae assignment",
            11: "   reserved for sae assignment",
            110: "   reserved for sae assignment",
            111: "   reserved for sae assignment",
            112: "   reserved for sae assignment",
            113: "   reserved for sae assignment",
            114: "   reserved for sae assignment",
            115: "   reserved for sae assignment",
            116: "   reserved for sae assignment",
            117: "   reserved for sae assignment",
            118: "   reserved for sae assignment",
            119: "   reserved for sae assignment",
            12: "   reserved for sae assignment",
            120: "   reserved for sae assignment",
            121: "   reserved for sae assignment",
            122: "   reserved for sae assignment",
            123: "   reserved for sae assignment",
            124: "   reserved for sae assignment",
            125: "   reserved for sae assignment",
            126: "   reserved for sae assignment",
            127: "   reserved for sae assignment",
            128: "   reserved for sae assignment",
            129: "   reserved for sae assignment",
            13: "   reserved for sae assignment",
            130: "   reserved for sae assignment",
            131: "   reserved for sae assignment",
            132: "   reserved for sae assignment",
            133: "   reserved for sae assignment",
            134: "   reserved for sae assignment",
            135: "   reserved for sae assignment",
            136: "   reserved for sae assignment",
            137: "   reserved for sae assignment",
            138: "   reserved for sae assignment",
            139: "   reserved for sae assignment",
            14: "   reserved for sae assignment",
            140: "   reserved for sae assignment",
            141: "   reserved for sae assignment",
            142: "   reserved for sae assignment",
            143: "   reserved for sae assignment",
            144: "   reserved for sae assignment",
            145: "   reserved for sae assignment",
            146: "   reserved for sae assignment",
            147: "   reserved for sae assignment",
            148: "   reserved for sae assignment",
            149: "   reserved for sae assignment",
            15: "   reserved for sae assignment",
            150: "   reserved for sae assignment",
            151: "   reserved for sae assignment",
            152: "   reserved for sae assignment",
            153: "   reserved for sae assignment",
            154: "   reserved for sae assignment",
            155: "   reserved for sae assignment",
            156: "   reserved for sae assignment",
            157: "   reserved for sae assignment",
            158: "   reserved for sae assignment",
            159: "   reserved for sae assignment",
            16: "   reserved for sae assignment",
            160: "   reserved for sae assignment",
            161: "   reserved for sae assignment",
            162: "   reserved for sae assignment",
            163: "   reserved for sae assignment",
            164: "   reserved for sae assignment",
            165: "   reserved for sae assignment",
            166: "   reserved for sae assignment",
            167: "   reserved for sae assignment",
            168: "   reserved for sae assignment",
            169: "   reserved for sae assignment",
            17: "   reserved for sae assignment",
            170: "   reserved for sae assignment",
            171: "   reserved for sae assignment",
            172: "   reserved for sae assignment",
            173: "   reserved for sae assignment",
            174: "   reserved for sae assignment",
            175: "   reserved for sae assignment",
            176: "   reserved for sae assignment",
            177: "   reserved for sae assignment",
            178: "   reserved for sae assignment",
            179: "   reserved for sae assignment",
            18: "   reserved for sae assignment",
            180: "   reserved for sae assignment",
            181: "   reserved for sae assignment",
            182: "   reserved for sae assignment",
            183: "   reserved for sae assignment",
            184: "   reserved for sae assignment",
            185: "   reserved for sae assignment",
            186: "   reserved for sae assignment",
            187: "   reserved for sae assignment",
            188: "   reserved for sae assignment",
            189: "   reserved for sae assignment",
            19: "   reserved for sae assignment",
            190: "   reserved for sae assignment",
            191: "   reserved for sae assignment",
            192: "   reserved for sae assignment",
            193: "   reserved for sae assignment",
            194: "   reserved for sae assignment",
            195: "   reserved for sae assignment",
            196: "   reserved for sae assignment",
            197: "   reserved for sae assignment",
            198: "   reserved for sae assignment",
            199: "   reserved for sae assignment",
            2: "               diagnostic trouble code unknown/does not exist",
            20: "   reserved for sae assignment",
            200: "   reserved for sae assignment",
            201: "   reserved for sae assignment",
            202: "   reserved for sae assignment",
            203: "   reserved for sae assignment",
            204: "   reserved for sae assignment",
            205: "   reserved for sae assignment",
            206: "   reserved for sae assignment",
            207: "   reserved for sae assignment",
            208: "   reserved for sae assignment",
            209: "   reserved for sae assignment",
            21: "   reserved for sae assignment",
            210: "   reserved for sae assignment",
            211: "   reserved for sae assignment",
            212: "   reserved for sae assignment",
            213: "   reserved for sae assignment",
            214: "   reserved for sae assignment",
            215: "   reserved for sae assignment",
            216: "   reserved for sae assignment",
            217: "   reserved for sae assignment",
            218: "   reserved for sae assignment",
            219: "   reserved for sae assignment",
            22: "   reserved for sae assignment",
            220: "   reserved for sae assignment",
            221: "   reserved for sae assignment",
            222: "   reserved for sae assignment",
            223: "   reserved for sae assignment",
            224: "   reserved for sae assignment",
            225: "   reserved for sae assignment",
            226: "   reserved for sae assignment",
            227: "   reserved for sae assignment",
            228: "   reserved for sae assignment",
            229: "   reserved for sae assignment",
            23: "   reserved for sae assignment",
            230: "   reserved for sae assignment",
            231: "   reserved for sae assignment",
            232: "   reserved for sae assignment",
            233: "   reserved for sae assignment",
            234: "   reserved for sae assignment",
            235: "   reserved for sae assignment",
            236: "   reserved for sae assignment",
            237: "   reserved for sae assignment",
            238: "   reserved for sae assignment",
            239: "   reserved for sae assignment",
            24: "   reserved for sae assignment",
            240: "   reserved for sae assignment",
            241: "   reserved for sae assignment",
            242: "   reserved for sae assignment",
            243: "   reserved for sae assignment",
            244: "   reserved for sae assignment",
            245: "   reserved for sae assignment",
            246: "   reserved for sae assignment",
            247: "   reserved for sae assignment",
            248: "   reserved for sae assignment",
            249: "   reserved for sae assignment",
            25: "   reserved for sae assignment",
            250: "   reserved for sae assignment",
            251: "   reserved for sae assignment",
            252: "   reserved for sae assignment",
            253: "   reserved for sae assignment",
            254: "   reserved for sae assignment",
            255: "             not available",
            26: "   reserved for sae assignment",
            27: "   reserved for sae assignment",
            28: "   reserved for sae assignment",
            29: "   reserved for sae assignment",
            3: "               diagnostic trouble code no longer previously active",
            30: "   reserved for sae assignment",
            31: "   reserved for sae assignment",
            32: "   reserved for sae assignment",
            33: "   reserved for sae assignment",
            34: "   reserved for sae assignment",
            35: "   reserved for sae assignment",
            36: "   reserved for sae assignment",
            37: "   reserved for sae assignment",
            38: "   reserved for sae assignment",
            39: "   reserved for sae assignment",
            4: "               diagnostic trouble code no longer active",
            40: "   reserved for sae assignment",
            41: "   reserved for sae assignment",
            42: "   reserved for sae assignment",
            43: "   reserved for sae assignment",
            44: "   reserved for sae assignment",
            45: "   reserved for sae assignment",
            46: "   reserved for sae assignment",
            47: "   reserved for sae assignment",
            48: "   reserved for sae assignment",
            49: "   reserved for sae assignment",
            5: "   reserved for sae assignment",
            50: "   reserved for sae assignment",
            51: "   reserved for sae assignment",
            52: "   reserved for sae assignment",
            53: "   reserved for sae assignment",
            54: "   reserved for sae assignment",
            55: "   reserved for sae assignment",
            56: "   reserved for sae assignment",
            57: "   reserved for sae assignment",
            58: "   reserved for sae assignment",
            59: "   reserved for sae assignment",
            6: "   reserved for sae assignment",
            60: "   reserved for sae assignment",
            61: "   reserved for sae assignment",
            62: "   reserved for sae assignment",
            63: "   reserved for sae assignment",
            64: "   reserved for sae assignment",
            65: "   reserved for sae assignment",
            66: "   reserved for sae assignment",
            67: "   reserved for sae assignment",
            68: "   reserved for sae assignment",
            69: "   reserved for sae assignment",
            7: "   reserved for sae assignment",
            70: "   reserved for sae assignment",
            71: "   reserved for sae assignment",
            72: "   reserved for sae assignment",
            73: "   reserved for sae assignment",
            74: "   reserved for sae assignment",
            75: "   reserved for sae assignment",
            76: "   reserved for sae assignment",
            77: "   reserved for sae assignment",
            78: "   reserved for sae assignment",
            79: "   reserved for sae assignment",
            8: "   reserved for sae assignment",
            80: "   reserved for sae assignment",
            81: "   reserved for sae assignment",
            82: "   reserved for sae assignment",
            83: "   reserved for sae assignment",
            84: "   reserved for sae assignment",
            85: "   reserved for sae assignment",
            86: "   reserved for sae assignment",
            87: "   reserved for sae assignment",
            88: "   reserved for sae assignment",
            89: "   reserved for sae assignment",
            9: "   reserved for sae assignment",
            90: "   reserved for sae assignment",
            91: "   reserved for sae assignment",
            92: "   reserved for sae assignment",
            93: "   reserved for sae assignment",
            94: "   reserved for sae assignment",
            95: "   reserved for sae assignment",
            96: "   reserved for sae assignment",
            97: "   reserved for sae assignment",
            98: "   reserved for sae assignment",
            99: "   reserved for sae assignment"
        },
        3038: {
            0: "slow flash (1 hz, 50 % duty cycle)",
            1: "fast flash (2 hz or faster, 50% duty cycle)",
            2: "class c dtc  ",
            3: "unavailable / do not flash"
        },
        3039: {
            0: "slow flash (1 hz, 50 % duty cycle)",
            1: "fast flash (2 hz or faster, 50% duty cycle)",
            2: "reserved",
            3: "unavailable / do not flash"
        },
        3040: {
            0: "slow flash (1 hz, 50 % duty cycle)",
            1: "fast flash (2 hz or faster, 50% duty cycle)",
            2: "reserved",
            3: "unavailable / do not flash"
        },
        3041: {
            0: "slow flash (1 hz, 50 % duty cycle)",
            1: "fast flash (2 hz or faster, 50% duty cycle)",
            2: "reserved",
            3: "unavailable / do not flash"
        },
        3043: {
            0: "absolute passenger count",
            1: "boarding passenger",
            2: "exiting passenger",
            251: " to 253 - reserved",
            254: "error indicator",
            255: "not available",
            3: "boarding passenger (second passenger stream)",
            4: "exiting passenger (second passenger stream)",
            5: " to 250 - reserved for future assignment"
        },
        3044: {
            0: "off",
            1: "on",
            2: "error condition",
            3: "not available"
        },
        3045: {
            0: "normal use",
            1: "unauthorized use",
            2: "error condition",
            3: "not available"
        },
        3046: {
            0: "off",
            1: "on",
            2: "error condition",
            3: "not available"
        },
        3079: {
            0: "message is a request directed to the emitter",
            1: "message is a response from the emitter",
            2: "error condition",
            3: "not available"
        },
        3080: {
            0: "transit route id not used for interleaved data",
            1: "transit route id used for interleaved data (if range code not enabled)",
            2: "error condition",
            3: "not available"
        },
        3081: {
            0: "range code not used for interleaved data",
            1: "range code used for interleaved data",
            2: "error condition",
            3: "not available"
        },
        3082: {
            0: "deactivate strobe",
            1: "activate strobe",
            2: "error condition",
            3: "not available"
        },
        3083: {
            0: "ignore transit door status",
            1: "transit door status will override strobe activation",
            2: "error condition",
            3: "not available"
        },
        3084: {
            0: "reserved",
            1: "low priority",
            10: "priority set by hardware to probe priority",
            11: "priority set by hardware to high priority",
            12: " to 1101 - reserved",
            14: "error condition",
            15: "not available",
            2: "probe priority",
            3: "high priority",
            4: " to 1000 - reserved",
            9: "priority set by hardware to low priority"
        },
        3086: {
            0: "transmission not ready for brake release",
            1: "transmission ready for brake release",
            2: "error",
            3: "not available"
        },
        3102: {
            0: "    off",
            1: "    on",
            2: "    reserved",
            3: "    don't care/take no action"
        },
        3103: {
            0: "    off",
            1: "    on",
            2: "    reserved",
            3: "    don't care/take no action"
        },
        3104: {
            0: "    off",
            1: "    on",
            2: "    reserved",
            3: "    don't care/take no action"
        },
        3105: {
            0: "    off",
            1: "    on",
            2: "    reserved",
            3: "    don't care/take no action"
        },
        3106: {
            0: "    off",
            1: "    on",
            2: "    reserved",
            3: "    don't care/take no action"
        },
        3107: {
            0: "    off",
            1: "    on",
            2: "    reserved",
            3: "    don't care/take no action"
        },
        3108: {
            0: "    off",
            1: "    on",
            2: "    reserved",
            3: "    don't care/take no action"
        },
        3109: {
            0: "    off",
            1: "    on",
            2: "    reserved",
            3: "    don't care/take no action"
        },
        3110: {
            0: "    off",
            1: "    on",
            2: "    reserved",
            3: "    don't care/take no action"
        },
        3111: {
            0: "    off",
            1: "    on",
            2: "    reserved",
            3: "    don't care/take no action"
        },
        3112: {
            0: "    off",
            1: "    on",
            2: "    reserved",
            3: "    don't care/take no action"
        },
        3113: {
            0: "    off",
            1: "    on",
            2: "    reserved",
            3: "    don't care/take no action"
        },
        3114: {
            0: "    off",
            1: "    on",
            2: "    reserved",
            3: "    don't care/take no action"
        },
        3115: {
            0: "    off",
            1: "    on",
            2: "    reserved",
            3: "    don't care/take no action"
        },
        3116: {
            0: "    off",
            1: "    on",
            2: "    reserved",
            3: "    don't care/take no action"
        },
        3117: {
            0: "    off",
            1: "    on",
            2: "    reserved",
            3: "    don't care/take no action"
        },
        3118: {
            0: "    off",
            1: "    on",
            2: "    reserved",
            3: "    don't care/take no action"
        },
        3119: {
            0: "    off",
            1: "    on",
            2: "    reserved",
            3: "    don't care/take no action"
        },
        3120: {
            0: "    off",
            1: "    on",
            2: "    reserved",
            3: "    don't care/take no action"
        },
        3121: {
            0: "    off",
            1: "    on",
            2: "    reserved",
            3: "    don't care/take no action"
        },
        3156: {
            0: " no button pressed",
            1: ".   the switch state can be read in two ways.  one method uses the direct analog switch input to determine the switch state, while the second method relies on a secondary control to read the analog input, then relay the information on the data link.  the parameter is designed to provide the actual switch state to other controls that need the information.",
            10: " manual button pressed",
            11: " automatic button pressed",
            1110: " error indicator",
            1111: " not installed",
            2: ".   other systems with automated blade controls should be able to use this parameter, since it is a measured switch value.",
            3: ".    the momentary contact button pair may be used, where the default is no buttons being pressed.  the user can press the manual button, and when pressure is released, it will return to the no button pressed position.  the user may press the auto button, and when pressure is released, it will return to the no button pressed position.  momentary contact buttons (button pair) usage is as follows:",
            4: ".    the momentary contact single button may be used, where the default is no button being pressed.  the user can press the button, which is the button pressed position, and when pressure is released, it will return to the no button pressed position.  the controller receiving this command will keep track of what mode the blade is in, auto or manual.  when the button is pressed, the \"mode button pressed\" state is reported and the receiving controller then changes the current mode to its opposite value (i.e. manual to automatic or vise-versa).  momentary contact button (single button) usage is as follows:"
        },
        3157: {
            0: " no button pressed",
            1: ".   the switch state can be read in two ways.  one method uses the direct analog switch input to determine the switch state, while the second method relies on a secondary control to read the analog input, then relay the information on the data link.  the parameter is designed to provide the actual switch state to other controls that need the information.",
            10: " decrement button pressed",
            1110: " error indicator",
            1111: " not installed",
            2: ".   other systems with automated blade controls should be able to use this parameter, since it is a measured switch value."
        },
        3158: {
            0: " auto allowed",
            1: " activity allows auto - must detect operator before auto allowed",
            14: " error indicator",
            15: " not supported",
            2: " suspend auto - temporarily suspend auto mode",
            3: " service mode - auto allowed only during service modes",
            4: " auto not allowed"
        },
        3170: {
            0: "cash",
            1: "token",
            10: "reserved for assignment",
            11: "reserved for assignment",
            12: "agency defined",
            13: "agency defined",
            14: "agency defined",
            15: "agency defined",
            2: "ticket",
            3: "pass",
            4: "card",
            5: "permit",
            6: "transfer",
            7: "free",
            8: "reserved for assignment",
            9: "reserved for assignment"
        },
        3172: {
            0: "local service",
            1: "express service",
            2: "agency defined",
            3: "agency defined",
            4: "agency defined",
            5: "agency defined",
            6: "agency defined",
            7: "agency defined"
        },
        3173: {
            0: "north",
            1: "south",
            10: "agency defined",
            11: "agency defined",
            12: "agency defined",
            13: "agency defined",
            14: "agency defined",
            15: "agency defined",
            16: "agency defined",
            17: "agency defined",
            18: "agency defined",
            19: "agency defined",
            2: "east",
            20: "agency defined",
            21: "agency defined",
            22: "agency defined",
            23: "agency defined",
            24: "agency defined",
            25: "agency defined",
            26: "agency defined",
            27: "agency defined",
            28: "agency defined",
            29: "agency defined",
            3: "west",
            30: "agency defined",
            31: "agency defined",
            4: "in",
            5: "out",
            6: "agency defined",
            7: "agency defined",
            8: "agency defined",
            9: "agency defined"
        },
        3174: {
            0: "north",
            1: "south",
            10: "agency defined",
            11: "agency defined",
            12: "agency defined",
            13: "agency defined",
            14: "agency defined",
            15: "agency defined",
            2: "east",
            3: "west",
            4: "in",
            5: "out",
            6: "agency defined",
            7: "agency defined",
            8: "agency defined",
            9: "agency defined"
        },
        3175: {
            0: " to 00001111 = agency defined",
            16: " to 11111111 = reserved for assignment"
        },
        3176: {
            0: "cash/no detail",
            1: "token a",
            10: "reserved for assignment",
            11: "agency defined",
            12: "agency defined",
            13: "agency defined",
            14: "agency defined",
            15: "agency defined",
            2: "token b",
            3: "ticket a",
            4: "ticket b",
            5: "pass a",
            6: "pass b",
            7: "reserved for assignment",
            8: "reserved for assignment",
            9: "reserved for assignment"
        },
        3177: {
            0: "not an upgrade",
            1: "cash",
            10: "reserved for assignment",
            11: "agency defined",
            12: "agency defined",
            13: "agency defined",
            14: "agency defined",
            15: "agency defined",
            2: "token",
            3: "ticket",
            4: "pass",
            5: "card",
            6: "reserved for assignment",
            7: "reserved for assignment",
            8: "reserved for assignment",
            9: "reserved for assignment"
        },
        3178: {
            0: "farebox out of service",
            1: "farebox in service",
            2: "error condition",
            3: "not available"
        },
        3179: {
            0: "non-emergency condition",
            1: "emergency condition",
            2: "error condition",
            3: "not available"
        },
        3180: {
            0: "undefined",
            1: "trip start",
            2: "trip end",
            3: "undefined",
            4: "undefined",
            5: "layover start",
            6: "layover end",
            7: "undefined"
        },
        3181: {
            0: "voltage dropout",
            1: "voltage restored",
            10: "coinbox full",
            100: "agency defined",
            101: "agency defined",
            102: "agency defined",
            103: "agency defined",
            104: "agency defined",
            105: "agency defined",
            106: "agency defined",
            107: "agency defined",
            108: "agency defined",
            109: "agency defined",
            11: "currency box 75% full",
            110: "agency defined",
            111: "agency defined",
            112: "agency defined",
            113: "agency defined",
            114: "agency defined",
            115: "agency defined",
            116: "agency defined",
            117: "agency defined",
            118: "agency defined",
            119: "agency defined",
            12: "currency box less than 75% full",
            120: "agency defined",
            121: "agency defined",
            122: "agency defined",
            123: "agency defined",
            124: "agency defined",
            125: "agency defined",
            126: "agency defined",
            127: "agency defined",
            13: "currency box full",
            14: "card/pass box 75% full",
            15: "card/pass box less than 75% full",
            16: "card/pass box full",
            17: "coin de-jam operated",
            18: "farebox set in manual bypass",
            19: "farebox reset to automatic mode",
            2: "probe started",
            20: "pass/transfer jam",
            21: "pass/transfer jam cleared",
            22: "paper currency jam",
            23: "paper currency jam cleared",
            24: "maintenance access in service, considered an alarm condition",
            25: "maintenance access out of service",
            26: "reserved - to be assigned",
            27: "reserved - to be assigned",
            28: "reserved - to be assigned",
            29: "reserved - to be assigned",
            3: "probe completed",
            30: "reserved - to be assigned",
            31: "reserved - to be assigned",
            32: "reserved - to be assigned",
            33: "reserved - to be assigned",
            34: "reserved - to be assigned",
            35: "reserved - to be assigned",
            36: "reserved - to be assigned",
            37: "reserved - to be assigned",
            38: "reserved - to be assigned",
            39: "reserved - to be assigned",
            4: "cashbox removed",
            40: "reserved - to be assigned",
            41: "reserved - to be assigned",
            42: "reserved - to be assigned",
            43: "reserved - to be assigned",
            44: "reserved - to be assigned",
            45: "reserved - to be assigned",
            46: "reserved - to be assigned",
            47: "reserved - to be assigned",
            48: "reserved - to be assigned",
            49: "reserved - to be assigned",
            5: "cashbox restored",
            50: "reserved - to be assigned",
            51: "reserved - to be assigned",
            52: "reserved - to be assigned",
            53: "reserved - to be assigned",
            54: "reserved - to be assigned",
            55: "reserved - to be assigned",
            56: "reserved - to be assigned",
            57: "reserved - to be assigned",
            58: "reserved - to be assigned",
            59: "reserved - to be assigned",
            6: "cashbox door timeout",
            60: "reserved - to be assigned",
            61: "reserved - to be assigned",
            62: "reserved - to be assigned",
            63: "reserved - to be assigned",
            64: "reserved - to be assigned",
            65: "reserved - to be assigned",
            66: "reserved - to be assigned",
            67: "reserved - to be assigned",
            68: "reserved - to be assigned",
            69: "reserved - to be assigned",
            7: "cashbox opened in service, considered an alarm condition",
            70: "reserved - to be assigned",
            71: "reserved - to be assigned",
            72: "reserved - to be assigned",
            73: "reserved - to be assigned",
            74: "reserved - to be assigned",
            75: "reserved - to be assigned",
            76: "reserved - to be assigned",
            77: "reserved - to be assigned",
            78: "reserved - to be assigned",
            79: "reserved - to be assigned",
            8: "insufficient fare accepted",
            80: "reserved - to be assigned",
            81: "reserved - to be assigned",
            82: "reserved - to be assigned",
            83: "reserved - to be assigned",
            84: "reserved - to be assigned",
            85: "reserved - to be assigned",
            86: "reserved - to be assigned",
            87: "reserved - to be assigned",
            88: "reserved - to be assigned",
            89: "reserved - to be assigned",
            9: "coinbox 75% full",
            90: "reserved - to be assigned",
            91: "reserved - to be assigned",
            92: "reserved - to be assigned",
            93: "reserved - to be assigned",
            94: "reserved - to be assigned",
            95: "reserved - to be assigned",
            96: "reserved - to be assigned",
            97: "agency defined",
            98: "agency defined",
            99: "agency defined"
        },
        3215: {
            0: " no request ",
            1: " request prohibit air suspension control ",
            2: " error indicator ",
            3: " not available"
        },
        3218: {
            0: "not in range",
            1: "in range",
            2: "error",
            3: "not available"
        },
        3219: {
            0: "not in range",
            1: "in range",
            2: "error",
            3: "not available"
        },
        3220: {
            0: "not stable",
            1: "stable",
            2: "error",
            3: "not available"
        },
        3221: {
            0: "not stable",
            1: "stable",
            2: "error",
            3: "not available"
        },
        3223: {
            0: "automatic",
            1: "preheat 2",
            2: "preheat 1",
            3: "heater off"
        },
        3228: {
            0: "not in range",
            1: "in range",
            2: "error",
            3: "not available"
        },
        3229: {
            0: "not in range",
            1: "in range",
            2: "error",
            3: "not available"
        },
        3230: {
            0: "not stable",
            1: "stable",
            2: "error",
            3: "not available"
        },
        3231: {
            0: "not stable",
            1: "stable",
            2: "error",
            3: "not available"
        },
        3233: {
            0: "automatic",
            1: "preheat 2",
            2: "preheat 1",
            3: "heater off"
        },
        3237: {
            0: "not exceeded the dew point",
            1: "exceeded the dew point",
            2: "error",
            3: "not available"
        },
        3238: {
            0: "not exceeded the dew point",
            1: "exceeded the dew point",
            2: "error",
            3: "not available"
        },
        3239: {
            0: "not exceeded the dew point",
            1: "exceeded the dew point",
            2: "error",
            3: "not available"
        },
        3240: {
            0: "not exceeded the dew point",
            1: "exceeded the dew point",
            2: "error",
            3: "not available"
        },
        3257: {
            0: "not in range",
            1: "in range",
            2: "error",
            3: "not available"
        },
        3258: {
            0: "not in range",
            1: "in range",
            2: "error",
            3: "not available"
        },
        3259: {
            0: "not stable",
            1: "stable",
            2: "error",
            3: "not available"
        },
        3260: {
            0: "not stable",
            1: "stable",
            2: "error",
            3: "not available"
        },
        3262: {
            0: "automatic",
            1: "preheat 2",
            2: "preheat 1",
            3: "heater off"
        },
        3267: {
            0: "not in range",
            1: "in range",
            2: "error",
            3: "not available"
        },
        3268: {
            0: "not in range",
            1: "in range",
            2: "error",
            3: "not available"
        },
        3269: {
            0: "not stable",
            1: "stable",
            2: "error",
            3: "not available"
        },
        3270: {
            0: "not stable",
            1: "stable",
            2: "error",
            3: "not available"
        },
        3272: {
            0: "automatic",
            1: "preheat 2",
            2: "preheat 1",
            3: "heater off"
        },
        3307: {
            0: "    0 - improper couple",
            1: "    1 - locks closed / no kingpin at startup",
            10: "    10 - not used",
            11: "    11 - not used",
            12: "    12 - not used",
            13: "    13 - not used",
            14: "    14 - not used",
            15: "    15 - not available",
            2: "    2 - kingpin / locks open at startup",
            3: "    3 - not used",
            4: "    4 - not used",
            5: "    5 - coupling sequence error",
            6: "    6 - coupling sequence error / kingpin",
            7: "    7 - coupling timing error",
            8: "    8 - kingpin missing more than 5 seconds",
            9: "    9 - kingpin / lock open (handle pulled)"
        },
        3312: {
            0: "   lock not open, not ready to couple",
            1: "   lock open, ready to couple",
            2: "   error",
            3: "   not available"
        },
        3313: {
            0: "   incomplete or bad couple, truck should not be moved",
            1: "   successful couple detected",
            2: "   error",
            3: "   not available"
        },
        3314: {
            0: "   coupler locked",
            1: "   coupler unlocked, trailer can be unhitched or hitched",
            2: "   error",
            3: "   not available"
        },
        3315: {
            0: "    fifth wheel release disabled",
            1: "    fifth wheel release enabled",
            2: "    error   ",
            3: "    not available"
        },
        3316: {
            0: " not locked",
            1: " locked",
            2: " error",
            3: " not available"
        },
        3317: {
            0: " limit not exceeded",
            1: " limit exceeded",
            2: " error",
            3: " not available"
        },
        3323: {
            0: "           1  pitch angle degraded.  data is suspect due to environmental conditions.",
            1: "           1  not available"
        },
        3324: {
            0: "        1  roll angle degraded.  data is suspect due to environmental conditions.",
            1: "        1  not available"
        },
        3325: {
            0: "        1  pitch rate degraded.  data is suspect due to environmental conditions.",
            1: "        1  not available"
        },
        3326: {
            0: "        1   compensation on",
            1: "        1   not available"
        },
        3332: {
            0: "  blade rotation angle fully functional.  data is within sensor specification.   ",
            1: "  blade rotation angle  degraded.  data is suspect due to environmental conditions.",
            2: "  blade rotation angle failed.  roll sensor failed to operate correctly.",
            3: "  blade rotation angle not available"
        },
        3334: {
            0: "    no change",
            1: ".  the operator control state can be read in two ways.  one method uses the direct operator control input to determine the operator control state, while the second method relies on a secondary control to read the operator control input, then relay the information on the data link.  the parameter is designed to provide the actual operator control state to other controls that need the information.",
            10: "    manual operator control request",
            11: "    automatic operator control request",
            1100: "   reserved",
            1101: "   reserved",
            1110: "    error indicator",
            1111: "    not installed",
            2: ".  other systems with automated blade controls should be able to use this parameter, since it is an actual operator control value.",
            3: ".  the momentary contact button pair may be used, where the default is no buttons being pressed.  the user can press the manual button, and when pressure is released, it will return to the no button pressed position.  the user may press the auto button, and when pressure is released, it will return to the no button pressed position.",
            4: ".  the momentary contact single button may be used, where the default is no button being pressed.  the user can press the button, which is the button pressed position, and when pressure is released, it will return to the no button pressed position.  the controller receiving this command will keep track of what mode the blade is in, auto or manual.  if the blade is in manual mode to begin with, one press of the button will change to mode to auto.  if the button is pressed again, the blade will be put into manual mode."
        },
        3335: {
            0: "    no change",
            1: "    momentary operator control request",
            10: "    manual operator control request",
            11: "    automatic operator control request",
            1110: "    error indicator",
            1111: "    not installed",
            2: ". the momentary rocker switch may be used, where the default is the no button pressed position.  the user can press the rocker into the manual position, and when pressure is released, it will return to the no button pressed position.  the user may also press the rocker into the auto position, and when pressure is released, it will return to the no button pressed position.",
            3: ".  the momentary contact button pair may be used, where the default is no buttons being pressed.  the user can press the manual button, and when pressure is released, it will return to the no button pressed position.  the user may press the auto button, and when pressure is released, it will return to the no button pressed position.",
            4: ".  the momentary contact single button may be used, where the default is no button being pressed.  the user can press the button, which is the button pressed position, and when pressure is released, it will return to the no button pressed position.  the controller receiving this command will keep track of what mode the blade is in, auto or manual.  if the blade is in manual mode to begin with, one press of the button will change to mode to auto.  if the button is pressed again, the blade will be put into manual mode."
        },
        3336: {
            0: " no operator control request",
            1: ".  the operator control state can be read in two ways.  one method uses the direct operator control input to determine the operator control state, while the second method relies on a secondary control to read the operator control input, then relay the information on the data link.  the parameter is designed to provide the actual operator control state to other controls that need the information.",
            10: " decrement operator control request",
            100: "  reserved",
            1000: "  reserved",
            1001: "  reserved",
            1002: "  reserved",
            1003: "  reserved",
            1004: "  reserved",
            1005: "  reserved",
            1006: "  reserved",
            1007: "  reserved",
            1008: "  reserved",
            1009: "  reserved",
            101: "  reserved",
            1010: "  reserved",
            1011: "  reserved",
            1012: "  reserved",
            1013: "  reserved",
            1014: "  reserved",
            1015: "  reserved",
            1016: "  reserved",
            1017: "  reserved",
            1018: "  reserved",
            1019: "  reserved",
            102: "  reserved",
            1020: "  reserved",
            1021: "  reserved",
            1022: "  reserved",
            1023: "  reserved",
            1024: "  reserved",
            1025: "  reserved",
            1026: "  reserved",
            1027: "  reserved",
            1028: "  reserved",
            1029: "  reserved",
            103: "  reserved",
            1030: "  reserved",
            1031: "  reserved",
            1032: "  reserved",
            1033: "  reserved",
            1034: "  reserved",
            1035: "  reserved",
            1036: "  reserved",
            1037: "  reserved",
            1038: "  reserved",
            1039: "  reserved",
            104: "  reserved",
            1040: "  reserved",
            1041: "  reserved",
            1042: "  reserved",
            1043: "  reserved",
            1044: "  reserved",
            1045: "  reserved",
            1046: "  reserved",
            1047: "  reserved",
            1048: "  reserved",
            1049: "  reserved",
            105: "  reserved",
            1050: "  reserved",
            1051: "  reserved",
            1052: "  reserved",
            1053: "  reserved",
            1054: "  reserved",
            1055: "  reserved",
            1056: "  reserved",
            1057: "  reserved",
            1058: "  reserved",
            1059: "  reserved",
            106: "  reserved",
            1060: "  reserved",
            1061: "  reserved",
            1062: "  reserved",
            1063: "  reserved",
            1064: "  reserved",
            1065: "  reserved",
            1066: "  reserved",
            1067: "  reserved",
            1068: "  reserved",
            1069: "  reserved",
            107: "  reserved",
            1070: "  reserved",
            1071: "  reserved",
            1072: "  reserved",
            1073: "  reserved",
            1074: "  reserved",
            1075: "  reserved",
            1076: "  reserved",
            1077: "  reserved",
            1078: "  reserved",
            1079: "  reserved",
            108: "  reserved",
            1080: "  reserved",
            1081: "  reserved",
            1082: "  reserved",
            1083: "  reserved",
            1084: "  reserved",
            1085: "  reserved",
            1086: "  reserved",
            1087: "  reserved",
            1088: "  reserved",
            1089: "  reserved",
            109: "  reserved",
            1090: "  reserved",
            1091: "  reserved",
            1092: "  reserved",
            1093: "  reserved",
            1094: "  reserved",
            1095: "  reserved",
            1096: "  reserved",
            1097: "  reserved",
            1098: "  reserved",
            1099: "  reserved",
            11: "  reserved",
            110: "  reserved",
            1100: "  reserved",
            1101: "  reserved",
            111: "  reserved",
            1110: " error indicator",
            1111: " not installed",
            112: "  reserved",
            113: "  reserved",
            114: "  reserved",
            115: "  reserved",
            116: "  reserved",
            117: "  reserved",
            118: "  reserved",
            119: "  reserved",
            12: "  reserved",
            120: "  reserved",
            121: "  reserved",
            122: "  reserved",
            123: "  reserved",
            124: "  reserved",
            125: "  reserved",
            126: "  reserved",
            127: "  reserved",
            128: "  reserved",
            129: "  reserved",
            13: "  reserved",
            130: "  reserved",
            131: "  reserved",
            132: "  reserved",
            133: "  reserved",
            134: "  reserved",
            135: "  reserved",
            136: "  reserved",
            137: "  reserved",
            138: "  reserved",
            139: "  reserved",
            14: "  reserved",
            140: "  reserved",
            141: "  reserved",
            142: "  reserved",
            143: "  reserved",
            144: "  reserved",
            145: "  reserved",
            146: "  reserved",
            147: "  reserved",
            148: "  reserved",
            149: "  reserved",
            15: "  reserved",
            150: "  reserved",
            151: "  reserved",
            152: "  reserved",
            153: "  reserved",
            154: "  reserved",
            155: "  reserved",
            156: "  reserved",
            157: "  reserved",
            158: "  reserved",
            159: "  reserved",
            16: "  reserved",
            160: "  reserved",
            161: "  reserved",
            162: "  reserved",
            163: "  reserved",
            164: "  reserved",
            165: "  reserved",
            166: "  reserved",
            167: "  reserved",
            168: "  reserved",
            169: "  reserved",
            17: "  reserved",
            170: "  reserved",
            171: "  reserved",
            172: "  reserved",
            173: "  reserved",
            174: "  reserved",
            175: "  reserved",
            176: "  reserved",
            177: "  reserved",
            178: "  reserved",
            179: "  reserved",
            18: "  reserved",
            180: "  reserved",
            181: "  reserved",
            182: "  reserved",
            183: "  reserved",
            184: "  reserved",
            185: "  reserved",
            186: "  reserved",
            187: "  reserved",
            188: "  reserved",
            189: "  reserved",
            19: "  reserved",
            190: "  reserved",
            191: "  reserved",
            192: "  reserved",
            193: "  reserved",
            194: "  reserved",
            195: "  reserved",
            196: "  reserved",
            197: "  reserved",
            198: "  reserved",
            199: "  reserved",
            2: ".  other systems with automated blade controls should be able to use this parameter, since it is an actual operator control value.",
            20: "  reserved",
            200: "  reserved",
            201: "  reserved",
            202: "  reserved",
            203: "  reserved",
            204: "  reserved",
            205: "  reserved",
            206: "  reserved",
            207: "  reserved",
            208: "  reserved",
            209: "  reserved",
            21: "  reserved",
            210: "  reserved",
            211: "  reserved",
            212: "  reserved",
            213: "  reserved",
            214: "  reserved",
            215: "  reserved",
            216: "  reserved",
            217: "  reserved",
            218: "  reserved",
            219: "  reserved",
            22: "  reserved",
            220: "  reserved",
            221: "  reserved",
            222: "  reserved",
            223: "  reserved",
            224: "  reserved",
            225: "  reserved",
            226: "  reserved",
            227: "  reserved",
            228: "  reserved",
            229: "  reserved",
            23: "  reserved",
            230: "  reserved",
            231: "  reserved",
            232: "  reserved",
            233: "  reserved",
            234: "  reserved",
            235: "  reserved",
            236: "  reserved",
            237: "  reserved",
            238: "  reserved",
            239: "  reserved",
            24: "  reserved",
            240: "  reserved",
            241: "  reserved",
            242: "  reserved",
            243: "  reserved",
            244: "  reserved",
            245: "  reserved",
            246: "  reserved",
            247: "  reserved",
            248: "  reserved",
            249: "  reserved",
            25: "  reserved",
            250: "  reserved",
            251: "  reserved",
            252: "  reserved",
            253: "  reserved",
            254: "  reserved",
            255: "  reserved",
            256: "  reserved",
            257: "  reserved",
            258: "  reserved",
            259: "  reserved",
            26: "  reserved",
            260: "  reserved",
            261: "  reserved",
            262: "  reserved",
            263: "  reserved",
            264: "  reserved",
            265: "  reserved",
            266: "  reserved",
            267: "  reserved",
            268: "  reserved",
            269: "  reserved",
            27: "  reserved",
            270: "  reserved",
            271: "  reserved",
            272: "  reserved",
            273: "  reserved",
            274: "  reserved",
            275: "  reserved",
            276: "  reserved",
            277: "  reserved",
            278: "  reserved",
            279: "  reserved",
            28: "  reserved",
            280: "  reserved",
            281: "  reserved",
            282: "  reserved",
            283: "  reserved",
            284: "  reserved",
            285: "  reserved",
            286: "  reserved",
            287: "  reserved",
            288: "  reserved",
            289: "  reserved",
            29: "  reserved",
            290: "  reserved",
            291: "  reserved",
            292: "  reserved",
            293: "  reserved",
            294: "  reserved",
            295: "  reserved",
            296: "  reserved",
            297: "  reserved",
            298: "  reserved",
            299: "  reserved",
            30: "  reserved",
            300: "  reserved",
            301: "  reserved",
            302: "  reserved",
            303: "  reserved",
            304: "  reserved",
            305: "  reserved",
            306: "  reserved",
            307: "  reserved",
            308: "  reserved",
            309: "  reserved",
            31: "  reserved",
            310: "  reserved",
            311: "  reserved",
            312: "  reserved",
            313: "  reserved",
            314: "  reserved",
            315: "  reserved",
            316: "  reserved",
            317: "  reserved",
            318: "  reserved",
            319: "  reserved",
            32: "  reserved",
            320: "  reserved",
            321: "  reserved",
            322: "  reserved",
            323: "  reserved",
            324: "  reserved",
            325: "  reserved",
            326: "  reserved",
            327: "  reserved",
            328: "  reserved",
            329: "  reserved",
            33: "  reserved",
            330: "  reserved",
            331: "  reserved",
            332: "  reserved",
            333: "  reserved",
            334: "  reserved",
            335: "  reserved",
            336: "  reserved",
            337: "  reserved",
            338: "  reserved",
            339: "  reserved",
            34: "  reserved",
            340: "  reserved",
            341: "  reserved",
            342: "  reserved",
            343: "  reserved",
            344: "  reserved",
            345: "  reserved",
            346: "  reserved",
            347: "  reserved",
            348: "  reserved",
            349: "  reserved",
            35: "  reserved",
            350: "  reserved",
            351: "  reserved",
            352: "  reserved",
            353: "  reserved",
            354: "  reserved",
            355: "  reserved",
            356: "  reserved",
            357: "  reserved",
            358: "  reserved",
            359: "  reserved",
            36: "  reserved",
            360: "  reserved",
            361: "  reserved",
            362: "  reserved",
            363: "  reserved",
            364: "  reserved",
            365: "  reserved",
            366: "  reserved",
            367: "  reserved",
            368: "  reserved",
            369: "  reserved",
            37: "  reserved",
            370: "  reserved",
            371: "  reserved",
            372: "  reserved",
            373: "  reserved",
            374: "  reserved",
            375: "  reserved",
            376: "  reserved",
            377: "  reserved",
            378: "  reserved",
            379: "  reserved",
            38: "  reserved",
            380: "  reserved",
            381: "  reserved",
            382: "  reserved",
            383: "  reserved",
            384: "  reserved",
            385: "  reserved",
            386: "  reserved",
            387: "  reserved",
            388: "  reserved",
            389: "  reserved",
            39: "  reserved",
            390: "  reserved",
            391: "  reserved",
            392: "  reserved",
            393: "  reserved",
            394: "  reserved",
            395: "  reserved",
            396: "  reserved",
            397: "  reserved",
            398: "  reserved",
            399: "  reserved",
            40: "  reserved",
            400: "  reserved",
            401: "  reserved",
            402: "  reserved",
            403: "  reserved",
            404: "  reserved",
            405: "  reserved",
            406: "  reserved",
            407: "  reserved",
            408: "  reserved",
            409: "  reserved",
            41: "  reserved",
            410: "  reserved",
            411: "  reserved",
            412: "  reserved",
            413: "  reserved",
            414: "  reserved",
            415: "  reserved",
            416: "  reserved",
            417: "  reserved",
            418: "  reserved",
            419: "  reserved",
            42: "  reserved",
            420: "  reserved",
            421: "  reserved",
            422: "  reserved",
            423: "  reserved",
            424: "  reserved",
            425: "  reserved",
            426: "  reserved",
            427: "  reserved",
            428: "  reserved",
            429: "  reserved",
            43: "  reserved",
            430: "  reserved",
            431: "  reserved",
            432: "  reserved",
            433: "  reserved",
            434: "  reserved",
            435: "  reserved",
            436: "  reserved",
            437: "  reserved",
            438: "  reserved",
            439: "  reserved",
            44: "  reserved",
            440: "  reserved",
            441: "  reserved",
            442: "  reserved",
            443: "  reserved",
            444: "  reserved",
            445: "  reserved",
            446: "  reserved",
            447: "  reserved",
            448: "  reserved",
            449: "  reserved",
            45: "  reserved",
            450: "  reserved",
            451: "  reserved",
            452: "  reserved",
            453: "  reserved",
            454: "  reserved",
            455: "  reserved",
            456: "  reserved",
            457: "  reserved",
            458: "  reserved",
            459: "  reserved",
            46: "  reserved",
            460: "  reserved",
            461: "  reserved",
            462: "  reserved",
            463: "  reserved",
            464: "  reserved",
            465: "  reserved",
            466: "  reserved",
            467: "  reserved",
            468: "  reserved",
            469: "  reserved",
            47: "  reserved",
            470: "  reserved",
            471: "  reserved",
            472: "  reserved",
            473: "  reserved",
            474: "  reserved",
            475: "  reserved",
            476: "  reserved",
            477: "  reserved",
            478: "  reserved",
            479: "  reserved",
            48: "  reserved",
            480: "  reserved",
            481: "  reserved",
            482: "  reserved",
            483: "  reserved",
            484: "  reserved",
            485: "  reserved",
            486: "  reserved",
            487: "  reserved",
            488: "  reserved",
            489: "  reserved",
            49: "  reserved",
            490: "  reserved",
            491: "  reserved",
            492: "  reserved",
            493: "  reserved",
            494: "  reserved",
            495: "  reserved",
            496: "  reserved",
            497: "  reserved",
            498: "  reserved",
            499: "  reserved",
            50: "  reserved",
            500: "  reserved",
            501: "  reserved",
            502: "  reserved",
            503: "  reserved",
            504: "  reserved",
            505: "  reserved",
            506: "  reserved",
            507: "  reserved",
            508: "  reserved",
            509: "  reserved",
            51: "  reserved",
            510: "  reserved",
            511: "  reserved",
            512: "  reserved",
            513: "  reserved",
            514: "  reserved",
            515: "  reserved",
            516: "  reserved",
            517: "  reserved",
            518: "  reserved",
            519: "  reserved",
            52: "  reserved",
            520: "  reserved",
            521: "  reserved",
            522: "  reserved",
            523: "  reserved",
            524: "  reserved",
            525: "  reserved",
            526: "  reserved",
            527: "  reserved",
            528: "  reserved",
            529: "  reserved",
            53: "  reserved",
            530: "  reserved",
            531: "  reserved",
            532: "  reserved",
            533: "  reserved",
            534: "  reserved",
            535: "  reserved",
            536: "  reserved",
            537: "  reserved",
            538: "  reserved",
            539: "  reserved",
            54: "  reserved",
            540: "  reserved",
            541: "  reserved",
            542: "  reserved",
            543: "  reserved",
            544: "  reserved",
            545: "  reserved",
            546: "  reserved",
            547: "  reserved",
            548: "  reserved",
            549: "  reserved",
            55: "  reserved",
            550: "  reserved",
            551: "  reserved",
            552: "  reserved",
            553: "  reserved",
            554: "  reserved",
            555: "  reserved",
            556: "  reserved",
            557: "  reserved",
            558: "  reserved",
            559: "  reserved",
            56: "  reserved",
            560: "  reserved",
            561: "  reserved",
            562: "  reserved",
            563: "  reserved",
            564: "  reserved",
            565: "  reserved",
            566: "  reserved",
            567: "  reserved",
            568: "  reserved",
            569: "  reserved",
            57: "  reserved",
            570: "  reserved",
            571: "  reserved",
            572: "  reserved",
            573: "  reserved",
            574: "  reserved",
            575: "  reserved",
            576: "  reserved",
            577: "  reserved",
            578: "  reserved",
            579: "  reserved",
            58: "  reserved",
            580: "  reserved",
            581: "  reserved",
            582: "  reserved",
            583: "  reserved",
            584: "  reserved",
            585: "  reserved",
            586: "  reserved",
            587: "  reserved",
            588: "  reserved",
            589: "  reserved",
            59: "  reserved",
            590: "  reserved",
            591: "  reserved",
            592: "  reserved",
            593: "  reserved",
            594: "  reserved",
            595: "  reserved",
            596: "  reserved",
            597: "  reserved",
            598: "  reserved",
            599: "  reserved",
            60: "  reserved",
            600: "  reserved",
            601: "  reserved",
            602: "  reserved",
            603: "  reserved",
            604: "  reserved",
            605: "  reserved",
            606: "  reserved",
            607: "  reserved",
            608: "  reserved",
            609: "  reserved",
            61: "  reserved",
            610: "  reserved",
            611: "  reserved",
            612: "  reserved",
            613: "  reserved",
            614: "  reserved",
            615: "  reserved",
            616: "  reserved",
            617: "  reserved",
            618: "  reserved",
            619: "  reserved",
            62: "  reserved",
            620: "  reserved",
            621: "  reserved",
            622: "  reserved",
            623: "  reserved",
            624: "  reserved",
            625: "  reserved",
            626: "  reserved",
            627: "  reserved",
            628: "  reserved",
            629: "  reserved",
            63: "  reserved",
            630: "  reserved",
            631: "  reserved",
            632: "  reserved",
            633: "  reserved",
            634: "  reserved",
            635: "  reserved",
            636: "  reserved",
            637: "  reserved",
            638: "  reserved",
            639: "  reserved",
            64: "  reserved",
            640: "  reserved",
            641: "  reserved",
            642: "  reserved",
            643: "  reserved",
            644: "  reserved",
            645: "  reserved",
            646: "  reserved",
            647: "  reserved",
            648: "  reserved",
            649: "  reserved",
            65: "  reserved",
            650: "  reserved",
            651: "  reserved",
            652: "  reserved",
            653: "  reserved",
            654: "  reserved",
            655: "  reserved",
            656: "  reserved",
            657: "  reserved",
            658: "  reserved",
            659: "  reserved",
            66: "  reserved",
            660: "  reserved",
            661: "  reserved",
            662: "  reserved",
            663: "  reserved",
            664: "  reserved",
            665: "  reserved",
            666: "  reserved",
            667: "  reserved",
            668: "  reserved",
            669: "  reserved",
            67: "  reserved",
            670: "  reserved",
            671: "  reserved",
            672: "  reserved",
            673: "  reserved",
            674: "  reserved",
            675: "  reserved",
            676: "  reserved",
            677: "  reserved",
            678: "  reserved",
            679: "  reserved",
            68: "  reserved",
            680: "  reserved",
            681: "  reserved",
            682: "  reserved",
            683: "  reserved",
            684: "  reserved",
            685: "  reserved",
            686: "  reserved",
            687: "  reserved",
            688: "  reserved",
            689: "  reserved",
            69: "  reserved",
            690: "  reserved",
            691: "  reserved",
            692: "  reserved",
            693: "  reserved",
            694: "  reserved",
            695: "  reserved",
            696: "  reserved",
            697: "  reserved",
            698: "  reserved",
            699: "  reserved",
            70: "  reserved",
            700: "  reserved",
            701: "  reserved",
            702: "  reserved",
            703: "  reserved",
            704: "  reserved",
            705: "  reserved",
            706: "  reserved",
            707: "  reserved",
            708: "  reserved",
            709: "  reserved",
            71: "  reserved",
            710: "  reserved",
            711: "  reserved",
            712: "  reserved",
            713: "  reserved",
            714: "  reserved",
            715: "  reserved",
            716: "  reserved",
            717: "  reserved",
            718: "  reserved",
            719: "  reserved",
            72: "  reserved",
            720: "  reserved",
            721: "  reserved",
            722: "  reserved",
            723: "  reserved",
            724: "  reserved",
            725: "  reserved",
            726: "  reserved",
            727: "  reserved",
            728: "  reserved",
            729: "  reserved",
            73: "  reserved",
            730: "  reserved",
            731: "  reserved",
            732: "  reserved",
            733: "  reserved",
            734: "  reserved",
            735: "  reserved",
            736: "  reserved",
            737: "  reserved",
            738: "  reserved",
            739: "  reserved",
            74: "  reserved",
            740: "  reserved",
            741: "  reserved",
            742: "  reserved",
            743: "  reserved",
            744: "  reserved",
            745: "  reserved",
            746: "  reserved",
            747: "  reserved",
            748: "  reserved",
            749: "  reserved",
            75: "  reserved",
            750: "  reserved",
            751: "  reserved",
            752: "  reserved",
            753: "  reserved",
            754: "  reserved",
            755: "  reserved",
            756: "  reserved",
            757: "  reserved",
            758: "  reserved",
            759: "  reserved",
            76: "  reserved",
            760: "  reserved",
            761: "  reserved",
            762: "  reserved",
            763: "  reserved",
            764: "  reserved",
            765: "  reserved",
            766: "  reserved",
            767: "  reserved",
            768: "  reserved",
            769: "  reserved",
            77: "  reserved",
            770: "  reserved",
            771: "  reserved",
            772: "  reserved",
            773: "  reserved",
            774: "  reserved",
            775: "  reserved",
            776: "  reserved",
            777: "  reserved",
            778: "  reserved",
            779: "  reserved",
            78: "  reserved",
            780: "  reserved",
            781: "  reserved",
            782: "  reserved",
            783: "  reserved",
            784: "  reserved",
            785: "  reserved",
            786: "  reserved",
            787: "  reserved",
            788: "  reserved",
            789: "  reserved",
            79: "  reserved",
            790: "  reserved",
            791: "  reserved",
            792: "  reserved",
            793: "  reserved",
            794: "  reserved",
            795: "  reserved",
            796: "  reserved",
            797: "  reserved",
            798: "  reserved",
            799: "  reserved",
            80: "  reserved",
            800: "  reserved",
            801: "  reserved",
            802: "  reserved",
            803: "  reserved",
            804: "  reserved",
            805: "  reserved",
            806: "  reserved",
            807: "  reserved",
            808: "  reserved",
            809: "  reserved",
            81: "  reserved",
            810: "  reserved",
            811: "  reserved",
            812: "  reserved",
            813: "  reserved",
            814: "  reserved",
            815: "  reserved",
            816: "  reserved",
            817: "  reserved",
            818: "  reserved",
            819: "  reserved",
            82: "  reserved",
            820: "  reserved",
            821: "  reserved",
            822: "  reserved",
            823: "  reserved",
            824: "  reserved",
            825: "  reserved",
            826: "  reserved",
            827: "  reserved",
            828: "  reserved",
            829: "  reserved",
            83: "  reserved",
            830: "  reserved",
            831: "  reserved",
            832: "  reserved",
            833: "  reserved",
            834: "  reserved",
            835: "  reserved",
            836: "  reserved",
            837: "  reserved",
            838: "  reserved",
            839: "  reserved",
            84: "  reserved",
            840: "  reserved",
            841: "  reserved",
            842: "  reserved",
            843: "  reserved",
            844: "  reserved",
            845: "  reserved",
            846: "  reserved",
            847: "  reserved",
            848: "  reserved",
            849: "  reserved",
            85: "  reserved",
            850: "  reserved",
            851: "  reserved",
            852: "  reserved",
            853: "  reserved",
            854: "  reserved",
            855: "  reserved",
            856: "  reserved",
            857: "  reserved",
            858: "  reserved",
            859: "  reserved",
            86: "  reserved",
            860: "  reserved",
            861: "  reserved",
            862: "  reserved",
            863: "  reserved",
            864: "  reserved",
            865: "  reserved",
            866: "  reserved",
            867: "  reserved",
            868: "  reserved",
            869: "  reserved",
            87: "  reserved",
            870: "  reserved",
            871: "  reserved",
            872: "  reserved",
            873: "  reserved",
            874: "  reserved",
            875: "  reserved",
            876: "  reserved",
            877: "  reserved",
            878: "  reserved",
            879: "  reserved",
            88: "  reserved",
            880: "  reserved",
            881: "  reserved",
            882: "  reserved",
            883: "  reserved",
            884: "  reserved",
            885: "  reserved",
            886: "  reserved",
            887: "  reserved",
            888: "  reserved",
            889: "  reserved",
            89: "  reserved",
            890: "  reserved",
            891: "  reserved",
            892: "  reserved",
            893: "  reserved",
            894: "  reserved",
            895: "  reserved",
            896: "  reserved",
            897: "  reserved",
            898: "  reserved",
            899: "  reserved",
            90: "  reserved",
            900: "  reserved",
            901: "  reserved",
            902: "  reserved",
            903: "  reserved",
            904: "  reserved",
            905: "  reserved",
            906: "  reserved",
            907: "  reserved",
            908: "  reserved",
            909: "  reserved",
            91: "  reserved",
            910: "  reserved",
            911: "  reserved",
            912: "  reserved",
            913: "  reserved",
            914: "  reserved",
            915: "  reserved",
            916: "  reserved",
            917: "  reserved",
            918: "  reserved",
            919: "  reserved",
            92: "  reserved",
            920: "  reserved",
            921: "  reserved",
            922: "  reserved",
            923: "  reserved",
            924: "  reserved",
            925: "  reserved",
            926: "  reserved",
            927: "  reserved",
            928: "  reserved",
            929: "  reserved",
            93: "  reserved",
            930: "  reserved",
            931: "  reserved",
            932: "  reserved",
            933: "  reserved",
            934: "  reserved",
            935: "  reserved",
            936: "  reserved",
            937: "  reserved",
            938: "  reserved",
            939: "  reserved",
            94: "  reserved",
            940: "  reserved",
            941: "  reserved",
            942: "  reserved",
            943: "  reserved",
            944: "  reserved",
            945: "  reserved",
            946: "  reserved",
            947: "  reserved",
            948: "  reserved",
            949: "  reserved",
            95: "  reserved",
            950: "  reserved",
            951: "  reserved",
            952: "  reserved",
            953: "  reserved",
            954: "  reserved",
            955: "  reserved",
            956: "  reserved",
            957: "  reserved",
            958: "  reserved",
            959: "  reserved",
            96: "  reserved",
            960: "  reserved",
            961: "  reserved",
            962: "  reserved",
            963: "  reserved",
            964: "  reserved",
            965: "  reserved",
            966: "  reserved",
            967: "  reserved",
            968: "  reserved",
            969: "  reserved",
            97: "  reserved",
            970: "  reserved",
            971: "  reserved",
            972: "  reserved",
            973: "  reserved",
            974: "  reserved",
            975: "  reserved",
            976: "  reserved",
            977: "  reserved",
            978: "  reserved",
            979: "  reserved",
            98: "  reserved",
            980: "  reserved",
            981: "  reserved",
            982: "  reserved",
            983: "  reserved",
            984: "  reserved",
            985: "  reserved",
            986: "  reserved",
            987: "  reserved",
            988: "  reserved",
            989: "  reserved",
            99: "  reserved",
            990: "  reserved",
            991: "  reserved",
            992: "  reserved",
            993: "  reserved",
            994: "  reserved",
            995: "  reserved",
            996: "  reserved",
            997: "  reserved",
            998: "  reserved",
            999: "  reserved"
        },
        3337: {
            0: " no operator control request",
            1: ".  the operator control state can be read in two ways.  one method uses the direct operator control input to determine the operator control state, while the second method relies on a secondary control to read the operator control input, then relay the information on the data link.  the parameter is designed to provide the actual operator control state to other controls that need the information.",
            10: " decrement operator control request",
            100: "  reserved",
            1000: "  reserved",
            1001: "  reserved",
            1002: "  reserved",
            1003: "  reserved",
            1004: "  reserved",
            1005: "  reserved",
            1006: "  reserved",
            1007: "  reserved",
            1008: "  reserved",
            1009: "  reserved",
            101: "  reserved",
            1010: "  reserved",
            1011: "  reserved",
            1012: "  reserved",
            1013: "  reserved",
            1014: "  reserved",
            1015: "  reserved",
            1016: "  reserved",
            1017: "  reserved",
            1018: "  reserved",
            1019: "  reserved",
            102: "  reserved",
            1020: "  reserved",
            1021: "  reserved",
            1022: "  reserved",
            1023: "  reserved",
            1024: "  reserved",
            1025: "  reserved",
            1026: "  reserved",
            1027: "  reserved",
            1028: "  reserved",
            1029: "  reserved",
            103: "  reserved",
            1030: "  reserved",
            1031: "  reserved",
            1032: "  reserved",
            1033: "  reserved",
            1034: "  reserved",
            1035: "  reserved",
            1036: "  reserved",
            1037: "  reserved",
            1038: "  reserved",
            1039: "  reserved",
            104: "  reserved",
            1040: "  reserved",
            1041: "  reserved",
            1042: "  reserved",
            1043: "  reserved",
            1044: "  reserved",
            1045: "  reserved",
            1046: "  reserved",
            1047: "  reserved",
            1048: "  reserved",
            1049: "  reserved",
            105: "  reserved",
            1050: "  reserved",
            1051: "  reserved",
            1052: "  reserved",
            1053: "  reserved",
            1054: "  reserved",
            1055: "  reserved",
            1056: "  reserved",
            1057: "  reserved",
            1058: "  reserved",
            1059: "  reserved",
            106: "  reserved",
            1060: "  reserved",
            1061: "  reserved",
            1062: "  reserved",
            1063: "  reserved",
            1064: "  reserved",
            1065: "  reserved",
            1066: "  reserved",
            1067: "  reserved",
            1068: "  reserved",
            1069: "  reserved",
            107: "  reserved",
            1070: "  reserved",
            1071: "  reserved",
            1072: "  reserved",
            1073: "  reserved",
            1074: "  reserved",
            1075: "  reserved",
            1076: "  reserved",
            1077: "  reserved",
            1078: "  reserved",
            1079: "  reserved",
            108: "  reserved",
            1080: "  reserved",
            1081: "  reserved",
            1082: "  reserved",
            1083: "  reserved",
            1084: "  reserved",
            1085: "  reserved",
            1086: "  reserved",
            1087: "  reserved",
            1088: "  reserved",
            1089: "  reserved",
            109: "  reserved",
            1090: "  reserved",
            1091: "  reserved",
            1092: "  reserved",
            1093: "  reserved",
            1094: "  reserved",
            1095: "  reserved",
            1096: "  reserved",
            1097: "  reserved",
            1098: "  reserved",
            1099: "  reserved",
            11: "  reserved",
            110: "  reserved",
            1100: "  reserved",
            1101: "  reserved",
            111: "  reserved",
            1110: " error indicator",
            1111: " not installed",
            112: "  reserved",
            113: "  reserved",
            114: "  reserved",
            115: "  reserved",
            116: "  reserved",
            117: "  reserved",
            118: "  reserved",
            119: "  reserved",
            12: "  reserved",
            120: "  reserved",
            121: "  reserved",
            122: "  reserved",
            123: "  reserved",
            124: "  reserved",
            125: "  reserved",
            126: "  reserved",
            127: "  reserved",
            128: "  reserved",
            129: "  reserved",
            13: "  reserved",
            130: "  reserved",
            131: "  reserved",
            132: "  reserved",
            133: "  reserved",
            134: "  reserved",
            135: "  reserved",
            136: "  reserved",
            137: "  reserved",
            138: "  reserved",
            139: "  reserved",
            14: "  reserved",
            140: "  reserved",
            141: "  reserved",
            142: "  reserved",
            143: "  reserved",
            144: "  reserved",
            145: "  reserved",
            146: "  reserved",
            147: "  reserved",
            148: "  reserved",
            149: "  reserved",
            15: "  reserved",
            150: "  reserved",
            151: "  reserved",
            152: "  reserved",
            153: "  reserved",
            154: "  reserved",
            155: "  reserved",
            156: "  reserved",
            157: "  reserved",
            158: "  reserved",
            159: "  reserved",
            16: "  reserved",
            160: "  reserved",
            161: "  reserved",
            162: "  reserved",
            163: "  reserved",
            164: "  reserved",
            165: "  reserved",
            166: "  reserved",
            167: "  reserved",
            168: "  reserved",
            169: "  reserved",
            17: "  reserved",
            170: "  reserved",
            171: "  reserved",
            172: "  reserved",
            173: "  reserved",
            174: "  reserved",
            175: "  reserved",
            176: "  reserved",
            177: "  reserved",
            178: "  reserved",
            179: "  reserved",
            18: "  reserved",
            180: "  reserved",
            181: "  reserved",
            182: "  reserved",
            183: "  reserved",
            184: "  reserved",
            185: "  reserved",
            186: "  reserved",
            187: "  reserved",
            188: "  reserved",
            189: "  reserved",
            19: "  reserved",
            190: "  reserved",
            191: "  reserved",
            192: "  reserved",
            193: "  reserved",
            194: "  reserved",
            195: "  reserved",
            196: "  reserved",
            197: "  reserved",
            198: "  reserved",
            199: "  reserved",
            2: ".  other systems with automated blade controls should be able to use this parameter, since it is an actual operator control value.",
            20: "  reserved",
            200: "  reserved",
            201: "  reserved",
            202: "  reserved",
            203: "  reserved",
            204: "  reserved",
            205: "  reserved",
            206: "  reserved",
            207: "  reserved",
            208: "  reserved",
            209: "  reserved",
            21: "  reserved",
            210: "  reserved",
            211: "  reserved",
            212: "  reserved",
            213: "  reserved",
            214: "  reserved",
            215: "  reserved",
            216: "  reserved",
            217: "  reserved",
            218: "  reserved",
            219: "  reserved",
            22: "  reserved",
            220: "  reserved",
            221: "  reserved",
            222: "  reserved",
            223: "  reserved",
            224: "  reserved",
            225: "  reserved",
            226: "  reserved",
            227: "  reserved",
            228: "  reserved",
            229: "  reserved",
            23: "  reserved",
            230: "  reserved",
            231: "  reserved",
            232: "  reserved",
            233: "  reserved",
            234: "  reserved",
            235: "  reserved",
            236: "  reserved",
            237: "  reserved",
            238: "  reserved",
            239: "  reserved",
            24: "  reserved",
            240: "  reserved",
            241: "  reserved",
            242: "  reserved",
            243: "  reserved",
            244: "  reserved",
            245: "  reserved",
            246: "  reserved",
            247: "  reserved",
            248: "  reserved",
            249: "  reserved",
            25: "  reserved",
            250: "  reserved",
            251: "  reserved",
            252: "  reserved",
            253: "  reserved",
            254: "  reserved",
            255: "  reserved",
            256: "  reserved",
            257: "  reserved",
            258: "  reserved",
            259: "  reserved",
            26: "  reserved",
            260: "  reserved",
            261: "  reserved",
            262: "  reserved",
            263: "  reserved",
            264: "  reserved",
            265: "  reserved",
            266: "  reserved",
            267: "  reserved",
            268: "  reserved",
            269: "  reserved",
            27: "  reserved",
            270: "  reserved",
            271: "  reserved",
            272: "  reserved",
            273: "  reserved",
            274: "  reserved",
            275: "  reserved",
            276: "  reserved",
            277: "  reserved",
            278: "  reserved",
            279: "  reserved",
            28: "  reserved",
            280: "  reserved",
            281: "  reserved",
            282: "  reserved",
            283: "  reserved",
            284: "  reserved",
            285: "  reserved",
            286: "  reserved",
            287: "  reserved",
            288: "  reserved",
            289: "  reserved",
            29: "  reserved",
            290: "  reserved",
            291: "  reserved",
            292: "  reserved",
            293: "  reserved",
            294: "  reserved",
            295: "  reserved",
            296: "  reserved",
            297: "  reserved",
            298: "  reserved",
            299: "  reserved",
            30: "  reserved",
            300: "  reserved",
            301: "  reserved",
            302: "  reserved",
            303: "  reserved",
            304: "  reserved",
            305: "  reserved",
            306: "  reserved",
            307: "  reserved",
            308: "  reserved",
            309: "  reserved",
            31: "  reserved",
            310: "  reserved",
            311: "  reserved",
            312: "  reserved",
            313: "  reserved",
            314: "  reserved",
            315: "  reserved",
            316: "  reserved",
            317: "  reserved",
            318: "  reserved",
            319: "  reserved",
            32: "  reserved",
            320: "  reserved",
            321: "  reserved",
            322: "  reserved",
            323: "  reserved",
            324: "  reserved",
            325: "  reserved",
            326: "  reserved",
            327: "  reserved",
            328: "  reserved",
            329: "  reserved",
            33: "  reserved",
            330: "  reserved",
            331: "  reserved",
            332: "  reserved",
            333: "  reserved",
            334: "  reserved",
            335: "  reserved",
            336: "  reserved",
            337: "  reserved",
            338: "  reserved",
            339: "  reserved",
            34: "  reserved",
            340: "  reserved",
            341: "  reserved",
            342: "  reserved",
            343: "  reserved",
            344: "  reserved",
            345: "  reserved",
            346: "  reserved",
            347: "  reserved",
            348: "  reserved",
            349: "  reserved",
            35: "  reserved",
            350: "  reserved",
            351: "  reserved",
            352: "  reserved",
            353: "  reserved",
            354: "  reserved",
            355: "  reserved",
            356: "  reserved",
            357: "  reserved",
            358: "  reserved",
            359: "  reserved",
            36: "  reserved",
            360: "  reserved",
            361: "  reserved",
            362: "  reserved",
            363: "  reserved",
            364: "  reserved",
            365: "  reserved",
            366: "  reserved",
            367: "  reserved",
            368: "  reserved",
            369: "  reserved",
            37: "  reserved",
            370: "  reserved",
            371: "  reserved",
            372: "  reserved",
            373: "  reserved",
            374: "  reserved",
            375: "  reserved",
            376: "  reserved",
            377: "  reserved",
            378: "  reserved",
            379: "  reserved",
            38: "  reserved",
            380: "  reserved",
            381: "  reserved",
            382: "  reserved",
            383: "  reserved",
            384: "  reserved",
            385: "  reserved",
            386: "  reserved",
            387: "  reserved",
            388: "  reserved",
            389: "  reserved",
            39: "  reserved",
            390: "  reserved",
            391: "  reserved",
            392: "  reserved",
            393: "  reserved",
            394: "  reserved",
            395: "  reserved",
            396: "  reserved",
            397: "  reserved",
            398: "  reserved",
            399: "  reserved",
            40: "  reserved",
            400: "  reserved",
            401: "  reserved",
            402: "  reserved",
            403: "  reserved",
            404: "  reserved",
            405: "  reserved",
            406: "  reserved",
            407: "  reserved",
            408: "  reserved",
            409: "  reserved",
            41: "  reserved",
            410: "  reserved",
            411: "  reserved",
            412: "  reserved",
            413: "  reserved",
            414: "  reserved",
            415: "  reserved",
            416: "  reserved",
            417: "  reserved",
            418: "  reserved",
            419: "  reserved",
            42: "  reserved",
            420: "  reserved",
            421: "  reserved",
            422: "  reserved",
            423: "  reserved",
            424: "  reserved",
            425: "  reserved",
            426: "  reserved",
            427: "  reserved",
            428: "  reserved",
            429: "  reserved",
            43: "  reserved",
            430: "  reserved",
            431: "  reserved",
            432: "  reserved",
            433: "  reserved",
            434: "  reserved",
            435: "  reserved",
            436: "  reserved",
            437: "  reserved",
            438: "  reserved",
            439: "  reserved",
            44: "  reserved",
            440: "  reserved",
            441: "  reserved",
            442: "  reserved",
            443: "  reserved",
            444: "  reserved",
            445: "  reserved",
            446: "  reserved",
            447: "  reserved",
            448: "  reserved",
            449: "  reserved",
            45: "  reserved",
            450: "  reserved",
            451: "  reserved",
            452: "  reserved",
            453: "  reserved",
            454: "  reserved",
            455: "  reserved",
            456: "  reserved",
            457: "  reserved",
            458: "  reserved",
            459: "  reserved",
            46: "  reserved",
            460: "  reserved",
            461: "  reserved",
            462: "  reserved",
            463: "  reserved",
            464: "  reserved",
            465: "  reserved",
            466: "  reserved",
            467: "  reserved",
            468: "  reserved",
            469: "  reserved",
            47: "  reserved",
            470: "  reserved",
            471: "  reserved",
            472: "  reserved",
            473: "  reserved",
            474: "  reserved",
            475: "  reserved",
            476: "  reserved",
            477: "  reserved",
            478: "  reserved",
            479: "  reserved",
            48: "  reserved",
            480: "  reserved",
            481: "  reserved",
            482: "  reserved",
            483: "  reserved",
            484: "  reserved",
            485: "  reserved",
            486: "  reserved",
            487: "  reserved",
            488: "  reserved",
            489: "  reserved",
            49: "  reserved",
            490: "  reserved",
            491: "  reserved",
            492: "  reserved",
            493: "  reserved",
            494: "  reserved",
            495: "  reserved",
            496: "  reserved",
            497: "  reserved",
            498: "  reserved",
            499: "  reserved",
            50: "  reserved",
            500: "  reserved",
            501: "  reserved",
            502: "  reserved",
            503: "  reserved",
            504: "  reserved",
            505: "  reserved",
            506: "  reserved",
            507: "  reserved",
            508: "  reserved",
            509: "  reserved",
            51: "  reserved",
            510: "  reserved",
            511: "  reserved",
            512: "  reserved",
            513: "  reserved",
            514: "  reserved",
            515: "  reserved",
            516: "  reserved",
            517: "  reserved",
            518: "  reserved",
            519: "  reserved",
            52: "  reserved",
            520: "  reserved",
            521: "  reserved",
            522: "  reserved",
            523: "  reserved",
            524: "  reserved",
            525: "  reserved",
            526: "  reserved",
            527: "  reserved",
            528: "  reserved",
            529: "  reserved",
            53: "  reserved",
            530: "  reserved",
            531: "  reserved",
            532: "  reserved",
            533: "  reserved",
            534: "  reserved",
            535: "  reserved",
            536: "  reserved",
            537: "  reserved",
            538: "  reserved",
            539: "  reserved",
            54: "  reserved",
            540: "  reserved",
            541: "  reserved",
            542: "  reserved",
            543: "  reserved",
            544: "  reserved",
            545: "  reserved",
            546: "  reserved",
            547: "  reserved",
            548: "  reserved",
            549: "  reserved",
            55: "  reserved",
            550: "  reserved",
            551: "  reserved",
            552: "  reserved",
            553: "  reserved",
            554: "  reserved",
            555: "  reserved",
            556: "  reserved",
            557: "  reserved",
            558: "  reserved",
            559: "  reserved",
            56: "  reserved",
            560: "  reserved",
            561: "  reserved",
            562: "  reserved",
            563: "  reserved",
            564: "  reserved",
            565: "  reserved",
            566: "  reserved",
            567: "  reserved",
            568: "  reserved",
            569: "  reserved",
            57: "  reserved",
            570: "  reserved",
            571: "  reserved",
            572: "  reserved",
            573: "  reserved",
            574: "  reserved",
            575: "  reserved",
            576: "  reserved",
            577: "  reserved",
            578: "  reserved",
            579: "  reserved",
            58: "  reserved",
            580: "  reserved",
            581: "  reserved",
            582: "  reserved",
            583: "  reserved",
            584: "  reserved",
            585: "  reserved",
            586: "  reserved",
            587: "  reserved",
            588: "  reserved",
            589: "  reserved",
            59: "  reserved",
            590: "  reserved",
            591: "  reserved",
            592: "  reserved",
            593: "  reserved",
            594: "  reserved",
            595: "  reserved",
            596: "  reserved",
            597: "  reserved",
            598: "  reserved",
            599: "  reserved",
            60: "  reserved",
            600: "  reserved",
            601: "  reserved",
            602: "  reserved",
            603: "  reserved",
            604: "  reserved",
            605: "  reserved",
            606: "  reserved",
            607: "  reserved",
            608: "  reserved",
            609: "  reserved",
            61: "  reserved",
            610: "  reserved",
            611: "  reserved",
            612: "  reserved",
            613: "  reserved",
            614: "  reserved",
            615: "  reserved",
            616: "  reserved",
            617: "  reserved",
            618: "  reserved",
            619: "  reserved",
            62: "  reserved",
            620: "  reserved",
            621: "  reserved",
            622: "  reserved",
            623: "  reserved",
            624: "  reserved",
            625: "  reserved",
            626: "  reserved",
            627: "  reserved",
            628: "  reserved",
            629: "  reserved",
            63: "  reserved",
            630: "  reserved",
            631: "  reserved",
            632: "  reserved",
            633: "  reserved",
            634: "  reserved",
            635: "  reserved",
            636: "  reserved",
            637: "  reserved",
            638: "  reserved",
            639: "  reserved",
            64: "  reserved",
            640: "  reserved",
            641: "  reserved",
            642: "  reserved",
            643: "  reserved",
            644: "  reserved",
            645: "  reserved",
            646: "  reserved",
            647: "  reserved",
            648: "  reserved",
            649: "  reserved",
            65: "  reserved",
            650: "  reserved",
            651: "  reserved",
            652: "  reserved",
            653: "  reserved",
            654: "  reserved",
            655: "  reserved",
            656: "  reserved",
            657: "  reserved",
            658: "  reserved",
            659: "  reserved",
            66: "  reserved",
            660: "  reserved",
            661: "  reserved",
            662: "  reserved",
            663: "  reserved",
            664: "  reserved",
            665: "  reserved",
            666: "  reserved",
            667: "  reserved",
            668: "  reserved",
            669: "  reserved",
            67: "  reserved",
            670: "  reserved",
            671: "  reserved",
            672: "  reserved",
            673: "  reserved",
            674: "  reserved",
            675: "  reserved",
            676: "  reserved",
            677: "  reserved",
            678: "  reserved",
            679: "  reserved",
            68: "  reserved",
            680: "  reserved",
            681: "  reserved",
            682: "  reserved",
            683: "  reserved",
            684: "  reserved",
            685: "  reserved",
            686: "  reserved",
            687: "  reserved",
            688: "  reserved",
            689: "  reserved",
            69: "  reserved",
            690: "  reserved",
            691: "  reserved",
            692: "  reserved",
            693: "  reserved",
            694: "  reserved",
            695: "  reserved",
            696: "  reserved",
            697: "  reserved",
            698: "  reserved",
            699: "  reserved",
            70: "  reserved",
            700: "  reserved",
            701: "  reserved",
            702: "  reserved",
            703: "  reserved",
            704: "  reserved",
            705: "  reserved",
            706: "  reserved",
            707: "  reserved",
            708: "  reserved",
            709: "  reserved",
            71: "  reserved",
            710: "  reserved",
            711: "  reserved",
            712: "  reserved",
            713: "  reserved",
            714: "  reserved",
            715: "  reserved",
            716: "  reserved",
            717: "  reserved",
            718: "  reserved",
            719: "  reserved",
            72: "  reserved",
            720: "  reserved",
            721: "  reserved",
            722: "  reserved",
            723: "  reserved",
            724: "  reserved",
            725: "  reserved",
            726: "  reserved",
            727: "  reserved",
            728: "  reserved",
            729: "  reserved",
            73: "  reserved",
            730: "  reserved",
            731: "  reserved",
            732: "  reserved",
            733: "  reserved",
            734: "  reserved",
            735: "  reserved",
            736: "  reserved",
            737: "  reserved",
            738: "  reserved",
            739: "  reserved",
            74: "  reserved",
            740: "  reserved",
            741: "  reserved",
            742: "  reserved",
            743: "  reserved",
            744: "  reserved",
            745: "  reserved",
            746: "  reserved",
            747: "  reserved",
            748: "  reserved",
            749: "  reserved",
            75: "  reserved",
            750: "  reserved",
            751: "  reserved",
            752: "  reserved",
            753: "  reserved",
            754: "  reserved",
            755: "  reserved",
            756: "  reserved",
            757: "  reserved",
            758: "  reserved",
            759: "  reserved",
            76: "  reserved",
            760: "  reserved",
            761: "  reserved",
            762: "  reserved",
            763: "  reserved",
            764: "  reserved",
            765: "  reserved",
            766: "  reserved",
            767: "  reserved",
            768: "  reserved",
            769: "  reserved",
            77: "  reserved",
            770: "  reserved",
            771: "  reserved",
            772: "  reserved",
            773: "  reserved",
            774: "  reserved",
            775: "  reserved",
            776: "  reserved",
            777: "  reserved",
            778: "  reserved",
            779: "  reserved",
            78: "  reserved",
            780: "  reserved",
            781: "  reserved",
            782: "  reserved",
            783: "  reserved",
            784: "  reserved",
            785: "  reserved",
            786: "  reserved",
            787: "  reserved",
            788: "  reserved",
            789: "  reserved",
            79: "  reserved",
            790: "  reserved",
            791: "  reserved",
            792: "  reserved",
            793: "  reserved",
            794: "  reserved",
            795: "  reserved",
            796: "  reserved",
            797: "  reserved",
            798: "  reserved",
            799: "  reserved",
            80: "  reserved",
            800: "  reserved",
            801: "  reserved",
            802: "  reserved",
            803: "  reserved",
            804: "  reserved",
            805: "  reserved",
            806: "  reserved",
            807: "  reserved",
            808: "  reserved",
            809: "  reserved",
            81: "  reserved",
            810: "  reserved",
            811: "  reserved",
            812: "  reserved",
            813: "  reserved",
            814: "  reserved",
            815: "  reserved",
            816: "  reserved",
            817: "  reserved",
            818: "  reserved",
            819: "  reserved",
            82: "  reserved",
            820: "  reserved",
            821: "  reserved",
            822: "  reserved",
            823: "  reserved",
            824: "  reserved",
            825: "  reserved",
            826: "  reserved",
            827: "  reserved",
            828: "  reserved",
            829: "  reserved",
            83: "  reserved",
            830: "  reserved",
            831: "  reserved",
            832: "  reserved",
            833: "  reserved",
            834: "  reserved",
            835: "  reserved",
            836: "  reserved",
            837: "  reserved",
            838: "  reserved",
            839: "  reserved",
            84: "  reserved",
            840: "  reserved",
            841: "  reserved",
            842: "  reserved",
            843: "  reserved",
            844: "  reserved",
            845: "  reserved",
            846: "  reserved",
            847: "  reserved",
            848: "  reserved",
            849: "  reserved",
            85: "  reserved",
            850: "  reserved",
            851: "  reserved",
            852: "  reserved",
            853: "  reserved",
            854: "  reserved",
            855: "  reserved",
            856: "  reserved",
            857: "  reserved",
            858: "  reserved",
            859: "  reserved",
            86: "  reserved",
            860: "  reserved",
            861: "  reserved",
            862: "  reserved",
            863: "  reserved",
            864: "  reserved",
            865: "  reserved",
            866: "  reserved",
            867: "  reserved",
            868: "  reserved",
            869: "  reserved",
            87: "  reserved",
            870: "  reserved",
            871: "  reserved",
            872: "  reserved",
            873: "  reserved",
            874: "  reserved",
            875: "  reserved",
            876: "  reserved",
            877: "  reserved",
            878: "  reserved",
            879: "  reserved",
            88: "  reserved",
            880: "  reserved",
            881: "  reserved",
            882: "  reserved",
            883: "  reserved",
            884: "  reserved",
            885: "  reserved",
            886: "  reserved",
            887: "  reserved",
            888: "  reserved",
            889: "  reserved",
            89: "  reserved",
            890: "  reserved",
            891: "  reserved",
            892: "  reserved",
            893: "  reserved",
            894: "  reserved",
            895: "  reserved",
            896: "  reserved",
            897: "  reserved",
            898: "  reserved",
            899: "  reserved",
            90: "  reserved",
            900: "  reserved",
            901: "  reserved",
            902: "  reserved",
            903: "  reserved",
            904: "  reserved",
            905: "  reserved",
            906: "  reserved",
            907: "  reserved",
            908: "  reserved",
            909: "  reserved",
            91: "  reserved",
            910: "  reserved",
            911: "  reserved",
            912: "  reserved",
            913: "  reserved",
            914: "  reserved",
            915: "  reserved",
            916: "  reserved",
            917: "  reserved",
            918: "  reserved",
            919: "  reserved",
            92: "  reserved",
            920: "  reserved",
            921: "  reserved",
            922: "  reserved",
            923: "  reserved",
            924: "  reserved",
            925: "  reserved",
            926: "  reserved",
            927: "  reserved",
            928: "  reserved",
            929: "  reserved",
            93: "  reserved",
            930: "  reserved",
            931: "  reserved",
            932: "  reserved",
            933: "  reserved",
            934: "  reserved",
            935: "  reserved",
            936: "  reserved",
            937: "  reserved",
            938: "  reserved",
            939: "  reserved",
            94: "  reserved",
            940: "  reserved",
            941: "  reserved",
            942: "  reserved",
            943: "  reserved",
            944: "  reserved",
            945: "  reserved",
            946: "  reserved",
            947: "  reserved",
            948: "  reserved",
            949: "  reserved",
            95: "  reserved",
            950: "  reserved",
            951: "  reserved",
            952: "  reserved",
            953: "  reserved",
            954: "  reserved",
            955: "  reserved",
            956: "  reserved",
            957: "  reserved",
            958: "  reserved",
            959: "  reserved",
            96: "  reserved",
            960: "  reserved",
            961: "  reserved",
            962: "  reserved",
            963: "  reserved",
            964: "  reserved",
            965: "  reserved",
            966: "  reserved",
            967: "  reserved",
            968: "  reserved",
            969: "  reserved",
            97: "  reserved",
            970: "  reserved",
            971: "  reserved",
            972: "  reserved",
            973: "  reserved",
            974: "  reserved",
            975: "  reserved",
            976: "  reserved",
            977: "  reserved",
            978: "  reserved",
            979: "  reserved",
            98: "  reserved",
            980: "  reserved",
            981: "  reserved",
            982: "  reserved",
            983: "  reserved",
            984: "  reserved",
            985: "  reserved",
            986: "  reserved",
            987: "  reserved",
            988: "  reserved",
            989: "  reserved",
            99: "  reserved",
            990: "  reserved",
            991: "  reserved",
            992: "  reserved",
            993: "  reserved",
            994: "  reserved",
            995: "  reserved",
            996: "  reserved",
            997: "  reserved",
            998: "  reserved",
            999: "  reserved"
        },
        3338: {
            0: " no operator control request",
            1: " increment operator control request",
            10: " decrement operator control request",
            100: " reserved",
            1000: " reserved",
            1001: " reserved",
            1002: " reserved",
            1003: " reserved",
            1004: " reserved",
            1005: " reserved",
            1006: " reserved",
            1007: " reserved",
            1008: " reserved",
            1009: " reserved",
            101: " reserved",
            1010: " reserved",
            1011: " reserved",
            1012: " reserved",
            1013: " reserved",
            1014: " reserved",
            1015: " reserved",
            1016: " reserved",
            1017: " reserved",
            1018: " reserved",
            1019: " reserved",
            102: " reserved",
            1020: " reserved",
            1021: " reserved",
            1022: " reserved",
            1023: " reserved",
            1024: " reserved",
            1025: " reserved",
            1026: " reserved",
            1027: " reserved",
            1028: " reserved",
            1029: " reserved",
            103: " reserved",
            1030: " reserved",
            1031: " reserved",
            1032: " reserved",
            1033: " reserved",
            1034: " reserved",
            1035: " reserved",
            1036: " reserved",
            1037: " reserved",
            1038: " reserved",
            1039: " reserved",
            104: " reserved",
            1040: " reserved",
            1041: " reserved",
            1042: " reserved",
            1043: " reserved",
            1044: " reserved",
            1045: " reserved",
            1046: " reserved",
            1047: " reserved",
            1048: " reserved",
            1049: " reserved",
            105: " reserved",
            1050: " reserved",
            1051: " reserved",
            1052: " reserved",
            1053: " reserved",
            1054: " reserved",
            1055: " reserved",
            1056: " reserved",
            1057: " reserved",
            1058: " reserved",
            1059: " reserved",
            106: " reserved",
            1060: " reserved",
            1061: " reserved",
            1062: " reserved",
            1063: " reserved",
            1064: " reserved",
            1065: " reserved",
            1066: " reserved",
            1067: " reserved",
            1068: " reserved",
            1069: " reserved",
            107: " reserved",
            1070: " reserved",
            1071: " reserved",
            1072: " reserved",
            1073: " reserved",
            1074: " reserved",
            1075: " reserved",
            1076: " reserved",
            1077: " reserved",
            1078: " reserved",
            1079: " reserved",
            108: " reserved",
            1080: " reserved",
            1081: " reserved",
            1082: " reserved",
            1083: " reserved",
            1084: " reserved",
            1085: " reserved",
            1086: " reserved",
            1087: " reserved",
            1088: " reserved",
            1089: " reserved",
            109: " reserved",
            1090: " reserved",
            1091: " reserved",
            1092: " reserved",
            1093: " reserved",
            1094: " reserved",
            1095: " reserved",
            1096: " reserved",
            1097: " reserved",
            1098: " reserved",
            1099: " reserved",
            11: " reserved",
            110: " reserved",
            1100: " reserved",
            1101: " reserved",
            111: " reserved",
            1110: " error indicator",
            1111: " not installed",
            112: " reserved",
            113: " reserved",
            114: " reserved",
            115: " reserved",
            116: " reserved",
            117: " reserved",
            118: " reserved",
            119: " reserved",
            12: " reserved",
            120: " reserved",
            121: " reserved",
            122: " reserved",
            123: " reserved",
            124: " reserved",
            125: " reserved",
            126: " reserved",
            127: " reserved",
            128: " reserved",
            129: " reserved",
            13: " reserved",
            130: " reserved",
            131: " reserved",
            132: " reserved",
            133: " reserved",
            134: " reserved",
            135: " reserved",
            136: " reserved",
            137: " reserved",
            138: " reserved",
            139: " reserved",
            14: " reserved",
            140: " reserved",
            141: " reserved",
            142: " reserved",
            143: " reserved",
            144: " reserved",
            145: " reserved",
            146: " reserved",
            147: " reserved",
            148: " reserved",
            149: " reserved",
            15: " reserved",
            150: " reserved",
            151: " reserved",
            152: " reserved",
            153: " reserved",
            154: " reserved",
            155: " reserved",
            156: " reserved",
            157: " reserved",
            158: " reserved",
            159: " reserved",
            16: " reserved",
            160: " reserved",
            161: " reserved",
            162: " reserved",
            163: " reserved",
            164: " reserved",
            165: " reserved",
            166: " reserved",
            167: " reserved",
            168: " reserved",
            169: " reserved",
            17: " reserved",
            170: " reserved",
            171: " reserved",
            172: " reserved",
            173: " reserved",
            174: " reserved",
            175: " reserved",
            176: " reserved",
            177: " reserved",
            178: " reserved",
            179: " reserved",
            18: " reserved",
            180: " reserved",
            181: " reserved",
            182: " reserved",
            183: " reserved",
            184: " reserved",
            185: " reserved",
            186: " reserved",
            187: " reserved",
            188: " reserved",
            189: " reserved",
            19: " reserved",
            190: " reserved",
            191: " reserved",
            192: " reserved",
            193: " reserved",
            194: " reserved",
            195: " reserved",
            196: " reserved",
            197: " reserved",
            198: " reserved",
            199: " reserved",
            2: ". the momentary contact button pair may be used, where the default is the neutral position with no buttons pressed. the user can press the increment button, and when pressure is released, it will return to the neutral position. the user may press the decrement button, and when pressure is released, it will return to the neutral position. momentary contact button (button pair) usage is as follows:",
            20: " reserved",
            200: " reserved",
            201: " reserved",
            202: " reserved",
            203: " reserved",
            204: " reserved",
            205: " reserved",
            206: " reserved",
            207: " reserved",
            208: " reserved",
            209: " reserved",
            21: " reserved",
            210: " reserved",
            211: " reserved",
            212: " reserved",
            213: " reserved",
            214: " reserved",
            215: " reserved",
            216: " reserved",
            217: " reserved",
            218: " reserved",
            219: " reserved",
            22: " reserved",
            220: " reserved",
            221: " reserved",
            222: " reserved",
            223: " reserved",
            224: " reserved",
            225: " reserved",
            226: " reserved",
            227: " reserved",
            228: " reserved",
            229: " reserved",
            23: " reserved",
            230: " reserved",
            231: " reserved",
            232: " reserved",
            233: " reserved",
            234: " reserved",
            235: " reserved",
            236: " reserved",
            237: " reserved",
            238: " reserved",
            239: " reserved",
            24: " reserved",
            240: " reserved",
            241: " reserved",
            242: " reserved",
            243: " reserved",
            244: " reserved",
            245: " reserved",
            246: " reserved",
            247: " reserved",
            248: " reserved",
            249: " reserved",
            25: " reserved",
            250: " reserved",
            251: " reserved",
            252: " reserved",
            253: " reserved",
            254: " reserved",
            255: " reserved",
            256: " reserved",
            257: " reserved",
            258: " reserved",
            259: " reserved",
            26: " reserved",
            260: " reserved",
            261: " reserved",
            262: " reserved",
            263: " reserved",
            264: " reserved",
            265: " reserved",
            266: " reserved",
            267: " reserved",
            268: " reserved",
            269: " reserved",
            27: " reserved",
            270: " reserved",
            271: " reserved",
            272: " reserved",
            273: " reserved",
            274: " reserved",
            275: " reserved",
            276: " reserved",
            277: " reserved",
            278: " reserved",
            279: " reserved",
            28: " reserved",
            280: " reserved",
            281: " reserved",
            282: " reserved",
            283: " reserved",
            284: " reserved",
            285: " reserved",
            286: " reserved",
            287: " reserved",
            288: " reserved",
            289: " reserved",
            29: " reserved",
            290: " reserved",
            291: " reserved",
            292: " reserved",
            293: " reserved",
            294: " reserved",
            295: " reserved",
            296: " reserved",
            297: " reserved",
            298: " reserved",
            299: " reserved",
            30: " reserved",
            300: " reserved",
            301: " reserved",
            302: " reserved",
            303: " reserved",
            304: " reserved",
            305: " reserved",
            306: " reserved",
            307: " reserved",
            308: " reserved",
            309: " reserved",
            31: " reserved",
            310: " reserved",
            311: " reserved",
            312: " reserved",
            313: " reserved",
            314: " reserved",
            315: " reserved",
            316: " reserved",
            317: " reserved",
            318: " reserved",
            319: " reserved",
            32: " reserved",
            320: " reserved",
            321: " reserved",
            322: " reserved",
            323: " reserved",
            324: " reserved",
            325: " reserved",
            326: " reserved",
            327: " reserved",
            328: " reserved",
            329: " reserved",
            33: " reserved",
            330: " reserved",
            331: " reserved",
            332: " reserved",
            333: " reserved",
            334: " reserved",
            335: " reserved",
            336: " reserved",
            337: " reserved",
            338: " reserved",
            339: " reserved",
            34: " reserved",
            340: " reserved",
            341: " reserved",
            342: " reserved",
            343: " reserved",
            344: " reserved",
            345: " reserved",
            346: " reserved",
            347: " reserved",
            348: " reserved",
            349: " reserved",
            35: " reserved",
            350: " reserved",
            351: " reserved",
            352: " reserved",
            353: " reserved",
            354: " reserved",
            355: " reserved",
            356: " reserved",
            357: " reserved",
            358: " reserved",
            359: " reserved",
            36: " reserved",
            360: " reserved",
            361: " reserved",
            362: " reserved",
            363: " reserved",
            364: " reserved",
            365: " reserved",
            366: " reserved",
            367: " reserved",
            368: " reserved",
            369: " reserved",
            37: " reserved",
            370: " reserved",
            371: " reserved",
            372: " reserved",
            373: " reserved",
            374: " reserved",
            375: " reserved",
            376: " reserved",
            377: " reserved",
            378: " reserved",
            379: " reserved",
            38: " reserved",
            380: " reserved",
            381: " reserved",
            382: " reserved",
            383: " reserved",
            384: " reserved",
            385: " reserved",
            386: " reserved",
            387: " reserved",
            388: " reserved",
            389: " reserved",
            39: " reserved",
            390: " reserved",
            391: " reserved",
            392: " reserved",
            393: " reserved",
            394: " reserved",
            395: " reserved",
            396: " reserved",
            397: " reserved",
            398: " reserved",
            399: " reserved",
            40: " reserved",
            400: " reserved",
            401: " reserved",
            402: " reserved",
            403: " reserved",
            404: " reserved",
            405: " reserved",
            406: " reserved",
            407: " reserved",
            408: " reserved",
            409: " reserved",
            41: " reserved",
            410: " reserved",
            411: " reserved",
            412: " reserved",
            413: " reserved",
            414: " reserved",
            415: " reserved",
            416: " reserved",
            417: " reserved",
            418: " reserved",
            419: " reserved",
            42: " reserved",
            420: " reserved",
            421: " reserved",
            422: " reserved",
            423: " reserved",
            424: " reserved",
            425: " reserved",
            426: " reserved",
            427: " reserved",
            428: " reserved",
            429: " reserved",
            43: " reserved",
            430: " reserved",
            431: " reserved",
            432: " reserved",
            433: " reserved",
            434: " reserved",
            435: " reserved",
            436: " reserved",
            437: " reserved",
            438: " reserved",
            439: " reserved",
            44: " reserved",
            440: " reserved",
            441: " reserved",
            442: " reserved",
            443: " reserved",
            444: " reserved",
            445: " reserved",
            446: " reserved",
            447: " reserved",
            448: " reserved",
            449: " reserved",
            45: " reserved",
            450: " reserved",
            451: " reserved",
            452: " reserved",
            453: " reserved",
            454: " reserved",
            455: " reserved",
            456: " reserved",
            457: " reserved",
            458: " reserved",
            459: " reserved",
            46: " reserved",
            460: " reserved",
            461: " reserved",
            462: " reserved",
            463: " reserved",
            464: " reserved",
            465: " reserved",
            466: " reserved",
            467: " reserved",
            468: " reserved",
            469: " reserved",
            47: " reserved",
            470: " reserved",
            471: " reserved",
            472: " reserved",
            473: " reserved",
            474: " reserved",
            475: " reserved",
            476: " reserved",
            477: " reserved",
            478: " reserved",
            479: " reserved",
            48: " reserved",
            480: " reserved",
            481: " reserved",
            482: " reserved",
            483: " reserved",
            484: " reserved",
            485: " reserved",
            486: " reserved",
            487: " reserved",
            488: " reserved",
            489: " reserved",
            49: " reserved",
            490: " reserved",
            491: " reserved",
            492: " reserved",
            493: " reserved",
            494: " reserved",
            495: " reserved",
            496: " reserved",
            497: " reserved",
            498: " reserved",
            499: " reserved",
            50: " reserved",
            500: " reserved",
            501: " reserved",
            502: " reserved",
            503: " reserved",
            504: " reserved",
            505: " reserved",
            506: " reserved",
            507: " reserved",
            508: " reserved",
            509: " reserved",
            51: " reserved",
            510: " reserved",
            511: " reserved",
            512: " reserved",
            513: " reserved",
            514: " reserved",
            515: " reserved",
            516: " reserved",
            517: " reserved",
            518: " reserved",
            519: " reserved",
            52: " reserved",
            520: " reserved",
            521: " reserved",
            522: " reserved",
            523: " reserved",
            524: " reserved",
            525: " reserved",
            526: " reserved",
            527: " reserved",
            528: " reserved",
            529: " reserved",
            53: " reserved",
            530: " reserved",
            531: " reserved",
            532: " reserved",
            533: " reserved",
            534: " reserved",
            535: " reserved",
            536: " reserved",
            537: " reserved",
            538: " reserved",
            539: " reserved",
            54: " reserved",
            540: " reserved",
            541: " reserved",
            542: " reserved",
            543: " reserved",
            544: " reserved",
            545: " reserved",
            546: " reserved",
            547: " reserved",
            548: " reserved",
            549: " reserved",
            55: " reserved",
            550: " reserved",
            551: " reserved",
            552: " reserved",
            553: " reserved",
            554: " reserved",
            555: " reserved",
            556: " reserved",
            557: " reserved",
            558: " reserved",
            559: " reserved",
            56: " reserved",
            560: " reserved",
            561: " reserved",
            562: " reserved",
            563: " reserved",
            564: " reserved",
            565: " reserved",
            566: " reserved",
            567: " reserved",
            568: " reserved",
            569: " reserved",
            57: " reserved",
            570: " reserved",
            571: " reserved",
            572: " reserved",
            573: " reserved",
            574: " reserved",
            575: " reserved",
            576: " reserved",
            577: " reserved",
            578: " reserved",
            579: " reserved",
            58: " reserved",
            580: " reserved",
            581: " reserved",
            582: " reserved",
            583: " reserved",
            584: " reserved",
            585: " reserved",
            586: " reserved",
            587: " reserved",
            588: " reserved",
            589: " reserved",
            59: " reserved",
            590: " reserved",
            591: " reserved",
            592: " reserved",
            593: " reserved",
            594: " reserved",
            595: " reserved",
            596: " reserved",
            597: " reserved",
            598: " reserved",
            599: " reserved",
            60: " reserved",
            600: " reserved",
            601: " reserved",
            602: " reserved",
            603: " reserved",
            604: " reserved",
            605: " reserved",
            606: " reserved",
            607: " reserved",
            608: " reserved",
            609: " reserved",
            61: " reserved",
            610: " reserved",
            611: " reserved",
            612: " reserved",
            613: " reserved",
            614: " reserved",
            615: " reserved",
            616: " reserved",
            617: " reserved",
            618: " reserved",
            619: " reserved",
            62: " reserved",
            620: " reserved",
            621: " reserved",
            622: " reserved",
            623: " reserved",
            624: " reserved",
            625: " reserved",
            626: " reserved",
            627: " reserved",
            628: " reserved",
            629: " reserved",
            63: " reserved",
            630: " reserved",
            631: " reserved",
            632: " reserved",
            633: " reserved",
            634: " reserved",
            635: " reserved",
            636: " reserved",
            637: " reserved",
            638: " reserved",
            639: " reserved",
            64: " reserved",
            640: " reserved",
            641: " reserved",
            642: " reserved",
            643: " reserved",
            644: " reserved",
            645: " reserved",
            646: " reserved",
            647: " reserved",
            648: " reserved",
            649: " reserved",
            65: " reserved",
            650: " reserved",
            651: " reserved",
            652: " reserved",
            653: " reserved",
            654: " reserved",
            655: " reserved",
            656: " reserved",
            657: " reserved",
            658: " reserved",
            659: " reserved",
            66: " reserved",
            660: " reserved",
            661: " reserved",
            662: " reserved",
            663: " reserved",
            664: " reserved",
            665: " reserved",
            666: " reserved",
            667: " reserved",
            668: " reserved",
            669: " reserved",
            67: " reserved",
            670: " reserved",
            671: " reserved",
            672: " reserved",
            673: " reserved",
            674: " reserved",
            675: " reserved",
            676: " reserved",
            677: " reserved",
            678: " reserved",
            679: " reserved",
            68: " reserved",
            680: " reserved",
            681: " reserved",
            682: " reserved",
            683: " reserved",
            684: " reserved",
            685: " reserved",
            686: " reserved",
            687: " reserved",
            688: " reserved",
            689: " reserved",
            69: " reserved",
            690: " reserved",
            691: " reserved",
            692: " reserved",
            693: " reserved",
            694: " reserved",
            695: " reserved",
            696: " reserved",
            697: " reserved",
            698: " reserved",
            699: " reserved",
            70: " reserved",
            700: " reserved",
            701: " reserved",
            702: " reserved",
            703: " reserved",
            704: " reserved",
            705: " reserved",
            706: " reserved",
            707: " reserved",
            708: " reserved",
            709: " reserved",
            71: " reserved",
            710: " reserved",
            711: " reserved",
            712: " reserved",
            713: " reserved",
            714: " reserved",
            715: " reserved",
            716: " reserved",
            717: " reserved",
            718: " reserved",
            719: " reserved",
            72: " reserved",
            720: " reserved",
            721: " reserved",
            722: " reserved",
            723: " reserved",
            724: " reserved",
            725: " reserved",
            726: " reserved",
            727: " reserved",
            728: " reserved",
            729: " reserved",
            73: " reserved",
            730: " reserved",
            731: " reserved",
            732: " reserved",
            733: " reserved",
            734: " reserved",
            735: " reserved",
            736: " reserved",
            737: " reserved",
            738: " reserved",
            739: " reserved",
            74: " reserved",
            740: " reserved",
            741: " reserved",
            742: " reserved",
            743: " reserved",
            744: " reserved",
            745: " reserved",
            746: " reserved",
            747: " reserved",
            748: " reserved",
            749: " reserved",
            75: " reserved",
            750: " reserved",
            751: " reserved",
            752: " reserved",
            753: " reserved",
            754: " reserved",
            755: " reserved",
            756: " reserved",
            757: " reserved",
            758: " reserved",
            759: " reserved",
            76: " reserved",
            760: " reserved",
            761: " reserved",
            762: " reserved",
            763: " reserved",
            764: " reserved",
            765: " reserved",
            766: " reserved",
            767: " reserved",
            768: " reserved",
            769: " reserved",
            77: " reserved",
            770: " reserved",
            771: " reserved",
            772: " reserved",
            773: " reserved",
            774: " reserved",
            775: " reserved",
            776: " reserved",
            777: " reserved",
            778: " reserved",
            779: " reserved",
            78: " reserved",
            780: " reserved",
            781: " reserved",
            782: " reserved",
            783: " reserved",
            784: " reserved",
            785: " reserved",
            786: " reserved",
            787: " reserved",
            788: " reserved",
            789: " reserved",
            79: " reserved",
            790: " reserved",
            791: " reserved",
            792: " reserved",
            793: " reserved",
            794: " reserved",
            795: " reserved",
            796: " reserved",
            797: " reserved",
            798: " reserved",
            799: " reserved",
            80: " reserved",
            800: " reserved",
            801: " reserved",
            802: " reserved",
            803: " reserved",
            804: " reserved",
            805: " reserved",
            806: " reserved",
            807: " reserved",
            808: " reserved",
            809: " reserved",
            81: " reserved",
            810: " reserved",
            811: " reserved",
            812: " reserved",
            813: " reserved",
            814: " reserved",
            815: " reserved",
            816: " reserved",
            817: " reserved",
            818: " reserved",
            819: " reserved",
            82: " reserved",
            820: " reserved",
            821: " reserved",
            822: " reserved",
            823: " reserved",
            824: " reserved",
            825: " reserved",
            826: " reserved",
            827: " reserved",
            828: " reserved",
            829: " reserved",
            83: " reserved",
            830: " reserved",
            831: " reserved",
            832: " reserved",
            833: " reserved",
            834: " reserved",
            835: " reserved",
            836: " reserved",
            837: " reserved",
            838: " reserved",
            839: " reserved",
            84: " reserved",
            840: " reserved",
            841: " reserved",
            842: " reserved",
            843: " reserved",
            844: " reserved",
            845: " reserved",
            846: " reserved",
            847: " reserved",
            848: " reserved",
            849: " reserved",
            85: " reserved",
            850: " reserved",
            851: " reserved",
            852: " reserved",
            853: " reserved",
            854: " reserved",
            855: " reserved",
            856: " reserved",
            857: " reserved",
            858: " reserved",
            859: " reserved",
            86: " reserved",
            860: " reserved",
            861: " reserved",
            862: " reserved",
            863: " reserved",
            864: " reserved",
            865: " reserved",
            866: " reserved",
            867: " reserved",
            868: " reserved",
            869: " reserved",
            87: " reserved",
            870: " reserved",
            871: " reserved",
            872: " reserved",
            873: " reserved",
            874: " reserved",
            875: " reserved",
            876: " reserved",
            877: " reserved",
            878: " reserved",
            879: " reserved",
            88: " reserved",
            880: " reserved",
            881: " reserved",
            882: " reserved",
            883: " reserved",
            884: " reserved",
            885: " reserved",
            886: " reserved",
            887: " reserved",
            888: " reserved",
            889: " reserved",
            89: " reserved",
            890: " reserved",
            891: " reserved",
            892: " reserved",
            893: " reserved",
            894: " reserved",
            895: " reserved",
            896: " reserved",
            897: " reserved",
            898: " reserved",
            899: " reserved",
            90: " reserved",
            900: " reserved",
            901: " reserved",
            902: " reserved",
            903: " reserved",
            904: " reserved",
            905: " reserved",
            906: " reserved",
            907: " reserved",
            908: " reserved",
            909: " reserved",
            91: " reserved",
            910: " reserved",
            911: " reserved",
            912: " reserved",
            913: " reserved",
            914: " reserved",
            915: " reserved",
            916: " reserved",
            917: " reserved",
            918: " reserved",
            919: " reserved",
            92: " reserved",
            920: " reserved",
            921: " reserved",
            922: " reserved",
            923: " reserved",
            924: " reserved",
            925: " reserved",
            926: " reserved",
            927: " reserved",
            928: " reserved",
            929: " reserved",
            93: " reserved",
            930: " reserved",
            931: " reserved",
            932: " reserved",
            933: " reserved",
            934: " reserved",
            935: " reserved",
            936: " reserved",
            937: " reserved",
            938: " reserved",
            939: " reserved",
            94: " reserved",
            940: " reserved",
            941: " reserved",
            942: " reserved",
            943: " reserved",
            944: " reserved",
            945: " reserved",
            946: " reserved",
            947: " reserved",
            948: " reserved",
            949: " reserved",
            95: " reserved",
            950: " reserved",
            951: " reserved",
            952: " reserved",
            953: " reserved",
            954: " reserved",
            955: " reserved",
            956: " reserved",
            957: " reserved",
            958: " reserved",
            959: " reserved",
            96: " reserved",
            960: " reserved",
            961: " reserved",
            962: " reserved",
            963: " reserved",
            964: " reserved",
            965: " reserved",
            966: " reserved",
            967: " reserved",
            968: " reserved",
            969: " reserved",
            97: " reserved",
            970: " reserved",
            971: " reserved",
            972: " reserved",
            973: " reserved",
            974: " reserved",
            975: " reserved",
            976: " reserved",
            977: " reserved",
            978: " reserved",
            979: " reserved",
            98: " reserved",
            980: " reserved",
            981: " reserved",
            982: " reserved",
            983: " reserved",
            984: " reserved",
            985: " reserved",
            986: " reserved",
            987: " reserved",
            988: " reserved",
            989: " reserved",
            99: " reserved",
            990: " reserved",
            991: " reserved",
            992: " reserved",
            993: " reserved",
            994: " reserved",
            995: " reserved",
            996: " reserved",
            997: " reserved",
            998: " reserved",
            999: " reserved"
        },
        3339: {
            0: " no operator control request",
            1: " momentary operator control request",
            10: " manual operator control request",
            100: "   reserved",
            1000: "   reserved",
            1001: "   reserved",
            1002: "   reserved",
            1003: "   reserved",
            1004: "   reserved",
            1005: "   reserved",
            1006: "   reserved",
            1007: "   reserved",
            1008: "   reserved",
            1009: "   reserved",
            101: "   reserved",
            1010: "   reserved",
            1011: "   reserved",
            1012: "   reserved",
            1013: "   reserved",
            1014: "   reserved",
            1015: "   reserved",
            1016: "   reserved",
            1017: "   reserved",
            1018: "   reserved",
            1019: "   reserved",
            102: "   reserved",
            1020: "   reserved",
            1021: "   reserved",
            1022: "   reserved",
            1023: "   reserved",
            1024: "   reserved",
            1025: "   reserved",
            1026: "   reserved",
            1027: "   reserved",
            1028: "   reserved",
            1029: "   reserved",
            103: "   reserved",
            1030: "   reserved",
            1031: "   reserved",
            1032: "   reserved",
            1033: "   reserved",
            1034: "   reserved",
            1035: "   reserved",
            1036: "   reserved",
            1037: "   reserved",
            1038: "   reserved",
            1039: "   reserved",
            104: "   reserved",
            1040: "   reserved",
            1041: "   reserved",
            1042: "   reserved",
            1043: "   reserved",
            1044: "   reserved",
            1045: "   reserved",
            1046: "   reserved",
            1047: "   reserved",
            1048: "   reserved",
            1049: "   reserved",
            105: "   reserved",
            1050: "   reserved",
            1051: "   reserved",
            1052: "   reserved",
            1053: "   reserved",
            1054: "   reserved",
            1055: "   reserved",
            1056: "   reserved",
            1057: "   reserved",
            1058: "   reserved",
            1059: "   reserved",
            106: "   reserved",
            1060: "   reserved",
            1061: "   reserved",
            1062: "   reserved",
            1063: "   reserved",
            1064: "   reserved",
            1065: "   reserved",
            1066: "   reserved",
            1067: "   reserved",
            1068: "   reserved",
            1069: "   reserved",
            107: "   reserved",
            1070: "   reserved",
            1071: "   reserved",
            1072: "   reserved",
            1073: "   reserved",
            1074: "   reserved",
            1075: "   reserved",
            1076: "   reserved",
            1077: "   reserved",
            1078: "   reserved",
            1079: "   reserved",
            108: "   reserved",
            1080: "   reserved",
            1081: "   reserved",
            1082: "   reserved",
            1083: "   reserved",
            1084: "   reserved",
            1085: "   reserved",
            1086: "   reserved",
            1087: "   reserved",
            1088: "   reserved",
            1089: "   reserved",
            109: "   reserved",
            1090: "   reserved",
            1091: "   reserved",
            1092: "   reserved",
            1093: "   reserved",
            1094: "   reserved",
            1095: "   reserved",
            1096: "   reserved",
            1097: "   reserved",
            1098: "   reserved",
            1099: "   reserved",
            11: " automatic operator control request",
            110: "   reserved",
            1100: "   reserved",
            1101: "   reserved",
            111: "   reserved",
            1110: " error indicator",
            1111: " not installed",
            112: "   reserved",
            113: "   reserved",
            114: "   reserved",
            115: "   reserved",
            116: "   reserved",
            117: "   reserved",
            118: "   reserved",
            119: "   reserved",
            120: "   reserved",
            121: "   reserved",
            122: "   reserved",
            123: "   reserved",
            124: "   reserved",
            125: "   reserved",
            126: "   reserved",
            127: "   reserved",
            128: "   reserved",
            129: "   reserved",
            130: "   reserved",
            131: "   reserved",
            132: "   reserved",
            133: "   reserved",
            134: "   reserved",
            135: "   reserved",
            136: "   reserved",
            137: "   reserved",
            138: "   reserved",
            139: "   reserved",
            140: "   reserved",
            141: "   reserved",
            142: "   reserved",
            143: "   reserved",
            144: "   reserved",
            145: "   reserved",
            146: "   reserved",
            147: "   reserved",
            148: "   reserved",
            149: "   reserved",
            150: "   reserved",
            151: "   reserved",
            152: "   reserved",
            153: "   reserved",
            154: "   reserved",
            155: "   reserved",
            156: "   reserved",
            157: "   reserved",
            158: "   reserved",
            159: "   reserved",
            160: "   reserved",
            161: "   reserved",
            162: "   reserved",
            163: "   reserved",
            164: "   reserved",
            165: "   reserved",
            166: "   reserved",
            167: "   reserved",
            168: "   reserved",
            169: "   reserved",
            170: "   reserved",
            171: "   reserved",
            172: "   reserved",
            173: "   reserved",
            174: "   reserved",
            175: "   reserved",
            176: "   reserved",
            177: "   reserved",
            178: "   reserved",
            179: "   reserved",
            180: "   reserved",
            181: "   reserved",
            182: "   reserved",
            183: "   reserved",
            184: "   reserved",
            185: "   reserved",
            186: "   reserved",
            187: "   reserved",
            188: "   reserved",
            189: "   reserved",
            190: "   reserved",
            191: "   reserved",
            192: "   reserved",
            193: "   reserved",
            194: "   reserved",
            195: "   reserved",
            196: "   reserved",
            197: "   reserved",
            198: "   reserved",
            199: "   reserved",
            2: ".  the momentary rocker switch may be used, where the default is the no button pressed position.  the user can press the rocker into the manual position, and when pressure is released, it will return to the no button pressed position.  the user may also press the rocker into the auto position, and when pressure is released, it will return to the no button pressed position.  momentary rocker button usage is as follows:",
            200: "   reserved",
            201: "   reserved",
            202: "   reserved",
            203: "   reserved",
            204: "   reserved",
            205: "   reserved",
            206: "   reserved",
            207: "   reserved",
            208: "   reserved",
            209: "   reserved",
            210: "   reserved",
            211: "   reserved",
            212: "   reserved",
            213: "   reserved",
            214: "   reserved",
            215: "   reserved",
            216: "   reserved",
            217: "   reserved",
            218: "   reserved",
            219: "   reserved",
            220: "   reserved",
            221: "   reserved",
            222: "   reserved",
            223: "   reserved",
            224: "   reserved",
            225: "   reserved",
            226: "   reserved",
            227: "   reserved",
            228: "   reserved",
            229: "   reserved",
            230: "   reserved",
            231: "   reserved",
            232: "   reserved",
            233: "   reserved",
            234: "   reserved",
            235: "   reserved",
            236: "   reserved",
            237: "   reserved",
            238: "   reserved",
            239: "   reserved",
            240: "   reserved",
            241: "   reserved",
            242: "   reserved",
            243: "   reserved",
            244: "   reserved",
            245: "   reserved",
            246: "   reserved",
            247: "   reserved",
            248: "   reserved",
            249: "   reserved",
            250: "   reserved",
            251: "   reserved",
            252: "   reserved",
            253: "   reserved",
            254: "   reserved",
            255: "   reserved",
            256: "   reserved",
            257: "   reserved",
            258: "   reserved",
            259: "   reserved",
            260: "   reserved",
            261: "   reserved",
            262: "   reserved",
            263: "   reserved",
            264: "   reserved",
            265: "   reserved",
            266: "   reserved",
            267: "   reserved",
            268: "   reserved",
            269: "   reserved",
            270: "   reserved",
            271: "   reserved",
            272: "   reserved",
            273: "   reserved",
            274: "   reserved",
            275: "   reserved",
            276: "   reserved",
            277: "   reserved",
            278: "   reserved",
            279: "   reserved",
            280: "   reserved",
            281: "   reserved",
            282: "   reserved",
            283: "   reserved",
            284: "   reserved",
            285: "   reserved",
            286: "   reserved",
            287: "   reserved",
            288: "   reserved",
            289: "   reserved",
            290: "   reserved",
            291: "   reserved",
            292: "   reserved",
            293: "   reserved",
            294: "   reserved",
            295: "   reserved",
            296: "   reserved",
            297: "   reserved",
            298: "   reserved",
            299: "   reserved",
            3: ".  the momentary contact button pair may be used, where the default is no buttons being pressed.  the user can press the manual button, and when pressure is released, it will return to the no button pressed position.  the user may press the auto button, and when pressure is released, it will return to the no button pressed position.  momentary contact buttons (button pair) usage is as follows:",
            300: "   reserved",
            301: "   reserved",
            302: "   reserved",
            303: "   reserved",
            304: "   reserved",
            305: "   reserved",
            306: "   reserved",
            307: "   reserved",
            308: "   reserved",
            309: "   reserved",
            310: "   reserved",
            311: "   reserved",
            312: "   reserved",
            313: "   reserved",
            314: "   reserved",
            315: "   reserved",
            316: "   reserved",
            317: "   reserved",
            318: "   reserved",
            319: "   reserved",
            320: "   reserved",
            321: "   reserved",
            322: "   reserved",
            323: "   reserved",
            324: "   reserved",
            325: "   reserved",
            326: "   reserved",
            327: "   reserved",
            328: "   reserved",
            329: "   reserved",
            330: "   reserved",
            331: "   reserved",
            332: "   reserved",
            333: "   reserved",
            334: "   reserved",
            335: "   reserved",
            336: "   reserved",
            337: "   reserved",
            338: "   reserved",
            339: "   reserved",
            340: "   reserved",
            341: "   reserved",
            342: "   reserved",
            343: "   reserved",
            344: "   reserved",
            345: "   reserved",
            346: "   reserved",
            347: "   reserved",
            348: "   reserved",
            349: "   reserved",
            350: "   reserved",
            351: "   reserved",
            352: "   reserved",
            353: "   reserved",
            354: "   reserved",
            355: "   reserved",
            356: "   reserved",
            357: "   reserved",
            358: "   reserved",
            359: "   reserved",
            360: "   reserved",
            361: "   reserved",
            362: "   reserved",
            363: "   reserved",
            364: "   reserved",
            365: "   reserved",
            366: "   reserved",
            367: "   reserved",
            368: "   reserved",
            369: "   reserved",
            370: "   reserved",
            371: "   reserved",
            372: "   reserved",
            373: "   reserved",
            374: "   reserved",
            375: "   reserved",
            376: "   reserved",
            377: "   reserved",
            378: "   reserved",
            379: "   reserved",
            380: "   reserved",
            381: "   reserved",
            382: "   reserved",
            383: "   reserved",
            384: "   reserved",
            385: "   reserved",
            386: "   reserved",
            387: "   reserved",
            388: "   reserved",
            389: "   reserved",
            390: "   reserved",
            391: "   reserved",
            392: "   reserved",
            393: "   reserved",
            394: "   reserved",
            395: "   reserved",
            396: "   reserved",
            397: "   reserved",
            398: "   reserved",
            399: "   reserved",
            4: ".  the momentary contact single button may be used, where the default is no button being pressed.  the user can press the button, which is the button pressed position, and when pressure is released, it will return to the no button pressed position.  the controller receiving this command will keep track of what mode the blade is in, auto or manual.  if the blade is in manual mode to begin with, one press of the button will change to mode to auto.  if the button is pressed again, the blade will be put into manual mode.  momentary contact button (single button) usage is as follows:",
            400: "   reserved",
            401: "   reserved",
            402: "   reserved",
            403: "   reserved",
            404: "   reserved",
            405: "   reserved",
            406: "   reserved",
            407: "   reserved",
            408: "   reserved",
            409: "   reserved",
            410: "   reserved",
            411: "   reserved",
            412: "   reserved",
            413: "   reserved",
            414: "   reserved",
            415: "   reserved",
            416: "   reserved",
            417: "   reserved",
            418: "   reserved",
            419: "   reserved",
            420: "   reserved",
            421: "   reserved",
            422: "   reserved",
            423: "   reserved",
            424: "   reserved",
            425: "   reserved",
            426: "   reserved",
            427: "   reserved",
            428: "   reserved",
            429: "   reserved",
            430: "   reserved",
            431: "   reserved",
            432: "   reserved",
            433: "   reserved",
            434: "   reserved",
            435: "   reserved",
            436: "   reserved",
            437: "   reserved",
            438: "   reserved",
            439: "   reserved",
            440: "   reserved",
            441: "   reserved",
            442: "   reserved",
            443: "   reserved",
            444: "   reserved",
            445: "   reserved",
            446: "   reserved",
            447: "   reserved",
            448: "   reserved",
            449: "   reserved",
            450: "   reserved",
            451: "   reserved",
            452: "   reserved",
            453: "   reserved",
            454: "   reserved",
            455: "   reserved",
            456: "   reserved",
            457: "   reserved",
            458: "   reserved",
            459: "   reserved",
            460: "   reserved",
            461: "   reserved",
            462: "   reserved",
            463: "   reserved",
            464: "   reserved",
            465: "   reserved",
            466: "   reserved",
            467: "   reserved",
            468: "   reserved",
            469: "   reserved",
            470: "   reserved",
            471: "   reserved",
            472: "   reserved",
            473: "   reserved",
            474: "   reserved",
            475: "   reserved",
            476: "   reserved",
            477: "   reserved",
            478: "   reserved",
            479: "   reserved",
            480: "   reserved",
            481: "   reserved",
            482: "   reserved",
            483: "   reserved",
            484: "   reserved",
            485: "   reserved",
            486: "   reserved",
            487: "   reserved",
            488: "   reserved",
            489: "   reserved",
            490: "   reserved",
            491: "   reserved",
            492: "   reserved",
            493: "   reserved",
            494: "   reserved",
            495: "   reserved",
            496: "   reserved",
            497: "   reserved",
            498: "   reserved",
            499: "   reserved",
            500: "   reserved",
            501: "   reserved",
            502: "   reserved",
            503: "   reserved",
            504: "   reserved",
            505: "   reserved",
            506: "   reserved",
            507: "   reserved",
            508: "   reserved",
            509: "   reserved",
            510: "   reserved",
            511: "   reserved",
            512: "   reserved",
            513: "   reserved",
            514: "   reserved",
            515: "   reserved",
            516: "   reserved",
            517: "   reserved",
            518: "   reserved",
            519: "   reserved",
            520: "   reserved",
            521: "   reserved",
            522: "   reserved",
            523: "   reserved",
            524: "   reserved",
            525: "   reserved",
            526: "   reserved",
            527: "   reserved",
            528: "   reserved",
            529: "   reserved",
            530: "   reserved",
            531: "   reserved",
            532: "   reserved",
            533: "   reserved",
            534: "   reserved",
            535: "   reserved",
            536: "   reserved",
            537: "   reserved",
            538: "   reserved",
            539: "   reserved",
            540: "   reserved",
            541: "   reserved",
            542: "   reserved",
            543: "   reserved",
            544: "   reserved",
            545: "   reserved",
            546: "   reserved",
            547: "   reserved",
            548: "   reserved",
            549: "   reserved",
            550: "   reserved",
            551: "   reserved",
            552: "   reserved",
            553: "   reserved",
            554: "   reserved",
            555: "   reserved",
            556: "   reserved",
            557: "   reserved",
            558: "   reserved",
            559: "   reserved",
            560: "   reserved",
            561: "   reserved",
            562: "   reserved",
            563: "   reserved",
            564: "   reserved",
            565: "   reserved",
            566: "   reserved",
            567: "   reserved",
            568: "   reserved",
            569: "   reserved",
            570: "   reserved",
            571: "   reserved",
            572: "   reserved",
            573: "   reserved",
            574: "   reserved",
            575: "   reserved",
            576: "   reserved",
            577: "   reserved",
            578: "   reserved",
            579: "   reserved",
            580: "   reserved",
            581: "   reserved",
            582: "   reserved",
            583: "   reserved",
            584: "   reserved",
            585: "   reserved",
            586: "   reserved",
            587: "   reserved",
            588: "   reserved",
            589: "   reserved",
            590: "   reserved",
            591: "   reserved",
            592: "   reserved",
            593: "   reserved",
            594: "   reserved",
            595: "   reserved",
            596: "   reserved",
            597: "   reserved",
            598: "   reserved",
            599: "   reserved",
            600: "   reserved",
            601: "   reserved",
            602: "   reserved",
            603: "   reserved",
            604: "   reserved",
            605: "   reserved",
            606: "   reserved",
            607: "   reserved",
            608: "   reserved",
            609: "   reserved",
            610: "   reserved",
            611: "   reserved",
            612: "   reserved",
            613: "   reserved",
            614: "   reserved",
            615: "   reserved",
            616: "   reserved",
            617: "   reserved",
            618: "   reserved",
            619: "   reserved",
            620: "   reserved",
            621: "   reserved",
            622: "   reserved",
            623: "   reserved",
            624: "   reserved",
            625: "   reserved",
            626: "   reserved",
            627: "   reserved",
            628: "   reserved",
            629: "   reserved",
            630: "   reserved",
            631: "   reserved",
            632: "   reserved",
            633: "   reserved",
            634: "   reserved",
            635: "   reserved",
            636: "   reserved",
            637: "   reserved",
            638: "   reserved",
            639: "   reserved",
            640: "   reserved",
            641: "   reserved",
            642: "   reserved",
            643: "   reserved",
            644: "   reserved",
            645: "   reserved",
            646: "   reserved",
            647: "   reserved",
            648: "   reserved",
            649: "   reserved",
            650: "   reserved",
            651: "   reserved",
            652: "   reserved",
            653: "   reserved",
            654: "   reserved",
            655: "   reserved",
            656: "   reserved",
            657: "   reserved",
            658: "   reserved",
            659: "   reserved",
            660: "   reserved",
            661: "   reserved",
            662: "   reserved",
            663: "   reserved",
            664: "   reserved",
            665: "   reserved",
            666: "   reserved",
            667: "   reserved",
            668: "   reserved",
            669: "   reserved",
            670: "   reserved",
            671: "   reserved",
            672: "   reserved",
            673: "   reserved",
            674: "   reserved",
            675: "   reserved",
            676: "   reserved",
            677: "   reserved",
            678: "   reserved",
            679: "   reserved",
            680: "   reserved",
            681: "   reserved",
            682: "   reserved",
            683: "   reserved",
            684: "   reserved",
            685: "   reserved",
            686: "   reserved",
            687: "   reserved",
            688: "   reserved",
            689: "   reserved",
            690: "   reserved",
            691: "   reserved",
            692: "   reserved",
            693: "   reserved",
            694: "   reserved",
            695: "   reserved",
            696: "   reserved",
            697: "   reserved",
            698: "   reserved",
            699: "   reserved",
            700: "   reserved",
            701: "   reserved",
            702: "   reserved",
            703: "   reserved",
            704: "   reserved",
            705: "   reserved",
            706: "   reserved",
            707: "   reserved",
            708: "   reserved",
            709: "   reserved",
            710: "   reserved",
            711: "   reserved",
            712: "   reserved",
            713: "   reserved",
            714: "   reserved",
            715: "   reserved",
            716: "   reserved",
            717: "   reserved",
            718: "   reserved",
            719: "   reserved",
            720: "   reserved",
            721: "   reserved",
            722: "   reserved",
            723: "   reserved",
            724: "   reserved",
            725: "   reserved",
            726: "   reserved",
            727: "   reserved",
            728: "   reserved",
            729: "   reserved",
            730: "   reserved",
            731: "   reserved",
            732: "   reserved",
            733: "   reserved",
            734: "   reserved",
            735: "   reserved",
            736: "   reserved",
            737: "   reserved",
            738: "   reserved",
            739: "   reserved",
            740: "   reserved",
            741: "   reserved",
            742: "   reserved",
            743: "   reserved",
            744: "   reserved",
            745: "   reserved",
            746: "   reserved",
            747: "   reserved",
            748: "   reserved",
            749: "   reserved",
            750: "   reserved",
            751: "   reserved",
            752: "   reserved",
            753: "   reserved",
            754: "   reserved",
            755: "   reserved",
            756: "   reserved",
            757: "   reserved",
            758: "   reserved",
            759: "   reserved",
            760: "   reserved",
            761: "   reserved",
            762: "   reserved",
            763: "   reserved",
            764: "   reserved",
            765: "   reserved",
            766: "   reserved",
            767: "   reserved",
            768: "   reserved",
            769: "   reserved",
            770: "   reserved",
            771: "   reserved",
            772: "   reserved",
            773: "   reserved",
            774: "   reserved",
            775: "   reserved",
            776: "   reserved",
            777: "   reserved",
            778: "   reserved",
            779: "   reserved",
            780: "   reserved",
            781: "   reserved",
            782: "   reserved",
            783: "   reserved",
            784: "   reserved",
            785: "   reserved",
            786: "   reserved",
            787: "   reserved",
            788: "   reserved",
            789: "   reserved",
            790: "   reserved",
            791: "   reserved",
            792: "   reserved",
            793: "   reserved",
            794: "   reserved",
            795: "   reserved",
            796: "   reserved",
            797: "   reserved",
            798: "   reserved",
            799: "   reserved",
            800: "   reserved",
            801: "   reserved",
            802: "   reserved",
            803: "   reserved",
            804: "   reserved",
            805: "   reserved",
            806: "   reserved",
            807: "   reserved",
            808: "   reserved",
            809: "   reserved",
            810: "   reserved",
            811: "   reserved",
            812: "   reserved",
            813: "   reserved",
            814: "   reserved",
            815: "   reserved",
            816: "   reserved",
            817: "   reserved",
            818: "   reserved",
            819: "   reserved",
            820: "   reserved",
            821: "   reserved",
            822: "   reserved",
            823: "   reserved",
            824: "   reserved",
            825: "   reserved",
            826: "   reserved",
            827: "   reserved",
            828: "   reserved",
            829: "   reserved",
            830: "   reserved",
            831: "   reserved",
            832: "   reserved",
            833: "   reserved",
            834: "   reserved",
            835: "   reserved",
            836: "   reserved",
            837: "   reserved",
            838: "   reserved",
            839: "   reserved",
            840: "   reserved",
            841: "   reserved",
            842: "   reserved",
            843: "   reserved",
            844: "   reserved",
            845: "   reserved",
            846: "   reserved",
            847: "   reserved",
            848: "   reserved",
            849: "   reserved",
            850: "   reserved",
            851: "   reserved",
            852: "   reserved",
            853: "   reserved",
            854: "   reserved",
            855: "   reserved",
            856: "   reserved",
            857: "   reserved",
            858: "   reserved",
            859: "   reserved",
            860: "   reserved",
            861: "   reserved",
            862: "   reserved",
            863: "   reserved",
            864: "   reserved",
            865: "   reserved",
            866: "   reserved",
            867: "   reserved",
            868: "   reserved",
            869: "   reserved",
            870: "   reserved",
            871: "   reserved",
            872: "   reserved",
            873: "   reserved",
            874: "   reserved",
            875: "   reserved",
            876: "   reserved",
            877: "   reserved",
            878: "   reserved",
            879: "   reserved",
            880: "   reserved",
            881: "   reserved",
            882: "   reserved",
            883: "   reserved",
            884: "   reserved",
            885: "   reserved",
            886: "   reserved",
            887: "   reserved",
            888: "   reserved",
            889: "   reserved",
            890: "   reserved",
            891: "   reserved",
            892: "   reserved",
            893: "   reserved",
            894: "   reserved",
            895: "   reserved",
            896: "   reserved",
            897: "   reserved",
            898: "   reserved",
            899: "   reserved",
            900: "   reserved",
            901: "   reserved",
            902: "   reserved",
            903: "   reserved",
            904: "   reserved",
            905: "   reserved",
            906: "   reserved",
            907: "   reserved",
            908: "   reserved",
            909: "   reserved",
            910: "   reserved",
            911: "   reserved",
            912: "   reserved",
            913: "   reserved",
            914: "   reserved",
            915: "   reserved",
            916: "   reserved",
            917: "   reserved",
            918: "   reserved",
            919: "   reserved",
            920: "   reserved",
            921: "   reserved",
            922: "   reserved",
            923: "   reserved",
            924: "   reserved",
            925: "   reserved",
            926: "   reserved",
            927: "   reserved",
            928: "   reserved",
            929: "   reserved",
            930: "   reserved",
            931: "   reserved",
            932: "   reserved",
            933: "   reserved",
            934: "   reserved",
            935: "   reserved",
            936: "   reserved",
            937: "   reserved",
            938: "   reserved",
            939: "   reserved",
            940: "   reserved",
            941: "   reserved",
            942: "   reserved",
            943: "   reserved",
            944: "   reserved",
            945: "   reserved",
            946: "   reserved",
            947: "   reserved",
            948: "   reserved",
            949: "   reserved",
            950: "   reserved",
            951: "   reserved",
            952: "   reserved",
            953: "   reserved",
            954: "   reserved",
            955: "   reserved",
            956: "   reserved",
            957: "   reserved",
            958: "   reserved",
            959: "   reserved",
            960: "   reserved",
            961: "   reserved",
            962: "   reserved",
            963: "   reserved",
            964: "   reserved",
            965: "   reserved",
            966: "   reserved",
            967: "   reserved",
            968: "   reserved",
            969: "   reserved",
            970: "   reserved",
            971: "   reserved",
            972: "   reserved",
            973: "   reserved",
            974: "   reserved",
            975: "   reserved",
            976: "   reserved",
            977: "   reserved",
            978: "   reserved",
            979: "   reserved",
            980: "   reserved",
            981: "   reserved",
            982: "   reserved",
            983: "   reserved",
            984: "   reserved",
            985: "   reserved",
            986: "   reserved",
            987: "   reserved",
            988: "   reserved",
            989: "   reserved",
            990: "   reserved",
            991: "   reserved",
            992: "   reserved",
            993: "   reserved",
            994: "   reserved",
            995: "   reserved",
            996: "   reserved",
            997: "   reserved",
            998: "   reserved",
            999: "   reserved"
        },
        3344: {
            1: "                      1000 ms transmission rate    ",
            2: "                        750 ms transmission rate",
            3: "                        500 ms transmission rate",
            4: "                        250 ms transmission rate",
            5: "                        100 ms transmission rate",
            6: "                          50 ms transmission rate",
            7: "                          20 ms transmission rate",
            8: "                         reserved for sae assignment (set to one)"
        },
        3345: {
            1: "                             p1                          accelerator pedal/operator selection",
            2: "                             p2                          cruise control",
            3: "                             p3                          pto governor",
            4: "                             p4                          road speed governor",
            5: "                             p5                          engine protection",
            6: "                             p6                          reserved for assignment by sae",
            7: "                             p7                          reserved for assignment by sae",
            8: "                             p8                          reserved for assignment by sae"
        },
        3346: {
            1: "                             p9                          reserved for assignment by sae",
            2: "                             p10                          reserved for assignment by sae",
            3: "                             p11                        reserved for assignment by sae",
            4: "                             p12                        reserved for assignment by sae",
            5: "                             p13                        reserved for assignment by sae",
            6: "                             p14                        reserved for assignment by sae",
            7: "                             p15                        reserved for assignment by sae",
            8: "                             p16                        reserved for assignment by sae"
        },
        3347: {
            1: "                             p17                        reserved for assignment by sae",
            2: "                             p18                        reserved for assignment by sae",
            3: "                             p19                        reserved for assignment by sae",
            4: "                             p20                        reserved for assignment by sae",
            5: "                             p21                        reserved for assignment by sae",
            6: "                             p22                        reserved for assignment by sae",
            7: "                             p23                        reserved for assignment by sae",
            8: "                             p24                        reserved for assignment by sae"
        },
        3348: {
            1: "                             p25                        reserved for assignment by sae",
            2: "                             p26                        reserved for assignment by sae",
            3: "                             p27                        reserved for assignment by sae",
            4: "                             p28                        reserved for assignment by sae",
            5: "                             p29                        reserved for assignment by sae",
            6: "                             p30                        reserved for assignment by sae",
            7: "                             p31                        reserved for assignment by sae",
            8: "                             see note               not assignable, must always be set to 1.       "
        },
        3349: {
            0: " ms transmission rate",
            1: " ms transmission rate",
            10: " ms transmission rate",
            100: " ms transmission rate",
            1000: " ms transmission rate",
            101: " ms transmission rate",
            102: " ms transmission rate",
            103: " ms transmission rate",
            104: " ms transmission rate",
            105: " ms transmission rate",
            106: " ms transmission rate",
            107: " ms transmission rate",
            108: " ms transmission rate",
            109: " ms transmission rate",
            11: " ms transmission rate",
            110: " ms transmission rate",
            111: " ms transmission rate",
            112: " ms transmission rate",
            113: " ms transmission rate",
            114: " ms transmission rate",
            115: " ms transmission rate",
            116: " ms transmission rate",
            117: " ms transmission rate",
            118: " ms transmission rate",
            119: " ms transmission rate",
            12: " ms transmission rate",
            120: " ms transmission rate",
            121: " ms transmission rate",
            122: " ms transmission rate",
            123: " ms transmission rate",
            124: " ms transmission rate",
            125: " ms transmission rate",
            126: " ms transmission rate",
            127: " ms transmission rate",
            128: " ms transmission rate",
            129: " ms transmission rate",
            13: " ms transmission rate",
            130: " ms transmission rate",
            131: " ms transmission rate",
            132: " ms transmission rate",
            133: " ms transmission rate",
            134: " ms transmission rate",
            135: " ms transmission rate",
            136: " ms transmission rate",
            137: " ms transmission rate",
            138: " ms transmission rate",
            139: " ms transmission rate",
            14: " ms transmission rate",
            140: " ms transmission rate",
            141: " ms transmission rate",
            142: " ms transmission rate",
            143: " ms transmission rate",
            144: " ms transmission rate",
            145: " ms transmission rate",
            146: " ms transmission rate",
            147: " ms transmission rate",
            148: " ms transmission rate",
            149: " ms transmission rate",
            15: " ms transmission rate",
            150: " ms transmission rate",
            151: " ms transmission rate",
            152: " ms transmission rate",
            153: " ms transmission rate",
            154: " ms transmission rate",
            155: " ms transmission rate",
            156: " ms transmission rate",
            157: " ms transmission rate",
            158: " ms transmission rate",
            159: " ms transmission rate",
            16: " ms transmission rate",
            160: " ms transmission rate",
            161: " ms transmission rate",
            162: " ms transmission rate",
            163: " ms transmission rate",
            164: " ms transmission rate",
            165: " ms transmission rate",
            166: " ms transmission rate",
            167: " ms transmission rate",
            168: " ms transmission rate",
            169: " ms transmission rate",
            17: " ms transmission rate",
            170: " ms transmission rate",
            171: " ms transmission rate",
            172: " ms transmission rate",
            173: " ms transmission rate",
            174: " ms transmission rate",
            175: " ms transmission rate",
            176: " ms transmission rate",
            177: " ms transmission rate",
            178: " ms transmission rate",
            179: " ms transmission rate",
            18: " ms transmission rate",
            180: " ms transmission rate",
            181: " ms transmission rate",
            182: " ms transmission rate",
            183: " ms transmission rate",
            184: " ms transmission rate",
            185: " ms transmission rate",
            186: " ms transmission rate",
            187: " ms transmission rate",
            188: " ms transmission rate",
            189: " ms transmission rate",
            19: " ms transmission rate",
            190: " ms transmission rate",
            191: " ms transmission rate",
            192: " ms transmission rate",
            193: " ms transmission rate",
            194: " ms transmission rate",
            195: " ms transmission rate",
            196: " ms transmission rate",
            197: " ms transmission rate",
            198: " ms transmission rate",
            199: " ms transmission rate",
            2: " ms transmission rate",
            20: " ms transmission rate",
            200: " ms transmission rate",
            201: " ms transmission rate",
            202: " ms transmission rate",
            203: " ms transmission rate",
            204: " ms transmission rate",
            205: " ms transmission rate",
            206: " ms transmission rate",
            207: " ms transmission rate",
            208: " ms transmission rate",
            209: " ms transmission rate",
            21: " ms transmission rate",
            210: " ms transmission rate",
            211: " ms transmission rate",
            212: " ms transmission rate",
            213: " ms transmission rate",
            214: " ms transmission rate",
            215: " ms transmission rate",
            216: " ms transmission rate",
            217: " ms transmission rate",
            218: " ms transmission rate",
            219: " ms transmission rate",
            22: " ms transmission rate",
            220: " ms transmission rate",
            221: " ms transmission rate",
            222: " ms transmission rate",
            223: " ms transmission rate",
            224: " ms transmission rate",
            225: " ms transmission rate",
            226: " ms transmission rate",
            227: " ms transmission rate",
            228: " ms transmission rate",
            229: " ms transmission rate",
            23: " ms transmission rate",
            230: " ms transmission rate",
            231: " ms transmission rate",
            232: " ms transmission rate",
            233: " ms transmission rate",
            234: " ms transmission rate",
            235: " ms transmission rate",
            236: " ms transmission rate",
            237: " ms transmission rate",
            238: " ms transmission rate",
            239: " ms transmission rate",
            24: " ms transmission rate",
            240: " ms transmission rate",
            241: " ms transmission rate",
            242: " ms transmission rate",
            243: " ms transmission rate",
            244: " ms transmission rate",
            245: " ms transmission rate",
            246: " ms transmission rate",
            247: " ms transmission rate",
            248: " ms transmission rate",
            249: " ms transmission rate",
            25: " ms transmission rate",
            250: " ms transmission rate",
            251: " ms transmission rate",
            252: " ms transmission rate",
            253: " ms transmission rate",
            254: " ms transmission rate",
            255: " ms transmission rate",
            256: " ms transmission rate",
            257: " ms transmission rate",
            258: " ms transmission rate",
            259: " ms transmission rate",
            26: " ms transmission rate",
            260: " ms transmission rate",
            261: " ms transmission rate",
            262: " ms transmission rate",
            263: " ms transmission rate",
            264: " ms transmission rate",
            265: " ms transmission rate",
            266: " ms transmission rate",
            267: " ms transmission rate",
            268: " ms transmission rate",
            269: " ms transmission rate",
            27: " ms transmission rate",
            270: " ms transmission rate",
            271: " ms transmission rate",
            272: " ms transmission rate",
            273: " ms transmission rate",
            274: " ms transmission rate",
            275: " ms transmission rate",
            276: " ms transmission rate",
            277: " ms transmission rate",
            278: " ms transmission rate",
            279: " ms transmission rate",
            28: " ms transmission rate",
            280: " ms transmission rate",
            281: " ms transmission rate",
            282: " ms transmission rate",
            283: " ms transmission rate",
            284: " ms transmission rate",
            285: " ms transmission rate",
            286: " ms transmission rate",
            287: " ms transmission rate",
            288: " ms transmission rate",
            289: " ms transmission rate",
            29: " ms transmission rate",
            290: " ms transmission rate",
            291: " ms transmission rate",
            292: " ms transmission rate",
            293: " ms transmission rate",
            294: " ms transmission rate",
            295: " ms transmission rate",
            296: " ms transmission rate",
            297: " ms transmission rate",
            298: " ms transmission rate",
            299: " ms transmission rate",
            3: " ms transmission rate",
            30: " ms transmission rate",
            300: " ms transmission rate",
            301: " ms transmission rate",
            302: " ms transmission rate",
            303: " ms transmission rate",
            304: " ms transmission rate",
            305: " ms transmission rate",
            306: " ms transmission rate",
            307: " ms transmission rate",
            308: " ms transmission rate",
            309: " ms transmission rate",
            31: " ms transmission rate",
            310: " ms transmission rate",
            311: " ms transmission rate",
            312: " ms transmission rate",
            313: " ms transmission rate",
            314: " ms transmission rate",
            315: " ms transmission rate",
            316: " ms transmission rate",
            317: " ms transmission rate",
            318: " ms transmission rate",
            319: " ms transmission rate",
            32: " ms transmission rate",
            320: " ms transmission rate",
            321: " ms transmission rate",
            322: " ms transmission rate",
            323: " ms transmission rate",
            324: " ms transmission rate",
            325: " ms transmission rate",
            326: " ms transmission rate",
            327: " ms transmission rate",
            328: " ms transmission rate",
            329: " ms transmission rate",
            33: " ms transmission rate",
            330: " ms transmission rate",
            331: " ms transmission rate",
            332: " ms transmission rate",
            333: " ms transmission rate",
            334: " ms transmission rate",
            335: " ms transmission rate",
            336: " ms transmission rate",
            337: " ms transmission rate",
            338: " ms transmission rate",
            339: " ms transmission rate",
            34: " ms transmission rate",
            340: " ms transmission rate",
            341: " ms transmission rate",
            342: " ms transmission rate",
            343: " ms transmission rate",
            344: " ms transmission rate",
            345: " ms transmission rate",
            346: " ms transmission rate",
            347: " ms transmission rate",
            348: " ms transmission rate",
            349: " ms transmission rate",
            35: " ms transmission rate",
            350: " ms transmission rate",
            351: " ms transmission rate",
            352: " ms transmission rate",
            353: " ms transmission rate",
            354: " ms transmission rate",
            355: " ms transmission rate",
            356: " ms transmission rate",
            357: " ms transmission rate",
            358: " ms transmission rate",
            359: " ms transmission rate",
            36: " ms transmission rate",
            360: " ms transmission rate",
            361: " ms transmission rate",
            362: " ms transmission rate",
            363: " ms transmission rate",
            364: " ms transmission rate",
            365: " ms transmission rate",
            366: " ms transmission rate",
            367: " ms transmission rate",
            368: " ms transmission rate",
            369: " ms transmission rate",
            37: " ms transmission rate",
            370: " ms transmission rate",
            371: " ms transmission rate",
            372: " ms transmission rate",
            373: " ms transmission rate",
            374: " ms transmission rate",
            375: " ms transmission rate",
            376: " ms transmission rate",
            377: " ms transmission rate",
            378: " ms transmission rate",
            379: " ms transmission rate",
            38: " ms transmission rate",
            380: " ms transmission rate",
            381: " ms transmission rate",
            382: " ms transmission rate",
            383: " ms transmission rate",
            384: " ms transmission rate",
            385: " ms transmission rate",
            386: " ms transmission rate",
            387: " ms transmission rate",
            388: " ms transmission rate",
            389: " ms transmission rate",
            39: " ms transmission rate",
            390: " ms transmission rate",
            391: " ms transmission rate",
            392: " ms transmission rate",
            393: " ms transmission rate",
            394: " ms transmission rate",
            395: " ms transmission rate",
            396: " ms transmission rate",
            397: " ms transmission rate",
            398: " ms transmission rate",
            399: " ms transmission rate",
            4: " ms transmission rate",
            40: " ms transmission rate",
            400: " ms transmission rate",
            401: " ms transmission rate",
            402: " ms transmission rate",
            403: " ms transmission rate",
            404: " ms transmission rate",
            405: " ms transmission rate",
            406: " ms transmission rate",
            407: " ms transmission rate",
            408: " ms transmission rate",
            409: " ms transmission rate",
            41: " ms transmission rate",
            410: " ms transmission rate",
            411: " ms transmission rate",
            412: " ms transmission rate",
            413: " ms transmission rate",
            414: " ms transmission rate",
            415: " ms transmission rate",
            416: " ms transmission rate",
            417: " ms transmission rate",
            418: " ms transmission rate",
            419: " ms transmission rate",
            42: " ms transmission rate",
            420: " ms transmission rate",
            421: " ms transmission rate",
            422: " ms transmission rate",
            423: " ms transmission rate",
            424: " ms transmission rate",
            425: " ms transmission rate",
            426: " ms transmission rate",
            427: " ms transmission rate",
            428: " ms transmission rate",
            429: " ms transmission rate",
            43: " ms transmission rate",
            430: " ms transmission rate",
            431: " ms transmission rate",
            432: " ms transmission rate",
            433: " ms transmission rate",
            434: " ms transmission rate",
            435: " ms transmission rate",
            436: " ms transmission rate",
            437: " ms transmission rate",
            438: " ms transmission rate",
            439: " ms transmission rate",
            44: " ms transmission rate",
            440: " ms transmission rate",
            441: " ms transmission rate",
            442: " ms transmission rate",
            443: " ms transmission rate",
            444: " ms transmission rate",
            445: " ms transmission rate",
            446: " ms transmission rate",
            447: " ms transmission rate",
            448: " ms transmission rate",
            449: " ms transmission rate",
            45: " ms transmission rate",
            450: " ms transmission rate",
            451: " ms transmission rate",
            452: " ms transmission rate",
            453: " ms transmission rate",
            454: " ms transmission rate",
            455: " ms transmission rate",
            456: " ms transmission rate",
            457: " ms transmission rate",
            458: " ms transmission rate",
            459: " ms transmission rate",
            46: " ms transmission rate",
            460: " ms transmission rate",
            461: " ms transmission rate",
            462: " ms transmission rate",
            463: " ms transmission rate",
            464: " ms transmission rate",
            465: " ms transmission rate",
            466: " ms transmission rate",
            467: " ms transmission rate",
            468: " ms transmission rate",
            469: " ms transmission rate",
            47: " ms transmission rate",
            470: " ms transmission rate",
            471: " ms transmission rate",
            472: " ms transmission rate",
            473: " ms transmission rate",
            474: " ms transmission rate",
            475: " ms transmission rate",
            476: " ms transmission rate",
            477: " ms transmission rate",
            478: " ms transmission rate",
            479: " ms transmission rate",
            48: " ms transmission rate",
            480: " ms transmission rate",
            481: " ms transmission rate",
            482: " ms transmission rate",
            483: " ms transmission rate",
            484: " ms transmission rate",
            485: " ms transmission rate",
            486: " ms transmission rate",
            487: " ms transmission rate",
            488: " ms transmission rate",
            489: " ms transmission rate",
            49: " ms transmission rate",
            490: " ms transmission rate",
            491: " ms transmission rate",
            492: " ms transmission rate",
            493: " ms transmission rate",
            494: " ms transmission rate",
            495: " ms transmission rate",
            496: " ms transmission rate",
            497: " ms transmission rate",
            498: " ms transmission rate",
            499: " ms transmission rate",
            5: " ms transmission rate",
            50: " ms transmission rate",
            500: " ms transmission rate",
            501: " ms transmission rate",
            502: " ms transmission rate",
            503: " ms transmission rate",
            504: " ms transmission rate",
            505: " ms transmission rate",
            506: " ms transmission rate",
            507: " ms transmission rate",
            508: " ms transmission rate",
            509: " ms transmission rate",
            51: " ms transmission rate",
            510: " ms transmission rate",
            511: " ms transmission rate",
            512: " ms transmission rate",
            513: " ms transmission rate",
            514: " ms transmission rate",
            515: " ms transmission rate",
            516: " ms transmission rate",
            517: " ms transmission rate",
            518: " ms transmission rate",
            519: " ms transmission rate",
            52: " ms transmission rate",
            520: " ms transmission rate",
            521: " ms transmission rate",
            522: " ms transmission rate",
            523: " ms transmission rate",
            524: " ms transmission rate",
            525: " ms transmission rate",
            526: " ms transmission rate",
            527: " ms transmission rate",
            528: " ms transmission rate",
            529: " ms transmission rate",
            53: " ms transmission rate",
            530: " ms transmission rate",
            531: " ms transmission rate",
            532: " ms transmission rate",
            533: " ms transmission rate",
            534: " ms transmission rate",
            535: " ms transmission rate",
            536: " ms transmission rate",
            537: " ms transmission rate",
            538: " ms transmission rate",
            539: " ms transmission rate",
            54: " ms transmission rate",
            540: " ms transmission rate",
            541: " ms transmission rate",
            542: " ms transmission rate",
            543: " ms transmission rate",
            544: " ms transmission rate",
            545: " ms transmission rate",
            546: " ms transmission rate",
            547: " ms transmission rate",
            548: " ms transmission rate",
            549: " ms transmission rate",
            55: " ms transmission rate",
            550: " ms transmission rate",
            551: " ms transmission rate",
            552: " ms transmission rate",
            553: " ms transmission rate",
            554: " ms transmission rate",
            555: " ms transmission rate",
            556: " ms transmission rate",
            557: " ms transmission rate",
            558: " ms transmission rate",
            559: " ms transmission rate",
            56: " ms transmission rate",
            560: " ms transmission rate",
            561: " ms transmission rate",
            562: " ms transmission rate",
            563: " ms transmission rate",
            564: " ms transmission rate",
            565: " ms transmission rate",
            566: " ms transmission rate",
            567: " ms transmission rate",
            568: " ms transmission rate",
            569: " ms transmission rate",
            57: " ms transmission rate",
            570: " ms transmission rate",
            571: " ms transmission rate",
            572: " ms transmission rate",
            573: " ms transmission rate",
            574: " ms transmission rate",
            575: " ms transmission rate",
            576: " ms transmission rate",
            577: " ms transmission rate",
            578: " ms transmission rate",
            579: " ms transmission rate",
            58: " ms transmission rate",
            580: " ms transmission rate",
            581: " ms transmission rate",
            582: " ms transmission rate",
            583: " ms transmission rate",
            584: " ms transmission rate",
            585: " ms transmission rate",
            586: " ms transmission rate",
            587: " ms transmission rate",
            588: " ms transmission rate",
            589: " ms transmission rate",
            59: " ms transmission rate",
            590: " ms transmission rate",
            591: " ms transmission rate",
            592: " ms transmission rate",
            593: " ms transmission rate",
            594: " ms transmission rate",
            595: " ms transmission rate",
            596: " ms transmission rate",
            597: " ms transmission rate",
            598: " ms transmission rate",
            599: " ms transmission rate",
            6: " ms transmission rate",
            60: " ms transmission rate",
            600: " ms transmission rate",
            601: " ms transmission rate",
            602: " ms transmission rate",
            603: " ms transmission rate",
            604: " ms transmission rate",
            605: " ms transmission rate",
            606: " ms transmission rate",
            607: " ms transmission rate",
            608: " ms transmission rate",
            609: " ms transmission rate",
            61: " ms transmission rate",
            610: " ms transmission rate",
            611: " ms transmission rate",
            612: " ms transmission rate",
            613: " ms transmission rate",
            614: " ms transmission rate",
            615: " ms transmission rate",
            616: " ms transmission rate",
            617: " ms transmission rate",
            618: " ms transmission rate",
            619: " ms transmission rate",
            62: " ms transmission rate",
            620: " ms transmission rate",
            621: " ms transmission rate",
            622: " ms transmission rate",
            623: " ms transmission rate",
            624: " ms transmission rate",
            625: " ms transmission rate",
            626: " ms transmission rate",
            627: " ms transmission rate",
            628: " ms transmission rate",
            629: " ms transmission rate",
            63: " ms transmission rate",
            630: " ms transmission rate",
            631: " ms transmission rate",
            632: " ms transmission rate",
            633: " ms transmission rate",
            634: " ms transmission rate",
            635: " ms transmission rate",
            636: " ms transmission rate",
            637: " ms transmission rate",
            638: " ms transmission rate",
            639: " ms transmission rate",
            64: " ms transmission rate",
            640: " ms transmission rate",
            641: " ms transmission rate",
            642: " ms transmission rate",
            643: " ms transmission rate",
            644: " ms transmission rate",
            645: " ms transmission rate",
            646: " ms transmission rate",
            647: " ms transmission rate",
            648: " ms transmission rate",
            649: " ms transmission rate",
            65: " ms transmission rate",
            650: " ms transmission rate",
            651: " ms transmission rate",
            652: " ms transmission rate",
            653: " ms transmission rate",
            654: " ms transmission rate",
            655: " ms transmission rate",
            656: " ms transmission rate",
            657: " ms transmission rate",
            658: " ms transmission rate",
            659: " ms transmission rate",
            66: " ms transmission rate",
            660: " ms transmission rate",
            661: " ms transmission rate",
            662: " ms transmission rate",
            663: " ms transmission rate",
            664: " ms transmission rate",
            665: " ms transmission rate",
            666: " ms transmission rate",
            667: " ms transmission rate",
            668: " ms transmission rate",
            669: " ms transmission rate",
            67: " ms transmission rate",
            670: " ms transmission rate",
            671: " ms transmission rate",
            672: " ms transmission rate",
            673: " ms transmission rate",
            674: " ms transmission rate",
            675: " ms transmission rate",
            676: " ms transmission rate",
            677: " ms transmission rate",
            678: " ms transmission rate",
            679: " ms transmission rate",
            68: " ms transmission rate",
            680: " ms transmission rate",
            681: " ms transmission rate",
            682: " ms transmission rate",
            683: " ms transmission rate",
            684: " ms transmission rate",
            685: " ms transmission rate",
            686: " ms transmission rate",
            687: " ms transmission rate",
            688: " ms transmission rate",
            689: " ms transmission rate",
            69: " ms transmission rate",
            690: " ms transmission rate",
            691: " ms transmission rate",
            692: " ms transmission rate",
            693: " ms transmission rate",
            694: " ms transmission rate",
            695: " ms transmission rate",
            696: " ms transmission rate",
            697: " ms transmission rate",
            698: " ms transmission rate",
            699: " ms transmission rate",
            7: " ms transmission rate",
            70: " ms transmission rate",
            700: " ms transmission rate",
            701: " ms transmission rate",
            702: " ms transmission rate",
            703: " ms transmission rate",
            704: " ms transmission rate",
            705: " ms transmission rate",
            706: " ms transmission rate",
            707: " ms transmission rate",
            708: " ms transmission rate",
            709: " ms transmission rate",
            71: " ms transmission rate",
            710: " ms transmission rate",
            711: " ms transmission rate",
            712: " ms transmission rate",
            713: " ms transmission rate",
            714: " ms transmission rate",
            715: " ms transmission rate",
            716: " ms transmission rate",
            717: " ms transmission rate",
            718: " ms transmission rate",
            719: " ms transmission rate",
            72: " ms transmission rate",
            720: " ms transmission rate",
            721: " ms transmission rate",
            722: " ms transmission rate",
            723: " ms transmission rate",
            724: " ms transmission rate",
            725: " ms transmission rate",
            726: " ms transmission rate",
            727: " ms transmission rate",
            728: " ms transmission rate",
            729: " ms transmission rate",
            73: " ms transmission rate",
            730: " ms transmission rate",
            731: " ms transmission rate",
            732: " ms transmission rate",
            733: " ms transmission rate",
            734: " ms transmission rate",
            735: " ms transmission rate",
            736: " ms transmission rate",
            737: " ms transmission rate",
            738: " ms transmission rate",
            739: " ms transmission rate",
            74: " ms transmission rate",
            740: " ms transmission rate",
            741: " ms transmission rate",
            742: " ms transmission rate",
            743: " ms transmission rate",
            744: " ms transmission rate",
            745: " ms transmission rate",
            746: " ms transmission rate",
            747: " ms transmission rate",
            748: " ms transmission rate",
            749: " ms transmission rate",
            75: " ms transmission rate",
            750: " ms transmission rate",
            751: " ms transmission rate",
            752: " ms transmission rate",
            753: " ms transmission rate",
            754: " ms transmission rate",
            755: " ms transmission rate",
            756: " ms transmission rate",
            757: " ms transmission rate",
            758: " ms transmission rate",
            759: " ms transmission rate",
            76: " ms transmission rate",
            760: " ms transmission rate",
            761: " ms transmission rate",
            762: " ms transmission rate",
            763: " ms transmission rate",
            764: " ms transmission rate",
            765: " ms transmission rate",
            766: " ms transmission rate",
            767: " ms transmission rate",
            768: " ms transmission rate",
            769: " ms transmission rate",
            77: " ms transmission rate",
            770: " ms transmission rate",
            771: " ms transmission rate",
            772: " ms transmission rate",
            773: " ms transmission rate",
            774: " ms transmission rate",
            775: " ms transmission rate",
            776: " ms transmission rate",
            777: " ms transmission rate",
            778: " ms transmission rate",
            779: " ms transmission rate",
            78: " ms transmission rate",
            780: " ms transmission rate",
            781: " ms transmission rate",
            782: " ms transmission rate",
            783: " ms transmission rate",
            784: " ms transmission rate",
            785: " ms transmission rate",
            786: " ms transmission rate",
            787: " ms transmission rate",
            788: " ms transmission rate",
            789: " ms transmission rate",
            79: " ms transmission rate",
            790: " ms transmission rate",
            791: " ms transmission rate",
            792: " ms transmission rate",
            793: " ms transmission rate",
            794: " ms transmission rate",
            795: " ms transmission rate",
            796: " ms transmission rate",
            797: " ms transmission rate",
            798: " ms transmission rate",
            799: " ms transmission rate",
            8: " ms transmission rate",
            80: " ms transmission rate",
            800: " ms transmission rate",
            801: " ms transmission rate",
            802: " ms transmission rate",
            803: " ms transmission rate",
            804: " ms transmission rate",
            805: " ms transmission rate",
            806: " ms transmission rate",
            807: " ms transmission rate",
            808: " ms transmission rate",
            809: " ms transmission rate",
            81: " ms transmission rate",
            810: " ms transmission rate",
            811: " ms transmission rate",
            812: " ms transmission rate",
            813: " ms transmission rate",
            814: " ms transmission rate",
            815: " ms transmission rate",
            816: " ms transmission rate",
            817: " ms transmission rate",
            818: " ms transmission rate",
            819: " ms transmission rate",
            82: " ms transmission rate",
            820: " ms transmission rate",
            821: " ms transmission rate",
            822: " ms transmission rate",
            823: " ms transmission rate",
            824: " ms transmission rate",
            825: " ms transmission rate",
            826: " ms transmission rate",
            827: " ms transmission rate",
            828: " ms transmission rate",
            829: " ms transmission rate",
            83: " ms transmission rate",
            830: " ms transmission rate",
            831: " ms transmission rate",
            832: " ms transmission rate",
            833: " ms transmission rate",
            834: " ms transmission rate",
            835: " ms transmission rate",
            836: " ms transmission rate",
            837: " ms transmission rate",
            838: " ms transmission rate",
            839: " ms transmission rate",
            84: " ms transmission rate",
            840: " ms transmission rate",
            841: " ms transmission rate",
            842: " ms transmission rate",
            843: " ms transmission rate",
            844: " ms transmission rate",
            845: " ms transmission rate",
            846: " ms transmission rate",
            847: " ms transmission rate",
            848: " ms transmission rate",
            849: " ms transmission rate",
            85: " ms transmission rate",
            850: " ms transmission rate",
            851: " ms transmission rate",
            852: " ms transmission rate",
            853: " ms transmission rate",
            854: " ms transmission rate",
            855: " ms transmission rate",
            856: " ms transmission rate",
            857: " ms transmission rate",
            858: " ms transmission rate",
            859: " ms transmission rate",
            86: " ms transmission rate",
            860: " ms transmission rate",
            861: " ms transmission rate",
            862: " ms transmission rate",
            863: " ms transmission rate",
            864: " ms transmission rate",
            865: " ms transmission rate",
            866: " ms transmission rate",
            867: " ms transmission rate",
            868: " ms transmission rate",
            869: " ms transmission rate",
            87: " ms transmission rate",
            870: " ms transmission rate",
            871: " ms transmission rate",
            872: " ms transmission rate",
            873: " ms transmission rate",
            874: " ms transmission rate",
            875: " ms transmission rate",
            876: " ms transmission rate",
            877: " ms transmission rate",
            878: " ms transmission rate",
            879: " ms transmission rate",
            88: " ms transmission rate",
            880: " ms transmission rate",
            881: " ms transmission rate",
            882: " ms transmission rate",
            883: " ms transmission rate",
            884: " ms transmission rate",
            885: " ms transmission rate",
            886: " ms transmission rate",
            887: " ms transmission rate",
            888: " ms transmission rate",
            889: " ms transmission rate",
            89: " ms transmission rate",
            890: " ms transmission rate",
            891: " ms transmission rate",
            892: " ms transmission rate",
            893: " ms transmission rate",
            894: " ms transmission rate",
            895: " ms transmission rate",
            896: " ms transmission rate",
            897: " ms transmission rate",
            898: " ms transmission rate",
            899: " ms transmission rate",
            9: " ms transmission rate",
            90: " ms transmission rate",
            900: " ms transmission rate",
            901: " ms transmission rate",
            902: " ms transmission rate",
            903: " ms transmission rate",
            904: " ms transmission rate",
            905: " ms transmission rate",
            906: " ms transmission rate",
            907: " ms transmission rate",
            908: " ms transmission rate",
            909: " ms transmission rate",
            91: " ms transmission rate",
            910: " ms transmission rate",
            911: " ms transmission rate",
            912: " ms transmission rate",
            913: " ms transmission rate",
            914: " ms transmission rate",
            915: " ms transmission rate",
            916: " ms transmission rate",
            917: " ms transmission rate",
            918: " ms transmission rate",
            919: " ms transmission rate",
            92: " ms transmission rate",
            920: " ms transmission rate",
            921: " ms transmission rate",
            922: " ms transmission rate",
            923: " ms transmission rate",
            924: " ms transmission rate",
            925: " ms transmission rate",
            926: " ms transmission rate",
            927: " ms transmission rate",
            928: " ms transmission rate",
            929: " ms transmission rate",
            93: " ms transmission rate",
            930: " ms transmission rate",
            931: " ms transmission rate",
            932: " ms transmission rate",
            933: " ms transmission rate",
            934: " ms transmission rate",
            935: " ms transmission rate",
            936: " ms transmission rate",
            937: " ms transmission rate",
            938: " ms transmission rate",
            939: " ms transmission rate",
            94: " ms transmission rate",
            940: " ms transmission rate",
            941: " ms transmission rate",
            942: " ms transmission rate",
            943: " ms transmission rate",
            944: " ms transmission rate",
            945: " ms transmission rate",
            946: " ms transmission rate",
            947: " ms transmission rate",
            948: " ms transmission rate",
            949: " ms transmission rate",
            95: " ms transmission rate",
            950: " ms transmission rate",
            951: " ms transmission rate",
            952: " ms transmission rate",
            953: " ms transmission rate",
            954: " ms transmission rate",
            955: " ms transmission rate",
            956: " ms transmission rate",
            957: " ms transmission rate",
            958: " ms transmission rate",
            959: " ms transmission rate",
            96: " ms transmission rate",
            960: " ms transmission rate",
            961: " ms transmission rate",
            962: " ms transmission rate",
            963: " ms transmission rate",
            964: " ms transmission rate",
            965: " ms transmission rate",
            966: " ms transmission rate",
            967: " ms transmission rate",
            968: " ms transmission rate",
            969: " ms transmission rate",
            97: " ms transmission rate",
            970: " ms transmission rate",
            971: " ms transmission rate",
            972: " ms transmission rate",
            973: " ms transmission rate",
            974: " ms transmission rate",
            975: " ms transmission rate",
            976: " ms transmission rate",
            977: " ms transmission rate",
            978: " ms transmission rate",
            979: " ms transmission rate",
            98: " ms transmission rate",
            980: " ms transmission rate",
            981: " ms transmission rate",
            982: " ms transmission rate",
            983: " ms transmission rate",
            984: " ms transmission rate",
            985: " ms transmission rate",
            986: " ms transmission rate",
            987: " ms transmission rate",
            988: " ms transmission rate",
            989: " ms transmission rate",
            99: " ms transmission rate",
            990: " ms transmission rate",
            991: " ms transmission rate",
            992: " ms transmission rate",
            993: " ms transmission rate",
            994: " ms transmission rate",
            995: " ms transmission rate",
            996: " ms transmission rate",
            997: " ms transmission rate",
            998: " ms transmission rate",
            999: " ms transmission rate"
        },
        3350: {
            0: "p1 = accelerator pedal/operator selection",
            1: "p2 = cruise control",
            10: "p6-p31 = reserved for sae assignment ",
            11: "p6-p31 = reserved for sae assignment ",
            12: "p6-p31 = reserved for sae assignment ",
            13: "p6-p31 = reserved for sae assignment ",
            14: "p6-p31 = reserved for sae assignment ",
            15: "p6-p31 = reserved for sae assignment ",
            16: "p6-p31 = reserved for sae assignment ",
            17: "p6-p31 = reserved for sae assignment ",
            18: "p6-p31 = reserved for sae assignment ",
            19: "p6-p31 = reserved for sae assignment ",
            2: "p3 = pto governor",
            20: "p6-p31 = reserved for sae assignment ",
            21: "p6-p31 = reserved for sae assignment ",
            22: "p6-p31 = reserved for sae assignment ",
            23: "p6-p31 = reserved for sae assignment ",
            24: "p6-p31 = reserved for sae assignment ",
            25: "p6-p31 = reserved for sae assignment ",
            26: "p6-p31 = reserved for sae assignment ",
            27: "p6-p31 = reserved for sae assignment ",
            28: "p6-p31 = reserved for sae assignment ",
            29: "p6-p31 = reserved for sae assignment ",
            3: "p4 = road speed governor",
            30: "p6-p31 = reserved for sae assignment ",
            31: "p32 = temporary power train control (original use of tsc1 command) ",
            4: "p5 = engine protection",
            5: "p6-p31 = reserved for sae assignment ",
            6: "p6-p31 = reserved for sae assignment ",
            7: "p6-p31 = reserved for sae assignment ",
            8: "p6-p31 = reserved for sae assignment ",
            9: "p6-p31 = reserved for sae assignment "
        },
        3353: {
            0: " alternator 1 not charging",
            1: " alternator 1 charging",
            2: " error",
            3: " not available"
        },
        3354: {
            0: " alternator 2 not charging",
            1: " alternator 2 charging",
            2: " error",
            3: " not available"
        },
        3355: {
            0: " alternator 3 not charging",
            1: " alternator 3 charging",
            2: " error",
            3: " not available"
        },
        3356: {
            0: " alternator 4 not charging",
            1: " alternator 4 charging",
            2: " error",
            3: " not available"
        },
        3359: {
            0: "no restriction",
            1: "restriction exists on oil filter",
            2: "error",
            3: "not available"
        },
        3367: {
            0: "  relative blade height fully functional.  data is within sensor specification.   ",
            1: "  relative blade height degraded.  data is suspect due to environmental conditions.",
            2: "  relative blade height failed.  ",
            3: "  relative blade height not available"
        },
        3368: {
            0: "off",
            1: "on",
            10: "reserved for future assignment",
            100: "reserved for future assignment",
            101: "reserved for future assignment",
            102: "reserved for future assignment",
            103: "reserved for future assignment",
            104: "reserved for future assignment",
            105: "reserved for future assignment",
            106: "reserved for future assignment",
            107: "reserved for future assignment",
            108: "reserved for future assignment",
            109: "reserved for future assignment",
            11: "reserved for future assignment",
            110: "reserved for future assignment",
            111: "reserved for future assignment",
            112: "reserved for future assignment",
            113: "reserved for future assignment",
            114: "reserved for future assignment",
            115: "reserved for future assignment",
            116: "reserved for future assignment",
            117: "reserved for future assignment",
            118: "reserved for future assignment",
            119: "reserved for future assignment",
            12: "reserved for future assignment",
            120: "reserved for future assignment",
            121: "reserved for future assignment",
            122: "reserved for future assignment",
            123: "reserved for future assignment",
            124: "reserved for future assignment",
            125: "reserved for future assignment",
            126: "reserved for future assignment",
            127: "reserved for future assignment",
            128: "reserved for future assignment",
            129: "reserved for future assignment",
            13: "reserved for future assignment",
            130: "reserved for future assignment",
            131: "reserved for future assignment",
            132: "reserved for future assignment",
            133: "reserved for future assignment",
            134: "reserved for future assignment",
            135: "reserved for future assignment",
            136: "reserved for future assignment",
            137: "reserved for future assignment",
            138: "reserved for future assignment",
            139: "reserved for future assignment",
            14: "reserved for future assignment",
            140: "reserved for future assignment",
            141: "reserved for future assignment",
            142: "reserved for future assignment",
            143: "reserved for future assignment",
            144: "reserved for future assignment",
            145: "reserved for future assignment",
            146: "reserved for future assignment",
            147: "reserved for future assignment",
            148: "reserved for future assignment",
            149: "reserved for future assignment",
            15: "reserved for future assignment",
            150: "reserved for future assignment",
            151: "reserved for future assignment",
            152: "reserved for future assignment",
            153: "reserved for future assignment",
            154: "reserved for future assignment",
            155: "reserved for future assignment",
            156: "reserved for future assignment",
            157: "reserved for future assignment",
            158: "reserved for future assignment",
            159: "reserved for future assignment",
            16: "reserved for future assignment",
            160: "reserved for future assignment",
            161: "reserved for future assignment",
            162: "reserved for future assignment",
            163: "reserved for future assignment",
            164: "reserved for future assignment",
            165: "reserved for future assignment",
            166: "reserved for future assignment",
            167: "reserved for future assignment",
            168: "reserved for future assignment",
            169: "reserved for future assignment",
            17: "reserved for future assignment",
            170: "reserved for future assignment",
            171: "reserved for future assignment",
            172: "reserved for future assignment",
            173: "reserved for future assignment",
            174: "reserved for future assignment",
            175: "reserved for future assignment",
            176: "reserved for future assignment",
            177: "reserved for future assignment",
            178: "reserved for future assignment",
            179: "reserved for future assignment",
            18: "reserved for future assignment",
            180: "reserved for future assignment",
            181: "reserved for future assignment",
            182: "reserved for future assignment",
            183: "reserved for future assignment",
            184: "reserved for future assignment",
            185: "reserved for future assignment",
            186: "reserved for future assignment",
            187: "reserved for future assignment",
            188: "reserved for future assignment",
            189: "reserved for future assignment",
            19: "reserved for future assignment",
            190: "reserved for future assignment",
            191: "reserved for future assignment",
            192: "reserved for future assignment",
            193: "reserved for future assignment",
            194: "reserved for future assignment",
            195: "reserved for future assignment",
            196: "reserved for future assignment",
            197: "reserved for future assignment",
            198: "reserved for future assignment",
            199: "reserved for future assignment",
            2: "not authorized to operate on network",
            20: "reserved for future assignment",
            200: "reserved for future assignment",
            201: "reserved for future assignment",
            202: "reserved for future assignment",
            203: "reserved for future assignment",
            204: "reserved for future assignment",
            205: "reserved for future assignment",
            206: "reserved for future assignment",
            207: "reserved for future assignment",
            208: "reserved for future assignment",
            209: "reserved for future assignment",
            21: "reserved for future assignment",
            210: "reserved for future assignment",
            211: "reserved for future assignment",
            212: "reserved for future assignment",
            213: "reserved for future assignment",
            214: "reserved for future assignment",
            215: "reserved for future assignment",
            216: "reserved for future assignment",
            217: "reserved for future assignment",
            218: "reserved for future assignment",
            219: "reserved for future assignment",
            22: "reserved for future assignment",
            220: "reserved for future assignment",
            221: "reserved for future assignment",
            222: "reserved for future assignment",
            223: "reserved for future assignment",
            224: "reserved for future assignment",
            225: "reserved for future assignment",
            226: "reserved for future assignment",
            227: "reserved for future assignment",
            228: "reserved for future assignment",
            229: "reserved for future assignment",
            23: "reserved for future assignment",
            230: "reserved for future assignment",
            231: "reserved for future assignment",
            232: "reserved for future assignment",
            233: "reserved for future assignment",
            234: "reserved for future assignment",
            235: "reserved for future assignment",
            236: "reserved for future assignment",
            237: "reserved for future assignment",
            238: "reserved for future assignment",
            239: "reserved for future assignment",
            24: "reserved for future assignment",
            240: "reserved for future assignment",
            241: "reserved for future assignment",
            242: "reserved for future assignment",
            243: "reserved for future assignment",
            244: "reserved for future assignment",
            245: "reserved for future assignment",
            246: "reserved for future assignment",
            247: "reserved for future assignment",
            248: "reserved for future assignment",
            249: "reserved for future assignment",
            25: "reserved for future assignment",
            250: "reserved for future assignment",
            251: "reserved",
            252: "reserved",
            253: "reserved",
            254: "error",
            255: "not available",
            26: "reserved for future assignment",
            27: "reserved for future assignment",
            28: "reserved for future assignment",
            29: "reserved for future assignment",
            3: "reserved for future assignment",
            30: "reserved for future assignment",
            31: "reserved for future assignment",
            32: "reserved for future assignment",
            33: "reserved for future assignment",
            34: "reserved for future assignment",
            35: "reserved for future assignment",
            36: "reserved for future assignment",
            37: "reserved for future assignment",
            38: "reserved for future assignment",
            39: "reserved for future assignment",
            4: "reserved for future assignment",
            40: "reserved for future assignment",
            41: "reserved for future assignment",
            42: "reserved for future assignment",
            43: "reserved for future assignment",
            44: "reserved for future assignment",
            45: "reserved for future assignment",
            46: "reserved for future assignment",
            47: "reserved for future assignment",
            48: "reserved for future assignment",
            49: "reserved for future assignment",
            5: "reserved for future assignment",
            50: "reserved for future assignment",
            51: "reserved for future assignment",
            52: "reserved for future assignment",
            53: "reserved for future assignment",
            54: "reserved for future assignment",
            55: "reserved for future assignment",
            56: "reserved for future assignment",
            57: "reserved for future assignment",
            58: "reserved for future assignment",
            59: "reserved for future assignment",
            6: "reserved for future assignment",
            60: "reserved for future assignment",
            61: "reserved for future assignment",
            62: "reserved for future assignment",
            63: "reserved for future assignment",
            64: "reserved for future assignment",
            65: "reserved for future assignment",
            66: "reserved for future assignment",
            67: "reserved for future assignment",
            68: "reserved for future assignment",
            69: "reserved for future assignment",
            7: "reserved for future assignment",
            70: "reserved for future assignment",
            71: "reserved for future assignment",
            72: "reserved for future assignment",
            73: "reserved for future assignment",
            74: "reserved for future assignment",
            75: "reserved for future assignment",
            76: "reserved for future assignment",
            77: "reserved for future assignment",
            78: "reserved for future assignment",
            79: "reserved for future assignment",
            8: "reserved for future assignment",
            80: "reserved for future assignment",
            81: "reserved for future assignment",
            82: "reserved for future assignment",
            83: "reserved for future assignment",
            84: "reserved for future assignment",
            85: "reserved for future assignment",
            86: "reserved for future assignment",
            87: "reserved for future assignment",
            88: "reserved for future assignment",
            89: "reserved for future assignment",
            9: "reserved for future assignment",
            90: "reserved for future assignment",
            91: "reserved for future assignment",
            92: "reserved for future assignment",
            93: "reserved for future assignment",
            94: "reserved for future assignment",
            95: "reserved for future assignment",
            96: "reserved for future assignment",
            97: "reserved for future assignment",
            98: "reserved for future assignment",
            99: "reserved for future assignment"
        },
        3369: {
            0: "local service",
            1: "roaming service",
            10: "reserved for future assignment",
            100: "reserved for future assignment",
            101: "reserved for future assignment",
            102: "reserved for future assignment",
            103: "reserved for future assignment",
            104: "reserved for future assignment",
            105: "reserved for future assignment",
            106: "reserved for future assignment",
            107: "reserved for future assignment",
            108: "reserved for future assignment",
            109: "reserved for future assignment",
            11: "reserved for future assignment",
            110: "reserved for future assignment",
            111: "reserved for future assignment",
            112: "reserved for future assignment",
            113: "reserved for future assignment",
            114: "reserved for future assignment",
            115: "reserved for future assignment",
            116: "reserved for future assignment",
            117: "reserved for future assignment",
            118: "reserved for future assignment",
            119: "reserved for future assignment",
            12: "reserved for future assignment",
            120: "reserved for future assignment",
            121: "reserved for future assignment",
            122: "reserved for future assignment",
            123: "reserved for future assignment",
            124: "reserved for future assignment",
            125: "reserved for future assignment",
            126: "reserved for future assignment",
            127: "reserved for future assignment",
            128: "reserved for future assignment",
            129: "reserved for future assignment",
            13: "reserved for future assignment",
            130: "reserved for future assignment",
            131: "reserved for future assignment",
            132: "reserved for future assignment",
            133: "reserved for future assignment",
            134: "reserved for future assignment",
            135: "reserved for future assignment",
            136: "reserved for future assignment",
            137: "reserved for future assignment",
            138: "reserved for future assignment",
            139: "reserved for future assignment",
            14: "reserved for future assignment",
            140: "reserved for future assignment",
            141: "reserved for future assignment",
            142: "reserved for future assignment",
            143: "reserved for future assignment",
            144: "reserved for future assignment",
            145: "reserved for future assignment",
            146: "reserved for future assignment",
            147: "reserved for future assignment",
            148: "reserved for future assignment",
            149: "reserved for future assignment",
            15: "reserved for future assignment",
            150: "reserved for future assignment",
            151: "reserved for future assignment",
            152: "reserved for future assignment",
            153: "reserved for future assignment",
            154: "reserved for future assignment",
            155: "reserved for future assignment",
            156: "reserved for future assignment",
            157: "reserved for future assignment",
            158: "reserved for future assignment",
            159: "reserved for future assignment",
            16: "reserved for future assignment",
            160: "reserved for future assignment",
            161: "reserved for future assignment",
            162: "reserved for future assignment",
            163: "reserved for future assignment",
            164: "reserved for future assignment",
            165: "reserved for future assignment",
            166: "reserved for future assignment",
            167: "reserved for future assignment",
            168: "reserved for future assignment",
            169: "reserved for future assignment",
            17: "reserved for future assignment",
            170: "reserved for future assignment",
            171: "reserved for future assignment",
            172: "reserved for future assignment",
            173: "reserved for future assignment",
            174: "reserved for future assignment",
            175: "reserved for future assignment",
            176: "reserved for future assignment",
            177: "reserved for future assignment",
            178: "reserved for future assignment",
            179: "reserved for future assignment",
            18: "reserved for future assignment",
            180: "reserved for future assignment",
            181: "reserved for future assignment",
            182: "reserved for future assignment",
            183: "reserved for future assignment",
            184: "reserved for future assignment",
            185: "reserved for future assignment",
            186: "reserved for future assignment",
            187: "reserved for future assignment",
            188: "reserved for future assignment",
            189: "reserved for future assignment",
            19: "reserved for future assignment",
            190: "reserved for future assignment",
            191: "reserved for future assignment",
            192: "reserved for future assignment",
            193: "reserved for future assignment",
            194: "reserved for future assignment",
            195: "reserved for future assignment",
            196: "reserved for future assignment",
            197: "reserved for future assignment",
            198: "reserved for future assignment",
            199: "reserved for future assignment",
            2: "service not available",
            20: "reserved for future assignment",
            200: "reserved for future assignment",
            201: "reserved for future assignment",
            202: "reserved for future assignment",
            203: "reserved for future assignment",
            204: "reserved for future assignment",
            205: "reserved for future assignment",
            206: "reserved for future assignment",
            207: "reserved for future assignment",
            208: "reserved for future assignment",
            209: "reserved for future assignment",
            21: "reserved for future assignment",
            210: "reserved for future assignment",
            211: "reserved for future assignment",
            212: "reserved for future assignment",
            213: "reserved for future assignment",
            214: "reserved for future assignment",
            215: "reserved for future assignment",
            216: "reserved for future assignment",
            217: "reserved for future assignment",
            218: "reserved for future assignment",
            219: "reserved for future assignment",
            22: "reserved for future assignment",
            220: "reserved for future assignment",
            221: "reserved for future assignment",
            222: "reserved for future assignment",
            223: "reserved for future assignment",
            224: "reserved for future assignment",
            225: "reserved for future assignment",
            226: "reserved for future assignment",
            227: "reserved for future assignment",
            228: "reserved for future assignment",
            229: "reserved for future assignment",
            23: "reserved for future assignment",
            230: "reserved for future assignment",
            231: "reserved for future assignment",
            232: "reserved for future assignment",
            233: "reserved for future assignment",
            234: "reserved for future assignment",
            235: "reserved for future assignment",
            236: "reserved for future assignment",
            237: "reserved for future assignment",
            238: "reserved for future assignment",
            239: "reserved for future assignment",
            24: "reserved for future assignment",
            240: "reserved for future assignment",
            241: "reserved for future assignment",
            242: "reserved for future assignment",
            243: "reserved for future assignment",
            244: "reserved for future assignment",
            245: "reserved for future assignment",
            246: "reserved for future assignment",
            247: "reserved for future assignment",
            248: "reserved for future assignment",
            249: "reserved for future assignment",
            25: "reserved for future assignment",
            250: "reserved for future assignment",
            251: "reserved",
            252: "reserved",
            253: "reserved",
            254: "error",
            255: "not available",
            26: "reserved for future assignment",
            27: "reserved for future assignment",
            28: "reserved for future assignment",
            29: "reserved for future assignment",
            3: "not authorized to operate on service",
            30: "reserved for future assignment",
            31: "reserved for future assignment",
            32: "reserved for future assignment",
            33: "reserved for future assignment",
            34: "reserved for future assignment",
            35: "reserved for future assignment",
            36: "reserved for future assignment",
            37: "reserved for future assignment",
            38: "reserved for future assignment",
            39: "reserved for future assignment",
            4: "reserved for future assignment",
            40: "reserved for future assignment",
            41: "reserved for future assignment",
            42: "reserved for future assignment",
            43: "reserved for future assignment",
            44: "reserved for future assignment",
            45: "reserved for future assignment",
            46: "reserved for future assignment",
            47: "reserved for future assignment",
            48: "reserved for future assignment",
            49: "reserved for future assignment",
            5: "reserved for future assignment",
            50: "reserved for future assignment",
            51: "reserved for future assignment",
            52: "reserved for future assignment",
            53: "reserved for future assignment",
            54: "reserved for future assignment",
            55: "reserved for future assignment",
            56: "reserved for future assignment",
            57: "reserved for future assignment",
            58: "reserved for future assignment",
            59: "reserved for future assignment",
            6: "reserved for future assignment",
            60: "reserved for future assignment",
            61: "reserved for future assignment",
            62: "reserved for future assignment",
            63: "reserved for future assignment",
            64: "reserved for future assignment",
            65: "reserved for future assignment",
            66: "reserved for future assignment",
            67: "reserved for future assignment",
            68: "reserved for future assignment",
            69: "reserved for future assignment",
            7: "reserved for future assignment",
            70: "reserved for future assignment",
            71: "reserved for future assignment",
            72: "reserved for future assignment",
            73: "reserved for future assignment",
            74: "reserved for future assignment",
            75: "reserved for future assignment",
            76: "reserved for future assignment",
            77: "reserved for future assignment",
            78: "reserved for future assignment",
            79: "reserved for future assignment",
            8: "reserved for future assignment",
            80: "reserved for future assignment",
            81: "reserved for future assignment",
            82: "reserved for future assignment",
            83: "reserved for future assignment",
            84: "reserved for future assignment",
            85: "reserved for future assignment",
            86: "reserved for future assignment",
            87: "reserved for future assignment",
            88: "reserved for future assignment",
            89: "reserved for future assignment",
            9: "reserved for future assignment",
            90: "reserved for future assignment",
            91: "reserved for future assignment",
            92: "reserved for future assignment",
            93: "reserved for future assignment",
            94: "reserved for future assignment",
            95: "reserved for future assignment",
            96: "reserved for future assignment",
            97: "reserved for future assignment",
            98: "reserved for future assignment",
            99: "reserved for future assignment"
        },
        3370: {
            0: "off",
            1: "on",
            10: "reserved for future assignment",
            100: "reserved for future assignment",
            101: "reserved for future assignment",
            102: "reserved for future assignment",
            103: "reserved for future assignment",
            104: "reserved for future assignment",
            105: "reserved for future assignment",
            106: "reserved for future assignment",
            107: "reserved for future assignment",
            108: "reserved for future assignment",
            109: "reserved for future assignment",
            11: "reserved for future assignment",
            110: "reserved for future assignment",
            111: "reserved for future assignment",
            112: "reserved for future assignment",
            113: "reserved for future assignment",
            114: "reserved for future assignment",
            115: "reserved for future assignment",
            116: "reserved for future assignment",
            117: "reserved for future assignment",
            118: "reserved for future assignment",
            119: "reserved for future assignment",
            12: "reserved for future assignment",
            120: "reserved for future assignment",
            121: "reserved for future assignment",
            122: "reserved for future assignment",
            123: "reserved for future assignment",
            124: "reserved for future assignment",
            125: "reserved for future assignment",
            126: "reserved for future assignment",
            127: "reserved for future assignment",
            128: "reserved for future assignment",
            129: "reserved for future assignment",
            13: "reserved for future assignment",
            130: "reserved for future assignment",
            131: "reserved for future assignment",
            132: "reserved for future assignment",
            133: "reserved for future assignment",
            134: "reserved for future assignment",
            135: "reserved for future assignment",
            136: "reserved for future assignment",
            137: "reserved for future assignment",
            138: "reserved for future assignment",
            139: "reserved for future assignment",
            14: "reserved for future assignment",
            140: "reserved for future assignment",
            141: "reserved for future assignment",
            142: "reserved for future assignment",
            143: "reserved for future assignment",
            144: "reserved for future assignment",
            145: "reserved for future assignment",
            146: "reserved for future assignment",
            147: "reserved for future assignment",
            148: "reserved for future assignment",
            149: "reserved for future assignment",
            15: "reserved for future assignment",
            150: "reserved for future assignment",
            151: "reserved for future assignment",
            152: "reserved for future assignment",
            153: "reserved for future assignment",
            154: "reserved for future assignment",
            155: "reserved for future assignment",
            156: "reserved for future assignment",
            157: "reserved for future assignment",
            158: "reserved for future assignment",
            159: "reserved for future assignment",
            16: "reserved for future assignment",
            160: "reserved for future assignment",
            161: "reserved for future assignment",
            162: "reserved for future assignment",
            163: "reserved for future assignment",
            164: "reserved for future assignment",
            165: "reserved for future assignment",
            166: "reserved for future assignment",
            167: "reserved for future assignment",
            168: "reserved for future assignment",
            169: "reserved for future assignment",
            17: "reserved for future assignment",
            170: "reserved for future assignment",
            171: "reserved for future assignment",
            172: "reserved for future assignment",
            173: "reserved for future assignment",
            174: "reserved for future assignment",
            175: "reserved for future assignment",
            176: "reserved for future assignment",
            177: "reserved for future assignment",
            178: "reserved for future assignment",
            179: "reserved for future assignment",
            18: "reserved for future assignment",
            180: "reserved for future assignment",
            181: "reserved for future assignment",
            182: "reserved for future assignment",
            183: "reserved for future assignment",
            184: "reserved for future assignment",
            185: "reserved for future assignment",
            186: "reserved for future assignment",
            187: "reserved for future assignment",
            188: "reserved for future assignment",
            189: "reserved for future assignment",
            19: "reserved for future assignment",
            190: "reserved for future assignment",
            191: "reserved for future assignment",
            192: "reserved for future assignment",
            193: "reserved for future assignment",
            194: "reserved for future assignment",
            195: "reserved for future assignment",
            196: "reserved for future assignment",
            197: "reserved for future assignment",
            198: "reserved for future assignment",
            199: "reserved for future assignment",
            2: "reserved for future assignment",
            20: "reserved for future assignment",
            200: "reserved for future assignment",
            201: "reserved for future assignment",
            202: "reserved for future assignment",
            203: "reserved for future assignment",
            204: "reserved for future assignment",
            205: "reserved for future assignment",
            206: "reserved for future assignment",
            207: "reserved for future assignment",
            208: "reserved for future assignment",
            209: "reserved for future assignment",
            21: "reserved for future assignment",
            210: "reserved for future assignment",
            211: "reserved for future assignment",
            212: "reserved for future assignment",
            213: "reserved for future assignment",
            214: "reserved for future assignment",
            215: "reserved for future assignment",
            216: "reserved for future assignment",
            217: "reserved for future assignment",
            218: "reserved for future assignment",
            219: "reserved for future assignment",
            22: "reserved for future assignment",
            220: "reserved for future assignment",
            221: "reserved for future assignment",
            222: "reserved for future assignment",
            223: "reserved for future assignment",
            224: "reserved for future assignment",
            225: "reserved for future assignment",
            226: "reserved for future assignment",
            227: "reserved for future assignment",
            228: "reserved for future assignment",
            229: "reserved for future assignment",
            23: "reserved for future assignment",
            230: "reserved for future assignment",
            231: "reserved for future assignment",
            232: "reserved for future assignment",
            233: "reserved for future assignment",
            234: "reserved for future assignment",
            235: "reserved for future assignment",
            236: "reserved for future assignment",
            237: "reserved for future assignment",
            238: "reserved for future assignment",
            239: "reserved for future assignment",
            24: "reserved for future assignment",
            240: "reserved for future assignment",
            241: "reserved for future assignment",
            242: "reserved for future assignment",
            243: "reserved for future assignment",
            244: "reserved for future assignment",
            245: "reserved for future assignment",
            246: "reserved for future assignment",
            247: "reserved for future assignment",
            248: "reserved for future assignment",
            249: "reserved for future assignment",
            25: "reserved for future assignment",
            250: "reserved for future assignment",
            251: "reserved",
            252: "reserved",
            253: "reserved",
            254: "error",
            255: "not available",
            26: "reserved for future assignment",
            27: "reserved for future assignment",
            28: "reserved for future assignment",
            29: "reserved for future assignment",
            3: "reserved for future assignment",
            30: "reserved for future assignment",
            31: "reserved for future assignment",
            32: "reserved for future assignment",
            33: "reserved for future assignment",
            34: "reserved for future assignment",
            35: "reserved for future assignment",
            36: "reserved for future assignment",
            37: "reserved for future assignment",
            38: "reserved for future assignment",
            39: "reserved for future assignment",
            4: "reserved for future assignment",
            40: "reserved for future assignment",
            41: "reserved for future assignment",
            42: "reserved for future assignment",
            43: "reserved for future assignment",
            44: "reserved for future assignment",
            45: "reserved for future assignment",
            46: "reserved for future assignment",
            47: "reserved for future assignment",
            48: "reserved for future assignment",
            49: "reserved for future assignment",
            5: "reserved for future assignment",
            50: "reserved for future assignment",
            51: "reserved for future assignment",
            52: "reserved for future assignment",
            53: "reserved for future assignment",
            54: "reserved for future assignment",
            55: "reserved for future assignment",
            56: "reserved for future assignment",
            57: "reserved for future assignment",
            58: "reserved for future assignment",
            59: "reserved for future assignment",
            6: "reserved for future assignment",
            60: "reserved for future assignment",
            61: "reserved for future assignment",
            62: "reserved for future assignment",
            63: "reserved for future assignment",
            64: "reserved for future assignment",
            65: "reserved for future assignment",
            66: "reserved for future assignment",
            67: "reserved for future assignment",
            68: "reserved for future assignment",
            69: "reserved for future assignment",
            7: "reserved for future assignment",
            70: "reserved for future assignment",
            71: "reserved for future assignment",
            72: "reserved for future assignment",
            73: "reserved for future assignment",
            74: "reserved for future assignment",
            75: "reserved for future assignment",
            76: "reserved for future assignment",
            77: "reserved for future assignment",
            78: "reserved for future assignment",
            79: "reserved for future assignment",
            8: "reserved for future assignment",
            80: "reserved for future assignment",
            81: "reserved for future assignment",
            82: "reserved for future assignment",
            83: "reserved for future assignment",
            84: "reserved for future assignment",
            85: "reserved for future assignment",
            86: "reserved for future assignment",
            87: "reserved for future assignment",
            88: "reserved for future assignment",
            89: "reserved for future assignment",
            9: "reserved for future assignment",
            90: "reserved for future assignment",
            91: "reserved for future assignment",
            92: "reserved for future assignment",
            93: "reserved for future assignment",
            94: "reserved for future assignment",
            95: "reserved for future assignment",
            96: "reserved for future assignment",
            97: "reserved for future assignment",
            98: "reserved for future assignment",
            99: "reserved for future assignment"
        },
        3372: {
            0: "cellular network #1",
            1: "cellular network #2",
            10: "reserved for future assignment",
            100: "reserved for future assignment",
            101: "reserved for future assignment",
            102: "reserved for future assignment",
            103: "reserved for future assignment",
            104: "reserved for future assignment",
            105: "reserved for future assignment",
            106: "reserved for future assignment",
            107: "reserved for future assignment",
            108: "reserved for future assignment",
            109: "reserved for future assignment",
            11: "reserved for future assignment",
            110: "reserved for future assignment",
            111: "reserved for future assignment",
            112: "reserved for future assignment",
            113: "reserved for future assignment",
            114: "reserved for future assignment",
            115: "reserved for future assignment",
            116: "reserved for future assignment",
            117: "reserved for future assignment",
            118: "reserved for future assignment",
            119: "reserved for future assignment",
            12: "reserved for future assignment",
            120: "reserved for future assignment",
            121: "reserved for future assignment",
            122: "reserved for future assignment",
            123: "reserved for future assignment",
            124: "reserved for future assignment",
            125: "reserved for future assignment",
            126: "reserved for future assignment",
            127: "reserved for future assignment",
            128: "reserved for future assignment",
            129: "reserved for future assignment",
            13: "reserved for future assignment",
            130: "reserved for future assignment",
            131: "reserved for future assignment",
            132: "reserved for future assignment",
            133: "reserved for future assignment",
            134: "reserved for future assignment",
            135: "reserved for future assignment",
            136: "reserved for future assignment",
            137: "reserved for future assignment",
            138: "reserved for future assignment",
            139: "reserved for future assignment",
            14: "reserved for future assignment",
            140: "reserved for future assignment",
            141: "reserved for future assignment",
            142: "reserved for future assignment",
            143: "reserved for future assignment",
            144: "reserved for future assignment",
            145: "reserved for future assignment",
            146: "reserved for future assignment",
            147: "reserved for future assignment",
            148: "reserved for future assignment",
            149: "reserved for future assignment",
            15: "reserved for future assignment",
            150: "reserved for future assignment",
            151: "reserved for future assignment",
            152: "reserved for future assignment",
            153: "reserved for future assignment",
            154: "reserved for future assignment",
            155: "reserved for future assignment",
            156: "reserved for future assignment",
            157: "reserved for future assignment",
            158: "reserved for future assignment",
            159: "reserved for future assignment",
            16: "reserved for future assignment",
            160: "reserved for future assignment",
            161: "reserved for future assignment",
            162: "reserved for future assignment",
            163: "reserved for future assignment",
            164: "reserved for future assignment",
            165: "reserved for future assignment",
            166: "reserved for future assignment",
            167: "reserved for future assignment",
            168: "reserved for future assignment",
            169: "reserved for future assignment",
            17: "reserved for future assignment",
            170: "reserved for future assignment",
            171: "reserved for future assignment",
            172: "reserved for future assignment",
            173: "reserved for future assignment",
            174: "reserved for future assignment",
            175: "reserved for future assignment",
            176: "reserved for future assignment",
            177: "reserved for future assignment",
            178: "reserved for future assignment",
            179: "reserved for future assignment",
            18: "reserved for future assignment",
            180: "reserved for future assignment",
            181: "reserved for future assignment",
            182: "reserved for future assignment",
            183: "reserved for future assignment",
            184: "reserved for future assignment",
            185: "reserved for future assignment",
            186: "reserved for future assignment",
            187: "reserved for future assignment",
            188: "reserved for future assignment",
            189: "reserved for future assignment",
            19: "reserved for future assignment",
            190: "reserved for future assignment",
            191: "reserved for future assignment",
            192: "reserved for future assignment",
            193: "reserved for future assignment",
            194: "reserved for future assignment",
            195: "reserved for future assignment",
            196: "reserved for future assignment",
            197: "reserved for future assignment",
            198: "reserved for future assignment",
            199: "reserved for future assignment",
            2: "wireless lan #1",
            20: "reserved for future assignment",
            200: "reserved for future assignment",
            201: "reserved for future assignment",
            202: "reserved for future assignment",
            203: "reserved for future assignment",
            204: "reserved for future assignment",
            205: "reserved for future assignment",
            206: "reserved for future assignment",
            207: "reserved for future assignment",
            208: "reserved for future assignment",
            209: "reserved for future assignment",
            21: "reserved for future assignment",
            210: "reserved for future assignment",
            211: "reserved for future assignment",
            212: "reserved for future assignment",
            213: "reserved for future assignment",
            214: "reserved for future assignment",
            215: "reserved for future assignment",
            216: "reserved for future assignment",
            217: "reserved for future assignment",
            218: "reserved for future assignment",
            219: "reserved for future assignment",
            22: "reserved for future assignment",
            220: "reserved for future assignment",
            221: "reserved for future assignment",
            222: "reserved for future assignment",
            223: "reserved for future assignment",
            224: "reserved for future assignment",
            225: "reserved for future assignment",
            226: "reserved for future assignment",
            227: "reserved for future assignment",
            228: "reserved for future assignment",
            229: "reserved for future assignment",
            23: "reserved for future assignment",
            230: "reserved for future assignment",
            231: "reserved for future assignment",
            232: "reserved for future assignment",
            233: "reserved for future assignment",
            234: "reserved for future assignment",
            235: "reserved for future assignment",
            236: "reserved for future assignment",
            237: "reserved for future assignment",
            238: "reserved for future assignment",
            239: "reserved for future assignment",
            24: "reserved for future assignment",
            240: "reserved for future assignment",
            241: "reserved for future assignment",
            242: "reserved for future assignment",
            243: "reserved for future assignment",
            244: "reserved for future assignment",
            245: "reserved for future assignment",
            246: "reserved for future assignment",
            247: "reserved for future assignment",
            248: "reserved for future assignment",
            249: "reserved for future assignment",
            25: "reserved for future assignment",
            250: "reserved for future assignment",
            251: "reserved",
            252: "reserved",
            253: "reserved",
            254: "error",
            255: "not available",
            26: "reserved for future assignment",
            27: "reserved for future assignment",
            28: "reserved for future assignment",
            29: "reserved for future assignment",
            3: "wireless lan #2",
            30: "reserved for future assignment",
            31: "reserved for future assignment",
            32: "reserved for future assignment",
            33: "reserved for future assignment",
            34: "reserved for future assignment",
            35: "reserved for future assignment",
            36: "reserved for future assignment",
            37: "reserved for future assignment",
            38: "reserved for future assignment",
            39: "reserved for future assignment",
            4: "satellite network",
            40: "reserved for future assignment",
            41: "reserved for future assignment",
            42: "reserved for future assignment",
            43: "reserved for future assignment",
            44: "reserved for future assignment",
            45: "reserved for future assignment",
            46: "reserved for future assignment",
            47: "reserved for future assignment",
            48: "reserved for future assignment",
            49: "reserved for future assignment",
            5: "reserved for future assignment",
            50: "reserved for future assignment",
            51: "reserved for future assignment",
            52: "reserved for future assignment",
            53: "reserved for future assignment",
            54: "reserved for future assignment",
            55: "reserved for future assignment",
            56: "reserved for future assignment",
            57: "reserved for future assignment",
            58: "reserved for future assignment",
            59: "reserved for future assignment",
            6: "reserved for future assignment",
            60: "reserved for future assignment",
            61: "reserved for future assignment",
            62: "reserved for future assignment",
            63: "reserved for future assignment",
            64: "reserved for future assignment",
            65: "reserved for future assignment",
            66: "reserved for future assignment",
            67: "reserved for future assignment",
            68: "reserved for future assignment",
            69: "reserved for future assignment",
            7: "reserved for future assignment",
            70: "reserved for future assignment",
            71: "reserved for future assignment",
            72: "reserved for future assignment",
            73: "reserved for future assignment",
            74: "reserved for future assignment",
            75: "reserved for future assignment",
            76: "reserved for future assignment",
            77: "reserved for future assignment",
            78: "reserved for future assignment",
            79: "reserved for future assignment",
            8: "reserved for future assignment",
            80: "reserved for future assignment",
            81: "reserved for future assignment",
            82: "reserved for future assignment",
            83: "reserved for future assignment",
            84: "reserved for future assignment",
            85: "reserved for future assignment",
            86: "reserved for future assignment",
            87: "reserved for future assignment",
            88: "reserved for future assignment",
            89: "reserved for future assignment",
            9: "reserved for future assignment",
            90: "reserved for future assignment",
            91: "reserved for future assignment",
            92: "reserved for future assignment",
            93: "reserved for future assignment",
            94: "reserved for future assignment",
            95: "reserved for future assignment",
            96: "reserved for future assignment",
            97: "reserved for future assignment",
            98: "reserved for future assignment",
            99: "reserved for future assignment"
        },
        3375: {
            0: "load compensation disabled",
            1: "voltage droop compensation enabled",
            2: "ir line drop compensation enabled",
            3: "reserved",
            4: "reserved",
            5: "reserved",
            6: "reserved",
            7: "not available"
        },
        3376: {
            0: "var/power factor regulation disabled",
            1: "power factor regulation enabled",
            2: "var regulation enabled",
            3: "reserved",
            4: "reserved",
            5: "reserved",
            6: "reserved",
            7: "not available"
        },
        3377: {
            0: "underfrequency compensation disabled",
            1: "underfrequency compensation enabled",
            2: "reserved",
            3: "not available"
        },
        3378: {
            0: "soft start is inactive",
            1: "soft start is active",
            2: "reserved",
            3: "not available"
        },
        3379: {
            0: "not regulating voltage - unit disabled",
            1: "regulating voltage",
            2: "reserved",
            3: "not available"
        },
        3385: {
            0: "leading",
            1: "lagging",
            2: "reserved",
            3: "not available or not installed"
        },
        3387: {
            0: "        no combustion",
            1: "        combustion exists",
            2: "        error",
            3: "        not available"
        },
        3388: {
            0: "        no combustion",
            1: "        combustion exists",
            2: "        error",
            3: "        not available"
        },
        3389: {
            0: "        no combustion",
            1: "        combustion exists",
            2: "        error",
            3: "        not available"
        },
        3390: {
            0: "        no combustion",
            1: "        combustion exists",
            2: "        error",
            3: "        not available"
        },
        3391: {
            0: "        no combustion",
            1: "        combustion exists",
            2: "        error",
            3: "        not available"
        },
        3392: {
            0: "        no combustion",
            1: "        combustion exists",
            2: "        error",
            3: "        not available"
        },
        3393: {
            0: "        no combustion",
            1: "        combustion exists",
            2: "        error",
            3: "        not available"
        },
        3394: {
            0: "        no combustion",
            1: "        combustion exists",
            2: "        error",
            3: "        not available"
        },
        3395: {
            0: "        no combustion",
            1: "        combustion exists",
            2: "        error",
            3: "        not available"
        },
        3396: {
            0: "        no combustion",
            1: "        combustion exists",
            2: "        error",
            3: "        not available"
        },
        3397: {
            0: "        no combustion",
            1: "        combustion exists",
            2: "        error",
            3: "        not available"
        },
        3398: {
            0: "        no combustion",
            1: "        combustion exists",
            2: "        error",
            3: "        not available"
        },
        3399: {
            0: "        no combustion",
            1: "        combustion exists",
            2: "        error",
            3: "        not available"
        },
        3400: {
            0: "        no combustion",
            1: "        combustion exists",
            2: "        error",
            3: "        not available"
        },
        3401: {
            0: "        no combustion",
            1: "        combustion exists",
            2: "        error",
            3: "        not available"
        },
        3402: {
            0: "        no combustion",
            1: "        combustion exists",
            2: "        error",
            3: "        not available"
        },
        3403: {
            0: "        no combustion",
            1: "        combustion exists",
            2: "        error",
            3: "        not available"
        },
        3404: {
            0: "        no combustion",
            1: "        combustion exists",
            2: "        error",
            3: "        not available"
        },
        3405: {
            0: "        no combustion",
            1: "        combustion exists",
            2: "        error",
            3: "        not available"
        },
        3406: {
            0: "        no combustion",
            1: "        combustion exists",
            2: "        error",
            3: "        not available"
        },
        3407: {
            0: "        no combustion",
            1: "        combustion exists",
            2: "        error",
            3: "        not available"
        },
        3408: {
            0: "        no combustion",
            1: "        combustion exists",
            2: "        error",
            3: "        not available"
        },
        3409: {
            0: "        no combustion",
            1: "        combustion exists",
            2: "        error",
            3: "        not available"
        },
        3410: {
            0: "        no combustion",
            1: "        combustion exists",
            2: "        error",
            3: "        not available"
        },
        3411: {
            0: "   all bus doors are disabled",
            1: "   at least 1 bus door enabled ",
            2: "   error",
            3: "   not available"
        },
        3412: {
            0: "  unlocked ",
            1: "  locked",
            2: "  error",
            3: "  not available"
        },
        3413: {
            0: "  door closed",
            1: "  door opened",
            2: "  error",
            3: "  not available"
        },
        3414: {
            0: "  door disabled",
            1: "  door enabled",
            2: "  error",
            3: "  not available"
        },
        3415: {
            0: "  unlocked ",
            1: "  locked",
            2: "  error",
            3: "  not available"
        },
        3416: {
            0: "  door closed",
            1: "  door opened",
            2: "  error",
            3: "  not available"
        },
        3417: {
            0: "  door disabled",
            1: "  door enabled",
            2: "  error",
            3: "  not available"
        },
        3418: {
            0: "  unlocked ",
            1: "  locked",
            2: "  error",
            3: "  not available"
        },
        3419: {
            0: "  door closed",
            1: "  door opened",
            2: "  error",
            3: "  not available"
        },
        3420: {
            0: "  door disabled",
            1: "  door enabled",
            2: "  error",
            3: "  not available"
        },
        3421: {
            0: "  unlocked ",
            1: "  locked",
            2: "  error",
            3: "  not available"
        },
        3422: {
            0: "  door closed",
            1: "  door opened",
            2: "  error",
            3: "  not available"
        },
        3423: {
            0: "  door disabled",
            1: "  door enabled",
            2: "  error",
            3: "  not available"
        },
        3424: {
            0: "  unlocked ",
            1: "  locked",
            2: "  error",
            3: "  not available"
        },
        3425: {
            0: "  door closed",
            1: "  door opened",
            2: "  error",
            3: "  not available"
        },
        3426: {
            0: "  door disabled",
            1: "  door enabled",
            2: "  error",
            3: "  not available"
        },
        3427: {
            0: "  unlocked ",
            1: "  locked",
            2: "  error",
            3: "  not available"
        },
        3428: {
            0: "  door closed",
            1: "  door opened",
            2: "  error",
            3: "  not available"
        },
        3429: {
            0: "  door disabled",
            1: "  door enabled",
            2: "  error",
            3: "  not available"
        },
        3430: {
            0: "  unlocked ",
            1: "  locked",
            2: "  error",
            3: "  not available"
        },
        3431: {
            0: "  door closed",
            1: "  door opened",
            2: "  error",
            3: "  not available"
        },
        3432: {
            0: "  door disabled",
            1: "  door enabled",
            2: "  error",
            3: "  not available"
        },
        3433: {
            0: "  unlocked ",
            1: "  locked",
            2: "  error",
            3: "  not available"
        },
        3434: {
            0: "  door closed",
            1: "  door opened",
            2: "  error",
            3: "  not available"
        },
        3435: {
            0: "  door disabled",
            1: "  door enabled",
            2: "  error",
            3: "  not available"
        },
        3436: {
            0: "  unlocked ",
            1: "  locked",
            2: "  error",
            3: "  not available"
        },
        3437: {
            0: "  door closed",
            1: "  door opened",
            2: "  error",
            3: "  not available"
        },
        3438: {
            0: "  door disabled",
            1: "  door enabled",
            2: "  error",
            3: "  not available"
        },
        3439: {
            0: "  unlocked ",
            1: "  locked",
            2: "  error",
            3: "  not available"
        },
        3440: {
            0: "  door closed",
            1: "  door opened",
            2: "  error",
            3: "  not available"
        },
        3441: {
            0: "  door disabled",
            1: "  door enabled",
            2: "  error",
            3: "  not available"
        },
        3442: {
            0: "off",
            1: "on",
            10: "reserved for future assignment",
            100: "reserved for future assignment",
            101: "reserved for future assignment",
            102: "reserved for future assignment",
            103: "reserved for future assignment",
            104: "reserved for future assignment",
            105: "reserved for future assignment",
            106: "reserved for future assignment",
            107: "reserved for future assignment",
            108: "reserved for future assignment",
            109: "reserved for future assignment",
            11: "reserved for future assignment",
            110: "reserved for future assignment",
            111: "reserved for future assignment",
            112: "reserved for future assignment",
            113: "reserved for future assignment",
            114: "reserved for future assignment",
            115: "reserved for future assignment",
            116: "reserved for future assignment",
            117: "reserved for future assignment",
            118: "reserved for future assignment",
            119: "reserved for future assignment",
            12: "reserved for future assignment",
            120: "reserved for future assignment",
            121: "reserved for future assignment",
            122: "reserved for future assignment",
            123: "reserved for future assignment",
            124: "reserved for future assignment",
            125: "reserved for future assignment",
            126: "reserved for future assignment",
            127: "reserved for future assignment",
            128: "reserved for future assignment",
            129: "reserved for future assignment",
            13: "reserved for future assignment",
            130: "reserved for future assignment",
            131: "reserved for future assignment",
            132: "reserved for future assignment",
            133: "reserved for future assignment",
            134: "reserved for future assignment",
            135: "reserved for future assignment",
            136: "reserved for future assignment",
            137: "reserved for future assignment",
            138: "reserved for future assignment",
            139: "reserved for future assignment",
            14: "reserved for future assignment",
            140: "reserved for future assignment",
            141: "reserved for future assignment",
            142: "reserved for future assignment",
            143: "reserved for future assignment",
            144: "reserved for future assignment",
            145: "reserved for future assignment",
            146: "reserved for future assignment",
            147: "reserved for future assignment",
            148: "reserved for future assignment",
            149: "reserved for future assignment",
            15: "reserved for future assignment",
            150: "reserved for future assignment",
            151: "reserved for future assignment",
            152: "reserved for future assignment",
            153: "reserved for future assignment",
            154: "reserved for future assignment",
            155: "reserved for future assignment",
            156: "reserved for future assignment",
            157: "reserved for future assignment",
            158: "reserved for future assignment",
            159: "reserved for future assignment",
            16: "reserved for future assignment",
            160: "reserved for future assignment",
            161: "reserved for future assignment",
            162: "reserved for future assignment",
            163: "reserved for future assignment",
            164: "reserved for future assignment",
            165: "reserved for future assignment",
            166: "reserved for future assignment",
            167: "reserved for future assignment",
            168: "reserved for future assignment",
            169: "reserved for future assignment",
            17: "reserved for future assignment",
            170: "reserved for future assignment",
            171: "reserved for future assignment",
            172: "reserved for future assignment",
            173: "reserved for future assignment",
            174: "reserved for future assignment",
            175: "reserved for future assignment",
            176: "reserved for future assignment",
            177: "reserved for future assignment",
            178: "reserved for future assignment",
            179: "reserved for future assignment",
            18: "reserved for future assignment",
            180: "reserved for future assignment",
            181: "reserved for future assignment",
            182: "reserved for future assignment",
            183: "reserved for future assignment",
            184: "reserved for future assignment",
            185: "reserved for future assignment",
            186: "reserved for future assignment",
            187: "reserved for future assignment",
            188: "reserved for future assignment",
            189: "reserved for future assignment",
            19: "reserved for future assignment",
            190: "reserved for future assignment",
            191: "reserved for future assignment",
            192: "reserved for future assignment",
            193: "reserved for future assignment",
            194: "reserved for future assignment",
            195: "reserved for future assignment",
            196: "reserved for future assignment",
            197: "reserved for future assignment",
            198: "reserved for future assignment",
            199: "reserved for future assignment",
            2: "not authorized to operate on network",
            20: "reserved for future assignment",
            200: "reserved for future assignment",
            201: "reserved for future assignment",
            202: "reserved for future assignment",
            203: "reserved for future assignment",
            204: "reserved for future assignment",
            205: "reserved for future assignment",
            206: "reserved for future assignment",
            207: "reserved for future assignment",
            208: "reserved for future assignment",
            209: "reserved for future assignment",
            21: "reserved for future assignment",
            210: "reserved for future assignment",
            211: "reserved for future assignment",
            212: "reserved for future assignment",
            213: "reserved for future assignment",
            214: "reserved for future assignment",
            215: "reserved for future assignment",
            216: "reserved for future assignment",
            217: "reserved for future assignment",
            218: "reserved for future assignment",
            219: "reserved for future assignment",
            22: "reserved for future assignment",
            220: "reserved for future assignment",
            221: "reserved for future assignment",
            222: "reserved for future assignment",
            223: "reserved for future assignment",
            224: "reserved for future assignment",
            225: "reserved for future assignment",
            226: "reserved for future assignment",
            227: "reserved for future assignment",
            228: "reserved for future assignment",
            229: "reserved for future assignment",
            23: "reserved for future assignment",
            230: "reserved for future assignment",
            231: "reserved for future assignment",
            232: "reserved for future assignment",
            233: "reserved for future assignment",
            234: "reserved for future assignment",
            235: "reserved for future assignment",
            236: "reserved for future assignment",
            237: "reserved for future assignment",
            238: "reserved for future assignment",
            239: "reserved for future assignment",
            24: "reserved for future assignment",
            240: "reserved for future assignment",
            241: "reserved for future assignment",
            242: "reserved for future assignment",
            243: "reserved for future assignment",
            244: "reserved for future assignment",
            245: "reserved for future assignment",
            246: "reserved for future assignment",
            247: "reserved for future assignment",
            248: "reserved for future assignment",
            249: "reserved for future assignment",
            25: "reserved for future assignment",
            250: "reserved for future assignment",
            251: "reserved",
            252: "reserved",
            253: "reserved",
            254: "error",
            255: "not available",
            26: "reserved for future assignment",
            27: "reserved for future assignment",
            28: "reserved for future assignment",
            29: "reserved for future assignment",
            3: "reserved for future assignment",
            30: "reserved for future assignment",
            31: "reserved for future assignment",
            32: "reserved for future assignment",
            33: "reserved for future assignment",
            34: "reserved for future assignment",
            35: "reserved for future assignment",
            36: "reserved for future assignment",
            37: "reserved for future assignment",
            38: "reserved for future assignment",
            39: "reserved for future assignment",
            4: "reserved for future assignment",
            40: "reserved for future assignment",
            41: "reserved for future assignment",
            42: "reserved for future assignment",
            43: "reserved for future assignment",
            44: "reserved for future assignment",
            45: "reserved for future assignment",
            46: "reserved for future assignment",
            47: "reserved for future assignment",
            48: "reserved for future assignment",
            49: "reserved for future assignment",
            5: "reserved for future assignment",
            50: "reserved for future assignment",
            51: "reserved for future assignment",
            52: "reserved for future assignment",
            53: "reserved for future assignment",
            54: "reserved for future assignment",
            55: "reserved for future assignment",
            56: "reserved for future assignment",
            57: "reserved for future assignment",
            58: "reserved for future assignment",
            59: "reserved for future assignment",
            6: "reserved for future assignment",
            60: "reserved for future assignment",
            61: "reserved for future assignment",
            62: "reserved for future assignment",
            63: "reserved for future assignment",
            64: "reserved for future assignment",
            65: "reserved for future assignment",
            66: "reserved for future assignment",
            67: "reserved for future assignment",
            68: "reserved for future assignment",
            69: "reserved for future assignment",
            7: "reserved for future assignment",
            70: "reserved for future assignment",
            71: "reserved for future assignment",
            72: "reserved for future assignment",
            73: "reserved for future assignment",
            74: "reserved for future assignment",
            75: "reserved for future assignment",
            76: "reserved for future assignment",
            77: "reserved for future assignment",
            78: "reserved for future assignment",
            79: "reserved for future assignment",
            8: "reserved for future assignment",
            80: "reserved for future assignment",
            81: "reserved for future assignment",
            82: "reserved for future assignment",
            83: "reserved for future assignment",
            84: "reserved for future assignment",
            85: "reserved for future assignment",
            86: "reserved for future assignment",
            87: "reserved for future assignment",
            88: "reserved for future assignment",
            89: "reserved for future assignment",
            9: "reserved for future assignment",
            90: "reserved for future assignment",
            91: "reserved for future assignment",
            92: "reserved for future assignment",
            93: "reserved for future assignment",
            94: "reserved for future assignment",
            95: "reserved for future assignment",
            96: "reserved for future assignment",
            97: "reserved for future assignment",
            98: "reserved for future assignment",
            99: "reserved for future assignment"
        },
        3443: {
            0: "local service",
            1: "roaming service",
            10: "reserved for future assignment",
            100: "reserved for future assignment",
            101: "reserved for future assignment",
            102: "reserved for future assignment",
            103: "reserved for future assignment",
            104: "reserved for future assignment",
            105: "reserved for future assignment",
            106: "reserved for future assignment",
            107: "reserved for future assignment",
            108: "reserved for future assignment",
            109: "reserved for future assignment",
            11: "reserved for future assignment",
            110: "reserved for future assignment",
            111: "reserved for future assignment",
            112: "reserved for future assignment",
            113: "reserved for future assignment",
            114: "reserved for future assignment",
            115: "reserved for future assignment",
            116: "reserved for future assignment",
            117: "reserved for future assignment",
            118: "reserved for future assignment",
            119: "reserved for future assignment",
            12: "reserved for future assignment",
            120: "reserved for future assignment",
            121: "reserved for future assignment",
            122: "reserved for future assignment",
            123: "reserved for future assignment",
            124: "reserved for future assignment",
            125: "reserved for future assignment",
            126: "reserved for future assignment",
            127: "reserved for future assignment",
            128: "reserved for future assignment",
            129: "reserved for future assignment",
            13: "reserved for future assignment",
            130: "reserved for future assignment",
            131: "reserved for future assignment",
            132: "reserved for future assignment",
            133: "reserved for future assignment",
            134: "reserved for future assignment",
            135: "reserved for future assignment",
            136: "reserved for future assignment",
            137: "reserved for future assignment",
            138: "reserved for future assignment",
            139: "reserved for future assignment",
            14: "reserved for future assignment",
            140: "reserved for future assignment",
            141: "reserved for future assignment",
            142: "reserved for future assignment",
            143: "reserved for future assignment",
            144: "reserved for future assignment",
            145: "reserved for future assignment",
            146: "reserved for future assignment",
            147: "reserved for future assignment",
            148: "reserved for future assignment",
            149: "reserved for future assignment",
            15: "reserved for future assignment",
            150: "reserved for future assignment",
            151: "reserved for future assignment",
            152: "reserved for future assignment",
            153: "reserved for future assignment",
            154: "reserved for future assignment",
            155: "reserved for future assignment",
            156: "reserved for future assignment",
            157: "reserved for future assignment",
            158: "reserved for future assignment",
            159: "reserved for future assignment",
            16: "reserved for future assignment",
            160: "reserved for future assignment",
            161: "reserved for future assignment",
            162: "reserved for future assignment",
            163: "reserved for future assignment",
            164: "reserved for future assignment",
            165: "reserved for future assignment",
            166: "reserved for future assignment",
            167: "reserved for future assignment",
            168: "reserved for future assignment",
            169: "reserved for future assignment",
            17: "reserved for future assignment",
            170: "reserved for future assignment",
            171: "reserved for future assignment",
            172: "reserved for future assignment",
            173: "reserved for future assignment",
            174: "reserved for future assignment",
            175: "reserved for future assignment",
            176: "reserved for future assignment",
            177: "reserved for future assignment",
            178: "reserved for future assignment",
            179: "reserved for future assignment",
            18: "reserved for future assignment",
            180: "reserved for future assignment",
            181: "reserved for future assignment",
            182: "reserved for future assignment",
            183: "reserved for future assignment",
            184: "reserved for future assignment",
            185: "reserved for future assignment",
            186: "reserved for future assignment",
            187: "reserved for future assignment",
            188: "reserved for future assignment",
            189: "reserved for future assignment",
            19: "reserved for future assignment",
            190: "reserved for future assignment",
            191: "reserved for future assignment",
            192: "reserved for future assignment",
            193: "reserved for future assignment",
            194: "reserved for future assignment",
            195: "reserved for future assignment",
            196: "reserved for future assignment",
            197: "reserved for future assignment",
            198: "reserved for future assignment",
            199: "reserved for future assignment",
            2: "service not available",
            20: "reserved for future assignment",
            200: "reserved for future assignment",
            201: "reserved for future assignment",
            202: "reserved for future assignment",
            203: "reserved for future assignment",
            204: "reserved for future assignment",
            205: "reserved for future assignment",
            206: "reserved for future assignment",
            207: "reserved for future assignment",
            208: "reserved for future assignment",
            209: "reserved for future assignment",
            21: "reserved for future assignment",
            210: "reserved for future assignment",
            211: "reserved for future assignment",
            212: "reserved for future assignment",
            213: "reserved for future assignment",
            214: "reserved for future assignment",
            215: "reserved for future assignment",
            216: "reserved for future assignment",
            217: "reserved for future assignment",
            218: "reserved for future assignment",
            219: "reserved for future assignment",
            22: "reserved for future assignment",
            220: "reserved for future assignment",
            221: "reserved for future assignment",
            222: "reserved for future assignment",
            223: "reserved for future assignment",
            224: "reserved for future assignment",
            225: "reserved for future assignment",
            226: "reserved for future assignment",
            227: "reserved for future assignment",
            228: "reserved for future assignment",
            229: "reserved for future assignment",
            23: "reserved for future assignment",
            230: "reserved for future assignment",
            231: "reserved for future assignment",
            232: "reserved for future assignment",
            233: "reserved for future assignment",
            234: "reserved for future assignment",
            235: "reserved for future assignment",
            236: "reserved for future assignment",
            237: "reserved for future assignment",
            238: "reserved for future assignment",
            239: "reserved for future assignment",
            24: "reserved for future assignment",
            240: "reserved for future assignment",
            241: "reserved for future assignment",
            242: "reserved for future assignment",
            243: "reserved for future assignment",
            244: "reserved for future assignment",
            245: "reserved for future assignment",
            246: "reserved for future assignment",
            247: "reserved for future assignment",
            248: "reserved for future assignment",
            249: "reserved for future assignment",
            25: "reserved for future assignment",
            250: "reserved for future assignment",
            251: "reserved",
            252: "reserved",
            253: "reserved",
            254: "error",
            255: "not available",
            26: "reserved for future assignment",
            27: "reserved for future assignment",
            28: "reserved for future assignment",
            29: "reserved for future assignment",
            3: "not authorized to operate on service",
            30: "reserved for future assignment",
            31: "reserved for future assignment",
            32: "reserved for future assignment",
            33: "reserved for future assignment",
            34: "reserved for future assignment",
            35: "reserved for future assignment",
            36: "reserved for future assignment",
            37: "reserved for future assignment",
            38: "reserved for future assignment",
            39: "reserved for future assignment",
            4: "reserved for future assignment",
            40: "reserved for future assignment",
            41: "reserved for future assignment",
            42: "reserved for future assignment",
            43: "reserved for future assignment",
            44: "reserved for future assignment",
            45: "reserved for future assignment",
            46: "reserved for future assignment",
            47: "reserved for future assignment",
            48: "reserved for future assignment",
            49: "reserved for future assignment",
            5: "reserved for future assignment",
            50: "reserved for future assignment",
            51: "reserved for future assignment",
            52: "reserved for future assignment",
            53: "reserved for future assignment",
            54: "reserved for future assignment",
            55: "reserved for future assignment",
            56: "reserved for future assignment",
            57: "reserved for future assignment",
            58: "reserved for future assignment",
            59: "reserved for future assignment",
            6: "reserved for future assignment",
            60: "reserved for future assignment",
            61: "reserved for future assignment",
            62: "reserved for future assignment",
            63: "reserved for future assignment",
            64: "reserved for future assignment",
            65: "reserved for future assignment",
            66: "reserved for future assignment",
            67: "reserved for future assignment",
            68: "reserved for future assignment",
            69: "reserved for future assignment",
            7: "reserved for future assignment",
            70: "reserved for future assignment",
            71: "reserved for future assignment",
            72: "reserved for future assignment",
            73: "reserved for future assignment",
            74: "reserved for future assignment",
            75: "reserved for future assignment",
            76: "reserved for future assignment",
            77: "reserved for future assignment",
            78: "reserved for future assignment",
            79: "reserved for future assignment",
            8: "reserved for future assignment",
            80: "reserved for future assignment",
            81: "reserved for future assignment",
            82: "reserved for future assignment",
            83: "reserved for future assignment",
            84: "reserved for future assignment",
            85: "reserved for future assignment",
            86: "reserved for future assignment",
            87: "reserved for future assignment",
            88: "reserved for future assignment",
            89: "reserved for future assignment",
            9: "reserved for future assignment",
            90: "reserved for future assignment",
            91: "reserved for future assignment",
            92: "reserved for future assignment",
            93: "reserved for future assignment",
            94: "reserved for future assignment",
            95: "reserved for future assignment",
            96: "reserved for future assignment",
            97: "reserved for future assignment",
            98: "reserved for future assignment",
            99: "reserved for future assignment"
        },
        3444: {
            0: "off",
            1: "on",
            10: "reserved for future assignment",
            100: "reserved for future assignment",
            101: "reserved for future assignment",
            102: "reserved for future assignment",
            103: "reserved for future assignment",
            104: "reserved for future assignment",
            105: "reserved for future assignment",
            106: "reserved for future assignment",
            107: "reserved for future assignment",
            108: "reserved for future assignment",
            109: "reserved for future assignment",
            11: "reserved for future assignment",
            110: "reserved for future assignment",
            111: "reserved for future assignment",
            112: "reserved for future assignment",
            113: "reserved for future assignment",
            114: "reserved for future assignment",
            115: "reserved for future assignment",
            116: "reserved for future assignment",
            117: "reserved for future assignment",
            118: "reserved for future assignment",
            119: "reserved for future assignment",
            12: "reserved for future assignment",
            120: "reserved for future assignment",
            121: "reserved for future assignment",
            122: "reserved for future assignment",
            123: "reserved for future assignment",
            124: "reserved for future assignment",
            125: "reserved for future assignment",
            126: "reserved for future assignment",
            127: "reserved for future assignment",
            128: "reserved for future assignment",
            129: "reserved for future assignment",
            13: "reserved for future assignment",
            130: "reserved for future assignment",
            131: "reserved for future assignment",
            132: "reserved for future assignment",
            133: "reserved for future assignment",
            134: "reserved for future assignment",
            135: "reserved for future assignment",
            136: "reserved for future assignment",
            137: "reserved for future assignment",
            138: "reserved for future assignment",
            139: "reserved for future assignment",
            14: "reserved for future assignment",
            140: "reserved for future assignment",
            141: "reserved for future assignment",
            142: "reserved for future assignment",
            143: "reserved for future assignment",
            144: "reserved for future assignment",
            145: "reserved for future assignment",
            146: "reserved for future assignment",
            147: "reserved for future assignment",
            148: "reserved for future assignment",
            149: "reserved for future assignment",
            15: "reserved for future assignment",
            150: "reserved for future assignment",
            151: "reserved for future assignment",
            152: "reserved for future assignment",
            153: "reserved for future assignment",
            154: "reserved for future assignment",
            155: "reserved for future assignment",
            156: "reserved for future assignment",
            157: "reserved for future assignment",
            158: "reserved for future assignment",
            159: "reserved for future assignment",
            16: "reserved for future assignment",
            160: "reserved for future assignment",
            161: "reserved for future assignment",
            162: "reserved for future assignment",
            163: "reserved for future assignment",
            164: "reserved for future assignment",
            165: "reserved for future assignment",
            166: "reserved for future assignment",
            167: "reserved for future assignment",
            168: "reserved for future assignment",
            169: "reserved for future assignment",
            17: "reserved for future assignment",
            170: "reserved for future assignment",
            171: "reserved for future assignment",
            172: "reserved for future assignment",
            173: "reserved for future assignment",
            174: "reserved for future assignment",
            175: "reserved for future assignment",
            176: "reserved for future assignment",
            177: "reserved for future assignment",
            178: "reserved for future assignment",
            179: "reserved for future assignment",
            18: "reserved for future assignment",
            180: "reserved for future assignment",
            181: "reserved for future assignment",
            182: "reserved for future assignment",
            183: "reserved for future assignment",
            184: "reserved for future assignment",
            185: "reserved for future assignment",
            186: "reserved for future assignment",
            187: "reserved for future assignment",
            188: "reserved for future assignment",
            189: "reserved for future assignment",
            19: "reserved for future assignment",
            190: "reserved for future assignment",
            191: "reserved for future assignment",
            192: "reserved for future assignment",
            193: "reserved for future assignment",
            194: "reserved for future assignment",
            195: "reserved for future assignment",
            196: "reserved for future assignment",
            197: "reserved for future assignment",
            198: "reserved for future assignment",
            199: "reserved for future assignment",
            2: "reserved for future assignment",
            20: "reserved for future assignment",
            200: "reserved for future assignment",
            201: "reserved for future assignment",
            202: "reserved for future assignment",
            203: "reserved for future assignment",
            204: "reserved for future assignment",
            205: "reserved for future assignment",
            206: "reserved for future assignment",
            207: "reserved for future assignment",
            208: "reserved for future assignment",
            209: "reserved for future assignment",
            21: "reserved for future assignment",
            210: "reserved for future assignment",
            211: "reserved for future assignment",
            212: "reserved for future assignment",
            213: "reserved for future assignment",
            214: "reserved for future assignment",
            215: "reserved for future assignment",
            216: "reserved for future assignment",
            217: "reserved for future assignment",
            218: "reserved for future assignment",
            219: "reserved for future assignment",
            22: "reserved for future assignment",
            220: "reserved for future assignment",
            221: "reserved for future assignment",
            222: "reserved for future assignment",
            223: "reserved for future assignment",
            224: "reserved for future assignment",
            225: "reserved for future assignment",
            226: "reserved for future assignment",
            227: "reserved for future assignment",
            228: "reserved for future assignment",
            229: "reserved for future assignment",
            23: "reserved for future assignment",
            230: "reserved for future assignment",
            231: "reserved for future assignment",
            232: "reserved for future assignment",
            233: "reserved for future assignment",
            234: "reserved for future assignment",
            235: "reserved for future assignment",
            236: "reserved for future assignment",
            237: "reserved for future assignment",
            238: "reserved for future assignment",
            239: "reserved for future assignment",
            24: "reserved for future assignment",
            240: "reserved for future assignment",
            241: "reserved for future assignment",
            242: "reserved for future assignment",
            243: "reserved for future assignment",
            244: "reserved for future assignment",
            245: "reserved for future assignment",
            246: "reserved for future assignment",
            247: "reserved for future assignment",
            248: "reserved for future assignment",
            249: "reserved for future assignment",
            25: "reserved for future assignment",
            250: "reserved for future assignment",
            251: "reserved",
            252: "reserved",
            253: "reserved",
            254: "error",
            255: "not available",
            26: "reserved for future assignment",
            27: "reserved for future assignment",
            28: "reserved for future assignment",
            29: "reserved for future assignment",
            3: "reserved for future assignment",
            30: "reserved for future assignment",
            31: "reserved for future assignment",
            32: "reserved for future assignment",
            33: "reserved for future assignment",
            34: "reserved for future assignment",
            35: "reserved for future assignment",
            36: "reserved for future assignment",
            37: "reserved for future assignment",
            38: "reserved for future assignment",
            39: "reserved for future assignment",
            4: "reserved for future assignment",
            40: "reserved for future assignment",
            41: "reserved for future assignment",
            42: "reserved for future assignment",
            43: "reserved for future assignment",
            44: "reserved for future assignment",
            45: "reserved for future assignment",
            46: "reserved for future assignment",
            47: "reserved for future assignment",
            48: "reserved for future assignment",
            49: "reserved for future assignment",
            5: "reserved for future assignment",
            50: "reserved for future assignment",
            51: "reserved for future assignment",
            52: "reserved for future assignment",
            53: "reserved for future assignment",
            54: "reserved for future assignment",
            55: "reserved for future assignment",
            56: "reserved for future assignment",
            57: "reserved for future assignment",
            58: "reserved for future assignment",
            59: "reserved for future assignment",
            6: "reserved for future assignment",
            60: "reserved for future assignment",
            61: "reserved for future assignment",
            62: "reserved for future assignment",
            63: "reserved for future assignment",
            64: "reserved for future assignment",
            65: "reserved for future assignment",
            66: "reserved for future assignment",
            67: "reserved for future assignment",
            68: "reserved for future assignment",
            69: "reserved for future assignment",
            7: "reserved for future assignment",
            70: "reserved for future assignment",
            71: "reserved for future assignment",
            72: "reserved for future assignment",
            73: "reserved for future assignment",
            74: "reserved for future assignment",
            75: "reserved for future assignment",
            76: "reserved for future assignment",
            77: "reserved for future assignment",
            78: "reserved for future assignment",
            79: "reserved for future assignment",
            8: "reserved for future assignment",
            80: "reserved for future assignment",
            81: "reserved for future assignment",
            82: "reserved for future assignment",
            83: "reserved for future assignment",
            84: "reserved for future assignment",
            85: "reserved for future assignment",
            86: "reserved for future assignment",
            87: "reserved for future assignment",
            88: "reserved for future assignment",
            89: "reserved for future assignment",
            9: "reserved for future assignment",
            90: "reserved for future assignment",
            91: "reserved for future assignment",
            92: "reserved for future assignment",
            93: "reserved for future assignment",
            94: "reserved for future assignment",
            95: "reserved for future assignment",
            96: "reserved for future assignment",
            97: "reserved for future assignment",
            98: "reserved for future assignment",
            99: "reserved for future assignment"
        },
        3446: {
            0: "cellular network #1",
            1: "cellular network #2",
            10: "reserved for future assignment",
            100: "reserved for future assignment",
            101: "reserved for future assignment",
            102: "reserved for future assignment",
            103: "reserved for future assignment",
            104: "reserved for future assignment",
            105: "reserved for future assignment",
            106: "reserved for future assignment",
            107: "reserved for future assignment",
            108: "reserved for future assignment",
            109: "reserved for future assignment",
            11: "reserved for future assignment",
            110: "reserved for future assignment",
            111: "reserved for future assignment",
            112: "reserved for future assignment",
            113: "reserved for future assignment",
            114: "reserved for future assignment",
            115: "reserved for future assignment",
            116: "reserved for future assignment",
            117: "reserved for future assignment",
            118: "reserved for future assignment",
            119: "reserved for future assignment",
            12: "reserved for future assignment",
            120: "reserved for future assignment",
            121: "reserved for future assignment",
            122: "reserved for future assignment",
            123: "reserved for future assignment",
            124: "reserved for future assignment",
            125: "reserved for future assignment",
            126: "reserved for future assignment",
            127: "reserved for future assignment",
            128: "reserved for future assignment",
            129: "reserved for future assignment",
            13: "reserved for future assignment",
            130: "reserved for future assignment",
            131: "reserved for future assignment",
            132: "reserved for future assignment",
            133: "reserved for future assignment",
            134: "reserved for future assignment",
            135: "reserved for future assignment",
            136: "reserved for future assignment",
            137: "reserved for future assignment",
            138: "reserved for future assignment",
            139: "reserved for future assignment",
            14: "reserved for future assignment",
            140: "reserved for future assignment",
            141: "reserved for future assignment",
            142: "reserved for future assignment",
            143: "reserved for future assignment",
            144: "reserved for future assignment",
            145: "reserved for future assignment",
            146: "reserved for future assignment",
            147: "reserved for future assignment",
            148: "reserved for future assignment",
            149: "reserved for future assignment",
            15: "reserved for future assignment",
            150: "reserved for future assignment",
            151: "reserved for future assignment",
            152: "reserved for future assignment",
            153: "reserved for future assignment",
            154: "reserved for future assignment",
            155: "reserved for future assignment",
            156: "reserved for future assignment",
            157: "reserved for future assignment",
            158: "reserved for future assignment",
            159: "reserved for future assignment",
            16: "reserved for future assignment",
            160: "reserved for future assignment",
            161: "reserved for future assignment",
            162: "reserved for future assignment",
            163: "reserved for future assignment",
            164: "reserved for future assignment",
            165: "reserved for future assignment",
            166: "reserved for future assignment",
            167: "reserved for future assignment",
            168: "reserved for future assignment",
            169: "reserved for future assignment",
            17: "reserved for future assignment",
            170: "reserved for future assignment",
            171: "reserved for future assignment",
            172: "reserved for future assignment",
            173: "reserved for future assignment",
            174: "reserved for future assignment",
            175: "reserved for future assignment",
            176: "reserved for future assignment",
            177: "reserved for future assignment",
            178: "reserved for future assignment",
            179: "reserved for future assignment",
            18: "reserved for future assignment",
            180: "reserved for future assignment",
            181: "reserved for future assignment",
            182: "reserved for future assignment",
            183: "reserved for future assignment",
            184: "reserved for future assignment",
            185: "reserved for future assignment",
            186: "reserved for future assignment",
            187: "reserved for future assignment",
            188: "reserved for future assignment",
            189: "reserved for future assignment",
            19: "reserved for future assignment",
            190: "reserved for future assignment",
            191: "reserved for future assignment",
            192: "reserved for future assignment",
            193: "reserved for future assignment",
            194: "reserved for future assignment",
            195: "reserved for future assignment",
            196: "reserved for future assignment",
            197: "reserved for future assignment",
            198: "reserved for future assignment",
            199: "reserved for future assignment",
            2: "wireless lan #1",
            20: "reserved for future assignment",
            200: "reserved for future assignment",
            201: "reserved for future assignment",
            202: "reserved for future assignment",
            203: "reserved for future assignment",
            204: "reserved for future assignment",
            205: "reserved for future assignment",
            206: "reserved for future assignment",
            207: "reserved for future assignment",
            208: "reserved for future assignment",
            209: "reserved for future assignment",
            21: "reserved for future assignment",
            210: "reserved for future assignment",
            211: "reserved for future assignment",
            212: "reserved for future assignment",
            213: "reserved for future assignment",
            214: "reserved for future assignment",
            215: "reserved for future assignment",
            216: "reserved for future assignment",
            217: "reserved for future assignment",
            218: "reserved for future assignment",
            219: "reserved for future assignment",
            22: "reserved for future assignment",
            220: "reserved for future assignment",
            221: "reserved for future assignment",
            222: "reserved for future assignment",
            223: "reserved for future assignment",
            224: "reserved for future assignment",
            225: "reserved for future assignment",
            226: "reserved for future assignment",
            227: "reserved for future assignment",
            228: "reserved for future assignment",
            229: "reserved for future assignment",
            23: "reserved for future assignment",
            230: "reserved for future assignment",
            231: "reserved for future assignment",
            232: "reserved for future assignment",
            233: "reserved for future assignment",
            234: "reserved for future assignment",
            235: "reserved for future assignment",
            236: "reserved for future assignment",
            237: "reserved for future assignment",
            238: "reserved for future assignment",
            239: "reserved for future assignment",
            24: "reserved for future assignment",
            240: "reserved for future assignment",
            241: "reserved for future assignment",
            242: "reserved for future assignment",
            243: "reserved for future assignment",
            244: "reserved for future assignment",
            245: "reserved for future assignment",
            246: "reserved for future assignment",
            247: "reserved for future assignment",
            248: "reserved for future assignment",
            249: "reserved for future assignment",
            25: "reserved for future assignment",
            250: "reserved for future assignment",
            251: "reserved",
            252: "reserved",
            253: "reserved",
            254: "error",
            255: "not available",
            26: "reserved for future assignment",
            27: "reserved for future assignment",
            28: "reserved for future assignment",
            29: "reserved for future assignment",
            3: "wireless lan #2",
            30: "reserved for future assignment",
            31: "reserved for future assignment",
            32: "reserved for future assignment",
            33: "reserved for future assignment",
            34: "reserved for future assignment",
            35: "reserved for future assignment",
            36: "reserved for future assignment",
            37: "reserved for future assignment",
            38: "reserved for future assignment",
            39: "reserved for future assignment",
            4: "satellite network",
            40: "reserved for future assignment",
            41: "reserved for future assignment",
            42: "reserved for future assignment",
            43: "reserved for future assignment",
            44: "reserved for future assignment",
            45: "reserved for future assignment",
            46: "reserved for future assignment",
            47: "reserved for future assignment",
            48: "reserved for future assignment",
            49: "reserved for future assignment",
            5: "reserved for future assignment",
            50: "reserved for future assignment",
            51: "reserved for future assignment",
            52: "reserved for future assignment",
            53: "reserved for future assignment",
            54: "reserved for future assignment",
            55: "reserved for future assignment",
            56: "reserved for future assignment",
            57: "reserved for future assignment",
            58: "reserved for future assignment",
            59: "reserved for future assignment",
            6: "reserved for future assignment",
            60: "reserved for future assignment",
            61: "reserved for future assignment",
            62: "reserved for future assignment",
            63: "reserved for future assignment",
            64: "reserved for future assignment",
            65: "reserved for future assignment",
            66: "reserved for future assignment",
            67: "reserved for future assignment",
            68: "reserved for future assignment",
            69: "reserved for future assignment",
            7: "reserved for future assignment",
            70: "reserved for future assignment",
            71: "reserved for future assignment",
            72: "reserved for future assignment",
            73: "reserved for future assignment",
            74: "reserved for future assignment",
            75: "reserved for future assignment",
            76: "reserved for future assignment",
            77: "reserved for future assignment",
            78: "reserved for future assignment",
            79: "reserved for future assignment",
            8: "reserved for future assignment",
            80: "reserved for future assignment",
            81: "reserved for future assignment",
            82: "reserved for future assignment",
            83: "reserved for future assignment",
            84: "reserved for future assignment",
            85: "reserved for future assignment",
            86: "reserved for future assignment",
            87: "reserved for future assignment",
            88: "reserved for future assignment",
            89: "reserved for future assignment",
            9: "reserved for future assignment",
            90: "reserved for future assignment",
            91: "reserved for future assignment",
            92: "reserved for future assignment",
            93: "reserved for future assignment",
            94: "reserved for future assignment",
            95: "reserved for future assignment",
            96: "reserved for future assignment",
            97: "reserved for future assignment",
            98: "reserved for future assignment",
            99: "reserved for future assignment"
        },
        3447: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available"
        },
        3448: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available"
        },
        3452: {
            0: "   enable switch off - pto operation not desired",
            1: "   enable switch off - pto operation desired",
            2: "   error",
            3: "   not available"
        },
        3453: {
            0: "   enable switch off - pto operation not desired",
            1: "   enable switch off - pto operation desired",
            2: "   error",
            3: "   not available"
        },
        3454: {
            0: "   enable switch off - pto operation not desired",
            1: "   enable switch off - pto operation desired",
            2: "   error",
            3: "   not available"
        },
        3455: {
            0: "   enable switch off - pto operation not desired",
            1: "   enable switch off - pto operation desired",
            2: "   error",
            3: "   not available"
        },
        3456: {
            0: "   consent not given - pto drive should not be engaged",
            1: "   consent given - pto drive may be engaged",
            2: "   error",
            3: "   not available"
        },
        3457: {
            0: "   consent not given - pto drive should not be engaged",
            1: "   consent given - pto drive may be engaged",
            2: "   error",
            3: "   not available"
        },
        3458: {
            0: "   consent not given - pto drive should not be engaged",
            1: "   consent given - pto drive may be engaged",
            2: "   error",
            3: "   not available"
        },
        3459: {
            0: "   consent not given - pto drive should not be engaged",
            1: "   consent given - pto drive may be engaged",
            2: "   error",
            3: "   not available"
        },
        3460: {
            0: "   drive not engaged",
            1: "   drive is engaged",
            2: "   error",
            3: "   not available"
        },
        3461: {
            0: "   drive not engaged",
            1: "   drive is engaged",
            2: "   error",
            3: "   not available"
        },
        3462: {
            0: "   drive not engaged",
            1: "   drive is engaged",
            2: "   error",
            3: "   not available"
        },
        3463: {
            0: "   drive not engaged",
            1: "   drive is engaged",
            2: "   error",
            3: "   not available"
        },
        3482: {
            0: "   not active",
            1: "   active",
            2: "   reserved for sae assignment",
            3: "   not available"
        },
        3483: {
            0: "   not active",
            1: "   active",
            2: "   reserved for sae assignment",
            3: "   not available"
        },
        3484: {
            0: "   not active",
            1: "   active",
            2: "   reserved for sae assignment",
            3: "   not available"
        },
        3489: {
            0: "   not active",
            1: "   active",
            2: "   reserved for sae assignment",
            3: "   not available"
        },
        3490: {
            0: "   not active",
            1: "   active",
            2: "   reserved for sae assignment",
            3: "   not available"
        },
        3491: {
            0: "   not active",
            1: "   active",
            2: "   reserved for sae assignment",
            3: "   not available"
        },
        3492: {
            0: "   not active",
            1: "   active",
            2: "   reserved for sae assignment",
            3: "   not available"
        },
        3496: {
            0: "   not active",
            1: "   active",
            2: "   reserved for sae assignment",
            3: "   not available"
        },
        3497: {
            0: "   not active",
            1: "   active",
            2: "   reserved for sae assignment",
            3: "   not available"
        },
        3498: {
            0: "   not active",
            1: "   active",
            2: "   reserved for sae assignment",
            3: "   not available"
        },
        3503: {
            0: "   not active",
            1: "   active",
            2: "   reserved for sae assignment",
            3: "   not available"
        },
        3504: {
            0: "   not active",
            1: "   active",
            2: "   reserved for sae assignment",
            3: "   not available"
        },
        3505: {
            0: "   not active",
            1: "   active",
            2: "   reserved for sae assignment",
            3: "   not available"
        },
        3506: {
            0: "   not active",
            1: "   active",
            2: "   reserved for sae assignment",
            3: "   not available"
        },
        3521: {
            0: "urea concentration too high",
            1: "urea concentration too low",
            13: "not able to determine diesel exhaust fluid property (fluid type unknown)",
            14: "error with diesel exhaust fluid property detection",
            15: "not available",
            2: "fluid is diesel",
            3: "diesel exhaust fluid is proper mixture",
            5: " to 1100 - reserved for sae assignment"
        },
        3533: {
            0: "empty",
            1: "full or not empty",
            2: "error",
            3: "not available"
        },
        3550: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available or not installed"
        },
        3551: {
            0: "not sufficiently lubricated",
            1: "sufficiently lubricated",
            2: "unable to determine if sufficiently lubricated",
            3: "not available or not installed"
        },
        3552: {
            0: "not sufficiently pre-heated",
            1: "sufficiently pre-heated",
            2: "unable to determine if engine oil is pre-heated sufficiently",
            3: "not available or not installed"
        },
        3553: {
            0: "not sufficiently pre-heated",
            1: "sufficiently pre-heated",
            2: "unable to determine  engine coolant pre-heated sufficiently",
            3: "not available or not installed"
        },
        3554: {
            0: "sufficiently vented (all combustible gases have been evacuated from intake and exhaust systems)",
            1: "ventilation needed (combustible gases may be present in intake and exhaust systems)",
            2: "ventilation delay (delay without cranking between ventilation and start engine states)",
            3: "currently venting (engine cranks w/o gas or ignition to clear intake and exhaust systems with air)",
            4: "reserved for sae assignment",
            5: "reserved for sae assignment ",
            6: "not able to determine if sufficiently vented",
            7: "not available or not installe"
        },
        3557: {
            0: "off",
            1: "on",
            2: "reserved ",
            3: "take no action"
        },
        3558: {
            0: "             sae j1995",
            1: "   not defined",
            10: "   not defined",
            11: "   not defined",
            12: "   not defined",
            13: "   not defined",
            15: "             not available",
            2: "   not defined",
            3: "   not defined",
            4: "   not defined",
            5: "   not defined",
            6: "   not defined",
            7: "   not defined",
            8: "   not defined",
            9: "   not defined"
        },
        3564: {
            0: "   disable lane departure warning",
            1: "   enable lane departure warning",
            2: "   reserved",
            3: "   don-t care"
        },
        3565: {
            0: "   middle of the vehicle does not depart the lane to the left side",
            1: "   middle of the vehicle departs the lane to the left side",
            2: "   reserved",
            3: "   take no action"
        },
        3566: {
            0: "   middle of the vehicle does not depart the lane to the right side",
            1: "   middle of the vehicle departs the lane to the right side",
            2: "   reserved",
            3: "   take no action"
        },
        3567: {
            0: "inactive (ready to start automatically)",
            1: "active (not ready to start automatically)",
            2: "error",
            3: "not available"
        },
        3568: {
            0: "inactive (ready to parallel)",
            1: "active (not ready to parallel)",
            2: "error",
            3: "not available"
        },
        3589: {
            0: "off",
            1: "on",
            2: "reserved",
            3: "unavailable"
        },
        3600: {
            0: " take no action - do not reset",
            1: " reset",
            2: " reserved, take no action",
            3: " not applicable"
        },
        3601: {
            0: "off (reset / no test in process)",
            1: "on (test in process / test complete)",
            2: "reserved",
            3: "don't care / take no action"
        },
        3602: {
            0: "off",
            1: "on",
            2: "reserved",
            3: "don't care / take no action"
        },
        3603: {
            0: "off (no power conservation; all systems with supply power available)",
            1: "on (power conservation active; certain systems have supply power witheld)",
            2: "reserved",
            3: "unavailable"
        },
        3604: {
            0: "off ",
            1: "on",
            2: "reserved",
            3: "unavailable"
        },
        3605: {
            0: "off",
            1: "on",
            2: "reserved",
            3: "don't care / take no action"
        },
        3606: {
            0: "off (no shutdown requested)",
            1: "on (shutdown requested)",
            2: "reserved",
            3: "don't care / take no action"
        },
        3607: {
            0: "off (no shutdown requested)",
            1: "on (shutdown requested)",
            2: "reserved",
            3: "don't care / take no action"
        },
        3608: {
            0: "closed (vent closed)",
            1: "open (vent open)",
            2: "reserved",
            3: "don't care / take no action"
        },
        3613: {
            0: " clear display - the displayed information should be deleted or not displayed any more.",
            1: " hold display - the presently displayed information should be retained",
            10: "  reserved",
            11: "  reserved",
            12: "  reserved",
            13: "  reserved",
            14: "  reserved",
            15: "not applicable",
            2: " append at end of display - the transmitted information should be displayed at the end of the presently displayed information",
            3: " overwrite display - the presently displayed information is to be overwritten with the transmitted information",
            4: " overwrite substring - a portion of the presently displayed information is to be overwritten with the transmitted information starting at display text index (spn 3614)",
            5: " highlight substring",
            6: " blink substring",
            7: "  reserved",
            8: "  reserved",
            9: "  reserved"
        },
        3621: {
            0: "dtc is previously active",
            1: "dtc is active"
        },
        3645: {
            0: " 2 wheel high (normal or `'on highway' range)",
            1: " 4 wheel high (normal or  'on highway' range",
            2: " neutral",
            3: " 2 wheel low (or `'off highway' range)",
            4: " 4 wheel low (or `'off highway' range)",
            5: " transfer case shift in progress or gear not confirmed",
            6: " error",
            7: " not available"
        },
        3667: {
            0: "air flow allowed to engine",
            1: "air flow restricted",
            2: "error",
            3: "not available"
        },
        3669: {
            0: "counterclockwise (ccw)",
            1: "clockwise (cw)",
            2: "reserved",
            3: "not available"
        },
        3685: {
            0: "relative number of counts",
            1: "absolute number of counts",
            2: "not defined",
            3: "signal not available"
        },
        3687: {
            0: "programming mode active",
            1: "normal mode active",
            2: "reserved",
            3: "not available"
        },
        3688: {
            0: "sas not calibrated",
            1: "sas calibrated",
            2: "reserved",
            3: "not available"
        },
        3691: {
            0: "   dynamic bending light deactivated.  there is no problem but for some reasons (e.g. daylight present) the bending light is deactivated.",
            1: "   dynamic bending light is working properly.  normal operation mode. bending light is working properly.",
            2: "   dynamic bending light fail safe position.  bending light is damaged, current position does not affect oncoming traffic.",
            3: "   dynamic bending light failure.  bending light does not work, current position may affect oncoming traffic.",
            4: " reserved",
            5: " reserved",
            6: " reserved",
            7: "   not available"
        },
        3692: {
            0: "   dynamic bending light deactivated.  there is no problem but for some reasons (e.g. daylight present) the bending light is deactivated.",
            1: "   dynamic bending light is working properly.  normal operation mode. bending light is working properly.",
            2: "   dynamic bending light fail safe position.  bending light is damaged, current position does not affect oncoming traffic.",
            3: "   dynamic bending light failure.  bending light does not work, current position may affect oncoming traffic.",
            4: " reserved",
            5: " reserved",
            6: " reserved",
            7: "   not available"
        },
        3693: {
            0: "   light distribution town - the basic light distribution for town light is generated with a symmetrical cut-off line (class v)",
            1: "   light distribution country - country light is mostly similar to the actual sae low beam pattern. (class c)",
            10: "   reserved",
            11: "   reserved",
            12: "   reserved",
            13: "   reserved",
            14: "   reserved",
            15: "   not available",
            2: "   light distribution motorway - the task of motorway light is an improved overall viewing distance and an optimum of illumination of the second lane on roads with more than one lane and without oncoming traffic. (class e)",
            3: "   light distribution high beam - the high beam distribution of the module generates an illumination level that an additional high beam spot is not necessary.",
            4: "   light distribution adverse weather - the adverse weather light improves the viewing situation for the driver on wet roads and reduces the amount of glare from reflections on wed reads for the oncoming traffic. (class w) ",
            5: "   light distribution fail safe position - module is damaged, current light distribution does not affect oncoming traffic.",
            6: "   light distribution failure - module is damaged, current position may affect oncoming traffic.",
            7: "   different light distributions not available -  no afs distributions available. head lamp has only one distribution.",
            8: "   reserved",
            9: "   reserved"
        },
        3694: {
            0: "   light distribution town - the basic light distribution for town light is generated with a symmetrical cut-off line (class v)",
            1: "   light distribution country - country light is mostly similar to the actual sae low beam pattern. (class c)",
            10: "   reserved",
            11: "   reserved",
            12: "   reserved",
            13: "   reserved",
            14: "   reserved",
            15: "   not available",
            2: "   light distribution motorway - the task of motorway light is an improved overall viewing distance and an optimum of illumination of the second lane on roads with more than one lane and without oncoming traffic. (class e)",
            3: "   light distribution high beam - the high beam distribution of the module generates an illumination level that an additional high beam spot is not necessary.",
            4: "   light distribution adverse weather - the adverse weather light improves the viewing situation for the driver on wet roads and reduces the amount of glare from reflections on wed reads for the oncoming traffic. (class w) ",
            5: "   light distribution fail safe position - module is damaged, current light distribution does not affect oncoming traffic.",
            6: "   light distribution failure - module is damaged, current position may affect oncoming traffic.",
            7: "   different light distributions not available -  no afs distributions available. head lamp has only one distribution.",
            8: "   reserved",
            9: "   reserved"
        },
        3695: {
            0: " not active",
            1: " active",
            2: " error",
            3: " not available"
        },
        3696: {
            0: " not active",
            1: " active",
            2: " error",
            3: " not available"
        },
        3697: {
            0: " off",
            1: " on - solid",
            2: " reserved for sae assignment ",
            3: " reserved for sae assignment",
            4: " on - fast blink (1 hz)",
            5: " reserved for sae assignment",
            6: " reserved for sae assignment",
            7: " not available"
        },
        3698: {
            0: " off",
            1: " on - solid ",
            2: " reserved for sae assignment ",
            3: " reserved for sae assignment",
            4: " reserved for sae assignment",
            5: " reserved for sae assignment",
            6: " reserved for sae assignment",
            7: " not available"
        },
        3699: {
            0: " not active",
            1: " active",
            2: " reserved for sae assignment",
            3: " not available"
        },
        3700: {
            0: " not active",
            1: " active",
            2: " regeneration needed - automatically initiated active regeneration imminent",
            3: " not available"
        },
        3701: {
            0: " regeneration not needed",
            1: " regeneration needed - lowest level",
            2: " regeneration needed - moderate level",
            3: " regeneration needed - highest level",
            4: " reserved for sae assignment",
            5: " reserved for sae assignment",
            6: " reserved for sae assignment",
            7: " not available"
        },
        3702: {
            0: " not inhibited",
            1: " inhibited",
            2: " reserved for sae assignment",
            3: " not available"
        },
        3703: {
            0: " not inhibited",
            1: " inhibited",
            2: " reserved for sae assignment",
            3: " not available"
        },
        3704: {
            0: " not inhibited",
            1: " inhibited",
            2: " reserved for sae assignment",
            3: " not available"
        },
        3705: {
            0: " not inhibited",
            1: " inhibited",
            2: " reserved for sae assignment",
            3: " not available"
        },
        3706: {
            0: " not inhibited",
            1: " inhibited",
            2: " reserved for sae assignment",
            3: " not available"
        },
        3707: {
            0: " not inhibited",
            1: " inhibited",
            2: " reserved for sae assignment",
            3: " not available"
        },
        3708: {
            0: " not inhibited",
            1: " inhibited",
            2: " reserved for sae assignment",
            3: " not available"
        },
        3709: {
            0: " not inhibited",
            1: " inhibited",
            2: " reserved for sae assignment",
            3: " not available"
        },
        3710: {
            0: " not inhibited",
            1: " inhibited",
            2: " reserved for sae assignment",
            3: " not available"
        },
        3711: {
            0: " not inhibited",
            1: " inhibited",
            2: " reserved for sae assignment",
            3: " not available"
        },
        3712: {
            0: " not inhibited",
            1: " inhibited",
            2: " reserved for sae assignment",
            3: " not available"
        },
        3713: {
            0: " not inhibited",
            1: " inhibited",
            2: " reserved for sae assignment",
            3: " not available"
        },
        3714: {
            0: " not inhibited",
            1: " inhibited",
            2: " reserved for sae assignment",
            3: " not available"
        },
        3715: {
            0: " not inhibited",
            1: " inhibited",
            2: " reserved for sae assignment",
            3: " not available"
        },
        3716: {
            0: " not inhibited",
            1: " inhibited",
            2: " reserved for sae assignment",
            3: " not available"
        },
        3717: {
            0: " not inhibited",
            1: " inhibited",
            2: " reserved for sae assignment",
            3: " not available"
        },
        3718: {
            0: " not enabled",
            1: " enabled",
            2: " reserved for sae assignment",
            3: " not available"
        },
        3750: {
            0: " active dpf regeneration not inhibited",
            1: " active dpf regeneration inhibited",
            2: " reserved for sae assignment",
            3: " not available"
        },
        3785: {
            0: "ok (normal operation)  ",
            1: "non-functioning",
            2: "overstroke ",
            3: "dragging brake",
            4: "reserved",
            5: "reserved",
            6: "sensor error",
            7: "not available"
        },
        3786: {
            0: "ok (normal operation)",
            1: "non-functioning",
            2: "overstroke",
            3: "dragging brake",
            4: "reserved",
            5: "reserved",
            6: "sensor error",
            7: "not available"
        },
        3787: {
            0: "ok (normal operation)",
            1: "non-functioning",
            2: "overstroke",
            3: "dragging brake",
            4: "reserved",
            5: "reserved",
            6: "sensor error",
            7: "not available"
        },
        3788: {
            0: "ok (normal operation)",
            1: "non-functioning",
            2: "overstroke",
            3: "dragging brake",
            4: "reserved",
            5: "reserved",
            6: "sensor error",
            7: "not available"
        },
        3789: {
            0: "ok (normal operation)",
            1: "non-functioning",
            2: "overstroke",
            3: "dragging brake",
            4: "reserved",
            5: "reserved",
            6: "sensor error",
            7: "not available"
        },
        3790: {
            0: "ok (normal operation)",
            1: "non-functioning",
            2: "overstroke",
            3: "dragging brake",
            4: "reserved",
            5: "reserved",
            6: "sensor error",
            7: "not available"
        },
        3791: {
            0: "ok (normal operation)",
            1: "non-functioning",
            2: "overstroke",
            3: "dragging brake",
            4: "reserved",
            5: "reserved",
            6: "sensor error",
            7: "not available"
        },
        3792: {
            0: "ok (normal operation)",
            1: "non-functioning",
            2: "overstroke",
            3: "dragging brake",
            4: "reserved",
            5: "reserved",
            6: "sensor error",
            7: "not available"
        },
        3793: {
            0: "ok (normal operation)",
            1: "non-functioning",
            2: "overstroke",
            3: "dragging brake",
            4: "reserved",
            5: "reserved",
            6: "sensor error",
            7: "not available"
        },
        3794: {
            0: "ok (normal operation)",
            1: "non-functioning",
            2: "overstroke",
            3: "dragging brake",
            4: "reserved",
            5: "reserved",
            6: "sensor error",
            7: "not available"
        },
        3795: {
            0: "ok (normal operation)",
            1: "non-functioning",
            2: "overstroke",
            3: "dragging brake",
            4: "reserved",
            5: "reserved",
            6: "sensor error",
            7: "not available"
        },
        3796: {
            0: "ok (normal operation)",
            1: "non-functioning",
            2: "overstroke",
            3: "dragging brake",
            4: "reserved",
            5: "reserved",
            6: "sensor error",
            7: "not available"
        },
        3797: {
            0: "ok (normal operation)",
            1: "non-functioning",
            2: "overstroke",
            3: "dragging brake",
            4: "reserved",
            5: "reserved",
            6: "sensor error",
            7: "not available"
        },
        3798: {
            0: "ok (normal operation)",
            1: "non-functioning",
            2: "overstroke",
            3: "dragging brake",
            4: "reserved",
            5: "reserved",
            6: "sensor error",
            7: "not available"
        },
        3799: {
            0: "ok (normal operation)",
            1: "non-functioning",
            2: "overstroke",
            3: "dragging brake",
            4: "reserved",
            5: "reserved",
            6: "sensor error",
            7: "not available"
        },
        3800: {
            0: "ok (normal operation)",
            1: "non-functioning",
            2: "overstroke",
            3: "dragging brake",
            4: "reserved",
            5: "reserved",
            6: "sensor error",
            7: "not available"
        },
        3801: {
            0: "ok (normal operation)",
            1: "non-functioning",
            2: "overstroke",
            3: "dragging brake",
            4: "reserved",
            5: "reserved",
            6: "sensor error",
            7: "not available"
        },
        3802: {
            0: "ok (normal operation)",
            1: "non-functioning",
            2: "overstroke",
            3: "dragging brake",
            4: "reserved",
            5: "reserved",
            6: "sensor error",
            7: "not available"
        },
        3803: {
            0: "ok (normal operation)",
            1: "non-functioning",
            2: "overstroke",
            3: "dragging brake",
            4: "reserved",
            5: "reserved",
            6: "sensor error",
            7: "not available"
        },
        3804: {
            0: "ok (normal operation)",
            1: "non-functioning",
            2: "overstroke",
            3: "dragging brake",
            4: "reserved",
            5: "reserved",
            6: "sensor error",
            7: "not available"
        },
        3807: {
            0: "park brake release inhibit not requested",
            1: "park brake release inhibit requested",
            2: "sae reserved",
            3: "unavailable"
        },
        3808: {
            0: "park brake release is not inhibited",
            1: "park brake release is inhibited",
            2: "error (for example: park brake release is not inhibited, but is requested to be inhibited)",
            3: "unavailable"
        },
        3809: {
            0: "no transmission oil level reading desired",
            1: "transmission oil level reading desired",
            2: "reserved",
            3: "don't care / take no action"
        },
        3810: {
            0: " ramp retracted",
            1: " ramp extended",
            2: " error",
            3: " not available"
        },
        3811: {
            0: " ramp disabled",
            1: " ramp enabled ",
            2: " error ",
            3: " not available"
        },
        3812: {
            0: " ramp not being moved",
            1: " ramp being moved",
            2: " error",
            3: " not available"
        },
        3813: {
            0: " ramp retracted",
            1: " ramp extended",
            2: " error",
            3: " not available"
        },
        3814: {
            0: " ramp disabled",
            1: " ramp enabled ",
            2: " error ",
            3: " not available"
        },
        3815: {
            0: " ramp not being moved",
            1: " ramp being moved",
            2: " error",
            3: " not available"
        },
        3816: {
            0: " ramp retracted",
            1: " ramp extended",
            2: " error",
            3: " not available"
        },
        3817: {
            0: " ramp disabled",
            1: " ramp enabled ",
            2: " error ",
            3: " not available"
        },
        3818: {
            0: " ramp not being moved",
            1: " ramp being moved",
            2: " error",
            3: " not available"
        },
        3819: {
            0: " front axle group disengaged ",
            1: " front axle group engaged",
            2: " error",
            3: " not available"
        },
        3820: {
            0: " rear axle group disengaged ",
            1: " rear axle group engaged",
            2: " error",
            3: " not available"
        },
        3839: {
            0: " not active",
            1: " active",
            2: " reserved",
            3: " not available"
        },
        3840: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3841: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3842: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3843: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3844: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3845: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3846: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3847: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3848: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3849: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3850: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3851: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3852: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3853: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3854: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3855: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3856: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3857: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3858: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3859: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3860: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3861: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3862: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3863: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3864: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3865: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3866: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3867: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3868: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3869: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3870: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3871: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3872: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3873: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3874: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3875: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3876: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3877: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3878: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3879: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3880: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3881: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3882: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3883: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3884: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3885: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3886: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3887: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3888: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3889: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3890: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3891: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3892: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3893: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3894: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3895: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3896: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3897: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3898: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3899: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3900: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3901: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3902: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3903: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3904: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3905: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3906: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3907: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3908: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3909: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3910: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3911: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3912: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3913: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3914: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3915: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3916: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3917: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3918: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3919: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3920: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3921: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3922: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3923: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3924: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3925: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3926: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3927: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3928: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3929: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3930: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3931: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3932: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3933: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3934: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3935: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error indicator or not used , depending on application",
            3: "not available or take no action, depending on application"
        },
        3939: {
            0: " enable switch off - pto operation not desired",
            1: " enable switch on - pto operation desired ",
            2: " error",
            3: " not available"
        },
        3940: {
            0: " consent not given - pto drive may not be engaged",
            1: " consent given - pto drive may be engaged",
            2: " error",
            3: " not available"
        },
        3941: {
            0: " drive not engaged",
            1: " drive engaged",
            2: " error",
            3: " not available"
        },
        3942: {
            0: " enable switch off - pto operation not desired",
            1: " enable switch on - pto operation desired",
            2: " error",
            3: " not available"
        },
        3943: {
            0: " consent not given - pto drive may not be engaged",
            1: " consent given - pto drive may be engaged",
            2: " error",
            3: " not available"
        },
        3944: {
            0: " drive not engaged",
            1: " drive engaged",
            2: " error",
            3: " not available"
        },
        3945: {
            0: " enable switch off - pto operation not desired",
            1: " enable switch on - pto operation desired",
            2: " error",
            3: " not available"
        },
        3946: {
            0: " consent not given - pto drive may not be engaged",
            1: " consent given - pto drive may be engaged",
            2: " error",
            3: " not available"
        },
        3947: {
            0: " drive not engaged",
            1: " drive engaged",
            2: " error",
            3: " not available"
        },
        3948: {
            0: " no pto drive is engaged",
            1: " at least one pto drive is engaged",
            2: " error",
            3: " not available"
        },
        3987: {
            0: "lamp off",
            1: "lamp on",
            2: "reserved",
            3: "not available"
        },
        4059: {
            0: "steer axle group not present",
            1: "steer axle group present",
            2: "reserved",
            3: "not available / not applicable"
        },
        4060: {
            0: "lift axle group not present",
            1: "lift axle group present",
            2: "reserved",
            3: "not available / not applicable"
        },
        4061: {
            0: "drive axle group not present",
            1: "drive axle group present",
            2: "reserved",
            3: "not available / not applicable"
        },
        4062: {
            0: "tag axle group not present",
            1: "tag axle group present",
            2: "reserved",
            3: "not available / not applicable"
        },
        4063: {
            0: "additional tractor axle group not present",
            1: "additional tractor axle group present",
            2: "reserved",
            3: "not available / not applicable"
        },
        4064: {
            0: "trailer a axle group not present",
            1: "trailer a axle group present",
            2: "reserved",
            3: "not available / not applicable"
        },
        4065: {
            0: "trailer b axle group not present",
            1: "trailer b axle group present",
            2: "reserved",
            3: "not available / not applicable"
        },
        4066: {
            0: "trailer c axle group not present",
            1: "trailer c axle group present",
            2: "reserved",
            3: "not available / not applicable"
        },
        4067: {
            0: "trailer d axle group not present",
            1: "trailer d axle group present",
            2: "reserved",
            3: "not available / not applicable"
        },
        4068: {
            0: "trailer e axle group not present",
            1: "trailer e axle group present",
            2: "reserved",
            3: "not available / not applicable"
        },
        4069: {
            0: "trailer f axle group not present",
            1: "trailer f axle group present",
            2: "reserved",
            3: "not available / not applicable"
        },
        4070: {
            0: "trailer g axle group not present",
            1: "trailer g axle group present",
            2: "reserved",
            3: "not available / not applicable"
        },
        4071: {
            0: "trailer h axle group not present",
            1: "trailer h axle group present",
            2: "reserved",
            3: "not available / not applicable"
        },
        4072: {
            0: "additional trailer axle group not present",
            1: "additional trailer axle group present",
            2: "reserved",
            3: "not available / not applicable"
        },
        4073: {
            0: "reserved",
            1: "steer axle group",
            10: "trailer e axle group",
            11: "trailer f axle group",
            12: "trailer g axle group",
            13: "trailer h axle group",
            14: "additional trailer axle group",
            15: "reserved",
            2: "lift axle group",
            3: "drive axle group",
            4: "tag axle group",
            5: "additional tractor axle group",
            6: "trailer a axle group",
            7: "trailer b axle group",
            8: "trailer c axle group",
            9: "trailer d axle group"
        },
        4074: {
            0: "reserved",
            1: "steer axle group",
            10: "trailer e axle group",
            11: "trailer f axle group",
            12: "trailer g axle group",
            13: "trailer h axle group",
            14: "additional trailer axle group",
            15: "reserved",
            2: "lift axle group",
            3: "drive axle group",
            4: "tag axle group",
            5: "additional tractor axle group",
            6: "trailer a axle group",
            7: "trailer b axle group",
            8: "trailer c axle group",
            9: "trailer d axle group"
        },
        4075: {
            0: "no change to net vehicle weight change ",
            1: "zero net vehicle weight change",
            2: "reserved",
            3: "not available / not applicable"
        },
        4079: {
            0: "rated speed",
            1: "low idle speed",
            2: "reserved",
            3: "don't care / take no action"
        },
        4080: {
            0: " hz",
            1: " hz",
            10: " hz",
            100: "sae reserved",
            1000: "sae reserved",
            1001: "sae reserved",
            1002: "sae reserved",
            1003: "sae reserved",
            1004: "sae reserved",
            1005: "sae reserved",
            1006: "sae reserved",
            1007: "sae reserved",
            1008: "sae reserved",
            1009: "sae reserved",
            101: "sae reserved",
            1010: "sae reserved",
            1011: "sae reserved",
            1012: "sae reserved",
            1013: "sae reserved",
            1014: "sae reserved",
            1015: "sae reserved",
            1016: "sae reserved",
            1017: "sae reserved",
            1018: "sae reserved",
            1019: "sae reserved",
            102: "sae reserved",
            1020: "sae reserved",
            1021: "sae reserved",
            1022: "sae reserved",
            1023: "sae reserved",
            1024: "sae reserved",
            1025: "sae reserved",
            1026: "sae reserved",
            1027: "sae reserved",
            1028: "sae reserved",
            1029: "sae reserved",
            103: "sae reserved",
            1030: "sae reserved",
            1031: "sae reserved",
            1032: "sae reserved",
            1033: "sae reserved",
            1034: "sae reserved",
            1035: "sae reserved",
            1036: "sae reserved",
            1037: "sae reserved",
            1038: "sae reserved",
            1039: "sae reserved",
            104: "sae reserved",
            1040: "sae reserved",
            1041: "sae reserved",
            1042: "sae reserved",
            1043: "sae reserved",
            1044: "sae reserved",
            1045: "sae reserved",
            1046: "sae reserved",
            1047: "sae reserved",
            1048: "sae reserved",
            1049: "sae reserved",
            105: "sae reserved",
            1050: "sae reserved",
            1051: "sae reserved",
            1052: "sae reserved",
            1053: "sae reserved",
            1054: "sae reserved",
            1055: "sae reserved",
            1056: "sae reserved",
            1057: "sae reserved",
            1058: "sae reserved",
            1059: "sae reserved",
            106: "sae reserved",
            1060: "sae reserved",
            1061: "sae reserved",
            1062: "sae reserved",
            1063: "sae reserved",
            1064: "sae reserved",
            1065: "sae reserved",
            1066: "sae reserved",
            1067: "sae reserved",
            1068: "sae reserved",
            1069: "sae reserved",
            107: "sae reserved",
            1070: "sae reserved",
            1071: "sae reserved",
            1072: "sae reserved",
            1073: "sae reserved",
            1074: "sae reserved",
            1075: "sae reserved",
            1076: "sae reserved",
            1077: "sae reserved",
            1078: "sae reserved",
            1079: "sae reserved",
            108: "sae reserved",
            1080: "sae reserved",
            1081: "sae reserved",
            1082: "sae reserved",
            1083: "sae reserved",
            1084: "sae reserved",
            1085: "sae reserved",
            1086: "sae reserved",
            1087: "sae reserved",
            1088: "sae reserved",
            1089: "sae reserved",
            109: "sae reserved",
            1090: "sae reserved",
            1091: "sae reserved",
            1092: "sae reserved",
            1093: "sae reserved",
            1094: "sae reserved",
            1095: "sae reserved",
            1096: "sae reserved",
            1097: "sae reserved",
            1098: "sae reserved",
            1099: "sae reserved",
            11: "sae reserved",
            110: "sae reserved",
            1100: "sae reserved",
            1101: "sae reserved",
            111: "sae reserved",
            1110: "unknown",
            1111: "don't care /take no action",
            112: "sae reserved",
            113: "sae reserved",
            114: "sae reserved",
            115: "sae reserved",
            116: "sae reserved",
            117: "sae reserved",
            118: "sae reserved",
            119: "sae reserved",
            12: "sae reserved",
            120: "sae reserved",
            121: "sae reserved",
            122: "sae reserved",
            123: "sae reserved",
            124: "sae reserved",
            125: "sae reserved",
            126: "sae reserved",
            127: "sae reserved",
            128: "sae reserved",
            129: "sae reserved",
            13: "sae reserved",
            130: "sae reserved",
            131: "sae reserved",
            132: "sae reserved",
            133: "sae reserved",
            134: "sae reserved",
            135: "sae reserved",
            136: "sae reserved",
            137: "sae reserved",
            138: "sae reserved",
            139: "sae reserved",
            14: "sae reserved",
            140: "sae reserved",
            141: "sae reserved",
            142: "sae reserved",
            143: "sae reserved",
            144: "sae reserved",
            145: "sae reserved",
            146: "sae reserved",
            147: "sae reserved",
            148: "sae reserved",
            149: "sae reserved",
            15: "sae reserved",
            150: "sae reserved",
            151: "sae reserved",
            152: "sae reserved",
            153: "sae reserved",
            154: "sae reserved",
            155: "sae reserved",
            156: "sae reserved",
            157: "sae reserved",
            158: "sae reserved",
            159: "sae reserved",
            16: "sae reserved",
            160: "sae reserved",
            161: "sae reserved",
            162: "sae reserved",
            163: "sae reserved",
            164: "sae reserved",
            165: "sae reserved",
            166: "sae reserved",
            167: "sae reserved",
            168: "sae reserved",
            169: "sae reserved",
            17: "sae reserved",
            170: "sae reserved",
            171: "sae reserved",
            172: "sae reserved",
            173: "sae reserved",
            174: "sae reserved",
            175: "sae reserved",
            176: "sae reserved",
            177: "sae reserved",
            178: "sae reserved",
            179: "sae reserved",
            18: "sae reserved",
            180: "sae reserved",
            181: "sae reserved",
            182: "sae reserved",
            183: "sae reserved",
            184: "sae reserved",
            185: "sae reserved",
            186: "sae reserved",
            187: "sae reserved",
            188: "sae reserved",
            189: "sae reserved",
            19: "sae reserved",
            190: "sae reserved",
            191: "sae reserved",
            192: "sae reserved",
            193: "sae reserved",
            194: "sae reserved",
            195: "sae reserved",
            196: "sae reserved",
            197: "sae reserved",
            198: "sae reserved",
            199: "sae reserved",
            2: " hz",
            20: "sae reserved",
            200: "sae reserved",
            201: "sae reserved",
            202: "sae reserved",
            203: "sae reserved",
            204: "sae reserved",
            205: "sae reserved",
            206: "sae reserved",
            207: "sae reserved",
            208: "sae reserved",
            209: "sae reserved",
            21: "sae reserved",
            210: "sae reserved",
            211: "sae reserved",
            212: "sae reserved",
            213: "sae reserved",
            214: "sae reserved",
            215: "sae reserved",
            216: "sae reserved",
            217: "sae reserved",
            218: "sae reserved",
            219: "sae reserved",
            22: "sae reserved",
            220: "sae reserved",
            221: "sae reserved",
            222: "sae reserved",
            223: "sae reserved",
            224: "sae reserved",
            225: "sae reserved",
            226: "sae reserved",
            227: "sae reserved",
            228: "sae reserved",
            229: "sae reserved",
            23: "sae reserved",
            230: "sae reserved",
            231: "sae reserved",
            232: "sae reserved",
            233: "sae reserved",
            234: "sae reserved",
            235: "sae reserved",
            236: "sae reserved",
            237: "sae reserved",
            238: "sae reserved",
            239: "sae reserved",
            24: "sae reserved",
            240: "sae reserved",
            241: "sae reserved",
            242: "sae reserved",
            243: "sae reserved",
            244: "sae reserved",
            245: "sae reserved",
            246: "sae reserved",
            247: "sae reserved",
            248: "sae reserved",
            249: "sae reserved",
            25: "sae reserved",
            250: "sae reserved",
            251: "sae reserved",
            252: "sae reserved",
            253: "sae reserved",
            254: "sae reserved",
            255: "sae reserved",
            256: "sae reserved",
            257: "sae reserved",
            258: "sae reserved",
            259: "sae reserved",
            26: "sae reserved",
            260: "sae reserved",
            261: "sae reserved",
            262: "sae reserved",
            263: "sae reserved",
            264: "sae reserved",
            265: "sae reserved",
            266: "sae reserved",
            267: "sae reserved",
            268: "sae reserved",
            269: "sae reserved",
            27: "sae reserved",
            270: "sae reserved",
            271: "sae reserved",
            272: "sae reserved",
            273: "sae reserved",
            274: "sae reserved",
            275: "sae reserved",
            276: "sae reserved",
            277: "sae reserved",
            278: "sae reserved",
            279: "sae reserved",
            28: "sae reserved",
            280: "sae reserved",
            281: "sae reserved",
            282: "sae reserved",
            283: "sae reserved",
            284: "sae reserved",
            285: "sae reserved",
            286: "sae reserved",
            287: "sae reserved",
            288: "sae reserved",
            289: "sae reserved",
            29: "sae reserved",
            290: "sae reserved",
            291: "sae reserved",
            292: "sae reserved",
            293: "sae reserved",
            294: "sae reserved",
            295: "sae reserved",
            296: "sae reserved",
            297: "sae reserved",
            298: "sae reserved",
            299: "sae reserved",
            3: " hz",
            30: "sae reserved",
            300: "sae reserved",
            301: "sae reserved",
            302: "sae reserved",
            303: "sae reserved",
            304: "sae reserved",
            305: "sae reserved",
            306: "sae reserved",
            307: "sae reserved",
            308: "sae reserved",
            309: "sae reserved",
            31: "sae reserved",
            310: "sae reserved",
            311: "sae reserved",
            312: "sae reserved",
            313: "sae reserved",
            314: "sae reserved",
            315: "sae reserved",
            316: "sae reserved",
            317: "sae reserved",
            318: "sae reserved",
            319: "sae reserved",
            32: "sae reserved",
            320: "sae reserved",
            321: "sae reserved",
            322: "sae reserved",
            323: "sae reserved",
            324: "sae reserved",
            325: "sae reserved",
            326: "sae reserved",
            327: "sae reserved",
            328: "sae reserved",
            329: "sae reserved",
            33: "sae reserved",
            330: "sae reserved",
            331: "sae reserved",
            332: "sae reserved",
            333: "sae reserved",
            334: "sae reserved",
            335: "sae reserved",
            336: "sae reserved",
            337: "sae reserved",
            338: "sae reserved",
            339: "sae reserved",
            34: "sae reserved",
            340: "sae reserved",
            341: "sae reserved",
            342: "sae reserved",
            343: "sae reserved",
            344: "sae reserved",
            345: "sae reserved",
            346: "sae reserved",
            347: "sae reserved",
            348: "sae reserved",
            349: "sae reserved",
            35: "sae reserved",
            350: "sae reserved",
            351: "sae reserved",
            352: "sae reserved",
            353: "sae reserved",
            354: "sae reserved",
            355: "sae reserved",
            356: "sae reserved",
            357: "sae reserved",
            358: "sae reserved",
            359: "sae reserved",
            36: "sae reserved",
            360: "sae reserved",
            361: "sae reserved",
            362: "sae reserved",
            363: "sae reserved",
            364: "sae reserved",
            365: "sae reserved",
            366: "sae reserved",
            367: "sae reserved",
            368: "sae reserved",
            369: "sae reserved",
            37: "sae reserved",
            370: "sae reserved",
            371: "sae reserved",
            372: "sae reserved",
            373: "sae reserved",
            374: "sae reserved",
            375: "sae reserved",
            376: "sae reserved",
            377: "sae reserved",
            378: "sae reserved",
            379: "sae reserved",
            38: "sae reserved",
            380: "sae reserved",
            381: "sae reserved",
            382: "sae reserved",
            383: "sae reserved",
            384: "sae reserved",
            385: "sae reserved",
            386: "sae reserved",
            387: "sae reserved",
            388: "sae reserved",
            389: "sae reserved",
            39: "sae reserved",
            390: "sae reserved",
            391: "sae reserved",
            392: "sae reserved",
            393: "sae reserved",
            394: "sae reserved",
            395: "sae reserved",
            396: "sae reserved",
            397: "sae reserved",
            398: "sae reserved",
            399: "sae reserved",
            4: " hz",
            40: "sae reserved",
            400: "sae reserved",
            401: "sae reserved",
            402: "sae reserved",
            403: "sae reserved",
            404: "sae reserved",
            405: "sae reserved",
            406: "sae reserved",
            407: "sae reserved",
            408: "sae reserved",
            409: "sae reserved",
            41: "sae reserved",
            410: "sae reserved",
            411: "sae reserved",
            412: "sae reserved",
            413: "sae reserved",
            414: "sae reserved",
            415: "sae reserved",
            416: "sae reserved",
            417: "sae reserved",
            418: "sae reserved",
            419: "sae reserved",
            42: "sae reserved",
            420: "sae reserved",
            421: "sae reserved",
            422: "sae reserved",
            423: "sae reserved",
            424: "sae reserved",
            425: "sae reserved",
            426: "sae reserved",
            427: "sae reserved",
            428: "sae reserved",
            429: "sae reserved",
            43: "sae reserved",
            430: "sae reserved",
            431: "sae reserved",
            432: "sae reserved",
            433: "sae reserved",
            434: "sae reserved",
            435: "sae reserved",
            436: "sae reserved",
            437: "sae reserved",
            438: "sae reserved",
            439: "sae reserved",
            44: "sae reserved",
            440: "sae reserved",
            441: "sae reserved",
            442: "sae reserved",
            443: "sae reserved",
            444: "sae reserved",
            445: "sae reserved",
            446: "sae reserved",
            447: "sae reserved",
            448: "sae reserved",
            449: "sae reserved",
            45: "sae reserved",
            450: "sae reserved",
            451: "sae reserved",
            452: "sae reserved",
            453: "sae reserved",
            454: "sae reserved",
            455: "sae reserved",
            456: "sae reserved",
            457: "sae reserved",
            458: "sae reserved",
            459: "sae reserved",
            46: "sae reserved",
            460: "sae reserved",
            461: "sae reserved",
            462: "sae reserved",
            463: "sae reserved",
            464: "sae reserved",
            465: "sae reserved",
            466: "sae reserved",
            467: "sae reserved",
            468: "sae reserved",
            469: "sae reserved",
            47: "sae reserved",
            470: "sae reserved",
            471: "sae reserved",
            472: "sae reserved",
            473: "sae reserved",
            474: "sae reserved",
            475: "sae reserved",
            476: "sae reserved",
            477: "sae reserved",
            478: "sae reserved",
            479: "sae reserved",
            48: "sae reserved",
            480: "sae reserved",
            481: "sae reserved",
            482: "sae reserved",
            483: "sae reserved",
            484: "sae reserved",
            485: "sae reserved",
            486: "sae reserved",
            487: "sae reserved",
            488: "sae reserved",
            489: "sae reserved",
            49: "sae reserved",
            490: "sae reserved",
            491: "sae reserved",
            492: "sae reserved",
            493: "sae reserved",
            494: "sae reserved",
            495: "sae reserved",
            496: "sae reserved",
            497: "sae reserved",
            498: "sae reserved",
            499: "sae reserved",
            5: " hz",
            50: "sae reserved",
            500: "sae reserved",
            501: "sae reserved",
            502: "sae reserved",
            503: "sae reserved",
            504: "sae reserved",
            505: "sae reserved",
            506: "sae reserved",
            507: "sae reserved",
            508: "sae reserved",
            509: "sae reserved",
            51: "sae reserved",
            510: "sae reserved",
            511: "sae reserved",
            512: "sae reserved",
            513: "sae reserved",
            514: "sae reserved",
            515: "sae reserved",
            516: "sae reserved",
            517: "sae reserved",
            518: "sae reserved",
            519: "sae reserved",
            52: "sae reserved",
            520: "sae reserved",
            521: "sae reserved",
            522: "sae reserved",
            523: "sae reserved",
            524: "sae reserved",
            525: "sae reserved",
            526: "sae reserved",
            527: "sae reserved",
            528: "sae reserved",
            529: "sae reserved",
            53: "sae reserved",
            530: "sae reserved",
            531: "sae reserved",
            532: "sae reserved",
            533: "sae reserved",
            534: "sae reserved",
            535: "sae reserved",
            536: "sae reserved",
            537: "sae reserved",
            538: "sae reserved",
            539: "sae reserved",
            54: "sae reserved",
            540: "sae reserved",
            541: "sae reserved",
            542: "sae reserved",
            543: "sae reserved",
            544: "sae reserved",
            545: "sae reserved",
            546: "sae reserved",
            547: "sae reserved",
            548: "sae reserved",
            549: "sae reserved",
            55: "sae reserved",
            550: "sae reserved",
            551: "sae reserved",
            552: "sae reserved",
            553: "sae reserved",
            554: "sae reserved",
            555: "sae reserved",
            556: "sae reserved",
            557: "sae reserved",
            558: "sae reserved",
            559: "sae reserved",
            56: "sae reserved",
            560: "sae reserved",
            561: "sae reserved",
            562: "sae reserved",
            563: "sae reserved",
            564: "sae reserved",
            565: "sae reserved",
            566: "sae reserved",
            567: "sae reserved",
            568: "sae reserved",
            569: "sae reserved",
            57: "sae reserved",
            570: "sae reserved",
            571: "sae reserved",
            572: "sae reserved",
            573: "sae reserved",
            574: "sae reserved",
            575: "sae reserved",
            576: "sae reserved",
            577: "sae reserved",
            578: "sae reserved",
            579: "sae reserved",
            58: "sae reserved",
            580: "sae reserved",
            581: "sae reserved",
            582: "sae reserved",
            583: "sae reserved",
            584: "sae reserved",
            585: "sae reserved",
            586: "sae reserved",
            587: "sae reserved",
            588: "sae reserved",
            589: "sae reserved",
            59: "sae reserved",
            590: "sae reserved",
            591: "sae reserved",
            592: "sae reserved",
            593: "sae reserved",
            594: "sae reserved",
            595: "sae reserved",
            596: "sae reserved",
            597: "sae reserved",
            598: "sae reserved",
            599: "sae reserved",
            6: " hz",
            60: "sae reserved",
            600: "sae reserved",
            601: "sae reserved",
            602: "sae reserved",
            603: "sae reserved",
            604: "sae reserved",
            605: "sae reserved",
            606: "sae reserved",
            607: "sae reserved",
            608: "sae reserved",
            609: "sae reserved",
            61: "sae reserved",
            610: "sae reserved",
            611: "sae reserved",
            612: "sae reserved",
            613: "sae reserved",
            614: "sae reserved",
            615: "sae reserved",
            616: "sae reserved",
            617: "sae reserved",
            618: "sae reserved",
            619: "sae reserved",
            62: "sae reserved",
            620: "sae reserved",
            621: "sae reserved",
            622: "sae reserved",
            623: "sae reserved",
            624: "sae reserved",
            625: "sae reserved",
            626: "sae reserved",
            627: "sae reserved",
            628: "sae reserved",
            629: "sae reserved",
            63: "sae reserved",
            630: "sae reserved",
            631: "sae reserved",
            632: "sae reserved",
            633: "sae reserved",
            634: "sae reserved",
            635: "sae reserved",
            636: "sae reserved",
            637: "sae reserved",
            638: "sae reserved",
            639: "sae reserved",
            64: "sae reserved",
            640: "sae reserved",
            641: "sae reserved",
            642: "sae reserved",
            643: "sae reserved",
            644: "sae reserved",
            645: "sae reserved",
            646: "sae reserved",
            647: "sae reserved",
            648: "sae reserved",
            649: "sae reserved",
            65: "sae reserved",
            650: "sae reserved",
            651: "sae reserved",
            652: "sae reserved",
            653: "sae reserved",
            654: "sae reserved",
            655: "sae reserved",
            656: "sae reserved",
            657: "sae reserved",
            658: "sae reserved",
            659: "sae reserved",
            66: "sae reserved",
            660: "sae reserved",
            661: "sae reserved",
            662: "sae reserved",
            663: "sae reserved",
            664: "sae reserved",
            665: "sae reserved",
            666: "sae reserved",
            667: "sae reserved",
            668: "sae reserved",
            669: "sae reserved",
            67: "sae reserved",
            670: "sae reserved",
            671: "sae reserved",
            672: "sae reserved",
            673: "sae reserved",
            674: "sae reserved",
            675: "sae reserved",
            676: "sae reserved",
            677: "sae reserved",
            678: "sae reserved",
            679: "sae reserved",
            68: "sae reserved",
            680: "sae reserved",
            681: "sae reserved",
            682: "sae reserved",
            683: "sae reserved",
            684: "sae reserved",
            685: "sae reserved",
            686: "sae reserved",
            687: "sae reserved",
            688: "sae reserved",
            689: "sae reserved",
            69: "sae reserved",
            690: "sae reserved",
            691: "sae reserved",
            692: "sae reserved",
            693: "sae reserved",
            694: "sae reserved",
            695: "sae reserved",
            696: "sae reserved",
            697: "sae reserved",
            698: "sae reserved",
            699: "sae reserved",
            7: " hz",
            70: "sae reserved",
            700: "sae reserved",
            701: "sae reserved",
            702: "sae reserved",
            703: "sae reserved",
            704: "sae reserved",
            705: "sae reserved",
            706: "sae reserved",
            707: "sae reserved",
            708: "sae reserved",
            709: "sae reserved",
            71: "sae reserved",
            710: "sae reserved",
            711: "sae reserved",
            712: "sae reserved",
            713: "sae reserved",
            714: "sae reserved",
            715: "sae reserved",
            716: "sae reserved",
            717: "sae reserved",
            718: "sae reserved",
            719: "sae reserved",
            72: "sae reserved",
            720: "sae reserved",
            721: "sae reserved",
            722: "sae reserved",
            723: "sae reserved",
            724: "sae reserved",
            725: "sae reserved",
            726: "sae reserved",
            727: "sae reserved",
            728: "sae reserved",
            729: "sae reserved",
            73: "sae reserved",
            730: "sae reserved",
            731: "sae reserved",
            732: "sae reserved",
            733: "sae reserved",
            734: "sae reserved",
            735: "sae reserved",
            736: "sae reserved",
            737: "sae reserved",
            738: "sae reserved",
            739: "sae reserved",
            74: "sae reserved",
            740: "sae reserved",
            741: "sae reserved",
            742: "sae reserved",
            743: "sae reserved",
            744: "sae reserved",
            745: "sae reserved",
            746: "sae reserved",
            747: "sae reserved",
            748: "sae reserved",
            749: "sae reserved",
            75: "sae reserved",
            750: "sae reserved",
            751: "sae reserved",
            752: "sae reserved",
            753: "sae reserved",
            754: "sae reserved",
            755: "sae reserved",
            756: "sae reserved",
            757: "sae reserved",
            758: "sae reserved",
            759: "sae reserved",
            76: "sae reserved",
            760: "sae reserved",
            761: "sae reserved",
            762: "sae reserved",
            763: "sae reserved",
            764: "sae reserved",
            765: "sae reserved",
            766: "sae reserved",
            767: "sae reserved",
            768: "sae reserved",
            769: "sae reserved",
            77: "sae reserved",
            770: "sae reserved",
            771: "sae reserved",
            772: "sae reserved",
            773: "sae reserved",
            774: "sae reserved",
            775: "sae reserved",
            776: "sae reserved",
            777: "sae reserved",
            778: "sae reserved",
            779: "sae reserved",
            78: "sae reserved",
            780: "sae reserved",
            781: "sae reserved",
            782: "sae reserved",
            783: "sae reserved",
            784: "sae reserved",
            785: "sae reserved",
            786: "sae reserved",
            787: "sae reserved",
            788: "sae reserved",
            789: "sae reserved",
            79: "sae reserved",
            790: "sae reserved",
            791: "sae reserved",
            792: "sae reserved",
            793: "sae reserved",
            794: "sae reserved",
            795: "sae reserved",
            796: "sae reserved",
            797: "sae reserved",
            798: "sae reserved",
            799: "sae reserved",
            8: " hz",
            80: "sae reserved",
            800: "sae reserved",
            801: "sae reserved",
            802: "sae reserved",
            803: "sae reserved",
            804: "sae reserved",
            805: "sae reserved",
            806: "sae reserved",
            807: "sae reserved",
            808: "sae reserved",
            809: "sae reserved",
            81: "sae reserved",
            810: "sae reserved",
            811: "sae reserved",
            812: "sae reserved",
            813: "sae reserved",
            814: "sae reserved",
            815: "sae reserved",
            816: "sae reserved",
            817: "sae reserved",
            818: "sae reserved",
            819: "sae reserved",
            82: "sae reserved",
            820: "sae reserved",
            821: "sae reserved",
            822: "sae reserved",
            823: "sae reserved",
            824: "sae reserved",
            825: "sae reserved",
            826: "sae reserved",
            827: "sae reserved",
            828: "sae reserved",
            829: "sae reserved",
            83: "sae reserved",
            830: "sae reserved",
            831: "sae reserved",
            832: "sae reserved",
            833: "sae reserved",
            834: "sae reserved",
            835: "sae reserved",
            836: "sae reserved",
            837: "sae reserved",
            838: "sae reserved",
            839: "sae reserved",
            84: "sae reserved",
            840: "sae reserved",
            841: "sae reserved",
            842: "sae reserved",
            843: "sae reserved",
            844: "sae reserved",
            845: "sae reserved",
            846: "sae reserved",
            847: "sae reserved",
            848: "sae reserved",
            849: "sae reserved",
            85: "sae reserved",
            850: "sae reserved",
            851: "sae reserved",
            852: "sae reserved",
            853: "sae reserved",
            854: "sae reserved",
            855: "sae reserved",
            856: "sae reserved",
            857: "sae reserved",
            858: "sae reserved",
            859: "sae reserved",
            86: "sae reserved",
            860: "sae reserved",
            861: "sae reserved",
            862: "sae reserved",
            863: "sae reserved",
            864: "sae reserved",
            865: "sae reserved",
            866: "sae reserved",
            867: "sae reserved",
            868: "sae reserved",
            869: "sae reserved",
            87: "sae reserved",
            870: "sae reserved",
            871: "sae reserved",
            872: "sae reserved",
            873: "sae reserved",
            874: "sae reserved",
            875: "sae reserved",
            876: "sae reserved",
            877: "sae reserved",
            878: "sae reserved",
            879: "sae reserved",
            88: "sae reserved",
            880: "sae reserved",
            881: "sae reserved",
            882: "sae reserved",
            883: "sae reserved",
            884: "sae reserved",
            885: "sae reserved",
            886: "sae reserved",
            887: "sae reserved",
            888: "sae reserved",
            889: "sae reserved",
            89: "sae reserved",
            890: "sae reserved",
            891: "sae reserved",
            892: "sae reserved",
            893: "sae reserved",
            894: "sae reserved",
            895: "sae reserved",
            896: "sae reserved",
            897: "sae reserved",
            898: "sae reserved",
            899: "sae reserved",
            9: " hz",
            90: "sae reserved",
            900: "sae reserved",
            901: "sae reserved",
            902: "sae reserved",
            903: "sae reserved",
            904: "sae reserved",
            905: "sae reserved",
            906: "sae reserved",
            907: "sae reserved",
            908: "sae reserved",
            909: "sae reserved",
            91: "sae reserved",
            910: "sae reserved",
            911: "sae reserved",
            912: "sae reserved",
            913: "sae reserved",
            914: "sae reserved",
            915: "sae reserved",
            916: "sae reserved",
            917: "sae reserved",
            918: "sae reserved",
            919: "sae reserved",
            92: "sae reserved",
            920: "sae reserved",
            921: "sae reserved",
            922: "sae reserved",
            923: "sae reserved",
            924: "sae reserved",
            925: "sae reserved",
            926: "sae reserved",
            927: "sae reserved",
            928: "sae reserved",
            929: "sae reserved",
            93: "sae reserved",
            930: "sae reserved",
            931: "sae reserved",
            932: "sae reserved",
            933: "sae reserved",
            934: "sae reserved",
            935: "sae reserved",
            936: "sae reserved",
            937: "sae reserved",
            938: "sae reserved",
            939: "sae reserved",
            94: "sae reserved",
            940: "sae reserved",
            941: "sae reserved",
            942: "sae reserved",
            943: "sae reserved",
            944: "sae reserved",
            945: "sae reserved",
            946: "sae reserved",
            947: "sae reserved",
            948: "sae reserved",
            949: "sae reserved",
            95: "sae reserved",
            950: "sae reserved",
            951: "sae reserved",
            952: "sae reserved",
            953: "sae reserved",
            954: "sae reserved",
            955: "sae reserved",
            956: "sae reserved",
            957: "sae reserved",
            958: "sae reserved",
            959: "sae reserved",
            96: "sae reserved",
            960: "sae reserved",
            961: "sae reserved",
            962: "sae reserved",
            963: "sae reserved",
            964: "sae reserved",
            965: "sae reserved",
            966: "sae reserved",
            967: "sae reserved",
            968: "sae reserved",
            969: "sae reserved",
            97: "sae reserved",
            970: "sae reserved",
            971: "sae reserved",
            972: "sae reserved",
            973: "sae reserved",
            974: "sae reserved",
            975: "sae reserved",
            976: "sae reserved",
            977: "sae reserved",
            978: "sae reserved",
            979: "sae reserved",
            98: "sae reserved",
            980: "sae reserved",
            981: "sae reserved",
            982: "sae reserved",
            983: "sae reserved",
            984: "sae reserved",
            985: "sae reserved",
            986: "sae reserved",
            987: "sae reserved",
            988: "sae reserved",
            989: "sae reserved",
            99: "sae reserved",
            990: "sae reserved",
            991: "sae reserved",
            992: "sae reserved",
            993: "sae reserved",
            994: "sae reserved",
            995: "sae reserved",
            996: "sae reserved",
            997: "sae reserved",
            998: "sae reserved",
            999: "sae reserved"
        },
        4082: {
            0: "deactivate",
            1: "activate",
            2: "reserved",
            3: "unavailable"
        },
        4083: {
            0: "off",
            1: "on",
            2: "error",
            3: "unavailable"
        },
        4097: {
            0: "   not active",
            1: "   active",
            2: "   reserved for sae assignment",
            3: "   not available"
        },
        4098: {
            0: "   not active",
            1: "   active",
            2: "   reserved for sae assignment",
            3: "   not available"
        },
        4175: {
            0: " not active",
            1: " active - forced by switch (see spn 3696)",
            2: " active - forced by service tool ",
            3: " reserved for sae assignment",
            4: " reserved for sae assignment",
            5: " reserved for sae assignment",
            6: " reserved for sae assignment",
            7: " not available"
        },
        4176: {
            0: "not blanked",
            1: "blanked",
            2: "error",
            3: "not supported"
        },
        4178: {
            0: "  transmission service indicator is off",
            1: "  transmission service indicator is on continuously",
            2: "  transmission service indicator is flashing",
            3: "  not available"
        },
        4195: {
            0: "temperature control mode",
            1: "position control mode (used for service only)",
            2: "reserved",
            3: "don't care/reserved"
        },
        4198: {
            0: "temperature control mode",
            1: "position control mode (used for service only)",
            2: "reserved",
            3: "don't care/reserved"
        },
        4203: {
            0: "not ok",
            1: "ok",
            2: "error",
            3: "not available"
        },
        4204: {
            0: "not ok",
            1: "ok",
            2: "error",
            3: "not available"
        },
        4205: {
            0: "not ok",
            1: "ok",
            2: "error",
            3: "not available"
        },
        4233: {
            0: "   road speed limiting by retarder is disabled",
            1: "   road speed limiting by retarder is enabled.  the road speed that the retarder will limit will be determined when the switch transitions from 00 to 01",
            2: "   error indicator",
            3: "   not available"
        },
        4234: {
            0: "   road speed is below threshold ",
            1: "   road speed is above threshold and retarder is allowed to be activated",
            2: "   reserved",
            3: "   don't care/take no action"
        },
        4240: {
            0: "open loop - has not yet satisfied conditions to go closed loop",
            1: "closed loop - using oxygen sensor(s) as feedback for fuel control",
            10: "reserved",
            11: "reserved",
            12: "reserved",
            13: "reserved",
            14: "reserved",
            15: "not supported",
            2: "open loop due to driving conditions (e.g., power enrichment, deceleration enleanment)",
            3: "open loop - due to detected system fault ",
            4: "closed loop, but fault with at least one oxygen sensor - may be using single oxygen sensor for fuel control",
            5: "reserved",
            6: "reserved",
            7: "reserved",
            8: "reserved",
            9: "reserved"
        },
        4241: {
            0: "open loop - has not yet satisfied conditions to go closed loop",
            1: "closed loop - using oxygen sensor(s) as feedback for fuel control",
            10: "reserved",
            11: "reserved",
            12: "reserved",
            13: "reserved",
            14: "reserved",
            15: "not supported",
            2: "open loop due to driving conditions (e.g., power enrichment, deceleration enleanment)",
            3: "open loop - due to detected system fault ",
            4: "closed loop, but fault with at least one oxygen sensor - may be using single oxygen sensor for fuel control",
            5: "reserved",
            6: "reserved",
            7: "reserved",
            8: "reserved",
            9: "reserved"
        },
        4242: {
            0: "allow shifts into reverse gear",
            1: "inhibit shifts into reverse gear, and shift transmission to neutral if already in reverse or attempting to shift to reverse",
            2: "reserved ",
            3: "take no action"
        },
        4246: {
            0: "  disable",
            1: "  enable",
            2: "  reserved",
            3: "  take no action"
        },
        4247: {
            0: "  disable",
            1: "  enable",
            2: "  reserved",
            3: "  take no action"
        },
        4248: {
            0: "  disable",
            1: "  enable",
            2: "  reserved",
            3: "  take no action"
        },
        4249: {
            0: "  disable",
            1: "  enable",
            2: "  reserved",
            3: "  take no action"
        },
        4250: {
            0: "  disable",
            1: "  enable",
            2: "  reserved",
            3: "  take no action"
        },
        4251: {
            0: "  disable",
            1: "  enable",
            2: "  reserved",
            3: "  take no action"
        },
        4252: {
            0: "  disable",
            1: "  enable",
            2: "  reserved",
            3: "  take no action"
        },
        4253: {
            0: "  disable",
            1: "  enable",
            2: "  reserved",
            3: "  take no action"
        },
        4255: {
            0: "no specific launch gear requested; use default launch gear",
            1: "launch the vehicle in 1st gear",
            10: "launch the vehicle in reverse 2",
            11: "launch the vehicle in reverse 3",
            12: "launch the vehicle in reverse 4 ",
            13: "reserved",
            14: "error",
            15: "not available",
            2: "launch the vehicle in 2nd gear",
            3: "launch the vehicle in 3rd gear",
            4: "launch the vehicle in 4th gear",
            5: "launch the vehicle in 5th gear",
            6: "launch the vehicle in 6th gear",
            7: "launch the vehicle in 7th gear",
            8: "launch the vehicle in 8th gear",
            9: "launch the vehicle in reverse 1"
        },
        4261: {
            0: "reverse gear shifts are currently allowed",
            1: "reverse gear shifts are currently inhibited",
            2: "error",
            3: "not available"
        },
        4307: {
            0: "not limited",
            1: "operator limited/controlled (request cannot be implemented)",
            2: "limited high (only lower command values result in change)",
            3: "limited low (only high command values result in change)",
            4: "reserved",
            5: "reserved",
            6: "non-recoverable fault",
            7: "not available (parameter not supported)"
        },
        4308: {
            0: "wheel-based speed",
            1: "ground-based speed",
            2: "navigation-based speed",
            3: "'blended' speed",
            4: "'simulated' speed",
            5: "reserved",
            6: "reserved",
            7: "not available"
        },
        4309: {
            0: "reverse",
            1: "forward",
            2: "error indication",
            3: "not available"
        },
        4312: {
            0: "reverse",
            1: "forward",
            2: "reserved",
            3: "don't care, take no action"
        },
        4314: {
            0: "first revision (no suffix)",
            1: "second revision (suffix a)",
            2: "third revision (suffix b)",
            3: "fourth revision (suffix c)",
            4: "reserved",
            5: "reserved",
            6: "reserved",
            7: "not available"
        },
        4315: {
            0: "non-certified laboratory/ self certification",
            1: "eu certified laboratory",
            2: "na certified laboratory",
            3: "reserved",
            4: "reserved",
            5: "reserved",
            6: "reserved",
            7: "not available (not certified)"
        },
        4317: {
            0: "    not certified ",
            1: "    certified"
        },
        4318: {
            0: "    not certified ",
            1: "    certified"
        },
        4319: {
            0: "    not certified ",
            1: "    certified"
        },
        4320: {
            0: "    not certified ",
            1: "    certified"
        },
        4321: {
            0: "    not certified ",
            1: "    certified"
        },
        4322: {
            0: "    not certified ",
            1: "    certified"
        },
        4323: {
            0: "    not certified ",
            1: "    certified"
        },
        4324: {
            0: "    not certified ",
            1: "    certified"
        },
        4325: {
            0: "    not certified ",
            1: "    certified"
        },
        4326: {
            0: "    not certified ",
            1: "    certified"
        },
        4327: {
            0: "    not certified ",
            1: "    certified"
        },
        4328: {
            0: "    not certified ",
            1: "    certified"
        },
        4329: {
            0: "    not certified ",
            1: "    certified"
        },
        4332: {
            0: "    dormant (sleep mode)",
            1: "    preparing dosing readiness (wake up; prepare to operate; wait for start)",
            10: "    service test mode, dosing not allowed",
            11: " reserved for future assignment by sae",
            12: " reserved for future assignment by sae",
            13: " reserved for future assignment by sae",
            14: "    error",
            15: "    not available",
            2: "    normal dosing operation",
            3: "    system error pending",
            4: "    reserved for future assignment by sae",
            5: "    protect mode against heat  (pressure buildup)",
            6: "    protect mode against cold  (defreeze)",
            7: "    shutoff  (wait for afterrun)",
            8: "    diagnosis (afterrun)",
            9: "    service test mode, dosing allowed "
        },
        4338: {
            0: "no request ",
            1: "reduction request stage 1",
            2: "reduction request stage 2",
            3: "reserved for future assignment by sae",
            4: "reserved for future assignment by sae",
            5: "reserved for future assignment by sae",
            6: "error",
            7: "not available"
        },
        4339: {
            0: " open loop control active ",
            1: " closed loop control active ",
            2: " reserved for future assignment by sae ",
            3: " reserved for future assignment by sae ",
            4: " reserved for future assignment by sae ",
            5: " reserved for future assignment by sae ",
            6: " error ",
            7: " not available"
        },
        4340: {
            0: " heater inactive ",
            1: " heater active ",
            2: " error ",
            3: " not available"
        },
        4342: {
            0: " heater inactive ",
            1: " heater active ",
            2: " error ",
            3: " not available"
        },
        4344: {
            0: " heater inactive ",
            1: " heater active ",
            2: " error ",
            3: " not available"
        },
        4346: {
            0: " heater inactive ",
            1: " heater active ",
            2: " error ",
            3: " not available"
        },
        4349: {
            0: " dormant (sleep mode) ",
            1: " preparing dosing readiness (wake up; prepare to operate; wait for start) ",
            10: " service test mode, dosing not allowed",
            11: " reserved for future assignment by sae",
            12: " reserved for future assignment by sae",
            13: " reserved for future assignment by sae",
            14: " error",
            15: " not available",
            2: " normal dosing operation ",
            3: " system error pending ",
            4: " reserved for future assignment by sae ",
            5: " protect mode against heat (pressure buildup) ",
            6: " protect mode against cold (defreeze) ",
            7: " shutoff (wait for afterrun) ",
            8: " diagnosis (afterrun) ",
            9: " service test mode, dosing allowed "
        },
        4352: {
            0: "    error suppression off",
            1: "    error suppression on",
            2: "    reserved",
            3: "    not available"
        },
        4353: {
            0: "   heater off",
            1: "   heater economy mode",
            2: "   heater automatic mode",
            3: "   heater on",
            4: "    reserved",
            5: "    reserved",
            6: "    reserved",
            7: "   not available"
        },
        4354: {
            0: " request for heater to be inactive ",
            1: " request for heater to be active ",
            2: " reserved for future assignment by sae ",
            3: " not available"
        },
        4355: {
            0: " request for heater to be inactive ",
            1: " request for heater to be active ",
            2: " reserved for future assignment by sae ",
            3: " not available"
        },
        4356: {
            0: " request for heater to be inactive ",
            1: " request for heater to be active ",
            2: " reserved for future assignment by sae ",
            3: " not available"
        },
        4357: {
            0: " request for heater to be inactive ",
            1: " request for heater to be active ",
            2: " reserved for future assignment by sae ",
            3: " not available"
        },
        4379: {
            0: "reading is not stable",
            1: "reading is stable",
            2: "error",
            3: "not available"
        },
        4380: {
            0: "power is not in range",
            1: "power is in range",
            2: "error",
            3: "not available"
        },
        4381: {
            0: "heater element is not in range",
            1: "heater element is in range",
            2: "error",
            3: "not available"
        },
        4383: {
            0: "automatic ",
            1: "preheat 2 ",
            2: "preheat 1 ",
            3: "heater off or not available"
        },
        4385: {
            0: "    dormant (sleep mode)",
            1: "    preparing dosing readiness (wake up; prepare to operate; wait for start)",
            10: "    service test mode, dosing not allowed",
            11: " reserved for future assignment by sae",
            12: " reserved for future assignment by sae",
            13: " reserved for future assignment by sae",
            14: "    error",
            15: "    not available",
            2: "    normal dosing operation",
            3: "    system error pending",
            4: "    reserved for future assignment by sae",
            5: "    protect mode against heat (pressure buildup) ",
            6: "    protect mode against cold (defreeze) ",
            7: "    shutoff (wait for afterrun) ",
            8: "    diagnosis (afterrun)",
            9: "    service test mode, dosing allowed "
        },
        4391: {
            0: "no request ",
            1: "reduction request stage 1",
            2: "reduction request stage 2",
            3: "reserved for future assignment by sae",
            4: "reserved for future assignment by sae",
            5: "reserved for future assignment by sae",
            6: "error",
            7: "not available"
        },
        4392: {
            0: " open loop control active ",
            1: " closed loop control active ",
            2: " reserved for future assignment by sae ",
            3: " reserved for future assignment by sae ",
            4: " reserved for future assignment by sae ",
            5: " reserved for future assignment by sae ",
            6: " error ",
            7: " not available"
        },
        4393: {
            0: " heater inactive ",
            1: " heater active ",
            2: " error ",
            3: " not available"
        },
        4395: {
            0: " heater inactive ",
            1: " heater active ",
            2: " error ",
            3: " not available"
        },
        4397: {
            0: " heater inactive ",
            1: " heater active ",
            2: " error ",
            3: " not available"
        },
        4399: {
            0: " heater inactive ",
            1: " heater active ",
            2: " error ",
            3: " not available"
        },
        4402: {
            0: " dormant (sleep mode) ",
            1: " preparing dosing readiness (wake up; prepare to operate; wait for start) ",
            10: " service test mode, dosing not allowed",
            11: " reserved for future assignment by sae",
            12: " reserved for future assignment by sae",
            13: " reserved for future assignment by sae",
            14: " error",
            15: " not available",
            2: " normal dosing operation ",
            3: " system error pending ",
            4: " reserved for future assignment by sae ",
            5: " protect mode against heat (pressure buildup) ",
            6: " protect mode against cold (defreeze) ",
            7: " shutoff (wait for afterrun) ",
            8: " diagnosis (afterrun)",
            9: " service test mode, dosing allowed "
        },
        4405: {
            0: "    error suppression off",
            1: "    error suppression on",
            2: "    reserved",
            3: "    not available"
        },
        4406: {
            0: "   heater off",
            1: "   heater economy mode",
            2: "   heater automatic mode",
            3: "   heater on",
            4: "    reserved",
            5: "    reserved",
            6: "    reserved",
            7: "   not available"
        },
        4407: {
            0: " request for heater to be inactive ",
            1: " request for heater to be active ",
            2: " reserved for future assignment by sae ",
            3: " not available"
        },
        4408: {
            0: " request for heater to be inactive ",
            1: " request for heater to be active ",
            2: " reserved for future assignment by sae ",
            3: " not available"
        },
        4409: {
            0: " request for heater to be inactive ",
            1: " request for heater to be active ",
            2: " reserved for future assignment by sae ",
            3: " not available"
        },
        4410: {
            0: " request for heater to be inactive ",
            1: " request for heater to be active ",
            2: " reserved for future assignment by sae ",
            3: " not available"
        },
        4425: {
            0: "urea concentration too high",
            1: "urea concentration too low",
            13: "not able to determine diesel exhaust fluid property (fluid type unknown)",
            14: "error with diesel exhaust fluid property detection",
            15: "not available",
            2: "fluid is diesel",
            3: "diesel exhaust fluid is proper mixture",
            5: " to 1100 - reserved for sae assignment"
        },
        4445: {
            0: "reading is not stable",
            1: "reading is stable",
            2: "error",
            3: "not available"
        },
        4446: {
            0: "heater element is not in range",
            1: "heater element is in range",
            2: "error",
            3: "not available"
        },
        4447: {
            0: "heater element is not in range",
            1: "heater element is in range",
            2: "error",
            3: "not available"
        },
        4449: {
            0: "automatic ",
            1: "preheat 2 ",
            2: "preheat 1 ",
            3: "heater off or not available"
        },
        4454: {
            0: " ramp retracted",
            1: " ramp extended",
            2: " error",
            3: " not available"
        },
        4455: {
            0: " ramp disabled",
            1: " ramp enabled ",
            2: " error ",
            3: " not available"
        },
        4456: {
            0: " ramp not being moved",
            1: " ramp being moved",
            2: " error",
            3: " not available"
        },
        4457: {
            0: " ramp retracted",
            1: " ramp extended",
            2: " error",
            3: " not available"
        },
        4458: {
            0: " ramp disabled",
            1: " ramp enabled ",
            2: " error ",
            3: " not available"
        },
        4459: {
            0: " ramp not being moved",
            1: " ramp being moved",
            2: " error",
            3: " not available"
        },
        4460: {
            0: "off",
            1: "on",
            2: "blinking",
            3: "not available or not installed"
        },
        4461: {
            0: "off",
            1: "on",
            2: "blinking",
            3: "not available or not installed"
        },
        4462: {
            0: "off",
            1: "on",
            2: "blinking",
            3: "not available or not installed"
        },
        4463: {
            0: "off",
            1: "on",
            2: "blinking",
            3: "not available or not installed"
        },
        4464: {
            0: "off",
            1: "on",
            2: "blinking",
            3: "not available or not installed"
        },
        4465: {
            0: "off",
            1: "on",
            2: "blinking",
            3: "not available or not installed"
        },
        4466: {
            0: "off",
            1: "on",
            2: "blinking",
            3: "not available or not installed"
        },
        4467: {
            0: "off",
            1: "on",
            2: "blinking",
            3: "not available or not installed"
        },
        4468: {
            0: "off",
            1: "on",
            2: "blinking",
            3: "not available or not installed"
        },
        4469: {
            0: "off",
            1: "on",
            2: "blinking",
            3: "not available or not installed"
        },
        4470: {
            0: "off",
            1: "on",
            2: "blinking",
            3: "not available or not installed"
        },
        4471: {
            0: "off",
            1: "on",
            2: "blinking",
            3: "not available or not installed"
        },
        4472: {
            0: "off",
            1: "on",
            2: "blinking",
            3: "not available or not installed"
        },
        4473: {
            0: "off",
            1: "on",
            2: "blinking",
            3: "not available or not installed"
        },
        4474: {
            0: "off",
            1: "on",
            2: "blinking",
            3: "not available or not installed"
        },
        4475: {
            0: "off",
            1: "on",
            2: "blinking",
            3: "not available or not installed"
        },
        4476: {
            0: "off",
            1: "on",
            2: "blinking",
            3: "not available or not installed"
        },
        4477: {
            0: "off",
            1: "on",
            2: "blinking",
            3: "not available or not installed"
        },
        4478: {
            0: "off",
            1: "on",
            2: "blinking",
            3: "not available or not installed"
        },
        4479: {
            0: "off",
            1: "on",
            2: "blinking",
            3: "not available or not installed"
        },
        4480: {
            0: "off",
            1: "on",
            2: "blinking",
            3: "not available or not installed"
        },
        4481: {
            0: "off",
            1: "on",
            2: "blinking",
            3: "not available or not installed"
        },
        4482: {
            0: "off",
            1: "on",
            2: "blinking",
            3: "not available or not installed"
        },
        4483: {
            0: "off",
            1: "on",
            2: "blinking",
            3: "not available or not installed"
        },
        4484: {
            0: "off",
            1: "on",
            2: "blinking",
            3: "not available or not installed"
        },
        4485: {
            0: "off",
            1: "on",
            2: "blinking",
            3: "not available or not installed"
        },
        4486: {
            0: "off",
            1: "on",
            2: "blinking",
            3: "not available or not installed"
        },
        4487: {
            0: "off",
            1: "on",
            2: "blinking",
            3: "not available or not installed"
        },
        4488: {
            0: "off",
            1: "on",
            2: "blinking",
            3: "not available or not installed"
        },
        4489: {
            0: "off",
            1: "on",
            2: "blinking",
            3: "not available or not installed"
        },
        4491: {
            0: " not in neutral position",
            1: " in neutral position",
            2: " error indicator",
            3: " not available"
        },
        4492: {
            0: " not on negative side of neutral",
            1: " on negative side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4493: {
            0: " not on positive side of neutral",
            1: " on positive side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4495: {
            0: " not in neutral position",
            1: " in neutral position",
            2: " error indicator",
            3: " not available"
        },
        4496: {
            0: " not on negative side of neutral",
            1: " on negative side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4497: {
            0: " not on positive side of neutral",
            1: " on positive side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4499: {
            0: " not in the detent position",
            1: " in the detent position",
            2: " error indicator",
            3: " not available"
        },
        4500: {
            0: " not in the detent position",
            1: " in the detent position",
            2: " error indicator",
            3: " not available"
        },
        4501: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4502: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4503: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4504: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4505: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4506: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4507: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4508: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4509: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4510: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4511: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4512: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4513: {
            0: " not in neutral position",
            1: " in neutral position",
            2: " error indicator",
            3: " not available"
        },
        4514: {
            0: " not on negative side of neutral",
            1: " on negative side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4515: {
            0: " not on positive side of neutral",
            1: " on positive side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4517: {
            0: " not in neutral position",
            1: " in neutral position",
            2: " error indicator",
            3: " not available"
        },
        4518: {
            0: " not on negative side of neutral",
            1: " on negative side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4519: {
            0: " not on positive side of neutral",
            1: " on positive side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4521: {
            0: " not in neutral position",
            1: " in neutral position",
            2: " error indicator",
            3: " not available"
        },
        4522: {
            0: " not on negative side of neutral",
            1: " on negative side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4523: {
            0: " not on positive side of neutral",
            1: " on positive side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4525: {
            0: " not in the detent position",
            1: " in the detent position",
            2: " error indicator",
            3: " not available"
        },
        4526: {
            0: " not in the detent position",
            1: " in the detent position",
            2: " error indicator",
            3: " not available"
        },
        4527: {
            0: " not in the detent position",
            1: " in the detent position",
            2: " error indicator",
            3: " not available"
        },
        4528: {
            0: " not in neutral position",
            1: " in neutral position",
            2: " error indicator",
            3: " not available"
        },
        4529: {
            0: " not on negative side of neutral",
            1: " on negative side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4530: {
            0: " not on positive side of neutral",
            1: " on positive side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4532: {
            0: " not in neutral position",
            1: " in neutral position",
            2: " error indicator",
            3: " not available"
        },
        4533: {
            0: " not on negative side of neutral",
            1: " on negative side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4534: {
            0: " not on positive side of neutral",
            1: " on positive side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4536: {
            0: " not in the detent position",
            1: " in the detent position",
            2: " error indicator",
            3: " not available"
        },
        4537: {
            0: " not in the detent position",
            1: " in the detent position",
            2: " error indicator",
            3: " not available"
        },
        4538: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4539: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4540: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4541: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4542: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4543: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4544: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4545: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4546: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4547: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4548: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4549: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4550: {
            0: " not in neutral position",
            1: " in neutral position",
            2: " error indicator",
            3: " not available"
        },
        4551: {
            0: " not on negative side of neutral",
            1: " on negative side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4552: {
            0: " not on positive side of neutral",
            1: " on positive side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4554: {
            0: " not in neutral position",
            1: " in neutral position",
            2: " error indicator",
            3: " not available"
        },
        4555: {
            0: " not on negative side of neutral",
            1: " on negative side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4556: {
            0: " not on positive side of neutral",
            1: " on positive side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4558: {
            0: " not in neutral position",
            1: " in neutral position",
            2: " error indicator",
            3: " not available"
        },
        4559: {
            0: " not on negative side of neutral",
            1: " on negative side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4560: {
            0: " not on positive side of neutral",
            1: " on positive side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4562: {
            0: " not in the detent position",
            1: " in the detent position",
            2: " error indicator",
            3: " not available"
        },
        4563: {
            0: " not in the detent position",
            1: " in the detent position",
            2: " error indicator",
            3: " not available"
        },
        4564: {
            0: " not in the detent position",
            1: " in the detent position",
            2: " error indicator",
            3: " not available"
        },
        4565: {
            0: " not in neutral position",
            1: " in neutral position",
            2: " error indicator",
            3: " not available"
        },
        4566: {
            0: " not on negative side of neutral",
            1: " on negative side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4567: {
            0: " not on positive side of neutral",
            1: " on positive side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4569: {
            0: " not in neutral position",
            1: " in neutral position",
            2: " error indicator",
            3: " not available"
        },
        4570: {
            0: " not on negative side of neutral",
            1: " on negative side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4571: {
            0: " not on positive side of neutral",
            1: " on positive side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4573: {
            0: " not in the detent position",
            1: " in the detent position",
            2: " error indicator",
            3: " not available"
        },
        4574: {
            0: " not in the detent position",
            1: " in the detent position",
            2: " error indicator",
            3: " not available"
        },
        4575: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4576: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4577: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4578: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4579: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4580: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4581: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4582: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4583: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4584: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4585: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4586: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4587: {
            0: " not in neutral position",
            1: " in neutral position",
            2: " error indicator",
            3: " not available"
        },
        4588: {
            0: " not on negative side of neutral",
            1: " on negative side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4589: {
            0: " not on positive side of neutral",
            1: " on positive side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4591: {
            0: " not in neutral position",
            1: " in neutral position",
            2: " error indicator",
            3: " not available"
        },
        4592: {
            0: " not on negative side of neutral",
            1: " on negative side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4593: {
            0: " not on positive side of neutral",
            1: " on positive side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4595: {
            0: " not in neutral position",
            1: " in neutral position",
            2: " error indicator",
            3: " not available"
        },
        4596: {
            0: " not on negative side of neutral",
            1: " on negative side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4597: {
            0: " not on positive side of neutral",
            1: " on positive side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4599: {
            0: " not in the detent position",
            1: " in the detent position",
            2: " error indicator",
            3: " not available"
        },
        4600: {
            0: " not in the detent position",
            1: " in the detent position",
            2: " error indicator",
            3: " not available"
        },
        4601: {
            0: " not in the detent position",
            1: " in the detent position",
            2: " error indicator",
            3: " not available"
        },
        4602: {
            0: " not in neutral position",
            1: " in neutral position",
            2: " error indicator",
            3: " not available"
        },
        4603: {
            0: " not on negative side of neutral",
            1: " on negative side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4604: {
            0: " not on positive side of neutral",
            1: " on positive side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4606: {
            0: " not in neutral position",
            1: " in neutral position",
            2: " error indicator",
            3: " not available"
        },
        4607: {
            0: " not on negative side of neutral",
            1: " on negative side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4608: {
            0: " not on positive side of neutral",
            1: " on positive side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4610: {
            0: " not in the detent position",
            1: " in the detent position",
            2: " error indicator",
            3: " not available"
        },
        4611: {
            0: " not in the detent position",
            1: " in the detent position",
            2: " error indicator",
            3: " not available"
        },
        4612: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4613: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4614: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4615: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4616: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4617: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4618: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4619: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4620: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4621: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4622: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4623: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4624: {
            0: " not in neutral position",
            1: " in neutral position",
            2: " error indicator",
            3: " not available"
        },
        4625: {
            0: " not on negative side of neutral",
            1: " on negative side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4626: {
            0: " not on positive side of neutral",
            1: " on positive side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4628: {
            0: " not in neutral position",
            1: " in neutral position",
            2: " error indicator",
            3: " not available"
        },
        4629: {
            0: " not on negative side of neutral",
            1: " on negative side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4630: {
            0: " not on positive side of neutral",
            1: " on positive side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4632: {
            0: " not in neutral position",
            1: " in neutral position",
            2: " error indicator",
            3: " not available"
        },
        4633: {
            0: " not on negative side of neutral",
            1: " on negative side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4634: {
            0: " not on positive side of neutral",
            1: " on positive side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4636: {
            0: " not in the detent position",
            1: " in the detent position",
            2: " error indicator",
            3: " not available"
        },
        4637: {
            0: " not in the detent position",
            1: " in the detent position",
            2: " error indicator",
            3: " not available"
        },
        4638: {
            0: " not in the detent position",
            1: " in the detent position",
            2: " error indicator",
            3: " not available"
        },
        4639: {
            0: " not in neutral position",
            1: " in neutral position",
            2: " error indicator",
            3: " not available"
        },
        4640: {
            0: " not on negative side of neutral",
            1: " on negative side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4641: {
            0: " not on positive side of neutral",
            1: " on positive side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4643: {
            0: " not in neutral position",
            1: " in neutral position",
            2: " error indicator",
            3: " not available"
        },
        4644: {
            0: " not on negative side of neutral",
            1: " on negative side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4645: {
            0: " not on positive side of neutral",
            1: " on positive side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4647: {
            0: " not in the detent position",
            1: " in the detent position",
            2: " error indicator",
            3: " not available"
        },
        4648: {
            0: " not in the detent position",
            1: " in the detent position",
            2: " error indicator",
            3: " not available"
        },
        4649: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4650: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4651: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4652: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4653: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4654: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4655: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4656: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4657: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4658: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4659: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4660: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4661: {
            0: " not in neutral position",
            1: " in neutral position",
            2: " error indicator",
            3: " not available"
        },
        4662: {
            0: " not on negative side of neutral",
            1: " on negative side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4663: {
            0: " not on positive side of neutral",
            1: " on positive side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4665: {
            0: " not in neutral position",
            1: " in neutral position",
            2: " error indicator",
            3: " not available"
        },
        4666: {
            0: " not on negative side of neutral",
            1: " on negative side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4667: {
            0: " not on positive side of neutral",
            1: " on positive side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4669: {
            0: " not in neutral position",
            1: " in neutral position",
            2: " error indicator",
            3: " not available"
        },
        4670: {
            0: " not on negative side of neutral",
            1: " on negative side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4671: {
            0: " not on positive side of neutral",
            1: " on positive side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4673: {
            0: " not in the detent position",
            1: " in the detent position",
            2: " error indicator",
            3: " not available"
        },
        4674: {
            0: " not in the detent position",
            1: " in the detent position",
            2: " error indicator",
            3: " not available"
        },
        4675: {
            0: " not in the detent position",
            1: " in the detent position",
            2: " error indicator",
            3: " not available"
        },
        4676: {
            0: " not in neutral position",
            1: " in neutral position",
            2: " error indicator",
            3: " not available"
        },
        4677: {
            0: " not on negative side of neutral",
            1: " on negative side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4678: {
            0: " not on positive side of neutral",
            1: " on positive side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4680: {
            0: " not in neutral position",
            1: " in neutral position",
            2: " error indicator",
            3: " not available"
        },
        4681: {
            0: " not on negative side of neutral",
            1: " on negative side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4682: {
            0: " not on positive side of neutral",
            1: " on positive side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4684: {
            0: " not in the detent position",
            1: " in the detent position",
            2: " error indicator",
            3: " not available"
        },
        4685: {
            0: " not in the detent position",
            1: " in the detent position",
            2: " error indicator",
            3: " not available"
        },
        4686: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4687: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4688: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4689: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4690: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4691: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4692: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4693: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4694: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4695: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4696: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4697: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4698: {
            0: " not in neutral position",
            1: " in neutral position",
            2: " error indicator",
            3: " not available"
        },
        4699: {
            0: " not on negative side of neutral",
            1: " on negative side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4700: {
            0: " not on positive side of neutral",
            1: " on positive side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4702: {
            0: " not in neutral position",
            1: " in neutral position",
            2: " error indicator",
            3: " not available"
        },
        4703: {
            0: " not on negative side of neutral",
            1: " on negative side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4704: {
            0: " not on positive side of neutral",
            1: " on positive side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4706: {
            0: " not in neutral position",
            1: " in neutral position",
            2: " error indicator",
            3: " not available"
        },
        4707: {
            0: " not on negative side of neutral",
            1: " on negative side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4708: {
            0: " not on positive side of neutral",
            1: " on positive side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4710: {
            0: " not in the detent position",
            1: " in the detent position",
            2: " error indicator",
            3: " not available"
        },
        4711: {
            0: " not in the detent position",
            1: " in the detent position",
            2: " error indicator",
            3: " not available"
        },
        4712: {
            0: " not in the detent position",
            1: " in the detent position",
            2: " error indicator",
            3: " not available"
        },
        4713: {
            0: " not in neutral position",
            1: " in neutral position",
            2: " error indicator",
            3: " not available"
        },
        4714: {
            0: " not on negative side of neutral",
            1: " on negative side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4715: {
            0: " not on positive side of neutral",
            1: " on positive side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4717: {
            0: " not in neutral position",
            1: " in neutral position",
            2: " error indicator",
            3: " not available"
        },
        4718: {
            0: " not on negative side of neutral",
            1: " on negative side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4719: {
            0: " not on positive side of neutral",
            1: " on positive side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4721: {
            0: " not in the detent position",
            1: " in the detent position",
            2: " error indicator",
            3: " not available"
        },
        4722: {
            0: " not in the detent position",
            1: " in the detent position",
            2: " error indicator",
            3: " not available"
        },
        4723: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4724: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4725: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4726: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4727: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4728: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4729: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4730: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4731: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4732: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4733: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4734: {
            0: " button not pressed",
            1: " button pressed",
            2: " error indicator",
            3: " not available"
        },
        4735: {
            0: " not in neutral position",
            1: " in neutral position",
            2: " error indicator",
            3: " not available"
        },
        4736: {
            0: " not on negative side of neutral",
            1: " on negative side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4737: {
            0: " not on positive side of neutral",
            1: " on positive side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4739: {
            0: " not in neutral position",
            1: " in neutral position",
            2: " error indicator",
            3: " not available"
        },
        4740: {
            0: " not on negative side of neutral",
            1: " on negative side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4741: {
            0: " not on positive side of neutral",
            1: " on positive side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4743: {
            0: " not in neutral position",
            1: " in neutral position",
            2: " error indicator",
            3: " not available"
        },
        4744: {
            0: " not on negative side of neutral",
            1: " on negative side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4745: {
            0: " not on positive side of neutral",
            1: " on positive side of neutral",
            2: " error indicator",
            3: " not available"
        },
        4747: {
            0: " not in the detent position",
            1: " in the detent position",
            2: " error indicator",
            3: " not available"
        },
        4748: {
            0: " not in the detent position",
            1: " in the detent position",
            2: " error indicator",
            3: " not available"
        },
        4749: {
            0: " not in the detent position",
            1: " in the detent position",
            2: " error indicator",
            3: " not available"
        },
        4815: {
            0: "off (cooling flow not needed)",
            1: "on (cooling flow demanded)",
            2: "error",
            3: "not available"
        },
        4816: {
            0: "transition is not in process",
            1: "transition is in process",
            2: "error",
            3: "not available"
        },
        4918: {
            0: "acceleration sensor",
            1: "pressure sensor",
            10: "reserved",
            11: "reserved",
            12: "reserved",
            13: "reserved",
            14: "error - type of sensor cannot be determined (unknown sensor)",
            15: "not available",
            2: "rollrate sensor",
            3: "sonic sensor",
            4: "fiber-optic sensor",
            5: "radar sensor",
            6: "lidar sensor",
            7: "optical sensor",
            8: "reserved",
            9: "reserved"
        },
        4919: {
            0: "acceleration sensor",
            1: "pressure sensor",
            10: "reserved",
            11: "reserved",
            12: "reserved",
            13: "reserved",
            14: "error - type of sensor cannot be determined (unknown sensor)",
            15: "not available",
            2: "rollrate sensor",
            3: "sonic sensor",
            4: "fiber-optic sensor",
            5: "radar sensor",
            6: "lidar sensor",
            7: "optical sensor",
            8: "reserved",
            9: "reserved"
        },
        4920: {
            0: "acceleration sensor",
            1: "pressure sensor",
            10: "reserved",
            11: "reserved",
            12: "reserved",
            13: "reserved",
            14: "error - type of sensor cannot be determined (unknown sensor)",
            15: "not available",
            2: "rollrate sensor",
            3: "sonic sensor",
            4: "fiber-optic sensor",
            5: "radar sensor",
            6: "lidar sensor",
            7: "optical sensor",
            8: "reserved",
            9: "reserved"
        },
        4921: {
            0: "acceleration sensor",
            1: "pressure sensor",
            10: "reserved",
            11: "reserved",
            12: "reserved",
            13: "reserved",
            14: "error - type of sensor cannot be determined (unknown sensor)",
            15: "not available",
            2: "rollrate sensor",
            3: "sonic sensor",
            4: "fiber-optic sensor",
            5: "radar sensor",
            6: "lidar sensor",
            7: "optical sensor",
            8: "reserved",
            9: "reserved"
        },
        4922: {
            0: "acceleration sensor",
            1: "pressure sensor",
            10: "reserved",
            11: "reserved",
            12: "reserved",
            13: "reserved",
            14: "error - type of sensor cannot be determined (unknown sensor)",
            15: "not available",
            2: "rollrate sensor",
            3: "sonic sensor",
            4: "fiber-optic sensor",
            5: "radar sensor",
            6: "lidar sensor",
            7: "optical sensor",
            8: "reserved",
            9: "reserved"
        },
        4923: {
            0: "acceleration sensor",
            1: "pressure sensor",
            10: "reserved",
            11: "reserved",
            12: "reserved",
            13: "reserved",
            14: "error - type of sensor cannot be determined (unknown sensor)",
            15: "not available",
            2: "rollrate sensor",
            3: "sonic sensor",
            4: "fiber-optic sensor",
            5: "radar sensor",
            6: "lidar sensor",
            7: "optical sensor",
            8: "reserved",
            9: "reserved"
        },
        4924: {
            0: "acceleration sensor",
            1: "pressure sensor",
            10: "reserved",
            11: "reserved",
            12: "reserved",
            13: "reserved",
            14: "error - type of sensor cannot be determined (unknown sensor)",
            15: "not available",
            2: "rollrate sensor",
            3: "sonic sensor",
            4: "fiber-optic sensor",
            5: "radar sensor",
            6: "lidar sensor",
            7: "optical sensor",
            8: "reserved",
            9: "reserved"
        },
        4925: {
            0: "acceleration sensor",
            1: "pressure sensor",
            10: "reserved",
            11: "reserved",
            12: "reserved",
            13: "reserved",
            14: "error - type of sensor cannot be determined (unknown sensor)",
            15: "not available",
            2: "rollrate sensor",
            3: "sonic sensor",
            4: "fiber-optic sensor",
            5: "radar sensor",
            6: "lidar sensor",
            7: "optical sensor",
            8: "reserved",
            9: "reserved"
        },
        4926: {
            0: "acceleration sensor",
            1: "pressure sensor",
            10: "reserved",
            11: "reserved",
            12: "reserved",
            13: "reserved",
            14: "error - type of sensor cannot be determined (unknown sensor)",
            15: "not available",
            2: "rollrate sensor",
            3: "sonic sensor",
            4: "fiber-optic sensor",
            5: "radar sensor",
            6: "lidar sensor",
            7: "optical sensor",
            8: "reserved",
            9: "reserved"
        },
        4927: {
            0: "acceleration sensor",
            1: "pressure sensor",
            10: "reserved",
            11: "reserved",
            12: "reserved",
            13: "reserved",
            14: "error - type of sensor cannot be determined (unknown sensor)",
            15: "not available",
            2: "rollrate sensor",
            3: "sonic sensor",
            4: "fiber-optic sensor",
            5: "radar sensor",
            6: "lidar sensor",
            7: "optical sensor",
            8: "reserved",
            9: "reserved"
        },
        4928: {
            0: "acceleration sensor",
            1: "pressure sensor",
            10: "reserved",
            11: "reserved",
            12: "reserved",
            13: "reserved",
            14: "error - type of sensor cannot be determined (unknown sensor)",
            15: "not available",
            2: "rollrate sensor",
            3: "sonic sensor",
            4: "fiber-optic sensor",
            5: "radar sensor",
            6: "lidar sensor",
            7: "optical sensor",
            8: "reserved",
            9: "reserved"
        },
        4929: {
            0: "acceleration sensor",
            1: "pressure sensor",
            10: "reserved",
            11: "reserved",
            12: "reserved",
            13: "reserved",
            14: "error - type of sensor cannot be determined (unknown sensor)",
            15: "not available",
            2: "rollrate sensor",
            3: "sonic sensor",
            4: "fiber-optic sensor",
            5: "radar sensor",
            6: "lidar sensor",
            7: "optical sensor",
            8: "reserved",
            9: "reserved"
        },
        4930: {
            0: "acceleration sensor",
            1: "pressure sensor",
            10: "reserved",
            11: "reserved",
            12: "reserved",
            13: "reserved",
            14: "error - type of sensor cannot be determined (unknown sensor)",
            15: "not available",
            2: "rollrate sensor",
            3: "sonic sensor",
            4: "fiber-optic sensor",
            5: "radar sensor",
            6: "lidar sensor",
            7: "optical sensor",
            8: "reserved",
            9: "reserved"
        },
        4931: {
            0: "acceleration sensor",
            1: "pressure sensor",
            10: "reserved",
            11: "reserved",
            12: "reserved",
            13: "reserved",
            14: "error - type of sensor cannot be determined (unknown sensor)",
            15: "not available",
            2: "rollrate sensor",
            3: "sonic sensor",
            4: "fiber-optic sensor",
            5: "radar sensor",
            6: "lidar sensor",
            7: "optical sensor",
            8: "reserved",
            9: "reserved"
        },
        4932: {
            0: "acceleration sensor",
            1: "pressure sensor",
            10: "reserved",
            11: "reserved",
            12: "reserved",
            13: "reserved",
            14: "error - type of sensor cannot be determined (unknown sensor)",
            15: "not available",
            2: "rollrate sensor",
            3: "sonic sensor",
            4: "fiber-optic sensor",
            5: "radar sensor",
            6: "lidar sensor",
            7: "optical sensor",
            8: "reserved",
            9: "reserved"
        },
        4933: {
            0: "acceleration sensor",
            1: "pressure sensor",
            10: "reserved",
            11: "reserved",
            12: "reserved",
            13: "reserved",
            14: "error - type of sensor cannot be determined (unknown sensor)",
            15: "not available",
            2: "rollrate sensor",
            3: "sonic sensor",
            4: "fiber-optic sensor",
            5: "radar sensor",
            6: "lidar sensor",
            7: "optical sensor",
            8: "reserved",
            9: "reserved"
        },
        4934: {
            0: "acceleration sensor",
            1: "pressure sensor",
            10: "reserved",
            11: "reserved",
            12: "reserved",
            13: "reserved",
            14: "error - type of sensor cannot be determined (unknown sensor)",
            15: "not available",
            2: "rollrate sensor",
            3: "sonic sensor",
            4: "fiber-optic sensor",
            5: "radar sensor",
            6: "lidar sensor",
            7: "optical sensor",
            8: "reserved",
            9: "reserved"
        },
        4952: {
            0: "unlocked",
            1: "locked",
            2: "error - beltlock state cannot be determined",
            3: "not available"
        },
        4953: {
            0: "unlocked",
            1: "locked",
            2: "error - beltlock state cannot be determined",
            3: "not available"
        },
        4954: {
            0: "unlocked",
            1: "locked",
            2: "error - beltlock state cannot be determined",
            3: "not available"
        },
        4955: {
            0: "unlocked",
            1: "locked",
            2: "error - beltlock state cannot be determined",
            3: "not available"
        },
        4956: {
            0: "unlocked",
            1: "locked",
            2: "error - beltlock state cannot be determined",
            3: "not available"
        },
        4957: {
            0: "unlocked",
            1: "locked",
            2: "error - beltlock state cannot be determined",
            3: "not available"
        },
        4958: {
            0: "unlocked",
            1: "locked",
            2: "error - beltlock state cannot be determined",
            3: "not available"
        },
        4959: {
            0: "unlocked",
            1: "locked",
            2: "error - beltlock state cannot be determined",
            3: "not available"
        },
        4960: {
            0: "unlocked",
            1: "locked",
            2: "error - beltlock state cannot be determined",
            3: "not available"
        },
        4961: {
            0: "unlocked",
            1: "locked",
            2: "error - beltlock state cannot be determined",
            3: "not available"
        },
        4962: {
            0: "passenger airbag activated",
            1: "passenger airbag deactivated",
            2: "error - state of deactivation switch cannot be determined",
            3: "not available"
        },
        4963: {
            0: "empty",
            1: "child",
            2: "small adult",
            3: "adult",
            4: "large adult",
            5: "reserved",
            6: "error - classification of occupant cannot be determined",
            7: "not available"
        },
        4964: {
            0: "empty",
            1: "child",
            2: "small adult",
            3: "adult",
            4: "large adult",
            5: "error - classification of occupant cannot be determined",
            7: "not available"
        },
        4965: {
            0: "empty",
            1: "occupied",
            2: "error - classification cannot be determined",
            3: "not available"
        },
        4966: {
            0: "empty",
            1: "occupied",
            2: "error - classification cannot be determined",
            3: "not available"
        },
        4967: {
            0: "empty",
            1: "occupied",
            2: "error - classification cannot be determined",
            3: "not available"
        },
        4968: {
            0: "empty",
            1: "occupied",
            2: "error - classification cannot be determined",
            3: "not available"
        },
        4969: {
            0: "empty",
            1: "occupied",
            2: "error - classification cannot be determined",
            3: "not available"
        },
        4970: {
            0: "empty",
            1: "occupied",
            2: "error - classification cannot be determined",
            3: "not available"
        },
        4971: {
            0: "empty",
            1: "occupied",
            2: "error - classification cannot be determined",
            3: "not available"
        },
        4972: {
            0: "empty",
            1: "occupied",
            2: "error - classification cannot be determined",
            3: "not available"
        },
        4978: {
            0: "on",
            1: "off",
            2: "error",
            3: "not available"
        },
        4979: {
            0: "pitch angle fully functional. data is within sensor specification.",
            1: "pitch angle degraded. data is suspect due to environmental conditions.",
            2: "error",
            3: "not available"
        },
        4980: {
            0: "on",
            1: "off",
            2: "error",
            3: "not available"
        },
        4981: {
            0: "roll angle fully functional. data is within sensor specification.",
            1: "roll angle degraded. data is suspect due to environmental conditions.",
            2: "error",
            3: "not available"
        },
        4986: {
            0: "pitch rate fully functional. data is within sensor specification.",
            1: "pitch rate degraded. data is suspect due to environmental conditions.",
            2: "error",
            3: "not available"
        },
        4987: {
            0: "roll rate fully functional. data is within sensor specification.",
            1: "roll rate degraded. data is suspect due to environmental conditions.",
            2: "error",
            3: "not available"
        },
        4988: {
            0: "yaw rate fully functional. data is within sensor specification.",
            1: "yaw rate degraded. data is suspect due to environmental conditions.",
            2: "error",
            3: "not available"
        },
        4990: {
            0: "idling. the power line or the battery is disconnected from the charger.",
            1: "charging the battery.",
            13: "battery failure. an error condition due to the battery state (high temperature, etc.)",
            14: "charger failure. an error condition due to the charger state.",
            15: "not available.",
            2: "stand-by or maintaining the battery charge.",
            3: " to 12 - reserved"
        },
        4991: {
            0: "disconnected. the power line is disconnected from the charger.",
            1: "connected. the power line is connected to the charger.",
            2: "error",
            3: "not available"
        },
        4994: {
            0: "idling. the power line or the battery is disconnected from the charger.",
            1: "charging the battery.",
            13: "battery failure. an error condition due to the battery state (high temperature, etc.)",
            14: "charger failure. an error condition due to the charger state.",
            15: "not available.",
            2: "stand-by or maintaining the battery charge.",
            3: " to 12 - reserved"
        },
        4995: {
            0: "disconnected. the power line is disconnected from the charger.",
            1: "connected. the power line is connected to the charger.",
            2: "error",
            3: "not available"
        },
        5000: {
            0: "not overheat",
            1: "overheat",
            2: "error",
            3: "not available"
        },
        5001: {
            0: "out of range ",
            1: "in range",
            2: "error",
            3: "not available"
        },
        5002: {
            0: "1",
            1: "2",
            10: "11",
            11: "12",
            12: "13",
            13: "14",
            14: "15",
            15: "16",
            16: "not used",
            17: "not used",
            18: "not used",
            19: "not used",
            2: "3",
            20: "not used",
            21: "not used",
            22: "not used",
            23: "not used",
            24: "not used",
            25: "not used",
            26: "not used",
            27: "not used",
            28: "not used",
            29: "not used",
            3: "4",
            30: "error",
            31: "not available",
            4: "5",
            5: "6",
            6: "7",
            7: "8",
            8: "9",
            9: "10"
        },
        5003: {
            0: "grapple selected",
            1: "magnet selected",
            2: "error",
            3: "not available"
        },
        5004: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available"
        },
        5005: {
            0: "not enabled",
            1: "enabled",
            2: "error",
            3: "not avaialble"
        },
        5006: {
            0: "monitor is off",
            1: "monitor is on",
            2: "error",
            3: "not available"
        },
        5007: {
            0: "duty cycle not exceed",
            1: "duty cycle exceed",
            2: "error",
            3: "not avaiable"
        },
        5010: {
            0: "don't lift",
            1: "lift",
            2: "error",
            3: "not available"
        },
        5011: {
            0: "don't drop",
            1: "drop",
            2: "error",
            3: "not available"
        },
        5012: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available"
        },
        5013: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available"
        },
        5014: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available"
        },
        5015: {
            0: " not requesting maximum power available",
            1: " momentarily requesting maximum power available",
            2: " fault",
            3: " not available"
        },
        5021: {
            0: " disabled",
            1: " supported ",
            2: " reserved",
            3: " don't care"
        },
        5022: {
            0: " no warning",
            1: " collision is imminent",
            2: " error indicator",
            3: " not available"
        },
        5023: {
            0: " driver demand for not using acc distance keeping control",
            1: " driver demand for using acc distance keeping control",
            2: " reserved",
            3: " don't care"
        },
        5057: {
            0: "take cab noise control system off line",
            1: "bring cab noise control system on line",
            2: "reserved",
            3: "don't care/not commanded"
        },
        5058: {
            0: "activate manual tuning, wait for save and exit commands",
            1: "activate automatic tuning, wait for save and exit commands",
            2: "activate automatic tuning state, saving parameters, exit on completion",
            3: "remain in tuning, save current parameters",
            4: "exit tuning, not saving parameters",
            5: "exit tuning, saving parameters",
            6: "reserved",
            7: "don't care, not commanded"
        },
        5059: {
            0: "cab noise control system on-line",
            1: "cab noise control system off-line",
            2: "error indicator",
            3: "not available/not requested"
        },
        5060: {
            0: "manual tuning mode",
            1: "manual tuning mode - command complete",
            10: "automatic tuning mode - saving data",
            11: "reserved ",
            12: "reserved",
            13: "reserved",
            14: "tuning mode - error",
            15: "not available/not tuning/not requested",
            2: "manual tuning mode - saving data",
            3: "reserved",
            4: "reserved ",
            5: "reserved",
            6: "reserved",
            7: "manual tuning denied - requires security verification",
            8: "automatic tuning mode",
            9: "automatic tuning mode - command complete"
        },
        5077: {
            0: "lamp off",
            1: "lamp on",
            2: "reserved",
            3: "not available"
        },
        5078: {
            0: "lamp off",
            1: "lamp on",
            2: "reserved",
            3: "not available"
        },
        5079: {
            0: "lamp off",
            1: "lamp on",
            2: "reserved",
            3: "not available"
        },
        5080: {
            0: "lamp off",
            1: "lamp on",
            2: "reserved",
            3: "not available"
        },
        5081: {
            0: "lamp off",
            1: "lamp on",
            2: "reserved",
            3: "not available"
        },
        5082: {
            0: "lamp off",
            1: "lamp on",
            2: "reserved",
            3: "not available"
        },
        5083: {
            0: "lamp off",
            1: "lamp on",
            2: "reserved",
            3: "not available"
        },
        5084: {
            0: "lamp off",
            1: "lamp on",
            2: "reserved",
            3: "not available"
        },
        5085: {
            0: "lamp off",
            1: "lamp on",
            2: "reserved",
            3: "not available"
        },
        5086: {
            0: "lamp off",
            1: "lamp on",
            2: "reserved",
            3: "not available"
        },
        5087: {
            0: "lamp off",
            1: "lamp on",
            2: "reserved",
            3: "not available"
        },
        5088: {
            0: "lamp off",
            1: "lamp on",
            2: "reserved",
            3: "not available"
        },
        5089: {
            0: "lamp off",
            1: "lamp on",
            2: "reserved",
            3: "not available"
        },
        5090: {
            0: "lamp off",
            1: "lamp on",
            2: "reserved",
            3: "not available"
        },
        5091: {
            0: "lamp off",
            1: "lamp on",
            2: "reserved",
            3: "not available"
        },
        5093: {
            0: "lamp deactivated (off)",
            1: "lamp activated (on)",
            2: "error",
            3: "not available"
        },
        5094: {
            0: "lamp deactivated (off)",
            1: "lamp activated (on)",
            2: "error",
            3: "not available"
        },
        5095: {
            0: "lamp deactivated (off)",
            1: "lamp activated (on)",
            2: "error",
            3: "not available"
        },
        5096: {
            0: "lamp deactivated (off)",
            1: "lamp activated (on)",
            2: "error",
            3: "not available"
        },
        5097: {
            0: "lamp deactivated (off)",
            1: "lamp activated (on)",
            2: "error",
            3: "not available"
        },
        5098: {
            0: "lamp deactivated (off)",
            1: "lamp activated (on)",
            2: "error",
            3: "not available"
        },
        5099: {
            0: "lamp deactivated (off)",
            1: "lamp activated (on)",
            2: "error",
            3: "not available"
        },
        5100: {
            0: "lamp deactivated (off)",
            1: "lamp activated (on)",
            2: "error",
            3: "not available"
        },
        5101: {
            0: "lamp deactivated (off)",
            1: "lamp activated (on)",
            2: "error",
            3: "not available"
        },
        5102: {
            0: "lamp deactivated (off)",
            1: "lamp activated (on)",
            2: "error",
            3: "not available"
        },
        5103: {
            0: "lamp deactivated (off)",
            1: "lamp activated (on)",
            2: "error",
            3: "not available"
        },
        5104: {
            0: "lamp deactivated (off)",
            1: "lamp activated (on)",
            2: "error",
            3: "not available"
        },
        5105: {
            0: "lamp deactivated (off)",
            1: "lamp activated (on)",
            2: "error",
            3: "not available"
        },
        5106: {
            0: "lamp deactivated (off)",
            1: "lamp activated (on)",
            2: "error",
            3: "not available"
        },
        5107: {
            0: "lamp deactivated (off)",
            1: "lamp activated (on)",
            2: "error",
            3: "not available"
        },
        5108: {
            0: "lamp deactivated (off)",
            1: "lamp activated (on)",
            2: "error",
            3: "not available"
        },
        5139: {
            0: "  set implement to field working state",
            1: "  set implement to transport state",
            2: "  set implement to park state",
            3: "  not available   [ed note, should this be \"no change/don't care\""
        },
        5140: {
            0: "  all implements stop operations switch off",
            1: "  all implements stop operations switch on",
            2: "  error indication",
            3: "  not available"
        },
        5141: {
            0: "input off",
            1: "input on",
            2: "error",
            3: "not installed"
        },
        5142: {
            0: "input off",
            1: "input on",
            2: "error",
            3: "not installed"
        },
        5143: {
            0: "automatic",
            1: "reserved",
            10: "reserved",
            11: "reserved",
            12: "reserved",
            13: "reserved",
            14: "reserved",
            15: "not available",
            2: "reserved",
            3: "reserved",
            4: "manual disconnect via j1939 request",
            5: "manual connect via j1939 request",
            6: "reserved",
            7: "reserved",
            8: "manual disconnect via hardwired input",
            9: "manual connect via hardwired input"
        },
        5144: {
            0: "output off or disconnected",
            1: "output on or connected",
            10: "reserved",
            11: "reserved",
            12: "reserved",
            13: "reserved",
            14: "output fault",
            15: "not available or not installed",
            2: "disconnect alarm expired, output off or disconnected",
            3: "disconnect pending, alarm active",
            4: "connect pending, alarm active",
            5: "reserved",
            6: "reserved",
            7: "reserved",
            8: "reserved",
            9: "reserved"
        },
        5145: {
            0: "output off or disconnected",
            1: "output on or connected",
            10: "reserved",
            11: "reserved",
            12: "reserved",
            13: "reserved",
            14: "output fault",
            15: "not available or not installed",
            2: "disconnect alarm expired, output off or disconnected",
            3: "disconnect pending, alarm active",
            4: "connect pending, alarm active",
            5: "reserved",
            6: "reserved",
            7: "reserved",
            8: "reserved",
            9: "reserved"
        },
        5146: {
            0: "output off or disconnected",
            1: "output on or connected",
            10: "reserved",
            11: "reserved",
            12: "reserved",
            13: "reserved",
            14: "output fault",
            15: "not available or not installed",
            2: "disconnect alarm expired, output off or disconnected",
            3: "disconnect pending, alarm active",
            4: "connect pending, alarm active",
            5: "reserved",
            6: "reserved",
            7: "reserved",
            8: "reserved",
            9: "reserved"
        },
        5147: {
            0: "output off or disconnected",
            1: "output on or connected",
            10: "reserved",
            11: "reserved",
            12: "reserved",
            13: "reserved",
            14: "output fault",
            15: "not available or not installed",
            2: "disconnect alarm expired, output off or disconnected",
            3: "disconnect pending, alarm active",
            4: "connect pending, alarm active",
            5: "reserved",
            6: "reserved",
            7: "reserved",
            8: "reserved",
            9: "reserved"
        },
        5149: {
            0: "automatic",
            1: "reserved",
            10: "reserved",
            11: "reserved",
            12: "reserved",
            13: "reserved",
            14: "reserved",
            15: "no change",
            2: "reserved",
            3: "reserved",
            4: "manual disconnect",
            5: "manual connect",
            6: "reserved",
            7: "reserved",
            8: "reserved",
            9: "reserved"
        },
        5245: {
            0: "off -  indicates adequate def level",
            1: "on solid -  indicates low def level",
            2: "reserved for sae assignment",
            3: "reserved for sae assignment",
            4: "on fast blink (1 hz) indicates the def level is lower than the level indicated by the solid illumination (state 001)",
            5: "reserved for sae assignment",
            6: "reserved for sae assignment",
            7: "not available"
        },
        5246: {
            0: "driver warning, low-level inducement, and severe inducement non-active",
            1: "inducement level 1 - epa defined scr inducement - def warning  ...................... fmi 15",
            2: "inducement level 2 - def warning, second level (optional) .................................. fmi 15",
            3: "inducement level 3 - epa defined scr inducement - engine derate ..................... fmi 16",
            4: "inducement level 4 ... severe inducement pre-trigger (optional) ---.................. fmi 16",
            5: "inducement level 5 ... epa defined scr inducement - severe inducement --......  fmi 0",
            6: "temporary override of inducement  -  the scr inducement has been temporarily interrupted.",
            7: "not available / not supported"
        },
        527: {
            0: "off/disabled",
            1: "hold",
            2: "accelerate",
            3: "decelerate",
            4: "resume",
            5: "set",
            6: "accelerator override",
            7: "not available"
        },
        5274: {
            0: " no request ",
            1: " request prohibit automatic air suspension control ",
            2: " error indicator ",
            3: " not available"
        },
        5275: {
            0: "parking brake actuator not fully activated",
            1: "parking brake actuator fully activated",
            2: "error",
            3: "not available"
        },
        5293: {
            0: "disallow level control during braking (normal operation)",
            1: "allow level control during braking",
            2: "reserved",
            3: "don-t' care/take no action"
        },
        5294: {
            0: "deactivated",
            1: "activated",
            2: "reserved",
            3: "don-t care/take no action"
        },
        5295: {
            0: "deactivate",
            1: "activate",
            2: "reserved",
            3: "don-t care/take no action"
        },
        5296: {
            0: "inactive",
            1: "active",
            2: "reserved",
            3: "don-t' care/take no action"
        },
        5323: {
            0: "     open loop",
            1: "     closed loop",
            2: "     sae reserved",
            3: "     not available"
        },
        5324: {
            0: "inactive",
            1: "active",
            2: "error",
            3: "not available"
        },
        5325: {
            0: "inactive",
            1: "active",
            2: "error",
            3: "not available"
        },
        5326: {
            0: "inactive",
            1: "active",
            2: "error",
            3: "not available"
        },
        5327: {
            0: "inactive",
            1: "active",
            2: "error",
            3: "not available"
        },
        5328: {
            0: "inactive",
            1: "active",
            2: "error",
            3: "not available"
        },
        5329: {
            0: "inactive",
            1: "active",
            2: "error",
            3: "not available"
        },
        5330: {
            0: "inactive",
            1: "active",
            2: "error",
            3: "not available"
        },
        5331: {
            0: "inactive",
            1: "active",
            2: "error",
            3: "not available"
        },
        5332: {
            0: "inactive",
            1: "active",
            2: "error",
            3: "not available"
        },
        5333: {
            0: "inactive",
            1: "active",
            2: "error",
            3: "not available"
        },
        5334: {
            0: "inactive",
            1: "active",
            2: "error",
            3: "not available"
        },
        5335: {
            0: "inactive",
            1: "active",
            2: "error",
            3: "not available"
        },
        5336: {
            0: "inactive",
            1: "active",
            2: "error",
            3: "not available"
        },
        5337: {
            0: "inactive",
            1: "active",
            2: "error",
            3: "not available"
        },
        5338: {
            0: "inactive",
            1: "active",
            2: "error",
            3: "not available"
        },
        5339: {
            0: "inactive",
            1: "active",
            2: "error",
            3: "not available"
        },
        5340: {
            0: "inactive",
            1: "active",
            2: "error",
            3: "not available"
        },
        5341: {
            0: "inactive",
            1: "active",
            2: "error",
            3: "not available"
        },
        5342: {
            0: "inactive",
            1: "active",
            2: "error",
            3: "not available"
        },
        5343: {
            0: "inactive",
            1: "active",
            2: "error",
            3: "not available"
        },
        5344: {
            0: "transmission warning indicator is off",
            1: "transmission warning indicator is on continuously",
            2: "transmission warning indicator is flashing",
            3: "not available"
        },
        5345: {
            0: "transmission overheat indicator is off",
            1: "transmission overheat indicator is on continuously",
            2: "transmission overheat indicator is flashing",
            3: "not available"
        },
        5346: {
            0: "driveline retarder overheat indicator is off",
            1: "driveline retarder overheat indicator is on continuously",
            2: "driveline retarder overheat indicator is flashing",
            3: "not available"
        },
        5350: {
            0: "lateral acceleration is fully functional.  data is within sensor specification",
            1: "lateral acceleration is degraded.  data is suspect due to enviromental conditions",
            2: "error",
            3: "not available"
        },
        5351: {
            0: "longitudinal acceleration is fully functional.  data is within sensor specification",
            1: "longitudinal acceleration is degraded.  data is suspect due to enviromental conditions",
            2: "error",
            3: "not available"
        },
        5352: {
            0: "vertical acceleration is fully functional.  data is within sensor specification",
            1: "verticalacceleration is degraded.  data is suspect due to enviromental conditions",
            2: "error",
            3: "not available"
        },
        5353: {
            1: " ms rate not supported ",
            10: " ms rate not supported ",
            11: " ms rate not supported ",
            12: " ms rate not supported ",
            13: " ms rate not supported ",
            14: " ms rate not supported ",
            15: " ms rate not supported ",
            16: " ms rate not supported ",
            17: " ms rate not supported ",
            18: " ms rate not supported ",
            19: " ms rate not supported ",
            2: " ms rate not supported ",
            20: " ms rate not supported ",
            3: " ms rate not supported ",
            4: " ms rate not supported ",
            5: " ms rate not supported ",
            6: " ms rate not supported ",
            7: " ms rate not supported ",
            8: " ms rate not supported ",
            9: " ms rate not supported "
        },
        5368: {
            0: "high most severe",
            1: "high least severe",
            2: "in range",
            3: "low least severe",
            4: "low most severe",
            5: "not defined",
            6: "error",
            7: "not available"
        },
        5372: {
            0: "high most severe",
            1: "high least severe",
            2: "in range",
            3: "low least severe",
            4: "low most severe",
            5: "not defined",
            6: "error",
            7: "not available"
        },
        5376: {
            0: "high most severe",
            1: "high least severe",
            2: "in range",
            3: "low least severe",
            4: "low most severe",
            5: "not defined",
            6: "error",
            7: "not available"
        },
        5379: {
            0: "high most severe",
            1: "high least severe",
            2: "in range",
            3: "low least severe",
            4: "low most severe",
            5: "not defined",
            6: "error",
            7: "not available"
        },
        5381: {
            0: "high most severe",
            1: "high least severe",
            2: "in range",
            3: "low least severe",
            4: "low most severe",
            5: "not defined",
            6: "error",
            7: "not available"
        },
        5383: {
            0: "high most severe",
            1: "high least severe",
            2: "in range",
            3: "low least severe",
            4: "low most severe",
            5: "not defined",
            6: "error",
            7: "not available"
        },
        5385: {
            0: "high most severe",
            1: "high least severe",
            2: "in range",
            3: "low least severe",
            4: "low most severe",
            5: "not defined",
            6: "error",
            7: "not available"
        },
        5391: {
            0: "high most severe",
            1: "high least severe",
            2: "in range",
            3: "low least severe",
            4: "low most severe",
            5: "not defined",
            6: "error",
            7: "not available"
        },
        5399: {
            0: "dpf thermal management is not active",
            1: "dpf thermal management is active",
            2: "reserved",
            3: "don't care"
        },
        5400: {
            0: "scr thermal management is not active",
            1: "scr thermal management is active",
            2: "reserved",
            3: "don't care"
        },
        5402: {
            0: "off, disable lift relay request ",
            1: "on, enable lift relay request",
            2: "reserved",
            3: "take no action"
        },
        5403: {
            0: "off, disable drop relay request",
            1: "on, enable drop relay request",
            2: "reserved",
            3: "take no action"
        },
        5404: {
            0: "no",
            1: "yes",
            2: "error",
            3: "not available"
        },
        5405: {
            0: "do not set reference elevation",
            1: "set reference elevation",
            2: "reserved",
            3: "don't care / take no action"
        },
        5406: {
            0: "do not set reference elevation",
            1: "set reference elevation",
            2: "reserved",
            3: "don't care / take no action"
        },
        5407: {
            0: "manual",
            1: "auto ",
            10: "reserved",
            11: "reserved",
            12: "reserved",
            13: "reserved",
            14: "error",
            15: "not available",
            2: "inactive auto",
            3: "reserved",
            4: "reserved",
            5: "reserved",
            6: "reserved",
            7: "reserved",
            8: "reserved",
            9: "reserved"
        },
        5408: {
            0: "manual",
            1: "auto ",
            10: "reserved",
            11: "reserved",
            12: "reserved",
            13: "reserved",
            14: "error",
            15: "not available",
            2: "inactive auto",
            3: "reserved",
            4: "reserved",
            5: "reserved",
            6: "reserved",
            7: "reserved",
            8: "reserved",
            9: "reserved"
        },
        5409: {
            0: "disabled",
            1: "enabled",
            2: "error",
            3: "not available / not installed"
        },
        5416: {
            0: "lamp deactived (off)",
            1: "lamp actived (on)",
            2: "error",
            3: "not available"
        },
        5423: {
            0: " off",
            1: " on",
            2: " reserved for sae assignment",
            3: " not available"
        },
        5424: {
            0: " off",
            1: " on",
            2: " reserved for sae assignment",
            3: " not available"
        },
        5426: {
            0: " off",
            1: " on",
            2: " reserved for sae assignment",
            3: " not available"
        },
        5427: {
            0: " off",
            1: " on",
            2: " reserved for sae assignment",
            3: " not available"
        },
        5432: {
            0: "no memory level ",
            1: "memory level 1 selected ",
            10: "memory level  2 stored",
            11: "memory level  3 stored",
            12: "memory level  4 stored",
            13: "memory level  5 stored",
            14: "error indicator",
            15: "not available",
            2: "memory level 2 selected",
            3: "memory level 3 selected",
            4: "memory level 4 selected",
            5: "memory level 5 selected",
            6: "reserved",
            7: "reserved",
            8: "reserved",
            9: "memory level  1 stored"
        },
        5433: {
            0: "no memory level request",
            1: "memory level 1 request",
            10: "store memory level  2",
            11: "store memory level  3",
            12: "store memory level  4",
            13: "store memory level  5",
            14: "reserved",
            15: "don't care",
            2: "memory level 2 request",
            3: "memory level 3 request",
            4: "memory level 4 request",
            5: "memory level 5 request",
            6: "reserved",
            7: "reserved",
            8: "reserved",
            9: "store memory level  1"
        },
        5435: {
            0: "  pump operational",
            1: "  pump error",
            2: "  reserved for sae assignment",
            3: " not available"
        },
        5438: {
            0: "  pump operational",
            1: "  pump error",
            2: "  reserved for sae assignment",
            3: " not available"
        },
        5445: {
            0: "normal",
            1: "alarm (fully operational but needs service)",
            10: "   reserved for future assignment",
            11: "   reserved for future assignment",
            12: "   reserved for future assignment",
            13: "   reserved for future assignment",
            14: "error",
            15: "not available",
            2: "alarm high severity (functional but transient performance may be reduced)",
            3: "derate active (torque output reduced due to environmental conditions)",
            4: "controlled shutdown active (driving to the default position - usually this means closed)",
            5: "uncontrolled shutdown active (actuator current is off so it is limp)",
            6: "   reserved for future assignment",
            7: "   reserved for future assignment",
            8: "   reserved for future assignment",
            9: "   reserved for future assignment"
        },
        5446: {
            0: "normal",
            1: "alarm (fully operational but needs service)",
            10: "   reserved for future assignment",
            11: "   reserved for future assignment",
            12: "   reserved for future assignment",
            13: "   reserved for future assignment",
            14: "error",
            2: "alarm high severity (functional but transient performance may be reduced)",
            3: "derate active (torque output reduced due to environmental conditions)",
            4: "controlled shutdown active (driving to the default position - usually this means closed)",
            5: "uncontrolled shutdown active (actuator current is off so it is limp)",
            6: "   reserved for future assignment",
            7: "   reserved for future assignment",
            8: "   reserved for future assignment",
            9: "   reserved for future assignment"
        },
        5447: {
            0: "normal",
            1: "alarm (fully operational but needs service)",
            10: "   reserved for future assignment",
            11: "   reserved for future assignment",
            12: "   reserved for future assignment",
            13: "   reserved for future assignment",
            14: "error",
            2: "alarm high severity (functional but transient performance may be reduced)",
            3: "derate active (torque output reduced due to environmental conditions)",
            4: "controlled shutdown active (driving to the default position - usually this means closed)",
            5: "uncontrolled shutdown active (actuator current is off so it is limp)",
            6: "   reserved for future assignment",
            7: "   reserved for future assignment",
            8: "   reserved for future assignment",
            9: "   reserved for future assignment"
        },
        5448: {
            0: "normal",
            1: "alarm (fully operational but needs service)",
            10: "   reserved for future assignment",
            11: "   reserved for future assignment",
            12: "   reserved for future assignment",
            13: "   reserved for future assignment",
            14: "error",
            2: "alarm high severity (functional but transient performance may be reduced)",
            3: "derate active (torque output reduced due to environmental conditions)",
            4: "controlled shutdown active (driving to the default position - usually this means closed)",
            5: "uncontrolled shutdown active (actuator current is off so it is limp)",
            6: "   reserved for future assignment",
            7: "   reserved for future assignment",
            8: "   reserved for future assignment",
            9: "   reserved for future assignment"
        },
        5449: {
            0: "normal",
            1: "alarm (fully operational but needs service)",
            10: "   reserved for future assignment",
            11: "   reserved for future assignment",
            12: "   reserved for future assignment",
            13: "   reserved for future assignment",
            14: "error",
            2: "alarm high severity (functional but transient performance may be reduced)",
            3: "derate active (torque output reduced due to environmental conditions)",
            4: "controlled shutdown active (driving to the default position - usually this means closed)",
            5: "uncontrolled shutdown active (actuator current is off so it is limp)",
            6: "   reserved for future assignment",
            7: "   reserved for future assignment",
            8: "   reserved for future assignment",
            9: "   reserved for future assignment"
        },
        5450: {
            0: "normal",
            1: "alarm (fully operational but needs service)",
            10: "   reserved for future assignment",
            11: "   reserved for future assignment",
            12: "   reserved for future assignment",
            13: "   reserved for future assignment",
            14: "error",
            2: "alarm high severity (functional but transient performance may be reduced)",
            3: "derate active (torque output reduced due to environmental conditions)",
            4: "controlled shutdown active (driving to the default position - usually this means closed)",
            5: "uncontrolled shutdown active (actuator current is off so it is limp)",
            6: "   reserved for future assignment",
            7: "   reserved for future assignment",
            8: "   reserved for future assignment",
            9: "   reserved for future assignment"
        },
        5451: {
            0: "normal",
            1: "alarm (fully operational but needs service)",
            10: "   reserved for future assignment",
            11: "   reserved for future assignment",
            12: "   reserved for future assignment",
            13: "   reserved for future assignment",
            14: "error",
            2: "alarm high severity (functional but transient performance may be reduced)",
            3: "derate active (torque output reduced due to environmental conditions)",
            4: "controlled shutdown active (driving to the default position - usually this means closed)",
            5: "uncontrolled shutdown active (actuator current is off so it is limp)",
            6: "   reserved for future assignment",
            7: "   reserved for future assignment",
            8: "   reserved for future assignment",
            9: "   reserved for future assignment"
        },
        5452: {
            0: "normal",
            1: "alarm (fully operational but needs service)",
            10: "   reserved for future assignment",
            11: "   reserved for future assignment",
            12: "   reserved for future assignment",
            13: "   reserved for future assignment",
            14: "error",
            2: "alarm high severity (functional but transient performance may be reduced)",
            3: "derate active (torque output reduced due to environmental conditions)",
            4: "controlled shutdown active (driving to the default position - usually this means closed)",
            5: "uncontrolled shutdown active (actuator current is off so it is limp)",
            6: "   reserved for future assignment",
            7: "   reserved for future assignment",
            8: "   reserved for future assignment",
            9: "   reserved for future assignment"
        },
        5457: {
            0: " open loop",
            1: " closed loop",
            2: " sae reserved",
            3: " not available"
        },
        5459: {
            0: "  inactive",
            1: "  active",
            2: "  error indicator",
            3: "  not available"
        },
        5460: {
            0: "  inactive",
            1: "  active",
            2: "  error indicator",
            3: "  not available"
        },
        5461: {
            0: "  inactive",
            1: "  active",
            2: "  error indicator",
            3: "  not available"
        },
        5462: {
            0: "  inactive",
            1: "  active",
            2: "  error indicator",
            3: "  not available"
        },
        5465: {
            0: "  open loop ",
            1: "  closed loop ",
            2: "  sae reserved",
            3: "  not available"
        },
        5469: {
            0: "lamp off",
            1: "lamp on",
            2: "flash (1 hz)",
            3: "not available"
        },
        5470: {
            0: "lamp deactivated (off)",
            1: "lamp activated (on)",
            2: "error",
            3: "not available"
        },
        5488: {
            0: " heater inactive ",
            1: " heater active ",
            2: " error ",
            3: " not available"
        },
        5490: {
            0: " request for heater to be inactive ",
            1: " request for heater to be active ",
            2: " reserved for future assignment by sae ",
            3: " not available"
        },
        5492: {
            0: "off",
            1: "on",
            2: "reserved",
            3: "don't care/take no action"
        },
        5493: {
            0: "off",
            1: "on",
            2: "reserved",
            3: "don't care/take no action"
        },
        5494: {
            0: "off",
            1: "on",
            2: "reserved",
            3: "don't care/take no action"
        },
        5499: {
            0: " off",
            1: " on",
            2: " reserved for sae assignment",
            3: " not available"
        },
        5500: {
            0: " off",
            1: " on",
            2: " reserved for sae assignment",
            3: " not available"
        },
        5504: {
            0: " purging not enabled",
            1: " purging enabled - less urgent",
            2: " purging enabled - urgent",
            3: " not available"
        },
        5506: {
            0: " hc doser disabled",
            1: " hc doser active",
            2: " sae reserved",
            3: " hc doser enabled, waiting for regeneration",
            4: " sae reserved",
            5: " sae reserved",
            6: " error",
            7: " not available"
        },
        5507: {
            0: " injecting not active",
            1: " injecting active",
            2: " error",
            3: " not available"
        },
        5508: {
            0: " system test has executed and passed or system test has not yet executed",
            1: " internal system test is in process",
            2: "  sae reserved",
            3: "  sae reserved",
            4: "  sae reserved",
            5: "  sae reserved",
            6: " error  --  system test has failed",
            7: " not available"
        },
        5509: {
            0: "   purging not required",
            1: "   purging required - less urgent",
            2: "   purging required - urgent",
            3: "   not available"
        },
        5510: {
            0: "not locked",
            1: "locked",
            2: "reserved",
            3: "no action"
        },
        5511: {
            0: "cycle mode # 1 (lowest lubrication frequency)",
            1: "cycle mode # 2",
            2: "cycle mode # 3 (highest lubrication frequency)",
            3: " to 110 - reserved ",
            7: "no action"
        },
        5512: {
            0: "automatic mode",
            1: "manual service mode",
            2: "test mode # 1",
            3: "test mode # 2",
            4: "test mode # 3",
            5: "reserved",
            6: "reserved",
            7: "take no action"
        },
        5513: {
            0: "no action (normal pump operation)",
            1: "reset pump operation  ",
            2: "reserved",
            3: "take no action"
        },
        5514: {
            0: "pause",
            1: "enable",
            2: "reserved ",
            3: "take no action     "
        },
        5515: {
            0: "not locked",
            1: "locked",
            2: "error",
            3: "not available"
        },
        5516: {
            0: "cycle mode # 1 (lowest lubrication frequency)",
            1: "cycle mode # 2",
            2: "cycle mode # 3 (highest lubrication frequency)",
            3: " to 101 - reserved ",
            6: "error",
            7: "not available"
        },
        5517: {
            0: "automatic mode",
            1: "manual service mode",
            2: "test mode # 1",
            3: "test mode # 2",
            4: "test mode # 3",
            5: "reserved",
            6: "error",
            7: "not available"
        },
        5518: {
            0: "disabled",
            1: "enabled",
            2: "error",
            3: "not available"
        },
        5519: {
            0: "reservoir not low",
            1: "reservoir low",
            2: "error ",
            3: "not available or not installed"
        },
        5520: {
            0: "paused ",
            1: "enabled",
            2: "error ",
            3: "not available or not installed"
        },
        5549: {
            0: "  off",
            1: "  on",
            2: "  reserved",
            3: "  don't care/take no action"
        },
        5550: {
            0: "  off",
            1: "  on",
            2: "  error",
            3: "  not available"
        },
        5551: {
            0: "  off, disable generator",
            1: "  on, enable generator",
            2: "  reserved",
            3: "  take no action"
        },
        5552: {
            0: "  disable ",
            1: "  enable",
            2: "  reserved ",
            3: "  don't care/take no action"
        },
        5553: {
            0: "   forward",
            1: "   reverse",
            2: "   reserved",
            3: "   don't care/take no action (leave function as is)"
        },
        5555: {
            0: "   disable",
            1: "   enable",
            2: "   reserved",
            3: "   don't care/take no action"
        },
        5556: {
            0: "   disabled",
            1: "   enabled",
            2: "   error",
            3: "   not available/not installed"
        },
        5557: {
            0: "   forward",
            1: "   reverse",
            2: "   error",
            3: "   not available/not installed"
        },
        5558: {
            0: "   at velocity (fan ok)",
            1: "   velocity error  (unable to achieve commanded speed)",
            10: "   reserved",
            11: "   reserved",
            12: "   reserved",
            13: "   reserved",
            14: "   error",
            15: "   not available",
            2: "   fan spin up",
            3: "   fan spin down",
            4: "   fan stall",
            5: "   reserved",
            6: "   reserved",
            7: "   reserved",
            8: "   reserved",
            9: "   reserved"
        },
        5559: {
            0: "   high most severe ",
            1: "   high least severe ",
            2: "   in range ",
            3: "   low least severe ",
            4: "   low most severe ",
            5: "   not defined ",
            6: "   error ",
            7: "   not available"
        },
        5560: {
            0: "   normal = \"ok\"",
            1: "   alarm = fully operational but needs service",
            10: "   reserved for future assignment = not used",
            11: "   reserved for future assignment = not used",
            12: "   reserved for future assignment = not used",
            13: "   reserved for future assignment = not used",
            14: "   error",
            15: "   not available",
            2: "   alarm high severity = functional but transient performance may be reduced",
            3: "   derate active = torque output reduced due to environmental conditions",
            4: "   controlled shutdown active = driving to the default position - usually this means closed",
            5: "   uncontrolled shutdown active = actuator current is off so it is limp",
            6: "   reserved for future assignment = not used",
            7: "   reserved for future assignment = not used",
            8: "   reserved for future assignment = not used",
            9: "   reserved for future assignment = not used"
        },
        5565: {
            0: "disable engine protection shutdown override due to low coolant level",
            1: "enable engine protection shutdown override due to low coolant level",
            2: "reserved",
            3: "don't care/take no action"
        },
        5566: {
            0: "disabled - calibration does not allow engine protection due to low coolant level or the override is currently active",
            1: "enabled - calibration allows engine protection due to low coolant level and the override is currently not active",
            2: "reserved",
            3: "not available"
        },
        558: {
            0: "accelerator pedal 1 not in low idle condition",
            1: "accelerator pedal 1 in low idle condition",
            2: "error",
            3: "not available"
        },
        5583: {
            0: "sensor busy",
            1: "sensor ok",
            2: "sensor defect",
            3: "sensor reboots",
            4: "angle data invalid",
            5: "reserved",
            6: "error",
            7: "not available"
        },
        559: {
            0: "kickdown passive",
            1: "kickdown active",
            2: "error",
            3: "not available"
        },
        560: {
            0: "driveline disengaged",
            1: "driveline engaged",
            2: "error",
            3: "not available"
        },
        5603: {
            0: "cruise control is allowed",
            1: "cruise control is not allowed",
            2: "reserved",
            3: "don't care/take no action"
        },
        5604: {
            0: "cruise control resume not requested",
            1: "cruise control resume requested",
            2: "reserved",
            3: "don't care/take no action"
        },
        5605: {
            0: "cruise control is allowed",
            1: "cruise control is not allowed",
            2: "reserved",
            3: "don't care/take no action"
        },
        5606: {
            0: "off -- cruise control system is not configured to accept commands from an adaptive cruise controller",
            1: "on -- cruise control system is configured to accept commands from an adaptive cruise controller.",
            2: "error condition -- prevents system from responding to adaptive cruise controller commands even though it is configured.",
            3: "not available/not supported"
        },
        5607: {
            0: "no cruise control system command",
            1: "cruise control has been disabled by cruise control disable command (spn 5603)",
            2: "cruise control  has been disabled by cruise control pause command (spn 5605)",
            3: "cruise control has been re-activated by cruise control resume command (spn 5604) and cruise control states (spn 527) is equal to resume (100).  cruise control system command state will be equal to 011 as long as cruise control states is equal to resume.  cruise control system command state will change to the appropriate value when cruise control states is no longer equal to resume.",
            4: "cruise control device has received a cruise control resume command (spn 5604), but there is no previous set speed.",
            5: "- sae reserved",
            6: "- sae reserved",
            7: "not supported"
        },
        561: {
            0: "asr engine control passive but installed",
            1: "asr engine control active",
            2: "reserved",
            3: "not available"
        },
        562: {
            0: "asr brake control passive but installed",
            1: "asr brake control active",
            2: "reserved",
            3: "not available"
        },
        5624: {
            0: "vdc passive, but installed",
            1: "vdc active",
            2: "reserved",
            3: "not availble or not installed"
        },
        5627: {
            0: "high most severe",
            1: "high least severe",
            2: "in range",
            3: "low least severe",
            4: "low most severe",
            5: "not defined",
            6: "error",
            7: "not available"
        },
        5628: {
            0: "normal",
            1: "alarm (fully operational but needs service)",
            10: " reserved for future assignment",
            11: " reserved for future assignment",
            12: " reserved for future assignment",
            13: " reserved for future assignment",
            14: "error",
            15: "not available",
            2: "alarm high severity (functional but transient performance may be reduced)",
            3: "derate active (torque output reduced due to environmental conditions)",
            4: "controlled shutdown active (driving to the default position - usually this means closed)",
            5: "uncontrolled shutdown active (actuator current is off so it is limp)",
            6: " reserved for future assignment",
            7: " reserved for future assignment",
            8: " reserved for future assignment",
            9: " reserved for future assignment"
        },
        5629: {
            0: " not inhibited",
            1: " inhibited",
            2: " reserved for sae assignment",
            3: " not available"
        },
        563: {
            0: "abs passive but installed",
            1: "abs active",
            2: "reserved",
            3: "not available"
        },
        5630: {
            0: "park brake apply not requested",
            1: "park brake apply requested",
            2: "sae reserved  ",
            3: "don't care/take no action"
        },
        564: {
            0: " differential lock disengaged",
            1: " differential lock engaged",
            2: " error",
            3: " not available"
        },
        565: {
            0: " differential lock disengaged",
            1: " differential lock engaged",
            2: " error",
            3: " not available"
        },
        5657: {
            0: "security not satisfied. different sa for adopt pending than set pending",
            1: "item(s) not allowed to change.  set qualifier flag of disallowed items.",
            10: "reserved ",
            100: "reserved ",
            101: "reserved ",
            102: "reserved ",
            103: "reserved ",
            104: "reserved ",
            105: "reserved ",
            106: "reserved ",
            107: "reserved ",
            108: "reserved ",
            109: "reserved ",
            11: "reserved ",
            110: "reserved ",
            111: "reserved ",
            112: "reserved ",
            113: "reserved ",
            114: "reserved ",
            115: "reserved ",
            116: "reserved ",
            117: "reserved ",
            118: "reserved ",
            119: "reserved ",
            12: "reserved ",
            120: "reserved ",
            121: "reserved ",
            122: "reserved ",
            123: "reserved ",
            124: "reserved ",
            125: "reserved ",
            126: "reserved ",
            127: "reserved ",
            128: "reserved ",
            129: "reserved ",
            13: "reserved ",
            130: "reserved ",
            131: "reserved ",
            132: "reserved ",
            133: "reserved ",
            134: "reserved ",
            135: "reserved ",
            136: "reserved ",
            137: "reserved ",
            138: "reserved ",
            139: "reserved ",
            14: "reserved ",
            140: "reserved ",
            141: "reserved ",
            142: "reserved ",
            143: "reserved ",
            144: "reserved ",
            145: "reserved ",
            146: "reserved ",
            147: "reserved ",
            148: "reserved ",
            149: "reserved ",
            15: "reserved ",
            150: "reserved ",
            151: "reserved ",
            152: "reserved ",
            153: "reserved ",
            154: "reserved ",
            155: "reserved ",
            156: "reserved ",
            157: "reserved ",
            158: "reserved ",
            159: "reserved ",
            16: "reserved ",
            160: "reserved ",
            161: "reserved ",
            162: "reserved ",
            163: "reserved ",
            164: "reserved ",
            165: "reserved ",
            166: "reserved ",
            167: "reserved ",
            168: "reserved ",
            169: "reserved ",
            17: "reserved ",
            170: "reserved ",
            171: "reserved ",
            172: "reserved ",
            173: "reserved ",
            174: "reserved ",
            175: "reserved ",
            176: "reserved ",
            177: "reserved ",
            178: "reserved ",
            179: "reserved ",
            18: "reserved ",
            180: "reserved ",
            181: "reserved ",
            182: "reserved ",
            183: "reserved ",
            184: "reserved ",
            185: "reserved ",
            186: "reserved ",
            187: "reserved ",
            188: "reserved ",
            189: "reserved ",
            19: "reserved ",
            190: "reserved ",
            191: "reserved ",
            192: "reserved ",
            193: "reserved ",
            194: "reserved ",
            195: "reserved ",
            196: "reserved ",
            197: "reserved ",
            198: "reserved ",
            199: "reserved ",
            2: "item conflict. cannot perform function assigned, cannot perform as arbitrary address capable, etc. qualifier flags of disallowed items are set.",
            20: "reserved ",
            200: "reserved ",
            201: "reserved ",
            202: "reserved ",
            203: "reserved ",
            204: "reserved ",
            205: "reserved ",
            206: "reserved ",
            207: "reserved ",
            208: "reserved ",
            209: "reserved ",
            21: "reserved ",
            210: "reserved ",
            211: "reserved ",
            212: "reserved ",
            213: "reserved ",
            214: "reserved ",
            215: "reserved ",
            216: "reserved ",
            217: "reserved ",
            218: "reserved ",
            219: "reserved ",
            22: "reserved ",
            220: "reserved ",
            221: "reserved ",
            222: "reserved ",
            223: "reserved ",
            224: "reserved ",
            225: "reserved ",
            226: "reserved ",
            227: "reserved ",
            228: "reserved ",
            229: "reserved ",
            23: "reserved ",
            230: "reserved ",
            231: "reserved ",
            232: "reserved ",
            233: "reserved ",
            234: "reserved ",
            235: "reserved ",
            236: "reserved ",
            237: "reserved ",
            238: "reserved ",
            239: "reserved ",
            24: "reserved ",
            240: "reserved ",
            241: "reserved ",
            242: "reserved ",
            243: "reserved ",
            244: "reserved ",
            245: "reserved ",
            246: "reserved ",
            247: "reserved ",
            248: "reserved ",
            249: "reserved ",
            25: "reserved ",
            250: "reserved ",
            251: "reserved ",
            252: "reserved ",
            253: "reserved ",
            254: "reserved ",
            255: "not available",
            26: "reserved ",
            27: "reserved ",
            28: "reserved ",
            29: "reserved ",
            3: "checksum does not match",
            30: "reserved ",
            31: "reserved ",
            32: "reserved ",
            33: "reserved ",
            34: "reserved ",
            35: "reserved ",
            36: "reserved ",
            37: "reserved ",
            38: "reserved ",
            39: "reserved ",
            4: "other",
            40: "reserved ",
            41: "reserved ",
            42: "reserved ",
            43: "reserved ",
            44: "reserved ",
            45: "reserved ",
            46: "reserved ",
            47: "reserved ",
            48: "reserved ",
            49: "reserved ",
            5: "reserved ",
            50: "reserved ",
            51: "reserved ",
            52: "reserved ",
            53: "reserved ",
            54: "reserved ",
            55: "reserved ",
            56: "reserved ",
            57: "reserved ",
            58: "reserved ",
            59: "reserved ",
            6: "reserved ",
            60: "reserved ",
            61: "reserved ",
            62: "reserved ",
            63: "reserved ",
            64: "reserved ",
            65: "reserved ",
            66: "reserved ",
            67: "reserved ",
            68: "reserved ",
            69: "reserved ",
            7: "reserved ",
            70: "reserved ",
            71: "reserved ",
            72: "reserved ",
            73: "reserved ",
            74: "reserved ",
            75: "reserved ",
            76: "reserved ",
            77: "reserved ",
            78: "reserved ",
            79: "reserved ",
            8: "reserved ",
            80: "reserved ",
            81: "reserved ",
            82: "reserved ",
            83: "reserved ",
            84: "reserved ",
            85: "reserved ",
            86: "reserved ",
            87: "reserved ",
            88: "reserved ",
            89: "reserved ",
            9: "reserved ",
            90: "reserved ",
            91: "reserved ",
            92: "reserved ",
            93: "reserved ",
            94: "reserved ",
            95: "reserved ",
            96: "reserved ",
            97: "reserved ",
            98: "reserved ",
            99: "reserved "
        },
        5658: {
            0: "manufacturer code is to be changed",
            1: "manufacturer code shall not be changed"
        },
        5659: {
            0: "ecu instance is to be changed",
            1: "ecu instance shall not be changed"
        },
        566: {
            0: " differential lock disengaged",
            1: " differential lock engaged",
            2: " error",
            3: " not available"
        },
        5660: {
            0: "function instance is to be changed",
            1: "function instance shall not be changed"
        },
        5661: {
            0: "function is to be changed",
            1: "function shall not be changed"
        },
        5662: {
            0: "vehicle system is to be changed",
            1: "vehicle system shall not be changed"
        },
        5663: {
            0: "vehicle system instance is to be changed",
            1: "vehicle system instance shall not be changed"
        },
        5664: {
            0: "industry group is to be changed",
            1: "industry group shall not be changed"
        },
        5665: {
            0: "arbitrary address capable is to be changed",
            1: "arbitrary address capable shall not be changed"
        },
        567: {
            0: " differential lock disengaged",
            1: " differential lock engaged",
            2: " error",
            3: " not available"
        },
        5675: {
            0: "indicator(s) inactive",
            1: "upshift indicator active",
            2: "downshift indicator active",
            3: "don't care/take no action"
        },
        5676: {
            0: "system is not ready (initialization not finished)",
            1: "system is temporarily not available (e.g. due to boundary conditions necessary for operation)",
            10: "reserved for future use",
            11: "reserved for future use",
            12: "reserved for future use",
            13: "reserved for future use",
            14: "error indication",
            15: "not available / not installed",
            2: "system is deactivated by driver",
            3: "system is ready (no warning and no braking active)",
            4: "driver overrides system",
            5: "collision warning active (not affecting vehicle dynamics)",
            6: "collision warning with braking (e.g. brake jerk or partial braking)",
            7: "emergency braking active",
            8: "reserved for future use",
            9: "reserved for future use"
        },
        5677: {
            0: "no warning",
            1: "warning level 1 - lowest",
            10: "reserved",
            11: "reserved",
            12: "reserved",
            13: "reserved",
            14: "reserved",
            15: "don-t care / take no action",
            2: "warning level 2",
            3: "warning level 3",
            4: "warning level 4",
            5: "warning level 5",
            6: "warning level 6",
            7: "warning level 7 - highest",
            8: "reserved",
            9: "reserved"
        },
        5678: {
            0: "no relevant object monitored",
            1: "relevant object is being monitored",
            2: "relevant object is not being detected reliably",
            3: "..101 - reserved",
            6: "error indication",
            7: "signal not available / not installed"
        },
        5679: {
            0: "no indication for a bend off maneuver of the relevant object",
            1: "low probability for a bend off maneuver of the relevant object",
            2: "medium probability for a bend off maneuver of the relevant object",
            3: "high probability for a bend off maneuver of the relevant object",
            4: "highest probability for a bend off maneuver of the relevant object",
            5: "relevant object is already in a turning maneuver",
            6: "error indication",
            7: "signal not available / not installed"
        },
        568: {
            0: " differential lock disengaged",
            1: " differential lock engaged",
            2: " error",
            3: " not available"
        },
        5681: {
            0: "the driver does not want the advanced emergency braking system to warn or intervene at any time (deactivation of system)",
            1: "the driver wants the advanced emergency braking system to warn or intervene if necessary (no deactivation of system)",
            2: "reserved",
            3: "don-t' care / take no action"
        },
        569: {
            0: " differential lock disengaged",
            1: " differential lock engaged",
            2: " error",
            3: " not available"
        },
        570: {
            0: " differential lock disengaged",
            1: " differential lock engaged",
            2: " error",
            3: " not available"
        },
        5709: {
            0: " diagnostic mode disabled",
            1: " nox sensor ato1 self diagnosis request at lambda=1.0",
            10: " reserved",
            11: " reserved",
            12: " reserved",
            13: " reserved",
            14: " reserved",
            15: " not supported",
            2: " nox sensor ato1 self diagnosis request at partial load",
            3: " nox sensor ato1 self diagnosis request at fuel cut",
            4: " nox sensor ato1 self diagnosis request at after run",
            5: " reserved",
            6: " reserved",
            7: " reserved",
            8: " reserved",
            9: " reserved"
        },
        571: {
            0: " retarder - brake assist disabled",
            1: " retarder - brake assist enabled",
            2: " error",
            3: " not available"
        },
        5710: {
            0: " diagnostic mode disabled",
            1: " nox sensor ati1 self diagnosis request at lambda=1.0",
            10: " reserved",
            11: " reserved",
            12: " reserved",
            13: " reserved",
            14: " reserved",
            15: " not supported",
            2: " nox sensor ati1 self diagnosis request at partial load",
            3: " nox sensor ati1 self diagnosis request at fuel cut",
            4: " nox sensor ati1 self diagnosis request at after run",
            5: " reserved",
            6: " reserved",
            7: " reserved",
            8: " reserved",
            9: " reserved"
        },
        5711: {
            0: " diagnostic mode disabled",
            1: " nox sensor ato2 self diagnosis request at lambda=1.0",
            10: " reserved",
            11: " reserved",
            12: " reserved",
            13: " reserved",
            14: " reserved",
            15: " not supported",
            2: " nox sensor ato2 self diagnosis request at partial load",
            3: " nox sensor ato2 self diagnosis request at fuel cut",
            4: " nox sensor ato2 self diagnosis request at after run",
            5: " reserved",
            6: " reserved",
            7: " reserved",
            8: " reserved",
            9: " reserved"
        },
        5712: {
            0: " diagnostic mode disabled",
            1: " nox sensor ati2 self diagnosis request at lambda=1.0",
            10: " reserved",
            11: " reserved",
            12: " reserved",
            13: " reserved",
            14: " reserved",
            15: " not supported",
            2: " nox sensor ati2 self diagnosis request at partial load",
            3: " nox sensor ati2 self diagnosis request at fuel cut",
            4: " nox sensor ati2 self diagnosis request at after run",
            5: " reserved",
            6: " reserved",
            7: " reserved",
            8: " reserved",
            9: " reserved"
        },
        5713: {
            0: " diagnosis not active",
            1: " nox sensor ato1 self diagnosis active flag",
            2: " nox sensor ato1 self diagnosis result complete",
            3: " nox sensor ato1 self diagnosis aborted",
            4: " nox sensor ato1 self diagnosis not possible",
            5: " reserved",
            6: " reserved",
            7: " not supported"
        },
        5714: {
            0: " diagnosis not active",
            1: " nox sensor ati1 self diagnosis active flag",
            2: " nox sensor ati1 self diagnosis result complete",
            3: " nox sensor ati1 self diagnosis aborted",
            4: " nox sensor ati1 self diagnosis not possible",
            5: " reserved",
            6: " reserved",
            7: " not supported"
        },
        5715: {
            0: " diagnosis not active",
            1: " nox sensor ato2 self diagnosis active flag",
            2: " nox sensor ato2 self diagnosis result complete",
            3: " nox sensor ato2 self diagnosis aborted",
            4: " nox sensor ato2 self diagnosis not possible",
            5: " reserved",
            6: " reserved",
            7: " not supported"
        },
        5716: {
            0: " diagnosis not active",
            1: " nox sensor ati2 self diagnosis active flag",
            2: " nox sensor ati2 self diagnosis result complete",
            3: " nox sensor ati2 self diagnosis aborted",
            4: " nox sensor ati2 self diagnosis not possible",
            5: " reserved",
            6: " reserved",
            7: " not supported"
        },
        572: {
            0: " retarder - shift assist disabled",
            1: " retarder - shift assist enabled",
            2: " error",
            3: " not available"
        },
        573: {
            0: " torque converter lockup disengaged",
            1: " torque converter lockup engaged",
            2: " error",
            3: " not available"
        },
        574: {
            0: "shift is not in process ",
            1: "shift in process",
            2: "error",
            3: "not available"
        },
        575: {
            0: "abs off-road switch passive",
            1: "abs off-road switch active",
            2: "error",
            3: "not available"
        },
        5751: {
            0: "b - stop - received an rf messsage from a transmitter requesting a machine level \"stop.\"",
            1: "b - clear - received an rf message from a transmitter clearing a machine level \"stop\" request.",
            14: "b - error - device error",
            15: "b - don't care",
            2: "b - field test - used to request an acknowledgement of an rf signal from the onboard ecu, without adversely affecting the operation of the vehicle",
            3: "b - transmitter test - a transmitter test rf message was requested by a transmitter, used to diagnose transmitter behaviors",
            4: "b - standby - no new status message received at the time the message was communicated",
            5: "b to 1101b - reserved"
        },
        5755: {
            0: "b - beacon request - an ecu requests that a beacon message be transmitted from a wireless device on a machine. this status will always originate from the ecu, and be sent to a wireless device. the message associated with this state will not have an associated transmitter id, or signal strength.",
            1: "b - beacon received acknowledgement - acknowledgement of a beacon transmission detected by an onboard wireless device. this status will always originate from the wireless device, and be sent to the ecu. the message associated with this state will have an associated transmitter id and signal strength.",
            2: "b - reserved",
            3: "b - don't care"
        },
        5758: {
            0: "off",
            1: "on: the ecu is powering the aftertreatment 1 intake gas sensor",
            2: "error",
            3: "unavailable"
        },
        5759: {
            0: "off",
            1: "on: the ecu is powering the aftertreatment 1 outlet gas sensor",
            2: "error",
            3: "unavailable"
        },
        576: {
            0: "asr off-road switch passive",
            1: "asr off-road switch active",
            2: "error",
            3: "not available"
        },
        5760: {
            0: "off",
            1: "on: the ecu is powering the aftertreatment 2 intake gas sensor",
            2: "error",
            3: "unavailable"
        },
        5761: {
            0: "off",
            1: "on: the ecu is powering the aftertreatment 2 outlet gas sensor",
            2: "error",
            3: "unavailable"
        },
        5762: {
            0: "allow or resume transmission load reduction functions",
            1: "inhibit or abort transmission load reduction functions",
            2: "reserved",
            3: "don't care/take no action"
        },
        5764: {
            0: "high most severe",
            1: "high least severe",
            2: "in range ",
            3: "low least severe",
            4: "low most severe",
            5: "not defined",
            6: "error",
            7: "not available"
        },
        5768: {
            0: "high most severe",
            1: "high least severe",
            2: "in range ",
            3: "low least severe",
            4: "low most severe",
            5: "not defined",
            6: "error",
            7: "not available"
        },
        577: {
            0: "asr -hill holder- switch passive",
            1: "asr -hill holder- switch active",
            2: "error",
            3: "not available"
        },
        5771: {
            0: "normal",
            1: "alarm (fully operational but needs service)",
            10: " reserved for future assignment",
            11: " reserved for future assignment",
            12: " reserved for future assignment",
            13: " reserved for future assignment",
            14: "error",
            15: "not available",
            2: "alarm high severity (functional but transient performance may be reduced)",
            3: "derate active (torque output reduced due to environmental conditions)",
            4: "controlled shutdown active (driving to the default position - usually this means closed)",
            5: "uncontrolled shutdown active (actuator current is off so it is limp)",
            6: " reserved for future assignment",
            7: " reserved for future assignment",
            8: " reserved for future assignment",
            9: " reserved for future assignment"
        },
        5772: {
            0: "normal",
            1: "alarm (fully operational but needs service)",
            10: " reserved for future assignment",
            11: " reserved for future assignment",
            12: " reserved for future assignment",
            13: " reserved for future assignment",
            14: "error",
            15: "not available",
            2: "alarm high severity (functional but transient performance may be reduced)",
            3: "derate active (torque output reduced due to environmental conditions)",
            4: "controlled shutdown active (driving to the default position - usually this means closed)",
            5: "uncontrolled shutdown active (actuator current is off so it is limp)",
            6: " reserved for future assignment",
            7: " reserved for future assignment",
            8: " reserved for future assignment",
            9: " reserved for future assignment"
        },
        5774: {
            0: "high most severe",
            1: "high least severe",
            2: "in range ",
            3: "low least severe",
            4: "low most severe",
            5: "not defined",
            6: "error",
            7: "not available"
        },
        5778: {
            0: "high most severe",
            1: "high least severe",
            2: "in range ",
            3: "low least severe",
            4: "low most severe",
            5: "not defined",
            6: "error",
            7: "not available"
        },
        5781: {
            0: "normal",
            1: "alarm (fully operational but needs service)",
            10: " reserved for future assignment",
            11: " reserved for future assignment",
            12: " reserved for future assignment",
            13: " reserved for future assignment",
            14: "error",
            15: "not available",
            2: "alarm high severity (functional but transient performance may be reduced)",
            3: "derate active (torque output reduced due to environmental conditions)",
            4: "controlled shutdown active (driving to the default position - usually this means closed)",
            5: "uncontrolled shutdown active (actuator current is off so it is limp)",
            6: " reserved for future assignment",
            7: " reserved for future assignment",
            8: " reserved for future assignment",
            9: " reserved for future assignment"
        },
        5782: {
            0: "normal",
            1: "alarm (fully operational but needs service)",
            10: " reserved for future assignment",
            11: " reserved for future assignment",
            12: " reserved for future assignment",
            13: " reserved for future assignment",
            14: "error",
            15: "not available",
            2: "alarm high severity (functional but transient performance may be reduced)",
            3: "derate active (torque output reduced due to environmental conditions)",
            4: "controlled shutdown active (driving to the default position - usually this means closed)",
            5: "uncontrolled shutdown active (actuator current is off so it is limp)",
            6: " reserved for future assignment",
            7: " reserved for future assignment",
            8: " reserved for future assignment",
            9: " reserved for future assignment"
        },
        5793: {
            0: " fueling not desired (shut off engine fueling)",
            1: " fueling desired (keep engine running)",
            2: " parameter supported, but no request",
            3: " don't care / take no action"
        },
        5794: {
            0: " fueling is or will be inhibited (spn 5793 desired engine fueling state = 00)",
            1: " engine will be kept running (spn 5793 desired engine fueling state = 01)",
            2: " no active request per spn 5793 desired engine fueling state",
            3: " not supported"
        },
        5795: {
            0: "   engine currently will not stop fueling in response to spn 5793 desired engine fueling state",
            1: "   engine will stop fueling in response to spn 5793 desired engine fueling state",
            2: "   sae reserved",
            3: "   not supported"
        },
        5800: {
            0: "no reason/all clear",
            1: "operator presence not detected",
            10: "reserved",
            11: "reserved",
            12: "reserved",
            13: "reserved",
            14: "reserved",
            15: "reserved",
            16: "reserved",
            17: "reserved",
            18: "reserved",
            19: "reserved",
            2: "implement released control of function",
            20: "hydraulic oil level too low",
            21: "valve locked out",
            22: "reserved",
            23: "reserved",
            24: "reserved",
            25: "reserved",
            26: "reserved",
            27: "reserved",
            28: "reserved",
            29: "reserved",
            3: "operator override of function",
            30: "reserved",
            31: "reserved",
            32: "reserved",
            33: "reserved",
            34: "reserved",
            35: "reserved",
            36: "reserved",
            37: "reserved",
            38: "reserved",
            39: "reserved",
            4: "operator control not in valid position",
            40: "reserved",
            41: "manufacturer specific",
            42: "manufacturer specific",
            43: "manufacturer specific",
            44: "manufacturer specific",
            45: "manufacturer specific",
            46: "manufacturer specific",
            47: "manufacturer specific",
            48: "manufacturer specific",
            49: "manufacturer specific",
            5: "remote command timeout",
            50: "manufacturer specific",
            51: "manufacturer specific",
            52: "manufacturer specific",
            53: "manufacturer specific",
            54: "manufacturer specific",
            55: "manufacturer specific",
            56: "manufacturer specific",
            57: "manufacturer specific",
            58: "manufacturer specific",
            59: "manufacturer specific",
            6: "remote command out of range/invalid",
            60: "manufacturer specific",
            61: "manufacturer specific",
            62: "error",
            63: "not available/not installed",
            7: "function not calibrated",
            8: "operator control fault",
            9: "function fault"
        },
        5801: {
            0: "no reason/all clear",
            1: "operator presence not detected",
            10: "reserved",
            11: "reserved",
            12: "reserved",
            13: "reserved",
            14: "reserved",
            15: "reserved",
            16: "reserved",
            17: "reserved",
            18: "reserved",
            19: "reserved",
            2: "implement released control of function",
            20: "hydraulic oil level too low",
            21: "valve locked out",
            22: "reserved",
            23: "reserved",
            24: "reserved",
            25: "reserved",
            26: "reserved",
            27: "reserved",
            28: "reserved",
            29: "reserved",
            3: "operator override of function",
            30: "reserved",
            31: "reserved",
            32: "reserved",
            33: "reserved",
            34: "reserved",
            35: "reserved",
            36: "reserved",
            37: "reserved",
            38: "reserved",
            39: "reserved",
            4: "operator control not in valid position",
            40: "reserved",
            41: "manufacturer specific",
            42: "manufacturer specific",
            43: "manufacturer specific",
            44: "manufacturer specific",
            45: "manufacturer specific",
            46: "manufacturer specific",
            47: "manufacturer specific",
            48: "manufacturer specific",
            49: "manufacturer specific",
            5: "remote command timeout",
            50: "manufacturer specific",
            51: "manufacturer specific",
            52: "manufacturer specific",
            53: "manufacturer specific",
            54: "manufacturer specific",
            55: "manufacturer specific",
            56: "manufacturer specific",
            57: "manufacturer specific",
            58: "manufacturer specific",
            59: "manufacturer specific",
            6: "remote command out of range/invalid",
            60: "manufacturer specific",
            61: "manufacturer specific",
            62: "error",
            63: "not available/not installed",
            7: "function not calibrated",
            8: "operator control fault",
            9: "function fault"
        },
        5802: {
            0: "no reason/all clear",
            1: "operator presence not detected",
            10: "reserved",
            11: "reserved",
            12: "reserved",
            13: "reserved",
            14: "reserved",
            15: "reserved",
            16: "reserved",
            17: "reserved",
            18: "reserved",
            19: "reserved",
            2: "implement released control of function",
            20: "hydraulic oil level too low",
            21: "valve locked out",
            22: "reserved",
            23: "reserved",
            24: "reserved",
            25: "reserved",
            26: "reserved",
            27: "reserved",
            28: "reserved",
            29: "reserved",
            3: "operator override of function",
            30: "reserved",
            31: "reserved",
            32: "reserved",
            33: "reserved",
            34: "reserved",
            35: "reserved",
            36: "reserved",
            37: "reserved",
            38: "reserved",
            39: "reserved",
            4: "operator control not in valid position",
            40: "reserved",
            41: "manufacturer specific",
            42: "manufacturer specific",
            43: "manufacturer specific",
            44: "manufacturer specific",
            45: "manufacturer specific",
            46: "manufacturer specific",
            47: "manufacturer specific",
            48: "manufacturer specific",
            49: "manufacturer specific",
            5: "remote command timeout",
            50: "manufacturer specific",
            51: "manufacturer specific",
            52: "manufacturer specific",
            53: "manufacturer specific",
            54: "manufacturer specific",
            55: "manufacturer specific",
            56: "manufacturer specific",
            57: "manufacturer specific",
            58: "manufacturer specific",
            59: "manufacturer specific",
            6: "remote command out of range/invalid",
            60: "manufacturer specific",
            61: "manufacturer specific",
            62: "error",
            63: "not available/not installed",
            7: "function not calibrated",
            8: "operator control fault",
            9: "function fault"
        },
        5803: {
            0: "no reason/all clear",
            1: "operator presence not detected",
            10: "reserved",
            11: "reserved",
            12: "reserved",
            13: "reserved",
            14: "reserved",
            15: "reserved",
            16: "reserved",
            17: "reserved",
            18: "reserved",
            19: "reserved",
            2: "implement released control of function",
            20: "hydraulic oil level too low",
            21: "valve locked out",
            22: "reserved",
            23: "reserved",
            24: "reserved",
            25: "reserved",
            26: "reserved",
            27: "reserved",
            28: "reserved",
            29: "reserved",
            3: "operator override of function",
            30: "reserved",
            31: "reserved",
            32: "reserved",
            33: "reserved",
            34: "reserved",
            35: "reserved",
            36: "reserved",
            37: "reserved",
            38: "reserved",
            39: "reserved",
            4: "operator control not in valid position",
            40: "reserved",
            41: "manufacturer specific",
            42: "manufacturer specific",
            43: "manufacturer specific",
            44: "manufacturer specific",
            45: "manufacturer specific",
            46: "manufacturer specific",
            47: "manufacturer specific",
            48: "manufacturer specific",
            49: "manufacturer specific",
            5: "remote command timeout",
            50: "manufacturer specific",
            51: "manufacturer specific",
            52: "manufacturer specific",
            53: "manufacturer specific",
            54: "manufacturer specific",
            55: "manufacturer specific",
            56: "manufacturer specific",
            57: "manufacturer specific",
            58: "manufacturer specific",
            59: "manufacturer specific",
            6: "remote command out of range/invalid",
            60: "manufacturer specific",
            61: "manufacturer specific",
            62: "error",
            63: "not available/not installed",
            7: "function not calibrated",
            8: "operator control fault",
            9: "function fault"
        },
        5804: {
            0: "no reason/all clear",
            1: "operator presence not detected",
            10: "reserved",
            11: "reserved",
            12: "reserved",
            13: "reserved",
            14: "reserved",
            15: "reserved",
            16: "reserved",
            17: "reserved",
            18: "reserved",
            19: "reserved",
            2: "implement released control of function",
            20: "hydraulic oil level too low",
            21: "valve locked out",
            22: "reserved",
            23: "reserved",
            24: "reserved",
            25: "reserved",
            26: "reserved",
            27: "reserved",
            28: "reserved",
            29: "reserved",
            3: "operator override of function",
            30: "reserved",
            31: "reserved",
            32: "reserved",
            33: "reserved",
            34: "reserved",
            35: "reserved",
            36: "reserved",
            37: "reserved",
            38: "reserved",
            39: "reserved",
            4: "operator control not in valid position",
            40: "reserved",
            41: "manufacturer specific",
            42: "manufacturer specific",
            43: "manufacturer specific",
            44: "manufacturer specific",
            45: "manufacturer specific",
            46: "manufacturer specific",
            47: "manufacturer specific",
            48: "manufacturer specific",
            49: "manufacturer specific",
            5: "remote command timeout",
            50: "manufacturer specific",
            51: "manufacturer specific",
            52: "manufacturer specific",
            53: "manufacturer specific",
            54: "manufacturer specific",
            55: "manufacturer specific",
            56: "manufacturer specific",
            57: "manufacturer specific",
            58: "manufacturer specific",
            59: "manufacturer specific",
            6: "remote command out of range/invalid",
            60: "manufacturer specific",
            61: "manufacturer specific",
            62: "error",
            63: "not available/not installed",
            7: "function not calibrated",
            8: "operator control fault",
            9: "function fault"
        },
        5805: {
            0: "no reason/all clear",
            1: "operator presence not detected",
            10: "reserved",
            11: "reserved",
            12: "reserved",
            13: "reserved",
            14: "reserved",
            15: "reserved",
            16: "reserved",
            17: "reserved",
            18: "reserved",
            19: "reserved",
            2: "implement released control of function",
            20: "hydraulic oil level too low",
            21: "valve locked out",
            22: "reserved",
            23: "reserved",
            24: "reserved",
            25: "reserved",
            26: "reserved",
            27: "reserved",
            28: "reserved",
            29: "reserved",
            3: "operator override of function",
            30: "reserved",
            31: "reserved",
            32: "reserved",
            33: "reserved",
            34: "reserved",
            35: "reserved",
            36: "reserved",
            37: "reserved",
            38: "reserved",
            39: "reserved",
            4: "operator control not in valid position",
            40: "reserved",
            41: "manufacturer specific",
            42: "manufacturer specific",
            43: "manufacturer specific",
            44: "manufacturer specific",
            45: "manufacturer specific",
            46: "manufacturer specific",
            47: "manufacturer specific",
            48: "manufacturer specific",
            49: "manufacturer specific",
            5: "remote command timeout",
            50: "manufacturer specific",
            51: "manufacturer specific",
            52: "manufacturer specific",
            53: "manufacturer specific",
            54: "manufacturer specific",
            55: "manufacturer specific",
            56: "manufacturer specific",
            57: "manufacturer specific",
            58: "manufacturer specific",
            59: "manufacturer specific",
            6: "remote command out of range/invalid",
            60: "manufacturer specific",
            61: "manufacturer specific",
            62: "error",
            63: "not available/not installed",
            7: "function not calibrated",
            8: "operator control fault",
            9: "function fault"
        },
        5806: {
            0: "no reason/all clear",
            1: "operator presence not detected",
            10: "reserved",
            11: "reserved",
            12: "reserved",
            13: "reserved",
            14: "reserved",
            15: "reserved",
            16: "reserved",
            17: "reserved",
            18: "reserved",
            19: "reserved",
            2: "implement released control of function",
            20: "hydraulic oil level too low",
            21: "valve locked out",
            22: "reserved",
            23: "reserved",
            24: "reserved",
            25: "reserved",
            26: "reserved",
            27: "reserved",
            28: "reserved",
            29: "reserved",
            3: "operator override of function",
            30: "reserved",
            31: "reserved",
            32: "reserved",
            33: "reserved",
            34: "reserved",
            35: "reserved",
            36: "reserved",
            37: "reserved",
            38: "reserved",
            39: "reserved",
            4: "operator control not in valid position",
            40: "reserved",
            41: "manufacturer specific",
            42: "manufacturer specific",
            43: "manufacturer specific",
            44: "manufacturer specific",
            45: "manufacturer specific",
            46: "manufacturer specific",
            47: "manufacturer specific",
            48: "manufacturer specific",
            49: "manufacturer specific",
            5: "remote command timeout",
            50: "manufacturer specific",
            51: "manufacturer specific",
            52: "manufacturer specific",
            53: "manufacturer specific",
            54: "manufacturer specific",
            55: "manufacturer specific",
            56: "manufacturer specific",
            57: "manufacturer specific",
            58: "manufacturer specific",
            59: "manufacturer specific",
            6: "remote command out of range/invalid",
            60: "manufacturer specific",
            61: "manufacturer specific",
            62: "error",
            63: "not available/not installed",
            7: "function not calibrated",
            8: "operator control fault",
            9: "function fault"
        },
        5807: {
            0: "no reason/all clear",
            1: "operator presence not detected",
            10: "reserved",
            11: "reserved",
            12: "reserved",
            13: "reserved",
            14: "reserved",
            15: "reserved",
            16: "reserved",
            17: "reserved",
            18: "reserved",
            19: "reserved",
            2: "implement released control of function",
            20: "hydraulic oil level too low",
            21: "valve locked out",
            22: "reserved",
            23: "reserved",
            24: "reserved",
            25: "reserved",
            26: "reserved",
            27: "reserved",
            28: "reserved",
            29: "reserved",
            3: "operator override of function",
            30: "reserved",
            31: "reserved",
            32: "reserved",
            33: "reserved",
            34: "reserved",
            35: "reserved",
            36: "reserved",
            37: "reserved",
            38: "reserved",
            39: "reserved",
            4: "operator control not in valid position",
            40: "reserved",
            41: "manufacturer specific",
            42: "manufacturer specific",
            43: "manufacturer specific",
            44: "manufacturer specific",
            45: "manufacturer specific",
            46: "manufacturer specific",
            47: "manufacturer specific",
            48: "manufacturer specific",
            49: "manufacturer specific",
            5: "remote command timeout",
            50: "manufacturer specific",
            51: "manufacturer specific",
            52: "manufacturer specific",
            53: "manufacturer specific",
            54: "manufacturer specific",
            55: "manufacturer specific",
            56: "manufacturer specific",
            57: "manufacturer specific",
            58: "manufacturer specific",
            59: "manufacturer specific",
            6: "remote command out of range/invalid",
            60: "manufacturer specific",
            61: "manufacturer specific",
            62: "error",
            63: "not available/not installed",
            7: "function not calibrated",
            8: "operator control fault",
            9: "function fault"
        },
        5808: {
            0: "no reason/all clear",
            1: "operator presence not detected",
            10: "reserved",
            11: "reserved",
            12: "reserved",
            13: "reserved",
            14: "reserved",
            15: "reserved",
            16: "reserved",
            17: "reserved",
            18: "reserved",
            19: "reserved",
            2: "implement released control of function",
            20: "hydraulic oil level too low",
            21: "valve locked out",
            22: "reserved",
            23: "reserved",
            24: "reserved",
            25: "reserved",
            26: "reserved",
            27: "reserved",
            28: "reserved",
            29: "reserved",
            3: "operator override of function",
            30: "reserved",
            31: "reserved",
            32: "reserved",
            33: "reserved",
            34: "reserved",
            35: "reserved",
            36: "reserved",
            37: "reserved",
            38: "reserved",
            39: "reserved",
            4: "operator control not in valid position",
            40: "reserved",
            41: "manufacturer specific",
            42: "manufacturer specific",
            43: "manufacturer specific",
            44: "manufacturer specific",
            45: "manufacturer specific",
            46: "manufacturer specific",
            47: "manufacturer specific",
            48: "manufacturer specific",
            49: "manufacturer specific",
            5: "remote command timeout",
            50: "manufacturer specific",
            51: "manufacturer specific",
            52: "manufacturer specific",
            53: "manufacturer specific",
            54: "manufacturer specific",
            55: "manufacturer specific",
            56: "manufacturer specific",
            57: "manufacturer specific",
            58: "manufacturer specific",
            59: "manufacturer specific",
            6: "remote command out of range/invalid",
            60: "manufacturer specific",
            61: "manufacturer specific",
            62: "error",
            63: "not available/not installed",
            7: "function not calibrated",
            8: "operator control fault",
            9: "function fault"
        },
        5809: {
            0: "no reason/all clear",
            1: "operator presence not detected",
            10: "reserved",
            11: "reserved",
            12: "reserved",
            13: "reserved",
            14: "reserved",
            15: "reserved",
            16: "reserved",
            17: "reserved",
            18: "reserved",
            19: "reserved",
            2: "implement released control of function",
            20: "hydraulic oil level too low",
            21: "valve locked out",
            22: "reserved",
            23: "reserved",
            24: "reserved",
            25: "reserved",
            26: "reserved",
            27: "reserved",
            28: "reserved",
            29: "reserved",
            3: "operator override of function",
            30: "reserved",
            31: "reserved",
            32: "reserved",
            33: "reserved",
            34: "reserved",
            35: "reserved",
            36: "reserved",
            37: "reserved",
            38: "reserved",
            39: "reserved",
            4: "operator control not in valid position",
            40: "reserved",
            41: "manufacturer specific",
            42: "manufacturer specific",
            43: "manufacturer specific",
            44: "manufacturer specific",
            45: "manufacturer specific",
            46: "manufacturer specific",
            47: "manufacturer specific",
            48: "manufacturer specific",
            49: "manufacturer specific",
            5: "remote command timeout",
            50: "manufacturer specific",
            51: "manufacturer specific",
            52: "manufacturer specific",
            53: "manufacturer specific",
            54: "manufacturer specific",
            55: "manufacturer specific",
            56: "manufacturer specific",
            57: "manufacturer specific",
            58: "manufacturer specific",
            59: "manufacturer specific",
            6: "remote command out of range/invalid",
            60: "manufacturer specific",
            61: "manufacturer specific",
            62: "error",
            63: "not available/not installed",
            7: "function not calibrated",
            8: "operator control fault",
            9: "function fault"
        },
        5810: {
            0: "no reason/all clear",
            1: "operator presence not detected",
            10: "reserved",
            11: "reserved",
            12: "reserved",
            13: "reserved",
            14: "reserved",
            15: "reserved",
            16: "reserved",
            17: "reserved",
            18: "reserved",
            19: "reserved",
            2: "implement released control of function",
            20: "hydraulic oil level too low",
            21: "valve locked out",
            22: "reserved",
            23: "reserved",
            24: "reserved",
            25: "reserved",
            26: "reserved",
            27: "reserved",
            28: "reserved",
            29: "reserved",
            3: "operator override of function",
            30: "reserved",
            31: "reserved",
            32: "reserved",
            33: "reserved",
            34: "reserved",
            35: "reserved",
            36: "reserved",
            37: "reserved",
            38: "reserved",
            39: "reserved",
            4: "operator control not in valid position",
            40: "reserved",
            41: "manufacturer specific",
            42: "manufacturer specific",
            43: "manufacturer specific",
            44: "manufacturer specific",
            45: "manufacturer specific",
            46: "manufacturer specific",
            47: "manufacturer specific",
            48: "manufacturer specific",
            49: "manufacturer specific",
            5: "remote command timeout",
            50: "manufacturer specific",
            51: "manufacturer specific",
            52: "manufacturer specific",
            53: "manufacturer specific",
            54: "manufacturer specific",
            55: "manufacturer specific",
            56: "manufacturer specific",
            57: "manufacturer specific",
            58: "manufacturer specific",
            59: "manufacturer specific",
            6: "remote command out of range/invalid",
            60: "manufacturer specific",
            61: "manufacturer specific",
            62: "error",
            63: "not available/not installed",
            7: "function not calibrated",
            8: "operator control fault",
            9: "function fault"
        },
        5811: {
            0: "no reason/all clear",
            1: "operator presence not detected",
            10: "reserved",
            11: "reserved",
            12: "reserved",
            13: "reserved",
            14: "reserved",
            15: "reserved",
            16: "reserved",
            17: "reserved",
            18: "reserved",
            19: "reserved",
            2: "implement released control of function",
            20: "hydraulic oil level too low",
            21: "valve locked out",
            22: "reserved",
            23: "reserved",
            24: "reserved",
            25: "reserved",
            26: "reserved",
            27: "reserved",
            28: "reserved",
            29: "reserved",
            3: "operator override of function",
            30: "reserved",
            31: "reserved",
            32: "reserved",
            33: "reserved",
            34: "reserved",
            35: "reserved",
            36: "reserved",
            37: "reserved",
            38: "reserved",
            39: "reserved",
            4: "operator control not in valid position",
            40: "reserved",
            41: "manufacturer specific",
            42: "manufacturer specific",
            43: "manufacturer specific",
            44: "manufacturer specific",
            45: "manufacturer specific",
            46: "manufacturer specific",
            47: "manufacturer specific",
            48: "manufacturer specific",
            49: "manufacturer specific",
            5: "remote command timeout",
            50: "manufacturer specific",
            51: "manufacturer specific",
            52: "manufacturer specific",
            53: "manufacturer specific",
            54: "manufacturer specific",
            55: "manufacturer specific",
            56: "manufacturer specific",
            57: "manufacturer specific",
            58: "manufacturer specific",
            59: "manufacturer specific",
            6: "remote command out of range/invalid",
            60: "manufacturer specific",
            61: "manufacturer specific",
            62: "error",
            63: "not available/not installed",
            7: "function not calibrated",
            8: "operator control fault",
            9: "function fault"
        },
        5812: {
            0: "no reason/all clear",
            1: "operator presence not detected",
            10: "reserved",
            11: "reserved",
            12: "reserved",
            13: "reserved",
            14: "reserved",
            15: "reserved",
            16: "reserved",
            17: "reserved",
            18: "reserved",
            19: "reserved",
            2: "implement released control of function",
            20: "hydraulic oil level too low",
            21: "valve locked out",
            22: "reserved",
            23: "reserved",
            24: "reserved",
            25: "reserved",
            26: "reserved",
            27: "reserved",
            28: "reserved",
            29: "reserved",
            3: "operator override of function",
            30: "reserved",
            31: "reserved",
            32: "reserved",
            33: "reserved",
            34: "reserved",
            35: "reserved",
            36: "reserved",
            37: "reserved",
            38: "reserved",
            39: "reserved",
            4: "operator control not in valid position",
            40: "reserved",
            41: "manufacturer specific",
            42: "manufacturer specific",
            43: "manufacturer specific",
            44: "manufacturer specific",
            45: "manufacturer specific",
            46: "manufacturer specific",
            47: "manufacturer specific",
            48: "manufacturer specific",
            49: "manufacturer specific",
            5: "remote command timeout",
            50: "manufacturer specific",
            51: "manufacturer specific",
            52: "manufacturer specific",
            53: "manufacturer specific",
            54: "manufacturer specific",
            55: "manufacturer specific",
            56: "manufacturer specific",
            57: "manufacturer specific",
            58: "manufacturer specific",
            59: "manufacturer specific",
            6: "remote command out of range/invalid",
            60: "manufacturer specific",
            61: "manufacturer specific",
            62: "error",
            63: "not available/not installed",
            7: "function not calibrated",
            8: "operator control fault",
            9: "function fault"
        },
        5813: {
            0: "no reason/all clear",
            1: "operator presence not detected",
            10: "reserved",
            11: "reserved",
            12: "reserved",
            13: "reserved",
            14: "reserved",
            15: "reserved",
            16: "reserved",
            17: "reserved",
            18: "reserved",
            19: "reserved",
            2: "implement released control of function",
            20: "hydraulic oil level too low",
            21: "valve locked out",
            22: "reserved",
            23: "reserved",
            24: "reserved",
            25: "reserved",
            26: "reserved",
            27: "reserved",
            28: "reserved",
            29: "reserved",
            3: "operator override of function",
            30: "reserved",
            31: "reserved",
            32: "reserved",
            33: "reserved",
            34: "reserved",
            35: "reserved",
            36: "reserved",
            37: "reserved",
            38: "reserved",
            39: "reserved",
            4: "operator control not in valid position",
            40: "reserved",
            41: "manufacturer specific",
            42: "manufacturer specific",
            43: "manufacturer specific",
            44: "manufacturer specific",
            45: "manufacturer specific",
            46: "manufacturer specific",
            47: "manufacturer specific",
            48: "manufacturer specific",
            49: "manufacturer specific",
            5: "remote command timeout",
            50: "manufacturer specific",
            51: "manufacturer specific",
            52: "manufacturer specific",
            53: "manufacturer specific",
            54: "manufacturer specific",
            55: "manufacturer specific",
            56: "manufacturer specific",
            57: "manufacturer specific",
            58: "manufacturer specific",
            59: "manufacturer specific",
            6: "remote command out of range/invalid",
            60: "manufacturer specific",
            61: "manufacturer specific",
            62: "error",
            63: "not available/not installed",
            7: "function not calibrated",
            8: "operator control fault",
            9: "function fault"
        },
        5814: {
            0: "no reason/all clear",
            1: "operator presence not detected",
            10: "reserved",
            11: "reserved",
            12: "reserved",
            13: "reserved",
            14: "reserved",
            15: "reserved",
            16: "reserved",
            17: "reserved",
            18: "reserved",
            19: "reserved",
            2: "implement released control of function",
            20: "hydraulic oil level too low",
            21: "valve locked out",
            22: "reserved",
            23: "reserved",
            24: "reserved",
            25: "reserved",
            26: "reserved",
            27: "reserved",
            28: "reserved",
            29: "reserved",
            3: "operator override of function",
            30: "reserved",
            31: "reserved",
            32: "reserved",
            33: "reserved",
            34: "reserved",
            35: "reserved",
            36: "reserved",
            37: "reserved",
            38: "reserved",
            39: "reserved",
            4: "operator control not in valid position",
            40: "reserved",
            41: "manufacturer specific",
            42: "manufacturer specific",
            43: "manufacturer specific",
            44: "manufacturer specific",
            45: "manufacturer specific",
            46: "manufacturer specific",
            47: "manufacturer specific",
            48: "manufacturer specific",
            49: "manufacturer specific",
            5: "remote command timeout",
            50: "manufacturer specific",
            51: "manufacturer specific",
            52: "manufacturer specific",
            53: "manufacturer specific",
            54: "manufacturer specific",
            55: "manufacturer specific",
            56: "manufacturer specific",
            57: "manufacturer specific",
            58: "manufacturer specific",
            59: "manufacturer specific",
            6: "remote command out of range/invalid",
            60: "manufacturer specific",
            61: "manufacturer specific",
            62: "error",
            63: "not available/not installed",
            7: "function not calibrated",
            8: "operator control fault",
            9: "function fault"
        },
        5815: {
            0: "no reason/all clear",
            1: "operator presence not detected",
            10: "reserved",
            11: "reserved",
            12: "reserved",
            13: "reserved",
            14: "reserved",
            15: "reserved",
            16: "reserved",
            17: "reserved",
            18: "reserved",
            19: "reserved",
            2: "implement released control of function",
            20: "hydraulic oil level too low",
            21: "valve locked out",
            22: "reserved",
            23: "reserved",
            24: "reserved",
            25: "reserved",
            26: "reserved",
            27: "reserved",
            28: "reserved",
            29: "reserved",
            3: "operator override of function",
            30: "reserved",
            31: "reserved",
            32: "reserved",
            33: "reserved",
            34: "reserved",
            35: "reserved",
            36: "reserved",
            37: "reserved",
            38: "reserved",
            39: "reserved",
            4: "operator control not in valid position",
            40: "reserved",
            41: "manufacturer specific",
            42: "manufacturer specific",
            43: "manufacturer specific",
            44: "manufacturer specific",
            45: "manufacturer specific",
            46: "manufacturer specific",
            47: "manufacturer specific",
            48: "manufacturer specific",
            49: "manufacturer specific",
            5: "remote command timeout",
            50: "manufacturer specific",
            51: "manufacturer specific",
            52: "manufacturer specific",
            53: "manufacturer specific",
            54: "manufacturer specific",
            55: "manufacturer specific",
            56: "manufacturer specific",
            57: "manufacturer specific",
            58: "manufacturer specific",
            59: "manufacturer specific",
            6: "remote command out of range/invalid",
            60: "manufacturer specific",
            61: "manufacturer specific",
            62: "error",
            63: "not available/not installed",
            7: "function not calibrated",
            8: "operator control fault",
            9: "function fault"
        },
        5816: {
            0: "no reason/all clear",
            1: "operator presence not detected",
            10: "reserved",
            11: "reserved",
            12: "reserved",
            13: "reserved",
            14: "reserved",
            15: "reserved",
            16: "reserved",
            17: "reserved",
            18: "reserved",
            19: "reserved",
            2: "implement released control of function",
            20: "hydraulic oil level too low",
            21: "valve locked out",
            22: "reserved",
            23: "reserved",
            24: "reserved",
            25: "reserved",
            26: "reserved",
            27: "reserved",
            28: "reserved",
            29: "reserved",
            3: "operator override of function",
            30: "reserved",
            31: "reserved",
            32: "reserved",
            33: "reserved",
            34: "reserved",
            35: "reserved",
            36: "reserved",
            37: "reserved",
            38: "reserved",
            39: "reserved",
            4: "operator control not in valid position",
            40: "reserved",
            41: "manufacturer specific",
            42: "manufacturer specific",
            43: "manufacturer specific",
            44: "manufacturer specific",
            45: "manufacturer specific",
            46: "manufacturer specific",
            47: "manufacturer specific",
            48: "manufacturer specific",
            49: "manufacturer specific",
            5: "remote command timeout",
            50: "manufacturer specific",
            51: "manufacturer specific",
            52: "manufacturer specific",
            53: "manufacturer specific",
            54: "manufacturer specific",
            55: "manufacturer specific",
            56: "manufacturer specific",
            57: "manufacturer specific",
            58: "manufacturer specific",
            59: "manufacturer specific",
            6: "remote command out of range/invalid",
            60: "manufacturer specific",
            61: "manufacturer specific",
            62: "error",
            63: "not available/not installed",
            7: "function not calibrated",
            8: "operator control fault",
            9: "function fault"
        },
        5817: {
            0: "no reason/all clear",
            1: "operator presence not detected",
            10: "reserved",
            11: "reserved",
            12: "reserved",
            13: "reserved",
            14: "reserved",
            15: "reserved",
            16: "reserved",
            17: "reserved",
            18: "reserved",
            19: "reserved",
            2: "implement released control of function",
            20: "hydraulic oil level too low",
            21: "valve locked out",
            22: "reserved",
            23: "reserved",
            24: "reserved",
            25: "reserved",
            26: "reserved",
            27: "reserved",
            28: "reserved",
            29: "reserved",
            3: "operator override of function",
            30: "reserved",
            31: "reserved",
            32: "reserved",
            33: "reserved",
            34: "reserved",
            35: "reserved",
            36: "reserved",
            37: "reserved",
            38: "reserved",
            39: "reserved",
            4: "operator control not in valid position",
            40: "reserved",
            41: "manufacturer specific",
            42: "manufacturer specific",
            43: "manufacturer specific",
            44: "manufacturer specific",
            45: "manufacturer specific",
            46: "manufacturer specific",
            47: "manufacturer specific",
            48: "manufacturer specific",
            49: "manufacturer specific",
            5: "remote command timeout",
            50: "manufacturer specific",
            51: "manufacturer specific",
            52: "manufacturer specific",
            53: "manufacturer specific",
            54: "manufacturer specific",
            55: "manufacturer specific",
            56: "manufacturer specific",
            57: "manufacturer specific",
            58: "manufacturer specific",
            59: "manufacturer specific",
            6: "remote command out of range/invalid",
            60: "manufacturer specific",
            61: "manufacturer specific",
            62: "error",
            63: "not available/not installed",
            7: "function not calibrated",
            8: "operator control fault",
            9: "function fault"
        },
        5818: {
            0: "no reason/all clear",
            1: "operator presence not detected",
            10: "reserved",
            11: "reserved",
            12: "reserved",
            13: "reserved",
            14: "reserved",
            15: "reserved",
            16: "reserved",
            17: "reserved",
            18: "reserved",
            19: "reserved",
            2: "implement released control of function",
            20: "hydraulic oil level too low",
            21: "valve locked out",
            22: "reserved",
            23: "reserved",
            24: "reserved",
            25: "reserved",
            26: "reserved",
            27: "reserved",
            28: "reserved",
            29: "reserved",
            3: "operator override of function",
            30: "reserved",
            31: "reserved",
            32: "reserved",
            33: "reserved",
            34: "reserved",
            35: "reserved",
            36: "reserved",
            37: "reserved",
            38: "reserved",
            39: "reserved",
            4: "operator control not in valid position",
            40: "reserved",
            41: "manufacturer specific",
            42: "manufacturer specific",
            43: "manufacturer specific",
            44: "manufacturer specific",
            45: "manufacturer specific",
            46: "manufacturer specific",
            47: "manufacturer specific",
            48: "manufacturer specific",
            49: "manufacturer specific",
            5: "remote command timeout",
            50: "manufacturer specific",
            51: "manufacturer specific",
            52: "manufacturer specific",
            53: "manufacturer specific",
            54: "manufacturer specific",
            55: "manufacturer specific",
            56: "manufacturer specific",
            57: "manufacturer specific",
            58: "manufacturer specific",
            59: "manufacturer specific",
            6: "remote command out of range/invalid",
            60: "manufacturer specific",
            61: "manufacturer specific",
            62: "error",
            63: "not available/not installed",
            7: "function not calibrated",
            8: "operator control fault",
            9: "function fault"
        },
        5819: {
            0: "no reason/all clear",
            1: "operator presence not detected",
            10: "reserved",
            11: "reserved",
            12: "reserved",
            13: "reserved",
            14: "reserved",
            15: "reserved",
            16: "reserved",
            17: "reserved",
            18: "reserved",
            19: "reserved",
            2: "implement released control of function",
            20: "hydraulic oil level too low",
            21: "valve locked out",
            22: "reserved",
            23: "reserved",
            24: "reserved",
            25: "reserved",
            26: "reserved",
            27: "reserved",
            28: "reserved",
            29: "reserved",
            3: "operator override of function",
            30: "reserved",
            31: "reserved",
            32: "reserved",
            33: "reserved",
            34: "reserved",
            35: "reserved",
            36: "reserved",
            37: "reserved",
            38: "reserved",
            39: "reserved",
            4: "operator control not in valid position",
            40: "reserved",
            41: "manufacturer specific",
            42: "manufacturer specific",
            43: "manufacturer specific",
            44: "manufacturer specific",
            45: "manufacturer specific",
            46: "manufacturer specific",
            47: "manufacturer specific",
            48: "manufacturer specific",
            49: "manufacturer specific",
            5: "remote command timeout",
            50: "manufacturer specific",
            51: "manufacturer specific",
            52: "manufacturer specific",
            53: "manufacturer specific",
            54: "manufacturer specific",
            55: "manufacturer specific",
            56: "manufacturer specific",
            57: "manufacturer specific",
            58: "manufacturer specific",
            59: "manufacturer specific",
            6: "remote command out of range/invalid",
            60: "manufacturer specific",
            61: "manufacturer specific",
            62: "error",
            63: "not available/not installed",
            7: "function not calibrated",
            8: "operator control fault",
            9: "function fault"
        },
        5820: {
            0: "no reason/all clear",
            1: "operator presence not detected",
            10: "reserved",
            11: "reserved",
            12: "reserved",
            13: "reserved",
            14: "reserved",
            15: "reserved",
            16: "reserved",
            17: "reserved",
            18: "reserved",
            19: "reserved",
            2: "implement released control of function",
            20: "hydraulic oil level too low",
            21: "valve locked out",
            22: "reserved",
            23: "reserved",
            24: "reserved",
            25: "reserved",
            26: "reserved",
            27: "reserved",
            28: "reserved",
            29: "reserved",
            3: "operator override of function",
            30: "reserved",
            31: "reserved",
            32: "reserved",
            33: "reserved",
            34: "reserved",
            35: "reserved",
            36: "reserved",
            37: "reserved",
            38: "reserved",
            39: "reserved",
            4: "operator control not in valid position",
            40: "reserved",
            41: "manufacturer specific",
            42: "manufacturer specific",
            43: "manufacturer specific",
            44: "manufacturer specific",
            45: "manufacturer specific",
            46: "manufacturer specific",
            47: "manufacturer specific",
            48: "manufacturer specific",
            49: "manufacturer specific",
            5: "remote command timeout",
            50: "manufacturer specific",
            51: "manufacturer specific",
            52: "manufacturer specific",
            53: "manufacturer specific",
            54: "manufacturer specific",
            55: "manufacturer specific",
            56: "manufacturer specific",
            57: "manufacturer specific",
            58: "manufacturer specific",
            59: "manufacturer specific",
            6: "remote command out of range/invalid",
            60: "manufacturer specific",
            61: "manufacturer specific",
            62: "error",
            63: "not available/not installed",
            7: "function not calibrated",
            8: "operator control fault",
            9: "function fault"
        },
        5822: {
            0: "    sensor fully functional",
            1: "    sensor degraded",
            2: "    sensor failed",
            3: "    sensor not available"
        },
        5824: {
            0: "    sensor fully functional",
            1: "    sensor degraded",
            2: "    sensor failed",
            3: "    sensor not available"
        },
        5825: {
            0: "off - used when driver warning system is non-active - which, by requirement, means the low-level and severe inducement systems are also non-active",
            1: "on - solid - used when driver warning system is activated for any relevant malfunction",
            2: "reserved for sae assignment",
            3: "reserved for sae assignment",
            4: "on - fast blink (1 hz) - used for the second driver warning system detection threshold for diesel exhaust fluid level, where the visual alarm of the driver warning system is required to \"escalate in intensity\"",
            5: "reserved for sae assignment",
            6: "reserved for sae assignment",
            7: "not available"
        },
        5826: {
            0: "driver warning, low-level inducement, and severe inducement non-active",
            1: "inducement level 1... euro vi defined driver warning system active ............................ fmi 15",
            2: "inducement level 2 ... euro vi defined low-level inducement enabled ......................... fmi 15",
            3: "inducement level 3 ... euro vi defined low-level inducement active ............................ fmi 16",
            4: "inducement level 4 ... euro vi defined severe inducement enabled .............................. fmi 16",
            5: "inducement level 5 ... euro vi defined severe inducement active ................................. fmi 0",
            6: "temporary override of inducement",
            7: "not available / not supported"
        },
        5837: {
            0: "xff - not available"
        },
        5843: {
            0: "a warm-up cycle has not occurred on this engine start",
            1: "a warm-up cycle has occurred on this engine start",
            2: "sae reserved",
            3: "not supported or not available"
        },
        5850: {
            0: "reading is not stable",
            1: "reading is stable",
            2: "error",
            3: "not available"
        },
        5851: {
            0: "power is not in range",
            1: "power is in range",
            2: "error",
            3: "not available"
        },
        5852: {
            0: "heater element is not in range",
            1: "heater element is in range",
            2: "error",
            3: "not available"
        },
        5854: {
            0: "automatic ",
            1: "preheat 2 ",
            2: "preheat 1 ",
            3: "heater off or not available"
        },
        5857: {
            0: "reading is not stable",
            1: "reading is stable",
            2: "error",
            3: "not available"
        },
        5858: {
            0: "heater element is not in range",
            1: "heater element is in range",
            2: "error",
            3: "not available"
        },
        5859: {
            0: "heater element is not in range",
            1: "heater element is in range",
            2: "error",
            3: "not available"
        },
        5861: {
            0: "automatic ",
            1: "preheat 2 ",
            2: "preheat 1 ",
            3: "heater off or not available"
        },
        5866: {
            0: "fueling inhibit allowed (spn 5795 = 01)",
            1: "dpf regeneration is active or preparing to be active",
            15: "not supported",
            2: "oil distribution in engine is incomplete",
            3: "engine coolant temperature too low",
            4: "service override",
            5: " to 14 - reserved"
        },
        5872: {
            0: "hybrid system warning indicator is off",
            1: "hybrid system warning indicator is on continuously",
            2: "hybrid system warning indicator is flashing",
            3: "not available"
        },
        5873: {
            0: "hybrid system temperature indicator is off",
            1: "hybrid system temperature indicator is on continuously",
            2: "hybrid system temperature indicator is flashing",
            3: "not available"
        },
        5874: {
            0: "hybrid system stop indicator is off",
            1: "hybrid system stop indicator is on continuously",
            2: "hybrid system stop indicator is flashing",
            3: "not available"
        },
        590: {
            0: "inactive",
            1: "active",
            2: "error",
            3: "not available"
        },
        591: {
            0: "disabled in calibration",
            1: "enabled in calibration",
            2: "error",
            3: "not available"
        },
        5915: {
            0: " less than 1 minute",
            1: " one minute",
            10: " ten minutes",
            11: " eleven minutes",
            12: " twelve minutes",
            13: " thirteen minutes",
            14: " error",
            15: " not available",
            2: " two minutes",
            3: " three minutes",
            4: " four minutes",
            5: " five minutes",
            6: " six minutes",
            7: " seven minutes",
            8: " eight minutes",
            9: " nine minutes"
        },
        5916: {
            0: " conditions valid for transmission oil level measurement",
            1: " conditions not valid - settling timer still counting down",
            10: " not defined",
            11: " not defined",
            12: " not defined",
            13: " conditions not valid - other",
            14: " error",
            15: " not available",
            2: " conditions not valid - transmission in gear",
            3: " conditions not valid - transmission fluid temperature too low",
            4: " conditions not valid - transmission fluid temperature too high",
            5: " conditions not valid - vehicle moving; output shaft speed too high",
            6: " conditions not valid - vehicle not level",
            7: " conditions not valid - engine speed too low",
            8: " conditions not valid - engine speed too high",
            9: " conditions not valid - no request for reading"
        },
        592: {
            0: "inactive",
            1: "active",
            2: "error",
            3: "not available"
        },
        593: {
            0: "no",
            1: "yes",
            2: "error",
            3: "not available"
        },
        594: {
            0: "inactive",
            1: "active",
            2: "error",
            3: "not available"
        },
        595: {
            0: "cruise control switched off",
            1: "cruise control switched on",
            2: "error",
            3: "not available"
        },
        596: {
            0: "cruise control disabled",
            1: "cruise control enabled",
            2: "error",
            3: "not available"
        },
        597: {
            0: "brake pedal released",
            1: "brake pedal depressed",
            2: "error",
            3: "not available"
        },
        598: {
            0: "clutch pedal released",
            1: "clutch pedal depressed",
            2: "error",
            3: "not available"
        },
        599: {
            0: "cruise control activator not in the position -set-",
            1: "cruise control activator in position -set-",
            2: "error",
            3: "not available"
        },
        600: {
            0: "cruise control activator not in the position -coast-",
            1: "cruise control activator in position -coast-",
            2: "error",
            3: "not available"
        },
        601: {
            0: "cruise control activator not in the position -resume-",
            1: "cruise control activator in position -resume-",
            2: "error",
            3: "not available"
        },
        602: {
            0: "cruise control activator not in the position -accelerate-",
            1: "cruise control activator in position -accelerate-",
            2: "error",
            3: "not available"
        },
        604: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available"
        },
        605: {
            0: "pressure normal",
            1: "pressure too high, compressor clutch may be disengaged",
            2: "error",
            3: "not available"
        },
        606: {
            0: " momentary engine overspeed is disabled",
            1: " momentary engine overspeed is enabled",
            2: " reserved",
            3: " take no action"
        },
        607: {
            0: " progressive shift is not disabled",
            1: " progressive shift is disabled",
            2: " reserved",
            3: " take no action"
        },
        608: {
            0: " stop broadcast",
            1: " start broadcast",
            2: " reserved",
            3: " don-t care/take no action (leave as is)"
        },
        619: {
            0: "parking brake actuator inactive",
            1: "parking brake actuator active",
            2: "error",
            3: "not available"
        },
        622: {
            0: " stop broadcast",
            1: " start broadcast",
            2: " reserved",
            3: " don-t care/take no action (leave as is)"
        },
        623: {
            0: "     lamp off",
            1: "     lamp on",
            2: "     reserved",
            3: "     not available"
        },
        624: {
            0: "     lamp off",
            1: "     lamp on",
            2: "     reserved",
            3: "     not available"
        },
        626: {
            0: "start enable off",
            1: "start enable on",
            2: "reserved",
            3: "not available"
        },
        632: {
            0: "open (fuel supplied to engine)",
            1: "closed (no fuel supplied to engine)",
            2: "reserved",
            3: "don't care / take no action"
        },
        639: {
            0: " stop broadcast",
            1: " start broadcast",
            2: " reserved",
            3: " don-t care/take no action (leave as is)"
        },
        681: {
            0: " gear shifts are allowed  (disable function)",
            1: " gear shifts are inhibited  (enable function)",
            2: " reserved",
            3: " take no action  (leave function as is)"
        },
        682: {
            0: " allow torque converter lockup",
            1: " disable torque converter lockup",
            2: " reserved",
            3: " take no action"
        },
        683: {
            0: " allow driveline engagement",
            1: " disengage driveline",
            2: " reserved",
            3: " take no action"
        },
        685: {
            0: " engage differential lock",
            1: " disengage differential lock",
            2: " reserved",
            3: " take no action"
        },
        686: {
            0: " engage differential lock",
            1: " disengage differential lock",
            2: " reserved",
            3: " take no action"
        },
        687: {
            0: " engage differential lock",
            1: " disengage differential lock",
            2: " reserved",
            3: " take no action"
        },
        688: {
            0: " engage differential lock",
            1: " disengage differential lock",
            2: " reserved",
            3: " take no action"
        },
        689: {
            0: " engage differential lock",
            1: " disengage differential lock",
            2: " reserved",
            3: " take no action"
        },
        69: {
            0: "low speed range",
            1: "high speed range",
            2: "error",
            3: "not available"
        },
        690: {
            0: " engage differential lock",
            1: " disengage differential lock",
            2: " reserved",
            3: " take no action"
        },
        691: {
            0: " engage differential lock",
            1: " disengage differential lock",
            2: " reserved",
            3: " take no action"
        },
        692: {
            0: " switch off abs offroad function",
            1: " switch on abs offroad function",
            2: " reserved",
            3: " take no action"
        },
        693: {
            0: " switch off asr offroad function",
            1: " switch on asr offroad function",
            2: " reserved",
            3: " take no action"
        },
        694: {
            0: " switch off asr special function",
            1: " switch on asr special function",
            2: " reserved",
            3: " take no action"
        },
        695: {
            0: " override disabled - disable any existing control commanded by the source of this command.",
            1: " speed control - govern speed to the included -desired speed- value.",
            2: " torque control - control torque to the included -desired torque- value.",
            3: " speed/torque limit control - limit speed and/or torque based on the included limit values.  the speed limit governor is a droop governor where the speed limit value defines the speed at the maximum torque available during this operation."
        },
        696: {
            0: "b - this speed governor gain selection is adjusted to provide rapid transition between speed setpoints.  rpm overshoot and undershoot may be greater than what is seen when the -speed control characteristic- is set to be stability optimized.",
            1: "b - this control condition has been optimized to minimize rpm overshoot and undershoot given an expected plant consisting of the engine and its accessory loads.  this gain adjustment is not intended to compensate for driveline characteristics.  this characteristic is most appropriate when no driveline is connected.",
            2: "b - this control condition has been optimized to minimize rpm overshoot and undershoot given a more complex plant.  for instance, the more complex plant would contain the engine, its accessory loads and the driveline characteristics.  as an example, the driveline characteristics might include the effective spring mass relationship of pumps, tires, clutches, axles, driveshafts, and multiple gear ratios.  this characteristic is most appropriate when a driveline is engaged.",
            3: "b - this speed control characteristic is available for applications requiring compensation for more than one driveline characteristic.  it has been optimized to minimize rpm overshoot and undershoot given a more complex plant of the second variety.  this more complex plant would again contain the engine, its accessory loads and a second driveline characteristic unique from the one described in speed control characteristic 10b."
        },
        70: {
            0: "parking brake not set",
            1: "parking brake set",
            2: "error",
            3: "not available"
        },
        701: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error",
            3: "not available"
        },
        702: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error",
            3: "not available"
        },
        703: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error",
            3: "not available"
        },
        704: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error",
            3: "not available"
        },
        705: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error",
            3: "not available"
        },
        706: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error",
            3: "not available"
        },
        707: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error",
            3: "not available"
        },
        708: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error",
            3: "not available"
        },
        709: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error",
            3: "not available"
        },
        710: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error",
            3: "not available"
        },
        711: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error",
            3: "not available"
        },
        712: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error",
            3: "not available"
        },
        713: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error",
            3: "not available"
        },
        714: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error",
            3: "not available"
        },
        715: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error",
            3: "not available"
        },
        716: {
            0: "auxiliary channel off",
            1: "auxiliary channel on",
            2: "error",
            3: "not available"
        },
        740: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available"
        },
        748: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available"
        },
        767: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available"
        },
        768: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available"
        },
        769: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available"
        },
        770: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available"
        },
        771: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available"
        },
        772: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available"
        },
        773: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available"
        },
        778: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available"
        },
        779: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available"
        },
        780: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available"
        },
        781: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available"
        },
        782: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available"
        },
        783: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available"
        },
        784: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available"
        },
        786: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available"
        },
        787: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available"
        },
        788: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available"
        },
        875: {
            0: "pressure normal",
            1: "pressure too low, compressor clutch may be disengaged",
            2: "error",
            3: "not available"
        },
        897: {
            0: "b - highest priority = used for situations that require immediate action by the receiving device in order to provide safe vehicle operation (i.e., braking systems).  this level of priority should only be used in safety critical conditions.",
            1: "b -  high priority  =  used for control situations that require prompt action in order to provide safe vehicle operation.  an example is when the transmission is performing a shift and requires control of the engine in order to control driveline reengagement.",
            2: "b - medium priority = used for powertrain control operations which are related to assuring that the vehicle is in a stable operating condition.  an example is when the traction control system is commanding the engine in order to achieve traction stability.",
            3: "b - low priority = used to indicate that the associated command desires powertrain control but is needed for function which improves the driver comfort which may be overridden by other devices.  an example is cruise control or the non-critical part of a transmission shift to a new gear."
        },
        899: {
            0: " low idle governor/no request (default mode)",
            1: " accelerator pedal/operator selection",
            10: " braking system",
            11: " remote accelerator",
            12: " service procedure",
            13: " not defined",
            14: " other",
            15: " not available",
            2: " cruise control",
            3: " pto governor",
            4: " road speed governor",
            5: " asr control",
            6: " transmission control",
            7: " abs control",
            8: " torque limiting",
            9: " high speed governor"
        },
        901: {
            0: "      electric/magnetic",
            1: "      hydraulic",
            10: "   not defined",
            11: "   not defined",
            12: "   not defined",
            13: "   not defined",
            14: "      other",
            15: "      not available",
            2: "      cooled friction",
            3: "      compression release (engine retarder)",
            4: "      exhaust",
            5: "   not defined",
            6: "   not defined",
            7: "   not defined",
            8: "   not defined",
            9: "   not defined"
        },
        902: {
            0: "   (primary)         engine compression release brake (engine rpm)",
            1: "   (primary)         engine exhaust brake (exhaust pressure)",
            10: "                  not defined",
            11: "                  not defined",
            12: "                  not defined",
            13: "                  not defined",
            14: "                           other",
            15: "                           not available",
            2: "   (primary)         transmission input (engine rpm)",
            3: "   (secondary)    transmission output (output shaft rpm)",
            4: "   (secondary)    driveline (output shaft rpm)",
            5: "                           trailer (vehicle speed)",
            6: "                  not defined",
            7: "                  not defined",
            8: "                  not defined",
            9: "                  not defined"
        },
        903: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available"
        },
        966: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available"
        },
        967: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available"
        },
        968: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available"
        },
        969: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available"
        },
        97: {
            0: "no",
            1: "yes",
            2: "error",
            3: "not available"
        },
        970: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available"
        },
        971: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available"
        },
        972: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available"
        },
        976: {
            0: "     off/disabled",
            1: "     hold",
            10: "     preprogrammed set speed 1",
            11: "     preprogrammed set speed 2",
            12: "     preprogrammed set speed 3",
            13: "     preprogrammed set speed 4",
            14: "     preprogrammed set speed 5",
            15: "     preprogrammed set speed 6",
            16: "     preprogrammed set speed 7",
            17: "     preprogrammed set speed 8",
            18: "     pto set speed memory 1",
            19: "     pto set speed memory 2",
            2: "     remote hold",
            20: "      not defined",
            21: "      not defined",
            22: "      not defined",
            23: "      not defined",
            24: "      not defined",
            25: "      not defined",
            26: "      not defined",
            27: "      not defined",
            28: "      not defined",
            29: "      not defined",
            3: "     standby",
            30: "      not defined",
            31: "     not available",
            4: "     remote standby",
            5: "     set",
            6: "     decelerate/coast",
            7: "     resume",
            8: "     accelerate",
            9: "     accelerator override"
        },
        977: {
            0: " fan off",
            1: " engine system-general",
            10: " transmission retarder",
            11: " a/c system",
            12: " timer",
            13: " engine brake",
            14: " other",
            15: " not available",
            2: " excessive engine air temperature",
            3: " excessive engine oil temperature",
            4: " excessive engine coolant temperature",
            5: " excessive transmission oil temperature",
            6: " excessive hydraulic oil temperature",
            7: " default operation",
            8: " reverse operation",
            9: " manual control"
        },
        978: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available"
        },
        979: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available"
        },
        980: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available"
        },
        981: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available"
        },
        982: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available"
        },
        983: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available"
        },
        984: {
            0: "off",
            1: "on",
            2: "error",
            3: "not available"
        },
        985: {
            0: "pressure normal",
            1: "pressure high, fan may be engaged",
            2: "error",
            3: "not available"
        },
        987: {
            0: "     lamp off",
            1: "     lamp on",
            2: "     reserved",
            3: "     not available"
        },
        988: {
            0: " take no action",
            1: " reset",
            2: " reserved",
            3: " not applicable"
        },
        989: {
            0: " take no action",
            1: " reset",
            2: " reserved",
            3: " not applicable"
        }
    }

J1939FMITabledbr = {
        0: {
            "Name": "Data Valid But Above Normal Operational Range",
            "Severity": "Most Severe Level"
        },
        1: {
            "Name": "Data Valid But Below Normal Operational Range",
            "Severity": "Most Severe Level"
        },
        10: {
            "Name": "Abnormal Rate Of Change",
            "Severity": "N/A"
        },
        11: {
            "Name": "Root Cause Not Known",
            "Severity": "N/A"
        },
        12: {
            "Name": "Bad Intelligent Device Or Component",
            "Severity": "N/A"
        },
        13: {
            "Name": "Out Of Calibration",
            "Severity": "N/A"
        },
        14: {
            "Name": "Special Instructions",
            "Severity": "N/A"
        },
        15: {
            "Name": "Data Valid But Above Normal Operating Range",
            "Severity": "Least Severe Level"
        },
        16: {
            "Name": "Data Valid But Above Normal Operating Range",
            "Severity": "Moderately Severe Level"
        },
        17: {
            "Name": "Data Valid But Below Normal Operating Range",
            "Severity": "Least Severe Level"
        },
        18: {
            "Name": "Data Valid But Below Normal Operating Range",
            "Severity": "Moderately Severe Level"
        },
        19: {
            "Name": "Received Network Data In Error",
            "Severity": "N/A"
        },
        2: {
            "Name": "Data Erratic, Intermittent Or Incorrect",
            "Severity": "N/A"
        },
        20: {
            "Name": "Data Drifted High",
            "Severity": "N/A"
        },
        21: {
            "Name": "Data Drifted Low",
            "Severity": "N/A"
        },
        22: {
            "Name": "Reserved For SAE Assignment",
            "Severity": "N/A"
        },
        23: {
            "Name": "Reserved For SAE Assignment",
            "Severity": "N/A"
        },
        24: {
            "Name": "Reserved For SAE Assignment",
            "Severity": "N/A"
        },
        25: {
            "Name": "Reserved For SAE Assignment",
            "Severity": "N/A"
        },
        26: {
            "Name": "Reserved For SAE Assignment",
            "Severity": "N/A"
        },
        27: {
            "Name": "Reserved For SAE Assignment",
            "Severity": "N/A"
        },
        28: {
            "Name": "Reserved For SAE Assignment",
            "Severity": "N/A"
        },
        29: {
            "Name": "Reserved For SAE Assignment",
            "Severity": "N/A"
        },
        3: {
            "Name": "Voltage Above Normal, Or Shorted To High Source",
            "Severity": "N/A"
        },
        30: {
            "Name": "Reserved For SAE Assignment",
            "Severity": "N/A"
        },
        31: {
            "Name": "Condition Exists",
            "Severity": "N/A"
        },
        4: {
            "Name": "Voltage Below Normal, Or Shorted To Low Source",
            "Severity": "N/A"
        },
        5: {
            "Name": "Current Below Normal Or Open Circuit",
            "Severity": "N/A"
        },
        6: {
            "Name": "Current Above Normal Or Grounded Circuit",
            "Severity": "N/A"
        },
        7: {
            "Name": "Mechanical System Not Responding Or Out Of Adjustment",
            "Severity": "N/A"
        },
        8: {
            "Name": "Abnormal Frequency Or Pulse Width Or Period",
            "Severity": "N/A"
        },
        9: {
            "Name": "Abnormal Update Rate",
            "Severity": "N/A"
        }
    }

J1939LampFlashTabledb = {
        0: {
            "AWL": "Off",
            "Flash": 0,
            "Lamp": 0,
            "MIL": "Off",
            "Protected": "Off",
            "RSL": "Off"
        },
        1: {
            "AWL": "Off",
            "Flash": 1,
            "Lamp": 0,
            "MIL": "Off",
            "Protected": "Off",
            "RSL": "Off"
        },
        10: {
            "AWL": "SAE reserved",
            "Flash": 2,
            "Lamp": 2,
            "MIL": "SAE reserved",
            "Protected": "SAE reserved",
            "RSL": "SAE reserved"
        },
        11: {
            "AWL": "SAE reserved",
            "Flash": 3,
            "Lamp": 2,
            "MIL": "SAE reserved",
            "Protected": "SAE reserved",
            "RSL": "SAE reserved"
        },
        12: {
            "AWL": "SAE reserved",
            "Flash": 0,
            "Lamp": 3,
            "MIL": "SAE reserved",
            "Protected": "SAE reserved",
            "RSL": "SAE reserved"
        },
        13: {
            "AWL": "SAE reserved",
            "Flash": 1,
            "Lamp": 3,
            "MIL": "SAE reserved",
            "Protected": "SAE reserved",
            "RSL": "SAE reserved"
        },
        14: {
            "AWL": "SAE reserved",
            "Flash": 2,
            "Lamp": 3,
            "MIL": "SAE reserved",
            "Protected": "SAE reserved",
            "RSL": "SAE reserved"
        },
        15: {
            "AWL": "Don't Care",
            "Flash": 3,
            "Lamp": 3,
            "MIL": "Don't Care",
            "Protected": "Don't Care",
            "RSL": "Don't Care"
        },
        2: {
            "AWL": "SAE reserved",
            "Flash": 2,
            "Lamp": 0,
            "MIL": "Off, class C not active",
            "Protected": "SAE reserved",
            "RSL": "SAE reserved"
        },
        3: {
            "AWL": "Off",
            "Flash": 3,
            "Lamp": 0,
            "MIL": "Off",
            "Protected": "Off",
            "RSL": "Off"
        },
        4: {
            "AWL": "On, slow flash",
            "Flash": 0,
            "Lamp": 1,
            "MIL": "On, slow flash",
            "Protected": "On, slow flash",
            "RSL": "On, slow flash"
        },
        5: {
            "AWL": "On, fast flash",
            "Flash": 1,
            "Lamp": 1,
            "MIL": "On, fast flash",
            "Protected": "On, fast flash",
            "RSL": "On, fast flash"
        },
        6: {
            "AWL": "SAE reserved",
            "Flash": 2,
            "Lamp": 1,
            "MIL": "Off, class C not active",
            "Protected": "SAE reserved",
            "RSL": "SAE reserved"
        },
        7: {
            "AWL": "On, don't flash",
            "Flash": 3,
            "Lamp": 1,
            "MIL": "On, don't flash",
            "Protected": "On, don't flash",
            "RSL": "On, don't flash"
        },
        8: {
            "AWL": "SAE reserved",
            "Flash": 0,
            "Lamp": 2,
            "MIL": "Short MI not active",
            "Protected": "SAE reserved",
            "RSL": "SAE reserved"
        },
        9: {
            "AWL": "SAE reserved",
            "Flash": 1,
            "Lamp": 2,
            "MIL": "Short MI active",
            "Protected": "SAE reserved",
            "RSL": "SAE reserved"
        }
    }

J1939OBDTabledb = {
        0: {
            "Acronym": "Reserved",
            "Description": "Reserved for assignment by SAE"
        },
        1: {
            "Acronym": "OBD II",
            "Description": "OBD II (California Air Resources Board)"
        },
        10: {
            "Acronym": "JOBD",
            "Description": "JOBD"
        },
        11: {
            "Acronym": "JOBD and OBD II",
            "Description": "JOBD and OBD II"
        },
        12: {
            "Acronym": "JOBD and EOBD",
            "Description": "JOBD and EOBD"
        },
        13: {
            "Acronym": "JOBD, EOBD and OBD II",
            "Description": "JOBD, EOBD and OBD II"
        },
        14: {
            "Acronym": "B1",
            "Description": "Heavy Duty Vehicles (EURO IV) B1"
        },
        15: {
            "Acronym": "B2",
            "Description": "Heavy Duty Vehicles (EURO V) B2"
        },
        16: {
            "Acronym": "C",
            "Description": "Heavy Duty Vehicles (EURO EEC) C (gas engines)"
        },
        17: {
            "Acronym": "EMD",
            "Description": "Engine Manufacturer Diagnostics (EMD)"
        },
        18: {
            "Acronym": "EMD+",
            "Description": "Engine Manufacturer Diagnostics Enhanced (EMD+),"
        },
        19: {
            "Acronym": "HD OBD p",
            "Description": "Heavy Duty/On-Board Diagnostics Partial"
        },
        2: {
            "Acronym": "OBD",
            "Description": "OBD (Federal, EPA)"
        },
        20: {
            "Acronym": "HD OBD",
            "Description": "Heavy Duty/On-Board Diagnostics"
        },
        21: {
            "Acronym": "WWH OBD",
            "Description": "World Wide Harmonized OBD"
        },
        22: {
            "Acronym": "OBD II",
            "Description": "OBD II (Califronia Air Resources Board, CCR 1968.2,2007/2008/2009/2010 revisions)"
        },
        23: {
            "Acronym": "OD EOBD",
            "Description": "Heavy Duty Vehicle"
        },
        24: {
            "Acronym": "Reserved",
            "Description": "SAE reserved to be assigned"
        },
        25: {
            "Acronym": "OBD-M (SI-SD/I)",
            "Description": "OBD-M Compliance for Spark-Ignition Sterndrive and Inboard Engines"
        },
        3: {
            "Acronym": "OBD and OBDII",
            "Description": "OBD and OBD II"
        },
        4: {
            "Acronym": "OBD I",
            "Description": "OBD I"
        },
        5: {
            "Acronym": "No OBD II",
            "Description": "Not intended to meet OBD II requirements"
        },
        6: {
            "Acronym": "EOBD",
            "Description": "EOBD"
        },
        7: {
            "Acronym": "EOBD and OBD II",
            "Description": "EOBD and OBD II"
        },
        8: {
            "Acronym": "EOBD and OBD",
            "Description": "EOBD and OBD"
        },
        9: {
            "Acronym": "EOBD, OBD and OBD II",
            "Description": "EOBD, OBD and OBD II"
        }
    }

J1939PGNdb = {
        0: {
            "Label": "TSC1",
            "Name": "Torque/Speed Control 1",
            "PGNLength": 8,
            "Rate": "To engine: Control Purpose dependent or 10 ms To retarder: 50 ms",
            "SPNs": [
                695,
                696,
                897,
                518,
                3349,
                3350,
                4191,
                4206,
                4207,
                898
            ]
        },
        1024: {
            "Label": "XBR",
            "Name": "External Brake Request",
            "PGNLength": 8,
            "Rate": "When active: 20 ms; else 200 ms",
            "SPNs": [
                2914,
                2915,
                2916,
                4099,
                3189,
                3188,
                2920
            ]
        },
        126208: {
            "Label": "",
            "Name": "NMEA - Request/Command/Acknowledge group function",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        126464: {
            "Label": "",
            "Name": "PGN List - Transmit/Receive PGN's group function",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        126720: {
            "Label": "PropA2",
            "Name": "Proprietary A2",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                3328
            ]
        },
        126992: {
            "Label": "",
            "Name": "System Time",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        126996: {
            "Label": "",
            "Name": "Product Information",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        126998: {
            "Label": "",
            "Name": "Configuration Information",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        127237: {
            "Label": "",
            "Name": "Heading/Track Control",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        127245: {
            "Label": "",
            "Name": "Rudder",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        127250: {
            "Label": "",
            "Name": "Vessel Heading",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        127251: {
            "Label": "",
            "Name": "Rate of Turn",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        127257: {
            "Label": "",
            "Name": "Attitude",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        127488: {
            "Label": "",
            "Name": "Engine Parameters: Rapid Update",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        127489: {
            "Label": "",
            "Name": "Engine Parameters: Dynamic",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        127493: {
            "Label": "",
            "Name": "Transmission Parameters: Dynamic",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        127497: {
            "Label": "",
            "Name": "Trip Parameters: Small Craft",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        127498: {
            "Label": "",
            "Name": "Engine Parameters: Static",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        127501: {
            "Label": "",
            "Name": "Binary Switch Bank Status",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        127502: {
            "Label": "",
            "Name": "Switch Bank Control",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        127505: {
            "Label": "",
            "Name": "Fluid Level",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        127508: {
            "Label": "",
            "Name": "Battery Status",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        1280: {
            "Label": "CAM11",
            "Name": "Reserved for CANopen",
            "PGNLength": 8,
            "Rate": "50 ms",
            "SPNs": []
        },
        128259: {
            "Label": "",
            "Name": "Speed",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        128267: {
            "Label": "",
            "Name": "Water Depth",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        128275: {
            "Label": "",
            "Name": "Distance Log",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        128520: {
            "Label": "",
            "Name": "Tracked Target Data",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        129025: {
            "Label": "",
            "Name": "Position: Rapid Update",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        129026: {
            "Label": "",
            "Name": "COG & SOG: Rapid Update",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        129029: {
            "Label": "",
            "Name": "GNSS Position Data",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        129033: {
            "Label": "",
            "Name": "Time & Date",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        129044: {
            "Label": "",
            "Name": "Datum",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        129045: {
            "Label": "",
            "Name": "User Datum Settings",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        129283: {
            "Label": "",
            "Name": "Cross Track Error",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        129284: {
            "Label": "",
            "Name": "Navigation Data",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        129291: {
            "Label": "",
            "Name": "Set & Drift: Rapid Update",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        129301: {
            "Label": "",
            "Name": "Time to/from Mark",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        129302: {
            "Label": "",
            "Name": "Bearing and Distance between two Marks",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        129538: {
            "Label": "",
            "Name": "GNSS Control Status",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        129539: {
            "Label": "",
            "Name": "GNSS DOPs",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        129540: {
            "Label": "",
            "Name": "GNSS Sats in View",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        129541: {
            "Label": "",
            "Name": "GPS Almanac Data",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        129542: {
            "Label": "",
            "Name": "GNSS Pseudorange Noise Statistics",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        129545: {
            "Label": "",
            "Name": "GNSS RAIM Output",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        129546: {
            "Label": "",
            "Name": "GNSS RAIM Settings",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        129547: {
            "Label": "",
            "Name": "GNSS Pseudorange Error Statistics",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        129549: {
            "Label": "",
            "Name": "DGNSS Corrections",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        129550: {
            "Label": "",
            "Name": "GNSS Differential Correction Receiver Interface",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        129551: {
            "Label": "",
            "Name": "GNSS Differential Correction Receiver Signal",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        129556: {
            "Label": "",
            "Name": "GLONASS Almanac Data",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        129799: {
            "Label": "",
            "Name": "Radio Frequency/Mode/Power",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        129808: {
            "Label": "",
            "Name": "DSC Call Information",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        130052: {
            "Label": "",
            "Name": "Loran-C TD Data",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        130053: {
            "Label": "",
            "Name": "Loran-C Range Data",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        130054: {
            "Label": "",
            "Name": "Loran-C Signal Data",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        130306: {
            "Label": "",
            "Name": "Wind Data",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        130310: {
            "Label": "",
            "Name": "Environmental Parameters",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        130320: {
            "Label": "",
            "Name": "Tide Station Data",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        130321: {
            "Label": "",
            "Name": "Salinity Station Data",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        130322: {
            "Label": "",
            "Name": "Current Station Data",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        130323: {
            "Label": "",
            "Name": "Meteorological Station Data",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        130324: {
            "Label": "",
            "Name": "Moored Buoy Station Data",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        130576: {
            "Label": "",
            "Name": "Small Craft Status",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        130577: {
            "Label": "",
            "Name": "Direction Data",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        130578: {
            "Label": "",
            "Name": "Vessel Speed Components",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        1536: {
            "Label": "CAM21",
            "Name": "Reserved for CANopen",
            "PGNLength": 8,
            "Rate": "50 ms",
            "SPNs": []
        },
        1792: {
            "Label": "GPV4",
            "Name": "General Purpose Valve Pressure",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                4087,
                4086,
                4088,
                4089
            ]
        },
        2048: {
            "Label": "AUXIO5",
            "Name": "Auxiliary Input/Output Status 5:8",
            "PGNLength": "Manufacturer defined",
            "Rate": " not faster than 20 ms",
            "SPNs": [
                4155,
                4156,
                4157,
                4158
            ]
        },
        2304: {
            "Label": "SRASI",
            "Name": "Static Roll Angle Sensor Information",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                5582,
                5583
            ]
        },
        256: {
            "Label": "TC1",
            "Name": "Transmission Control 1",
            "PGNLength": 8,
            "Rate": "When active; 50 ms to transmission and axles",
            "SPNs": [
                681,
                682,
                683,
                4242,
                684,
                525,
                685,
                686,
                687,
                688,
                689,
                690,
                691,
                5762,
                1852,
                1853,
                1854,
                1855,
                4255,
                2985,
                4246,
                4247,
                4248,
                4249
            ]
        },
        2560: {
            "Label": "CCVS2",
            "Name": "Cruise Control / Vehicle Speed 2",
            "PGNLength": 8,
            "Rate": "Every 100 ms and on change but no faster than 20 ms",
            "SPNs": [
                5603,
                5604,
                5605
            ]
        },
        2816: {
            "Label": "AEBS2",
            "Name": "Advanced Emergency Braking System 2",
            "PGNLength": 8,
            "Rate": "Every 50ms and if new SPN Driver activation demand for Advanced Emergency Braking System has changed but no faster than every 10 ms",
            "SPNs": [
                5681,
                5682,
                5683
            ]
        },
        34560: {
            "Label": "ESR",
            "Name": "Engine State Requests",
            "PGNLength": 8,
            "Rate": "Every 1 sec and on change but no faster than 100 ms",
            "SPNs": [
                5793
            ]
        },
        34816: {
            "Label": "RBR",
            "Name": "Radio Beacon Request",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                5755,
                5757,
                5754,
                5756
            ]
        },
        35072: {
            "Label": "RGE24",
            "Name": "Reserved for ISO 11992",
            "PGNLength": 8,
            "Rate": "1 sec",
            "SPNs": []
        },
        35328: {
            "Label": "RGE12",
            "Name": "Reserved for ISO 11992",
            "PGNLength": 8,
            "Rate": "1 sec",
            "SPNs": []
        },
        35584: {
            "Label": "CCP DTO",
            "Name": "CCP Data Transmission Object",
            "PGNLength": 8,
            "Rate": "As defined in protocol",
            "SPNs": []
        },
        35840: {
            "Label": "CCP CRO",
            "Name": "CCP Command Receive Object",
            "PGNLength": 8,
            "Rate": "As defined in protocol",
            "SPNs": []
        },
        36096: {
            "Label": "SCCCM",
            "Name": "Sequence Control Client-to-Sequence Control Master",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        36352: {
            "Label": "SCMCC",
            "Name": "Sequence Control Master-to-Sequence Control Client",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        36608: {
            "Label": "ECC",
            "Name": "Engine Configuration Commands",
            "PGNLength": 8,
            "Rate": "Every 10 seconds and on change but no faster than 1 second.",
            "SPNs": [
                5565
            ]
        },
        36864: {
            "Label": "SFC",
            "Name": "Supplemental Fan Command",
            "PGNLength": 8,
            "Rate": "1 second",
            "SPNs": [
                5552,
                5553,
                5555,
                5554
            ]
        },
        37120: {
            "Label": "VLS1",
            "Name": "Vehicle/Chassis Lubrication System 1",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                5510,
                5511,
                5512,
                5513,
                5514
            ]
        },
        37376: {
            "Label": "HCDI1",
            "Name": "Aftertreatment 1 Hydrocarbon Doser Information 1",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                5506,
                5507,
                5508,
                5509,
                5505
            ]
        },
        37632: {
            "Label": "NM",
            "Name": "NAME Management Message",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                5657,
                5658,
                5659,
                5660,
                5661,
                5662,
                5663,
                5664,
                5665,
                5666,
                5668,
                5669,
                5670,
                5671,
                5672,
                5673,
                5674,
                5667
            ]
        },
        37888: {
            "Label": "AT1DPFSSC",
            "Name": "Aftertreatment 1 DPF Soot Sensor Calibration",
            "PGNLength": 8,
            "Rate": "On powerup and on request",
            "SPNs": [
                5495,
                5496
            ]
        },
        38144: {
            "Label": "AT2DPFSSC",
            "Name": "Aftertreatment 2 DPF Soot Sensor Calibration",
            "PGNLength": 8,
            "Rate": "On powerup and on request",
            "SPNs": [
                5497,
                5498
            ]
        },
        38400: {
            "Label": "LVDSOM",
            "Name": "Low Voltage Disconnect Set Operating Mode",
            "PGNLength": 8,
            "Rate": "As required but no more often than 500 ms",
            "SPNs": [
                5149
            ]
        },
        38656: {
            "Label": "NCS",
            "Name": "Noise Control Status:8",
            "PGNLength": "Every 10 seconds and on change of state",
            "Rate": " but not faster than 1 second.  Every second when in tuning mode.",
            "SPNs": [
                5059,
                5060
            ]
        },
        38912: {
            "Label": "NC1",
            "Name": "Noise Control 1:8",
            "PGNLength": "Every 10 seconds and on change of state",
            "Rate": " but not faster than every 1 second.",
            "SPNs": [
                5057,
                5058
            ]
        },
        39168: {
            "Label": "JLCM",
            "Name": "Joystick Lamp Command Message",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                4460,
                4461,
                4462,
                4463,
                4464,
                4465,
                4466,
                4467,
                4468,
                4469,
                4470,
                4471,
                4472,
                4473,
                4474,
                4475,
                4476,
                4477,
                4478,
                4479,
                4480,
                4481,
                4482,
                4483,
                4484,
                4485,
                4486,
                4487,
                4488,
                4489
            ]
        },
        39424: {
            "Label": "GPM18",
            "Name": "Reserved for ISO 11992",
            "PGNLength": 8,
            "Rate": "5 sec",
            "SPNs": []
        },
        39680: {
            "Label": "PMI",
            "Name": "Proprietary Method Identification",
            "PGNLength": "Variable",
            "Rate": "On request",
            "SPNs": [
                4180,
                4181
            ]
        },
        39936: {
            "Label": "AUXIO7",
            "Name": "Auxiliary Input/Output Status 7:8",
            "PGNLength": "manufacturer defined",
            "Rate": " not faster than 100 ms",
            "SPNs": [
                4167,
                4168,
                4169,
                4170,
                4171,
                4172,
                4173,
                4174
            ]
        },
        40192: {
            "Label": "AUXIO6",
            "Name": "Auxiliary Input/Output Status 6:8",
            "PGNLength": "manufacturer defined",
            "Rate": " not faster than 100 ms",
            "SPNs": [
                4159,
                4160,
                4161,
                4162,
                4163,
                4164,
                4165,
                4166
            ]
        },
        40448: {
            "Label": "DM29",
            "Name": "DTC Counts",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        40704: {
            "Label": "DM35",
            "Name": "Immediate Fault Status",
            "PGNLength": "Variable",
            "Rate": "On request",
            "SPNs": [
                987,
                624,
                623,
                1213,
                3041,
                3040,
                3039,
                3038,
                1215,
                1216,
                1706,
                1214
            ]
        },
        40960: {
            "Label": "DM34",
            "Name": "NTE Status",
            "PGNLength": "Variable",
            "Rate": "On request",
            "SPNs": []
        },
        41216: {
            "Label": "DM33",
            "Name": "Emission Increasing Auxiliary Emission Control Device Active Time",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        41472: {
            "Label": "DM32",
            "Name": "Regulated Exhaust Emission Level Exceedance",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        41728: {
            "Label": "DM31",
            "Name": "DTC To Lamp Association",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        41984: {
            "Label": "DM30",
            "Name": "Scaled Test Results",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        42240: {
            "Label": "AUXIO4",
            "Name": "Auxiliary Input/Output Status 4:8",
            "PGNLength": "Manufacturer defined",
            "Rate": " not faster than 100 ms",
            "SPNs": [
                3907,
                3906,
                3905,
                3904,
                3911,
                3910,
                3909,
                3908,
                3915,
                3914,
                3913,
                3912,
                3919,
                3918,
                3917,
                3916,
                3923,
                3922,
                3921,
                3920,
                3927,
                3926,
                3925,
                3924,
                3931,
                3930,
                3929,
                3928,
                3935,
                3934,
                3933,
                3932
            ]
        },
        42496: {
            "Label": "AUXIO3",
            "Name": "Auxiliary Input/Output Status 3:8",
            "PGNLength": "Manufacturer defined",
            "Rate": " not faster than 100 ms",
            "SPNs": [
                3875,
                3874,
                3873,
                3872,
                3879,
                3878,
                3877,
                3876,
                3883,
                3882,
                3881,
                3880,
                3887,
                3886,
                3885,
                3884,
                3891,
                3890,
                3889,
                3888,
                3895,
                3894,
                3893,
                3892,
                3899,
                3898,
                3897,
                3896,
                3903,
                3902,
                3901,
                3900
            ]
        },
        42752: {
            "Label": "AUXIO2",
            "Name": "Auxiliary Input/Output Status 2:8",
            "PGNLength": "Manufacturer defined",
            "Rate": " not faster than 100 ms",
            "SPNs": [
                3843,
                3842,
                3841,
                3840,
                3847,
                3846,
                3845,
                3844,
                3851,
                3850,
                3849,
                3848,
                3855,
                3854,
                3853,
                3852,
                3859,
                3858,
                3857,
                3856,
                3863,
                3862,
                3861,
                3860,
                3867,
                3866,
                3865,
                3864,
                3871,
                3870,
                3869,
                3868
            ]
        },
        43008: {
            "Label": "DISP1",
            "Name": "Text Display",
            "PGNLength": "Variable - up to 200 characters",
            "Rate": "As required",
            "SPNs": [
                3613,
                3614,
                3615
            ]
        },
        43264: {
            "Label": "FLIC",
            "Name": "Forward Lane Image Command",
            "PGNLength": 8,
            "Rate": "On event",
            "SPNs": [
                3564
            ]
        },
        43520: {
            "Label": "CFS",
            "Name": "Client (ECU) to File Server message",
            "PGNLength": "",
            "Rate": "On request",
            "SPNs": []
        },
        43776: {
            "Label": "FSC",
            "Name": "File Server to Client (ECU) message",
            "PGNLength": "",
            "Rate": "On request",
            "SPNs": []
        },
        44032: {
            "Label": "GMS",
            "Name": "Agricultural Guidance Machine Info",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": []
        },
        44288: {
            "Label": "GSC",
            "Name": "Agricultural Guidance System Command",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": []
        },
        44544: {
            "Label": "TPRS",
            "Name": "Tire Pressure Reference Setting",
            "PGNLength": 8,
            "Rate": "As needed",
            "SPNs": [
                3192,
                3193
            ]
        },
        44800: {
            "Label": "PLM",
            "Name": "Parameter Locate Message",
            "PGNLength": 8,
            "Rate": "As required",
            "SPNs": []
        },
        45056: {
            "Label": "CIM",
            "Name": "Configuration Identification Message",
            "PGNLength": 8,
            "Rate": "As required",
            "SPNs": []
        },
        45312: {
            "Label": "PCM1",
            "Name": "Proprietarily Configurable Message #1",
            "PGNLength": "Variable",
            "Rate": "As required",
            "SPNs": []
        },
        45568: {
            "Label": "PCM2",
            "Name": "Proprietarily Configurable Message #2",
            "PGNLength": "Variable",
            "Rate": "As required",
            "SPNs": []
        },
        45824: {
            "Label": "PCM3",
            "Name": "Proprietarily Configurable Message #3",
            "PGNLength": "Variable",
            "Rate": "As required",
            "SPNs": []
        },
        46080: {
            "Label": "PCM4",
            "Name": "Proprietarily Configurable Message #4",
            "PGNLength": "Variable",
            "Rate": "As required",
            "SPNs": []
        },
        46336: {
            "Label": "PCM5",
            "Name": "Proprietarily Configurable Message #5",
            "PGNLength": "Variable",
            "Rate": "As required",
            "SPNs": []
        },
        46592: {
            "Label": "PCM6",
            "Name": "Proprietarily Configurable Message #6",
            "PGNLength": "Variable",
            "Rate": "As required",
            "SPNs": []
        },
        46848: {
            "Label": "PCM7",
            "Name": "Proprietarily Configurable Message #7",
            "PGNLength": "Variable",
            "Rate": "As required",
            "SPNs": []
        },
        47104: {
            "Label": "PCM8",
            "Name": "Proprietarily Configurable Message #8",
            "PGNLength": "Variable",
            "Rate": "As required",
            "SPNs": []
        },
        47360: {
            "Label": "PCM9",
            "Name": "Proprietarily Configurable Message #9",
            "PGNLength": "Variable",
            "Rate": "As required",
            "SPNs": []
        },
        47616: {
            "Label": "PCM10",
            "Name": "Proprietarily Configurable Message #10",
            "PGNLength": "Variable",
            "Rate": "As required",
            "SPNs": []
        },
        47872: {
            "Label": "PCM11",
            "Name": "Proprietarily Configurable Message #11",
            "PGNLength": "Variable",
            "Rate": "As required",
            "SPNs": []
        },
        48128: {
            "Label": "PCM12",
            "Name": "Proprietarily Configurable Message #12",
            "PGNLength": "Variable",
            "Rate": "As required",
            "SPNs": []
        },
        48384: {
            "Label": "PCM13",
            "Name": "Proprietarily Configurable Message #13",
            "PGNLength": "Variable",
            "Rate": "As required",
            "SPNs": []
        },
        48640: {
            "Label": "PCM14",
            "Name": "Proprietarily Configurable Message #14",
            "PGNLength": "Variable",
            "Rate": "As required",
            "SPNs": []
        },
        48896: {
            "Label": "PCM15",
            "Name": "Proprietarily Configurable Message #15",
            "PGNLength": "Variable",
            "Rate": "As required",
            "SPNs": []
        },
        49152: {
            "Label": "PCM16",
            "Name": "Proprietarily Configurable Message #16",
            "PGNLength": "Variable",
            "Rate": "As required",
            "SPNs": []
        },
        49408: {
            "Label": "DM21",
            "Name": "Diagnostic Readiness 2",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                3069,
                3294,
                3295,
                3296
            ]
        },
        49664: {
            "Label": "DM20",
            "Name": "Monitor Performance Ratio",
            "PGNLength": "Variable",
            "Rate": "On request",
            "SPNs": [
                3048,
                3049,
                3067,
                3068,
                3066
            ]
        },
        49920: {
            "Label": "DM22",
            "Name": "Individual Clear/Reset Of Active And Previously Active DTC",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                3034,
                3035,
                3037,
                3036
            ]
        },
        50176: {
            "Label": "GPV3",
            "Name": "General Purpose Valve Command",
            "PGNLength": 8,
            "Rate": "1 s for each valve or on change of state.",
            "SPNs": [
                2944
            ]
        },
        50432: {
            "Label": "GPV2",
            "Name": "General Purpose Valve Measured Flow",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                2937,
                2938,
                2943,
                2941,
                2942
            ]
        },
        50688: {
            "Label": "GPV1",
            "Name": "General Purpose Valve Estimated Flow",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                2939,
                2940,
                2932,
                2934
            ]
        },
        50944: {
            "Label": "ETP.DT",
            "Name": "Extended Transport Protocol - Data Transfer",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        512: {
            "Label": "EBS11",
            "Name": "Reserved for ISO 11992",
            "PGNLength": 8,
            "Rate": "10 ms +/- 1 ms",
            "SPNs": []
        },
        51200: {
            "Label": "ETP.CM",
            "Name": "Extended Transport Protocol - Connection Management",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        51456: {
            "Label": " see PGN 59392",
            "Name": "Request 2:8 bytes (multipacket  supported)",
            "PGNLength": "Per user requirements ",
            "Rate": " generally recommended that requests occur no more than 2 or 3 times per second. A NACK is required if the PG being asked for with a destination specifc address and it is not supported",
            "SPNs": [
                2575,
                2574
            ]
        },
        51712: {
            "Label": "XFER",
            "Name": "Transfer",
            "PGNLength": "9 to 1785 bytes (multipacket supported)",
            "Rate": "In response to a Request2 PGN with \u201cUse Transfer PGN\u201d = 01",
            "SPNs": [
                2553,
                2552,
                2554,
                2555
            ]
        },
        51968: {
            "Label": "PD",
            "Name": "Process Data Message",
            "PGNLength": 8,
            "Rate": "As Events Require",
            "SPNs": [
                2421,
                2420,
                2419,
                2422,
                2426,
                2425,
                2428,
                2427,
                2429
            ]
        },
        52224: {
            "Label": "REQRR",
            "Name": "Request for Repetition Rate",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                2421,
                2420,
                2419,
                2426,
                2425,
                2428,
                2427,
                2418
            ]
        },
        52480: {
            "Label": "KWP4",
            "Name": "Reserved for ISO 15765",
            "PGNLength": 8,
            "Rate": "User requirements",
            "SPNs": []
        },
        52736: {
            "Label": "KWP3",
            "Name": "Reserved for ISO 15765",
            "PGNLength": 8,
            "Rate": "User requirements",
            "SPNs": []
        },
        52992: {
            "Label": "CTL",
            "Name": "Continuous Torque & Speed Limit Request",
            "PGNLength": 8,
            "Rate": "5 s",
            "SPNs": [
                1784,
                1785,
                1786,
                1787,
                1788,
                1789,
                1790,
                1791
            ]
        },
        53248: {
            "Label": "CL",
            "Name": "Cab Illumination Message",
            "PGNLength": 8,
            "Rate": "Every 5 s and on change of state but no faster than every 100 ms",
            "SPNs": [
                1487,
                5532
            ]
        },
        53504: {
            "Label": "ASC6",
            "Name": "Air Suspension Control 6",
            "PGNLength": 8,
            "Rate": "100 ms when active",
            "SPNs": [
                1732,
                1757,
                1758,
                1735
            ]
        },
        53760: {
            "Label": "ASC2",
            "Name": "Air Suspension Control 2",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                2984,
                1749,
                1748,
                1747,
                1751,
                1750,
                1753,
                1752,
                1828,
                1718,
                1719,
                1720,
                1830,
                1829,
                3215,
                5293,
                5295,
                5274,
                5433
            ]
        },
        54016: {
            "Label": "DM19",
            "Name": "Calibration Information",
            "PGNLength": "Variable",
            "Rate": "On request",
            "SPNs": [
                1634,
                1635
            ]
        },
        54272: {
            "Label": "DM18",
            "Name": "Data Security",
            "PGNLength": "Variable",
            "Rate": "As needed",
            "SPNs": [
                1479,
                1596,
                1597
            ]
        },
        54528: {
            "Label": "TDA",
            "Name": "Time/Date Adjust",
            "PGNLength": 8,
            "Rate": "As needed",
            "SPNs": [
                1603,
                1604,
                1605,
                1606,
                1607,
                1608,
                1609,
                1610
            ]
        },
        54784: {
            "Label": "DM17",
            "Name": "Boot Load Data",
            "PGNLength": 8,
            "Rate": "As needed",
            "SPNs": [
                1652
            ]
        },
        55040: {
            "Label": "DM16",
            "Name": "Binary Data Transfer",
            "PGNLength": "Variable",
            "Rate": "As needed",
            "SPNs": [
                1650,
                1651,
                1651
            ]
        },
        55296: {
            "Label": "DM15",
            "Name": "Memory Access Response",
            "PGNLength": 8,
            "Rate": "As needed",
            "SPNs": [
                1646,
                1647,
                1648,
                1599,
                1649
            ]
        },
        55552: {
            "Label": "DM14",
            "Name": "Memory Access Request",
            "PGNLength": 8,
            "Rate": "As needed",
            "SPNs": [
                1642,
                1641,
                1643,
                1644,
                1645,
                1640
            ]
        },
        55808: {
            "Label": "KWP2",
            "Name": "Reserved for ISO 15765",
            "PGNLength": 8,
            "Rate": "",
            "SPNs": []
        },
        56064: {
            "Label": "KWP1",
            "Name": "Reserved for ISO 15765",
            "PGNLength": 8,
            "Rate": "",
            "SPNs": []
        },
        56320: {
            "Label": "ATS",
            "Name": "Anti-theft Status",
            "PGNLength": 8,
            "Rate": "This message is transmitted in response to an Anti-Theft Request message. This message is also sent when the component has an abnormal power interruption.  In this situation the Anti-Theft Status Report is sent without the Anti-Theft Request.",
            "SPNs": [
                1194,
                1195,
                1196,
                1197,
                1198
            ]
        },
        56576: {
            "Label": "ATR",
            "Name": "Anti-theft Request",
            "PGNLength": 8,
            "Rate": "Transmission of this message is interrupt driven.  This message is also transmitted upon power-up of the interfacing device sending this message.",
            "SPNs": [
                1199,
                1200,
                1201,
                1202
            ]
        },
        56832: {
            "Label": "RESET",
            "Name": "Reset",
            "PGNLength": 8,
            "Rate": "When needed",
            "SPNs": [
                988,
                989,
                1584,
                1211,
                3600
            ]
        },
        57088: {
            "Label": "DM13",
            "Name": "Stop Start Broadcast",
            "PGNLength": 8,
            "Rate": "Sent when ever a Stop or Start broadcast event is necessary. To maintain the modified state of the vehicle network(s) the commanding device must send the Hold Signal once every 5 seconds.",
            "SPNs": [
                639,
                622,
                608,
                1230,
                1234,
                1233,
                1232,
                1231,
                1235,
                2618,
                1236,
                2619
            ]
        },
        57344: {
            "Label": "CM1",
            "Name": "Cab Message 1",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                986,
                1684,
                1682,
                1714,
                1856,
                5630,
                1655,
                1654,
                1653,
                3695,
                3696,
                1666,
                1656,
                1683,
                1685,
                1686,
                2596,
                1691
            ]
        },
        57600: {
            "Label": "GPM21",
            "Name": "Reserved for ISO 11992",
            "PGNLength": 8,
            "Rate": "",
            "SPNs": []
        },
        57856: {
            "Label": "GPM11",
            "Name": "Reserved for ISO 11992",
            "PGNLength": 8,
            "Rate": "",
            "SPNs": []
        },
        58112: {
            "Label": "DM07",
            "Name": "Command Non-continuously Monitored Test",
            "PGNLength": 8,
            "Rate": "Sent whenever a test is desired",
            "SPNs": [
                1224,
                4149,
                4148
            ]
        },
        58368: {
            "Label": "RGE11",
            "Name": "Reserved for ISO 11992",
            "PGNLength": 8,
            "Rate": "100 ms +/- 10 ms",
            "SPNs": []
        },
        58624: {
            "Label": "RGE21",
            "Name": "Reserved for ISO 11992",
            "PGNLength": 8,
            "Rate": "100 ms +/- 10 ms",
            "SPNs": []
        },
        58880: {
            "Label": "VT12",
            "Name": "Virtual Terminal-to-Node",
            "PGNLength": "Variable",
            "Rate": "",
            "SPNs": []
        },
        59136: {
            "Label": "VT21",
            "Name": "Node-to-Virtual Terminal",
            "PGNLength": "Variable",
            "Rate": "",
            "SPNs": []
        },
        59392: {
            "Label": "ACKM",
            "Name": "Acknowledgment Message",
            "PGNLength": 8,
            "Rate": "Upon reception of a Parameter Group Number that requires this form of acknowledgment.",
            "SPNs": [
                2541,
                2542,
                2543,
                2544,
                2545,
                2546,
                2547,
                2548,
                2549
            ]
        },
        59904: {
            "Label": "RQST",
            "Name": "Request:3",
            "PGNLength": "Per user requirements",
            "Rate": " generally recommended that requests occur no more than 2 or 3 times per second.",
            "SPNs": [
                2540
            ]
        },
        60160: {
            "Label": "TP.DT",
            "Name": "Transport Protocol - Data Transfer",
            "PGNLength": 8,
            "Rate": "Per the Parameter Group to be transferred",
            "SPNs": [
                2572,
                2573
            ]
        },
        60416: {
            "Label": "TP.CM.xx",
            "Name": "Transport Protocol - Connection Mgmt",
            "PGNLength": 8,
            "Rate": "Per the Parameter Group to be transferred",
            "SPNs": [
                2556,
                2560,
                2561,
                2563,
                2566,
                2569,
                2571
            ]
        },
        60672: {
            "Label": "N.xx",
            "Name": "Network Layer",
            "PGNLength": "Variable",
            "Rate": "100 ms",
            "SPNs": []
        },
        60928: {
            "Label": "AC",
            "Name": "Address Claimed",
            "PGNLength": 8,
            "Rate": "As required",
            "SPNs": [
                2848
            ]
        },
        61184: {
            "Label": "PropA",
            "Name": "Proprietary A",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2550
            ]
        },
        61440: {
            "Label": "ERC1",
            "Name": "Electronic Retarder Controller 1",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                900,
                571,
                572,
                520,
                1085,
                1082,
                1667,
                4233,
                4234,
                1480,
                1715,
                1716,
                1717
            ]
        },
        61441: {
            "Label": "EBC1",
            "Name": "Electronic Brake Controller 1",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                561,
                562,
                563,
                1121,
                521,
                575,
                576,
                577,
                1238,
                972,
                971,
                970,
                969,
                973,
                1243,
                1439,
                1438,
                1793,
                1481,
                2911,
                1836,
                1792
            ]
        },
        61442: {
            "Label": "ETC1",
            "Name": "Electronic Transmission Controller 1",
            "PGNLength": 8,
            "Rate": "10 ms",
            "SPNs": [
                560,
                573,
                574,
                4816,
                522,
                606,
                607,
                5015,
                1482,
                191,
                161
            ]
        },
        61443: {
            "Label": "EEC2",
            "Name": "Electronic Engine Controller 2",
            "PGNLength": 8,
            "Rate": "50 msec (preferred) or Engine Speed Dependent (if required by application)",
            "SPNs": [
                558,
                559,
                1437,
                2970,
                91,
                92,
                974,
                29,
                2979,
                5021,
                5399,
                5400,
                3357,
                5398
            ]
        },
        61444: {
            "Label": "EEC1",
            "Name": "Electronic Engine Controller 1",
            "PGNLength": 8,
            "Rate": "engine speed dependent",
            "SPNs": [
                899,
                4154,
                512,
                513,
                1483,
                1675,
                2432,
                190
            ]
        },
        61445: {
            "Label": "ETC2",
            "Name": "Electronic Transmission Controller 2",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                524,
                523,
                526,
                162,
                163
            ]
        },
        61446: {
            "Label": "EAC1",
            "Name": "Electronic Axle Controller 1",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                927,
                567,
                568,
                569,
                570,
                564,
                565,
                566,
                3819,
                3820
            ]
        },
        61447: {
            "Label": "FLI1",
            "Name": "Forward Lane Image urgent msg",
            "PGNLength": 8,
            "Rate": "50 ms (only when active)",
            "SPNs": [
                3565,
                3566,
                1701,
                1700
            ]
        },
        61448: {
            "Label": "HPG",
            "Name": "Hydraulic Pressure Governor Info",
            "PGNLength": 8,
            "Rate": "50 ms",
            "SPNs": [
                1763,
                1764,
                2599,
                1762
            ]
        },
        61449: {
            "Label": "VDC2",
            "Name": "Vehicle Dynamic Stability Control 2",
            "PGNLength": 8,
            "Rate": "10 ms",
            "SPNs": [
                1811,
                1812,
                1810,
                1807,
                1808,
                1809
            ]
        },
        61450: {
            "Label": "EGF1",
            "Name": "Engine Gas Flow Rate",
            "PGNLength": 8,
            "Rate": "50 ms",
            "SPNs": [
                2659,
                132,
                5257
            ]
        },
        61451: {
            "Label": "ESC1",
            "Name": "Electronic Steering Control",
            "PGNLength": 8,
            "Rate": "20 ms",
            "SPNs": [
                2928,
                2923,
                2922,
                2924,
                2925,
                2926,
                2927
            ]
        },
        61452: {
            "Label": "ETC8",
            "Name": "Electronic Transmission Controller #8:8",
            "PGNLength": "20 ms when torque converter unlocked",
            "Rate": " 100 ms when torque converter locked",
            "SPNs": [
                3030,
                5052
            ]
        },
        61453: {
            "Label": "LOI",
            "Name": "Land Leveling System Operational Information",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                3156,
                3157,
                3158,
                3334,
                3335,
                3336,
                3337,
                3338,
                3339,
                5405,
                5406
            ]
        },
        61454: {
            "Label": "AT1IG1",
            "Name": "Aftertreatment 1 Intake Gas 1",
            "PGNLength": 8,
            "Rate": "50 ms",
            "SPNs": [
                3218,
                3219,
                3220,
                3221,
                3222,
                3223,
                3224,
                5713,
                3225,
                3216,
                3217
            ]
        },
        61455: {
            "Label": "AT1OG1",
            "Name": "Aftertreatment 1 Outlet Gas 1",
            "PGNLength": 8,
            "Rate": "50 ms",
            "SPNs": [
                3228,
                3229,
                3230,
                3231,
                3232,
                3233,
                3234,
                5714,
                3235,
                3226,
                3227
            ]
        },
        61456: {
            "Label": "AT2IG1",
            "Name": "Aftertreatment 2 Intake Gas 1",
            "PGNLength": 8,
            "Rate": "50 ms",
            "SPNs": [
                3257,
                3258,
                3259,
                3260,
                3261,
                3262,
                3263,
                5715,
                3264,
                3255,
                3256
            ]
        },
        61457: {
            "Label": "AT2OG1",
            "Name": "Aftertreatment 2 Outlet Gas 1",
            "PGNLength": 8,
            "Rate": "50 ms",
            "SPNs": [
                3267,
                3268,
                3269,
                3270,
                3271,
                3272,
                3273,
                5716,
                3274,
                3265,
                3266
            ]
        },
        61458: {
            "Label": "FWSS1",
            "Name": "Fifth Wheel Smart Systems 1",
            "PGNLength": 8,
            "Rate": "50 ms",
            "SPNs": [
                3317,
                3308,
                3309,
                3310
            ]
        },
        61459: {
            "Label": "SSI",
            "Name": "Slope Sensor Information",
            "PGNLength": 8,
            "Rate": "10 ms",
            "SPNs": [
                3323,
                3324,
                3325,
                3326,
                3327,
                3318,
                3319,
                3322
            ]
        },
        61460: {
            "Label": "BI",
            "Name": "Blade Information",
            "PGNLength": 8,
            "Rate": "50 ms",
            "SPNs": [
                3366,
                3367,
                3332,
                3365,
                3331
            ]
        },
        61461: {
            "Label": "RGTACRP",
            "Name": "Requested Generator Total AC Reactive Power",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                3385,
                3383,
                3384
            ]
        },
        61462: {
            "Label": "CCS",
            "Name": "Cylinder Combustion Status:8",
            "PGNLength": "Engine speed dependent when there is no combustion",
            "Rate": " once every 5 seconds otherwise",
            "SPNs": [
                3387,
                3388,
                3389,
                3390,
                3391,
                3392,
                3393,
                3394,
                3395,
                3396,
                3397,
                3398,
                3399,
                3400,
                3401,
                3402,
                3403,
                3404,
                3405,
                3406,
                3407,
                3408,
                3409,
                3410
            ]
        },
        61463: {
            "Label": "KL1",
            "Name": "Engine Knock Level #1:8",
            "PGNLength": "Engine speed dependent when knock present",
            "Rate": " once every 5 seconds otherwise",
            "SPNs": [
                1352,
                1353,
                1354,
                1355,
                1356,
                1357,
                1358,
                1359
            ]
        },
        61464: {
            "Label": "KL2",
            "Name": "Engine Knock Level #2:8",
            "PGNLength": "Engine speed dependent when knock present",
            "Rate": " once every 5 seconds otherwise",
            "SPNs": [
                1360,
                1361,
                1362,
                1363,
                1364,
                1365,
                1366,
                1367
            ]
        },
        61465: {
            "Label": "KL3",
            "Name": "Engine Knock Level #3:8",
            "PGNLength": "Engine speed dependent when knock present",
            "Rate": " once every 5 seconds otherwise",
            "SPNs": [
                1368,
                1369,
                1370,
                1371,
                1372,
                1373,
                1374,
                1375
            ]
        },
        61466: {
            "Label": "TFAC",
            "Name": "Engine Throttle / Fuel Actuator Control Command",
            "PGNLength": 8,
            "Rate": "50 msec (preferred) or Engine Speed Dependent (if required by application)",
            "SPNs": [
                3464,
                3465,
                633,
                1244
            ]
        },
        61467: {
            "Label": "GPM17",
            "Name": "General Purpose Message #1/7",
            "PGNLength": 8,
            "Rate": "10 ms",
            "SPNs": []
        },
        61468: {
            "Label": "RGAAC",
            "Name": "Requested Generator Average Basic AC Quantities",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                3386
            ]
        },
        61469: {
            "Label": "SAS",
            "Name": "Steering Angle Sensor Information",
            "PGNLength": 8,
            "Rate": "10 ms",
            "SPNs": [
                3684,
                3685,
                3687,
                3688,
                3689,
                3690,
                3683,
                3686
            ]
        },
        61470: {
            "Label": "GC2",
            "Name": "Generator Control 2",
            "PGNLength": 8,
            "Rate": "20 ms",
            "SPNs": [
                3938
            ]
        },
        61471: {
            "Label": "EBS26",
            "Name": "Electronic Brake System #2/6",
            "PGNLength": 8,
            "Rate": "10 ms +/- 1 ms",
            "SPNs": []
        },
        61472: {
            "Label": "EBS25",
            "Name": "Electronic Brake System #2/5",
            "PGNLength": 8,
            "Rate": "10 ms +/- 1 ms",
            "SPNs": []
        },
        61473: {
            "Label": "ESSI",
            "Name": "Engine Speed Sensor Information:8",
            "PGNLength": "Transmitted only after requested.  After request",
            "Rate": " broadcast rate is engine speed dependent.  Update stopped after key switch cycle.",
            "SPNs": [
                4205,
                4204,
                4203,
                4201,
                723,
                4202
            ]
        },
        61474: {
            "Label": "MSS",
            "Name": "Machine Selected Speed",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                4309,
                4308,
                4307,
                4305,
                4306
            ]
        },
        61475: {
            "Label": "A1SCRDSI1",
            "Name": "Aftertreatment 1 SCR Dosing System Information 1",
            "PGNLength": 8,
            "Rate": "50 ms",
            "SPNs": [
                4332,
                4334,
                4331,
                4333
            ]
        },
        61476: {
            "Label": "A1SCRDSR1",
            "Name": "Aftertreatment 1 SCR Dosing System Requests 1",
            "PGNLength": 8,
            "Rate": "50 ms",
            "SPNs": [
                4349,
                4348,
                4350
            ]
        },
        61477: {
            "Label": "A1SCRAI",
            "Name": "Aftertreatment 1 SCR Ammonia Information",
            "PGNLength": 8,
            "Rate": "50 ms",
            "SPNs": [
                4378,
                4379,
                4380,
                4381,
                4382,
                4383,
                4377
            ]
        },
        61478: {
            "Label": "A2SCRDSI1",
            "Name": "Aftertreatment 2 SCR Dosing System Information 1",
            "PGNLength": 8,
            "Rate": "50 ms",
            "SPNs": [
                4385,
                4387,
                4384,
                4386
            ]
        },
        61479: {
            "Label": "A2SCRDSR1",
            "Name": "Aftertreatment 2 SCR Dosing System Requests 1",
            "PGNLength": 8,
            "Rate": "50 ms",
            "SPNs": [
                4402,
                4401,
                4403
            ]
        },
        61480: {
            "Label": "A2SCRAI",
            "Name": "Aftertreatment 2 SCR Ammonia Information",
            "PGNLength": 8,
            "Rate": "50 ms",
            "SPNs": [
                4444,
                4445,
                4446,
                4447,
                4448,
                4449,
                4443
            ]
        },
        61481: {
            "Label": "SSI2",
            "Name": "Slope Sensor Information 2",
            "PGNLength": 8,
            "Rate": "10 ms",
            "SPNs": [
                4978,
                4979,
                4980,
                4981,
                4982,
                4976,
                4977
            ]
        },
        61482: {
            "Label": "ARI",
            "Name": "Angular Rate Information",
            "PGNLength": 8,
            "Rate": "10 ms",
            "SPNs": [
                4986,
                4987,
                4988,
                4989,
                4983,
                4984,
                4985
            ]
        },
        61483: {
            "Label": "CN",
            "Name": "Crash Notification",
            "PGNLength": 8,
            "Rate": "Transmitted every 20 msec for the first 100 msec and then broadcast every 1 sec for 10 sec in case of a crash event",
            "SPNs": [
                4973,
                4974,
                4975
            ]
        },
        61484: {
            "Label": "MSI2",
            "Name": "Magnet Status Information 2",
            "PGNLength": 8,
            "Rate": "50 ms",
            "SPNs": [
                5009,
                5010,
                5011,
                5012,
                5013,
                5014,
                5402,
                5403,
                5551,
                5008
            ]
        },
        61485: {
            "Label": "ACCS",
            "Name": "Acceleration Sensor",
            "PGNLength": 8,
            "Rate": "10 ms (default) or 20 ms",
            "SPNs": [
                5350,
                5351,
                5352,
                5353,
                5347,
                5348,
                5349
            ]
        },
        61486: {
            "Label": "ETWAC",
            "Name": "Engine Turbocharger Wastegate Actuator and Exhaust Back Pressure Regulator Command",
            "PGNLength": 8,
            "Rate": "Engine speed dependent",
            "SPNs": [
                5386,
                5387,
                649
            ]
        },
        61487: {
            "Label": "AEBS1",
            "Name": "Advanced Emergency Braking System 1",
            "PGNLength": 8,
            "Rate": "Every 50ms and on change of AEBS system state or change of Collision warning status but no faster than every 10 ms",
            "SPNs": [
                5676,
                5677,
                5678,
                5679,
                5680
            ]
        },
        61488: {
            "Label": "WS",
            "Name": "Wireless Status",
            "PGNLength": 8,
            "Rate": "50 ms",
            "SPNs": [
                5751,
                5753,
                5750,
                5752
            ]
        },
        61489: {
            "Label": "WAND",
            "Name": "Wand Sensor",
            "PGNLength": 8,
            "Rate": "20 ms",
            "SPNs": [
                5822,
                5821
            ]
        },
        61490: {
            "Label": "LDISP",
            "Name": "Linear Displacement Sensor",
            "PGNLength": 8,
            "Rate": "20 ms",
            "SPNs": [
                5824,
                5823
            ]
        },
        61491: {
            "Label": "A1SCRAI2",
            "Name": "Aftertreatment 1 SCR Ammonia Information 2",
            "PGNLength": 8,
            "Rate": "50 ms",
            "SPNs": [
                5849,
                5850,
                5851,
                5852,
                5853,
                5854,
                5848
            ]
        },
        61492: {
            "Label": "A2SCRAI2",
            "Name": "Aftertreatment 2 SCR Ammonia Information 2",
            "PGNLength": 8,
            "Rate": "50 ms",
            "SPNs": [
                5856,
                5857,
                5858,
                5859,
                5860,
                5861,
                5855
            ]
        },
        64706: {
            "Label": "HSS",
            "Name": "Hybrid System Status",
            "PGNLength": 8,
            "Rate": "Every 1 sec and on change but no faster than 100 ms",
            "SPNs": [
                5872,
                5873,
                5874
            ]
        },
        64707: {
            "Label": "GFI4",
            "Name": "Fuel Information 4 (Gaseous)",
            "PGNLength": "8 bytes",
            "Rate": "200 ms",
            "SPNs": [
                5867
            ]
        },
        64708: {
            "Label": "A2SCREGT2",
            "Name": "Aftertreatment 2 SCR Exhaust Gas Temperature 2",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                5865,
                5864
            ]
        },
        64709: {
            "Label": "A1SCREGT2",
            "Name": "Aftertreatment 1 SCR Exhaust Gas Temperature 2",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                5863,
                5862
            ]
        },
        64710: {
            "Label": "DM57",
            "Name": "OBD Information",
            "PGNLength": 0,
            "Rate": "On request",
            "SPNs": [
                5843
            ]
        },
        64711: {
            "Label": "DM56",
            "Name": "Engine Emissions Certification Information",
            "PGNLength": 0,
            "Rate": "On request",
            "SPNs": [
                5844,
                5845
            ]
        },
        64712: {
            "Label": "EEC13",
            "Name": "Electronic Engine Controller 13",
            "PGNLength": 8,
            "Rate": "Every 1 sec and on change but no faster than 100 ms",
            "SPNs": [
                5794,
                5795,
                5866
            ]
        },
        64713: {
            "Label": "EF/TVI2",
            "Name": "Engine Fuel/Throttle Valve Information 2",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                5783,
                5784,
                5785,
                5786,
                5787
            ]
        },
        64714: {
            "Label": "EEGR1A",
            "Name": "Engine Exhaust Gas Recirculation 2 Actuator",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                5773,
                5774,
                5775,
                5776,
                5777,
                5778,
                5779,
                5780,
                5781,
                5782
            ]
        },
        64715: {
            "Label": "EEGR1A",
            "Name": "Engine Exhaust Gas Recirculation 1 Actuator",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                5763,
                5764,
                5765,
                5766,
                5767,
                5768,
                5769,
                5770,
                5771,
                5772
            ]
        },
        64716: {
            "Label": "EEC12",
            "Name": "Electronic Engine Controller 12",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                5758,
                5759,
                5760,
                5761
            ]
        },
        64717: {
            "Label": "DSS",
            "Name": "Drive Strategy Status",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": []
        },
        64718: {
            "Label": "DSC",
            "Name": "Drive Strategy Command",
            "PGNLength": 8,
            "Rate": "Every 500 ms and on change of data in the PGN but no faster than 100 ms",
            "SPNs": []
        },
        64719: {
            "Label": "NSSR",
            "Name": "NOx Sensor Self-Diagnosis Request",
            "PGNLength": 8,
            "Rate": "As required",
            "SPNs": [
                5709,
                5710,
                5711,
                5712
            ]
        },
        64720: {
            "Label": "EPSI",
            "Name": "Engine Particulate Sensor Information",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                5835,
                5836
            ]
        },
        64721: {
            "Label": "DM53",
            "Name": "Active Service Only DTCs",
            "PGNLength": "Variable",
            "Rate": "On request",
            "SPNs": []
        },
        64722: {
            "Label": "DM54",
            "Name": "Previously Active Service Only DTCs",
            "PGNLength": "Variable",
            "Rate": "On request",
            "SPNs": []
        },
        64723: {
            "Label": "DM55",
            "Name": "Diagnostic Data Clear/Reset for All Service Only DTCs",
            "PGNLength": 0,
            "Rate": "On request",
            "SPNs": []
        },
        64724: {
            "Label": "XCP ID",
            "Name": "XCP Get Slave ID",
            "PGNLength": 8,
            "Rate": "As defined in protocol",
            "SPNs": []
        },
        64725: {
            "Label": "AT1P1I",
            "Name": "Aftertreatment 1 Diesel Exhaust Fluid Pump 1 Information",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                5706
            ]
        },
        64726: {
            "Label": "AT1DPF2S5",
            "Name": "Aftertreatment 1 DPF 2 Soot 5",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                5652,
                5653,
                5654,
                5655
            ]
        },
        64727: {
            "Label": "AT1DPF2S4",
            "Name": "Aftertreatment 1 DPF 2 Soot 4",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                5648,
                5649,
                5650,
                5651
            ]
        },
        64728: {
            "Label": "AT1DPF2S3",
            "Name": "Aftertreatment 1 DPF 2 Soot 3",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                5644,
                5645,
                5646,
                5647
            ]
        },
        64729: {
            "Label": "AT1DPF1S5",
            "Name": "Aftertreatment 1 DPF 1 Soot 5",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                5640,
                5641,
                5642,
                5643
            ]
        },
        64730: {
            "Label": "AT1DPF1S4",
            "Name": "Aftertreatment 1 DPF 1 Soot 4",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                5636,
                5637,
                5638,
                5639
            ]
        },
        64731: {
            "Label": "AT1DPF1S3",
            "Name": "Aftertreatment 1 DPF 1 Soot 3",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                5632,
                5633,
                5634,
                5635
            ]
        },
        64732: {
            "Label": "CCVS3",
            "Name": "Cruise Control 3",
            "PGNLength": 8,
            "Rate": "Every 1 sec and on change but no faster than 100 ms",
            "SPNs": [
                5606,
                5607,
                5608,
                5609
            ]
        },
        64733: {
            "Label": "AT2AC3",
            "Name": "Aftertreatment 2 Air Control 3",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                5590
            ]
        },
        64734: {
            "Label": "AT1AC3",
            "Name": "Aftertreatment 1 Air Control 3",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                5589
            ]
        },
        64735: {
            "Label": "EFL/P12",
            "Name": "Engine Fluid Level/Pressure 12",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                5578,
                5579,
                5580,
                5584
            ]
        },
        64736: {
            "Label": "DPFC2",
            "Name": "DPF Control 2",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                5501,
                5502,
                5503,
                5834
            ]
        },
        64737: {
            "Label": "LFE2",
            "Name": "Fuel Economy 2 (Liquid)",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                1600
            ]
        },
        64738: {
            "Label": "SFS",
            "Name": "Supplemental Fan Status",
            "PGNLength": 8,
            "Rate": "1 second",
            "SPNs": [
                5556,
                5557,
                5558,
                5559,
                5560,
                5561,
                5563,
                5562,
                5564
            ]
        },
        64739: {
            "Label": "EBCC",
            "Name": "Engine Exhaust Brake Control",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                5541,
                5542,
                5543,
                5544,
                5545
            ]
        },
        64740: {
            "Label": "EFP",
            "Name": "Engine Fuel Properties",
            "PGNLength": 8,
            "Rate": "30 seconds",
            "SPNs": [
                5537,
                5538,
                5539,
                5540
            ]
        },
        64741: {
            "Label": "HOP",
            "Name": "Hydraulic Oil Properties",
            "PGNLength": 8,
            "Rate": "30 seconds",
            "SPNs": [
                5533,
                5534,
                5535,
                5536
            ]
        },
        64742: {
            "Label": "VLS2",
            "Name": "Vehicle/Chassis Lubrication System 2",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                5515,
                5516,
                5517,
                5518,
                5519,
                5520
            ]
        },
        64743: {
            "Label": "EC3",
            "Name": "Engine Configuration 3",
            "PGNLength": 8,
            "Rate": "Every 5 s and on change of torque/speed points of more than 10% since last transmission but no faster than every 500 ms",
            "SPNs": [
                5471,
                5472,
                5473,
                5474,
                5475,
                5476,
                5477
            ]
        },
        64744: {
            "Label": "IVAC",
            "Name": "Intake Valve Actuation Control",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                5499,
                5500
            ]
        },
        64745: {
            "Label": "ARMSWIMC",
            "Name": "Armrest Switch Matrix Commands",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                1503,
                5492,
                5493,
                5494
            ]
        },
        64746: {
            "Label": "VEP4",
            "Name": "Vehicle Electrical Power #4",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                5464
            ]
        },
        64747: {
            "Label": "AFT2NAI",
            "Name": "Aftertreatment 2 NOx Adsorber Information",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                5460,
                5462
            ]
        },
        64748: {
            "Label": "AFT1NAI",
            "Name": "Aftertreatment 1 NOx Adsorber Information",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                5459,
                5461
            ]
        },
        64749: {
            "Label": "AT2WUDOC",
            "Name": "Aftertreatment 2 Warm Up Diesel Oxidation Catalyst Information",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                5315,
                5316
            ]
        },
        64750: {
            "Label": "LLRE",
            "Name": "Land Leveling System Reference Elevation Data",
            "PGNLength": 8,
            "Rate": "200 ms",
            "SPNs": [
                5410,
                5411,
                5412,
                5413
            ]
        },
        64751: {
            "Label": "EFL/P11",
            "Name": "Engine Fluid Level/Pressure 11",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                5430,
                5431
            ]
        },
        64752: {
            "Label": "EFS2",
            "Name": "Engine Fuel/lube systems 2",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                5417
            ]
        },
        64753: {
            "Label": "ETWAI",
            "Name": "Engine Turbocharger Wastegate Actuator and Exhaust Back Pressure Regulator Information",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                5384,
                5385,
                5451,
                5452,
                5625,
                5626,
                5627,
                5628,
                5788,
                5789,
                5790
            ]
        },
        64754: {
            "Label": "EFTVI",
            "Name": "Engine Fuel/Throttle Valve Information",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                5374,
                5375,
                5376,
                5377,
                5378,
                5379,
                5380,
                5381,
                5382,
                5383,
                5445,
                5446,
                5447,
                5448
            ]
        },
        64755: {
            "Label": "ETCBI",
            "Name": "Engine Turbocharger Compressor Bypass Information",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                5388,
                5389,
                5390,
                5391,
                5449,
                5450,
                5791,
                5792
            ]
        },
        64756: {
            "Label": "CAC2",
            "Name": "Charge Air Cooler 2",
            "PGNLength": 8,
            "Rate": "1 sec",
            "SPNs": [
                5292,
                5289,
                5290,
                5291
            ]
        },
        64757: {
            "Label": "CAC2P",
            "Name": "Charge Air Cooler 2 Precooler",
            "PGNLength": 8,
            "Rate": "1 sec",
            "SPNs": [
                5288,
                5286,
                5287
            ]
        },
        64758: {
            "Label": "CAC1",
            "Name": "Charge Air Cooler 1",
            "PGNLength": 8,
            "Rate": "1 sec",
            "SPNs": [
                5285,
                5283,
                5284
            ]
        },
        64759: {
            "Label": "CAC1P",
            "Name": "Charge Air Cooler 1 Precooler",
            "PGNLength": 8,
            "Rate": "1 sec",
            "SPNs": [
                5282,
                5280,
                5281
            ]
        },
        64760: {
            "Label": "DPF2S2",
            "Name": "DPF 2 Soot 2",
            "PGNLength": 8,
            "Rate": "1 sec",
            "SPNs": [
                5269,
                5270,
                5271
            ]
        },
        64761: {
            "Label": "DPF1S2",
            "Name": "DPF 1 Soot 2",
            "PGNLength": 8,
            "Rate": "1 sec",
            "SPNs": [
                5266,
                5267,
                5268
            ]
        },
        64762: {
            "Label": "EEC11",
            "Name": "Electronic Engine Controller 11",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                5264,
                5265,
                5831,
                5832
            ]
        },
        64763: {
            "Label": "EMAP",
            "Name": "Engine Manifold Actuator Position",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                5277,
                5279
            ]
        },
        64764: {
            "Label": "EMAC",
            "Name": "Engine Manifold Actuator Control",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                5276,
                5278
            ]
        },
        64765: {
            "Label": "EEC9",
            "Name": "Electronic Engine Controller 9",
            "PGNLength": 8,
            "Rate": "100 msec",
            "SPNs": [
                5262,
                5263,
                5313,
                5314
            ]
        },
        64766: {
            "Label": "EEC10",
            "Name": "Electronic Engine Controller 10",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                5260,
                5261,
                5258,
                5259
            ]
        },
        64767: {
            "Label": "ET5",
            "Name": "Engine Temperature 5",
            "PGNLength": 8,
            "Rate": "1 sec",
            "SPNs": [
                5255,
                5256
            ]
        },
        64768: {
            "Label": "EFL/P10",
            "Name": "Engine Fluid Level/Pressure 10",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                5253,
                5708,
                5252,
                5254,
                5429
            ]
        },
        64769: {
            "Label": "LVDS",
            "Name": "Low Voltage Disconnect  Status",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                5141,
                5142,
                5143,
                5144,
                5145,
                5146,
                5147,
                5148
            ]
        },
        64770: {
            "Label": "AISO",
            "Name": "All implements stop operations switch state",
            "PGNLength": 8,
            "Rate": "Every 1 s and on change of state but no faster than every 100 ms",
            "SPNs": [
                5140
            ]
        },
        64771: {
            "Label": "IOS",
            "Name": "Implement Operating State Command",
            "PGNLength": 8,
            "Rate": "Every 1 s and on change of state but no faster than every 100 ms",
            "SPNs": [
                5139
            ]
        },
        64772: {
            "Label": "DLCD2",
            "Name": "Direct Lamp Control Data 2",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                5104,
                5105,
                5106,
                5107,
                5108
            ]
        },
        64773: {
            "Label": "DLCD1",
            "Name": "Direct Lamp Control Data 1",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                5093,
                5094,
                5095,
                5096,
                5097,
                5098,
                5099,
                5100,
                5101,
                5102,
                5103,
                5470,
                5416
            ]
        },
        64774: {
            "Label": "DLCC2",
            "Name": "Direct Lamp Control Command 2",
            "PGNLength": 8,
            "Rate": "Every 1 s and on change of state but no faster than every 100 ms",
            "SPNs": [
                5087,
                5088,
                5089,
                5090,
                5091
            ]
        },
        64775: {
            "Label": "DLCC1",
            "Name": "Direct Lamp Control Command 1",
            "PGNLength": 8,
            "Rate": "Every 1 s and on change of state but no faster than every 100 ms",
            "SPNs": [
                5077,
                5078,
                5079,
                5080,
                5081,
                3987,
                5082,
                5083,
                5084,
                5085,
                5086,
                5469
            ]
        },
        64776: {
            "Label": "EOM",
            "Name": "Engine oil message",
            "PGNLength": 8,
            "Rate": "30 seconds",
            "SPNs": [
                5055,
                5056,
                5468
            ]
        },
        64777: {
            "Label": "HRLFC",
            "Name": "High Resolution Fuel Consumption (Liquid)",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                5053,
                5054
            ]
        },
        64778: {
            "Label": "AT2OGC2",
            "Name": "Aftertreatment 2 Outlet Gas NOx Sensor correction data 2",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                5048,
                5049,
                5050,
                5051,
                5719
            ]
        },
        64779: {
            "Label": "AT2OGC1",
            "Name": "Aftertreatment 2 Outlet Gas NOx Sensor correction data:8",
            "PGNLength": "On start-up",
            "Rate": " and every second until the dewpoint signal state = 1 (SPN 3240) has been received by the transmitter",
            "SPNs": [
                5047,
                5045,
                5046,
                5723
            ]
        },
        64780: {
            "Label": "AT2IGC2",
            "Name": "Aftertreatment 2 Intake Gas NOx Sensor correction data 2",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                5041,
                5042,
                5043,
                5044,
                5720
            ]
        },
        64781: {
            "Label": "AT2IGC1",
            "Name": "Aftertreatment 2 Intake Gas NOx Sensor correction data:8",
            "PGNLength": "On start-up",
            "Rate": " and every second until the dewpoint signal state = 1 (SPN 3239) has been received by the transmitter",
            "SPNs": [
                5040,
                5038,
                5039,
                5724
            ]
        },
        64782: {
            "Label": "AT1OGC2",
            "Name": "Aftertreatment 1 Outlet Gas NOx Sensor correction data 2",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                5034,
                5035,
                5036,
                5037,
                5717
            ]
        },
        64783: {
            "Label": "AT1OGC1",
            "Name": "Aftertreatment 1 Outlet Gas NOx Sensor correction data:8",
            "PGNLength": "On start-up",
            "Rate": " and every second until the dewpoint signal state = 1 (SPN 3238) has been received by the transmitter",
            "SPNs": [
                5033,
                5031,
                5032,
                5721
            ]
        },
        64784: {
            "Label": "AT1IGC2",
            "Name": "Aftertreatment 1 Intake Gas NOx Sensor correction data 2",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                5027,
                5028,
                5029,
                5030,
                5718
            ]
        },
        64785: {
            "Label": "AT1IGC1",
            "Name": "Aftertreatment 1 Intake Gas  NOx Sensor correction data:8",
            "PGNLength": "On start-up",
            "Rate": " and every second until the dewpoint signal state = 1 (SPN 3237) has been received by the transmitter",
            "SPNs": [
                5026,
                5024,
                5025,
                5722
            ]
        },
        64786: {
            "Label": "MSI1",
            "Name": "Magnet Status Information 1",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                5000,
                5001,
                5003,
                5004,
                5002,
                5005,
                5006,
                5007
            ]
        },
        64787: {
            "Label": "MSCI",
            "Name": "Magnet System Configuration Information",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                4998,
                4999
            ]
        },
        64788: {
            "Label": "BCH2",
            "Name": "Battery Charger 2",
            "PGNLength": 8,
            "Rate": "1 sec",
            "SPNs": [
                4994,
                4995,
                4996,
                4997
            ]
        },
        64789: {
            "Label": "BCH1",
            "Name": "Battery Charger 1",
            "PGNLength": 8,
            "Rate": "1 sec",
            "SPNs": [
                4990,
                4991,
                4992,
                4993
            ]
        },
        64790: {
            "Label": "OCSI",
            "Name": "Occupant Classification System Information",
            "PGNLength": 8,
            "Rate": "Transmitted every 5 sec and on change of PGN 64791 but no faster than every 250 msec",
            "SPNs": [
                4963,
                4964,
                4965,
                4966,
                4967,
                4968,
                4969,
                4970,
                4971,
                4972
            ]
        },
        64791: {
            "Label": "BDS",
            "Name": "Beltlock and Airbag Deactivation Switch Information",
            "PGNLength": 8,
            "Rate": "250 msec",
            "SPNs": [
                4952,
                4953,
                4954,
                4955,
                4956,
                4957,
                4958,
                4959,
                4960,
                4961,
                4962
            ]
        },
        64792: {
            "Label": "CSI",
            "Name": "Collision Sensor Information",
            "PGNLength": "Variable",
            "Rate": "On request",
            "SPNs": [
                4918,
                4919,
                4920,
                4921,
                4922,
                4923,
                4924,
                4925,
                4926,
                4927,
                4928,
                4929,
                4930,
                4931,
                4932,
                4933,
                4934,
                4935,
                4936,
                4937,
                4938,
                4939,
                4940,
                4941,
                4942,
                4943,
                4944,
                4945,
                4946,
                4947,
                4948,
                4949,
                4950,
                4951
            ]
        },
        64793: {
            "Label": "ILI",
            "Name": "Ignitor Loop Information",
            "PGNLength": "Variable",
            "Rate": "On request",
            "SPNs": [
                4818,
                4819,
                4820,
                4821,
                4822,
                4823,
                4824,
                4825,
                4826,
                4827,
                4828,
                4829,
                4830,
                4831,
                4832,
                4833,
                4834,
                4835,
                4836,
                4837,
                4838,
                4839,
                4840,
                4841,
                4842,
                4843,
                4844,
                4845,
                4846,
                4847,
                4848,
                4849,
                4850,
                4851,
                4852,
                4853,
                4854,
                4855,
                4856,
                4857,
                4858,
                4859,
                4860,
                4861,
                4862,
                4863,
                4864,
                4865,
                4866,
                4867,
                4868,
                4869,
                4870,
                4871,
                4872,
                4873,
                4874,
                4875,
                4876,
                4877,
                4878,
                4879,
                4880,
                4881,
                4882,
                4883,
                4884,
                4885,
                4886,
                4887,
                4888,
                4889,
                4890,
                4891,
                4892,
                4893,
                4894,
                4895,
                4896,
                4897,
                4898,
                4899,
                4900,
                4901,
                4902,
                4903,
                4904,
                4905,
                4906,
                4907,
                4908,
                4909,
                4910,
                4911,
                4912,
                4913,
                4914,
                4915,
                4916,
                4917
            ]
        },
        64794: {
            "Label": "AT1WUDOC",
            "Name": "Aftertreatment 1 Warm Up Diesel Oxidation Catalyst Information",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                4809,
                4810
            ]
        },
        64795: {
            "Label": "DPF2S",
            "Name": "DPF 2 Soot",
            "PGNLength": 8,
            "Rate": "1 sec",
            "SPNs": [
                4786,
                4787,
                4790,
                5570,
                4788,
                4789
            ]
        },
        64796: {
            "Label": "DPF1S",
            "Name": "DPF 1 Soot",
            "PGNLength": 8,
            "Rate": "1 sec",
            "SPNs": [
                4781,
                4782,
                4785,
                5569,
                4783,
                4784
            ]
        },
        64797: {
            "Label": "ATDT2",
            "Name": "Aftertreatment Differential Temperature 2",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                4779,
                4780
            ]
        },
        64798: {
            "Label": "ATDT1",
            "Name": "Aftertreatment Differential Temperature 1",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                4777,
                4778
            ]
        },
        64799: {
            "Label": "A2DOC",
            "Name": "Aftertreatment 2 Diesel Oxidation Catalyst",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                4774,
                4776,
                4771,
                4772,
                4773,
                4775
            ]
        },
        64800: {
            "Label": "A1DOC",
            "Name": "Aftertreatment 1 Diesel Oxidation Catalyst",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                4768,
                4770,
                4765,
                4766,
                4767,
                4769
            ]
        },
        64801: {
            "Label": "A2GOC",
            "Name": "Aftertreatment 2 Gas Oxidation Catalyst",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                4762,
                4764,
                4759,
                4760,
                4761,
                4763
            ]
        },
        64802: {
            "Label": "A1GOC",
            "Name": "Aftertreatment 1 Gas Oxidation Catalyst",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                4756,
                4758,
                4753,
                4754,
                4755,
                4757
            ]
        },
        64803: {
            "Label": " 2010: Every 100 ms or on change of state",
            "Name": "Extended Joystick Message 10:8",
            "PGNLength": "Every 100 ms and on change of state",
            "Rate": " but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July",
            "SPNs": [
                4735,
                4736,
                4737,
                4739,
                4740,
                4741,
                4743,
                4744,
                4745,
                4747,
                4748,
                4749,
                4738,
                4742,
                4746
            ]
        },
        64804: {
            "Label": " 2010: Every 100 ms or on change of state",
            "Name": "Basic Joystick Message 10:8",
            "PGNLength": "Every 100 ms and on change of state",
            "Rate": " but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July",
            "SPNs": [
                4713,
                4714,
                4715,
                4717,
                4718,
                4719,
                4721,
                4722,
                4723,
                4724,
                4725,
                4726,
                4727,
                4728,
                4729,
                4730,
                4731,
                4732,
                4733,
                4734,
                4716,
                4720
            ]
        },
        64805: {
            "Label": " 2010: Every 100 ms or on change of state",
            "Name": "Extended Joystick Message 9:8",
            "PGNLength": "Every 100 ms and on change of state",
            "Rate": " but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July",
            "SPNs": [
                4698,
                4699,
                4700,
                4702,
                4703,
                4704,
                4706,
                4707,
                4708,
                4710,
                4711,
                4712,
                4701,
                4705,
                4709
            ]
        },
        64806: {
            "Label": " 2010: Every 100 ms or on change of state",
            "Name": "Basic Joystick Message 9:8",
            "PGNLength": "Every 100 ms and on change of state",
            "Rate": " but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July",
            "SPNs": [
                4676,
                4677,
                4678,
                4680,
                4681,
                4682,
                4684,
                4685,
                4686,
                4687,
                4688,
                4689,
                4690,
                4691,
                4692,
                4693,
                4694,
                4695,
                4696,
                4697,
                4679,
                4683
            ]
        },
        64807: {
            "Label": " 2010: Every 100 ms or on change of state",
            "Name": "Extended Joystick Message 8:8",
            "PGNLength": "Every 100 ms and on change of state",
            "Rate": " but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July",
            "SPNs": [
                4661,
                4662,
                4663,
                4665,
                4666,
                4667,
                4669,
                4670,
                4671,
                4673,
                4674,
                4675,
                4664,
                4668,
                4672
            ]
        },
        64808: {
            "Label": " 2010: Every 100 ms or on change of state",
            "Name": "Basic Joystick Message 8:8",
            "PGNLength": "Every 100 ms and on change of state",
            "Rate": " but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July",
            "SPNs": [
                4639,
                4640,
                4641,
                4643,
                4644,
                4645,
                4647,
                4648,
                4649,
                4650,
                4651,
                4652,
                4653,
                4654,
                4655,
                4656,
                4657,
                4658,
                4659,
                4660,
                4642,
                4646
            ]
        },
        64809: {
            "Label": " 2010: Every 100 ms or on change of state",
            "Name": "Extended Joystick Message 7:8",
            "PGNLength": "Every 100 ms and on change of state",
            "Rate": " but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July",
            "SPNs": [
                4624,
                4625,
                4626,
                4628,
                4629,
                4630,
                4632,
                4633,
                4634,
                4636,
                4637,
                4638,
                4627,
                4631,
                4635
            ]
        },
        64810: {
            "Label": " 2010: Every 100 ms or on change of state",
            "Name": "Basic Joystick Message 7:8",
            "PGNLength": "Every 100 ms and on change of state",
            "Rate": " but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July",
            "SPNs": [
                4602,
                4603,
                4604,
                4606,
                4607,
                4608,
                4610,
                4611,
                4612,
                4613,
                4614,
                4615,
                4616,
                4617,
                4618,
                4619,
                4620,
                4621,
                4622,
                4623,
                4605,
                4609
            ]
        },
        64811: {
            "Label": " 2010: Every 100 ms or on change of state",
            "Name": "Extended Joystick Message 6:8",
            "PGNLength": "Every 100 ms and on change of state",
            "Rate": " but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July",
            "SPNs": [
                4587,
                4588,
                4589,
                4591,
                4592,
                4593,
                4595,
                4596,
                4597,
                4599,
                4600,
                4601,
                4590,
                4594,
                4598
            ]
        },
        64812: {
            "Label": " 2010: Every 100 ms or on change of state",
            "Name": "Basic Joystick Message 6:8",
            "PGNLength": "Every 100 ms and on change of state",
            "Rate": " but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July",
            "SPNs": [
                4565,
                4566,
                4567,
                4569,
                4570,
                4571,
                4573,
                4574,
                4575,
                4576,
                4577,
                4578,
                4579,
                4580,
                4581,
                4582,
                4583,
                4584,
                4585,
                4586,
                4568,
                4572
            ]
        },
        64813: {
            "Label": " 2010: Every 100 ms or on change of state",
            "Name": "Extended Joystick Message 5:8",
            "PGNLength": "Every 100 ms and on change of state",
            "Rate": " but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July",
            "SPNs": [
                4550,
                4551,
                4552,
                4554,
                4555,
                4556,
                4558,
                4559,
                4560,
                4562,
                4563,
                4564,
                4553,
                4557,
                4561
            ]
        },
        64814: {
            "Label": " 2010: Every 100 ms or on change of state",
            "Name": "Basic Joystick Message 5:8",
            "PGNLength": "Every 100 ms and on change of state",
            "Rate": " but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July",
            "SPNs": [
                4528,
                4529,
                4530,
                4532,
                4533,
                4534,
                4536,
                4537,
                4538,
                4539,
                4540,
                4541,
                4542,
                4543,
                4544,
                4545,
                4546,
                4547,
                4548,
                4549,
                4531,
                4535
            ]
        },
        64815: {
            "Label": " 2010: Every 100 ms or on change of state",
            "Name": "Extended Joystick Message 4:8",
            "PGNLength": "Every 100 ms and on change of state",
            "Rate": " but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July",
            "SPNs": [
                4513,
                4514,
                4515,
                4517,
                4518,
                4519,
                4521,
                4522,
                4523,
                4525,
                4526,
                4527,
                4516,
                4520,
                4524
            ]
        },
        64816: {
            "Label": " 2010: Every 100 ms or on change of state",
            "Name": "Basic Joystick Message 4:8",
            "PGNLength": "Every 100 ms and on change of state",
            "Rate": " but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July",
            "SPNs": [
                4491,
                4492,
                4493,
                4495,
                4496,
                4497,
                4499,
                4500,
                4501,
                4502,
                4503,
                4504,
                4505,
                4506,
                4507,
                4508,
                4509,
                4510,
                4511,
                4512,
                4494,
                4498
            ]
        },
        64817: {
            "Label": "FD2",
            "Name": "Fan Drive #2",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                1550,
                1557,
                1708,
                1598,
                1707
            ]
        },
        64818: {
            "Label": "TBD",
            "Name": "Reserved for ISOBUS ECU diagnostic protocol",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": []
        },
        64819: {
            "Label": "A2DEFSI",
            "Name": "Aftertreatment 2 Diesel Exhaust Fluid Supply Information",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                4441,
                4442,
                5437,
                5438,
                5439,
                4440
            ]
        },
        64820: {
            "Label": "A2SCRRT2I",
            "Name": "Aftertreatment 2 SCR Reagant Tank 2 Information",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                4433,
                4434,
                4436,
                4437,
                4438,
                4439,
                4435
            ]
        },
        64821: {
            "Label": "A2SCRRT1I",
            "Name": "Aftertreatment 2 SCR Reagant Tank 1 Information",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                4426,
                4427,
                4429,
                4430,
                4431,
                4432,
                4428
            ]
        },
        64822: {
            "Label": "A2DEFI",
            "Name": "Aftertreatment 2 Diesel Exhaust Fluid Information",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                4420,
                4421,
                4422,
                4423,
                4424,
                4425
            ]
        },
        64823: {
            "Label": "A2SCRSI",
            "Name": "Aftertreatment 2 SCR Service Information",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                4419,
                4417,
                4418
            ]
        },
        64824: {
            "Label": "A2SCREGT",
            "Name": "Aftertreatment 2 SCR Exhaust Gas Temperature",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                4414,
                4416,
                4413,
                4415
            ]
        },
        64825: {
            "Label": "A2SCREGP",
            "Name": "Aftertreatment 2 SCR Exhaust Gas Pressures",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                4412,
                4411
            ]
        },
        64826: {
            "Label": "A2SCRDSR2",
            "Name": "Aftertreatment 2 SCR Dosing System Requests 2",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                4405,
                4406,
                4407,
                4408,
                4409,
                4410,
                5138,
                5415
            ]
        },
        64827: {
            "Label": "A2SCRDSI2",
            "Name": "Aftertreatment 2 SCR Dosing System Information 2",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                4388,
                4389,
                4390,
                4391,
                4392,
                4393,
                4394,
                4395,
                4396,
                4397,
                4398,
                4399,
                4400
            ]
        },
        64828: {
            "Label": "A1DEFSI",
            "Name": "Aftertreatment 1 Diesel Exhaust Fluid Supply Information",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                4375,
                4376,
                5434,
                5435,
                5436,
                4374
            ]
        },
        64829: {
            "Label": "A1SCRRT2I",
            "Name": "Aftertreatment 1 SCR Reagant Tank 2 Information",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                4367,
                4368,
                4370,
                4371,
                4372,
                4373,
                4369
            ]
        },
        64830: {
            "Label": "A1SCREGT",
            "Name": "Aftertreatment 1 SCR Exhaust Gas Temperature",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                4361,
                4362,
                4360,
                4363
            ]
        },
        64831: {
            "Label": "A1SCREGP",
            "Name": "Aftertreatment 1 SCR Exhaust Gas Pressures",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                4359,
                4358
            ]
        },
        64832: {
            "Label": "A1SCRDSR2",
            "Name": "Aftertreatment 1 SCR Dosing System Requests 2",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                4352,
                4353,
                4354,
                4355,
                4356,
                4357,
                5137,
                5414,
                5707
            ]
        },
        64833: {
            "Label": "A1SCRDSI2",
            "Name": "Aftertreatment 1 SCR Dosing System Information 2",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                4335,
                4336,
                4337,
                4338,
                4339,
                4340,
                4341,
                4342,
                4343,
                4344,
                4345,
                4346,
                4347
            ]
        },
        64834: {
            "Label": "ICC",
            "Name": "ISOBUS compliance certification",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                4313,
                4315,
                4324,
                4323,
                4322,
                4321,
                4320,
                4319,
                4318,
                4317,
                4329,
                4328,
                4327,
                4326,
                4325,
                4330,
                4314,
                4316
            ]
        },
        64835: {
            "Label": "MSSC",
            "Name": "Machine Selected Speed Command",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                5818,
                4312,
                4310,
                4311
            ]
        },
        64836: {
            "Label": "AT2FC2",
            "Name": "Aftertreatment 2 Fuel Control 2",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                5426,
                5427,
                4303,
                5428
            ]
        },
        64837: {
            "Label": "AT2TWCC",
            "Name": "Aftertreatment 2 Three Way Catalytic Converter",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                4298,
                4300,
                4295,
                4296,
                4297,
                4299
            ]
        },
        64838: {
            "Label": "AT1TWCC",
            "Name": "Aftertreatment 1 Three Way Catalytic Converter",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                4292,
                4294,
                4289,
                4290,
                4291,
                4293
            ]
        },
        64839: {
            "Label": "TML",
            "Name": "Transmission Mode Labels",
            "PGNLength": "Variable",
            "Rate": "On request",
            "SPNs": [
                4254
            ]
        },
        64840: {
            "Label": "O2FT2",
            "Name": "Engine Exhaust Bank 2 O2 Fuel Trim",
            "PGNLength": 8,
            "Rate": "On Request",
            "SPNs": [
                4241,
                4239,
                4238
            ]
        },
        64841: {
            "Label": "O2FT1",
            "Name": "Engine Exhaust Bank 1 O2 Fuel Trim",
            "PGNLength": 8,
            "Rate": "On Request",
            "SPNs": [
                4240,
                4237,
                4236
            ]
        },
        64842: {
            "Label": "GPM211",
            "Name": "Reserved for ISO 11992",
            "PGNLength": 8,
            "Rate": "5 sec",
            "SPNs": []
        },
        64843: {
            "Label": "GPM210",
            "Name": "Reserved for ISO 11992",
            "PGNLength": 8,
            "Rate": "10 sec",
            "SPNs": []
        },
        64844: {
            "Label": "GPM29",
            "Name": "Reserved for ISO 11992",
            "PGNLength": 8,
            "Rate": "10 sec",
            "SPNs": []
        },
        64845: {
            "Label": "GPM28",
            "Name": "Reserved for ISO 11992",
            "PGNLength": 8,
            "Rate": "5 sec",
            "SPNs": []
        },
        64846: {
            "Label": "GPM27",
            "Name": "Reserved for ISO 11992",
            "PGNLength": 8,
            "Rate": "5 sec",
            "SPNs": []
        },
        64847: {
            "Label": "GPM26",
            "Name": "Reserved for ISO 11992",
            "PGNLength": 8,
            "Rate": "5 sec",
            "SPNs": []
        },
        64848: {
            "Label": "GPM19",
            "Name": "Reserved for ISO 11992",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": []
        },
        64849: {
            "Label": "ACCVC",
            "Name": "Aftercooler Coolant Control Valve Command",
            "PGNLength": 8,
            "Rate": "1 sec",
            "SPNs": [
                4198,
                4199,
                4200
            ]
        },
        64850: {
            "Label": "ECCVC",
            "Name": "Engine Coolant Control Valve Command",
            "PGNLength": 8,
            "Rate": "1 sec",
            "SPNs": [
                4195,
                4196,
                4197,
                5546,
                5547
            ]
        },
        64851: {
            "Label": "EAI",
            "Name": "Engine Average Information",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                4151,
                4153,
                4152
            ]
        },
        64852: {
            "Label": "DM52",
            "Name": "DTCs - C: Previously Active",
            "PGNLength": "Variable",
            "Rate": "On request",
            "SPNs": [
                987,
                624,
                623,
                1213,
                3041,
                3040,
                3039,
                3038,
                1215,
                1216,
                1706,
                1214
            ]
        },
        64853: {
            "Label": "DM51",
            "Name": "DTCs - C: Confirmed and Active",
            "PGNLength": "Variable",
            "Rate": "On request",
            "SPNs": [
                987,
                624,
                623,
                1213,
                3041,
                3040,
                3039,
                3038,
                1215,
                1216,
                1706,
                1214
            ]
        },
        64854: {
            "Label": "DM50",
            "Name": "DTCs - C: Pending",
            "PGNLength": "Variable",
            "Rate": "On request",
            "SPNs": [
                987,
                624,
                623,
                1213,
                3041,
                3040,
                3039,
                3038,
                1215,
                1216,
                1706,
                1214
            ]
        },
        64855: {
            "Label": "DM49",
            "Name": "DTCs - B2: Previously Active",
            "PGNLength": "Variable",
            "Rate": "On request",
            "SPNs": [
                987,
                624,
                623,
                1213,
                3041,
                3040,
                3039,
                3038,
                1215,
                1216,
                1706,
                1214
            ]
        },
        64856: {
            "Label": "DM48",
            "Name": "DTCs - B2: Confirmed and Active",
            "PGNLength": "Variable",
            "Rate": "On request",
            "SPNs": [
                987,
                624,
                623,
                1213,
                3041,
                3040,
                3039,
                3038,
                1215,
                1216,
                1706,
                1214
            ]
        },
        64857: {
            "Label": "DM47",
            "Name": "DTCs - B2: Pending",
            "PGNLength": "Variable",
            "Rate": "On request",
            "SPNs": [
                987,
                624,
                623,
                1213,
                3041,
                3040,
                3039,
                3038,
                1215,
                1216,
                1706,
                1214
            ]
        },
        64858: {
            "Label": "DM46",
            "Name": "DTCs - B1: Previously Active",
            "PGNLength": "Variable",
            "Rate": "On request",
            "SPNs": [
                987,
                624,
                623,
                1213,
                3041,
                3040,
                3039,
                3038,
                1215,
                1216,
                1706,
                1214
            ]
        },
        64859: {
            "Label": "DM45",
            "Name": "DTCs - B1: Confirmed and Active",
            "PGNLength": "Variable",
            "Rate": "On request",
            "SPNs": [
                987,
                624,
                623,
                1213,
                3041,
                3040,
                3039,
                3038,
                1215,
                1216,
                1706,
                1214
            ]
        },
        64860: {
            "Label": "DM44",
            "Name": "DTCs - B1: Pending",
            "PGNLength": "Variable",
            "Rate": "On request",
            "SPNs": [
                987,
                624,
                623,
                1213,
                3041,
                3040,
                3039,
                3038,
                1215,
                1216,
                1706,
                1214
            ]
        },
        64861: {
            "Label": "DM43",
            "Name": "DTCs - A: Previously Active",
            "PGNLength": "Variable",
            "Rate": "On request",
            "SPNs": [
                987,
                624,
                623,
                1213,
                3041,
                3040,
                3039,
                3038,
                1215,
                1216,
                1706,
                1214
            ]
        },
        64862: {
            "Label": "DM42",
            "Name": "DTCs - A: Confirmed and Active",
            "PGNLength": "Variable",
            "Rate": "On request",
            "SPNs": [
                987,
                624,
                623,
                1213,
                3041,
                3040,
                3039,
                3038,
                1215,
                1216,
                1706,
                1214
            ]
        },
        64863: {
            "Label": "DM41",
            "Name": "DTCs - A: Pending",
            "PGNLength": "Variable",
            "Rate": "On request",
            "SPNs": [
                987,
                624,
                623,
                1213,
                3041,
                3040,
                3039,
                3038,
                1215,
                1216,
                1706,
                1214
            ]
        },
        64864: {
            "Label": "DM40",
            "Name": "Harmonized B1 Failure Counts (HB1C)",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        64865: {
            "Label": "DM39",
            "Name": "Cumulative Continuous MI \u2013 System (HCMI)",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        64866: {
            "Label": "DM38",
            "Name": "Harmonized Global Regulation Description (HGRD)",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        64867: {
            "Label": "DM37",
            "Name": "Harmonized Roadworthiness \u2013 System (HRWS)",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        64868: {
            "Label": "DM36",
            "Name": "Harmonized Roadworthiness - Vehicle (HRWV)",
            "PGNLength": "",
            "Rate": "",
            "SPNs": [
                4138
            ]
        },
        64869: {
            "Label": "AT1FC2",
            "Name": "Aftertreatment 1 Fuel Control 2",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                5423,
                5424,
                5456,
                4077,
                5425
            ]
        },
        64870: {
            "Label": "ET4",
            "Name": "Engine Temperature 4",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                4076,
                4193,
                4194,
                4288,
                5020
            ]
        },
        64871: {
            "Label": "ZNVW",
            "Name": "Zero Net Vehicle Weight Change",
            "PGNLength": 8,
            "Rate": "As needed",
            "SPNs": [
                4075
            ]
        },
        64872: {
            "Label": "GCVW",
            "Name": "Gross Combination Vehicle Weight",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                417,
                413
            ]
        },
        64873: {
            "Label": "AGCW",
            "Name": "Axle Group Calibration Weights:8",
            "PGNLength": "On request.  Upon request",
            "Rate": " will be broadcast as many times as required to transmit all available axle groups.",
            "SPNs": [
                4074,
                408,
                407
            ]
        },
        64874: {
            "Label": "AGW",
            "Name": "Axle Group Weight:8",
            "PGNLength": "On request.  Upon request",
            "Rate": " will be broadcast as many times as required to transmit all available axle groups.",
            "SPNs": [
                4073,
                409
            ]
        },
        64875: {
            "Label": "AAGW",
            "Name": "Available Axle Group Weights",
            "PGNLength": 8,
            "Rate": "As needed.  Broadcast whenever an axle group equipped with an on-board scale joined or left the on-board scale subset.",
            "SPNs": [
                4059,
                4060,
                4061,
                4062,
                4063,
                4064,
                4065,
                4066,
                4067,
                4068,
                4069,
                4070,
                4071,
                4072
            ]
        },
        64876: {
            "Label": "AT2AC2",
            "Name": "Aftertreatment 2 Air Control 2",
            "PGNLength": 8,
            "Rate": "500 msec",
            "SPNs": [
                3833,
                3834,
                3835,
                3838
            ]
        },
        64877: {
            "Label": "AT1AC2",
            "Name": "Aftertreatment 1 Air Control 2",
            "PGNLength": 8,
            "Rate": "500 msec",
            "SPNs": [
                3830,
                3831,
                3832,
                3837
            ]
        },
        64878: {
            "Label": "SCR1",
            "Name": "Aftertreatment 1 SCR Service Information",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                4364,
                3826,
                3828,
                5463
            ]
        },
        64879: {
            "Label": "EEC8",
            "Name": "Electronic Engine Controller 8",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                4752,
                3821,
                4750,
                4751
            ]
        },
        64880: {
            "Label": "DRC",
            "Name": "Door ramp control:8",
            "PGNLength": "Every 1 s while active and on change of state but no faster than every 100 ms.  Grandfathered definition for systems that implemented this message prior to July",
            "Rate": " 2010:  1 s when active and on change of state",
            "SPNs": [
                3810,
                3811,
                3812,
                3813,
                3814,
                3815,
                3816,
                3817,
                3818,
                4454,
                4455,
                4456,
                4457,
                4458,
                4459
            ]
        },
        64881: {
            "Label": "BSA",
            "Name": "Brake actuator stroke status",
            "PGNLength": 8,
            "Rate": "1 sec",
            "SPNs": [
                3785,
                3786,
                3788,
                3789,
                3791,
                3792,
                3793,
                3794,
                3796,
                3797,
                3799,
                3800,
                3801,
                3802,
                3804,
                3787,
                3790,
                3795,
                3798,
                3803
            ]
        },
        64882: {
            "Label": "ESV6",
            "Name": "Engine Spark Voltage 6",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                1314,
                1315,
                1316,
                1317
            ]
        },
        64883: {
            "Label": "ESV5",
            "Name": "Engine Spark Voltage 5",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                1310,
                1311,
                1312,
                1313
            ]
        },
        64884: {
            "Label": "ESV4",
            "Name": "Engine Spark Voltage 4",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                1306,
                1307,
                1308,
                1309
            ]
        },
        64885: {
            "Label": "ESV3",
            "Name": "Engine Spark Voltage 3",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                1302,
                1303,
                1304,
                1305
            ]
        },
        64886: {
            "Label": "ESV2",
            "Name": "Engine Spark Voltage 2",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                1298,
                1299,
                1300,
                1301
            ]
        },
        64887: {
            "Label": "ESV1",
            "Name": "Engine Spark Voltage 1",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                1294,
                1295,
                1296,
                1297
            ]
        },
        64888: {
            "Label": "AT2TI",
            "Name": "Aftertreatment 2 Trip Information",
            "PGNLength": 32,
            "Rate": "On request",
            "SPNs": [
                3741,
                3742,
                3743,
                3744,
                3745,
                3746,
                3747,
                3748
            ]
        },
        64889: {
            "Label": "AT1TI",
            "Name": "Aftertreatment 1 Trip Information",
            "PGNLength": 32,
            "Rate": "On request",
            "SPNs": [
                3733,
                3734,
                3735,
                3736,
                3737,
                3738,
                3739,
                3740
            ]
        },
        64890: {
            "Label": "AT2S",
            "Name": "Aftertreatment 2 Service",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                3722,
                3723,
                5467,
                3724
            ]
        },
        64891: {
            "Label": "AT1S",
            "Name": "Aftertreatment 1 Service",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                3719,
                3720,
                5466,
                3721
            ]
        },
        64892: {
            "Label": "DPFC1",
            "Name": "DPF Control 1:8",
            "PGNLength": "Every 1 s and on change of state but no faster than every 100 ms.  Grandfathered definition for systems that implemented this message prior to July",
            "Rate": " 2010:  1 s and on change",
            "SPNs": [
                3697,
                3699,
                3700,
                3701,
                3702,
                3703,
                3704,
                3705,
                3706,
                3707,
                3708,
                3709,
                3710,
                3711,
                3712,
                3713,
                3714,
                3715,
                3716,
                3717,
                3718,
                3698,
                4175,
                5504,
                5629,
                3750
            ]
        },
        64893: {
            "Label": "FMS1",
            "Name": "Reserved for FMS Telltale status",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        64894: {
            "Label": "AFSS",
            "Name": "Adaptive Front-Lighting System Status",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                3691,
                3692,
                3693,
                3694
            ]
        },
        64895: {
            "Label": "EC2",
            "Name": "Engine Configuration 2",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                3670
            ]
        },
        64896: {
            "Label": "DM28",
            "Name": "Permanent DTCs",
            "PGNLength": "Variable",
            "Rate": "On request",
            "SPNs": [
                987,
                624,
                623,
                1213,
                3041,
                3040,
                3039,
                3038,
                1215,
                1216,
                1706,
                1214
            ]
        },
        64897: {
            "Label": "EGRBV",
            "Name": "EGR Cooler Bypass",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                3672
            ]
        },
        64898: {
            "Label": "DM27",
            "Name": "All Pending DTCs",
            "PGNLength": "Variable",
            "Rate": "On request",
            "SPNs": [
                987,
                624,
                623,
                1213,
                3041,
                3040,
                3039,
                3038,
                1215,
                1216,
                1706,
                1214
            ]
        },
        64899: {
            "Label": "TCI",
            "Name": "Transfer Case Information:8",
            "PGNLength": "Every 1 s and on change of state but no faster than every 100 ms.  Grandfathered definition for systems that implemented this message prior to July",
            "Rate": " 2010:  1 s or on change",
            "SPNs": [
                3645
            ]
        },
        64900: {
            "Label": "EFL/P9",
            "Name": "Engine Fluid Level/Pressure 9",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                3640,
                3641,
                3642,
                3643
            ]
        },
        64901: {
            "Label": "EFL/P8",
            "Name": "Engine Fluid Level/Pressure 8",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                3636,
                3637,
                3638,
                3639
            ]
        },
        64902: {
            "Label": "EFL/P7",
            "Name": "Engine Fluid Level/Pressure 7",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                3632,
                3633,
                3634,
                3635
            ]
        },
        64903: {
            "Label": "EFL/P6",
            "Name": "Engine Fluid Level/Pressure 6",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                3628,
                3629,
                3630,
                3631
            ]
        },
        64904: {
            "Label": "EFL/P5",
            "Name": "Engine Fluid Level/Pressure 5",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                3624,
                3625,
                3626,
                3627
            ]
        },
        64905: {
            "Label": "VDS2",
            "Name": "Vehicle Direction/Speed 2",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                3623
            ]
        },
        64906: {
            "Label": "J2012",
            "Name": "SAE J2012 DTC Display",
            "PGNLength": "Variable",
            "Rate": "On request",
            "SPNs": [
                3619,
                3621,
                3622,
                3620
            ]
        },
        64907: {
            "Label": "AT2GP",
            "Name": "Aftertreatment 2 Gas Parameters",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                3611,
                3612
            ]
        },
        64908: {
            "Label": "AT1GP",
            "Name": "Aftertreatment 1 Gas Parameters",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                3609,
                3610
            ]
        },
        64909: {
            "Label": "UTACER",
            "Name": "Utility Total AC Reactive Energy",
            "PGNLength": 8,
            "Rate": "250 ms",
            "SPNs": [
                3595,
                3596
            ]
        },
        64910: {
            "Label": "GTACER",
            "Name": "Generator Total AC Reactive Energy",
            "PGNLength": 8,
            "Rate": "250 ms",
            "SPNs": [
                3593,
                3594
            ]
        },
        64911: {
            "Label": "GTACPP",
            "Name": "Generator Total AC Percent Power",
            "PGNLength": 8,
            "Rate": "250 ms",
            "SPNs": [
                3590,
                3591,
                3592
            ]
        },
        64912: {
            "Label": "AETC",
            "Name": "Advertised Engine Torque Curve",
            "PGNLength": "Variable",
            "Rate": "On request",
            "SPNs": [
                3558,
                3559,
                3560,
                3561
            ]
        },
        64913: {
            "Label": "ACS",
            "Name": "AC Switching Device Status",
            "PGNLength": 8,
            "Rate": "250 ms",
            "SPNs": [
                3545,
                3546,
                3547
            ]
        },
        64914: {
            "Label": "EOI",
            "Name": "Engine Operating Information",
            "PGNLength": 8,
            "Rate": "250 ms",
            "SPNs": [
                3543,
                4082,
                3608,
                632,
                2807,
                3601,
                3589,
                3602,
                3603,
                3604,
                3605,
                3606,
                3607,
                3644,
                3544
            ]
        },
        64915: {
            "Label": "GC1",
            "Name": "Generator Control 1",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                3542,
                3567,
                3568,
                4079,
                4080,
                5568,
                4078,
                5567
            ]
        },
        64916: {
            "Label": "EEC7",
            "Name": "Electronic Engine Controller 7",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                27,
                3822,
                5444,
                5312
            ]
        },
        64917: {
            "Label": "TRF2",
            "Name": "Transmission Fluids 2",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                3359,
                3533,
                5345,
                4177,
                5914,
                5915,
                5916,
                3823,
                5913
            ]
        },
        64920: {
            "Label": "AT1HI",
            "Name": "Aftertreatment 1 Historical Information",
            "PGNLength": "Variable",
            "Rate": "On request",
            "SPNs": [
                3522,
                3523,
                3524,
                3525,
                3725,
                3726,
                3727,
                3728,
                5454,
                5827
            ]
        },
        64921: {
            "Label": "AT2HI",
            "Name": "Aftertreatment 2 Historical information",
            "PGNLength": "Variable",
            "Rate": "On request",
            "SPNs": [
                3526,
                3527,
                3528,
                3529,
                3729,
                3730,
                3731,
                3732,
                5455,
                5828
            ]
        },
        64922: {
            "Label": "EBS24",
            "Name": "Electronic Brake System #2/4",
            "PGNLength": 8,
            "Rate": "1 sec",
            "SPNs": []
        },
        64923: {
            "Label": "A1DEFI",
            "Name": "Aftertreatment 1 Diesel Exhaust Fluid Information",
            "PGNLength": 8,
            "Rate": "1 sec",
            "SPNs": [
                3515,
                3516,
                3518,
                3519,
                3520,
                3521
            ]
        },
        64924: {
            "Label": "SEP2",
            "Name": "Sensor Electrical Power #2",
            "PGNLength": 8,
            "Rate": "1 sec",
            "SPNs": [
                3513,
                3514,
                5125,
                5126
            ]
        },
        64925: {
            "Label": "SEP1",
            "Name": "Sensor Electrical Power #1",
            "PGNLength": 8,
            "Rate": "1 sec",
            "SPNs": [
                3509,
                3510,
                3511,
                3512
            ]
        },
        64926: {
            "Label": "AT2AC1",
            "Name": "Aftertreatment 2 Air Control 1",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                3502,
                3506,
                3505,
                3504,
                3503,
                3499,
                3500,
                3501
            ]
        },
        64927: {
            "Label": "AT1AC1",
            "Name": "Aftertreatment 1 Air Control 1",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                3488,
                3492,
                3491,
                3490,
                3489,
                3485,
                3486,
                3487
            ]
        },
        64928: {
            "Label": "AT2FC1",
            "Name": "Aftertreatment 2 Fuel Control 1",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                4098,
                3498,
                3497,
                3496,
                4302,
                3494,
                3495,
                3493
            ]
        },
        64929: {
            "Label": "AT1FC1",
            "Name": "Aftertreatment 1 Fuel Control 1",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                4097,
                3484,
                3483,
                3482,
                4301,
                3480,
                3481,
                3479
            ]
        },
        64930: {
            "Label": "GFI",
            "Name": "Fuel Information 3 (Gaseous)",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                3468,
                3466,
                3467,
                3469
            ]
        },
        64931: {
            "Label": "EEC6",
            "Name": "Electronic Engine Controller 6",
            "PGNLength": 8,
            "Rate": "100 msec (preferred) or Engine Speed Dependent (if required by application)",
            "SPNs": [
                641,
                3675,
                5366,
                5367,
                5368,
                3470,
                5369
            ]
        },
        64932: {
            "Label": "PTODE",
            "Name": "PTO Drive Engagement",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                3455,
                3454,
                3453,
                3452,
                3939,
                3942,
                3945,
                3459,
                3458,
                3457,
                3456,
                3940,
                3943,
                3946,
                3463,
                3462,
                3461,
                3460,
                3941,
                3944,
                3947,
                3948
            ]
        },
        64933: {
            "Label": "DC2",
            "Name": "Door Control 2",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                3412,
                3413,
                3414,
                3415,
                3416,
                3417,
                3418,
                3419,
                3420,
                3421,
                3422,
                3423,
                3424,
                3425,
                3426,
                3427,
                3428,
                3429,
                3430,
                3431,
                3432,
                3433,
                3434,
                3435,
                3436,
                3437,
                3438,
                3439,
                3440,
                3441
            ]
        },
        64934: {
            "Label": "VREP",
            "Name": "Voltage Regulator Excitation Status",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                3380,
                3381,
                3382
            ]
        },
        64935: {
            "Label": "VROM",
            "Name": "Voltage Regulator Operating Mode",
            "PGNLength": 8,
            "Rate": "1 sec",
            "SPNs": [
                3375,
                3376,
                3377,
                3378,
                3379
            ]
        },
        64936: {
            "Label": "WCM2",
            "Name": "Wireless Communications Message 2",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                3442,
                3443,
                3444,
                3445,
                3446
            ]
        },
        64937: {
            "Label": "WCM1",
            "Name": "Wireless Communications Message 1",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                3368,
                3369,
                3370,
                3371,
                3372
            ]
        },
        64938: {
            "Label": "EFL/P4",
            "Name": "Engine Fluid Level/Pressure 4",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                3340,
                3341,
                3342,
                3668,
                3676,
                2631,
                3343
            ]
        },
        64939: {
            "Label": "RCMS",
            "Name": "Request For Complete Configurable Message Set (no longer used)",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": []
        },
        64941: {
            "Label": "RCMS",
            "Name": "Request For Complete Configurable Message Set",
            "PGNLength": 8,
            "Rate": "As Needed",
            "SPNs": [
                3329,
                3330
            ]
        },
        64942: {
            "Label": "FWSS2",
            "Name": "Fifth Wheel Smart Systems 2",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                3307,
                3312,
                3313,
                3311,
                3316
            ]
        },
        64943: {
            "Label": "AT2IMG",
            "Name": "Aftertreatment 2 Intermediate Gas",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                3286,
                3288,
                3283,
                3284,
                3285,
                3287
            ]
        },
        64944: {
            "Label": "AT2OG2",
            "Name": "Aftertreatment 2 Outlet Gas 2",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                3281,
                3282,
                3279,
                3280
            ]
        },
        64945: {
            "Label": "AT2IG2",
            "Name": "Aftertreatment 2 Intake Gas 2",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                3277,
                3278,
                3275,
                3276
            ]
        },
        64946: {
            "Label": "AT1IMG",
            "Name": "Aftertreatment 1 Intermediate Gas",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                3252,
                3254,
                3249,
                3250,
                3251,
                3253
            ]
        },
        64947: {
            "Label": "AT1OG2",
            "Name": "Aftertreatment 1 Outlet Gas 2",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                3247,
                3248,
                3245,
                3246
            ]
        },
        64948: {
            "Label": "AT1IG2",
            "Name": "Aftertreatment 1 Intake Gas 2",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                3243,
                3244,
                3241,
                3242
            ]
        },
        64949: {
            "Label": "DM23",
            "Name": "Previously Active Emission Related Faults",
            "PGNLength": "Variable",
            "Rate": "On request",
            "SPNs": [
                987,
                624,
                623,
                1213,
                3041,
                3040,
                3039,
                3038,
                1215,
                1216,
                1706,
                1214
            ]
        },
        64950: {
            "Label": "DM24",
            "Name": "SPN Support",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                4100,
                4101,
                4102,
                4103,
                3297
            ]
        },
        64951: {
            "Label": "DM25",
            "Name": "Expanded Freeze Frame",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                3300,
                1215,
                1216,
                1706,
                1214,
                1545
            ]
        },
        64952: {
            "Label": "DM26",
            "Name": "Diagnostic Readiness 3",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                3302,
                3303,
                3301,
                3304,
                3305
            ]
        },
        64953: {
            "Label": "TPRI",
            "Name": "Tire Pressure Reference Information",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                3190,
                3191
            ]
        },
        64954: {
            "Label": "TR6",
            "Name": "Farebox Status",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                3179,
                3181
            ]
        },
        64955: {
            "Label": "TR5",
            "Name": "Farebox Point of Sale",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                3170,
                3171,
                3176,
                3177,
                3165,
                3166,
                3167,
                3172,
                3173,
                3168,
                3169
            ]
        },
        64956: {
            "Label": "TR4",
            "Name": "Farebox Service Detail",
            "PGNLength": 15,
            "Rate": "On request",
            "SPNs": [
                3178,
                3180,
                3174,
                3175,
                3159,
                3161,
                3160,
                3162,
                3163,
                3164
            ]
        },
        64957: {
            "Label": "TR3",
            "Name": "Signal Preemption",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                3081,
                3080,
                3079,
                3084,
                3083,
                3082,
                3085
            ]
        },
        64958: {
            "Label": "TR1",
            "Name": "Transit Route",
            "PGNLength": "Variable",
            "Rate": "On request",
            "SPNs": [
                3078,
                3071,
                3072,
                3073,
                3074,
                3075,
                3076
            ]
        },
        64959: {
            "Label": "TR2",
            "Name": "Transit Milepost",
            "PGNLength": "Variable",
            "Rate": "On request",
            "SPNs": [
                3070,
                509
            ]
        },
        64960: {
            "Label": "TR7",
            "Name": "Passenger Counter",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                3043,
                3047,
                3044,
                3045,
                3046
            ]
        },
        64961: {
            "Label": "EFL/P3",
            "Name": "Engine Fluid Level/Pressure 3",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                3358,
                5019,
                2948,
                4287,
                5631
            ]
        },
        64962: {
            "Label": "EEC14",
            "Name": "Electronic Engine Controller 14",
            "PGNLength": 8,
            "Rate": "On Request",
            "SPNs": [
                5837,
                5829,
                5830,
                5833
            ]
        },
        64963: {
            "Label": "EEC15",
            "Name": "Electronic Engine Controller 15",
            "PGNLength": 8,
            "Rate": "On Request",
            "SPNs": [
                2623,
                2624,
                2625,
                2626
            ]
        },
        64964: {
            "Label": "EBC5",
            "Name": "Electronic Brake Controller 5",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                3839,
                2913,
                2912,
                2919,
                2917,
                2918,
                2921,
                5275
            ]
        },
        64965: {
            "Label": "ECUID",
            "Name": "ECU Identification Information",
            "PGNLength": "Variable",
            "Rate": "On request",
            "SPNs": [
                2901,
                2902,
                2903,
                2904,
                4304
            ]
        },
        64966: {
            "Label": "CSA",
            "Name": "Cold Start Aids",
            "PGNLength": 8,
            "Rate": "As required",
            "SPNs": [
                626,
                1804,
                2899,
                2898,
                5548,
                5549,
                5550
            ]
        },
        64967: {
            "Label": "OHCSS",
            "Name": "Off-Highway Engine Control Selection States",
            "PGNLength": 8,
            "Rate": "0.5 s",
            "SPNs": [
                2896,
                2890,
                2891,
                2888,
                2889,
                2893,
                2894,
                2895
            ]
        },
        64968: {
            "Label": "ISCS",
            "Name": "Operator Primary Intermediate Speed Control state",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                2892
            ]
        },
        64969: {
            "Label": "CMI",
            "Name": "Electronic Control Module Information",
            "PGNLength": 8,
            "Rate": "As required",
            "SPNs": [
                2887
            ]
        },
        64970: {
            "Label": "ISC",
            "Name": "Intermediate Speed Control",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                2880
            ]
        },
        64971: {
            "Label": "OHECS",
            "Name": "Off-Highway Engine Control Selection",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                2884,
                1377,
                2883,
                2882,
                2881,
                2879,
                2886,
                2885
            ]
        },
        64972: {
            "Label": "OEL",
            "Name": "Operators External Light Controls Message",
            "PGNLength": 8,
            "Rate": "Every 1 s and on change of switch state but no faster than every 100 ms",
            "SPNs": [
                2873,
                2872,
                2876,
                2875,
                2874,
                2878,
                2877
            ]
        },
        64973: {
            "Label": "OWW",
            "Name": "Operator Wiper and Washer Controls Message",
            "PGNLength": 8,
            "Rate": "200 ms",
            "SPNs": [
                2864,
                2863,
                2865,
                2869,
                2870,
                2871,
                2867,
                2866,
                2868
            ]
        },
        64974: {
            "Label": "WSMM",
            "Name": "Working Set Member Message",
            "PGNLength": 8,
            "Rate": "As required",
            "SPNs": [
                2845
            ]
        },
        64975: {
            "Label": "WSM",
            "Name": "Working Set Master Message",
            "PGNLength": 8,
            "Rate": "As required",
            "SPNs": [
                2409
            ]
        },
        64976: {
            "Label": "IC2",
            "Name": "Intake/Exhaust Conditions 2",
            "PGNLength": 8,
            "Rate": "0.5 s",
            "SPNs": [
                2809,
                2810,
                2811,
                3562,
                3563,
                5422,
                4817
            ]
        },
        64977: {
            "Label": "FMS",
            "Name": "FMS-standard Interface Identity/Capabilities",
            "PGNLength": 8,
            "Rate": "10 s",
            "SPNs": [
                2804,
                2805,
                2806
            ]
        },
        64978: {
            "Label": "EP",
            "Name": "ECU Performance",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                2802,
                2803
            ]
        },
        64979: {
            "Label": "TCI6",
            "Name": "Turbocharger Information 6",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                2629,
                2799,
                2800,
                2801
            ]
        },
        64980: {
            "Label": "CM3",
            "Name": "Cab Message 3:8",
            "PGNLength": "Every 10 s and on change of state but no faster than every 100 ms.  Grandfathered definition for systems that implemented this message prior to July",
            "Rate": " 2010: On change or every 10 s",
            "SPNs": [
                2796,
                3314,
                3315,
                3809
            ]
        },
        64981: {
            "Label": "EEC5",
            "Name": "Electronic Engine Controller 5",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                2792,
                5323,
                5457,
                2795,
                2789,
                2790,
                2791
            ]
        },
        64982: {
            "Label": "BJM1",
            "Name": "Basic Joystick Message 1",
            "PGNLength": 8,
            "Rate": "Every 100 ms and on change of state but no faster than every 20 ms.",
            "SPNs": [
                2675,
                2670,
                2665,
                2676,
                2671,
                2666,
                2681,
                2680,
                2688,
                2687,
                2686,
                2685,
                2692,
                2691,
                2690,
                2689,
                2696,
                2695,
                2694,
                2693,
                2660,
                2661
            ]
        },
        64983: {
            "Label": "EJM1",
            "Name": "Extended Joystick Message 1",
            "PGNLength": 8,
            "Rate": "Every 100 ms and on change of state but no faster than every 20 ms.",
            "SPNs": [
                2677,
                2672,
                2667,
                2678,
                2673,
                2668,
                2679,
                2674,
                2669,
                2684,
                2683,
                2682,
                2662,
                2663,
                2664
            ]
        },
        64984: {
            "Label": "BJM2",
            "Name": "Basic Joystick Message 2",
            "PGNLength": 8,
            "Rate": "Every 100 ms and on change of state but no faster than every 20 ms.",
            "SPNs": [
                2712,
                2707,
                2702,
                2713,
                2708,
                2703,
                2718,
                2717,
                2725,
                2724,
                2723,
                2722,
                2729,
                2728,
                2727,
                2726,
                2733,
                2732,
                2731,
                2730,
                2697,
                2698
            ]
        },
        64985: {
            "Label": "EJM2",
            "Name": "Extended Joystick Message 2",
            "PGNLength": 8,
            "Rate": "Every 100 ms and on change of state but no faster than every 20 ms.",
            "SPNs": [
                2714,
                2709,
                2704,
                2715,
                2710,
                2705,
                2716,
                2711,
                2706,
                2721,
                2720,
                2719,
                2699,
                2700,
                2701
            ]
        },
        64986: {
            "Label": "BJM3",
            "Name": "Basic Joystick Message 3",
            "PGNLength": 8,
            "Rate": "Every 100 ms and on change of state but no faster than every 20 ms.",
            "SPNs": [
                2749,
                2744,
                2739,
                2750,
                2745,
                2740,
                2755,
                2754,
                2762,
                2761,
                2760,
                2759,
                2766,
                2765,
                2764,
                2763,
                2770,
                2769,
                2768,
                2767,
                2734,
                2735
            ]
        },
        64987: {
            "Label": "EJM3",
            "Name": "Extended Joystick Message 3",
            "PGNLength": 8,
            "Rate": "Every 100 ms and on change of state but no faster than every 20 ms.",
            "SPNs": [
                2751,
                2746,
                2741,
                2752,
                2747,
                2742,
                2753,
                2748,
                2743,
                2758,
                2757,
                2756,
                2736,
                2737,
                2738
            ]
        },
        64988: {
            "Label": "MCI",
            "Name": "Marine Control Information",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                2615,
                2616,
                2617
            ]
        },
        64989: {
            "Label": "MAM11",
            "Name": "ISO 11992 Military Application Tractor - Trailer Message",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": []
        },
        64990: {
            "Label": "MAM21",
            "Name": "ISO 11992 Military Application Trailer - Tractor Feedback Message",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": []
        },
        64991: {
            "Label": "FWD",
            "Name": "Front Wheel Drive Status",
            "PGNLength": 8,
            "Rate": "0.5 s",
            "SPNs": [
                2612
            ]
        },
        64992: {
            "Label": "AMB2",
            "Name": "Ambient Conditions 2",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                2610,
                2611,
                4490,
                5581,
                5685
            ]
        },
        64993: {
            "Label": "CACI",
            "Name": "Cab A/C Climate System Information",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                2609
            ]
        },
        64994: {
            "Label": "SPR",
            "Name": "Supply Pressure Demand:8",
            "PGNLength": "1 s",
            "Rate": " when active",
            "SPNs": [
                2603,
                2604,
                2605,
                2606,
                2607,
                2608
            ]
        },
        64995: {
            "Label": "EOAC",
            "Name": "Equipment Operation and Control",
            "PGNLength": 8,
            "Rate": "250 mS",
            "SPNs": [
                2601
            ]
        },
        64996: {
            "Label": "EPD",
            "Name": "Equipment Performance Data",
            "PGNLength": 8,
            "Rate": "500 mS",
            "SPNs": [
                2600
            ]
        },
        64997: {
            "Label": "MVS",
            "Name": "Maximum Vehicle Speed Limit Status",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                2588,
                2589,
                2590,
                2591,
                2592,
                2593,
                2594,
                2595
            ]
        },
        64998: {
            "Label": "HBS",
            "Name": "Hydraulic Braking System",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                2580,
                2581,
                2584,
                2585,
                2582,
                2583,
                2930,
                2931
            ]
        },
        64999: {
            "Label": "BUSC",
            "Name": "Bus #1/Utility Sync Check Status",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                2535,
                2531,
                2533,
                2532,
                2534,
                2517
            ]
        },
        65000: {
            "Label": "BGSC",
            "Name": "Bus #1/Generator Sync Check Status",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                2530,
                2526,
                2528,
                2527,
                2529,
                2516
            ]
        },
        65001: {
            "Label": "BPCAC",
            "Name": "Bus #1 Phase C  Basic AC Quantities",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                2511,
                2515,
                2507
            ]
        },
        65002: {
            "Label": "BPBAC",
            "Name": "Bus #1 Phase B  Basic AC Quantities",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                2510,
                2514,
                2506
            ]
        },
        65003: {
            "Label": "BPAAC",
            "Name": "Bus #1 Phase A  Basic AC Quantities",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                2509,
                2513,
                2505
            ]
        },
        65004: {
            "Label": "BAAC",
            "Name": "Bus #1 Average Basic AC Quantities",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                2508,
                2512,
                2504
            ]
        },
        65005: {
            "Label": "UTACE",
            "Name": "Utility Total AC Energy",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                2502,
                2503
            ]
        },
        65006: {
            "Label": "UPCACR",
            "Name": "Utility Phase C AC Reactive Power",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                2525,
                2493,
                2501
            ]
        },
        65007: {
            "Label": "UPCACP",
            "Name": "Utility Phase C AC Power",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                2489,
                2497
            ]
        },
        65008: {
            "Label": "UPCAC",
            "Name": "Utility Phase C AC Basic Quantities",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                2477,
                2481,
                2473,
                2485
            ]
        },
        65009: {
            "Label": "UPBACR",
            "Name": "Utility Phase B AC Reactive Power",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                2524,
                2492,
                2500
            ]
        },
        65010: {
            "Label": "UPBACP",
            "Name": "Utility Phase B AC Power",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                2488,
                2496
            ]
        },
        65011: {
            "Label": "UPBAC",
            "Name": "Utility Phase B AC Basic Quantities",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                2476,
                2480,
                2472,
                2484
            ]
        },
        65012: {
            "Label": "UPACCR",
            "Name": "Utility Phase A  AC Reactive Power",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                2523,
                2491,
                2499
            ]
        },
        65013: {
            "Label": "UPAACP",
            "Name": "Utility Phase A  AC Power",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                2487,
                2495
            ]
        },
        65014: {
            "Label": "UPAAC",
            "Name": "Utility Phase A Basic AC Quantities",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                2475,
                2479,
                2471,
                2483
            ]
        },
        65015: {
            "Label": "UTACR",
            "Name": "Utility Total AC Reactive Power",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                2522,
                2490,
                2498
            ]
        },
        65016: {
            "Label": "UTACP",
            "Name": "Utility Total AC Power",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                2486,
                2494
            ]
        },
        65017: {
            "Label": "UAAC",
            "Name": "Utility Average Basic AC Quantities",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                2474,
                2478,
                2470,
                2482
            ]
        },
        65018: {
            "Label": "GTACE",
            "Name": "Generator Total AC Energy",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                2468,
                2469
            ]
        },
        65019: {
            "Label": "GPCACR",
            "Name": "Generator Phase C AC Reactive Power",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                2521,
                2459,
                2467
            ]
        },
        65020: {
            "Label": "GPCACP",
            "Name": "Generator Phase C AC Power",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                2455,
                2463
            ]
        },
        65021: {
            "Label": "GPCAC",
            "Name": "Generator Phase C Basic AC Quantities",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                2443,
                2447,
                2439,
                2451
            ]
        },
        65022: {
            "Label": "GPBACRP",
            "Name": "Generator Phase B AC Reactive Power",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                2520,
                2458,
                2466
            ]
        },
        65023: {
            "Label": "GPBACP",
            "Name": "Generator Phase B AC Power",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                2454,
                2462
            ]
        },
        65024: {
            "Label": "GPBAC",
            "Name": "Generator Phase B Basic AC Quantities",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                2442,
                2446,
                2438,
                2450
            ]
        },
        65025: {
            "Label": "GPAACR",
            "Name": "Generator Phase A AC Reactive Power",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                2519,
                2457,
                2465
            ]
        },
        65026: {
            "Label": "GPAACP",
            "Name": "Generator Phase A AC Power",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                2453,
                2461
            ]
        },
        65027: {
            "Label": "GPAAC",
            "Name": "Generator Phase A Basic AC Quantities",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                2441,
                2445,
                2437,
                2449
            ]
        },
        65028: {
            "Label": "GTACR",
            "Name": "Generator Total AC Reactive Power",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                2518,
                2456,
                2464
            ]
        },
        65029: {
            "Label": "GTACP",
            "Name": "Generator Total AC Power",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                2452,
                2460
            ]
        },
        65030: {
            "Label": "GAAC",
            "Name": "Generator Average Basic AC Quantities",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                2440,
                2444,
                2436,
                2448
            ]
        },
        65031: {
            "Label": "ET",
            "Name": "Exhaust Temperature",
            "PGNLength": 8,
            "Rate": "0.5 s",
            "SPNs": [
                2433,
                2434
            ]
        },
        65032: {
            "Label": "RTF",
            "Name": "Required Tractor Facilities message",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": []
        },
        65033: {
            "Label": "TFR",
            "Name": "Tractor Facilities response message",
            "PGNLength": 8,
            "Rate": "On power up and then on request",
            "SPNs": []
        },
        65034: {
            "Label": "IRCR",
            "Name": "Implement Remote Control Command Tractor Response",
            "PGNLength": 8,
            "Rate": "on every command",
            "SPNs": []
        },
        65035: {
            "Label": "IRC",
            "Name": "Implement Remote Control Command:8",
            "PGNLength": "default is 100 ms",
            "Rate": " maximum is 10 ms for doing closed loop cruise control.  For setting slip control functions a repetition rate of 1 in 10 seconds and on change of state is required.",
            "SPNs": []
        },
        65036: {
            "Label": "WSMEM",
            "Name": "Working Set Member",
            "PGNLength": 8,
            "Rate": "As Required",
            "SPNs": []
        },
        65037: {
            "Label": "WSMSTR",
            "Name": "Working Set Master",
            "PGNLength": 8,
            "Rate": "As Required",
            "SPNs": [
                2409
            ]
        },
        65038: {
            "Label": "RESRR",
            "Name": "Response for Repetition Rate",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                2421,
                2420,
                2419,
                2426,
                2425,
                2428,
                2427,
                2418
            ]
        },
        65039: {
            "Label": "LC",
            "Name": "Language Command",
            "PGNLength": 8,
            "Rate": "On system initialisation and on request",
            "SPNs": [
                2413,
                2411,
                2412,
                2417,
                2416,
                2415,
                2414,
                2410
            ]
        },
        65040: {
            "Label": "AV00EF",
            "Name": "Auxiliary Valve Number 0 Estimated Flow",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                1901,
                1902,
                1903,
                1910,
                5800
            ]
        },
        65041: {
            "Label": "AV01EF",
            "Name": "Auxiliary Valve Number 1 Estimated Flow",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                1913,
                1914,
                1915,
                1922,
                5801
            ]
        },
        65042: {
            "Label": "AV02EF",
            "Name": "Auxiliary Valve Number 2 Estimated Flow",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                1925,
                1926,
                1927,
                1934,
                5802
            ]
        },
        65043: {
            "Label": "AV03EF",
            "Name": "Auxiliary Valve Number 3 Estimated Flow",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                1937,
                1938,
                1939,
                1946,
                5803
            ]
        },
        65044: {
            "Label": "AV04EF",
            "Name": "Auxiliary Valve Number 4 Estimated Flow",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                1949,
                1950,
                1951,
                1958,
                5804
            ]
        },
        65045: {
            "Label": "AV05EF",
            "Name": "Auxiliary Valve Number 5 Estimated Flow",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                1961,
                1962,
                1963,
                1970,
                5805
            ]
        },
        65046: {
            "Label": "AV06EF",
            "Name": "Auxiliary Valve Number 6 Estimated Flow",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                1973,
                1974,
                1975,
                1982,
                5806
            ]
        },
        65047: {
            "Label": "AV07EF",
            "Name": "Auxiliary Valve Number 7 Estimated Flow",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                1985,
                1986,
                1987,
                1994,
                5807
            ]
        },
        65048: {
            "Label": "AV08EF",
            "Name": "Auxiliary Valve Number 8 Estimated Flow",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                1997,
                1998,
                1999,
                2262,
                5808
            ]
        },
        65049: {
            "Label": "AV09EF",
            "Name": "Auxiliary Valve Number 9 Estimated Flow",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                2265,
                2266,
                2267,
                2274,
                5809
            ]
        },
        65050: {
            "Label": "AV10EF",
            "Name": "Auxiliary Valve Number 10 Estimated Flow",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                2277,
                2278,
                2279,
                2286,
                5810
            ]
        },
        65051: {
            "Label": "AV11EF",
            "Name": "Auxiliary Valve Number 11 Estimated Flow",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                2289,
                2290,
                2291,
                2298,
                5811
            ]
        },
        65052: {
            "Label": "AV12EF",
            "Name": "Auxiliary Valve Number 12 Estimated Flow",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                2301,
                2302,
                2303,
                2310,
                5812
            ]
        },
        65053: {
            "Label": "AV13EF",
            "Name": "Auxiliary Valve Number 13 Estimated Flow",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                2313,
                2314,
                2315,
                2322,
                5813
            ]
        },
        65054: {
            "Label": "AV14EF",
            "Name": "Auxiliary Valve Number 14 Estimated Flow",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                2325,
                2326,
                2327,
                2334,
                5814
            ]
        },
        65055: {
            "Label": "AV15EF",
            "Name": "Auxiliary Valve Number 15 Estimated Flow",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                2337,
                2338,
                2339,
                2346,
                5815
            ]
        },
        65056: {
            "Label": "AV00MF",
            "Name": "Auxiliary Valve Number 0 Measured Flow",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                1899,
                1900,
                1906,
                1904,
                1905
            ]
        },
        65057: {
            "Label": "AV01MF",
            "Name": "Auxiliary Valve Number 1 Measured Flow",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                1911,
                1912,
                1918,
                1916,
                1917
            ]
        },
        65058: {
            "Label": "AV02MF",
            "Name": "Auxiliary Valve Number 2 Measured Flow",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                1923,
                1924,
                1930,
                1928,
                1929
            ]
        },
        65059: {
            "Label": "AV03MF",
            "Name": "Auxiliary Valve Number 3 Measured Flow",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                1935,
                1936,
                1942,
                1940,
                1941
            ]
        },
        65060: {
            "Label": "AV04MF",
            "Name": "Auxiliary Valve Number 4 Measured Flow",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                1947,
                1948,
                1954,
                1952,
                1953
            ]
        },
        65061: {
            "Label": "AV05MF",
            "Name": "Auxiliary Valve Number 5 Measured Flow",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                1959,
                1960,
                1966,
                1964,
                1965
            ]
        },
        65062: {
            "Label": "AV06MF",
            "Name": "Auxiliary Valve Number 6 Measured Flow",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                1971,
                1972,
                1978,
                1976,
                1977
            ]
        },
        65063: {
            "Label": "AV07MF",
            "Name": "Auxiliary Valve Number 7 Measured Flow",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                1983,
                1984,
                1990,
                1988,
                1989
            ]
        },
        65064: {
            "Label": "AV08MF",
            "Name": "Auxiliary Valve Number 8 Measured Flow",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                1995,
                1996,
                2258,
                2256,
                2257
            ]
        },
        65065: {
            "Label": "AV09MF",
            "Name": "Auxiliary Valve Number 9 Measured Flow",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                2263,
                2264,
                2270,
                2268,
                2269
            ]
        },
        65066: {
            "Label": "AV10MF",
            "Name": "Auxiliary Valve Number 10 Measured Flow",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                2275,
                2276,
                2282,
                2280,
                2281
            ]
        },
        65067: {
            "Label": "AV11MF",
            "Name": "Auxiliary Valve Number 11 Measured Flow",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                2287,
                2288,
                2294,
                2292,
                2293
            ]
        },
        65068: {
            "Label": "AV12MF",
            "Name": "Auxiliary Valve Number 12 Measured Flow",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                2299,
                2300,
                2306,
                2304,
                2305
            ]
        },
        65069: {
            "Label": "AV13MF",
            "Name": "Auxiliary Valve Number 13 Measured Flow",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                2311,
                2312,
                2318,
                2316,
                2317
            ]
        },
        65070: {
            "Label": "AV14MF",
            "Name": "Auxiliary Valve Number 14 Measured Flow",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                2323,
                2324,
                2330,
                2328,
                2329
            ]
        },
        65071: {
            "Label": "AV15MF",
            "Name": "Auxiliary Valve Number 15 Measured Flow",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                2335,
                2336,
                2342,
                2340,
                2341
            ]
        },
        65072: {
            "Label": "AV00C",
            "Name": "Auxiliary Valve Number 0 Command",
            "PGNLength": 8,
            "Rate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.",
            "SPNs": [
                1907,
                1908,
                1909
            ]
        },
        65073: {
            "Label": "AV01C",
            "Name": "Auxiliary Valve Number 1 Command",
            "PGNLength": 8,
            "Rate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.",
            "SPNs": [
                1919,
                1920,
                1921
            ]
        },
        65074: {
            "Label": "AV02C",
            "Name": "Auxiliary Valve Number 2 Command",
            "PGNLength": 8,
            "Rate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.",
            "SPNs": [
                1931,
                1932,
                1933
            ]
        },
        65075: {
            "Label": "AV03C",
            "Name": "Auxiliary Valve Number 3 Command",
            "PGNLength": 8,
            "Rate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.",
            "SPNs": [
                1943,
                1944,
                1945
            ]
        },
        65076: {
            "Label": "AV04C",
            "Name": "Auxiliary Valve Number 4 Command",
            "PGNLength": 8,
            "Rate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.",
            "SPNs": [
                1955,
                1956,
                1957
            ]
        },
        65077: {
            "Label": "AV05C",
            "Name": "Auxiliary Valve Number 5 Command",
            "PGNLength": 8,
            "Rate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.",
            "SPNs": [
                1967,
                1968,
                1969
            ]
        },
        65078: {
            "Label": "AV06C",
            "Name": "Auxiliary Valve Number 6 Command",
            "PGNLength": 8,
            "Rate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.",
            "SPNs": [
                1979,
                1980,
                1981
            ]
        },
        65079: {
            "Label": "AV07C",
            "Name": "Auxiliary Valve Number 7 Command",
            "PGNLength": 8,
            "Rate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.",
            "SPNs": [
                1991,
                1992,
                1993
            ]
        },
        65080: {
            "Label": "AV08C",
            "Name": "Auxiliary Valve Number 8 Command",
            "PGNLength": 8,
            "Rate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.",
            "SPNs": [
                2259,
                2260,
                2261
            ]
        },
        65081: {
            "Label": "AV09C",
            "Name": "Auxiliary Valve Number 9 Command",
            "PGNLength": 8,
            "Rate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.",
            "SPNs": [
                2271,
                2272,
                2273
            ]
        },
        65082: {
            "Label": "AV10C",
            "Name": "Auxiliary Valve Number 10 Command",
            "PGNLength": 8,
            "Rate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.",
            "SPNs": [
                2283,
                2284,
                2285
            ]
        },
        65083: {
            "Label": "AV11C",
            "Name": "Auxiliary Valve Number 11 Command",
            "PGNLength": 8,
            "Rate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.",
            "SPNs": [
                2295,
                2296,
                2297
            ]
        },
        65084: {
            "Label": "AV12C",
            "Name": "Auxiliary Valve Number 12 Command",
            "PGNLength": 8,
            "Rate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.",
            "SPNs": [
                2307,
                2308,
                2309
            ]
        },
        65085: {
            "Label": "AV13C",
            "Name": "Auxiliary Valve Number 13 Command",
            "PGNLength": 8,
            "Rate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.",
            "SPNs": [
                2319,
                2320,
                2321
            ]
        },
        65086: {
            "Label": "AV14C",
            "Name": "Auxiliary Valve Number 14 Command",
            "PGNLength": 8,
            "Rate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.",
            "SPNs": [
                2331,
                2332,
                2333
            ]
        },
        65087: {
            "Label": "AV15C",
            "Name": "Auxiliary Valve Number 15 Command",
            "PGNLength": 8,
            "Rate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.",
            "SPNs": [
                2343,
                2344,
                2345
            ]
        },
        65088: {
            "Label": "LD",
            "Name": "Lighting Data",
            "PGNLength": 8,
            "Rate": "As requested.",
            "SPNs": [
                2404,
                2352,
                2350,
                2348,
                2388,
                2386,
                2370,
                2368,
                2392,
                2376,
                2374,
                2372,
                2384,
                2382,
                2380,
                2378,
                2390,
                2358,
                2360,
                2362,
                2364,
                2366,
                2354,
                2356,
                2398,
                2396,
                2407,
                2598,
                2402,
                2400,
                2394
            ]
        },
        65089: {
            "Label": "LCMD",
            "Name": "Lighting Command",
            "PGNLength": 8,
            "Rate": "Every 1 s and on change of state but no faster than every 100 ms",
            "SPNs": [
                2403,
                2351,
                2349,
                2347,
                2387,
                2385,
                2369,
                2367,
                2391,
                2375,
                2373,
                2371,
                2383,
                2381,
                2379,
                2377,
                2389,
                2357,
                2359,
                2361,
                2363,
                2365,
                2353,
                2355,
                2397,
                2395,
                2406,
                2597,
                2393,
                2401,
                2399,
                2405
            ]
        },
        65090: {
            "Label": "HPTOC",
            "Name": "Hitch and PTO Commands",
            "PGNLength": 8,
            "Rate": "Minimum rate of 1 second between messages or when a parameter is required to change state.",
            "SPNs": [
                1874,
                1875,
                1894,
                1893,
                1898,
                1897,
                1896,
                1895,
                1886,
                1887
            ]
        },
        65091: {
            "Label": "RPTO",
            "Name": "Primary or Rear Power Take off Output Shaft:8",
            "PGNLength": "100ms when engage",
            "Rate": " otherwise on request",
            "SPNs": [
                1892,
                1890,
                2408,
                5820,
                1883,
                1885
            ]
        },
        65092: {
            "Label": "FPTO",
            "Name": "Secondary or Front Power Take off Output Shaft:8",
            "PGNLength": "100ms when engage",
            "Rate": " otherwise on request",
            "SPNs": [
                1891,
                1889,
                1888,
                5816,
                5817,
                1882,
                1884
            ]
        },
        65093: {
            "Label": "RHS",
            "Name": "Primary or Rear Hitch Status",
            "PGNLength": 8,
            "Rate": "100ms",
            "SPNs": [
                1873,
                1877,
                1881,
                5819,
                1879
            ]
        },
        65094: {
            "Label": "FHS",
            "Name": "Secondary or Front Hitch Status",
            "PGNLength": 8,
            "Rate": "100ms",
            "SPNs": [
                1872,
                1876,
                1880,
                1878
            ]
        },
        65095: {
            "Label": "MP",
            "Name": "Maintain Power",
            "PGNLength": 8,
            "Rate": "As required after receiving the message indicating that the ignition switch as change from ON state to OFF state or on change of state of parameters",
            "SPNs": [
                1868,
                1867,
                1871,
                1870,
                1869
            ]
        },
        65096: {
            "Label": "WBSD",
            "Name": "Wheel-based Speed and Distance",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                1866,
                1864,
                1865,
                1862,
                1863
            ]
        },
        65097: {
            "Label": "GBSD",
            "Name": "Ground-based Speed and Distance",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                1861
            ]
        },
        65098: {
            "Label": "ETC7",
            "Name": "Electronic Transmission Controller 7",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                4176,
                4178,
                1850,
                1849,
                3086,
                2945,
                2900,
                1851,
                2539,
                2538,
                2537,
                2536,
                3289,
                4250,
                4251,
                4252,
                4253,
                4261,
                5344
            ]
        },
        65099: {
            "Label": "TCFG2",
            "Name": "Transmission Configuration 2",
            "PGNLength": "Variable",
            "Rate": "On request or sender may transmit every 5 seconds until acknowledged by reception of the engine configuration message PGN 65251 SPN 1846.",
            "SPNs": [
                1845
            ]
        },
        65100: {
            "Label": "ML",
            "Name": "Military Lighting Command:8",
            "PGNLength": "500ms or upon state change",
            "Rate": " but not faster than 100 ms.",
            "SPNs": [
                1840,
                1839,
                1838,
                1837,
                1841,
                1843,
                1842,
                1844
            ]
        },
        65101: {
            "Label": "TAVG",
            "Name": "Total Averaged Information",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                1834,
                1835
            ]
        },
        65102: {
            "Label": "DC1",
            "Name": "Door Control 1",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                1821,
                1820,
                3411
            ]
        },
        65103: {
            "Label": "VDC1",
            "Name": "Vehicle Dynamic Stability Control 1",
            "PGNLength": 8,
            "Rate": "100ms",
            "SPNs": [
                1813,
                1814,
                1815,
                1816,
                1818,
                1817,
                1819,
                5624
            ]
        },
        65104: {
            "Label": "BT1",
            "Name": "Battery Temperature",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                1800,
                1801
            ]
        },
        65105: {
            "Label": "ACC2",
            "Name": "Adaptive Cruise Control: Operator Input",
            "PGNLength": 8,
            "Rate": "250 ms",
            "SPNs": [
                5023,
                1799
            ]
        },
        65106: {
            "Label": "VEP3",
            "Name": "Vehicle Electrical Power #3",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                1795,
                2579
            ]
        },
        65107: {
            "Label": "RTC1",
            "Name": "Retarder Continuous Torque & Speed Limit",
            "PGNLength": 8,
            "Rate": "5 s",
            "SPNs": [
                1776,
                1777,
                1778,
                1779,
                1780,
                1781,
                1782,
                1783
            ]
        },
        65108: {
            "Label": "ECT1",
            "Name": "Engine Continuous Torque & Speed Limit",
            "PGNLength": 8,
            "Rate": "5 s",
            "SPNs": [
                1768,
                1769,
                1770,
                1771,
                1772,
                1773,
                1774,
                1775
            ]
        },
        65109: {
            "Label": "GFD",
            "Name": "Gaseous Fuel Properties",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                1767,
                5684,
                4245
            ]
        },
        65110: {
            "Label": "AT1T1I",
            "Name": "Aftertreatment 1 Diesel Exhaust Fluid Tank 1 Information",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                1761,
                3031,
                3532,
                5245,
                4365,
                5246,
                3363,
                4366,
                3517
            ]
        },
        65111: {
            "Label": "ASC5",
            "Name": "Air Suspension Control 5",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                1729,
                1730,
                1731,
                1833,
                1832,
                1831
            ]
        },
        65112: {
            "Label": "ASC4",
            "Name": "Air Suspension Control 4",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                1725,
                1726,
                1727,
                1728
            ]
        },
        65113: {
            "Label": "ASC3",
            "Name": "Air Suspension Control 3",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                1721,
                1722,
                1724,
                1723
            ]
        },
        65114: {
            "Label": "ASC1",
            "Name": "Air Suspension Control 1",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                1734,
                1733,
                1738,
                1754,
                1737,
                1736,
                1740,
                1755,
                1739,
                1756,
                1742,
                1741,
                1746,
                1745,
                1744,
                1743,
                1824,
                1823,
                1822,
                1826,
                1825,
                5294,
                5296,
                1827,
                5432
            ]
        },
        65115: {
            "Label": "FLI2",
            "Name": "Forward Lane Image",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                1711,
                1710,
                1702
            ]
        },
        65116: {
            "Label": "RGE22",
            "Name": "ISO 11992 (even) - Running Gear Equipment #2/2",
            "PGNLength": 8,
            "Rate": "100 ms +/- 10 ms",
            "SPNs": []
        },
        65117: {
            "Label": "GPM12",
            "Name": "ISO 11992 (odd) - General Purpose Message #1/2",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": []
        },
        65118: {
            "Label": "RGE23",
            "Name": "ISO 11992 (even) - Running Gear Equipment #2/3",
            "PGNLength": 8,
            "Rate": "1000 ms +/- 10 ms",
            "SPNs": []
        },
        65119: {
            "Label": "GPM13",
            "Name": "ISO 11992 (odd) - General Purpose Message #1/3",
            "PGNLength": 8,
            "Rate": "50 ms",
            "SPNs": []
        },
        65120: {
            "Label": "GPM23",
            "Name": "ISO 11992 (even) - General Purpose Message #2/3",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": []
        },
        65121: {
            "Label": "GPM14",
            "Name": "ISO 11992 (odd) - General Purpose Message #1/4",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": []
        },
        65122: {
            "Label": "GPM24",
            "Name": "ISO 11992 (even) - General Purpose Message #2/4",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": []
        },
        65123: {
            "Label": "GPM15",
            "Name": "ISO 11992 (odd) - General Purpose Message #1/5",
            "PGNLength": 8,
            "Rate": "1000 ms",
            "SPNs": []
        },
        65124: {
            "Label": "GPM25",
            "Name": "ISO 11992 (even) - General Purpose Message #2/5",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": []
        },
        65125: {
            "Label": "GPM16",
            "Name": "ISO 11992 (odd) - General Purpose Message #1/6",
            "PGNLength": 8,
            "Rate": "1000 ms",
            "SPNs": []
        },
        65126: {
            "Label": "BM",
            "Name": "Battery Main Switch Information",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                1681
            ]
        },
        65127: {
            "Label": "CCC",
            "Name": "Climate Control Configuration",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                1690
            ]
        },
        65128: {
            "Label": "VF",
            "Name": "Vehicle Fluids",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                1638,
                1713,
                1857,
                2602
            ]
        },
        65129: {
            "Label": "ET3",
            "Name": "Engine Temperature 3",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                1636,
                1637,
                2986,
                2630
            ]
        },
        65130: {
            "Label": "EFS",
            "Name": "Engine Fuel/lube systems",
            "PGNLength": 8,
            "Rate": "0.5 s",
            "SPNs": [
                1380,
                1381,
                1382,
                3548,
                3549,
                3550,
                3551,
                3552,
                3553,
                3554,
                4083
            ]
        },
        65131: {
            "Label": "DI",
            "Name": "Driver's Identification",
            "PGNLength": "Variable",
            "Rate": "On request",
            "SPNs": [
                1625,
                1626
            ]
        },
        65132: {
            "Label": "TCO1",
            "Name": "Tachograph",
            "PGNLength": 8,
            "Rate": "50 ms",
            "SPNs": [
                1612,
                1613,
                1611,
                1617,
                1615,
                1614,
                1618,
                1616,
                1622,
                1621,
                1620,
                1619,
                1623,
                1624
            ]
        },
        65133: {
            "Label": "HTR",
            "Name": "Heater Information",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                1687,
                1688,
                1689,
                1677,
                1676,
                1678,
                1679,
                1680
            ]
        },
        65134: {
            "Label": "HRW",
            "Name": "High Resolution Wheel Speed",
            "PGNLength": 8,
            "Rate": "20 ms",
            "SPNs": [
                1592,
                1593,
                1594,
                1595
            ]
        },
        65135: {
            "Label": "ACC1",
            "Name": "Adaptive Cruise Control 1:8",
            "PGNLength": "100ms or upon state change",
            "Rate": " but not faster than 20 ms.",
            "SPNs": [
                1586,
                1587,
                1588,
                1590,
                1589,
                1798,
                1797,
                1796,
                5022,
                1591
            ]
        },
        65136: {
            "Label": "CVW",
            "Name": "Combination Vehicle Weight",
            "PGNLength": "Variable",
            "Rate": "On request",
            "SPNs": [
                1585,
                1760
            ]
        },
        65137: {
            "Label": "LTP",
            "Name": "Laser Tracer Position",
            "PGNLength": 8,
            "Rate": "50 ms",
            "SPNs": [
                1581,
                1582,
                1583,
                1579,
                1580
            ]
        },
        65138: {
            "Label": "LBC",
            "Name": "Laser Leveling System Blade Control",
            "PGNLength": 8,
            "Rate": "50 ms",
            "SPNs": [
                1578,
                5407,
                5408,
                5409,
                1577
            ]
        },
        65139: {
            "Label": "LMP",
            "Name": "Laser Receiver Mast Position",
            "PGNLength": 8,
            "Rate": "50 ms",
            "SPNs": [
                1576
            ]
        },
        65140: {
            "Label": "LSP",
            "Name": "Modify Leveling System Control Set Point",
            "PGNLength": 8,
            "Rate": "50 ms",
            "SPNs": [
                1575,
                1759
            ]
        },
        65141: {
            "Label": "LVD",
            "Name": "Laser Leveling System Vertical Deviation",
            "PGNLength": 8,
            "Rate": "50 ms",
            "SPNs": [
                2576,
                1574,
                2793,
                2794
            ]
        },
        65142: {
            "Label": "LVDD",
            "Name": "Laser Leveling System Vertical Position Display Data",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                1573,
                1805,
                1806,
                2578,
                2577
            ]
        },
        65143: {
            "Label": "AP",
            "Name": "Auxiliary Pressures",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                136,
                137,
                138
            ]
        },
        65144: {
            "Label": "TP1",
            "Name": "Tire Pressure Control Unit Mode and Status",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                39,
                1466,
                1467,
                1468,
                1469,
                1470,
                1471,
                1472,
                1473,
                1474,
                1475
            ]
        },
        65145: {
            "Label": "TP2",
            "Name": "Tire Pressure Control Unit Target Pressures",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                141,
                142,
                143
            ]
        },
        65146: {
            "Label": "TP3",
            "Name": "Tire Pressure Control Unit Current Pressures",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                144,
                145,
                146
            ]
        },
        65147: {
            "Label": "CT1",
            "Name": "Combustion Time 1",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                1444,
                1445,
                1446,
                1447
            ]
        },
        65148: {
            "Label": "CT2",
            "Name": "Combustion Time 2",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                1448,
                1449,
                1450,
                1451
            ]
        },
        65149: {
            "Label": "CT3",
            "Name": "Combustion Time 3",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                1452,
                1453,
                1454,
                1455
            ]
        },
        65150: {
            "Label": "CT4",
            "Name": "Combustion Time 4",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                1456,
                1457,
                1458,
                1459
            ]
        },
        65151: {
            "Label": "CT5",
            "Name": "Combustion Time 5",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                1460,
                1461,
                1462,
                1463
            ]
        },
        65152: {
            "Label": "CT6",
            "Name": "Combustion Time 6",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                1464,
                1465
            ]
        },
        65153: {
            "Label": "GFI2",
            "Name": "Fuel Information 2 (Gaseous)",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                1442,
                1443,
                1765,
                1766,
                1440,
                1441
            ]
        },
        65154: {
            "Label": "IT1",
            "Name": "Ignition Timing 1",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                1413,
                1414,
                1415,
                1416
            ]
        },
        65155: {
            "Label": "IT2",
            "Name": "Ignition Timing 2",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                1417,
                1418,
                1419,
                1420
            ]
        },
        65156: {
            "Label": "IT3",
            "Name": "Ignition Timing 3",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                1421,
                1422,
                1423,
                1424
            ]
        },
        65157: {
            "Label": "IT4",
            "Name": "Ignition Timing 4",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                1425,
                1426,
                1427,
                1428
            ]
        },
        65158: {
            "Label": "IT5",
            "Name": "Ignition Timing 5",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                1429,
                1430,
                1431,
                1432
            ]
        },
        65159: {
            "Label": "IT6",
            "Name": "Ignition Timing 6",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                1433,
                1434,
                1435,
                1436
            ]
        },
        65160: {
            "Label": "ISO1",
            "Name": "Ignition Transformer Secondary Output 1",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                1393,
                1394,
                1395,
                1396,
                1397,
                1398,
                1399,
                1400
            ]
        },
        65161: {
            "Label": "ISO2",
            "Name": "Ignition Transformer Secondary Output 2",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                1401,
                1402,
                1403,
                1404,
                1405,
                1406,
                1407,
                1408
            ]
        },
        65162: {
            "Label": "ISO3",
            "Name": "Ignition Transformer Secondary Output 3",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                1409,
                1410,
                1411,
                1412
            ]
        },
        65163: {
            "Label": "GFP",
            "Name": "Gaseous Fuel Pressure",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                1390,
                1391,
                1392,
                2980
            ]
        },
        65164: {
            "Label": "AAI",
            "Name": "Auxiliary Analog Information",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                441,
                442,
                1387,
                1388,
                354,
                3087
            ]
        },
        65165: {
            "Label": "VEP2",
            "Name": "Vehicle Electrical Power #2",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                444,
                3597,
                3598,
                3599
            ]
        },
        65166: {
            "Label": "S2",
            "Name": "Service 2",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                1379,
                1350
            ]
        },
        65167: {
            "Label": "SP2",
            "Name": "Supply Pressure 2",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                1320
            ]
        },
        65168: {
            "Label": "ETH",
            "Name": "Engine Torque History",
            "PGNLength": "Variable",
            "Rate": "On request",
            "SPNs": [
                1246,
                1250,
                1251,
                1252,
                1254,
                1632,
                1247,
                1248,
                1249,
                1253,
                1255,
                1256,
                1257,
                1258,
                1259,
                1260,
                1261,
                1262,
                1263
            ]
        },
        65169: {
            "Label": "FL",
            "Name": "Fuel Leakage",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                1239,
                1240
            ]
        },
        65170: {
            "Label": "EI",
            "Name": "Engine Information",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                1208,
                1210,
                1209,
                1241,
                1242
            ]
        },
        65171: {
            "Label": "EES",
            "Name": "Engine Electrical System/Module Information",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                1205,
                1204
            ]
        },
        65172: {
            "Label": "EAC",
            "Name": "Engine Auxiliary Coolant",
            "PGNLength": 8,
            "Rate": "0.5 s",
            "SPNs": [
                1203,
                1212,
                2435,
                20
            ]
        },
        65173: {
            "Label": "RBI",
            "Name": "Rebuild Information",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                1193
            ]
        },
        65174: {
            "Label": "TCW",
            "Name": "Turbocharger Wastegate",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                1188,
                1189,
                1190,
                1191,
                1192,
                5370,
                5371,
                5372,
                5373
            ]
        },
        65175: {
            "Label": "TCI5",
            "Name": "Turbocharger Information 5",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                1184,
                1185,
                1186,
                1187
            ]
        },
        65176: {
            "Label": "TCI4",
            "Name": "Turbocharger Information 4",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                1180,
                1181,
                1182,
                1183
            ]
        },
        65177: {
            "Label": "TCI3",
            "Name": "Turbocharger Information 3",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                1176,
                1177,
                1178,
                1179
            ]
        },
        65178: {
            "Label": "TCI2",
            "Name": "Turbocharger Information 2",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                1172,
                1173,
                1174,
                1175
            ]
        },
        65179: {
            "Label": "TCI1",
            "Name": "Turbocharger Information 1",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                1168,
                1169,
                1170,
                1171
            ]
        },
        65180: {
            "Label": "MBT3",
            "Name": "Main Bearing Temperature 3",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                1165,
                1166,
                1167
            ]
        },
        65181: {
            "Label": "MBT2",
            "Name": "Main Bearing Temperature 2",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                1161,
                1162,
                1163,
                1164
            ]
        },
        65182: {
            "Label": "MBT1",
            "Name": "Main Bearing Temperature 1",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                1157,
                1158,
                1159,
                1160
            ]
        },
        65183: {
            "Label": "EPT5",
            "Name": "Exhaust Port Temperature 5",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                1153,
                1154,
                1155,
                1156
            ]
        },
        65184: {
            "Label": "EPT4",
            "Name": "Exhaust Port Temperature 4",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                1149,
                1150,
                1151,
                1152
            ]
        },
        65185: {
            "Label": "EPT3",
            "Name": "Exhaust Port Temperature 3",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                1145,
                1146,
                1147,
                1148
            ]
        },
        65186: {
            "Label": "EPT2",
            "Name": "Exhaust Port Temperature 2",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                1141,
                1142,
                1143,
                1144
            ]
        },
        65187: {
            "Label": "EPT1",
            "Name": "Exhaust Port Temperature 1",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                1137,
                1138,
                1139,
                1140
            ]
        },
        65188: {
            "Label": "ET2",
            "Name": "Engine Temperature 2",
            "PGNLength": 8,
            "Rate": "1s",
            "SPNs": [
                1135,
                1136,
                411,
                412
            ]
        },
        65189: {
            "Label": "IMT2",
            "Name": "Intake Manifold Information 2",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                1131,
                1132,
                1133,
                1802,
                1803
            ]
        },
        65190: {
            "Label": "IMT1",
            "Name": "Intake Manifold Information 1",
            "PGNLength": 8,
            "Rate": "0.5 s",
            "SPNs": [
                1127,
                1128,
                1129,
                1130
            ]
        },
        65191: {
            "Label": "AT",
            "Name": "Alternator Temperature",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                1122,
                1123,
                1124,
                1125,
                1126
            ]
        },
        65192: {
            "Label": "ACTL",
            "Name": "Articulation Control",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                1120
            ]
        },
        65193: {
            "Label": "EO1",
            "Name": "Exhaust Oxygen 1",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                1695,
                1696,
                1117,
                1118,
                1119
            ]
        },
        65194: {
            "Label": "AF2",
            "Name": "Alternate Fuel 2",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                1116,
                1693,
                1694,
                1692
            ]
        },
        65195: {
            "Label": "ETC6",
            "Name": "Electronic Transmission Controller 6",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                1113,
                1115,
                1114,
                2983
            ]
        },
        65196: {
            "Label": "EBC4",
            "Name": "Wheel Brake Lining Remaining Information",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                1099,
                1100,
                1101,
                1102,
                1103,
                1104,
                1105,
                1106
            ]
        },
        65197: {
            "Label": "EBC3",
            "Name": "Wheel Application Pressure High Range Information",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                1091,
                1092,
                1093,
                1094,
                1095,
                1096,
                1097,
                1098
            ]
        },
        65198: {
            "Label": "AIR1",
            "Name": "Air Supply Pressure",
            "PGNLength": 8,
            "Rate": "1 sec",
            "SPNs": [
                46,
                1086,
                1087,
                1088,
                1089,
                1090,
                1351
            ]
        },
        65199: {
            "Label": "GFC",
            "Name": "Fuel Consumption (Gaseous)",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                1039,
                1040
            ]
        },
        65200: {
            "Label": "TTI2",
            "Name": "Trip Time Information 2",
            "PGNLength": 20,
            "Rate": "On request",
            "SPNs": [
                1034,
                1035,
                1036,
                1037,
                1038
            ]
        },
        65201: {
            "Label": "EH",
            "Name": "ECU History",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                1032,
                1033
            ]
        },
        65202: {
            "Label": "GFI1",
            "Name": "Fuel Information 1 (Gaseous)",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                1030,
                1031,
                1389
            ]
        },
        65203: {
            "Label": "LFI",
            "Name": "Fuel Information (Liquid)",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                1028,
                1029,
                5458
            ]
        },
        65204: {
            "Label": "TTI1",
            "Name": "Trip Time Information 1",
            "PGNLength": 16,
            "Rate": "On request",
            "SPNs": [
                1024,
                1025,
                1026,
                1027
            ]
        },
        65205: {
            "Label": "TSI",
            "Name": "Trip Shutdown Information",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                1020,
                1021,
                1022,
                1023
            ]
        },
        65206: {
            "Label": "TVI",
            "Name": "Trip Vehicle Speed/Cruise Distance Information",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                1018,
                1019
            ]
        },
        65207: {
            "Label": "LF",
            "Name": "Engine Speed/Load Factor Information",
            "PGNLength": 10,
            "Rate": "On request",
            "SPNs": [
                1015,
                1016,
                1013,
                1014,
                1017
            ]
        },
        65208: {
            "Label": "GTFI",
            "Name": "Trip Fuel Information (Gaseous)",
            "PGNLength": 22,
            "Rate": "On request",
            "SPNs": [
                1007,
                1008,
                1009,
                1010,
                1011,
                1012
            ]
        },
        65209: {
            "Label": "LTFI",
            "Name": "Trip Fuel Information (Liquid)",
            "PGNLength": 22,
            "Rate": "On request",
            "SPNs": [
                1001,
                1002,
                1003,
                1004,
                1005,
                1006
            ]
        },
        65210: {
            "Label": "TDI",
            "Name": "Trip Distance Information",
            "PGNLength": 12,
            "Rate": "On request",
            "SPNs": [
                998,
                999,
                1000
            ]
        },
        65211: {
            "Label": "TFI",
            "Name": "Trip Fan Information",
            "PGNLength": 16,
            "Rate": "On request",
            "SPNs": [
                994,
                995,
                996,
                997
            ]
        },
        65212: {
            "Label": "CBI",
            "Name": "Compression/Service Brake Information",
            "PGNLength": 16,
            "Rate": "On request",
            "SPNs": [
                990,
                991,
                992,
                993
            ]
        },
        65213: {
            "Label": "FD1",
            "Name": "Fan Drive #1",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                975,
                977,
                4212,
                1639,
                4211
            ]
        },
        65214: {
            "Label": "EEC4",
            "Name": "Electronic Engine Controller 4",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                3669,
                5465,
                3671,
                166,
                189
            ]
        },
        65215: {
            "Label": "EBC2",
            "Name": "Wheel Speed Information",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                905,
                906,
                907,
                908,
                909,
                910,
                904
            ]
        },
        65216: {
            "Label": "SERV",
            "Name": "Service Information",
            "PGNLength": "8 bytes or Variable",
            "Rate": "On request",
            "SPNs": [
                911,
                912,
                915,
                913,
                914,
                916
            ]
        },
        65217: {
            "Label": "VDHR",
            "Name": "High Resolution Vehicle Distance",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                917,
                918
            ]
        },
        65218: {
            "Label": "ERC2",
            "Name": "Electronic Retarder Controller 2:8",
            "PGNLength": "Every 1 s and on change of state but no faster than every 100 ms.  Grandfathered definition for systems that implemented this message prior to July",
            "Rate": " 2010: 1 s when active; or on change of state",
            "SPNs": [
                748
            ]
        },
        65219: {
            "Label": "ETC5",
            "Name": "Electronic Transmission Controller 5",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                778,
                779,
                767,
                604,
                903
            ]
        },
        65220: {
            "Label": "EBS22",
            "Name": "Reserved for ISO 11992",
            "PGNLength": 8,
            "Rate": "100 ms +/- 10 ms",
            "SPNs": []
        },
        65221: {
            "Label": "ETC4",
            "Name": "Electronic Transmission Controller 4",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                53,
                54
            ]
        },
        65222: {
            "Label": "EBS23",
            "Name": "Reserved for ISO 11992",
            "PGNLength": 8,
            "Rate": "100 ms +/- 10 ms",
            "SPNs": []
        },
        65223: {
            "Label": "ETC3",
            "Name": "Electronic Transmission Controller 3",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                59,
                60,
                780,
                781,
                782,
                772,
                773,
                783,
                784,
                768,
                769,
                770,
                771,
                788,
                740,
                786,
                787
            ]
        },
        65224: {
            "Label": "GPM22",
            "Name": "Reserved for ISO 11992",
            "PGNLength": 8,
            "Rate": "100 ms +/- 10 ms",
            "SPNs": []
        },
        65225: {
            "Label": "EBS12",
            "Name": "Reserved for ISO 11992",
            "PGNLength": 8,
            "Rate": "100 ms +/- 10 ms",
            "SPNs": []
        },
        65226: {
            "Label": "DM01",
            "Name": "Active Diagnostic Trouble Codes",
            "PGNLength": "Variable",
            "Rate": "1 s or on change",
            "SPNs": [
                987,
                624,
                623,
                1213,
                3041,
                3040,
                3039,
                3038,
                1215,
                1216,
                1706,
                1214
            ]
        },
        65227: {
            "Label": "DM02",
            "Name": "Previously Active Diagnostic Trouble Codes",
            "PGNLength": "Variable",
            "Rate": "On request",
            "SPNs": [
                987,
                624,
                623,
                1213,
                3041,
                3040,
                3039,
                3038,
                1215,
                1216,
                1706,
                1214
            ]
        },
        65228: {
            "Label": "DM03",
            "Name": "Diagnostics Data Clear/Reset for Previously Active DTCs",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        65229: {
            "Label": "DM04",
            "Name": "Freeze Frame Parameters",
            "PGNLength": "Variable",
            "Rate": "On request",
            "SPNs": [
                1217,
                1215,
                1216,
                1706,
                1214,
                4263
            ]
        },
        65230: {
            "Label": "DM05",
            "Name": "Diagnostic Readiness 1",
            "PGNLength": "Variable",
            "Rate": "On request",
            "SPNs": [
                1218,
                1219,
                1220,
                1221,
                1222,
                1223
            ]
        },
        65231: {
            "Label": "DM06",
            "Name": "Pending DTCs",
            "PGNLength": "Variable",
            "Rate": "On request",
            "SPNs": [
                987,
                624,
                623,
                1213,
                3041,
                3040,
                3039,
                3038,
                1215,
                1216,
                1706,
                1214
            ]
        },
        65232: {
            "Label": "DM08",
            "Name": "Test Results for Non-continuously Monitored Systems",
            "PGNLength": "Variable",
            "Rate": "Sent in response to PGN 58112 when the results are available",
            "SPNs": [
                1224,
                1225,
                1226,
                1227,
                1228
            ]
        },
        65233: {
            "Label": "DM09",
            "Name": "Oxygen Sensor Test Results",
            "PGNLength": "N/A",
            "Rate": "N/A",
            "SPNs": []
        },
        65234: {
            "Label": "DM10",
            "Name": "Non-continuously Monitored System Test Identifiers Support",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                1229
            ]
        },
        65235: {
            "Label": "DM11",
            "Name": "Diagnostic Data Clear/Reset for Active DTCs",
            "PGNLength": 0,
            "Rate": "On request",
            "SPNs": []
        },
        65236: {
            "Label": "DM12",
            "Name": "Emissions Related Active DTCs",
            "PGNLength": "Variable",
            "Rate": "On request",
            "SPNs": [
                987,
                624,
                623,
                1213,
                3041,
                3040,
                3039,
                3038,
                1215,
                1216,
                1706,
                1214
            ]
        },
        65237: {
            "Label": "AS",
            "Name": "Alternator Information",
            "PGNLength": 8,
            "Rate": "1 sec",
            "SPNs": [
                3353,
                3354,
                3355,
                3356,
                589
            ]
        },
        65238: {
            "Label": "RESV1",
            "Name": "Reserved for Network Management",
            "PGNLength": "",
            "Rate": "",
            "SPNs": []
        },
        65240: {
            "Label": "CA",
            "Name": "Commanded Address",
            "PGNLength": 9,
            "Rate": "As required",
            "SPNs": [
                2847,
                2849
            ]
        },
        65241: {
            "Label": "AUXIO1",
            "Name": "Auxiliary Input/Output Status 1:8",
            "PGNLength": "manufacturer defined",
            "Rate": " not faster than 100 ms",
            "SPNs": [
                704,
                703,
                702,
                701,
                708,
                707,
                706,
                705,
                712,
                711,
                710,
                709,
                716,
                715,
                714,
                713,
                1083,
                1084
            ]
        },
        65242: {
            "Label": "SOFT",
            "Name": "Software Identification",
            "PGNLength": "Variable",
            "Rate": "On request",
            "SPNs": [
                965,
                234
            ]
        },
        65243: {
            "Label": "EFL/P2",
            "Name": "Engine Fluid Level/Pressure 2",
            "PGNLength": 8,
            "Rate": "0.5 s",
            "SPNs": [
                164,
                157,
                156,
                1349
            ]
        },
        65244: {
            "Label": "IO",
            "Name": "Idle Operation",
            "PGNLength": "Variable",
            "Rate": "On request",
            "SPNs": [
                236,
                235
            ]
        },
        65245: {
            "Label": "TC",
            "Name": "Turbocharger",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                104,
                1665,
                103
            ]
        },
        65246: {
            "Label": "AIR2",
            "Name": "Air Start Pressure",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                82
            ]
        },
        65247: {
            "Label": "EEC3",
            "Name": "Electronic Engine Controller 3",
            "PGNLength": 8,
            "Rate": "250 msec (preferred) or Engine Speed Dependent (if required by application)",
            "SPNs": [
                514,
                519,
                2978,
                3237,
                3238,
                3239,
                3240,
                515,
                3236
            ]
        },
        65248: {
            "Label": "VD",
            "Name": "Vehicle Distance",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                244,
                245
            ]
        },
        65249: {
            "Label": "RC",
            "Name": "Retarder Configuration",
            "PGNLength": 19,
            "Rate": "Every 5 s and on change of torque/speed points of more than 10% since last transmission but no faster than every 500 ms",
            "SPNs": [
                901,
                902,
                557,
                551,
                552,
                553,
                554,
                555,
                546,
                548,
                549,
                550,
                547,
                556
            ]
        },
        65250: {
            "Label": "TCFG",
            "Name": "Transmission Configuration",
            "PGNLength": "Variable",
            "Rate": "On request",
            "SPNs": [
                958,
                957,
                581
            ]
        },
        65251: {
            "Label": "EC1",
            "Name": "Engine Configuration 1",
            "PGNLength": 39,
            "Rate": "Every 5 s and on change of torque/speed points of more than 10% since last transmission but no faster than every 500 ms",
            "SPNs": [
                539,
                540,
                541,
                542,
                543,
                534,
                535,
                536,
                537,
                538,
                3344,
                3345,
                3346,
                3347,
                3348,
                188,
                528,
                529,
                530,
                531,
                532,
                545,
                544,
                533,
                1712,
                1794,
                1846
            ]
        },
        65252: {
            "Label": "SHUTDN",
            "Name": "Shutdown",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                593,
                594,
                592,
                590,
                591,
                985,
                875,
                605,
                1081,
                1110,
                1109,
                1108,
                1107,
                1111,
                2815,
                2814,
                2813,
                2812,
                3667,
                5404,
                5566
            ]
        },
        65253: {
            "Label": "HOURS",
            "Name": "Engine Hours: Revolutions",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                247,
                249
            ]
        },
        65254: {
            "Label": "TD",
            "Name": "Time/Date",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                959,
                960,
                961,
                963,
                962,
                964,
                1601,
                1602
            ]
        },
        65255: {
            "Label": "VH",
            "Name": "Vehicle Hours",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                246,
                248
            ]
        },
        65256: {
            "Label": "VDS",
            "Name": "Vehicle Direction/Speed",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                165,
                517,
                583,
                580
            ]
        },
        65257: {
            "Label": "LFC",
            "Name": "Fuel Consumption (Liquid)",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                182,
                250
            ]
        },
        65258: {
            "Label": "VW",
            "Name": "Vehicle Weight",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                928,
                582,
                180,
                181
            ]
        },
        65259: {
            "Label": "CI",
            "Name": "Component Identification",
            "PGNLength": "Variable",
            "Rate": "On request",
            "SPNs": [
                586,
                587,
                588,
                233
            ]
        },
        65260: {
            "Label": "VI",
            "Name": "Vehicle Identification",
            "PGNLength": "Variable",
            "Rate": "On request",
            "SPNs": [
                237
            ]
        },
        65261: {
            "Label": "CCSS",
            "Name": "Cruise Control/Vehicle Speed Setup",
            "PGNLength": 8,
            "Rate": "On request",
            "SPNs": [
                74,
                87,
                88
            ]
        },
        65262: {
            "Label": "ET1",
            "Name": "Engine Temperature 1",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                110,
                174,
                52,
                1134,
                175,
                176
            ]
        },
        65263: {
            "Label": "EFL/P1",
            "Name": "Engine Fluid Level/Pressure 1",
            "PGNLength": 8,
            "Rate": "0.5 s",
            "SPNs": [
                94,
                22,
                98,
                100,
                109,
                111,
                101
            ]
        },
        65264: {
            "Label": "PTO",
            "Name": "Power Takeoff Information",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                90,
                980,
                979,
                978,
                984,
                983,
                982,
                981,
                2897,
                3447,
                3448,
                186,
                187
            ]
        },
        65265: {
            "Label": "CCVS1",
            "Name": "Cruise Control/Vehicle Speed 1",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                69,
                70,
                1633,
                3807,
                595,
                596,
                597,
                598,
                599,
                600,
                601,
                602,
                86,
                976,
                527,
                968,
                967,
                966,
                1237,
                84
            ]
        },
        65266: {
            "Label": "LFE1",
            "Name": "Fuel Economy (Liquid)",
            "PGNLength": 8,
            "Rate": "100 ms",
            "SPNs": [
                51,
                3673,
                183,
                184,
                185
            ]
        },
        65267: {
            "Label": "VP",
            "Name": "Vehicle Position",
            "PGNLength": 8,
            "Rate": "5 s",
            "SPNs": [
                584,
                585
            ]
        },
        65268: {
            "Label": "TIRE",
            "Name": "Tire Condition",
            "PGNLength": 8,
            "Rate": "10 s",
            "SPNs": [
                929,
                241,
                1699,
                1698,
                1697,
                2587,
                242,
                2586
            ]
        },
        65269: {
            "Label": "AMB",
            "Name": "Ambient Conditions",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                108,
                172,
                170,
                171,
                79
            ]
        },
        65270: {
            "Label": "IC1",
            "Name": "Intake/Exhaust Conditions 1",
            "PGNLength": 8,
            "Rate": "0.5 s",
            "SPNs": [
                81,
                102,
                105,
                106,
                107,
                112,
                173
            ]
        },
        65271: {
            "Label": "VEP1",
            "Name": "Vehicle Electrical Power 1",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                114,
                115,
                167,
                168,
                158
            ]
        },
        65272: {
            "Label": "TRF1",
            "Name": "Transmission Fluids 1",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                123,
                124,
                126,
                127,
                3027,
                3028,
                3026,
                177
            ]
        },
        65273: {
            "Label": "AI",
            "Name": "Axle Information",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                75,
                930,
                579,
                578,
                2613,
                2614
            ]
        },
        65274: {
            "Label": "B",
            "Name": "Brakes",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                116,
                117,
                118,
                619,
                3557,
                3808
            ]
        },
        65275: {
            "Label": "RF",
            "Name": "Retarder fluids",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                119,
                120,
                5346
            ]
        },
        65276: {
            "Label": "DD",
            "Name": "Dash Display",
            "PGNLength": 8,
            "Rate": "1s",
            "SPNs": [
                80,
                96,
                95,
                99,
                38,
                169
            ]
        },
        65277: {
            "Label": "A1",
            "Name": "Alternate Fuel 1",
            "PGNLength": 8,
            "Rate": "500 ms",
            "SPNs": [
                72,
                159
            ]
        },
        65278: {
            "Label": "AWPP",
            "Name": "Auxiliary Water Pump Pressure",
            "PGNLength": 8,
            "Rate": "1 s",
            "SPNs": [
                73
            ]
        },
        65279: {
            "Label": "OI",
            "Name": "Operator indicators",
            "PGNLength": 8,
            "Rate": "10 s or on change but no more often than 1s",
            "SPNs": [
                97,
                5675,
                5825,
                5826
            ]
        },
        65280: {
            "Label": "PropB_00",
            "Name": "Proprietary B (first entry)",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65281: {
            "Label": "PropB_01",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65282: {
            "Label": "PropB_02",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65283: {
            "Label": "PropB_03",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65284: {
            "Label": "PropB_04",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65285: {
            "Label": "PropB_05",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65286: {
            "Label": "PropB_06",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65287: {
            "Label": "PropB_07",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65288: {
            "Label": "PropB_08",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65289: {
            "Label": "PropB_09",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65290: {
            "Label": "PropB_0A",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65291: {
            "Label": "PropB_0B",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65292: {
            "Label": "PropB_0C",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65293: {
            "Label": "PropB_0D",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65294: {
            "Label": "PropB_0E",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65295: {
            "Label": "PropB_0F",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65296: {
            "Label": "PropB_10",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65297: {
            "Label": "PropB_11",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65298: {
            "Label": "PropB_12",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65299: {
            "Label": "PropB_13",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65300: {
            "Label": "PropB_14",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65301: {
            "Label": "PropB_15",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65302: {
            "Label": "PropB_16",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65303: {
            "Label": "PropB_17",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65304: {
            "Label": "PropB_18",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65305: {
            "Label": "PropB_19",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65306: {
            "Label": "PropB_1A",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65307: {
            "Label": "PropB_1B",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65308: {
            "Label": "PropB_1C",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65309: {
            "Label": "PropB_1D",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65310: {
            "Label": "PropB_1E",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65311: {
            "Label": "PropB_1F",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65312: {
            "Label": "PropB_20",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65313: {
            "Label": "PropB_21",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65314: {
            "Label": "PropB_22",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65315: {
            "Label": "PropB_23",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65316: {
            "Label": "PropB_24",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65317: {
            "Label": "PropB_25",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65318: {
            "Label": "PropB_26",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65319: {
            "Label": "PropB_27",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65320: {
            "Label": "PropB_28",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65321: {
            "Label": "PropB_29",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65322: {
            "Label": "PropB_2A",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65323: {
            "Label": "PropB_2B",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65324: {
            "Label": "PropB_2C",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65325: {
            "Label": "PropB_2D",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65326: {
            "Label": "PropB_2E",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65327: {
            "Label": "PropB_2F",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65328: {
            "Label": "PropB_30",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65329: {
            "Label": "PropB_31",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65330: {
            "Label": "PropB_32",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65331: {
            "Label": "PropB_33",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65332: {
            "Label": "PropB_34",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65333: {
            "Label": "PropB_35",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65334: {
            "Label": "PropB_36",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65335: {
            "Label": "PropB_37",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65336: {
            "Label": "PropB_38",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65337: {
            "Label": "PropB_39",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65338: {
            "Label": "PropB_3A",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65339: {
            "Label": "PropB_3B",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65340: {
            "Label": "PropB_3C",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65341: {
            "Label": "PropB_3D",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65342: {
            "Label": "PropB_3E",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65343: {
            "Label": "PropB_3F",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65344: {
            "Label": "PropB_40",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65345: {
            "Label": "PropB_41",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65346: {
            "Label": "PropB_42",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65347: {
            "Label": "PropB_43",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65348: {
            "Label": "PropB_44",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65349: {
            "Label": "PropB_45",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65350: {
            "Label": "PropB_46",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65351: {
            "Label": "PropB_47",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65352: {
            "Label": "PropB_48",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65353: {
            "Label": "PropB_49",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65354: {
            "Label": "PropB_4A",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65355: {
            "Label": "PropB_4B",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65356: {
            "Label": "PropB_4C",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65357: {
            "Label": "PropB_4D",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65358: {
            "Label": "PropB_4E",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65359: {
            "Label": "PropB_4F",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65360: {
            "Label": "PropB_50",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65361: {
            "Label": "PropB_51",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65362: {
            "Label": "PropB_52",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65363: {
            "Label": "PropB_53",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65364: {
            "Label": "PropB_54",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65365: {
            "Label": "PropB_55",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65366: {
            "Label": "PropB_56",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65367: {
            "Label": "PropB_57",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65368: {
            "Label": "PropB_58",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65369: {
            "Label": "PropB_59",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65370: {
            "Label": "PropB_5A",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65371: {
            "Label": "PropB_5B",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65372: {
            "Label": "PropB_5C",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65373: {
            "Label": "PropB_5D",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65374: {
            "Label": "PropB_5E",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65375: {
            "Label": "PropB_5F",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65376: {
            "Label": "PropB_60",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65377: {
            "Label": "PropB_61",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65378: {
            "Label": "PropB_62",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65379: {
            "Label": "PropB_63",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65380: {
            "Label": "PropB_64",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65381: {
            "Label": "PropB_65",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65382: {
            "Label": "PropB_66",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65383: {
            "Label": "PropB_67",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65384: {
            "Label": "PropB_68",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65385: {
            "Label": "PropB_69",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65386: {
            "Label": "PropB_6A",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65387: {
            "Label": "PropB_6B",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65388: {
            "Label": "PropB_6C",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65389: {
            "Label": "PropB_6D",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65390: {
            "Label": "PropB_6E",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65391: {
            "Label": "PropB_6F",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65392: {
            "Label": "PropB_70",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65393: {
            "Label": "PropB_71",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65394: {
            "Label": "PropB_72",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65395: {
            "Label": "PropB_73",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65396: {
            "Label": "PropB_74",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65397: {
            "Label": "PropB_75",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65398: {
            "Label": "PropB_76",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65399: {
            "Label": "PropB_77",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65400: {
            "Label": "PropB_78",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65401: {
            "Label": "PropB_79",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65402: {
            "Label": "PropB_7A",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65403: {
            "Label": "PropB_7B",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65404: {
            "Label": "PropB_7C",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65405: {
            "Label": "PropB_7D",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65406: {
            "Label": "PropB_7E",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65407: {
            "Label": "PropB_7F",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65408: {
            "Label": "PropB_80",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65409: {
            "Label": "PropB_81",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65410: {
            "Label": "PropB_82",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65411: {
            "Label": "PropB_83",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65412: {
            "Label": "PropB_84",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65413: {
            "Label": "PropB_85",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65414: {
            "Label": "PropB_86",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65415: {
            "Label": "PropB_87",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65416: {
            "Label": "PropB_88",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65417: {
            "Label": "PropB_89",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65418: {
            "Label": "PropB_8A",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65419: {
            "Label": "PropB_8B",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65420: {
            "Label": "PropB_8C",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65421: {
            "Label": "PropB_8D",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65422: {
            "Label": "PropB_8E",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65423: {
            "Label": "PropB_8F",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65424: {
            "Label": "PropB_90",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65425: {
            "Label": "PropB_91",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65426: {
            "Label": "PropB_92",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65427: {
            "Label": "PropB_93",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65428: {
            "Label": "PropB_94",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65429: {
            "Label": "PropB_95",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65430: {
            "Label": "PropB_96",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65431: {
            "Label": "PropB_97",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65432: {
            "Label": "PropB_98",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65433: {
            "Label": "PropB_99",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65434: {
            "Label": "PropB_9A",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65435: {
            "Label": "PropB_9B",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65436: {
            "Label": "PropB_9C",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65437: {
            "Label": "PropB_9D",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65438: {
            "Label": "PropB_9E",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65439: {
            "Label": "PropB_9F",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65440: {
            "Label": "PropB_A0",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65441: {
            "Label": "PropB_A1",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65442: {
            "Label": "PropB_A2",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65443: {
            "Label": "PropB_A3",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65444: {
            "Label": "PropB_A4",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65445: {
            "Label": "PropB_A5",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65446: {
            "Label": "PropB_A6",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65447: {
            "Label": "PropB_A7",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65448: {
            "Label": "PropB_A8",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65449: {
            "Label": "PropB_A9",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65450: {
            "Label": "PropB_AA",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65451: {
            "Label": "PropB_AB",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65452: {
            "Label": "PropB_AC",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65453: {
            "Label": "PropB_AD",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65454: {
            "Label": "PropB_AE",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65455: {
            "Label": "PropB_AF",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65456: {
            "Label": "PropB_B0",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65457: {
            "Label": "PropB_B1",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65458: {
            "Label": "PropB_B2",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65459: {
            "Label": "PropB_B3",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65460: {
            "Label": "PropB_B4",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65461: {
            "Label": "PropB_B5",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65462: {
            "Label": "PropB_B6",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65463: {
            "Label": "PropB_B7",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65464: {
            "Label": "PropB_B8",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65465: {
            "Label": "PropB_B9",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65466: {
            "Label": "PropB_BA",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65467: {
            "Label": "PropB_BB",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65468: {
            "Label": "PropB_BC",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65469: {
            "Label": "PropB_BD",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65470: {
            "Label": "PropB_BE",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65471: {
            "Label": "PropB_BF",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65472: {
            "Label": "PropB_C0",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65473: {
            "Label": "PropB_C1",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65474: {
            "Label": "PropB_C2",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65475: {
            "Label": "PropB_C3",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65476: {
            "Label": "PropB_C4",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65477: {
            "Label": "PropB_C5",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65478: {
            "Label": "PropB_C6",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65479: {
            "Label": "PropB_C7",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65480: {
            "Label": "PropB_C8",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65481: {
            "Label": "PropB_C9",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65482: {
            "Label": "PropB_CA",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65483: {
            "Label": "PropB_CB",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65484: {
            "Label": "PropB_CC",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65485: {
            "Label": "PropB_CD",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65486: {
            "Label": "PropB_CE",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65487: {
            "Label": "PropB_CF",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65488: {
            "Label": "PropB_D0",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65489: {
            "Label": "PropB_D1",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65490: {
            "Label": "PropB_D2",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65491: {
            "Label": "PropB_D3",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65492: {
            "Label": "PropB_D4",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65493: {
            "Label": "PropB_D5",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65494: {
            "Label": "PropB_D6",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65495: {
            "Label": "PropB_D7",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65496: {
            "Label": "PropB_D8",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65497: {
            "Label": "PropB_D9",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65498: {
            "Label": "PropB_DA",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65499: {
            "Label": "PropB_DB",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65500: {
            "Label": "PropB_DC",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65501: {
            "Label": "PropB_DD",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65502: {
            "Label": "PropB_DE",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65503: {
            "Label": "PropB_DF",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65504: {
            "Label": "PropB_E0",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65505: {
            "Label": "PropB_E1",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65506: {
            "Label": "PropB_E2",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65507: {
            "Label": "PropB_E3",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65508: {
            "Label": "PropB_E4",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65509: {
            "Label": "PropB_E5",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65510: {
            "Label": "PropB_E6",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65511: {
            "Label": "PropB_E7",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65512: {
            "Label": "PropB_E8",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65513: {
            "Label": "PropB_E9",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65514: {
            "Label": "PropB_EA",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65515: {
            "Label": "PropB_EB",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65516: {
            "Label": "PropB_EC",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65517: {
            "Label": "PropB_ED",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65518: {
            "Label": "PropB_EE",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65519: {
            "Label": "PropB_EF",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65520: {
            "Label": "PropB_F0",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65521: {
            "Label": "PropB_F1",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65522: {
            "Label": "PropB_F2",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65523: {
            "Label": "PropB_F3",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65524: {
            "Label": "PropB_F4",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65525: {
            "Label": "PropB_F5",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65526: {
            "Label": "PropB_F6",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65527: {
            "Label": "PropB_F7",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65528: {
            "Label": "PropB_F8",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65529: {
            "Label": "PropB_F9",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65530: {
            "Label": "PropB_FA",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65531: {
            "Label": "PropB_FB",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65532: {
            "Label": "PropB_FC",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65533: {
            "Label": "PropB_FD",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65534: {
            "Label": "PropB_FE",
            "Name": "Proprietary B",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        65535: {
            "Label": "PropB_FF",
            "Name": "Proprietary B (last entry)",
            "PGNLength": "Variable",
            "Rate": "Per user requirements",
            "SPNs": [
                2551
            ]
        },
        768: {
            "Label": "EBS21",
            "Name": "Reserved for ISO 11992",
            "PGNLength": 8,
            "Rate": "10 ms +/- 1 ms",
            "SPNs": []
        }
    }

J1939SAHWTabledb = {
        128: "thru 159 are reserved for future assignment by SAE but available for use by self configurable ECUs",
        160: "Advanced emergency braking system",
        161: "Fifth Wheel Smart Systems",
        162: "Slope Sensor",
        163: "Catalyst Fluid Sensor",
        164: "On Board Diagnostic Unit #2",
        165: "Rear Steering Axle Controller #2",
        166: "Rear Steering Axle Controller #3",
        167: "Instrument Cluster #2",
        168: "Trailer #5 Bridge",
        169: "Trailer #5 Lighting-electrical",
        170: "Trailer #5 Brakes (ABS-EBS)",
        171: "Trailer #5 Reefer",
        172: "Trailer #5 Cargo",
        173: "Trailer #5 Chassis-Suspension",
        174: "Other Trailer #5 Devices",
        175: "Other Trailer #5 Devices",
        176: "Trailer #4 Bridge",
        177: "Trailer #4 Lighting-electrical",
        178: "Trailer #4 Brakes (ABS-EBS)",
        179: "Trailer #4 Reefer",
        180: "Trailer #4 Cargo",
        181: "Trailer #4 Chassis-Suspension",
        182: "Other Trailer #4 Devices",
        183: "Other Trailer #4 Devices",
        184: "Trailer #3 Bridge",
        185: "Trailer #3 Lighting-electrical",
        186: "Trailer #3 Brakes (ABS-EBS)",
        187: "Trailer #3 Reefer",
        188: "Trailer #3 Cargo",
        189: "Trailer #3 Chassis-Suspension",
        190: "Other Trailer #3 Devices",
        191: "Other Trailer #3 Devices",
        192: "Trailer #2 Bridge",
        193: "Trailer #2 Lighting-electrical",
        194: "Trailer #2 Brakes (ABS-EBS)",
        195: "Trailer #2 Reefer",
        196: "Trailer #2 Cargo",
        197: "Trailer #2 Chassis-Suspension",
        198: "Other Trailer #2 Devices",
        199: "Other Trailer #2 Devices",
        200: "Trailer #1 Bridge",
        201: "Trailer #1 Lighting-electrical",
        202: "Trailer #1 Brakes (ABS-EBS)",
        203: "Trailer #1 Reefer",
        204: "Trailer #1 Cargo",
        205: "Trailer #1 Chassis-Suspension",
        206: "Other Trailer #1 Devices",
        207: "Other Trailer #1 Devices",
        208: "thru 227 are reserved for future assignment by SAE",
        228: "Steering Input Unit",
        229: "Body Controller #2",
        230: "Body-to-Vehicle Interface Control",
        231: "Articulation Turntable Control",
        232: "Forward Road Image Processor",
        233: "Door Controller #3",
        234: "Door Controller #4",
        235: "Tractor/Trailer Bridge #2",
        236: "Door Controller #1",
        237: "Door Controller #2",
        238: "Tachograph",
        239: "Electric Propulsion Control Unit #1",
        240: "Electric Propulsion Control Unit #2",
        241: "Electric Propulsion Control Unit #3",
        242: "Electric Propulsion Control Unit #4",
        243: "Battery Pack Monitor #1",
        244: "Battery Pack Monitor #2 / APU #4",
        245: "Battery Pack Monitor #3 / APU #3",
        246: "Battery Pack Monitor #4 / APU #2",
        247: "Auxiliary Power Unit (APU) #1"
    }

J1939SATabledb = {
        0: "Engine #1",
        1: "Engine #2",
        10: "Axle - Drive #2",
        11: "Brakes - System Controller",
        12: "Brakes - Steer Axle",
        13: "Brakes - Drive axle #1",
        14: "Brakes - Drive Axle #2",
        15: "Retarder - Engine",
        16: "Retarder - Driveline",
        17: "Cruise Control",
        18: "Fuel System",
        19: "Steering Controller",
        2: "Turbocharger",
        20: "Suspension - Steer Axle",
        21: "Suspension - Drive Axle #1",
        22: "Suspension - Drive Axle #2",
        23: "Instrument Cluster #1",
        24: "Trip Recorder",
        248: "File Server / Printer",
        249: "Off Board Diagnostic-Service Tool #1",
        25: "Passenger-Operator Climate Control #1",
        250: "Off Board Diagnostic-Service Tool #2",
        251: "On-Board Data Logger",
        252: "Reserved for Experimental Use",
        253: "Reserved for OEM",
        254: "Null Address",
        255: "GLOBAL (All-Any Node)",
        26: "Alternator/Electrical Charging System",
        27: "Aerodynamic Control",
        28: "Vehicle Navigation",
        29: "Vehicle Security",
        3: "Transmission #1",
        30: "Electrical System",
        31: "Starter System",
        32: "Tractor-Trailer Bridge #1",
        33: "Body Controller",
        34: "Auxiliary Valve Control or Engine Air System Valve Control",
        35: "Hitch Control",
        36: "Power TakeOff (Front or Secondary)",
        37: "Off Vehicle Gateway",
        38: "Virtual Terminal (in cab)",
        39: "Management Computer #1",
        4: "Transmission #2",
        40: "Cab Display #1",
        41: "Retarder, Exhaust, Engine #1",
        42: "Headway Controller",
        43: "On-Board Diagnostic Unit",
        44: "Retarder, Exhaust, Engine #2",
        45: "Endurance Braking System",
        46: "Hydraulic Pump Controller",
        47: "Suspension - System Controller #1",
        48: "Pneumatic - System Controller",
        49: "Cab Controller - Primary",
        5: "Shift Console - Primary",
        50: "Cab Controller - Secondary",
        51: "Tire Pressure Controller",
        52: "Ignition Control Module #1",
        53: "Ignition Control Module #2",
        54: "Seat Control #1",
        55: "Lighting - Operator Controls",
        56: "Rear Axle Steering Controller #1",
        57: "Water Pump Controller",
        58: "Passenger-Operator Climate Control #2",
        59: "Transmission Display - Primary",
        6: "Shift Console - Secondary",
        60: "Transmission Display - Secondary",
        61: "Exhaust Emission Controller",
        62: "Vehicle Dynamic Stability Controller",
        63: "Oil Sensor",
        64: "Suspension - System Controller #2",
        65: "Information System Controller #1",
        66: "Ramp Control",
        67: "Clutch/Converter Unit",
        68: "Auxiliary Heater #1",
        69: "Auxiliary Heater #2",
        7: "Power TakeOff - (Main or Rear)",
        70: "Engine Valve Controller",
        71: "Chassis Controller #1",
        72: "Chassis Controller #2",
        73: "Propulsion Battery Charger",
        74: "Communications Unit, Cellular",
        75: "Communications Unit, Satellite",
        76: "Communications Unit, Radio",
        77: "Steering Column Unit",
        78: "Fan Drive Controller",
        79: "Seat Control #2",
        8: "Axle - Steering",
        80: "Parking brake controller",
        81: "Aftertreatment #1 system gas intake",
        82: "Aftertreatment #1 system gas outlet",
        83: "Safety Restraint System",
        84: "Cab Display #2",
        85: "DPF Controller",
        86: "Aftertreatment #2 system gas intake",
        87: "Aftertreatment #2 system gas outlet",
        88: "Safety Restraint System #S",
        89: "Atmospheric Sensor",
        9: "Axle - Drive #1",
        90: "thru 127 are reserved for future assignment by SAE"
    }

J1939SPNdb = {
        100: {
            "Acronym": "EFL/P1",
            "DataRange": "0.0 to 145.037738 psi",
            "EndBit": 31,
            "Name": "Engine Oil Pressure",
            "Offset": 0.0,
            "OperationalHigh": 145.037738,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.580150952,
            "SPN": 100,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "0.5 s",
            "Units": "psi"
        },
        1000: {
            "Acronym": "TDI",
            "DataRange": "0.0 to 327080468.2212549 mile",
            "EndBit": 95,
            "Name": "Trip Distance in Top Gear",
            "Offset": 0.0,
            "OperationalHigh": 327080468.2212549,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 12,
            "Resolution": 0.077671375,
            "SPN": 1000,
            "SPNLength": 32,
            "StartBit": 64,
            "TransmissionRate": "On request",
            "Units": "miles"
        },
        1001: {
            "Acronym": "LTFI",
            "DataRange": "0.0 to 556224873.36449 US Gal",
            "EndBit": 31,
            "Name": "Trip Drive Fuel Used",
            "Offset": 0.0,
            "OperationalHigh": 556224873.36449,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 22,
            "Resolution": 0.132086,
            "SPN": 1001,
            "SPNLength": 32,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "gallons"
        },
        1002: {
            "Acronym": "LTFI",
            "DataRange": "0.0 to 556224873.36449 US Gal",
            "EndBit": 63,
            "Name": "Trip PTO Governor Moving Fuel Used",
            "Offset": 0.0,
            "OperationalHigh": 556224873.36449,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 22,
            "Resolution": 0.132086,
            "SPN": 1002,
            "SPNLength": 32,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "gallons"
        },
        1003: {
            "Acronym": "LTFI",
            "DataRange": "0.0 to 556224873.36449 US Gal",
            "EndBit": 95,
            "Name": "Trip PTO Governor Non-moving Fuel Used",
            "Offset": 0.0,
            "OperationalHigh": 556224873.36449,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 22,
            "Resolution": 0.132086,
            "SPN": 1003,
            "SPNLength": 32,
            "StartBit": 64,
            "TransmissionRate": "On request",
            "Units": "gallons"
        },
        1004: {
            "Acronym": "LTFI",
            "DataRange": "0.0 to 556224873.36449 US Gal",
            "EndBit": 127,
            "Name": "Trip Vehicle Idle Fuel Used",
            "Offset": 0.0,
            "OperationalHigh": 556224873.36449,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 22,
            "Resolution": 0.132086,
            "SPN": 1004,
            "SPNLength": 32,
            "StartBit": 96,
            "TransmissionRate": "On request",
            "Units": "gallons"
        },
        1005: {
            "Acronym": "LTFI",
            "DataRange": "0.0 to 556224873.36449 US Gal",
            "EndBit": 159,
            "Name": "Trip Cruise Fuel Used",
            "Offset": 0.0,
            "OperationalHigh": 556224873.36449,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 22,
            "Resolution": 0.132086,
            "SPN": 1005,
            "SPNLength": 32,
            "StartBit": 128,
            "TransmissionRate": "On request",
            "Units": "gallons"
        },
        1006: {
            "Acronym": "LTFI",
            "DataRange": "0.0 to 295.1897076301758 mile/US Gal",
            "EndBit": 175,
            "Name": "Trip Drive Fuel Economy",
            "Offset": 0.0,
            "OperationalHigh": 295.1897076301758,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 22,
            "Resolution": 0.0045940348242187,
            "SPN": 1006,
            "SPNLength": 16,
            "StartBit": 160,
            "TransmissionRate": "On request",
            "Units": "miles/gallon"
        },
        1007: {
            "Acronym": "GTFI",
            "DataRange": "0.0 to 4641916934.106649 lb",
            "EndBit": 31,
            "Name": "Trip Drive Fuel Used (Gaseous)",
            "Offset": 0.0,
            "OperationalHigh": 4641916934.106649,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 22,
            "Resolution": 1.10231,
            "SPN": 1007,
            "SPNLength": 32,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "lb"
        },
        1008: {
            "Acronym": "GTFI",
            "DataRange": "0.0 to 4641916934.106649 lb",
            "EndBit": 63,
            "Name": "Trip PTO Governor Moving Fuel Used (Gaseous)",
            "Offset": 0.0,
            "OperationalHigh": 4641916934.106649,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 22,
            "Resolution": 1.10231,
            "SPN": 1008,
            "SPNLength": 32,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "lb"
        },
        1009: {
            "Acronym": "GTFI",
            "DataRange": "0.0 to 4641916934.106649 lb",
            "EndBit": 95,
            "Name": "Trip PTO Governor Non-moving Fuel Used (Gaseous)",
            "Offset": 0.0,
            "OperationalHigh": 4641916934.106649,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 22,
            "Resolution": 1.10231,
            "SPN": 1009,
            "SPNLength": 32,
            "StartBit": 64,
            "TransmissionRate": "On request",
            "Units": "lb"
        },
        101: {
            "Acronym": "EFL/P1",
            "DataRange": "-36.2594345 to 36.54805959862 psi",
            "EndBit": 47,
            "Name": "Engine Crankcase Pressure",
            "Offset": -36.2594345,
            "OperationalHigh": 36.54805959862,
            "OperationalLow": -36.2594345,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.001133107,
            "SPN": 101,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "0.5 s",
            "Units": "psi"
        },
        1010: {
            "Acronym": "GTFI",
            "DataRange": "0.0 to 4641916934.106649 lb",
            "EndBit": 127,
            "Name": "Trip Vehicle Idle Fuel Used (Gaseous)",
            "Offset": 0.0,
            "OperationalHigh": 4641916934.106649,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 22,
            "Resolution": 1.10231,
            "SPN": 1010,
            "SPNLength": 32,
            "StartBit": 96,
            "TransmissionRate": "On request",
            "Units": "lb"
        },
        1011: {
            "Acronym": "GTFI",
            "DataRange": "0.0 to 4641916934.106649 lb",
            "EndBit": 159,
            "Name": "Trip Cruise Fuel Used (Gaseous)",
            "Offset": 0.0,
            "OperationalHigh": 4641916934.106649,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 22,
            "Resolution": 1.10231,
            "SPN": 1011,
            "SPNLength": 32,
            "StartBit": 128,
            "TransmissionRate": "On request",
            "Units": "lb"
        },
        1012: {
            "Acronym": "GTFI",
            "DataRange": "0.0 to 35.37155925273437 mile/lb",
            "EndBit": 175,
            "Name": "Trip Drive Fuel Economy (Gaseous)",
            "Offset": 0.0,
            "OperationalHigh": 35.37155925273437,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 22,
            "Resolution": 0.000550487265625,
            "SPN": 1012,
            "SPNLength": 16,
            "StartBit": 160,
            "TransmissionRate": "On request",
            "Units": "miles/lb"
        },
        1013: {
            "Acronym": "LF",
            "DataRange": "0 to 8,031.875 rpm",
            "EndBit": 15,
            "Name": "Trip Maximum Engine Speed",
            "Offset": 0.0,
            "OperationalHigh": 8031.875,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 10,
            "Resolution": 0.125,
            "SPN": 1013,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "rpm"
        },
        1014: {
            "Acronym": "LF",
            "DataRange": "0 to 8,031.875 rpm",
            "EndBit": 31,
            "Name": "Trip Average Engine Speed",
            "Offset": 0.0,
            "OperationalHigh": 8031.875,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 10,
            "Resolution": 0.125,
            "SPN": 1014,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "rpm"
        },
        1015: {
            "Acronym": "LF",
            "DataRange": "0 to 100 %",
            "EndBit": 39,
            "Name": "Trip Drive Average Load Factor",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 10,
            "Resolution": 0.4,
            "SPN": 1015,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        1016: {
            "Acronym": "LF",
            "DataRange": "0 to 100 %",
            "EndBit": 47,
            "Name": "Total Drive Average Load Factor",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 10,
            "Resolution": 0.4,
            "SPN": 1016,
            "SPNLength": 8,
            "StartBit": 40,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        1017: {
            "Acronym": "LF",
            "DataRange": "0 to 210,554,060.75 hr",
            "EndBit": 79,
            "Name": "Total Engine Cruise Time",
            "Offset": 0.0,
            "OperationalHigh": 210554060.75,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 10,
            "Resolution": 0.05,
            "SPN": 1017,
            "SPNLength": 32,
            "StartBit": 48,
            "TransmissionRate": "On request",
            "Units": "hours"
        },
        1018: {
            "Acronym": "TVI",
            "DataRange": "0.0 to 155.961635516 mph",
            "EndBit": 15,
            "Name": "Trip Maximum Vehicle Speed",
            "Offset": 0.0,
            "OperationalHigh": 155.961635516,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.00242723046875,
            "SPN": 1018,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "mph"
        },
        1019: {
            "Acronym": "TVI",
            "DataRange": "0.0 to 327080468.2212549 mile",
            "EndBit": 47,
            "Name": "Trip Cruise Distance",
            "Offset": 0.0,
            "OperationalHigh": 327080468.2212549,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.077671375,
            "SPN": 1019,
            "SPNLength": 32,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "miles"
        },
        102: {
            "Acronym": "IC1",
            "DataRange": "0.0 to 72.518869 psi",
            "EndBit": 15,
            "Name": "Engine Intake Manifold #1 Pressure",
            "Offset": 0.0,
            "OperationalHigh": 72.518869,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.290075476,
            "SPN": 102,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "0.5 s",
            "Units": "psi"
        },
        1020: {
            "Acronym": "TSI",
            "DataRange": "0 to 64,255",
            "EndBit": 15,
            "Name": "Trip Number of Hot Shutdowns",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1020,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "count"
        },
        1021: {
            "Acronym": "TSI",
            "DataRange": "0 to 64,255",
            "EndBit": 31,
            "Name": "Trip Number of Idle Shutdowns",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1021,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "count"
        },
        1022: {
            "Acronym": "TSI",
            "DataRange": "0 to 64,255",
            "EndBit": 47,
            "Name": "Trip Number of Idle Shutdown Overrides",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1022,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "count"
        },
        1023: {
            "Acronym": "TSI",
            "DataRange": "0 to 64,255",
            "EndBit": 63,
            "Name": "Trip Sudden Decelerations",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1023,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "On request",
            "Units": "count"
        },
        1024: {
            "Acronym": "TTI1",
            "DataRange": "0 to 210,554,060.75 hr",
            "EndBit": 31,
            "Name": "Trip Time in VSL",
            "Offset": 0.0,
            "OperationalHigh": 210554060.75,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 16,
            "Resolution": 0.05,
            "SPN": 1024,
            "SPNLength": 32,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "hours"
        },
        1025: {
            "Acronym": "TTI1",
            "DataRange": "0 to 210,554,060.75 hr",
            "EndBit": 63,
            "Name": "Trip Time in Top Gear",
            "Offset": 0.0,
            "OperationalHigh": 210554060.75,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 16,
            "Resolution": 0.05,
            "SPN": 1025,
            "SPNLength": 32,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "hours"
        },
        1026: {
            "Acronym": "TTI1",
            "DataRange": "0 to 210,554,060.75 hr",
            "EndBit": 95,
            "Name": "Trip Time in Gear Down",
            "Offset": 0.0,
            "OperationalHigh": 210554060.75,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 16,
            "Resolution": 0.05,
            "SPN": 1026,
            "SPNLength": 32,
            "StartBit": 64,
            "TransmissionRate": "On request",
            "Units": "hours"
        },
        1027: {
            "Acronym": "TTI1",
            "DataRange": "0 to 210,554,060.75 hr",
            "EndBit": 127,
            "Name": "Trip Time in Derate by Engine",
            "Offset": 0.0,
            "OperationalHigh": 210554060.75,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 16,
            "Resolution": 0.05,
            "SPN": 1027,
            "SPNLength": 32,
            "StartBit": 96,
            "TransmissionRate": "On request",
            "Units": "hours"
        },
        1028: {
            "Acronym": "LFI",
            "DataRange": "0.0 to 556224873.36449 US Gal",
            "EndBit": 31,
            "Name": "Total Engine PTO Governor Fuel Used",
            "Offset": 0.0,
            "OperationalHigh": 556224873.36449,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.132086,
            "SPN": 1028,
            "SPNLength": 32,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "gallons"
        },
        1029: {
            "Acronym": "LFI",
            "DataRange": "0.0 to 848.718593 US Gal/h",
            "EndBit": 47,
            "Name": "Trip Average Fuel Rate",
            "Offset": 0.0,
            "OperationalHigh": 848.718593,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0132086,
            "SPN": 1029,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "gallons/h"
        },
        103: {
            "Acronym": "TC",
            "DataRange": "0 to 257,020 rpm",
            "EndBit": 23,
            "Name": "Engine Turbocharger 1 Speed",
            "Offset": 0.0,
            "OperationalHigh": 257020.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 4,
            "SPN": 103,
            "SPNLength": 16,
            "StartBit": 8,
            "TransmissionRate": "1 s",
            "Units": "rpm"
        },
        1030: {
            "Acronym": "GFI1",
            "DataRange": "0 to 2,105,540,607.5 kg",
            "EndBit": 31,
            "Name": "Total Engine PTO Governor Fuel Used (Gaseous)",
            "Offset": 0.0,
            "OperationalHigh": 2105540607.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.5,
            "SPN": 1030,
            "SPNLength": 32,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "kg"
        },
        1031: {
            "Acronym": "GFI1",
            "DataRange": "0.0 to 7082.892905 lb/h",
            "EndBit": 47,
            "Name": "Trip Average Fuel Rate (Gaseous)",
            "Offset": 0.0,
            "OperationalHigh": 7082.892905,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.110231,
            "SPN": 1031,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "lb/h"
        },
        1032: {
            "Acronym": "EH",
            "DataRange": "0 to 526,385,151.9 km",
            "EndBit": 31,
            "Name": "Total ECU Distance",
            "Offset": 0.0,
            "OperationalHigh": 526385151.9,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.125,
            "SPN": 1032,
            "SPNLength": 32,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "m"
        },
        1033: {
            "Acronym": "EH",
            "DataRange": "0 to 210,554,060.75 hr",
            "EndBit": 63,
            "Name": "Total ECU Run Time",
            "Offset": 0.0,
            "OperationalHigh": 210554060.75,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05,
            "SPN": 1033,
            "SPNLength": 32,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "hours"
        },
        1034: {
            "Acronym": "TTI2",
            "DataRange": "0 to 210,554,060.75 hr",
            "EndBit": 31,
            "Name": "Trip Cruise Time",
            "Offset": 0.0,
            "OperationalHigh": 210554060.75,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 20,
            "Resolution": 0.05,
            "SPN": 1034,
            "SPNLength": 32,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "hours"
        },
        1035: {
            "Acronym": "TTI2",
            "DataRange": "0 to 210,554,060.75 hr",
            "EndBit": 63,
            "Name": "Trip PTO Governor Time",
            "Offset": 0.0,
            "OperationalHigh": 210554060.75,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 20,
            "Resolution": 0.05,
            "SPN": 1035,
            "SPNLength": 32,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "hours"
        },
        1036: {
            "Acronym": "TTI2",
            "DataRange": "0 to 210,554,060.75 hr",
            "EndBit": 95,
            "Name": "Trip Engine Running Time",
            "Offset": 0.0,
            "OperationalHigh": 210554060.75,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 20,
            "Resolution": 0.05,
            "SPN": 1036,
            "SPNLength": 32,
            "StartBit": 64,
            "TransmissionRate": "On request",
            "Units": "hours"
        },
        1037: {
            "Acronym": "TTI2",
            "DataRange": "0 to 210,554,060.75 hr",
            "EndBit": 127,
            "Name": "Trip Idle Time",
            "Offset": 0.0,
            "OperationalHigh": 210554060.75,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 20,
            "Resolution": 0.05,
            "SPN": 1037,
            "SPNLength": 32,
            "StartBit": 96,
            "TransmissionRate": "On request",
            "Units": "hours"
        },
        1038: {
            "Acronym": "TTI2",
            "DataRange": "0 to 210,554,060.75 hr",
            "EndBit": 159,
            "Name": "Trip Air Compressor On Time",
            "Offset": 0.0,
            "OperationalHigh": 210554060.75,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 20,
            "Resolution": 0.05,
            "SPN": 1038,
            "SPNLength": 32,
            "StartBit": 128,
            "TransmissionRate": "On request",
            "Units": "hours"
        },
        1039: {
            "Acronym": "GFC",
            "DataRange": "0.0 to 4641916934.106649 lb",
            "EndBit": 31,
            "Name": "Trip Fuel (Gaseous)",
            "Offset": 0.0,
            "OperationalHigh": 4641916934.106649,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.10231,
            "SPN": 1039,
            "SPNLength": 32,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "lb"
        },
        104: {
            "Acronym": "TC",
            "DataRange": "0.0 to 145.037738 psi",
            "EndBit": 7,
            "Name": "Engine Turbocharger Lube Oil Pressure 1",
            "Offset": 0.0,
            "OperationalHigh": 145.037738,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.580150952,
            "SPN": 104,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "psi"
        },
        1040: {
            "Acronym": "GFC",
            "DataRange": "0 to 2,105,540,607.5 kg",
            "EndBit": 63,
            "Name": "Total Fuel Used (Gaseous)",
            "Offset": 0.0,
            "OperationalHigh": 2105540607.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.5,
            "SPN": 1040,
            "SPNLength": 32,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "kg"
        },
        105: {
            "Acronym": "IC1",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 23,
            "Name": "Engine Intake Manifold 1 Temperature",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 105,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "0.5 s",
            "Units": "deg F"
        },
        106: {
            "Acronym": "IC1",
            "DataRange": "0.0 to 72.518869 psi",
            "EndBit": 31,
            "Name": "Engine Air Intake Pressure",
            "Offset": 0.0,
            "OperationalHigh": 72.518869,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.290075476,
            "SPN": 106,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "0.5 s",
            "Units": "psi"
        },
        107: {
            "Acronym": "IC1",
            "DataRange": "0.0 to 1.812971725 psi",
            "EndBit": 39,
            "Name": "Engine Air Filter 1 Differential Pressure",
            "Offset": 0.0,
            "OperationalHigh": 1.812971725,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0072518869,
            "SPN": 107,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "0.5 s",
            "Units": "psi"
        },
        108: {
            "Acronym": "AMB",
            "DataRange": "0.0 to 18.12971725 psi",
            "EndBit": 7,
            "Name": "Barometric Pressure",
            "Offset": 0.0,
            "OperationalHigh": 18.12971725,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.072518869,
            "SPN": 108,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "psi"
        },
        1081: {
            "Acronym": "SHUTDN",
            "DataRange": "0 to 3",
            "EndBit": 25,
            "Name": "Engine Wait to Start Lamp",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1081,
            "SPNLength": 2,
            "StartBit": 24,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        1082: {
            "Acronym": "ERC1",
            "DataRange": "0 to 3",
            "EndBit": 25,
            "Name": "Engine Coolant Load Increase",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1082,
            "SPNLength": 2,
            "StartBit": 24,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1083: {
            "Acronym": "AUXIO1",
            "DataRange": "0 to 64,255",
            "EndBit": 47,
            "Name": "Auxiliary I/O Channel #1",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1083,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "manufacturer defined, not faster than 100 ms",
            "Units": "count"
        },
        1084: {
            "Acronym": "AUXIO1",
            "DataRange": "0 to 64,255",
            "EndBit": 63,
            "Name": "Auxiliary I/O Channel #2",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1084,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "manufacturer defined, not faster than 100 ms",
            "Units": "count"
        },
        1085: {
            "Acronym": "ERC1",
            "DataRange": "-125 to 125 %",
            "EndBit": 23,
            "Name": "Intended Retarder Percent Torque",
            "Offset": -125.0,
            "OperationalHigh": 0.0,
            "OperationalLow": -125.0,
            "OperationalRange": "-125 to 0%",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1085,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        1086: {
            "Acronym": "AIR1",
            "DataRange": "0.0 to 290.075476 psi",
            "EndBit": 15,
            "Name": "Parking and/or Trailer Air Pressure",
            "Offset": 0.0,
            "OperationalHigh": 290.075476,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.160301904,
            "SPN": 1086,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "1 sec",
            "Units": "psi"
        },
        1087: {
            "Acronym": "AIR1",
            "DataRange": "0.0 to 290.075476 psi",
            "EndBit": 23,
            "Name": "Service Brake Circuit 1 Air Pressure",
            "Offset": 0.0,
            "OperationalHigh": 290.075476,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.160301904,
            "SPN": 1087,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "1 sec",
            "Units": "psi"
        },
        1088: {
            "Acronym": "AIR1",
            "DataRange": "0.0 to 290.075476 psi",
            "EndBit": 31,
            "Name": "Service Brake Circuit 2 Air Pressure",
            "Offset": 0.0,
            "OperationalHigh": 290.075476,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.160301904,
            "SPN": 1088,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "1 sec",
            "Units": "psi"
        },
        1089: {
            "Acronym": "AIR1",
            "DataRange": "0.0 to 290.075476 psi",
            "EndBit": 39,
            "Name": "Auxiliary Equipment Supply Pressure",
            "Offset": 0.0,
            "OperationalHigh": 290.075476,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.160301904,
            "SPN": 1089,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "1 sec",
            "Units": "psi"
        },
        109: {
            "Acronym": "EFL/P1",
            "DataRange": "0.0 to 72.518869 psi",
            "EndBit": 55,
            "Name": "Engine Coolant Pressure",
            "Offset": 0.0,
            "OperationalHigh": 72.518869,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.290075476,
            "SPN": 109,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "0.5 s",
            "Units": "psi"
        },
        1090: {
            "Acronym": "AIR1",
            "DataRange": "0.0 to 290.075476 psi",
            "EndBit": 47,
            "Name": "Air Suspension Supply Pressure",
            "Offset": 0.0,
            "OperationalHigh": 290.075476,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.160301904,
            "SPN": 1090,
            "SPNLength": 8,
            "StartBit": 40,
            "TransmissionRate": "1 sec",
            "Units": "psi"
        },
        1091: {
            "Acronym": "EBC3",
            "DataRange": "0.0 to 181.2971725 psi",
            "EndBit": 7,
            "Name": "Brake Application Pressure High Range, Front Axle, Left Wheel",
            "Offset": 0.0,
            "OperationalHigh": 181.2971725,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.72518869,
            "SPN": 1091,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        1092: {
            "Acronym": "EBC3",
            "DataRange": "0.0 to 181.2971725 psi",
            "EndBit": 15,
            "Name": "Brake Application Pressure High Range, Front Axle, Right Wheel",
            "Offset": 0.0,
            "OperationalHigh": 181.2971725,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.72518869,
            "SPN": 1092,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        1093: {
            "Acronym": "EBC3",
            "DataRange": "0.0 to 181.2971725 psi",
            "EndBit": 23,
            "Name": "Brake Application Pressure High Range, Rear Axle #1, Left Wheel",
            "Offset": 0.0,
            "OperationalHigh": 181.2971725,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.72518869,
            "SPN": 1093,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        1094: {
            "Acronym": "EBC3",
            "DataRange": "0.0 to 181.2971725 psi",
            "EndBit": 31,
            "Name": "Brake Application Pressure High Range, Rear Axle #1, Right Wheel",
            "Offset": 0.0,
            "OperationalHigh": 181.2971725,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.72518869,
            "SPN": 1094,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        1095: {
            "Acronym": "EBC3",
            "DataRange": "0.0 to 181.2971725 psi",
            "EndBit": 39,
            "Name": "Brake Application Pressure High Range, Rear Axle #2, Left Wheel",
            "Offset": 0.0,
            "OperationalHigh": 181.2971725,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.72518869,
            "SPN": 1095,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        1096: {
            "Acronym": "EBC3",
            "DataRange": "0.0 to 181.2971725 psi",
            "EndBit": 47,
            "Name": "Brake Application Pressure High Range, Rear Axle #2, Right Wheel",
            "Offset": 0.0,
            "OperationalHigh": 181.2971725,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.72518869,
            "SPN": 1096,
            "SPNLength": 8,
            "StartBit": 40,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        1097: {
            "Acronym": "EBC3",
            "DataRange": "0.0 to 181.2971725 psi",
            "EndBit": 55,
            "Name": "Brake Application Pressure High Range, Rear Axle #3, Left Wheel",
            "Offset": 0.0,
            "OperationalHigh": 181.2971725,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.72518869,
            "SPN": 1097,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        1098: {
            "Acronym": "EBC3",
            "DataRange": "0.0 to 181.2971725 psi",
            "EndBit": 63,
            "Name": "Brake Application Pressure High Range, Rear Axle #3, Right Wheel",
            "Offset": 0.0,
            "OperationalHigh": 181.2971725,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.72518869,
            "SPN": 1098,
            "SPNLength": 8,
            "StartBit": 56,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        1099: {
            "Acronym": "EBC4",
            "DataRange": "0 to 100 %",
            "EndBit": 7,
            "Name": "Brake Lining Remaining, Front Axle, Left Wheel",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 1099,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        110: {
            "Acronym": "ET1",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 7,
            "Name": "Engine Coolant Temperature",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 110,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1100: {
            "Acronym": "EBC4",
            "DataRange": "0 to 100 %",
            "EndBit": 15,
            "Name": "Brake Lining Remaining, Front Axle, Right Wheel",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 1100,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        1101: {
            "Acronym": "EBC4",
            "DataRange": "0 to 100 %",
            "EndBit": 23,
            "Name": "Brake Lining Remaining, Rear Axle #1, Left Wheel",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 1101,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        1102: {
            "Acronym": "EBC4",
            "DataRange": "0 to 100 %",
            "EndBit": 31,
            "Name": "Brake Lining Remaining, Rear Axle #1, Right Wheel",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 1102,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        1103: {
            "Acronym": "EBC4",
            "DataRange": "0 to 100 %",
            "EndBit": 39,
            "Name": "Brake Lining Remaining, Rear Axle #2, Left Wheel",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 1103,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        1104: {
            "Acronym": "EBC4",
            "DataRange": "0 to 100 %",
            "EndBit": 47,
            "Name": "Brake Lining Remaining, Rear Axle #2, Right Wheel",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 1104,
            "SPNLength": 8,
            "StartBit": 40,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        1105: {
            "Acronym": "EBC4",
            "DataRange": "0 to 100 %",
            "EndBit": 55,
            "Name": "Brake Lining Remaining, Rear Axle #3, Left Wheel",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 1105,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        1106: {
            "Acronym": "EBC4",
            "DataRange": "0 to 100 %",
            "EndBit": 63,
            "Name": "Brake Lining Remaining, Rear Axle #3, Right Wheel",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 1106,
            "SPNLength": 8,
            "StartBit": 56,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        1107: {
            "Acronym": "SHUTDN",
            "DataRange": "0 to 3",
            "EndBit": 39,
            "Name": "Engine Protection System Timer State",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1107,
            "SPNLength": 2,
            "StartBit": 38,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        1108: {
            "Acronym": "SHUTDN",
            "DataRange": "0 to 3",
            "EndBit": 37,
            "Name": "Engine Protection System Timer Override",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1108,
            "SPNLength": 2,
            "StartBit": 36,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        1109: {
            "Acronym": "SHUTDN",
            "DataRange": "0 to 3",
            "EndBit": 35,
            "Name": "Engine Protection System Approaching Shutdown",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1109,
            "SPNLength": 2,
            "StartBit": 34,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        111: {
            "Acronym": "EFL/P1",
            "DataRange": "0 to 100 %",
            "EndBit": 63,
            "Name": "Engine Coolant Level",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 111,
            "SPNLength": 8,
            "StartBit": 56,
            "TransmissionRate": "0.5 s",
            "Units": "%"
        },
        1110: {
            "Acronym": "SHUTDN",
            "DataRange": "0 to 3",
            "EndBit": 33,
            "Name": "Engine Protection System has Shutdown Engine",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1110,
            "SPNLength": 2,
            "StartBit": 32,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        1111: {
            "Acronym": "SHUTDN",
            "DataRange": "0 to 3",
            "EndBit": 47,
            "Name": "Engine Protection System Configuration",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1111,
            "SPNLength": 2,
            "StartBit": 46,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        1113: {
            "Acronym": "ETC6",
            "DataRange": "-125 to 125",
            "EndBit": 7,
            "Name": "Recommended Gear",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "-125 to +125",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1113,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "gear value"
        },
        1114: {
            "Acronym": "ETC6",
            "DataRange": "-125 to 125",
            "EndBit": 23,
            "Name": "Lowest Possible Gear",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "-125 to +125",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1114,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "gear value"
        },
        1115: {
            "Acronym": "ETC6",
            "DataRange": "-125 to 125",
            "EndBit": 15,
            "Name": "Highest Possible Gear",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "-125 to +125",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1115,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "On request",
            "Units": "gear value"
        },
        1116: {
            "Acronym": "AF2",
            "DataRange": "0 to 250 %",
            "EndBit": 7,
            "Name": "Engine Gaseous Fuel Correction Factor",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1116,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        1117: {
            "Acronym": "EO1",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 15,
            "Name": "Engine Desired Rated Exhaust Oxygen",
            "Offset": 0.0,
            "OperationalHigh": 160.6375,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 1117,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        1118: {
            "Acronym": "EO1",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 31,
            "Name": "Engine Desired Exhaust Oxygen",
            "Offset": 0.0,
            "OperationalHigh": 160.6375,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 1118,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        1119: {
            "Acronym": "EO1",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 47,
            "Name": "Engine Actual Exhaust Oxygen",
            "Offset": 0.0,
            "OperationalHigh": 160.6375,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 1119,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        112: {
            "Acronym": "IC1",
            "DataRange": "0.0 to 18.12971725 psi",
            "EndBit": 63,
            "Name": "Engine Coolant Filter Differential Pressure",
            "Offset": 0.0,
            "OperationalHigh": 18.12971725,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.072518869,
            "SPN": 112,
            "SPNLength": 8,
            "StartBit": 56,
            "TransmissionRate": "0.5 s",
            "Units": "psi"
        },
        1120: {
            "Acronym": "ACTL",
            "DataRange": "-125 to 125 deg",
            "EndBit": 7,
            "Name": "Articulation Angle",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1120,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "deg"
        },
        1121: {
            "Acronym": "EBC1",
            "DataRange": "0 to 3",
            "EndBit": 7,
            "Name": "EBS Brake Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1121,
            "SPNLength": 2,
            "StartBit": 6,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1122: {
            "Acronym": "AT",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 7,
            "Name": "Engine Alternator Bearing 1 Temperature",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 1122,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1123: {
            "Acronym": "AT",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 15,
            "Name": "Engine Alternator Bearing 2 Temperature",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 1123,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1124: {
            "Acronym": "AT",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 23,
            "Name": "Engine Alternator Winding 1 Temperature",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 1124,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1125: {
            "Acronym": "AT",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 31,
            "Name": "Engine Alternator Winding 2 Temperature",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 1125,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1126: {
            "Acronym": "AT",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 39,
            "Name": "Engine Alternator Winding 3 Temperature",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 1126,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1127: {
            "Acronym": "IMT1",
            "DataRange": "0.0 to 1164.92498189875 psi",
            "EndBit": 15,
            "Name": "Engine Turbocharger 1 Boost Pressure",
            "Offset": 0.0,
            "OperationalHigh": 1164.92498189875,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.01812971725,
            "SPN": 1127,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "0.5 s",
            "Units": "psi"
        },
        1128: {
            "Acronym": "IMT1",
            "DataRange": "0.0 to 1164.92498189875 psi",
            "EndBit": 31,
            "Name": "Engine Turbocharger 2 Boost Pressure",
            "Offset": 0.0,
            "OperationalHigh": 1164.92498189875,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.01812971725,
            "SPN": 1128,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "0.5 s",
            "Units": "psi"
        },
        1129: {
            "Acronym": "IMT1",
            "DataRange": "0.0 to 1164.92498189875 psi",
            "EndBit": 47,
            "Name": "Engine Turbocharger 3 Boost Pressure",
            "Offset": 0.0,
            "OperationalHigh": 1164.92498189875,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.01812971725,
            "SPN": 1129,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "0.5 s",
            "Units": "psi"
        },
        1130: {
            "Acronym": "IMT1",
            "DataRange": "0.0 to 1164.92498189875 psi",
            "EndBit": 63,
            "Name": "Engine Turbocharger 4 Boost Pressure",
            "Offset": 0.0,
            "OperationalHigh": 1164.92498189875,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.01812971725,
            "SPN": 1130,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "0.5 s",
            "Units": "psi"
        },
        1131: {
            "Acronym": "IMT2",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 7,
            "Name": "Engine Intake Manifold 2 Temperature",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 1131,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1132: {
            "Acronym": "IMT2",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 15,
            "Name": "Engine Intake Manifold 3 Temperature",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 1132,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1133: {
            "Acronym": "IMT2",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 23,
            "Name": "Engine Intake Manifold 4 Temperature",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 1133,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1134: {
            "Acronym": "ET1",
            "DataRange": "0 to 100 %",
            "EndBit": 63,
            "Name": "Engine Intercooler Thermostat Opening",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 1134,
            "SPNLength": 8,
            "StartBit": 56,
            "TransmissionRate": "1 s",
            "Units": "%"
        },
        1135: {
            "Acronym": "ET2",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 15,
            "Name": "Engine Oil Temperature 2",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 1135,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "1s",
            "Units": "deg F"
        },
        1136: {
            "Acronym": "ET2",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 31,
            "Name": "Engine ECU Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 1136,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "1s",
            "Units": "deg F"
        },
        1137: {
            "Acronym": "EPT1",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 15,
            "Name": "Engine Exhaust Gas Port 1 Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 1137,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1138: {
            "Acronym": "EPT1",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 31,
            "Name": "Engine Exhaust Gas Port 2 Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 1138,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1139: {
            "Acronym": "EPT1",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 47,
            "Name": "Engine Exhaust Gas Port 3 Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 1139,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        114: {
            "Acronym": "VEP1",
            "DataRange": "-125 to 125 A",
            "EndBit": 7,
            "Name": "Net Battery Current",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 114,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "Amps"
        },
        1140: {
            "Acronym": "EPT1",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 63,
            "Name": "Engine Exhaust Gas Port 4 Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 1140,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1141: {
            "Acronym": "EPT2",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 15,
            "Name": "Engine Exhaust Gas Port 5 Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 1141,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1142: {
            "Acronym": "EPT2",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 31,
            "Name": "Engine Exhaust Gas Port 6 Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 1142,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1143: {
            "Acronym": "EPT2",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 47,
            "Name": "Engine Exhaust Gas Port 7 Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 1143,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1144: {
            "Acronym": "EPT2",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 63,
            "Name": "Engine Exhaust Gas Port 8 Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 1144,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1145: {
            "Acronym": "EPT3",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 15,
            "Name": "Engine Exhaust Gas Port 9 Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 1145,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1146: {
            "Acronym": "EPT3",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 31,
            "Name": "Engine Exhaust Gas Port 10 Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 1146,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1147: {
            "Acronym": "EPT3",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 47,
            "Name": "Engine Exhaust Gas Port 11 Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 1147,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1148: {
            "Acronym": "EPT3",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 63,
            "Name": "Engine Exhaust Gas Port 12 Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 1148,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1149: {
            "Acronym": "EPT4",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 15,
            "Name": "Engine Exhaust Gas Port 13 Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 1149,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        115: {
            "Acronym": "VEP1",
            "DataRange": "0 to 250 A",
            "EndBit": 15,
            "Name": "Alternator Current",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 115,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "1 s",
            "Units": "Amps"
        },
        1150: {
            "Acronym": "EPT4",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 31,
            "Name": "Engine Exhaust Gas Port 14 Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 1150,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1151: {
            "Acronym": "EPT4",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 47,
            "Name": "Engine Exhaust Gas Port 15 Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 1151,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1152: {
            "Acronym": "EPT4",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 63,
            "Name": "Engine Exhaust Gas Port 16 Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 1152,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1153: {
            "Acronym": "EPT5",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 15,
            "Name": "Engine Exhaust Gas Port 17 Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 1153,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1154: {
            "Acronym": "EPT5",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 31,
            "Name": "Engine Exhaust Gas Port 18 Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 1154,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1155: {
            "Acronym": "EPT5",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 47,
            "Name": "Engine Exhaust Gas Port 19 Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 1155,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1156: {
            "Acronym": "EPT5",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 63,
            "Name": "Engine Exhaust Gas Port 20 Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 1156,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1157: {
            "Acronym": "MBT1",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 15,
            "Name": "Engine Main Bearing 1 Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 1157,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1158: {
            "Acronym": "MBT1",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 31,
            "Name": "Engine Main Bearing 2 Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 1158,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1159: {
            "Acronym": "MBT1",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 47,
            "Name": "Engine Main Bearing 3 Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 1159,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        116: {
            "Acronym": "B",
            "DataRange": "0.0 to 145.037738 psi",
            "EndBit": 7,
            "Name": "Brake Application Pressure",
            "Offset": 0.0,
            "OperationalHigh": 145.037738,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.580150952,
            "SPN": 116,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "psi"
        },
        1160: {
            "Acronym": "MBT1",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 63,
            "Name": "Engine Main Bearing 4 Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 1160,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1161: {
            "Acronym": "MBT2",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 15,
            "Name": "Engine Main Bearing 5 Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 1161,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1162: {
            "Acronym": "MBT2",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 31,
            "Name": "Engine Main Bearing 6 Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 1162,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1163: {
            "Acronym": "MBT2",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 47,
            "Name": "Engine Main Bearing 7 Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 1163,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1164: {
            "Acronym": "MBT2",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 63,
            "Name": "Engine Main Bearing 8 Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 1164,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1165: {
            "Acronym": "MBT3",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 15,
            "Name": "Engine Main Bearing 9 Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 1165,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1166: {
            "Acronym": "MBT3",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 31,
            "Name": "Engine Main Bearing 10 Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 1166,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1167: {
            "Acronym": "MBT3",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 47,
            "Name": "Engine Main Bearing 11 Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 1167,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1168: {
            "Acronym": "TCI1",
            "DataRange": "0.0 to 145.037738 psi",
            "EndBit": 7,
            "Name": "Engine Turbocharger Lube Oil Pressure 2",
            "Offset": 0.0,
            "OperationalHigh": 145.037738,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.580150952,
            "SPN": 1168,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "psi"
        },
        1169: {
            "Acronym": "TCI1",
            "DataRange": "0 to 257,020 rpm",
            "EndBit": 23,
            "Name": "Engine Turbocharger 2 Speed",
            "Offset": 0.0,
            "OperationalHigh": 257020.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 4,
            "SPN": 1169,
            "SPNLength": 16,
            "StartBit": 8,
            "TransmissionRate": "1 s",
            "Units": "rpm"
        },
        117: {
            "Acronym": "B",
            "DataRange": "0.0 to 145.037738 psi",
            "EndBit": 15,
            "Name": "Brake Primary Pressure",
            "Offset": 0.0,
            "OperationalHigh": 145.037738,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.580150952,
            "SPN": 117,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "1 s",
            "Units": "psi"
        },
        1170: {
            "Acronym": "TCI1",
            "DataRange": "0 to 257,020 rpm",
            "EndBit": 39,
            "Name": "Engine Turbocharger 3 Speed",
            "Offset": 0.0,
            "OperationalHigh": 257020.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 4,
            "SPN": 1170,
            "SPNLength": 16,
            "StartBit": 24,
            "TransmissionRate": "1 s",
            "Units": "rpm"
        },
        1171: {
            "Acronym": "TCI1",
            "DataRange": "0 to 257,020 rpm",
            "EndBit": 55,
            "Name": "Engine Turbocharger 4 Speed",
            "Offset": 0.0,
            "OperationalHigh": 257020.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 4,
            "SPN": 1171,
            "SPNLength": 16,
            "StartBit": 40,
            "TransmissionRate": "1 s",
            "Units": "rpm"
        },
        1172: {
            "Acronym": "TCI2",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 15,
            "Name": "Engine Turbocharger 1 Compressor Intake Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 1172,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1173: {
            "Acronym": "TCI2",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 31,
            "Name": "Engine Turbocharger 2 Compressor Intake Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 1173,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1174: {
            "Acronym": "TCI2",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 47,
            "Name": "Engine Turbocharger 3 Compressor Intake Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 1174,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1175: {
            "Acronym": "TCI2",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 63,
            "Name": "Engine Turbocharger 4 Compressor Intake Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 1175,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1176: {
            "Acronym": "TCI3",
            "DataRange": "-36.2594345 to 36.54805959862 psi",
            "EndBit": 15,
            "Name": "Engine Turbocharger 1 Compressor Intake Pressure",
            "Offset": -36.2594345,
            "OperationalHigh": 36.54805959862,
            "OperationalLow": -36.2594345,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.001133107,
            "SPN": 1176,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "psi"
        },
        1177: {
            "Acronym": "TCI3",
            "DataRange": "-36.2594345 to 36.54805959862 psi",
            "EndBit": 31,
            "Name": "Engine Turbocharger 2 Compressor Intake Pressure",
            "Offset": -36.2594345,
            "OperationalHigh": 36.54805959862,
            "OperationalLow": -36.2594345,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.001133107,
            "SPN": 1177,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "1 s",
            "Units": "psi"
        },
        1178: {
            "Acronym": "TCI3",
            "DataRange": "-36.2594345 to 36.54805959862 psi",
            "EndBit": 47,
            "Name": "Engine Turbocharger 3 Compressor Intake Pressure",
            "Offset": -36.2594345,
            "OperationalHigh": 36.54805959862,
            "OperationalLow": -36.2594345,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.001133107,
            "SPN": 1178,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "1 s",
            "Units": "psi"
        },
        1179: {
            "Acronym": "TCI3",
            "DataRange": "-36.2594345 to 36.54805959862 psi",
            "EndBit": 63,
            "Name": "Engine Turbocharger 4 Compressor Intake Pressure",
            "Offset": -36.2594345,
            "OperationalHigh": 36.54805959862,
            "OperationalLow": -36.2594345,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.001133107,
            "SPN": 1179,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "1 s",
            "Units": "psi"
        },
        118: {
            "Acronym": "B",
            "DataRange": "0.0 to 145.037738 psi",
            "EndBit": 23,
            "Name": "Brake Secondary Pressure",
            "Offset": 0.0,
            "OperationalHigh": 145.037738,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.580150952,
            "SPN": 118,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "1 s",
            "Units": "psi"
        },
        1180: {
            "Acronym": "TCI4",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 15,
            "Name": "Engine Turbocharger 1 Turbine Intake Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 1180,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1181: {
            "Acronym": "TCI4",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 31,
            "Name": "Engine Turbocharger 2 Turbine Intake Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 1181,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1182: {
            "Acronym": "TCI4",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 47,
            "Name": "Engine Turbocharger 3 Turbine Intake Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 1182,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1183: {
            "Acronym": "TCI4",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 63,
            "Name": "Engine Turbocharger 4 Turbine Intake Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 1183,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1184: {
            "Acronym": "TCI5",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 15,
            "Name": "Engine Turbocharger 1 Turbine Outlet Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 1184,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1185: {
            "Acronym": "TCI5",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 31,
            "Name": "Engine Turbocharger 2 Turbine Outlet Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 1185,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1186: {
            "Acronym": "TCI5",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 47,
            "Name": "Engine Turbocharger 3 Turbine Outlet Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 1186,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1187: {
            "Acronym": "TCI5",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 63,
            "Name": "Engine Turbocharger 4 Turbine Outlet Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 1187,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1188: {
            "Acronym": "TCW",
            "DataRange": "0 to 100 %",
            "EndBit": 7,
            "Name": "Engine Turbocharger Wastegate Actuator 1 Position",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 1188,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        1189: {
            "Acronym": "TCW",
            "DataRange": "0 to 100 %",
            "EndBit": 15,
            "Name": "Engine Turbocharger Wastegate Actuator 2 Position",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 1189,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        119: {
            "Acronym": "RF",
            "DataRange": "0.0 to 580.150952 psi",
            "EndBit": 7,
            "Name": "Hydraulic Retarder Pressure",
            "Offset": 0.0,
            "OperationalHigh": 580.150952,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 2.320603808,
            "SPN": 119,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "psi"
        },
        1190: {
            "Acronym": "TCW",
            "DataRange": "0 to 100 %",
            "EndBit": 23,
            "Name": "Engine Turbocharger Wastegate Actuator 3 Position",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 1190,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        1191: {
            "Acronym": "TCW",
            "DataRange": "0 to 100 %",
            "EndBit": 31,
            "Name": "Engine Turbocharger Wastegate Actuator 4 Position",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 1191,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        1192: {
            "Acronym": "TCW",
            "DataRange": "0.0 to 145.037738 psi",
            "EndBit": 39,
            "Name": "Engine Turbocharger Wastegate Actuator Control Air Pressure",
            "Offset": 0.0,
            "OperationalHigh": 145.037738,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.580150952,
            "SPN": 1192,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        1193: {
            "Acronym": "RBI",
            "DataRange": "0 to 4,211,081,215 s",
            "EndBit": 31,
            "Name": "Engine Operation Time Since Rebuild",
            "Offset": 0.0,
            "OperationalHigh": 4211081215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1193,
            "SPNLength": 32,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "Seconds"
        },
        1194: {
            "Acronym": "ATS",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Anti-theft Encryption Seed Present Indicator",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1194,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "This message is transmitted in response to an Anti-Theft Request message. This message is also sent when the component has an abnormal power interruption.  In this situation the Anti-Theft Status Report is sent without the Anti-Theft Request.",
            "Units": "bit"
        },
        1195: {
            "Acronym": "ATS",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Anti-theft Password Valid Indicator",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1195,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "This message is transmitted in response to an Anti-Theft Request message. This message is also sent when the component has an abnormal power interruption.  In this situation the Anti-Theft Status Report is sent without the Anti-Theft Request.",
            "Units": "bit"
        },
        1196: {
            "Acronym": "ATS",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Anti-theft Component Status States",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1196,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "This message is transmitted in response to an Anti-Theft Request message. This message is also sent when the component has an abnormal power interruption.  In this situation the Anti-Theft Status Report is sent without the Anti-Theft Request.",
            "Units": "bit"
        },
        1197: {
            "Acronym": "ATS",
            "DataRange": "0 to 3",
            "EndBit": 7,
            "Name": "Anti-theft Modify Password States",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1197,
            "SPNLength": 2,
            "StartBit": 6,
            "TransmissionRate": "This message is transmitted in response to an Anti-Theft Request message. This message is also sent when the component has an abnormal power interruption.  In this situation the Anti-Theft Status Report is sent without the Anti-Theft Request.",
            "Units": "bit"
        },
        1198: {
            "Acronym": "ATS",
            "DataRange": "0 to 255 per byte",
            "EndBit": 63,
            "Name": "Anti-theft Random Number",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -2,
            "SPN": 1198,
            "SPNLength": 56,
            "StartBit": 8,
            "TransmissionRate": "This message is transmitted in response to an Anti-Theft Request message. This message is also sent when the component has an abnormal power interruption.  In this situation the Anti-Theft Status Report is sent without the Anti-Theft Request.",
            "Units": "ASCII"
        },
        1199: {
            "Acronym": "ATR",
            "DataRange": "0 to 3",
            "EndBit": 2,
            "Name": "Anti-theft Encryption Indicator States",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1199,
            "SPNLength": 2,
            "StartBit": 1,
            "TransmissionRate": "Transmission of this message is interrupt driven.  This message is also transmitted upon power-up of the interfacing device sending this message.",
            "Units": "bit"
        },
        120: {
            "Acronym": "RF",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 15,
            "Name": "Hydraulic Retarder Oil Temperature",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 120,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1200: {
            "Acronym": "ATR",
            "DataRange": "0 to 3",
            "EndBit": 4,
            "Name": "Anti-theft Desired Exit Mode States",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1200,
            "SPNLength": 2,
            "StartBit": 3,
            "TransmissionRate": "Transmission of this message is interrupt driven.  This message is also transmitted upon power-up of the interfacing device sending this message.",
            "Units": "bit"
        },
        1201: {
            "Acronym": "ATR",
            "DataRange": "0 to 7",
            "EndBit": 7,
            "Name": "Anti-theft Command States",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1201,
            "SPNLength": 3,
            "StartBit": 5,
            "TransmissionRate": "Transmission of this message is interrupt driven.  This message is also transmitted upon power-up of the interfacing device sending this message.",
            "Units": "bit"
        },
        1202: {
            "Acronym": "ATR",
            "DataRange": "0 to 255 per byte",
            "EndBit": 63,
            "Name": "Anti-theft Password Representation",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -2,
            "SPN": 1202,
            "SPNLength": 56,
            "StartBit": 8,
            "TransmissionRate": "Transmission of this message is interrupt driven.  This message is also transmitted upon power-up of the interfacing device sending this message.",
            "Units": "ASCII"
        },
        1203: {
            "Acronym": "EAC",
            "DataRange": "0.0 to 145.037738 psi",
            "EndBit": 7,
            "Name": "Engine Auxiliary Coolant Pressure",
            "Offset": 0.0,
            "OperationalHigh": 145.037738,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.580150952,
            "SPN": 1203,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "0.5 s",
            "Units": "psi"
        },
        1204: {
            "Acronym": "EES",
            "DataRange": "0.0 to 43083.687196475 hp",
            "EndBit": 15,
            "Name": "Electrical Load",
            "Offset": 0.0,
            "OperationalHigh": 43083.687196475,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.670511045,
            "SPN": 1204,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "hp"
        },
        1205: {
            "Acronym": "EES",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Safety Wire Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1205,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1208: {
            "Acronym": "EI",
            "DataRange": "0.0 to 145.037738 psi",
            "EndBit": 7,
            "Name": "Engine Pre-filter Oil Pressure",
            "Offset": 0.0,
            "OperationalHigh": 145.037738,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.580150952,
            "SPN": 1208,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        1209: {
            "Acronym": "EI",
            "DataRange": "-36.2594345 to 36.54805959862 psi",
            "EndBit": 23,
            "Name": "Engine Exhaust Gas Pressure",
            "Offset": -36.2594345,
            "OperationalHigh": 36.54805959862,
            "OperationalLow": -36.2594345,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.001133107,
            "SPN": 1209,
            "SPNLength": 16,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        1210: {
            "Acronym": "EI",
            "DataRange": "0 to 100 %",
            "EndBit": 31,
            "Name": "Engine Fuel Rack Position",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 1210,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        1211: {
            "Acronym": "RESET",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Engine Build Hours Reset",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1211,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "When needed",
            "Units": "bit"
        },
        1212: {
            "Acronym": "EAC",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 15,
            "Name": "Engine Auxiliary Coolant Temperature",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 1212,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "0.5 s",
            "Units": "deg F"
        },
        1213: {
            "Acronym": "DM12",
            "DataRange": "0 to 3",
            "EndBit": 7,
            "Name": "Malfunction Indicator Lamp",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 1213,
            "SPNLength": 2,
            "StartBit": 6,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        1214: {
            "Acronym": "DM12",
            "DataRange": "0 to 524,287",
            "EndBit": 34,
            "Name": "Suspect Parameter Number",
            "Offset": 0.0,
            "OperationalHigh": 524287.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": -3,
            "SPN": 1214,
            "SPNLength": 19,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "binary"
        },
        1215: {
            "Acronym": "DM12",
            "DataRange": "00 to 31",
            "EndBit": 39,
            "Name": "Failure Mode Identifier",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": -3,
            "SPN": 1215,
            "SPNLength": 5,
            "StartBit": 35,
            "TransmissionRate": "On request",
            "Units": "binary"
        },
        1216: {
            "Acronym": "DM12",
            "DataRange": "0 to 127",
            "EndBit": 46,
            "Name": "Occurrence Count",
            "Offset": 0.0,
            "OperationalHigh": 127.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": -3,
            "SPN": 1216,
            "SPNLength": 7,
            "StartBit": 40,
            "TransmissionRate": "On request",
            "Units": "binary"
        },
        1217: {
            "Acronym": "DM04",
            "DataRange": "0 to 255",
            "EndBit": 7,
            "Name": "Freeze Frame Length",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 1217,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "count"
        },
        1218: {
            "Acronym": "DM05",
            "DataRange": "0 to 250",
            "EndBit": 7,
            "Name": "Active Trouble Codes",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 1218,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "trouble code"
        },
        1219: {
            "Acronym": "DM05",
            "DataRange": "0 to 250",
            "EndBit": 15,
            "Name": "Previously Active Trouble Codes",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 1219,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "On request",
            "Units": "trouble code"
        },
        1220: {
            "Acronym": "DM05",
            "DataRange": "0 to 250",
            "EndBit": 23,
            "Name": "OBD Compliance",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 1220,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "count"
        },
        1221: {
            "Acronym": "DM05",
            "DataRange": "bit-mapped",
            "EndBit": 31,
            "Name": "Continuously Monitored Systems Support/Status",
            "Offset": 0.0,
            "OperationalHigh": -1,
            "OperationalLow": -1,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 8,
            "SPN": 1221,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "On request",
            "Units": "bit-mapped"
        },
        1222: {
            "Acronym": "DM05",
            "DataRange": "bit-mapped",
            "EndBit": 47,
            "Name": "Non-continuously Monitored Systems Support",
            "Offset": 0.0,
            "OperationalHigh": -1,
            "OperationalLow": -1,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 16,
            "SPN": 1222,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "bit-mapped"
        },
        1223: {
            "Acronym": "DM05",
            "DataRange": "bit-mapped",
            "EndBit": 63,
            "Name": "Non-continuously Monitored Systems Status",
            "Offset": 0.0,
            "OperationalHigh": -1,
            "OperationalLow": -1,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 16,
            "SPN": 1223,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "On request",
            "Units": "bit-mapped"
        },
        1224: {
            "Acronym": "DM08",
            "DataRange": "0 to 255",
            "EndBit": 7,
            "Name": "Test Identifier",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 1224,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "Sent in response to PGN 58112 when the results are available",
            "Units": "count"
        },
        1225: {
            "Acronym": "DM08",
            "DataRange": "0 to 250",
            "EndBit": 15,
            "Name": "Test Type/Component Identifier",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 1225,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "Sent in response to PGN 58112 when the results are available",
            "Units": "count"
        },
        1226: {
            "Acronym": "DM08",
            "DataRange": "Not defined",
            "EndBit": 31,
            "Name": "Test Value",
            "Offset": 0.0,
            "OperationalHigh": -1,
            "OperationalLow": -1,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": -1,
            "SPN": 1226,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "Sent in response to PGN 58112 when the results are available",
            "Units": "Request Dependent"
        },
        1227: {
            "Acronym": "DM08",
            "DataRange": "Not defined",
            "EndBit": 47,
            "Name": "Test Limit Maximum",
            "Offset": 0.0,
            "OperationalHigh": -1,
            "OperationalLow": -1,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": -1,
            "SPN": 1227,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "Sent in response to PGN 58112 when the results are available",
            "Units": "Request Dependent"
        },
        1228: {
            "Acronym": "DM08",
            "DataRange": "Not defined",
            "EndBit": 63,
            "Name": "Test Limit Minimum",
            "Offset": 0.0,
            "OperationalHigh": -1,
            "OperationalLow": -1,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": -1,
            "SPN": 1228,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "Sent in response to PGN 58112 when the results are available",
            "Units": "Request Dependent"
        },
        1229: {
            "Acronym": "DM10",
            "DataRange": "bit-mapped",
            "EndBit": 63,
            "Name": "Test Identifiers Supported",
            "Offset": 0.0,
            "OperationalHigh": -1,
            "OperationalLow": -1,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 64,
            "SPN": 1229,
            "SPNLength": 64,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "bit-mapped"
        },
        123: {
            "Acronym": "TRF1",
            "DataRange": "0.0 to 580.150952 psi",
            "EndBit": 7,
            "Name": "Clutch Pressure",
            "Offset": 0.0,
            "OperationalHigh": 580.150952,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 2.320603808,
            "SPN": 123,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "psi"
        },
        1230: {
            "Acronym": "DM13",
            "DataRange": "0 to 3",
            "EndBit": 7,
            "Name": "Current Data Link",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1230,
            "SPNLength": 2,
            "StartBit": 6,
            "TransmissionRate": "Sent when ever a Stop or Start broadcast event is necessary. To maintain the modified state of the vehicle network(s) the commanding device must send the Hold Signal once every 5 seconds.",
            "Units": "bit"
        },
        1231: {
            "Acronym": "DM13",
            "DataRange": "0 to 3",
            "EndBit": 15,
            "Name": "J1939 Network #2",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1231,
            "SPNLength": 2,
            "StartBit": 14,
            "TransmissionRate": "Sent when ever a Stop or Start broadcast event is necessary. To maintain the modified state of the vehicle network(s) the commanding device must send the Hold Signal once every 5 seconds.",
            "Units": "bit"
        },
        1232: {
            "Acronym": "DM13",
            "DataRange": "0 to 3",
            "EndBit": 13,
            "Name": "ISO 9141",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1232,
            "SPNLength": 2,
            "StartBit": 12,
            "TransmissionRate": "Sent when ever a Stop or Start broadcast event is necessary. To maintain the modified state of the vehicle network(s) the commanding device must send the Hold Signal once every 5 seconds.",
            "Units": "bit"
        },
        1233: {
            "Acronym": "DM13",
            "DataRange": "0 to 3",
            "EndBit": 11,
            "Name": "J1850",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1233,
            "SPNLength": 2,
            "StartBit": 10,
            "TransmissionRate": "Sent when ever a Stop or Start broadcast event is necessary. To maintain the modified state of the vehicle network(s) the commanding device must send the Hold Signal once every 5 seconds.",
            "Units": "bit"
        },
        1234: {
            "Acronym": "DM13",
            "DataRange": "0 to 3",
            "EndBit": 9,
            "Name": "Other, Manufacturer Specified Port",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1234,
            "SPNLength": 2,
            "StartBit": 8,
            "TransmissionRate": "Sent when ever a Stop or Start broadcast event is necessary. To maintain the modified state of the vehicle network(s) the commanding device must send the Hold Signal once every 5 seconds.",
            "Units": "bit"
        },
        1235: {
            "Acronym": "DM13",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "J1939 Network #3",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1235,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "Sent when ever a Stop or Start broadcast event is necessary. To maintain the modified state of the vehicle network(s) the commanding device must send the Hold Signal once every 5 seconds.",
            "Units": "bit"
        },
        1236: {
            "Acronym": "DM13",
            "DataRange": "0 to 15",
            "EndBit": 31,
            "Name": "Hold Signal",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1236,
            "SPNLength": 4,
            "StartBit": 28,
            "TransmissionRate": "Sent when ever a Stop or Start broadcast event is necessary. To maintain the modified state of the vehicle network(s) the commanding device must send the Hold Signal once every 5 seconds.",
            "Units": "bit"
        },
        1237: {
            "Acronym": "CCVS1",
            "DataRange": "0 to 3",
            "EndBit": 63,
            "Name": "Engine Shutdown Override Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1237,
            "SPNLength": 2,
            "StartBit": 62,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1238: {
            "Acronym": "EBC1",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Traction Control Override Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1238,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1239: {
            "Acronym": "FL",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Engine Fuel Leakage 1",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1239,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        124: {
            "Acronym": "TRF1",
            "DataRange": "0 to 100 %",
            "EndBit": 15,
            "Name": "Transmission Oil Level 1",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 124,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "1 s",
            "Units": "%"
        },
        1240: {
            "Acronym": "FL",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Engine Fuel Leakage 2",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1240,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        1241: {
            "Acronym": "EI",
            "DataRange": "0 to 3212.75 kg/h",
            "EndBit": 47,
            "Name": "Engine Gas Mass Flow Rate 1",
            "Offset": 0.0,
            "OperationalHigh": 3212.75,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05,
            "SPN": 1241,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "kg/h"
        },
        1242: {
            "Acronym": "EI",
            "DataRange": "0.0 to 43083.687196475 hp",
            "EndBit": 63,
            "Name": "Instantaneous Estimated Brake Power",
            "Offset": 0.0,
            "OperationalHigh": 43083.687196475,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.670511045,
            "SPN": 1242,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "hp"
        },
        1243: {
            "Acronym": "EBC1",
            "DataRange": "0 to 3",
            "EndBit": 41,
            "Name": "ABS Fully Operational",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1243,
            "SPNLength": 2,
            "StartBit": 40,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1244: {
            "Acronym": "TFAC",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 63,
            "Name": "Engine Fuel Actuator 2 Control Command",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 100%",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 1244,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "50 msec (preferred) or Engine Speed Dependent (if required by application)",
            "Units": "%"
        },
        1246: {
            "Acronym": "ETH",
            "DataRange": "0 to 250 records",
            "EndBit": 7,
            "Name": "Number of Engine Torque History Records",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 1246,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "Record"
        },
        1247: {
            "Acronym": "ETH",
            "DataRange": "0.0 to 43083.687196475 hp",
            "EndBit": 23,
            "Name": "Engine Power",
            "Offset": 0.0,
            "OperationalHigh": 43083.687196475,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.670511045,
            "SPN": 1247,
            "SPNLength": 16,
            "StartBit": 8,
            "TransmissionRate": "On request",
            "Units": "hp"
        },
        1248: {
            "Acronym": "ETH",
            "DataRange": "0 to 64,255 Nm",
            "EndBit": 39,
            "Name": "Engine Peak Torque 1",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 1248,
            "SPNLength": 16,
            "StartBit": 24,
            "TransmissionRate": "On request",
            "Units": "Nm"
        },
        1249: {
            "Acronym": "ETH",
            "DataRange": "0 to 64,255 Nm",
            "EndBit": 55,
            "Name": "Engine Peak Torque 2",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 1249,
            "SPNLength": 16,
            "StartBit": 40,
            "TransmissionRate": "On request",
            "Units": "Nm"
        },
        1250: {
            "Acronym": "ETH",
            "DataRange": "0 to 250 months",
            "EndBit": 63,
            "Name": "Calibration Record Start Month",
            "Offset": 0.0,
            "OperationalHigh": 12.0,
            "OperationalLow": 1.0,
            "OperationalRange": "1 to 12 month",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 1250,
            "SPNLength": 8,
            "StartBit": 56,
            "TransmissionRate": "On request",
            "Units": "Months"
        },
        1251: {
            "Acronym": "ETH",
            "DataRange": "0 to 62.5 days",
            "EndBit": 71,
            "Name": "Calibration Record Start Day",
            "Offset": 0.0,
            "OperationalHigh": 31.75,
            "OperationalLow": 0.25,
            "OperationalRange": "0.25 to 31.75 day",
            "PGNLength": "Variable",
            "Resolution": 0.25,
            "SPN": 1251,
            "SPNLength": 8,
            "StartBit": 64,
            "TransmissionRate": "On request",
            "Units": "Days"
        },
        1252: {
            "Acronym": "ETH",
            "DataRange": "1985 to 2235 years",
            "EndBit": 79,
            "Name": "Calibration Record Start Year",
            "Offset": 1985.0,
            "OperationalHigh": 2235.0,
            "OperationalLow": 1985.0,
            "OperationalRange": "1985 to 2235 year",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 1252,
            "SPNLength": 8,
            "StartBit": 72,
            "TransmissionRate": "On request",
            "Units": "Years"
        },
        1253: {
            "Acronym": "ETH",
            "DataRange": "0 to 210,554,060.75 hr",
            "EndBit": 111,
            "Name": "Calibration Record Duration Time",
            "Offset": 0.0,
            "OperationalHigh": 210554060.75,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.05,
            "SPN": 1253,
            "SPNLength": 32,
            "StartBit": 80,
            "TransmissionRate": "On request",
            "Units": "hours"
        },
        1254: {
            "Acronym": "ETH",
            "DataRange": "0 to 3",
            "EndBit": 113,
            "Name": "Torque Limiting Feature Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 1254,
            "SPNLength": 2,
            "StartBit": 112,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        1255: {
            "Acronym": "ETH",
            "DataRange": "0 to 642.55",
            "EndBit": 135,
            "Name": "Transmission Gear Ratio 1",
            "Offset": 0.0,
            "OperationalHigh": 642.55,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.01,
            "SPN": 1255,
            "SPNLength": 16,
            "StartBit": 120,
            "TransmissionRate": "On request",
            "Units": ""
        },
        1256: {
            "Acronym": "ETH",
            "DataRange": "0 to 64,255 Nm",
            "EndBit": 151,
            "Name": "Engine Torque Limit 1, Transmission",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 1256,
            "SPNLength": 16,
            "StartBit": 136,
            "TransmissionRate": "On request",
            "Units": "Nm"
        },
        1257: {
            "Acronym": "ETH",
            "DataRange": "0 to 642.55",
            "EndBit": 167,
            "Name": "Transmission Gear Ratio 2",
            "Offset": 0.0,
            "OperationalHigh": 642.55,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.01,
            "SPN": 1257,
            "SPNLength": 16,
            "StartBit": 152,
            "TransmissionRate": "On request",
            "Units": ""
        },
        1258: {
            "Acronym": "ETH",
            "DataRange": "0 to 64,255 Nm",
            "EndBit": 183,
            "Name": "Engine Torque Limit 2, Transmission",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 1258,
            "SPNLength": 16,
            "StartBit": 168,
            "TransmissionRate": "On request",
            "Units": "Nm"
        },
        1259: {
            "Acronym": "ETH",
            "DataRange": "0 to 642.55",
            "EndBit": 199,
            "Name": "Transmission Gear Ratio 3",
            "Offset": 0.0,
            "OperationalHigh": 642.55,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.01,
            "SPN": 1259,
            "SPNLength": 16,
            "StartBit": 184,
            "TransmissionRate": "On request",
            "Units": ""
        },
        126: {
            "Acronym": "TRF1",
            "DataRange": "0.0 to 72.518869 psi",
            "EndBit": 23,
            "Name": "Transmission Filter Differential Pressure",
            "Offset": 0.0,
            "OperationalHigh": 72.518869,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.290075476,
            "SPN": 126,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "1 s",
            "Units": "psi"
        },
        1260: {
            "Acronym": "ETH",
            "DataRange": "0 to 64,255 Nm",
            "EndBit": 215,
            "Name": "Engine Torque Limit 3, Transmission",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 1260,
            "SPNLength": 16,
            "StartBit": 200,
            "TransmissionRate": "On request",
            "Units": "Nm"
        },
        1261: {
            "Acronym": "ETH",
            "DataRange": "0 to 64,255 Nm",
            "EndBit": 231,
            "Name": "Engine Torque Limit 4, Transmission",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 1261,
            "SPNLength": 16,
            "StartBit": 216,
            "TransmissionRate": "On request",
            "Units": "Nm"
        },
        1262: {
            "Acronym": "ETH",
            "DataRange": "0 to 64,255 Nm",
            "EndBit": 247,
            "Name": "Engine Torque Limit 5, Switch",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 1262,
            "SPNLength": 16,
            "StartBit": 232,
            "TransmissionRate": "On request",
            "Units": "Nm"
        },
        1263: {
            "Acronym": "ETH",
            "DataRange": "0 to 128,510 Nm",
            "EndBit": 263,
            "Name": "Engine Torque Limit 6, Axle Input",
            "Offset": 0.0,
            "OperationalHigh": 128510.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 2,
            "SPN": 1263,
            "SPNLength": 16,
            "StartBit": 248,
            "TransmissionRate": "On request",
            "Units": "Nm"
        },
        127: {
            "Acronym": "TRF1",
            "DataRange": "0.0 to 580.150952 psi",
            "EndBit": 31,
            "Name": "Transmission Oil Pressure",
            "Offset": 0.0,
            "OperationalHigh": 580.150952,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 2.320603808,
            "SPN": 127,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "1 s",
            "Units": "psi"
        },
        1294: {
            "Acronym": "ESV1",
            "DataRange": "0 to 64,255 Volts",
            "EndBit": 15,
            "Name": "Engine Spark Plug 1",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1294,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "Volts"
        },
        1295: {
            "Acronym": "ESV1",
            "DataRange": "0 to 64,255 Volts",
            "EndBit": 31,
            "Name": "Engine Spark Plug 2",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1295,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "Volts"
        },
        1296: {
            "Acronym": "ESV1",
            "DataRange": "0 to 64,255 Volts",
            "EndBit": 47,
            "Name": "Engine Spark Plug 3",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1296,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "Volts"
        },
        1297: {
            "Acronym": "ESV1",
            "DataRange": "0 to 64,255 Volts",
            "EndBit": 63,
            "Name": "Engine Spark Plug 4",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1297,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "On request",
            "Units": "Volts"
        },
        1298: {
            "Acronym": "ESV2",
            "DataRange": "0 to 64,255 Volts",
            "EndBit": 15,
            "Name": "Engine Spark Plug 5",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1298,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "Volts"
        },
        1299: {
            "Acronym": "ESV2",
            "DataRange": "0 to 64,255 Volts",
            "EndBit": 31,
            "Name": "Engine Spark Plug 6",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1299,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "Volts"
        },
        1300: {
            "Acronym": "ESV2",
            "DataRange": "0 to 64,255 Volts",
            "EndBit": 47,
            "Name": "Engine Spark Plug 7",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1300,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "Volts"
        },
        1301: {
            "Acronym": "ESV2",
            "DataRange": "0 to 64,255 Volts",
            "EndBit": 63,
            "Name": "Engine Spark Plug 8",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1301,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "On request",
            "Units": "Volts"
        },
        1302: {
            "Acronym": "ESV3",
            "DataRange": "0 to 64,255 Volts",
            "EndBit": 15,
            "Name": "Engine Spark Plug 9",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1302,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "Volts"
        },
        1303: {
            "Acronym": "ESV3",
            "DataRange": "0 to 64,255 Volts",
            "EndBit": 31,
            "Name": "Engine Spark Plug 10",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1303,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "Volts"
        },
        1304: {
            "Acronym": "ESV3",
            "DataRange": "0 to 64,255 Volts",
            "EndBit": 47,
            "Name": "Engine Spark Plug 11",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1304,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "Volts"
        },
        1305: {
            "Acronym": "ESV3",
            "DataRange": "0 to 64,255 Volts",
            "EndBit": 63,
            "Name": "Engine Spark Plug 12",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1305,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "On request",
            "Units": "Volts"
        },
        1306: {
            "Acronym": "ESV4",
            "DataRange": "0 to 64,255 Volts",
            "EndBit": 15,
            "Name": "Engine Spark Plug 13",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1306,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "Volts"
        },
        1307: {
            "Acronym": "ESV4",
            "DataRange": "0 to 64,255 Volts",
            "EndBit": 31,
            "Name": "Engine Spark Plug 14",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1307,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "Volts"
        },
        1308: {
            "Acronym": "ESV4",
            "DataRange": "0 to 64,255 Volts",
            "EndBit": 47,
            "Name": "Engine Spark Plug 15",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1308,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "Volts"
        },
        1309: {
            "Acronym": "ESV4",
            "DataRange": "0 to 64,255 Volts",
            "EndBit": 63,
            "Name": "Engine Spark Plug 16",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1309,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "On request",
            "Units": "Volts"
        },
        1310: {
            "Acronym": "ESV5",
            "DataRange": "0 to 64,255 Volts",
            "EndBit": 15,
            "Name": "Engine Spark Plug 17",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1310,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "Volts"
        },
        1311: {
            "Acronym": "ESV5",
            "DataRange": "0 to 64,255 Volts",
            "EndBit": 31,
            "Name": "Engine Spark Plug 18",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1311,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "Volts"
        },
        1312: {
            "Acronym": "ESV5",
            "DataRange": "0 to 64,255 Volts",
            "EndBit": 47,
            "Name": "Engine Spark Plug 19",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1312,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "Volts"
        },
        1313: {
            "Acronym": "ESV5",
            "DataRange": "0 to 64,255 Volts",
            "EndBit": 63,
            "Name": "Engine Spark Plug 20",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1313,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "On request",
            "Units": "Volts"
        },
        1314: {
            "Acronym": "ESV6",
            "DataRange": "0 to 64,255 Volts",
            "EndBit": 15,
            "Name": "Engine Spark Plug 21",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1314,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "Volts"
        },
        1315: {
            "Acronym": "ESV6",
            "DataRange": "0 to 64,255 Volts",
            "EndBit": 31,
            "Name": "Engine Spark Plug 22",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1315,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "Volts"
        },
        1316: {
            "Acronym": "ESV6",
            "DataRange": "0 to 64,255 Volts",
            "EndBit": 47,
            "Name": "Engine Spark Plug 23",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1316,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "Volts"
        },
        1317: {
            "Acronym": "ESV6",
            "DataRange": "0 to 64,255 Volts",
            "EndBit": 63,
            "Name": "Engine Spark Plug 24",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1317,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "On request",
            "Units": "Volts"
        },
        132: {
            "Acronym": "EGF1",
            "DataRange": "0 to 3212.75 kg/h",
            "EndBit": 31,
            "Name": "Engine Intake Air Mass Flow Rate",
            "Offset": 0.0,
            "OperationalHigh": 3212.75,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05,
            "SPN": 132,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "50 ms",
            "Units": "kg/h"
        },
        1320: {
            "Acronym": "SP2",
            "DataRange": "0.0 to 4659.699927595 psi",
            "EndBit": 15,
            "Name": "Engine External Shutdown Air Supply Pressure",
            "Offset": 0.0,
            "OperationalHigh": 4659.699927595,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.072518869,
            "SPN": 1320,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "psi"
        },
        1349: {
            "Acronym": "EFL/P2",
            "DataRange": "0.0 to 36404.472237999995 psi",
            "EndBit": 63,
            "Name": "Engine Injector Metering Rail 2 Pressure",
            "Offset": 0.0,
            "OperationalHigh": 36404.472237999995,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.5665536640625,
            "SPN": 1349,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "0.5 s",
            "Units": "psi"
        },
        1350: {
            "Acronym": "S2",
            "DataRange": "-32,127 to 32,128 hr",
            "EndBit": 23,
            "Name": "Time Since Last Service",
            "Offset": -115657200.0,
            "OperationalHigh": 32128.0,
            "OperationalLow": -32127.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1350,
            "SPNLength": 16,
            "StartBit": 8,
            "TransmissionRate": "On request",
            "Units": "hours"
        },
        1351: {
            "Acronym": "AIR1",
            "DataRange": "0 to 3",
            "EndBit": 49,
            "Name": "Air Compressor Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1351,
            "SPNLength": 2,
            "StartBit": 48,
            "TransmissionRate": "1 sec",
            "Units": "bit"
        },
        1352: {
            "Acronym": "KL1",
            "DataRange": "0 to 250 %",
            "EndBit": 7,
            "Name": "Engine Cylinder 1 Knock Level",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 100%",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1352,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "Engine speed dependent when knock present, once every 5 seconds otherwise",
            "Units": "%"
        },
        1353: {
            "Acronym": "KL1",
            "DataRange": "0 to 250 %",
            "EndBit": 15,
            "Name": "Engine Cylinder 2 Knock Level",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1353,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "Engine speed dependent when knock present, once every 5 seconds otherwise",
            "Units": "%"
        },
        1354: {
            "Acronym": "KL1",
            "DataRange": "0 to 250 %",
            "EndBit": 23,
            "Name": "Engine Cylinder 3 Knock Level",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1354,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "Engine speed dependent when knock present, once every 5 seconds otherwise",
            "Units": "%"
        },
        1355: {
            "Acronym": "KL1",
            "DataRange": "0 to 250 %",
            "EndBit": 31,
            "Name": "Engine Cylinder 4 Knock Level",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1355,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "Engine speed dependent when knock present, once every 5 seconds otherwise",
            "Units": "%"
        },
        1356: {
            "Acronym": "KL1",
            "DataRange": "0 to 250 %",
            "EndBit": 39,
            "Name": "Engine Cylinder 5 Knock Level",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1356,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "Engine speed dependent when knock present, once every 5 seconds otherwise",
            "Units": "%"
        },
        1357: {
            "Acronym": "KL1",
            "DataRange": "0 to 250 %",
            "EndBit": 47,
            "Name": "Engine Cylinder 6 Knock Level",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1357,
            "SPNLength": 8,
            "StartBit": 40,
            "TransmissionRate": "Engine speed dependent when knock present, once every 5 seconds otherwise",
            "Units": "%"
        },
        1358: {
            "Acronym": "KL1",
            "DataRange": "0 to 250 %",
            "EndBit": 55,
            "Name": "Engine Cylinder 7 Knock Level",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1358,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "Engine speed dependent when knock present, once every 5 seconds otherwise",
            "Units": "%"
        },
        1359: {
            "Acronym": "KL1",
            "DataRange": "0 to 250 %",
            "EndBit": 63,
            "Name": "Engine Cylinder 8 Knock Level",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1359,
            "SPNLength": 8,
            "StartBit": 56,
            "TransmissionRate": "Engine speed dependent when knock present, once every 5 seconds otherwise",
            "Units": "%"
        },
        136: {
            "Acronym": "AP",
            "DataRange": "0.0 to 4659.699927595 psi",
            "EndBit": 15,
            "Name": "Auxiliary Vacuum Pressure Reading",
            "Offset": 0.0,
            "OperationalHigh": 4659.699927595,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.072518869,
            "SPN": 136,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "psi"
        },
        1360: {
            "Acronym": "KL2",
            "DataRange": "0 to 250 %",
            "EndBit": 7,
            "Name": "Engine Cylinder 9 Knock Level",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1360,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "Engine speed dependent when knock present, once every 5 seconds otherwise",
            "Units": "%"
        },
        1361: {
            "Acronym": "KL2",
            "DataRange": "0 to 250 %",
            "EndBit": 15,
            "Name": "Engine Cylinder 10 Knock Level",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1361,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "Engine speed dependent when knock present, once every 5 seconds otherwise",
            "Units": "%"
        },
        1362: {
            "Acronym": "KL2",
            "DataRange": "0 to 250 %",
            "EndBit": 23,
            "Name": "Engine Cylinder 11 Knock Level",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1362,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "Engine speed dependent when knock present, once every 5 seconds otherwise",
            "Units": "%"
        },
        1363: {
            "Acronym": "KL2",
            "DataRange": "0 to 250 %",
            "EndBit": 31,
            "Name": "Engine Cylinder 12 Knock Level",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1363,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "Engine speed dependent when knock present, once every 5 seconds otherwise",
            "Units": "%"
        },
        1364: {
            "Acronym": "KL2",
            "DataRange": "0 to 250 %",
            "EndBit": 39,
            "Name": "Engine Cylinder 13 Knock Level",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1364,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "Engine speed dependent when knock present, once every 5 seconds otherwise",
            "Units": "%"
        },
        1365: {
            "Acronym": "KL2",
            "DataRange": "0 to 250 %",
            "EndBit": 47,
            "Name": "Engine Cylinder 14 Knock Level",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1365,
            "SPNLength": 8,
            "StartBit": 40,
            "TransmissionRate": "Engine speed dependent when knock present, once every 5 seconds otherwise",
            "Units": "%"
        },
        1366: {
            "Acronym": "KL2",
            "DataRange": "0 to 250 %",
            "EndBit": 55,
            "Name": "Engine Cylinder 15 Knock Level",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1366,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "Engine speed dependent when knock present, once every 5 seconds otherwise",
            "Units": "%"
        },
        1367: {
            "Acronym": "KL2",
            "DataRange": "0 to 250 %",
            "EndBit": 63,
            "Name": "Engine Cylinder 16 Knock Level",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1367,
            "SPNLength": 8,
            "StartBit": 56,
            "TransmissionRate": "Engine speed dependent when knock present, once every 5 seconds otherwise",
            "Units": "%"
        },
        1368: {
            "Acronym": "KL3",
            "DataRange": "0 to 250 %",
            "EndBit": 7,
            "Name": "Engine Cylinder 17 Knock Level",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1368,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "Engine speed dependent when knock present, once every 5 seconds otherwise",
            "Units": "%"
        },
        1369: {
            "Acronym": "KL3",
            "DataRange": "0 to 250 %",
            "EndBit": 15,
            "Name": "Engine Cylinder 18 Knock Level",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1369,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "Engine speed dependent when knock present, once every 5 seconds otherwise",
            "Units": "%"
        },
        137: {
            "Acronym": "AP",
            "DataRange": "0.0 to 4659.699927595 psi",
            "EndBit": 31,
            "Name": "Auxiliary Gage Pressure Reading 1",
            "Offset": 0.0,
            "OperationalHigh": 4659.699927595,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.072518869,
            "SPN": 137,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "psi"
        },
        1370: {
            "Acronym": "KL3",
            "DataRange": "0 to 250 %",
            "EndBit": 23,
            "Name": "Engine Cylinder 19 Knock Level",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1370,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "Engine speed dependent when knock present, once every 5 seconds otherwise",
            "Units": "%"
        },
        1371: {
            "Acronym": "KL3",
            "DataRange": "0 to 250 %",
            "EndBit": 31,
            "Name": "Engine Cylinder 20 Knock Level",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1371,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "Engine speed dependent when knock present, once every 5 seconds otherwise",
            "Units": "%"
        },
        1372: {
            "Acronym": "KL3",
            "DataRange": "0 to 250 %",
            "EndBit": 39,
            "Name": "Engine Cylinder 21 Knock Level",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1372,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "Engine speed dependent when knock present, once every 5 seconds otherwise",
            "Units": "%"
        },
        1373: {
            "Acronym": "KL3",
            "DataRange": "0 to 250 %",
            "EndBit": 47,
            "Name": "Engine Cylinder 22 Knock Level",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1373,
            "SPNLength": 8,
            "StartBit": 40,
            "TransmissionRate": "Engine speed dependent when knock present, once every 5 seconds otherwise",
            "Units": "%"
        },
        1374: {
            "Acronym": "KL3",
            "DataRange": "0 to 250 %",
            "EndBit": 55,
            "Name": "Engine Cylinder 23 Knock Level",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1374,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "Engine speed dependent when knock present, once every 5 seconds otherwise",
            "Units": "%"
        },
        1375: {
            "Acronym": "KL3",
            "DataRange": "0 to 250 %",
            "EndBit": 63,
            "Name": "Engine Cylinder 24 Knock Level",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1375,
            "SPNLength": 8,
            "StartBit": 56,
            "TransmissionRate": "Engine speed dependent when knock present, once every 5 seconds otherwise",
            "Units": "%"
        },
        1377: {
            "Acronym": "OHECS",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Engine Synchronization Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1377,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        1379: {
            "Acronym": "S2",
            "DataRange": "0 to 250 ID",
            "EndBit": 7,
            "Name": "Service Component Identification",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1379,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "ID"
        },
        138: {
            "Acronym": "AP",
            "DataRange": "0.0 to 4659.699927595 psi",
            "EndBit": 47,
            "Name": "Auxiliary Absolute Pressure Reading",
            "Offset": 0.0,
            "OperationalHigh": 4659.699927595,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.072518869,
            "SPN": 138,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "psi"
        },
        1380: {
            "Acronym": "EFS",
            "DataRange": "0 to 100 %",
            "EndBit": 7,
            "Name": "Engine Oil Level Remote Reservoir",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 1380,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "0.5 s",
            "Units": "%"
        },
        1381: {
            "Acronym": "EFS",
            "DataRange": "0.0 to 72.518869 psi",
            "EndBit": 15,
            "Name": "Engine Fuel Supply Pump Intake Pressure",
            "Offset": 0.0,
            "OperationalHigh": 72.518869,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.290075476,
            "SPN": 1381,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "0.5 s",
            "Units": "psi"
        },
        1382: {
            "Acronym": "EFS",
            "DataRange": "0.0 to 72.518869 psi",
            "EndBit": 23,
            "Name": "Engine Fuel Filter (suction side) Differential Pressure",
            "Offset": 0.0,
            "OperationalHigh": 72.518869,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.290075476,
            "SPN": 1382,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "0.5 s",
            "Units": "psi"
        },
        1387: {
            "Acronym": "AAI",
            "DataRange": "0.0 to 580.150952 psi",
            "EndBit": 23,
            "Name": "Auxiliary Pressure #1",
            "Offset": 0.0,
            "OperationalHigh": 580.150952,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 2.320603808,
            "SPN": 1387,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "psi"
        },
        1388: {
            "Acronym": "AAI",
            "DataRange": "0.0 to 580.150952 psi",
            "EndBit": 31,
            "Name": "Auxiliary Pressure #2",
            "Offset": 0.0,
            "OperationalHigh": 580.150952,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 2.320603808,
            "SPN": 1388,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "On request",
            "Units": "psi"
        },
        1389: {
            "Acronym": "GFI1",
            "DataRange": "0 to 6.4255",
            "EndBit": 63,
            "Name": "Engine Fuel Specific Gravity",
            "Offset": 0.0,
            "OperationalHigh": 2.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 2.0000",
            "PGNLength": 8,
            "Resolution": 0.0001,
            "SPN": 1389,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "On request",
            "Units": ""
        },
        1390: {
            "Acronym": "GFP",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 15,
            "Name": "Engine Fuel Valve 1 Intake Absolute Pressure",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 1390,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "psi"
        },
        1391: {
            "Acronym": "GFP",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 31,
            "Name": "Engine Fuel Valve Differential Pressure",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 1391,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "psi"
        },
        1392: {
            "Acronym": "GFP",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 47,
            "Name": "Engine Air to Fuel Differential Pressure",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 1392,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "psi"
        },
        1393: {
            "Acronym": "ISO1",
            "DataRange": "-125 to 125 %",
            "EndBit": 7,
            "Name": "Engine Cylinder #1 Ignition Transformer Secondary Output",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1393,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        1394: {
            "Acronym": "ISO1",
            "DataRange": "-125 to 125 %",
            "EndBit": 15,
            "Name": "Engine Cylinder #2 Ignition Transformer Secondary Output",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1394,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        1395: {
            "Acronym": "ISO1",
            "DataRange": "-125 to 125 %",
            "EndBit": 23,
            "Name": "Engine Cylinder #3 Ignition Transformer Secondary Output",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1395,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        1396: {
            "Acronym": "ISO1",
            "DataRange": "-125 to 125 %",
            "EndBit": 31,
            "Name": "Engine Cylinder #4 Ignition Transformer Secondary Output",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1396,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        1397: {
            "Acronym": "ISO1",
            "DataRange": "-125 to 125 %",
            "EndBit": 39,
            "Name": "Engine Cylinder #5 Ignition Transformer Secondary Output",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1397,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        1398: {
            "Acronym": "ISO1",
            "DataRange": "-125 to 125 %",
            "EndBit": 47,
            "Name": "Engine Cylinder #6 Ignition Transformer Secondary Output",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1398,
            "SPNLength": 8,
            "StartBit": 40,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        1399: {
            "Acronym": "ISO1",
            "DataRange": "-125 to 125 %",
            "EndBit": 55,
            "Name": "Engine Cylinder #7 Ignition Transformer Secondary Output",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1399,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        1400: {
            "Acronym": "ISO1",
            "DataRange": "-125 to 125 %",
            "EndBit": 63,
            "Name": "Engine Cylinder #8 Ignition Transformer Secondary Output",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1400,
            "SPNLength": 8,
            "StartBit": 56,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        1401: {
            "Acronym": "ISO2",
            "DataRange": "-125 to 125 %",
            "EndBit": 7,
            "Name": "Engine Cylinder #9 Ignition Transformer Secondary Output",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1401,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        1402: {
            "Acronym": "ISO2",
            "DataRange": "-125 to 125 %",
            "EndBit": 15,
            "Name": "Engine Cylinder #10 Ignition Transformer Secondary Output",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1402,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        1403: {
            "Acronym": "ISO2",
            "DataRange": "-125 to 125 %",
            "EndBit": 23,
            "Name": "Engine Cylinder #11 Ignition Transformer Secondary Output",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1403,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        1404: {
            "Acronym": "ISO2",
            "DataRange": "-125 to 125 %",
            "EndBit": 31,
            "Name": "Engine Cylinder #12 Ignition Transformer Secondary Output",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1404,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        1405: {
            "Acronym": "ISO2",
            "DataRange": "-125 to 125 %",
            "EndBit": 39,
            "Name": "Engine Cylinder #13 Ignition Transformer Secondary Output",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1405,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        1406: {
            "Acronym": "ISO2",
            "DataRange": "-125 to 125 %",
            "EndBit": 47,
            "Name": "Engine Cylinder #14 Ignition Transformer Secondary Output",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1406,
            "SPNLength": 8,
            "StartBit": 40,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        1407: {
            "Acronym": "ISO2",
            "DataRange": "-125 to 125 %",
            "EndBit": 55,
            "Name": "Engine Cylinder #15 Ignition Transformer Secondary Output",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1407,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        1408: {
            "Acronym": "ISO2",
            "DataRange": "-125 to 125 %",
            "EndBit": 63,
            "Name": "Engine Cylinder #16 Ignition Transformer Secondary Output",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1408,
            "SPNLength": 8,
            "StartBit": 56,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        1409: {
            "Acronym": "ISO3",
            "DataRange": "-125 to 125 %",
            "EndBit": 7,
            "Name": "Engine Cylinder #17 Ignition Transformer Secondary Output",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1409,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        141: {
            "Acronym": "TP2",
            "DataRange": "0.0 to 4659.699927595 psi",
            "EndBit": 15,
            "Name": "Trailer, Tag Or Push Channel Tire Pressure Target",
            "Offset": 0.0,
            "OperationalHigh": 4659.699927595,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.072518869,
            "SPN": 141,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "psi"
        },
        1410: {
            "Acronym": "ISO3",
            "DataRange": "-125 to 125 %",
            "EndBit": 15,
            "Name": "Engine Cylinder #18 Ignition Transformer Secondary Output",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1410,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        1411: {
            "Acronym": "ISO3",
            "DataRange": "-125 to 125 %",
            "EndBit": 23,
            "Name": "Engine Cylinder #19 Ignition Transformer Secondary Output",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1411,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        1412: {
            "Acronym": "ISO3",
            "DataRange": "-125 to 125 %",
            "EndBit": 31,
            "Name": "Engine Cylinder #20 Ignition Transformer Secondary Output",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1412,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        1413: {
            "Acronym": "IT1",
            "DataRange": "-200 to 301.99 deg",
            "EndBit": 15,
            "Name": "Engine Cylinder #1 Ignition Timing",
            "Offset": -200.0,
            "OperationalHigh": 301.99,
            "OperationalLow": -200.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1413,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "deg"
        },
        1414: {
            "Acronym": "IT1",
            "DataRange": "-200 to 301.99 deg",
            "EndBit": 31,
            "Name": "Engine Cylinder #2 Ignition Timing",
            "Offset": -200.0,
            "OperationalHigh": 301.99,
            "OperationalLow": -200.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1414,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "deg"
        },
        1415: {
            "Acronym": "IT1",
            "DataRange": "-200 to 301.99 deg",
            "EndBit": 47,
            "Name": "Engine Cylinder #3 Ignition Timing",
            "Offset": -200.0,
            "OperationalHigh": 301.99,
            "OperationalLow": -200.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1415,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "deg"
        },
        1416: {
            "Acronym": "IT1",
            "DataRange": "-200 to 301.99 deg",
            "EndBit": 63,
            "Name": "Engine Cylinder #4 Ignition Timing",
            "Offset": -200.0,
            "OperationalHigh": 301.99,
            "OperationalLow": -200.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1416,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "On request",
            "Units": "deg"
        },
        1417: {
            "Acronym": "IT2",
            "DataRange": "-200 to 301.99 deg",
            "EndBit": 15,
            "Name": "Engine Cylinder #5 Ignition Timing",
            "Offset": -200.0,
            "OperationalHigh": 301.99,
            "OperationalLow": -200.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1417,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "deg"
        },
        1418: {
            "Acronym": "IT2",
            "DataRange": "-200 to 301.99 deg",
            "EndBit": 31,
            "Name": "Engine Cylinder #6 Ignition Timing",
            "Offset": -200.0,
            "OperationalHigh": 301.99,
            "OperationalLow": -200.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1418,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "deg"
        },
        1419: {
            "Acronym": "IT2",
            "DataRange": "-200 to 301.99 deg",
            "EndBit": 47,
            "Name": "Engine Cylinder #7 Ignition Timing",
            "Offset": -200.0,
            "OperationalHigh": 301.99,
            "OperationalLow": -200.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1419,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "deg"
        },
        142: {
            "Acronym": "TP2",
            "DataRange": "0.0 to 4659.699927595 psi",
            "EndBit": 31,
            "Name": "Drive Channel Tire Pressure Target",
            "Offset": 0.0,
            "OperationalHigh": 4659.699927595,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.072518869,
            "SPN": 142,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "psi"
        },
        1420: {
            "Acronym": "IT2",
            "DataRange": "-200 to 301.99 deg",
            "EndBit": 63,
            "Name": "Engine Cylinder #8 Ignition Timing",
            "Offset": -200.0,
            "OperationalHigh": 301.99,
            "OperationalLow": -200.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1420,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "On request",
            "Units": "deg"
        },
        1421: {
            "Acronym": "IT3",
            "DataRange": "-200 to 301.99 deg",
            "EndBit": 15,
            "Name": "Engine Cylinder #9 Ignition Timing",
            "Offset": -200.0,
            "OperationalHigh": 301.99,
            "OperationalLow": -200.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1421,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "deg"
        },
        1422: {
            "Acronym": "IT3",
            "DataRange": "-200 to 301.99 deg",
            "EndBit": 31,
            "Name": "Engine Cylinder #10 Ignition Timing",
            "Offset": -200.0,
            "OperationalHigh": 301.99,
            "OperationalLow": -200.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1422,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "deg"
        },
        1423: {
            "Acronym": "IT3",
            "DataRange": "-200 to 301.99 deg",
            "EndBit": 47,
            "Name": "Engine Cylinder #11 Ignition Timing",
            "Offset": -200.0,
            "OperationalHigh": 301.99,
            "OperationalLow": -200.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1423,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "deg"
        },
        1424: {
            "Acronym": "IT3",
            "DataRange": "-200 to 301.99 deg",
            "EndBit": 63,
            "Name": "Engine Cylinder #12 Ignition Timing",
            "Offset": -200.0,
            "OperationalHigh": 301.99,
            "OperationalLow": -200.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1424,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "On request",
            "Units": "deg"
        },
        1425: {
            "Acronym": "IT4",
            "DataRange": "-200 to 301.99 deg",
            "EndBit": 15,
            "Name": "Engine Cylinder #13 Ignition Timing",
            "Offset": -200.0,
            "OperationalHigh": 301.99,
            "OperationalLow": -200.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1425,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "deg"
        },
        1426: {
            "Acronym": "IT4",
            "DataRange": "-200 to 301.99 deg",
            "EndBit": 31,
            "Name": "Engine Cylinder #14 Ignition Timing",
            "Offset": -200.0,
            "OperationalHigh": 301.99,
            "OperationalLow": -200.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1426,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "deg"
        },
        1427: {
            "Acronym": "IT4",
            "DataRange": "-200 to 301.99 deg",
            "EndBit": 47,
            "Name": "Engine Cylinder #15 Ignition Timing",
            "Offset": -200.0,
            "OperationalHigh": 301.99,
            "OperationalLow": -200.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1427,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "deg"
        },
        1428: {
            "Acronym": "IT4",
            "DataRange": "-200 to 301.99 deg",
            "EndBit": 63,
            "Name": "Engine Cylinder #16 Ignition Timing",
            "Offset": -200.0,
            "OperationalHigh": 301.99,
            "OperationalLow": -200.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1428,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "On request",
            "Units": "deg"
        },
        1429: {
            "Acronym": "IT5",
            "DataRange": "-200 to 301.99 deg",
            "EndBit": 15,
            "Name": "Engine Cylinder #17 Ignition Timing",
            "Offset": -200.0,
            "OperationalHigh": 301.99,
            "OperationalLow": -200.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1429,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "deg"
        },
        143: {
            "Acronym": "TP2",
            "DataRange": "0.0 to 4659.699927595 psi",
            "EndBit": 47,
            "Name": "Steer Channel Tire Pressure Target",
            "Offset": 0.0,
            "OperationalHigh": 4659.699927595,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.072518869,
            "SPN": 143,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "psi"
        },
        1430: {
            "Acronym": "IT5",
            "DataRange": "-200 to 301.99 deg",
            "EndBit": 31,
            "Name": "Engine Cylinder #18 Ignition Timing",
            "Offset": -200.0,
            "OperationalHigh": 301.99,
            "OperationalLow": -200.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1430,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "deg"
        },
        1431: {
            "Acronym": "IT5",
            "DataRange": "-200 to 301.99 deg",
            "EndBit": 47,
            "Name": "Engine Cylinder #19 Ignition Timing",
            "Offset": -200.0,
            "OperationalHigh": 301.99,
            "OperationalLow": -200.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1431,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "deg"
        },
        1432: {
            "Acronym": "IT5",
            "DataRange": "-200 to 301.99 deg",
            "EndBit": 63,
            "Name": "Engine Cylinder #20 Ignition Timing",
            "Offset": -200.0,
            "OperationalHigh": 301.99,
            "OperationalLow": -200.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1432,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "On request",
            "Units": "deg"
        },
        1433: {
            "Acronym": "IT6",
            "DataRange": "-200 to 301.99 deg",
            "EndBit": 15,
            "Name": "Engine Desired Ignition Timing #1",
            "Offset": -200.0,
            "OperationalHigh": 301.99,
            "OperationalLow": -200.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1433,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "deg"
        },
        1434: {
            "Acronym": "IT6",
            "DataRange": "-200 to 301.99 deg",
            "EndBit": 31,
            "Name": "Engine Desired Ignition Timing #2",
            "Offset": -200.0,
            "OperationalHigh": 301.99,
            "OperationalLow": -200.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1434,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "deg"
        },
        1435: {
            "Acronym": "IT6",
            "DataRange": "-200 to 301.99 deg",
            "EndBit": 47,
            "Name": "Engine Desired Ignition Timing #3",
            "Offset": -200.0,
            "OperationalHigh": 301.99,
            "OperationalLow": -200.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1435,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "deg"
        },
        1436: {
            "Acronym": "IT6",
            "DataRange": "-200 to 301.99 deg",
            "EndBit": 63,
            "Name": "Engine Actual Ignition Timing",
            "Offset": -200.0,
            "OperationalHigh": 301.99,
            "OperationalLow": -200.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1436,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "On request",
            "Units": "deg"
        },
        1437: {
            "Acronym": "EEC2",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Road Speed Limit Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1437,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "50 msec (preferred) or Engine Speed Dependent (if required by application)",
            "Units": "bit"
        },
        1438: {
            "Acronym": "EBC1",
            "DataRange": "0 to 3",
            "EndBit": 45,
            "Name": "ABS/EBS Amber Warning Signal (Powered Vehicle)",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1438,
            "SPNLength": 2,
            "StartBit": 44,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1439: {
            "Acronym": "EBC1",
            "DataRange": "0 to 3",
            "EndBit": 43,
            "Name": "EBS Red Warning Signal",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1439,
            "SPNLength": 2,
            "StartBit": 42,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        144: {
            "Acronym": "TP3",
            "DataRange": "0.0 to 4659.699927595 psi",
            "EndBit": 15,
            "Name": "Trailer, Tag Or Push Channel Tire Pressure",
            "Offset": 0.0,
            "OperationalHigh": 4659.699927595,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.072518869,
            "SPN": 144,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "psi"
        },
        1440: {
            "Acronym": "GFI2",
            "DataRange": "0 to 6425.5 m^3/h",
            "EndBit": 15,
            "Name": "Engine Fuel Flow Rate 1",
            "Offset": 0.0,
            "OperationalHigh": 6425.53,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 1440,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "m^3/h"
        },
        1441: {
            "Acronym": "GFI2",
            "DataRange": "0 to 6425.5 m^3/h",
            "EndBit": 31,
            "Name": "Engine Fuel Flow Rate 2",
            "Offset": 0.0,
            "OperationalHigh": 6425.53,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 1441,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "m^3/h"
        },
        1442: {
            "Acronym": "GFI2",
            "DataRange": "0 to 100 %",
            "EndBit": 39,
            "Name": "Engine Fuel Valve 1 Position",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 1442,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        1443: {
            "Acronym": "GFI2",
            "DataRange": "0 to 100 %",
            "EndBit": 47,
            "Name": "Engine Fuel Valve 2 Position",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 1443,
            "SPNLength": 8,
            "StartBit": 40,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        1444: {
            "Acronym": "CT1",
            "DataRange": "0 to 642.55 ms",
            "EndBit": 15,
            "Name": "Engine Cylinder #1 Combustion Time",
            "Offset": 0.0,
            "OperationalHigh": 642.55,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.01,
            "SPN": 1444,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "ms"
        },
        1445: {
            "Acronym": "CT1",
            "DataRange": "0 to 642.55 ms",
            "EndBit": 31,
            "Name": "Engine Cylinder #2 Combustion Time",
            "Offset": 0.0,
            "OperationalHigh": 642.55,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.01,
            "SPN": 1445,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "ms"
        },
        1446: {
            "Acronym": "CT1",
            "DataRange": "0 to 642.55 ms",
            "EndBit": 47,
            "Name": "Engine Cylinder #3 Combustion Time",
            "Offset": 0.0,
            "OperationalHigh": 642.55,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.01,
            "SPN": 1446,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "ms"
        },
        1447: {
            "Acronym": "CT1",
            "DataRange": "0 to 642.55 ms",
            "EndBit": 63,
            "Name": "Engine Cylinder #4 Combustion Time",
            "Offset": 0.0,
            "OperationalHigh": 642.55,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.01,
            "SPN": 1447,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "On request",
            "Units": "ms"
        },
        1448: {
            "Acronym": "CT2",
            "DataRange": "0 to 642.55 ms",
            "EndBit": 15,
            "Name": "Engine Cylinder #5 Combustion Time",
            "Offset": 0.0,
            "OperationalHigh": 642.55,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.01,
            "SPN": 1448,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "ms"
        },
        1449: {
            "Acronym": "CT2",
            "DataRange": "0 to 642.55 ms",
            "EndBit": 31,
            "Name": "Engine Cylinder #6 Combustion Time",
            "Offset": 0.0,
            "OperationalHigh": 642.55,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.01,
            "SPN": 1449,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "ms\n"
        },
        145: {
            "Acronym": "TP3",
            "DataRange": "0.0 to 4659.699927595 psi",
            "EndBit": 31,
            "Name": "Drive Channel Tire Pressure",
            "Offset": 0.0,
            "OperationalHigh": 4659.699927595,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.072518869,
            "SPN": 145,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "psi"
        },
        1450: {
            "Acronym": "CT2",
            "DataRange": "0 to 642.55 ms",
            "EndBit": 47,
            "Name": "Engine Cylinder #7 Combustion Time",
            "Offset": 0.0,
            "OperationalHigh": 642.55,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.01,
            "SPN": 1450,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "ms"
        },
        1451: {
            "Acronym": "CT2",
            "DataRange": "0 to 642.55 ms",
            "EndBit": 63,
            "Name": "Engine Cylinder #8 Combustion Time",
            "Offset": 0.0,
            "OperationalHigh": 642.55,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.01,
            "SPN": 1451,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "On request",
            "Units": "ms"
        },
        1452: {
            "Acronym": "CT3",
            "DataRange": "0 to 642.55 ms",
            "EndBit": 15,
            "Name": "Engine Cylinder #9 Combustion Time",
            "Offset": 0.0,
            "OperationalHigh": 642.55,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.01,
            "SPN": 1452,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "ms"
        },
        1453: {
            "Acronym": "CT3",
            "DataRange": "0 to 642.55 ms",
            "EndBit": 31,
            "Name": "Engine Cylinder #10 Combustion Time",
            "Offset": 0.0,
            "OperationalHigh": 642.55,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.01,
            "SPN": 1453,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "ms"
        },
        1454: {
            "Acronym": "CT3",
            "DataRange": "0 to 642.55 ms",
            "EndBit": 47,
            "Name": "Engine Cylinder #11 Combustion Time",
            "Offset": 0.0,
            "OperationalHigh": 642.55,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.01,
            "SPN": 1454,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "ms"
        },
        1455: {
            "Acronym": "CT3",
            "DataRange": "0 to 642.55 ms",
            "EndBit": 63,
            "Name": "Engine Cylinder #12 Combustion Time",
            "Offset": 0.0,
            "OperationalHigh": 642.55,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.01,
            "SPN": 1455,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "On request",
            "Units": "ms"
        },
        1456: {
            "Acronym": "CT4",
            "DataRange": "0 to 642.55 ms",
            "EndBit": 15,
            "Name": "Engine Cylinder #13 Combustion Time",
            "Offset": 0.0,
            "OperationalHigh": 642.55,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.01,
            "SPN": 1456,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "ms"
        },
        1457: {
            "Acronym": "CT4",
            "DataRange": "0 to 642.55 ms",
            "EndBit": 31,
            "Name": "Engine Cylinder #14 Combustion Time",
            "Offset": 0.0,
            "OperationalHigh": 642.55,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.01,
            "SPN": 1457,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "ms"
        },
        1458: {
            "Acronym": "CT4",
            "DataRange": "0 to 642.55 ms",
            "EndBit": 47,
            "Name": "Engine Cylinder #15 Combustion Time",
            "Offset": 0.0,
            "OperationalHigh": 642.55,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.01,
            "SPN": 1458,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "ms"
        },
        1459: {
            "Acronym": "CT4",
            "DataRange": "0 to 642.55 ms",
            "EndBit": 63,
            "Name": "Engine Cylinder #16 Combustion Time",
            "Offset": 0.0,
            "OperationalHigh": 642.55,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.01,
            "SPN": 1459,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "On request",
            "Units": "ms"
        },
        146: {
            "Acronym": "TP3",
            "DataRange": "0.0 to 4659.699927595 psi",
            "EndBit": 47,
            "Name": "Steer Channel Tire Pressure",
            "Offset": 0.0,
            "OperationalHigh": 4659.699927595,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.072518869,
            "SPN": 146,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "psi"
        },
        1460: {
            "Acronym": "CT5",
            "DataRange": "0 to 642.55 ms",
            "EndBit": 15,
            "Name": "Engine Cylinder #17 Combustion Time",
            "Offset": 0.0,
            "OperationalHigh": 642.55,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.01,
            "SPN": 1460,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "ms"
        },
        1461: {
            "Acronym": "CT5",
            "DataRange": "0 to 642.55 ms",
            "EndBit": 31,
            "Name": "Engine Cylinder #18 Combustion Time",
            "Offset": 0.0,
            "OperationalHigh": 642.55,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.01,
            "SPN": 1461,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "ms"
        },
        1462: {
            "Acronym": "CT5",
            "DataRange": "0 to 642.55 ms",
            "EndBit": 47,
            "Name": "Engine Cylinder #19 Combustion Time",
            "Offset": 0.0,
            "OperationalHigh": 642.55,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.01,
            "SPN": 1462,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "ms"
        },
        1463: {
            "Acronym": "CT5",
            "DataRange": "0 to 642.55 ms",
            "EndBit": 63,
            "Name": "Engine Cylinder #20 Combustion Time",
            "Offset": 0.0,
            "OperationalHigh": 642.55,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.01,
            "SPN": 1463,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "On request",
            "Units": "ms"
        },
        1464: {
            "Acronym": "CT6",
            "DataRange": "0 to 642.55 ms",
            "EndBit": 15,
            "Name": "Engine Desired Combustion Time",
            "Offset": 0.0,
            "OperationalHigh": 642.55,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.01,
            "SPN": 1464,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "ms"
        },
        1465: {
            "Acronym": "CT6",
            "DataRange": "0 to 642.55 ms",
            "EndBit": 31,
            "Name": "Engine Average Combustion Time",
            "Offset": 0.0,
            "OperationalHigh": 642.55,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.01,
            "SPN": 1465,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "ms"
        },
        1466: {
            "Acronym": "TP1",
            "DataRange": "0 to 15",
            "EndBit": 11,
            "Name": "Steer Channel Mode",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1466,
            "SPNLength": 4,
            "StartBit": 8,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        1467: {
            "Acronym": "TP1",
            "DataRange": "0 to 15",
            "EndBit": 19,
            "Name": "Trailer/tag Channel Mode",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1467,
            "SPNLength": 4,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        1468: {
            "Acronym": "TP1",
            "DataRange": "0 to 15",
            "EndBit": 23,
            "Name": "Drive Channel Mode",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1468,
            "SPNLength": 4,
            "StartBit": 20,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        1469: {
            "Acronym": "TP1",
            "DataRange": "0 to 3",
            "EndBit": 25,
            "Name": "PCU Drive Solenoid Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1469,
            "SPNLength": 2,
            "StartBit": 24,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        1470: {
            "Acronym": "TP1",
            "DataRange": "0 to 3",
            "EndBit": 27,
            "Name": "PCU Steer Solenoid Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1470,
            "SPNLength": 2,
            "StartBit": 26,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        1471: {
            "Acronym": "TP1",
            "DataRange": "0 to 3",
            "EndBit": 29,
            "Name": "Tire Pressure Supply Switch Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1471,
            "SPNLength": 2,
            "StartBit": 28,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        1472: {
            "Acronym": "TP1",
            "DataRange": "0 to 3",
            "EndBit": 33,
            "Name": "PCU Deflate Solenoid Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1472,
            "SPNLength": 2,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        1473: {
            "Acronym": "TP1",
            "DataRange": "0 to 3",
            "EndBit": 35,
            "Name": "PCU Control Solenoid Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1473,
            "SPNLength": 2,
            "StartBit": 34,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        1474: {
            "Acronym": "TP1",
            "DataRange": "0 to 3",
            "EndBit": 37,
            "Name": "PCU Supply Solenoid Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1474,
            "SPNLength": 2,
            "StartBit": 36,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        1475: {
            "Acronym": "TP1",
            "DataRange": "0 to 3",
            "EndBit": 39,
            "Name": "PCU Trailer, Tag or Push Solenoid Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1475,
            "SPNLength": 2,
            "StartBit": 38,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        1479: {
            "Acronym": "DM18",
            "DataRange": "0 to 15",
            "EndBit": 11,
            "Name": "Security Entity Type",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 1479,
            "SPNLength": 4,
            "StartBit": 8,
            "TransmissionRate": "As needed",
            "Units": "bit"
        },
        1480: {
            "Acronym": "ERC1",
            "DataRange": "0 to 255",
            "EndBit": 39,
            "Name": "Source Address of Controlling Device for Retarder Control",
            "Offset": 0.0,
            "OperationalHigh": 253.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 253",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1480,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "SA"
        },
        1481: {
            "Acronym": "EBC1",
            "DataRange": "0 to 255",
            "EndBit": 55,
            "Name": "Source Address of Controlling Device for Brake Control",
            "Offset": 0.0,
            "OperationalHigh": 253.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 253",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1481,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "SA"
        },
        1482: {
            "Acronym": "ETC1",
            "DataRange": "0 to 255",
            "EndBit": 63,
            "Name": "Source Address of Controlling Device for Transmission Control",
            "Offset": 0.0,
            "OperationalHigh": 253.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 253",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1482,
            "SPNLength": 8,
            "StartBit": 56,
            "TransmissionRate": "10 ms",
            "Units": "SA"
        },
        1483: {
            "Acronym": "EEC1",
            "DataRange": "0 to 255",
            "EndBit": 47,
            "Name": "Source Address of Controlling Device for Engine Control",
            "Offset": 0.0,
            "OperationalHigh": 253.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 253",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1483,
            "SPNLength": 8,
            "StartBit": 40,
            "TransmissionRate": "engine speed dependent",
            "Units": "SA"
        },
        1487: {
            "Acronym": "CL",
            "DataRange": "0 to 100 %",
            "EndBit": 7,
            "Name": "Illumination Brightness Percent",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 1487,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "Every 5 s and on change of state but no faster than every 100 ms",
            "Units": "%"
        },
        1503: {
            "Acronym": "ARMSWIMC",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Armrest 1 Switch 1",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1503,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1545: {
            "Acronym": "DM25",
            "DataRange": "manufacturer defined",
            "EndBit": -1,
            "Name": "SPN Data",
            "Offset": 0,
            "OperationalHigh": -1,
            "OperationalLow": -1,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0,
            "SPN": 1545,
            "SPNLength": 8,
            "StartBit": 5,
            "TransmissionRate": "On request",
            "Units": ""
        },
        1550: {
            "Acronym": "FD2",
            "DataRange": "0 to 100 %",
            "EndBit": 7,
            "Name": "Estimated Percent Fan 2 Speed",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 1550,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "%"
        },
        1557: {
            "Acronym": "FD2",
            "DataRange": "0 to 15",
            "EndBit": 11,
            "Name": "Fan 2 Drive State",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1557,
            "SPNLength": 4,
            "StartBit": 8,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        156: {
            "Acronym": "EFL/P2",
            "DataRange": "0.0 to 36404.472237999995 psi",
            "EndBit": 47,
            "Name": "Engine Injector Timing Rail 1 Pressure",
            "Offset": 0.0,
            "OperationalHigh": 36404.472237999995,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.5665536640625,
            "SPN": 156,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "0.5 s",
            "Units": "psi"
        },
        157: {
            "Acronym": "EFL/P2",
            "DataRange": "0.0 to 36404.472237999995 psi",
            "EndBit": 31,
            "Name": "Engine Injector Metering Rail 1 Pressure",
            "Offset": 0.0,
            "OperationalHigh": 36404.472237999995,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.5665536640625,
            "SPN": 157,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "0.5 s",
            "Units": "psi"
        },
        1573: {
            "Acronym": "LVDD",
            "DataRange": "0 to 255",
            "EndBit": 7,
            "Name": "LED Display Data #1",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1573,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1574: {
            "Acronym": "LVD",
            "DataRange": "-3,200 to 3,225.5 mm",
            "EndBit": 15,
            "Name": "Laser Strike Vertical Deviation",
            "Offset": -3.2,
            "OperationalHigh": 3200.0,
            "OperationalLow": -3200.0,
            "OperationalRange": "-3200 to +3200",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 1574,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "50 ms",
            "Units": "m"
        },
        1575: {
            "Acronym": "LSP",
            "DataRange": "-3,200 to 3,225.5 mm",
            "EndBit": 15,
            "Name": "Modify Leveling System Set Point",
            "Offset": -3.2,
            "OperationalHigh": 3225.5,
            "OperationalLow": -3200.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 1575,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "50 ms",
            "Units": "m"
        },
        1576: {
            "Acronym": "LMP",
            "DataRange": "-3,200 to 3,225.5 mm",
            "EndBit": 15,
            "Name": "Mast Position",
            "Offset": -3.2,
            "OperationalHigh": 3200.0,
            "OperationalLow": -3200.0,
            "OperationalRange": "-3200 to +3200 mm",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 1576,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "50 ms",
            "Units": "m"
        },
        1577: {
            "Acronym": "LBC",
            "DataRange": "-3276.8 to 3148.7 sec",
            "EndBit": 15,
            "Name": "Blade Duration and Direction",
            "Offset": -3276.8,
            "OperationalHigh": 3276.8,
            "OperationalLow": -3276.8,
            "OperationalRange": "-3276.8 to 3276.8 sec",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 1577,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "50 ms",
            "Units": "sec"
        },
        1578: {
            "Acronym": "LBC",
            "DataRange": "0 to 255",
            "EndBit": 23,
            "Name": "Blade Control Mode",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1578,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        1579: {
            "Acronym": "LTP",
            "DataRange": "-3,200 to 3,225.5 mm",
            "EndBit": 15,
            "Name": "Laser Tracer Target Deviation",
            "Offset": -3.2,
            "OperationalHigh": 3200.0,
            "OperationalLow": -3200.0,
            "OperationalRange": "-3200 to +3200",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 1579,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "50 ms",
            "Units": "m"
        },
        158: {
            "Acronym": "VEP1",
            "DataRange": "0 to 3212.75 V",
            "EndBit": 63,
            "Name": "Keyswitch Battery Potential",
            "Offset": 0.0,
            "OperationalHigh": 3212.75,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05,
            "SPN": 158,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "1 s",
            "Units": "Volts"
        },
        1580: {
            "Acronym": "LTP",
            "DataRange": "0 to 6,425.5 mm",
            "EndBit": 31,
            "Name": "Laser Tracer Vertical Distance",
            "Offset": 0.0,
            "OperationalHigh": 6400.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 6400 mm",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 1580,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "50 ms",
            "Units": "m"
        },
        1581: {
            "Acronym": "LTP",
            "DataRange": "0 to 250 %",
            "EndBit": 39,
            "Name": "Laser Tracer Horizontal Deviation",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1581,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "50 ms",
            "Units": "%"
        },
        1582: {
            "Acronym": "LTP",
            "DataRange": "0 to 255",
            "EndBit": 47,
            "Name": "LED Display Data #2",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1582,
            "SPNLength": 8,
            "StartBit": 40,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        1583: {
            "Acronym": "LTP",
            "DataRange": "0 to 255",
            "EndBit": 55,
            "Name": "Laser Tracer Information",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1583,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        1584: {
            "Acronym": "RESET",
            "DataRange": "0 to 250 ID",
            "EndBit": 15,
            "Name": "Service Component Identification",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1584,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "When needed",
            "Units": "ID"
        },
        1585: {
            "Acronym": "CVW",
            "DataRange": "0 to 642,550 kg",
            "EndBit": 15,
            "Name": "Powered Vehicle Weight",
            "Offset": 0.0,
            "OperationalHigh": 642550.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 10,
            "SPN": 1585,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "kg"
        },
        1586: {
            "Acronym": "ACC1",
            "DataRange": "0.0 to 155.34275 mph",
            "EndBit": 7,
            "Name": "Speed of forward vehicle",
            "Offset": 0.0,
            "OperationalHigh": 155.34275,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.621371,
            "SPN": 1586,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "100ms or upon state change, but not faster than 20 ms.",
            "Units": "mph"
        },
        1587: {
            "Acronym": "ACC1",
            "DataRange": "0 to 250 m",
            "EndBit": 15,
            "Name": "Distance to forward vehicle",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1587,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "100ms or upon state change, but not faster than 20 ms.",
            "Units": "m"
        },
        1588: {
            "Acronym": "ACC1",
            "DataRange": "0 to 155.343 mph",
            "EndBit": 23,
            "Name": "Adaptive Cruise Control Set Speed",
            "Offset": 0.0,
            "OperationalHigh": 74.5645,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 74.5645 mph",
            "PGNLength": 8,
            "Resolution": 0.621371,
            "SPN": 1588,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "100ms or upon state change, but not faster than 20 ms.",
            "Units": "mph"
        },
        1589: {
            "Acronym": "ACC1",
            "DataRange": "0 to 7",
            "EndBit": 29,
            "Name": "Adaptive cruise control set distance mode",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1589,
            "SPNLength": 3,
            "StartBit": 27,
            "TransmissionRate": "100ms or upon state change, but not faster than 20 ms.",
            "Units": "bit"
        },
        159: {
            "Acronym": "A1",
            "DataRange": "0.0 to 4659.699927595 psi",
            "EndBit": 23,
            "Name": "Engine Gas Supply Pressure",
            "Offset": 0.0,
            "OperationalHigh": 4659.699927595,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.072518869,
            "SPN": 159,
            "SPNLength": 16,
            "StartBit": 8,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        1590: {
            "Acronym": "ACC1",
            "DataRange": "0 to 7",
            "EndBit": 26,
            "Name": "Adaptive Cruise Control Mode",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1590,
            "SPNLength": 3,
            "StartBit": 24,
            "TransmissionRate": "100ms or upon state change, but not faster than 20 ms.",
            "Units": "bit"
        },
        1591: {
            "Acronym": "ACC1",
            "DataRange": "-250 to 251.992 1/km",
            "EndBit": 47,
            "Name": "Road curvature",
            "Offset": -250.0,
            "OperationalHigh": 251.9921,
            "OperationalLow": -250.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1591,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "100ms or upon state change, but not faster than 20 ms.",
            "Units": "1/km"
        },
        1592: {
            "Acronym": "HRW",
            "DataRange": "0.0 to 155.961635516 mph",
            "EndBit": 15,
            "Name": "Front Axle, Left Wheel Speed",
            "Offset": 0.0,
            "OperationalHigh": 155.961635516,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.00242723046875,
            "SPN": 1592,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "20 ms",
            "Units": "mph"
        },
        1593: {
            "Acronym": "HRW",
            "DataRange": "0.0 to 155.961635516 mph",
            "EndBit": 31,
            "Name": "Front axle, right wheel speed",
            "Offset": 0.0,
            "OperationalHigh": 155.961635516,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.00242723046875,
            "SPN": 1593,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "20 ms",
            "Units": "mph"
        },
        1594: {
            "Acronym": "HRW",
            "DataRange": "0.0 to 155.961635516 mph",
            "EndBit": 47,
            "Name": "Rear axle, left wheel speed",
            "Offset": 0.0,
            "OperationalHigh": 155.961635516,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.00242723046875,
            "SPN": 1594,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "20 ms",
            "Units": "mph"
        },
        1595: {
            "Acronym": "HRW",
            "DataRange": "0.0 to 155.961635516 mph",
            "EndBit": 63,
            "Name": "Rear axle, right wheel speed",
            "Offset": 0.0,
            "OperationalHigh": 155.961635516,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.00242723046875,
            "SPN": 1595,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "20 ms",
            "Units": "mph"
        },
        1596: {
            "Acronym": "DM18",
            "DataRange": "0 to 4095",
            "EndBit": 11,
            "Name": "Security Entity Length",
            "Offset": 0.0,
            "OperationalHigh": 4095.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 1596,
            "SPNLength": 12,
            "StartBit": 0,
            "TransmissionRate": "As needed",
            "Units": "bit"
        },
        1597: {
            "Acronym": "DM18",
            "DataRange": "manufacturer defined",
            "EndBit": -1,
            "Name": "Data Security Parameter",
            "Offset": 0,
            "OperationalHigh": -1,
            "OperationalLow": -1,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0,
            "SPN": 1597,
            "SPNLength": 8,
            "StartBit": 2,
            "TransmissionRate": "As needed",
            "Units": ""
        },
        1598: {
            "Acronym": "FD2",
            "DataRange": "0 to 8,031.875 rpm",
            "EndBit": 31,
            "Name": "Fan 2 Speed",
            "Offset": 0.0,
            "OperationalHigh": 8031.875,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.125,
            "SPN": 1598,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "1 s",
            "Units": "rpm"
        },
        1599: {
            "Acronym": "DM15",
            "DataRange": "0 to 65,535",
            "EndBit": 63,
            "Name": "Seed",
            "Offset": 0.0,
            "OperationalHigh": 65535.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1599,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "As needed",
            "Units": "count"
        },
        1600: {
            "Acronym": "LFE2",
            "DataRange": "0 to 4,211,081.215 L/h",
            "EndBit": 31,
            "Name": "High Resolution Engine Fuel Rate",
            "Offset": 0.0,
            "OperationalHigh": 4211081.215,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.001,
            "SPN": 1600,
            "SPNLength": 32,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "l/h"
        },
        1601: {
            "Acronym": "TD",
            "DataRange": "-125 to 125 mins",
            "EndBit": 55,
            "Name": "Local minute offset",
            "Offset": -7500.0,
            "OperationalHigh": 59.0,
            "OperationalLow": -59.0,
            "OperationalRange": "-59 to +59 min",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1601,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "On request",
            "Units": "minutes"
        },
        1602: {
            "Acronym": "TD",
            "DataRange": "-125 to 125 hr",
            "EndBit": 63,
            "Name": "Local hour offset",
            "Offset": -450000.0,
            "OperationalHigh": 23.0,
            "OperationalLow": -23.0,
            "OperationalRange": "-23 to +23 h",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1602,
            "SPNLength": 8,
            "StartBit": 56,
            "TransmissionRate": "On request",
            "Units": "hours"
        },
        1603: {
            "Acronym": "TDA",
            "DataRange": "0 to 62.5 s",
            "EndBit": 7,
            "Name": "Adjust seconds",
            "Offset": 0.0,
            "OperationalHigh": 59.75,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 59.75 s",
            "PGNLength": 8,
            "Resolution": 0.25,
            "SPN": 1603,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "As needed",
            "Units": "Seconds"
        },
        1604: {
            "Acronym": "TDA",
            "DataRange": "0 to 250 mins",
            "EndBit": 15,
            "Name": "Adjust minutes",
            "Offset": 0.0,
            "OperationalHigh": 59.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 59 min",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1604,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "As needed",
            "Units": "minutes"
        },
        1605: {
            "Acronym": "TDA",
            "DataRange": "0 to 250 hr",
            "EndBit": 23,
            "Name": "Adjust hours",
            "Offset": 0.0,
            "OperationalHigh": 23.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 23 h",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1605,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "As needed",
            "Units": "hours"
        },
        1606: {
            "Acronym": "TDA",
            "DataRange": "0 to 250 months",
            "EndBit": 31,
            "Name": "Adjust month",
            "Offset": 0.0,
            "OperationalHigh": 12.0,
            "OperationalLow": 1.0,
            "OperationalRange": "1 to 12 month",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1606,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "As needed",
            "Units": "Months"
        },
        1607: {
            "Acronym": "TDA",
            "DataRange": "0 to 62.5 days",
            "EndBit": 39,
            "Name": "Adjust day",
            "Offset": 0.0,
            "OperationalHigh": 31.75,
            "OperationalLow": 0.25,
            "OperationalRange": "0.25 to 31.75 day",
            "PGNLength": 8,
            "Resolution": 0.25,
            "SPN": 1607,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "As needed",
            "Units": "Days"
        },
        1608: {
            "Acronym": "TDA",
            "DataRange": "1985 to 2235 years",
            "EndBit": 47,
            "Name": "Adjust year",
            "Offset": 1985.0,
            "OperationalHigh": 2235.0,
            "OperationalLow": 1985.0,
            "OperationalRange": "1985 to 2235 year",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1608,
            "SPNLength": 8,
            "StartBit": 40,
            "TransmissionRate": "As needed",
            "Units": "Years"
        },
        1609: {
            "Acronym": "TDA",
            "DataRange": "-125 to 125 mins",
            "EndBit": 55,
            "Name": "Adjust local minute offset",
            "Offset": -7500.0,
            "OperationalHigh": 59.0,
            "OperationalLow": -59.0,
            "OperationalRange": "-59 to +59 min",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1609,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "As needed",
            "Units": "minutes"
        },
        161: {
            "Acronym": "ETC1",
            "DataRange": "0 to 8,031.875 rpm",
            "EndBit": 55,
            "Name": "Transmission Input Shaft Speed",
            "Offset": 0.0,
            "OperationalHigh": 8031.875,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.125,
            "SPN": 161,
            "SPNLength": 16,
            "StartBit": 40,
            "TransmissionRate": "10 ms",
            "Units": "rpm"
        },
        1610: {
            "Acronym": "TDA",
            "DataRange": "-125 to 125 hr",
            "EndBit": 63,
            "Name": "Adjust local hour offset",
            "Offset": -450000.0,
            "OperationalHigh": 23.0,
            "OperationalLow": -24.0,
            "OperationalRange": "-24 to +23 h",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1610,
            "SPNLength": 8,
            "StartBit": 56,
            "TransmissionRate": "As needed",
            "Units": "hours"
        },
        1611: {
            "Acronym": "TCO1",
            "DataRange": "0 to 3",
            "EndBit": 7,
            "Name": "Vehicle motion",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1611,
            "SPNLength": 2,
            "StartBit": 6,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        1612: {
            "Acronym": "TCO1",
            "DataRange": "0 to 7",
            "EndBit": 2,
            "Name": "Driver 1 working state",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1612,
            "SPNLength": 3,
            "StartBit": 0,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        1613: {
            "Acronym": "TCO1",
            "DataRange": "0 to 7",
            "EndBit": 5,
            "Name": "Driver 2 working state",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1613,
            "SPNLength": 3,
            "StartBit": 3,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        1614: {
            "Acronym": "TCO1",
            "DataRange": "0 to 3",
            "EndBit": 15,
            "Name": "Vehicle Overspeed",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1614,
            "SPNLength": 2,
            "StartBit": 14,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        1615: {
            "Acronym": "TCO1",
            "DataRange": "0 to 3",
            "EndBit": 13,
            "Name": "Driver card, driver 1",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1615,
            "SPNLength": 2,
            "StartBit": 12,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        1616: {
            "Acronym": "TCO1",
            "DataRange": "0 to 3",
            "EndBit": 21,
            "Name": "Driver card, driver 2",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1616,
            "SPNLength": 2,
            "StartBit": 20,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        1617: {
            "Acronym": "TCO1",
            "DataRange": "0 to 15",
            "EndBit": 11,
            "Name": "Driver 1 Time Related States",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1617,
            "SPNLength": 4,
            "StartBit": 8,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        1618: {
            "Acronym": "TCO1",
            "DataRange": "0 to 15",
            "EndBit": 19,
            "Name": "Driver 2 Time Related States",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1618,
            "SPNLength": 4,
            "StartBit": 16,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        1619: {
            "Acronym": "TCO1",
            "DataRange": "0 to 3",
            "EndBit": 31,
            "Name": "Direction indicator",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1619,
            "SPNLength": 2,
            "StartBit": 30,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        162: {
            "Acronym": "ETC2",
            "DataRange": "0 to 255 per byte",
            "EndBit": 47,
            "Name": "Transmission Requested Range",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -2,
            "SPN": 162,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "ASCII"
        },
        1620: {
            "Acronym": "TCO1",
            "DataRange": "0 to 3",
            "EndBit": 29,
            "Name": "Tachograph performance",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1620,
            "SPNLength": 2,
            "StartBit": 28,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        1621: {
            "Acronym": "TCO1",
            "DataRange": "0 to 3",
            "EndBit": 27,
            "Name": "Handling information",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1621,
            "SPNLength": 2,
            "StartBit": 26,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        1622: {
            "Acronym": "TCO1",
            "DataRange": "0 to 3",
            "EndBit": 25,
            "Name": "System event",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1622,
            "SPNLength": 2,
            "StartBit": 24,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        1623: {
            "Acronym": "TCO1",
            "DataRange": "0 to 8,031.875 rpm",
            "EndBit": 47,
            "Name": "Tachograph output shaft speed",
            "Offset": 0.0,
            "OperationalHigh": 8031.875,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.125,
            "SPN": 1623,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "50 ms",
            "Units": "rpm"
        },
        1624: {
            "Acronym": "TCO1",
            "DataRange": "0.0 to 155.961635516 mph",
            "EndBit": 63,
            "Name": "Tachograph vehicle speed",
            "Offset": 0.0,
            "OperationalHigh": 155.961635516,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.00242723046875,
            "SPN": 1624,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "50 ms",
            "Units": "mph"
        },
        1625: {
            "Acronym": "DI",
            "DataRange": "0 to 255 per byte",
            "EndBit": -1,
            "Name": "Driver 1 identification",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": -2,
            "SPN": 1625,
            "SPNLength": 13824,
            "StartBit": -1,
            "TransmissionRate": "On request",
            "Units": "ASCII"
        },
        1626: {
            "Acronym": "DI",
            "DataRange": "0 to 255 per byte",
            "EndBit": -2,
            "Name": "Driver 2 identification",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": -2,
            "SPN": 1626,
            "SPNLength": 13824,
            "StartBit": -1,
            "TransmissionRate": "On request",
            "Units": "ASCII"
        },
        163: {
            "Acronym": "ETC2",
            "DataRange": "0 to 255 per byte",
            "EndBit": 63,
            "Name": "Transmission Current Range",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -2,
            "SPN": 163,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "ASCII"
        },
        1632: {
            "Acronym": "ETH",
            "DataRange": "0 to 7",
            "EndBit": 116,
            "Name": "Engine Torque Limit Feature",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 1632,
            "SPNLength": 3,
            "StartBit": 114,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        1633: {
            "Acronym": "CCVS1",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Cruise Control Pause Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1633,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1634: {
            "Acronym": "DM19",
            "DataRange": "0 to 4,294,967,295",
            "EndBit": 31,
            "Name": "Calibration Verification Number",
            "Offset": 0.0,
            "OperationalHigh": 4294967295.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 1634,
            "SPNLength": 32,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "count"
        },
        1635: {
            "Acronym": "DM19",
            "DataRange": "0 to 255 per byte",
            "EndBit": 159,
            "Name": "Calibration Identification",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": -2,
            "SPN": 1635,
            "SPNLength": 128,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "ASCII"
        },
        1636: {
            "Acronym": "ET3",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 15,
            "Name": "Engine Intake Manifold 1 Air Temperature (High Resolution)",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 1636,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1637: {
            "Acronym": "ET3",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 31,
            "Name": "Engine Coolant Temperature (High Resolution)",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 1637,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1638: {
            "Acronym": "VF",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 7,
            "Name": "Hydraulic Temperature",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 1638,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1639: {
            "Acronym": "FD1",
            "DataRange": "0 to 8,031.875 rpm",
            "EndBit": 31,
            "Name": "Fan Speed",
            "Offset": 0.0,
            "OperationalHigh": 8031.875,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.125,
            "SPN": 1639,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "1 s",
            "Units": "rpm"
        },
        164: {
            "Acronym": "EFL/P2",
            "DataRange": "0.0 to 36404.472237999995 psi",
            "EndBit": 15,
            "Name": "Engine Injection Control Pressure",
            "Offset": 0.0,
            "OperationalHigh": 36404.472237999995,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.5665536640625,
            "SPN": 164,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "0.5 s",
            "Units": "psi"
        },
        1640: {
            "Acronym": "DM14",
            "DataRange": "0 to 2047",
            "EndBit": 10,
            "Name": "Length/Number Requested",
            "Offset": 0.0,
            "OperationalHigh": 2047.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1640,
            "SPNLength": 11,
            "StartBit": 0,
            "TransmissionRate": "As needed",
            "Units": "bit"
        },
        1641: {
            "Acronym": "DM14",
            "DataRange": "0 to 1",
            "EndBit": 12,
            "Name": "Pointer Type",
            "Offset": 0.0,
            "OperationalHigh": 1.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1641,
            "SPNLength": 1,
            "StartBit": 12,
            "TransmissionRate": "As needed",
            "Units": "bit"
        },
        1642: {
            "Acronym": "DM14",
            "DataRange": "0 to 7",
            "EndBit": 11,
            "Name": "Command",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1642,
            "SPNLength": 3,
            "StartBit": 9,
            "TransmissionRate": "As needed",
            "Units": "bit"
        },
        1643: {
            "Acronym": "DM14",
            "DataRange": "0 to 255",
            "EndBit": 47,
            "Name": "Pointer Extension",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1643,
            "SPNLength": 8,
            "StartBit": 40,
            "TransmissionRate": "As needed",
            "Units": "bit"
        },
        1644: {
            "Acronym": "DM14",
            "DataRange": "0 to 16,777,216",
            "EndBit": 39,
            "Name": "Pointer ",
            "Offset": 0.0,
            "OperationalHigh": 16777216.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1644,
            "SPNLength": 24,
            "StartBit": 16,
            "TransmissionRate": "As needed",
            "Units": "count"
        },
        1645: {
            "Acronym": "DM14",
            "DataRange": "0 to 65,535",
            "EndBit": 63,
            "Name": "Key/User_Level",
            "Offset": 0.0,
            "OperationalHigh": 65535.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1645,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "As needed",
            "Units": "count"
        },
        1646: {
            "Acronym": "DM15",
            "DataRange": "0 to 7",
            "EndBit": 11,
            "Name": "Status (for DM 15)",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1646,
            "SPNLength": 3,
            "StartBit": 9,
            "TransmissionRate": "As needed",
            "Units": "bit"
        },
        1647: {
            "Acronym": "DM15",
            "DataRange": "0 to 255",
            "EndBit": 47,
            "Name": "EDCP Extension",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1647,
            "SPNLength": 8,
            "StartBit": 40,
            "TransmissionRate": "As needed",
            "Units": "bit"
        },
        1648: {
            "Acronym": "DM15",
            "DataRange": "0 to 16777215",
            "EndBit": 39,
            "Name": "Error Indicator/EDC Parameter",
            "Offset": 0.0,
            "OperationalHigh": 16777215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1648,
            "SPNLength": 24,
            "StartBit": 16,
            "TransmissionRate": "As needed",
            "Units": "bit"
        },
        1649: {
            "Acronym": "DM15",
            "DataRange": "0 to 2047",
            "EndBit": 10,
            "Name": "Length/Number Allowed",
            "Offset": 0.0,
            "OperationalHigh": 2047.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1649,
            "SPNLength": 11,
            "StartBit": 0,
            "TransmissionRate": "As needed",
            "Units": "bit"
        },
        165: {
            "Acronym": "VDS",
            "DataRange": "0 to 501.99 deg",
            "EndBit": 15,
            "Name": "Compass Bearing",
            "Offset": 0.0,
            "OperationalHigh": 501.99,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0078125,
            "SPN": 165,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "deg"
        },
        1650: {
            "Acronym": "DM16",
            "DataRange": "0 to 255",
            "EndBit": 7,
            "Name": "Number of Occurrences of Raw Binary Data",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 1650,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "As needed",
            "Units": "count"
        },
        1651: {
            "Acronym": "DM16",
            "DataRange": "manufacturer defined",
            "EndBit": -1,
            "Name": "Raw Binary Data",
            "Offset": 0,
            "OperationalHigh": -1,
            "OperationalLow": -1,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0,
            "SPN": 1651,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "As needed",
            "Units": ""
        },
        1652: {
            "Acronym": "DM17",
            "DataRange": "manufacturer defined",
            "EndBit": 63,
            "Name": "Boot Load Data",
            "Offset": 0,
            "OperationalHigh": -1,
            "OperationalLow": -1,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0,
            "SPN": 1652,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "As needed",
            "Units": ""
        },
        1653: {
            "Acronym": "CM1",
            "DataRange": "0 to 3",
            "EndBit": 39,
            "Name": "Vehicle Limiting Speed Governor Enable Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1653,
            "SPNLength": 2,
            "StartBit": 38,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        1654: {
            "Acronym": "CM1",
            "DataRange": "0 to 3",
            "EndBit": 37,
            "Name": "Vehicle Limiting Speed Governor Increment Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1654,
            "SPNLength": 2,
            "StartBit": 36,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        1655: {
            "Acronym": "CM1",
            "DataRange": "0 to 3",
            "EndBit": 35,
            "Name": "Vehicle Limiting Speed Governor Decrement Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1655,
            "SPNLength": 2,
            "StartBit": 34,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        1656: {
            "Acronym": "CM1",
            "DataRange": "0 to 3",
            "EndBit": 47,
            "Name": "Engine Automatic Start Enable Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1656,
            "SPNLength": 2,
            "StartBit": 46,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        166: {
            "Acronym": "EEC4",
            "DataRange": "0.0 to 43083.687196475 hp",
            "EndBit": 15,
            "Name": "Engine Rated Power",
            "Offset": 0.0,
            "OperationalHigh": 43083.687196475,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.670511045,
            "SPN": 166,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "hp"
        },
        1665: {
            "Acronym": "TC",
            "DataRange": "0 to 3",
            "EndBit": 31,
            "Name": "Engine Turbocharger Oil Level Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1665,
            "SPNLength": 2,
            "StartBit": 30,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        1666: {
            "Acronym": "CM1",
            "DataRange": "0 to 3",
            "EndBit": 45,
            "Name": "Automatic Gear Shifting Enable Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1666,
            "SPNLength": 2,
            "StartBit": 44,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        1667: {
            "Acronym": "ERC1",
            "DataRange": "0 to 3",
            "EndBit": 27,
            "Name": "Retarder Requesting Brake Light",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1667,
            "SPNLength": 2,
            "StartBit": 26,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        167: {
            "Acronym": "VEP1",
            "DataRange": "0 to 3212.75 V",
            "EndBit": 31,
            "Name": "Charging System Potential (Voltage)",
            "Offset": 0.0,
            "OperationalHigh": 3212.75,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05,
            "SPN": 167,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "1 s",
            "Units": "Volts"
        },
        1675: {
            "Acronym": "EEC1",
            "DataRange": "0 to 15",
            "EndBit": 51,
            "Name": "Engine Starter Mode",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1675,
            "SPNLength": 4,
            "StartBit": 48,
            "TransmissionRate": "engine speed dependent",
            "Units": "bit"
        },
        1676: {
            "Acronym": "HTR",
            "DataRange": "0 to 3",
            "EndBit": 33,
            "Name": "Auxiliary Heater Water Pump Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1676,
            "SPNLength": 2,
            "StartBit": 32,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        1677: {
            "Acronym": "HTR",
            "DataRange": "0 to 15",
            "EndBit": 27,
            "Name": "Auxiliary Heater Mode",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1677,
            "SPNLength": 4,
            "StartBit": 24,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        1678: {
            "Acronym": "HTR",
            "DataRange": "0 to 3",
            "EndBit": 35,
            "Name": "Cab Ventilation",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1678,
            "SPNLength": 2,
            "StartBit": 34,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        1679: {
            "Acronym": "HTR",
            "DataRange": "0 to 3",
            "EndBit": 37,
            "Name": "Engine Heating Zone",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1679,
            "SPNLength": 2,
            "StartBit": 36,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        168: {
            "Acronym": "VEP1",
            "DataRange": "0 to 3212.75 V",
            "EndBit": 47,
            "Name": "Battery Potential / Power Input 1",
            "Offset": 0.0,
            "OperationalHigh": 3212.75,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05,
            "SPN": 168,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "1 s",
            "Units": "Volts"
        },
        1680: {
            "Acronym": "HTR",
            "DataRange": "0 to 3",
            "EndBit": 39,
            "Name": "Cab Heating Zone",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1680,
            "SPNLength": 2,
            "StartBit": 38,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        1681: {
            "Acronym": "BM",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Battery Main Switch Hold State",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1681,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        1682: {
            "Acronym": "CM1",
            "DataRange": "0 to 3",
            "EndBit": 27,
            "Name": "Battery Main Switch Hold Request",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1682,
            "SPNLength": 2,
            "StartBit": 26,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        1683: {
            "Acronym": "CM1",
            "DataRange": "0 to 15",
            "EndBit": 51,
            "Name": "Auxiliary Heater Mode Request",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1683,
            "SPNLength": 4,
            "StartBit": 48,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        1684: {
            "Acronym": "CM1",
            "DataRange": "0 to 3",
            "EndBit": 25,
            "Name": "Auxiliary Heater Coolant Pump Request",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1684,
            "SPNLength": 2,
            "StartBit": 24,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        1685: {
            "Acronym": "CM1",
            "DataRange": "0 to 3",
            "EndBit": 53,
            "Name": "Request Engine Zone Heating",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1685,
            "SPNLength": 2,
            "StartBit": 52,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        1686: {
            "Acronym": "CM1",
            "DataRange": "0 to 3",
            "EndBit": 55,
            "Name": "Request Cab Zone Heating",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1686,
            "SPNLength": 2,
            "StartBit": 54,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        1687: {
            "Acronym": "HTR",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 7,
            "Name": "Auxiliary Heater Output Coolant Temperature",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 1687,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1688: {
            "Acronym": "HTR",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 15,
            "Name": "Auxiliary Heater Input Air Temperature",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 1688,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1689: {
            "Acronym": "HTR",
            "DataRange": "0 to 100 %",
            "EndBit": 23,
            "Name": "Auxiliary Heater Output Power Percent",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 1689,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "1 s",
            "Units": "%"
        },
        169: {
            "Acronym": "DD",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 47,
            "Name": "Cargo Ambient Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 169,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "1s",
            "Units": "deg F"
        },
        1690: {
            "Acronym": "CCC",
            "DataRange": "0 to 128,510 W",
            "EndBit": 15,
            "Name": "Auxiliary Heater Maximum Output Power",
            "Offset": 0.0,
            "OperationalHigh": 128510.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 2,
            "SPN": 1690,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "W"
        },
        1691: {
            "Acronym": "CM1",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 23,
            "Name": "Cab Interior Temperature Command",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 1691,
            "SPNLength": 16,
            "StartBit": 8,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1692: {
            "Acronym": "AF2",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 23,
            "Name": "Engine Desired Absolute Intake Manifold Pressure (Turbo Boost Limit)",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 1692,
            "SPNLength": 16,
            "StartBit": 8,
            "TransmissionRate": "On request",
            "Units": "psi"
        },
        1693: {
            "Acronym": "AF2",
            "DataRange": "0 to 100 %",
            "EndBit": 31,
            "Name": "Engine Turbocharger Wastegate Valve Position",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 1693,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        1694: {
            "Acronym": "AF2",
            "DataRange": "-125 to 125 %",
            "EndBit": 39,
            "Name": "Engine Gas Mass Flow Sensor Fueling Correction",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1694,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        1695: {
            "Acronym": "EO1",
            "DataRange": "-125 to 125 %",
            "EndBit": 55,
            "Name": "Engine Exhaust Gas Oxygen Sensor Fueling Correction",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1695,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        1696: {
            "Acronym": "EO1",
            "DataRange": "0 to 3",
            "EndBit": 63,
            "Name": "Engine Exhaust Gas Oxygen Sensor Closed Loop Operation",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1696,
            "SPNLength": 2,
            "StartBit": 62,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        1697: {
            "Acronym": "TIRE",
            "DataRange": "0 to 3",
            "EndBit": 37,
            "Name": "CTI Wheel End Electrical Fault",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1697,
            "SPNLength": 2,
            "StartBit": 36,
            "TransmissionRate": "10 s",
            "Units": "bit"
        },
        1698: {
            "Acronym": "TIRE",
            "DataRange": "0 to 3",
            "EndBit": 35,
            "Name": "CTI Tire Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1698,
            "SPNLength": 2,
            "StartBit": 34,
            "TransmissionRate": "10 s",
            "Units": "bit"
        },
        1699: {
            "Acronym": "TIRE",
            "DataRange": "0 to 3",
            "EndBit": 33,
            "Name": "CTI Wheel Sensor Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1699,
            "SPNLength": 2,
            "StartBit": 32,
            "TransmissionRate": "10 s",
            "Units": "bit"
        },
        170: {
            "Acronym": "AMB",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 23,
            "Name": "Cab Interior Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 170,
            "SPNLength": 16,
            "StartBit": 8,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1700: {
            "Acronym": "FLI1",
            "DataRange": "0 to 3",
            "EndBit": 7,
            "Name": "Lane Departure Imminent, Left Side",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1700,
            "SPNLength": 2,
            "StartBit": 6,
            "TransmissionRate": "50 ms (only when active)",
            "Units": "bit"
        },
        1701: {
            "Acronym": "FLI1",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Lane Departure Imminent, Right Side",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1701,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "50 ms (only when active)",
            "Units": "bit"
        },
        1702: {
            "Acronym": "FLI2",
            "DataRange": "0 to 3",
            "EndBit": 7,
            "Name": "Lane Departure Indication Enable Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1702,
            "SPNLength": 2,
            "StartBit": 6,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1706: {
            "Acronym": "DM12",
            "DataRange": "0 to 1",
            "EndBit": 47,
            "Name": "SPN Conversion Method",
            "Offset": 0.0,
            "OperationalHigh": 1.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 1706,
            "SPNLength": 1,
            "StartBit": 47,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        1707: {
            "Acronym": "FD2",
            "DataRange": "0.0 to 4659.699927595 psi",
            "EndBit": 47,
            "Name": "Hydraulic Fan 2 Motor Pressure",
            "Offset": 0.0,
            "OperationalHigh": 4659.699927595,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.072518869,
            "SPN": 1707,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "1 s",
            "Units": "psi"
        },
        1708: {
            "Acronym": "FD2",
            "DataRange": "0 to 100 %",
            "EndBit": 55,
            "Name": "Fan 2 Drive Bypass Command Status",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 1708,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "1 s",
            "Units": "%"
        },
        171: {
            "Acronym": "AMB",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 39,
            "Name": "Ambient Air Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 171,
            "SPNLength": 16,
            "StartBit": 24,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1710: {
            "Acronym": "FLI2",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Lane Tracking Status Left Side",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1710,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1711: {
            "Acronym": "FLI2",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Lane Tracking Status Right Side",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1711,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1712: {
            "Acronym": "EC1",
            "DataRange": "0 to 8,031.875 rpm",
            "EndBit": 239,
            "Name": "Engine Extended Range Requested Speed Control Range Upper Limit (Engine configuration)",
            "Offset": 0.0,
            "OperationalHigh": 8031.875,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 39,
            "Resolution": 0.125,
            "SPN": 1712,
            "SPNLength": 16,
            "StartBit": 224,
            "TransmissionRate": "Every 5 s and on change of torque/speed points of more than 10% since last transmission but no faster than every 500 ms",
            "Units": "rpm"
        },
        1713: {
            "Acronym": "VF",
            "DataRange": "0 to 3",
            "EndBit": 9,
            "Name": "Hydraulic Oil Filter Restriction Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1713,
            "SPNLength": 2,
            "StartBit": 8,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        1714: {
            "Acronym": "CM1",
            "DataRange": "0 to 3",
            "EndBit": 29,
            "Name": "Operator Seat Direction Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1714,
            "SPNLength": 2,
            "StartBit": 28,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        1715: {
            "Acronym": "ERC1",
            "DataRange": "-125 to 125 %",
            "EndBit": 47,
            "Name": "Drivers Demand Retarder -  Percent Torque",
            "Offset": -125.0,
            "OperationalHigh": 0.0,
            "OperationalLow": -125.0,
            "OperationalRange": "-125 to 0%",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1715,
            "SPNLength": 8,
            "StartBit": 40,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        1716: {
            "Acronym": "ERC1",
            "DataRange": "0 to 100 %",
            "EndBit": 55,
            "Name": "Retarder Selection, non-engine",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 1716,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        1717: {
            "Acronym": "ERC1",
            "DataRange": "-125 to 125 %",
            "EndBit": 63,
            "Name": "Actual Maximum Available Retarder - Percent Torque",
            "Offset": -125.0,
            "OperationalHigh": 0.0,
            "OperationalLow": -125.0,
            "OperationalRange": "-125 to 0",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1717,
            "SPNLength": 8,
            "StartBit": 56,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        1718: {
            "Acronym": "ASC2",
            "DataRange": "0 to 100 %",
            "EndBit": 31,
            "Name": "Damper Stiffness Request Front Axle",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 1718,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        1719: {
            "Acronym": "ASC2",
            "DataRange": "0 to 100 %",
            "EndBit": 39,
            "Name": "Damper Stiffness Request Rear Axle",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 1719,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        172: {
            "Acronym": "AMB",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 47,
            "Name": "Engine Air Intake Temperature",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 172,
            "SPNLength": 8,
            "StartBit": 40,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1720: {
            "Acronym": "ASC2",
            "DataRange": "0 to 100 %",
            "EndBit": 47,
            "Name": "Damper Stiffness Request Lift / Tag Axle",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 1720,
            "SPNLength": 8,
            "StartBit": 40,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        1721: {
            "Acronym": "ASC3",
            "DataRange": "-3,200 to 3,225.5 mm",
            "EndBit": 15,
            "Name": "Relative Level Front Axle Left",
            "Offset": -3.2,
            "OperationalHigh": 3225.5,
            "OperationalLow": -3200.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 1721,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "m"
        },
        1722: {
            "Acronym": "ASC3",
            "DataRange": "-3,200 to 3,225.5 mm",
            "EndBit": 31,
            "Name": "Relative Level Front Axle Right",
            "Offset": -3.2,
            "OperationalHigh": 3225.5,
            "OperationalLow": -3200.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 1722,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "m"
        },
        1723: {
            "Acronym": "ASC3",
            "DataRange": "-3,200 to 3,225.5 mm",
            "EndBit": 63,
            "Name": "Relative Level Rear Axle Right",
            "Offset": -3.2,
            "OperationalHigh": 3225.5,
            "OperationalLow": -3200.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 1723,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "m"
        },
        1724: {
            "Acronym": "ASC3",
            "DataRange": "-3,200 to 3,225.5 mm",
            "EndBit": 47,
            "Name": "Relative Level Rear Axle Left",
            "Offset": -3.2,
            "OperationalHigh": 3225.5,
            "OperationalLow": -3200.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 1724,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "m"
        },
        1725: {
            "Acronym": "ASC4",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 15,
            "Name": "Bellow Pressure Front Axle Left",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 1725,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        1726: {
            "Acronym": "ASC4",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 31,
            "Name": "Bellow Pressure Front Axle Right",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 1726,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        1727: {
            "Acronym": "ASC4",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 47,
            "Name": "Bellow Pressure Rear Axle Left",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 1727,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        1728: {
            "Acronym": "ASC4",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 63,
            "Name": "Bellow Pressure Rear Axle Right",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 1728,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        1729: {
            "Acronym": "ASC5",
            "DataRange": "0 to 100 %",
            "EndBit": 7,
            "Name": "Damper Stiffness Front Axle",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 1729,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        173: {
            "Acronym": "IC1",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 55,
            "Name": "Engine Exhaust Gas Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 173,
            "SPNLength": 16,
            "StartBit": 40,
            "TransmissionRate": "0.5 s",
            "Units": "deg F"
        },
        1730: {
            "Acronym": "ASC5",
            "DataRange": "0 to 100 %",
            "EndBit": 15,
            "Name": "Damper Stiffness Rear Axle",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 1730,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        1731: {
            "Acronym": "ASC5",
            "DataRange": "0 to 100 %",
            "EndBit": 23,
            "Name": "Damper Stiffness Lift / Tag Axle",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 1731,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        1732: {
            "Acronym": "ASC6",
            "DataRange": "-3,200 to 3,225.5 mm",
            "EndBit": 15,
            "Name": "Level Preset Front Axle Left",
            "Offset": -3.2,
            "OperationalHigh": 3225.5,
            "OperationalLow": -3200.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 1732,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "100 ms when active",
            "Units": "m"
        },
        1733: {
            "Acronym": "ASC1",
            "DataRange": "0 to 15",
            "EndBit": 7,
            "Name": "Nominal Level Rear Axle",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1733,
            "SPNLength": 4,
            "StartBit": 4,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1734: {
            "Acronym": "ASC1",
            "DataRange": "0 to 15",
            "EndBit": 3,
            "Name": "Nominal Level Front Axle",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1734,
            "SPNLength": 4,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1735: {
            "Acronym": "ASC6",
            "DataRange": "-3,200 to 3,225.5 mm",
            "EndBit": 63,
            "Name": "Level Preset Rear Axle Right",
            "Offset": -3.2,
            "OperationalHigh": 3225.5,
            "OperationalLow": -3200.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 1735,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "100 ms when active",
            "Units": "m"
        },
        1736: {
            "Acronym": "ASC1",
            "DataRange": "0 to 3",
            "EndBit": 15,
            "Name": "Above Nominal Level Rear Axle",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1736,
            "SPNLength": 2,
            "StartBit": 14,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1737: {
            "Acronym": "ASC1",
            "DataRange": "0 to 3",
            "EndBit": 13,
            "Name": "Above Nominal Level Front Axle",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1737,
            "SPNLength": 2,
            "StartBit": 12,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1738: {
            "Acronym": "ASC1",
            "DataRange": "0 to 3",
            "EndBit": 9,
            "Name": "Below Nominal Level Front Axle",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1738,
            "SPNLength": 2,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1739: {
            "Acronym": "ASC1",
            "DataRange": "0 to 3",
            "EndBit": 21,
            "Name": "Lifting Control Mode Front Axle",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1739,
            "SPNLength": 2,
            "StartBit": 20,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        174: {
            "Acronym": "ET1",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 15,
            "Name": "Engine Fuel Temperature 1",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 174,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1740: {
            "Acronym": "ASC1",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Lowering Control Mode Front Axle",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1740,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1741: {
            "Acronym": "ASC1",
            "DataRange": "0 to 15",
            "EndBit": 31,
            "Name": "Level Control Mode",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1741,
            "SPNLength": 4,
            "StartBit": 28,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1742: {
            "Acronym": "ASC1",
            "DataRange": "0 to 15",
            "EndBit": 27,
            "Name": "Kneeling Information",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1742,
            "SPNLength": 4,
            "StartBit": 24,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1743: {
            "Acronym": "ASC1",
            "DataRange": "0 to 3",
            "EndBit": 39,
            "Name": "Lift Axle 1 Position",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1743,
            "SPNLength": 2,
            "StartBit": 38,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1744: {
            "Acronym": "ASC1",
            "DataRange": "0 to 3",
            "EndBit": 37,
            "Name": "Door Release",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1744,
            "SPNLength": 2,
            "StartBit": 36,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1745: {
            "Acronym": "ASC1",
            "DataRange": "0 to 3",
            "EndBit": 35,
            "Name": "Vehicle Motion Inhibit",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1745,
            "SPNLength": 2,
            "StartBit": 34,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1746: {
            "Acronym": "ASC1",
            "DataRange": "0 to 3",
            "EndBit": 33,
            "Name": "Security Device",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1746,
            "SPNLength": 2,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1747: {
            "Acronym": "ASC2",
            "DataRange": "0 to 3",
            "EndBit": 7,
            "Name": "Kneeling Control Mode Request",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1747,
            "SPNLength": 2,
            "StartBit": 6,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1748: {
            "Acronym": "ASC2",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Kneeling Request Right Side",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1748,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1749: {
            "Acronym": "ASC2",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Kneeling Request Left Side",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1749,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        175: {
            "Acronym": "ET1",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 31,
            "Name": "Engine Oil Temperature 1",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 175,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1750: {
            "Acronym": "ASC2",
            "DataRange": "0 to 15",
            "EndBit": 15,
            "Name": "Nominal Level Request Rear Axle",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1750,
            "SPNLength": 4,
            "StartBit": 12,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1751: {
            "Acronym": "ASC2",
            "DataRange": "0 to 15",
            "EndBit": 11,
            "Name": "Nominal Level Request Front Axle",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1751,
            "SPNLength": 4,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1752: {
            "Acronym": "ASC2",
            "DataRange": "0 to 3",
            "EndBit": 21,
            "Name": "Lift Axle 1 Position Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1752,
            "SPNLength": 2,
            "StartBit": 20,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1753: {
            "Acronym": "ASC2",
            "DataRange": "0 to 15",
            "EndBit": 19,
            "Name": "Level Control Mode Request",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1753,
            "SPNLength": 4,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1754: {
            "Acronym": "ASC1",
            "DataRange": "0 to 3",
            "EndBit": 11,
            "Name": "Below Nominal Level Rear Axle",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1754,
            "SPNLength": 2,
            "StartBit": 10,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1755: {
            "Acronym": "ASC1",
            "DataRange": "0 to 3",
            "EndBit": 19,
            "Name": "Lowering Control Mode Rear Axle",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1755,
            "SPNLength": 2,
            "StartBit": 18,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1756: {
            "Acronym": "ASC1",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Lifting Control Mode Rear Axle",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1756,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1757: {
            "Acronym": "ASC6",
            "DataRange": "-3,200 to 3,225.5 mm",
            "EndBit": 31,
            "Name": "Level Preset Front Axle Right",
            "Offset": -3.2,
            "OperationalHigh": 3225.5,
            "OperationalLow": -3200.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 1757,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "100 ms when active",
            "Units": "m"
        },
        1758: {
            "Acronym": "ASC6",
            "DataRange": "-3,200 to 3,225.5 mm",
            "EndBit": 47,
            "Name": "Level Preset Rear Axle Left",
            "Offset": -3.2,
            "OperationalHigh": 3225.5,
            "OperationalLow": -3200.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 1758,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "100 ms when active",
            "Units": "m"
        },
        1759: {
            "Acronym": "LSP",
            "DataRange": "-209.7152 to 211.3929215 m",
            "EndBit": 47,
            "Name": "Blade Height Set Point - High Resolution",
            "Offset": -209.7152,
            "OperationalHigh": 209.7152,
            "OperationalLow": -209.7152,
            "OperationalRange": "-209.7152 to 209.7152m",
            "PGNLength": 8,
            "Resolution": 100,
            "SPN": 1759,
            "SPNLength": 32,
            "StartBit": 16,
            "TransmissionRate": "50 ms",
            "Units": "m"
        },
        176: {
            "Acronym": "ET1",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 47,
            "Name": "Engine Turbocharger Oil Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 176,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1760: {
            "Acronym": "CVW",
            "DataRange": "0 to 642,550 kg",
            "EndBit": 31,
            "Name": "Gross Combination Vehicle Weight",
            "Offset": 0.0,
            "OperationalHigh": 642550.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 10,
            "SPN": 1760,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "kg"
        },
        1761: {
            "Acronym": "AT1T1I",
            "DataRange": "0 to 100 %",
            "EndBit": 7,
            "Name": "Aftertreatment 1 Diesel Exhaust Fluid Tank Level",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 1761,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "%"
        },
        1762: {
            "Acronym": "HPG",
            "DataRange": "0.0 to 18638.79971038 psi",
            "EndBit": 15,
            "Name": "Hydraulic Pressure",
            "Offset": 0.0,
            "OperationalHigh": 18638.79971038,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.290075476,
            "SPN": 1762,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "50 ms",
            "Units": "psi"
        },
        1763: {
            "Acronym": "HPG",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Engine Hydraulic Pressure Governor Mode Indicator",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1763,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        1764: {
            "Acronym": "HPG",
            "DataRange": "0 to 3",
            "EndBit": 19,
            "Name": "Engine Hydraulic Pressure Governor Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1764,
            "SPNLength": 2,
            "StartBit": 18,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        1765: {
            "Acronym": "GFI2",
            "DataRange": "0 to 100 %",
            "EndBit": 55,
            "Name": "Engine Requested Fuel Valve 1 Position",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 1765,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        1766: {
            "Acronym": "GFI2",
            "DataRange": "0 to 100 %",
            "EndBit": 63,
            "Name": "Engine Requested Fuel Valve 2 Position",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 1766,
            "SPNLength": 8,
            "StartBit": 56,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        1767: {
            "Acronym": "GFD",
            "DataRange": "0 to 64.255",
            "EndBit": 15,
            "Name": "Specific Heat Ratio",
            "Offset": 0.0,
            "OperationalHigh": 2.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 2.0000",
            "PGNLength": 8,
            "Resolution": 0.001,
            "SPN": 1767,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "Ratio"
        },
        1768: {
            "Acronym": "ECT1",
            "DataRange": "0 to 8,000 rpm",
            "EndBit": 7,
            "Name": "Engine Low Limit Threshold for Maximum RPM from Engine",
            "Offset": 0.0,
            "OperationalHigh": 8000.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 32,
            "SPN": 1768,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "5 s",
            "Units": "rpm"
        },
        1769: {
            "Acronym": "ECT1",
            "DataRange": "0 to 8,000 rpm",
            "EndBit": 15,
            "Name": "Engine High Limit Threshold for Minimum Continuous Engine RPM",
            "Offset": 0.0,
            "OperationalHigh": 8000.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 32,
            "SPN": 1769,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "5 s",
            "Units": "rpm"
        },
        177: {
            "Acronym": "TRF1",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 47,
            "Name": "Transmission Oil Temperature 1",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 177,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1770: {
            "Acronym": "ECT1",
            "DataRange": "-125 to 125 %",
            "EndBit": 23,
            "Name": "Engine Low Limit Threshold for Maximum Torque from Engine",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1770,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "5 s",
            "Units": "%"
        },
        1771: {
            "Acronym": "ECT1",
            "DataRange": "-125 to 125 %",
            "EndBit": 31,
            "Name": "Engine High Limit Threshold for Minimum Continuous Torque from Engine",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1771,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "5 s",
            "Units": "%"
        },
        1772: {
            "Acronym": "ECT1",
            "DataRange": "0 to 8,000 rpm",
            "EndBit": 39,
            "Name": "Engine Maximum Continuous RPM",
            "Offset": 0.0,
            "OperationalHigh": 8000.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 32,
            "SPN": 1772,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "5 s",
            "Units": "rpm"
        },
        1773: {
            "Acronym": "ECT1",
            "DataRange": "0 to 8,000 rpm",
            "EndBit": 47,
            "Name": "Engine Minimum Continuous RPM",
            "Offset": 0.0,
            "OperationalHigh": 8000.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 32,
            "SPN": 1773,
            "SPNLength": 8,
            "StartBit": 40,
            "TransmissionRate": "5 s",
            "Units": "rpm"
        },
        1774: {
            "Acronym": "ECT1",
            "DataRange": "-125 to 125 %",
            "EndBit": 55,
            "Name": "Engine Maximum Continuous Torque",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1774,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "5 s",
            "Units": "%"
        },
        1775: {
            "Acronym": "ECT1",
            "DataRange": "-125 to 125 %",
            "EndBit": 63,
            "Name": "Engine Minimum Continuous Torque",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1775,
            "SPNLength": 8,
            "StartBit": 56,
            "TransmissionRate": "5 s",
            "Units": "%"
        },
        1776: {
            "Acronym": "RTC1",
            "DataRange": "0 to 8,000 rpm",
            "EndBit": 7,
            "Name": "Low Limit Threshold for Maximum RPM from Retarder",
            "Offset": 0.0,
            "OperationalHigh": 8000.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 32,
            "SPN": 1776,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "5 s",
            "Units": "rpm"
        },
        1777: {
            "Acronym": "RTC1",
            "DataRange": "0 to 8,000 rpm",
            "EndBit": 15,
            "Name": "High Limit Threshold for Minimum Continuous RPM from Retarder",
            "Offset": 0.0,
            "OperationalHigh": 8000.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 32,
            "SPN": 1777,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "5 s",
            "Units": "rpm"
        },
        1778: {
            "Acronym": "RTC1",
            "DataRange": "-125 to 125 %",
            "EndBit": 23,
            "Name": "Low Limit Threshold for Maximum Torque from Retarder",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1778,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "5 s",
            "Units": "%"
        },
        1779: {
            "Acronym": "RTC1",
            "DataRange": "-125 to 125 %",
            "EndBit": 31,
            "Name": "High Limit Threshold for Minimum Continuous Torque from Retarder",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1779,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "5 s",
            "Units": "%"
        },
        1780: {
            "Acronym": "RTC1",
            "DataRange": "0 to 8,000 rpm",
            "EndBit": 39,
            "Name": "Maximum Continuous Retarder Speed ",
            "Offset": 0.0,
            "OperationalHigh": 8000.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 32,
            "SPN": 1780,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "5 s",
            "Units": "rpm"
        },
        1781: {
            "Acronym": "RTC1",
            "DataRange": "0 to 8,000 rpm",
            "EndBit": 47,
            "Name": "Minimum Continuous Retarder Speed",
            "Offset": 0.0,
            "OperationalHigh": 8000.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 32,
            "SPN": 1781,
            "SPNLength": 8,
            "StartBit": 40,
            "TransmissionRate": "5 s",
            "Units": "rpm"
        },
        1782: {
            "Acronym": "RTC1",
            "DataRange": "-125 to 125 %",
            "EndBit": 55,
            "Name": "Maximum Continuous Retarder Torque",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1782,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "5 s",
            "Units": "%"
        },
        1783: {
            "Acronym": "RTC1",
            "DataRange": "-125 to 125 %",
            "EndBit": 63,
            "Name": "Minimum Continuous Retarder Torque",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1783,
            "SPNLength": 8,
            "StartBit": 56,
            "TransmissionRate": "5 s",
            "Units": "%"
        },
        1784: {
            "Acronym": "CTL",
            "DataRange": "0 to 8,000 rpm",
            "EndBit": 7,
            "Name": "Engine Speed Limit Request - Minimum Continuous",
            "Offset": 0.0,
            "OperationalHigh": 8000.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 32,
            "SPN": 1784,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "5 s",
            "Units": "rpm"
        },
        1785: {
            "Acronym": "CTL",
            "DataRange": "0 to 8,000 rpm",
            "EndBit": 15,
            "Name": "Engine Speed Limit Request - Maximum Continuous",
            "Offset": 0.0,
            "OperationalHigh": 8000.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 32,
            "SPN": 1785,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "5 s",
            "Units": "rpm"
        },
        1786: {
            "Acronym": "CTL",
            "DataRange": "-125 to 125 %",
            "EndBit": 23,
            "Name": "Engine Torque Limit Request - Minimum Continuous",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 125%",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1786,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "5 s",
            "Units": "%"
        },
        1787: {
            "Acronym": "CTL",
            "DataRange": "-125 to 125 %",
            "EndBit": 31,
            "Name": "Engine Torque Limit Request - Maximum Continuous",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 125%",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1787,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "5 s",
            "Units": "%"
        },
        1788: {
            "Acronym": "CTL",
            "DataRange": "0 to 8,000 rpm",
            "EndBit": 39,
            "Name": "Minimum Continuous Retarder Speed Limit Request",
            "Offset": 0.0,
            "OperationalHigh": 8000.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 32,
            "SPN": 1788,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "5 s",
            "Units": "rpm"
        },
        1789: {
            "Acronym": "CTL",
            "DataRange": "0 to 8,000 rpm",
            "EndBit": 47,
            "Name": "Maximum Continuous Retarder Speed Limit Request",
            "Offset": 0.0,
            "OperationalHigh": 8000.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 32,
            "SPN": 1789,
            "SPNLength": 8,
            "StartBit": 40,
            "TransmissionRate": "5 s",
            "Units": "rpm"
        },
        1790: {
            "Acronym": "CTL",
            "DataRange": "-125 to 125 %",
            "EndBit": 55,
            "Name": "Minimum Continuous Retarder Torque Limit Request",
            "Offset": -125.0,
            "OperationalHigh": 0.0,
            "OperationalLow": -125.0,
            "OperationalRange": "-125 to 0%",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1790,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "5 s",
            "Units": "%"
        },
        1791: {
            "Acronym": "CTL",
            "DataRange": "-125 to 125 %",
            "EndBit": 63,
            "Name": "Maximum Continuous Retarder Torque Limit Request",
            "Offset": -125.0,
            "OperationalHigh": 0.0,
            "OperationalLow": -125.0,
            "OperationalRange": "-125 to 0%",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1791,
            "SPNLength": 8,
            "StartBit": 56,
            "TransmissionRate": "5 s",
            "Units": "%"
        },
        1792: {
            "Acronym": "EBC1",
            "DataRange": "0 to 3",
            "EndBit": 63,
            "Name": "Tractor-Mounted Trailer ABS Warning Signal",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1792,
            "SPNLength": 2,
            "StartBit": 62,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1793: {
            "Acronym": "EBC1",
            "DataRange": "0 to 3",
            "EndBit": 47,
            "Name": "ATC/ASR Information Signal",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1793,
            "SPNLength": 2,
            "StartBit": 46,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1794: {
            "Acronym": "EC1",
            "DataRange": "0 to 257.02 kg-m^2",
            "EndBit": 255,
            "Name": "Engine Moment of Inertia",
            "Offset": 0.0,
            "OperationalHigh": 257.022,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 39,
            "Resolution": 0.004,
            "SPN": 1794,
            "SPNLength": 16,
            "StartBit": 240,
            "TransmissionRate": "Every 5 s and on change of torque/speed points of more than 10% since last transmission but no faster than every 500 ms",
            "Units": "kg/m2"
        },
        1795: {
            "Acronym": "VEP3",
            "DataRange": "-1600 to 1612.75 A",
            "EndBit": 15,
            "Name": "Alternator Current (High Range/Resolution)",
            "Offset": -1600.0,
            "OperationalHigh": 1612.75,
            "OperationalLow": -1600.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05,
            "SPN": 1795,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "Amps"
        },
        1796: {
            "Acronym": "ACC1",
            "DataRange": "0 to 3",
            "EndBit": 53,
            "Name": "ACC Distance Alert Signal",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1796,
            "SPNLength": 2,
            "StartBit": 52,
            "TransmissionRate": "100ms or upon state change, but not faster than 20 ms.",
            "Units": "bit"
        },
        1797: {
            "Acronym": "ACC1",
            "DataRange": "0 to 3",
            "EndBit": 51,
            "Name": "ACC System Shutoff Warning",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1797,
            "SPNLength": 2,
            "StartBit": 50,
            "TransmissionRate": "100ms or upon state change, but not faster than 20 ms.",
            "Units": "bit"
        },
        1798: {
            "Acronym": "ACC1",
            "DataRange": "0 to 3",
            "EndBit": 49,
            "Name": "ACC Target Detected",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1798,
            "SPNLength": 2,
            "StartBit": 48,
            "TransmissionRate": "100ms or upon state change, but not faster than 20 ms.",
            "Units": "bit"
        },
        1799: {
            "Acronym": "ACC2",
            "DataRange": "0 to 7",
            "EndBit": 7,
            "Name": "Requested ACC Distance Mode",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1799,
            "SPNLength": 3,
            "StartBit": 5,
            "TransmissionRate": "250 ms",
            "Units": "bit"
        },
        180: {
            "Acronym": "VW",
            "DataRange": "0 to 128,510 kg",
            "EndBit": 39,
            "Name": "Trailer Weight",
            "Offset": 0.0,
            "OperationalHigh": 128510.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 2,
            "SPN": 180,
            "SPNLength": 16,
            "StartBit": 24,
            "TransmissionRate": "On request",
            "Units": "kg"
        },
        1800: {
            "Acronym": "BT1",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 7,
            "Name": "Battery 1 Temperature",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 1800,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1801: {
            "Acronym": "BT1",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 15,
            "Name": "Battery 2 Temperature",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 1801,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1802: {
            "Acronym": "IMT2",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 31,
            "Name": "Engine Intake Manifold 5 Temperature",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 1802,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1803: {
            "Acronym": "IMT2",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 39,
            "Name": "Engine Intake Manifold 6 Temperature",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 1803,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        1804: {
            "Acronym": "CSA",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Engine Start Enable Device 2",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1804,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "As required",
            "Units": "bit"
        },
        1805: {
            "Acronym": "LVDD",
            "DataRange": "0 to 15",
            "EndBit": 11,
            "Name": "LED Display Mode Control",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1805,
            "SPNLength": 4,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1806: {
            "Acronym": "LVDD",
            "DataRange": "0 to 15",
            "EndBit": 15,
            "Name": "LED Display Deadband Control",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1806,
            "SPNLength": 4,
            "StartBit": 12,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1807: {
            "Acronym": "VDC2",
            "DataRange": "-31.374 to +31.374 rad",
            "EndBit": 15,
            "Name": "Steering Wheel Angle",
            "Offset": -31.374,
            "OperationalHigh": 31.374,
            "OperationalLow": -31.374,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1807,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "10 ms",
            "Units": "rad"
        },
        1808: {
            "Acronym": "VDC2",
            "DataRange": "-3.92 to +3.92 rad/s",
            "EndBit": 39,
            "Name": "Yaw Rate",
            "Offset": -3.92,
            "OperationalHigh": 3.92,
            "OperationalLow": -3.92,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1808,
            "SPNLength": 16,
            "StartBit": 24,
            "TransmissionRate": "10 ms",
            "Units": "rad/s"
        },
        1809: {
            "Acronym": "VDC2",
            "DataRange": "-15.687 to +15.687 m/s^2",
            "EndBit": 55,
            "Name": "Lateral Acceleration",
            "Offset": -15.687,
            "OperationalHigh": 15.6872,
            "OperationalLow": -15.687,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1809,
            "SPNLength": 16,
            "StartBit": 40,
            "TransmissionRate": "10 ms",
            "Units": "m/s^2"
        },
        181: {
            "Acronym": "VW",
            "DataRange": "0 to 128,510 kg",
            "EndBit": 55,
            "Name": "Cargo Weight",
            "Offset": 0.0,
            "OperationalHigh": 128510.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 2,
            "SPN": 181,
            "SPNLength": 16,
            "StartBit": 40,
            "TransmissionRate": "On request",
            "Units": "kg"
        },
        1810: {
            "Acronym": "VDC2",
            "DataRange": "-12.5 to +12.5 m/s^2",
            "EndBit": 63,
            "Name": "Longitudinal Acceleration",
            "Offset": -12.5,
            "OperationalHigh": 12.52,
            "OperationalLow": -12.5,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 1810,
            "SPNLength": 8,
            "StartBit": 56,
            "TransmissionRate": "10 ms",
            "Units": "m/s^2"
        },
        1811: {
            "Acronym": "VDC2",
            "DataRange": "-32 to 29 turns",
            "EndBit": 21,
            "Name": "Steering Wheel Turn Counter",
            "Offset": -32.0,
            "OperationalHigh": 10.0,
            "OperationalLow": -10.0,
            "OperationalRange": "-10 to +10 Turns",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1811,
            "SPNLength": 6,
            "StartBit": 16,
            "TransmissionRate": "10 ms",
            "Units": "count"
        },
        1812: {
            "Acronym": "VDC2",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Steering Wheel Angle Sensor Type",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1812,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "10 ms",
            "Units": "bit"
        },
        1813: {
            "Acronym": "VDC1",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "VDC Information Signal",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1813,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "100ms",
            "Units": "bit"
        },
        1814: {
            "Acronym": "VDC1",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "VDC Fully Operational",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1814,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "100ms",
            "Units": "bit"
        },
        1815: {
            "Acronym": "VDC1",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "VDC brake light request",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1815,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "100ms",
            "Units": "bit"
        },
        1816: {
            "Acronym": "VDC1",
            "DataRange": "0 to 3",
            "EndBit": 9,
            "Name": "ROP Engine Control active",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1816,
            "SPNLength": 2,
            "StartBit": 8,
            "TransmissionRate": "100ms",
            "Units": "bit"
        },
        1817: {
            "Acronym": "VDC1",
            "DataRange": "0 to 3",
            "EndBit": 13,
            "Name": "YC Engine Control active",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1817,
            "SPNLength": 2,
            "StartBit": 12,
            "TransmissionRate": "100ms",
            "Units": "bit"
        },
        1818: {
            "Acronym": "VDC1",
            "DataRange": "0 to 3",
            "EndBit": 11,
            "Name": "ROP Brake Control active",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1818,
            "SPNLength": 2,
            "StartBit": 10,
            "TransmissionRate": "100ms",
            "Units": "bit"
        },
        1819: {
            "Acronym": "VDC1",
            "DataRange": "0 to 3",
            "EndBit": 15,
            "Name": "YC Brake Control active",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1819,
            "SPNLength": 2,
            "StartBit": 14,
            "TransmissionRate": "100ms",
            "Units": "bit"
        },
        182: {
            "Acronym": "LFC",
            "DataRange": "0.0 to 556224873.36449 US Gal",
            "EndBit": 31,
            "Name": "Engine Trip Fuel",
            "Offset": 0.0,
            "OperationalHigh": 556224873.36449,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.132086,
            "SPN": 182,
            "SPNLength": 32,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "gallons"
        },
        1820: {
            "Acronym": "DC1",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Ramp / Wheel Chair Lift Position",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1820,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1821: {
            "Acronym": "DC1",
            "DataRange": "0 to 15",
            "EndBit": 3,
            "Name": "Position of doors",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1821,
            "SPNLength": 4,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1822: {
            "Acronym": "ASC1",
            "DataRange": "0 to 3",
            "EndBit": 47,
            "Name": "Lift Axle 2 Position",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1822,
            "SPNLength": 2,
            "StartBit": 46,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1823: {
            "Acronym": "ASC1",
            "DataRange": "0 to 3",
            "EndBit": 43,
            "Name": "Rear Axle in Bumper Range",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1823,
            "SPNLength": 2,
            "StartBit": 42,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1824: {
            "Acronym": "ASC1",
            "DataRange": "0 to 3",
            "EndBit": 41,
            "Name": "Front Axle in Bumper Range",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1824,
            "SPNLength": 2,
            "StartBit": 40,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1825: {
            "Acronym": "ASC1",
            "DataRange": "0 to 3",
            "EndBit": 51,
            "Name": "Suspension Remote control 2",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1825,
            "SPNLength": 2,
            "StartBit": 50,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1826: {
            "Acronym": "ASC1",
            "DataRange": "0 to 3",
            "EndBit": 49,
            "Name": "Suspension Remote Control 1",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1826,
            "SPNLength": 2,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1827: {
            "Acronym": "ASC1",
            "DataRange": "0 to 15",
            "EndBit": 59,
            "Name": "Suspension Control Refusal Information",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1827,
            "SPNLength": 4,
            "StartBit": 56,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1828: {
            "Acronym": "ASC2",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Lift Axle 2 Position Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1828,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1829: {
            "Acronym": "ASC2",
            "DataRange": "0 to 3",
            "EndBit": 51,
            "Name": "Kneeling Command - Rear Axle",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1829,
            "SPNLength": 2,
            "StartBit": 50,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        183: {
            "Acronym": "LFE1",
            "DataRange": "0.0 to 848.718593 US Gal/h",
            "EndBit": 15,
            "Name": "Engine Fuel Rate",
            "Offset": 0.0,
            "OperationalHigh": 848.718593,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0132086,
            "SPN": 183,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "gallons/h"
        },
        1830: {
            "Acronym": "ASC2",
            "DataRange": "0 to 3",
            "EndBit": 49,
            "Name": "Kneeling Command - Front Axle",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1830,
            "SPNLength": 2,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1831: {
            "Acronym": "ASC5",
            "DataRange": "0 to 3",
            "EndBit": 29,
            "Name": "Electronic Shock Absorber Control Mode - Lift/Tag Axle",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1831,
            "SPNLength": 2,
            "StartBit": 28,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1832: {
            "Acronym": "ASC5",
            "DataRange": "0 to 3",
            "EndBit": 27,
            "Name": "Electronic Shock Absorber Control Mode - Rear Axle",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1832,
            "SPNLength": 2,
            "StartBit": 26,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1833: {
            "Acronym": "ASC5",
            "DataRange": "0 to 3",
            "EndBit": 25,
            "Name": "Electronic Shock Absorber Control Mode -  Front Axle",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1833,
            "SPNLength": 2,
            "StartBit": 24,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1834: {
            "Acronym": "TAVG",
            "DataRange": "0.0 to 848.718593 US Gal/h",
            "EndBit": 15,
            "Name": "Engine Total Average Fuel Rate",
            "Offset": 0.0,
            "OperationalHigh": 848.718593,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0132086,
            "SPN": 1834,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "gallons/h"
        },
        1835: {
            "Acronym": "TAVG",
            "DataRange": "0.0 to 295.1897076301758 mile/US Gal",
            "EndBit": 31,
            "Name": "Engine Total Average Fuel Economy",
            "Offset": 0.0,
            "OperationalHigh": 295.1897076301758,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0045940348242187,
            "SPN": 1835,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "miles/gallon"
        },
        1836: {
            "Acronym": "EBC1",
            "DataRange": "0 to 3",
            "EndBit": 61,
            "Name": "Trailer ABS Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1836,
            "SPNLength": 2,
            "StartBit": 60,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1837: {
            "Acronym": "ML",
            "DataRange": "0 to 3",
            "EndBit": 7,
            "Name": "Convoy Driving Lamp Select",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1837,
            "SPNLength": 2,
            "StartBit": 6,
            "TransmissionRate": "500ms or upon state change, but not faster than 100 ms.",
            "Units": "bit"
        },
        1838: {
            "Acronym": "ML",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Convoy Lamp Select",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1838,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "500ms or upon state change, but not faster than 100 ms.",
            "Units": "bit"
        },
        1839: {
            "Acronym": "ML",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Front Black Out Marker Lamp Select",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1839,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "500ms or upon state change, but not faster than 100 ms.",
            "Units": "bit"
        },
        184: {
            "Acronym": "LFE1",
            "DataRange": "0.0 to 295.1897076301758 mile/US Gal",
            "EndBit": 31,
            "Name": "Engine Instantaneous Fuel Economy",
            "Offset": 0.0,
            "OperationalHigh": 295.1897076301758,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0045940348242187,
            "SPN": 184,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "miles/gallon"
        },
        1840: {
            "Acronym": "ML",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Rear Black Out Marker Select",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1840,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "500ms or upon state change, but not faster than 100 ms.",
            "Units": "bit"
        },
        1841: {
            "Acronym": "ML",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Black Out Brake/Stop Lamp Select",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1841,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "500ms or upon state change, but not faster than 100 ms.",
            "Units": "bit"
        },
        1842: {
            "Acronym": "ML",
            "DataRange": "0 to 3",
            "EndBit": 31,
            "Name": "Black Out Work Lamp Select",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1842,
            "SPNLength": 2,
            "StartBit": 30,
            "TransmissionRate": "500ms or upon state change, but not faster than 100 ms.",
            "Units": "bit"
        },
        1843: {
            "Acronym": "ML",
            "DataRange": "0 to 3",
            "EndBit": 25,
            "Name": "Night Vision Illuminator Select",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1843,
            "SPNLength": 2,
            "StartBit": 24,
            "TransmissionRate": "500ms or upon state change, but not faster than 100 ms.",
            "Units": "bit"
        },
        1844: {
            "Acronym": "ML",
            "DataRange": "0 to 100 %",
            "EndBit": 63,
            "Name": "Operators Black Out Intensity Selection",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 1844,
            "SPNLength": 8,
            "StartBit": 56,
            "TransmissionRate": "500ms or upon state change, but not faster than 100 ms.",
            "Units": "%"
        },
        1845: {
            "Acronym": "TCFG2",
            "DataRange": "0 to 64,255 Nm",
            "EndBit": 15,
            "Name": "Transmission Torque Limit",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 1845,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On request or sender may transmit every 5 seconds until acknowledged by reception of the engine configuration message PGN 65251 SPN 1846.",
            "Units": "Nm"
        },
        1846: {
            "Acronym": "EC1",
            "DataRange": "0 to 64,255 Nm",
            "EndBit": 271,
            "Name": "Engine Default Torque Limit",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 39,
            "Resolution": 1,
            "SPN": 1846,
            "SPNLength": 16,
            "StartBit": 256,
            "TransmissionRate": "Every 5 s and on change of torque/speed points of more than 10% since last transmission but no faster than every 500 ms",
            "Units": "Nm"
        },
        1849: {
            "Acronym": "ETC7",
            "DataRange": "0 to 3",
            "EndBit": 7,
            "Name": "Transmission Requested Range Display Flash State",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1849,
            "SPNLength": 2,
            "StartBit": 6,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        185: {
            "Acronym": "LFE1",
            "DataRange": "0.0 to 295.1897076301758 mile/US Gal",
            "EndBit": 47,
            "Name": "Engine Average Fuel Economy",
            "Offset": 0.0,
            "OperationalHigh": 295.1897076301758,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0045940348242187,
            "SPN": 185,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "miles/gallon"
        },
        1850: {
            "Acronym": "ETC7",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Transmission Requested Range Display Blank State",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1850,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1851: {
            "Acronym": "ETC7",
            "DataRange": "0 to 3",
            "EndBit": 15,
            "Name": "Transmission Shift Inhibit Indicator",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1851,
            "SPNLength": 2,
            "StartBit": 14,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1852: {
            "Acronym": "TC1",
            "DataRange": "0 to 3",
            "EndBit": 41,
            "Name": "Transmission Mode 1",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1852,
            "SPNLength": 2,
            "StartBit": 40,
            "TransmissionRate": "When active; 50 ms to transmission and axles",
            "Units": "bit"
        },
        1853: {
            "Acronym": "TC1",
            "DataRange": "0 to 3",
            "EndBit": 43,
            "Name": "Transmission Mode 2",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1853,
            "SPNLength": 2,
            "StartBit": 42,
            "TransmissionRate": "When active; 50 ms to transmission and axles",
            "Units": "bit"
        },
        1854: {
            "Acronym": "TC1",
            "DataRange": "0 to 3",
            "EndBit": 45,
            "Name": "Transmission Mode 3",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1854,
            "SPNLength": 2,
            "StartBit": 44,
            "TransmissionRate": "When active; 50 ms to transmission and axles",
            "Units": "bit"
        },
        1855: {
            "Acronym": "TC1",
            "DataRange": "0 to 3",
            "EndBit": 47,
            "Name": "Transmission Mode 4",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1855,
            "SPNLength": 2,
            "StartBit": 46,
            "TransmissionRate": "When active; 50 ms to transmission and axles",
            "Units": "bit"
        },
        1856: {
            "Acronym": "CM1",
            "DataRange": "0 to 3",
            "EndBit": 31,
            "Name": "Seat Belt Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1856,
            "SPNLength": 2,
            "StartBit": 30,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        1857: {
            "Acronym": "VF",
            "DataRange": "0 to 3",
            "EndBit": 11,
            "Name": "Winch Oil Pressure Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1857,
            "SPNLength": 2,
            "StartBit": 10,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        186: {
            "Acronym": "PTO",
            "DataRange": "0 to 8,031.875 rpm",
            "EndBit": 23,
            "Name": "Power Takeoff Speed",
            "Offset": 0.0,
            "OperationalHigh": 8031.875,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.125,
            "SPN": 186,
            "SPNLength": 16,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "rpm"
        },
        1861: {
            "Acronym": "GBSD",
            "DataRange": "0 to 3",
            "EndBit": 57,
            "Name": "Ground Based Direction",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1861,
            "SPNLength": 2,
            "StartBit": 56,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1862: {
            "Acronym": "WBSD",
            "DataRange": "0 to 64.255 m/s",
            "EndBit": 15,
            "Name": "Wheel Based Speed",
            "Offset": 0.0,
            "OperationalHigh": 64.255,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.001,
            "SPN": 1862,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "m/s"
        },
        1863: {
            "Acronym": "WBSD",
            "DataRange": "0 to 4,211,081.215 m",
            "EndBit": 47,
            "Name": "Wheel Based Distance",
            "Offset": 0.0,
            "OperationalHigh": 4211081.215,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.001,
            "SPN": 1863,
            "SPNLength": 32,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "m"
        },
        1864: {
            "Acronym": "WBSD",
            "DataRange": "0 to 3",
            "EndBit": 57,
            "Name": "Wheel Based Direction",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1864,
            "SPNLength": 2,
            "StartBit": 56,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1865: {
            "Acronym": "WBSD",
            "DataRange": "0 to 3",
            "EndBit": 59,
            "Name": "Key Switch NOT OFF",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1865,
            "SPNLength": 2,
            "StartBit": 58,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1866: {
            "Acronym": "WBSD",
            "DataRange": "0 to 250 mins",
            "EndBit": 55,
            "Name": "Maximum Time of Tractor Power",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1866,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "minutes"
        },
        1867: {
            "Acronym": "MP",
            "DataRange": "0 to 3",
            "EndBit": 7,
            "Name": "Maintain ECU Power",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1867,
            "SPNLength": 2,
            "StartBit": 6,
            "TransmissionRate": "As required after receiving the message indicating that the ignition switch as change from ON state to OFF state or on change of state of parameters",
            "Units": "bit"
        },
        1868: {
            "Acronym": "MP",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Maintain Acuator Power",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1868,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "As required after receiving the message indicating that the ignition switch as change from ON state to OFF state or on change of state of parameters",
            "Units": "bit"
        },
        1869: {
            "Acronym": "MP",
            "DataRange": "0 to 3",
            "EndBit": 15,
            "Name": "Implement Transport State",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1869,
            "SPNLength": 2,
            "StartBit": 14,
            "TransmissionRate": "As required after receiving the message indicating that the ignition switch as change from ON state to OFF state or on change of state of parameters",
            "Units": "bit"
        },
        187: {
            "Acronym": "PTO",
            "DataRange": "0 to 8,031.875 rpm",
            "EndBit": 39,
            "Name": "Power Takeoff Set Speed",
            "Offset": 0.0,
            "OperationalHigh": 8031.875,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.125,
            "SPN": 187,
            "SPNLength": 16,
            "StartBit": 24,
            "TransmissionRate": "100 ms",
            "Units": "rpm"
        },
        1870: {
            "Acronym": "MP",
            "DataRange": "0 to 3",
            "EndBit": 13,
            "Name": "Implement Park State",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1870,
            "SPNLength": 2,
            "StartBit": 12,
            "TransmissionRate": "As required after receiving the message indicating that the ignition switch as change from ON state to OFF state or on change of state of parameters",
            "Units": "bit"
        },
        1871: {
            "Acronym": "MP",
            "DataRange": "0 to 3",
            "EndBit": 11,
            "Name": "Implement Work State",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1871,
            "SPNLength": 2,
            "StartBit": 10,
            "TransmissionRate": "As required after receiving the message indicating that the ignition switch as change from ON state to OFF state or on change of state of parameters",
            "Units": "bit"
        },
        1872: {
            "Acronym": "FHS",
            "DataRange": "0 to 100 %",
            "EndBit": 7,
            "Name": "Front Hitch Position",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 1872,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "100ms",
            "Units": "%"
        },
        1873: {
            "Acronym": "RHS",
            "DataRange": "0 to 100 %",
            "EndBit": 7,
            "Name": "Rear Hitch Position",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 1873,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "100ms",
            "Units": "%"
        },
        1874: {
            "Acronym": "HPTOC",
            "DataRange": "0 to 100 %",
            "EndBit": 7,
            "Name": "Front Hitch Position Command",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 1874,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "Minimum rate of 1 second between messages or when a parameter is required to change state.",
            "Units": "%"
        },
        1875: {
            "Acronym": "HPTOC",
            "DataRange": "0 to 100 %",
            "EndBit": 15,
            "Name": "Rear Hitch Position Command",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 1875,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "Minimum rate of 1 second between messages or when a parameter is required to change state.",
            "Units": "%"
        },
        1876: {
            "Acronym": "FHS",
            "DataRange": "0 to 3",
            "EndBit": 15,
            "Name": "Front Hitch In-work Indication",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1876,
            "SPNLength": 2,
            "StartBit": 14,
            "TransmissionRate": "100ms",
            "Units": "bit"
        },
        1877: {
            "Acronym": "RHS",
            "DataRange": "0 to 3",
            "EndBit": 15,
            "Name": "Rear Hitch In-work Indication",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1877,
            "SPNLength": 2,
            "StartBit": 14,
            "TransmissionRate": "100ms",
            "Units": "bit"
        },
        1878: {
            "Acronym": "FHS",
            "DataRange": "-320,000 to 322,550 N",
            "EndBit": 39,
            "Name": "Front Draft",
            "Offset": -320000.0,
            "OperationalHigh": 322550.0,
            "OperationalLow": -320000.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 10,
            "SPN": 1878,
            "SPNLength": 16,
            "StartBit": 24,
            "TransmissionRate": "100ms",
            "Units": "N"
        },
        1879: {
            "Acronym": "RHS",
            "DataRange": "-320,000 to 322,550 N",
            "EndBit": 39,
            "Name": "Rear Draft",
            "Offset": -320000.0,
            "OperationalHigh": 322550.0,
            "OperationalLow": -320000.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 10,
            "SPN": 1879,
            "SPNLength": 16,
            "StartBit": 24,
            "TransmissionRate": "100ms",
            "Units": "N"
        },
        188: {
            "Acronym": "EC1",
            "DataRange": "0 to 8,031.875 rpm",
            "EndBit": 15,
            "Name": "Engine Speed At Idle, Point 1 (Engine Configuration)",
            "Offset": 0.0,
            "OperationalHigh": 8031.875,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 39,
            "Resolution": 0.125,
            "SPN": 188,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "Every 5 s and on change of torque/speed points of more than 10% since last transmission but no faster than every 500 ms",
            "Units": "rpm"
        },
        1880: {
            "Acronym": "FHS",
            "DataRange": "-100 to 100 %",
            "EndBit": 23,
            "Name": "Front Nominal Lower Link Force",
            "Offset": -100.0,
            "OperationalHigh": 100.0,
            "OperationalLow": -100.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.8,
            "SPN": 1880,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "100ms",
            "Units": "%"
        },
        1881: {
            "Acronym": "RHS",
            "DataRange": "-100 to 100 %",
            "EndBit": 23,
            "Name": "Rear  Nominal Lower Link Force",
            "Offset": -100.0,
            "OperationalHigh": 100.0,
            "OperationalLow": -100.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.8,
            "SPN": 1881,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "100ms",
            "Units": "%"
        },
        1882: {
            "Acronym": "FPTO",
            "DataRange": "0 to 8,031.875 rpm",
            "EndBit": 15,
            "Name": "Front PTO output shaft speed",
            "Offset": 0.0,
            "OperationalHigh": 8031.875,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.125,
            "SPN": 1882,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "100ms when engage, otherwise on request",
            "Units": "rpm"
        },
        1883: {
            "Acronym": "RPTO",
            "DataRange": "0 to 8,031.875 rpm",
            "EndBit": 15,
            "Name": "Rear PTO output shaft speed",
            "Offset": 0.0,
            "OperationalHigh": 8031.875,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.125,
            "SPN": 1883,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "100ms when engage, otherwise on request",
            "Units": "rpm"
        },
        1884: {
            "Acronym": "FPTO",
            "DataRange": "0 to 8,031.875 rpm",
            "EndBit": 31,
            "Name": "Front PTO Output Shaft Speed Set Point",
            "Offset": 0.0,
            "OperationalHigh": 8031.875,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.125,
            "SPN": 1884,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "100ms when engage, otherwise on request",
            "Units": "rpm"
        },
        1885: {
            "Acronym": "RPTO",
            "DataRange": "0 to 8,031.875 rpm",
            "EndBit": 31,
            "Name": "Rear PTO Output Shaft Speed Set Point",
            "Offset": 0.0,
            "OperationalHigh": 8031.875,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.125,
            "SPN": 1885,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "100ms when engage, otherwise on request",
            "Units": "rpm"
        },
        1886: {
            "Acronym": "HPTOC",
            "DataRange": "0 to 8,031.875 rpm",
            "EndBit": 31,
            "Name": "Front PTO Output Shaft Speed Set Point Command",
            "Offset": 0.0,
            "OperationalHigh": 8031.875,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.125,
            "SPN": 1886,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "Minimum rate of 1 second between messages or when a parameter is required to change state.",
            "Units": "rpm"
        },
        1887: {
            "Acronym": "HPTOC",
            "DataRange": "0 to 8,031.875 rpm",
            "EndBit": 47,
            "Name": "Rear PTO Output Shaft Speed Set Point Command",
            "Offset": 0.0,
            "OperationalHigh": 8031.875,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.125,
            "SPN": 1887,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "Minimum rate of 1 second between messages or when a parameter is required to change state.",
            "Units": "rpm"
        },
        1888: {
            "Acronym": "FPTO",
            "DataRange": "0 to 3",
            "EndBit": 39,
            "Name": "Front Power Take Off Engagement",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1888,
            "SPNLength": 2,
            "StartBit": 38,
            "TransmissionRate": "100ms when engage, otherwise on request",
            "Units": "bit"
        },
        1889: {
            "Acronym": "FPTO",
            "DataRange": "0 to 3",
            "EndBit": 37,
            "Name": "Front Power Take Off Mode",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1889,
            "SPNLength": 2,
            "StartBit": 36,
            "TransmissionRate": "100ms when engage, otherwise on request",
            "Units": "bit"
        },
        189: {
            "Acronym": "EEC4",
            "DataRange": "0 to 8,031.875 rpm",
            "EndBit": 31,
            "Name": "Engine Rated Speed",
            "Offset": 0.0,
            "OperationalHigh": 8031.875,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.125,
            "SPN": 189,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "rpm"
        },
        1890: {
            "Acronym": "RPTO",
            "DataRange": "0 to 3",
            "EndBit": 37,
            "Name": "Rear Power Take Off Mode",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1890,
            "SPNLength": 2,
            "StartBit": 36,
            "TransmissionRate": "100ms when engage, otherwise on request",
            "Units": "bit"
        },
        1891: {
            "Acronym": "FPTO",
            "DataRange": "0 to 3",
            "EndBit": 35,
            "Name": "Front Power Take Off Economy Mode",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1891,
            "SPNLength": 2,
            "StartBit": 34,
            "TransmissionRate": "100ms when engage, otherwise on request",
            "Units": "bit"
        },
        1892: {
            "Acronym": "RPTO",
            "DataRange": "0 to 3",
            "EndBit": 35,
            "Name": "Rear Power Take Off Economy Mode",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1892,
            "SPNLength": 2,
            "StartBit": 34,
            "TransmissionRate": "100ms when engage, otherwise on request",
            "Units": "bit"
        },
        1893: {
            "Acronym": "HPTOC",
            "DataRange": "0 to 3",
            "EndBit": 55,
            "Name": "Front Power Take Off Engagement Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1893,
            "SPNLength": 2,
            "StartBit": 54,
            "TransmissionRate": "Minimum rate of 1 second between messages or when a parameter is required to change state.",
            "Units": "bit"
        },
        1894: {
            "Acronym": "HPTOC",
            "DataRange": "0 to 3",
            "EndBit": 53,
            "Name": "Rear Power Take Off Engagement Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1894,
            "SPNLength": 2,
            "StartBit": 52,
            "TransmissionRate": "Minimum rate of 1 second between messages or when a parameter is required to change state.",
            "Units": "bit"
        },
        1895: {
            "Acronym": "HPTOC",
            "DataRange": "0 to 3",
            "EndBit": 63,
            "Name": "Front Power Take Off Mode Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1895,
            "SPNLength": 2,
            "StartBit": 62,
            "TransmissionRate": "Minimum rate of 1 second between messages or when a parameter is required to change state.",
            "Units": "bit"
        },
        1896: {
            "Acronym": "HPTOC",
            "DataRange": "0 to 3",
            "EndBit": 61,
            "Name": "Rear Power Take Off Mode Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1896,
            "SPNLength": 2,
            "StartBit": 60,
            "TransmissionRate": "Minimum rate of 1 second between messages or when a parameter is required to change state.",
            "Units": "bit"
        },
        1897: {
            "Acronym": "HPTOC",
            "DataRange": "0 to 3",
            "EndBit": 59,
            "Name": "Front Power Take Off Economy Mode Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1897,
            "SPNLength": 2,
            "StartBit": 58,
            "TransmissionRate": "Minimum rate of 1 second between messages or when a parameter is required to change state.",
            "Units": "bit"
        },
        1898: {
            "Acronym": "HPTOC",
            "DataRange": "0 to 3",
            "EndBit": 57,
            "Name": "Rear Power Take Off Economy Mode Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1898,
            "SPNLength": 2,
            "StartBit": 56,
            "TransmissionRate": "Minimum rate of 1 second between messages or when a parameter is required to change state.",
            "Units": "bit"
        },
        1899: {
            "Acronym": "AV00MF",
            "DataRange": "-125 to 125 %",
            "EndBit": 7,
            "Name": "Aux Valve 0 Extend Port Measured Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1899,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        190: {
            "Acronym": "EEC1",
            "DataRange": "0 to 8,031.875 rpm",
            "EndBit": 39,
            "Name": "Engine Speed",
            "Offset": 0.0,
            "OperationalHigh": 8031.875,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.125,
            "SPN": 190,
            "SPNLength": 16,
            "StartBit": 24,
            "TransmissionRate": "engine speed dependent",
            "Units": "rpm"
        },
        1900: {
            "Acronym": "AV00MF",
            "DataRange": "-125 to 125 %",
            "EndBit": 15,
            "Name": "Aux Valve 0 Retract Port Measured Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1900,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        1901: {
            "Acronym": "AV00EF",
            "DataRange": "-125 to 125 %",
            "EndBit": 7,
            "Name": "Aux Valve 0 Extend Port Estimated Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1901,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        1902: {
            "Acronym": "AV00EF",
            "DataRange": "-125 to 125 %",
            "EndBit": 15,
            "Name": "Aux Valve 0 Retract Port Estimated Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1902,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        1903: {
            "Acronym": "AV00EF",
            "DataRange": "0 to 15",
            "EndBit": 19,
            "Name": "Aux Valve 0 State",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1903,
            "SPNLength": 4,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1904: {
            "Acronym": "AV00MF",
            "DataRange": "0.0 to 46596.99927595 psi",
            "EndBit": 31,
            "Name": "Aux Valve 0 Extend Port Pressure",
            "Offset": 0.0,
            "OperationalHigh": 46596.99927595,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.72518869,
            "SPN": 1904,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        1905: {
            "Acronym": "AV00MF",
            "DataRange": "0.0 to 46596.99927595 psi",
            "EndBit": 47,
            "Name": "Aux Valve 0 Retract Port Pressure",
            "Offset": 0.0,
            "OperationalHigh": 46596.99927595,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.72518869,
            "SPN": 1905,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        1906: {
            "Acronym": "AV00MF",
            "DataRange": "0.0 to 580.150952 psi",
            "EndBit": 55,
            "Name": "Aux Valve 0 Return Port Pressure",
            "Offset": 0.0,
            "OperationalHigh": 580.150952,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 2.320603808,
            "SPN": 1906,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        1907: {
            "Acronym": "AV00C",
            "DataRange": "0 to 100 %",
            "EndBit": 7,
            "Name": "Aux Valve 0 Port Flow Command",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 1907,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.\n\n??? What's the fastest Rate > or < 100 ms ??",
            "Units": "%"
        },
        1908: {
            "Acronym": "AV00C",
            "DataRange": "0 to 15",
            "EndBit": 19,
            "Name": "Aux Valve 0 State command",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1908,
            "SPNLength": 4,
            "StartBit": 16,
            "TransmissionRate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.\n\n??? What's the fastest Rate > or < 100 ms ??",
            "Units": "bit"
        },
        1909: {
            "Acronym": "AV00C",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Aux Valve 0 Fail Safe Mode Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1909,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.\n\n??? What's the fastest Rate > or < 100 ms ??",
            "Units": "bit"
        },
        191: {
            "Acronym": "ETC1",
            "DataRange": "0 to 8,031.875 rpm",
            "EndBit": 23,
            "Name": "Transmission Output Shaft Speed",
            "Offset": 0.0,
            "OperationalHigh": 8031.875,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.125,
            "SPN": 191,
            "SPNLength": 16,
            "StartBit": 8,
            "TransmissionRate": "10 ms",
            "Units": "rpm"
        },
        1910: {
            "Acronym": "AV00EF",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Aux Valve 0 Fail Safe Mode",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1910,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1911: {
            "Acronym": "AV01MF",
            "DataRange": "-125 to 125 %",
            "EndBit": 7,
            "Name": "Aux Valve 1 Extend Port Measured Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1911,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        1912: {
            "Acronym": "AV01MF",
            "DataRange": "-125 to 125 %",
            "EndBit": 15,
            "Name": "Aux Valve 1 Retract Port Measured Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1912,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        1913: {
            "Acronym": "AV01EF",
            "DataRange": "-125 to 125 %",
            "EndBit": 7,
            "Name": "Aux Valve 1 Extend Port Estimated Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1913,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        1914: {
            "Acronym": "AV01EF",
            "DataRange": "-125 to 125 %",
            "EndBit": 15,
            "Name": "Aux Valve 1 Retract Port Estimated Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1914,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        1915: {
            "Acronym": "AV01EF",
            "DataRange": "0 to 15",
            "EndBit": 19,
            "Name": "Aux Valve 1 State",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1915,
            "SPNLength": 4,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1916: {
            "Acronym": "AV01MF",
            "DataRange": "0.0 to 46596.99927595 psi",
            "EndBit": 31,
            "Name": "Aux Valve 1 Extend Port Pressure",
            "Offset": 0.0,
            "OperationalHigh": 46596.99927595,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.72518869,
            "SPN": 1916,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        1917: {
            "Acronym": "AV01MF",
            "DataRange": "0.0 to 46596.99927595 psi",
            "EndBit": 47,
            "Name": "Aux Valve 1 Retract Port Pressure",
            "Offset": 0.0,
            "OperationalHigh": 46596.99927595,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.72518869,
            "SPN": 1917,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        1918: {
            "Acronym": "AV01MF",
            "DataRange": "0.0 to 580.150952 psi",
            "EndBit": 55,
            "Name": "Aux Valve 1 Return Port Pressure",
            "Offset": 0.0,
            "OperationalHigh": 580.150952,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 2.320603808,
            "SPN": 1918,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        1919: {
            "Acronym": "AV01C",
            "DataRange": "0 to 100 %",
            "EndBit": 7,
            "Name": "Aux Valve 1 Port Flow Command",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 1919,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.\n\n??? What's the fastest Rate > or < 100 ms ??",
            "Units": "%"
        },
        1920: {
            "Acronym": "AV01C",
            "DataRange": "0 to 15",
            "EndBit": 19,
            "Name": "Aux Valve 1 State Command",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1920,
            "SPNLength": 4,
            "StartBit": 16,
            "TransmissionRate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.\n\n??? What's the fastest Rate > or < 100 ms ??",
            "Units": "bit"
        },
        1921: {
            "Acronym": "AV01C",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Aux Valve 1 Fail Safe Mode Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1921,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.\n\n??? What's the fastest Rate > or < 100 ms ??",
            "Units": "bit"
        },
        1922: {
            "Acronym": "AV01EF",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Aux Valve 1 Fail Safe Mode",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1922,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1923: {
            "Acronym": "AV02MF",
            "DataRange": "-125 to 125 %",
            "EndBit": 7,
            "Name": "Aux Valve 2 Extend Port Measured Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1923,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        1924: {
            "Acronym": "AV02MF",
            "DataRange": "-125 to 125 %",
            "EndBit": 15,
            "Name": "Aux Valve 2 Retract Port Measured Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1924,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        1925: {
            "Acronym": "AV02EF",
            "DataRange": "-125 to 125 %",
            "EndBit": 7,
            "Name": "Aux Valve 2 Extend Port Estimated Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1925,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        1926: {
            "Acronym": "AV02EF",
            "DataRange": "-125 to 125 %",
            "EndBit": 15,
            "Name": "Aux Valve 2 Retract Port Estimated Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1926,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        1927: {
            "Acronym": "AV02EF",
            "DataRange": "0 to 15",
            "EndBit": 19,
            "Name": "Aux Valve 2 State",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1927,
            "SPNLength": 4,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1928: {
            "Acronym": "AV02MF",
            "DataRange": "0.0 to 46596.99927595 psi",
            "EndBit": 31,
            "Name": "Aux Valve 2 Extend Port Pressure",
            "Offset": 0.0,
            "OperationalHigh": 46596.99927595,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.72518869,
            "SPN": 1928,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        1929: {
            "Acronym": "AV02MF",
            "DataRange": "0.0 to 46596.99927595 psi",
            "EndBit": 47,
            "Name": "Aux Valve 2 Retract Port Pressure",
            "Offset": 0.0,
            "OperationalHigh": 46596.99927595,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.72518869,
            "SPN": 1929,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        1930: {
            "Acronym": "AV02MF",
            "DataRange": "0.0 to 580.150952 psi",
            "EndBit": 55,
            "Name": "Aux Valve 2 Return Port Pressure",
            "Offset": 0.0,
            "OperationalHigh": 580.150952,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 2.320603808,
            "SPN": 1930,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        1931: {
            "Acronym": "AV02C",
            "DataRange": "0 to 100 %",
            "EndBit": 7,
            "Name": "Aux Valve 2 Port Flow Command",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 1931,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.\n\n??? What's the fastest Rate > or < 100 ms ??",
            "Units": "%"
        },
        1932: {
            "Acronym": "AV02C",
            "DataRange": "0 to 15",
            "EndBit": 19,
            "Name": "Aux Valve 2 State Command",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1932,
            "SPNLength": 4,
            "StartBit": 16,
            "TransmissionRate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.\n\n??? What's the fastest Rate > or < 100 ms ??",
            "Units": "bit"
        },
        1933: {
            "Acronym": "AV02C",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Aux Valve 2 Fail Safe Mode Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1933,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.\n\n??? What's the fastest Rate > or < 100 ms ??",
            "Units": "bit"
        },
        1934: {
            "Acronym": "AV02EF",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Aux Valve 2 Fail Safe Mode",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1934,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1935: {
            "Acronym": "AV03MF",
            "DataRange": "-125 to 125 %",
            "EndBit": 7,
            "Name": "Aux Valve 3 Extend Port Measured Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1935,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        1936: {
            "Acronym": "AV03MF",
            "DataRange": "-125 to 125 %",
            "EndBit": 15,
            "Name": "Aux Valve 3 Retract Port Measured Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1936,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        1937: {
            "Acronym": "AV03EF",
            "DataRange": "-125 to 125 %",
            "EndBit": 7,
            "Name": "Aux Valve 3 Extend Port Estimated Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1937,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        1938: {
            "Acronym": "AV03EF",
            "DataRange": "-125 to 125 %",
            "EndBit": 15,
            "Name": "Aux Valve 3 Retract Port Estimated Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1938,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        1939: {
            "Acronym": "AV03EF",
            "DataRange": "0 to 15",
            "EndBit": 19,
            "Name": "Aux Valve 3 State",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1939,
            "SPNLength": 4,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1940: {
            "Acronym": "AV03MF",
            "DataRange": "0.0 to 46596.99927595 psi",
            "EndBit": 31,
            "Name": "Aux Valve 3 Extend Port Pressure",
            "Offset": 0.0,
            "OperationalHigh": 46596.99927595,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.72518869,
            "SPN": 1940,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        1941: {
            "Acronym": "AV03MF",
            "DataRange": "0.0 to 46596.99927595 psi",
            "EndBit": 47,
            "Name": "Aux Valve 3 Retract Port Pressure",
            "Offset": 0.0,
            "OperationalHigh": 46596.99927595,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.72518869,
            "SPN": 1941,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        1942: {
            "Acronym": "AV03MF",
            "DataRange": "0.0 to 580.150952 psi",
            "EndBit": 55,
            "Name": "Aux Valve 3 Return Port Pressure",
            "Offset": 0.0,
            "OperationalHigh": 580.150952,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 2.320603808,
            "SPN": 1942,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        1943: {
            "Acronym": "AV03C",
            "DataRange": "0 to 100 %",
            "EndBit": 7,
            "Name": "Aux Valve 3 Port Flow Command",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 1943,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.\n\n??? What's the fastest Rate > or < 100 ms ??",
            "Units": "%"
        },
        1944: {
            "Acronym": "AV03C",
            "DataRange": "0 to 15",
            "EndBit": 19,
            "Name": "Aux Valve 3 State Command",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1944,
            "SPNLength": 4,
            "StartBit": 16,
            "TransmissionRate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.\n\n??? What's the fastest Rate > or < 100 ms ??",
            "Units": "bit"
        },
        1945: {
            "Acronym": "AV03C",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Aux Valve 3 Fail Safe Mode Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1945,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.\n\n??? What's the fastest Rate > or < 100 ms ??",
            "Units": "bit"
        },
        1946: {
            "Acronym": "AV03EF",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Aux Valve 3 Fail Safe Mode",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1946,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1947: {
            "Acronym": "AV04MF",
            "DataRange": "-125 to 125 %",
            "EndBit": 7,
            "Name": "Aux Valve 4 Extend Port Measured Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1947,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        1948: {
            "Acronym": "AV04MF",
            "DataRange": "-125 to 125 %",
            "EndBit": 15,
            "Name": "Aux Valve 4 Retract Port Measured Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1948,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        1949: {
            "Acronym": "AV04EF",
            "DataRange": "-125 to 125 %",
            "EndBit": 7,
            "Name": "Aux Valve 4 Extend Port Estimated Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1949,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        1950: {
            "Acronym": "AV04EF",
            "DataRange": "-125 to 125 %",
            "EndBit": 15,
            "Name": "Aux Valve 4 Retract Port Estimated Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1950,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        1951: {
            "Acronym": "AV04EF",
            "DataRange": "0 to 15",
            "EndBit": 19,
            "Name": "Aux Valve 4 State",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1951,
            "SPNLength": 4,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1952: {
            "Acronym": "AV04MF",
            "DataRange": "0.0 to 46596.99927595 psi",
            "EndBit": 31,
            "Name": "Aux Valve 4 Extend Port Pressure",
            "Offset": 0.0,
            "OperationalHigh": 46596.99927595,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.72518869,
            "SPN": 1952,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        1953: {
            "Acronym": "AV04MF",
            "DataRange": "0.0 to 46596.99927595 psi",
            "EndBit": 47,
            "Name": "Aux Valve 4 Retract Port Pressure",
            "Offset": 0.0,
            "OperationalHigh": 46596.99927595,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.72518869,
            "SPN": 1953,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        1954: {
            "Acronym": "AV04MF",
            "DataRange": "0.0 to 580.150952 psi",
            "EndBit": 55,
            "Name": "Aux Valve 4 Return Port Pressure",
            "Offset": 0.0,
            "OperationalHigh": 580.150952,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 2.320603808,
            "SPN": 1954,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        1955: {
            "Acronym": "AV04C",
            "DataRange": "0 to 100 %",
            "EndBit": 7,
            "Name": "Aux Valve 4 Port Flow Command",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 1955,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.\n\n??? What's the fastest Rate > or < 100 ms ??",
            "Units": "%"
        },
        1956: {
            "Acronym": "AV04C",
            "DataRange": "0 to 15",
            "EndBit": 19,
            "Name": "Aux Valve 4 State Command",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1956,
            "SPNLength": 4,
            "StartBit": 16,
            "TransmissionRate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.\n\n??? What's the fastest Rate > or < 100 ms ??",
            "Units": "bit"
        },
        1957: {
            "Acronym": "AV04C",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Aux Valve 4 Fail Safe Mode Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1957,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.\n\n??? What's the fastest Rate > or < 100 ms ??",
            "Units": "bit"
        },
        1958: {
            "Acronym": "AV04EF",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Aux Valve 4 Fail Safe Mode",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1958,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1959: {
            "Acronym": "AV05MF",
            "DataRange": "-125 to 125 %",
            "EndBit": 7,
            "Name": "Aux Valve 5 Extend Port Measured Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1959,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        1960: {
            "Acronym": "AV05MF",
            "DataRange": "-125 to 125 %",
            "EndBit": 15,
            "Name": "Aux Valve 5 Retract Port Measured Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1960,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        1961: {
            "Acronym": "AV05EF",
            "DataRange": "-125 to 125 %",
            "EndBit": 7,
            "Name": "Aux Valve 5 Extend Port Estimated Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1961,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        1962: {
            "Acronym": "AV05EF",
            "DataRange": "-125 to 125 %",
            "EndBit": 15,
            "Name": "Aux Valve 5 Retract Port Estimated Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1962,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        1963: {
            "Acronym": "AV05EF",
            "DataRange": "0 to 15",
            "EndBit": 19,
            "Name": "Aux Valve 5 State",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1963,
            "SPNLength": 4,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1964: {
            "Acronym": "AV05MF",
            "DataRange": "0.0 to 46596.99927595 psi",
            "EndBit": 31,
            "Name": "Aux Valve 5 Extend Port Pressure",
            "Offset": 0.0,
            "OperationalHigh": 46596.99927595,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.72518869,
            "SPN": 1964,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        1965: {
            "Acronym": "AV05MF",
            "DataRange": "0.0 to 46596.99927595 psi",
            "EndBit": 47,
            "Name": "Aux Valve 5 Retract Port Pressure",
            "Offset": 0.0,
            "OperationalHigh": 46596.99927595,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.72518869,
            "SPN": 1965,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        1966: {
            "Acronym": "AV05MF",
            "DataRange": "0.0 to 580.150952 psi",
            "EndBit": 55,
            "Name": "Aux Valve 5 Return Port Pressure",
            "Offset": 0.0,
            "OperationalHigh": 580.150952,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 2.320603808,
            "SPN": 1966,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        1967: {
            "Acronym": "AV05C",
            "DataRange": "0 to 100 %",
            "EndBit": 7,
            "Name": "Aux Valve 5 Port Flow Command",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 1967,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.\n\n??? What's the fastest Rate > or < 100 ms ??",
            "Units": "%"
        },
        1968: {
            "Acronym": "AV05C",
            "DataRange": "0 to 15",
            "EndBit": 19,
            "Name": "Aux Valve 5 State Command",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1968,
            "SPNLength": 4,
            "StartBit": 16,
            "TransmissionRate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.\n\n??? What's the fastest Rate > or < 100 ms ??",
            "Units": "bit"
        },
        1969: {
            "Acronym": "AV05C",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Aux Valve 5 Fail Safe Mode Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1969,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.\n\n??? What's the fastest Rate > or < 100 ms ??",
            "Units": "bit"
        },
        1970: {
            "Acronym": "AV05EF",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Aux Valve 5 Fail Safe Mode",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1970,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1971: {
            "Acronym": "AV06MF",
            "DataRange": "-125 to 125 %",
            "EndBit": 7,
            "Name": "Aux Valve 6 Extend Port Measured Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1971,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        1972: {
            "Acronym": "AV06MF",
            "DataRange": "-125 to 125 %",
            "EndBit": 15,
            "Name": "Aux Valve 6 Retract Port Measured Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1972,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        1973: {
            "Acronym": "AV06EF",
            "DataRange": "-125 to 125 %",
            "EndBit": 7,
            "Name": "Aux Valve 6 Extend Port Estimated Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1973,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        1974: {
            "Acronym": "AV06EF",
            "DataRange": "-125 to 125 %",
            "EndBit": 15,
            "Name": "Aux Valve 6 Retract Port Estimated Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1974,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        1975: {
            "Acronym": "AV06EF",
            "DataRange": "0 to 15",
            "EndBit": 19,
            "Name": "Aux Valve 6 State",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1975,
            "SPNLength": 4,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1976: {
            "Acronym": "AV06MF",
            "DataRange": "0.0 to 46596.99927595 psi",
            "EndBit": 31,
            "Name": "Aux Valve 6 Extend Port Pressure",
            "Offset": 0.0,
            "OperationalHigh": 46596.99927595,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.72518869,
            "SPN": 1976,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        1977: {
            "Acronym": "AV06MF",
            "DataRange": "0.0 to 46596.99927595 psi",
            "EndBit": 47,
            "Name": "Aux Valve 6 Retract Port Pressure",
            "Offset": 0.0,
            "OperationalHigh": 46596.99927595,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.72518869,
            "SPN": 1977,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        1978: {
            "Acronym": "AV06MF",
            "DataRange": "0.0 to 580.150952 psi",
            "EndBit": 55,
            "Name": "Aux Valve 6 Return Port Pressure",
            "Offset": 0.0,
            "OperationalHigh": 580.150952,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 2.320603808,
            "SPN": 1978,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        1979: {
            "Acronym": "AV06C",
            "DataRange": "0 to 100 %",
            "EndBit": 7,
            "Name": "Aux Valve 6 Port Flow Command",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 1979,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.\n\n??? What's the fastest Rate > or < 100 ms ??",
            "Units": "%"
        },
        1980: {
            "Acronym": "AV06C",
            "DataRange": "0 to 15",
            "EndBit": 19,
            "Name": "Aux Valve 6 State Command",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1980,
            "SPNLength": 4,
            "StartBit": 16,
            "TransmissionRate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.\n\n??? What's the fastest Rate > or < 100 ms ??",
            "Units": "bit"
        },
        1981: {
            "Acronym": "AV06C",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Aux Valve 6 Fail Safe Mode Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1981,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.\n\n??? What's the fastest Rate > or < 100 ms ??",
            "Units": "bit"
        },
        1982: {
            "Acronym": "AV06EF",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Aux Valve 6 Fail Safe Mode",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1982,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1983: {
            "Acronym": "AV07MF",
            "DataRange": "-125 to 125 %",
            "EndBit": 7,
            "Name": "Aux Valve 7 Extend Port Measured Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1983,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        1984: {
            "Acronym": "AV07MF",
            "DataRange": "-125 to 125 %",
            "EndBit": 15,
            "Name": "Aux Valve 7 Retract Port Measured Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1984,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        1985: {
            "Acronym": "AV07EF",
            "DataRange": "-125 to 125 %",
            "EndBit": 7,
            "Name": "Aux Valve 7 Extend Port Estimated Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1985,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        1986: {
            "Acronym": "AV07EF",
            "DataRange": "-125 to 125 %",
            "EndBit": 15,
            "Name": "Aux Valve 7 Retract Port Estimated Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1986,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        1987: {
            "Acronym": "AV07EF",
            "DataRange": "0 to 15",
            "EndBit": 19,
            "Name": "Aux Valve 7 State",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1987,
            "SPNLength": 4,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1988: {
            "Acronym": "AV07MF",
            "DataRange": "0.0 to 46596.99927595 psi",
            "EndBit": 31,
            "Name": "Aux Valve 7 Extend Port Pressure",
            "Offset": 0.0,
            "OperationalHigh": 46596.99927595,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.72518869,
            "SPN": 1988,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        1989: {
            "Acronym": "AV07MF",
            "DataRange": "0.0 to 46596.99927595 psi",
            "EndBit": 47,
            "Name": "Aux Valve 7 Retract Port Pressure",
            "Offset": 0.0,
            "OperationalHigh": 46596.99927595,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.72518869,
            "SPN": 1989,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        1990: {
            "Acronym": "AV07MF",
            "DataRange": "0.0 to 580.150952 psi",
            "EndBit": 55,
            "Name": "Aux Valve 7 Return Port Pressure",
            "Offset": 0.0,
            "OperationalHigh": 580.150952,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 2.320603808,
            "SPN": 1990,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        1991: {
            "Acronym": "AV07C",
            "DataRange": "0 to 100 %",
            "EndBit": 7,
            "Name": "Aux Valve 7 Port Flow Command",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 1991,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.\n\n??? What's the fastest Rate > or < 100 ms ??",
            "Units": "%"
        },
        1992: {
            "Acronym": "AV07C",
            "DataRange": "0 to 15",
            "EndBit": 19,
            "Name": "Aux Valve 7 State command",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1992,
            "SPNLength": 4,
            "StartBit": 16,
            "TransmissionRate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.\n\n??? What's the fastest Rate > or < 100 ms ??",
            "Units": "bit"
        },
        1993: {
            "Acronym": "AV07C",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Aux Valve 7 Fail Safe Mode Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1993,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.\n\n??? What's the fastest Rate > or < 100 ms ??",
            "Units": "bit"
        },
        1994: {
            "Acronym": "AV07EF",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Aux Valve 7 Fail Safe Mode",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1994,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        1995: {
            "Acronym": "AV08MF",
            "DataRange": "-125 to 125 %",
            "EndBit": 7,
            "Name": "Aux Valve 8 Extend Port Measured Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1995,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        1996: {
            "Acronym": "AV08MF",
            "DataRange": "-125 to 125 %",
            "EndBit": 15,
            "Name": "Aux Valve 8 Retract Port Measured Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1996,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        1997: {
            "Acronym": "AV08EF",
            "DataRange": "-125 to 125 %",
            "EndBit": 7,
            "Name": "Aux Valve 8 Extend Port Estimated Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1997,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        1998: {
            "Acronym": "AV08EF",
            "DataRange": "-125 to 125 %",
            "EndBit": 15,
            "Name": "Aux Valve 8 Retract Port Estimated Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1998,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        1999: {
            "Acronym": "AV08EF",
            "DataRange": "0 to 15",
            "EndBit": 19,
            "Name": "Aux Valve 8 State",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 1999,
            "SPNLength": 4,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        20: {
            "Acronym": "EAC",
            "DataRange": "0.0 to 145.037738 psi",
            "EndBit": 31,
            "Name": "Engine Extended Range Coolant Pressure",
            "Offset": 0.0,
            "OperationalHigh": 145.037738,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.580150952,
            "SPN": 20,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "0.5 s",
            "Units": "psi"
        },
        22: {
            "Acronym": "EFL/P1",
            "DataRange": "0.0 to 1.812971725 psi",
            "EndBit": 15,
            "Name": "Engine Extended Crankcase Blow-by Pressure",
            "Offset": 0.0,
            "OperationalHigh": 1.812971725,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0072518869,
            "SPN": 22,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "0.5 s",
            "Units": "psi"
        },
        2256: {
            "Acronym": "AV08MF",
            "DataRange": "0.0 to 46596.99927595 psi",
            "EndBit": 31,
            "Name": "Aux Valve 8 Extend Port Pressure",
            "Offset": 0.0,
            "OperationalHigh": 46596.99927595,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.72518869,
            "SPN": 2256,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        2257: {
            "Acronym": "AV08MF",
            "DataRange": "0.0 to 46596.99927595 psi",
            "EndBit": 47,
            "Name": "Aux Valve 8 Retract Port Pressure",
            "Offset": 0.0,
            "OperationalHigh": 46596.99927595,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.72518869,
            "SPN": 2257,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        2258: {
            "Acronym": "AV08MF",
            "DataRange": "0.0 to 580.150952 psi",
            "EndBit": 55,
            "Name": "Aux Valve 8 Return Port Pressure",
            "Offset": 0.0,
            "OperationalHigh": 580.150952,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 2.320603808,
            "SPN": 2258,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        2259: {
            "Acronym": "AV08C",
            "DataRange": "0 to 100 %",
            "EndBit": 7,
            "Name": "Aux Valve 8 Port Flow Command",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 2259,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.\n\n??? What's the fastest Rate > or < 100 ms ??",
            "Units": "%"
        },
        2260: {
            "Acronym": "AV08C",
            "DataRange": "0 to 15",
            "EndBit": 19,
            "Name": "Aux Valve 8 State Command",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2260,
            "SPNLength": 4,
            "StartBit": 16,
            "TransmissionRate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.\n\n??? What's the fastest Rate > or < 100 ms ??",
            "Units": "bit"
        },
        2261: {
            "Acronym": "AV08C",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Aux Valve 8 Fail Safe Mode Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2261,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.\n\n??? What's the fastest Rate > or < 100 ms ??",
            "Units": "bit"
        },
        2262: {
            "Acronym": "AV08EF",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Aux Valve 8 Fail Safe Mode",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2262,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2263: {
            "Acronym": "AV09MF",
            "DataRange": "-125 to 125 %",
            "EndBit": 7,
            "Name": "Aux Valve 9 Extend Port Measured Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2263,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        2264: {
            "Acronym": "AV09MF",
            "DataRange": "-125 to 125 %",
            "EndBit": 15,
            "Name": "Aux Valve 9 Retract Port Measured Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2264,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        2265: {
            "Acronym": "AV09EF",
            "DataRange": "-125 to 125 %",
            "EndBit": 7,
            "Name": "Aux Valve 9 Extend Port Estimated Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2265,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        2266: {
            "Acronym": "AV09EF",
            "DataRange": "-125 to 125 %",
            "EndBit": 15,
            "Name": "Aux Valve 9 Retract Port Estimated Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2266,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        2267: {
            "Acronym": "AV09EF",
            "DataRange": "0 to 15",
            "EndBit": 19,
            "Name": "Aux Valve 9 State",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2267,
            "SPNLength": 4,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2268: {
            "Acronym": "AV09MF",
            "DataRange": "0.0 to 46596.99927595 psi",
            "EndBit": 31,
            "Name": "Aux Valve 9 Extend Port Pressure",
            "Offset": 0.0,
            "OperationalHigh": 46596.99927595,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.72518869,
            "SPN": 2268,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        2269: {
            "Acronym": "AV09MF",
            "DataRange": "0.0 to 46596.99927595 psi",
            "EndBit": 47,
            "Name": "Aux Valve 9 Retract Port Pressure",
            "Offset": 0.0,
            "OperationalHigh": 46596.99927595,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.72518869,
            "SPN": 2269,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        2270: {
            "Acronym": "AV09MF",
            "DataRange": "0.0 to 580.150952 psi",
            "EndBit": 55,
            "Name": "Aux Valve 9 Return Port Pressure",
            "Offset": 0.0,
            "OperationalHigh": 580.150952,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 2.320603808,
            "SPN": 2270,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        2271: {
            "Acronym": "AV09C",
            "DataRange": "0 to 100 %",
            "EndBit": 7,
            "Name": "Aux Valve 9 Port Flow Command",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 2271,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.\n\n??? What's the fastest Rate > or < 100 ms ??",
            "Units": "%"
        },
        2272: {
            "Acronym": "AV09C",
            "DataRange": "0 to 15",
            "EndBit": 19,
            "Name": "Aux Valve 9 State Command",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2272,
            "SPNLength": 4,
            "StartBit": 16,
            "TransmissionRate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.\n\n??? What's the fastest Rate > or < 100 ms ??",
            "Units": "bit"
        },
        2273: {
            "Acronym": "AV09C",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Aux Valve 9 Fail Safe Mode Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2273,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.\n\n??? What's the fastest Rate > or < 100 ms ??",
            "Units": "bit"
        },
        2274: {
            "Acronym": "AV09EF",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Aux Valve 9 Fail Safe Mode",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2274,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2275: {
            "Acronym": "AV10MF",
            "DataRange": "-125 to 125 %",
            "EndBit": 7,
            "Name": "Aux Valve 10 Extend Port Measured Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2275,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        2276: {
            "Acronym": "AV10MF",
            "DataRange": "-125 to 125 %",
            "EndBit": 15,
            "Name": "Aux Valve 10 Retract Port Measured Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2276,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        2277: {
            "Acronym": "AV10EF",
            "DataRange": "-125 to 125 %",
            "EndBit": 7,
            "Name": "Aux Valve 10 Extend Port Estimated Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2277,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        2278: {
            "Acronym": "AV10EF",
            "DataRange": "-125 to 125 %",
            "EndBit": 15,
            "Name": "Aux Valve 10 Retract Port Estimated Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2278,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        2279: {
            "Acronym": "AV10EF",
            "DataRange": "0 to 15",
            "EndBit": 19,
            "Name": "Aux Valve 10 State",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2279,
            "SPNLength": 4,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2280: {
            "Acronym": "AV10MF",
            "DataRange": "0.0 to 46596.99927595 psi",
            "EndBit": 31,
            "Name": "Aux Valve 10 Extend Port Pressure",
            "Offset": 0.0,
            "OperationalHigh": 46596.99927595,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.72518869,
            "SPN": 2280,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        2281: {
            "Acronym": "AV10MF",
            "DataRange": "0.0 to 46596.99927595 psi",
            "EndBit": 47,
            "Name": "Aux Valve 10 Retract Port Pressure",
            "Offset": 0.0,
            "OperationalHigh": 46596.99927595,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.72518869,
            "SPN": 2281,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        2282: {
            "Acronym": "AV10MF",
            "DataRange": "0.0 to 580.150952 psi",
            "EndBit": 55,
            "Name": "Aux Valve 10 Return Port Pressure",
            "Offset": 0.0,
            "OperationalHigh": 580.150952,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 2.320603808,
            "SPN": 2282,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        2283: {
            "Acronym": "AV10C",
            "DataRange": "0 to 100 %",
            "EndBit": 7,
            "Name": "Aux Valve 10 Port Flow Command",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 2283,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.\n\n??? What's the fastest Rate > or < 100 ms ??",
            "Units": "%"
        },
        2284: {
            "Acronym": "AV10C",
            "DataRange": "0 to 15",
            "EndBit": 19,
            "Name": "Aux Valve 10 State Command",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2284,
            "SPNLength": 4,
            "StartBit": 16,
            "TransmissionRate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.\n\n??? What's the fastest Rate > or < 100 ms ??",
            "Units": "bit"
        },
        2285: {
            "Acronym": "AV10C",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Aux Valve 10 Fail Safe Mode Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2285,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.\n\n??? What's the fastest Rate > or < 100 ms ??",
            "Units": "bit"
        },
        2286: {
            "Acronym": "AV10EF",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Aux Valve 10 Fail Safe Mode",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2286,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2287: {
            "Acronym": "AV11MF",
            "DataRange": "-125 to 125 %",
            "EndBit": 7,
            "Name": "Aux Valve 11 Extend Port Measured Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2287,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        2288: {
            "Acronym": "AV11MF",
            "DataRange": "-125 to 125 %",
            "EndBit": 15,
            "Name": "Aux Valve 11 Retract Port Measured Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2288,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        2289: {
            "Acronym": "AV11EF",
            "DataRange": "-125 to 125 %",
            "EndBit": 7,
            "Name": "Aux Valve 11 Extend Port Estimated Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2289,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        2290: {
            "Acronym": "AV11EF",
            "DataRange": "-125 to 125 %",
            "EndBit": 15,
            "Name": "Aux Valve 11 Retract Port Estimated Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2290,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        2291: {
            "Acronym": "AV11EF",
            "DataRange": "0 to 15",
            "EndBit": 19,
            "Name": "Aux Valve 11 State",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2291,
            "SPNLength": 4,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2292: {
            "Acronym": "AV11MF",
            "DataRange": "0.0 to 46596.99927595 psi",
            "EndBit": 31,
            "Name": "Aux Valve 11 Extend Port Pressure",
            "Offset": 0.0,
            "OperationalHigh": 46596.99927595,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.72518869,
            "SPN": 2292,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        2293: {
            "Acronym": "AV11MF",
            "DataRange": "0.0 to 46596.99927595 psi",
            "EndBit": 47,
            "Name": "Aux Valve 11 Retract Port Pressure",
            "Offset": 0.0,
            "OperationalHigh": 46596.99927595,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.72518869,
            "SPN": 2293,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        2294: {
            "Acronym": "AV11MF",
            "DataRange": "0.0 to 580.150952 psi",
            "EndBit": 55,
            "Name": "Aux Valve 11 Return Port Pressure",
            "Offset": 0.0,
            "OperationalHigh": 580.150952,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 2.320603808,
            "SPN": 2294,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        2295: {
            "Acronym": "AV11C",
            "DataRange": "0 to 100 %",
            "EndBit": 7,
            "Name": "Aux Valve 11 Port Flow Command",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 2295,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.\n\n??? What's the fastest Rate > or < 100 ms ??",
            "Units": "%"
        },
        2296: {
            "Acronym": "AV11C",
            "DataRange": "0 to 15",
            "EndBit": 19,
            "Name": "Aux Valve 11 State Command",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2296,
            "SPNLength": 4,
            "StartBit": 16,
            "TransmissionRate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.\n\n??? What's the fastest Rate > or < 100 ms ??",
            "Units": "bit"
        },
        2297: {
            "Acronym": "AV11C",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Aux Valve 11 Fail Safe Mode Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2297,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.\n\n??? What's the fastest Rate > or < 100 ms ??",
            "Units": "bit"
        },
        2298: {
            "Acronym": "AV11EF",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Aux Valve 11 Fail Safe Mode",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2298,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2299: {
            "Acronym": "AV12MF",
            "DataRange": "-125 to 125 %",
            "EndBit": 7,
            "Name": "Aux Valve 12 Extend Port Measured Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2299,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        2300: {
            "Acronym": "AV12MF",
            "DataRange": "-125 to 125 %",
            "EndBit": 15,
            "Name": "Aux Valve 12 Retract Port Measured Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2300,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        2301: {
            "Acronym": "AV12EF",
            "DataRange": "-125 to 125 %",
            "EndBit": 7,
            "Name": "Aux Valve 12 Extend Port Estimated Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2301,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        2302: {
            "Acronym": "AV12EF",
            "DataRange": "-125 to 125 %",
            "EndBit": 15,
            "Name": "Aux Valve 12 Retract Port Estimated Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2302,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        2303: {
            "Acronym": "AV12EF",
            "DataRange": "0 to 15",
            "EndBit": 19,
            "Name": "Aux Valve 12 State",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2303,
            "SPNLength": 4,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2304: {
            "Acronym": "AV12MF",
            "DataRange": "0.0 to 46596.99927595 psi",
            "EndBit": 31,
            "Name": "Aux Valve 12 Extend Port Pressure",
            "Offset": 0.0,
            "OperationalHigh": 46596.99927595,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.72518869,
            "SPN": 2304,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        2305: {
            "Acronym": "AV12MF",
            "DataRange": "0.0 to 46596.99927595 psi",
            "EndBit": 47,
            "Name": "Aux Valve 12 Retract Port Pressure",
            "Offset": 0.0,
            "OperationalHigh": 46596.99927595,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.72518869,
            "SPN": 2305,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        2306: {
            "Acronym": "AV12MF",
            "DataRange": "0.0 to 580.150952 psi",
            "EndBit": 55,
            "Name": "Aux Valve 12 Return Port Pressure",
            "Offset": 0.0,
            "OperationalHigh": 580.150952,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 2.320603808,
            "SPN": 2306,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        2307: {
            "Acronym": "AV12C",
            "DataRange": "0 to 100 %",
            "EndBit": 7,
            "Name": "Aux Valve 12 Port Flow Command",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 2307,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.\n\n??? What's the fastest Rate > or < 100 ms ??",
            "Units": "%"
        },
        2308: {
            "Acronym": "AV12C",
            "DataRange": "0 to 15",
            "EndBit": 19,
            "Name": "Aux Valve 12 State Command",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2308,
            "SPNLength": 4,
            "StartBit": 16,
            "TransmissionRate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.\n\n??? What's the fastest Rate > or < 100 ms ??",
            "Units": "bit"
        },
        2309: {
            "Acronym": "AV12C",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Aux Valve 12 Fail Safe Mode Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2309,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.\n\n??? What's the fastest Rate > or < 100 ms ??",
            "Units": "bit"
        },
        2310: {
            "Acronym": "AV12EF",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Aux Valve 12 Fail Safe Mode",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2310,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2311: {
            "Acronym": "AV13MF",
            "DataRange": "-125 to 125 %",
            "EndBit": 7,
            "Name": "Aux Valve 13 Extend Port Measured Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2311,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        2312: {
            "Acronym": "AV13MF",
            "DataRange": "-125 to 125 %",
            "EndBit": 15,
            "Name": "Aux Valve 13 Retract Port Measured Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2312,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        2313: {
            "Acronym": "AV13EF",
            "DataRange": "-125 to 125 %",
            "EndBit": 7,
            "Name": "Aux Valve 13 Extend Port Estimated Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2313,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        2314: {
            "Acronym": "AV13EF",
            "DataRange": "-125 to 125 %",
            "EndBit": 15,
            "Name": "Aux Valve 13 Retract Port Estimated Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2314,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        2315: {
            "Acronym": "AV13EF",
            "DataRange": "0 to 15",
            "EndBit": 19,
            "Name": "Aux Valve 13 State",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2315,
            "SPNLength": 4,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2316: {
            "Acronym": "AV13MF",
            "DataRange": "0.0 to 46596.99927595 psi",
            "EndBit": 31,
            "Name": "Aux Valve 13 Extend Port Pressure",
            "Offset": 0.0,
            "OperationalHigh": 46596.99927595,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.72518869,
            "SPN": 2316,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        2317: {
            "Acronym": "AV13MF",
            "DataRange": "0.0 to 46596.99927595 psi",
            "EndBit": 47,
            "Name": "Aux Valve 13 Retract Port Pressure",
            "Offset": 0.0,
            "OperationalHigh": 46596.99927595,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.72518869,
            "SPN": 2317,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        2318: {
            "Acronym": "AV13MF",
            "DataRange": "0.0 to 580.150952 psi",
            "EndBit": 55,
            "Name": "Aux Valve 13 Return Port Pressure",
            "Offset": 0.0,
            "OperationalHigh": 580.150952,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 2.320603808,
            "SPN": 2318,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        2319: {
            "Acronym": "AV13C",
            "DataRange": "0 to 100 %",
            "EndBit": 7,
            "Name": "Aux Valve 13 Port Flow Command",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 2319,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.\n\n??? What's the fastest Rate > or < 100 ms ??",
            "Units": "%"
        },
        2320: {
            "Acronym": "AV13C",
            "DataRange": "0 to 15",
            "EndBit": 19,
            "Name": "Aux Valve 13 State Command",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2320,
            "SPNLength": 4,
            "StartBit": 16,
            "TransmissionRate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.\n\n??? What's the fastest Rate > or < 100 ms ??",
            "Units": "bit"
        },
        2321: {
            "Acronym": "AV13C",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Aux Valve 13 Fail Safe Mode Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2321,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.\n\n??? What's the fastest Rate > or < 100 ms ??",
            "Units": "bit"
        },
        2322: {
            "Acronym": "AV13EF",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Aux Valve 13 Fail Safe Mode",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2322,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2323: {
            "Acronym": "AV14MF",
            "DataRange": "-125 to 125 %",
            "EndBit": 7,
            "Name": "Aux Valve 14 Extend Port Measured Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2323,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        2324: {
            "Acronym": "AV14MF",
            "DataRange": "-125 to 125 %",
            "EndBit": 15,
            "Name": "Aux Valve 14 Retract Port Measured Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2324,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        2325: {
            "Acronym": "AV14EF",
            "DataRange": "-125 to 125 %",
            "EndBit": 7,
            "Name": "Aux Valve 14 Extend Port Estimated Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2325,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        2326: {
            "Acronym": "AV14EF",
            "DataRange": "-125 to 125 %",
            "EndBit": 15,
            "Name": "Aux Valve 14 Retract Port Estimated Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2326,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        2327: {
            "Acronym": "AV14EF",
            "DataRange": "0 to 15",
            "EndBit": 19,
            "Name": "Aux Valve 14 State",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2327,
            "SPNLength": 4,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2328: {
            "Acronym": "AV14MF",
            "DataRange": "0.0 to 46596.99927595 psi",
            "EndBit": 31,
            "Name": "Aux Valve 14 Extend Port Pressure",
            "Offset": 0.0,
            "OperationalHigh": 46596.99927595,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.72518869,
            "SPN": 2328,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        2329: {
            "Acronym": "AV14MF",
            "DataRange": "0.0 to 46596.99927595 psi",
            "EndBit": 47,
            "Name": "Aux Valve 14 Retract Port Pressure",
            "Offset": 0.0,
            "OperationalHigh": 46596.99927595,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.72518869,
            "SPN": 2329,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        233: {
            "Acronym": "CI",
            "DataRange": "0 to 255 per byte",
            "EndBit": -4,
            "Name": "Unit Number (Power Unit)",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": -2,
            "SPN": 233,
            "SPNLength": 1600,
            "StartBit": -1,
            "TransmissionRate": "On request",
            "Units": "ASCII"
        },
        2330: {
            "Acronym": "AV14MF",
            "DataRange": "0.0 to 580.150952 psi",
            "EndBit": 55,
            "Name": "Aux Valve 14 Return Port Pressure",
            "Offset": 0.0,
            "OperationalHigh": 580.150952,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 2.320603808,
            "SPN": 2330,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        2331: {
            "Acronym": "AV14C",
            "DataRange": "0 to 100 %",
            "EndBit": 7,
            "Name": "Aux Valve 14 Port Flow Command",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 2331,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.\n\n??? What's the fastest Rate > or < 100 ms ??",
            "Units": "%"
        },
        2332: {
            "Acronym": "AV14C",
            "DataRange": "0 to 15",
            "EndBit": 19,
            "Name": "Aux Valve 14 State Command",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2332,
            "SPNLength": 4,
            "StartBit": 16,
            "TransmissionRate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.\n\n??? What's the fastest Rate > or < 100 ms ??",
            "Units": "bit"
        },
        2333: {
            "Acronym": "AV14C",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Aux Valve 14 Fail Safe Mode Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2333,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.\n\n??? What's the fastest Rate > or < 100 ms ??",
            "Units": "bit"
        },
        2334: {
            "Acronym": "AV14EF",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Aux Valve 14 Fail Safe Mode",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2334,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2335: {
            "Acronym": "AV15MF",
            "DataRange": "-125 to 125 %",
            "EndBit": 7,
            "Name": "Aux Valve 15 Extend Port Measured Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2335,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        2336: {
            "Acronym": "AV15MF",
            "DataRange": "-125 to 125 %",
            "EndBit": 15,
            "Name": "Aux Valve 15 Retract Port Measured Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2336,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        2337: {
            "Acronym": "AV15EF",
            "DataRange": "-125 to 125 %",
            "EndBit": 7,
            "Name": "Aux Valve 15 Extend Port Estimated Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2337,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        2338: {
            "Acronym": "AV15EF",
            "DataRange": "-125 to 125 %",
            "EndBit": 15,
            "Name": "Aux Valve 15 Retract Port Estimated Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2338,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        2339: {
            "Acronym": "AV15EF",
            "DataRange": "0 to 15",
            "EndBit": 19,
            "Name": "Aux Valve 15 State",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2339,
            "SPNLength": 4,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        234: {
            "Acronym": "SOFT",
            "DataRange": "0 to 255 per byte",
            "EndBit": -1,
            "Name": "Software Identification",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": -2,
            "SPN": 234,
            "SPNLength": 1600,
            "StartBit": 1,
            "TransmissionRate": "On request",
            "Units": "ASCII"
        },
        2340: {
            "Acronym": "AV15MF",
            "DataRange": "0.0 to 46596.99927595 psi",
            "EndBit": 31,
            "Name": "Aux Valve 15 Extend Port Pressure",
            "Offset": 0.0,
            "OperationalHigh": 46596.99927595,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.72518869,
            "SPN": 2340,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        2341: {
            "Acronym": "AV15MF",
            "DataRange": "0.0 to 46596.99927595 psi",
            "EndBit": 47,
            "Name": "Aux Valve 15 Retract Port Pressure",
            "Offset": 0.0,
            "OperationalHigh": 46596.99927595,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.72518869,
            "SPN": 2341,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        2342: {
            "Acronym": "AV15MF",
            "DataRange": "0.0 to 580.150952 psi",
            "EndBit": 55,
            "Name": "Aux Valve 15 Return Port Pressure",
            "Offset": 0.0,
            "OperationalHigh": 580.150952,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 2.320603808,
            "SPN": 2342,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        2343: {
            "Acronym": "AV15C",
            "DataRange": "0 to 100 %",
            "EndBit": 7,
            "Name": "Aux Valve 15 Port Flow Command",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 2343,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.\n\n??? What's the fastest Rate > or < 100 ms ??",
            "Units": "%"
        },
        2344: {
            "Acronym": "AV15C",
            "DataRange": "0 to 15",
            "EndBit": 19,
            "Name": "Aux Valve 15 State Command",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2344,
            "SPNLength": 4,
            "StartBit": 16,
            "TransmissionRate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.\n\n??? What's the fastest Rate > or < 100 ms ??",
            "Units": "bit"
        },
        2345: {
            "Acronym": "AV15C",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Aux Valve 15 Fail Safe Mode Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2345,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "Minimum rate of 1 second between messages for each valve or when a parameter is required to change state.\n\n??? What's the fastest Rate > or < 100 ms ??",
            "Units": "bit"
        },
        2346: {
            "Acronym": "AV15EF",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Aux Valve 15 Fail Safe Mode",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2346,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2347: {
            "Acronym": "LCMD",
            "DataRange": "0 to 3",
            "EndBit": 7,
            "Name": "High Beam Head Light Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2347,
            "SPNLength": 2,
            "StartBit": 6,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms",
            "Units": "bit"
        },
        2348: {
            "Acronym": "LD",
            "DataRange": "0 to 3",
            "EndBit": 7,
            "Name": "High Beam Head Light Data",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2348,
            "SPNLength": 2,
            "StartBit": 6,
            "TransmissionRate": "As requested.",
            "Units": "bit"
        },
        2349: {
            "Acronym": "LCMD",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Low Beam Head Light Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2349,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms",
            "Units": "bit"
        },
        235: {
            "Acronym": "IO",
            "DataRange": "0 to 210,554,060.75 hr",
            "EndBit": 63,
            "Name": "Engine Total Idle Hours",
            "Offset": 0.0,
            "OperationalHigh": 210554060.75,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.05,
            "SPN": 235,
            "SPNLength": 32,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "hours"
        },
        2350: {
            "Acronym": "LD",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Low Beam Head Light Data",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2350,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "As requested.",
            "Units": "bit"
        },
        2351: {
            "Acronym": "LCMD",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Alternate Beam Head Light Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2351,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms",
            "Units": "bit"
        },
        2352: {
            "Acronym": "LD",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Alternate Beam Head Light Data",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2352,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "As requested.",
            "Units": "bit"
        },
        2353: {
            "Acronym": "LCMD",
            "DataRange": "0 to 3",
            "EndBit": 45,
            "Name": "Tractor Front Low Mounted Work Lights Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2353,
            "SPNLength": 2,
            "StartBit": 44,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms",
            "Units": "bit"
        },
        2354: {
            "Acronym": "LD",
            "DataRange": "0 to 3",
            "EndBit": 45,
            "Name": "Tractor Front Low Mounted Work Lights",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2354,
            "SPNLength": 2,
            "StartBit": 44,
            "TransmissionRate": "As requested.",
            "Units": "bit"
        },
        2355: {
            "Acronym": "LCMD",
            "DataRange": "0 to 3",
            "EndBit": 47,
            "Name": "Tractor Front High Mounted Work Lights Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2355,
            "SPNLength": 2,
            "StartBit": 46,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms",
            "Units": "bit"
        },
        2356: {
            "Acronym": "LD",
            "DataRange": "0 to 3",
            "EndBit": 47,
            "Name": "Tractor Front High Mounted Work Lights",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2356,
            "SPNLength": 2,
            "StartBit": 46,
            "TransmissionRate": "As requested.",
            "Units": "bit"
        },
        2357: {
            "Acronym": "LCMD",
            "DataRange": "0 to 3",
            "EndBit": 35,
            "Name": "Tractor Underside Mounted Work Lights Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2357,
            "SPNLength": 2,
            "StartBit": 34,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms",
            "Units": "bit"
        },
        2358: {
            "Acronym": "LD",
            "DataRange": "0 to 3",
            "EndBit": 35,
            "Name": "Tractor Underside Mounted Work Lights",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2358,
            "SPNLength": 2,
            "StartBit": 34,
            "TransmissionRate": "As requested.",
            "Units": "bit"
        },
        2359: {
            "Acronym": "LCMD",
            "DataRange": "0 to 3",
            "EndBit": 37,
            "Name": "Tractor Rear Low Mounted Work Lights Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2359,
            "SPNLength": 2,
            "StartBit": 36,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms",
            "Units": "bit"
        },
        236: {
            "Acronym": "IO",
            "DataRange": "0 to 556224873.36449 US Gal",
            "EndBit": 31,
            "Name": "Engine Total Idle Fuel Used",
            "Offset": 0.0,
            "OperationalHigh": 556224873.36449,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.132086,
            "SPN": 236,
            "SPNLength": 32,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "gallons"
        },
        2360: {
            "Acronym": "LD",
            "DataRange": "0 to 3",
            "EndBit": 37,
            "Name": "Tractor Rear Low Mounted Work Lights",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2360,
            "SPNLength": 2,
            "StartBit": 36,
            "TransmissionRate": "As requested.",
            "Units": "bit"
        },
        2361: {
            "Acronym": "LCMD",
            "DataRange": "0 to 3",
            "EndBit": 39,
            "Name": "Tractor Rear High Mounted Work Lights Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2361,
            "SPNLength": 2,
            "StartBit": 38,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms",
            "Units": "bit"
        },
        2362: {
            "Acronym": "LD",
            "DataRange": "0 to 3",
            "EndBit": 39,
            "Name": "Tractor Rear High Mounted Work Lights",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2362,
            "SPNLength": 2,
            "StartBit": 38,
            "TransmissionRate": "As requested.",
            "Units": "bit"
        },
        2363: {
            "Acronym": "LCMD",
            "DataRange": "0 to 3",
            "EndBit": 41,
            "Name": "Tractor Side Low Mounted Work Lights Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2363,
            "SPNLength": 2,
            "StartBit": 40,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms",
            "Units": "bit"
        },
        2364: {
            "Acronym": "LD",
            "DataRange": "0 to 3",
            "EndBit": 41,
            "Name": "Tractor Side Low Mounted Work Lights",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2364,
            "SPNLength": 2,
            "StartBit": 40,
            "TransmissionRate": "As requested.",
            "Units": "bit"
        },
        2365: {
            "Acronym": "LCMD",
            "DataRange": "0 to 3",
            "EndBit": 43,
            "Name": "Tractor Side High Mounted Work Lights Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2365,
            "SPNLength": 2,
            "StartBit": 42,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms",
            "Units": "bit"
        },
        2366: {
            "Acronym": "LD",
            "DataRange": "0 to 3",
            "EndBit": 43,
            "Name": "Tractor Side High Mounted Work Lights",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2366,
            "SPNLength": 2,
            "StartBit": 42,
            "TransmissionRate": "As requested.",
            "Units": "bit"
        },
        2367: {
            "Acronym": "LCMD",
            "DataRange": "0 to 3",
            "EndBit": 15,
            "Name": "Left Turn Signal Lights Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2367,
            "SPNLength": 2,
            "StartBit": 14,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms",
            "Units": "bit"
        },
        2368: {
            "Acronym": "LD",
            "DataRange": "0 to 3",
            "EndBit": 15,
            "Name": "Left Turn Signal Lights",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2368,
            "SPNLength": 2,
            "StartBit": 14,
            "TransmissionRate": "As requested.",
            "Units": "bit"
        },
        2369: {
            "Acronym": "LCMD",
            "DataRange": "0 to 3",
            "EndBit": 13,
            "Name": "Right Turn Signal Lights Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2369,
            "SPNLength": 2,
            "StartBit": 12,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms",
            "Units": "bit"
        },
        237: {
            "Acronym": "VI",
            "DataRange": "0 to 255 per byte",
            "EndBit": 1599,
            "Name": "Vehicle Identification Number",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": -2,
            "SPN": 237,
            "SPNLength": 1600,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "ASCII"
        },
        2370: {
            "Acronym": "LD",
            "DataRange": "0 to 3",
            "EndBit": 13,
            "Name": "Right Turn Signal Lights",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2370,
            "SPNLength": 2,
            "StartBit": 12,
            "TransmissionRate": "As requested.",
            "Units": "bit"
        },
        2371: {
            "Acronym": "LCMD",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Left Stop Light Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2371,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms",
            "Units": "bit"
        },
        2372: {
            "Acronym": "LD",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Left Stop Light",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2372,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "As requested.",
            "Units": "bit"
        },
        2373: {
            "Acronym": "LCMD",
            "DataRange": "0 to 3",
            "EndBit": 21,
            "Name": "Right Stop Light Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2373,
            "SPNLength": 2,
            "StartBit": 20,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms",
            "Units": "bit"
        },
        2374: {
            "Acronym": "LD",
            "DataRange": "0 to 3",
            "EndBit": 21,
            "Name": "Right Stop Light",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2374,
            "SPNLength": 2,
            "StartBit": 20,
            "TransmissionRate": "As requested.",
            "Units": "bit"
        },
        2375: {
            "Acronym": "LCMD",
            "DataRange": "0 to 3",
            "EndBit": 19,
            "Name": "Center Stop Light Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2375,
            "SPNLength": 2,
            "StartBit": 18,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms",
            "Units": "bit"
        },
        2376: {
            "Acronym": "LD",
            "DataRange": "0 to 3",
            "EndBit": 19,
            "Name": "Center Stop Light",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2376,
            "SPNLength": 2,
            "StartBit": 18,
            "TransmissionRate": "As requested.",
            "Units": "bit"
        },
        2377: {
            "Acronym": "LCMD",
            "DataRange": "0 to 3",
            "EndBit": 31,
            "Name": "Tractor Marker Light Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2377,
            "SPNLength": 2,
            "StartBit": 30,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms",
            "Units": "bit"
        },
        2378: {
            "Acronym": "LD",
            "DataRange": "0 to 3",
            "EndBit": 31,
            "Name": "Tractor Marker Light",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2378,
            "SPNLength": 2,
            "StartBit": 30,
            "TransmissionRate": "As requested.",
            "Units": "bit"
        },
        2379: {
            "Acronym": "LCMD",
            "DataRange": "0 to 3",
            "EndBit": 29,
            "Name": "Implement Marker Light Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2379,
            "SPNLength": 2,
            "StartBit": 28,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms",
            "Units": "bit"
        },
        2380: {
            "Acronym": "LD",
            "DataRange": "0 to 3",
            "EndBit": 29,
            "Name": "Implement Marker Light",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2380,
            "SPNLength": 2,
            "StartBit": 28,
            "TransmissionRate": "As requested.",
            "Units": "bit"
        },
        2381: {
            "Acronym": "LCMD",
            "DataRange": "0 to 3",
            "EndBit": 27,
            "Name": "Tractor Clearance Light Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2381,
            "SPNLength": 2,
            "StartBit": 26,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms",
            "Units": "bit"
        },
        2382: {
            "Acronym": "LD",
            "DataRange": "0 to 3",
            "EndBit": 27,
            "Name": "Tractor Clearance Light",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2382,
            "SPNLength": 2,
            "StartBit": 26,
            "TransmissionRate": "As requested.",
            "Units": "bit"
        },
        2383: {
            "Acronym": "LCMD",
            "DataRange": "0 to 3",
            "EndBit": 25,
            "Name": "Implement Clearance Light Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2383,
            "SPNLength": 2,
            "StartBit": 24,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms",
            "Units": "bit"
        },
        2384: {
            "Acronym": "LD",
            "DataRange": "0 to 3",
            "EndBit": 25,
            "Name": "Implement Clearance Light",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2384,
            "SPNLength": 2,
            "StartBit": 24,
            "TransmissionRate": "As requested.",
            "Units": "bit"
        },
        2385: {
            "Acronym": "LCMD",
            "DataRange": "0 to 3",
            "EndBit": 11,
            "Name": "Rotating Beacon Light Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2385,
            "SPNLength": 2,
            "StartBit": 10,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms",
            "Units": "bit"
        },
        2386: {
            "Acronym": "LD",
            "DataRange": "0 to 3",
            "EndBit": 11,
            "Name": "Rotating Beacon Light",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2386,
            "SPNLength": 2,
            "StartBit": 10,
            "TransmissionRate": "As requested.",
            "Units": "bit"
        },
        2387: {
            "Acronym": "LCMD",
            "DataRange": "0 to 3",
            "EndBit": 9,
            "Name": "Tractor Front Fog Lights Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2387,
            "SPNLength": 2,
            "StartBit": 8,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms",
            "Units": "bit"
        },
        2388: {
            "Acronym": "LD",
            "DataRange": "0 to 3",
            "EndBit": 9,
            "Name": "Tractor Front Fog Lights",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2388,
            "SPNLength": 2,
            "StartBit": 8,
            "TransmissionRate": "As requested.",
            "Units": "bit"
        },
        2389: {
            "Acronym": "LCMD",
            "DataRange": "0 to 3",
            "EndBit": 33,
            "Name": "Rear Fog Light Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2389,
            "SPNLength": 2,
            "StartBit": 32,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms",
            "Units": "bit"
        },
        2390: {
            "Acronym": "LD",
            "DataRange": "0 to 3",
            "EndBit": 33,
            "Name": "Rear Fog Lights",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2390,
            "SPNLength": 2,
            "StartBit": 32,
            "TransmissionRate": "As requested.",
            "Units": "bit"
        },
        2391: {
            "Acronym": "LCMD",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Back Up Light and Alarm Horn Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2391,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms",
            "Units": "bit"
        },
        2392: {
            "Acronym": "LD",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Back Up Light and Alarm Horn",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2392,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "As requested.",
            "Units": "bit"
        },
        2393: {
            "Acronym": "LCMD",
            "DataRange": "0 to 3",
            "EndBit": 57,
            "Name": "Lighting Data Request Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2393,
            "SPNLength": 2,
            "StartBit": 56,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms",
            "Units": "bit"
        },
        2394: {
            "Acronym": "LD",
            "DataRange": "0 to 3",
            "EndBit": 63,
            "Name": "Implement Rear Work Light",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2394,
            "SPNLength": 2,
            "StartBit": 62,
            "TransmissionRate": "As requested.",
            "Units": "bit"
        },
        2395: {
            "Acronym": "LCMD",
            "DataRange": "0 to 3",
            "EndBit": 51,
            "Name": "Implement OEM Option 1 Light Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2395,
            "SPNLength": 2,
            "StartBit": 50,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms",
            "Units": "bit"
        },
        2396: {
            "Acronym": "LD",
            "DataRange": "0 to 3",
            "EndBit": 51,
            "Name": "Implement OEM Option 1 Light",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2396,
            "SPNLength": 2,
            "StartBit": 50,
            "TransmissionRate": "As requested.",
            "Units": "bit"
        },
        2397: {
            "Acronym": "LCMD",
            "DataRange": "0 to 3",
            "EndBit": 49,
            "Name": "Implement OEM Option 2 Light Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2397,
            "SPNLength": 2,
            "StartBit": 48,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms",
            "Units": "bit"
        },
        2398: {
            "Acronym": "LD",
            "DataRange": "0 to 3",
            "EndBit": 49,
            "Name": "Implement OEM Option 2 Light",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2398,
            "SPNLength": 2,
            "StartBit": 48,
            "TransmissionRate": "As requested.",
            "Units": "bit"
        },
        2399: {
            "Acronym": "LCMD",
            "DataRange": "0 to 3",
            "EndBit": 61,
            "Name": "Implement Left Facing Work Light Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2399,
            "SPNLength": 2,
            "StartBit": 60,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms",
            "Units": "bit"
        },
        2400: {
            "Acronym": "LD",
            "DataRange": "0 to 3",
            "EndBit": 61,
            "Name": "Implement Left Facing Work Light",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2400,
            "SPNLength": 2,
            "StartBit": 60,
            "TransmissionRate": "As requested.",
            "Units": "bit"
        },
        2401: {
            "Acronym": "LCMD",
            "DataRange": "0 to 3",
            "EndBit": 59,
            "Name": "Implement Right Forward Work Light Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2401,
            "SPNLength": 2,
            "StartBit": 58,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms",
            "Units": "bit"
        },
        2402: {
            "Acronym": "LD",
            "DataRange": "0 to 3",
            "EndBit": 59,
            "Name": "Implement Right Forward Work Light",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2402,
            "SPNLength": 2,
            "StartBit": 58,
            "TransmissionRate": "As requested.",
            "Units": "bit"
        },
        2403: {
            "Acronym": "LCMD",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Running Light Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2403,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms",
            "Units": "bit"
        },
        2404: {
            "Acronym": "LD",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Running Light",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2404,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "As requested.",
            "Units": "bit"
        },
        2405: {
            "Acronym": "LCMD",
            "DataRange": "0 to 3",
            "EndBit": 63,
            "Name": "Implement Rear Work Light Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2405,
            "SPNLength": 2,
            "StartBit": 62,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms",
            "Units": "bit"
        },
        2406: {
            "Acronym": "LCMD",
            "DataRange": "0 to 3",
            "EndBit": 53,
            "Name": "Implement Right Facing Work Light Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2406,
            "SPNLength": 2,
            "StartBit": 52,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms",
            "Units": "bit"
        },
        2407: {
            "Acronym": "LD",
            "DataRange": "0 to 3",
            "EndBit": 53,
            "Name": "Implement Right Facing Work Light",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2407,
            "SPNLength": 2,
            "StartBit": 52,
            "TransmissionRate": "As requested.",
            "Units": "bit"
        },
        2408: {
            "Acronym": "RPTO",
            "DataRange": "0 to 3",
            "EndBit": 39,
            "Name": "Rear Power Take Off Engagement",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2408,
            "SPNLength": 2,
            "StartBit": 38,
            "TransmissionRate": "100ms when engage, otherwise on request",
            "Units": "bit"
        },
        2409: {
            "Acronym": "WSMSTR",
            "DataRange": "0 to 250",
            "EndBit": 7,
            "Name": "Number of Members in Working Set",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2409,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "As Required",
            "Units": "count"
        },
        241: {
            "Acronym": "TIRE",
            "DataRange": "0.0 to 145.037738 psi",
            "EndBit": 15,
            "Name": "Tire Pressure",
            "Offset": 0.0,
            "OperationalHigh": 145.037738,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.580150952,
            "SPN": 241,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "10 s",
            "Units": "psi"
        },
        2410: {
            "Acronym": "LC",
            "DataRange": "0 to 255 per byte",
            "EndBit": 15,
            "Name": "Language Code Command",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -2,
            "SPN": 2410,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On system initialisation and on request",
            "Units": "ASCII"
        },
        2411: {
            "Acronym": "LC",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Decimal Symbol Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2411,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "On system initialisation and on request",
            "Units": "bit"
        },
        2412: {
            "Acronym": "LC",
            "DataRange": "0 to 255",
            "EndBit": 31,
            "Name": "Date Command",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2412,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "On system initialisation and on request",
            "Units": "binary"
        },
        2413: {
            "Acronym": "LC",
            "DataRange": "0 to 3",
            "EndBit": 21,
            "Name": "Time Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2413,
            "SPNLength": 2,
            "StartBit": 20,
            "TransmissionRate": "On system initialisation and on request",
            "Units": "bit"
        },
        2414: {
            "Acronym": "LC",
            "DataRange": "0 to 3",
            "EndBit": 39,
            "Name": "Distance Unit Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2414,
            "SPNLength": 2,
            "StartBit": 38,
            "TransmissionRate": "On system initialisation and on request",
            "Units": "bit"
        },
        2415: {
            "Acronym": "LC",
            "DataRange": "0 to 3",
            "EndBit": 37,
            "Name": "Area Unit Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2415,
            "SPNLength": 2,
            "StartBit": 36,
            "TransmissionRate": "On system initialisation and on request",
            "Units": "bit"
        },
        2416: {
            "Acronym": "LC",
            "DataRange": "0 to 3",
            "EndBit": 35,
            "Name": "Volume Unit Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2416,
            "SPNLength": 2,
            "StartBit": 34,
            "TransmissionRate": "On system initialisation and on request",
            "Units": "bit"
        },
        2417: {
            "Acronym": "LC",
            "DataRange": "0 to 3",
            "EndBit": 33,
            "Name": "Mass Unit Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2417,
            "SPNLength": 2,
            "StartBit": 32,
            "TransmissionRate": "On system initialisation and on request",
            "Units": "bit"
        },
        2418: {
            "Acronym": "RESRR",
            "DataRange": "0 to 64.255 s",
            "EndBit": 39,
            "Name": "Repetition Rate Parameter",
            "Offset": 0.0,
            "OperationalHigh": 64.255,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2418,
            "SPNLength": 16,
            "StartBit": 24,
            "TransmissionRate": "On request",
            "Units": "ms"
        },
        2419: {
            "Acronym": "RESRR",
            "DataRange": "0 to 3",
            "EndBit": 46,
            "Name": "Data Format/Error Condition",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2419,
            "SPNLength": 2,
            "StartBit": 45,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        242: {
            "Acronym": "TIRE",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 31,
            "Name": "Tire Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 242,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "10 s",
            "Units": "deg F"
        },
        2420: {
            "Acronym": "RESRR",
            "DataRange": "0 to 3",
            "EndBit": 44,
            "Name": "Process Data Type",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2420,
            "SPNLength": 2,
            "StartBit": 43,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        2421: {
            "Acronym": "RESRR",
            "DataRange": "0 to 7",
            "EndBit": 42,
            "Name": "Process Data Modifier",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2421,
            "SPNLength": 3,
            "StartBit": 40,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        2422: {
            "Acronym": "PD",
            "DataRange": "0 to 255",
            "EndBit": 15,
            "Name": "Count Number",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2422,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "As Events Require",
            "Units": "count"
        },
        2425: {
            "Acronym": "RESRR",
            "DataRange": "0 to 15",
            "EndBit": 55,
            "Name": "Implement Type",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2425,
            "SPNLength": 4,
            "StartBit": 52,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        2426: {
            "Acronym": "RESRR",
            "DataRange": "0 to 15",
            "EndBit": 51,
            "Name": "Implement Position",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2426,
            "SPNLength": 4,
            "StartBit": 48,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        2427: {
            "Acronym": "RESRR",
            "DataRange": "0 to 15",
            "EndBit": 63,
            "Name": "Data Dictionary Row",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2427,
            "SPNLength": 4,
            "StartBit": 60,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        2428: {
            "Acronym": "RESRR",
            "DataRange": "0 to 15",
            "EndBit": 59,
            "Name": "Data Dictionary Column",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2428,
            "SPNLength": 4,
            "StartBit": 56,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        2429: {
            "Acronym": "PD",
            "DataRange": "Variant Determined",
            "EndBit": 63,
            "Name": "Process Variable Value",
            "Offset": 0.0,
            "OperationalHigh": -1,
            "OperationalLow": -1,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -1,
            "SPN": 2429,
            "SPNLength": 32,
            "StartBit": 32,
            "TransmissionRate": "As Events Require",
            "Units": "VariantDetermined"
        },
        2432: {
            "Acronym": "EEC1",
            "DataRange": "-125 to 125 %",
            "EndBit": 63,
            "Name": "Engine Demand - Percent Torque",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "-125 to +125%",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2432,
            "SPNLength": 8,
            "StartBit": 56,
            "TransmissionRate": "engine speed dependent",
            "Units": "%"
        },
        2433: {
            "Acronym": "ET",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 15,
            "Name": "Engine Exhaust Gas Temperature - Right Manifold",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 2433,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "0.5 s",
            "Units": "deg F"
        },
        2434: {
            "Acronym": "ET",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 31,
            "Name": "Engine Exhaust Gas Temperature - Left Manifold",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 2434,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "0.5 s",
            "Units": "deg F"
        },
        2435: {
            "Acronym": "EAC",
            "DataRange": "0.0 to 72.518869 psi",
            "EndBit": 23,
            "Name": "Sea Water Pump Outlet Pressure",
            "Offset": 0.0,
            "OperationalHigh": 72.518869,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.290075476,
            "SPN": 2435,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "0.5 s",
            "Units": "psi"
        },
        2436: {
            "Acronym": "GAAC",
            "DataRange": "0 to 501.9921875 Hz",
            "EndBit": 47,
            "Name": "Generator Average AC Frequency",
            "Offset": 0.0,
            "OperationalHigh": 501.9921875,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2436,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "Hz"
        },
        2437: {
            "Acronym": "GPAAC",
            "DataRange": "0 to 501.9921875 Hz",
            "EndBit": 47,
            "Name": "Generator Phase A AC Frequency",
            "Offset": 0.0,
            "OperationalHigh": 501.9921875,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2437,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "Hz"
        },
        2438: {
            "Acronym": "GPBAC",
            "DataRange": "0 to 501.9921875 Hz",
            "EndBit": 47,
            "Name": "Generator Phase B AC Frequency",
            "Offset": 0.0,
            "OperationalHigh": 501.9921875,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2438,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "Hz"
        },
        2439: {
            "Acronym": "GPCAC",
            "DataRange": "0 to 501.9921875 Hz",
            "EndBit": 47,
            "Name": "Generator Phase C AC Frequency",
            "Offset": 0.0,
            "OperationalHigh": 501.9921875,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2439,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "Hz"
        },
        244: {
            "Acronym": "VD",
            "DataRange": "0.0 to 327080468.2212549 mile",
            "EndBit": 31,
            "Name": "Trip Distance",
            "Offset": 0.0,
            "OperationalHigh": 327080468.2212549,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.077671375,
            "SPN": 244,
            "SPNLength": 32,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "miles"
        },
        2440: {
            "Acronym": "GAAC",
            "DataRange": "0 to 64,255 Volts",
            "EndBit": 15,
            "Name": "Generator Average Line-Line AC RMS Voltage",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2440,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "Volts"
        },
        2441: {
            "Acronym": "GPAAC",
            "DataRange": "0 to 64,255 Volts",
            "EndBit": 15,
            "Name": "Generator Phase AB Line-Line AC RMS Voltage",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2441,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "Volts"
        },
        2442: {
            "Acronym": "GPBAC",
            "DataRange": "0 to 64,255 Volts",
            "EndBit": 15,
            "Name": "Generator Phase BC Line-Line AC RMS Voltage",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2442,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "Volts"
        },
        2443: {
            "Acronym": "GPCAC",
            "DataRange": "0 to 64,255 Volts",
            "EndBit": 15,
            "Name": "Generator Phase CA Line-Line AC RMS Voltage",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2443,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "Volts"
        },
        2444: {
            "Acronym": "GAAC",
            "DataRange": "0 to 64,255 Volts",
            "EndBit": 31,
            "Name": "Generator Average Line-Neutral AC RMS Voltage",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2444,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "Volts"
        },
        2445: {
            "Acronym": "GPAAC",
            "DataRange": "0 to 64,255 Volts",
            "EndBit": 31,
            "Name": "Generator Phase A Line-Neutral AC RMS Voltage",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2445,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "Volts"
        },
        2446: {
            "Acronym": "GPBAC",
            "DataRange": "0 to 64,255 Volts",
            "EndBit": 31,
            "Name": "Generator Phase B Line-Neutral AC RMS Voltage",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2446,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "Volts"
        },
        2447: {
            "Acronym": "GPCAC",
            "DataRange": "0 to 64,255 Volts",
            "EndBit": 31,
            "Name": "Generator Phase C Line-Neutral AC RMS Voltage",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2447,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "Volts"
        },
        2448: {
            "Acronym": "GAAC",
            "DataRange": "0 to 64,255 Amps",
            "EndBit": 63,
            "Name": "Generator Average AC RMS Current",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2448,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "Amps"
        },
        2449: {
            "Acronym": "GPAAC",
            "DataRange": "0 to 64,255 Amps",
            "EndBit": 63,
            "Name": "Generator Phase A AC RMS Current",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2449,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "Amps"
        },
        245: {
            "Acronym": "VD",
            "DataRange": "0.0 to 327080468.2212549 mile",
            "EndBit": 63,
            "Name": "Total Vehicle Distance",
            "Offset": 0.0,
            "OperationalHigh": 327080468.2212549,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.077671375,
            "SPN": 245,
            "SPNLength": 32,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "miles"
        },
        2450: {
            "Acronym": "GPBAC",
            "DataRange": "0 to 64,255 Amps",
            "EndBit": 63,
            "Name": "Generator Phase B AC RMS Current",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2450,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "Amps"
        },
        2451: {
            "Acronym": "GPCAC",
            "DataRange": "0 to 64,255 Amps",
            "EndBit": 63,
            "Name": "Generator Phase C AC RMS Current",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2451,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "Amps"
        },
        2452: {
            "Acronym": "GTACP",
            "DataRange": "-2,000,000,000 to +2,211,081,215 Watts",
            "EndBit": 31,
            "Name": "Generator Total Real Power",
            "Offset": -2000000000.0,
            "OperationalHigh": 2211081215.0,
            "OperationalLow": -2000000000.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2452,
            "SPNLength": 32,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "W"
        },
        2453: {
            "Acronym": "GPAACP",
            "DataRange": "-2,000,000,000 to +2,211,081,215 Watts",
            "EndBit": 31,
            "Name": "Generator Phase A Real Power",
            "Offset": -2000000000.0,
            "OperationalHigh": 2211081215.0,
            "OperationalLow": -2000000000.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2453,
            "SPNLength": 32,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "W"
        },
        2454: {
            "Acronym": "GPBACP",
            "DataRange": "-2,000,000,000 to +2,211,081,215 Watts",
            "EndBit": 31,
            "Name": "Generator Phase B Real Power",
            "Offset": -2000000000.0,
            "OperationalHigh": 2211081215.0,
            "OperationalLow": -2000000000.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2454,
            "SPNLength": 32,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "W"
        },
        2455: {
            "Acronym": "GPCACP",
            "DataRange": "-2,000,000,000 to +2,211,081,215 Watts",
            "EndBit": 31,
            "Name": "Generator Phase C Real Power",
            "Offset": -2000000000.0,
            "OperationalHigh": 2211081215.0,
            "OperationalLow": -2000000000.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2455,
            "SPNLength": 32,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "W"
        },
        2456: {
            "Acronym": "GTACR",
            "DataRange": "-2,000,000,000 to +2,211,081,215 VAr",
            "EndBit": 31,
            "Name": "Generator Total Reactive Power",
            "Offset": -2000000000.0,
            "OperationalHigh": 2211081215.0,
            "OperationalLow": -2000000000.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2456,
            "SPNLength": 32,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "VAr"
        },
        2457: {
            "Acronym": "GPAACR",
            "DataRange": "-2,000,000,000 to +2,211,081,215 VAr",
            "EndBit": 31,
            "Name": "Generator Phase A Reactive Power",
            "Offset": -2000000000.0,
            "OperationalHigh": 2211081215.0,
            "OperationalLow": -2000000000.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2457,
            "SPNLength": 32,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "VAr"
        },
        2458: {
            "Acronym": "GPBACRP",
            "DataRange": "-2,000,000,000 to +2,211,081,215 VAr",
            "EndBit": 31,
            "Name": "Generator Phase B Reactive Power",
            "Offset": -2000000000.0,
            "OperationalHigh": 2211081215.0,
            "OperationalLow": -2000000000.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2458,
            "SPNLength": 32,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "VAr"
        },
        2459: {
            "Acronym": "GPCACR",
            "DataRange": "-2,000,000,000 to +2,211,081,215 VAr",
            "EndBit": 31,
            "Name": "Generator Phase C Reactive Power",
            "Offset": -2000000000.0,
            "OperationalHigh": 2211081215.0,
            "OperationalLow": -2000000000.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2459,
            "SPNLength": 32,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "VAr"
        },
        246: {
            "Acronym": "VH",
            "DataRange": "0 to 210,554,060.75 hr",
            "EndBit": 31,
            "Name": "Total Vehicle Hours",
            "Offset": 0.0,
            "OperationalHigh": 210554060.75,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05,
            "SPN": 246,
            "SPNLength": 32,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "hours"
        },
        2460: {
            "Acronym": "GTACP",
            "DataRange": "-2,000,000,000 to +2,211,081,215 VA",
            "EndBit": 63,
            "Name": "Generator Total Apparent Power",
            "Offset": -2000000000.0,
            "OperationalHigh": 2211081215.0,
            "OperationalLow": -2000000000.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2460,
            "SPNLength": 32,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "VA"
        },
        2461: {
            "Acronym": "GPAACP",
            "DataRange": "-2,000,000,000 to +2,211,081,215 VA",
            "EndBit": 63,
            "Name": "Generator Phase A Apparent Power",
            "Offset": -2000000000.0,
            "OperationalHigh": 2211081215.0,
            "OperationalLow": -2000000000.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2461,
            "SPNLength": 32,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "VA"
        },
        2462: {
            "Acronym": "GPBACP",
            "DataRange": "-2,000,000,000 to +2,211,081,215 VA",
            "EndBit": 63,
            "Name": "Generator Phase B Apparent Power",
            "Offset": -2000000000.0,
            "OperationalHigh": 2211081215.0,
            "OperationalLow": -2000000000.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2462,
            "SPNLength": 32,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "VA"
        },
        2463: {
            "Acronym": "GPCACP",
            "DataRange": "-2,000,000,000 to +2,211,081,215 VA",
            "EndBit": 63,
            "Name": "Generator Phase C Apparent Power",
            "Offset": -2000000000.0,
            "OperationalHigh": 2211081215.0,
            "OperationalLow": -2000000000.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2463,
            "SPNLength": 32,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "VA"
        },
        2464: {
            "Acronym": "GTACR",
            "DataRange": "-1.00000 to +2.921814",
            "EndBit": 47,
            "Name": "Generator Overall Power Factor",
            "Offset": -1.0,
            "OperationalHigh": 1.0,
            "OperationalLow": -1.0,
            "OperationalRange": "-1 to +1",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2464,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": ""
        },
        2465: {
            "Acronym": "GPAACR",
            "DataRange": "-1.00000 to +2.921814",
            "EndBit": 47,
            "Name": "Generator Phase A Power Factor",
            "Offset": -1.0,
            "OperationalHigh": 1.0,
            "OperationalLow": -1.0,
            "OperationalRange": "-1 to +1",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2465,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": ""
        },
        2466: {
            "Acronym": "GPBACRP",
            "DataRange": "-1.00000 to +2.921814",
            "EndBit": 47,
            "Name": "Generator Phase B Power Factor",
            "Offset": -1.0,
            "OperationalHigh": 1.0,
            "OperationalLow": -1.0,
            "OperationalRange": "-1 to +1",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2466,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": ""
        },
        2467: {
            "Acronym": "GPCACR",
            "DataRange": "-1.00000 to +2.921814",
            "EndBit": 47,
            "Name": "Generator Phase C Power Factor",
            "Offset": -1.0,
            "OperationalHigh": 1.0,
            "OperationalLow": -1.0,
            "OperationalRange": "-1 to +1",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2467,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": ""
        },
        2468: {
            "Acronym": "GTACE",
            "DataRange": "0 to 4211081215 kWh",
            "EndBit": 31,
            "Name": "Generator Total kW Hours Export",
            "Offset": 0.0,
            "OperationalHigh": 4211081215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2468,
            "SPNLength": 32,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "kWh"
        },
        2469: {
            "Acronym": "GTACE",
            "DataRange": "0 to 4211081215 kWh",
            "EndBit": 63,
            "Name": "Generator Total kW Hours Import",
            "Offset": 0.0,
            "OperationalHigh": 4211081215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2469,
            "SPNLength": 32,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "kWh"
        },
        247: {
            "Acronym": "HOURS",
            "DataRange": "0 to 210,554,060.75 hr",
            "EndBit": 31,
            "Name": "Engine Total Hours of Operation",
            "Offset": 0.0,
            "OperationalHigh": 210554060.75,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05,
            "SPN": 247,
            "SPNLength": 32,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "hours"
        },
        2470: {
            "Acronym": "UAAC",
            "DataRange": "0 to 501.9921875 Hz",
            "EndBit": 47,
            "Name": "Utility Average AC Frequency",
            "Offset": 0.0,
            "OperationalHigh": 501.9921875,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2470,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "Hz"
        },
        2471: {
            "Acronym": "UPAAC",
            "DataRange": "0 to 501.9921875 Hz",
            "EndBit": 47,
            "Name": "Utility Phase A AC Frequency",
            "Offset": 0.0,
            "OperationalHigh": 501.9921875,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2471,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "Hz"
        },
        2472: {
            "Acronym": "UPBAC",
            "DataRange": "0 to 501.9921875 Hz",
            "EndBit": 47,
            "Name": "Utility Phase B AC Frequency",
            "Offset": 0.0,
            "OperationalHigh": 501.9921875,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2472,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "Hz"
        },
        2473: {
            "Acronym": "UPCAC",
            "DataRange": "0 to 501.9921875 Hz",
            "EndBit": 47,
            "Name": "Utility Phase C AC Frequency",
            "Offset": 0.0,
            "OperationalHigh": 501.9921875,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2473,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "Hz"
        },
        2474: {
            "Acronym": "UAAC",
            "DataRange": "0 to 64,255 Volts",
            "EndBit": 15,
            "Name": "Utility Average Line-Line AC RMS Voltage",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2474,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "Volts"
        },
        2475: {
            "Acronym": "UPAAC",
            "DataRange": "0 to 64,255 Volts",
            "EndBit": 15,
            "Name": "Utility Phase AB Line-Line AC RMS Voltage",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2475,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "Volts"
        },
        2476: {
            "Acronym": "UPBAC",
            "DataRange": "0 to 64,255 Volts",
            "EndBit": 15,
            "Name": "Utility Phase BC Line-Line AC RMS Voltage",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2476,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "Volts"
        },
        2477: {
            "Acronym": "UPCAC",
            "DataRange": "0 to 64,255 Volts",
            "EndBit": 15,
            "Name": "Utility Phase CA Line-Line AC RMS Voltage",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2477,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "Volts"
        },
        2478: {
            "Acronym": "UAAC",
            "DataRange": "0 to 64,255 Volts",
            "EndBit": 31,
            "Name": "Utility Average Line-Neutral AC RMS Voltage",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2478,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "Volts"
        },
        2479: {
            "Acronym": "UPAAC",
            "DataRange": "0 to 64,255 Volts",
            "EndBit": 31,
            "Name": "Utility Phase A Line-Neutral AC RMS Voltage",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2479,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "Volts"
        },
        248: {
            "Acronym": "VH",
            "DataRange": "0 to 210,554,060.75 hr",
            "EndBit": 63,
            "Name": "Total Power Takeoff Hours",
            "Offset": 0.0,
            "OperationalHigh": 210554060.75,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05,
            "SPN": 248,
            "SPNLength": 32,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "hours"
        },
        2480: {
            "Acronym": "UPBAC",
            "DataRange": "0 to 64,255 Volts",
            "EndBit": 31,
            "Name": "Utility Phase B Line-Neutral AC RMS Voltage",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2480,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "Volts"
        },
        2481: {
            "Acronym": "UPCAC",
            "DataRange": "0 to 64,255 Volts",
            "EndBit": 31,
            "Name": "Utility Phase C Line-Neutral AC RMS Voltage",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2481,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "Volts"
        },
        2482: {
            "Acronym": "UAAC",
            "DataRange": "0 to 64,255 Amps",
            "EndBit": 63,
            "Name": "Utility Average AC RMS Current",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2482,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "Amps"
        },
        2483: {
            "Acronym": "UPAAC",
            "DataRange": "0 to 64,255 Amps",
            "EndBit": 63,
            "Name": "Utility Phase A AC RMS Current",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2483,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "Amps"
        },
        2484: {
            "Acronym": "UPBAC",
            "DataRange": "0 to 64,255 Amps",
            "EndBit": 63,
            "Name": "Utility Phase B AC RMS Current",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2484,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "Amps"
        },
        2485: {
            "Acronym": "UPCAC",
            "DataRange": "0 to 64,255 Amps",
            "EndBit": 63,
            "Name": "Utility Phase C AC RMS Current",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2485,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "Amps"
        },
        2486: {
            "Acronym": "UTACP",
            "DataRange": "-2,000,000,000 to +2,211,081,215 Watts",
            "EndBit": 31,
            "Name": "Utility Total Real Power",
            "Offset": -2000000000.0,
            "OperationalHigh": 2211081215.0,
            "OperationalLow": -2000000000.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2486,
            "SPNLength": 32,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "W"
        },
        2487: {
            "Acronym": "UPAACP",
            "DataRange": "-2,000,000,000 to +2,211,081,215 Watts",
            "EndBit": 31,
            "Name": "Utility Phase A Real Power",
            "Offset": -2000000000.0,
            "OperationalHigh": 2211081215.0,
            "OperationalLow": -2000000000.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2487,
            "SPNLength": 32,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "W"
        },
        2488: {
            "Acronym": "UPBACP",
            "DataRange": "-2,000,000,000 to +2,211,081,215 Watts",
            "EndBit": 31,
            "Name": "Utility Phase B Real Power",
            "Offset": -2000000000.0,
            "OperationalHigh": 2211081215.0,
            "OperationalLow": -2000000000.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2488,
            "SPNLength": 32,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "W"
        },
        2489: {
            "Acronym": "UPCACP",
            "DataRange": "-2,000,000,000 to +2,211,081,215 Watts",
            "EndBit": 31,
            "Name": "Utility Phase C Real Power",
            "Offset": -2000000000.0,
            "OperationalHigh": 2211081215.0,
            "OperationalLow": -2000000000.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2489,
            "SPNLength": 32,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "W"
        },
        249: {
            "Acronym": "HOURS",
            "DataRange": "0 to 4,211,081,215,000 r",
            "EndBit": 63,
            "Name": "Engine Total Revolutions",
            "Offset": 0.0,
            "OperationalHigh": 4211081215000.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1000,
            "SPN": 249,
            "SPNLength": 32,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "r"
        },
        2490: {
            "Acronym": "UTACR",
            "DataRange": "-2,000,000,000 to +2,211,081,215 VAr",
            "EndBit": 31,
            "Name": "Utility Total Reactive Power",
            "Offset": -2000000000.0,
            "OperationalHigh": 2211081215.0,
            "OperationalLow": -2000000000.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2490,
            "SPNLength": 32,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "VAr"
        },
        2491: {
            "Acronym": "UPACCR",
            "DataRange": "-2,000,000,000 to +2,211,081,215 VAr",
            "EndBit": 31,
            "Name": "Utility Phase A Reactive Power",
            "Offset": -2000000000.0,
            "OperationalHigh": 2211081215.0,
            "OperationalLow": -2000000000.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2491,
            "SPNLength": 32,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "VAr"
        },
        2492: {
            "Acronym": "UPBACR",
            "DataRange": "-2,000,000,000 to +2,211,081,215 VAr",
            "EndBit": 31,
            "Name": "Utility Phase B Reactive Power",
            "Offset": -2000000000.0,
            "OperationalHigh": 2211081215.0,
            "OperationalLow": -2000000000.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2492,
            "SPNLength": 32,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "VAr"
        },
        2493: {
            "Acronym": "UPCACR",
            "DataRange": "-2,000,000,000 to +2,211,081,215 VAr",
            "EndBit": 31,
            "Name": "Utility Phase C Reactive Power",
            "Offset": -2000000000.0,
            "OperationalHigh": 2211081215.0,
            "OperationalLow": -2000000000.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2493,
            "SPNLength": 32,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "VAr"
        },
        2494: {
            "Acronym": "UTACP",
            "DataRange": "-2,000,000,000 to +2,211,081,215 VA",
            "EndBit": 63,
            "Name": "Utility Total Apparent Power",
            "Offset": -2000000000.0,
            "OperationalHigh": 2211081215.0,
            "OperationalLow": -2000000000.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2494,
            "SPNLength": 32,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "VA"
        },
        2495: {
            "Acronym": "UPAACP",
            "DataRange": "-2,000,000,000 to +2,211,081,215 VA",
            "EndBit": 63,
            "Name": "Utility Phase A Apparent Power",
            "Offset": -2000000000.0,
            "OperationalHigh": 2211081215.0,
            "OperationalLow": -2000000000.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2495,
            "SPNLength": 32,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "VA"
        },
        2496: {
            "Acronym": "UPBACP",
            "DataRange": "-2,000,000,000 to +2,211,081,215 VA",
            "EndBit": 63,
            "Name": "Utility Phase B Apparent Power",
            "Offset": -2000000000.0,
            "OperationalHigh": 2211081215.0,
            "OperationalLow": -2000000000.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2496,
            "SPNLength": 32,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "VA"
        },
        2497: {
            "Acronym": "UPCACP",
            "DataRange": "-2,000,000,000 to +2,211,081,215 VA",
            "EndBit": 63,
            "Name": "Utility Phase C Apparent Power",
            "Offset": -2000000000.0,
            "OperationalHigh": 2211081215.0,
            "OperationalLow": -2000000000.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2497,
            "SPNLength": 32,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "VA"
        },
        2498: {
            "Acronym": "UTACR",
            "DataRange": "-1.00000 to +2.921814",
            "EndBit": 47,
            "Name": "Utility Overall Power Factor",
            "Offset": -1.0,
            "OperationalHigh": 2.921814,
            "OperationalLow": -1.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2498,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": ""
        },
        2499: {
            "Acronym": "UPACCR",
            "DataRange": "-1.00000 to +2.921814",
            "EndBit": 47,
            "Name": "Utility Phase A Power Factor",
            "Offset": -1.0,
            "OperationalHigh": 2.921814,
            "OperationalLow": -1.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2499,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": ""
        },
        250: {
            "Acronym": "LFC",
            "DataRange": "0.0 to 556224873.36449 US Gal",
            "EndBit": 63,
            "Name": "Engine Total Fuel Used",
            "Offset": 0.0,
            "OperationalHigh": 556224873.36449,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.132086,
            "SPN": 250,
            "SPNLength": 32,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "gallons"
        },
        2500: {
            "Acronym": "UPBACR",
            "DataRange": "-1.00000 to +2.921814",
            "EndBit": 47,
            "Name": "Utility Phase B Power Factor",
            "Offset": -1.0,
            "OperationalHigh": 2.921814,
            "OperationalLow": -1.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2500,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": ""
        },
        2501: {
            "Acronym": "UPCACR",
            "DataRange": "-1.00000 to +2.921814",
            "EndBit": 47,
            "Name": "Utility Phase C Power Factor",
            "Offset": -1.0,
            "OperationalHigh": 2.921814,
            "OperationalLow": -1.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2501,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": ""
        },
        2502: {
            "Acronym": "UTACE",
            "DataRange": "0 to 4211081215 kWh",
            "EndBit": 31,
            "Name": "Utility Total kW Hours Export",
            "Offset": 0.0,
            "OperationalHigh": 4211081215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2502,
            "SPNLength": 32,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "kWh"
        },
        2503: {
            "Acronym": "UTACE",
            "DataRange": "0 to 4211081215 kWh",
            "EndBit": 63,
            "Name": "Utility Total kW Hours Import",
            "Offset": 0.0,
            "OperationalHigh": 4211081215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2503,
            "SPNLength": 32,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "kWh"
        },
        2504: {
            "Acronym": "BAAC",
            "DataRange": "0 to 501.9921875 Hz",
            "EndBit": 47,
            "Name": "Bus #1 Average AC Frequency",
            "Offset": 0.0,
            "OperationalHigh": 501.9921875,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2504,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "Hz"
        },
        2505: {
            "Acronym": "BPAAC",
            "DataRange": "0 to 501.9921875 Hz",
            "EndBit": 47,
            "Name": "Bus #1 Phase A AC Frequency",
            "Offset": 0.0,
            "OperationalHigh": 501.9921875,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2505,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "Hz"
        },
        2506: {
            "Acronym": "BPBAC",
            "DataRange": "0 to 501.9921875 Hz",
            "EndBit": 47,
            "Name": "Bus #1 Phase B AC Frequency",
            "Offset": 0.0,
            "OperationalHigh": 501.9921875,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2506,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "Hz"
        },
        2507: {
            "Acronym": "BPCAC",
            "DataRange": "0 to 501.9921875 Hz",
            "EndBit": 47,
            "Name": "Bus #1 Phase C AC Frequency",
            "Offset": 0.0,
            "OperationalHigh": 501.9921875,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2507,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "Hz"
        },
        2508: {
            "Acronym": "BAAC",
            "DataRange": "0 to 64,255 Volts",
            "EndBit": 15,
            "Name": "Bus #1 Average Line-Line AC RMS Voltage",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2508,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "Volts"
        },
        2509: {
            "Acronym": "BPAAC",
            "DataRange": "0 to 64,255 Volts",
            "EndBit": 15,
            "Name": "Bus #1 Phase AB Line-Line AC RMS Voltage",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2509,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "Volts"
        },
        2510: {
            "Acronym": "BPBAC",
            "DataRange": "0 to 64,255 Volts",
            "EndBit": 15,
            "Name": "Bus #1 Phase BC Line-Line AC RMS Voltage",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2510,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "Volts"
        },
        2511: {
            "Acronym": "BPCAC",
            "DataRange": "0 to 64,255 Volts",
            "EndBit": 15,
            "Name": "Bus #1 Phase CA Line-Line AC RMS Voltage",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2511,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "Volts"
        },
        2512: {
            "Acronym": "BAAC",
            "DataRange": "0 to 64,255 Volts",
            "EndBit": 31,
            "Name": "Bus #1 Average Line-Neutral AC RMS Voltage",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2512,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "Volts"
        },
        2513: {
            "Acronym": "BPAAC",
            "DataRange": "0 to 64,255 Volts",
            "EndBit": 31,
            "Name": "Bus #1 Phase A Line-Neutral AC RMS Voltage",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2513,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "Volts"
        },
        2514: {
            "Acronym": "BPBAC",
            "DataRange": "0 to 64,255 Volts",
            "EndBit": 31,
            "Name": "Bus #1 Phase B Line-Neutral AC RMS Voltage",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2514,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "Volts"
        },
        2515: {
            "Acronym": "BPCAC",
            "DataRange": "0 to 64,255 Volts",
            "EndBit": 31,
            "Name": "Bus #1 Phase C Line-Neutral AC RMS Voltage",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2515,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "Volts"
        },
        2516: {
            "Acronym": "BGSC",
            "DataRange": "-200 to 301.99 deg",
            "EndBit": 31,
            "Name": "Bus #1/Generator AC Phase Difference",
            "Offset": -200.0,
            "OperationalHigh": 301.99,
            "OperationalLow": -200.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2516,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "deg"
        },
        2517: {
            "Acronym": "BUSC",
            "DataRange": "-200 to 301.99 deg",
            "EndBit": 31,
            "Name": "Bus #1/Utility AC Phase Difference",
            "Offset": -200.0,
            "OperationalHigh": 301.99,
            "OperationalLow": -200.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2517,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "deg"
        },
        2518: {
            "Acronym": "GTACR",
            "DataRange": "0 to 3",
            "EndBit": 49,
            "Name": "Generator Overall Power Factor Lagging",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2518,
            "SPNLength": 2,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2519: {
            "Acronym": "GPAACR",
            "DataRange": "0 to 3",
            "EndBit": 49,
            "Name": "Generator Phase A Power Factor Lagging",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2519,
            "SPNLength": 2,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2520: {
            "Acronym": "GPBACRP",
            "DataRange": "0 to 3",
            "EndBit": 49,
            "Name": "Generator Phase B Power Factor Lagging",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2520,
            "SPNLength": 2,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2521: {
            "Acronym": "GPCACR",
            "DataRange": "0 to 3",
            "EndBit": 49,
            "Name": "Generator Phase C Power Factor Lagging",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2521,
            "SPNLength": 2,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2522: {
            "Acronym": "UTACR",
            "DataRange": "0 to 3",
            "EndBit": 49,
            "Name": "Utility Overall Power Factor Lagging",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2522,
            "SPNLength": 2,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2523: {
            "Acronym": "UPACCR",
            "DataRange": "0 to 3",
            "EndBit": 49,
            "Name": "Utility Phase A Power Factor Lagging",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2523,
            "SPNLength": 2,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2524: {
            "Acronym": "UPBACR",
            "DataRange": "0 to 3",
            "EndBit": 49,
            "Name": "Utility Phase B Power Factor Lagging",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2524,
            "SPNLength": 2,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2525: {
            "Acronym": "UPCACR",
            "DataRange": "0 to 3",
            "EndBit": 49,
            "Name": "Utility Phase C Power Factor Lagging",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2525,
            "SPNLength": 2,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2526: {
            "Acronym": "BGSC",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Bus #1/Generator Phase Match",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2526,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2527: {
            "Acronym": "BGSC",
            "DataRange": "0 to 3",
            "EndBit": 7,
            "Name": "Bus #1/Generator Voltage Match",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2527,
            "SPNLength": 2,
            "StartBit": 6,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2528: {
            "Acronym": "BGSC",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Bus #1/Generator Frequency Match",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2528,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2529: {
            "Acronym": "BGSC",
            "DataRange": "0 to 3",
            "EndBit": 9,
            "Name": "Bus #1/Generator In Sync",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2529,
            "SPNLength": 2,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2530: {
            "Acronym": "BGSC",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Bus #1/Generator Dead Bus",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2530,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2531: {
            "Acronym": "BUSC",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Bus #1/Utility Phase Match",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2531,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2532: {
            "Acronym": "BUSC",
            "DataRange": "0 to 3",
            "EndBit": 7,
            "Name": "Bus #1/Utility Voltage Match",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2532,
            "SPNLength": 2,
            "StartBit": 6,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2533: {
            "Acronym": "BUSC",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Bus #1/Utility Frequency Match",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2533,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2534: {
            "Acronym": "BUSC",
            "DataRange": "0 to 3",
            "EndBit": 9,
            "Name": "Bus #1/Utility In Sync",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2534,
            "SPNLength": 2,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2535: {
            "Acronym": "BUSC",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Bus #1/Utility Dead Bus",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2535,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2536: {
            "Acronym": "ETC7",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Transmission Mode 1 Indicator",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2536,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2537: {
            "Acronym": "ETC7",
            "DataRange": "0 to 3",
            "EndBit": 21,
            "Name": "Transmission Mode 2 Indicator",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2537,
            "SPNLength": 2,
            "StartBit": 20,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2538: {
            "Acronym": "ETC7",
            "DataRange": "0 to 3",
            "EndBit": 19,
            "Name": "Transmission Mode 3 Indicator",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2538,
            "SPNLength": 2,
            "StartBit": 18,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2539: {
            "Acronym": "ETC7",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Transmission Mode 4 Indicator",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2539,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2540: {
            "Acronym": "RQST",
            "DataRange": "0 to 16,777,215",
            "EndBit": 23,
            "Name": "Parameter Group Number (RQST)",
            "Offset": 0.0,
            "OperationalHigh": 16777215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 3,
            "Resolution": -3,
            "SPN": 2540,
            "SPNLength": 24,
            "StartBit": 0,
            "TransmissionRate": "Per user requirements, generally recommended that requests occur no more than 2 or 3 times per second.",
            "Units": "binary"
        },
        2541: {
            "Acronym": "ACKM",
            "DataRange": "0 to 255",
            "EndBit": 7,
            "Name": "Control Byte (ACKM)",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 3",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2541,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "Upon reception of a Parameter Group Number that requires this form of acknowledgment.",
            "Units": "binary"
        },
        2542: {
            "Acronym": "ACKM",
            "DataRange": "0 to 255",
            "EndBit": 0,
            "Name": "Group Function Value (ACK)",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 250",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2542,
            "SPNLength": 8,
            "StartBit": -1,
            "TransmissionRate": "Upon reception of a Parameter Group Number that requires this form of acknowledgment.",
            "Units": "binary"
        },
        2543: {
            "Acronym": "ACKM",
            "DataRange": "0 to 16,777,215",
            "EndBit": 0,
            "Name": "Parameter Group Number (ACK)",
            "Offset": 0.0,
            "OperationalHigh": 16777215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 2543,
            "SPNLength": 24,
            "StartBit": -1,
            "TransmissionRate": "Upon reception of a Parameter Group Number that requires this form of acknowledgment.",
            "Units": "binary"
        },
        2544: {
            "Acronym": "ACKM",
            "DataRange": "0 to 255",
            "EndBit": 0,
            "Name": "Group Function Value (NACK)",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 250",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2544,
            "SPNLength": 8,
            "StartBit": -1,
            "TransmissionRate": "Upon reception of a Parameter Group Number that requires this form of acknowledgment.",
            "Units": "binary"
        },
        2545: {
            "Acronym": "ACKM",
            "DataRange": "0 to 16,777,215",
            "EndBit": 0,
            "Name": "Parameter Group Number (NACK)",
            "Offset": 0.0,
            "OperationalHigh": 16777215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 2545,
            "SPNLength": 24,
            "StartBit": -1,
            "TransmissionRate": "Upon reception of a Parameter Group Number that requires this form of acknowledgment.",
            "Units": "binary"
        },
        2546: {
            "Acronym": "ACKM",
            "DataRange": "0 to 255",
            "EndBit": 0,
            "Name": "Group Function Value (NACK_AD)",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 250",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2546,
            "SPNLength": 8,
            "StartBit": -1,
            "TransmissionRate": "Upon reception of a Parameter Group Number that requires this form of acknowledgment.",
            "Units": "binary"
        },
        2547: {
            "Acronym": "ACKM",
            "DataRange": "0 to 16,777,215",
            "EndBit": 0,
            "Name": "Parameter Group Number  (NACK_AD)",
            "Offset": 0.0,
            "OperationalHigh": 16777215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 2547,
            "SPNLength": 24,
            "StartBit": -1,
            "TransmissionRate": "Upon reception of a Parameter Group Number that requires this form of acknowledgment.",
            "Units": "binary"
        },
        2548: {
            "Acronym": "ACKM",
            "DataRange": "0 to 255",
            "EndBit": 0,
            "Name": "Group Function Value (NACK_Busy)",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 250",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2548,
            "SPNLength": 8,
            "StartBit": -1,
            "TransmissionRate": "Upon reception of a Parameter Group Number that requires this form of acknowledgment.",
            "Units": "binary"
        },
        2549: {
            "Acronym": "ACKM",
            "DataRange": "0 to 16,777,215",
            "EndBit": 0,
            "Name": "Parameter Group Number (NACK_Busy)",
            "Offset": 0.0,
            "OperationalHigh": 16777215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 2549,
            "SPNLength": 24,
            "StartBit": -1,
            "TransmissionRate": "Upon reception of a Parameter Group Number that requires this form of acknowledgment.",
            "Units": "binary"
        },
        2550: {
            "Acronym": "PropA",
            "DataRange": "Manufacturer Determined",
            "EndBit": 63,
            "Name": "Manufacturer Specific Information (PropA_PDU1)",
            "Offset": 0.0,
            "OperationalHigh": -1,
            "OperationalLow": -1,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": -1,
            "SPN": 2550,
            "SPNLength": 14280,
            "StartBit": 0,
            "TransmissionRate": "Per user requirements",
            "Units": "Manufacturer Determined"
        },
        2551: {
            "Acronym": "PropB_FF",
            "DataRange": "Manufacturer Determined",
            "EndBit": 63,
            "Name": "Manufacturer Defined Usage (PropB_PDU2)",
            "Offset": 0.0,
            "OperationalHigh": -1,
            "OperationalLow": -1,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": -1,
            "SPN": 2551,
            "SPNLength": 14280,
            "StartBit": 0,
            "TransmissionRate": "Per user requirements",
            "Units": "Manufacturer Determined"
        },
        2552: {
            "Acronym": "XFER",
            "DataRange": "0 to 16,777,215",
            "EndBit": 23,
            "Name": "Parameter Group Number of Requested Information (XFER)",
            "Offset": 0.0,
            "OperationalHigh": 16777215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "9 to 1785 bytes (multipacket supported)",
            "Resolution": -3,
            "SPN": 2552,
            "SPNLength": 24,
            "StartBit": 0,
            "TransmissionRate": "In response to a Request2 PGN with \"Use Transfer PGN\" = 01",
            "Units": "binary"
        },
        2553: {
            "Acronym": "XFER",
            "DataRange": "0 to 255",
            "EndBit": 31,
            "Name": "Length of data for the reported PGN (XFER)",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "9 to 1785 bytes (multipacket supported)",
            "Resolution": 1,
            "SPN": 2553,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "In response to a Request2 PGN with \"Use Transfer PGN\" = 01",
            "Units": "bit"
        },
        2554: {
            "Acronym": "XFER",
            "DataRange": "0 to 4294967296",
            "EndBit": 63,
            "Name": "Controller Application Identity (CAI)",
            "Offset": 0.0,
            "OperationalHigh": 4294967296.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "9 to 1785 bytes (multipacket supported)",
            "Resolution": 1,
            "SPN": 2554,
            "SPNLength": 32,
            "StartBit": 32,
            "TransmissionRate": "In response to a Request2 PGN with \"Use Transfer PGN\" = 01",
            "Units": "bit"
        },
        2555: {
            "Acronym": "XFER",
            "DataRange": "9 to 1777 bytes of data",
            "EndBit": -1,
            "Name": "Transfer Data",
            "Offset": 0.0,
            "OperationalHigh": 1777.0,
            "OperationalLow": 9.0,
            "OperationalRange": "",
            "PGNLength": "9 to 1785 bytes (multipacket supported)",
            "Resolution": -1,
            "SPN": 2555,
            "SPNLength": 14216,
            "StartBit": 8,
            "TransmissionRate": "In response to a Request2 PGN with \"Use Transfer PGN\" = 01",
            "Units": "Request Dependent"
        },
        2556: {
            "Acronym": "TP.CM.xx",
            "DataRange": "0 to 255",
            "EndBit": 7,
            "Name": "Control Byte (TP.CM)",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2556,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "Per the Parameter Group to be transferred",
            "Units": "binary"
        },
        2560: {
            "Acronym": "TP.CM.xx",
            "DataRange": "0 to 16,777,215",
            "EndBit": 0,
            "Name": "Parameter Group Number of the packeted message (TP.CM_RTS)",
            "Offset": 0.0,
            "OperationalHigh": 16777215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 2560,
            "SPNLength": 24,
            "StartBit": -1,
            "TransmissionRate": "Per the Parameter Group to be transferred",
            "Units": "binary"
        },
        2561: {
            "Acronym": "TP.CM.xx",
            "DataRange": "0 to 255",
            "EndBit": 0,
            "Name": "Number of Packets that can be sent (TP.CM_CTS)",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 255",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2561,
            "SPNLength": 8,
            "StartBit": -1,
            "TransmissionRate": "Per the Parameter Group to be transferred",
            "Units": "count"
        },
        2563: {
            "Acronym": "TP.CM.xx",
            "DataRange": "0 to 16,777,215",
            "EndBit": 0,
            "Name": "Parameter Group Number of the packeted message (TP.CM_CTS)",
            "Offset": 0.0,
            "OperationalHigh": 16777215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 2563,
            "SPNLength": 24,
            "StartBit": -1,
            "TransmissionRate": "Per the Parameter Group to be transferred",
            "Units": "binary"
        },
        2566: {
            "Acronym": "TP.CM.xx",
            "DataRange": "0 to 16,777,215",
            "EndBit": 0,
            "Name": "Parameter Group Number of the packeted message (TP.CM_EndofMsgACK)",
            "Offset": 0.0,
            "OperationalHigh": 16777215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 2566,
            "SPNLength": 24,
            "StartBit": -1,
            "TransmissionRate": "Per the Parameter Group to be transferred",
            "Units": "binary"
        },
        2569: {
            "Acronym": "TP.CM.xx",
            "DataRange": "0 to 16,777,215",
            "EndBit": 0,
            "Name": "Parameter Group Number of the packeted message (TP.CM_BAM)",
            "Offset": 0.0,
            "OperationalHigh": 16777215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 2569,
            "SPNLength": 24,
            "StartBit": -1,
            "TransmissionRate": "Per the Parameter Group to be transferred",
            "Units": "binary"
        },
        2571: {
            "Acronym": "TP.CM.xx",
            "DataRange": "0 to 16,777,215",
            "EndBit": 0,
            "Name": "Parameter Group Number of packeted message (TP.CM_Conn_Abort)",
            "Offset": 0.0,
            "OperationalHigh": 16777215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 2571,
            "SPNLength": 24,
            "StartBit": -1,
            "TransmissionRate": "Per the Parameter Group to be transferred",
            "Units": "binary"
        },
        2572: {
            "Acronym": "TP.DT",
            "DataRange": "0 to 255",
            "EndBit": 7,
            "Name": "Sequence Number (TP.DT)",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 1.0,
            "OperationalRange": "1 to 255",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2572,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "Per the Parameter Group to be transferred",
            "Units": "count"
        },
        2573: {
            "Acronym": "TP.DT",
            "DataRange": "9 to 1784 bytes of data",
            "EndBit": -1,
            "Name": "Packetized Data (TP.DT)",
            "Offset": 0.0,
            "OperationalHigh": 1784.0,
            "OperationalLow": 9.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -1,
            "SPN": 2573,
            "SPNLength": 14272,
            "StartBit": 1,
            "TransmissionRate": "Per the Parameter Group to be transferred",
            "Units": "Request Dependent"
        },
        2574: {
            "Acronym": "RQST2",
            "DataRange": "0 to 16,777,215",
            "EndBit": 23,
            "Name": "Parameter Group Number (RQST2)",
            "Offset": 0.0,
            "OperationalHigh": 16777215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "8 bytes (multipacket  supported)",
            "Resolution": -3,
            "SPN": 2574,
            "SPNLength": 24,
            "StartBit": 0,
            "TransmissionRate": "Per user requirements , generally recommended that requests occur \nno more than 2 or 3 times per second.\nA NACK is required if the PG being asked for with a destination specifc address and it is not supported, see PGN 59392",
            "Units": "binary"
        },
        2575: {
            "Acronym": "RQST2",
            "DataRange": "0 to 3",
            "EndBit": 25,
            "Name": "Use Transfer Mode",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "8 bytes (multipacket  supported)",
            "Resolution": 1,
            "SPN": 2575,
            "SPNLength": 2,
            "StartBit": 24,
            "TransmissionRate": "Per user requirements , generally recommended that requests occur \nno more than 2 or 3 times per second.\nA NACK is required if the PG being asked for with a destination specifc address and it is not supported, see PGN 59392",
            "Units": "bit"
        },
        2576: {
            "Acronym": "LVD",
            "DataRange": "0 to 250",
            "EndBit": 23,
            "Name": "Laser Receiver Type",
            "Offset": 0.0,
            "OperationalHigh": 4.0,
            "OperationalLow": 1.0,
            "OperationalRange": "1 to 4",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2576,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "50 ms",
            "Units": "count"
        },
        2577: {
            "Acronym": "LVDD",
            "DataRange": "0 to 15",
            "EndBit": 23,
            "Name": "Display Deadbands",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2577,
            "SPNLength": 4,
            "StartBit": 20,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2578: {
            "Acronym": "LVDD",
            "DataRange": "0 to 15",
            "EndBit": 19,
            "Name": "LED Pattern Control",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2578,
            "SPNLength": 4,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2579: {
            "Acronym": "VEP3",
            "DataRange": "-1600 to 1612.75 A",
            "EndBit": 31,
            "Name": "Net Battery Current (High Range/Resolution)",
            "Offset": -1600.0,
            "OperationalHigh": 1612.75,
            "OperationalLow": -1600.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05,
            "SPN": 2579,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "1 s",
            "Units": "Amps"
        },
        2580: {
            "Acronym": "HBS",
            "DataRange": "0.0 to 3625.9434499999998 psi",
            "EndBit": 7,
            "Name": "Hydraulic Brake Pressure Circuit 1",
            "Offset": 0.0,
            "OperationalHigh": 3625.9434499999998,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 14.5037738,
            "SPN": 2580,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        2581: {
            "Acronym": "HBS",
            "DataRange": "0.0 to 3625.9434499999998 psi",
            "EndBit": 15,
            "Name": "Hydraulic Brake Pressure Circuit 2",
            "Offset": 0.0,
            "OperationalHigh": 3625.9434499999998,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 14.5037738,
            "SPN": 2581,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        2582: {
            "Acronym": "HBS",
            "DataRange": "0 to 3",
            "EndBit": 21,
            "Name": "Hydraulic Brake Pressure Supply State Circuit 1",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2582,
            "SPNLength": 2,
            "StartBit": 20,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2583: {
            "Acronym": "HBS",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Hydraulic Brake Pressure Supply State Circuit 2",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2583,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2584: {
            "Acronym": "HBS",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Hydraulic Brake Pressure Warning State Circuit 1",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2584,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2585: {
            "Acronym": "HBS",
            "DataRange": "0 to 3",
            "EndBit": 19,
            "Name": "Hydraulic Brake Pressure Warning State Circuit 2",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2585,
            "SPNLength": 2,
            "StartBit": 18,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2586: {
            "Acronym": "TIRE",
            "DataRange": "0 to 6425.5 Pa/s",
            "EndBit": 55,
            "Name": "Tire Air Leakage Rate",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 2586,
            "SPNLength": 16,
            "StartBit": 40,
            "TransmissionRate": "10 s",
            "Units": "Pa/s"
        },
        2587: {
            "Acronym": "TIRE",
            "DataRange": "0 to 7",
            "EndBit": 63,
            "Name": "Tire Pressure Threshold Detection",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2587,
            "SPNLength": 3,
            "StartBit": 61,
            "TransmissionRate": "10 s",
            "Units": "bit"
        },
        2588: {
            "Acronym": "MVS",
            "DataRange": "0.0 to 155.34275 mph",
            "EndBit": 7,
            "Name": "Maximum Vehicle Speed Limit 1",
            "Offset": 0.0,
            "OperationalHigh": 155.34275,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.621371,
            "SPN": 2588,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "mph"
        },
        2589: {
            "Acronym": "MVS",
            "DataRange": "0.0 to 155.34275 mph",
            "EndBit": 15,
            "Name": "Maximum Vehicle Speed Limit 2",
            "Offset": 0.0,
            "OperationalHigh": 155.34275,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.621371,
            "SPN": 2589,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "1 s",
            "Units": "mph"
        },
        2590: {
            "Acronym": "MVS",
            "DataRange": "0.0 to 155.34275 mph",
            "EndBit": 23,
            "Name": "Maximum Vehicle Speed Limit 3",
            "Offset": 0.0,
            "OperationalHigh": 155.34275,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.621371,
            "SPN": 2590,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "1 s",
            "Units": "mph"
        },
        2591: {
            "Acronym": "MVS",
            "DataRange": "0.0 to 155.34275 mph",
            "EndBit": 31,
            "Name": "Maximum Vehicle Speed Limit 4",
            "Offset": 0.0,
            "OperationalHigh": 155.34275,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.621371,
            "SPN": 2591,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "1 s",
            "Units": "mph"
        },
        2592: {
            "Acronym": "MVS",
            "DataRange": "0.0 to 155.34275 mph",
            "EndBit": 39,
            "Name": "Maximum Vehicle Speed Limit 5",
            "Offset": 0.0,
            "OperationalHigh": 155.34275,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.621371,
            "SPN": 2592,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "1 s",
            "Units": "mph"
        },
        2593: {
            "Acronym": "MVS",
            "DataRange": "0.0 to 155.34275 mph",
            "EndBit": 47,
            "Name": "Maximum Vehicle Speed Limit 6",
            "Offset": 0.0,
            "OperationalHigh": 155.34275,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.621371,
            "SPN": 2593,
            "SPNLength": 8,
            "StartBit": 40,
            "TransmissionRate": "1 s",
            "Units": "mph"
        },
        2594: {
            "Acronym": "MVS",
            "DataRange": "0.0 to 155.34275 mph",
            "EndBit": 55,
            "Name": "Maximum Vehicle Speed Limit 7",
            "Offset": 0.0,
            "OperationalHigh": 155.34275,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.621371,
            "SPN": 2594,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "1 s",
            "Units": "mph"
        },
        2595: {
            "Acronym": "MVS",
            "DataRange": "0.0 to 155.34275 mph",
            "EndBit": 63,
            "Name": "Applied Vehicle Speed Limit",
            "Offset": 0.0,
            "OperationalHigh": 155.34275,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.621371,
            "SPN": 2595,
            "SPNLength": 8,
            "StartBit": 56,
            "TransmissionRate": "1 s",
            "Units": "mph"
        },
        2596: {
            "Acronym": "CM1",
            "DataRange": "0 to 250",
            "EndBit": 63,
            "Name": "Selected Maximum Vehicle Speed Limit",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 7",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2596,
            "SPNLength": 8,
            "StartBit": 56,
            "TransmissionRate": "1 s",
            "Units": "count"
        },
        2597: {
            "Acronym": "LCMD",
            "DataRange": "0 to 3",
            "EndBit": 55,
            "Name": "Implement Left Forward Work Light Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2597,
            "SPNLength": 2,
            "StartBit": 54,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms",
            "Units": "bit"
        },
        2598: {
            "Acronym": "LD",
            "DataRange": "0 to 3",
            "EndBit": 55,
            "Name": "Implement Left Forward Work Light",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2598,
            "SPNLength": 2,
            "StartBit": 54,
            "TransmissionRate": "As requested.",
            "Units": "bit"
        },
        2599: {
            "Acronym": "HPG",
            "DataRange": "0 to 3",
            "EndBit": 21,
            "Name": "Fire Apparatus Pump Engagement",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2599,
            "SPNLength": 2,
            "StartBit": 20,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        2600: {
            "Acronym": "EPD",
            "DataRange": "0 to 250 %",
            "EndBit": 7,
            "Name": "Payload Percentage",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2600,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "500 mS",
            "Units": "%"
        },
        2601: {
            "Acronym": "EOAC",
            "DataRange": "-125 to 125 %",
            "EndBit": 7,
            "Name": "Travel Velocity Control Position",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2601,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "250 mS",
            "Units": "%"
        },
        2602: {
            "Acronym": "VF",
            "DataRange": "0 to 100 %",
            "EndBit": 23,
            "Name": "Hydraulic Oil Level",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 2602,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "1 s",
            "Units": "%"
        },
        2603: {
            "Acronym": "SPR",
            "DataRange": "0.0 to 290.075476 psi",
            "EndBit": 7,
            "Name": "Pneumatic Supply Pressure Request",
            "Offset": 0.0,
            "OperationalHigh": 290.075476,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.160301904,
            "SPN": 2603,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "1 s, when active",
            "Units": "psi"
        },
        2604: {
            "Acronym": "SPR",
            "DataRange": "0.0 to 290.075476 psi",
            "EndBit": 15,
            "Name": "Parking and/or Trailer Air Pressure Request",
            "Offset": 0.0,
            "OperationalHigh": 290.075476,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.160301904,
            "SPN": 2604,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "1 s, when active",
            "Units": "psi"
        },
        2605: {
            "Acronym": "SPR",
            "DataRange": "0.0 to 290.075476 psi",
            "EndBit": 23,
            "Name": "Service Brake Air Pressure Request, Circuit #1",
            "Offset": 0.0,
            "OperationalHigh": 290.075476,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.160301904,
            "SPN": 2605,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "1 s, when active",
            "Units": "psi"
        },
        2606: {
            "Acronym": "SPR",
            "DataRange": "0.0 to 290.075476 psi",
            "EndBit": 31,
            "Name": "Service Brake Air Pressure Request, Circuit #2",
            "Offset": 0.0,
            "OperationalHigh": 290.075476,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.160301904,
            "SPN": 2606,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "1 s, when active",
            "Units": "psi"
        },
        2607: {
            "Acronym": "SPR",
            "DataRange": "0.0 to 290.075476 psi",
            "EndBit": 39,
            "Name": "Auxiliary Equipment Supply Pressure Request",
            "Offset": 0.0,
            "OperationalHigh": 290.075476,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.160301904,
            "SPN": 2607,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "1 s, when active",
            "Units": "psi"
        },
        2608: {
            "Acronym": "SPR",
            "DataRange": "0.0 to 290.075476 psi",
            "EndBit": 47,
            "Name": "Air Suspension Supply Pressure Request",
            "Offset": 0.0,
            "OperationalHigh": 290.075476,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.160301904,
            "SPN": 2608,
            "SPNLength": 8,
            "StartBit": 40,
            "TransmissionRate": "1 s, when active",
            "Units": "psi"
        },
        2609: {
            "Acronym": "CACI",
            "DataRange": "0.0 to 580.150952 psi",
            "EndBit": 7,
            "Name": "Cab A/C Refrigerant Compressor Outlet Pressure",
            "Offset": 0.0,
            "OperationalHigh": 580.150952,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 2.320603808,
            "SPN": 2609,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "psi"
        },
        2610: {
            "Acronym": "AMB2",
            "DataRange": "0 to 100 %",
            "EndBit": 7,
            "Name": "Solar Intensity Percent",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 2610,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "%"
        },
        2611: {
            "Acronym": "AMB2",
            "DataRange": "0 to 100 mW/cm^2",
            "EndBit": 15,
            "Name": "Solar Sensor Maximum",
            "Offset": 0.0,
            "OperationalHigh": 1002.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 2611,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "1 s",
            "Units": "mW/cm^2"
        },
        2612: {
            "Acronym": "FWD",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Front Wheel Drive Actuator Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2612,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "0.5 s",
            "Units": "bit"
        },
        2613: {
            "Acronym": "AI",
            "DataRange": "0.0 to 145.037738 psi",
            "EndBit": 39,
            "Name": "Drive Axle Lube Pressure",
            "Offset": 0.0,
            "OperationalHigh": 145.037738,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.580150952,
            "SPN": 2613,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "1 s",
            "Units": "psi"
        },
        2614: {
            "Acronym": "AI",
            "DataRange": "0.0 to 145.037738 psi",
            "EndBit": 63,
            "Name": "Steering Axle Lube Pressure",
            "Offset": 0.0,
            "OperationalHigh": 145.037738,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.580150952,
            "SPN": 2614,
            "SPNLength": 8,
            "StartBit": 56,
            "TransmissionRate": "1 s",
            "Units": "psi"
        },
        2615: {
            "Acronym": "MCI",
            "DataRange": "0 to 15",
            "EndBit": 3,
            "Name": "Engine Throttle Synchronization Mode Status",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2615,
            "SPNLength": 4,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        2616: {
            "Acronym": "MCI",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Trolling Mode Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2616,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        2617: {
            "Acronym": "MCI",
            "DataRange": "0 to 3",
            "EndBit": 7,
            "Name": "Slow Vessel Mode Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2617,
            "SPNLength": 2,
            "StartBit": 6,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        2618: {
            "Acronym": "DM13",
            "DataRange": "0 to 15",
            "EndBit": 27,
            "Name": "Suspend Signal",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2618,
            "SPNLength": 4,
            "StartBit": 24,
            "TransmissionRate": "Sent when ever a Stop or Start broadcast event is necessary. To maintain the modified state of the vehicle network(s) the commanding device must send the Hold Signal once every 5 seconds.",
            "Units": "bit"
        },
        2619: {
            "Acronym": "DM13",
            "DataRange": "0 to 64,255 s",
            "EndBit": 47,
            "Name": "Suspend Duration",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2619,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "Sent when ever a Stop or Start broadcast event is necessary. To maintain the modified state of the vehicle network(s) the commanding device must send the Hold Signal once every 5 seconds.",
            "Units": "Seconds"
        },
        2623: {
            "Acronym": "EEC15",
            "DataRange": "0 to 100 %",
            "EndBit": 7,
            "Name": "Accelerator Pedal #1 Channel 2",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 2623,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "On Request",
            "Units": "%"
        },
        2624: {
            "Acronym": "EEC15",
            "DataRange": "0 to 100 %",
            "EndBit": 15,
            "Name": "Accelerator Pedal #1 Channel 3",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 2624,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "On Request",
            "Units": "%"
        },
        2625: {
            "Acronym": "EEC15",
            "DataRange": "0 to 100 %",
            "EndBit": 23,
            "Name": "Accelerator Pedal #2 Channel 2",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 2625,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "On Request",
            "Units": "%"
        },
        2626: {
            "Acronym": "EEC15",
            "DataRange": "0 to 100 %",
            "EndBit": 31,
            "Name": "Accelerator Pedal #2 Channel 3",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 2626,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "On Request",
            "Units": "%"
        },
        2629: {
            "Acronym": "TCI6",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 15,
            "Name": "Engine Turbocharger 1 Compressor Outlet Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 2629,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        2630: {
            "Acronym": "ET3",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 63,
            "Name": "Engine Charge Air Cooler 1 Outlet Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 2630,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        2631: {
            "Acronym": "EFL/P4",
            "DataRange": "0.0 to 72.518869 psi",
            "EndBit": 63,
            "Name": "Engine Charge Air Cooler Outlet Pressure",
            "Offset": 0.0,
            "OperationalHigh": 72.518869,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.290075476,
            "SPN": 2631,
            "SPNLength": 8,
            "StartBit": 56,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        2659: {
            "Acronym": "EGF1",
            "DataRange": "0 to 3212.75 kg/h",
            "EndBit": 15,
            "Name": "Engine Exhaust Gas Recirculation 1 (EGR1) Mass Flow Rate",
            "Offset": 0.0,
            "OperationalHigh": 3212.75,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05,
            "SPN": 2659,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "50 ms",
            "Units": "kg/h"
        },
        2660: {
            "Acronym": "BJM1",
            "DataRange": "0 to 102 %",
            "EndBit": 15,
            "Name": "Joystick 1 X-Axis Position",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0.0 to 100.0%",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 2660,
            "SPNLength": 10,
            "StartBit": 6,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "%"
        },
        2661: {
            "Acronym": "BJM1",
            "DataRange": "0 to 102 %",
            "EndBit": 31,
            "Name": "Joystick 1 Y-Axis Position",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0.0 to 100.0%",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 2661,
            "SPNLength": 10,
            "StartBit": 22,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "%"
        },
        2662: {
            "Acronym": "EJM1",
            "DataRange": "0 to 102 %",
            "EndBit": 15,
            "Name": "Joystick 1 Grip X-Axis Position",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0.0 to 100.0%",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 2662,
            "SPNLength": 10,
            "StartBit": 6,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "%"
        },
        2663: {
            "Acronym": "EJM1",
            "DataRange": "0 to 102 %",
            "EndBit": 31,
            "Name": "Joystick 1 Grip Y-Axis Position",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0.0 to 100.0%",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 2663,
            "SPNLength": 10,
            "StartBit": 22,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "%"
        },
        2664: {
            "Acronym": "EJM1",
            "DataRange": "0 to 102 %",
            "EndBit": 47,
            "Name": "Joystick 1 Theta-Axis Position",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0.0 to 100.0%",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 2664,
            "SPNLength": 10,
            "StartBit": 38,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "%"
        },
        2665: {
            "Acronym": "BJM1",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Joystick 1 X-Axis Lever Right Positive Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2665,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2666: {
            "Acronym": "BJM1",
            "DataRange": "0 to 3",
            "EndBit": 21,
            "Name": "Joystick 1 Y-Axis Lever Forward Positive Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2666,
            "SPNLength": 2,
            "StartBit": 20,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2667: {
            "Acronym": "EJM1",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Joystick 1 Grip X-Axis Lever Right Positive Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2667,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2668: {
            "Acronym": "EJM1",
            "DataRange": "0 to 3",
            "EndBit": 21,
            "Name": "Joystick 1 Grip Y-Axis Lever Forward Positive Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2668,
            "SPNLength": 2,
            "StartBit": 20,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2669: {
            "Acronym": "EJM1",
            "DataRange": "0 to 3",
            "EndBit": 37,
            "Name": "Joystick 1 Theta-Axis Clockwise Positive Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2669,
            "SPNLength": 2,
            "StartBit": 36,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2670: {
            "Acronym": "BJM1",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Joystick 1 X-Axis Lever Left Negative Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2670,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2671: {
            "Acronym": "BJM1",
            "DataRange": "0 to 3",
            "EndBit": 19,
            "Name": "Joystick 1 Y-Axis Lever Back Negative Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2671,
            "SPNLength": 2,
            "StartBit": 18,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2672: {
            "Acronym": "EJM1",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Joystick 1 Grip X-Axis Lever Left Negative Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2672,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2673: {
            "Acronym": "EJM1",
            "DataRange": "0 to 3",
            "EndBit": 19,
            "Name": "Joystick 1 Grip Y-Axis Lever Back Negative Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2673,
            "SPNLength": 2,
            "StartBit": 18,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2674: {
            "Acronym": "EJM1",
            "DataRange": "0 to 3",
            "EndBit": 35,
            "Name": "Joystick 1 Theta-Axis Counter Clockwise Negative Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2674,
            "SPNLength": 2,
            "StartBit": 34,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2675: {
            "Acronym": "BJM1",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Joystick 1 X-Axis Neutral Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2675,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2676: {
            "Acronym": "BJM1",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Joystick 1 Y-Axis Neutral Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2676,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2677: {
            "Acronym": "EJM1",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Joystick 1 Grip X-Axis Neutral Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2677,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2678: {
            "Acronym": "EJM1",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Joystick 1 Grip Y-Axis Neutral Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2678,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2679: {
            "Acronym": "EJM1",
            "DataRange": "0 to 3",
            "EndBit": 33,
            "Name": "Joystick 1 Theta-Axis Neutral Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2679,
            "SPNLength": 2,
            "StartBit": 32,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2680: {
            "Acronym": "BJM1",
            "DataRange": "0 to 3",
            "EndBit": 39,
            "Name": "Joystick 1 X-Axis Detent Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2680,
            "SPNLength": 2,
            "StartBit": 38,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2681: {
            "Acronym": "BJM1",
            "DataRange": "0 to 3",
            "EndBit": 37,
            "Name": "Joystick 1 Y-Axis Detent Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2681,
            "SPNLength": 2,
            "StartBit": 36,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2682: {
            "Acronym": "EJM1",
            "DataRange": "0 to 3",
            "EndBit": 55,
            "Name": "Joystick 1 Grip X-Axis Detent Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2682,
            "SPNLength": 2,
            "StartBit": 54,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2683: {
            "Acronym": "EJM1",
            "DataRange": "0 to 3",
            "EndBit": 53,
            "Name": "Joystick 1 Grip Y-Axis Detent Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2683,
            "SPNLength": 2,
            "StartBit": 52,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2684: {
            "Acronym": "EJM1",
            "DataRange": "0 to 3",
            "EndBit": 51,
            "Name": "Joystick 1 Theta-Axis Detent Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2684,
            "SPNLength": 2,
            "StartBit": 50,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2685: {
            "Acronym": "BJM1",
            "DataRange": "0 to 3",
            "EndBit": 47,
            "Name": "Joystick 1 Button 1 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2685,
            "SPNLength": 2,
            "StartBit": 46,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2686: {
            "Acronym": "BJM1",
            "DataRange": "0 to 3",
            "EndBit": 45,
            "Name": "Joystick 1 Button 2 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2686,
            "SPNLength": 2,
            "StartBit": 44,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2687: {
            "Acronym": "BJM1",
            "DataRange": "0 to 3",
            "EndBit": 43,
            "Name": "Joystick 1 Button 3 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2687,
            "SPNLength": 2,
            "StartBit": 42,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2688: {
            "Acronym": "BJM1",
            "DataRange": "0 to 3",
            "EndBit": 41,
            "Name": "Joystick 1 Button 4 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2688,
            "SPNLength": 2,
            "StartBit": 40,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2689: {
            "Acronym": "BJM1",
            "DataRange": "0 to 3",
            "EndBit": 55,
            "Name": "Joystick 1 Button 5 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2689,
            "SPNLength": 2,
            "StartBit": 54,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2690: {
            "Acronym": "BJM1",
            "DataRange": "0 to 3",
            "EndBit": 53,
            "Name": "Joystick 1 Button 6 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2690,
            "SPNLength": 2,
            "StartBit": 52,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2691: {
            "Acronym": "BJM1",
            "DataRange": "0 to 3",
            "EndBit": 51,
            "Name": "Joystick 1 Button 7 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2691,
            "SPNLength": 2,
            "StartBit": 50,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2692: {
            "Acronym": "BJM1",
            "DataRange": "0 to 3",
            "EndBit": 49,
            "Name": "Joystick 1 Button 8 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2692,
            "SPNLength": 2,
            "StartBit": 48,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2693: {
            "Acronym": "BJM1",
            "DataRange": "0 to 3",
            "EndBit": 63,
            "Name": "Joystick 1 Button 9 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2693,
            "SPNLength": 2,
            "StartBit": 62,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2694: {
            "Acronym": "BJM1",
            "DataRange": "0 to 3",
            "EndBit": 61,
            "Name": "Joystick 1 Button 10 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2694,
            "SPNLength": 2,
            "StartBit": 60,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2695: {
            "Acronym": "BJM1",
            "DataRange": "0 to 3",
            "EndBit": 59,
            "Name": "Joystick 1 Button 11 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2695,
            "SPNLength": 2,
            "StartBit": 58,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2696: {
            "Acronym": "BJM1",
            "DataRange": "0 to 3",
            "EndBit": 57,
            "Name": "Joystick 1 Button 12 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2696,
            "SPNLength": 2,
            "StartBit": 56,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2697: {
            "Acronym": "BJM2",
            "DataRange": "0 to 102 %",
            "EndBit": 15,
            "Name": "Joystick 2 X-Axis Position",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0.0 to 100.0%",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 2697,
            "SPNLength": 10,
            "StartBit": 6,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "%"
        },
        2698: {
            "Acronym": "BJM2",
            "DataRange": "0 to 102 %",
            "EndBit": 31,
            "Name": "Joystick 2 Y-Axis Position",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0.0 to 100.0%",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 2698,
            "SPNLength": 10,
            "StartBit": 22,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "%"
        },
        2699: {
            "Acronym": "EJM2",
            "DataRange": "0 to 102 %",
            "EndBit": 15,
            "Name": "Joystick 2 Grip X-Axis Position",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0.0 to 100.0%",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 2699,
            "SPNLength": 10,
            "StartBit": 6,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "%"
        },
        27: {
            "Acronym": "EEC7",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 15,
            "Name": "Engine Exhaust Gas Recirculation 1 Valve Position",
            "Offset": 0.0,
            "OperationalHigh": 160.6375,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 27,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        2700: {
            "Acronym": "EJM2",
            "DataRange": "0 to 102 %",
            "EndBit": 31,
            "Name": "Joystick 2 Grip Y-Axis Position",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0.0 to 100.0%",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 2700,
            "SPNLength": 10,
            "StartBit": 22,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "%"
        },
        2701: {
            "Acronym": "EJM2",
            "DataRange": "0 to 102 %",
            "EndBit": 47,
            "Name": "Joystick 2 Theta-Axis Position",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0.0 to 100.0%",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 2701,
            "SPNLength": 10,
            "StartBit": 38,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "%"
        },
        2702: {
            "Acronym": "BJM2",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Joystick 2 X-Axis Lever Right Positive Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2702,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2703: {
            "Acronym": "BJM2",
            "DataRange": "0 to 3",
            "EndBit": 21,
            "Name": "Joystick 2 Y-Axis Lever Forward Positive Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2703,
            "SPNLength": 2,
            "StartBit": 20,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2704: {
            "Acronym": "EJM2",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Joystick 2 Grip X-Axis Lever Right Positive Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2704,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2705: {
            "Acronym": "EJM2",
            "DataRange": "0 to 3",
            "EndBit": 21,
            "Name": "Joystick 2 Grip Y-Axis Lever Forward Positive Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2705,
            "SPNLength": 2,
            "StartBit": 20,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2706: {
            "Acronym": "EJM2",
            "DataRange": "0 to 3",
            "EndBit": 37,
            "Name": "Joystick 2 Theta-Axis Clockwise Positive Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2706,
            "SPNLength": 2,
            "StartBit": 36,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2707: {
            "Acronym": "BJM2",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Joystick 2 X-Axis Lever Left Negative Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2707,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2708: {
            "Acronym": "BJM2",
            "DataRange": "0 to 3",
            "EndBit": 19,
            "Name": "Joystick 2 Y-Axis Lever Back Negative Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2708,
            "SPNLength": 2,
            "StartBit": 18,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2709: {
            "Acronym": "EJM2",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Joystick 2 Grip X-Axis Lever Left Negative Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2709,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2710: {
            "Acronym": "EJM2",
            "DataRange": "0 to 3",
            "EndBit": 19,
            "Name": "Joystick 2 Grip Y-Axis Lever Back Negative Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2710,
            "SPNLength": 2,
            "StartBit": 18,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2711: {
            "Acronym": "EJM2",
            "DataRange": "0 to 3",
            "EndBit": 35,
            "Name": "Joystick 2 Theta-Axis Counter Clockwise Negative Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2711,
            "SPNLength": 2,
            "StartBit": 34,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2712: {
            "Acronym": "BJM2",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Joystick 2 X-Axis Neutral Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2712,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2713: {
            "Acronym": "BJM2",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Joystick 2 Y-Axis Neutral Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2713,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2714: {
            "Acronym": "EJM2",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Joystick 2 Grip X-Axis Neutral Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2714,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2715: {
            "Acronym": "EJM2",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Joystick 2 Grip Y-Axis Neutral Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2715,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2716: {
            "Acronym": "EJM2",
            "DataRange": "0 to 3",
            "EndBit": 33,
            "Name": "Joystick 2 Theta-Axis Neutral Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2716,
            "SPNLength": 2,
            "StartBit": 32,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2717: {
            "Acronym": "BJM2",
            "DataRange": "0 to 3",
            "EndBit": 39,
            "Name": "Joystick 2 X-Axis Detent Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2717,
            "SPNLength": 2,
            "StartBit": 38,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2718: {
            "Acronym": "BJM2",
            "DataRange": "0 to 3",
            "EndBit": 37,
            "Name": "Joystick 2 Y-Axis Detent Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2718,
            "SPNLength": 2,
            "StartBit": 36,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2719: {
            "Acronym": "EJM2",
            "DataRange": "0 to 3",
            "EndBit": 55,
            "Name": "Joystick 2 Grip X-Axis Detent Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2719,
            "SPNLength": 2,
            "StartBit": 54,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2720: {
            "Acronym": "EJM2",
            "DataRange": "0 to 3",
            "EndBit": 53,
            "Name": "Joystick 2 Grip Y-Axis Detent Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2720,
            "SPNLength": 2,
            "StartBit": 52,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2721: {
            "Acronym": "EJM2",
            "DataRange": "0 to 3",
            "EndBit": 51,
            "Name": "Joystick 2 Theta-Axis Detent Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2721,
            "SPNLength": 2,
            "StartBit": 50,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2722: {
            "Acronym": "BJM2",
            "DataRange": "0 to 3",
            "EndBit": 47,
            "Name": "Joystick 2 Button 1 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2722,
            "SPNLength": 2,
            "StartBit": 46,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2723: {
            "Acronym": "BJM2",
            "DataRange": "0 to 3",
            "EndBit": 45,
            "Name": "Joystick 2 Button 2 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2723,
            "SPNLength": 2,
            "StartBit": 44,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2724: {
            "Acronym": "BJM2",
            "DataRange": "0 to 3",
            "EndBit": 43,
            "Name": "Joystick 2 Button 3 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2724,
            "SPNLength": 2,
            "StartBit": 42,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2725: {
            "Acronym": "BJM2",
            "DataRange": "0 to 3",
            "EndBit": 41,
            "Name": "Joystick 2 Button 4 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2725,
            "SPNLength": 2,
            "StartBit": 40,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2726: {
            "Acronym": "BJM2",
            "DataRange": "0 to 3",
            "EndBit": 55,
            "Name": "Joystick 2 Button 5 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2726,
            "SPNLength": 2,
            "StartBit": 54,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2727: {
            "Acronym": "BJM2",
            "DataRange": "0 to 3",
            "EndBit": 53,
            "Name": "Joystick 2 Button 6 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2727,
            "SPNLength": 2,
            "StartBit": 52,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2728: {
            "Acronym": "BJM2",
            "DataRange": "0 to 3",
            "EndBit": 51,
            "Name": "Joystick 2 Button 7 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2728,
            "SPNLength": 2,
            "StartBit": 50,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2729: {
            "Acronym": "BJM2",
            "DataRange": "0 to 3",
            "EndBit": 49,
            "Name": "Joystick 2 Button 8 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2729,
            "SPNLength": 2,
            "StartBit": 48,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2730: {
            "Acronym": "BJM2",
            "DataRange": "0 to 3",
            "EndBit": 63,
            "Name": "Joystick 2 Button 9 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2730,
            "SPNLength": 2,
            "StartBit": 62,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2731: {
            "Acronym": "BJM2",
            "DataRange": "0 to 3",
            "EndBit": 61,
            "Name": "Joystick 2 Button 10 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2731,
            "SPNLength": 2,
            "StartBit": 60,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2732: {
            "Acronym": "BJM2",
            "DataRange": "0 to 3",
            "EndBit": 59,
            "Name": "Joystick 2 Button 11 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2732,
            "SPNLength": 2,
            "StartBit": 58,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2733: {
            "Acronym": "BJM2",
            "DataRange": "0 to 3",
            "EndBit": 57,
            "Name": "Joystick 2 Button 12 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2733,
            "SPNLength": 2,
            "StartBit": 56,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2734: {
            "Acronym": "BJM3",
            "DataRange": "0 to 102 %",
            "EndBit": 15,
            "Name": "Joystick 3 X-Axis Position",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0.0 to 100.0%",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 2734,
            "SPNLength": 10,
            "StartBit": 6,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "%"
        },
        2735: {
            "Acronym": "BJM3",
            "DataRange": "0 to 102 %",
            "EndBit": 31,
            "Name": "Joystick 3 Y-Axis Position",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0.0 to 100.0%",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 2735,
            "SPNLength": 10,
            "StartBit": 22,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "%"
        },
        2736: {
            "Acronym": "EJM3",
            "DataRange": "0 to 102 %",
            "EndBit": 15,
            "Name": "Joystick 3 Grip X-Axis Position",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0.0 to 100.0%",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 2736,
            "SPNLength": 10,
            "StartBit": 6,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "%"
        },
        2737: {
            "Acronym": "EJM3",
            "DataRange": "0 to 102 %",
            "EndBit": 31,
            "Name": "Joystick 3 Grip Y-Axis Position",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0.0 to 100.0%",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 2737,
            "SPNLength": 10,
            "StartBit": 22,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "%"
        },
        2738: {
            "Acronym": "EJM3",
            "DataRange": "0 to 102 %",
            "EndBit": 47,
            "Name": "Joystick 3 Theta-Axis Position",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0.0 to 100.0%",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 2738,
            "SPNLength": 10,
            "StartBit": 38,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "%"
        },
        2739: {
            "Acronym": "BJM3",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Joystick 3 X-Axis Lever Right Positive Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2739,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2740: {
            "Acronym": "BJM3",
            "DataRange": "0 to 3",
            "EndBit": 21,
            "Name": "Joystick 3 Y-Axis Lever Forward Positive Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2740,
            "SPNLength": 2,
            "StartBit": 20,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2741: {
            "Acronym": "EJM3",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Joystick 3 Grip X-Axis Lever Right Positive Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2741,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2742: {
            "Acronym": "EJM3",
            "DataRange": "0 to 3",
            "EndBit": 21,
            "Name": "Joystick 3 Grip Y-Axis Lever Forward Positive Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2742,
            "SPNLength": 2,
            "StartBit": 20,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2743: {
            "Acronym": "EJM3",
            "DataRange": "0 to 3",
            "EndBit": 37,
            "Name": "Joystick 3 Theta-Axis Clockwise Positive Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2743,
            "SPNLength": 2,
            "StartBit": 36,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2744: {
            "Acronym": "BJM3",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Joystick 3 X-Axis Lever Left Negative Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2744,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2745: {
            "Acronym": "BJM3",
            "DataRange": "0 to 3",
            "EndBit": 19,
            "Name": "Joystick 3 Y-Axis Lever Back Negative Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2745,
            "SPNLength": 2,
            "StartBit": 18,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2746: {
            "Acronym": "EJM3",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Joystick 3 Grip X-Axis Lever Left Negative Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2746,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2747: {
            "Acronym": "EJM3",
            "DataRange": "0 to 3",
            "EndBit": 19,
            "Name": "Joystick 3 Grip Y-Axis Lever Back Negative Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2747,
            "SPNLength": 2,
            "StartBit": 18,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2748: {
            "Acronym": "EJM3",
            "DataRange": "0 to 3",
            "EndBit": 35,
            "Name": "Joystick 3 Theta-Axis Counter Clockwise Negative Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2748,
            "SPNLength": 2,
            "StartBit": 34,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2749: {
            "Acronym": "BJM3",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Joystick 3 X-Axis Neutral Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2749,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2750: {
            "Acronym": "BJM3",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Joystick 3 Y-Axis Neutral Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2750,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2751: {
            "Acronym": "EJM3",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Joystick 3 Grip X-Axis Neutral Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2751,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2752: {
            "Acronym": "EJM3",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Joystick 3 Grip Y-Axis Neutral Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2752,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2753: {
            "Acronym": "EJM3",
            "DataRange": "0 to 3",
            "EndBit": 33,
            "Name": "Joystick 3 Theta-Axis Neutral Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2753,
            "SPNLength": 2,
            "StartBit": 32,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2754: {
            "Acronym": "BJM3",
            "DataRange": "0 to 3",
            "EndBit": 39,
            "Name": "Joystick 3 X-Axis Detent Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2754,
            "SPNLength": 2,
            "StartBit": 38,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2755: {
            "Acronym": "BJM3",
            "DataRange": "0 to 3",
            "EndBit": 37,
            "Name": "Joystick 3 Y-Axis Detent Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2755,
            "SPNLength": 2,
            "StartBit": 36,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2756: {
            "Acronym": "EJM3",
            "DataRange": "0 to 3",
            "EndBit": 55,
            "Name": "Joystick 3 Grip X-Axis Detent Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2756,
            "SPNLength": 2,
            "StartBit": 54,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2757: {
            "Acronym": "EJM3",
            "DataRange": "0 to 3",
            "EndBit": 53,
            "Name": "Joystick 3 Grip Y-Axis Detent Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2757,
            "SPNLength": 2,
            "StartBit": 52,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2758: {
            "Acronym": "EJM3",
            "DataRange": "0 to 3",
            "EndBit": 51,
            "Name": "Joystick 3 Theta-Axis Detent Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2758,
            "SPNLength": 2,
            "StartBit": 50,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2759: {
            "Acronym": "BJM3",
            "DataRange": "0 to 3",
            "EndBit": 47,
            "Name": "Joystick 3 Button 1 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2759,
            "SPNLength": 2,
            "StartBit": 46,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2760: {
            "Acronym": "BJM3",
            "DataRange": "0 to 3",
            "EndBit": 45,
            "Name": "Joystick 3 Button 2 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2760,
            "SPNLength": 2,
            "StartBit": 44,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2761: {
            "Acronym": "BJM3",
            "DataRange": "0 to 3",
            "EndBit": 43,
            "Name": "Joystick 3 Button 3 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2761,
            "SPNLength": 2,
            "StartBit": 42,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2762: {
            "Acronym": "BJM3",
            "DataRange": "0 to 3",
            "EndBit": 41,
            "Name": "Joystick 3 Button 4 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2762,
            "SPNLength": 2,
            "StartBit": 40,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2763: {
            "Acronym": "BJM3",
            "DataRange": "0 to 3",
            "EndBit": 55,
            "Name": "Joystick 3 Button 5 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2763,
            "SPNLength": 2,
            "StartBit": 54,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2764: {
            "Acronym": "BJM3",
            "DataRange": "0 to 3",
            "EndBit": 53,
            "Name": "Joystick 3 Button 6 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2764,
            "SPNLength": 2,
            "StartBit": 52,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2765: {
            "Acronym": "BJM3",
            "DataRange": "0 to 3",
            "EndBit": 51,
            "Name": "Joystick 3 Button 7 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2765,
            "SPNLength": 2,
            "StartBit": 50,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2766: {
            "Acronym": "BJM3",
            "DataRange": "0 to 3",
            "EndBit": 49,
            "Name": "Joystick 3 Button 8 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2766,
            "SPNLength": 2,
            "StartBit": 48,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2767: {
            "Acronym": "BJM3",
            "DataRange": "0 to 3",
            "EndBit": 63,
            "Name": "Joystick 3 Button 9 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2767,
            "SPNLength": 2,
            "StartBit": 62,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2768: {
            "Acronym": "BJM3",
            "DataRange": "0 to 3",
            "EndBit": 61,
            "Name": "Joystick 3 Button 10 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2768,
            "SPNLength": 2,
            "StartBit": 60,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2769: {
            "Acronym": "BJM3",
            "DataRange": "0 to 3",
            "EndBit": 59,
            "Name": "Joystick 3 Button 11 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2769,
            "SPNLength": 2,
            "StartBit": 58,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2770: {
            "Acronym": "BJM3",
            "DataRange": "0 to 3",
            "EndBit": 57,
            "Name": "Joystick 3 Button 12 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2770,
            "SPNLength": 2,
            "StartBit": 56,
            "TransmissionRate": "Every 100 ms and on change of state but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 100 ms or on change, not to exceed 20 ms",
            "Units": "bit"
        },
        2789: {
            "Acronym": "EEC5",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 15,
            "Name": "Engine Turbocharger 1 Calculated Turbine Intake Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 2789,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "deg F"
        },
        2790: {
            "Acronym": "EEC5",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 31,
            "Name": "Engine Turbocharger 1 Calculated Turbine Outlet Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 2790,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "deg F"
        },
        2791: {
            "Acronym": "EEC5",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 47,
            "Name": "Engine Exhaust Gas Recirculation 1 (EGR1) Valve Control",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 100%",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 2791,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        2792: {
            "Acronym": "EEC5",
            "DataRange": "0 to 3",
            "EndBit": 49,
            "Name": "Engine Variable Geometry Turbocharger (VGT) Air Control Shutoff Valve",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2792,
            "SPNLength": 2,
            "StartBit": 48,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        2793: {
            "Acronym": "LVD",
            "DataRange": "0 to 3.289856 s",
            "EndBit": 39,
            "Name": "Laser Strike Data Latency",
            "Offset": 0.0,
            "OperationalHigh": 3.289856,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 51.2,
            "SPN": 2793,
            "SPNLength": 16,
            "StartBit": 24,
            "TransmissionRate": "50 ms",
            "Units": "us"
        },
        2794: {
            "Acronym": "LVD",
            "DataRange": "0 to 6,425.5 mm (0 to 6.4255 m)",
            "EndBit": 55,
            "Name": "Absolute Laser Strike Position",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 2794,
            "SPNLength": 16,
            "StartBit": 40,
            "TransmissionRate": "50 ms",
            "Units": "m"
        },
        2795: {
            "Acronym": "EEC5",
            "DataRange": "0 to 100 %",
            "EndBit": 63,
            "Name": "Engine Variable Geometry Turbocharger (VGT) 1 Actuator Position",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 2795,
            "SPNLength": 8,
            "StartBit": 56,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        2796: {
            "Acronym": "CM3",
            "DataRange": "0 to 7",
            "EndBit": 2,
            "Name": "Transfer Case Selector Switch",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2796,
            "SPNLength": 3,
            "StartBit": 0,
            "TransmissionRate": "Every 10 s and on change of state but no faster than every 100 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: On change or every 10 s",
            "Units": "bit"
        },
        2799: {
            "Acronym": "TCI6",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 31,
            "Name": "Engine Turbocharger 2 Compressor Outlet Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 2799,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        2800: {
            "Acronym": "TCI6",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 47,
            "Name": "Engine Turbocharger 3 Compressor Outlet Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 2800,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        2801: {
            "Acronym": "TCI6",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 63,
            "Name": "Engine Turbocharger 4 Compressor Outlet Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 2801,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        2802: {
            "Acronym": "EP",
            "DataRange": "0 to 100 %",
            "EndBit": 23,
            "Name": "Data Memory Usage",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 2802,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        2803: {
            "Acronym": "EP",
            "DataRange": "0 to 64255mAhr (64.255Ahr)",
            "EndBit": 15,
            "Name": "Keep-Alive Battery Consumption",
            "Offset": 0.0,
            "OperationalHigh": 6425564.255,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2803,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "mAhr"
        },
        2804: {
            "Acronym": "FMS",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "FMS-standard Diagnostics Supported",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2804,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "10 s",
            "Units": "bit"
        },
        2805: {
            "Acronym": "FMS",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "FMS-standard Requests Supported",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2805,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "10 s",
            "Units": "bit"
        },
        2806: {
            "Acronym": "FMS",
            "DataRange": "0 to 255 per byte",
            "EndBit": 39,
            "Name": "FMS-standard SW-version supported.",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -2,
            "SPN": 2806,
            "SPNLength": 32,
            "StartBit": 8,
            "TransmissionRate": "10 s",
            "Units": "ASCII"
        },
        2807: {
            "Acronym": "EOI",
            "DataRange": "0 to 3",
            "EndBit": 29,
            "Name": "Engine Fuel Shutoff 2 Control",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2807,
            "SPNLength": 2,
            "StartBit": 28,
            "TransmissionRate": "250 ms",
            "Units": "bit"
        },
        2809: {
            "Acronym": "IC2",
            "DataRange": "0.0 to 1.812971725 psi",
            "EndBit": 7,
            "Name": "Engine Air Filter 2 Differential Pressure",
            "Offset": 0.0,
            "OperationalHigh": 1.812971725,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0072518869,
            "SPN": 2809,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "0.5 s",
            "Units": "psi"
        },
        2810: {
            "Acronym": "IC2",
            "DataRange": "0.0 to 1.812971725 psi",
            "EndBit": 15,
            "Name": "Engine Air Filter 3 Differential Pressure",
            "Offset": 0.0,
            "OperationalHigh": 1.812971725,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0072518869,
            "SPN": 2810,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "0.5 s",
            "Units": "psi"
        },
        2811: {
            "Acronym": "IC2",
            "DataRange": "0.0 to 1.812971725 psi",
            "EndBit": 23,
            "Name": "Engine Air Filter 4 Differential Pressure",
            "Offset": 0.0,
            "OperationalHigh": 1.812971725,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0072518869,
            "SPN": 2811,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "0.5 s",
            "Units": "psi"
        },
        2812: {
            "Acronym": "SHUTDN",
            "DataRange": "0 to 3",
            "EndBit": 55,
            "Name": "Engine Overspeed Test",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2812,
            "SPNLength": 2,
            "StartBit": 54,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        2813: {
            "Acronym": "SHUTDN",
            "DataRange": "0 to 3",
            "EndBit": 53,
            "Name": "Engine Air Shutoff Command Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2813,
            "SPNLength": 2,
            "StartBit": 52,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        2814: {
            "Acronym": "SHUTDN",
            "DataRange": "0 to 3",
            "EndBit": 51,
            "Name": "Engine Alarm Output Command Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2814,
            "SPNLength": 2,
            "StartBit": 50,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        2815: {
            "Acronym": "SHUTDN",
            "DataRange": "0 to 3",
            "EndBit": 49,
            "Name": "Engine Alarm Acknowledge",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2815,
            "SPNLength": 2,
            "StartBit": 48,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        2845: {
            "Acronym": "WSMM",
            "DataRange": "0 to 18446744073709551615",
            "EndBit": 63,
            "Name": "NAME of Working Set Member",
            "Offset": 0.0,
            "OperationalHigh": 1.8446744073709552e+19,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2845,
            "SPNLength": 64,
            "StartBit": 0,
            "TransmissionRate": "As required",
            "Units": "binary"
        },
        2847: {
            "Acronym": "CA",
            "DataRange": "0 to 255",
            "EndBit": 71,
            "Name": "Address Assignment (new source address)",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 9,
            "Resolution": 1,
            "SPN": 2847,
            "SPNLength": 8,
            "StartBit": 64,
            "TransmissionRate": "As required",
            "Units": "SA"
        },
        2848: {
            "Acronym": "AC",
            "DataRange": "0 to 18446744073709551615",
            "EndBit": 63,
            "Name": "NAME of Controller Application (for address claimed)",
            "Offset": 0.0,
            "OperationalHigh": 1.8446744073709552e+19,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2848,
            "SPNLength": 64,
            "StartBit": 0,
            "TransmissionRate": "As required",
            "Units": "binary"
        },
        2849: {
            "Acronym": "CA",
            "DataRange": "0 to 18446744073709551615",
            "EndBit": 63,
            "Name": "NAME of Commanded Address Target",
            "Offset": 0.0,
            "OperationalHigh": 1.8446744073709552e+19,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 9,
            "Resolution": 1,
            "SPN": 2849,
            "SPNLength": 64,
            "StartBit": 0,
            "TransmissionRate": "As required",
            "Units": "binary"
        },
        2863: {
            "Acronym": "OWW",
            "DataRange": "0 to 15",
            "EndBit": 7,
            "Name": "Front Operator Wiper Switch",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2863,
            "SPNLength": 4,
            "StartBit": 4,
            "TransmissionRate": "200 ms",
            "Units": "bit"
        },
        2864: {
            "Acronym": "OWW",
            "DataRange": "0 to 15",
            "EndBit": 3,
            "Name": "Front Non-operator Wiper Switch",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2864,
            "SPNLength": 4,
            "StartBit": 0,
            "TransmissionRate": "200 ms",
            "Units": "bit"
        },
        2865: {
            "Acronym": "OWW",
            "DataRange": "0 to 15",
            "EndBit": 15,
            "Name": "Rear Wiper Switch",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2865,
            "SPNLength": 4,
            "StartBit": 12,
            "TransmissionRate": "200 ms",
            "Units": "bit"
        },
        2866: {
            "Acronym": "OWW",
            "DataRange": "0 to 7",
            "EndBit": 47,
            "Name": "Front Operator Washer Switch",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2866,
            "SPNLength": 3,
            "StartBit": 45,
            "TransmissionRate": "200 ms",
            "Units": "bit"
        },
        2867: {
            "Acronym": "OWW",
            "DataRange": "0 to 7",
            "EndBit": 44,
            "Name": "Front Non-operator Washer Switch",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2867,
            "SPNLength": 3,
            "StartBit": 42,
            "TransmissionRate": "200 ms",
            "Units": "bit"
        },
        2868: {
            "Acronym": "OWW",
            "DataRange": "0 to 7",
            "EndBit": 55,
            "Name": "Rear Washer Function",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2868,
            "SPNLength": 3,
            "StartBit": 53,
            "TransmissionRate": "200 ms",
            "Units": "bit"
        },
        2869: {
            "Acronym": "OWW",
            "DataRange": "0 to 100 %",
            "EndBit": 23,
            "Name": "Front Operator Wiper Delay Control",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 2869,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "200 ms",
            "Units": "%"
        },
        2870: {
            "Acronym": "OWW",
            "DataRange": "0 to 100 %",
            "EndBit": 31,
            "Name": "Front Non-operator Wiper Delay Control",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 2870,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "200 ms",
            "Units": "%"
        },
        2871: {
            "Acronym": "OWW",
            "DataRange": "0 to 100 %",
            "EndBit": 39,
            "Name": "Rear Wiper Delay Control",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 2871,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "200 ms",
            "Units": "%"
        },
        2872: {
            "Acronym": "OEL",
            "DataRange": "0 to 15",
            "EndBit": 7,
            "Name": "Main Light Switch",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2872,
            "SPNLength": 4,
            "StartBit": 4,
            "TransmissionRate": "Every 1 s and on change of switch state but no faster than every 100 ms",
            "Units": "bit"
        },
        2873: {
            "Acronym": "OEL",
            "DataRange": "0 to 15",
            "EndBit": 3,
            "Name": "Work Light Switch",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2873,
            "SPNLength": 4,
            "StartBit": 0,
            "TransmissionRate": "Every 1 s and on change of switch state but no faster than every 100 ms",
            "Units": "bit"
        },
        2874: {
            "Acronym": "OEL",
            "DataRange": "0 to 3",
            "EndBit": 15,
            "Name": "High-Low Beam Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2874,
            "SPNLength": 2,
            "StartBit": 14,
            "TransmissionRate": "Every 1 s and on change of switch state but no faster than every 100 ms",
            "Units": "bit"
        },
        2875: {
            "Acronym": "OEL",
            "DataRange": "0 to 3",
            "EndBit": 13,
            "Name": "Hazard Light Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2875,
            "SPNLength": 2,
            "StartBit": 12,
            "TransmissionRate": "Every 1 s and on change of switch state but no faster than every 100 ms",
            "Units": "bit"
        },
        2876: {
            "Acronym": "OEL",
            "DataRange": "0 to 15",
            "EndBit": 11,
            "Name": "Turn Signal Switch",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2876,
            "SPNLength": 4,
            "StartBit": 8,
            "TransmissionRate": "Every 1 s and on change of switch state but no faster than every 100 ms",
            "Units": "bit"
        },
        2877: {
            "Acronym": "OEL",
            "DataRange": "0 to 64,255 s",
            "EndBit": 39,
            "Name": "Operators Desired - Delayed Lamp Off Time",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2877,
            "SPNLength": 16,
            "StartBit": 24,
            "TransmissionRate": "Every 1 s and on change of switch state but no faster than every 100 ms",
            "Units": "Seconds"
        },
        2878: {
            "Acronym": "OEL",
            "DataRange": "0 to 100 %",
            "EndBit": 23,
            "Name": "Operators Desired Back-light",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 2878,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "Every 1 s and on change of switch state but no faster than every 100 ms",
            "Units": "%"
        },
        2879: {
            "Acronym": "OHECS",
            "DataRange": "0 to 15",
            "EndBit": 23,
            "Name": "Engine Alternate Droop Accelerator 2 Select",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2879,
            "SPNLength": 4,
            "StartBit": 20,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        2880: {
            "Acronym": "ISC",
            "DataRange": "0 to 15",
            "EndBit": 3,
            "Name": "Engine Operator Primary Intermediate Speed Select",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2880,
            "SPNLength": 4,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        2881: {
            "Acronym": "OHECS",
            "DataRange": "0 to 15",
            "EndBit": 19,
            "Name": "Engine Alternate Droop Accelerator 1 Select",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2881,
            "SPNLength": 4,
            "StartBit": 16,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        2882: {
            "Acronym": "OHECS",
            "DataRange": "0 to 255",
            "EndBit": 15,
            "Name": "Engine Alternate Rating Select",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2882,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "500 ms",
            "Units": "selection"
        },
        2883: {
            "Acronym": "OHECS",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Engine Alternate Low Idle Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2883,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        2884: {
            "Acronym": "OHECS",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Engine Auxiliary Governor Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2884,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        2885: {
            "Acronym": "OHECS",
            "DataRange": "0 to 15",
            "EndBit": 31,
            "Name": "Engine Alternate Droop Auxiliary Input Select",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2885,
            "SPNLength": 4,
            "StartBit": 28,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        2886: {
            "Acronym": "OHECS",
            "DataRange": "0 to 15",
            "EndBit": 27,
            "Name": "Engine Alternate Droop Remote Accelerator Select",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2886,
            "SPNLength": 4,
            "StartBit": 24,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        2887: {
            "Acronym": "CMI",
            "DataRange": "0 to 64,255",
            "EndBit": 15,
            "Name": "Total Count of Configuration Changes Made",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2887,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "As required",
            "Units": "count"
        },
        2888: {
            "Acronym": "OHCSS",
            "DataRange": "0 to 255",
            "EndBit": 15,
            "Name": "Engine Alternate Rating Select State",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2888,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "0.5 s",
            "Units": "bit"
        },
        2889: {
            "Acronym": "OHCSS",
            "DataRange": "0 to 15",
            "EndBit": 19,
            "Name": "Engine Alternate Droop Accelerator 1 Select State",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2889,
            "SPNLength": 4,
            "StartBit": 16,
            "TransmissionRate": "0.5 s",
            "Units": "bit"
        },
        2890: {
            "Acronym": "OHCSS",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Engine Multi-Unit Sync State",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2890,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "0.5 s",
            "Units": "bit"
        },
        2891: {
            "Acronym": "OHCSS",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Engine Alternate Low Idle Select State",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2891,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "0.5 s",
            "Units": "bit"
        },
        2892: {
            "Acronym": "ISCS",
            "DataRange": "0 to 15",
            "EndBit": 3,
            "Name": "Engine Operator Primary Intermediate Speed Select State",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2892,
            "SPNLength": 4,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        2893: {
            "Acronym": "OHCSS",
            "DataRange": "0 to 15",
            "EndBit": 23,
            "Name": "Engine Alternate Droop Accelerator 2 Select State",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2893,
            "SPNLength": 4,
            "StartBit": 20,
            "TransmissionRate": "0.5 s",
            "Units": "bit"
        },
        2894: {
            "Acronym": "OHCSS",
            "DataRange": "0 to 15",
            "EndBit": 27,
            "Name": "Engine Alternate Droop Remote Accelerator Select State",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2894,
            "SPNLength": 4,
            "StartBit": 24,
            "TransmissionRate": "0.5 s",
            "Units": "bit"
        },
        2895: {
            "Acronym": "OHCSS",
            "DataRange": "0 to 15",
            "EndBit": 31,
            "Name": "Engine Alternate Droop Auxiliary Input Select State",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2895,
            "SPNLength": 4,
            "StartBit": 28,
            "TransmissionRate": "0.5 s",
            "Units": "bit"
        },
        2896: {
            "Acronym": "OHCSS",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Engine Auxiliary Governor State",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2896,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "0.5 s",
            "Units": "bit"
        },
        2897: {
            "Acronym": "PTO",
            "DataRange": "0 to 3",
            "EndBit": 57,
            "Name": "Operator Engine PTO Governor Memory Select Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2897,
            "SPNLength": 2,
            "StartBit": 56,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2898: {
            "Acronym": "CSA",
            "DataRange": "0 to 15",
            "EndBit": 15,
            "Name": "Engine Start Enable Device 2 Configuration",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2898,
            "SPNLength": 4,
            "StartBit": 12,
            "TransmissionRate": "As required",
            "Units": "bit"
        },
        2899: {
            "Acronym": "CSA",
            "DataRange": "0 to 15",
            "EndBit": 11,
            "Name": "Engine Start Enable Device 1 Configuration",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2899,
            "SPNLength": 4,
            "StartBit": 8,
            "TransmissionRate": "As required",
            "Units": "bit"
        },
        29: {
            "Acronym": "EEC2",
            "DataRange": "0 to 100 %",
            "EndBit": 39,
            "Name": "Accelerator Pedal Position 2",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 29,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "50 msec (preferred) or Engine Speed Dependent (if required by application)",
            "Units": "%"
        },
        2900: {
            "Acronym": "ETC7",
            "DataRange": "0 to 3",
            "EndBit": 13,
            "Name": "Transmission Engine Crank Enable",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2900,
            "SPNLength": 2,
            "StartBit": 12,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2901: {
            "Acronym": "ECUID",
            "DataRange": "0 to 255 per byte",
            "EndBit": -1,
            "Name": "ECU Part Number",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": -2,
            "SPN": 2901,
            "SPNLength": 1600,
            "StartBit": -1,
            "TransmissionRate": "On request",
            "Units": "ASCII"
        },
        2902: {
            "Acronym": "ECUID",
            "DataRange": "0 to 255 per byte",
            "EndBit": -2,
            "Name": "ECU Serial Number",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": -2,
            "SPN": 2902,
            "SPNLength": 1600,
            "StartBit": -1,
            "TransmissionRate": "On request",
            "Units": "ASCII"
        },
        2903: {
            "Acronym": "ECUID",
            "DataRange": "0 to 255 per byte",
            "EndBit": -3,
            "Name": "ECU Location",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": -2,
            "SPN": 2903,
            "SPNLength": 1600,
            "StartBit": -1,
            "TransmissionRate": "On request",
            "Units": "ASCII"
        },
        2904: {
            "Acronym": "ECUID",
            "DataRange": "0 to 255 per byte",
            "EndBit": -4,
            "Name": "ECU Type",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": -2,
            "SPN": 2904,
            "SPNLength": 1600,
            "StartBit": -1,
            "TransmissionRate": "On request",
            "Units": "ASCII"
        },
        2911: {
            "Acronym": "EBC1",
            "DataRange": "0 to 3",
            "EndBit": 59,
            "Name": "Halt brake switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2911,
            "SPNLength": 2,
            "StartBit": 58,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2912: {
            "Acronym": "EBC5",
            "DataRange": "0 to 7",
            "EndBit": 7,
            "Name": "Hill holder mode",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2912,
            "SPNLength": 3,
            "StartBit": 5,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2913: {
            "Acronym": "EBC5",
            "DataRange": "0 to 7",
            "EndBit": 4,
            "Name": "Halt brake mode",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2913,
            "SPNLength": 3,
            "StartBit": 2,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2914: {
            "Acronym": "XBR",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "XBR EBI Mode",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2914,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "When active: 20 ms; else 200 ms",
            "Units": "bit"
        },
        2915: {
            "Acronym": "XBR",
            "DataRange": "0 to 3",
            "EndBit": 19,
            "Name": "XBR Priority",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2915,
            "SPNLength": 2,
            "StartBit": 18,
            "TransmissionRate": "When active: 20 ms; else 200 ms",
            "Units": "bit"
        },
        2916: {
            "Acronym": "XBR",
            "DataRange": "0 to 3",
            "EndBit": 21,
            "Name": "XBR Control Mode",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2916,
            "SPNLength": 2,
            "StartBit": 20,
            "TransmissionRate": "When active: 20 ms; else 200 ms",
            "Units": "bit"
        },
        2917: {
            "Acronym": "EBC5",
            "DataRange": "0 to 3",
            "EndBit": 11,
            "Name": "XBR System State",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2917,
            "SPNLength": 2,
            "StartBit": 10,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2918: {
            "Acronym": "EBC5",
            "DataRange": "0 to 15",
            "EndBit": 15,
            "Name": "XBR Active Control Mode",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2918,
            "SPNLength": 4,
            "StartBit": 12,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2919: {
            "Acronym": "EBC5",
            "DataRange": "0 to 3",
            "EndBit": 9,
            "Name": "Foundation Brake Use",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2919,
            "SPNLength": 2,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2920: {
            "Acronym": "XBR",
            "DataRange": "-15.687 to +15.687 m/s^2",
            "EndBit": 15,
            "Name": "External Acceleration Demand",
            "Offset": -15.687,
            "OperationalHigh": 10.02,
            "OperationalLow": -10.0,
            "OperationalRange": "-10.0 to +10.0 m/s^2",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2920,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "When active: 20 ms; else 200 ms",
            "Units": "m/s^2"
        },
        2921: {
            "Acronym": "EBC5",
            "DataRange": "-12.5 to +12.5 m/s^2",
            "EndBit": 23,
            "Name": "XBR Acceleration Limit",
            "Offset": -12.5,
            "OperationalHigh": 10.02,
            "OperationalLow": -10.0,
            "OperationalRange": "-10.0 to +10.0 m/s^2",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 2921,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "m/s^2"
        },
        2922: {
            "Acronym": "ESC1",
            "DataRange": "0 to 3",
            "EndBit": 29,
            "Name": "Steerable Lift Axle Lowering Inhibit",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2922,
            "SPNLength": 2,
            "StartBit": 28,
            "TransmissionRate": "20 ms",
            "Units": "bit"
        },
        2923: {
            "Acronym": "ESC1",
            "DataRange": "0 to 15",
            "EndBit": 27,
            "Name": "Status of Steering Axle",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2923,
            "SPNLength": 4,
            "StartBit": 24,
            "TransmissionRate": "20 ms",
            "Units": "bit"
        },
        2924: {
            "Acronym": "ESC1",
            "DataRange": "0 to 15",
            "EndBit": 35,
            "Name": "Steering Type",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2924,
            "SPNLength": 4,
            "StartBit": 32,
            "TransmissionRate": "20 ms",
            "Units": "bit"
        },
        2925: {
            "Acronym": "ESC1",
            "DataRange": "0 to 15",
            "EndBit": 39,
            "Name": "Type of Steering Forces",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2925,
            "SPNLength": 4,
            "StartBit": 36,
            "TransmissionRate": "20 ms",
            "Units": "bit"
        },
        2926: {
            "Acronym": "ESC1",
            "DataRange": "0 to 15",
            "EndBit": 43,
            "Name": "Type of Steering Transmission",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2926,
            "SPNLength": 4,
            "StartBit": 40,
            "TransmissionRate": "20 ms",
            "Units": "bit"
        },
        2927: {
            "Acronym": "ESC1",
            "DataRange": "-125 to 125 deg",
            "EndBit": 15,
            "Name": "Actual Inner wheel steering angle",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2927,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "20 ms",
            "Units": "deg"
        },
        2928: {
            "Acronym": "ESC1",
            "DataRange": "0 to 255",
            "EndBit": 23,
            "Name": "Axle Location",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2928,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "20 ms",
            "Units": "bit"
        },
        2930: {
            "Acronym": "HBS",
            "DataRange": "0 to 3",
            "EndBit": 25,
            "Name": "Hydraulic Brake System Audible Warning Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2930,
            "SPNLength": 2,
            "StartBit": 24,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2931: {
            "Acronym": "HBS",
            "DataRange": "0 to 3",
            "EndBit": 27,
            "Name": "Hydraulic Brake Fluid Level Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2931,
            "SPNLength": 2,
            "StartBit": 26,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2932: {
            "Acronym": "GPV1",
            "DataRange": "0 to 15",
            "EndBit": 19,
            "Name": "Valve State",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2932,
            "SPNLength": 4,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2934: {
            "Acronym": "GPV1",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Valve Fail Safe Mode",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2934,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2937: {
            "Acronym": "GPV2",
            "DataRange": "-125 to 125 %",
            "EndBit": 7,
            "Name": "Extend Port Measured Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2937,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        2938: {
            "Acronym": "GPV2",
            "DataRange": "-125 to 125 %",
            "EndBit": 15,
            "Name": "Retract Port Measured Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2938,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        2939: {
            "Acronym": "GPV1",
            "DataRange": "-125 to 125 %",
            "EndBit": 7,
            "Name": "Extend Port Estimated Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2939,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        2940: {
            "Acronym": "GPV1",
            "DataRange": "-125 to 125 %",
            "EndBit": 15,
            "Name": "Retract Port Estimated Flow",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2940,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        2941: {
            "Acronym": "GPV2",
            "DataRange": "0.0 to 46596.99927595 psi",
            "EndBit": 31,
            "Name": "Extend Port Pressure",
            "Offset": 0.0,
            "OperationalHigh": 46596.99927595,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.72518869,
            "SPN": 2941,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        2942: {
            "Acronym": "GPV2",
            "DataRange": "0.0 to 46596.99927595 psi",
            "EndBit": 47,
            "Name": "Retract Port Pressure",
            "Offset": 0.0,
            "OperationalHigh": 46596.99927595,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.72518869,
            "SPN": 2942,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        2943: {
            "Acronym": "GPV2",
            "DataRange": "0.0 to 580.150952 psi",
            "EndBit": 55,
            "Name": "Return Port Pressure",
            "Offset": 0.0,
            "OperationalHigh": 580.150952,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 2.320603808,
            "SPN": 2943,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        2944: {
            "Acronym": "GPV3",
            "DataRange": "0 to 100 %",
            "EndBit": 7,
            "Name": "Port Flow Command",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 2944,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "1 s for each valve or on change of state.",
            "Units": "%"
        },
        2945: {
            "Acronym": "ETC7",
            "DataRange": "0 to 3",
            "EndBit": 11,
            "Name": "Active Shift Console Indicator",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2945,
            "SPNLength": 2,
            "StartBit": 10,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2948: {
            "Acronym": "EFL/P3",
            "DataRange": "0.0 to 36404.472237999995 psi",
            "EndBit": 15,
            "Name": "Engine Intake Valve Actuation System Oil Pressure",
            "Offset": 0.0,
            "OperationalHigh": 36404.472237999995,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.5665536640625,
            "SPN": 2948,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        2970: {
            "Acronym": "EEC2",
            "DataRange": "0 to 3",
            "EndBit": 7,
            "Name": "Accelerator Pedal 2 Low Idle Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2970,
            "SPNLength": 2,
            "StartBit": 6,
            "TransmissionRate": "50 msec (preferred) or Engine Speed Dependent (if required by application)",
            "Units": "bit"
        },
        2978: {
            "Acronym": "EEC3",
            "DataRange": "-125 to 125 %",
            "EndBit": 39,
            "Name": "Estimated Engine Parasitic Losses - Percent Torque",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 125%",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2978,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "250 msec (preferred) or Engine Speed Dependent (if required by application)",
            "Units": "%"
        },
        2979: {
            "Acronym": "EEC2",
            "DataRange": "0 to 3",
            "EndBit": 41,
            "Name": "Vehicle Acceleration Rate Limit Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2979,
            "SPNLength": 2,
            "StartBit": 40,
            "TransmissionRate": "50 msec (preferred) or Engine Speed Dependent (if required by application)",
            "Units": "bit"
        },
        2980: {
            "Acronym": "GFP",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 63,
            "Name": "Engine Fuel  Valve 1 Outlet Absolute Pressure",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 2980,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "On request",
            "Units": "psi"
        },
        2983: {
            "Acronym": "ETC6",
            "DataRange": "0 to 100 %",
            "EndBit": 31,
            "Name": "Clutch Life Remaining",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 2983,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        2984: {
            "Acronym": "ASC2",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Automatic traction help (load transfer)",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2984,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        2985: {
            "Acronym": "TC1",
            "DataRange": "0 to 3",
            "EndBit": 55,
            "Name": "Transmission Shift Selector Display Mode Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 2985,
            "SPNLength": 2,
            "StartBit": 54,
            "TransmissionRate": "When active; 50 ms to transmission and axles",
            "Units": "bit"
        },
        2986: {
            "Acronym": "ET3",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 47,
            "Name": "Engine Intake Valve Actuation System Oil Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 2986,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        3026: {
            "Acronym": "TRF1",
            "DataRange": "0 to 15",
            "EndBit": 63,
            "Name": "Transmission Oil Level 1 Measurement Status",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3026,
            "SPNLength": 4,
            "StartBit": 60,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        3027: {
            "Acronym": "TRF1",
            "DataRange": "-62.5 to 62.5 L",
            "EndBit": 55,
            "Name": "Transmission Oil Level 1 High / Low",
            "Offset": -62.5,
            "OperationalHigh": 62.5,
            "OperationalLow": -62.5,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.5,
            "SPN": 3027,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "1 s",
            "Units": "liters"
        },
        3028: {
            "Acronym": "TRF1",
            "DataRange": "0 to 15",
            "EndBit": 59,
            "Name": "Transmission Oil Level 1 Countdown Timer",
            "Offset": 0.0,
            "OperationalHigh": 13.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 13 minutes",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3028,
            "SPNLength": 4,
            "StartBit": 56,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        3030: {
            "Acronym": "ETC8",
            "DataRange": "0 to 64.255",
            "EndBit": 15,
            "Name": "Transmission Torque Converter Ratio",
            "Offset": 0.0,
            "OperationalHigh": 64.255,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.001,
            "SPN": 3030,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "20 ms when torque converter unlocked, 100 ms when torque converter locked",
            "Units": "Ratio"
        },
        3031: {
            "Acronym": "AT1T1I",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 15,
            "Name": "Aftertreatment 1 Diesel Exhaust Fluid Tank Temperature",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 3031,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        3034: {
            "Acronym": "DM22",
            "DataRange": "0 to 255",
            "EndBit": 7,
            "Name": "DM22 Control Byte - Individual DTC Clear/Reset Control Byte",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3034,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        3035: {
            "Acronym": "DM22",
            "DataRange": "0 to 255",
            "EndBit": 15,
            "Name": "DM22 - Control Byte Specific Indicator For Individual DTC Clear",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3035,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        3036: {
            "Acronym": "DM22",
            "DataRange": "0 to 524,287",
            "EndBit": 61,
            "Name": "DM22 - SPN",
            "Offset": 0.0,
            "OperationalHigh": 524287.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 3036,
            "SPNLength": 19,
            "StartBit": 40,
            "TransmissionRate": "On request",
            "Units": "binary"
        },
        3037: {
            "Acronym": "DM22",
            "DataRange": "0 to 31",
            "EndBit": 60,
            "Name": "DM22 - FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 3037,
            "SPNLength": 5,
            "StartBit": 56,
            "TransmissionRate": "On request",
            "Units": "binary"
        },
        3038: {
            "Acronym": "DM12",
            "DataRange": "0 to 3",
            "EndBit": 15,
            "Name": "Flash Malfunction Indicator Lamp (MIL)",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 3038,
            "SPNLength": 2,
            "StartBit": 14,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        3039: {
            "Acronym": "DM12",
            "DataRange": "0 to 3",
            "EndBit": 13,
            "Name": "Flash Red Stop Lamp (RSL)",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 3039,
            "SPNLength": 2,
            "StartBit": 12,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        3040: {
            "Acronym": "DM12",
            "DataRange": "0 to 3",
            "EndBit": 11,
            "Name": "Flash Amber Warning Lamp (AWL)",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 3040,
            "SPNLength": 2,
            "StartBit": 10,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        3041: {
            "Acronym": "DM12",
            "DataRange": "0 to 3",
            "EndBit": 9,
            "Name": "Flash Protect Lamp",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 3041,
            "SPNLength": 2,
            "StartBit": 8,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        3043: {
            "Acronym": "TR7",
            "DataRange": "0 to 255",
            "EndBit": 7,
            "Name": "Type of Passenger Count",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3043,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        3044: {
            "Acronym": "TR7",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Silent Alarm Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3044,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        3045: {
            "Acronym": "TR7",
            "DataRange": "0 to 3",
            "EndBit": 19,
            "Name": "Vehicle Use Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3045,
            "SPNLength": 2,
            "StartBit": 18,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        3046: {
            "Acronym": "TR7",
            "DataRange": "0 to 3",
            "EndBit": 21,
            "Name": "Transit Run Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3046,
            "SPNLength": 2,
            "StartBit": 20,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        3047: {
            "Acronym": "TR7",
            "DataRange": "0 to 250",
            "EndBit": 15,
            "Name": "Patron Count",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3047,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "On request",
            "Units": "count"
        },
        3048: {
            "Acronym": "DM20",
            "DataRange": "0 to 65,535",
            "EndBit": 15,
            "Name": "Ignition Cycle Counter",
            "Offset": 0.0,
            "OperationalHigh": 65535.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 3048,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "count"
        },
        3049: {
            "Acronym": "DM20",
            "DataRange": "0 to 65,535",
            "EndBit": 31,
            "Name": "OBD Monitoring Conditions Encountered",
            "Offset": 0.0,
            "OperationalHigh": 65535.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 3049,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "count"
        },
        3066: {
            "Acronym": "DM20",
            "DataRange": "0 to 524,287",
            "EndBit": 50,
            "Name": "SPN of Applicable System Monitor",
            "Offset": 0.0,
            "OperationalHigh": 524287.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": -3,
            "SPN": 3066,
            "SPNLength": 19,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "binary"
        },
        3067: {
            "Acronym": "DM20",
            "DataRange": "0 to 65,535",
            "EndBit": 71,
            "Name": "Applicable System Monitor Numerator",
            "Offset": 0.0,
            "OperationalHigh": 65535.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 3067,
            "SPNLength": 16,
            "StartBit": 56,
            "TransmissionRate": "On request",
            "Units": "count"
        },
        3068: {
            "Acronym": "DM20",
            "DataRange": "0 to 65,535",
            "EndBit": 87,
            "Name": "Applicable System Monitor Denominator",
            "Offset": 0.0,
            "OperationalHigh": 65535.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 3068,
            "SPNLength": 16,
            "StartBit": 72,
            "TransmissionRate": "On request",
            "Units": "count"
        },
        3069: {
            "Acronym": "DM21",
            "DataRange": "0.0 to 39926.193605 mile",
            "EndBit": 15,
            "Name": "Distance Travelled While MIL is Activated",
            "Offset": 0.0,
            "OperationalHigh": 39926.193605,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.621371,
            "SPN": 3069,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "miles"
        },
        3070: {
            "Acronym": "TR2",
            "DataRange": "0 to 250 bytes",
            "EndBit": 7,
            "Name": "Number of bytes in the Milepost Identification",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 100 bytes",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 3070,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "byte"
        },
        3071: {
            "Acronym": "TR1",
            "DataRange": "0 to 250 bytes",
            "EndBit": 15,
            "Name": "Number of bytes in the Transit Assigned Route Identity",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 1.0,
            "OperationalRange": "1 to 100 bytes",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 3071,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "On request",
            "Units": "byte"
        },
        3072: {
            "Acronym": "TR1",
            "DataRange": "0 to 250 bytes",
            "EndBit": 23,
            "Name": "Number of bytes in the Transit Assigned Run Identity",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 1.0,
            "OperationalRange": "1 to 100 bytes",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 3072,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "byte"
        },
        3073: {
            "Acronym": "TR1",
            "DataRange": "0 to 250 bytes",
            "EndBit": 31,
            "Name": "Number of bytes in the Transit Assigned Block Identity",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 1.0,
            "OperationalRange": "1 to 100 bytes",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 3073,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "On request",
            "Units": "byte"
        },
        3074: {
            "Acronym": "TR1",
            "DataRange": "0 to 255 per byte",
            "EndBit": 0,
            "Name": "Transit Assigned Route Identity",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": -2,
            "SPN": 3074,
            "SPNLength": 800,
            "StartBit": -1,
            "TransmissionRate": "On request",
            "Units": "ASCII"
        },
        3075: {
            "Acronym": "TR1",
            "DataRange": "0 to 255 per byte",
            "EndBit": 0,
            "Name": "Transit Assigned Run Identity",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": -2,
            "SPN": 3075,
            "SPNLength": 800,
            "StartBit": -1,
            "TransmissionRate": "On request",
            "Units": "ASCII"
        },
        3076: {
            "Acronym": "TR1",
            "DataRange": "0 to 255 per byte",
            "EndBit": 0,
            "Name": "Transit Assigned Block Identity",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": -2,
            "SPN": 3076,
            "SPNLength": 800,
            "StartBit": -1,
            "TransmissionRate": "On request",
            "Units": "ASCII"
        },
        3078: {
            "Acronym": "TR1",
            "DataRange": "0 to 255",
            "EndBit": 7,
            "Name": "Agency",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 3078,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "count"
        },
        3079: {
            "Acronym": "TR3",
            "DataRange": "0 to 3",
            "EndBit": 7,
            "Name": "Intersection Preemption Request/Response",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3079,
            "SPNLength": 2,
            "StartBit": 6,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        3080: {
            "Acronym": "TR3",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Transit Route ID Usage",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3080,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        3081: {
            "Acronym": "TR3",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Range Code Enable",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3081,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        3082: {
            "Acronym": "TR3",
            "DataRange": "0 to 3",
            "EndBit": 15,
            "Name": "Strobe Activation Control Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3082,
            "SPNLength": 2,
            "StartBit": 14,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        3083: {
            "Acronym": "TR3",
            "DataRange": "0 to 3",
            "EndBit": 13,
            "Name": "Transit Door Enable",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3083,
            "SPNLength": 2,
            "StartBit": 12,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        3084: {
            "Acronym": "TR3",
            "DataRange": "0 to 15",
            "EndBit": 11,
            "Name": "Priority of Response Sent by Emitter",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3084,
            "SPNLength": 4,
            "StartBit": 8,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        3085: {
            "Acronym": "TR3",
            "DataRange": "0 to 65,535",
            "EndBit": 31,
            "Name": "Vehicle ID",
            "Offset": 0.0,
            "OperationalHigh": 65534.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 65,534",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3085,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "count"
        },
        3086: {
            "Acronym": "ETC7",
            "DataRange": "0 to 3",
            "EndBit": 9,
            "Name": "Transmission Ready for Brake Release",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3086,
            "SPNLength": 2,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3087: {
            "Acronym": "AAI",
            "DataRange": "0 to 6,425.5 mm (0 to 6.4255 m)",
            "EndBit": 47,
            "Name": "Auxiliary Level",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 3087,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "m"
        },
        3156: {
            "Acronym": "LOI",
            "DataRange": "0 to 15",
            "EndBit": 3,
            "Name": "Blade Control Mode Switch",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3156,
            "SPNLength": 4,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3157: {
            "Acronym": "LOI",
            "DataRange": "0 to 15",
            "EndBit": 7,
            "Name": "Desired Grade Offset Switch",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3157,
            "SPNLength": 4,
            "StartBit": 4,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3158: {
            "Acronym": "LOI",
            "DataRange": "0 to 15",
            "EndBit": 11,
            "Name": "Blade Auto Mode Command",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3158,
            "SPNLength": 4,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3159: {
            "Acronym": "TR4",
            "DataRange": "0 to 65,535",
            "EndBit": 39,
            "Name": "Trip Number",
            "Offset": 0.0,
            "OperationalHigh": 65535.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 15,
            "Resolution": 1,
            "SPN": 3159,
            "SPNLength": 16,
            "StartBit": 24,
            "TransmissionRate": "On request",
            "Units": "count"
        },
        3160: {
            "Acronym": "TR4",
            "DataRange": "0 to 65,535",
            "EndBit": 71,
            "Name": "Assigned Route",
            "Offset": 0.0,
            "OperationalHigh": 65535.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 15,
            "Resolution": 1,
            "SPN": 3160,
            "SPNLength": 16,
            "StartBit": 56,
            "TransmissionRate": "On request",
            "Units": "count"
        },
        3161: {
            "Acronym": "TR4",
            "DataRange": "0 to 65,535",
            "EndBit": 55,
            "Name": "Pattern Number ",
            "Offset": 0.0,
            "OperationalHigh": 65535.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 15,
            "Resolution": 1,
            "SPN": 3161,
            "SPNLength": 16,
            "StartBit": 40,
            "TransmissionRate": "On request",
            "Units": "count"
        },
        3162: {
            "Acronym": "TR4",
            "DataRange": "0 to 65,535",
            "EndBit": 87,
            "Name": "Assigned Run",
            "Offset": 0.0,
            "OperationalHigh": 65535.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 15,
            "Resolution": 1,
            "SPN": 3162,
            "SPNLength": 16,
            "StartBit": 72,
            "TransmissionRate": "On request",
            "Units": "count"
        },
        3163: {
            "Acronym": "TR4",
            "DataRange": "0 to 65,535",
            "EndBit": 103,
            "Name": "Assigned Block",
            "Offset": 0.0,
            "OperationalHigh": 65535.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 15,
            "Resolution": 1,
            "SPN": 3163,
            "SPNLength": 16,
            "StartBit": 88,
            "TransmissionRate": "On request",
            "Units": "count"
        },
        3164: {
            "Acronym": "TR4",
            "DataRange": "0 to 65,535",
            "EndBit": 119,
            "Name": "Driver's farebox security code",
            "Offset": 0.0,
            "OperationalHigh": 65535.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 15,
            "Resolution": 1,
            "SPN": 3164,
            "SPNLength": 16,
            "StartBit": 104,
            "TransmissionRate": "On request",
            "Units": "count"
        },
        3165: {
            "Acronym": "TR5",
            "DataRange": "0 to 15",
            "EndBit": 19,
            "Name": "Fare Validity",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3165,
            "SPNLength": 4,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "count"
        },
        3166: {
            "Acronym": "TR5",
            "DataRange": "0 to 15",
            "EndBit": 23,
            "Name": "Pass Category",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3166,
            "SPNLength": 4,
            "StartBit": 20,
            "TransmissionRate": "On request",
            "Units": "count"
        },
        3167: {
            "Acronym": "TR5",
            "DataRange": "0 to 31",
            "EndBit": 28,
            "Name": "Initial Fare Agency",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 1.0,
            "OperationalRange": "1 to 31",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3167,
            "SPNLength": 5,
            "StartBit": 24,
            "TransmissionRate": "On request",
            "Units": "count"
        },
        3168: {
            "Acronym": "TR5",
            "DataRange": "0 to 255",
            "EndBit": 63,
            "Name": "Transfer Sold",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3168,
            "SPNLength": 8,
            "StartBit": 56,
            "TransmissionRate": "On request",
            "Units": "count"
        },
        3169: {
            "Acronym": "TR5",
            "DataRange": "0 to 4095",
            "EndBit": 55,
            "Name": "Route Number",
            "Offset": 0.0,
            "OperationalHigh": 4095.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3169,
            "SPNLength": 12,
            "StartBit": 40,
            "TransmissionRate": "On request",
            "Units": "count"
        },
        3170: {
            "Acronym": "TR5",
            "DataRange": "0 to 15",
            "EndBit": 3,
            "Name": "Transaction Type",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3170,
            "SPNLength": 4,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        3171: {
            "Acronym": "TR5",
            "DataRange": "0 to 15",
            "EndBit": 7,
            "Name": "Passenger Type",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3171,
            "SPNLength": 4,
            "StartBit": 4,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        3172: {
            "Acronym": "TR5",
            "DataRange": "0 to 7",
            "EndBit": 34,
            "Name": "Type of Service",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3172,
            "SPNLength": 3,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        3173: {
            "Acronym": "TR5",
            "DataRange": "0 to 31",
            "EndBit": 39,
            "Name": "Transfer Type",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3173,
            "SPNLength": 5,
            "StartBit": 35,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        3174: {
            "Acronym": "TR4",
            "DataRange": "0 to 15",
            "EndBit": 11,
            "Name": "Trip Direction",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 15,
            "Resolution": 1,
            "SPN": 3174,
            "SPNLength": 4,
            "StartBit": 8,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        3175: {
            "Acronym": "TR4",
            "DataRange": "0 to 255",
            "EndBit": 23,
            "Name": "Fare Presets",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 15,
            "Resolution": 1,
            "SPN": 3175,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        3176: {
            "Acronym": "TR5",
            "DataRange": "0 to 15",
            "EndBit": 11,
            "Name": "Type of Fare",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3176,
            "SPNLength": 4,
            "StartBit": 8,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        3177: {
            "Acronym": "TR5",
            "DataRange": "0 to 15",
            "EndBit": 15,
            "Name": "Payment Details",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3177,
            "SPNLength": 4,
            "StartBit": 12,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        3178: {
            "Acronym": "TR4",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Farebox Service Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 15,
            "Resolution": 1,
            "SPN": 3178,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        3179: {
            "Acronym": "TR6",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Farebox Emergency Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3179,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        3180: {
            "Acronym": "TR4",
            "DataRange": "0 to 7",
            "EndBit": 4,
            "Name": "Trip Status",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 15,
            "Resolution": 1,
            "SPN": 3180,
            "SPNLength": 3,
            "StartBit": 2,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        3181: {
            "Acronym": "TR6",
            "DataRange": "0 to 127",
            "EndBit": 14,
            "Name": "Farebox Alarm Identifier",
            "Offset": 0.0,
            "OperationalHigh": 127.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3181,
            "SPNLength": 7,
            "StartBit": 8,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        3188: {
            "Acronym": "XBR",
            "DataRange": "0 to 15",
            "EndBit": 63,
            "Name": "XBR Message Checksum",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3188,
            "SPNLength": 4,
            "StartBit": 60,
            "TransmissionRate": "When active: 20 ms; else 200 ms",
            "Units": "count"
        },
        3189: {
            "Acronym": "XBR",
            "DataRange": "0 to 15",
            "EndBit": 59,
            "Name": "XBR Message Counter",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3189,
            "SPNLength": 4,
            "StartBit": 56,
            "TransmissionRate": "When active: 20 ms; else 200 ms",
            "Units": "count"
        },
        3190: {
            "Acronym": "TPRI",
            "DataRange": "0 to 255",
            "EndBit": 7,
            "Name": "Tire Location",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3190,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        3191: {
            "Acronym": "TPRI",
            "DataRange": "0.0 to 290.075476 psi",
            "EndBit": 15,
            "Name": "Reference Tire Pressure",
            "Offset": 0.0,
            "OperationalHigh": 290.075476,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.160301904,
            "SPN": 3191,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "On request",
            "Units": "psi"
        },
        3192: {
            "Acronym": "TPRS",
            "DataRange": "0 to 255",
            "EndBit": 7,
            "Name": "Tire Location",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3192,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "As needed",
            "Units": "bit"
        },
        3193: {
            "Acronym": "TPRS",
            "DataRange": "0.0 to 290.075476 psi",
            "EndBit": 15,
            "Name": "Reference Tire Pressure Setting",
            "Offset": 0.0,
            "OperationalHigh": 290.075476,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.160301904,
            "SPN": 3193,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "As needed",
            "Units": "psi"
        },
        3215: {
            "Acronym": "ASC2",
            "DataRange": "0 to 3",
            "EndBit": 53,
            "Name": "Prohibit air suspension control",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3215,
            "SPNLength": 2,
            "StartBit": 52,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3216: {
            "Acronym": "AT1IG1",
            "DataRange": "-200 to 3012.75 ppm",
            "EndBit": 15,
            "Name": "Aftertreatment 1 Intake NOx",
            "Offset": -200.0,
            "OperationalHigh": 3012.75,
            "OperationalLow": -200.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05,
            "SPN": 3216,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "50 ms",
            "Units": "ppm"
        },
        3217: {
            "Acronym": "AT1IG1",
            "DataRange": "-12 to 21%",
            "EndBit": 31,
            "Name": "Aftertreatment 1 Intake O2",
            "Offset": -12.0,
            "OperationalHigh": 21.0,
            "OperationalLow": -12.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.000514,
            "SPN": 3217,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "50 ms",
            "Units": "%"
        },
        3218: {
            "Acronym": "AT1IG1",
            "DataRange": "0 to 3",
            "EndBit": 33,
            "Name": "Aftertreatment 1 Intake Gas Sensor Power Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3218,
            "SPNLength": 2,
            "StartBit": 32,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        3219: {
            "Acronym": "AT1IG1",
            "DataRange": "0 to 3",
            "EndBit": 35,
            "Name": "Aftertreatment 1 Intake Gas Sensor at Temperature",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3219,
            "SPNLength": 2,
            "StartBit": 34,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        3220: {
            "Acronym": "AT1IG1",
            "DataRange": "0 to 3",
            "EndBit": 37,
            "Name": "Aftertreatment 1 Intake NOx Reading Stable",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3220,
            "SPNLength": 2,
            "StartBit": 36,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        3221: {
            "Acronym": "AT1IG1",
            "DataRange": "0 to 3",
            "EndBit": 39,
            "Name": "Aftertreatment 1 Intake Wide-Range % O2 Reading Stable",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3221,
            "SPNLength": 2,
            "StartBit": 38,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        3222: {
            "Acronym": "AT1IG1",
            "DataRange": "0 to 31",
            "EndBit": 44,
            "Name": "Aftertreatment 1 Intake Gas Sensor Heater Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 3222,
            "SPNLength": 5,
            "StartBit": 40,
            "TransmissionRate": "50 ms",
            "Units": "binary"
        },
        3223: {
            "Acronym": "AT1IG1",
            "DataRange": "0 to 3",
            "EndBit": 46,
            "Name": "Aftertreatment 1 Intake Gas Sensor Heater Control",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3223,
            "SPNLength": 2,
            "StartBit": 45,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        3224: {
            "Acronym": "AT1IG1",
            "DataRange": "0 to 31",
            "EndBit": 52,
            "Name": "Aftertreatment 1 Intake NOx Sensor Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 3224,
            "SPNLength": 5,
            "StartBit": 48,
            "TransmissionRate": "50 ms",
            "Units": "binary"
        },
        3225: {
            "Acronym": "AT1IG1",
            "DataRange": "0 to 31",
            "EndBit": 60,
            "Name": "Aftertreatment 1 Intake Oxygen Sensor Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 3225,
            "SPNLength": 5,
            "StartBit": 56,
            "TransmissionRate": "50 ms",
            "Units": "binary"
        },
        3226: {
            "Acronym": "AT1OG1",
            "DataRange": "-200 to 3012.75 ppm",
            "EndBit": 15,
            "Name": "Aftertreatment 1 Outlet NOx",
            "Offset": -200.0,
            "OperationalHigh": 3012.75,
            "OperationalLow": -200.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05,
            "SPN": 3226,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "50 ms",
            "Units": "ppm"
        },
        3227: {
            "Acronym": "AT1OG1",
            "DataRange": "-12 to 21%",
            "EndBit": 31,
            "Name": "Aftertreatment 1 Outlet O2",
            "Offset": -12.0,
            "OperationalHigh": 21.0,
            "OperationalLow": -12.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.000514,
            "SPN": 3227,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "50 ms",
            "Units": "%"
        },
        3228: {
            "Acronym": "AT1OG1",
            "DataRange": "0 to 3",
            "EndBit": 33,
            "Name": "Aftertreatment 1 Outlet Gas Sensor Power Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3228,
            "SPNLength": 2,
            "StartBit": 32,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        3229: {
            "Acronym": "AT1OG1",
            "DataRange": "0 to 3",
            "EndBit": 35,
            "Name": "Aftertreatment 1 Outlet Gas Sensor at Temperature",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3229,
            "SPNLength": 2,
            "StartBit": 34,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        3230: {
            "Acronym": "AT1OG1",
            "DataRange": "0 to 3",
            "EndBit": 37,
            "Name": "Aftertreatment 1 Outlet NOx Reading Stable",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3230,
            "SPNLength": 2,
            "StartBit": 36,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        3231: {
            "Acronym": "AT1OG1",
            "DataRange": "0 to 3",
            "EndBit": 39,
            "Name": "Aftertreatment 1 Outlet Wide-Range %O2 Reading Stable",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3231,
            "SPNLength": 2,
            "StartBit": 38,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        3232: {
            "Acronym": "AT1OG1",
            "DataRange": "0 to 31",
            "EndBit": 44,
            "Name": "Aftertreatment 1 Outlet Gas Sensor Heater Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 3232,
            "SPNLength": 5,
            "StartBit": 40,
            "TransmissionRate": "50 ms",
            "Units": "binary"
        },
        3233: {
            "Acronym": "AT1OG1",
            "DataRange": "0 to 3",
            "EndBit": 46,
            "Name": "Aftertreatment 1 Outlet Gas Sensor Heater Control",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3233,
            "SPNLength": 2,
            "StartBit": 45,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        3234: {
            "Acronym": "AT1OG1",
            "DataRange": "0 to 31",
            "EndBit": 52,
            "Name": "Aftertreatment 1 Outlet NOx Sensor Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 3234,
            "SPNLength": 5,
            "StartBit": 48,
            "TransmissionRate": "50 ms",
            "Units": "binary"
        },
        3235: {
            "Acronym": "AT1OG1",
            "DataRange": "0 to 31",
            "EndBit": 60,
            "Name": "Aftertreatment 1 Outlet Oxygen Sensor Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 3235,
            "SPNLength": 5,
            "StartBit": 56,
            "TransmissionRate": "50 ms",
            "Units": "binary"
        },
        3236: {
            "Acronym": "EEC3",
            "DataRange": "0 to 12851 kg/h per bit",
            "EndBit": 55,
            "Name": "Aftertreatment 1 Exhaust Gas Mass Flow",
            "Offset": 0.0,
            "OperationalHigh": 12851.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.2,
            "SPN": 3236,
            "SPNLength": 16,
            "StartBit": 40,
            "TransmissionRate": "250 msec (preferred) or Engine Speed Dependent (if required by application)",
            "Units": "kg/h"
        },
        3237: {
            "Acronym": "EEC3",
            "DataRange": "0 to 3",
            "EndBit": 57,
            "Name": "Aftertreatment 1 Intake Dew Point",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3237,
            "SPNLength": 2,
            "StartBit": 56,
            "TransmissionRate": "250 msec (preferred) or Engine Speed Dependent (if required by application)",
            "Units": "bit"
        },
        3238: {
            "Acronym": "EEC3",
            "DataRange": "0 to 3",
            "EndBit": 59,
            "Name": "Aftertreatment 1 Exhaust Dew Point",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3238,
            "SPNLength": 2,
            "StartBit": 58,
            "TransmissionRate": "250 msec (preferred) or Engine Speed Dependent (if required by application)",
            "Units": "bit"
        },
        3239: {
            "Acronym": "EEC3",
            "DataRange": "0 to 3",
            "EndBit": 61,
            "Name": "Aftertreatment 2 Intake Dew Point",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3239,
            "SPNLength": 2,
            "StartBit": 60,
            "TransmissionRate": "250 msec (preferred) or Engine Speed Dependent (if required by application)",
            "Units": "bit"
        },
        3240: {
            "Acronym": "EEC3",
            "DataRange": "0 to 3",
            "EndBit": 63,
            "Name": "Aftertreatment 2 Exhaust Dew Point",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3240,
            "SPNLength": 2,
            "StartBit": 62,
            "TransmissionRate": "250 msec (preferred) or Engine Speed Dependent (if required by application)",
            "Units": "bit"
        },
        3241: {
            "Acronym": "AT1IG2",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 15,
            "Name": "Aftertreatment 1 Exhaust Gas Temperature 1",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 3241,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "500 ms",
            "Units": "deg F"
        },
        3242: {
            "Acronym": "AT1IG2",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 31,
            "Name": "Aftertreatment 1 DPF Intake Gas Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 3242,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "500 ms",
            "Units": "deg F"
        },
        3243: {
            "Acronym": "AT1IG2",
            "DataRange": "0 to 31",
            "EndBit": 36,
            "Name": "Aftertreatment 1 Exhaust Gas Temperature 1 Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 3243,
            "SPNLength": 5,
            "StartBit": 32,
            "TransmissionRate": "500 ms",
            "Units": "binary"
        },
        3244: {
            "Acronym": "AT1IG2",
            "DataRange": "0 to 31",
            "EndBit": 44,
            "Name": "Aftertreatment 1 DPF Intake Gas Temperature Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 3244,
            "SPNLength": 5,
            "StartBit": 40,
            "TransmissionRate": "500 ms",
            "Units": "binary"
        },
        3245: {
            "Acronym": "AT1OG2",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 15,
            "Name": "Aftertreatment 1 Exhaust Gas Temperature 3",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 3245,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "500 ms",
            "Units": "deg F"
        },
        3246: {
            "Acronym": "AT1OG2",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 31,
            "Name": "Aftertreatment 1 DPF Outlet Gas Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 3246,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "500 ms",
            "Units": "deg F"
        },
        3247: {
            "Acronym": "AT1OG2",
            "DataRange": "0 to 31",
            "EndBit": 36,
            "Name": "Aftertreatment 1 Exhaust Gas Temperature 3 Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 3247,
            "SPNLength": 5,
            "StartBit": 32,
            "TransmissionRate": "500 ms",
            "Units": "binary"
        },
        3248: {
            "Acronym": "AT1OG2",
            "DataRange": "0 to 31",
            "EndBit": 44,
            "Name": "Aftertreatment 1 DPF Outlet Exhaust Gas Temperature  Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 3248,
            "SPNLength": 5,
            "StartBit": 40,
            "TransmissionRate": "500 ms",
            "Units": "binary"
        },
        3249: {
            "Acronym": "AT1IMG",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 15,
            "Name": "Aftertreatment 1 Exhaust Gas Temperature 2",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 3249,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "500 ms",
            "Units": "deg F"
        },
        3250: {
            "Acronym": "AT1IMG",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 31,
            "Name": "Aftertreatment 1 DPF Intermediate Gas Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 3250,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "500 ms",
            "Units": "deg F"
        },
        3251: {
            "Acronym": "AT1IMG",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 47,
            "Name": "Aftertreatment 1 DPF Differential Pressure",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 3251,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        3252: {
            "Acronym": "AT1IMG",
            "DataRange": "0 to 31",
            "EndBit": 52,
            "Name": "Aftertreatment 1 Exhaust Gas Temperature 2 Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 3252,
            "SPNLength": 5,
            "StartBit": 48,
            "TransmissionRate": "500 ms",
            "Units": "binary"
        },
        3253: {
            "Acronym": "AT1IMG",
            "DataRange": "0 to 31",
            "EndBit": 56,
            "Name": "Aftertreatment 1 DPF Delta Pressure Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 3253,
            "SPNLength": 5,
            "StartBit": 53,
            "TransmissionRate": "500 ms",
            "Units": "binary"
        },
        3254: {
            "Acronym": "AT1IMG",
            "DataRange": "0 to 31",
            "EndBit": 62,
            "Name": "Aftertreatment 1 DPF Intermediate Gas Temperature Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 3254,
            "SPNLength": 5,
            "StartBit": 58,
            "TransmissionRate": "500 ms",
            "Units": "binary"
        },
        3255: {
            "Acronym": "AT2IG1",
            "DataRange": "-200 to 3012.75 ppm",
            "EndBit": 15,
            "Name": "Aftertreatment 2 Intake NOx",
            "Offset": -200.0,
            "OperationalHigh": 3012.75,
            "OperationalLow": -200.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05,
            "SPN": 3255,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "50 ms",
            "Units": "ppm"
        },
        3256: {
            "Acronym": "AT2IG1",
            "DataRange": "-12 to 21%",
            "EndBit": 31,
            "Name": "Aftertreatment 2 Intake O2",
            "Offset": -12.0,
            "OperationalHigh": 21.0,
            "OperationalLow": -12.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.000514,
            "SPN": 3256,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "50 ms",
            "Units": "%"
        },
        3257: {
            "Acronym": "AT2IG1",
            "DataRange": "0 to 3",
            "EndBit": 33,
            "Name": "Aftertreatment 2 Intake Gas Sensor Power Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3257,
            "SPNLength": 2,
            "StartBit": 32,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        3258: {
            "Acronym": "AT2IG1",
            "DataRange": "0 to 3",
            "EndBit": 35,
            "Name": "Aftertreatment 2 Intake Gas Sensor at Temperature",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3258,
            "SPNLength": 2,
            "StartBit": 34,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        3259: {
            "Acronym": "AT2IG1",
            "DataRange": "0 to 3",
            "EndBit": 37,
            "Name": "Aftertreatment 2 Intake NOx Reading Stable",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3259,
            "SPNLength": 2,
            "StartBit": 36,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        3260: {
            "Acronym": "AT2IG1",
            "DataRange": "0 to 3",
            "EndBit": 39,
            "Name": "Aftertreatment 2 Intake Wide-Range % O2 Reading Stable",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3260,
            "SPNLength": 2,
            "StartBit": 38,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        3261: {
            "Acronym": "AT2IG1",
            "DataRange": "0 to 31",
            "EndBit": 44,
            "Name": "Aftertreatment 2 Intake Gas Sensor Heater Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 3261,
            "SPNLength": 5,
            "StartBit": 40,
            "TransmissionRate": "50 ms",
            "Units": "binary"
        },
        3262: {
            "Acronym": "AT2IG1",
            "DataRange": "0 to 3",
            "EndBit": 46,
            "Name": "Aftertreatment 2 Intake Gas Sensor Heater Control",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3262,
            "SPNLength": 2,
            "StartBit": 45,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        3263: {
            "Acronym": "AT2IG1",
            "DataRange": "0 to 31",
            "EndBit": 52,
            "Name": "Aftertreatment 2 Intake NOx Sensor Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 3263,
            "SPNLength": 5,
            "StartBit": 48,
            "TransmissionRate": "50 ms",
            "Units": "binary"
        },
        3264: {
            "Acronym": "AT2IG1",
            "DataRange": "0 to 31",
            "EndBit": 60,
            "Name": "Aftertreatment 2 Intake Oxygen Sensor Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 3264,
            "SPNLength": 5,
            "StartBit": 56,
            "TransmissionRate": "50 ms",
            "Units": "binary"
        },
        3265: {
            "Acronym": "AT2OG1",
            "DataRange": "-200 to 3012.75 ppm",
            "EndBit": 15,
            "Name": "Aftertreatment 2 Outlet NOx",
            "Offset": -200.0,
            "OperationalHigh": 3012.75,
            "OperationalLow": -200.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05,
            "SPN": 3265,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "50 ms",
            "Units": "ppm"
        },
        3266: {
            "Acronym": "AT2OG1",
            "DataRange": "-12 to 21%",
            "EndBit": 31,
            "Name": "Aftertreatment 2 Outlet O2",
            "Offset": -12.0,
            "OperationalHigh": 21.0,
            "OperationalLow": -12.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.000514,
            "SPN": 3266,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "50 ms",
            "Units": "%"
        },
        3267: {
            "Acronym": "AT2OG1",
            "DataRange": "0 to 3",
            "EndBit": 33,
            "Name": "Aftertreatment 2 Outlet Gas Sensor Power Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3267,
            "SPNLength": 2,
            "StartBit": 32,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        3268: {
            "Acronym": "AT2OG1",
            "DataRange": "0 to 3",
            "EndBit": 35,
            "Name": "Aftertreatment 2 Outlet Gas Sensor at Temperature",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3268,
            "SPNLength": 2,
            "StartBit": 34,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        3269: {
            "Acronym": "AT2OG1",
            "DataRange": "0 to 3",
            "EndBit": 37,
            "Name": "Aftertreatment 2 Outlet NOx Reading Stable",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3269,
            "SPNLength": 2,
            "StartBit": 36,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        3270: {
            "Acronym": "AT2OG1",
            "DataRange": "0 to 3",
            "EndBit": 39,
            "Name": "Aftertreatment 2 Outlet Wide-Range % O2 Reading Stable",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3270,
            "SPNLength": 2,
            "StartBit": 38,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        3271: {
            "Acronym": "AT2OG1",
            "DataRange": "0 to 31",
            "EndBit": 44,
            "Name": "Aftertreatment 2 Outlet Gas Sensor Heater Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 3271,
            "SPNLength": 5,
            "StartBit": 40,
            "TransmissionRate": "50 ms",
            "Units": "binary"
        },
        3272: {
            "Acronym": "AT2OG1",
            "DataRange": "0 to 3",
            "EndBit": 46,
            "Name": "Aftertreatment 2 Outlet Gas Sensor Heater Control",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3272,
            "SPNLength": 2,
            "StartBit": 45,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        3273: {
            "Acronym": "AT2OG1",
            "DataRange": "0 to 31",
            "EndBit": 52,
            "Name": "Aftertreatment 2 Outlet NOx Sensor Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 3273,
            "SPNLength": 5,
            "StartBit": 48,
            "TransmissionRate": "50 ms",
            "Units": "binary"
        },
        3274: {
            "Acronym": "AT2OG1",
            "DataRange": "0 to 31",
            "EndBit": 60,
            "Name": "Aftertreatment 2 Outlet Oxygen Sensor Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 3274,
            "SPNLength": 5,
            "StartBit": 56,
            "TransmissionRate": "50 ms",
            "Units": "binary"
        },
        3275: {
            "Acronym": "AT2IG2",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 15,
            "Name": "Aftertreatment 2 Exhaust Gas Temperature 1",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 3275,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "500 ms",
            "Units": "deg F"
        },
        3276: {
            "Acronym": "AT2IG2",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 31,
            "Name": "Aftertreatment 2 DPF Intake Gas Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 3276,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "500 ms",
            "Units": "deg F"
        },
        3277: {
            "Acronym": "AT2IG2",
            "DataRange": "0 to 31",
            "EndBit": 36,
            "Name": "Aftertreatment 2 Exhaust Gas Temperature 1 Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 3277,
            "SPNLength": 5,
            "StartBit": 32,
            "TransmissionRate": "500 ms",
            "Units": "binary"
        },
        3278: {
            "Acronym": "AT2IG2",
            "DataRange": "0 to 31",
            "EndBit": 44,
            "Name": "Aftertreatment 2 DPF Intake Gas Temperature Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 3278,
            "SPNLength": 5,
            "StartBit": 40,
            "TransmissionRate": "500 ms",
            "Units": "binary"
        },
        3279: {
            "Acronym": "AT2OG2",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 15,
            "Name": "Aftertreatment 2 Exhaust Gas Temperature 3",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 3279,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "500 ms",
            "Units": "deg F"
        },
        3280: {
            "Acronym": "AT2OG2",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 31,
            "Name": "Aftertreatment 2 DPF Outlet Gas Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 3280,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "500 ms",
            "Units": "deg F"
        },
        3281: {
            "Acronym": "AT2OG2",
            "DataRange": "0 to 31",
            "EndBit": 36,
            "Name": "Aftertreatment 2 Exhaust Gas Temperature 3 Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 3281,
            "SPNLength": 5,
            "StartBit": 32,
            "TransmissionRate": "500 ms",
            "Units": "binary"
        },
        3282: {
            "Acronym": "AT2OG2",
            "DataRange": "0 to 31",
            "EndBit": 44,
            "Name": "Aftertreatment 2 DPF Exhaust Gas Temperature  Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 3282,
            "SPNLength": 5,
            "StartBit": 40,
            "TransmissionRate": "500 ms",
            "Units": "binary"
        },
        3283: {
            "Acronym": "AT2IMG",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 15,
            "Name": "Aftertreatment 2 Exhaust Gas Temperature 2",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 3283,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "500 ms",
            "Units": "deg F"
        },
        3284: {
            "Acronym": "AT2IMG",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 31,
            "Name": "Aftertreatment 2 DPF Intermediate Gas Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 3284,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "500 ms",
            "Units": "deg F"
        },
        3285: {
            "Acronym": "AT2IMG",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 47,
            "Name": "Aftertreatment 2 DPF Differential Pressure",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 3285,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        3286: {
            "Acronym": "AT2IMG",
            "DataRange": "0 to 31",
            "EndBit": 52,
            "Name": "Aftertreatment 2 Exhaust Gas Temperature 2 Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 3286,
            "SPNLength": 5,
            "StartBit": 48,
            "TransmissionRate": "500 ms",
            "Units": "binary"
        },
        3287: {
            "Acronym": "AT2IMG",
            "DataRange": "0 to 31",
            "EndBit": 56,
            "Name": "Aftertreatment 2 DPF Delta Pressure Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 3287,
            "SPNLength": 5,
            "StartBit": 53,
            "TransmissionRate": "500 ms",
            "Units": "binary"
        },
        3288: {
            "Acronym": "AT2IMG",
            "DataRange": "0 to 31",
            "EndBit": 62,
            "Name": "Aftertreatment 2 DPF Intermediate Gas Temperature Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 3288,
            "SPNLength": 5,
            "StartBit": 58,
            "TransmissionRate": "500 ms",
            "Units": "binary"
        },
        3289: {
            "Acronym": "ETC7",
            "DataRange": "-125 to 125",
            "EndBit": 31,
            "Name": "Transmission Requested Gear Feedback",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3289,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "100 ms",
            "Units": "gear value"
        },
        3294: {
            "Acronym": "DM21",
            "DataRange": "0.0 to 39926.193605 mile",
            "EndBit": 31,
            "Name": "Distance Since Diagnostic Trouble Codes Cleared",
            "Offset": 0.0,
            "OperationalHigh": 39926.193605,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.621371,
            "SPN": 3294,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "miles"
        },
        3295: {
            "Acronym": "DM21",
            "DataRange": "0 to 64,255 mins",
            "EndBit": 47,
            "Name": "Minutes Accumulated While MIL Activated",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3295,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "minutes"
        },
        3296: {
            "Acronym": "DM21",
            "DataRange": "0 to 64,255 mins",
            "EndBit": 63,
            "Name": "Time Since Diagnostic Trouble Codes Cleared",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3296,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "On request",
            "Units": "minutes"
        },
        3297: {
            "Acronym": "DM24",
            "DataRange": "0 to 524,287",
            "EndBit": 18,
            "Name": "SPN Supported",
            "Offset": 0.0,
            "OperationalHigh": 524287.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 3297,
            "SPNLength": 19,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "binary"
        },
        3300: {
            "Acronym": "DM25",
            "DataRange": "0 to 255",
            "EndBit": 7,
            "Name": "Expanded Freeze Frame Length",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3300,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        3301: {
            "Acronym": "DM26",
            "DataRange": "0 to 64,255 s",
            "EndBit": 15,
            "Name": "Time Since Engine Start",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3301,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "Seconds"
        },
        3302: {
            "Acronym": "DM26",
            "DataRange": "0 to 250",
            "EndBit": 23,
            "Name": "Number of Warm-Ups Since DTCs Cleared",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3302,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "count"
        },
        3303: {
            "Acronym": "DM26",
            "DataRange": "bit-mapped",
            "EndBit": 31,
            "Name": "Continuously Monitored Systems Enabled/Completed Status",
            "Offset": 0.0,
            "OperationalHigh": -1,
            "OperationalLow": -1,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 8,
            "SPN": 3303,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "On request",
            "Units": "bit-mapped"
        },
        3304: {
            "Acronym": "DM26",
            "DataRange": "bit-mapped",
            "EndBit": 47,
            "Name": "Non-Continuously Monitored Systems Enabled Status",
            "Offset": 0.0,
            "OperationalHigh": -1,
            "OperationalLow": -1,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 16,
            "SPN": 3304,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "bit-mapped"
        },
        3305: {
            "Acronym": "DM26",
            "DataRange": "bit-mapped",
            "EndBit": 63,
            "Name": "Non-Continuously Monitored Systems Complete Status",
            "Offset": 0.0,
            "OperationalHigh": -1,
            "OperationalLow": -1,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 16,
            "SPN": 3305,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "On request",
            "Units": "bit-mapped"
        },
        3307: {
            "Acronym": "FWSS2",
            "DataRange": "0 to 15",
            "EndBit": 3,
            "Name": "Fifth Wheel Error Status",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3307,
            "SPNLength": 4,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        3308: {
            "Acronym": "FWSS1",
            "DataRange": "0 to 321,275 N",
            "EndBit": 15,
            "Name": "Fifth Wheel Vertical Force",
            "Offset": 0.0,
            "OperationalHigh": 321275.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 5,
            "SPN": 3308,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "50 ms",
            "Units": "N"
        },
        3309: {
            "Acronym": "FWSS1",
            "DataRange": "-320,000 to 322,550 N",
            "EndBit": 31,
            "Name": "Fifth Wheel Drawbar Force",
            "Offset": -320000.0,
            "OperationalHigh": 322550.0,
            "OperationalLow": -320000.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 10,
            "SPN": 3309,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "50 ms",
            "Units": "N"
        },
        3310: {
            "Acronym": "FWSS1",
            "DataRange": "-320,000 to 322,550 N",
            "EndBit": 47,
            "Name": "Fifth Wheel Roll Moment ",
            "Offset": -320000.0,
            "OperationalHigh": 322550.0,
            "OperationalLow": -320000.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 10,
            "SPN": 3310,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "50 ms",
            "Units": "N"
        },
        3311: {
            "Acronym": "FWSS2",
            "DataRange": "0 to 2500 mm (0 to 2.500 m)",
            "EndBit": 15,
            "Name": "Fifth Wheel Slider Position",
            "Offset": 0.0,
            "OperationalHigh": 25000.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 10,
            "SPN": 3311,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "On request",
            "Units": "mm"
        },
        3312: {
            "Acronym": "FWSS2",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Fifth Wheel Lock Ready to Couple Indicator",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3312,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        3313: {
            "Acronym": "FWSS2",
            "DataRange": "0 to 3",
            "EndBit": 7,
            "Name": "Fifth Wheel Lock Couple Status Indicator",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3313,
            "SPNLength": 2,
            "StartBit": 6,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        3314: {
            "Acronym": "CM3",
            "DataRange": "0 to 3",
            "EndBit": 9,
            "Name": "Fifth Wheel Release Control",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3314,
            "SPNLength": 2,
            "StartBit": 8,
            "TransmissionRate": "Every 10 s and on change of state but no faster than every 100 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: On change or every 10 s",
            "Units": "bit"
        },
        3315: {
            "Acronym": "CM3",
            "DataRange": "0 to 3",
            "EndBit": 11,
            "Name": "Fifth Wheel Release Control Security Lockout",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3315,
            "SPNLength": 2,
            "StartBit": 10,
            "TransmissionRate": "Every 10 s and on change of state but no faster than every 100 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: On change or every 10 s",
            "Units": "bit"
        },
        3316: {
            "Acronym": "FWSS2",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Fifth Wheel Slider Lock Indicator",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3316,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        3317: {
            "Acronym": "FWSS1",
            "DataRange": "0 to 3",
            "EndBit": 49,
            "Name": "Fifth Wheel Roll Warning Indicator",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3317,
            "SPNLength": 2,
            "StartBit": 48,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        3318: {
            "Acronym": "SSI",
            "DataRange": "-64 to 64.51 deg",
            "EndBit": 15,
            "Name": "Pitch Angle",
            "Offset": -64.0,
            "OperationalHigh": 64.51,
            "OperationalLow": -64.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.002,
            "SPN": 3318,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "10 ms",
            "Units": "deg"
        },
        3319: {
            "Acronym": "SSI",
            "DataRange": "-64 to 64.51 deg",
            "EndBit": 31,
            "Name": "Roll Angle",
            "Offset": -64.0,
            "OperationalHigh": 64.51,
            "OperationalLow": -64.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.002,
            "SPN": 3319,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "10 ms",
            "Units": "deg"
        },
        3322: {
            "Acronym": "SSI",
            "DataRange": "-64 to 64.51 deg/sec",
            "EndBit": 47,
            "Name": "Pitch Rate",
            "Offset": -64.0,
            "OperationalHigh": 64.51,
            "OperationalLow": -64.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.002,
            "SPN": 3322,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "10 ms",
            "Units": "deg/sec"
        },
        3323: {
            "Acronym": "SSI",
            "DataRange": "0 to 3",
            "EndBit": 49,
            "Name": "Pitch Angle Figure of Merit",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3323,
            "SPNLength": 2,
            "StartBit": 48,
            "TransmissionRate": "10 ms",
            "Units": "bit"
        },
        3324: {
            "Acronym": "SSI",
            "DataRange": "0 to 3",
            "EndBit": 51,
            "Name": "Roll Angle Figure of Merit",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3324,
            "SPNLength": 2,
            "StartBit": 50,
            "TransmissionRate": "10 ms",
            "Units": "bit"
        },
        3325: {
            "Acronym": "SSI",
            "DataRange": "0 to 3",
            "EndBit": 53,
            "Name": "Pitch Rate Figure of Merit",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3325,
            "SPNLength": 2,
            "StartBit": 52,
            "TransmissionRate": "10 ms",
            "Units": "bit"
        },
        3326: {
            "Acronym": "SSI",
            "DataRange": "0 to 3",
            "EndBit": 55,
            "Name": "Pitch and Roll Compensated",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3326,
            "SPNLength": 2,
            "StartBit": 54,
            "TransmissionRate": "10 ms",
            "Units": "bit"
        },
        3327: {
            "Acronym": "SSI",
            "DataRange": "0 to 125 ms",
            "EndBit": 63,
            "Name": "Roll and Pitch Measurement Latency",
            "Offset": 0.0,
            "OperationalHigh": 125.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.5,
            "SPN": 3327,
            "SPNLength": 8,
            "StartBit": 56,
            "TransmissionRate": "10 ms",
            "Units": "Milliseconds"
        },
        3328: {
            "Acronym": "PropA2",
            "DataRange": "Manufacturer Determined",
            "EndBit": 63,
            "Name": "Manufacturer Specific Information (PropA2_PDU1)",
            "Offset": 0.0,
            "OperationalHigh": -1,
            "OperationalLow": -1,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": -1,
            "SPN": 3328,
            "SPNLength": 14280,
            "StartBit": 0,
            "TransmissionRate": "Per user requirements",
            "Units": "Manufacturer Determined"
        },
        3329: {
            "Acronym": "RCMS",
            "DataRange": "0 to 250",
            "EndBit": 7,
            "Name": "Message Selection Control",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3329,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "As Needed",
            "Units": "count"
        },
        3330: {
            "Acronym": "RCMS",
            "DataRange": "0 to 16,777,215",
            "EndBit": 31,
            "Name": "PGN of Configurable Message Desired",
            "Offset": 0.0,
            "OperationalHigh": 16777215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 3330,
            "SPNLength": 24,
            "StartBit": 8,
            "TransmissionRate": "As Needed",
            "Units": "binary"
        },
        3331: {
            "Acronym": "BI",
            "DataRange": "-200 to 301.99 deg",
            "EndBit": 31,
            "Name": "Blade Rotation Angle",
            "Offset": -200.0,
            "OperationalHigh": 301.99,
            "OperationalLow": -200.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3331,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "50 ms",
            "Units": "deg"
        },
        3332: {
            "Acronym": "BI",
            "DataRange": "0 to 3",
            "EndBit": 43,
            "Name": "Blade Rotation Angle Figure of Merit",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3332,
            "SPNLength": 2,
            "StartBit": 42,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        3334: {
            "Acronym": "LOI",
            "DataRange": "0 to 15",
            "EndBit": 15,
            "Name": "Left Blade Control Mode Operator Control",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3334,
            "SPNLength": 4,
            "StartBit": 12,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3335: {
            "Acronym": "LOI",
            "DataRange": "0 to 15",
            "EndBit": 19,
            "Name": "Right Blade Control Mode Operator Control",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3335,
            "SPNLength": 4,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3336: {
            "Acronym": "LOI",
            "DataRange": "0 to 15",
            "EndBit": 23,
            "Name": "Left Desired Blade Offset Operator Control",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3336,
            "SPNLength": 4,
            "StartBit": 20,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3337: {
            "Acronym": "LOI",
            "DataRange": "0 to 15",
            "EndBit": 27,
            "Name": "Right Desired Blade Offset Operator Control",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3337,
            "SPNLength": 4,
            "StartBit": 24,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3338: {
            "Acronym": "LOI",
            "DataRange": "0 to 15",
            "EndBit": 31,
            "Name": "Side-shift Blade Control Mode Operator Control",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3338,
            "SPNLength": 4,
            "StartBit": 28,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3339: {
            "Acronym": "LOI",
            "DataRange": "0 to 15",
            "EndBit": 35,
            "Name": "Side-shift Desired Blade Offset Operator Control",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3339,
            "SPNLength": 4,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3340: {
            "Acronym": "EFL/P4",
            "DataRange": "0.0 to 72.518869 psi",
            "EndBit": 7,
            "Name": "Engine Charge Air Cooler 1 Intake Pressure",
            "Offset": 0.0,
            "OperationalHigh": 72.518869,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.290075476,
            "SPN": 3340,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        3341: {
            "Acronym": "EFL/P4",
            "DataRange": "0.0 to 72.518869 psi",
            "EndBit": 15,
            "Name": "Engine Charge Air Cooler 2 Intake Pressure",
            "Offset": 0.0,
            "OperationalHigh": 72.518869,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.290075476,
            "SPN": 3341,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        3342: {
            "Acronym": "EFL/P4",
            "DataRange": "-1.015264166 to 58.450208414 psi",
            "EndBit": 23,
            "Name": "Engine Coolant Pump Differential Pressure",
            "Offset": -1.015264166,
            "OperationalHigh": 58.450208414,
            "OperationalLow": -1.015264166,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.23786189032,
            "SPN": 3342,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        3343: {
            "Acronym": "EFL/P4",
            "DataRange": "0 to 257,020 rpm",
            "EndBit": 39,
            "Name": "Engine Centrifugal Oil Filter speed",
            "Offset": 0.0,
            "OperationalHigh": 257020.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 4,
            "SPN": 3343,
            "SPNLength": 16,
            "StartBit": 24,
            "TransmissionRate": "500 ms",
            "Units": "rpm"
        },
        3344: {
            "Acronym": "EC1",
            "DataRange": "0 to 255",
            "EndBit": 279,
            "Name": "Support Variable Rate TSC1 Message",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 39,
            "Resolution": 1,
            "SPN": 3344,
            "SPNLength": 8,
            "StartBit": 272,
            "TransmissionRate": "Every 5 s and on change of torque/speed points of more than 10% since last transmission but no faster than every 500 ms",
            "Units": "bit"
        },
        3345: {
            "Acronym": "EC1",
            "DataRange": "0 to 255",
            "EndBit": 287,
            "Name": "Support TSC1 Control Purpose Group 1",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 39,
            "Resolution": 1,
            "SPN": 3345,
            "SPNLength": 8,
            "StartBit": 280,
            "TransmissionRate": "Every 5 s and on change of torque/speed points of more than 10% since last transmission but no faster than every 500 ms",
            "Units": "bit"
        },
        3346: {
            "Acronym": "EC1",
            "DataRange": "0 to 255",
            "EndBit": 295,
            "Name": "Support TSC1 Control Purpose Group 2",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 39,
            "Resolution": 1,
            "SPN": 3346,
            "SPNLength": 8,
            "StartBit": 288,
            "TransmissionRate": "Every 5 s and on change of torque/speed points of more than 10% since last transmission but no faster than every 500 ms",
            "Units": "bit"
        },
        3347: {
            "Acronym": "EC1",
            "DataRange": "0 to 255",
            "EndBit": 303,
            "Name": "Support TSC1 Control Purpose Group 3",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 39,
            "Resolution": 1,
            "SPN": 3347,
            "SPNLength": 8,
            "StartBit": 296,
            "TransmissionRate": "Every 5 s and on change of torque/speed points of more than 10% since last transmission but no faster than every 500 ms",
            "Units": "bit"
        },
        3348: {
            "Acronym": "EC1",
            "DataRange": "0 to 255",
            "EndBit": 311,
            "Name": "Support TSC1 Control Purpose Group 4",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 39,
            "Resolution": 1,
            "SPN": 3348,
            "SPNLength": 8,
            "StartBit": 304,
            "TransmissionRate": "Every 5 s and on change of torque/speed points of more than 10% since last transmission but no faster than every 500 ms",
            "Units": "bit"
        },
        3349: {
            "Acronym": "TSC1",
            "DataRange": "0 to 7",
            "EndBit": 34,
            "Name": "TSC1 Transmission Rate",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3349,
            "SPNLength": 3,
            "StartBit": 32,
            "TransmissionRate": "To engine: Control Purpose dependent or 10 ms\nTo retarder: 50 ms",
            "Units": "bit"
        },
        3350: {
            "Acronym": "TSC1",
            "DataRange": "0 to 31",
            "EndBit": 39,
            "Name": "TSC1 Control Purpose",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3350,
            "SPNLength": 5,
            "StartBit": 35,
            "TransmissionRate": "To engine: Control Purpose dependent or 10 ms\nTo retarder: 50 ms",
            "Units": "bit"
        },
        3353: {
            "Acronym": "AS",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Alternator 1 Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3353,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "1 sec",
            "Units": "bit"
        },
        3354: {
            "Acronym": "AS",
            "DataRange": "0 to 3",
            "EndBit": 19,
            "Name": "Alternator 2 Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3354,
            "SPNLength": 2,
            "StartBit": 18,
            "TransmissionRate": "1 sec",
            "Units": "bit"
        },
        3355: {
            "Acronym": "AS",
            "DataRange": "0 to 3",
            "EndBit": 21,
            "Name": "Alternator 3 Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3355,
            "SPNLength": 2,
            "StartBit": 20,
            "TransmissionRate": "1 sec",
            "Units": "bit"
        },
        3356: {
            "Acronym": "AS",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Alternator 4 Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3356,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "1 sec",
            "Units": "bit"
        },
        3357: {
            "Acronym": "EEC2",
            "DataRange": "0 to 100 %",
            "EndBit": 55,
            "Name": "Actual Maximum Available Engine - Percent Torque",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 3357,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "50 msec (preferred) or Engine Speed Dependent (if required by application)",
            "Units": "%"
        },
        3358: {
            "Acronym": "EFL/P3",
            "DataRange": "0.0 to 72.518869 psi",
            "EndBit": 23,
            "Name": "Engine Exhaust Gas Recirculation 1 Intake Pressure",
            "Offset": 0.0,
            "OperationalHigh": 72.518869,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.290075476,
            "SPN": 3358,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        3359: {
            "Acronym": "TRF2",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Transmission Oil Filter Restriction Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3359,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        3363: {
            "Acronym": "AT1T1I",
            "DataRange": "0 to 100 %",
            "EndBit": 55,
            "Name": "Aftertreatment 1 Diesel Exhaust Fluid Tank Heater",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 3363,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "1 s",
            "Units": "%"
        },
        3365: {
            "Acronym": "BI",
            "DataRange": "-3,200 to 3,225.5 mm",
            "EndBit": 15,
            "Name": "Relative Blade Height",
            "Offset": -3.2,
            "OperationalHigh": 3225.5,
            "OperationalLow": -3200.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 3365,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "50 ms",
            "Units": "m"
        },
        3366: {
            "Acronym": "BI",
            "DataRange": "0 to 125 ms",
            "EndBit": 39,
            "Name": "Relative Blade Height and Blade Rotation Angle Measurement Latency",
            "Offset": 0.0,
            "OperationalHigh": 125.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.5,
            "SPN": 3366,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "50 ms",
            "Units": "Milliseconds"
        },
        3367: {
            "Acronym": "BI",
            "DataRange": "0 to 3",
            "EndBit": 41,
            "Name": "Relative Blade Height Figure of Merit",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3367,
            "SPNLength": 2,
            "StartBit": 40,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        3368: {
            "Acronym": "WCM1",
            "DataRange": "0 to 255",
            "EndBit": 7,
            "Name": "Network Transceiver Status 1",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3368,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        3369: {
            "Acronym": "WCM1",
            "DataRange": "0 to 255",
            "EndBit": 15,
            "Name": "Network Service Status 1",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3369,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        3370: {
            "Acronym": "WCM1",
            "DataRange": "0 to 255",
            "EndBit": 23,
            "Name": "Network Antenna Status 1",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3370,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        3371: {
            "Acronym": "WCM1",
            "DataRange": "0 to 100 %",
            "EndBit": 31,
            "Name": "Network Signal Strength 1",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 3371,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        3372: {
            "Acronym": "WCM1",
            "DataRange": "0 to 255",
            "EndBit": 39,
            "Name": "Wireless Communication Network Type 1",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3372,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        3375: {
            "Acronym": "VROM",
            "DataRange": "0 to 7",
            "EndBit": 2,
            "Name": "Voltage Regulator Load Compensation Mode",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3375,
            "SPNLength": 3,
            "StartBit": 0,
            "TransmissionRate": "1 sec",
            "Units": "bit"
        },
        3376: {
            "Acronym": "VROM",
            "DataRange": "0 to 7",
            "EndBit": 5,
            "Name": "Voltage Regulator VAr/Power Factor operating mode",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3376,
            "SPNLength": 3,
            "StartBit": 3,
            "TransmissionRate": "1 sec",
            "Units": "bit"
        },
        3377: {
            "Acronym": "VROM",
            "DataRange": "0 to 3",
            "EndBit": 7,
            "Name": "Voltage Regulator Underfrequency Compensation enabled",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3377,
            "SPNLength": 2,
            "StartBit": 6,
            "TransmissionRate": "1 sec",
            "Units": "bit"
        },
        3378: {
            "Acronym": "VROM",
            "DataRange": "0 to 3",
            "EndBit": 9,
            "Name": "Voltage Regulator Soft Start State",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3378,
            "SPNLength": 2,
            "StartBit": 8,
            "TransmissionRate": "1 sec",
            "Units": "bit"
        },
        3379: {
            "Acronym": "VROM",
            "DataRange": "0 to 3",
            "EndBit": 11,
            "Name": "Voltage Regulator Enabled",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3379,
            "SPNLength": 2,
            "StartBit": 10,
            "TransmissionRate": "1 sec",
            "Units": "bit"
        },
        3380: {
            "Acronym": "VREP",
            "DataRange": "-1606.00 to 1606.75 V",
            "EndBit": 15,
            "Name": "Generator Excitation Field Voltage",
            "Offset": -1606.0,
            "OperationalHigh": 1606.75,
            "OperationalLow": -1606.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05,
            "SPN": 3380,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "Volts"
        },
        3381: {
            "Acronym": "VREP",
            "DataRange": "0 to 3212.75 A",
            "EndBit": 31,
            "Name": "Generator Excitation Field Current",
            "Offset": 0.0,
            "OperationalHigh": 3212.75,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05,
            "SPN": 3381,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "Amps"
        },
        3382: {
            "Acronym": "VREP",
            "DataRange": "-100 to 100 %",
            "EndBit": 47,
            "Name": "Generator Output Voltage Bias Percentage",
            "Offset": -3212.75,
            "OperationalHigh": 100.0,
            "OperationalLow": -100.0,
            "OperationalRange": "-100 to 100%",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 3382,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        3383: {
            "Acronym": "RGTACRP",
            "DataRange": "-2,000,000,000 to +2,211,081,215 VA",
            "EndBit": 31,
            "Name": "Requested Generator Total AC Reactive Power",
            "Offset": -2000000000.0,
            "OperationalHigh": 2211081215.0,
            "OperationalLow": -2000000000.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3383,
            "SPNLength": 32,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "VA"
        },
        3384: {
            "Acronym": "RGTACRP",
            "DataRange": "-1.00000 to +2.921814",
            "EndBit": 47,
            "Name": "Requested Generator Overall Power Factor",
            "Offset": -1.0,
            "OperationalHigh": 1.0,
            "OperationalLow": -1.0,
            "OperationalRange": "-1 to 1",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3384,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": ""
        },
        3385: {
            "Acronym": "RGTACRP",
            "DataRange": "0 to 3",
            "EndBit": 49,
            "Name": "Requested Generator Overall Power Factor Lagging",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3385,
            "SPNLength": 2,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3386: {
            "Acronym": "RGAAC",
            "DataRange": "0 to 32,899,071 V",
            "EndBit": 31,
            "Name": "Requested Generator Average Line-Line AC RMS Voltage",
            "Offset": 0.0,
            "OperationalHigh": 32899071.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3386,
            "SPNLength": 32,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "Volts"
        },
        3387: {
            "Acronym": "CCS",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Engine Cylinder 1 Combustion Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3387,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "Engine speed dependent when there is no combustion, once every 5 seconds otherwise",
            "Units": "bit"
        },
        3388: {
            "Acronym": "CCS",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Engine Cylinder 2 Combustion Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3388,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "Engine speed dependent when there is no combustion, once every 5 seconds otherwise",
            "Units": "bit"
        },
        3389: {
            "Acronym": "CCS",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Engine Cylinder 3 Combustion Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3389,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "Engine speed dependent when there is no combustion, once every 5 seconds otherwise",
            "Units": "bit"
        },
        3390: {
            "Acronym": "CCS",
            "DataRange": "0 to 3",
            "EndBit": 7,
            "Name": "Engine Cylinder 4 Combustion Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3390,
            "SPNLength": 2,
            "StartBit": 6,
            "TransmissionRate": "Engine speed dependent when there is no combustion, once every 5 seconds otherwise",
            "Units": "bit"
        },
        3391: {
            "Acronym": "CCS",
            "DataRange": "0 to 3",
            "EndBit": 9,
            "Name": "Engine Cylinder 5 Combustion Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3391,
            "SPNLength": 2,
            "StartBit": 8,
            "TransmissionRate": "Engine speed dependent when there is no combustion, once every 5 seconds otherwise",
            "Units": "bit"
        },
        3392: {
            "Acronym": "CCS",
            "DataRange": "0 to 3",
            "EndBit": 11,
            "Name": "Engine Cylinder 6 Combustion Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3392,
            "SPNLength": 2,
            "StartBit": 10,
            "TransmissionRate": "Engine speed dependent when there is no combustion, once every 5 seconds otherwise",
            "Units": "bit"
        },
        3393: {
            "Acronym": "CCS",
            "DataRange": "0 to 3",
            "EndBit": 13,
            "Name": "Engine Cylinder 7 Combustion Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3393,
            "SPNLength": 2,
            "StartBit": 12,
            "TransmissionRate": "Engine speed dependent when there is no combustion, once every 5 seconds otherwise",
            "Units": "bit"
        },
        3394: {
            "Acronym": "CCS",
            "DataRange": "0 to 3",
            "EndBit": 15,
            "Name": "Engine Cylinder 8 Combustion Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3394,
            "SPNLength": 2,
            "StartBit": 14,
            "TransmissionRate": "Engine speed dependent when there is no combustion, once every 5 seconds otherwise",
            "Units": "bit"
        },
        3395: {
            "Acronym": "CCS",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Engine Cylinder 9 Combustion Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3395,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "Engine speed dependent when there is no combustion, once every 5 seconds otherwise",
            "Units": "bit"
        },
        3396: {
            "Acronym": "CCS",
            "DataRange": "0 to 3",
            "EndBit": 19,
            "Name": "Engine Cylinder 10 Combustion Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3396,
            "SPNLength": 2,
            "StartBit": 18,
            "TransmissionRate": "Engine speed dependent when there is no combustion, once every 5 seconds otherwise",
            "Units": "bit"
        },
        3397: {
            "Acronym": "CCS",
            "DataRange": "0 to 3",
            "EndBit": 21,
            "Name": "Engine Cylinder 11 Combustion Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3397,
            "SPNLength": 2,
            "StartBit": 20,
            "TransmissionRate": "Engine speed dependent when there is no combustion, once every 5 seconds otherwise",
            "Units": "bit"
        },
        3398: {
            "Acronym": "CCS",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Engine Cylinder 12 Combustion Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3398,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "Engine speed dependent when there is no combustion, once every 5 seconds otherwise",
            "Units": "bit"
        },
        3399: {
            "Acronym": "CCS",
            "DataRange": "0 to 3",
            "EndBit": 25,
            "Name": "Engine Cylinder 13 Combustion Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3399,
            "SPNLength": 2,
            "StartBit": 24,
            "TransmissionRate": "Engine speed dependent when there is no combustion, once every 5 seconds otherwise",
            "Units": "bit"
        },
        3400: {
            "Acronym": "CCS",
            "DataRange": "0 to 3",
            "EndBit": 27,
            "Name": "Engine Cylinder 14 Combustion Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3400,
            "SPNLength": 2,
            "StartBit": 26,
            "TransmissionRate": "Engine speed dependent when there is no combustion, once every 5 seconds otherwise",
            "Units": "bit"
        },
        3401: {
            "Acronym": "CCS",
            "DataRange": "0 to 3",
            "EndBit": 29,
            "Name": "Engine Cylinder 15 Combustion Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3401,
            "SPNLength": 2,
            "StartBit": 28,
            "TransmissionRate": "Engine speed dependent when there is no combustion, once every 5 seconds otherwise",
            "Units": "bit"
        },
        3402: {
            "Acronym": "CCS",
            "DataRange": "0 to 3",
            "EndBit": 31,
            "Name": "Engine Cylinder 16 Combustion Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3402,
            "SPNLength": 2,
            "StartBit": 30,
            "TransmissionRate": "Engine speed dependent when there is no combustion, once every 5 seconds otherwise",
            "Units": "bit"
        },
        3403: {
            "Acronym": "CCS",
            "DataRange": "0 to 3",
            "EndBit": 33,
            "Name": "Engine Cylinder 17 Combustion Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3403,
            "SPNLength": 2,
            "StartBit": 32,
            "TransmissionRate": "Engine speed dependent when there is no combustion, once every 5 seconds otherwise",
            "Units": "bit"
        },
        3404: {
            "Acronym": "CCS",
            "DataRange": "0 to 3",
            "EndBit": 35,
            "Name": "Engine Cylinder 18 Combustion Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3404,
            "SPNLength": 2,
            "StartBit": 34,
            "TransmissionRate": "Engine speed dependent when there is no combustion, once every 5 seconds otherwise",
            "Units": "bit"
        },
        3405: {
            "Acronym": "CCS",
            "DataRange": "0 to 3",
            "EndBit": 37,
            "Name": "Engine Cylinder 19 Combustion Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3405,
            "SPNLength": 2,
            "StartBit": 36,
            "TransmissionRate": "Engine speed dependent when there is no combustion, once every 5 seconds otherwise",
            "Units": "bit"
        },
        3406: {
            "Acronym": "CCS",
            "DataRange": "0 to 3",
            "EndBit": 39,
            "Name": "Engine Cylinder 20 Combustion Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3406,
            "SPNLength": 2,
            "StartBit": 38,
            "TransmissionRate": "Engine speed dependent when there is no combustion, once every 5 seconds otherwise",
            "Units": "bit"
        },
        3407: {
            "Acronym": "CCS",
            "DataRange": "0 to 3",
            "EndBit": 41,
            "Name": "Engine Cylinder 21 Combustion Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3407,
            "SPNLength": 2,
            "StartBit": 40,
            "TransmissionRate": "Engine speed dependent when there is no combustion, once every 5 seconds otherwise",
            "Units": "bit"
        },
        3408: {
            "Acronym": "CCS",
            "DataRange": "0 to 3",
            "EndBit": 43,
            "Name": "Engine Cylinder 22 Combustion Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3408,
            "SPNLength": 2,
            "StartBit": 42,
            "TransmissionRate": "Engine speed dependent when there is no combustion, once every 5 seconds otherwise",
            "Units": "bit"
        },
        3409: {
            "Acronym": "CCS",
            "DataRange": "0 to 3",
            "EndBit": 45,
            "Name": "Engine Cylinder 23 Combustion Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3409,
            "SPNLength": 2,
            "StartBit": 44,
            "TransmissionRate": "Engine speed dependent when there is no combustion, once every 5 seconds otherwise",
            "Units": "bit"
        },
        3410: {
            "Acronym": "CCS",
            "DataRange": "0 to 3",
            "EndBit": 47,
            "Name": "Engine Cylinder 24 Combustion Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3410,
            "SPNLength": 2,
            "StartBit": 46,
            "TransmissionRate": "Engine speed dependent when there is no combustion, once every 5 seconds otherwise",
            "Units": "bit"
        },
        3411: {
            "Acronym": "DC1",
            "DataRange": "0 to 3",
            "EndBit": 7,
            "Name": "Status 2 of doors",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3411,
            "SPNLength": 2,
            "StartBit": 6,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3412: {
            "Acronym": "DC2",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Lock Status of Door 1",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3412,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3413: {
            "Acronym": "DC2",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Open Status of Door 1",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3413,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3414: {
            "Acronym": "DC2",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Enable Status of Door 1",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3414,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3415: {
            "Acronym": "DC2",
            "DataRange": "0 to 3",
            "EndBit": 7,
            "Name": "Lock Status of Door 2",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3415,
            "SPNLength": 2,
            "StartBit": 6,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3416: {
            "Acronym": "DC2",
            "DataRange": "0 to 3",
            "EndBit": 9,
            "Name": "Open Status of Door 2",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3416,
            "SPNLength": 2,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3417: {
            "Acronym": "DC2",
            "DataRange": "0 to 3",
            "EndBit": 11,
            "Name": "Enable Status of Door 2",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3417,
            "SPNLength": 2,
            "StartBit": 10,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3418: {
            "Acronym": "DC2",
            "DataRange": "0 to 3",
            "EndBit": 13,
            "Name": "Lock Status of Door 3",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3418,
            "SPNLength": 2,
            "StartBit": 12,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3419: {
            "Acronym": "DC2",
            "DataRange": "0 to 3",
            "EndBit": 15,
            "Name": "Open Status of Door 3",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3419,
            "SPNLength": 2,
            "StartBit": 14,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3420: {
            "Acronym": "DC2",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Enable Status of Door 3",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3420,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3421: {
            "Acronym": "DC2",
            "DataRange": "0 to 3",
            "EndBit": 19,
            "Name": "Lock Status of Door 4",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3421,
            "SPNLength": 2,
            "StartBit": 18,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3422: {
            "Acronym": "DC2",
            "DataRange": "0 to 3",
            "EndBit": 21,
            "Name": "Open Status of Door 4",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3422,
            "SPNLength": 2,
            "StartBit": 20,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3423: {
            "Acronym": "DC2",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Enable Status of Door 4",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3423,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3424: {
            "Acronym": "DC2",
            "DataRange": "0 to 3",
            "EndBit": 25,
            "Name": "Lock Status of Door 5",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3424,
            "SPNLength": 2,
            "StartBit": 24,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3425: {
            "Acronym": "DC2",
            "DataRange": "0 to 3",
            "EndBit": 27,
            "Name": "Open Status of Door 5",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3425,
            "SPNLength": 2,
            "StartBit": 26,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3426: {
            "Acronym": "DC2",
            "DataRange": "0 to 3",
            "EndBit": 29,
            "Name": "Enable Status of Door 5",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3426,
            "SPNLength": 2,
            "StartBit": 28,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3427: {
            "Acronym": "DC2",
            "DataRange": "0 to 3",
            "EndBit": 31,
            "Name": "Lock Status of Door 6",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3427,
            "SPNLength": 2,
            "StartBit": 30,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3428: {
            "Acronym": "DC2",
            "DataRange": "0 to 3",
            "EndBit": 33,
            "Name": "Open Status of Door 6",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3428,
            "SPNLength": 2,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3429: {
            "Acronym": "DC2",
            "DataRange": "0 to 3",
            "EndBit": 35,
            "Name": "Enable Status of Door 6",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3429,
            "SPNLength": 2,
            "StartBit": 34,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3430: {
            "Acronym": "DC2",
            "DataRange": "0 to 3",
            "EndBit": 37,
            "Name": "Lock Status of Door 7",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3430,
            "SPNLength": 2,
            "StartBit": 36,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3431: {
            "Acronym": "DC2",
            "DataRange": "0 to 3",
            "EndBit": 39,
            "Name": "Open Status of Door 7",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3431,
            "SPNLength": 2,
            "StartBit": 38,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3432: {
            "Acronym": "DC2",
            "DataRange": "0 to 3",
            "EndBit": 41,
            "Name": "Enable Status of Door 7",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3432,
            "SPNLength": 2,
            "StartBit": 40,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3433: {
            "Acronym": "DC2",
            "DataRange": "0 to 3",
            "EndBit": 43,
            "Name": "Lock Status of Door 8",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3433,
            "SPNLength": 2,
            "StartBit": 42,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3434: {
            "Acronym": "DC2",
            "DataRange": "0 to 3",
            "EndBit": 45,
            "Name": "Open Status of Door 8",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3434,
            "SPNLength": 2,
            "StartBit": 44,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3435: {
            "Acronym": "DC2",
            "DataRange": "0 to 3",
            "EndBit": 47,
            "Name": "Enable Status of Door 8",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3435,
            "SPNLength": 2,
            "StartBit": 46,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3436: {
            "Acronym": "DC2",
            "DataRange": "0 to 3",
            "EndBit": 49,
            "Name": "Lock Status of Door 9",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3436,
            "SPNLength": 2,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3437: {
            "Acronym": "DC2",
            "DataRange": "0 to 3",
            "EndBit": 51,
            "Name": "Open Status of Door 9",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3437,
            "SPNLength": 2,
            "StartBit": 50,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3438: {
            "Acronym": "DC2",
            "DataRange": "0 to 3",
            "EndBit": 53,
            "Name": "Enable Status of Door 9",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3438,
            "SPNLength": 2,
            "StartBit": 52,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3439: {
            "Acronym": "DC2",
            "DataRange": "0 to 3",
            "EndBit": 55,
            "Name": "Lock Status of Door 10",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3439,
            "SPNLength": 2,
            "StartBit": 54,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3440: {
            "Acronym": "DC2",
            "DataRange": "0 to 3",
            "EndBit": 57,
            "Name": "Open Status of Door 10",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3440,
            "SPNLength": 2,
            "StartBit": 56,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3441: {
            "Acronym": "DC2",
            "DataRange": "0 to 3",
            "EndBit": 59,
            "Name": "Enable Status of Door 10",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3441,
            "SPNLength": 2,
            "StartBit": 58,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3442: {
            "Acronym": "WCM2",
            "DataRange": "0 to 255",
            "EndBit": 7,
            "Name": "Network Transceiver Status 2",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3442,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        3443: {
            "Acronym": "WCM2",
            "DataRange": "0 to 255",
            "EndBit": 15,
            "Name": "Network Service Status 2",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3443,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        3444: {
            "Acronym": "WCM2",
            "DataRange": "0 to 255",
            "EndBit": 23,
            "Name": "Network Antenna Status 2",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3444,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        3445: {
            "Acronym": "WCM2",
            "DataRange": "0 to 100 %",
            "EndBit": 31,
            "Name": "Network Signal Strength 2",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 3445,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        3446: {
            "Acronym": "WCM2",
            "DataRange": "0 to 255",
            "EndBit": 39,
            "Name": "Wireless Communication Network Type 2",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3446,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        3447: {
            "Acronym": "PTO",
            "DataRange": "0 to 3",
            "EndBit": 59,
            "Name": "Remote PTO Governor Preprogrammed Speed Control Switch #2",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3447,
            "SPNLength": 2,
            "StartBit": 58,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3448: {
            "Acronym": "PTO",
            "DataRange": "0 to 3",
            "EndBit": 61,
            "Name": "Auxiliary Input Ignore Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3448,
            "SPNLength": 2,
            "StartBit": 60,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3452: {
            "Acronym": "PTODE",
            "DataRange": "0 to 3",
            "EndBit": 7,
            "Name": "Enable Switch - Transmission input shaft PTO 1",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3452,
            "SPNLength": 2,
            "StartBit": 6,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3453: {
            "Acronym": "PTODE",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Enable Switch - Transmission input shaft PTO 2",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3453,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3454: {
            "Acronym": "PTODE",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Enable Switch - Transmission output shaft PTO",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3454,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3455: {
            "Acronym": "PTODE",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Enable Switch - Transfer case output shaft PTO",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3455,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3456: {
            "Acronym": "PTODE",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Engagement Consent - Transmission input shaft PTO 1",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3456,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3457: {
            "Acronym": "PTODE",
            "DataRange": "0 to 3",
            "EndBit": 21,
            "Name": "Engagement Consent - Transmission input shaft PTO 2",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3457,
            "SPNLength": 2,
            "StartBit": 20,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3458: {
            "Acronym": "PTODE",
            "DataRange": "0 to 3",
            "EndBit": 19,
            "Name": "Engagement Consent - Transmission output shaft PTO",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3458,
            "SPNLength": 2,
            "StartBit": 18,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3459: {
            "Acronym": "PTODE",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Engagement Consent - Transfer case output shaft PTO",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3459,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3460: {
            "Acronym": "PTODE",
            "DataRange": "0 to 3",
            "EndBit": 39,
            "Name": "Engagement Status - Transmission input shaft PTO 1",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3460,
            "SPNLength": 2,
            "StartBit": 38,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3461: {
            "Acronym": "PTODE",
            "DataRange": "0 to 3",
            "EndBit": 37,
            "Name": "Engagement Status - Transmission input shaft PTO 2",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3461,
            "SPNLength": 2,
            "StartBit": 36,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3462: {
            "Acronym": "PTODE",
            "DataRange": "0 to 3",
            "EndBit": 35,
            "Name": "Engagement Status - Transmission output shaft PTO",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3462,
            "SPNLength": 2,
            "StartBit": 34,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3463: {
            "Acronym": "PTODE",
            "DataRange": "0 to 3",
            "EndBit": 33,
            "Name": "Engagement Status - Transfer case output shaft PTO",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3463,
            "SPNLength": 2,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3464: {
            "Acronym": "TFAC",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 15,
            "Name": "Engine Throttle Actuator 1 Control Command",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 100%",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 3464,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "50 msec (preferred) or Engine Speed Dependent (if required by application)",
            "Units": "%"
        },
        3465: {
            "Acronym": "TFAC",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 31,
            "Name": "Engine Throttle Actuator 2 Control Command",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 100%",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 3465,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "50 msec (preferred) or Engine Speed Dependent (if required by application)",
            "Units": "%"
        },
        3466: {
            "Acronym": "GFI",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 15,
            "Name": "Engine Fuel Valve 2 Intake Absolute Pressure",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 3466,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        3467: {
            "Acronym": "GFI",
            "DataRange": "0 to 3212.75 kg/h",
            "EndBit": 31,
            "Name": "Engine Gas 2 Mass Flow Rate",
            "Offset": 0.0,
            "OperationalHigh": 3212.75,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05,
            "SPN": 3467,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "500 ms",
            "Units": "kg/h"
        },
        3468: {
            "Acronym": "GFI",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 39,
            "Name": "Engine Fuel Temperature 2",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 3468,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "500 ms",
            "Units": "deg F"
        },
        3469: {
            "Acronym": "GFI",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 63,
            "Name": "Engine Fuel  Valve 2 Outlet Absolute Pressure",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 3469,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        3470: {
            "Acronym": "EEC6",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 15,
            "Name": "Engine Turbocharger Compressor Bypass Actuator 1 Command",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 100%",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 3470,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "100 msec (preferred) or Engine Speed Dependent (if required by application)",
            "Units": "%"
        },
        3479: {
            "Acronym": "AT1FC1",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 47,
            "Name": "Aftertreatment 1 Fuel Pressure 1 Control",
            "Offset": 0.0,
            "OperationalHigh": 160.6375,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 3479,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "500 ms",
            "Units": "%"
        },
        3480: {
            "Acronym": "AT1FC1",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 15,
            "Name": "Aftertreatment 1 Fuel Pressure 1",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 3480,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        3481: {
            "Acronym": "AT1FC1",
            "DataRange": "0 to 3,212.75 L/h",
            "EndBit": 31,
            "Name": "Aftertreatment 1 Fuel Rate",
            "Offset": 0.0,
            "OperationalHigh": 3212.75,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05,
            "SPN": 3481,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "500 ms",
            "Units": "l/h"
        },
        3482: {
            "Acronym": "AT1FC1",
            "DataRange": "0 to 3",
            "EndBit": 55,
            "Name": "Aftertreatment 1 Fuel Enable Actuator",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3482,
            "SPNLength": 2,
            "StartBit": 54,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        3483: {
            "Acronym": "AT1FC1",
            "DataRange": "0 to 3",
            "EndBit": 53,
            "Name": "Aftertreatment 1 Regeneration Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3483,
            "SPNLength": 2,
            "StartBit": 52,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        3484: {
            "Acronym": "AT1FC1",
            "DataRange": "0 to 3",
            "EndBit": 51,
            "Name": "Aftertreatment 1 Ignition",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3484,
            "SPNLength": 2,
            "StartBit": 50,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        3485: {
            "Acronym": "AT1AC1",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 15,
            "Name": "Aftertreatment 1 Supply Air Pressure",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 3485,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        3486: {
            "Acronym": "AT1AC1",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 31,
            "Name": "Aftertreatment 1 Purge Air Pressure",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 3486,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        3487: {
            "Acronym": "AT1AC1",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 47,
            "Name": "Aftertreatment 1 Air Pressure Control",
            "Offset": 0.0,
            "OperationalHigh": 160.6375,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 3487,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "500 ms",
            "Units": "%"
        },
        3488: {
            "Acronym": "AT1AC1",
            "DataRange": "0 to 100 %",
            "EndBit": 55,
            "Name": "Aftertreatment 1 Air Pressure Actuator Position",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 3488,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "500 ms",
            "Units": "%"
        },
        3489: {
            "Acronym": "AT1AC1",
            "DataRange": "0 to 3",
            "EndBit": 63,
            "Name": "Aftertreatment 1 Air Enable Actuator",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3489,
            "SPNLength": 2,
            "StartBit": 62,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        3490: {
            "Acronym": "AT1AC1",
            "DataRange": "0 to 3",
            "EndBit": 61,
            "Name": "Aftertreatment 1 Purge Air Actuator",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3490,
            "SPNLength": 2,
            "StartBit": 60,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        3491: {
            "Acronym": "AT1AC1",
            "DataRange": "0 to 3",
            "EndBit": 59,
            "Name": "Aftertreatment 1 Atomization Air Actuator",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3491,
            "SPNLength": 2,
            "StartBit": 58,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        3492: {
            "Acronym": "AT1AC1",
            "DataRange": "0 to 3",
            "EndBit": 57,
            "Name": "Aftertreatment 1 Air System Relay",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3492,
            "SPNLength": 2,
            "StartBit": 56,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        3493: {
            "Acronym": "AT2FC1",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 47,
            "Name": "Aftertreatment 2 Fuel Pressure 1 Control",
            "Offset": 0.0,
            "OperationalHigh": 160.6375,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 3493,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "500 ms",
            "Units": "%"
        },
        3494: {
            "Acronym": "AT2FC1",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 15,
            "Name": "Aftertreatment 2 Fuel Pressure 1",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 3494,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        3495: {
            "Acronym": "AT2FC1",
            "DataRange": "0 to 3,212.75 L/h",
            "EndBit": 31,
            "Name": "Aftertreatment 2 Fuel Rate",
            "Offset": 0.0,
            "OperationalHigh": 3212.75,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05,
            "SPN": 3495,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "500 ms",
            "Units": "l/h"
        },
        3496: {
            "Acronym": "AT2FC1",
            "DataRange": "0 to 3",
            "EndBit": 55,
            "Name": "Aftertreatment 2 Fuel Enable Actuator",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3496,
            "SPNLength": 2,
            "StartBit": 54,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        3497: {
            "Acronym": "AT2FC1",
            "DataRange": "0 to 3",
            "EndBit": 53,
            "Name": "Aftertreatment 2 Regeneration Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3497,
            "SPNLength": 2,
            "StartBit": 52,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        3498: {
            "Acronym": "AT2FC1",
            "DataRange": "0 to 3",
            "EndBit": 51,
            "Name": "Aftertreatment 2 Ignition",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3498,
            "SPNLength": 2,
            "StartBit": 50,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        3499: {
            "Acronym": "AT2AC1",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 15,
            "Name": "Aftertreatment 2 Supply Air Pressure",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 3499,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        3500: {
            "Acronym": "AT2AC1",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 31,
            "Name": "Aftertreatment 2 Purge Air Pressure",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 3500,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        3501: {
            "Acronym": "AT2AC1",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 47,
            "Name": "Aftertreatment 2 Air Pressure Control",
            "Offset": 0.0,
            "OperationalHigh": 160.6375,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 3501,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "500 ms",
            "Units": "%"
        },
        3502: {
            "Acronym": "AT2AC1",
            "DataRange": "0 to 100 %",
            "EndBit": 55,
            "Name": "Aftertreatment 2 Air Pressure Actuator Position",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 3502,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "500 ms",
            "Units": "%"
        },
        3503: {
            "Acronym": "AT2AC1",
            "DataRange": "0 to 3",
            "EndBit": 63,
            "Name": "Aftertreatment 2 Air Enable Actuator",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3503,
            "SPNLength": 2,
            "StartBit": 62,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        3504: {
            "Acronym": "AT2AC1",
            "DataRange": "0 to 3",
            "EndBit": 61,
            "Name": "Aftertreatment 2 Purge Air Actuator",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3504,
            "SPNLength": 2,
            "StartBit": 60,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        3505: {
            "Acronym": "AT2AC1",
            "DataRange": "0 to 3",
            "EndBit": 59,
            "Name": "Aftertreatment 2 Atomization Air Actuator",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3505,
            "SPNLength": 2,
            "StartBit": 58,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        3506: {
            "Acronym": "AT2AC1",
            "DataRange": "0 to 3",
            "EndBit": 57,
            "Name": "Aftertreatment 2 Air System Relay",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3506,
            "SPNLength": 2,
            "StartBit": 56,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        3509: {
            "Acronym": "SEP1",
            "DataRange": "0 to 3212.75 V",
            "EndBit": 15,
            "Name": "Sensor supply voltage 1",
            "Offset": 0.0,
            "OperationalHigh": 3212.75,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05,
            "SPN": 3509,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "1 sec",
            "Units": "Volts"
        },
        3510: {
            "Acronym": "SEP1",
            "DataRange": "0 to 3212.75 V",
            "EndBit": 31,
            "Name": "Sensor supply voltage 2",
            "Offset": 0.0,
            "OperationalHigh": 3212.75,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05,
            "SPN": 3510,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "1 sec",
            "Units": "Volts"
        },
        3511: {
            "Acronym": "SEP1",
            "DataRange": "0 to 3212.75 V",
            "EndBit": 47,
            "Name": "Sensor supply voltage 3",
            "Offset": 0.0,
            "OperationalHigh": 3212.75,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05,
            "SPN": 3511,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "1 sec",
            "Units": "Volts"
        },
        3512: {
            "Acronym": "SEP1",
            "DataRange": "0 to 3212.75 V",
            "EndBit": 63,
            "Name": "Sensor supply voltage 4",
            "Offset": 0.0,
            "OperationalHigh": 3212.75,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05,
            "SPN": 3512,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "1 sec",
            "Units": "Volts"
        },
        3513: {
            "Acronym": "SEP2",
            "DataRange": "0 to 3212.75 V",
            "EndBit": 15,
            "Name": "Sensor supply voltage 5",
            "Offset": 0.0,
            "OperationalHigh": 3212.75,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05,
            "SPN": 3513,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "1 sec",
            "Units": "Volts"
        },
        3514: {
            "Acronym": "SEP2",
            "DataRange": "0 to 3212.75 V",
            "EndBit": 31,
            "Name": "Sensor supply voltage 6",
            "Offset": 0.0,
            "OperationalHigh": 3212.75,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05,
            "SPN": 3514,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "1 sec",
            "Units": "Volts"
        },
        3515: {
            "Acronym": "A1DEFI",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 7,
            "Name": "Aftertreatment 1 Diesel Exhaust Fluid Temperature 2",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 3515,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "1 sec",
            "Units": "deg F"
        },
        3516: {
            "Acronym": "A1DEFI",
            "DataRange": "0 to 62.5 %",
            "EndBit": 15,
            "Name": "Aftertreatment 1 Diesel Exhaust Fluid Concentration",
            "Offset": 0.0,
            "OperationalHigh": 62.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.25,
            "SPN": 3516,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "1 sec",
            "Units": "%"
        },
        3517: {
            "Acronym": "AT1T1I",
            "DataRange": "0 to 6,425.5 mm (0 to 6.4255 m)",
            "EndBit": 31,
            "Name": "Aftertreatment 1 Diesel Exhaust Fluid Tank Level 2",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 3517,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "1 s",
            "Units": "m"
        },
        3518: {
            "Acronym": "A1DEFI",
            "DataRange": "0 to 1250 microSiemens/mm",
            "EndBit": 23,
            "Name": "Aftertreatment 1 Diesel Exhaust Fluid Conductivity",
            "Offset": 0.0,
            "OperationalHigh": 1250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 5,
            "SPN": 3518,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "1 sec",
            "Units": "microSiemens/mm"
        },
        3519: {
            "Acronym": "A1DEFI",
            "DataRange": "0 to 31",
            "EndBit": 28,
            "Name": "Aftertreatment 1 SCR Catalyst Reagent Temperature 2 Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 3519,
            "SPNLength": 5,
            "StartBit": 24,
            "TransmissionRate": "1 sec",
            "Units": "binary"
        },
        3520: {
            "Acronym": "A1DEFI",
            "DataRange": "0 to 31",
            "EndBit": 36,
            "Name": "Aftertreatment 1 SCR Catalyst Reagent Properties Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 3520,
            "SPNLength": 5,
            "StartBit": 32,
            "TransmissionRate": "1 sec",
            "Units": "binary"
        },
        3521: {
            "Acronym": "A1DEFI",
            "DataRange": "0 to 15",
            "EndBit": 43,
            "Name": "Aftertreatment 1 Diesel Exhaust Fluid Property",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3521,
            "SPNLength": 4,
            "StartBit": 40,
            "TransmissionRate": "1 sec",
            "Units": "bit"
        },
        3522: {
            "Acronym": "AT1HI",
            "DataRange": "0 to 2,105,540,607.5 L",
            "EndBit": 31,
            "Name": "Aftertreatment 1 Total Fuel Used",
            "Offset": 0.0,
            "OperationalHigh": 2105540607.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.5,
            "SPN": 3522,
            "SPNLength": 32,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "liters"
        },
        3523: {
            "Acronym": "AT1HI",
            "DataRange": "0 to 4,211,081,215 s",
            "EndBit": 63,
            "Name": "Aftertreatment 1 Total Regeneration Time",
            "Offset": 0.0,
            "OperationalHigh": 4211081215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 3523,
            "SPNLength": 32,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "Seconds"
        },
        3524: {
            "Acronym": "AT1HI",
            "DataRange": "0 to 4,211,081,215 s",
            "EndBit": 95,
            "Name": "Aftertreatment 1 Total Disabled Time",
            "Offset": 0.0,
            "OperationalHigh": 4211081215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 3524,
            "SPNLength": 32,
            "StartBit": 64,
            "TransmissionRate": "On request",
            "Units": "Seconds"
        },
        3525: {
            "Acronym": "AT1HI",
            "DataRange": "0 to 4,211,081,215",
            "EndBit": 127,
            "Name": "Aftertreatment 1 Total Number of Active Regenerations",
            "Offset": 0.0,
            "OperationalHigh": 4211081215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 3525,
            "SPNLength": 32,
            "StartBit": 96,
            "TransmissionRate": "On request",
            "Units": "count"
        },
        3526: {
            "Acronym": "AT2HI",
            "DataRange": "0 to 2,105,540,607.5 L",
            "EndBit": 31,
            "Name": "Aftertreatment 2 Total Fuel Used",
            "Offset": 0.0,
            "OperationalHigh": 2105540607.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.5,
            "SPN": 3526,
            "SPNLength": 32,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "liters"
        },
        3527: {
            "Acronym": "AT2HI",
            "DataRange": "0 to 4,211,081,215 s",
            "EndBit": 63,
            "Name": "Aftertreatment 2 Total Regeneration Time",
            "Offset": 0.0,
            "OperationalHigh": 4211081215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 3527,
            "SPNLength": 32,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "Seconds"
        },
        3528: {
            "Acronym": "AT2HI",
            "DataRange": "0 to 4,211,081,215 s",
            "EndBit": 95,
            "Name": "Aftertreatment 2 Total Disabled Time",
            "Offset": 0.0,
            "OperationalHigh": 4211081215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 3528,
            "SPNLength": 32,
            "StartBit": 64,
            "TransmissionRate": "On request",
            "Units": "Seconds"
        },
        3529: {
            "Acronym": "AT2HI",
            "DataRange": "0 to 4,211,081,215",
            "EndBit": 127,
            "Name": "Aftertreatment 2 Total Number of Active Regenerations",
            "Offset": 0.0,
            "OperationalHigh": 4211081215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 3529,
            "SPNLength": 32,
            "StartBit": 96,
            "TransmissionRate": "On request",
            "Units": "count"
        },
        3532: {
            "Acronym": "AT1T1I",
            "DataRange": "0 to 31",
            "EndBit": 36,
            "Name": "Aftertreatment 1 Diesel Exhaust Fluid Tank Level Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 3532,
            "SPNLength": 5,
            "StartBit": 32,
            "TransmissionRate": "1 s",
            "Units": "binary"
        },
        3533: {
            "Acronym": "TRF2",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Transmission Oil Level Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3533,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        354: {
            "Acronym": "AAI",
            "DataRange": "0 to 100 %",
            "EndBit": 55,
            "Name": "Relative Humidity",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 354,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        3542: {
            "Acronym": "GC1",
            "DataRange": "0 to 15",
            "EndBit": 3,
            "Name": "Requested Engine Control Mode",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3542,
            "SPNLength": 4,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3543: {
            "Acronym": "EOI",
            "DataRange": "0 to 15",
            "EndBit": 3,
            "Name": "Engine Operating State",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3543,
            "SPNLength": 4,
            "StartBit": 0,
            "TransmissionRate": "250 ms",
            "Units": "bit"
        },
        3544: {
            "Acronym": "EOI",
            "DataRange": "0 to 64,255 s",
            "EndBit": 23,
            "Name": "Time Remaining in Engine Operating State",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3544,
            "SPNLength": 16,
            "StartBit": 8,
            "TransmissionRate": "250 ms",
            "Units": "Seconds"
        },
        3545: {
            "Acronym": "ACS",
            "DataRange": "0 to 7",
            "EndBit": 2,
            "Name": "Generator Circuit Breaker Status",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3545,
            "SPNLength": 3,
            "StartBit": 0,
            "TransmissionRate": "250 ms",
            "Units": "bit"
        },
        3546: {
            "Acronym": "ACS",
            "DataRange": "0 to 7",
            "EndBit": 5,
            "Name": "Utility Circuit Breaker Status",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3546,
            "SPNLength": 3,
            "StartBit": 3,
            "TransmissionRate": "250 ms",
            "Units": "bit"
        },
        3547: {
            "Acronym": "ACS",
            "DataRange": "0 to 7",
            "EndBit": 10,
            "Name": "Automatic Transfer Switch Status",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3547,
            "SPNLength": 3,
            "StartBit": 8,
            "TransmissionRate": "250 ms",
            "Units": "bit"
        },
        3548: {
            "Acronym": "EFS",
            "DataRange": "0 to 100 %",
            "EndBit": 31,
            "Name": "Engine Waste Oil Reservoir Level",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 3548,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "0.5 s",
            "Units": "%"
        },
        3549: {
            "Acronym": "EFS",
            "DataRange": "0.0 to 145.037738 psi",
            "EndBit": 39,
            "Name": "Engine Oil-Filter Outlet Pressure",
            "Offset": 0.0,
            "OperationalHigh": 145.037738,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.580150952,
            "SPN": 3549,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "0.5 s",
            "Units": "psi"
        },
        3550: {
            "Acronym": "EFS",
            "DataRange": "0 to 3",
            "EndBit": 41,
            "Name": "Engine Oil Priming Pump Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3550,
            "SPNLength": 2,
            "StartBit": 40,
            "TransmissionRate": "0.5 s",
            "Units": "bit"
        },
        3551: {
            "Acronym": "EFS",
            "DataRange": "0 to 3",
            "EndBit": 43,
            "Name": "Engine Oil Priming State",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3551,
            "SPNLength": 2,
            "StartBit": 42,
            "TransmissionRate": "0.5 s",
            "Units": "bit"
        },
        3552: {
            "Acronym": "EFS",
            "DataRange": "0 to 3",
            "EndBit": 45,
            "Name": "Engine Oil Pre-Heated State",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3552,
            "SPNLength": 2,
            "StartBit": 44,
            "TransmissionRate": "0.5 s",
            "Units": "bit"
        },
        3553: {
            "Acronym": "EFS",
            "DataRange": "0 to 3",
            "EndBit": 47,
            "Name": "Engine Coolant Pre-heated State",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3553,
            "SPNLength": 2,
            "StartBit": 46,
            "TransmissionRate": "0.5 s",
            "Units": "bit"
        },
        3554: {
            "Acronym": "EFS",
            "DataRange": "0 to 7",
            "EndBit": 50,
            "Name": "Engine Ventilation Status",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3554,
            "SPNLength": 3,
            "StartBit": 48,
            "TransmissionRate": "0.5 s",
            "Units": "bit"
        },
        3557: {
            "Acronym": "B",
            "DataRange": "0 to 3",
            "EndBit": 27,
            "Name": "Parking Brake Red Warning Signal",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3557,
            "SPNLength": 2,
            "StartBit": 26,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        3558: {
            "Acronym": "AETC",
            "DataRange": "0 to 15",
            "EndBit": 3,
            "Name": "AETC Data Collection Standard",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 3558,
            "SPNLength": 4,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        3559: {
            "Acronym": "AETC",
            "DataRange": "0 to 15",
            "EndBit": 7,
            "Name": "Number of AETC data points",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 5.0,
            "OperationalRange": "5 to 15 points",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 3559,
            "SPNLength": 4,
            "StartBit": 4,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        3560: {
            "Acronym": "AETC",
            "DataRange": "0 to 8,031.875 rpm",
            "EndBit": -1,
            "Name": "AETC Speed Value",
            "Offset": 0.0,
            "OperationalHigh": 8031.875,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.125,
            "SPN": 3560,
            "SPNLength": 16,
            "StartBit": -1,
            "TransmissionRate": "On request",
            "Units": "rpm"
        },
        3561: {
            "Acronym": "AETC",
            "DataRange": "0 to 64,255 Nm",
            "EndBit": -2,
            "Name": "AETC Torque value",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 3561,
            "SPNLength": 16,
            "StartBit": -1,
            "TransmissionRate": "On request",
            "Units": "Nm"
        },
        3562: {
            "Acronym": "IC2",
            "DataRange": "0.0 to 72.518869 psi",
            "EndBit": 31,
            "Name": "Engine Intake Manifold #2 Pressure",
            "Offset": 0.0,
            "OperationalHigh": 72.518869,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.290075476,
            "SPN": 3562,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "0.5 s",
            "Units": "psi"
        },
        3563: {
            "Acronym": "IC2",
            "DataRange": "0.0 to 72.518869 psi",
            "EndBit": 39,
            "Name": "Engine Intake Manifold #1 Absolute Pressure",
            "Offset": 0.0,
            "OperationalHigh": 72.518869,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.290075476,
            "SPN": 3563,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "0.5 s",
            "Units": "psi"
        },
        3564: {
            "Acronym": "FLIC",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Lane Departure Warning Enable Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3564,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "On event",
            "Units": "bit"
        },
        3565: {
            "Acronym": "FLI1",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Lane Departure Left",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3565,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "50 ms (only when active)",
            "Units": "bit"
        },
        3566: {
            "Acronym": "FLI1",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Lane Departure Right",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3566,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "50 ms (only when active)",
            "Units": "bit"
        },
        3567: {
            "Acronym": "GC1",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Generator Control Not In Automatic Start State",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3567,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3568: {
            "Acronym": "GC1",
            "DataRange": "0 to 3",
            "EndBit": 7,
            "Name": "Generator Not Ready to Automatically Parallel State",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3568,
            "SPNLength": 2,
            "StartBit": 6,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3589: {
            "Acronym": "EOI",
            "DataRange": "0 to 3",
            "EndBit": 33,
            "Name": "Engine Oil Priming Pump Control",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3589,
            "SPNLength": 2,
            "StartBit": 32,
            "TransmissionRate": "250 ms",
            "Units": "bit"
        },
        3590: {
            "Acronym": "GTACPP",
            "DataRange": "-251 to 250.99 %",
            "EndBit": 15,
            "Name": "Generator Total Percent kW",
            "Offset": -251.0,
            "OperationalHigh": 250.99,
            "OperationalLow": -251.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0078125,
            "SPN": 3590,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "250 ms",
            "Units": "%"
        },
        3591: {
            "Acronym": "GTACPP",
            "DataRange": "-251 to 250.99 %",
            "EndBit": 31,
            "Name": "Generator Total Percent kVA",
            "Offset": -251.0,
            "OperationalHigh": 250.99,
            "OperationalLow": -251.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0078125,
            "SPN": 3591,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "250 ms",
            "Units": "%"
        },
        3592: {
            "Acronym": "GTACPP",
            "DataRange": "-251 to 250.99 %",
            "EndBit": 47,
            "Name": "Generator Total Percent kVAr",
            "Offset": -251.0,
            "OperationalHigh": 250.99,
            "OperationalLow": -251.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0078125,
            "SPN": 3592,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "250 ms",
            "Units": "%"
        },
        3593: {
            "Acronym": "GTACER",
            "DataRange": "0 to 4211081215 kVArh",
            "EndBit": 31,
            "Name": "Generator Total kVAr Hours Export",
            "Offset": 0.0,
            "OperationalHigh": 4211081215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3593,
            "SPNLength": 32,
            "StartBit": 0,
            "TransmissionRate": "250 ms",
            "Units": "kVArh"
        },
        3594: {
            "Acronym": "GTACER",
            "DataRange": "0 to 4211081215 kVArh",
            "EndBit": 63,
            "Name": "Generator Total kVAr Hours Import",
            "Offset": 0.0,
            "OperationalHigh": 4211081215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3594,
            "SPNLength": 32,
            "StartBit": 32,
            "TransmissionRate": "250 ms",
            "Units": "kVArh"
        },
        3595: {
            "Acronym": "UTACER",
            "DataRange": "0 to 4211081215 kVArh",
            "EndBit": 31,
            "Name": "Utility Total kVAr Hours Export",
            "Offset": 0.0,
            "OperationalHigh": 4211081215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3595,
            "SPNLength": 32,
            "StartBit": 0,
            "TransmissionRate": "250 ms",
            "Units": "kVArh"
        },
        3596: {
            "Acronym": "UTACER",
            "DataRange": "0 to 4211081215 kVArh",
            "EndBit": 63,
            "Name": "Utility Total kVAr Hours Import",
            "Offset": 0.0,
            "OperationalHigh": 4211081215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3596,
            "SPNLength": 32,
            "StartBit": 32,
            "TransmissionRate": "250 ms",
            "Units": "kVArh"
        },
        3597: {
            "Acronym": "VEP2",
            "DataRange": "0 to 3212.75 V",
            "EndBit": 31,
            "Name": "ECU Power Output Supply Voltage #1",
            "Offset": 0.0,
            "OperationalHigh": 3212.75,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05,
            "SPN": 3597,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "Volts"
        },
        3598: {
            "Acronym": "VEP2",
            "DataRange": "0 to 3212.75 V",
            "EndBit": 47,
            "Name": "ECU Power Output Supply Voltage #2",
            "Offset": 0.0,
            "OperationalHigh": 3212.75,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05,
            "SPN": 3598,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "Volts"
        },
        3599: {
            "Acronym": "VEP2",
            "DataRange": "0 to 3212.75 V",
            "EndBit": 63,
            "Name": "ECU Power Output Supply Voltage #3",
            "Offset": 0.0,
            "OperationalHigh": 3212.75,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05,
            "SPN": 3599,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "On request",
            "Units": "Volts"
        },
        3600: {
            "Acronym": "RESET",
            "DataRange": "0 to 3",
            "EndBit": 19,
            "Name": "Steering Straight Ahead Position Reset",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3600,
            "SPNLength": 2,
            "StartBit": 18,
            "TransmissionRate": "When needed",
            "Units": "bit"
        },
        3601: {
            "Acronym": "EOI",
            "DataRange": "0 to 3",
            "EndBit": 31,
            "Name": "Engine Fuel Shutoff Valve Leak Test Control",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3601,
            "SPNLength": 2,
            "StartBit": 30,
            "TransmissionRate": "250 ms",
            "Units": "bit"
        },
        3602: {
            "Acronym": "EOI",
            "DataRange": "0 to 3",
            "EndBit": 35,
            "Name": "Engine Oil Pre-heater Control",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3602,
            "SPNLength": 2,
            "StartBit": 34,
            "TransmissionRate": "250 ms",
            "Units": "bit"
        },
        3603: {
            "Acronym": "EOI",
            "DataRange": "0 to 3",
            "EndBit": 37,
            "Name": "Engine Electrical System Power Conservation Control",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3603,
            "SPNLength": 2,
            "StartBit": 36,
            "TransmissionRate": "250 ms",
            "Units": "bit"
        },
        3604: {
            "Acronym": "EOI",
            "DataRange": "0 to 3",
            "EndBit": 39,
            "Name": "Engine Block / Coolant Pre-heater Control",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3604,
            "SPNLength": 2,
            "StartBit": 38,
            "TransmissionRate": "250 ms",
            "Units": "bit"
        },
        3605: {
            "Acronym": "EOI",
            "DataRange": "0 to 3",
            "EndBit": 41,
            "Name": "Engine Coolant Circulating Pump Control",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3605,
            "SPNLength": 2,
            "StartBit": 40,
            "TransmissionRate": "250 ms",
            "Units": "bit"
        },
        3606: {
            "Acronym": "EOI",
            "DataRange": "0 to 3",
            "EndBit": 43,
            "Name": "Engine Controlled Shutdown Request",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3606,
            "SPNLength": 2,
            "StartBit": 42,
            "TransmissionRate": "250 ms",
            "Units": "bit"
        },
        3607: {
            "Acronym": "EOI",
            "DataRange": "0 to 3",
            "EndBit": 45,
            "Name": "Engine Emergency (Immediate) Shutdown Indication",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3607,
            "SPNLength": 2,
            "StartBit": 44,
            "TransmissionRate": "250 ms",
            "Units": "bit"
        },
        3608: {
            "Acronym": "EOI",
            "DataRange": "0 to 3",
            "EndBit": 25,
            "Name": "Engine Fuel Shutoff Vent Control",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3608,
            "SPNLength": 2,
            "StartBit": 24,
            "TransmissionRate": "250 ms",
            "Units": "bit"
        },
        3609: {
            "Acronym": "AT1GP",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 15,
            "Name": "DPF Intake Pressure 1",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 3609,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        3610: {
            "Acronym": "AT1GP",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 31,
            "Name": "DPF Outlet Pressure 1",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 3610,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        3611: {
            "Acronym": "AT2GP",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 15,
            "Name": "DPF Intake Pressure 2",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 3611,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        3612: {
            "Acronym": "AT2GP",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 31,
            "Name": "DPF Outlet Pressure 2",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 3612,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        3613: {
            "Acronym": "DISP1",
            "DataRange": "0 to 15",
            "EndBit": 3,
            "Name": "Text Display Instructions",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable - up to 200 characters",
            "Resolution": 1,
            "SPN": 3613,
            "SPNLength": 4,
            "StartBit": 0,
            "TransmissionRate": "As required",
            "Units": "bit"
        },
        3614: {
            "Acronym": "DISP1",
            "DataRange": "0 to 250",
            "EndBit": 23,
            "Name": "Text Display Index",
            "Offset": 0.0,
            "OperationalHigh": 200.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 200",
            "PGNLength": "Variable - up to 200 characters",
            "Resolution": 1,
            "SPN": 3614,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "As required",
            "Units": "count"
        },
        3615: {
            "Acronym": "DISP1",
            "DataRange": "0 to 255 per byte",
            "EndBit": 0,
            "Name": "Text Display Character",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable - up to 200 characters",
            "Resolution": -2,
            "SPN": 3615,
            "SPNLength": 1600,
            "StartBit": -1,
            "TransmissionRate": "As required",
            "Units": "ASCII"
        },
        3619: {
            "Acronym": "J2012",
            "DataRange": "0 to 250",
            "EndBit": 7,
            "Name": "Number of J2012 DTCs",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 3619,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "count"
        },
        3620: {
            "Acronym": "J2012",
            "DataRange": "0 to 255 per byte",
            "EndBit": 47,
            "Name": "J2012 DTC",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": -2,
            "SPN": 3620,
            "SPNLength": 40,
            "StartBit": 8,
            "TransmissionRate": "On request",
            "Units": "ASCII"
        },
        3621: {
            "Acronym": "J2012",
            "DataRange": "0 to 1",
            "EndBit": 48,
            "Name": "J2012 DTC Status",
            "Offset": 0.0,
            "OperationalHigh": 1.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 3621,
            "SPNLength": 1,
            "StartBit": 48,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        3622: {
            "Acronym": "J2012",
            "DataRange": "0 to 127",
            "EndBit": 55,
            "Name": "J2012 DTC Occurrence Count",
            "Offset": 0.0,
            "OperationalHigh": 127.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 3622,
            "SPNLength": 7,
            "StartBit": 49,
            "TransmissionRate": "On request",
            "Units": "count"
        },
        3623: {
            "Acronym": "VDS2",
            "DataRange": "-200 to 301.99 deg",
            "EndBit": 15,
            "Name": "Vehicle Roll",
            "Offset": -200.0,
            "OperationalHigh": 90.0,
            "OperationalLow": -90.0,
            "OperationalRange": "-90 to 90 degrees",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3623,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "deg"
        },
        3624: {
            "Acronym": "EFL/P5",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 15,
            "Name": "Engine Intake Valve Actuation Oil Pressure for Cylinder #1",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 3624,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        3625: {
            "Acronym": "EFL/P5",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 31,
            "Name": "Engine Intake Valve Actuation Oil Pressure for Cylinder #2",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 3625,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        3626: {
            "Acronym": "EFL/P5",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 47,
            "Name": "Engine Intake Valve Actuation Oil Pressure for Cylinder #3",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 3626,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        3627: {
            "Acronym": "EFL/P5",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 63,
            "Name": "Engine Intake Valve Actuation Oil Pressure for Cylinder #4",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 3627,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        3628: {
            "Acronym": "EFL/P6",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 15,
            "Name": "Engine Intake Valve Actuation Oil Pressure for Cylinder #5",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 3628,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        3629: {
            "Acronym": "EFL/P6",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 31,
            "Name": "Engine Intake Valve Actuation Oil Pressure for Cylinder #6",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 3629,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        3630: {
            "Acronym": "EFL/P6",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 47,
            "Name": "Engine Intake Valve Actuation Oil Pressure for Cylinder #7",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 3630,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        3631: {
            "Acronym": "EFL/P6",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 63,
            "Name": "Engine Intake Valve Actuation Oil Pressure for Cylinder #8",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 3631,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        3632: {
            "Acronym": "EFL/P7",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 15,
            "Name": "Engine Intake Valve Actuation Oil Pressure for Cylinder #9",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 3632,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        3633: {
            "Acronym": "EFL/P7",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 31,
            "Name": "Engine Intake Valve Actuation Oil Pressure for Cylinder #10",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 3633,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        3634: {
            "Acronym": "EFL/P7",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 47,
            "Name": "Engine Intake Valve Actuation Oil Pressure for Cylinder #11",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 3634,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        3635: {
            "Acronym": "EFL/P7",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 63,
            "Name": "Engine Intake Valve Actuation Oil Pressure for Cylinder #12",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 3635,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        3636: {
            "Acronym": "EFL/P8",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 15,
            "Name": "Engine Intake Valve Actuation Oil Pressure for Cylinder #13",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 3636,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        3637: {
            "Acronym": "EFL/P8",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 31,
            "Name": "Engine Intake Valve Actuation Oil Pressure for Cylinder #14",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 3637,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        3638: {
            "Acronym": "EFL/P8",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 47,
            "Name": "Engine Intake Valve Actuation Oil Pressure for Cylinder #15",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 3638,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        3639: {
            "Acronym": "EFL/P8",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 63,
            "Name": "Engine Intake Valve Actuation Oil Pressure for Cylinder #16",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 3639,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        3640: {
            "Acronym": "EFL/P9",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 15,
            "Name": "Engine Intake Valve Actuation Oil Pressure for Cylinder #17",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 3640,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        3641: {
            "Acronym": "EFL/P9",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 31,
            "Name": "Engine Intake Valve Actuation Oil Pressure for Cylinder #18",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 3641,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        3642: {
            "Acronym": "EFL/P9",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 47,
            "Name": "Engine Intake Valve Actuation Oil Pressure for Cylinder #19",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 3642,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        3643: {
            "Acronym": "EFL/P9",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 63,
            "Name": "Engine Intake Valve Actuation Oil Pressure for Cylinder #20",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 3643,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        3644: {
            "Acronym": "EOI",
            "DataRange": "0 to 100 %",
            "EndBit": 63,
            "Name": "Engine Derate Request",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 3644,
            "SPNLength": 8,
            "StartBit": 56,
            "TransmissionRate": "250 ms",
            "Units": "%"
        },
        3645: {
            "Acronym": "TCI",
            "DataRange": "0 to 7",
            "EndBit": 2,
            "Name": "Transfer case status",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3645,
            "SPNLength": 3,
            "StartBit": 0,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010:  1 s or on change",
            "Units": "bit"
        },
        3667: {
            "Acronym": "SHUTDN",
            "DataRange": "0 to 3",
            "EndBit": 57,
            "Name": "Engine Air Shutoff Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3667,
            "SPNLength": 2,
            "StartBit": 56,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        3668: {
            "Acronym": "EFL/P4",
            "DataRange": "0 to 100 %",
            "EndBit": 47,
            "Name": "Engine Intercooler Coolant Level",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 3668,
            "SPNLength": 8,
            "StartBit": 40,
            "TransmissionRate": "500 ms",
            "Units": "%"
        },
        3669: {
            "Acronym": "EEC4",
            "DataRange": "0 to 3",
            "EndBit": 33,
            "Name": "Engine Rotation Direction",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3669,
            "SPNLength": 2,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        3670: {
            "Acronym": "EC2",
            "DataRange": "0 to 250",
            "EndBit": 7,
            "Name": "Maximum Crank Attempts per Start Attempt",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3670,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "count"
        },
        3671: {
            "Acronym": "EEC4",
            "DataRange": "0 to 250",
            "EndBit": 47,
            "Name": "Crank Attempt Count on Present Start Attempt",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3671,
            "SPNLength": 8,
            "StartBit": 40,
            "TransmissionRate": "On request",
            "Units": "count"
        },
        3672: {
            "Acronym": "EGRBV",
            "DataRange": "0 to 100 %",
            "EndBit": 7,
            "Name": "EGR1 Cooler Bypass Actuator Postion",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 3672,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "500 ms",
            "Units": "%"
        },
        3673: {
            "Acronym": "LFE1",
            "DataRange": "0 to 100 %",
            "EndBit": 63,
            "Name": "Engine Throttle Valve 2 Position",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 3673,
            "SPNLength": 8,
            "StartBit": 56,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        3675: {
            "Acronym": "EEC6",
            "DataRange": "0 to 100 %",
            "EndBit": 31,
            "Name": "Engine Turbocharger Compressor Bypass Actuator 1  Position",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 3675,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "100 msec (preferred) or Engine Speed Dependent (if required by application)",
            "Units": "%"
        },
        3676: {
            "Acronym": "EFL/P4",
            "DataRange": "0 to 100 %",
            "EndBit": 55,
            "Name": "Engine Aftercooler Coolant Level",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 3676,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "500 ms",
            "Units": "%"
        },
        3683: {
            "Acronym": "SAS",
            "DataRange": "-31.374 to +31.374 rad",
            "EndBit": 15,
            "Name": "Steering Wheel Angle",
            "Offset": -31.374,
            "OperationalHigh": 31.374,
            "OperationalLow": -31.374,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3683,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "10 ms",
            "Units": "rad"
        },
        3684: {
            "Acronym": "SAS",
            "DataRange": "-32 to 29 ranges",
            "EndBit": 21,
            "Name": "Steering Wheel Angle Range Counter",
            "Offset": -32.0,
            "OperationalHigh": 29.0,
            "OperationalLow": -32.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3684,
            "SPNLength": 6,
            "StartBit": 16,
            "TransmissionRate": "10 ms",
            "Units": "range"
        },
        3685: {
            "Acronym": "SAS",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Steering Wheel Angle Range Counter Type",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3685,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "10 ms",
            "Units": "bit"
        },
        3686: {
            "Acronym": "SAS",
            "DataRange": "-31.374 to +31.374 rad",
            "EndBit": 47,
            "Name": "Steering Wheel Angle Range",
            "Offset": -31.374,
            "OperationalHigh": 31.374,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to +31.374 rad",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3686,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "10 ms",
            "Units": "rad"
        },
        3687: {
            "Acronym": "SAS",
            "DataRange": "0 to 3",
            "EndBit": 49,
            "Name": "Steering Angle Sensor Active Mode",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3687,
            "SPNLength": 2,
            "StartBit": 48,
            "TransmissionRate": "10 ms",
            "Units": "bit"
        },
        3688: {
            "Acronym": "SAS",
            "DataRange": "0 to 3",
            "EndBit": 51,
            "Name": "Steering Angle Sensor Calibrated",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3688,
            "SPNLength": 2,
            "StartBit": 50,
            "TransmissionRate": "10 ms",
            "Units": "bit"
        },
        3689: {
            "Acronym": "SAS",
            "DataRange": "0 to 15",
            "EndBit": 59,
            "Name": "Message Counter",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3689,
            "SPNLength": 4,
            "StartBit": 56,
            "TransmissionRate": "10 ms",
            "Units": "count"
        },
        3690: {
            "Acronym": "SAS",
            "DataRange": "0 to 15",
            "EndBit": 63,
            "Name": "Message Checksum",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3690,
            "SPNLength": 4,
            "StartBit": 60,
            "TransmissionRate": "10 ms",
            "Units": "count"
        },
        3691: {
            "Acronym": "AFSS",
            "DataRange": "0 to 7",
            "EndBit": 2,
            "Name": "Left Headlamp Dynamic Bending Light",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3691,
            "SPNLength": 3,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        3692: {
            "Acronym": "AFSS",
            "DataRange": "0 to 7",
            "EndBit": 5,
            "Name": "Right Headlamp Dynamic Bending Light",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3692,
            "SPNLength": 3,
            "StartBit": 3,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        3693: {
            "Acronym": "AFSS",
            "DataRange": "0 to 15",
            "EndBit": 11,
            "Name": "Left Headlamp Light Distribution",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3693,
            "SPNLength": 4,
            "StartBit": 8,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        3694: {
            "Acronym": "AFSS",
            "DataRange": "0 to 15",
            "EndBit": 15,
            "Name": "Right Headlamp Light Distribution",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3694,
            "SPNLength": 4,
            "StartBit": 12,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        3695: {
            "Acronym": "CM1",
            "DataRange": "0 to 3",
            "EndBit": 41,
            "Name": "DPF Regeneration Inhibit Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3695,
            "SPNLength": 2,
            "StartBit": 40,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        3696: {
            "Acronym": "CM1",
            "DataRange": "0 to 3",
            "EndBit": 43,
            "Name": "DPF Regeneration Force Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3696,
            "SPNLength": 2,
            "StartBit": 42,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        3697: {
            "Acronym": "DPFC1",
            "DataRange": "0 to 7",
            "EndBit": 2,
            "Name": "DPF Lamp Command",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3697,
            "SPNLength": 3,
            "StartBit": 0,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010:  1 s and on change",
            "Units": "bit"
        },
        3698: {
            "Acronym": "DPFC1",
            "DataRange": "0 to 7",
            "EndBit": 52,
            "Name": "Exhaust System High Temperature Lamp Command",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3698,
            "SPNLength": 3,
            "StartBit": 50,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010:  1 s and on change",
            "Units": "bit"
        },
        3699: {
            "Acronym": "DPFC1",
            "DataRange": "0 to 3",
            "EndBit": 9,
            "Name": "DPF Passive Regeneration Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3699,
            "SPNLength": 2,
            "StartBit": 8,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010:  1 s and on change",
            "Units": "bit"
        },
        3700: {
            "Acronym": "DPFC1",
            "DataRange": "0 to 3",
            "EndBit": 11,
            "Name": "DPF Active Regeneration Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3700,
            "SPNLength": 2,
            "StartBit": 10,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010:  1 s and on change",
            "Units": "bit"
        },
        3701: {
            "Acronym": "DPFC1",
            "DataRange": "0 to 7",
            "EndBit": 14,
            "Name": "DPF Status",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3701,
            "SPNLength": 3,
            "StartBit": 12,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010:  1 s and on change",
            "Units": "bit"
        },
        3702: {
            "Acronym": "DPFC1",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "DPF Active Regeneration Inhibited Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3702,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010:  1 s and on change",
            "Units": "bit"
        },
        3703: {
            "Acronym": "DPFC1",
            "DataRange": "0 to 3",
            "EndBit": 19,
            "Name": "DPF Active Regeneration Inhibited Due to Inhibit Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3703,
            "SPNLength": 2,
            "StartBit": 18,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010:  1 s and on change",
            "Units": "bit"
        },
        3704: {
            "Acronym": "DPFC1",
            "DataRange": "0 to 3",
            "EndBit": 21,
            "Name": "DPF Active Regeneration Inhibited Due to Clutch Disengaged",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3704,
            "SPNLength": 2,
            "StartBit": 20,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010:  1 s and on change",
            "Units": "bit"
        },
        3705: {
            "Acronym": "DPFC1",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "DPF Active Regeneration Inhibited Due to Service Brake Active",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3705,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010:  1 s and on change",
            "Units": "bit"
        },
        3706: {
            "Acronym": "DPFC1",
            "DataRange": "0 to 3",
            "EndBit": 25,
            "Name": "DPF Active Regeneration Inhibited Due to PTO Active",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3706,
            "SPNLength": 2,
            "StartBit": 24,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010:  1 s and on change",
            "Units": "bit"
        },
        3707: {
            "Acronym": "DPFC1",
            "DataRange": "0 to 3",
            "EndBit": 27,
            "Name": "DPF Active Regeneration Inhibited Due to Accelerator Pedal Off Idle",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3707,
            "SPNLength": 2,
            "StartBit": 26,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010:  1 s and on change",
            "Units": "bit"
        },
        3708: {
            "Acronym": "DPFC1",
            "DataRange": "0 to 3",
            "EndBit": 29,
            "Name": "DPF Active Regeneration Inhibited Due to Out of Neutral",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3708,
            "SPNLength": 2,
            "StartBit": 28,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010:  1 s and on change",
            "Units": "bit"
        },
        3709: {
            "Acronym": "DPFC1",
            "DataRange": "0 to 3",
            "EndBit": 31,
            "Name": "DPF Active Regeneration Inhibited Due to Vehicle Speed Above Allowed Speed",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3709,
            "SPNLength": 2,
            "StartBit": 30,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010:  1 s and on change",
            "Units": "bit"
        },
        3710: {
            "Acronym": "DPFC1",
            "DataRange": "0 to 3",
            "EndBit": 33,
            "Name": "DPF Active Regeneration Inhibited Due to Parking Brake Not Set",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3710,
            "SPNLength": 2,
            "StartBit": 32,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010:  1 s and on change",
            "Units": "bit"
        },
        3711: {
            "Acronym": "DPFC1",
            "DataRange": "0 to 3",
            "EndBit": 35,
            "Name": "DPF Active Regeneration Inhibited Due to Low Exhaust Gas Temperature",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3711,
            "SPNLength": 2,
            "StartBit": 34,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010:  1 s and on change",
            "Units": "bit"
        },
        3712: {
            "Acronym": "DPFC1",
            "DataRange": "0 to 3",
            "EndBit": 37,
            "Name": "DPF Active Regeneration Inhibited Due to System Fault Active",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3712,
            "SPNLength": 2,
            "StartBit": 36,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010:  1 s and on change",
            "Units": "bit"
        },
        3713: {
            "Acronym": "DPFC1",
            "DataRange": "0 to 3",
            "EndBit": 39,
            "Name": "DPF Active Regeneration Inhibited Due to System Timeout",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3713,
            "SPNLength": 2,
            "StartBit": 38,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010:  1 s and on change",
            "Units": "bit"
        },
        3714: {
            "Acronym": "DPFC1",
            "DataRange": "0 to 3",
            "EndBit": 41,
            "Name": "DPF Active Regeneration Inhibited Due to Temporary System Lockout",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3714,
            "SPNLength": 2,
            "StartBit": 40,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010:  1 s and on change",
            "Units": "bit"
        },
        3715: {
            "Acronym": "DPFC1",
            "DataRange": "0 to 3",
            "EndBit": 43,
            "Name": "DPF Active Regeneration Inhibited Due to Permanent System Lockout",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3715,
            "SPNLength": 2,
            "StartBit": 42,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010:  1 s and on change",
            "Units": "bit"
        },
        3716: {
            "Acronym": "DPFC1",
            "DataRange": "0 to 3",
            "EndBit": 45,
            "Name": "DPF Active Regeneration Inhibited Due to Engine Not Warmed Up",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3716,
            "SPNLength": 2,
            "StartBit": 44,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010:  1 s and on change",
            "Units": "bit"
        },
        3717: {
            "Acronym": "DPFC1",
            "DataRange": "0 to 3",
            "EndBit": 47,
            "Name": "DPF Active Regeneration Inhibited Due to Vehicle Speed Below Allowed Speed",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3717,
            "SPNLength": 2,
            "StartBit": 46,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010:  1 s and on change",
            "Units": "bit"
        },
        3718: {
            "Acronym": "DPFC1",
            "DataRange": "0 to 3",
            "EndBit": 49,
            "Name": "DPF Automatic Active Regeneration Initiation Configuration",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3718,
            "SPNLength": 2,
            "StartBit": 48,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010:  1 s and on change",
            "Units": "bit"
        },
        3719: {
            "Acronym": "AT1S",
            "DataRange": "0 to 250 %",
            "EndBit": 7,
            "Name": "DPF 1 Soot Load Percent",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3719,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        3720: {
            "Acronym": "AT1S",
            "DataRange": "0 to 250 %",
            "EndBit": 15,
            "Name": "DPF 1 Ash Load Percent",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3720,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        3721: {
            "Acronym": "AT1S",
            "DataRange": "0 to 4,211,081,215 s",
            "EndBit": 47,
            "Name": "DPF 1 Time Since Last Active Regeneration",
            "Offset": 0.0,
            "OperationalHigh": 4211081215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3721,
            "SPNLength": 32,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "Seconds"
        },
        3722: {
            "Acronym": "AT2S",
            "DataRange": "0 to 250 %",
            "EndBit": 7,
            "Name": "DPF 2 Soot Load Percent",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3722,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        3723: {
            "Acronym": "AT2S",
            "DataRange": "0 to 250 %",
            "EndBit": 15,
            "Name": "DPF 2 Ash Load Percent",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3723,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        3724: {
            "Acronym": "AT2S",
            "DataRange": "0 to 4,211,081,215 s",
            "EndBit": 47,
            "Name": "DPF 2 Time Since Last Active Regeneration",
            "Offset": 0.0,
            "OperationalHigh": 4211081215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3724,
            "SPNLength": 32,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "Seconds"
        },
        3725: {
            "Acronym": "AT1HI",
            "DataRange": "0 to 4,211,081,215 s",
            "EndBit": 159,
            "Name": "Aftertreatment 1 Total Passive Regeneration Time",
            "Offset": 0.0,
            "OperationalHigh": 4211081215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 3725,
            "SPNLength": 32,
            "StartBit": 128,
            "TransmissionRate": "On request",
            "Units": "Seconds"
        },
        3726: {
            "Acronym": "AT1HI",
            "DataRange": "0 to 4,211,081,215",
            "EndBit": 191,
            "Name": "Aftertreatment 1 Total Number of Passive Regenerations",
            "Offset": 0.0,
            "OperationalHigh": 4211081215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 3726,
            "SPNLength": 32,
            "StartBit": 160,
            "TransmissionRate": "On request",
            "Units": "count"
        },
        3727: {
            "Acronym": "AT1HI",
            "DataRange": "0 to 4,211,081,215",
            "EndBit": 223,
            "Name": "Aftertreatment 1 Total Number of Active Regeneration Inhibit Requests",
            "Offset": 0.0,
            "OperationalHigh": 4211081215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 3727,
            "SPNLength": 32,
            "StartBit": 192,
            "TransmissionRate": "On request",
            "Units": "count"
        },
        3728: {
            "Acronym": "AT1HI",
            "DataRange": "0 to 4,211,081,215",
            "EndBit": 255,
            "Name": "Aftertreatment 1 Total Number of Active Regeneration Manual Requests",
            "Offset": 0.0,
            "OperationalHigh": 4211081215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 3728,
            "SPNLength": 32,
            "StartBit": 224,
            "TransmissionRate": "On request",
            "Units": "count"
        },
        3729: {
            "Acronym": "AT2HI",
            "DataRange": "0 to 4,211,081,215 s",
            "EndBit": 159,
            "Name": "Aftertreatment 2 Total Passive Regeneration Time",
            "Offset": 0.0,
            "OperationalHigh": 4211081215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 3729,
            "SPNLength": 32,
            "StartBit": 128,
            "TransmissionRate": "On request",
            "Units": "Seconds"
        },
        3730: {
            "Acronym": "AT2HI",
            "DataRange": "0 to 4,211,081,215",
            "EndBit": 191,
            "Name": "Aftertreatment 2 Total Number of Passive Regenerations",
            "Offset": 0.0,
            "OperationalHigh": 4211081215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 3730,
            "SPNLength": 32,
            "StartBit": 160,
            "TransmissionRate": "On request",
            "Units": "count"
        },
        3731: {
            "Acronym": "AT2HI",
            "DataRange": "0 to 4,211,081,215",
            "EndBit": 223,
            "Name": "Aftertreatment 2 Total Number of Active Regeneration Inhibit Requests",
            "Offset": 0.0,
            "OperationalHigh": 4211081215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 3731,
            "SPNLength": 32,
            "StartBit": 192,
            "TransmissionRate": "On request",
            "Units": "count"
        },
        3732: {
            "Acronym": "AT2HI",
            "DataRange": "0 to 4,211,081,215",
            "EndBit": 255,
            "Name": "Aftertreatment 2 Total Number of Active Regeneration Manual Requests",
            "Offset": 0.0,
            "OperationalHigh": 4211081215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 3732,
            "SPNLength": 32,
            "StartBit": 224,
            "TransmissionRate": "On request",
            "Units": "count"
        },
        3733: {
            "Acronym": "AT1TI",
            "DataRange": "0.0 to 556224873.36449 US Gal",
            "EndBit": 31,
            "Name": "Aftertreatment 1 Trip Fuel Used",
            "Offset": 0.0,
            "OperationalHigh": 556224873.36449,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 32,
            "Resolution": 0.132086,
            "SPN": 3733,
            "SPNLength": 32,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "gallons"
        },
        3734: {
            "Acronym": "AT1TI",
            "DataRange": "0 to 4,211,081,215 s",
            "EndBit": 63,
            "Name": "Aftertreatment 1 Trip Active Regeneration Time",
            "Offset": 0.0,
            "OperationalHigh": 4211081215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 32,
            "Resolution": 1,
            "SPN": 3734,
            "SPNLength": 32,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "Seconds"
        },
        3735: {
            "Acronym": "AT1TI",
            "DataRange": "0 to 4,211,081,215 s",
            "EndBit": 95,
            "Name": "Aftertreatment 1 Trip Disabled Time",
            "Offset": 0.0,
            "OperationalHigh": 4211081215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 32,
            "Resolution": 1,
            "SPN": 3735,
            "SPNLength": 32,
            "StartBit": 64,
            "TransmissionRate": "On request",
            "Units": "Seconds"
        },
        3736: {
            "Acronym": "AT1TI",
            "DataRange": "0 to 4,211,081,215",
            "EndBit": 127,
            "Name": "Aftertreatment 1 Trip Number of Active Regenerations",
            "Offset": 0.0,
            "OperationalHigh": 4211081215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 32,
            "Resolution": 1,
            "SPN": 3736,
            "SPNLength": 32,
            "StartBit": 96,
            "TransmissionRate": "On request",
            "Units": "count"
        },
        3737: {
            "Acronym": "AT1TI",
            "DataRange": "0 to 4,211,081,215 s",
            "EndBit": 159,
            "Name": "Aftertreatment 1 Trip Passive Regeneration Time",
            "Offset": 0.0,
            "OperationalHigh": 4211081215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 32,
            "Resolution": 1,
            "SPN": 3737,
            "SPNLength": 32,
            "StartBit": 128,
            "TransmissionRate": "On request",
            "Units": "Seconds"
        },
        3738: {
            "Acronym": "AT1TI",
            "DataRange": "0 to 4,211,081,215",
            "EndBit": 191,
            "Name": "Aftertreatment 1 Trip Number of Passive Regenerations",
            "Offset": 0.0,
            "OperationalHigh": 4211081215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 32,
            "Resolution": 1,
            "SPN": 3738,
            "SPNLength": 32,
            "StartBit": 160,
            "TransmissionRate": "On request",
            "Units": "count"
        },
        3739: {
            "Acronym": "AT1TI",
            "DataRange": "0 to 4,211,081,215",
            "EndBit": 223,
            "Name": "Aftertreatment 1 Trip Number of Active Regeneration Inhibit Requests",
            "Offset": 0.0,
            "OperationalHigh": 4211081215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 32,
            "Resolution": 1,
            "SPN": 3739,
            "SPNLength": 32,
            "StartBit": 192,
            "TransmissionRate": "On request",
            "Units": "count"
        },
        3740: {
            "Acronym": "AT1TI",
            "DataRange": "0 to 4,211,081,215",
            "EndBit": 255,
            "Name": "Aftertreatment 1 Trip Number of Active Regeneration Manual Requests",
            "Offset": 0.0,
            "OperationalHigh": 4211081215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 32,
            "Resolution": 1,
            "SPN": 3740,
            "SPNLength": 32,
            "StartBit": 224,
            "TransmissionRate": "On request",
            "Units": "count"
        },
        3741: {
            "Acronym": "AT2TI",
            "DataRange": "0.0 to 556224873.36449 US Gal",
            "EndBit": 31,
            "Name": "Aftertreatment 2 Trip Fuel Used",
            "Offset": 0.0,
            "OperationalHigh": 556224873.36449,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 32,
            "Resolution": 0.132086,
            "SPN": 3741,
            "SPNLength": 32,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "gallons"
        },
        3742: {
            "Acronym": "AT2TI",
            "DataRange": "0 to 4,211,081,215 s",
            "EndBit": 63,
            "Name": "Aftertreatment 2 Trip Active Regeneration Time",
            "Offset": 0.0,
            "OperationalHigh": 4211081215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 32,
            "Resolution": 1,
            "SPN": 3742,
            "SPNLength": 32,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "Seconds"
        },
        3743: {
            "Acronym": "AT2TI",
            "DataRange": "0 to 4,211,081,215 s",
            "EndBit": 95,
            "Name": "Aftertreatment 2 Trip Disabled Time",
            "Offset": 0.0,
            "OperationalHigh": 4211081215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 32,
            "Resolution": 1,
            "SPN": 3743,
            "SPNLength": 32,
            "StartBit": 64,
            "TransmissionRate": "On request",
            "Units": "Seconds"
        },
        3744: {
            "Acronym": "AT2TI",
            "DataRange": "0 to 4,211,081,215",
            "EndBit": 127,
            "Name": "Aftertreatment 2 Trip Number of Active Regenerations",
            "Offset": 0.0,
            "OperationalHigh": 4211081215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 32,
            "Resolution": 1,
            "SPN": 3744,
            "SPNLength": 32,
            "StartBit": 96,
            "TransmissionRate": "On request",
            "Units": "count"
        },
        3745: {
            "Acronym": "AT2TI",
            "DataRange": "0 to 4,211,081,215 s",
            "EndBit": 159,
            "Name": "Aftertreatment 2 Trip Passive Regeneration Time",
            "Offset": 0.0,
            "OperationalHigh": 4211081215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 32,
            "Resolution": 1,
            "SPN": 3745,
            "SPNLength": 32,
            "StartBit": 128,
            "TransmissionRate": "On request",
            "Units": "Seconds"
        },
        3746: {
            "Acronym": "AT2TI",
            "DataRange": "0 to 4,211,081,215",
            "EndBit": 191,
            "Name": "Aftertreatment 2 Trip Number of Passive Regenerations",
            "Offset": 0.0,
            "OperationalHigh": 4211081215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 32,
            "Resolution": 1,
            "SPN": 3746,
            "SPNLength": 32,
            "StartBit": 160,
            "TransmissionRate": "On request",
            "Units": "count"
        },
        3747: {
            "Acronym": "AT2TI",
            "DataRange": "0 to 4,211,081,215",
            "EndBit": 223,
            "Name": "Aftertreatment 2 Trip Number of Active Regeneration Inhibit Requests",
            "Offset": 0.0,
            "OperationalHigh": 4211081215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 32,
            "Resolution": 1,
            "SPN": 3747,
            "SPNLength": 32,
            "StartBit": 192,
            "TransmissionRate": "On request",
            "Units": "count"
        },
        3748: {
            "Acronym": "AT2TI",
            "DataRange": "0 to 4,211,081,215",
            "EndBit": 255,
            "Name": "Aftertreatment 2 Trip Number of Active Regeneration Manual Requests",
            "Offset": 0.0,
            "OperationalHigh": 4211081215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 32,
            "Resolution": 1,
            "SPN": 3748,
            "SPNLength": 32,
            "StartBit": 224,
            "TransmissionRate": "On request",
            "Units": "count"
        },
        3750: {
            "Acronym": "DPFC1",
            "DataRange": "0 to 3",
            "EndBit": 61,
            "Name": "DPF 1 Conditions Not Met for Active Regeneration",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3750,
            "SPNLength": 2,
            "StartBit": 60,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010:  1 s and on change",
            "Units": "bit"
        },
        3785: {
            "Acronym": "BSA",
            "DataRange": "0 to 7",
            "EndBit": 2,
            "Name": "Tractor Brake Stroke Axle 1 Left",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3785,
            "SPNLength": 3,
            "StartBit": 0,
            "TransmissionRate": "1 sec",
            "Units": "bit"
        },
        3786: {
            "Acronym": "BSA",
            "DataRange": "0 to 7",
            "EndBit": 5,
            "Name": "Tractor Brake Stroke Axle 1 Right",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3786,
            "SPNLength": 3,
            "StartBit": 3,
            "TransmissionRate": "1 sec",
            "Units": "bit"
        },
        3787: {
            "Acronym": "BSA",
            "DataRange": "0 to 7",
            "EndBit": 8,
            "Name": "Tractor Brake Stroke Axle 2 Left",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3787,
            "SPNLength": 3,
            "StartBit": 6,
            "TransmissionRate": "1 sec",
            "Units": "bit"
        },
        3788: {
            "Acronym": "BSA",
            "DataRange": "0 to 7",
            "EndBit": 11,
            "Name": "Tractor Brake Stroke Axle 2 Right",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3788,
            "SPNLength": 3,
            "StartBit": 9,
            "TransmissionRate": "1 sec",
            "Units": "bit"
        },
        3789: {
            "Acronym": "BSA",
            "DataRange": "0 to 7",
            "EndBit": 14,
            "Name": "Tractor Brake Stroke Axle 3 Left",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3789,
            "SPNLength": 3,
            "StartBit": 12,
            "TransmissionRate": "1 sec",
            "Units": "bit"
        },
        3790: {
            "Acronym": "BSA",
            "DataRange": "0 to 7",
            "EndBit": 16,
            "Name": "Tractor Brake Stroke Axle 3 Right",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3790,
            "SPNLength": 3,
            "StartBit": 15,
            "TransmissionRate": "1 sec",
            "Units": "bit"
        },
        3791: {
            "Acronym": "BSA",
            "DataRange": "0 to 7",
            "EndBit": 20,
            "Name": "Tractor Brake Stroke Axle 4 Left",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3791,
            "SPNLength": 3,
            "StartBit": 18,
            "TransmissionRate": "1 sec",
            "Units": "bit"
        },
        3792: {
            "Acronym": "BSA",
            "DataRange": "0 to 7",
            "EndBit": 23,
            "Name": "Tractor Brake Stroke Axle 4 Right",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3792,
            "SPNLength": 3,
            "StartBit": 21,
            "TransmissionRate": "1 sec",
            "Units": "bit"
        },
        3793: {
            "Acronym": "BSA",
            "DataRange": "0 to 7",
            "EndBit": 26,
            "Name": "Tractor Brake Stroke Axle 5 Left",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3793,
            "SPNLength": 3,
            "StartBit": 24,
            "TransmissionRate": "1 sec",
            "Units": "bit"
        },
        3794: {
            "Acronym": "BSA",
            "DataRange": "0 to 7",
            "EndBit": 29,
            "Name": "Tractor Brake Stroke Axle 5 Right",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3794,
            "SPNLength": 3,
            "StartBit": 27,
            "TransmissionRate": "1 sec",
            "Units": "bit"
        },
        3795: {
            "Acronym": "BSA",
            "DataRange": "0 to 7",
            "EndBit": 32,
            "Name": "Trailer Brake Stroke Axle 1 Left",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3795,
            "SPNLength": 3,
            "StartBit": 30,
            "TransmissionRate": "1 sec",
            "Units": "bit"
        },
        3796: {
            "Acronym": "BSA",
            "DataRange": "0 to 7",
            "EndBit": 35,
            "Name": "Trailer Brake Stroke Axle 1 Right",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3796,
            "SPNLength": 3,
            "StartBit": 33,
            "TransmissionRate": "1 sec",
            "Units": "bit"
        },
        3797: {
            "Acronym": "BSA",
            "DataRange": "0 to 7",
            "EndBit": 38,
            "Name": "Trailer Brake Stroke Axle 2 Left",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3797,
            "SPNLength": 3,
            "StartBit": 36,
            "TransmissionRate": "1 sec",
            "Units": "bit"
        },
        3798: {
            "Acronym": "BSA",
            "DataRange": "0 to 7",
            "EndBit": 40,
            "Name": "Trailer Brake Stroke Axle 2 Right",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3798,
            "SPNLength": 3,
            "StartBit": 39,
            "TransmissionRate": "1 sec",
            "Units": "bit"
        },
        3799: {
            "Acronym": "BSA",
            "DataRange": "0 to 7",
            "EndBit": 44,
            "Name": "Trailer Brake Stroke Axle 3 Left",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3799,
            "SPNLength": 3,
            "StartBit": 42,
            "TransmissionRate": "1 sec",
            "Units": "bit"
        },
        38: {
            "Acronym": "DD",
            "DataRange": "0 to 100 %",
            "EndBit": 55,
            "Name": "Fuel Level 2",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 38,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "1s",
            "Units": "%"
        },
        3800: {
            "Acronym": "BSA",
            "DataRange": "0 to 7",
            "EndBit": 47,
            "Name": "Trailer Brake Stroke Axle 3 Right",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3800,
            "SPNLength": 3,
            "StartBit": 45,
            "TransmissionRate": "1 sec",
            "Units": "bit"
        },
        3801: {
            "Acronym": "BSA",
            "DataRange": "0 to 7",
            "EndBit": 50,
            "Name": "Trailer Brake Stroke Axle 4 Left",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3801,
            "SPNLength": 3,
            "StartBit": 48,
            "TransmissionRate": "1 sec",
            "Units": "bit"
        },
        3802: {
            "Acronym": "BSA",
            "DataRange": "0 to 7",
            "EndBit": 53,
            "Name": "Trailer Brake Stroke Axle 4 Right",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3802,
            "SPNLength": 3,
            "StartBit": 51,
            "TransmissionRate": "1 sec",
            "Units": "bit"
        },
        3803: {
            "Acronym": "BSA",
            "DataRange": "0 to 7",
            "EndBit": 56,
            "Name": "Trailer Brake Stroke Axle 5 Left",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3803,
            "SPNLength": 3,
            "StartBit": 54,
            "TransmissionRate": "1 sec",
            "Units": "bit"
        },
        3804: {
            "Acronym": "BSA",
            "DataRange": "0 to 7",
            "EndBit": 59,
            "Name": "Trailer Brake Stroke Axle 5 Right",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3804,
            "SPNLength": 3,
            "StartBit": 57,
            "TransmissionRate": "1 sec",
            "Units": "bit"
        },
        3807: {
            "Acronym": "CCVS1",
            "DataRange": "0 to 3",
            "EndBit": 7,
            "Name": "Park Brake Release Inhibit Request",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3807,
            "SPNLength": 2,
            "StartBit": 6,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3808: {
            "Acronym": "B",
            "DataRange": "0 to 3",
            "EndBit": 29,
            "Name": "Park Brake Release Inhibit Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3808,
            "SPNLength": 2,
            "StartBit": 28,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        3809: {
            "Acronym": "CM3",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Transmission Oil Level Request",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3809,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "Every 10 s and on change of state but no faster than every 100 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: On change or every 10 s",
            "Units": "bit"
        },
        3810: {
            "Acronym": "DRC",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Retract Status of ramp 1",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3810,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "Every 1 s while active and on change of state but no faster than every 100 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010:  1 s when active and on change of state",
            "Units": "bit"
        },
        3811: {
            "Acronym": "DRC",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Enable status of ramp 1",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3811,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "Every 1 s while active and on change of state but no faster than every 100 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010:  1 s when active and on change of state",
            "Units": "bit"
        },
        3812: {
            "Acronym": "DRC",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Movement status of ramp 1",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3812,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "Every 1 s while active and on change of state but no faster than every 100 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010:  1 s when active and on change of state",
            "Units": "bit"
        },
        3813: {
            "Acronym": "DRC",
            "DataRange": "0 to 3",
            "EndBit": 9,
            "Name": "Retract Status of ramp 2",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3813,
            "SPNLength": 2,
            "StartBit": 8,
            "TransmissionRate": "Every 1 s while active and on change of state but no faster than every 100 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010:  1 s when active and on change of state",
            "Units": "bit"
        },
        3814: {
            "Acronym": "DRC",
            "DataRange": "0 to 3",
            "EndBit": 11,
            "Name": "Enable status of ramp 2",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3814,
            "SPNLength": 2,
            "StartBit": 10,
            "TransmissionRate": "Every 1 s while active and on change of state but no faster than every 100 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010:  1 s when active and on change of state",
            "Units": "bit"
        },
        3815: {
            "Acronym": "DRC",
            "DataRange": "0 to 3",
            "EndBit": 13,
            "Name": "Movement status of ramp 2",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3815,
            "SPNLength": 2,
            "StartBit": 12,
            "TransmissionRate": "Every 1 s while active and on change of state but no faster than every 100 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010:  1 s when active and on change of state",
            "Units": "bit"
        },
        3816: {
            "Acronym": "DRC",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Retract Status of ramp 3",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3816,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "Every 1 s while active and on change of state but no faster than every 100 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010:  1 s when active and on change of state",
            "Units": "bit"
        },
        3817: {
            "Acronym": "DRC",
            "DataRange": "0 to 3",
            "EndBit": 19,
            "Name": "Enable status of ramp 3",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3817,
            "SPNLength": 2,
            "StartBit": 18,
            "TransmissionRate": "Every 1 s while active and on change of state but no faster than every 100 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010:  1 s when active and on change of state",
            "Units": "bit"
        },
        3818: {
            "Acronym": "DRC",
            "DataRange": "0 to 3",
            "EndBit": 21,
            "Name": "Movement status of ramp 3",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3818,
            "SPNLength": 2,
            "StartBit": 20,
            "TransmissionRate": "Every 1 s while active and on change of state but no faster than every 100 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010:  1 s when active and on change of state",
            "Units": "bit"
        },
        3819: {
            "Acronym": "EAC1",
            "DataRange": "0 to 3",
            "EndBit": 25,
            "Name": "Front axle group engagement status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3819,
            "SPNLength": 2,
            "StartBit": 24,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        3820: {
            "Acronym": "EAC1",
            "DataRange": "0 to 3",
            "EndBit": 27,
            "Name": "Rear axle group engagement status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3820,
            "SPNLength": 2,
            "StartBit": 26,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        3821: {
            "Acronym": "EEC8",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 15,
            "Name": "Engine Exhaust Gas Recirculation 1 (EGR1) Valve 2 Control",
            "Offset": 0.0,
            "OperationalHigh": 160.6375,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 3821,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        3822: {
            "Acronym": "EEC7",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 31,
            "Name": "Engine Exhaust Gas Recirculation 1 Valve 2 Position",
            "Offset": 0.0,
            "OperationalHigh": 160.6375,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 3822,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        3823: {
            "Acronym": "TRF2",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 23,
            "Name": "Transmission Torque Converter Oil Outlet Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 3823,
            "SPNLength": 16,
            "StartBit": 8,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        3826: {
            "Acronym": "SCR1",
            "DataRange": "0 to 3,212.75 L/h",
            "EndBit": 15,
            "Name": "Aftertreatment 1 Diesel Exhaust Fluid Average Consumption",
            "Offset": 0.0,
            "OperationalHigh": 3212.75,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05,
            "SPN": 3826,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "l/h"
        },
        3828: {
            "Acronym": "SCR1",
            "DataRange": "0 to 3,212.75 L/h",
            "EndBit": 31,
            "Name": "Aftertreatment 1 Commanded Diesel Exhaust Fluid Consumption",
            "Offset": 0.0,
            "OperationalHigh": 3212.75,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05,
            "SPN": 3828,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "l/h"
        },
        3830: {
            "Acronym": "AT1AC2",
            "DataRange": "-36.2594345 to 36.54805959862 psi",
            "EndBit": 15,
            "Name": "Aftertreatment 1 Secondary Air Differential Pressure",
            "Offset": -36.2594345,
            "OperationalHigh": 36.54805959862,
            "OperationalLow": -36.2594345,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.001133107,
            "SPN": 3830,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "500 msec",
            "Units": "psi"
        },
        3831: {
            "Acronym": "AT1AC2",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 31,
            "Name": "Aftertreatment 1 Secondary Air Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 3831,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "500 msec",
            "Units": "deg F"
        },
        3832: {
            "Acronym": "AT1AC2",
            "DataRange": "0 to 12851 kg/h per bit",
            "EndBit": 47,
            "Name": "Aftertreatment 1 Secondary Air Mass Flow",
            "Offset": 0.0,
            "OperationalHigh": 12851.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.2,
            "SPN": 3832,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "500 msec",
            "Units": "kg/h"
        },
        3833: {
            "Acronym": "AT2AC2",
            "DataRange": "-36.2594345 to 36.54805959862 psi",
            "EndBit": 15,
            "Name": "Aftertreatment 2 Secondary Air Differential Pressure",
            "Offset": -36.2594345,
            "OperationalHigh": 36.54805959862,
            "OperationalLow": -36.2594345,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.001133107,
            "SPN": 3833,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "500 msec",
            "Units": "psi"
        },
        3834: {
            "Acronym": "AT2AC2",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 31,
            "Name": "Aftertreatment 2 Secondary Air Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 3834,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "500 msec",
            "Units": "deg F"
        },
        3835: {
            "Acronym": "AT2AC2",
            "DataRange": "0 to 12851 kg/h per bit",
            "EndBit": 47,
            "Name": "Aftertreatment 2 Secondary Air Mass Flow",
            "Offset": 0.0,
            "OperationalHigh": 12851.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.2,
            "SPN": 3835,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "500 msec",
            "Units": "kg/h"
        },
        3837: {
            "Acronym": "AT1AC2",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 63,
            "Name": "Aftertreatment 1 Secondary Air Pressure",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 3837,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "500 msec",
            "Units": "psi"
        },
        3838: {
            "Acronym": "AT2AC2",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 63,
            "Name": "Aftertreatment 2 Secondary Air Pressure",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 3838,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "500 msec",
            "Units": "psi"
        },
        3839: {
            "Acronym": "EBC5",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Brake Temperature Warning",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3839,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3840: {
            "Acronym": "AUXIO2",
            "DataRange": "0 to 3",
            "EndBit": 7,
            "Name": "Auxiliary I/O #17",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3840,
            "SPNLength": 2,
            "StartBit": 6,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3841: {
            "Acronym": "AUXIO2",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Auxiliary I/O #18",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3841,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3842: {
            "Acronym": "AUXIO2",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Auxiliary I/O #19",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3842,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3843: {
            "Acronym": "AUXIO2",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Auxiliary I/O #20",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3843,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3844: {
            "Acronym": "AUXIO2",
            "DataRange": "0 to 3",
            "EndBit": 15,
            "Name": "Auxiliary I/O #21",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3844,
            "SPNLength": 2,
            "StartBit": 14,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3845: {
            "Acronym": "AUXIO2",
            "DataRange": "0 to 3",
            "EndBit": 13,
            "Name": "Auxiliary I/O #22",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3845,
            "SPNLength": 2,
            "StartBit": 12,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3846: {
            "Acronym": "AUXIO2",
            "DataRange": "0 to 3",
            "EndBit": 11,
            "Name": "Auxiliary I/O #23",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3846,
            "SPNLength": 2,
            "StartBit": 10,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3847: {
            "Acronym": "AUXIO2",
            "DataRange": "0 to 3",
            "EndBit": 9,
            "Name": "Auxiliary I/O #24",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3847,
            "SPNLength": 2,
            "StartBit": 8,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3848: {
            "Acronym": "AUXIO2",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Auxiliary I/O #25",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3848,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3849: {
            "Acronym": "AUXIO2",
            "DataRange": "0 to 3",
            "EndBit": 21,
            "Name": "Auxiliary I/O #26",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3849,
            "SPNLength": 2,
            "StartBit": 20,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3850: {
            "Acronym": "AUXIO2",
            "DataRange": "0 to 3",
            "EndBit": 19,
            "Name": "Auxiliary I/O #27",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3850,
            "SPNLength": 2,
            "StartBit": 18,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3851: {
            "Acronym": "AUXIO2",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Auxiliary I/O #28",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3851,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3852: {
            "Acronym": "AUXIO2",
            "DataRange": "0 to 3",
            "EndBit": 31,
            "Name": "Auxiliary I/O #29",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3852,
            "SPNLength": 2,
            "StartBit": 30,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3853: {
            "Acronym": "AUXIO2",
            "DataRange": "0 to 3",
            "EndBit": 29,
            "Name": "Auxiliary I/O #30",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3853,
            "SPNLength": 2,
            "StartBit": 28,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3854: {
            "Acronym": "AUXIO2",
            "DataRange": "0 to 3",
            "EndBit": 27,
            "Name": "Auxiliary I/O #31",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3854,
            "SPNLength": 2,
            "StartBit": 26,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3855: {
            "Acronym": "AUXIO2",
            "DataRange": "0 to 3",
            "EndBit": 25,
            "Name": "Auxiliary I/O #32",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3855,
            "SPNLength": 2,
            "StartBit": 24,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3856: {
            "Acronym": "AUXIO2",
            "DataRange": "0 to 3",
            "EndBit": 39,
            "Name": "Auxiliary I/O #33",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3856,
            "SPNLength": 2,
            "StartBit": 38,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3857: {
            "Acronym": "AUXIO2",
            "DataRange": "0 to 3",
            "EndBit": 37,
            "Name": "Auxiliary I/O #34",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3857,
            "SPNLength": 2,
            "StartBit": 36,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3858: {
            "Acronym": "AUXIO2",
            "DataRange": "0 to 3",
            "EndBit": 35,
            "Name": "Auxiliary I/O #35",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3858,
            "SPNLength": 2,
            "StartBit": 34,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3859: {
            "Acronym": "AUXIO2",
            "DataRange": "0 to 3",
            "EndBit": 33,
            "Name": "Auxiliary I/O #36",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3859,
            "SPNLength": 2,
            "StartBit": 32,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3860: {
            "Acronym": "AUXIO2",
            "DataRange": "0 to 3",
            "EndBit": 47,
            "Name": "Auxiliary I/O #37",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3860,
            "SPNLength": 2,
            "StartBit": 46,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3861: {
            "Acronym": "AUXIO2",
            "DataRange": "0 to 3",
            "EndBit": 45,
            "Name": "Auxiliary I/O #38",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3861,
            "SPNLength": 2,
            "StartBit": 44,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3862: {
            "Acronym": "AUXIO2",
            "DataRange": "0 to 3",
            "EndBit": 43,
            "Name": "Auxiliary I/O #39",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3862,
            "SPNLength": 2,
            "StartBit": 42,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3863: {
            "Acronym": "AUXIO2",
            "DataRange": "0 to 3",
            "EndBit": 41,
            "Name": "Auxiliary I/O #40",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3863,
            "SPNLength": 2,
            "StartBit": 40,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3864: {
            "Acronym": "AUXIO2",
            "DataRange": "0 to 3",
            "EndBit": 55,
            "Name": "Auxiliary I/O #41",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3864,
            "SPNLength": 2,
            "StartBit": 54,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3865: {
            "Acronym": "AUXIO2",
            "DataRange": "0 to 3",
            "EndBit": 53,
            "Name": "Auxiliary I/O #42",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3865,
            "SPNLength": 2,
            "StartBit": 52,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3866: {
            "Acronym": "AUXIO2",
            "DataRange": "0 to 3",
            "EndBit": 51,
            "Name": "Auxiliary I/O #43",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3866,
            "SPNLength": 2,
            "StartBit": 50,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3867: {
            "Acronym": "AUXIO2",
            "DataRange": "0 to 3",
            "EndBit": 49,
            "Name": "Auxiliary I/O #44",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3867,
            "SPNLength": 2,
            "StartBit": 48,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3868: {
            "Acronym": "AUXIO2",
            "DataRange": "0 to 3",
            "EndBit": 63,
            "Name": "Auxiliary I/O #45",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3868,
            "SPNLength": 2,
            "StartBit": 62,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3869: {
            "Acronym": "AUXIO2",
            "DataRange": "0 to 3",
            "EndBit": 61,
            "Name": "Auxiliary I/O #46",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3869,
            "SPNLength": 2,
            "StartBit": 60,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3870: {
            "Acronym": "AUXIO2",
            "DataRange": "0 to 3",
            "EndBit": 59,
            "Name": "Auxiliary I/O #47",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3870,
            "SPNLength": 2,
            "StartBit": 58,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3871: {
            "Acronym": "AUXIO2",
            "DataRange": "0 to 3",
            "EndBit": 57,
            "Name": "Auxiliary I/O #48",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3871,
            "SPNLength": 2,
            "StartBit": 56,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3872: {
            "Acronym": "AUXIO3",
            "DataRange": "0 to 3",
            "EndBit": 7,
            "Name": "Auxiliary I/O #49",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3872,
            "SPNLength": 2,
            "StartBit": 6,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3873: {
            "Acronym": "AUXIO3",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Auxiliary I/O #50",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3873,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3874: {
            "Acronym": "AUXIO3",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Auxiliary I/O #51",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3874,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3875: {
            "Acronym": "AUXIO3",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Auxiliary I/O #52",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3875,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3876: {
            "Acronym": "AUXIO3",
            "DataRange": "0 to 3",
            "EndBit": 15,
            "Name": "Auxiliary I/O #53",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3876,
            "SPNLength": 2,
            "StartBit": 14,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3877: {
            "Acronym": "AUXIO3",
            "DataRange": "0 to 3",
            "EndBit": 13,
            "Name": "Auxiliary I/O #54",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3877,
            "SPNLength": 2,
            "StartBit": 12,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3878: {
            "Acronym": "AUXIO3",
            "DataRange": "0 to 3",
            "EndBit": 11,
            "Name": "Auxiliary I/O #55",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3878,
            "SPNLength": 2,
            "StartBit": 10,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3879: {
            "Acronym": "AUXIO3",
            "DataRange": "0 to 3",
            "EndBit": 9,
            "Name": "Auxiliary I/O #56",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3879,
            "SPNLength": 2,
            "StartBit": 8,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3880: {
            "Acronym": "AUXIO3",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Auxiliary I/O #57",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3880,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3881: {
            "Acronym": "AUXIO3",
            "DataRange": "0 to 3",
            "EndBit": 21,
            "Name": "Auxiliary I/O #58",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3881,
            "SPNLength": 2,
            "StartBit": 20,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3882: {
            "Acronym": "AUXIO3",
            "DataRange": "0 to 3",
            "EndBit": 19,
            "Name": "Auxiliary I/O #59",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3882,
            "SPNLength": 2,
            "StartBit": 18,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3883: {
            "Acronym": "AUXIO3",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Auxiliary I/O #60",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3883,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3884: {
            "Acronym": "AUXIO3",
            "DataRange": "0 to 3",
            "EndBit": 31,
            "Name": "Auxiliary I/O #61",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3884,
            "SPNLength": 2,
            "StartBit": 30,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3885: {
            "Acronym": "AUXIO3",
            "DataRange": "0 to 3",
            "EndBit": 29,
            "Name": "Auxiliary I/O #62",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3885,
            "SPNLength": 2,
            "StartBit": 28,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3886: {
            "Acronym": "AUXIO3",
            "DataRange": "0 to 3",
            "EndBit": 27,
            "Name": "Auxiliary I/O #63",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3886,
            "SPNLength": 2,
            "StartBit": 26,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3887: {
            "Acronym": "AUXIO3",
            "DataRange": "0 to 3",
            "EndBit": 25,
            "Name": "Auxiliary I/O #64",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3887,
            "SPNLength": 2,
            "StartBit": 24,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3888: {
            "Acronym": "AUXIO3",
            "DataRange": "0 to 3",
            "EndBit": 39,
            "Name": "Auxiliary I/O #65",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3888,
            "SPNLength": 2,
            "StartBit": 38,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3889: {
            "Acronym": "AUXIO3",
            "DataRange": "0 to 3",
            "EndBit": 37,
            "Name": "Auxiliary I/O #66",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3889,
            "SPNLength": 2,
            "StartBit": 36,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3890: {
            "Acronym": "AUXIO3",
            "DataRange": "0 to 3",
            "EndBit": 35,
            "Name": "Auxiliary I/O #67",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3890,
            "SPNLength": 2,
            "StartBit": 34,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3891: {
            "Acronym": "AUXIO3",
            "DataRange": "0 to 3",
            "EndBit": 33,
            "Name": "Auxiliary I/O #68",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3891,
            "SPNLength": 2,
            "StartBit": 32,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3892: {
            "Acronym": "AUXIO3",
            "DataRange": "0 to 3",
            "EndBit": 47,
            "Name": "Auxiliary I/O #69",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3892,
            "SPNLength": 2,
            "StartBit": 46,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3893: {
            "Acronym": "AUXIO3",
            "DataRange": "0 to 3",
            "EndBit": 45,
            "Name": "Auxiliary I/O #70",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3893,
            "SPNLength": 2,
            "StartBit": 44,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3894: {
            "Acronym": "AUXIO3",
            "DataRange": "0 to 3",
            "EndBit": 43,
            "Name": "Auxiliary I/O #71",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3894,
            "SPNLength": 2,
            "StartBit": 42,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3895: {
            "Acronym": "AUXIO3",
            "DataRange": "0 to 3",
            "EndBit": 41,
            "Name": "Auxiliary I/O #72",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3895,
            "SPNLength": 2,
            "StartBit": 40,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3896: {
            "Acronym": "AUXIO3",
            "DataRange": "0 to 3",
            "EndBit": 55,
            "Name": "Auxiliary I/O #73",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3896,
            "SPNLength": 2,
            "StartBit": 54,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3897: {
            "Acronym": "AUXIO3",
            "DataRange": "0 to 3",
            "EndBit": 53,
            "Name": "Auxiliary I/O #74",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3897,
            "SPNLength": 2,
            "StartBit": 52,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3898: {
            "Acronym": "AUXIO3",
            "DataRange": "0 to 3",
            "EndBit": 51,
            "Name": "Auxiliary I/O #75",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3898,
            "SPNLength": 2,
            "StartBit": 50,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3899: {
            "Acronym": "AUXIO3",
            "DataRange": "0 to 3",
            "EndBit": 49,
            "Name": "Auxiliary I/O #76",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3899,
            "SPNLength": 2,
            "StartBit": 48,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        39: {
            "Acronym": "TP1",
            "DataRange": "0 to 250 mins",
            "EndBit": 7,
            "Name": "Tire Pressure Check Interval",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 39,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "minutes"
        },
        3900: {
            "Acronym": "AUXIO3",
            "DataRange": "0 to 3",
            "EndBit": 63,
            "Name": "Auxiliary I/O #77",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3900,
            "SPNLength": 2,
            "StartBit": 62,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3901: {
            "Acronym": "AUXIO3",
            "DataRange": "0 to 3",
            "EndBit": 61,
            "Name": "Auxiliary I/O #78",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3901,
            "SPNLength": 2,
            "StartBit": 60,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3902: {
            "Acronym": "AUXIO3",
            "DataRange": "0 to 3",
            "EndBit": 59,
            "Name": "Auxiliary I/O #79",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3902,
            "SPNLength": 2,
            "StartBit": 58,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3903: {
            "Acronym": "AUXIO3",
            "DataRange": "0 to 3",
            "EndBit": 57,
            "Name": "Auxiliary I/O #80",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3903,
            "SPNLength": 2,
            "StartBit": 56,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3904: {
            "Acronym": "AUXIO4",
            "DataRange": "0 to 3",
            "EndBit": 7,
            "Name": "Auxiliary I/O #81",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3904,
            "SPNLength": 2,
            "StartBit": 6,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3905: {
            "Acronym": "AUXIO4",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Auxiliary I/O #82",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3905,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3906: {
            "Acronym": "AUXIO4",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Auxiliary I/O #83",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3906,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3907: {
            "Acronym": "AUXIO4",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Auxiliary I/O #84",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3907,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3908: {
            "Acronym": "AUXIO4",
            "DataRange": "0 to 3",
            "EndBit": 15,
            "Name": "Auxiliary I/O #85",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3908,
            "SPNLength": 2,
            "StartBit": 14,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3909: {
            "Acronym": "AUXIO4",
            "DataRange": "0 to 3",
            "EndBit": 13,
            "Name": "Auxiliary I/O #86",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3909,
            "SPNLength": 2,
            "StartBit": 12,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3910: {
            "Acronym": "AUXIO4",
            "DataRange": "0 to 3",
            "EndBit": 11,
            "Name": "Auxiliary I/O #87",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3910,
            "SPNLength": 2,
            "StartBit": 10,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3911: {
            "Acronym": "AUXIO4",
            "DataRange": "0 to 3",
            "EndBit": 9,
            "Name": "Auxiliary I/O #88",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3911,
            "SPNLength": 2,
            "StartBit": 8,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3912: {
            "Acronym": "AUXIO4",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Auxiliary I/O #89",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3912,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3913: {
            "Acronym": "AUXIO4",
            "DataRange": "0 to 3",
            "EndBit": 21,
            "Name": "Auxiliary I/O #90",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3913,
            "SPNLength": 2,
            "StartBit": 20,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3914: {
            "Acronym": "AUXIO4",
            "DataRange": "0 to 3",
            "EndBit": 19,
            "Name": "Auxiliary I/O #91",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3914,
            "SPNLength": 2,
            "StartBit": 18,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3915: {
            "Acronym": "AUXIO4",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Auxiliary I/O #92",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3915,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3916: {
            "Acronym": "AUXIO4",
            "DataRange": "0 to 3",
            "EndBit": 31,
            "Name": "Auxiliary I/O #93",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3916,
            "SPNLength": 2,
            "StartBit": 30,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3917: {
            "Acronym": "AUXIO4",
            "DataRange": "0 to 3",
            "EndBit": 29,
            "Name": "Auxiliary I/O #94",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3917,
            "SPNLength": 2,
            "StartBit": 28,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3918: {
            "Acronym": "AUXIO4",
            "DataRange": "0 to 3",
            "EndBit": 27,
            "Name": "Auxiliary I/O #95",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3918,
            "SPNLength": 2,
            "StartBit": 26,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3919: {
            "Acronym": "AUXIO4",
            "DataRange": "0 to 3",
            "EndBit": 25,
            "Name": "Auxiliary I/O #96",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3919,
            "SPNLength": 2,
            "StartBit": 24,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3920: {
            "Acronym": "AUXIO4",
            "DataRange": "0 to 3",
            "EndBit": 39,
            "Name": "Auxiliary I/O #97",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3920,
            "SPNLength": 2,
            "StartBit": 38,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3921: {
            "Acronym": "AUXIO4",
            "DataRange": "0 to 3",
            "EndBit": 37,
            "Name": "Auxiliary I/O #98",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3921,
            "SPNLength": 2,
            "StartBit": 36,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3922: {
            "Acronym": "AUXIO4",
            "DataRange": "0 to 3",
            "EndBit": 35,
            "Name": "Auxiliary I/O #99",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3922,
            "SPNLength": 2,
            "StartBit": 34,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3923: {
            "Acronym": "AUXIO4",
            "DataRange": "0 to 3",
            "EndBit": 33,
            "Name": "Auxiliary I/O #100",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3923,
            "SPNLength": 2,
            "StartBit": 32,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3924: {
            "Acronym": "AUXIO4",
            "DataRange": "0 to 3",
            "EndBit": 47,
            "Name": "Auxiliary I/O #101",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3924,
            "SPNLength": 2,
            "StartBit": 46,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3925: {
            "Acronym": "AUXIO4",
            "DataRange": "0 to 3",
            "EndBit": 45,
            "Name": "Auxiliary I/O #102",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3925,
            "SPNLength": 2,
            "StartBit": 44,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3926: {
            "Acronym": "AUXIO4",
            "DataRange": "0 to 3",
            "EndBit": 43,
            "Name": "Auxiliary I/O #103",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3926,
            "SPNLength": 2,
            "StartBit": 42,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3927: {
            "Acronym": "AUXIO4",
            "DataRange": "0 to 3",
            "EndBit": 41,
            "Name": "Auxiliary I/O #104",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3927,
            "SPNLength": 2,
            "StartBit": 40,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3928: {
            "Acronym": "AUXIO4",
            "DataRange": "0 to 3",
            "EndBit": 55,
            "Name": "Auxiliary I/O #105",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3928,
            "SPNLength": 2,
            "StartBit": 54,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3929: {
            "Acronym": "AUXIO4",
            "DataRange": "0 to 3",
            "EndBit": 53,
            "Name": "Auxiliary I/O #106",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3929,
            "SPNLength": 2,
            "StartBit": 52,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3930: {
            "Acronym": "AUXIO4",
            "DataRange": "0 to 3",
            "EndBit": 51,
            "Name": "Auxiliary I/O #107",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3930,
            "SPNLength": 2,
            "StartBit": 50,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3931: {
            "Acronym": "AUXIO4",
            "DataRange": "0 to 3",
            "EndBit": 49,
            "Name": "Auxiliary I/O #108",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3931,
            "SPNLength": 2,
            "StartBit": 48,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3932: {
            "Acronym": "AUXIO4",
            "DataRange": "0 to 3",
            "EndBit": 63,
            "Name": "Auxiliary I/O #109",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3932,
            "SPNLength": 2,
            "StartBit": 62,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3933: {
            "Acronym": "AUXIO4",
            "DataRange": "0 to 3",
            "EndBit": 61,
            "Name": "Auxiliary I/O #110",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3933,
            "SPNLength": 2,
            "StartBit": 60,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3934: {
            "Acronym": "AUXIO4",
            "DataRange": "0 to 3",
            "EndBit": 59,
            "Name": "Auxiliary I/O #111",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3934,
            "SPNLength": 2,
            "StartBit": 58,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3935: {
            "Acronym": "AUXIO4",
            "DataRange": "0 to 3",
            "EndBit": 57,
            "Name": "Auxiliary I/O #112",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3935,
            "SPNLength": 2,
            "StartBit": 56,
            "TransmissionRate": "Manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        3938: {
            "Acronym": "GC2",
            "DataRange": "-125 to 132.02%",
            "EndBit": 15,
            "Name": "Generator Governing Bias",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "-125 to 125%",
            "PGNLength": 8,
            "Resolution": 0.004,
            "SPN": 3938,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "20 ms",
            "Units": "%"
        },
        3939: {
            "Acronym": "PTODE",
            "DataRange": "0 to 3",
            "EndBit": 9,
            "Name": "Enable Switch - PTO Engine Flywheel",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3939,
            "SPNLength": 2,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3940: {
            "Acronym": "PTODE",
            "DataRange": "0 to 3",
            "EndBit": 25,
            "Name": "Engagement Consent - PTO Engine Flywheel",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3940,
            "SPNLength": 2,
            "StartBit": 24,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3941: {
            "Acronym": "PTODE",
            "DataRange": "0 to 3",
            "EndBit": 41,
            "Name": "Engagement Status - PTO Engine Flywheel",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3941,
            "SPNLength": 2,
            "StartBit": 40,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3942: {
            "Acronym": "PTODE",
            "DataRange": "0 to 3",
            "EndBit": 11,
            "Name": "Enable Switch - PTO Engine Accessory Drive 1",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3942,
            "SPNLength": 2,
            "StartBit": 10,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3943: {
            "Acronym": "PTODE",
            "DataRange": "0 to 3",
            "EndBit": 27,
            "Name": "Engagement Consent - PTO Engine Accessory Drive 1",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3943,
            "SPNLength": 2,
            "StartBit": 26,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3944: {
            "Acronym": "PTODE",
            "DataRange": "0 to 3",
            "EndBit": 43,
            "Name": "Engagement Status - PTO Engine Accessory Drive 1",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3944,
            "SPNLength": 2,
            "StartBit": 42,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3945: {
            "Acronym": "PTODE",
            "DataRange": "0 to 3",
            "EndBit": 13,
            "Name": "Enable Switch - PTO Engine Accessory Drive 2",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3945,
            "SPNLength": 2,
            "StartBit": 12,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3946: {
            "Acronym": "PTODE",
            "DataRange": "0 to 3",
            "EndBit": 29,
            "Name": "Engagement Consent - PTO Engine Accessory Drive 2",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3946,
            "SPNLength": 2,
            "StartBit": 28,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3947: {
            "Acronym": "PTODE",
            "DataRange": "0 to 3",
            "EndBit": 45,
            "Name": "Engagement Status - PTO Engine Accessory Drive 2",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3947,
            "SPNLength": 2,
            "StartBit": 44,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3948: {
            "Acronym": "PTODE",
            "DataRange": "0 to 3",
            "EndBit": 49,
            "Name": "At least one PTO engaged",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3948,
            "SPNLength": 2,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        3987: {
            "Acronym": "DLCC1",
            "DataRange": "0 to 3",
            "EndBit": 11,
            "Name": "Compression Brake Enable Switch Indicator Lamp Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 3987,
            "SPNLength": 2,
            "StartBit": 10,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms",
            "Units": "bit"
        },
        4059: {
            "Acronym": "AAGW",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Steer Axle Group Weight Available",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4059,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "As needed.  Broadcast whenever an axle group equipped with an on-board scale joined or left the on-board scale subset.",
            "Units": "bit"
        },
        4060: {
            "Acronym": "AAGW",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Lift Axle Group Weight Available",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4060,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "As needed.  Broadcast whenever an axle group equipped with an on-board scale joined or left the on-board scale subset.",
            "Units": "bit"
        },
        4061: {
            "Acronym": "AAGW",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Drive Axle Group Weight Available",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4061,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "As needed.  Broadcast whenever an axle group equipped with an on-board scale joined or left the on-board scale subset.",
            "Units": "bit"
        },
        4062: {
            "Acronym": "AAGW",
            "DataRange": "0 to 3",
            "EndBit": 7,
            "Name": "Tag Axle Group Weight Available",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4062,
            "SPNLength": 2,
            "StartBit": 6,
            "TransmissionRate": "As needed.  Broadcast whenever an axle group equipped with an on-board scale joined or left the on-board scale subset.",
            "Units": "bit"
        },
        4063: {
            "Acronym": "AAGW",
            "DataRange": "0 to 3",
            "EndBit": 9,
            "Name": "Additional Tractor Axle Group Weight Available",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4063,
            "SPNLength": 2,
            "StartBit": 8,
            "TransmissionRate": "As needed.  Broadcast whenever an axle group equipped with an on-board scale joined or left the on-board scale subset.",
            "Units": "bit"
        },
        4064: {
            "Acronym": "AAGW",
            "DataRange": "0 to 3",
            "EndBit": 11,
            "Name": "Trailer A Axle Group Weight Available",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4064,
            "SPNLength": 2,
            "StartBit": 10,
            "TransmissionRate": "As needed.  Broadcast whenever an axle group equipped with an on-board scale joined or left the on-board scale subset.",
            "Units": "bit"
        },
        4065: {
            "Acronym": "AAGW",
            "DataRange": "0 to 3",
            "EndBit": 13,
            "Name": "Trailer B Axle Group Weight Available",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4065,
            "SPNLength": 2,
            "StartBit": 12,
            "TransmissionRate": "As needed.  Broadcast whenever an axle group equipped with an on-board scale joined or left the on-board scale subset.",
            "Units": "bit"
        },
        4066: {
            "Acronym": "AAGW",
            "DataRange": "0 to 3",
            "EndBit": 15,
            "Name": "Trailer C Axle Group Weight Available",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4066,
            "SPNLength": 2,
            "StartBit": 14,
            "TransmissionRate": "As needed.  Broadcast whenever an axle group equipped with an on-board scale joined or left the on-board scale subset.",
            "Units": "bit"
        },
        4067: {
            "Acronym": "AAGW",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Trailer D Axle Group Weight Available",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4067,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "As needed.  Broadcast whenever an axle group equipped with an on-board scale joined or left the on-board scale subset.",
            "Units": "bit"
        },
        4068: {
            "Acronym": "AAGW",
            "DataRange": "0 to 3",
            "EndBit": 19,
            "Name": "Trailer E Axle Group Weight Available",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4068,
            "SPNLength": 2,
            "StartBit": 18,
            "TransmissionRate": "As needed.  Broadcast whenever an axle group equipped with an on-board scale joined or left the on-board scale subset.",
            "Units": "bit"
        },
        4069: {
            "Acronym": "AAGW",
            "DataRange": "0 to 3",
            "EndBit": 21,
            "Name": "Trailer F Axle Group Weight Available",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4069,
            "SPNLength": 2,
            "StartBit": 20,
            "TransmissionRate": "As needed.  Broadcast whenever an axle group equipped with an on-board scale joined or left the on-board scale subset.",
            "Units": "bit"
        },
        407: {
            "Acronym": "AGCW",
            "DataRange": "0 to 128,510 kg",
            "EndBit": 39,
            "Name": "Axle Group Full Weight Calibration",
            "Offset": 0.0,
            "OperationalHigh": 128510.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 2,
            "SPN": 407,
            "SPNLength": 16,
            "StartBit": 24,
            "TransmissionRate": "On request.  Upon request, will be broadcast as many times as required to transmit all available axle groups.",
            "Units": "kg"
        },
        4070: {
            "Acronym": "AAGW",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Trailer G Axle Group Weight Available",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4070,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "As needed.  Broadcast whenever an axle group equipped with an on-board scale joined or left the on-board scale subset.",
            "Units": "bit"
        },
        4071: {
            "Acronym": "AAGW",
            "DataRange": "0 to 3",
            "EndBit": 25,
            "Name": "Trailer H Axle Group Weight Available",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4071,
            "SPNLength": 2,
            "StartBit": 24,
            "TransmissionRate": "As needed.  Broadcast whenever an axle group equipped with an on-board scale joined or left the on-board scale subset.",
            "Units": "bit"
        },
        4072: {
            "Acronym": "AAGW",
            "DataRange": "0 to 3",
            "EndBit": 27,
            "Name": "Additional Trailer Axle Group Weight Available",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4072,
            "SPNLength": 2,
            "StartBit": 26,
            "TransmissionRate": "As needed.  Broadcast whenever an axle group equipped with an on-board scale joined or left the on-board scale subset.",
            "Units": "bit"
        },
        4073: {
            "Acronym": "AGW",
            "DataRange": "0 to 15",
            "EndBit": 3,
            "Name": "Axle Group Location",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4073,
            "SPNLength": 4,
            "StartBit": 0,
            "TransmissionRate": "On request.  Upon request, will be broadcast as many times as required to transmit all available axle groups.",
            "Units": "bit"
        },
        4074: {
            "Acronym": "AGCW",
            "DataRange": "0 to 15",
            "EndBit": 3,
            "Name": "Axle Group Location",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4074,
            "SPNLength": 4,
            "StartBit": 0,
            "TransmissionRate": "On request.  Upon request, will be broadcast as many times as required to transmit all available axle groups.",
            "Units": "bit"
        },
        4075: {
            "Acronym": "ZNVW",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Zero Net Vehicle Weight Change",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4075,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "As needed",
            "Units": "bit"
        },
        4076: {
            "Acronym": "ET4",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 7,
            "Name": "Engine Coolant Temperature 2",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 4076,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        4077: {
            "Acronym": "AT1FC2",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 15,
            "Name": "Aftertreatment 1 Fuel Pressure 2",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 4077,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        4078: {
            "Acronym": "GC1",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 23,
            "Name": "Generator Alternator Efficiency",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 100%",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 4078,
            "SPNLength": 16,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        4079: {
            "Acronym": "GC1",
            "DataRange": "0 to 3",
            "EndBit": 25,
            "Name": "Generator Governing Speed Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4079,
            "SPNLength": 2,
            "StartBit": 24,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        408: {
            "Acronym": "AGCW",
            "DataRange": "0 to 128,510 kg",
            "EndBit": 23,
            "Name": "Axle Group Empty Weight Calibration",
            "Offset": 0.0,
            "OperationalHigh": 128510.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 2,
            "SPN": 408,
            "SPNLength": 16,
            "StartBit": 8,
            "TransmissionRate": "On request.  Upon request, will be broadcast as many times as required to transmit all available axle groups.",
            "Units": "kg"
        },
        4080: {
            "Acronym": "GC1",
            "DataRange": "0 to 15",
            "EndBit": 29,
            "Name": "Generator Frequency Selection",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4080,
            "SPNLength": 4,
            "StartBit": 26,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        4082: {
            "Acronym": "EOI",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Fuel Pump Primer Control",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4082,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "250 ms",
            "Units": "bit"
        },
        4083: {
            "Acronym": "EFS",
            "DataRange": "0 to 3",
            "EndBit": 52,
            "Name": "Fuel Pump Primer Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4083,
            "SPNLength": 2,
            "StartBit": 51,
            "TransmissionRate": "0.5 s",
            "Units": "bit"
        },
        4086: {
            "Acronym": "GPV4",
            "DataRange": "0.0 to 46596.99927595 psi",
            "EndBit": 15,
            "Name": "Valve Load Sense Pressure",
            "Offset": 0.0,
            "OperationalHigh": 46596.99927595,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.72518869,
            "SPN": 4086,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        4087: {
            "Acronym": "GPV4",
            "DataRange": "0.0 to 580.150952 psi",
            "EndBit": 23,
            "Name": "Valve Pilot Pressure",
            "Offset": 0.0,
            "OperationalHigh": 580.150952,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 2.320603808,
            "SPN": 4087,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        4088: {
            "Acronym": "GPV4",
            "DataRange": "0.0 to 46596.99927595 psi",
            "EndBit": 39,
            "Name": "Valve Assembly Load sense Pressure",
            "Offset": 0.0,
            "OperationalHigh": 46596.99927595,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.72518869,
            "SPN": 4088,
            "SPNLength": 16,
            "StartBit": 24,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        4089: {
            "Acronym": "GPV4",
            "DataRange": "0.0 to 46596.99927595 psi",
            "EndBit": 55,
            "Name": "Valve Assembly Supply Pressure",
            "Offset": 0.0,
            "OperationalHigh": 46596.99927595,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.72518869,
            "SPN": 4089,
            "SPNLength": 16,
            "StartBit": 40,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        409: {
            "Acronym": "AGW",
            "DataRange": "0 to 128,510 kg",
            "EndBit": 23,
            "Name": "Axle Group Weight",
            "Offset": 0.0,
            "OperationalHigh": 128510.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 2,
            "SPN": 409,
            "SPNLength": 16,
            "StartBit": 8,
            "TransmissionRate": "On request.  Upon request, will be broadcast as many times as required to transmit all available axle groups.",
            "Units": "kg"
        },
        4097: {
            "Acronym": "AT1FC1",
            "DataRange": "0 to 3",
            "EndBit": 49,
            "Name": "Aftertreatment 1 Fuel Drain Actuator",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4097,
            "SPNLength": 2,
            "StartBit": 48,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        4098: {
            "Acronym": "AT2FC1",
            "DataRange": "0 to 3",
            "EndBit": 49,
            "Name": "Aftertreatment 2 Fuel Drain Actuator",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4098,
            "SPNLength": 2,
            "StartBit": 48,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        4099: {
            "Acronym": "XBR",
            "DataRange": "0 to 100 %",
            "EndBit": 31,
            "Name": "XBR urgency",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 4099,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "When active: 20 ms; else 200 ms",
            "Units": "%"
        },
        4100: {
            "Acronym": "DM24",
            "DataRange": "0 to 1",
            "EndBit": 16,
            "Name": "Supported in Expanded Freeze Frame (e.g., DM25)",
            "Offset": 0.0,
            "OperationalHigh": 1.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4100,
            "SPNLength": 1,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        4101: {
            "Acronym": "DM24",
            "DataRange": "0 to 1",
            "EndBit": 17,
            "Name": "Supported in Data Stream",
            "Offset": 0.0,
            "OperationalHigh": 1.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4101,
            "SPNLength": 1,
            "StartBit": 17,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        4102: {
            "Acronym": "DM24",
            "DataRange": "0 to 1",
            "EndBit": 18,
            "Name": "Supported in Scaled Test Results (e.g., DM30)",
            "Offset": 0.0,
            "OperationalHigh": 1.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4102,
            "SPNLength": 1,
            "StartBit": 18,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        4103: {
            "Acronym": "DM24",
            "DataRange": "0 to 250",
            "EndBit": 31,
            "Name": "SPN Data Length",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4103,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "On request",
            "Units": "count"
        },
        411: {
            "Acronym": "ET2",
            "DataRange": "-36.2594345 to 36.54805959862 psi",
            "EndBit": 47,
            "Name": "Engine Exhaust Gas Recirculation 1 Differential Pressure",
            "Offset": -36.2594345,
            "OperationalHigh": 36.54805959862,
            "OperationalLow": -36.2594345,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.001133107,
            "SPN": 411,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "1s",
            "Units": "psi"
        },
        412: {
            "Acronym": "ET2",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 63,
            "Name": "Engine Exhaust Gas Recirculation 1 Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 412,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "1s",
            "Units": "deg F"
        },
        413: {
            "Acronym": "GCVW",
            "DataRange": "-16,449,535 to 16,449,535 kg",
            "EndBit": 47,
            "Name": "Net Vehicle Weight Change",
            "Offset": -16449535.0,
            "OperationalHigh": 16449535.0,
            "OperationalLow": -16449535.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 2,
            "SPN": 413,
            "SPNLength": 24,
            "StartBit": 24,
            "TransmissionRate": "On request",
            "Units": "kg"
        },
        4138: {
            "Acronym": "DM36",
            "DataRange": "0 to 64,255 mins",
            "EndBit": 47,
            "Name": "Vehicle Current MI Accumulated Time",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "",
            "Resolution": 1,
            "SPN": 4138,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "",
            "Units": "minutes"
        },
        4148: {
            "Acronym": "DM07",
            "DataRange": "0 to 524,287",
            "EndBit": 26,
            "Name": "SPN Identifying Component / System",
            "Offset": 0.0,
            "OperationalHigh": 524287.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 4148,
            "SPNLength": 19,
            "StartBit": 8,
            "TransmissionRate": "Sent whenever a test is desired",
            "Units": "binary"
        },
        4149: {
            "Acronym": "DM07",
            "DataRange": "0 to 31",
            "EndBit": 28,
            "Name": "FMI Identifying Component / System Specific Test",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 4149,
            "SPNLength": 5,
            "StartBit": 24,
            "TransmissionRate": "Sent whenever a test is desired",
            "Units": "binary"
        },
        4151: {
            "Acronym": "EAI",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 15,
            "Name": "Engine Exhaust Gas Temperature Average",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 4151,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "500 ms",
            "Units": "deg F"
        },
        4152: {
            "Acronym": "EAI",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 47,
            "Name": "Engine Exhaust Gas Temperature Average - Bank 2",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 4152,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "500 ms",
            "Units": "deg F"
        },
        4153: {
            "Acronym": "EAI",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 31,
            "Name": "Engine Exhaust Gas Temperature Average - Bank 1",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 4153,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "500 ms",
            "Units": "deg F"
        },
        4154: {
            "Acronym": "EEC1",
            "DataRange": "0 to 0.875%",
            "EndBit": 7,
            "Name": "Actual Engine - Percent Torque High Resolution",
            "Offset": 0.0,
            "OperationalHigh": 0.875,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.125,
            "SPN": 4154,
            "SPNLength": 4,
            "StartBit": 4,
            "TransmissionRate": "engine speed dependent",
            "Units": "%"
        },
        4155: {
            "Acronym": "AUXIO5",
            "DataRange": "0 to 64,255",
            "EndBit": 15,
            "Name": "Auxiliary I/O Channel #6",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4155,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "Manufacturer defined, not faster than 20 ms",
            "Units": "count"
        },
        4156: {
            "Acronym": "AUXIO5",
            "DataRange": "0 to 64,255",
            "EndBit": 31,
            "Name": "Auxiliary I/O Channel #5",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4156,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "Manufacturer defined, not faster than 20 ms",
            "Units": "count"
        },
        4157: {
            "Acronym": "AUXIO5",
            "DataRange": "0 to 64,255",
            "EndBit": 47,
            "Name": "Auxiliary I/O Channel #4",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4157,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "Manufacturer defined, not faster than 20 ms",
            "Units": "count"
        },
        4158: {
            "Acronym": "AUXIO5",
            "DataRange": "0 to 64,255",
            "EndBit": 63,
            "Name": "Auxiliary I/O Channel #3",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4158,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "Manufacturer defined, not faster than 20 ms",
            "Units": "count"
        },
        4159: {
            "Acronym": "AUXIO6",
            "DataRange": "0 to 250",
            "EndBit": 7,
            "Name": "Auxiliary I/O Channel #10",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4159,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "manufacturer defined, not faster than 100 ms",
            "Units": "count"
        },
        4160: {
            "Acronym": "AUXIO6",
            "DataRange": "0 to 250",
            "EndBit": 15,
            "Name": "Auxiliary I/O Channel #9",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4160,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "manufacturer defined, not faster than 100 ms",
            "Units": "count"
        },
        4161: {
            "Acronym": "AUXIO6",
            "DataRange": "0 to 250",
            "EndBit": 23,
            "Name": "Auxiliary I/O Channel #8",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4161,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "manufacturer defined, not faster than 100 ms",
            "Units": "count"
        },
        4162: {
            "Acronym": "AUXIO6",
            "DataRange": "0 to 250",
            "EndBit": 31,
            "Name": "Auxiliary I/O Channel #7",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4162,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "manufacturer defined, not faster than 100 ms",
            "Units": "count"
        },
        4163: {
            "Acronym": "AUXIO6",
            "DataRange": "0 to 250",
            "EndBit": 39,
            "Name": "Auxiliary I/O Channel #14",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4163,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "manufacturer defined, not faster than 100 ms",
            "Units": "count"
        },
        4164: {
            "Acronym": "AUXIO6",
            "DataRange": "0 to 250",
            "EndBit": 47,
            "Name": "Auxiliary I/O Channel #13",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4164,
            "SPNLength": 8,
            "StartBit": 40,
            "TransmissionRate": "manufacturer defined, not faster than 100 ms",
            "Units": "count"
        },
        4165: {
            "Acronym": "AUXIO6",
            "DataRange": "0 to 250",
            "EndBit": 55,
            "Name": "Auxiliary I/O Channel #12",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4165,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "manufacturer defined, not faster than 100 ms",
            "Units": "count"
        },
        4166: {
            "Acronym": "AUXIO6",
            "DataRange": "0 to 250",
            "EndBit": 63,
            "Name": "Auxiliary I/O Channel #11",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4166,
            "SPNLength": 8,
            "StartBit": 56,
            "TransmissionRate": "manufacturer defined, not faster than 100 ms",
            "Units": "count"
        },
        4167: {
            "Acronym": "AUXIO7",
            "DataRange": "0 to 250",
            "EndBit": 7,
            "Name": "Auxiliary I/O Channel #18",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4167,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "manufacturer defined, not faster than 100 ms",
            "Units": "count"
        },
        4168: {
            "Acronym": "AUXIO7",
            "DataRange": "0 to 250",
            "EndBit": 15,
            "Name": "Auxiliary I/O Channel #17",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4168,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "manufacturer defined, not faster than 100 ms",
            "Units": "count"
        },
        4169: {
            "Acronym": "AUXIO7",
            "DataRange": "0 to 250",
            "EndBit": 23,
            "Name": "Auxiliary I/O Channel #16",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4169,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "manufacturer defined, not faster than 100 ms",
            "Units": "count"
        },
        417: {
            "Acronym": "GCVW",
            "DataRange": "0 to 32,899,070 kg",
            "EndBit": 23,
            "Name": "Gross Combination Weight",
            "Offset": 0.0,
            "OperationalHigh": 32899070.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 2,
            "SPN": 417,
            "SPNLength": 24,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "kg"
        },
        4170: {
            "Acronym": "AUXIO7",
            "DataRange": "0 to 250",
            "EndBit": 31,
            "Name": "Auxiliary I/O Channel #15",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4170,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "manufacturer defined, not faster than 100 ms",
            "Units": "count"
        },
        4171: {
            "Acronym": "AUXIO7",
            "DataRange": "0 to 250",
            "EndBit": 39,
            "Name": "Auxiliary I/O Channel #22",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4171,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "manufacturer defined, not faster than 100 ms",
            "Units": "count"
        },
        4172: {
            "Acronym": "AUXIO7",
            "DataRange": "0 to 250",
            "EndBit": 47,
            "Name": "Auxiliary I/O Channel #21",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4172,
            "SPNLength": 8,
            "StartBit": 40,
            "TransmissionRate": "manufacturer defined, not faster than 100 ms",
            "Units": "count"
        },
        4173: {
            "Acronym": "AUXIO7",
            "DataRange": "0 to 250",
            "EndBit": 55,
            "Name": "Auxiliary I/O Channel #20",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4173,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "manufacturer defined, not faster than 100 ms",
            "Units": "count"
        },
        4174: {
            "Acronym": "AUXIO7",
            "DataRange": "0 to 250",
            "EndBit": 63,
            "Name": "Auxiliary I/O Channel #19",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4174,
            "SPNLength": 8,
            "StartBit": 56,
            "TransmissionRate": "manufacturer defined, not faster than 100 ms",
            "Units": "count"
        },
        4175: {
            "Acronym": "DPFC1",
            "DataRange": "0 to 7",
            "EndBit": 55,
            "Name": "DPF Active Regeneration Forced Status",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4175,
            "SPNLength": 3,
            "StartBit": 53,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010:  1 s and on change",
            "Units": "bit"
        },
        4176: {
            "Acronym": "ETC7",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Transmission Current Range Display Blank State",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4176,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        4177: {
            "Acronym": "TRF2",
            "DataRange": "0 to 100 %",
            "EndBit": 31,
            "Name": "Transmission Oil Life Remaining",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 4177,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "1 s",
            "Units": "%"
        },
        4178: {
            "Acronym": "ETC7",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Transmission Service Indicator",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4178,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        4180: {
            "Acronym": "PMI",
            "DataRange": "0 to 2047",
            "EndBit": 8,
            "Name": "Data Dictionary Manufacturer Code",
            "Offset": 0.0,
            "OperationalHigh": 2047.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 4180,
            "SPNLength": 11,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        4181: {
            "Acronym": "PMI",
            "DataRange": "0 to 2097151",
            "EndBit": 31,
            "Name": "Data Dictionary Method",
            "Offset": 0.0,
            "OperationalHigh": 2097151.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 4181,
            "SPNLength": 21,
            "StartBit": 11,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        4191: {
            "Acronym": "TSC1",
            "DataRange": "0 to 0.875%",
            "EndBit": 43,
            "Name": "Engine Requested Torque -  High Resolution",
            "Offset": 0.0,
            "OperationalHigh": 0.875,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.125,
            "SPN": 4191,
            "SPNLength": 4,
            "StartBit": 40,
            "TransmissionRate": "To engine: Control Purpose dependent or 10 ms\nTo retarder: 50 ms",
            "Units": "%"
        },
        4193: {
            "Acronym": "ET4",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 15,
            "Name": "Engine Coolant Pump Outlet Temperature",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 4193,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        4194: {
            "Acronym": "ET4",
            "DataRange": "0 to 100 %",
            "EndBit": 23,
            "Name": "Engine Coolant Thermostat Opening",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 4194,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "1 s",
            "Units": "%"
        },
        4195: {
            "Acronym": "ECCVC",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Engine Coolant Thermostat Mode",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4195,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "1 sec",
            "Units": "bit"
        },
        4196: {
            "Acronym": "ECCVC",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 15,
            "Name": "Desired Engine Coolant Pump Outlet Temperature",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 4196,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "1 sec",
            "Units": "deg F"
        },
        4197: {
            "Acronym": "ECCVC",
            "DataRange": "0 to 100 %",
            "EndBit": 23,
            "Name": "Desired Engine Coolant Thermostat Opening",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 4197,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "1 sec",
            "Units": "%"
        },
        4198: {
            "Acronym": "ACCVC",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Aftercooler Coolant Thermostat Mode",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4198,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "1 sec",
            "Units": "bit"
        },
        4199: {
            "Acronym": "ACCVC",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 15,
            "Name": "Desired Aftercooler Coolant Intake Temperature",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 4199,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "1 sec",
            "Units": "deg F"
        },
        4200: {
            "Acronym": "ACCVC",
            "DataRange": "0 to 100 %",
            "EndBit": 23,
            "Name": "Desired Aftercooler Coolant Thermostat Opening",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 4200,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "1 sec",
            "Units": "%"
        },
        4201: {
            "Acronym": "ESSI",
            "DataRange": "0 to 32,127.5 rpm",
            "EndBit": 15,
            "Name": "Engine Speed 1",
            "Offset": 0.0,
            "OperationalHigh": 32127.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.5,
            "SPN": 4201,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "Transmitted only after requested.  After request, broadcast rate is engine speed dependent.  Update stopped after key switch cycle.",
            "Units": "rpm"
        },
        4202: {
            "Acronym": "ESSI",
            "DataRange": "0 to 32,127.5 rpm",
            "EndBit": 47,
            "Name": "Engine Speed 3",
            "Offset": 0.0,
            "OperationalHigh": 32127.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.5,
            "SPN": 4202,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "Transmitted only after requested.  After request, broadcast rate is engine speed dependent.  Update stopped after key switch cycle.",
            "Units": "rpm"
        },
        4203: {
            "Acronym": "ESSI",
            "DataRange": "0 to 3",
            "EndBit": 55,
            "Name": "Engine Speed Sensor 1 Timing Pattern Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4203,
            "SPNLength": 2,
            "StartBit": 54,
            "TransmissionRate": "Transmitted only after requested.  After request, broadcast rate is engine speed dependent.  Update stopped after key switch cycle.",
            "Units": "bit"
        },
        4204: {
            "Acronym": "ESSI",
            "DataRange": "0 to 3",
            "EndBit": 53,
            "Name": "Engine Speed Sensor 2 Timing Pattern Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4204,
            "SPNLength": 2,
            "StartBit": 52,
            "TransmissionRate": "Transmitted only after requested.  After request, broadcast rate is engine speed dependent.  Update stopped after key switch cycle.",
            "Units": "bit"
        },
        4205: {
            "Acronym": "ESSI",
            "DataRange": "0 to 3",
            "EndBit": 51,
            "Name": "Engine Speed Sensor 3 Timing Pattern Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4205,
            "SPNLength": 2,
            "StartBit": 50,
            "TransmissionRate": "Transmitted only after requested.  After request, broadcast rate is engine speed dependent.  Update stopped after key switch cycle.",
            "Units": "bit"
        },
        4206: {
            "Acronym": "TSC1",
            "DataRange": "0 to 15",
            "EndBit": 59,
            "Name": "Message Counter",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4206,
            "SPNLength": 4,
            "StartBit": 56,
            "TransmissionRate": "To engine: Control Purpose dependent or 10 ms\nTo retarder: 50 ms",
            "Units": "count"
        },
        4207: {
            "Acronym": "TSC1",
            "DataRange": "0 to 15",
            "EndBit": 63,
            "Name": "Message Checksum",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4207,
            "SPNLength": 4,
            "StartBit": 60,
            "TransmissionRate": "To engine: Control Purpose dependent or 10 ms\nTo retarder: 50 ms",
            "Units": "count"
        },
        4211: {
            "Acronym": "FD1",
            "DataRange": "0.0 to 4659.699927595 psi",
            "EndBit": 47,
            "Name": "Hydraulic Fan Motor Pressure",
            "Offset": 0.0,
            "OperationalHigh": 4659.699927595,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.072518869,
            "SPN": 4211,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "1 s",
            "Units": "psi"
        },
        4212: {
            "Acronym": "FD1",
            "DataRange": "0 to 100 %",
            "EndBit": 55,
            "Name": "Fan Drive Bypass Command Status",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 4212,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "1 s",
            "Units": "%"
        },
        4233: {
            "Acronym": "ERC1",
            "DataRange": "0 to 3",
            "EndBit": 29,
            "Name": "Retarder Road Speed Limit Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4233,
            "SPNLength": 2,
            "StartBit": 28,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        4234: {
            "Acronym": "ERC1",
            "DataRange": "0 to 3",
            "EndBit": 31,
            "Name": "Retarder Road Speed Exceeded Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4234,
            "SPNLength": 2,
            "StartBit": 30,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        4236: {
            "Acronym": "O2FT1",
            "DataRange": "-100 to 100 %",
            "EndBit": 31,
            "Name": "Short-term Fuel Trim - Bank 1",
            "Offset": -3212.75,
            "OperationalHigh": 100.0,
            "OperationalLow": -100.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 4236,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On Request",
            "Units": "%"
        },
        4237: {
            "Acronym": "O2FT1",
            "DataRange": "-100 to 100 %",
            "EndBit": 15,
            "Name": "Long-term Fuel Trim - Bank 1",
            "Offset": -3212.75,
            "OperationalHigh": 100.0,
            "OperationalLow": -100.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 4237,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On Request",
            "Units": "%"
        },
        4238: {
            "Acronym": "O2FT2",
            "DataRange": "-100 to 100 %",
            "EndBit": 31,
            "Name": "Short-term Fuel Trim - Bank 2",
            "Offset": -3212.75,
            "OperationalHigh": 100.0,
            "OperationalLow": -100.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 4238,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On Request",
            "Units": "%"
        },
        4239: {
            "Acronym": "O2FT2",
            "DataRange": "-100 to 100 %",
            "EndBit": 15,
            "Name": "Long-term Fuel Trim - Bank 2",
            "Offset": -3212.75,
            "OperationalHigh": 100.0,
            "OperationalLow": -100.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 4239,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On Request",
            "Units": "%"
        },
        4240: {
            "Acronym": "O2FT1",
            "DataRange": "0 to 15",
            "EndBit": 35,
            "Name": "Engine Exhaust Gas Oxygen Sensor Closed Loop Operation, Bank 1",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4240,
            "SPNLength": 4,
            "StartBit": 32,
            "TransmissionRate": "On Request",
            "Units": "bit"
        },
        4241: {
            "Acronym": "O2FT2",
            "DataRange": "0 to 15",
            "EndBit": 35,
            "Name": "Engine Exhaust Gas Oxygen Sensor Closed Loop Operation, Bank 2",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4241,
            "SPNLength": 4,
            "StartBit": 32,
            "TransmissionRate": "On Request",
            "Units": "bit"
        },
        4242: {
            "Acronym": "TC1",
            "DataRange": "0 to 3",
            "EndBit": 7,
            "Name": "Transmission Reverse Gear Shift Inhibit Request",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4242,
            "SPNLength": 2,
            "StartBit": 6,
            "TransmissionRate": "When active; 50 ms to transmission and axles",
            "Units": "bit"
        },
        4245: {
            "Acronym": "GFD",
            "DataRange": "0 to 250.996 MJ/Nm^3",
            "EndBit": 47,
            "Name": "Fuel energy content",
            "Offset": 0.0,
            "OperationalHigh": 250.9963,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4245,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "MJ/Nm^3"
        },
        4246: {
            "Acronym": "TC1",
            "DataRange": "0 to 3",
            "EndBit": 57,
            "Name": "Transmission Mode 5",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4246,
            "SPNLength": 2,
            "StartBit": 56,
            "TransmissionRate": "When active; 50 ms to transmission and axles",
            "Units": "bit"
        },
        4247: {
            "Acronym": "TC1",
            "DataRange": "0 to 3",
            "EndBit": 59,
            "Name": "Transmission Mode 6",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4247,
            "SPNLength": 2,
            "StartBit": 58,
            "TransmissionRate": "When active; 50 ms to transmission and axles",
            "Units": "bit"
        },
        4248: {
            "Acronym": "TC1",
            "DataRange": "0 to 3",
            "EndBit": 61,
            "Name": "Transmission Mode 7",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4248,
            "SPNLength": 2,
            "StartBit": 60,
            "TransmissionRate": "When active; 50 ms to transmission and axles",
            "Units": "bit"
        },
        4249: {
            "Acronym": "TC1",
            "DataRange": "0 to 3",
            "EndBit": 63,
            "Name": "Transmission Mode 8",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4249,
            "SPNLength": 2,
            "StartBit": 62,
            "TransmissionRate": "When active; 50 ms to transmission and axles",
            "Units": "bit"
        },
        4250: {
            "Acronym": "ETC7",
            "DataRange": "0 to 3",
            "EndBit": 33,
            "Name": "Transmission Mode 5 Indicator",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4250,
            "SPNLength": 2,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        4251: {
            "Acronym": "ETC7",
            "DataRange": "0 to 3",
            "EndBit": 35,
            "Name": "Transmission Mode 6 Indicator",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4251,
            "SPNLength": 2,
            "StartBit": 34,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        4252: {
            "Acronym": "ETC7",
            "DataRange": "0 to 3",
            "EndBit": 37,
            "Name": "Transmission Mode 7 Indicator",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4252,
            "SPNLength": 2,
            "StartBit": 36,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        4253: {
            "Acronym": "ETC7",
            "DataRange": "0 to 3",
            "EndBit": 39,
            "Name": "Transmission Mode 8 Indicator",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4253,
            "SPNLength": 2,
            "StartBit": 38,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        4254: {
            "Acronym": "TML",
            "DataRange": "0 to 255 per byte",
            "EndBit": 0,
            "Name": "Transmission Mode Label",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": -2,
            "SPN": 4254,
            "SPNLength": 200,
            "StartBit": -1,
            "TransmissionRate": "On request",
            "Units": "ASCII"
        },
        4255: {
            "Acronym": "TC1",
            "DataRange": "0 to 15",
            "EndBit": 53,
            "Name": "Transmission Requested Launch Gear",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4255,
            "SPNLength": 4,
            "StartBit": 50,
            "TransmissionRate": "When active; 50 ms to transmission and axles",
            "Units": "bit"
        },
        4261: {
            "Acronym": "ETC7",
            "DataRange": "0 to 3",
            "EndBit": 41,
            "Name": "Transmission Reverse Gear Shift Inhibit Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4261,
            "SPNLength": 2,
            "StartBit": 40,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        4263: {
            "Acronym": "DM04",
            "DataRange": "manufacturer defined",
            "EndBit": -1,
            "Name": "Manufacturer Specified Information",
            "Offset": 0,
            "OperationalHigh": -1,
            "OperationalLow": -1,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": -5,
            "SPN": 4263,
            "SPNLength": 8,
            "StartBit": 13,
            "TransmissionRate": "On request",
            "Units": ""
        },
        4287: {
            "Acronym": "EFL/P3",
            "DataRange": "0.0 to 36404.472237999995 psi",
            "EndBit": 39,
            "Name": "Engine Exhaust Valve Actuation System Oil Pressure",
            "Offset": 0.0,
            "OperationalHigh": 36404.472237999995,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.5665536640625,
            "SPN": 4287,
            "SPNLength": 16,
            "StartBit": 24,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        4288: {
            "Acronym": "ET4",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 39,
            "Name": "Engine Exhaust Valve Actuation System Oil Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 4288,
            "SPNLength": 16,
            "StartBit": 24,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        4289: {
            "Acronym": "AT1TWCC",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 15,
            "Name": "Aftertreatment 1 Three Way Catalytic Converter Intake Gas Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 4289,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "500 ms",
            "Units": "deg F"
        },
        4290: {
            "Acronym": "AT1TWCC",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 31,
            "Name": "Aftertreatment 1 Three Way Catalytic Converter Outlet Gas Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 4290,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "500 ms",
            "Units": "deg F"
        },
        4291: {
            "Acronym": "AT1TWCC",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 47,
            "Name": "Aftertreatment 1 Three Way Catalytic Converter Differential Pressure",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 4291,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        4292: {
            "Acronym": "AT1TWCC",
            "DataRange": "0 to 31",
            "EndBit": 52,
            "Name": "Aftertreatment 1 Three Way Catalytic Converter Intake Gas Temperature Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 4292,
            "SPNLength": 5,
            "StartBit": 48,
            "TransmissionRate": "500 ms",
            "Units": "binary"
        },
        4293: {
            "Acronym": "AT1TWCC",
            "DataRange": "0 to 31",
            "EndBit": 56,
            "Name": "Aftertreatment 1 Three Way Catalytic Converter Outlet Gas Temperature Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 4293,
            "SPNLength": 5,
            "StartBit": 53,
            "TransmissionRate": "500 ms",
            "Units": "binary"
        },
        4294: {
            "Acronym": "AT1TWCC",
            "DataRange": "0 to 31",
            "EndBit": 62,
            "Name": "Aftertreatment 1 Three Way Catalytic Converter Differential Pressure Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 4294,
            "SPNLength": 5,
            "StartBit": 58,
            "TransmissionRate": "500 ms",
            "Units": "binary"
        },
        4295: {
            "Acronym": "AT2TWCC",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 15,
            "Name": "Aftertreatment 2 Three Way Catalytic Converter Intake Gas Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 4295,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "500 ms",
            "Units": "deg F"
        },
        4296: {
            "Acronym": "AT2TWCC",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 31,
            "Name": "Aftertreatment 2 Three Way Catalytic Converter Outlet Gas Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 4296,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "500 ms",
            "Units": "deg F"
        },
        4297: {
            "Acronym": "AT2TWCC",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 47,
            "Name": "Aftertreatment 2 Three Way Catalytic Converter Differential Pressure",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 4297,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        4298: {
            "Acronym": "AT2TWCC",
            "DataRange": "0 to 31",
            "EndBit": 52,
            "Name": "Aftertreatment 2 Three Way Catalytic Converter Intake Gas Temperature Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 4298,
            "SPNLength": 5,
            "StartBit": 48,
            "TransmissionRate": "500 ms",
            "Units": "binary"
        },
        4299: {
            "Acronym": "AT2TWCC",
            "DataRange": "0 to 31",
            "EndBit": 56,
            "Name": "Aftertreatment 2 Three Way Catalytic Converter Outlet Gas Temperature Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 4299,
            "SPNLength": 5,
            "StartBit": 53,
            "TransmissionRate": "500 ms",
            "Units": "binary"
        },
        4300: {
            "Acronym": "AT2TWCC",
            "DataRange": "0 to 31",
            "EndBit": 62,
            "Name": "Aftertreatment 2 Three Way Catalytic Converter Differential Pressure Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 4300,
            "SPNLength": 5,
            "StartBit": 58,
            "TransmissionRate": "500 ms",
            "Units": "binary"
        },
        4301: {
            "Acronym": "AT1FC1",
            "DataRange": "0 to 100 %",
            "EndBit": 63,
            "Name": "Aftertreatment 1 Fuel Injector 1 Heater Control",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 4301,
            "SPNLength": 8,
            "StartBit": 56,
            "TransmissionRate": "500 ms",
            "Units": "%"
        },
        4302: {
            "Acronym": "AT2FC1",
            "DataRange": "0 to 100 %",
            "EndBit": 63,
            "Name": "Aftertreatment 2 Fuel Injector 1 Heater Control",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 4302,
            "SPNLength": 8,
            "StartBit": 56,
            "TransmissionRate": "500 ms",
            "Units": "%"
        },
        4303: {
            "Acronym": "AT2FC2",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 15,
            "Name": "Aftertreatment 2 Fuel Pressure 2",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 4303,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        4304: {
            "Acronym": "ECUID",
            "DataRange": "0 to 255 per byte",
            "EndBit": 0,
            "Name": "ECU Manufacturer Name",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": -2,
            "SPN": 4304,
            "SPNLength": 1600,
            "StartBit": -1,
            "TransmissionRate": "On request",
            "Units": "ASCII"
        },
        4305: {
            "Acronym": "MSS",
            "DataRange": "0 to 64.255 m/s",
            "EndBit": 15,
            "Name": "Machine selected speed",
            "Offset": 0.0,
            "OperationalHigh": 64.255,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.001,
            "SPN": 4305,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "m/s"
        },
        4306: {
            "Acronym": "MSS",
            "DataRange": "0 to 4,211,081.215 m",
            "EndBit": 47,
            "Name": "Machine selected distance",
            "Offset": 0.0,
            "OperationalHigh": 4211081.215,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.001,
            "SPN": 4306,
            "SPNLength": 32,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "m"
        },
        4307: {
            "Acronym": "MSS",
            "DataRange": "0 to 7",
            "EndBit": 63,
            "Name": "Persistent limit status",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4307,
            "SPNLength": 3,
            "StartBit": 61,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        4308: {
            "Acronym": "MSS",
            "DataRange": "0 to 7",
            "EndBit": 60,
            "Name": "Machine selected speed source",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4308,
            "SPNLength": 3,
            "StartBit": 58,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        4309: {
            "Acronym": "MSS",
            "DataRange": "0 to 3",
            "EndBit": 57,
            "Name": "Machine selected direction",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4309,
            "SPNLength": 2,
            "StartBit": 56,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        4310: {
            "Acronym": "MSSC",
            "DataRange": "0 to 64.255 m/s",
            "EndBit": 15,
            "Name": "Machine selected speed setpoint command",
            "Offset": 0.0,
            "OperationalHigh": 64.255,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.001,
            "SPN": 4310,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "m/s"
        },
        4311: {
            "Acronym": "MSSC",
            "DataRange": "0 to 64.255 m/s",
            "EndBit": 31,
            "Name": "Machine selected speed setpoint limit",
            "Offset": 0.0,
            "OperationalHigh": 64.255,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.001,
            "SPN": 4311,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "m/s"
        },
        4312: {
            "Acronym": "MSSC",
            "DataRange": "0 to 3",
            "EndBit": 57,
            "Name": "Machine selected direction command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4312,
            "SPNLength": 2,
            "StartBit": 56,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        4313: {
            "Acronym": "ICC",
            "DataRange": "2000 to 2061 years",
            "EndBit": 5,
            "Name": "ISOBUS compliance certification year",
            "Offset": 2000.0,
            "OperationalHigh": 2061.0,
            "OperationalLow": 2000.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4313,
            "SPNLength": 6,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "Years"
        },
        4314: {
            "Acronym": "ICC",
            "DataRange": "0 to 7",
            "EndBit": 8,
            "Name": "ISOBUS compliance certificate revision",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4314,
            "SPNLength": 3,
            "StartBit": 6,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        4315: {
            "Acronym": "ICC",
            "DataRange": "0 to 7",
            "EndBit": 11,
            "Name": "ISOBUS compliance certificate laboratory type",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4315,
            "SPNLength": 3,
            "StartBit": 9,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        4316: {
            "Acronym": "ICC",
            "DataRange": "0 to 2047",
            "EndBit": 23,
            "Name": "ISOBUS Compliance certification laboratory ID",
            "Offset": 0.0,
            "OperationalHigh": 2047.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4316,
            "SPNLength": 11,
            "StartBit": 13,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        4317: {
            "Acronym": "ICC",
            "DataRange": "0 to 1",
            "EndBit": 31,
            "Name": "ISOBUS Compliance certification type - Minimum ECU",
            "Offset": 0.0,
            "OperationalHigh": 1.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4317,
            "SPNLength": 1,
            "StartBit": 31,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        4318: {
            "Acronym": "ICC",
            "DataRange": "0 to 1",
            "EndBit": 30,
            "Name": "ISOBUS Compliance certification type - TECU Class 1",
            "Offset": 0.0,
            "OperationalHigh": 1.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4318,
            "SPNLength": 1,
            "StartBit": 30,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        4319: {
            "Acronym": "ICC",
            "DataRange": "0 to 1",
            "EndBit": 29,
            "Name": "ISOBUS Compliance certification type - TECU Class 2",
            "Offset": 0.0,
            "OperationalHigh": 1.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4319,
            "SPNLength": 1,
            "StartBit": 29,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        4320: {
            "Acronym": "ICC",
            "DataRange": "0 to 1",
            "EndBit": 28,
            "Name": "ISOBUS Compliance certification type - TECU Class 3",
            "Offset": 0.0,
            "OperationalHigh": 1.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4320,
            "SPNLength": 1,
            "StartBit": 28,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        4321: {
            "Acronym": "ICC",
            "DataRange": "0 to 1",
            "EndBit": 27,
            "Name": "ISOBUS Compliance certification type - Class 3 ECU",
            "Offset": 0.0,
            "OperationalHigh": 1.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4321,
            "SPNLength": 1,
            "StartBit": 27,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        4322: {
            "Acronym": "ICC",
            "DataRange": "0 to 1",
            "EndBit": 26,
            "Name": "ISOBUS Compliance certification type - Virtual terminal",
            "Offset": 0.0,
            "OperationalHigh": 1.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4322,
            "SPNLength": 1,
            "StartBit": 26,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        4323: {
            "Acronym": "ICC",
            "DataRange": "0 to 1",
            "EndBit": 25,
            "Name": "ISOBUS Compliance certification type - VT working set master",
            "Offset": 0.0,
            "OperationalHigh": 1.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4323,
            "SPNLength": 1,
            "StartBit": 25,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        4324: {
            "Acronym": "ICC",
            "DataRange": "0 to 1",
            "EndBit": 24,
            "Name": "ISOBUS Compliance certification type - VT working set member",
            "Offset": 0.0,
            "OperationalHigh": 1.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4324,
            "SPNLength": 1,
            "StartBit": 24,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        4325: {
            "Acronym": "ICC",
            "DataRange": "0 to 1",
            "EndBit": 39,
            "Name": "ISOBUS Compliance certification type - Task Controller",
            "Offset": 0.0,
            "OperationalHigh": 1.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4325,
            "SPNLength": 1,
            "StartBit": 39,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        4326: {
            "Acronym": "ICC",
            "DataRange": "0 to 1",
            "EndBit": 38,
            "Name": "ISOBUS Compliance certification type - TC working set master",
            "Offset": 0.0,
            "OperationalHigh": 1.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4326,
            "SPNLength": 1,
            "StartBit": 38,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        4327: {
            "Acronym": "ICC",
            "DataRange": "0 to 1",
            "EndBit": 37,
            "Name": "ISOBUS Compliance certification type - TC working set member",
            "Offset": 0.0,
            "OperationalHigh": 1.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4327,
            "SPNLength": 1,
            "StartBit": 37,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        4328: {
            "Acronym": "ICC",
            "DataRange": "0 to 1",
            "EndBit": 36,
            "Name": "ISOBUS Compliance certification type - File Server",
            "Offset": 0.0,
            "OperationalHigh": 1.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4328,
            "SPNLength": 1,
            "StartBit": 36,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        4329: {
            "Acronym": "ICC",
            "DataRange": "0 to 1",
            "EndBit": 35,
            "Name": "ISOBUS Compliance certification type - GPS Reciever",
            "Offset": 0.0,
            "OperationalHigh": 1.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4329,
            "SPNLength": 1,
            "StartBit": 35,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        4330: {
            "Acronym": "ICC",
            "DataRange": "0 to 65,535",
            "EndBit": 63,
            "Name": "ISOBUS Compliance certification reference number",
            "Offset": 0.0,
            "OperationalHigh": 65535.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4330,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "On request",
            "Units": "count"
        },
        4331: {
            "Acronym": "A1SCRDSI1",
            "DataRange": "0 to 19276.5 g/h",
            "EndBit": 15,
            "Name": "Aftertreatment 1 Diesel Exhaust Fluid Actual Dosing Quantity",
            "Offset": 0.0,
            "OperationalHigh": 19276.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.3,
            "SPN": 4331,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "50 ms",
            "Units": "g/h"
        },
        4332: {
            "Acronym": "A1SCRDSI1",
            "DataRange": "0 to 15",
            "EndBit": 19,
            "Name": "Aftertreatment 1 SCR System State",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4332,
            "SPNLength": 4,
            "StartBit": 16,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        4333: {
            "Acronym": "A1SCRDSI1",
            "DataRange": "0 to 6425.5 g",
            "EndBit": 39,
            "Name": "Aftertreatment 1 Diesel Exhaust Fluid Actual Quantity of Integrator",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 4333,
            "SPNLength": 16,
            "StartBit": 24,
            "TransmissionRate": "50 ms",
            "Units": "g"
        },
        4334: {
            "Acronym": "A1SCRDSI1",
            "DataRange": "0.0 to 290.075476 psi",
            "EndBit": 47,
            "Name": "Aftertreatment 1 Diesel Exhaust Fluid Doser Absolute Pressure",
            "Offset": 0.0,
            "OperationalHigh": 290.075476,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.160301904,
            "SPN": 4334,
            "SPNLength": 8,
            "StartBit": 40,
            "TransmissionRate": "50 ms",
            "Units": "psi"
        },
        4335: {
            "Acronym": "A1SCRDSI2",
            "DataRange": "0.0 to 290.075476 psi",
            "EndBit": 7,
            "Name": "Aftertreatment 1 SCR Dosing Air Assist Absolute Pressure",
            "Offset": 0.0,
            "OperationalHigh": 290.075476,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.160301904,
            "SPN": 4335,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        4336: {
            "Acronym": "A1SCRDSI2",
            "DataRange": "0 to 100 %",
            "EndBit": 15,
            "Name": "Aftertreatment 1 SCR Dosing Air Assist Valve",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 4336,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "500 ms",
            "Units": "%"
        },
        4337: {
            "Acronym": "A1SCRDSI2",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 23,
            "Name": "Aftertreatment 1 Diesel Exhaust Fluid Dosing Temperature",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 4337,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "500 ms",
            "Units": "deg F"
        },
        4338: {
            "Acronym": "A1SCRDSI2",
            "DataRange": "0 to 7",
            "EndBit": 26,
            "Name": "Aftertreatment 1 SCR Dosing Valve Exhaust Temperature Reduction Request",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4338,
            "SPNLength": 3,
            "StartBit": 24,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        4339: {
            "Acronym": "A1SCRDSI2",
            "DataRange": "0 to 7",
            "EndBit": 29,
            "Name": "Aftertreatment 1 SCR Feedback Control Status",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4339,
            "SPNLength": 3,
            "StartBit": 27,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        4340: {
            "Acronym": "A1SCRDSI2",
            "DataRange": "0 to 3",
            "EndBit": 33,
            "Name": "Aftertreatment 1 Diesel Exhaust Fluid Line Heater 1 State",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4340,
            "SPNLength": 2,
            "StartBit": 32,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        4341: {
            "Acronym": "A1SCRDSI2",
            "DataRange": "0 to 31",
            "EndBit": 38,
            "Name": "Aftertreatment 1 Diesel Exhaust Fluid Line Heater 1 Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 4341,
            "SPNLength": 5,
            "StartBit": 34,
            "TransmissionRate": "500 ms",
            "Units": "binary"
        },
        4342: {
            "Acronym": "A1SCRDSI2",
            "DataRange": "0 to 3",
            "EndBit": 41,
            "Name": "Aftertreatment 1 Diesel Exhaust Fluid Line Heater 2 State",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4342,
            "SPNLength": 2,
            "StartBit": 40,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        4343: {
            "Acronym": "A1SCRDSI2",
            "DataRange": "0 to 31",
            "EndBit": 46,
            "Name": "Aftertreatment 1 Diesel Exhaust Fluid Line Heater 2 Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 4343,
            "SPNLength": 5,
            "StartBit": 42,
            "TransmissionRate": "500 ms",
            "Units": "binary"
        },
        4344: {
            "Acronym": "A1SCRDSI2",
            "DataRange": "0 to 3",
            "EndBit": 49,
            "Name": "Aftertreatment 1 Diesel Exhaust Fluid Line Heater 3 State",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4344,
            "SPNLength": 2,
            "StartBit": 48,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        4345: {
            "Acronym": "A1SCRDSI2",
            "DataRange": "0 to 31",
            "EndBit": 54,
            "Name": "Aftertreatment 1 Diesel Exhaust Fluid Line Heater 3 Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 4345,
            "SPNLength": 5,
            "StartBit": 50,
            "TransmissionRate": "500 ms",
            "Units": "binary"
        },
        4346: {
            "Acronym": "A1SCRDSI2",
            "DataRange": "0 to 3",
            "EndBit": 57,
            "Name": "Aftertreatment 1 Diesel Exhaust Fluid Line Heater 4 State",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4346,
            "SPNLength": 2,
            "StartBit": 56,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        4347: {
            "Acronym": "A1SCRDSI2",
            "DataRange": "0 to 31",
            "EndBit": 62,
            "Name": "Aftertreatment 1 Diesel Exhaust Fluid Line Heater 4 Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 4347,
            "SPNLength": 5,
            "StartBit": 58,
            "TransmissionRate": "500 ms",
            "Units": "binary"
        },
        4348: {
            "Acronym": "A1SCRDSR1",
            "DataRange": "0 to 19276.5 g/h",
            "EndBit": 15,
            "Name": "Aftertreatment 1 Diesel Exhaust Fluid Dosing Requested Quantity",
            "Offset": 0.0,
            "OperationalHigh": 19276.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.3,
            "SPN": 4348,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "50 ms",
            "Units": "g/h"
        },
        4349: {
            "Acronym": "A1SCRDSR1",
            "DataRange": "0 to 15",
            "EndBit": 19,
            "Name": "Aftertreatment 1 SCR System Requested State",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4349,
            "SPNLength": 4,
            "StartBit": 16,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        4350: {
            "Acronym": "A1SCRDSR1",
            "DataRange": "0 to 6425.5 g",
            "EndBit": 39,
            "Name": "Aftertreatment 1 Diesel Exhaust Fluid Requested Quantity of Integrator",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 4350,
            "SPNLength": 16,
            "StartBit": 24,
            "TransmissionRate": "50 ms",
            "Units": "g"
        },
        4352: {
            "Acronym": "A1SCRDSR2",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Aftertreatment 1 Diesel Exhaust Fluid Doser Fault Suppression Request",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4352,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        4353: {
            "Acronym": "A1SCRDSR2",
            "DataRange": "0 to 7",
            "EndBit": 6,
            "Name": "Aftertreatment 1 Diesel Exhaust Fluid Doser Heating Mode Request",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4353,
            "SPNLength": 3,
            "StartBit": 4,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        4354: {
            "Acronym": "A1SCRDSR2",
            "DataRange": "0 to 3",
            "EndBit": 9,
            "Name": "Aftertreatment 1 Diesel Exhaust Fluid Line Heater 1",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4354,
            "SPNLength": 2,
            "StartBit": 8,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        4355: {
            "Acronym": "A1SCRDSR2",
            "DataRange": "0 to 3",
            "EndBit": 11,
            "Name": "Aftertreatment 1 Diesel Exhaust Fluid Line Heater 2",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4355,
            "SPNLength": 2,
            "StartBit": 10,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        4356: {
            "Acronym": "A1SCRDSR2",
            "DataRange": "0 to 3",
            "EndBit": 13,
            "Name": "Aftertreatment 1 Diesel Exhaust Fluid Line Heater 3",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4356,
            "SPNLength": 2,
            "StartBit": 12,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        4357: {
            "Acronym": "A1SCRDSR2",
            "DataRange": "0 to 3",
            "EndBit": 15,
            "Name": "Aftertreatment 1 Diesel Exhaust Fluid Line Heater 4",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4357,
            "SPNLength": 2,
            "StartBit": 14,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        4358: {
            "Acronym": "A1SCREGP",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 15,
            "Name": "Aftertreatment 1 SCR Exhaust Gas Differential Pressure",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 4358,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        4359: {
            "Acronym": "A1SCREGP",
            "DataRange": "0 to 31",
            "EndBit": 20,
            "Name": "Aftertreatment 1 SCR Exhaust Gas Differential Pressure Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4359,
            "SPNLength": 5,
            "StartBit": 16,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        4360: {
            "Acronym": "A1SCREGT",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 15,
            "Name": "Aftertreatment 1 SCR Catalyst Intake Gas Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 4360,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "500 ms",
            "Units": "deg F"
        },
        4361: {
            "Acronym": "A1SCREGT",
            "DataRange": "0 to 31",
            "EndBit": 20,
            "Name": "Aftertreatment 1 SCR Catalyst Intake Gas Temperature Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4361,
            "SPNLength": 5,
            "StartBit": 16,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        4362: {
            "Acronym": "A1SCREGT",
            "DataRange": "0 to 31",
            "EndBit": 44,
            "Name": "Aftertreatment 1 SCR Catalyst Outlet Gas Temperature Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4362,
            "SPNLength": 5,
            "StartBit": 40,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        4363: {
            "Acronym": "A1SCREGT",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 39,
            "Name": "Aftertreatment 1 SCR Catalyst Outlet Gas Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 4363,
            "SPNLength": 16,
            "StartBit": 24,
            "TransmissionRate": "500 ms",
            "Units": "deg F"
        },
        4364: {
            "Acronym": "SCR1",
            "DataRange": "0 to 100 %",
            "EndBit": 39,
            "Name": "Aftertreatment 1 SCR Conversion Efficiency",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 4364,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        4365: {
            "Acronym": "AT1T1I",
            "DataRange": "0 to 31",
            "EndBit": 44,
            "Name": "Aftertreatment 1 SCR Catalyst Reagent Tank 1 Temperature Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 4365,
            "SPNLength": 5,
            "StartBit": 40,
            "TransmissionRate": "1 s",
            "Units": "binary"
        },
        4366: {
            "Acronym": "AT1T1I",
            "DataRange": "0 to 31",
            "EndBit": 60,
            "Name": "Aftertreatment 1 SCR Catalyst Reagent Tank 1 Heater Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 4366,
            "SPNLength": 5,
            "StartBit": 56,
            "TransmissionRate": "1 s",
            "Units": "binary"
        },
        4367: {
            "Acronym": "A1SCRRT2I",
            "DataRange": "0 to 100 %",
            "EndBit": 7,
            "Name": "Aftertreatment 1 Diesel Exhaust Fluid Tank 2 Level",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 4367,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "%"
        },
        4368: {
            "Acronym": "A1SCRRT2I",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 15,
            "Name": "Aftertreatment 1 Diesel Exhaust Fluid Tank 2 Temperature",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 4368,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        4369: {
            "Acronym": "A1SCRRT2I",
            "DataRange": "0 to 6,425.5 mm (0 to 6.4255 m)",
            "EndBit": 31,
            "Name": "Aftertreatment 1 Diesel Exhaust Fluid Tank 2 Level 2",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 4369,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "1 s",
            "Units": "m"
        },
        4370: {
            "Acronym": "A1SCRRT2I",
            "DataRange": "0 to 31",
            "EndBit": 36,
            "Name": "Aftertreatment 1 Diesel Exhaust Fluid Tank 2 Level Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 4370,
            "SPNLength": 5,
            "StartBit": 32,
            "TransmissionRate": "1 s",
            "Units": "binary"
        },
        4371: {
            "Acronym": "A1SCRRT2I",
            "DataRange": "0 to 31",
            "EndBit": 44,
            "Name": "Aftertreatment 1 Diesel Exhaust Fluid Tank 2 Temperature Prelminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 4371,
            "SPNLength": 5,
            "StartBit": 40,
            "TransmissionRate": "1 s",
            "Units": "binary"
        },
        4372: {
            "Acronym": "A1SCRRT2I",
            "DataRange": "0 to 100 %",
            "EndBit": 55,
            "Name": "Aftertreatment 1 Diesel Exhaust Fluid Tank 2 Heater",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 4372,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "1 s",
            "Units": "%"
        },
        4373: {
            "Acronym": "A1SCRRT2I",
            "DataRange": "0 to 31",
            "EndBit": 60,
            "Name": "Aftertreatment 1 Diesel Exhaust Fluid Tank 2 Heater Prelminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 4373,
            "SPNLength": 5,
            "StartBit": 56,
            "TransmissionRate": "1 s",
            "Units": "binary"
        },
        4374: {
            "Acronym": "A1DEFSI",
            "DataRange": "0 to 32,127.5 rpm",
            "EndBit": 15,
            "Name": "Aftertreatment 1 Diesel Exhaust Fluid Pump Motor Speed",
            "Offset": 0.0,
            "OperationalHigh": 32127.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.5,
            "SPN": 4374,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "rpm"
        },
        4375: {
            "Acronym": "A1DEFSI",
            "DataRange": "0 to 100 %",
            "EndBit": 23,
            "Name": "Aftertreatment 1 Diesel Exhaust Fluid Pump Drive Percentage",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 4375,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "1 s",
            "Units": "%"
        },
        4376: {
            "Acronym": "A1DEFSI",
            "DataRange": "0 to 100 %",
            "EndBit": 31,
            "Name": "Aftertreatment 1 Diesel Exhaust Fluid Return Valve",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 4376,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "1 s",
            "Units": "%"
        },
        4377: {
            "Acronym": "A1SCRAI",
            "DataRange": "-200 to 3012.75 ppm",
            "EndBit": 15,
            "Name": "Aftertreatment 1 Outlet NH3",
            "Offset": -200.0,
            "OperationalHigh": 3012.75,
            "OperationalLow": -200.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05,
            "SPN": 4377,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "50 ms",
            "Units": "ppm"
        },
        4378: {
            "Acronym": "A1SCRAI",
            "DataRange": "0 to 31",
            "EndBit": 20,
            "Name": "Aftertreatment 1 Outlet NH3 Sensor Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 4378,
            "SPNLength": 5,
            "StartBit": 16,
            "TransmissionRate": "50 ms",
            "Units": "binary"
        },
        4379: {
            "Acronym": "A1SCRAI",
            "DataRange": "0 to 3",
            "EndBit": 25,
            "Name": "Aftertreatment 1 Outlet NH3 Reading Stable",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4379,
            "SPNLength": 2,
            "StartBit": 24,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        4380: {
            "Acronym": "A1SCRAI",
            "DataRange": "0 to 3",
            "EndBit": 27,
            "Name": "Aftertreatment 1 Outlet NH3 Gas Sensor Power In Range",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4380,
            "SPNLength": 2,
            "StartBit": 26,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        4381: {
            "Acronym": "A1SCRAI",
            "DataRange": "0 to 3",
            "EndBit": 29,
            "Name": "Aftertreatment 1 Outlet NH3 Gas Sensor at Temperature",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4381,
            "SPNLength": 2,
            "StartBit": 28,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        4382: {
            "Acronym": "A1SCRAI",
            "DataRange": "0 to 31",
            "EndBit": 36,
            "Name": "Aftertreatment 1 Outlet NH3 Gas Sensor Heater Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 4382,
            "SPNLength": 5,
            "StartBit": 32,
            "TransmissionRate": "50 ms",
            "Units": "binary"
        },
        4383: {
            "Acronym": "A1SCRAI",
            "DataRange": "0 to 3",
            "EndBit": 41,
            "Name": "Aftertreatment 1 Outlet NH3 Gas Sensor Heater Control",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4383,
            "SPNLength": 2,
            "StartBit": 40,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        4384: {
            "Acronym": "A2SCRDSI1",
            "DataRange": "0 to 19276.5 g/h",
            "EndBit": 15,
            "Name": "Aftertreatment 2 Diesel Exhaust Fluid Actual Dosing Quantity",
            "Offset": 0.0,
            "OperationalHigh": 19276.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.3,
            "SPN": 4384,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "50 ms",
            "Units": "g/h"
        },
        4385: {
            "Acronym": "A2SCRDSI1",
            "DataRange": "0 to 15",
            "EndBit": 19,
            "Name": "Aftertreatment 2 SCR System State",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4385,
            "SPNLength": 4,
            "StartBit": 16,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        4386: {
            "Acronym": "A2SCRDSI1",
            "DataRange": "0 to 6425.5 g",
            "EndBit": 39,
            "Name": "Aftertreatment 2 Diesel Exhaust Fluid Actual Quantity of Integrator",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 4386,
            "SPNLength": 16,
            "StartBit": 24,
            "TransmissionRate": "50 ms",
            "Units": "g"
        },
        4387: {
            "Acronym": "A2SCRDSI1",
            "DataRange": "0.0 to 290.075476 psi",
            "EndBit": 47,
            "Name": "Aftertreatment 2 Diesel Exhaust Fluid Dosing Absolute Pressure",
            "Offset": 0.0,
            "OperationalHigh": 290.075476,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.160301904,
            "SPN": 4387,
            "SPNLength": 8,
            "StartBit": 40,
            "TransmissionRate": "50 ms",
            "Units": "psi"
        },
        4388: {
            "Acronym": "A2SCRDSI2",
            "DataRange": "0.0 to 290.075476 psi",
            "EndBit": 7,
            "Name": "Aftertreatment 2 SCR Dosing Air Assist Absolute Pressure",
            "Offset": 0.0,
            "OperationalHigh": 290.075476,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.160301904,
            "SPN": 4388,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        4389: {
            "Acronym": "A2SCRDSI2",
            "DataRange": "0 to 100 %",
            "EndBit": 15,
            "Name": "Aftertreatment 2 SCR Dosing Air Assist Valve",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 4389,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "500 ms",
            "Units": "%"
        },
        4390: {
            "Acronym": "A2SCRDSI2",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 23,
            "Name": "Aftertreatment 2 Diesel Exhaust Fluid Dosing Temperature",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 4390,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "500 ms",
            "Units": "deg F"
        },
        4391: {
            "Acronym": "A2SCRDSI2",
            "DataRange": "0 to 7",
            "EndBit": 26,
            "Name": "Aftertreatment 2 SCR Dosing Valve Exhaust Temp. Reduction Request",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4391,
            "SPNLength": 3,
            "StartBit": 24,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        4392: {
            "Acronym": "A2SCRDSI2",
            "DataRange": "0 to 7",
            "EndBit": 29,
            "Name": "Aftertreatment 2 SCR Feedback Control Status",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4392,
            "SPNLength": 3,
            "StartBit": 27,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        4393: {
            "Acronym": "A2SCRDSI2",
            "DataRange": "0 to 3",
            "EndBit": 33,
            "Name": "Aftertreatment 2 Diesel Exhaust Fluid Line Heater 1 State",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4393,
            "SPNLength": 2,
            "StartBit": 32,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        4394: {
            "Acronym": "A2SCRDSI2",
            "DataRange": "0 to 31",
            "EndBit": 38,
            "Name": "Aftertreatment 2 Diesel Exhaust Fluid Line Heater 1 Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 4394,
            "SPNLength": 5,
            "StartBit": 34,
            "TransmissionRate": "500 ms",
            "Units": "binary"
        },
        4395: {
            "Acronym": "A2SCRDSI2",
            "DataRange": "0 to 3",
            "EndBit": 41,
            "Name": "Aftertreatment 2 Diesel Exhaust Fluid Line Heater 2 State",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4395,
            "SPNLength": 2,
            "StartBit": 40,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        4396: {
            "Acronym": "A2SCRDSI2",
            "DataRange": "0 to 31",
            "EndBit": 46,
            "Name": "Aftertreatment 2 Diesel Exhaust Fluid Line Heater 2 Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 4396,
            "SPNLength": 5,
            "StartBit": 42,
            "TransmissionRate": "500 ms",
            "Units": "binary"
        },
        4397: {
            "Acronym": "A2SCRDSI2",
            "DataRange": "0 to 3",
            "EndBit": 49,
            "Name": "Aftertreatment 2 Diesel Exhaust Fluid Line Heater 3 State",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4397,
            "SPNLength": 2,
            "StartBit": 48,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        4398: {
            "Acronym": "A2SCRDSI2",
            "DataRange": "0 to 31",
            "EndBit": 54,
            "Name": "Aftertreatment 2 Diesel Exhaust Fluid Line Heater 3 Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 4398,
            "SPNLength": 5,
            "StartBit": 50,
            "TransmissionRate": "500 ms",
            "Units": "binary"
        },
        4399: {
            "Acronym": "A2SCRDSI2",
            "DataRange": "0 to 3",
            "EndBit": 57,
            "Name": "Aftertreatment 2 Diesel Exhaust Fluid Line Heater 4 State",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4399,
            "SPNLength": 2,
            "StartBit": 56,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        4400: {
            "Acronym": "A2SCRDSI2",
            "DataRange": "0 to 31",
            "EndBit": 62,
            "Name": "Aftertreatment 2 Diesel Exhaust Fluid Line Heater 4 Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 4400,
            "SPNLength": 5,
            "StartBit": 58,
            "TransmissionRate": "500 ms",
            "Units": "binary"
        },
        4401: {
            "Acronym": "A2SCRDSR1",
            "DataRange": "0 to 19276.5 g/h",
            "EndBit": 15,
            "Name": "Aftertreatment 2 Diesel Exhaust Fluid Dosing Requested Quantity",
            "Offset": 0.0,
            "OperationalHigh": 19276.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.3,
            "SPN": 4401,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "50 ms",
            "Units": "g/h"
        },
        4402: {
            "Acronym": "A2SCRDSR1",
            "DataRange": "0 to 15",
            "EndBit": 19,
            "Name": "Aftertreatment 2 SCR System Requested State",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4402,
            "SPNLength": 4,
            "StartBit": 16,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        4403: {
            "Acronym": "A2SCRDSR1",
            "DataRange": "0 to 6425.5 g",
            "EndBit": 39,
            "Name": "Aftertreatment 2 Diesel Exhaust Fluid Requested Quantity of Integrator",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 4403,
            "SPNLength": 16,
            "StartBit": 24,
            "TransmissionRate": "50 ms",
            "Units": "g"
        },
        4405: {
            "Acronym": "A2SCRDSR2",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Aftertreatment 2 Diesel Exhaust Fluid  Doser Fault Suppression Request",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4405,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        4406: {
            "Acronym": "A2SCRDSR2",
            "DataRange": "0 to 7",
            "EndBit": 6,
            "Name": "Aftertreatment 2 Diesel Exhaust Fluid Doser Heating Mode Request",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4406,
            "SPNLength": 3,
            "StartBit": 4,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        4407: {
            "Acronym": "A2SCRDSR2",
            "DataRange": "0 to 3",
            "EndBit": 9,
            "Name": "Aftertreatment 2 Diesel Exhaust Fluid Line Heater 1",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4407,
            "SPNLength": 2,
            "StartBit": 8,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        4408: {
            "Acronym": "A2SCRDSR2",
            "DataRange": "0 to 3",
            "EndBit": 11,
            "Name": "Aftertreatment 2 Diesel Exhaust Fluid Line Heater 2",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4408,
            "SPNLength": 2,
            "StartBit": 10,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        4409: {
            "Acronym": "A2SCRDSR2",
            "DataRange": "0 to 3",
            "EndBit": 13,
            "Name": "Aftertreatment 2 Diesel Exhaust Fluid Line Heater 3",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4409,
            "SPNLength": 2,
            "StartBit": 12,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        441: {
            "Acronym": "AAI",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 7,
            "Name": "Auxiliary Temperature 1",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 441,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "deg F"
        },
        4410: {
            "Acronym": "A2SCRDSR2",
            "DataRange": "0 to 3",
            "EndBit": 15,
            "Name": "Aftertreatment 2 Diesel Exhaust Fluid Line Heater 4",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4410,
            "SPNLength": 2,
            "StartBit": 14,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        4411: {
            "Acronym": "A2SCREGP",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 15,
            "Name": "Aftertreatment 2 SCR Exhaust Gas Differential Pressure",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 4411,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        4412: {
            "Acronym": "A2SCREGP",
            "DataRange": "0 to 31",
            "EndBit": 20,
            "Name": "Aftertreatment 2 SCR Exhaust Gas Differential Pressure Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4412,
            "SPNLength": 5,
            "StartBit": 16,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        4413: {
            "Acronym": "A2SCREGT",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 15,
            "Name": "Aftertreatment 2 SCR Catalyst Intake Gas Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 4413,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "500 ms",
            "Units": "deg F"
        },
        4414: {
            "Acronym": "A2SCREGT",
            "DataRange": "0 to 31",
            "EndBit": 20,
            "Name": "Aftertreatment 2 SCR Catalyst Intake Gas Temperature Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4414,
            "SPNLength": 5,
            "StartBit": 16,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        4415: {
            "Acronym": "A2SCREGT",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 39,
            "Name": "Aftertreatment 2 SCR Catalyst Outlet Gas Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 4415,
            "SPNLength": 16,
            "StartBit": 24,
            "TransmissionRate": "500 ms",
            "Units": "deg F"
        },
        4416: {
            "Acronym": "A2SCREGT",
            "DataRange": "0 to 31",
            "EndBit": 44,
            "Name": "Aftertreatment 2 SCR Catalyst Outlet Gas Temperature Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4416,
            "SPNLength": 5,
            "StartBit": 40,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        4417: {
            "Acronym": "A2SCRSI",
            "DataRange": "0 to 3,212.75 L/h",
            "EndBit": 15,
            "Name": "Aftertreatment 2 Diesel Exhaust Fluid Average Consumption",
            "Offset": 0.0,
            "OperationalHigh": 3212.75,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05,
            "SPN": 4417,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "l/h"
        },
        4418: {
            "Acronym": "A2SCRSI",
            "DataRange": "0 to 3,212.75 L/h",
            "EndBit": 31,
            "Name": "Aftertreatment 2 SCR Commanded Catalyst Diesel Exhaust Fluid Consumption",
            "Offset": 0.0,
            "OperationalHigh": 3212.75,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05,
            "SPN": 4418,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "l/h"
        },
        4419: {
            "Acronym": "A2SCRSI",
            "DataRange": "0 to 100 %",
            "EndBit": 39,
            "Name": "Aftertreatment 2 SCR Conversion Efficiency",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 4419,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        442: {
            "Acronym": "AAI",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 15,
            "Name": "Auxiliary Temperature 2",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 442,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "On request",
            "Units": "deg F"
        },
        4420: {
            "Acronym": "A2DEFI",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 7,
            "Name": "Aftertreatment 2 Diesel Exhaust Fluid Temperature 2",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 4420,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        4421: {
            "Acronym": "A2DEFI",
            "DataRange": "0 to 62.5 %",
            "EndBit": 15,
            "Name": "Aftertreatment 2 Diesel Exhaust Fluid Concentration",
            "Offset": 0.0,
            "OperationalHigh": 62.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.25,
            "SPN": 4421,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "1 s",
            "Units": "%"
        },
        4422: {
            "Acronym": "A2DEFI",
            "DataRange": "0 to 1250 microSiemens/mm",
            "EndBit": 23,
            "Name": "Aftertreatment 2 Diesel Exhaust Fluid Conductivity",
            "Offset": 0.0,
            "OperationalHigh": 1250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 5,
            "SPN": 4422,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "1 s",
            "Units": "microSiemens/mm"
        },
        4423: {
            "Acronym": "A2DEFI",
            "DataRange": "0 to 31",
            "EndBit": 28,
            "Name": "Aftertreatment 2 SCR Catalyst Reagent Temperature 2 Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 4423,
            "SPNLength": 5,
            "StartBit": 24,
            "TransmissionRate": "1 s",
            "Units": "binary"
        },
        4424: {
            "Acronym": "A2DEFI",
            "DataRange": "0 to 31",
            "EndBit": 36,
            "Name": "Aftertreatment 2 SCR Catalyst Reagent Properties Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 4424,
            "SPNLength": 5,
            "StartBit": 32,
            "TransmissionRate": "1 s",
            "Units": "binary"
        },
        4425: {
            "Acronym": "A2DEFI",
            "DataRange": "0 to 15",
            "EndBit": 43,
            "Name": "Aftertreatment 2 Diesel Exhaust Fluid Type",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4425,
            "SPNLength": 4,
            "StartBit": 40,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        4426: {
            "Acronym": "A2SCRRT1I",
            "DataRange": "0 to 100 %",
            "EndBit": 7,
            "Name": "Aftertreatment 2 Diesel Exhaust Fluid Tank Level",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 4426,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "%"
        },
        4427: {
            "Acronym": "A2SCRRT1I",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 15,
            "Name": "Aftertreatment 2 Diesel Exhaust Fluid Tank Temperature",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 4427,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        4428: {
            "Acronym": "A2SCRRT1I",
            "DataRange": "0 to 6,425.5 mm (0 to 6.4255 m)",
            "EndBit": 31,
            "Name": "Aftertreatment 2 Diesel Exhaust Fluid Tank Level 2",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 4428,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "1 s",
            "Units": "m"
        },
        4429: {
            "Acronym": "A2SCRRT1I",
            "DataRange": "0 to 31",
            "EndBit": 36,
            "Name": "Aftertreatment 2 Diesel Exhaust Fluid Tank Level Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 4429,
            "SPNLength": 5,
            "StartBit": 32,
            "TransmissionRate": "1 s",
            "Units": "binary"
        },
        4430: {
            "Acronym": "A2SCRRT1I",
            "DataRange": "0 to 31",
            "EndBit": 44,
            "Name": "Aftertreatment 2 Diesel Exhaust Fluid Tank 1 Temperature Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 4430,
            "SPNLength": 5,
            "StartBit": 40,
            "TransmissionRate": "1 s",
            "Units": "binary"
        },
        4431: {
            "Acronym": "A2SCRRT1I",
            "DataRange": "0 to 100 %",
            "EndBit": 55,
            "Name": "Aftertreatment 2 Diesel Exhaust Fluid Tank Heater",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 4431,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "1 s",
            "Units": "%"
        },
        4432: {
            "Acronym": "A2SCRRT1I",
            "DataRange": "0 to 31",
            "EndBit": 60,
            "Name": "Aftertreatment 2 Diesel Exhaust Fluid Tank 1 Heater Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 4432,
            "SPNLength": 5,
            "StartBit": 56,
            "TransmissionRate": "1 s",
            "Units": "binary"
        },
        4433: {
            "Acronym": "A2SCRRT2I",
            "DataRange": "0 to 100 %",
            "EndBit": 7,
            "Name": "Aftertreatment 2 Diesel Exhaust Fluid Tank 2 Level",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 4433,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "%"
        },
        4434: {
            "Acronym": "A2SCRRT2I",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 15,
            "Name": "Aftertreatment 2 Diesel Exhaust Fluid Tank 2 Temperature",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 4434,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        4435: {
            "Acronym": "A2SCRRT2I",
            "DataRange": "0 to 6,425.5 mm (0 to 6.4255 m)",
            "EndBit": 31,
            "Name": "Aftertreatment 2 Diesel Exhaust Fluid Tank 2 Level 2",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 4435,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "1 s",
            "Units": "m"
        },
        4436: {
            "Acronym": "A2SCRRT2I",
            "DataRange": "0 to 31",
            "EndBit": 36,
            "Name": "Aftertreatment 2 Diesel Exhaust Fluid Tank 2 Level Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 4436,
            "SPNLength": 5,
            "StartBit": 32,
            "TransmissionRate": "1 s",
            "Units": "binary"
        },
        4437: {
            "Acronym": "A2SCRRT2I",
            "DataRange": "0 to 31",
            "EndBit": 44,
            "Name": "Aftertreatment 2 Diesel Exhaust Fluid Tank 2 Temperature Prelminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 4437,
            "SPNLength": 5,
            "StartBit": 40,
            "TransmissionRate": "1 s",
            "Units": "binary"
        },
        4438: {
            "Acronym": "A2SCRRT2I",
            "DataRange": "0 to 15",
            "EndBit": 51,
            "Name": "Aftertreatment 2 Diesel Exhaust Fluid Tank 2 Heater",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4438,
            "SPNLength": 4,
            "StartBit": 48,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        4439: {
            "Acronym": "A2SCRRT2I",
            "DataRange": "0 to 31",
            "EndBit": 60,
            "Name": "Aftertreatment 2 Diesel Exhaust Fluid Tank 2 Heater Prelminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 4439,
            "SPNLength": 5,
            "StartBit": 56,
            "TransmissionRate": "1 s",
            "Units": "binary"
        },
        444: {
            "Acronym": "VEP2",
            "DataRange": "0 to 3212.75 V",
            "EndBit": 15,
            "Name": "Battery Potential / Power Input 2",
            "Offset": 0.0,
            "OperationalHigh": 3212.75,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05,
            "SPN": 444,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "Volts"
        },
        4440: {
            "Acronym": "A2DEFSI",
            "DataRange": "0 to 32,127.5 rpm",
            "EndBit": 15,
            "Name": "Aftertreatment 2 Diesel Exhaust Fluid Pump Motor Speed",
            "Offset": 0.0,
            "OperationalHigh": 32127.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.5,
            "SPN": 4440,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "rpm"
        },
        4441: {
            "Acronym": "A2DEFSI",
            "DataRange": "0 to 100 %",
            "EndBit": 23,
            "Name": "Aftertreatment 2 Diesel Exhaust Fluid Pump Drive Percentage",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 4441,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "1 s",
            "Units": "%"
        },
        4442: {
            "Acronym": "A2DEFSI",
            "DataRange": "0 to 100 %",
            "EndBit": 31,
            "Name": "Aftertreatment 2 Diesel Exhaust Fluid Return Valve",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 4442,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "1 s",
            "Units": "%"
        },
        4443: {
            "Acronym": "A2SCRAI",
            "DataRange": "-200 to 3012.75 ppm",
            "EndBit": 15,
            "Name": "Aftertreatment 2 Outlet NH3",
            "Offset": -200.0,
            "OperationalHigh": 3012.75,
            "OperationalLow": -200.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05,
            "SPN": 4443,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "50 ms",
            "Units": "ppm"
        },
        4444: {
            "Acronym": "A2SCRAI",
            "DataRange": "0 to 31",
            "EndBit": 20,
            "Name": "Aftertreatment 2 Outlet NH3 Sensor Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 4444,
            "SPNLength": 5,
            "StartBit": 16,
            "TransmissionRate": "50 ms",
            "Units": "binary"
        },
        4445: {
            "Acronym": "A2SCRAI",
            "DataRange": "0 to 3",
            "EndBit": 25,
            "Name": "Aftertreatment 2 Outlet NH3 Reading Stable",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4445,
            "SPNLength": 2,
            "StartBit": 24,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        4446: {
            "Acronym": "A2SCRAI",
            "DataRange": "0 to 3",
            "EndBit": 27,
            "Name": "Aftertreatment 2 Outlet NH3 Gas Sensor Power In Range",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4446,
            "SPNLength": 2,
            "StartBit": 26,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        4447: {
            "Acronym": "A2SCRAI",
            "DataRange": "0 to 3",
            "EndBit": 29,
            "Name": "Aftertreatment 2 Outlet NH3 Gas Sensor at Temperature",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4447,
            "SPNLength": 2,
            "StartBit": 28,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        4448: {
            "Acronym": "A2SCRAI",
            "DataRange": "0 to 31",
            "EndBit": 36,
            "Name": "Aftertreatment 2 Outlet NH3 Gas Sensor Heater Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 4448,
            "SPNLength": 5,
            "StartBit": 32,
            "TransmissionRate": "50 ms",
            "Units": "binary"
        },
        4449: {
            "Acronym": "A2SCRAI",
            "DataRange": "0 to 3",
            "EndBit": 41,
            "Name": "Aftertreatment 2 Outlet NH3 Gas Sensor Heater Control",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4449,
            "SPNLength": 2,
            "StartBit": 40,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        4454: {
            "Acronym": "DRC",
            "DataRange": "0 to 3",
            "EndBit": 25,
            "Name": "Retract Status of ramp 4",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4454,
            "SPNLength": 2,
            "StartBit": 24,
            "TransmissionRate": "Every 1 s while active and on change of state but no faster than every 100 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010:  1 s when active and on change of state",
            "Units": "bit"
        },
        4455: {
            "Acronym": "DRC",
            "DataRange": "0 to 3",
            "EndBit": 27,
            "Name": "Enable status of ramp 4",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4455,
            "SPNLength": 2,
            "StartBit": 26,
            "TransmissionRate": "Every 1 s while active and on change of state but no faster than every 100 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010:  1 s when active and on change of state",
            "Units": "bit"
        },
        4456: {
            "Acronym": "DRC",
            "DataRange": "0 to 3",
            "EndBit": 29,
            "Name": "Movement status of ramp 4",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4456,
            "SPNLength": 2,
            "StartBit": 28,
            "TransmissionRate": "Every 1 s while active and on change of state but no faster than every 100 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010:  1 s when active and on change of state",
            "Units": "bit"
        },
        4457: {
            "Acronym": "DRC",
            "DataRange": "0 to 3",
            "EndBit": 33,
            "Name": "Retract Status of ramp 5",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4457,
            "SPNLength": 2,
            "StartBit": 32,
            "TransmissionRate": "Every 1 s while active and on change of state but no faster than every 100 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010:  1 s when active and on change of state",
            "Units": "bit"
        },
        4458: {
            "Acronym": "DRC",
            "DataRange": "0 to 3",
            "EndBit": 35,
            "Name": "Enable status of ramp 5",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4458,
            "SPNLength": 2,
            "StartBit": 34,
            "TransmissionRate": "Every 1 s while active and on change of state but no faster than every 100 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010:  1 s when active and on change of state",
            "Units": "bit"
        },
        4459: {
            "Acronym": "DRC",
            "DataRange": "0 to 3",
            "EndBit": 37,
            "Name": "Movement status of ramp 5",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4459,
            "SPNLength": 2,
            "StartBit": 36,
            "TransmissionRate": "Every 1 s while active and on change of state but no faster than every 100 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010:  1 s when active and on change of state",
            "Units": "bit"
        },
        4460: {
            "Acronym": "JLCM",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Joystick 1 Lamp 1 Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4460,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        4461: {
            "Acronym": "JLCM",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Joystick 1 Lamp 2 Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4461,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        4462: {
            "Acronym": "JLCM",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Joystick 1 Lamp 3 Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4462,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        4463: {
            "Acronym": "JLCM",
            "DataRange": "0 to 3",
            "EndBit": 7,
            "Name": "Joystick 1 Lamp 4 Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4463,
            "SPNLength": 2,
            "StartBit": 6,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        4464: {
            "Acronym": "JLCM",
            "DataRange": "0 to 3",
            "EndBit": 9,
            "Name": "Joystick 1 Lamp 5 Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4464,
            "SPNLength": 2,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        4465: {
            "Acronym": "JLCM",
            "DataRange": "0 to 3",
            "EndBit": 11,
            "Name": "Joystick 1 Lamp 6 Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4465,
            "SPNLength": 2,
            "StartBit": 10,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        4466: {
            "Acronym": "JLCM",
            "DataRange": "0 to 3",
            "EndBit": 13,
            "Name": "Joystick 1 Lamp 7 Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4466,
            "SPNLength": 2,
            "StartBit": 12,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        4467: {
            "Acronym": "JLCM",
            "DataRange": "0 to 3",
            "EndBit": 15,
            "Name": "Joystick 1 Lamp 8 Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4467,
            "SPNLength": 2,
            "StartBit": 14,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        4468: {
            "Acronym": "JLCM",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Joystick 1 Lamp 9 Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4468,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        4469: {
            "Acronym": "JLCM",
            "DataRange": "0 to 3",
            "EndBit": 19,
            "Name": "Joystick 1 Lamp 10 Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4469,
            "SPNLength": 2,
            "StartBit": 18,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        4470: {
            "Acronym": "JLCM",
            "DataRange": "0 to 3",
            "EndBit": 21,
            "Name": "Joystick 2 Lamp 1 Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4470,
            "SPNLength": 2,
            "StartBit": 20,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        4471: {
            "Acronym": "JLCM",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Joystick 2 Lamp 2 Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4471,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        4472: {
            "Acronym": "JLCM",
            "DataRange": "0 to 3",
            "EndBit": 25,
            "Name": "Joystick 2 Lamp 3 Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4472,
            "SPNLength": 2,
            "StartBit": 24,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        4473: {
            "Acronym": "JLCM",
            "DataRange": "0 to 3",
            "EndBit": 27,
            "Name": "Joystick 2 Lamp 4 Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4473,
            "SPNLength": 2,
            "StartBit": 26,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        4474: {
            "Acronym": "JLCM",
            "DataRange": "0 to 3",
            "EndBit": 29,
            "Name": "Joystick 2 Lamp 5 Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4474,
            "SPNLength": 2,
            "StartBit": 28,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        4475: {
            "Acronym": "JLCM",
            "DataRange": "0 to 3",
            "EndBit": 31,
            "Name": "Joystick 2 Lamp 6 Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4475,
            "SPNLength": 2,
            "StartBit": 30,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        4476: {
            "Acronym": "JLCM",
            "DataRange": "0 to 3",
            "EndBit": 33,
            "Name": "Joystick 2 Lamp 7 Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4476,
            "SPNLength": 2,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        4477: {
            "Acronym": "JLCM",
            "DataRange": "0 to 3",
            "EndBit": 35,
            "Name": "Joystick 2 Lamp 8 Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4477,
            "SPNLength": 2,
            "StartBit": 34,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        4478: {
            "Acronym": "JLCM",
            "DataRange": "0 to 3",
            "EndBit": 37,
            "Name": "Joystick 2 Lamp 9 Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4478,
            "SPNLength": 2,
            "StartBit": 36,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        4479: {
            "Acronym": "JLCM",
            "DataRange": "0 to 3",
            "EndBit": 39,
            "Name": "Joystick 2 Lamp 10 Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4479,
            "SPNLength": 2,
            "StartBit": 38,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        4480: {
            "Acronym": "JLCM",
            "DataRange": "0 to 3",
            "EndBit": 41,
            "Name": "Joystick 3 Lamp 1 Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4480,
            "SPNLength": 2,
            "StartBit": 40,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        4481: {
            "Acronym": "JLCM",
            "DataRange": "0 to 3",
            "EndBit": 43,
            "Name": "Joystick 3 Lamp 2 Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4481,
            "SPNLength": 2,
            "StartBit": 42,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        4482: {
            "Acronym": "JLCM",
            "DataRange": "0 to 3",
            "EndBit": 45,
            "Name": "Joystick 3 Lamp 3 Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4482,
            "SPNLength": 2,
            "StartBit": 44,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        4483: {
            "Acronym": "JLCM",
            "DataRange": "0 to 3",
            "EndBit": 47,
            "Name": "Joystick 3 Lamp 4 Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4483,
            "SPNLength": 2,
            "StartBit": 46,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        4484: {
            "Acronym": "JLCM",
            "DataRange": "0 to 3",
            "EndBit": 49,
            "Name": "Joystick 3 Lamp 5 Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4484,
            "SPNLength": 2,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        4485: {
            "Acronym": "JLCM",
            "DataRange": "0 to 3",
            "EndBit": 51,
            "Name": "Joystick 3 Lamp 6 Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4485,
            "SPNLength": 2,
            "StartBit": 50,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        4486: {
            "Acronym": "JLCM",
            "DataRange": "0 to 3",
            "EndBit": 53,
            "Name": "Joystick 3 Lamp 7 Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4486,
            "SPNLength": 2,
            "StartBit": 52,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        4487: {
            "Acronym": "JLCM",
            "DataRange": "0 to 3",
            "EndBit": 55,
            "Name": "Joystick 3 Lamp 8 Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4487,
            "SPNLength": 2,
            "StartBit": 54,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        4488: {
            "Acronym": "JLCM",
            "DataRange": "0 to 3",
            "EndBit": 57,
            "Name": "Joystick 3 Lamp 9 Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4488,
            "SPNLength": 2,
            "StartBit": 56,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        4489: {
            "Acronym": "JLCM",
            "DataRange": "0 to 3",
            "EndBit": 59,
            "Name": "Joystick 3 Lamp 10 Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4489,
            "SPNLength": 2,
            "StartBit": 58,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        4490: {
            "Acronym": "AMB2",
            "DataRange": "0 to 642.55 g/kg",
            "EndBit": 31,
            "Name": "Specific Humidity",
            "Offset": 0.0,
            "OperationalHigh": 642.55,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.01,
            "SPN": 4490,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "1 s",
            "Units": "g/kg"
        },
        4491: {
            "Acronym": "BJM4",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Joystick 4 X-Axis Neutral Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4491,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4492: {
            "Acronym": "BJM4",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Joystick 4 X-Axis Lever Left Negative Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4492,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4493: {
            "Acronym": "BJM4",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Joystick 4 X-Axis Lever Right Positive Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4493,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4494: {
            "Acronym": "BJM4",
            "DataRange": "0 to 102 %",
            "EndBit": 15,
            "Name": "Joystick 4 X-Axis Position",
            "Offset": 0.0,
            "OperationalHigh": 102.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 4494,
            "SPNLength": 10,
            "StartBit": 6,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "%"
        },
        4495: {
            "Acronym": "BJM4",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Joystick 4 Y-Axis Neutral Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4495,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4496: {
            "Acronym": "BJM4",
            "DataRange": "0 to 3",
            "EndBit": 19,
            "Name": "Joystick 4 Y-Axis Lever Back Negative Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4496,
            "SPNLength": 2,
            "StartBit": 18,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4497: {
            "Acronym": "BJM4",
            "DataRange": "0 to 3",
            "EndBit": 21,
            "Name": "Joystick 4 Y-Axis Lever Forward Positive Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4497,
            "SPNLength": 2,
            "StartBit": 20,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4498: {
            "Acronym": "BJM4",
            "DataRange": "0 to 102 %",
            "EndBit": 31,
            "Name": "Joystick 4 Y-Axis Position",
            "Offset": 0.0,
            "OperationalHigh": 102.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 4498,
            "SPNLength": 10,
            "StartBit": 22,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "%"
        },
        4499: {
            "Acronym": "BJM4",
            "DataRange": "0 to 3",
            "EndBit": 37,
            "Name": "Joystick 4 Y-Axis Detent Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4499,
            "SPNLength": 2,
            "StartBit": 36,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4500: {
            "Acronym": "BJM4",
            "DataRange": "0 to 3",
            "EndBit": 39,
            "Name": "Joystick 4 X-Axis Detent Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4500,
            "SPNLength": 2,
            "StartBit": 38,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4501: {
            "Acronym": "BJM4",
            "DataRange": "0 to 3",
            "EndBit": 41,
            "Name": "Joystick 4 Button 4 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4501,
            "SPNLength": 2,
            "StartBit": 40,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4502: {
            "Acronym": "BJM4",
            "DataRange": "0 to 3",
            "EndBit": 43,
            "Name": "Joystick 4 Button 3 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4502,
            "SPNLength": 2,
            "StartBit": 42,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4503: {
            "Acronym": "BJM4",
            "DataRange": "0 to 3",
            "EndBit": 45,
            "Name": "Joystick 4 Button 2 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4503,
            "SPNLength": 2,
            "StartBit": 44,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4504: {
            "Acronym": "BJM4",
            "DataRange": "0 to 3",
            "EndBit": 47,
            "Name": "Joystick 4 Button 1 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4504,
            "SPNLength": 2,
            "StartBit": 46,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4505: {
            "Acronym": "BJM4",
            "DataRange": "0 to 3",
            "EndBit": 49,
            "Name": "Joystick 4 Button 8 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4505,
            "SPNLength": 2,
            "StartBit": 48,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4506: {
            "Acronym": "BJM4",
            "DataRange": "0 to 3",
            "EndBit": 51,
            "Name": "Joystick 4 Button 7 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4506,
            "SPNLength": 2,
            "StartBit": 50,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4507: {
            "Acronym": "BJM4",
            "DataRange": "0 to 3",
            "EndBit": 53,
            "Name": "Joystick 4 Button 6 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4507,
            "SPNLength": 2,
            "StartBit": 52,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4508: {
            "Acronym": "BJM4",
            "DataRange": "0 to 3",
            "EndBit": 55,
            "Name": "Joystick 4 Button 5 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4508,
            "SPNLength": 2,
            "StartBit": 54,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4509: {
            "Acronym": "BJM4",
            "DataRange": "0 to 3",
            "EndBit": 57,
            "Name": "Joystick 4 Button 12 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4509,
            "SPNLength": 2,
            "StartBit": 56,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4510: {
            "Acronym": "BJM4",
            "DataRange": "0 to 3",
            "EndBit": 59,
            "Name": "Joystick 4 Button 11 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4510,
            "SPNLength": 2,
            "StartBit": 58,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4511: {
            "Acronym": "BJM4",
            "DataRange": "0 to 3",
            "EndBit": 61,
            "Name": "Joystick 4 Button 10 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4511,
            "SPNLength": 2,
            "StartBit": 60,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4512: {
            "Acronym": "BJM4",
            "DataRange": "0 to 3",
            "EndBit": 63,
            "Name": "Joystick 4 Button 9 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4512,
            "SPNLength": 2,
            "StartBit": 62,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4513: {
            "Acronym": "EJM4",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Joystick 4 Grip X-Axis Neutral Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4513,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4514: {
            "Acronym": "EJM4",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Joystick 4 Grip X-Axis Lever Left Negative Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4514,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4515: {
            "Acronym": "EJM4",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Joystick 4 Grip X-Axis Lever Right Positive Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4515,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4516: {
            "Acronym": "EJM4",
            "DataRange": "0 to 102 %",
            "EndBit": 15,
            "Name": "Joystick 4 Grip X-Axis Position",
            "Offset": 0.0,
            "OperationalHigh": 102.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 4516,
            "SPNLength": 10,
            "StartBit": 6,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "%"
        },
        4517: {
            "Acronym": "EJM4",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Joystick 4 Grip Y-Axis Neutral Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4517,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4518: {
            "Acronym": "EJM4",
            "DataRange": "0 to 3",
            "EndBit": 19,
            "Name": "Joystick 4 Grip Y-Axis Lever Back Negative Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4518,
            "SPNLength": 2,
            "StartBit": 18,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4519: {
            "Acronym": "EJM4",
            "DataRange": "0 to 3",
            "EndBit": 21,
            "Name": "Joystick 4 Grip Y-Axis Lever Forward Positive Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4519,
            "SPNLength": 2,
            "StartBit": 20,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4520: {
            "Acronym": "EJM4",
            "DataRange": "0 to 102 %",
            "EndBit": 31,
            "Name": "Joystick 4 Grip Y-Axis Position",
            "Offset": 0.0,
            "OperationalHigh": 102.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 4520,
            "SPNLength": 10,
            "StartBit": 22,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "%"
        },
        4521: {
            "Acronym": "EJM4",
            "DataRange": "0 to 3",
            "EndBit": 33,
            "Name": "Joystick 4 Theta-Axis Neutral Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4521,
            "SPNLength": 2,
            "StartBit": 32,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4522: {
            "Acronym": "EJM4",
            "DataRange": "0 to 3",
            "EndBit": 35,
            "Name": "Joystick 4 Theta-Axis Counter Clockwise Negative Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4522,
            "SPNLength": 2,
            "StartBit": 34,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4523: {
            "Acronym": "EJM4",
            "DataRange": "0 to 3",
            "EndBit": 37,
            "Name": "Joystick 4 Theta-Axis Clockwise Positive Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4523,
            "SPNLength": 2,
            "StartBit": 36,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4524: {
            "Acronym": "EJM4",
            "DataRange": "0 to 102 %",
            "EndBit": 47,
            "Name": "Joystick 4 Theta-Axis Position",
            "Offset": 0.0,
            "OperationalHigh": 102.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 4524,
            "SPNLength": 10,
            "StartBit": 38,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "%"
        },
        4525: {
            "Acronym": "EJM4",
            "DataRange": "0 to 3",
            "EndBit": 51,
            "Name": "Joystick 4 Theta-Axis Detent Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4525,
            "SPNLength": 2,
            "StartBit": 50,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4526: {
            "Acronym": "EJM4",
            "DataRange": "0 to 3",
            "EndBit": 53,
            "Name": "Joystick 4 Grip Y-Axis Detent Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4526,
            "SPNLength": 2,
            "StartBit": 52,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4527: {
            "Acronym": "EJM4",
            "DataRange": "0 to 3",
            "EndBit": 55,
            "Name": "Joystick 4 Grip X-Axis Detent Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4527,
            "SPNLength": 2,
            "StartBit": 54,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4528: {
            "Acronym": "BJM5",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Joystick 5 X-Axis Neutral Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4528,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4529: {
            "Acronym": "BJM5",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Joystick 5 X-Axis Lever Left Negative Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4529,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4530: {
            "Acronym": "BJM5",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Joystick 5 X-Axis Lever Right Positive Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4530,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4531: {
            "Acronym": "BJM5",
            "DataRange": "0 to 102 %",
            "EndBit": 15,
            "Name": "Joystick 5 X-Axis Position",
            "Offset": 0.0,
            "OperationalHigh": 102.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 4531,
            "SPNLength": 10,
            "StartBit": 6,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "%"
        },
        4532: {
            "Acronym": "BJM5",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Joystick 5 Y-Axis Neutral Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4532,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4533: {
            "Acronym": "BJM5",
            "DataRange": "0 to 3",
            "EndBit": 19,
            "Name": "Joystick 5 Y-Axis Lever Back Negative Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4533,
            "SPNLength": 2,
            "StartBit": 18,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4534: {
            "Acronym": "BJM5",
            "DataRange": "0 to 3",
            "EndBit": 21,
            "Name": "Joystick 5 Y-Axis Lever Forward Positive Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4534,
            "SPNLength": 2,
            "StartBit": 20,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4535: {
            "Acronym": "BJM5",
            "DataRange": "0 to 102 %",
            "EndBit": 31,
            "Name": "Joystick 5 Y-Axis Position",
            "Offset": 0.0,
            "OperationalHigh": 102.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 4535,
            "SPNLength": 10,
            "StartBit": 22,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "%"
        },
        4536: {
            "Acronym": "BJM5",
            "DataRange": "0 to 3",
            "EndBit": 37,
            "Name": "Joystick 5 Y-Axis Detent Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4536,
            "SPNLength": 2,
            "StartBit": 36,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4537: {
            "Acronym": "BJM5",
            "DataRange": "0 to 3",
            "EndBit": 39,
            "Name": "Joystick 5 X-Axis Detent Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4537,
            "SPNLength": 2,
            "StartBit": 38,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4538: {
            "Acronym": "BJM5",
            "DataRange": "0 to 3",
            "EndBit": 41,
            "Name": "Joystick 5 Button 4 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4538,
            "SPNLength": 2,
            "StartBit": 40,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4539: {
            "Acronym": "BJM5",
            "DataRange": "0 to 3",
            "EndBit": 43,
            "Name": "Joystick 5 Button 3 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4539,
            "SPNLength": 2,
            "StartBit": 42,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4540: {
            "Acronym": "BJM5",
            "DataRange": "0 to 3",
            "EndBit": 45,
            "Name": "Joystick 5 Button 2 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4540,
            "SPNLength": 2,
            "StartBit": 44,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4541: {
            "Acronym": "BJM5",
            "DataRange": "0 to 3",
            "EndBit": 47,
            "Name": "Joystick 5 Button 1 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4541,
            "SPNLength": 2,
            "StartBit": 46,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4542: {
            "Acronym": "BJM5",
            "DataRange": "0 to 3",
            "EndBit": 49,
            "Name": "Joystick 5 Button 8 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4542,
            "SPNLength": 2,
            "StartBit": 48,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4543: {
            "Acronym": "BJM5",
            "DataRange": "0 to 3",
            "EndBit": 51,
            "Name": "Joystick 5 Button 7 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4543,
            "SPNLength": 2,
            "StartBit": 50,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4544: {
            "Acronym": "BJM5",
            "DataRange": "0 to 3",
            "EndBit": 53,
            "Name": "Joystick 5 Button 6 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4544,
            "SPNLength": 2,
            "StartBit": 52,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4545: {
            "Acronym": "BJM5",
            "DataRange": "0 to 3",
            "EndBit": 55,
            "Name": "Joystick 5 Button 5 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4545,
            "SPNLength": 2,
            "StartBit": 54,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4546: {
            "Acronym": "BJM5",
            "DataRange": "0 to 3",
            "EndBit": 57,
            "Name": "Joystick 5 Button 12 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4546,
            "SPNLength": 2,
            "StartBit": 56,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4547: {
            "Acronym": "BJM5",
            "DataRange": "0 to 3",
            "EndBit": 59,
            "Name": "Joystick 5 Button 11 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4547,
            "SPNLength": 2,
            "StartBit": 58,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4548: {
            "Acronym": "BJM5",
            "DataRange": "0 to 3",
            "EndBit": 61,
            "Name": "Joystick 5 Button 10 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4548,
            "SPNLength": 2,
            "StartBit": 60,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4549: {
            "Acronym": "BJM5",
            "DataRange": "0 to 3",
            "EndBit": 63,
            "Name": "Joystick 5 Button 9 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4549,
            "SPNLength": 2,
            "StartBit": 62,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4550: {
            "Acronym": "EJM5",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Joystick 5 Grip X-Axis Neutral Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4550,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4551: {
            "Acronym": "EJM5",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Joystick 5 Grip X-Axis Lever Left Negative Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4551,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4552: {
            "Acronym": "EJM5",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Joystick 5 Grip X-Axis Lever Right Positive Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4552,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4553: {
            "Acronym": "EJM5",
            "DataRange": "0 to 102 %",
            "EndBit": 15,
            "Name": "Joystick 5 Grip X-Axis Position",
            "Offset": 0.0,
            "OperationalHigh": 102.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 4553,
            "SPNLength": 10,
            "StartBit": 6,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "%"
        },
        4554: {
            "Acronym": "EJM5",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Joystick 5 Grip Y-Axis Neutral Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4554,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4555: {
            "Acronym": "EJM5",
            "DataRange": "0 to 3",
            "EndBit": 19,
            "Name": "Joystick 5 Grip Y-Axis Lever Back Negative Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4555,
            "SPNLength": 2,
            "StartBit": 18,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4556: {
            "Acronym": "EJM5",
            "DataRange": "0 to 3",
            "EndBit": 21,
            "Name": "Joystick 5 Grip Y-Axis Lever Forward Positive Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4556,
            "SPNLength": 2,
            "StartBit": 20,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4557: {
            "Acronym": "EJM5",
            "DataRange": "0 to 102 %",
            "EndBit": 31,
            "Name": "Joystick 5 Grip Y-Axis Position",
            "Offset": 0.0,
            "OperationalHigh": 102.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 4557,
            "SPNLength": 10,
            "StartBit": 22,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "%"
        },
        4558: {
            "Acronym": "EJM5",
            "DataRange": "0 to 3",
            "EndBit": 33,
            "Name": "Joystick 5 Theta-Axis Neutral Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4558,
            "SPNLength": 2,
            "StartBit": 32,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4559: {
            "Acronym": "EJM5",
            "DataRange": "0 to 3",
            "EndBit": 35,
            "Name": "Joystick 5 Theta-Axis Counter Clockwise Negative Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4559,
            "SPNLength": 2,
            "StartBit": 34,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4560: {
            "Acronym": "EJM5",
            "DataRange": "0 to 3",
            "EndBit": 37,
            "Name": "Joystick 5 Theta-Axis Clockwise Positive Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4560,
            "SPNLength": 2,
            "StartBit": 36,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4561: {
            "Acronym": "EJM5",
            "DataRange": "0 to 102 %",
            "EndBit": 47,
            "Name": "Joystick 5 Theta-Axis Position",
            "Offset": 0.0,
            "OperationalHigh": 102.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 4561,
            "SPNLength": 10,
            "StartBit": 38,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "%"
        },
        4562: {
            "Acronym": "EJM5",
            "DataRange": "0 to 3",
            "EndBit": 51,
            "Name": "Joystick 5 Theta-Axis Detent Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4562,
            "SPNLength": 2,
            "StartBit": 50,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4563: {
            "Acronym": "EJM5",
            "DataRange": "0 to 3",
            "EndBit": 53,
            "Name": "Joystick 5 Grip Y-Axis Detent Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4563,
            "SPNLength": 2,
            "StartBit": 52,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4564: {
            "Acronym": "EJM5",
            "DataRange": "0 to 3",
            "EndBit": 55,
            "Name": "Joystick 5 Grip X-Axis Detent Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4564,
            "SPNLength": 2,
            "StartBit": 54,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4565: {
            "Acronym": "BJM6",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Joystick 6 X-Axis Neutral Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4565,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4566: {
            "Acronym": "BJM6",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Joystick 6 X-Axis Lever Left Negative Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4566,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4567: {
            "Acronym": "BJM6",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Joystick 6 X-Axis Lever Right Positive Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4567,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4568: {
            "Acronym": "BJM6",
            "DataRange": "0 to 102 %",
            "EndBit": 15,
            "Name": "Joystick 6 X-Axis Position",
            "Offset": 0.0,
            "OperationalHigh": 102.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 4568,
            "SPNLength": 10,
            "StartBit": 6,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "%"
        },
        4569: {
            "Acronym": "BJM6",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Joystick 6 Y-Axis Neutral Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4569,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4570: {
            "Acronym": "BJM6",
            "DataRange": "0 to 3",
            "EndBit": 19,
            "Name": "Joystick 6 Y-Axis Lever Back Negative Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4570,
            "SPNLength": 2,
            "StartBit": 18,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4571: {
            "Acronym": "BJM6",
            "DataRange": "0 to 3",
            "EndBit": 21,
            "Name": "Joystick 6 Y-Axis Lever Forward Positive Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4571,
            "SPNLength": 2,
            "StartBit": 20,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4572: {
            "Acronym": "BJM6",
            "DataRange": "0 to 102 %",
            "EndBit": 31,
            "Name": "Joystick 6 Y-Axis Position",
            "Offset": 0.0,
            "OperationalHigh": 102.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 4572,
            "SPNLength": 10,
            "StartBit": 22,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "%"
        },
        4573: {
            "Acronym": "BJM6",
            "DataRange": "0 to 3",
            "EndBit": 37,
            "Name": "Joystick 6 Y-Axis Detent Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4573,
            "SPNLength": 2,
            "StartBit": 36,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4574: {
            "Acronym": "BJM6",
            "DataRange": "0 to 3",
            "EndBit": 39,
            "Name": "Joystick 6 X-Axis Detent Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4574,
            "SPNLength": 2,
            "StartBit": 38,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4575: {
            "Acronym": "BJM6",
            "DataRange": "0 to 3",
            "EndBit": 41,
            "Name": "Joystick 6 Button 4 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4575,
            "SPNLength": 2,
            "StartBit": 40,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4576: {
            "Acronym": "BJM6",
            "DataRange": "0 to 3",
            "EndBit": 43,
            "Name": "Joystick 6 Button 3 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4576,
            "SPNLength": 2,
            "StartBit": 42,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4577: {
            "Acronym": "BJM6",
            "DataRange": "0 to 3",
            "EndBit": 45,
            "Name": "Joystick 6 Button 2 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4577,
            "SPNLength": 2,
            "StartBit": 44,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4578: {
            "Acronym": "BJM6",
            "DataRange": "0 to 3",
            "EndBit": 47,
            "Name": "Joystick 6 Button 1 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4578,
            "SPNLength": 2,
            "StartBit": 46,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4579: {
            "Acronym": "BJM6",
            "DataRange": "0 to 3",
            "EndBit": 49,
            "Name": "Joystick 6 Button 8 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4579,
            "SPNLength": 2,
            "StartBit": 48,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4580: {
            "Acronym": "BJM6",
            "DataRange": "0 to 3",
            "EndBit": 51,
            "Name": "Joystick 6 Button 7 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4580,
            "SPNLength": 2,
            "StartBit": 50,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4581: {
            "Acronym": "BJM6",
            "DataRange": "0 to 3",
            "EndBit": 53,
            "Name": "Joystick 6 Button 6 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4581,
            "SPNLength": 2,
            "StartBit": 52,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4582: {
            "Acronym": "BJM6",
            "DataRange": "0 to 3",
            "EndBit": 55,
            "Name": "Joystick 6 Button 5 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4582,
            "SPNLength": 2,
            "StartBit": 54,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4583: {
            "Acronym": "BJM6",
            "DataRange": "0 to 3",
            "EndBit": 57,
            "Name": "Joystick 6 Button 12 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4583,
            "SPNLength": 2,
            "StartBit": 56,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4584: {
            "Acronym": "BJM6",
            "DataRange": "0 to 3",
            "EndBit": 59,
            "Name": "Joystick 6 Button 11 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4584,
            "SPNLength": 2,
            "StartBit": 58,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4585: {
            "Acronym": "BJM6",
            "DataRange": "0 to 3",
            "EndBit": 61,
            "Name": "Joystick 6 Button 10 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4585,
            "SPNLength": 2,
            "StartBit": 60,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4586: {
            "Acronym": "BJM6",
            "DataRange": "0 to 3",
            "EndBit": 63,
            "Name": "Joystick 6 Button 9 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4586,
            "SPNLength": 2,
            "StartBit": 62,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4587: {
            "Acronym": "EJM6",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Joystick 6 Grip X-Axis Neutral Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4587,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4588: {
            "Acronym": "EJM6",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Joystick 6 Grip X-Axis Lever Left Negative Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4588,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4589: {
            "Acronym": "EJM6",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Joystick 6 Grip X-Axis Lever Right Positive Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4589,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4590: {
            "Acronym": "EJM6",
            "DataRange": "0 to 102 %",
            "EndBit": 15,
            "Name": "Joystick 6 Grip X-Axis Position",
            "Offset": 0.0,
            "OperationalHigh": 102.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 4590,
            "SPNLength": 10,
            "StartBit": 6,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "%"
        },
        4591: {
            "Acronym": "EJM6",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Joystick 6 Grip Y-Axis Neutral Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4591,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4592: {
            "Acronym": "EJM6",
            "DataRange": "0 to 3",
            "EndBit": 19,
            "Name": "Joystick 6 Grip Y-Axis Lever Back Negative Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4592,
            "SPNLength": 2,
            "StartBit": 18,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4593: {
            "Acronym": "EJM6",
            "DataRange": "0 to 3",
            "EndBit": 21,
            "Name": "Joystick 6 Grip Y-Axis Lever Forward Positive Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4593,
            "SPNLength": 2,
            "StartBit": 20,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4594: {
            "Acronym": "EJM6",
            "DataRange": "0 to 102 %",
            "EndBit": 31,
            "Name": "Joystick 6 Grip Y-Axis Position",
            "Offset": 0.0,
            "OperationalHigh": 102.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 4594,
            "SPNLength": 10,
            "StartBit": 22,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "%"
        },
        4595: {
            "Acronym": "EJM6",
            "DataRange": "0 to 3",
            "EndBit": 33,
            "Name": "Joystick 6 Theta-Axis Neutral Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4595,
            "SPNLength": 2,
            "StartBit": 32,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4596: {
            "Acronym": "EJM6",
            "DataRange": "0 to 3",
            "EndBit": 35,
            "Name": "Joystick 6 Theta-Axis Counter Clockwise Negative Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4596,
            "SPNLength": 2,
            "StartBit": 34,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4597: {
            "Acronym": "EJM6",
            "DataRange": "0 to 3",
            "EndBit": 37,
            "Name": "Joystick 6 Theta-Axis Clockwise Positive Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4597,
            "SPNLength": 2,
            "StartBit": 36,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4598: {
            "Acronym": "EJM6",
            "DataRange": "0 to 102 %",
            "EndBit": 47,
            "Name": "Joystick 6 Theta-Axis Position",
            "Offset": 0.0,
            "OperationalHigh": 102.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 4598,
            "SPNLength": 10,
            "StartBit": 38,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "%"
        },
        4599: {
            "Acronym": "EJM6",
            "DataRange": "0 to 3",
            "EndBit": 51,
            "Name": "Joystick 6 Theta-Axis Detent Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4599,
            "SPNLength": 2,
            "StartBit": 50,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        46: {
            "Acronym": "AIR1",
            "DataRange": "0.0 to 290.075476 psi",
            "EndBit": 7,
            "Name": "Pneumatic Supply Pressure",
            "Offset": 0.0,
            "OperationalHigh": 290.075476,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.160301904,
            "SPN": 46,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "1 sec",
            "Units": "psi"
        },
        4600: {
            "Acronym": "EJM6",
            "DataRange": "0 to 3",
            "EndBit": 53,
            "Name": "Joystick 6 Grip Y-Axis Detent Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4600,
            "SPNLength": 2,
            "StartBit": 52,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4601: {
            "Acronym": "EJM6",
            "DataRange": "0 to 3",
            "EndBit": 55,
            "Name": "Joystick 6 Grip X-Axis Detent Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4601,
            "SPNLength": 2,
            "StartBit": 54,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4602: {
            "Acronym": "BJM7",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Joystick 7 X-Axis Neutral Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4602,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4603: {
            "Acronym": "BJM7",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Joystick 7 X-Axis Lever Left Negative Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4603,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4604: {
            "Acronym": "BJM7",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Joystick 7 X-Axis Lever Right Positive Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4604,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4605: {
            "Acronym": "BJM7",
            "DataRange": "0 to 102 %",
            "EndBit": 15,
            "Name": "Joystick 7 X-Axis Position",
            "Offset": 0.0,
            "OperationalHigh": 102.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 4605,
            "SPNLength": 10,
            "StartBit": 6,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "%"
        },
        4606: {
            "Acronym": "BJM7",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Joystick 7 Y-Axis Neutral Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4606,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4607: {
            "Acronym": "BJM7",
            "DataRange": "0 to 3",
            "EndBit": 19,
            "Name": "Joystick 7 Y-Axis Lever Back Negative Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4607,
            "SPNLength": 2,
            "StartBit": 18,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4608: {
            "Acronym": "BJM7",
            "DataRange": "0 to 3",
            "EndBit": 21,
            "Name": "Joystick 7 Y-Axis Lever Forward Positive Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4608,
            "SPNLength": 2,
            "StartBit": 20,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4609: {
            "Acronym": "BJM7",
            "DataRange": "0 to 102 %",
            "EndBit": 31,
            "Name": "Joystick 7 Y-Axis Position",
            "Offset": 0.0,
            "OperationalHigh": 102.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 4609,
            "SPNLength": 10,
            "StartBit": 22,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "%"
        },
        4610: {
            "Acronym": "BJM7",
            "DataRange": "0 to 3",
            "EndBit": 37,
            "Name": "Joystick 7 Y-Axis Detent Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4610,
            "SPNLength": 2,
            "StartBit": 36,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4611: {
            "Acronym": "BJM7",
            "DataRange": "0 to 3",
            "EndBit": 39,
            "Name": "Joystick 7 X-Axis Detent Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4611,
            "SPNLength": 2,
            "StartBit": 38,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4612: {
            "Acronym": "BJM7",
            "DataRange": "0 to 3",
            "EndBit": 41,
            "Name": "Joystick 7 Button 4 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4612,
            "SPNLength": 2,
            "StartBit": 40,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4613: {
            "Acronym": "BJM7",
            "DataRange": "0 to 3",
            "EndBit": 43,
            "Name": "Joystick 7 Button 3 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4613,
            "SPNLength": 2,
            "StartBit": 42,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4614: {
            "Acronym": "BJM7",
            "DataRange": "0 to 3",
            "EndBit": 45,
            "Name": "Joystick 7 Button 2 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4614,
            "SPNLength": 2,
            "StartBit": 44,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4615: {
            "Acronym": "BJM7",
            "DataRange": "0 to 3",
            "EndBit": 47,
            "Name": "Joystick 7 Button 1 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4615,
            "SPNLength": 2,
            "StartBit": 46,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4616: {
            "Acronym": "BJM7",
            "DataRange": "0 to 3",
            "EndBit": 49,
            "Name": "Joystick 7 Button 8 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4616,
            "SPNLength": 2,
            "StartBit": 48,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4617: {
            "Acronym": "BJM7",
            "DataRange": "0 to 3",
            "EndBit": 51,
            "Name": "Joystick 7 Button 7 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4617,
            "SPNLength": 2,
            "StartBit": 50,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4618: {
            "Acronym": "BJM7",
            "DataRange": "0 to 3",
            "EndBit": 53,
            "Name": "Joystick 7 Button 6 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4618,
            "SPNLength": 2,
            "StartBit": 52,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4619: {
            "Acronym": "BJM7",
            "DataRange": "0 to 3",
            "EndBit": 55,
            "Name": "Joystick 7 Button 5 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4619,
            "SPNLength": 2,
            "StartBit": 54,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4620: {
            "Acronym": "BJM7",
            "DataRange": "0 to 3",
            "EndBit": 57,
            "Name": "Joystick 7 Button 12 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4620,
            "SPNLength": 2,
            "StartBit": 56,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4621: {
            "Acronym": "BJM7",
            "DataRange": "0 to 3",
            "EndBit": 59,
            "Name": "Joystick 7 Button 11 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4621,
            "SPNLength": 2,
            "StartBit": 58,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4622: {
            "Acronym": "BJM7",
            "DataRange": "0 to 3",
            "EndBit": 61,
            "Name": "Joystick 7 Button 10 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4622,
            "SPNLength": 2,
            "StartBit": 60,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4623: {
            "Acronym": "BJM7",
            "DataRange": "0 to 3",
            "EndBit": 63,
            "Name": "Joystick 7 Button 9 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4623,
            "SPNLength": 2,
            "StartBit": 62,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4624: {
            "Acronym": "EJM7",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Joystick 7 Grip X-Axis Neutral Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4624,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4625: {
            "Acronym": "EJM7",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Joystick 7 Grip X-Axis Lever Left Negative Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4625,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4626: {
            "Acronym": "EJM7",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Joystick 7 Grip X-Axis Lever Right Positive Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4626,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4627: {
            "Acronym": "EJM7",
            "DataRange": "0 to 102 %",
            "EndBit": 15,
            "Name": "Joystick 7 Grip X-Axis Position",
            "Offset": 0.0,
            "OperationalHigh": 102.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 4627,
            "SPNLength": 10,
            "StartBit": 6,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "%"
        },
        4628: {
            "Acronym": "EJM7",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Joystick 7 Grip Y-Axis Neutral Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4628,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4629: {
            "Acronym": "EJM7",
            "DataRange": "0 to 3",
            "EndBit": 19,
            "Name": "Joystick 7 Grip Y-Axis Lever Back Negative Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4629,
            "SPNLength": 2,
            "StartBit": 18,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4630: {
            "Acronym": "EJM7",
            "DataRange": "0 to 3",
            "EndBit": 21,
            "Name": "Joystick 7 Grip Y-Axis Lever Forward Positive Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4630,
            "SPNLength": 2,
            "StartBit": 20,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4631: {
            "Acronym": "EJM7",
            "DataRange": "0 to 102 %",
            "EndBit": 31,
            "Name": "Joystick 7 Grip Y-Axis Position",
            "Offset": 0.0,
            "OperationalHigh": 102.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 4631,
            "SPNLength": 10,
            "StartBit": 22,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "%"
        },
        4632: {
            "Acronym": "EJM7",
            "DataRange": "0 to 3",
            "EndBit": 33,
            "Name": "Joystick 7 Theta-Axis Neutral Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4632,
            "SPNLength": 2,
            "StartBit": 32,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4633: {
            "Acronym": "EJM7",
            "DataRange": "0 to 3",
            "EndBit": 35,
            "Name": "Joystick 7 Theta-Axis Counter Clockwise Negative Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4633,
            "SPNLength": 2,
            "StartBit": 34,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4634: {
            "Acronym": "EJM7",
            "DataRange": "0 to 3",
            "EndBit": 37,
            "Name": "Joystick 7 Theta-Axis Clockwise Positive Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4634,
            "SPNLength": 2,
            "StartBit": 36,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4635: {
            "Acronym": "EJM7",
            "DataRange": "0 to 102 %",
            "EndBit": 47,
            "Name": "Joystick 7 Theta-Axis Position",
            "Offset": 0.0,
            "OperationalHigh": 102.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 4635,
            "SPNLength": 10,
            "StartBit": 38,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "%"
        },
        4636: {
            "Acronym": "EJM7",
            "DataRange": "0 to 3",
            "EndBit": 51,
            "Name": "Joystick 7 Theta-Axis Detent Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4636,
            "SPNLength": 2,
            "StartBit": 50,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4637: {
            "Acronym": "EJM7",
            "DataRange": "0 to 3",
            "EndBit": 53,
            "Name": "Joystick 7 Grip Y-Axis Detent Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4637,
            "SPNLength": 2,
            "StartBit": 52,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4638: {
            "Acronym": "EJM7",
            "DataRange": "0 to 3",
            "EndBit": 55,
            "Name": "Joystick 7 Grip X-Axis Detent Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4638,
            "SPNLength": 2,
            "StartBit": 54,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4639: {
            "Acronym": "BJM8",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Joystick 8 X-Axis Neutral Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4639,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4640: {
            "Acronym": "BJM8",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Joystick 8 X-Axis Lever Left Negative Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4640,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4641: {
            "Acronym": "BJM8",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Joystick 8 X-Axis Lever Right Positive Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4641,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4642: {
            "Acronym": "BJM8",
            "DataRange": "0 to 102 %",
            "EndBit": 15,
            "Name": "Joystick 8 X-Axis Position",
            "Offset": 0.0,
            "OperationalHigh": 102.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 4642,
            "SPNLength": 10,
            "StartBit": 6,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "%"
        },
        4643: {
            "Acronym": "BJM8",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Joystick 8 Y-Axis Neutral Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4643,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4644: {
            "Acronym": "BJM8",
            "DataRange": "0 to 3",
            "EndBit": 19,
            "Name": "Joystick 8 Y-Axis Lever Back Negative Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4644,
            "SPNLength": 2,
            "StartBit": 18,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4645: {
            "Acronym": "BJM8",
            "DataRange": "0 to 3",
            "EndBit": 21,
            "Name": "Joystick 8 Y-Axis Lever Forward Positive Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4645,
            "SPNLength": 2,
            "StartBit": 20,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4646: {
            "Acronym": "BJM8",
            "DataRange": "0 to 102 %",
            "EndBit": 31,
            "Name": "Joystick 8 Y-Axis Position",
            "Offset": 0.0,
            "OperationalHigh": 102.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 4646,
            "SPNLength": 10,
            "StartBit": 22,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "%"
        },
        4647: {
            "Acronym": "BJM8",
            "DataRange": "0 to 3",
            "EndBit": 37,
            "Name": "Joystick 8 Y-Axis Detent Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4647,
            "SPNLength": 2,
            "StartBit": 36,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4648: {
            "Acronym": "BJM8",
            "DataRange": "0 to 3",
            "EndBit": 39,
            "Name": "Joystick 8 X-Axis Detent Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4648,
            "SPNLength": 2,
            "StartBit": 38,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4649: {
            "Acronym": "BJM8",
            "DataRange": "0 to 3",
            "EndBit": 41,
            "Name": "Joystick 8 Button 4 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4649,
            "SPNLength": 2,
            "StartBit": 40,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4650: {
            "Acronym": "BJM8",
            "DataRange": "0 to 3",
            "EndBit": 43,
            "Name": "Joystick 8 Button 3 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4650,
            "SPNLength": 2,
            "StartBit": 42,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4651: {
            "Acronym": "BJM8",
            "DataRange": "0 to 3",
            "EndBit": 45,
            "Name": "Joystick 8 Button 2 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4651,
            "SPNLength": 2,
            "StartBit": 44,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4652: {
            "Acronym": "BJM8",
            "DataRange": "0 to 3",
            "EndBit": 47,
            "Name": "Joystick 8 Button 1 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4652,
            "SPNLength": 2,
            "StartBit": 46,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4653: {
            "Acronym": "BJM8",
            "DataRange": "0 to 3",
            "EndBit": 49,
            "Name": "Joystick 8 Button 8 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4653,
            "SPNLength": 2,
            "StartBit": 48,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4654: {
            "Acronym": "BJM8",
            "DataRange": "0 to 3",
            "EndBit": 51,
            "Name": "Joystick 8 Button 7 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4654,
            "SPNLength": 2,
            "StartBit": 50,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4655: {
            "Acronym": "BJM8",
            "DataRange": "0 to 3",
            "EndBit": 53,
            "Name": "Joystick 8 Button 6 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4655,
            "SPNLength": 2,
            "StartBit": 52,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4656: {
            "Acronym": "BJM8",
            "DataRange": "0 to 3",
            "EndBit": 55,
            "Name": "Joystick 8 Button 5 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4656,
            "SPNLength": 2,
            "StartBit": 54,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4657: {
            "Acronym": "BJM8",
            "DataRange": "0 to 3",
            "EndBit": 57,
            "Name": "Joystick 8 Button 12 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4657,
            "SPNLength": 2,
            "StartBit": 56,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4658: {
            "Acronym": "BJM8",
            "DataRange": "0 to 3",
            "EndBit": 59,
            "Name": "Joystick 8 Button 11 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4658,
            "SPNLength": 2,
            "StartBit": 58,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4659: {
            "Acronym": "BJM8",
            "DataRange": "0 to 3",
            "EndBit": 61,
            "Name": "Joystick 8 Button 10 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4659,
            "SPNLength": 2,
            "StartBit": 60,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4660: {
            "Acronym": "BJM8",
            "DataRange": "0 to 3",
            "EndBit": 63,
            "Name": "Joystick 8 Button 9 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4660,
            "SPNLength": 2,
            "StartBit": 62,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4661: {
            "Acronym": "EJM8",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Joystick 8 Grip X-Axis Neutral Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4661,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4662: {
            "Acronym": "EJM8",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Joystick 8 Grip X-Axis Lever Left Negative Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4662,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4663: {
            "Acronym": "EJM8",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Joystick 8 Grip X-Axis Lever Right Positive Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4663,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4664: {
            "Acronym": "EJM8",
            "DataRange": "0 to 102 %",
            "EndBit": 15,
            "Name": "Joystick 8 Grip X-Axis Position",
            "Offset": 0.0,
            "OperationalHigh": 102.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 4664,
            "SPNLength": 10,
            "StartBit": 6,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "%"
        },
        4665: {
            "Acronym": "EJM8",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Joystick 8 Grip Y-Axis Neutral Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4665,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4666: {
            "Acronym": "EJM8",
            "DataRange": "0 to 3",
            "EndBit": 19,
            "Name": "Joystick 8 Grip Y-Axis Lever Back Negative Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4666,
            "SPNLength": 2,
            "StartBit": 18,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4667: {
            "Acronym": "EJM8",
            "DataRange": "0 to 3",
            "EndBit": 21,
            "Name": "Joystick 8 Grip Y-Axis Lever Forward Positive Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4667,
            "SPNLength": 2,
            "StartBit": 20,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4668: {
            "Acronym": "EJM8",
            "DataRange": "0 to 102 %",
            "EndBit": 31,
            "Name": "Joystick 8 Grip Y-Axis Position",
            "Offset": 0.0,
            "OperationalHigh": 102.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 4668,
            "SPNLength": 10,
            "StartBit": 22,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "%"
        },
        4669: {
            "Acronym": "EJM8",
            "DataRange": "0 to 3",
            "EndBit": 33,
            "Name": "Joystick 8 Theta-Axis Neutral Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4669,
            "SPNLength": 2,
            "StartBit": 32,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4670: {
            "Acronym": "EJM8",
            "DataRange": "0 to 3",
            "EndBit": 35,
            "Name": "Joystick 8 Theta-Axis Counter Clockwise Negative Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4670,
            "SPNLength": 2,
            "StartBit": 34,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4671: {
            "Acronym": "EJM8",
            "DataRange": "0 to 3",
            "EndBit": 37,
            "Name": "Joystick 8 Theta-Axis Clockwise Positive Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4671,
            "SPNLength": 2,
            "StartBit": 36,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4672: {
            "Acronym": "EJM8",
            "DataRange": "0 to 102 %",
            "EndBit": 47,
            "Name": "Joystick 8 Theta-Axis Position",
            "Offset": 0.0,
            "OperationalHigh": 102.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 4672,
            "SPNLength": 10,
            "StartBit": 38,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "%"
        },
        4673: {
            "Acronym": "EJM8",
            "DataRange": "0 to 3",
            "EndBit": 51,
            "Name": "Joystick 8 Theta-Axis Detent Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4673,
            "SPNLength": 2,
            "StartBit": 50,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4674: {
            "Acronym": "EJM8",
            "DataRange": "0 to 3",
            "EndBit": 53,
            "Name": "Joystick 8 Grip Y-Axis Detent Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4674,
            "SPNLength": 2,
            "StartBit": 52,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4675: {
            "Acronym": "EJM8",
            "DataRange": "0 to 3",
            "EndBit": 55,
            "Name": "Joystick 8 Grip X-Axis Detent Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4675,
            "SPNLength": 2,
            "StartBit": 54,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4676: {
            "Acronym": "BJM9",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Joystick 9 X-Axis Neutral Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4676,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4677: {
            "Acronym": "BJM9",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Joystick 9 X-Axis Lever Left Negative Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4677,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4678: {
            "Acronym": "BJM9",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Joystick 9 X-Axis Lever Right Positive Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4678,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4679: {
            "Acronym": "BJM9",
            "DataRange": "0 to 102 %",
            "EndBit": 15,
            "Name": "Joystick 9 X-Axis Position",
            "Offset": 0.0,
            "OperationalHigh": 102.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 4679,
            "SPNLength": 10,
            "StartBit": 6,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "%"
        },
        4680: {
            "Acronym": "BJM9",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Joystick 9 Y-Axis Neutral Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4680,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4681: {
            "Acronym": "BJM9",
            "DataRange": "0 to 3",
            "EndBit": 19,
            "Name": "Joystick 9 Y-Axis Lever Back Negative Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4681,
            "SPNLength": 2,
            "StartBit": 18,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4682: {
            "Acronym": "BJM9",
            "DataRange": "0 to 3",
            "EndBit": 21,
            "Name": "Joystick 9 Y-Axis Lever Forward Positive Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4682,
            "SPNLength": 2,
            "StartBit": 20,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4683: {
            "Acronym": "BJM9",
            "DataRange": "0 to 102 %",
            "EndBit": 31,
            "Name": "Joystick 9 Y-Axis Position",
            "Offset": 0.0,
            "OperationalHigh": 102.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 4683,
            "SPNLength": 10,
            "StartBit": 22,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "%"
        },
        4684: {
            "Acronym": "BJM9",
            "DataRange": "0 to 3",
            "EndBit": 37,
            "Name": "Joystick 9 Y-Axis Detent Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4684,
            "SPNLength": 2,
            "StartBit": 36,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4685: {
            "Acronym": "BJM9",
            "DataRange": "0 to 3",
            "EndBit": 39,
            "Name": "Joystick 9 X-Axis Detent Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4685,
            "SPNLength": 2,
            "StartBit": 38,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4686: {
            "Acronym": "BJM9",
            "DataRange": "0 to 3",
            "EndBit": 41,
            "Name": "Joystick 9 Button 4 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4686,
            "SPNLength": 2,
            "StartBit": 40,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4687: {
            "Acronym": "BJM9",
            "DataRange": "0 to 3",
            "EndBit": 43,
            "Name": "Joystick 9 Button 3 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4687,
            "SPNLength": 2,
            "StartBit": 42,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4688: {
            "Acronym": "BJM9",
            "DataRange": "0 to 3",
            "EndBit": 45,
            "Name": "Joystick 9 Button 2 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4688,
            "SPNLength": 2,
            "StartBit": 44,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4689: {
            "Acronym": "BJM9",
            "DataRange": "0 to 3",
            "EndBit": 47,
            "Name": "Joystick 9 Button 1 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4689,
            "SPNLength": 2,
            "StartBit": 46,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4690: {
            "Acronym": "BJM9",
            "DataRange": "0 to 3",
            "EndBit": 49,
            "Name": "Joystick 9 Button 8 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4690,
            "SPNLength": 2,
            "StartBit": 48,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4691: {
            "Acronym": "BJM9",
            "DataRange": "0 to 3",
            "EndBit": 51,
            "Name": "Joystick 9 Button 7 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4691,
            "SPNLength": 2,
            "StartBit": 50,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4692: {
            "Acronym": "BJM9",
            "DataRange": "0 to 3",
            "EndBit": 53,
            "Name": "Joystick 9 Button 6 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4692,
            "SPNLength": 2,
            "StartBit": 52,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4693: {
            "Acronym": "BJM9",
            "DataRange": "0 to 3",
            "EndBit": 55,
            "Name": "Joystick 9 Button 5 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4693,
            "SPNLength": 2,
            "StartBit": 54,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4694: {
            "Acronym": "BJM9",
            "DataRange": "0 to 3",
            "EndBit": 57,
            "Name": "Joystick 9 Button 12 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4694,
            "SPNLength": 2,
            "StartBit": 56,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4695: {
            "Acronym": "BJM9",
            "DataRange": "0 to 3",
            "EndBit": 59,
            "Name": "Joystick 9 Button 11 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4695,
            "SPNLength": 2,
            "StartBit": 58,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4696: {
            "Acronym": "BJM9",
            "DataRange": "0 to 3",
            "EndBit": 61,
            "Name": "Joystick 9 Button 10 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4696,
            "SPNLength": 2,
            "StartBit": 60,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4697: {
            "Acronym": "BJM9",
            "DataRange": "0 to 3",
            "EndBit": 63,
            "Name": "Joystick 9 Button 9 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4697,
            "SPNLength": 2,
            "StartBit": 62,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4698: {
            "Acronym": "EJM9",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Joystick 9 Grip X-Axis Neutral Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4698,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4699: {
            "Acronym": "EJM9",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Joystick 9 Grip X-Axis Lever Left Negative Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4699,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4700: {
            "Acronym": "EJM9",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Joystick 9 Grip X-Axis Lever Right Positive Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4700,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4701: {
            "Acronym": "EJM9",
            "DataRange": "0 to 102 %",
            "EndBit": 15,
            "Name": "Joystick 9 Grip X-Axis Position",
            "Offset": 0.0,
            "OperationalHigh": 102.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 4701,
            "SPNLength": 10,
            "StartBit": 6,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "%"
        },
        4702: {
            "Acronym": "EJM9",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Joystick 9 Grip Y-Axis Neutral Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4702,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4703: {
            "Acronym": "EJM9",
            "DataRange": "0 to 3",
            "EndBit": 19,
            "Name": "Joystick 9 Grip Y-Axis Lever Back Negative Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4703,
            "SPNLength": 2,
            "StartBit": 18,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4704: {
            "Acronym": "EJM9",
            "DataRange": "0 to 3",
            "EndBit": 21,
            "Name": "Joystick 9 Grip Y-Axis Lever Forward Positive Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4704,
            "SPNLength": 2,
            "StartBit": 20,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4705: {
            "Acronym": "EJM9",
            "DataRange": "0 to 102 %",
            "EndBit": 31,
            "Name": "Joystick 9 Grip Y-Axis Position",
            "Offset": 0.0,
            "OperationalHigh": 102.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 4705,
            "SPNLength": 10,
            "StartBit": 22,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "%"
        },
        4706: {
            "Acronym": "EJM9",
            "DataRange": "0 to 3",
            "EndBit": 33,
            "Name": "Joystick 9 Theta-Axis Neutral Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4706,
            "SPNLength": 2,
            "StartBit": 32,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4707: {
            "Acronym": "EJM9",
            "DataRange": "0 to 3",
            "EndBit": 35,
            "Name": "Joystick 9 Theta-Axis Counter Clockwise Negative Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4707,
            "SPNLength": 2,
            "StartBit": 34,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4708: {
            "Acronym": "EJM9",
            "DataRange": "0 to 3",
            "EndBit": 37,
            "Name": "Joystick 9 Theta-Axis Clockwise Positive Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4708,
            "SPNLength": 2,
            "StartBit": 36,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4709: {
            "Acronym": "EJM9",
            "DataRange": "0 to 102 %",
            "EndBit": 47,
            "Name": "Joystick 9 Theta-Axis Position",
            "Offset": 0.0,
            "OperationalHigh": 102.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 4709,
            "SPNLength": 10,
            "StartBit": 38,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "%"
        },
        4710: {
            "Acronym": "EJM9",
            "DataRange": "0 to 3",
            "EndBit": 51,
            "Name": "Joystick 9 Theta-Axis Detent Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4710,
            "SPNLength": 2,
            "StartBit": 50,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4711: {
            "Acronym": "EJM9",
            "DataRange": "0 to 3",
            "EndBit": 53,
            "Name": "Joystick 9 Grip Y-Axis Detent Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4711,
            "SPNLength": 2,
            "StartBit": 52,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4712: {
            "Acronym": "EJM9",
            "DataRange": "0 to 3",
            "EndBit": 55,
            "Name": "Joystick 9 Grip X-Axis Detent Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4712,
            "SPNLength": 2,
            "StartBit": 54,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4713: {
            "Acronym": "BJM10",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Joystick 10 X-Axis Neutral Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4713,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4714: {
            "Acronym": "BJM10",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Joystick 10 X-Axis Lever Left Negative Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4714,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4715: {
            "Acronym": "BJM10",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Joystick 10 X-Axis Lever Right Positive Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4715,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4716: {
            "Acronym": "BJM10",
            "DataRange": "0 to 102 %",
            "EndBit": 15,
            "Name": "Joystick 10 X-Axis Position",
            "Offset": 0.0,
            "OperationalHigh": 102.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 4716,
            "SPNLength": 10,
            "StartBit": 6,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "%"
        },
        4717: {
            "Acronym": "BJM10",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Joystick 10 Y-Axis Neutral Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4717,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4718: {
            "Acronym": "BJM10",
            "DataRange": "0 to 3",
            "EndBit": 19,
            "Name": "Joystick 10 Y-Axis Lever Back Negative Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4718,
            "SPNLength": 2,
            "StartBit": 18,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4719: {
            "Acronym": "BJM10",
            "DataRange": "0 to 3",
            "EndBit": 21,
            "Name": "Joystick 10 Y-Axis Lever Forward Positive Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4719,
            "SPNLength": 2,
            "StartBit": 20,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4720: {
            "Acronym": "BJM10",
            "DataRange": "0 to 102 %",
            "EndBit": 31,
            "Name": "Joystick 10 Y-Axis Position",
            "Offset": 0.0,
            "OperationalHigh": 102.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 4720,
            "SPNLength": 10,
            "StartBit": 22,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "%"
        },
        4721: {
            "Acronym": "BJM10",
            "DataRange": "0 to 3",
            "EndBit": 37,
            "Name": "Joystick 10 Y-Axis Detent Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4721,
            "SPNLength": 2,
            "StartBit": 36,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4722: {
            "Acronym": "BJM10",
            "DataRange": "0 to 3",
            "EndBit": 39,
            "Name": "Joystick 10 X-Axis Detent Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4722,
            "SPNLength": 2,
            "StartBit": 38,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4723: {
            "Acronym": "BJM10",
            "DataRange": "0 to 3",
            "EndBit": 41,
            "Name": "Joystick 10 Button 4 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4723,
            "SPNLength": 2,
            "StartBit": 40,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4724: {
            "Acronym": "BJM10",
            "DataRange": "0 to 3",
            "EndBit": 43,
            "Name": "Joystick 10 Button 3 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4724,
            "SPNLength": 2,
            "StartBit": 42,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4725: {
            "Acronym": "BJM10",
            "DataRange": "0 to 3",
            "EndBit": 45,
            "Name": "Joystick 10 Button 2 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4725,
            "SPNLength": 2,
            "StartBit": 44,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4726: {
            "Acronym": "BJM10",
            "DataRange": "0 to 3",
            "EndBit": 47,
            "Name": "Joystick 10 Button 1 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4726,
            "SPNLength": 2,
            "StartBit": 46,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4727: {
            "Acronym": "BJM10",
            "DataRange": "0 to 3",
            "EndBit": 49,
            "Name": "Joystick 10 Button 8 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4727,
            "SPNLength": 2,
            "StartBit": 48,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4728: {
            "Acronym": "BJM10",
            "DataRange": "0 to 3",
            "EndBit": 51,
            "Name": "Joystick 10 Button 7 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4728,
            "SPNLength": 2,
            "StartBit": 50,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4729: {
            "Acronym": "BJM10",
            "DataRange": "0 to 3",
            "EndBit": 53,
            "Name": "Joystick 10 Button 6 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4729,
            "SPNLength": 2,
            "StartBit": 52,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4730: {
            "Acronym": "BJM10",
            "DataRange": "0 to 3",
            "EndBit": 55,
            "Name": "Joystick 10 Button 5 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4730,
            "SPNLength": 2,
            "StartBit": 54,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4731: {
            "Acronym": "BJM10",
            "DataRange": "0 to 3",
            "EndBit": 57,
            "Name": "Joystick 10 Button 12 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4731,
            "SPNLength": 2,
            "StartBit": 56,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4732: {
            "Acronym": "BJM10",
            "DataRange": "0 to 3",
            "EndBit": 59,
            "Name": "Joystick 10 Button 11 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4732,
            "SPNLength": 2,
            "StartBit": 58,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4733: {
            "Acronym": "BJM10",
            "DataRange": "0 to 3",
            "EndBit": 61,
            "Name": "Joystick 10 Button 10 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4733,
            "SPNLength": 2,
            "StartBit": 60,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4734: {
            "Acronym": "BJM10",
            "DataRange": "0 to 3",
            "EndBit": 63,
            "Name": "Joystick 10 Button 9 Pressed Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4734,
            "SPNLength": 2,
            "StartBit": 62,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms. Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4735: {
            "Acronym": "EJM10",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Joystick 10 Grip X-Axis Neutral Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4735,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4736: {
            "Acronym": "EJM10",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Joystick 10 Grip X-Axis Lever Left Negative Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4736,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4737: {
            "Acronym": "EJM10",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Joystick 10 Grip X-Axis Lever Right Positive Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4737,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4738: {
            "Acronym": "EJM10",
            "DataRange": "0 to 102 %",
            "EndBit": 15,
            "Name": "Joystick 10 Grip X-Axis Position",
            "Offset": 0.0,
            "OperationalHigh": 102.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 4738,
            "SPNLength": 10,
            "StartBit": 6,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "%"
        },
        4739: {
            "Acronym": "EJM10",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Joystick 10 Grip Y-Axis Neutral Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4739,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4740: {
            "Acronym": "EJM10",
            "DataRange": "0 to 3",
            "EndBit": 19,
            "Name": "Joystick 10 Grip Y-Axis Lever Back Negative Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4740,
            "SPNLength": 2,
            "StartBit": 18,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4741: {
            "Acronym": "EJM10",
            "DataRange": "0 to 3",
            "EndBit": 21,
            "Name": "Joystick 10 Grip Y-Axis Lever Forward Positive Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4741,
            "SPNLength": 2,
            "StartBit": 20,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4742: {
            "Acronym": "EJM10",
            "DataRange": "0 to 102 %",
            "EndBit": 31,
            "Name": "Joystick 10 Grip Y-Axis Position",
            "Offset": 0.0,
            "OperationalHigh": 102.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 4742,
            "SPNLength": 10,
            "StartBit": 22,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "%"
        },
        4743: {
            "Acronym": "EJM10",
            "DataRange": "0 to 3",
            "EndBit": 33,
            "Name": "Joystick 10 Theta-Axis Neutral Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4743,
            "SPNLength": 2,
            "StartBit": 32,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4744: {
            "Acronym": "EJM10",
            "DataRange": "0 to 3",
            "EndBit": 35,
            "Name": "Joystick 10 Theta-Axis Counter Clockwise Negative Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4744,
            "SPNLength": 2,
            "StartBit": 34,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4745: {
            "Acronym": "EJM10",
            "DataRange": "0 to 3",
            "EndBit": 37,
            "Name": "Joystick 10 Theta-Axis Clockwise Positive Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4745,
            "SPNLength": 2,
            "StartBit": 36,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4746: {
            "Acronym": "EJM10",
            "DataRange": "0 to 102 %",
            "EndBit": 47,
            "Name": "Joystick 10 Theta-Axis Position",
            "Offset": 0.0,
            "OperationalHigh": 102.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 4746,
            "SPNLength": 10,
            "StartBit": 38,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "%"
        },
        4747: {
            "Acronym": "EJM10",
            "DataRange": "0 to 3",
            "EndBit": 51,
            "Name": "Joystick 10 Theta-Axis Detent Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4747,
            "SPNLength": 2,
            "StartBit": 50,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4748: {
            "Acronym": "EJM10",
            "DataRange": "0 to 3",
            "EndBit": 53,
            "Name": "Joystick 10 Grip Y-Axis Detent Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4748,
            "SPNLength": 2,
            "StartBit": 52,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4749: {
            "Acronym": "EJM10",
            "DataRange": "0 to 3",
            "EndBit": 55,
            "Name": "Joystick 10 Grip X-Axis Detent Position Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4749,
            "SPNLength": 2,
            "StartBit": 54,
            "TransmissionRate": "Every 100 ms and on change of state, but no faster than every 20 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: Every 100 ms or on change of state, but no faster than every 20 ms",
            "Units": "bit"
        },
        4750: {
            "Acronym": "EEC8",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 31,
            "Name": "Engine Exhaust Gas Recirculation 1 (EGR1) Cooler Intake Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 4750,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "deg F"
        },
        4751: {
            "Acronym": "EEC8",
            "DataRange": "0.0 to 4659.699927595 psi",
            "EndBit": 47,
            "Name": "Engine Exhaust Gas Recirculation 1 (EGR1) Cooler Intake Gas Absolute Pressure",
            "Offset": 0.0,
            "OperationalHigh": 4659.699927595,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.072518869,
            "SPN": 4751,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "psi"
        },
        4752: {
            "Acronym": "EEC8",
            "DataRange": "0 to 100 %",
            "EndBit": 55,
            "Name": "Engine Exhaust Gas Recirculation 1 (EGR1) Cooler Efficiency",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 4752,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        4753: {
            "Acronym": "A1GOC",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 15,
            "Name": "Aftertreatment 1 Gas Oxidation Catalyst Intake Gas Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 4753,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "500 ms",
            "Units": "deg F"
        },
        4754: {
            "Acronym": "A1GOC",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 31,
            "Name": "Aftertreatment 1 Gas Oxidation Catalyst Outlet Gas Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 4754,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "500 ms",
            "Units": "deg F"
        },
        4755: {
            "Acronym": "A1GOC",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 47,
            "Name": "Aftertreatment 1 Gas Oxidation Catalyst Differential Pressure",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 4755,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        4756: {
            "Acronym": "A1GOC",
            "DataRange": "0 to 31",
            "EndBit": 52,
            "Name": "Aftertreatment 1 Gas Oxidation Catalyst Intake Gas Temperature Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 4756,
            "SPNLength": 5,
            "StartBit": 48,
            "TransmissionRate": "500 ms",
            "Units": "binary"
        },
        4757: {
            "Acronym": "A1GOC",
            "DataRange": "0 to 31",
            "EndBit": 56,
            "Name": "Aftertreatment 1 Gas Oxidation Catalyst Outlet Gas Temperature Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 4757,
            "SPNLength": 5,
            "StartBit": 53,
            "TransmissionRate": "500 ms",
            "Units": "binary"
        },
        4758: {
            "Acronym": "A1GOC",
            "DataRange": "0 to 31",
            "EndBit": 62,
            "Name": "Aftertreatment 1 Gas Oxidation Catalyst Differential Pressure Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 4758,
            "SPNLength": 5,
            "StartBit": 58,
            "TransmissionRate": "500 ms",
            "Units": "binary"
        },
        4759: {
            "Acronym": "A2GOC",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 15,
            "Name": "Aftertreatment 2 Gas Oxidation Catalyst Intake Gas Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 4759,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "500 ms",
            "Units": "deg F"
        },
        4760: {
            "Acronym": "A2GOC",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 31,
            "Name": "Aftertreatment 2 Gas Oxidation Catalyst Outlet Gas Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 4760,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "500 ms",
            "Units": "deg F"
        },
        4761: {
            "Acronym": "A2GOC",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 47,
            "Name": "Aftertreatment 2 Gas Oxidation Catalyst Differential Pressure",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 4761,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        4762: {
            "Acronym": "A2GOC",
            "DataRange": "0 to 31",
            "EndBit": 52,
            "Name": "Aftertreatment 2 Gas Oxidation Catalyst Intake Gas Temperature Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 4762,
            "SPNLength": 5,
            "StartBit": 48,
            "TransmissionRate": "500 ms",
            "Units": "binary"
        },
        4763: {
            "Acronym": "A2GOC",
            "DataRange": "0 to 31",
            "EndBit": 56,
            "Name": "Aftertreatment 2 Gas Oxidation Catalyst Outlet Gas Temperature Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 4763,
            "SPNLength": 5,
            "StartBit": 53,
            "TransmissionRate": "500 ms",
            "Units": "binary"
        },
        4764: {
            "Acronym": "A2GOC",
            "DataRange": "0 to 31",
            "EndBit": 62,
            "Name": "Aftertreatment 2 Gas Oxidation Catalyst Differential Pressure Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 4764,
            "SPNLength": 5,
            "StartBit": 58,
            "TransmissionRate": "500 ms",
            "Units": "binary"
        },
        4765: {
            "Acronym": "A1DOC",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 15,
            "Name": "Aftertreatment 1 Diesel Oxidation Catalyst Intake Gas Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 4765,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "500 ms",
            "Units": "deg F"
        },
        4766: {
            "Acronym": "A1DOC",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 31,
            "Name": "Aftertreatment 1 Diesel Oxidation Catalyst Outlet Gas Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 4766,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "500 ms",
            "Units": "deg F"
        },
        4767: {
            "Acronym": "A1DOC",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 47,
            "Name": "Aftertreatment 1 Diesel Oxidation Catalyst Differential Pressure",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 4767,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        4768: {
            "Acronym": "A1DOC",
            "DataRange": "0 to 31",
            "EndBit": 52,
            "Name": "Aftertreatment 1 Diesel Oxidation Catalyst Intake Gas Temperature Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 4768,
            "SPNLength": 5,
            "StartBit": 48,
            "TransmissionRate": "500 ms",
            "Units": "binary"
        },
        4769: {
            "Acronym": "A1DOC",
            "DataRange": "0 to 31",
            "EndBit": 56,
            "Name": "Aftertreatment 1 Diesel Oxidation Catalyst Outlet Gas Temperature Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 4769,
            "SPNLength": 5,
            "StartBit": 53,
            "TransmissionRate": "500 ms",
            "Units": "binary"
        },
        4770: {
            "Acronym": "A1DOC",
            "DataRange": "0 to 31",
            "EndBit": 62,
            "Name": "Aftertreatment 1 Diesel Oxidation Catalyst Differential Pressure Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 4770,
            "SPNLength": 5,
            "StartBit": 58,
            "TransmissionRate": "500 ms",
            "Units": "binary"
        },
        4771: {
            "Acronym": "A2DOC",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 15,
            "Name": "Aftertreatment 2 Diesel Oxidation Catalyst Intake Gas Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 4771,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "500 ms",
            "Units": "deg F"
        },
        4772: {
            "Acronym": "A2DOC",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 31,
            "Name": "Aftertreatment 2 Diesel Oxidation Catalyst Outlet Gas Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 4772,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "500 ms",
            "Units": "deg F"
        },
        4773: {
            "Acronym": "A2DOC",
            "DataRange": "0.0 to 4659.699927595 psi",
            "EndBit": 47,
            "Name": "Aftertreatment 2 Diesel Oxidation Catalyst Differential Pressure",
            "Offset": 0.0,
            "OperationalHigh": 4659.699927595,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.072518869,
            "SPN": 4773,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        4774: {
            "Acronym": "A2DOC",
            "DataRange": "0 to 31",
            "EndBit": 52,
            "Name": "Aftertreatment 2 Diesel Oxidation Catalyst Intake Gas Temperature Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 4774,
            "SPNLength": 5,
            "StartBit": 48,
            "TransmissionRate": "500 ms",
            "Units": "binary"
        },
        4775: {
            "Acronym": "A2DOC",
            "DataRange": "0 to 31",
            "EndBit": 56,
            "Name": "Aftertreatment 2 Diesel Oxidation Catalyst Outlet Gas Temperature Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 4775,
            "SPNLength": 5,
            "StartBit": 53,
            "TransmissionRate": "500 ms",
            "Units": "binary"
        },
        4776: {
            "Acronym": "A2DOC",
            "DataRange": "0 to 31",
            "EndBit": 62,
            "Name": "Aftertreatment 2 Diesel Oxidation Catalyst Differential Pressure Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 4776,
            "SPNLength": 5,
            "StartBit": 58,
            "TransmissionRate": "500 ms",
            "Units": "binary"
        },
        4777: {
            "Acronym": "ATDT1",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 15,
            "Name": "Aftertreatment 1 Gas Oxidation Catalyst Differential Gas Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 4777,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "deg F"
        },
        4778: {
            "Acronym": "ATDT1",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 31,
            "Name": "Aftertreatment 2 Gas Oxidation Catalyst Differential Gas Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 4778,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "deg F"
        },
        4779: {
            "Acronym": "ATDT2",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 15,
            "Name": "Aftertreatment 1 Three Way Catalyst Differential Gas Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 4779,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "deg F"
        },
        4780: {
            "Acronym": "ATDT2",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 31,
            "Name": "Aftertreatment 2 Three Way Catalyst Differential Gas Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 4780,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "deg F"
        },
        4781: {
            "Acronym": "DPF1S",
            "DataRange": "0 to 1000 g",
            "EndBit": 7,
            "Name": "DPF 1 Soot Mass",
            "Offset": 0.0,
            "OperationalHigh": 1000.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 4,
            "SPN": 4781,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "1 sec",
            "Units": "g"
        },
        4782: {
            "Acronym": "DPF1S",
            "DataRange": "0 to 20.0 g/L",
            "EndBit": 15,
            "Name": "DPF 1 Soot Density",
            "Offset": 0.0,
            "OperationalHigh": 20.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.08,
            "SPN": 4782,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "1 sec",
            "Units": "g/L"
        },
        4783: {
            "Acronym": "DPF1S",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 31,
            "Name": "DPF 1 Mean Soot Signal",
            "Offset": 0.0,
            "OperationalHigh": 160.6375,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 4783,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "1 sec",
            "Units": "%"
        },
        4784: {
            "Acronym": "DPF1S",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 47,
            "Name": "DPF 1 Median Soot Signal",
            "Offset": 0.0,
            "OperationalHigh": 160.6375,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 4784,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "1 sec",
            "Units": "%"
        },
        4785: {
            "Acronym": "DPF1S",
            "DataRange": "0 to 31",
            "EndBit": 52,
            "Name": "DPF 1 Soot Sensor Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 4785,
            "SPNLength": 5,
            "StartBit": 48,
            "TransmissionRate": "1 sec",
            "Units": "binary"
        },
        4786: {
            "Acronym": "DPF2S",
            "DataRange": "0 to 1000 g",
            "EndBit": 7,
            "Name": "DPF 2 Soot Mass",
            "Offset": 0.0,
            "OperationalHigh": 1000.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 4,
            "SPN": 4786,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "1 sec",
            "Units": "g"
        },
        4787: {
            "Acronym": "DPF2S",
            "DataRange": "0 to 20.0 g/L",
            "EndBit": 15,
            "Name": "DPF 2 Soot Density",
            "Offset": 0.0,
            "OperationalHigh": 20.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.08,
            "SPN": 4787,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "1 sec",
            "Units": "g/L"
        },
        4788: {
            "Acronym": "DPF2S",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 31,
            "Name": "DPF 2 Mean Soot Signal",
            "Offset": 0.0,
            "OperationalHigh": 160.6375,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 4788,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "1 sec",
            "Units": "%"
        },
        4789: {
            "Acronym": "DPF2S",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 47,
            "Name": "DPF 2 Median Soot Signal",
            "Offset": 0.0,
            "OperationalHigh": 160.6375,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 4789,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "1 sec",
            "Units": "%"
        },
        4790: {
            "Acronym": "DPF2S",
            "DataRange": "0 to 31",
            "EndBit": 52,
            "Name": "DPF 2 Soot Sensor Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 4790,
            "SPNLength": 5,
            "StartBit": 48,
            "TransmissionRate": "1 sec",
            "Units": "binary"
        },
        4809: {
            "Acronym": "AT1WUDOC",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 15,
            "Name": "Aftertreatment 1 Warm Up Diesel Oxidation Catalyst Intake Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 4809,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "500 ms",
            "Units": "deg F"
        },
        4810: {
            "Acronym": "AT1WUDOC",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 31,
            "Name": "Aftertreatment 1 Warm Up Diesel Oxidation Catalyst Outlet Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 4810,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "500 ms",
            "Units": "deg F"
        },
        4816: {
            "Acronym": "ETC1",
            "DataRange": "0 to 3",
            "EndBit": 7,
            "Name": "Transmission Torque Converter Lockup Transition in Process",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4816,
            "SPNLength": 2,
            "StartBit": 6,
            "TransmissionRate": "10 ms",
            "Units": "bit"
        },
        4817: {
            "Acronym": "IC2",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 55,
            "Name": "Engine Intake Manifold #1 Absolute Pressure (High Resolution)",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 4817,
            "SPNLength": 16,
            "StartBit": 40,
            "TransmissionRate": "0.5 s",
            "Units": "psi"
        },
        4818: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 15,
            "Name": "Driver Airbag Ignitor Loop 1st Stage - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4818,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4819: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 31,
            "Name": "Passenger Airbag Ignitor Loop 1st Stage - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4819,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4820: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 47,
            "Name": "Driver Airbag Ignitor Loop 2nd Stage - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4820,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4821: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 63,
            "Name": "Passenger Airbag Ignitor Loop 2nd Stage - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4821,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4822: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 79,
            "Name": "Driver Belt Tensioner Ignitior Loop - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4822,
            "SPNLength": 16,
            "StartBit": 64,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4823: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 95,
            "Name": "Passenger Belt Tensioner Ignitor Loop - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4823,
            "SPNLength": 16,
            "StartBit": 80,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4824: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 111,
            "Name": "Side Bag Ignitor Loop 1 - Left - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4824,
            "SPNLength": 16,
            "StartBit": 96,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4825: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 127,
            "Name": "Side Bag Ignitor Loop 2 - Left - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4825,
            "SPNLength": 16,
            "StartBit": 112,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4826: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 143,
            "Name": "Side Bag Ignitor Loop 1 - Right - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4826,
            "SPNLength": 16,
            "StartBit": 128,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4827: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 159,
            "Name": "Side Bag Ignitor Loop 2 - Right - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4827,
            "SPNLength": 16,
            "StartBit": 144,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4828: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 175,
            "Name": "Special Ignitor Loop 1 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4828,
            "SPNLength": 16,
            "StartBit": 160,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4829: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 191,
            "Name": "Special Ignitor Loop 2 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4829,
            "SPNLength": 16,
            "StartBit": 176,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4830: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 207,
            "Name": "Special Ignitor Loop 3 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4830,
            "SPNLength": 16,
            "StartBit": 192,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4831: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 223,
            "Name": "Special Ignitor Loop 4 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4831,
            "SPNLength": 16,
            "StartBit": 208,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4832: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 239,
            "Name": "Special Ignitor Loop 5 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4832,
            "SPNLength": 16,
            "StartBit": 224,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4833: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 255,
            "Name": "Special Ignitor Loop 6 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4833,
            "SPNLength": 16,
            "StartBit": 240,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4834: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 271,
            "Name": "Special Ignitor Loop 7 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4834,
            "SPNLength": 16,
            "StartBit": 256,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4835: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 287,
            "Name": "Special Ignitor Loop 8 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4835,
            "SPNLength": 16,
            "StartBit": 272,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4836: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 303,
            "Name": "Special Ignitor Loop 9 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4836,
            "SPNLength": 16,
            "StartBit": 288,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4837: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 319,
            "Name": "Special Ignitor Loop 10 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4837,
            "SPNLength": 16,
            "StartBit": 304,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4838: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 335,
            "Name": "Special Ignitor Loop 11 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4838,
            "SPNLength": 16,
            "StartBit": 320,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4839: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 351,
            "Name": "Special Ignitor Loop 12 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4839,
            "SPNLength": 16,
            "StartBit": 336,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4840: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 367,
            "Name": "Special Ignitor Loop 13 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4840,
            "SPNLength": 16,
            "StartBit": 352,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4841: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 383,
            "Name": "Special Ignitor Loop 14 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4841,
            "SPNLength": 16,
            "StartBit": 368,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4842: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 399,
            "Name": "Special Ignitor Loop 15 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4842,
            "SPNLength": 16,
            "StartBit": 384,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4843: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 415,
            "Name": "Special Ignitor Loop 16 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4843,
            "SPNLength": 16,
            "StartBit": 400,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4844: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 431,
            "Name": "Special Ignitor Loop 17 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4844,
            "SPNLength": 16,
            "StartBit": 416,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4845: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 447,
            "Name": "Special Ignitor Loop 18 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4845,
            "SPNLength": 16,
            "StartBit": 432,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4846: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 463,
            "Name": "Special Ignitor Loop 19 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4846,
            "SPNLength": 16,
            "StartBit": 448,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4847: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 479,
            "Name": "Special Ignitor Loop 20 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4847,
            "SPNLength": 16,
            "StartBit": 464,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4848: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 495,
            "Name": "Special Ignitor Loop 21 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4848,
            "SPNLength": 16,
            "StartBit": 480,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4849: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 511,
            "Name": "Special Ignitor Loop 22 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4849,
            "SPNLength": 16,
            "StartBit": 496,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4850: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 527,
            "Name": "Special Ignitor Loop 23 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4850,
            "SPNLength": 16,
            "StartBit": 512,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4851: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 543,
            "Name": "Special Ignitor Loop 24 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4851,
            "SPNLength": 16,
            "StartBit": 528,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4852: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 559,
            "Name": "Special Ignitor Loop 25 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4852,
            "SPNLength": 16,
            "StartBit": 544,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4853: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 575,
            "Name": "Special Ignitor Loop 26 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4853,
            "SPNLength": 16,
            "StartBit": 560,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4854: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 591,
            "Name": "Special Ignitor Loop 27 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4854,
            "SPNLength": 16,
            "StartBit": 576,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4855: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 607,
            "Name": "Special Ignitor Loop 28 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4855,
            "SPNLength": 16,
            "StartBit": 592,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4856: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 623,
            "Name": "Special Ignitor Loop 29 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4856,
            "SPNLength": 16,
            "StartBit": 608,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4857: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 639,
            "Name": "Special Ignitor Loop 30 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4857,
            "SPNLength": 16,
            "StartBit": 624,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4858: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 655,
            "Name": "Special Ignitor Loop 31 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4858,
            "SPNLength": 16,
            "StartBit": 640,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4859: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 671,
            "Name": "Special Ignitor Loop 32 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4859,
            "SPNLength": 16,
            "StartBit": 656,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4860: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 687,
            "Name": "Special Ignitor Loop 33 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4860,
            "SPNLength": 16,
            "StartBit": 672,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4861: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 703,
            "Name": "Special Ignitor Loop 34 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4861,
            "SPNLength": 16,
            "StartBit": 688,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4862: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 719,
            "Name": "Special Ignitor Loop 35 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4862,
            "SPNLength": 16,
            "StartBit": 704,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4863: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 735,
            "Name": "Special Ignitor Loop 36 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4863,
            "SPNLength": 16,
            "StartBit": 720,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4864: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 751,
            "Name": "Special Ignitor Loop 37 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4864,
            "SPNLength": 16,
            "StartBit": 736,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4865: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 767,
            "Name": "Special Ignitor Loop 38 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4865,
            "SPNLength": 16,
            "StartBit": 752,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4866: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 783,
            "Name": "Special Ignitor Loop 39 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4866,
            "SPNLength": 16,
            "StartBit": 768,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4867: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 799,
            "Name": "Special Ignitor Loop 40 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4867,
            "SPNLength": 16,
            "StartBit": 784,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4868: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 815,
            "Name": "Special Ignitor Loop 41 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4868,
            "SPNLength": 16,
            "StartBit": 800,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4869: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 831,
            "Name": "Special Ignitor Loop 42 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4869,
            "SPNLength": 16,
            "StartBit": 816,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4870: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 847,
            "Name": "Special Ignitor Loop 43 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4870,
            "SPNLength": 16,
            "StartBit": 832,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4871: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 863,
            "Name": "Special Ignitor Loop 44 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4871,
            "SPNLength": 16,
            "StartBit": 848,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4872: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 879,
            "Name": "Special Ignitor Loop 45 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4872,
            "SPNLength": 16,
            "StartBit": 864,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4873: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 895,
            "Name": "Special Ignitor Loop 46 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4873,
            "SPNLength": 16,
            "StartBit": 880,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4874: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 911,
            "Name": "Special Ignitor Loop 47 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4874,
            "SPNLength": 16,
            "StartBit": 896,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4875: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 927,
            "Name": "Special Ignitor Loop 48 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4875,
            "SPNLength": 16,
            "StartBit": 912,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4876: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 943,
            "Name": "Special Ignitor Loop 49 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4876,
            "SPNLength": 16,
            "StartBit": 928,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4877: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 959,
            "Name": "Special Ignitor Loop 50 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4877,
            "SPNLength": 16,
            "StartBit": 944,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4878: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 975,
            "Name": "Special Ignitor Loop 51 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4878,
            "SPNLength": 16,
            "StartBit": 960,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4879: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 991,
            "Name": "Special Ignitor Loop 52 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4879,
            "SPNLength": 16,
            "StartBit": 976,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4880: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 1007,
            "Name": "Special Ignitor Loop 53 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4880,
            "SPNLength": 16,
            "StartBit": 992,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4881: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 1023,
            "Name": "Special Ignitor Loop 54 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4881,
            "SPNLength": 16,
            "StartBit": 1008,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4882: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 1039,
            "Name": "Special Ignitor Loop 55 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4882,
            "SPNLength": 16,
            "StartBit": 1024,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4883: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 1055,
            "Name": "Special Ignitor Loop 56 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4883,
            "SPNLength": 16,
            "StartBit": 1040,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4884: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 1071,
            "Name": "Special Ignitor Loop 57 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4884,
            "SPNLength": 16,
            "StartBit": 1056,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4885: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 1087,
            "Name": "Special Ignitor Loop 58 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4885,
            "SPNLength": 16,
            "StartBit": 1072,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4886: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 1103,
            "Name": "Special Ignitor Loop 59 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4886,
            "SPNLength": 16,
            "StartBit": 1088,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4887: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 1119,
            "Name": "Special Ignitor Loop 60 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4887,
            "SPNLength": 16,
            "StartBit": 1104,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4888: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 1135,
            "Name": "Special Ignitor Loop 61 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4888,
            "SPNLength": 16,
            "StartBit": 1120,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4889: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 1151,
            "Name": "Special Ignitor Loop 62 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4889,
            "SPNLength": 16,
            "StartBit": 1136,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4890: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 1167,
            "Name": "Special Ignitor Loop 63 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4890,
            "SPNLength": 16,
            "StartBit": 1152,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4891: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 1183,
            "Name": "Special Ignitor Loop 64 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4891,
            "SPNLength": 16,
            "StartBit": 1168,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4892: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 1199,
            "Name": "Special Ignitor Loop 65 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4892,
            "SPNLength": 16,
            "StartBit": 1184,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4893: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 1215,
            "Name": "Special Ignitor Loop 66 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4893,
            "SPNLength": 16,
            "StartBit": 1200,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4894: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 1231,
            "Name": "Special Ignitor Loop 67 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4894,
            "SPNLength": 16,
            "StartBit": 1216,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4895: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 1247,
            "Name": "Special Ignitor Loop 68 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4895,
            "SPNLength": 16,
            "StartBit": 1232,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4896: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 1263,
            "Name": "Special Ignitor Loop 69 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4896,
            "SPNLength": 16,
            "StartBit": 1248,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4897: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 1279,
            "Name": "Special Ignitor Loop 70 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4897,
            "SPNLength": 16,
            "StartBit": 1264,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4898: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 1295,
            "Name": "Special Ignitor Loop 71 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4898,
            "SPNLength": 16,
            "StartBit": 1280,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4899: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 1311,
            "Name": "Special Ignitor Loop 72 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4899,
            "SPNLength": 16,
            "StartBit": 1296,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4900: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 1327,
            "Name": "Special Ignitor Loop 73 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4900,
            "SPNLength": 16,
            "StartBit": 1312,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4901: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 1343,
            "Name": "Special Ignitor Loop 74 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4901,
            "SPNLength": 16,
            "StartBit": 1328,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4902: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 1359,
            "Name": "Special Ignitor Loop 75 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4902,
            "SPNLength": 16,
            "StartBit": 1344,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4903: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 1375,
            "Name": "Special Ignitor Loop 76 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4903,
            "SPNLength": 16,
            "StartBit": 1360,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4904: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 1391,
            "Name": "Special Ignitor Loop 77 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4904,
            "SPNLength": 16,
            "StartBit": 1376,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4905: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 1407,
            "Name": "Special Ignitor Loop 78 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4905,
            "SPNLength": 16,
            "StartBit": 1392,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4906: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 1423,
            "Name": "Special Ignitor Loop 79 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4906,
            "SPNLength": 16,
            "StartBit": 1408,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4907: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 1439,
            "Name": "Special Ignitor Loop 80 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4907,
            "SPNLength": 16,
            "StartBit": 1424,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4908: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 1455,
            "Name": "Special Ignitor Loop 81 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4908,
            "SPNLength": 16,
            "StartBit": 1440,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4909: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 1471,
            "Name": "Special Ignitor Loop 82 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4909,
            "SPNLength": 16,
            "StartBit": 1456,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4910: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 1487,
            "Name": "Special Ignitor Loop 83 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4910,
            "SPNLength": 16,
            "StartBit": 1472,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4911: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 1503,
            "Name": "Special Ignitor Loop 84 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4911,
            "SPNLength": 16,
            "StartBit": 1488,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4912: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 1519,
            "Name": "Special Ignitor Loop 85 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4912,
            "SPNLength": 16,
            "StartBit": 1504,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4913: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 1535,
            "Name": "Special Ignitor Loop 86 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4913,
            "SPNLength": 16,
            "StartBit": 1520,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4914: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 1551,
            "Name": "Special Ignitor Loop 87 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4914,
            "SPNLength": 16,
            "StartBit": 1536,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4915: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 1567,
            "Name": "Special Ignitor Loop 88 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4915,
            "SPNLength": 16,
            "StartBit": 1552,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4916: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 1583,
            "Name": "Special Ignitor Loop 89 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4916,
            "SPNLength": 16,
            "StartBit": 1568,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4917: {
            "Acronym": "ILI",
            "DataRange": "0 to 6425.5 ohm",
            "EndBit": 1599,
            "Name": "Special Ignitor Loop 90 - Resistance",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.1,
            "SPN": 4917,
            "SPNLength": 16,
            "StartBit": 1584,
            "TransmissionRate": "On request",
            "Units": "ohm"
        },
        4918: {
            "Acronym": "CSI",
            "DataRange": "0 to 15",
            "EndBit": 3,
            "Name": "Frontal Collision Sensor 1 - Type",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 4918,
            "SPNLength": 4,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        4919: {
            "Acronym": "CSI",
            "DataRange": "0 to 15",
            "EndBit": 7,
            "Name": "Frontal Collision Sensor 2 - Type",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 4919,
            "SPNLength": 4,
            "StartBit": 4,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        4920: {
            "Acronym": "CSI",
            "DataRange": "0 to 15",
            "EndBit": 11,
            "Name": "Frontal Collision Sensor 3 - Type",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 4920,
            "SPNLength": 4,
            "StartBit": 8,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        4921: {
            "Acronym": "CSI",
            "DataRange": "0 to 15",
            "EndBit": 15,
            "Name": "Frontal Collision Sensor 4 - Type",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 4921,
            "SPNLength": 4,
            "StartBit": 12,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        4922: {
            "Acronym": "CSI",
            "DataRange": "0 to 15",
            "EndBit": 19,
            "Name": "Side Collision Sensor Front Left - Type",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 4922,
            "SPNLength": 4,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        4923: {
            "Acronym": "CSI",
            "DataRange": "0 to 15",
            "EndBit": 23,
            "Name": "Side Collision Sensor Mid Front Left - Type",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 4923,
            "SPNLength": 4,
            "StartBit": 20,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        4924: {
            "Acronym": "CSI",
            "DataRange": "0 to 15",
            "EndBit": 27,
            "Name": "Side Collision Sensor Mid Rear Left - Type",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 4924,
            "SPNLength": 4,
            "StartBit": 24,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        4925: {
            "Acronym": "CSI",
            "DataRange": "0 to 15",
            "EndBit": 31,
            "Name": "Side Collision Sensor Rear Left - Type",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 4925,
            "SPNLength": 4,
            "StartBit": 28,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        4926: {
            "Acronym": "CSI",
            "DataRange": "0 to 15",
            "EndBit": 35,
            "Name": "Side Collision Sensor Front Right - Type",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 4926,
            "SPNLength": 4,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        4927: {
            "Acronym": "CSI",
            "DataRange": "0 to 15",
            "EndBit": 39,
            "Name": "Side Collision Sensor Mid Front Right - Type",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 4927,
            "SPNLength": 4,
            "StartBit": 36,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        4928: {
            "Acronym": "CSI",
            "DataRange": "0 to 15",
            "EndBit": 43,
            "Name": "Side Collision Sensor Mid Rear Right - Type",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 4928,
            "SPNLength": 4,
            "StartBit": 40,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        4929: {
            "Acronym": "CSI",
            "DataRange": "0 to 15",
            "EndBit": 47,
            "Name": "Side Collision Sensor Rear Right - Type",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 4929,
            "SPNLength": 4,
            "StartBit": 44,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        4930: {
            "Acronym": "CSI",
            "DataRange": "0 to 15",
            "EndBit": 51,
            "Name": "Rear Collision Sensor 1 - Type",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 4930,
            "SPNLength": 4,
            "StartBit": 48,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        4931: {
            "Acronym": "CSI",
            "DataRange": "0 to 15",
            "EndBit": 55,
            "Name": "Rear Collision Sensor 2 - Type",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 4931,
            "SPNLength": 4,
            "StartBit": 52,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        4932: {
            "Acronym": "CSI",
            "DataRange": "0 to 15",
            "EndBit": 59,
            "Name": "Rear Collision Sensor 3 - Type",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 4932,
            "SPNLength": 4,
            "StartBit": 56,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        4933: {
            "Acronym": "CSI",
            "DataRange": "0 to 15",
            "EndBit": 63,
            "Name": "Rear Collision Sensor 4 - Type",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 4933,
            "SPNLength": 4,
            "StartBit": 60,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        4934: {
            "Acronym": "CSI",
            "DataRange": "0 to 15",
            "EndBit": 67,
            "Name": "Rollover Sensor - Type",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 4934,
            "SPNLength": 4,
            "StartBit": 64,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        4935: {
            "Acronym": "CSI",
            "DataRange": "0 to 255 per byte",
            "EndBit": 0,
            "Name": "Frontal Collision Sensor 1 - Serial Number",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": -2,
            "SPN": 4935,
            "SPNLength": 256,
            "StartBit": -1,
            "TransmissionRate": "On request",
            "Units": "ASCII"
        },
        4936: {
            "Acronym": "CSI",
            "DataRange": "0 to 255 per byte",
            "EndBit": -2,
            "Name": "Frontal Collision Sensor 2 - Serial Number",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": -2,
            "SPN": 4936,
            "SPNLength": 256,
            "StartBit": -1,
            "TransmissionRate": "On request",
            "Units": "ASCII"
        },
        4937: {
            "Acronym": "CSI",
            "DataRange": "0 to 255 per byte",
            "EndBit": -3,
            "Name": "Frontal Collision Sensor 3 - Serial Number",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": -2,
            "SPN": 4937,
            "SPNLength": 256,
            "StartBit": -1,
            "TransmissionRate": "On request",
            "Units": "ASCII"
        },
        4938: {
            "Acronym": "CSI",
            "DataRange": "0 to 255 per byte",
            "EndBit": -4,
            "Name": "Frontal Collision Sensor 4 - Serial Number",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": -2,
            "SPN": 4938,
            "SPNLength": 256,
            "StartBit": -1,
            "TransmissionRate": "On request",
            "Units": "ASCII"
        },
        4939: {
            "Acronym": "CSI",
            "DataRange": "0 to 255 per byte",
            "EndBit": 0,
            "Name": "Side Collision Sensor Front Left - Serial Number",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": -2,
            "SPN": 4939,
            "SPNLength": 256,
            "StartBit": -1,
            "TransmissionRate": "On request",
            "Units": "ASCII"
        },
        4940: {
            "Acronym": "CSI",
            "DataRange": "0 to 255 per byte",
            "EndBit": 0,
            "Name": "Side Collision Sensor Mid Front Left - Serial Number",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": -2,
            "SPN": 4940,
            "SPNLength": 256,
            "StartBit": -1,
            "TransmissionRate": "On request",
            "Units": "ASCII"
        },
        4941: {
            "Acronym": "CSI",
            "DataRange": "0 to 255 per byte",
            "EndBit": 0,
            "Name": "Side Collision Sensor Mid Rear Left - Serial Number",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": -2,
            "SPN": 4941,
            "SPNLength": 256,
            "StartBit": -1,
            "TransmissionRate": "On request",
            "Units": "ASCII"
        },
        4942: {
            "Acronym": "CSI",
            "DataRange": "0 to 255 per byte",
            "EndBit": 0,
            "Name": "Side Collision Sensor Rear Left - Serial Number",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": -2,
            "SPN": 4942,
            "SPNLength": 256,
            "StartBit": -1,
            "TransmissionRate": "On request",
            "Units": "ASCII"
        },
        4943: {
            "Acronym": "CSI",
            "DataRange": "0 to 255 per byte",
            "EndBit": 0,
            "Name": "Side Collision Sensor Front Right - Serial Number",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": -2,
            "SPN": 4943,
            "SPNLength": 256,
            "StartBit": -1,
            "TransmissionRate": "On request",
            "Units": "ASCII"
        },
        4944: {
            "Acronym": "CSI",
            "DataRange": "0 to 255 per byte",
            "EndBit": 0,
            "Name": "Side Collision Sensor Mid Front Right - Serial Number",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": -2,
            "SPN": 4944,
            "SPNLength": 256,
            "StartBit": -1,
            "TransmissionRate": "On request",
            "Units": "ASCII"
        },
        4945: {
            "Acronym": "CSI",
            "DataRange": "0 to 255 per byte",
            "EndBit": 0,
            "Name": "Side Collision Sensor Mid Rear Right - Serial Number",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": -2,
            "SPN": 4945,
            "SPNLength": 256,
            "StartBit": -1,
            "TransmissionRate": "On request",
            "Units": "ASCII"
        },
        4946: {
            "Acronym": "CSI",
            "DataRange": "0 to 255 per byte",
            "EndBit": 0,
            "Name": "Side Collision Sensor Rear Right - Serial Number",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": -2,
            "SPN": 4946,
            "SPNLength": 256,
            "StartBit": -1,
            "TransmissionRate": "On request",
            "Units": "ASCII"
        },
        4947: {
            "Acronym": "CSI",
            "DataRange": "0 to 255 per byte",
            "EndBit": 0,
            "Name": "Rear Collision Sensor 1 - Serial Number",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": -2,
            "SPN": 4947,
            "SPNLength": 256,
            "StartBit": -1,
            "TransmissionRate": "On request",
            "Units": "ASCII"
        },
        4948: {
            "Acronym": "CSI",
            "DataRange": "0 to 255 per byte",
            "EndBit": 0,
            "Name": "Rear Collision Sensor 2 - Serial Number",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": -2,
            "SPN": 4948,
            "SPNLength": 256,
            "StartBit": -1,
            "TransmissionRate": "On request",
            "Units": "ASCII"
        },
        4949: {
            "Acronym": "CSI",
            "DataRange": "0 to 255 per byte",
            "EndBit": 0,
            "Name": "Rear Collision Sensor 3 - Serial Number",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": -2,
            "SPN": 4949,
            "SPNLength": 256,
            "StartBit": -1,
            "TransmissionRate": "On request",
            "Units": "ASCII"
        },
        4950: {
            "Acronym": "CSI",
            "DataRange": "0 to 255 per byte",
            "EndBit": 0,
            "Name": "Rear Collision Sensor 4 - Serial Number",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": -2,
            "SPN": 4950,
            "SPNLength": 256,
            "StartBit": -1,
            "TransmissionRate": "On request",
            "Units": "ASCII"
        },
        4951: {
            "Acronym": "CSI",
            "DataRange": "0 to 255 per byte",
            "EndBit": 0,
            "Name": "Rollover Sensor - Serial Number",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": -2,
            "SPN": 4951,
            "SPNLength": 256,
            "StartBit": -1,
            "TransmissionRate": "On request",
            "Units": "ASCII"
        },
        4952: {
            "Acronym": "BDS",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Driver Beltlock Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4952,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "250 msec",
            "Units": "bit"
        },
        4953: {
            "Acronym": "BDS",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Passenger Beltlock Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4953,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "250 msec",
            "Units": "bit"
        },
        4954: {
            "Acronym": "BDS",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Beltlock 3 Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4954,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "250 msec",
            "Units": "bit"
        },
        4955: {
            "Acronym": "BDS",
            "DataRange": "0 to 3",
            "EndBit": 7,
            "Name": "Beltlock 4 Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4955,
            "SPNLength": 2,
            "StartBit": 6,
            "TransmissionRate": "250 msec",
            "Units": "bit"
        },
        4956: {
            "Acronym": "BDS",
            "DataRange": "0 to 3",
            "EndBit": 9,
            "Name": "Beltlock 5 Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4956,
            "SPNLength": 2,
            "StartBit": 8,
            "TransmissionRate": "250 msec",
            "Units": "bit"
        },
        4957: {
            "Acronym": "BDS",
            "DataRange": "0 to 3",
            "EndBit": 11,
            "Name": "Beltlock 6 Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4957,
            "SPNLength": 2,
            "StartBit": 10,
            "TransmissionRate": "250 msec",
            "Units": "bit"
        },
        4958: {
            "Acronym": "BDS",
            "DataRange": "0 to 3",
            "EndBit": 13,
            "Name": "Beltlock 7 Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4958,
            "SPNLength": 2,
            "StartBit": 12,
            "TransmissionRate": "250 msec",
            "Units": "bit"
        },
        4959: {
            "Acronym": "BDS",
            "DataRange": "0 to 3",
            "EndBit": 15,
            "Name": "Beltlock 8 Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4959,
            "SPNLength": 2,
            "StartBit": 14,
            "TransmissionRate": "250 msec",
            "Units": "bit"
        },
        4960: {
            "Acronym": "BDS",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Beltlock 9 Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4960,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "250 msec",
            "Units": "bit"
        },
        4961: {
            "Acronym": "BDS",
            "DataRange": "0 to 3",
            "EndBit": 19,
            "Name": "Beltlock 10 Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4961,
            "SPNLength": 2,
            "StartBit": 18,
            "TransmissionRate": "250 msec",
            "Units": "bit"
        },
        4962: {
            "Acronym": "BDS",
            "DataRange": "0 to 3",
            "EndBit": 21,
            "Name": "Passenger Airbag Deactivation Switch Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4962,
            "SPNLength": 2,
            "StartBit": 20,
            "TransmissionRate": "250 msec",
            "Units": "bit"
        },
        4963: {
            "Acronym": "OCSI",
            "DataRange": "0 to 7",
            "EndBit": 2,
            "Name": "Driver Occupant Classification System Status",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4963,
            "SPNLength": 3,
            "StartBit": 0,
            "TransmissionRate": "Transmitted every 5 sec and on change of PGN 64791 but no faster than every 250 msec",
            "Units": "bit"
        },
        4964: {
            "Acronym": "OCSI",
            "DataRange": "0 to 7",
            "EndBit": 5,
            "Name": "Passenger Occupant Classification System Status",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4964,
            "SPNLength": 3,
            "StartBit": 3,
            "TransmissionRate": "Transmitted every 5 sec and on change of PGN 64791 but no faster than every 250 msec",
            "Units": "bit"
        },
        4965: {
            "Acronym": "OCSI",
            "DataRange": "0 to 3",
            "EndBit": 9,
            "Name": "Occupant Classification System 3 Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4965,
            "SPNLength": 2,
            "StartBit": 8,
            "TransmissionRate": "Transmitted every 5 sec and on change of PGN 64791 but no faster than every 250 msec",
            "Units": "bit"
        },
        4966: {
            "Acronym": "OCSI",
            "DataRange": "0 to 3",
            "EndBit": 11,
            "Name": "Occupant Classification System 4 Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4966,
            "SPNLength": 2,
            "StartBit": 10,
            "TransmissionRate": "Transmitted every 5 sec and on change of PGN 64791 but no faster than every 250 msec",
            "Units": "bit"
        },
        4967: {
            "Acronym": "OCSI",
            "DataRange": "0 to 3",
            "EndBit": 13,
            "Name": "Occupant Classification System 5 Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4967,
            "SPNLength": 2,
            "StartBit": 12,
            "TransmissionRate": "Transmitted every 5 sec and on change of PGN 64791 but no faster than every 250 msec",
            "Units": "bit"
        },
        4968: {
            "Acronym": "OCSI",
            "DataRange": "0 to 3",
            "EndBit": 15,
            "Name": "Occupant Classification System 6 Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4968,
            "SPNLength": 2,
            "StartBit": 14,
            "TransmissionRate": "Transmitted every 5 sec and on change of PGN 64791 but no faster than every 250 msec",
            "Units": "bit"
        },
        4969: {
            "Acronym": "OCSI",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Occupant Classification System 7 Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4969,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "Transmitted every 5 sec and on change of PGN 64791 but no faster than every 250 msec",
            "Units": "bit"
        },
        4970: {
            "Acronym": "OCSI",
            "DataRange": "0 to 3",
            "EndBit": 19,
            "Name": "Occupant Classification System 8 Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4970,
            "SPNLength": 2,
            "StartBit": 18,
            "TransmissionRate": "Transmitted every 5 sec and on change of PGN 64791 but no faster than every 250 msec",
            "Units": "bit"
        },
        4971: {
            "Acronym": "OCSI",
            "DataRange": "0 to 3",
            "EndBit": 21,
            "Name": "Occupant Classification System 9 Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4971,
            "SPNLength": 2,
            "StartBit": 20,
            "TransmissionRate": "Transmitted every 5 sec and on change of PGN 64791 but no faster than every 250 msec",
            "Units": "bit"
        },
        4972: {
            "Acronym": "OCSI",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Occupant Classification System 10 Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4972,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "Transmitted every 5 sec and on change of PGN 64791 but no faster than every 250 msec",
            "Units": "bit"
        },
        4973: {
            "Acronym": "CN",
            "DataRange": "bit-mapped",
            "EndBit": 4,
            "Name": "Crash Type",
            "Offset": 0.0,
            "OperationalHigh": -1,
            "OperationalLow": -1,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 5,
            "SPN": 4973,
            "SPNLength": 5,
            "StartBit": 0,
            "TransmissionRate": "Transmitted every 20 msec for the first 100 msec and then broadcast every 1 sec for 10 sec in case of a crash event",
            "Units": "bit-mapped"
        },
        4974: {
            "Acronym": "CN",
            "DataRange": "0 to 15",
            "EndBit": 59,
            "Name": "Crash Counter",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4974,
            "SPNLength": 4,
            "StartBit": 56,
            "TransmissionRate": "Transmitted every 20 msec for the first 100 msec and then broadcast every 1 sec for 10 sec in case of a crash event",
            "Units": "count"
        },
        4975: {
            "Acronym": "CN",
            "DataRange": "0 to 15",
            "EndBit": 63,
            "Name": "Crash Checksum",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4975,
            "SPNLength": 4,
            "StartBit": 60,
            "TransmissionRate": "Transmitted every 20 msec for the first 100 msec and then broadcast every 1 sec for 10 sec in case of a crash event",
            "Units": "count"
        },
        4976: {
            "Acronym": "SSI2",
            "DataRange": "-250 to 250.9999 deg",
            "EndBit": 23,
            "Name": "Pitch Angle Extended Range",
            "Offset": -250.0,
            "OperationalHigh": 250.9999,
            "OperationalLow": -250.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4976,
            "SPNLength": 24,
            "StartBit": 0,
            "TransmissionRate": "10 ms",
            "Units": "deg"
        },
        4977: {
            "Acronym": "SSI2",
            "DataRange": "-250 to 250.9999 deg",
            "EndBit": 47,
            "Name": "Roll Angle Extended Range",
            "Offset": -250.0,
            "OperationalHigh": 250.9999,
            "OperationalLow": -250.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4977,
            "SPNLength": 24,
            "StartBit": 24,
            "TransmissionRate": "10 ms",
            "Units": "deg"
        },
        4978: {
            "Acronym": "SSI2",
            "DataRange": "0 to 3",
            "EndBit": 49,
            "Name": "Pitch Angle Extended Range Compensation",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4978,
            "SPNLength": 2,
            "StartBit": 48,
            "TransmissionRate": "10 ms",
            "Units": "bit"
        },
        4979: {
            "Acronym": "SSI2",
            "DataRange": "0 to 3",
            "EndBit": 51,
            "Name": "Pitch Angle Extended Range Figure of Merit",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4979,
            "SPNLength": 2,
            "StartBit": 50,
            "TransmissionRate": "10 ms",
            "Units": "bit"
        },
        4980: {
            "Acronym": "SSI2",
            "DataRange": "0 to 3",
            "EndBit": 53,
            "Name": "Roll Angle Extended Range Compensation",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4980,
            "SPNLength": 2,
            "StartBit": 52,
            "TransmissionRate": "10 ms",
            "Units": "bit"
        },
        4981: {
            "Acronym": "SSI2",
            "DataRange": "0 to 3",
            "EndBit": 55,
            "Name": "Roll Angle Extended Range Figure of Merit",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4981,
            "SPNLength": 2,
            "StartBit": 54,
            "TransmissionRate": "10 ms",
            "Units": "bit"
        },
        4982: {
            "Acronym": "SSI2",
            "DataRange": "0 to 125 ms",
            "EndBit": 63,
            "Name": "Roll and Pitch Extended Range Measurement Latency",
            "Offset": 0.0,
            "OperationalHigh": 125.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.5,
            "SPN": 4982,
            "SPNLength": 8,
            "StartBit": 56,
            "TransmissionRate": "10 ms",
            "Units": "Milliseconds"
        },
        4983: {
            "Acronym": "ARI",
            "DataRange": "-250 to 250.992 deg/sec",
            "EndBit": 15,
            "Name": "Pitch Rate Extended Range",
            "Offset": -250.0,
            "OperationalHigh": 250.992,
            "OperationalLow": -250.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4983,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "10 ms",
            "Units": "deg/sec"
        },
        4984: {
            "Acronym": "ARI",
            "DataRange": "-250 to 250.992 deg/sec",
            "EndBit": 31,
            "Name": "Roll Rate Extended Range",
            "Offset": -250.0,
            "OperationalHigh": 250.992,
            "OperationalLow": -250.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4984,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "10 ms",
            "Units": "deg/sec"
        },
        4985: {
            "Acronym": "ARI",
            "DataRange": "-250 to 250.992 deg/sec",
            "EndBit": 47,
            "Name": "Yaw Rate Extended Range",
            "Offset": -250.0,
            "OperationalHigh": 250.992,
            "OperationalLow": -250.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4985,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "10 ms",
            "Units": "deg/sec"
        },
        4986: {
            "Acronym": "ARI",
            "DataRange": "0 to 3",
            "EndBit": 49,
            "Name": "Pitch Rate Extended Range Figure of Merit",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4986,
            "SPNLength": 2,
            "StartBit": 48,
            "TransmissionRate": "10 ms",
            "Units": "bit"
        },
        4987: {
            "Acronym": "ARI",
            "DataRange": "0 to 3",
            "EndBit": 51,
            "Name": "Roll Rate Extended Range Figure of Merit",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4987,
            "SPNLength": 2,
            "StartBit": 50,
            "TransmissionRate": "10 ms",
            "Units": "bit"
        },
        4988: {
            "Acronym": "ARI",
            "DataRange": "0 to 3",
            "EndBit": 53,
            "Name": "Yaw Rate Extended Range Figure of Merit",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4988,
            "SPNLength": 2,
            "StartBit": 52,
            "TransmissionRate": "10 ms",
            "Units": "bit"
        },
        4989: {
            "Acronym": "ARI",
            "DataRange": "0 to 125 ms",
            "EndBit": 63,
            "Name": "Angular Rate Measurement Latency",
            "Offset": 0.0,
            "OperationalHigh": 125.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.5,
            "SPN": 4989,
            "SPNLength": 8,
            "StartBit": 56,
            "TransmissionRate": "10 ms",
            "Units": "Milliseconds"
        },
        4990: {
            "Acronym": "BCH1",
            "DataRange": "0 to 15",
            "EndBit": 3,
            "Name": "Battery Charger 1 State",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4990,
            "SPNLength": 4,
            "StartBit": 0,
            "TransmissionRate": "1 sec",
            "Units": "bit"
        },
        4991: {
            "Acronym": "BCH1",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Battery Charger 1 Power Line State",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4991,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "1 sec",
            "Units": "bit"
        },
        4992: {
            "Acronym": "BCH1",
            "DataRange": "0 to 3212.75 V",
            "EndBit": 23,
            "Name": "Battery Charger 1 Output Voltage",
            "Offset": 0.0,
            "OperationalHigh": 3212.75,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05,
            "SPN": 4992,
            "SPNLength": 16,
            "StartBit": 8,
            "TransmissionRate": "1 sec",
            "Units": "Volts"
        },
        4993: {
            "Acronym": "BCH1",
            "DataRange": "-1600 to 1612.75 A",
            "EndBit": 39,
            "Name": "Battery Charger 1 Output Current",
            "Offset": -1600.0,
            "OperationalHigh": 1612.75,
            "OperationalLow": -1600.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05,
            "SPN": 4993,
            "SPNLength": 16,
            "StartBit": 24,
            "TransmissionRate": "1 sec",
            "Units": "Amps"
        },
        4994: {
            "Acronym": "BCH2",
            "DataRange": "0 to 15",
            "EndBit": 3,
            "Name": "Battery Charger 2 State",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4994,
            "SPNLength": 4,
            "StartBit": 0,
            "TransmissionRate": "1 sec",
            "Units": "bit"
        },
        4995: {
            "Acronym": "BCH2",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Battery Charger 2 Power Line State",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 4995,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "1 sec",
            "Units": "bit"
        },
        4996: {
            "Acronym": "BCH2",
            "DataRange": "0 to 3212.75 V",
            "EndBit": 23,
            "Name": "Battery Charger 2 Output Voltage",
            "Offset": 0.0,
            "OperationalHigh": 3212.75,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05,
            "SPN": 4996,
            "SPNLength": 16,
            "StartBit": 8,
            "TransmissionRate": "1 sec",
            "Units": "Volts"
        },
        4997: {
            "Acronym": "BCH2",
            "DataRange": "-1600 to 1612.75 A",
            "EndBit": 39,
            "Name": "Battery Charger 2 Output Current",
            "Offset": -1600.0,
            "OperationalHigh": 1612.75,
            "OperationalLow": -1600.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05,
            "SPN": 4997,
            "SPNLength": 16,
            "StartBit": 24,
            "TransmissionRate": "1 sec",
            "Units": "Amps"
        },
        4998: {
            "Acronym": "MSCI",
            "DataRange": "0 to 62.5 s",
            "EndBit": 7,
            "Name": "Magnet Boost Time",
            "Offset": 0.0,
            "OperationalHigh": 62.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.25,
            "SPN": 4998,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "Seconds"
        },
        4999: {
            "Acronym": "MSCI",
            "DataRange": "0.0 to 43083.687196475 hp",
            "EndBit": 23,
            "Name": "Magnet Rated Power",
            "Offset": 0.0,
            "OperationalHigh": 43083.687196475,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.670511045,
            "SPN": 4999,
            "SPNLength": 16,
            "StartBit": 8,
            "TransmissionRate": "On request",
            "Units": "hp"
        },
        5000: {
            "Acronym": "MSI1",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Generator Overheat Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5000,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        5001: {
            "Acronym": "MSI1",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Genset System Output Voltage Range Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5001,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        5002: {
            "Acronym": "MSI1",
            "DataRange": "0 to 31",
            "EndBit": 12,
            "Name": "Reverse Current Range Setting",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5002,
            "SPNLength": 5,
            "StartBit": 8,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        5003: {
            "Acronym": "MSI1",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Grapple/Magnet Selection Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5003,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        5004: {
            "Acronym": "MSI1",
            "DataRange": "0 to 3",
            "EndBit": 7,
            "Name": "Genset Softstart Active Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5004,
            "SPNLength": 2,
            "StartBit": 6,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        5005: {
            "Acronym": "MSI1",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Genset Enable Active Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5005,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        5006: {
            "Acronym": "MSI1",
            "DataRange": "0 to 3",
            "EndBit": 19,
            "Name": "Voltage Monitor Active Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5006,
            "SPNLength": 2,
            "StartBit": 18,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        5007: {
            "Acronym": "MSI1",
            "DataRange": "0 to 3",
            "EndBit": 21,
            "Name": "Generator Duty Cycle Exceeded Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5007,
            "SPNLength": 2,
            "StartBit": 20,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        5008: {
            "Acronym": "MSI2",
            "DataRange": "0 to 64,255 Amps",
            "EndBit": 15,
            "Name": "Magnet Forward Current",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5008,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "50 ms",
            "Units": "Amps"
        },
        5009: {
            "Acronym": "MSI2",
            "DataRange": "0 to 250 A",
            "EndBit": 23,
            "Name": "Magnet Reverse Current",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5009,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "50 ms",
            "Units": "Amps"
        },
        5010: {
            "Acronym": "MSI2",
            "DataRange": "0 to 3",
            "EndBit": 25,
            "Name": "Material Lift Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5010,
            "SPNLength": 2,
            "StartBit": 24,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        5011: {
            "Acronym": "MSI2",
            "DataRange": "0 to 3",
            "EndBit": 27,
            "Name": "Material Drop Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5011,
            "SPNLength": 2,
            "StartBit": 26,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        5012: {
            "Acronym": "MSI2",
            "DataRange": "0 to 3",
            "EndBit": 29,
            "Name": "Generator Current Boost Active Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5012,
            "SPNLength": 2,
            "StartBit": 28,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        5013: {
            "Acronym": "MSI2",
            "DataRange": "0 to 3",
            "EndBit": 31,
            "Name": "Material Lift Active Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5013,
            "SPNLength": 2,
            "StartBit": 30,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        5014: {
            "Acronym": "MSI2",
            "DataRange": "0 to 3",
            "EndBit": 33,
            "Name": "Material Drop Active Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5014,
            "SPNLength": 2,
            "StartBit": 32,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        5015: {
            "Acronym": "ETC1",
            "DataRange": "0 to 3",
            "EndBit": 37,
            "Name": "Momentary Engine Maximum Power Enable",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5015,
            "SPNLength": 2,
            "StartBit": 36,
            "TransmissionRate": "10 ms",
            "Units": "bit"
        },
        5019: {
            "Acronym": "EFL/P3",
            "DataRange": "0.0 to 72.518869 psi",
            "EndBit": 47,
            "Name": "Engine Exhaust Gas Recirculation 1 Outlet Pressure",
            "Offset": 0.0,
            "OperationalHigh": 72.518869,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.290075476,
            "SPN": 5019,
            "SPNLength": 8,
            "StartBit": 40,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        5020: {
            "Acronym": "ET4",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 55,
            "Name": "Engine Exhaust Gas Recirculation 1 Mixer Intake Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 5020,
            "SPNLength": 16,
            "StartBit": 40,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        5021: {
            "Acronym": "EEC2",
            "DataRange": "0 to 3",
            "EndBit": 43,
            "Name": "Momentary Engine Maximum Power Enable Feedback",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5021,
            "SPNLength": 2,
            "StartBit": 42,
            "TransmissionRate": "50 msec (preferred) or Engine Speed Dependent (if required by application)",
            "Units": "bit"
        },
        5022: {
            "Acronym": "ACC1",
            "DataRange": "0 to 3",
            "EndBit": 55,
            "Name": "Forward Collision Warning",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5022,
            "SPNLength": 2,
            "StartBit": 54,
            "TransmissionRate": "100ms or upon state change, but not faster than 20 ms.",
            "Units": "bit"
        },
        5023: {
            "Acronym": "ACC2",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "ACC usage demand",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5023,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "250 ms",
            "Units": "bit"
        },
        5024: {
            "Acronym": "AT1IGC1",
            "DataRange": "0 to 64.255",
            "EndBit": 15,
            "Name": "Aftertreatment 1 Intake Gas NOx Sensor Heater Ratio",
            "Offset": 0.0,
            "OperationalHigh": 64.255,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.001,
            "SPN": 5024,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On start-up, and every second until the dewpoint signal state = 1 (SPN 3237) has been received by the transmitter",
            "Units": "Ratio"
        },
        5025: {
            "Acronym": "AT1IGC1",
            "DataRange": "-100 to 100 %",
            "EndBit": 31,
            "Name": "Aftertreatment 1 Intake Gas NOx Sensor New part deviation NOx Gain",
            "Offset": -3212.75,
            "OperationalHigh": 100.0,
            "OperationalLow": -100.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 5025,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On start-up, and every second until the dewpoint signal state = 1 (SPN 3237) has been received by the transmitter",
            "Units": "%"
        },
        5026: {
            "Acronym": "AT1IGC1",
            "DataRange": "-125 to 125 ppm",
            "EndBit": 39,
            "Name": "Aftertreatment 1 Intake Gas NOx Sensor New part deviation NOx Offset",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5026,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "On start-up, and every second until the dewpoint signal state = 1 (SPN 3237) has been received by the transmitter",
            "Units": "ppm"
        },
        5027: {
            "Acronym": "AT1IGC2",
            "DataRange": "0 to 125%/bar",
            "EndBit": 7,
            "Name": "Aftertreatment 1 Intake Gas NOx Sensor Correction of Pressure Lambda",
            "Offset": 0.0,
            "OperationalHigh": 125.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.5,
            "SPN": 5027,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "%/bar"
        },
        5028: {
            "Acronym": "AT1IGC2",
            "DataRange": "0 to 125%/bar",
            "EndBit": 15,
            "Name": "Aftertreatment 1 Intake Gas NOx Sensor Correction of Pressure NOx",
            "Offset": 0.0,
            "OperationalHigh": 125.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.5,
            "SPN": 5028,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "On request",
            "Units": "%/bar"
        },
        5029: {
            "Acronym": "AT1IGC2",
            "DataRange": "0 to 125%",
            "EndBit": 23,
            "Name": "Aftertreatment 1 Intake Gas NOx Sensor NO2 Correction",
            "Offset": 0.0,
            "OperationalHigh": 125.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.5,
            "SPN": 5029,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        5030: {
            "Acronym": "AT1IGC2",
            "DataRange": "0 to 125%",
            "EndBit": 31,
            "Name": "Aftertreatment 1 Intake Gas NOx Sensor NH3 Correction",
            "Offset": 0.0,
            "OperationalHigh": 125.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.5,
            "SPN": 5030,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        5031: {
            "Acronym": "AT1OGC1",
            "DataRange": "0 to 64.255",
            "EndBit": 15,
            "Name": "Aftertreatment 1 Outlet Gas NOx Sensor Heater Ratio",
            "Offset": 0.0,
            "OperationalHigh": 64.255,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.001,
            "SPN": 5031,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On start-up, and every second until the dewpoint signal state = 1 (SPN 3238) has been received by the transmitter",
            "Units": "Ratio"
        },
        5032: {
            "Acronym": "AT1OGC1",
            "DataRange": "-100 to 100 %",
            "EndBit": 31,
            "Name": "Aftertreatment 1 Outlet Gas NOx Sensor New Part Deviation NOx Gain",
            "Offset": -3212.75,
            "OperationalHigh": 100.0,
            "OperationalLow": -100.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 5032,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On start-up, and every second until the dewpoint signal state = 1 (SPN 3238) has been received by the transmitter",
            "Units": "%"
        },
        5033: {
            "Acronym": "AT1OGC1",
            "DataRange": "-125 to 125 ppm",
            "EndBit": 39,
            "Name": "Aftertreatment 1 Outlet Gas NOx Sensor New Part Deviation NOx Offset",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5033,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "On start-up, and every second until the dewpoint signal state = 1 (SPN 3238) has been received by the transmitter",
            "Units": "ppm"
        },
        5034: {
            "Acronym": "AT1OGC2",
            "DataRange": "0 to 125%/bar",
            "EndBit": 7,
            "Name": "Aftertreatment 1 Outlet Gas NOx Sensor Correction of Pressure Lambda",
            "Offset": 0.0,
            "OperationalHigh": 125.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.5,
            "SPN": 5034,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "%/bar"
        },
        5035: {
            "Acronym": "AT1OGC2",
            "DataRange": "0 to 125%/bar",
            "EndBit": 15,
            "Name": "Aftertreatment 1 Outlet Gas NOx Sensor Correction of Pressure NOx",
            "Offset": 0.0,
            "OperationalHigh": 125.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.5,
            "SPN": 5035,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "On request",
            "Units": "%/bar"
        },
        5036: {
            "Acronym": "AT1OGC2",
            "DataRange": "0 to 125%",
            "EndBit": 23,
            "Name": "Aftertreatment 1 Outlet Gas NOx Sensor NO2 Correction",
            "Offset": 0.0,
            "OperationalHigh": 125.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.5,
            "SPN": 5036,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        5037: {
            "Acronym": "AT1OGC2",
            "DataRange": "0 to 125%",
            "EndBit": 31,
            "Name": "Aftertreatment 1 Outlet Gas NOx Sensor NH3 Correction",
            "Offset": 0.0,
            "OperationalHigh": 125.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.5,
            "SPN": 5037,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        5038: {
            "Acronym": "AT2IGC1",
            "DataRange": "0 to 64.255",
            "EndBit": 15,
            "Name": "Aftertreatment 2 Intake Gas NOx Sensor Heater Ratio",
            "Offset": 0.0,
            "OperationalHigh": 64.255,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.001,
            "SPN": 5038,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On start-up, and every second until the dewpoint signal state = 1 (SPN 3239) has been received by the transmitter",
            "Units": "Ratio"
        },
        5039: {
            "Acronym": "AT2IGC1",
            "DataRange": "-100 to 100 %",
            "EndBit": 31,
            "Name": "Aftertreatment 2 Intake Gas NOx Sensor New part deviation NOx Gain",
            "Offset": -3212.75,
            "OperationalHigh": 100.0,
            "OperationalLow": -100.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 5039,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On start-up, and every second until the dewpoint signal state = 1 (SPN 3239) has been received by the transmitter",
            "Units": "%"
        },
        5040: {
            "Acronym": "AT2IGC1",
            "DataRange": "-125 to 125 ppm",
            "EndBit": 39,
            "Name": "Aftertreatment 2 Intake Gas NOx Sensor New part deviation NOx Offset",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5040,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "On start-up, and every second until the dewpoint signal state = 1 (SPN 3239) has been received by the transmitter",
            "Units": "ppm"
        },
        5041: {
            "Acronym": "AT2IGC2",
            "DataRange": "0 to 125%/bar",
            "EndBit": 7,
            "Name": "Aftertreatment 2 Intake Gas NOx Sensor Correction of pressure Lambda",
            "Offset": 0.0,
            "OperationalHigh": 125.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.5,
            "SPN": 5041,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "%/bar"
        },
        5042: {
            "Acronym": "AT2IGC2",
            "DataRange": "0 to 125%/bar",
            "EndBit": 15,
            "Name": "Aftertreatment 2 Intake Gas NOx Sensor Correction of pressure Nox",
            "Offset": 0.0,
            "OperationalHigh": 125.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.5,
            "SPN": 5042,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "On request",
            "Units": "%/bar"
        },
        5043: {
            "Acronym": "AT2IGC2",
            "DataRange": "0 to 125%",
            "EndBit": 23,
            "Name": "Aftertreatment 2 Intake Gas NOx Sensor NO2 Correction",
            "Offset": 0.0,
            "OperationalHigh": 125.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.5,
            "SPN": 5043,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        5044: {
            "Acronym": "AT2IGC2",
            "DataRange": "0 to 125%",
            "EndBit": 31,
            "Name": "Aftertreatment 2 Intake Gas NOx Sensor NH3 Correction",
            "Offset": 0.0,
            "OperationalHigh": 125.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.5,
            "SPN": 5044,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        5045: {
            "Acronym": "AT2OGC1",
            "DataRange": "0 to 64.255",
            "EndBit": 15,
            "Name": "Aftertreatment 2 Outlet Gas NOx Sensor Heater Ratio",
            "Offset": 0.0,
            "OperationalHigh": 64.255,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.001,
            "SPN": 5045,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On start-up, and every second until the dewpoint signal state = 1 (SPN 3240) has been received by the transmitter",
            "Units": "Ratio"
        },
        5046: {
            "Acronym": "AT2OGC1",
            "DataRange": "-100 to 100 %",
            "EndBit": 31,
            "Name": "Aftertreatment 2 Outlet Gas NOx Sensor New part deviation NOx Gain",
            "Offset": -3212.75,
            "OperationalHigh": 100.0,
            "OperationalLow": -100.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 5046,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On start-up, and every second until the dewpoint signal state = 1 (SPN 3240) has been received by the transmitter",
            "Units": "%"
        },
        5047: {
            "Acronym": "AT2OGC1",
            "DataRange": "-125 to 125 ppm",
            "EndBit": 39,
            "Name": "Aftertreatment 2 Outlet Gas NOx Sensor New part deviation NOx Offset",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5047,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "On start-up, and every second until the dewpoint signal state = 1 (SPN 3240) has been received by the transmitter",
            "Units": "ppm"
        },
        5048: {
            "Acronym": "AT2OGC2",
            "DataRange": "0 to 125%/bar",
            "EndBit": 7,
            "Name": "Aftertreatment 2 Outlet Gas NOx Sensor Correction of pressure Lambda",
            "Offset": 0.0,
            "OperationalHigh": 125.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.5,
            "SPN": 5048,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "%/bar"
        },
        5049: {
            "Acronym": "AT2OGC2",
            "DataRange": "0 to 125%/bar",
            "EndBit": 15,
            "Name": "Aftertreatment 2 Outlet Gas NOx Sensor Correction of pressure Nox",
            "Offset": 0.0,
            "OperationalHigh": 125.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.5,
            "SPN": 5049,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "On request",
            "Units": "%/bar"
        },
        5050: {
            "Acronym": "AT2OGC2",
            "DataRange": "0 to 125%",
            "EndBit": 23,
            "Name": "Aftertreatment 2 Outlet Gas NOx Sensor NO2 Correction",
            "Offset": 0.0,
            "OperationalHigh": 125.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.5,
            "SPN": 5050,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        5051: {
            "Acronym": "AT2OGC2",
            "DataRange": "0 to 125%",
            "EndBit": 31,
            "Name": "Aftertreatment 2 Outlet Gas NOx Sensor NH3 Correction",
            "Offset": 0.0,
            "OperationalHigh": 125.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.5,
            "SPN": 5051,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        5052: {
            "Acronym": "ETC8",
            "DataRange": "0 to 8,031.875 rpm",
            "EndBit": 31,
            "Name": "Transmission Clutch/Converter Input Speed",
            "Offset": 0.0,
            "OperationalHigh": 8031.875,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.125,
            "SPN": 5052,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "20 ms when torque converter unlocked, 100 ms when torque converter locked",
            "Units": "rpm"
        },
        5053: {
            "Acronym": "HRLFC",
            "DataRange": "0 to 4,211,081.215 L",
            "EndBit": 31,
            "Name": "High Resolution Engine Trip Fuel",
            "Offset": 0.0,
            "OperationalHigh": 4211081.215,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.001,
            "SPN": 5053,
            "SPNLength": 32,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "l"
        },
        5054: {
            "Acronym": "HRLFC",
            "DataRange": "0 to 4,211,081.215 L",
            "EndBit": 63,
            "Name": "High Resolution Engine Total Fuel Used",
            "Offset": 0.0,
            "OperationalHigh": 4211081.215,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.001,
            "SPN": 5054,
            "SPNLength": 32,
            "StartBit": 32,
            "TransmissionRate": "1 s",
            "Units": "l"
        },
        5055: {
            "Acronym": "EOM",
            "DataRange": "0 to 1003.984375 Cp",
            "EndBit": 15,
            "Name": "Engine Oil Viscosity",
            "Offset": 0.0,
            "OperationalHigh": 1003.984375,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.015625,
            "SPN": 5055,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "30 seconds",
            "Units": "Cp"
        },
        5056: {
            "Acronym": "EOM",
            "DataRange": "0 to 1.961 g/cc",
            "EndBit": 31,
            "Name": "Engine Oil Density",
            "Offset": 0.0,
            "OperationalHigh": 1.961,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 3.052e-05,
            "SPN": 5056,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "30 seconds",
            "Units": "g/cc"
        },
        5057: {
            "Acronym": "NC1",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Cab Noise Control Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5057,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "Every 10 seconds and on change of state, but not faster than every 1 second.",
            "Units": "bit"
        },
        5058: {
            "Acronym": "NC1",
            "DataRange": "0 to 7",
            "EndBit": 4,
            "Name": "Cab Noise Control Tuning Command",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5058,
            "SPNLength": 3,
            "StartBit": 2,
            "TransmissionRate": "Every 10 seconds and on change of state, but not faster than every 1 second.",
            "Units": "bit"
        },
        5059: {
            "Acronym": "NCS",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Cab Noise Control Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5059,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "Every 10 seconds and on change of state, but not faster than 1 second.  Every second when in tuning mode.",
            "Units": "bit"
        },
        5060: {
            "Acronym": "NCS",
            "DataRange": "0 to 15",
            "EndBit": 5,
            "Name": "Cab Noise Control Tuning Status",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5060,
            "SPNLength": 4,
            "StartBit": 2,
            "TransmissionRate": "Every 10 seconds and on change of state, but not faster than 1 second.  Every second when in tuning mode.",
            "Units": "bit"
        },
        5077: {
            "Acronym": "DLCC1",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Engine Protect Lamp Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5077,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms",
            "Units": "bit"
        },
        5078: {
            "Acronym": "DLCC1",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Engine Amber Warning Lamp Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5078,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms",
            "Units": "bit"
        },
        5079: {
            "Acronym": "DLCC1",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Engine Red Stop Lamp Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5079,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms",
            "Units": "bit"
        },
        5080: {
            "Acronym": "DLCC1",
            "DataRange": "0 to 3",
            "EndBit": 7,
            "Name": "OBD Malfunction Indicator Lamp Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5080,
            "SPNLength": 2,
            "StartBit": 6,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms",
            "Units": "bit"
        },
        5081: {
            "Acronym": "DLCC1",
            "DataRange": "0 to 3",
            "EndBit": 9,
            "Name": "Engine Brake Active Lamp Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5081,
            "SPNLength": 2,
            "StartBit": 8,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms",
            "Units": "bit"
        },
        5082: {
            "Acronym": "DLCC1",
            "DataRange": "0 to 3",
            "EndBit": 13,
            "Name": "Engine Oil Pressure Low Lamp Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5082,
            "SPNLength": 2,
            "StartBit": 12,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms",
            "Units": "bit"
        },
        5083: {
            "Acronym": "DLCC1",
            "DataRange": "0 to 3",
            "EndBit": 15,
            "Name": "Engine Coolant Temperature High Lamp Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5083,
            "SPNLength": 2,
            "StartBit": 14,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms",
            "Units": "bit"
        },
        5084: {
            "Acronym": "DLCC1",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Engine Coolant Level Low Lamp Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5084,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms",
            "Units": "bit"
        },
        5085: {
            "Acronym": "DLCC1",
            "DataRange": "0 to 3",
            "EndBit": 19,
            "Name": "Engine Idle Management Active Lamp Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5085,
            "SPNLength": 2,
            "StartBit": 18,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms",
            "Units": "bit"
        },
        5086: {
            "Acronym": "DLCC1",
            "DataRange": "0 to 3",
            "EndBit": 21,
            "Name": "Engine Air Filter Restriction Lamp Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5086,
            "SPNLength": 2,
            "StartBit": 20,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms",
            "Units": "bit"
        },
        5087: {
            "Acronym": "DLCC2",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Vehicle Battery Voltage Low Lamp Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5087,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms",
            "Units": "bit"
        },
        5088: {
            "Acronym": "DLCC2",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Vehicle Fuel Level Low Lamp Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5088,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms",
            "Units": "bit"
        },
        5089: {
            "Acronym": "DLCC2",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Vehicle Air Pressure Low Lamp Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5089,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms",
            "Units": "bit"
        },
        509: {
            "Acronym": "TR2",
            "DataRange": "0 to 255 per byte",
            "EndBit": 0,
            "Name": "Milepost Identification",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": -2,
            "SPN": 509,
            "SPNLength": 800,
            "StartBit": -1,
            "TransmissionRate": "On request",
            "Units": "ASCII"
        },
        5090: {
            "Acronym": "DLCC2",
            "DataRange": "0 to 3",
            "EndBit": 7,
            "Name": "Vehicle HVAC Recirculation Lamp Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5090,
            "SPNLength": 2,
            "StartBit": 6,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms",
            "Units": "bit"
        },
        5091: {
            "Acronym": "DLCC2",
            "DataRange": "0 to 3",
            "EndBit": 9,
            "Name": "Vehicle Battery Charging Lamp Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5091,
            "SPNLength": 2,
            "StartBit": 8,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms",
            "Units": "bit"
        },
        5093: {
            "Acronym": "DLCD1",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Engine Protect Lamp Data",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5093,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        5094: {
            "Acronym": "DLCD1",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Engine Amber Warning Lamp Data",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5094,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        5095: {
            "Acronym": "DLCD1",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Engine Red Stop Lamp Data",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5095,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        5096: {
            "Acronym": "DLCD1",
            "DataRange": "0 to 3",
            "EndBit": 7,
            "Name": "OBD Malfunction Indicator Lamp Data",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5096,
            "SPNLength": 2,
            "StartBit": 6,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        5097: {
            "Acronym": "DLCD1",
            "DataRange": "0 to 3",
            "EndBit": 9,
            "Name": "Engine Brake Active Lamp Data",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5097,
            "SPNLength": 2,
            "StartBit": 8,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        5098: {
            "Acronym": "DLCD1",
            "DataRange": "0 to 3",
            "EndBit": 11,
            "Name": "Compression Brake Enable Switch Indicator Lamp Data",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5098,
            "SPNLength": 2,
            "StartBit": 10,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        5099: {
            "Acronym": "DLCD1",
            "DataRange": "0 to 3",
            "EndBit": 13,
            "Name": "Engine Oil Pressure Low Lamp Data",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5099,
            "SPNLength": 2,
            "StartBit": 12,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        51: {
            "Acronym": "LFE1",
            "DataRange": "0 to 100 %",
            "EndBit": 55,
            "Name": "Engine Throttle Valve 1 Position",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 51,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        5100: {
            "Acronym": "DLCD1",
            "DataRange": "0 to 3",
            "EndBit": 15,
            "Name": "Engine Coolant Temperature High Lamp Data",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5100,
            "SPNLength": 2,
            "StartBit": 14,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        5101: {
            "Acronym": "DLCD1",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Engine Coolant Level Low Lamp Data",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5101,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        5102: {
            "Acronym": "DLCD1",
            "DataRange": "0 to 3",
            "EndBit": 19,
            "Name": "Engine Idle Management Active Lamp Data",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5102,
            "SPNLength": 2,
            "StartBit": 18,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        5103: {
            "Acronym": "DLCD1",
            "DataRange": "0 to 3",
            "EndBit": 21,
            "Name": "Engine Air Filter Restriction Lamp Data",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5103,
            "SPNLength": 2,
            "StartBit": 20,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        5104: {
            "Acronym": "DLCD2",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Vehicle Battery Voltage Low Lamp Data",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5104,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        5105: {
            "Acronym": "DLCD2",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Vehicle Fuel Level Low Lamp Data",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5105,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        5106: {
            "Acronym": "DLCD2",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Vehicle Air Pressure Low Lamp Data",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5106,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        5107: {
            "Acronym": "DLCD2",
            "DataRange": "0 to 3",
            "EndBit": 7,
            "Name": "Vehicle HVAC Recirculation Lamp Data",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5107,
            "SPNLength": 2,
            "StartBit": 6,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        5108: {
            "Acronym": "DLCD2",
            "DataRange": "0 to 3",
            "EndBit": 9,
            "Name": "Vehicle Battery Charging Lamp Data",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5108,
            "SPNLength": 2,
            "StartBit": 8,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        512: {
            "Acronym": "EEC1",
            "DataRange": "-125 to 125 %",
            "EndBit": 15,
            "Name": "Driver's Demand Engine - Percent Torque",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 125%",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 512,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "engine speed dependent",
            "Units": "%"
        },
        5125: {
            "Acronym": "SEP2",
            "DataRange": "0 to 3212.75 V",
            "EndBit": 47,
            "Name": "Sensor supply voltage 7",
            "Offset": 0.0,
            "OperationalHigh": 3212.75,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05,
            "SPN": 5125,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "1 sec",
            "Units": "Volts"
        },
        5126: {
            "Acronym": "SEP2",
            "DataRange": "0 to 3212.75 V",
            "EndBit": 63,
            "Name": "Sensor supply voltage 8",
            "Offset": 0.0,
            "OperationalHigh": 3212.75,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05,
            "SPN": 5126,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "1 sec",
            "Units": "Volts"
        },
        513: {
            "Acronym": "EEC1",
            "DataRange": "-125 to 125 %",
            "EndBit": 23,
            "Name": "Actual Engine - Percent Torque",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 125%",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 513,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "engine speed dependent",
            "Units": "%"
        },
        5137: {
            "Acronym": "A1SCRDSR2",
            "DataRange": "0 to 100 %",
            "EndBit": 23,
            "Name": "Aftertreatment 1 Diesel Exhaust Fluid Tank Heater Command",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 5137,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "500 ms",
            "Units": "%"
        },
        5138: {
            "Acronym": "A2SCRDSR2",
            "DataRange": "0 to 100 %",
            "EndBit": 23,
            "Name": "Aftertreatment 2 Diesel Exhaust Fluid Tank Heater Command",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 5138,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "500 ms",
            "Units": "%"
        },
        5139: {
            "Acronym": "IOS",
            "DataRange": "0 to 3",
            "EndBit": 57,
            "Name": "Implement operating state - command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5139,
            "SPNLength": 2,
            "StartBit": 56,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms",
            "Units": "bit"
        },
        514: {
            "Acronym": "EEC3",
            "DataRange": "-125 to 125 %",
            "EndBit": 7,
            "Name": "Nominal Friction - Percent Torque",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 514,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "250 msec (preferred) or Engine Speed Dependent (if required by application)",
            "Units": "%"
        },
        5140: {
            "Acronym": "AISO",
            "DataRange": "0 to 3",
            "EndBit": 57,
            "Name": "All Implement stop operating operations switch state",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5140,
            "SPNLength": 2,
            "StartBit": 56,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms",
            "Units": "bit"
        },
        5141: {
            "Acronym": "LVDS",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Low Voltage Disconnect Manual Disconnect State",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5141,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        5142: {
            "Acronym": "LVDS",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Low Voltage Disconnect Manual Connect State",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5142,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        5143: {
            "Acronym": "LVDS",
            "DataRange": "0 to 15",
            "EndBit": 7,
            "Name": "Low Voltage Disconnect Current Operating Mode",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5143,
            "SPNLength": 4,
            "StartBit": 4,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        5144: {
            "Acronym": "LVDS",
            "DataRange": "0 to 15",
            "EndBit": 11,
            "Name": "Low Voltage Disconnect Output #3 State",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5144,
            "SPNLength": 4,
            "StartBit": 8,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        5145: {
            "Acronym": "LVDS",
            "DataRange": "0 to 15",
            "EndBit": 15,
            "Name": "Low Voltage Disconnect Output #2 State",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5145,
            "SPNLength": 4,
            "StartBit": 12,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        5146: {
            "Acronym": "LVDS",
            "DataRange": "0 to 15",
            "EndBit": 19,
            "Name": "Low Voltage Disconnect Output #1 State",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5146,
            "SPNLength": 4,
            "StartBit": 16,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        5147: {
            "Acronym": "LVDS",
            "DataRange": "0 to 15",
            "EndBit": 23,
            "Name": "Low Voltage Disconnect Vout Output State",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5147,
            "SPNLength": 4,
            "StartBit": 20,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        5148: {
            "Acronym": "LVDS",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 31,
            "Name": "Low Voltage Disconnect Temperature",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 5148,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "500 ms",
            "Units": "deg F"
        },
        5149: {
            "Acronym": "LVDSOM",
            "DataRange": "0 to 15",
            "EndBit": 3,
            "Name": "Low Voltage Disconnect Desired Operating Mode",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5149,
            "SPNLength": 4,
            "StartBit": 0,
            "TransmissionRate": "As required but no more often than 500 ms",
            "Units": "bit"
        },
        515: {
            "Acronym": "EEC3",
            "DataRange": "0 to 8,031.875 rpm",
            "EndBit": 23,
            "Name": "Engine's Desired Operating Speed",
            "Offset": 0.0,
            "OperationalHigh": 8031.875,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.125,
            "SPN": 515,
            "SPNLength": 16,
            "StartBit": 8,
            "TransmissionRate": "250 msec (preferred) or Engine Speed Dependent (if required by application)",
            "Units": "rpm"
        },
        517: {
            "Acronym": "VDS",
            "DataRange": "0.0 to 155.961635516 mph",
            "EndBit": 31,
            "Name": "Navigation-Based Vehicle Speed",
            "Offset": 0.0,
            "OperationalHigh": 155.961635516,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.00242723046875,
            "SPN": 517,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "mph"
        },
        518: {
            "Acronym": "TSC1",
            "DataRange": "-125 to 125 %",
            "EndBit": 31,
            "Name": "Engine Requested Torque/Torque Limit",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 518,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "To engine: Control Purpose dependent or 10 ms\nTo retarder: 50 ms",
            "Units": "%"
        },
        519: {
            "Acronym": "EEC3",
            "DataRange": "0 to 250",
            "EndBit": 31,
            "Name": "Engine's Desired Operating Speed Asymmetry Adjustment",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 519,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "250 msec (preferred) or Engine Speed Dependent (if required by application)",
            "Units": "Ratio"
        },
        52: {
            "Acronym": "ET1",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 55,
            "Name": "Engine Intercooler Temperature",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 52,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        520: {
            "Acronym": "ERC1",
            "DataRange": "-125 to 125 %",
            "EndBit": 15,
            "Name": "Actual Retarder - Percent Torque",
            "Offset": -125.0,
            "OperationalHigh": 0.0,
            "OperationalLow": -125.0,
            "OperationalRange": "-125 to 0 %",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 520,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        521: {
            "Acronym": "EBC1",
            "DataRange": "0 to 100 %",
            "EndBit": 15,
            "Name": "Brake Pedal Position",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 521,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        522: {
            "Acronym": "ETC1",
            "DataRange": "0 to 100 %",
            "EndBit": 31,
            "Name": "Percent Clutch Slip",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 522,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "10 ms",
            "Units": "%"
        },
        523: {
            "Acronym": "ETC2",
            "DataRange": "-125 to 125",
            "EndBit": 31,
            "Name": "Transmission Current Gear",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "-125 to +125",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 523,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "100 ms",
            "Units": "gear value"
        },
        524: {
            "Acronym": "ETC2",
            "DataRange": "-125 to 125",
            "EndBit": 7,
            "Name": "Transmission Selected Gear",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "-125 to +125",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 524,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "gear value"
        },
        5245: {
            "Acronym": "AT1T1I",
            "DataRange": "0 to 7",
            "EndBit": 39,
            "Name": "Aftertreatment 1 DEF Tank Low Level Indicator",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5245,
            "SPNLength": 3,
            "StartBit": 37,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        5246: {
            "Acronym": "AT1T1I",
            "DataRange": "0 to 7",
            "EndBit": 47,
            "Name": "Aftertreatment SCR Operator Inducement Severity",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5246,
            "SPNLength": 3,
            "StartBit": 45,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        525: {
            "Acronym": "TC1",
            "DataRange": "-125 to 125",
            "EndBit": 23,
            "Name": "Transmission Requested Gear",
            "Offset": -125.0,
            "OperationalHigh": 64.0,
            "OperationalLow": -64.0,
            "OperationalRange": "-64 to 64",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 525,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "When active; 50 ms to transmission and axles",
            "Units": "gear value"
        },
        5252: {
            "Acronym": "EFL/P10",
            "DataRange": "-36.2594345 to 36.54805959862 psi",
            "EndBit": 15,
            "Name": "Engine Exhaust Gas Recirculation 2 Differential Pressure",
            "Offset": -36.2594345,
            "OperationalHigh": 36.54805959862,
            "OperationalLow": -36.2594345,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.001133107,
            "SPN": 5252,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        5253: {
            "Acronym": "EFL/P10",
            "DataRange": "0.0 to 72.518869 psi",
            "EndBit": 23,
            "Name": "Engine Exhaust Gas Recirculation 2 Intake Pressure",
            "Offset": 0.0,
            "OperationalHigh": 72.518869,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.290075476,
            "SPN": 5253,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        5254: {
            "Acronym": "EFL/P10",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 39,
            "Name": "Engine Exhaust Gas Recirculation 2 Outlet Absolute Pressure",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 5254,
            "SPNLength": 16,
            "StartBit": 24,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        5255: {
            "Acronym": "ET5",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 15,
            "Name": "Engine Exhaust Gas Recirculation 2 Temperature ",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 5255,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "1 sec",
            "Units": "deg F"
        },
        5256: {
            "Acronym": "ET5",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 31,
            "Name": "Engine Exhaust Gas Recirculation 2 Mixer Intake Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 5256,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "1 sec",
            "Units": "deg F"
        },
        5257: {
            "Acronym": "EGF1",
            "DataRange": "0 to 3212.75 kg/h",
            "EndBit": 47,
            "Name": "Engine Exhaust Gas Recirculation 2 (EGR2) Mass Flow Rate",
            "Offset": 0.0,
            "OperationalHigh": 3212.75,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05,
            "SPN": 5257,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "50 ms",
            "Units": "kg/h"
        },
        5258: {
            "Acronym": "EEC10",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 15,
            "Name": "Engine Exhaust Gas Recirculation 2 (EGR2) Cooler  Intake Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 5258,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "deg F"
        },
        5259: {
            "Acronym": "EEC10",
            "DataRange": "0.0 to 4659.699927595 psi",
            "EndBit": 31,
            "Name": "Engine Exhaust Gas Recirculation 2 (EGR2) Cooler  Intake Gas Absolute Pressure",
            "Offset": 0.0,
            "OperationalHigh": 4659.699927595,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.072518869,
            "SPN": 5259,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "psi"
        },
        526: {
            "Acronym": "ETC2",
            "DataRange": "0 to 64.255",
            "EndBit": 23,
            "Name": "Transmission Actual Gear Ratio",
            "Offset": 0.0,
            "OperationalHigh": 64.255,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.001,
            "SPN": 526,
            "SPNLength": 16,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "Ratio"
        },
        5260: {
            "Acronym": "EEC10",
            "DataRange": "0 to 100 %",
            "EndBit": 39,
            "Name": "Engine Exhaust Gas Recirculation 2 (EGR2) Cooler  Efficiency",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 5260,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        5261: {
            "Acronym": "EEC10",
            "DataRange": "0 to 100 %",
            "EndBit": 47,
            "Name": "EGR 2 Cooler Bypass Actuator Position",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 5261,
            "SPNLength": 8,
            "StartBit": 40,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        5262: {
            "Acronym": "EEC9",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 15,
            "Name": "Engine Exhaust Gas Recirculation 2 Valve Position",
            "Offset": 0.0,
            "OperationalHigh": 160.6375,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 5262,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "100 msec",
            "Units": "%"
        },
        5263: {
            "Acronym": "EEC9",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 31,
            "Name": "Engine Exhaust Gas Recirculation 2 Valve 2 Position",
            "Offset": 0.0,
            "OperationalHigh": 160.6375,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 5263,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "100 msec",
            "Units": "%"
        },
        5264: {
            "Acronym": "EEC11",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 15,
            "Name": "Engine Exhaust Gas Recirculation 2 (EGR2) Valve Control",
            "Offset": 0.0,
            "OperationalHigh": 160.6375,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 5264,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        5265: {
            "Acronym": "EEC11",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 31,
            "Name": "Engine Exhaust Gas Recirculation 2 (EGR2) Valve 2 Control",
            "Offset": 0.0,
            "OperationalHigh": 160.6375,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 5265,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        5266: {
            "Acronym": "DPF1S2",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 15,
            "Name": "DPF 1 Soot Signal Standard Deviation",
            "Offset": 0.0,
            "OperationalHigh": 160.6375,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 5266,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "1 sec",
            "Units": "%"
        },
        5267: {
            "Acronym": "DPF1S2",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 31,
            "Name": "DPF 1 Soot Signal Maximum",
            "Offset": 0.0,
            "OperationalHigh": 160.6375,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 5267,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "1 sec",
            "Units": "%"
        },
        5268: {
            "Acronym": "DPF1S2",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 47,
            "Name": "DPF 1 Soot Signal Minimum",
            "Offset": 0.0,
            "OperationalHigh": 160.6375,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 5268,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "1 sec",
            "Units": "%"
        },
        5269: {
            "Acronym": "DPF2S2",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 15,
            "Name": "DPF 2 Soot Signal Standard Deviation",
            "Offset": 0.0,
            "OperationalHigh": 160.6375,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 5269,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "1 sec",
            "Units": "%"
        },
        527: {
            "Acronym": "CCVS1",
            "DataRange": "0 to 7",
            "EndBit": 55,
            "Name": "Cruise Control States",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 527,
            "SPNLength": 3,
            "StartBit": 53,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5270: {
            "Acronym": "DPF2S2",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 31,
            "Name": "DPF 2 Soot Signal Maximum",
            "Offset": 0.0,
            "OperationalHigh": 160.6375,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 5270,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "1 sec",
            "Units": "%"
        },
        5271: {
            "Acronym": "DPF2S2",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 47,
            "Name": "DPF 2 Soot Signal Minimum",
            "Offset": 0.0,
            "OperationalHigh": 160.6375,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 5271,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "1 sec",
            "Units": "%"
        },
        5274: {
            "Acronym": "ASC2",
            "DataRange": "0 to 3",
            "EndBit": 59,
            "Name": "Prohibit Automatic Air Suspension Control",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5274,
            "SPNLength": 2,
            "StartBit": 58,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5275: {
            "Acronym": "EBC5",
            "DataRange": "0 to 3",
            "EndBit": 25,
            "Name": "Parking Brake Actuator Fully Activated",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5275,
            "SPNLength": 2,
            "StartBit": 24,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5276: {
            "Acronym": "EMAC",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 15,
            "Name": "Engine Exhaust Manifold Bank 1 Flow Balance Valve Actuator Control",
            "Offset": 0.0,
            "OperationalHigh": 160.6375,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 5276,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        5277: {
            "Acronym": "EMAP",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 15,
            "Name": "Engine Exhaust Manifold Bank 1 Flow Balance Valve Actuator Position",
            "Offset": 0.0,
            "OperationalHigh": 160.6375,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 5277,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        5278: {
            "Acronym": "EMAC",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 31,
            "Name": "Engine Exhaust Manifold Bank 2 Flow Balance Valve Actuator Control",
            "Offset": 0.0,
            "OperationalHigh": 160.6375,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 5278,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        5279: {
            "Acronym": "EMAP",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 31,
            "Name": "Engine Exhaust Manifold Bank 2 Flow Balance Valve Actuator Position",
            "Offset": 0.0,
            "OperationalHigh": 160.6375,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 5279,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        528: {
            "Acronym": "EC1",
            "DataRange": "0 to 8,031.875 rpm",
            "EndBit": 39,
            "Name": "Engine Speed At Point 2 (Engine Configuration)",
            "Offset": 0.0,
            "OperationalHigh": 8031.875,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 39,
            "Resolution": 0.125,
            "SPN": 528,
            "SPNLength": 16,
            "StartBit": 24,
            "TransmissionRate": "Every 5 s and on change of torque/speed points of more than 10% since last transmission but no faster than every 500 ms",
            "Units": "rpm"
        },
        5280: {
            "Acronym": "CAC1P",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 15,
            "Name": "Engine Charge Air Cooler 1 Precooler Intake Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 5280,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "1 sec",
            "Units": "deg F"
        },
        5281: {
            "Acronym": "CAC1P",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 31,
            "Name": "Engine Charge Air Cooler 1 Precooler Outlet Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 5281,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "1 sec",
            "Units": "deg F"
        },
        5282: {
            "Acronym": "CAC1P",
            "DataRange": "0 to 100 %",
            "EndBit": 39,
            "Name": "Engine Charge Air Cooler 1 Precooler Efficiency",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 5282,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "1 sec",
            "Units": "%"
        },
        5283: {
            "Acronym": "CAC1",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 15,
            "Name": "Engine Charge Air Cooler 1 Intake Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 5283,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "1 sec",
            "Units": "deg F"
        },
        5284: {
            "Acronym": "CAC1",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 31,
            "Name": "Engine Charge Air Cooler 1 Ambient Air Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 5284,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "1 sec",
            "Units": "deg F"
        },
        5285: {
            "Acronym": "CAC1",
            "DataRange": "0 to 100 %",
            "EndBit": 39,
            "Name": "Engine Charge Air Cooler 1 Efficiency",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 5285,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "1 sec",
            "Units": "%"
        },
        5286: {
            "Acronym": "CAC2P",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 15,
            "Name": "Engine Charge Air Cooler 2 Precooler Intake Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 5286,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "1 sec",
            "Units": "deg F"
        },
        5287: {
            "Acronym": "CAC2P",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 31,
            "Name": "Engine Charge Air Cooler 2 Precooler Outlet Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 5287,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "1 sec",
            "Units": "deg F"
        },
        5288: {
            "Acronym": "CAC2P",
            "DataRange": "0 to 100 %",
            "EndBit": 39,
            "Name": "Engine Charge Air Cooler 2 Precooler Efficiency",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 5288,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "1 sec",
            "Units": "%"
        },
        5289: {
            "Acronym": "CAC2",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 15,
            "Name": "Engine Charge Air Cooler 2 Intake Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 5289,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "1 sec",
            "Units": "deg F"
        },
        529: {
            "Acronym": "EC1",
            "DataRange": "0 to 8,031.875 rpm",
            "EndBit": 63,
            "Name": "Engine Speed At Point 3 (Engine Configuration)",
            "Offset": 0.0,
            "OperationalHigh": 8031.875,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 39,
            "Resolution": 0.125,
            "SPN": 529,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "Every 5 s and on change of torque/speed points of more than 10% since last transmission but no faster than every 500 ms",
            "Units": "rpm"
        },
        5290: {
            "Acronym": "CAC2",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 31,
            "Name": "Engine Charge Air Cooler 2 Outlet Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 5290,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "1 sec",
            "Units": "deg F"
        },
        5291: {
            "Acronym": "CAC2",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 47,
            "Name": "Engine Charge Air Cooler 2 Ambient Air Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 5291,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "1 sec",
            "Units": "deg F"
        },
        5292: {
            "Acronym": "CAC2",
            "DataRange": "0 to 100 %",
            "EndBit": 55,
            "Name": "Engine Charge Air Cooler 2 Efficiency",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 5292,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "1 sec",
            "Units": "%"
        },
        5293: {
            "Acronym": "ASC2",
            "DataRange": "0 to 3",
            "EndBit": 55,
            "Name": "Allow Level Control During Braking Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5293,
            "SPNLength": 2,
            "StartBit": 54,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5294: {
            "Acronym": "ASC1",
            "DataRange": "0 to 3",
            "EndBit": 53,
            "Name": "Allow Level Control During Braking Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5294,
            "SPNLength": 2,
            "StartBit": 52,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5295: {
            "Acronym": "ASC2",
            "DataRange": "0 to 3",
            "EndBit": 57,
            "Name": "Speed Dependent Level Control Request",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5295,
            "SPNLength": 2,
            "StartBit": 56,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5296: {
            "Acronym": "ASC1",
            "DataRange": "0 to 3",
            "EndBit": 55,
            "Name": "Speed Dependant Level Control Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5296,
            "SPNLength": 2,
            "StartBit": 54,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        53: {
            "Acronym": "ETC4",
            "DataRange": "0 to 100 %",
            "EndBit": 7,
            "Name": "Transmission Synchronizer Clutch Value",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 53,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        530: {
            "Acronym": "EC1",
            "DataRange": "0 to 8,031.875 rpm",
            "EndBit": 87,
            "Name": "Engine Speed At Point 4 (Engine Configuration)",
            "Offset": 0.0,
            "OperationalHigh": 8031.875,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 39,
            "Resolution": 0.125,
            "SPN": 530,
            "SPNLength": 16,
            "StartBit": 72,
            "TransmissionRate": "Every 5 s and on change of torque/speed points of more than 10% since last transmission but no faster than every 500 ms",
            "Units": "rpm"
        },
        531: {
            "Acronym": "EC1",
            "DataRange": "0 to 8,031.875 rpm",
            "EndBit": 111,
            "Name": "Engine Speed At Point 5 (Engine Configuration)",
            "Offset": 0.0,
            "OperationalHigh": 8031.875,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 39,
            "Resolution": 0.125,
            "SPN": 531,
            "SPNLength": 16,
            "StartBit": 96,
            "TransmissionRate": "Every 5 s and on change of torque/speed points of more than 10% since last transmission but no faster than every 500 ms",
            "Units": "rpm"
        },
        5312: {
            "Acronym": "EEC7",
            "DataRange": "0.0 to 1164.92498189875 psi",
            "EndBit": 63,
            "Name": "Commanded Engine Intake Manifold Pressure",
            "Offset": 0.0,
            "OperationalHigh": 1164.92498189875,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.01812971725,
            "SPN": 5312,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        5313: {
            "Acronym": "EEC9",
            "DataRange": "0.0 to 36404.472237999995 psi",
            "EndBit": 47,
            "Name": "Commanded Engine Fuel Rail Pressure",
            "Offset": 0.0,
            "OperationalHigh": 36404.472237999995,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.5665536640625,
            "SPN": 5313,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "100 msec",
            "Units": "psi"
        },
        5314: {
            "Acronym": "EEC9",
            "DataRange": "0.0 to 36404.472237999995 psi",
            "EndBit": 63,
            "Name": "Commanded Engine Fuel Injection Control Pressure",
            "Offset": 0.0,
            "OperationalHigh": 36404.472237999995,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.5665536640625,
            "SPN": 5314,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "100 msec",
            "Units": "psi"
        },
        5315: {
            "Acronym": "AT2WUDOC",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 15,
            "Name": "Aftertreatment 2 Warm Up Diesel Oxidation Catalyst Intake Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 5315,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "500 ms",
            "Units": "deg F"
        },
        5316: {
            "Acronym": "AT2WUDOC",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 31,
            "Name": "Aftertreatment 2 Warm Up Diesel Oxidation Catalyst Outlet Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 5316,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "500 ms",
            "Units": "deg F"
        },
        532: {
            "Acronym": "EC1",
            "DataRange": "0 to 8,031.875 rpm",
            "EndBit": 135,
            "Name": "Engine Speed At High Idle, Point 6 (Engine Configuration)",
            "Offset": 0.0,
            "OperationalHigh": 8031.875,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 39,
            "Resolution": 0.125,
            "SPN": 532,
            "SPNLength": 16,
            "StartBit": 120,
            "TransmissionRate": "Every 5 s and on change of torque/speed points of more than 10% since last transmission but no faster than every 500 ms",
            "Units": "rpm"
        },
        5323: {
            "Acronym": "EEC5",
            "DataRange": "0 to 3",
            "EndBit": 51,
            "Name": "Engine Fuel Control Control Mode",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5323,
            "SPNLength": 2,
            "StartBit": 50,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        533: {
            "Acronym": "EC1",
            "DataRange": "0 to 8,031.875 rpm",
            "EndBit": 183,
            "Name": "Engine Maximum Momentary Override Speed, Point 7 (Engine Configuration)",
            "Offset": 0.0,
            "OperationalHigh": 8031.875,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 39,
            "Resolution": 0.125,
            "SPN": 533,
            "SPNLength": 16,
            "StartBit": 168,
            "TransmissionRate": "Every 5 s and on change of torque/speed points of more than 10% since last transmission but no faster than every 500 ms",
            "Units": "rpm"
        },
        534: {
            "Acronym": "EC1",
            "DataRange": "0 to 25 s",
            "EndBit": 191,
            "Name": "Engine Maximum Momentary Override Time Limit (Engine Configuration)",
            "Offset": 0.0,
            "OperationalHigh": 25.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 39,
            "Resolution": 0.1,
            "SPN": 534,
            "SPNLength": 8,
            "StartBit": 184,
            "TransmissionRate": "Every 5 s and on change of torque/speed points of more than 10% since last transmission but no faster than every 500 ms",
            "Units": "Seconds"
        },
        5344: {
            "Acronym": "ETC7",
            "DataRange": "0 to 3",
            "EndBit": 43,
            "Name": "Transmission Warning Indicator",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5344,
            "SPNLength": 2,
            "StartBit": 42,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5345: {
            "Acronym": "TRF2",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Transmission Overheat Indicator",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5345,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        5346: {
            "Acronym": "RF",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Driveline Retarder Overheat Indicator",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5346,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        5347: {
            "Acronym": "ACCS",
            "DataRange": "-320 to +322.55 m/s^2",
            "EndBit": 15,
            "Name": "Lateral Acceleration Extended Range",
            "Offset": -320.0,
            "OperationalHigh": 322.552,
            "OperationalLow": -320.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.01,
            "SPN": 5347,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "10 ms (default) or 20 ms",
            "Units": "m/s^2"
        },
        5348: {
            "Acronym": "ACCS",
            "DataRange": "-320 to +322.55 m/s^2",
            "EndBit": 31,
            "Name": "Longitudinal Acceleration Extended Range",
            "Offset": -320.0,
            "OperationalHigh": 322.552,
            "OperationalLow": -320.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.01,
            "SPN": 5348,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "10 ms (default) or 20 ms",
            "Units": "m/s^2"
        },
        5349: {
            "Acronym": "ACCS",
            "DataRange": "-320 to +322.55 m/s^2",
            "EndBit": 47,
            "Name": "Vertical Acceleration Extended Range",
            "Offset": -320.0,
            "OperationalHigh": 322.552,
            "OperationalLow": -320.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.01,
            "SPN": 5349,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "10 ms (default) or 20 ms",
            "Units": "m/s^2"
        },
        535: {
            "Acronym": "EC1",
            "DataRange": "0 to 2,500 rpm",
            "EndBit": 199,
            "Name": "Engine Requested Speed Control Range Lower Limit (Engine Configuration)",
            "Offset": 0.0,
            "OperationalHigh": 2500.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 39,
            "Resolution": 10,
            "SPN": 535,
            "SPNLength": 8,
            "StartBit": 192,
            "TransmissionRate": "Every 5 s and on change of torque/speed points of more than 10% since last transmission but no faster than every 500 ms",
            "Units": "rpm"
        },
        5350: {
            "Acronym": "ACCS",
            "DataRange": "0 to 3",
            "EndBit": 49,
            "Name": "Lateral Acceleration Extended Range Figure of Merit",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5350,
            "SPNLength": 2,
            "StartBit": 48,
            "TransmissionRate": "10 ms (default) or 20 ms",
            "Units": "bit"
        },
        5351: {
            "Acronym": "ACCS",
            "DataRange": "0 to 3",
            "EndBit": 51,
            "Name": "Longitudinal Acceleration Extended Range Figure of Merit",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5351,
            "SPNLength": 2,
            "StartBit": 50,
            "TransmissionRate": "10 ms (default) or 20 ms",
            "Units": "bit"
        },
        5352: {
            "Acronym": "ACCS",
            "DataRange": "0 to 3",
            "EndBit": 53,
            "Name": "Vertical Acceleration Extended Range Figure of Merit",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5352,
            "SPNLength": 2,
            "StartBit": 52,
            "TransmissionRate": "10 ms (default) or 20 ms",
            "Units": "bit"
        },
        5353: {
            "Acronym": "ACCS",
            "DataRange": "0 to 3",
            "EndBit": 55,
            "Name": "Support Variable Transmission Repetition Rate for Acceleration Sensor",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5353,
            "SPNLength": 2,
            "StartBit": 54,
            "TransmissionRate": "10 ms (default) or 20 ms",
            "Units": "bit"
        },
        536: {
            "Acronym": "EC1",
            "DataRange": "0 to 2,500 rpm",
            "EndBit": 207,
            "Name": "Engine Requested Speed Control Range Upper Limit (Engine Configuration)",
            "Offset": 0.0,
            "OperationalHigh": 2500.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 39,
            "Resolution": 10,
            "SPN": 536,
            "SPNLength": 8,
            "StartBit": 200,
            "TransmissionRate": "Every 5 s and on change of torque/speed points of more than 10% since last transmission but no faster than every 500 ms",
            "Units": "rpm"
        },
        5366: {
            "Acronym": "EEC6",
            "DataRange": "0 to 100 %",
            "EndBit": 55,
            "Name": "Engine Desired Turbocharger Compressor Bypass Actuator 1 Position",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 5366,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "100 msec (preferred) or Engine Speed Dependent (if required by application)",
            "Units": "%"
        },
        5367: {
            "Acronym": "EEC6",
            "DataRange": "0 to 31",
            "EndBit": 60,
            "Name": "Engine Turbocharger Compressor Bypass Actuator 1 Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 5367,
            "SPNLength": 5,
            "StartBit": 56,
            "TransmissionRate": "100 msec (preferred) or Engine Speed Dependent (if required by application)",
            "Units": "binary"
        },
        5368: {
            "Acronym": "EEC6",
            "DataRange": "0 to 7",
            "EndBit": 63,
            "Name": "Engine Turbocharger Compressor Bypass Actuator 1 Temperature Status",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5368,
            "SPNLength": 3,
            "StartBit": 61,
            "TransmissionRate": "100 msec (preferred) or Engine Speed Dependent (if required by application)",
            "Units": "bit"
        },
        5369: {
            "Acronym": "EEC6",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 47,
            "Name": "Engine Turbocharger Compressor Bypass Actuator 2 command",
            "Offset": 0.0,
            "OperationalHigh": 160.6375,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 5369,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "100 msec (preferred) or Engine Speed Dependent (if required by application)",
            "Units": "%"
        },
        537: {
            "Acronym": "EC1",
            "DataRange": "-125 to 125 %",
            "EndBit": 215,
            "Name": "Engine Requested Torque Control Range Lower Limit (Engine Configuration)",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 125%",
            "PGNLength": 39,
            "Resolution": 1,
            "SPN": 537,
            "SPNLength": 8,
            "StartBit": 208,
            "TransmissionRate": "Every 5 s and on change of torque/speed points of more than 10% since last transmission but no faster than every 500 ms",
            "Units": "%"
        },
        5370: {
            "Acronym": "TCW",
            "DataRange": "0 to 100 %",
            "EndBit": 47,
            "Name": "Engine Desired Turbocharger Wastegate Actuator 1 Position",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 5370,
            "SPNLength": 8,
            "StartBit": 40,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        5371: {
            "Acronym": "TCW",
            "DataRange": "0 to 31",
            "EndBit": 52,
            "Name": "Engine Turbocharger Wastegate Actuator 1 Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 5371,
            "SPNLength": 5,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "binary"
        },
        5372: {
            "Acronym": "TCW",
            "DataRange": "0 to 7",
            "EndBit": 55,
            "Name": "Engine Turbocharger Wastegate Actuator 1 Temperature Status",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5372,
            "SPNLength": 3,
            "StartBit": 53,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5373: {
            "Acronym": "TCW",
            "DataRange": "0 to 100 %",
            "EndBit": 63,
            "Name": "Engine Desired Turbocharger Wastegate Actuator 2 Position",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 5373,
            "SPNLength": 8,
            "StartBit": 56,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        5374: {
            "Acronym": "EFTVI",
            "DataRange": "0 to 100 %",
            "EndBit": 7,
            "Name": "Engine Desired Throttle Valve 1 Position ",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 5374,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        5375: {
            "Acronym": "EFTVI",
            "DataRange": "0 to 31",
            "EndBit": 12,
            "Name": "Engine Throttle Valve 1 Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 5375,
            "SPNLength": 5,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "binary"
        },
        5376: {
            "Acronym": "EFTVI",
            "DataRange": "0 to 7",
            "EndBit": 15,
            "Name": "Engine Throttle Valve 1 Temperature Status",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5376,
            "SPNLength": 3,
            "StartBit": 13,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5377: {
            "Acronym": "EFTVI",
            "DataRange": "0 to 100 %",
            "EndBit": 23,
            "Name": "Engine Desired Throttle Valve 2 Position",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 5377,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        5378: {
            "Acronym": "EFTVI",
            "DataRange": "0 to 31",
            "EndBit": 28,
            "Name": "Engine Throttle Valve 2 Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 5378,
            "SPNLength": 5,
            "StartBit": 24,
            "TransmissionRate": "100 ms",
            "Units": "binary"
        },
        5379: {
            "Acronym": "EFTVI",
            "DataRange": "0 to 7",
            "EndBit": 31,
            "Name": "Engine Throttle Valve 2 Temperature status",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5379,
            "SPNLength": 3,
            "StartBit": 29,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        538: {
            "Acronym": "EC1",
            "DataRange": "-125 to 125 %",
            "EndBit": 223,
            "Name": "Engine Requested Torque Control Range Upper Limit (Engine Configuration)",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 125%",
            "PGNLength": 39,
            "Resolution": 1,
            "SPN": 538,
            "SPNLength": 8,
            "StartBit": 216,
            "TransmissionRate": "Every 5 s and on change of torque/speed points of more than 10% since last transmission but no faster than every 500 ms",
            "Units": "%"
        },
        5380: {
            "Acronym": "EFTVI",
            "DataRange": "0 to 31",
            "EndBit": 36,
            "Name": "Engine Fuel Valve 1 Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 5380,
            "SPNLength": 5,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "binary"
        },
        5381: {
            "Acronym": "EFTVI",
            "DataRange": "0 to 7",
            "EndBit": 39,
            "Name": "Engine Fuel Valve 1 Temperature Status",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5381,
            "SPNLength": 3,
            "StartBit": 37,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5382: {
            "Acronym": "EFTVI",
            "DataRange": "0 to 31",
            "EndBit": 44,
            "Name": "Engine Fuel Valve 2 Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 5382,
            "SPNLength": 5,
            "StartBit": 40,
            "TransmissionRate": "100 ms",
            "Units": "binary"
        },
        5383: {
            "Acronym": "EFTVI",
            "DataRange": "0 to 7",
            "EndBit": 47,
            "Name": "Engine Fuel Valve 2 Temperature Status",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5383,
            "SPNLength": 3,
            "StartBit": 45,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5384: {
            "Acronym": "ETWAI",
            "DataRange": "0 to 31",
            "EndBit": 4,
            "Name": "Engine Turbocharger Wastegate Actuator 2 Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 5384,
            "SPNLength": 5,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "binary"
        },
        5385: {
            "Acronym": "ETWAI",
            "DataRange": "0 to 7",
            "EndBit": 7,
            "Name": "Engine Turbocharger Wastegate Actuator 2 Temperature Status",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5385,
            "SPNLength": 3,
            "StartBit": 5,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5386: {
            "Acronym": "ETWAC",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 15,
            "Name": "Engine Turbocharger Wastegate Actuator 1 Command",
            "Offset": 0.0,
            "OperationalHigh": 160.6375,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 5386,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "Engine speed dependent",
            "Units": "%"
        },
        5387: {
            "Acronym": "ETWAC",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 31,
            "Name": "Engine Turbocharger Wastegate Actuator 2 Command",
            "Offset": 0.0,
            "OperationalHigh": 160.6375,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 5387,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "Engine speed dependent",
            "Units": "%"
        },
        5388: {
            "Acronym": "ETCBI",
            "DataRange": "0 to 100 %",
            "EndBit": 7,
            "Name": "Engine Turbocharger Compressor Bypass Actuator 2 Position",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 5388,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        5389: {
            "Acronym": "ETCBI",
            "DataRange": "0 to 100 %",
            "EndBit": 15,
            "Name": "Engine Desired Turbocharger Compressor Bypass Actuator 2 Position",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 5389,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        539: {
            "Acronym": "EC1",
            "DataRange": "-125 to 125 %",
            "EndBit": 23,
            "Name": "Engine Percent Torque At Idle, Point 1 (Engine Configuration)",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 125%",
            "PGNLength": 39,
            "Resolution": 1,
            "SPN": 539,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "Every 5 s and on change of torque/speed points of more than 10% since last transmission but no faster than every 500 ms",
            "Units": "%"
        },
        5390: {
            "Acronym": "ETCBI",
            "DataRange": "0 to 31",
            "EndBit": 20,
            "Name": "Engine Turbocharger compressor Bypass Actuator 2 Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 5390,
            "SPNLength": 5,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "binary"
        },
        5391: {
            "Acronym": "ETCBI",
            "DataRange": "0 to 7",
            "EndBit": 23,
            "Name": "Engine Turbocharger Compressor Bypass Actuator 2 Temperature Status",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5391,
            "SPNLength": 3,
            "StartBit": 21,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5398: {
            "Acronym": "EEC2",
            "DataRange": "-125 to 125 %",
            "EndBit": 63,
            "Name": "Estimated Pumping - Percent Torque",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5398,
            "SPNLength": 8,
            "StartBit": 56,
            "TransmissionRate": "50 msec (preferred) or Engine Speed Dependent (if required by application)",
            "Units": "%"
        },
        5399: {
            "Acronym": "EEC2",
            "DataRange": "0 to 3",
            "EndBit": 45,
            "Name": "DPF Thermal Management Active",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5399,
            "SPNLength": 2,
            "StartBit": 44,
            "TransmissionRate": "50 msec (preferred) or Engine Speed Dependent (if required by application)",
            "Units": "bit"
        },
        54: {
            "Acronym": "ETC4",
            "DataRange": "0 to 100 %",
            "EndBit": 15,
            "Name": "Transmission Synchronizer Brake Value",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 54,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        540: {
            "Acronym": "EC1",
            "DataRange": "-125 to 125 %",
            "EndBit": 47,
            "Name": "Engine Percent Torque At Point 2 (Engine Configuration)",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 125%",
            "PGNLength": 39,
            "Resolution": 1,
            "SPN": 540,
            "SPNLength": 8,
            "StartBit": 40,
            "TransmissionRate": "Every 5 s and on change of torque/speed points of more than 10% since last transmission but no faster than every 500 ms",
            "Units": "%"
        },
        5400: {
            "Acronym": "EEC2",
            "DataRange": "0 to 3",
            "EndBit": 47,
            "Name": "SCR Thermal Management Active",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5400,
            "SPNLength": 2,
            "StartBit": 46,
            "TransmissionRate": "50 msec (preferred) or Engine Speed Dependent (if required by application)",
            "Units": "bit"
        },
        5402: {
            "Acronym": "MSI2",
            "DataRange": "0 to 3",
            "EndBit": 35,
            "Name": "Lift Relay Control",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5402,
            "SPNLength": 2,
            "StartBit": 34,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        5403: {
            "Acronym": "MSI2",
            "DataRange": "0 to 3",
            "EndBit": 37,
            "Name": "Drop Relay Control",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5403,
            "SPNLength": 2,
            "StartBit": 36,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        5404: {
            "Acronym": "SHUTDN",
            "DataRange": "0 to 3",
            "EndBit": 59,
            "Name": "PTO Shutdown has Shutdown Engine",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5404,
            "SPNLength": 2,
            "StartBit": 58,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        5405: {
            "Acronym": "LOI",
            "DataRange": "0 to 3",
            "EndBit": 37,
            "Name": "Set Elevation Sensor #1 Reference Elevation",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5405,
            "SPNLength": 2,
            "StartBit": 36,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5406: {
            "Acronym": "LOI",
            "DataRange": "0 to 3",
            "EndBit": 39,
            "Name": "Set Elevation Sensor #2 Reference Elevation",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5406,
            "SPNLength": 2,
            "StartBit": 38,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5407: {
            "Acronym": "LBC",
            "DataRange": "0 to 15",
            "EndBit": 27,
            "Name": "Blade Control Mode - Left",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5407,
            "SPNLength": 4,
            "StartBit": 24,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        5408: {
            "Acronym": "LBC",
            "DataRange": "0 to 15",
            "EndBit": 31,
            "Name": "Blade Control Mode - Right",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5408,
            "SPNLength": 4,
            "StartBit": 28,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        5409: {
            "Acronym": "LBC",
            "DataRange": "0 to 3",
            "EndBit": 33,
            "Name": "Land Leveling System Enable Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5409,
            "SPNLength": 2,
            "StartBit": 32,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        541: {
            "Acronym": "EC1",
            "DataRange": "-125 to 125 %",
            "EndBit": 71,
            "Name": "Engine Percent Torque At Point 3 (Engine Configuration)",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 125%",
            "PGNLength": 39,
            "Resolution": 1,
            "SPN": 541,
            "SPNLength": 8,
            "StartBit": 64,
            "TransmissionRate": "Every 5 s and on change of torque/speed points of more than 10% since last transmission but no faster than every 500 ms",
            "Units": "%"
        },
        5410: {
            "Acronym": "LLRE",
            "DataRange": "-32000 to 32255 mm (-32.000 to 32.255 m)",
            "EndBit": 15,
            "Name": "Blade Elevation Deviation - Left",
            "Offset": -32000.0,
            "OperationalHigh": 3225532.0,
            "OperationalLow": -32000.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5410,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "200 ms",
            "Units": "mm"
        },
        5411: {
            "Acronym": "LLRE",
            "DataRange": "-32000 to 32255 mm (-32.000 to 32.255 m)",
            "EndBit": 31,
            "Name": "Blade Elevation Deviation - Right",
            "Offset": -32000.0,
            "OperationalHigh": 3225532.0,
            "OperationalLow": -32000.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5411,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "200 ms",
            "Units": "mm"
        },
        5412: {
            "Acronym": "LLRE",
            "DataRange": "-32000 to 32255 mm (-32.000 to 32.255 m)",
            "EndBit": 47,
            "Name": "Blade Reference Elevation Offset - Left",
            "Offset": -32000.0,
            "OperationalHigh": 3225532.0,
            "OperationalLow": -32000.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5412,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "200 ms",
            "Units": "mm"
        },
        5413: {
            "Acronym": "LLRE",
            "DataRange": "-32000 to 32255 mm (-32.000 to 32.255 m)",
            "EndBit": 63,
            "Name": "Blade Reference Elevation Offset - Right",
            "Offset": -32000.0,
            "OperationalHigh": 3225532.0,
            "OperationalLow": -32000.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5413,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "200 ms",
            "Units": "mm"
        },
        5414: {
            "Acronym": "A1SCRDSR2",
            "DataRange": "0 to 100 %",
            "EndBit": 31,
            "Name": "Aftertreatment 1 Diesel Exhaust Fluid Tank 2 Heater Command",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 5414,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "500 ms",
            "Units": "%"
        },
        5415: {
            "Acronym": "A2SCRDSR2",
            "DataRange": "0 to 100 %",
            "EndBit": 31,
            "Name": "Aftertreatment 2 Diesel Exhaust Fluid Tank 2 Heater Command",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 5415,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "500 ms",
            "Units": "%"
        },
        5416: {
            "Acronym": "DLCD1",
            "DataRange": "0 to 3",
            "EndBit": 63,
            "Name": "Engine Wait To Start Lamp Data",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5416,
            "SPNLength": 2,
            "StartBit": 62,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        5417: {
            "Acronym": "EFS2",
            "DataRange": "0.0 to 72.518869 psi",
            "EndBit": 7,
            "Name": "Engine Fuel Filter (Suction Side) Intake Absolute Pressure",
            "Offset": 0.0,
            "OperationalHigh": 72.518869,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.290075476,
            "SPN": 5417,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        542: {
            "Acronym": "EC1",
            "DataRange": "-125 to 125 %",
            "EndBit": 95,
            "Name": "Engine Percent Torque At Point 4 (Engine Configuration)",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 125%",
            "PGNLength": 39,
            "Resolution": 1,
            "SPN": 542,
            "SPNLength": 8,
            "StartBit": 88,
            "TransmissionRate": "Every 5 s and on change of torque/speed points of more than 10% since last transmission but no faster than every 500 ms",
            "Units": "%"
        },
        5422: {
            "Acronym": "IC2",
            "DataRange": "0.0 to 72.518869 psi",
            "EndBit": 63,
            "Name": "Engine Intake Manifold 2 Absolute Pressure",
            "Offset": 0.0,
            "OperationalHigh": 72.518869,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.290075476,
            "SPN": 5422,
            "SPNLength": 8,
            "StartBit": 56,
            "TransmissionRate": "0.5 s",
            "Units": "psi"
        },
        5423: {
            "Acronym": "AT1FC2",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Aftertreatment 1 Fuel Pump Relay Control",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5423,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        5424: {
            "Acronym": "AT1FC2",
            "DataRange": "0 to 3",
            "EndBit": 19,
            "Name": "Aftertreatment 1 Fuel Flow Diverter Valve Control",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5424,
            "SPNLength": 2,
            "StartBit": 18,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        5425: {
            "Acronym": "AT1FC2",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 39,
            "Name": "Aftertreatment 1 Fuel Pressure 2 Actuator Control",
            "Offset": 0.0,
            "OperationalHigh": 160.6375,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 5425,
            "SPNLength": 16,
            "StartBit": 24,
            "TransmissionRate": "500 ms",
            "Units": "%"
        },
        5426: {
            "Acronym": "AT2FC2",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Aftertreatment 2 Fuel Pump Relay Control",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5426,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        5427: {
            "Acronym": "AT2FC2",
            "DataRange": "0 to 3",
            "EndBit": 19,
            "Name": "Aftertreatment 2 Fuel Flow Diverter Valve Control",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5427,
            "SPNLength": 2,
            "StartBit": 18,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        5428: {
            "Acronym": "AT2FC2",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 39,
            "Name": "Aftertreatment 2 Fuel Pressure 2 Control",
            "Offset": 0.0,
            "OperationalHigh": 160.6375,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 5428,
            "SPNLength": 16,
            "StartBit": 24,
            "TransmissionRate": "500 ms",
            "Units": "%"
        },
        5429: {
            "Acronym": "EFL/P10",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 55,
            "Name": "Engine Exhaust Gas Recirculation 2 Intake Absolute Pressure",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 5429,
            "SPNLength": 16,
            "StartBit": 40,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        543: {
            "Acronym": "EC1",
            "DataRange": "-125 to 125 %",
            "EndBit": 119,
            "Name": "Engine Percent Torque At Point 5 (Engine Configuration)",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 125%",
            "PGNLength": 39,
            "Resolution": 1,
            "SPN": 543,
            "SPNLength": 8,
            "StartBit": 112,
            "TransmissionRate": "Every 5 s and on change of torque/speed points of more than 10% since last transmission but no faster than every 500 ms",
            "Units": "%"
        },
        5430: {
            "Acronym": "EFL/P11",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 15,
            "Name": "Engine Exhaust Gas Recirculation 1 Intake Absolute Pressure",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 5430,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        5431: {
            "Acronym": "EFL/P11",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 31,
            "Name": "Engine Exhaust Gas Recirculation 1 Outlet Absolute Pressure (High Resolution)",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 5431,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        5432: {
            "Acronym": "ASC1",
            "DataRange": "0 to 15",
            "EndBit": 63,
            "Name": "Memory level",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5432,
            "SPNLength": 4,
            "StartBit": 60,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5433: {
            "Acronym": "ASC2",
            "DataRange": "0 to 15",
            "EndBit": 63,
            "Name": "Memory level request",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5433,
            "SPNLength": 4,
            "StartBit": 60,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5434: {
            "Acronym": "A1DEFSI",
            "DataRange": "0 to 100 %",
            "EndBit": 39,
            "Name": "Aftertreatment 1 Diesel Exhaust Fluid Tank Fill Valve Command",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 5434,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "1 s",
            "Units": "%"
        },
        5435: {
            "Acronym": "A1DEFSI",
            "DataRange": "0 to 3",
            "EndBit": 41,
            "Name": "Aftertreatment 1 Diesel Exhaust Fluid Pump State",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5435,
            "SPNLength": 2,
            "StartBit": 40,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        5436: {
            "Acronym": "A1DEFSI",
            "DataRange": "0 to 100 %",
            "EndBit": 55,
            "Name": "Aftertreatment 1 Diesel Exhaust Fluid Tank Drain Valve Command",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 5436,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "1 s",
            "Units": "%"
        },
        5437: {
            "Acronym": "A2DEFSI",
            "DataRange": "0 to 100 %",
            "EndBit": 39,
            "Name": "Aftertreatment 2 Diesel Exhaust Fluid Tank Fill Valve Command",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 5437,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "1 s",
            "Units": "%"
        },
        5438: {
            "Acronym": "A2DEFSI",
            "DataRange": "0 to 3",
            "EndBit": 41,
            "Name": "Aftertreatment 2 Diesel Exhaust Fluid Pump State",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5438,
            "SPNLength": 2,
            "StartBit": 40,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        5439: {
            "Acronym": "A2DEFSI",
            "DataRange": "0 to 100 %",
            "EndBit": 55,
            "Name": "Aftertreatment 2 Diesel Exhaust Fluid Tank Drain Valve Command",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 5439,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "1 s",
            "Units": "%"
        },
        544: {
            "Acronym": "EC1",
            "DataRange": "0 to 64,255 Nm",
            "EndBit": 167,
            "Name": "Engine Reference Torque (Engine Configuration)",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 39,
            "Resolution": 1,
            "SPN": 544,
            "SPNLength": 16,
            "StartBit": 152,
            "TransmissionRate": "Every 5 s and on change of torque/speed points of more than 10% since last transmission but no faster than every 500 ms",
            "Units": "Nm"
        },
        5444: {
            "Acronym": "EEC7",
            "DataRange": "0 to 64255 rpm",
            "EndBit": 47,
            "Name": "Engine Crankcase Breather Oil Separator Speed",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5444,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "rpm"
        },
        5445: {
            "Acronym": "EFTVI",
            "DataRange": "0 to 15",
            "EndBit": 51,
            "Name": "Engine Throttle Valve 1 Operation Status",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5445,
            "SPNLength": 4,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5446: {
            "Acronym": "EFTVI",
            "DataRange": "0 to 15",
            "EndBit": 55,
            "Name": "Engine Throttle Valve 2 Operation Status",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5446,
            "SPNLength": 4,
            "StartBit": 52,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5447: {
            "Acronym": "EFTVI",
            "DataRange": "0 to 15",
            "EndBit": 59,
            "Name": "Engine Fuel Valve 1 Operation Status",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5447,
            "SPNLength": 4,
            "StartBit": 56,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5448: {
            "Acronym": "EFTVI",
            "DataRange": "0 to 15",
            "EndBit": 63,
            "Name": "Engine Fuel Valve 2 Operation Status",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5448,
            "SPNLength": 4,
            "StartBit": 60,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5449: {
            "Acronym": "ETCBI",
            "DataRange": "0 to 15",
            "EndBit": 27,
            "Name": "Engine Turbocharger Compressor Bypass Actuator 1 Operation Status",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5449,
            "SPNLength": 4,
            "StartBit": 24,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        545: {
            "Acronym": "EC1",
            "DataRange": "0 to 50.2 %/rpm",
            "EndBit": 151,
            "Name": "Engine Gain (Kp) Of The Endspeed Governor (Engine Configuration)",
            "Offset": 0.0,
            "OperationalHigh": 50.2,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 39,
            "Resolution": 1,
            "SPN": 545,
            "SPNLength": 16,
            "StartBit": 136,
            "TransmissionRate": "Every 5 s and on change of torque/speed points of more than 10% since last transmission but no faster than every 500 ms",
            "Units": "%/rpm"
        },
        5450: {
            "Acronym": "ETCBI",
            "DataRange": "0 to 15",
            "EndBit": 31,
            "Name": "Engine Turbocharger Compressor Bypass Actuator 2 Operation Status",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5450,
            "SPNLength": 4,
            "StartBit": 28,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5451: {
            "Acronym": "ETWAI",
            "DataRange": "0 to 15",
            "EndBit": 11,
            "Name": "Engine Turbocharger Wastegate Actuator 1 Operation Status",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5451,
            "SPNLength": 4,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5452: {
            "Acronym": "ETWAI",
            "DataRange": "0 to 15",
            "EndBit": 15,
            "Name": "Engine Turbocharger Wastegate Actuator 2 Operation Status",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5452,
            "SPNLength": 4,
            "StartBit": 12,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5454: {
            "Acronym": "AT1HI",
            "DataRange": "0 to 4,211,081,215 s",
            "EndBit": 287,
            "Name": "Aftertreatment 1 Average Time Between Active Regenerations",
            "Offset": 0.0,
            "OperationalHigh": 4211081215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 5454,
            "SPNLength": 32,
            "StartBit": 256,
            "TransmissionRate": "On request",
            "Units": "Seconds"
        },
        5455: {
            "Acronym": "AT2HI",
            "DataRange": "0 to 4,211,081,215 s",
            "EndBit": 287,
            "Name": "Aftertreatment 2 Average Time Between Active Regenerations",
            "Offset": 0.0,
            "OperationalHigh": 4211081215.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 5455,
            "SPNLength": 32,
            "StartBit": 256,
            "TransmissionRate": "On request",
            "Units": "Seconds"
        },
        5456: {
            "Acronym": "AT1FC2",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 47,
            "Name": "Aftertreatment 1 Hydrocarbon Doser Intake Fuel Temperature",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 5456,
            "SPNLength": 8,
            "StartBit": 40,
            "TransmissionRate": "500 ms",
            "Units": "deg F"
        },
        5457: {
            "Acronym": "EEC5",
            "DataRange": "0 to 3",
            "EndBit": 53,
            "Name": "Engine Variable Geometry Turbocharger 1 Control Mode",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5457,
            "SPNLength": 2,
            "StartBit": 52,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        5458: {
            "Acronym": "LFI",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 63,
            "Name": "Flexible Fuel Percentage",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 100%",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 5458,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        5459: {
            "Acronym": "AFT1NAI",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Aftertreatment 1 NOx Adsorber deNOx Regeneration Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5459,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        546: {
            "Acronym": "RC",
            "DataRange": "0 to 8,031.875 rpm",
            "EndBit": 31,
            "Name": "Retarder Speed At Idle, Point 1 (Retarder Configuration)",
            "Offset": 0.0,
            "OperationalHigh": 8031.875,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 19,
            "Resolution": 0.125,
            "SPN": 546,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "Every 5 s and on change of torque/speed points of more than 10% since last transmission but no faster than every 500 ms",
            "Units": "rpm"
        },
        5460: {
            "Acronym": "AFT2NAI",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Aftertreatment 2 NOx Adsorber deNOx Regeneration Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5460,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        5461: {
            "Acronym": "AFT1NAI",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Aftertreatment 1 NOx Adsorber deSOx Regeneration Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5461,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        5462: {
            "Acronym": "AFT2NAI",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Aftertreatment 2 NOx Adsorber deSOx Regeneration Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5462,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        5463: {
            "Acronym": "SCR1",
            "DataRange": "0 to 64,255 km",
            "EndBit": 55,
            "Name": "Aftertreatment 1 SCR Operator Inducement Active Traveled Distance",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5463,
            "SPNLength": 16,
            "StartBit": 40,
            "TransmissionRate": "On request",
            "Units": "km"
        },
        5464: {
            "Acronym": "VEP4",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 15,
            "Name": "Hybrid Battery Pack Remaining Charge",
            "Offset": 0.0,
            "OperationalHigh": 160.6375,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 5464,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        5465: {
            "Acronym": "EEC4",
            "DataRange": "0 to 3",
            "EndBit": 35,
            "Name": "Engine Intake Manifold Pressure Control Mode",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5465,
            "SPNLength": 2,
            "StartBit": 34,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        5466: {
            "Acronym": "AT1S",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 63,
            "Name": "Aftertreatment 1 DPF Soot Load Regeneration Threshold",
            "Offset": 0.0,
            "OperationalHigh": 160.6375,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 5466,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        5467: {
            "Acronym": "AT2S",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 63,
            "Name": "Aftertreatment 2 DPF Soot Load Regeneration Threshold",
            "Offset": 0.0,
            "OperationalHigh": 160.6375,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 5467,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        5468: {
            "Acronym": "EOM",
            "DataRange": "0 to 7.844",
            "EndBit": 47,
            "Name": "Engine Oil Relative Dielectricity (high resolution)",
            "Offset": 0.0,
            "OperationalHigh": 7.844,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5468,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "30 seconds",
            "Units": ""
        },
        5469: {
            "Acronym": "DLCC1",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Engine Fuel Filter Restricted Lamp Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5469,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms",
            "Units": "bit"
        },
        547: {
            "Acronym": "RC",
            "DataRange": "0 to 8,031.875 rpm",
            "EndBit": 127,
            "Name": "Retarder Speed At Peak Torque, Point 5 (Retarder Configuration)",
            "Offset": 0.0,
            "OperationalHigh": 8031.875,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 19,
            "Resolution": 0.125,
            "SPN": 547,
            "SPNLength": 16,
            "StartBit": 112,
            "TransmissionRate": "Every 5 s and on change of torque/speed points of more than 10% since last transmission but no faster than every 500 ms",
            "Units": "rpm"
        },
        5470: {
            "Acronym": "DLCD1",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Engine Fuel Filter Restricted Lamp Data",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5470,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        5471: {
            "Acronym": "EC3",
            "DataRange": "-125 to 125 %",
            "EndBit": 7,
            "Name": "Engine Friction Percent Torque At Idle, Point 1",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 125%",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5471,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "Every 5 s and on change of torque/speed points of more than 10% since last transmission but no faster than every 500 ms",
            "Units": "%"
        },
        5472: {
            "Acronym": "EC3",
            "DataRange": "-125 to 125 %",
            "EndBit": 15,
            "Name": "Engine Friction Percent Torque At Idle, Point 2",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 125%",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5472,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "Every 5 s and on change of torque/speed points of more than 10% since last transmission but no faster than every 500 ms",
            "Units": "%"
        },
        5473: {
            "Acronym": "EC3",
            "DataRange": "-125 to 125 %",
            "EndBit": 23,
            "Name": "Engine Friction Percent Torque At Idle, Point 3",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 125%",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5473,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "Every 5 s and on change of torque/speed points of more than 10% since last transmission but no faster than every 500 ms",
            "Units": "%"
        },
        5474: {
            "Acronym": "EC3",
            "DataRange": "-125 to 125 %",
            "EndBit": 31,
            "Name": "Engine Friction Percent Torque At Idle, Point 4",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 125%",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5474,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "Every 5 s and on change of torque/speed points of more than 10% since last transmission but no faster than every 500 ms",
            "Units": "%"
        },
        5475: {
            "Acronym": "EC3",
            "DataRange": "-125 to 125 %",
            "EndBit": 39,
            "Name": "Engine Friction Percent Torque At Idle, Point 5",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 125%",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5475,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "Every 5 s and on change of torque/speed points of more than 10% since last transmission but no faster than every 500 ms",
            "Units": "%"
        },
        5476: {
            "Acronym": "EC3",
            "DataRange": "-125 to 125 %",
            "EndBit": 47,
            "Name": "Engine Friction Percent Torque At Idle, Point 6",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 125%",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5476,
            "SPNLength": 8,
            "StartBit": 40,
            "TransmissionRate": "Every 5 s and on change of torque/speed points of more than 10% since last transmission but no faster than every 500 ms",
            "Units": "%"
        },
        5477: {
            "Acronym": "EC3",
            "DataRange": "-125 to 125 %",
            "EndBit": 55,
            "Name": "Engine Friction Percent Torque At Idle, Point 7",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 125%",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5477,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "Every 5 s and on change of torque/speed points of more than 10% since last transmission but no faster than every 500 ms",
            "Units": "%"
        },
        548: {
            "Acronym": "RC",
            "DataRange": "0 to 8,031.875 rpm",
            "EndBit": 55,
            "Name": "Maximum Retarder Speed, Point 2 (Retarder Configuration)",
            "Offset": 0.0,
            "OperationalHigh": 8031.875,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 19,
            "Resolution": 0.125,
            "SPN": 548,
            "SPNLength": 16,
            "StartBit": 40,
            "TransmissionRate": "Every 5 s and on change of torque/speed points of more than 10% since last transmission but no faster than every 500 ms",
            "Units": "rpm"
        },
        549: {
            "Acronym": "RC",
            "DataRange": "0 to 8,031.875 rpm",
            "EndBit": 79,
            "Name": "Retarder Speed At Point 3 (Retarder Configuration)",
            "Offset": 0.0,
            "OperationalHigh": 8031.875,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 19,
            "Resolution": 0.125,
            "SPN": 549,
            "SPNLength": 16,
            "StartBit": 64,
            "TransmissionRate": "Every 5 s and on change of torque/speed points of more than 10% since last transmission but no faster than every 500 ms",
            "Units": "rpm"
        },
        5492: {
            "Acronym": "ARMSWIMC",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Armrest 1 Switch 2",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5492,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5493: {
            "Acronym": "ARMSWIMC",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Armrest 2 Switch 1",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5493,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5494: {
            "Acronym": "ARMSWIMC",
            "DataRange": "0 to 3",
            "EndBit": 7,
            "Name": "Armrest 2 Switch 2",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5494,
            "SPNLength": 2,
            "StartBit": 6,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5495: {
            "Acronym": "AT1DPFSSC",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 15,
            "Name": "Aftertreatment 1 DPF Soot Mean Calibration Offset",
            "Offset": 0.0,
            "OperationalHigh": 160.6375,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 5495,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On powerup and on request",
            "Units": "%"
        },
        5496: {
            "Acronym": "AT1DPFSSC",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 31,
            "Name": "Aftertreatment 1 DPF Soot Standard Deviation Calibration Offset",
            "Offset": 0.0,
            "OperationalHigh": 160.6375,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 5496,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On powerup and on request",
            "Units": "%"
        },
        5497: {
            "Acronym": "AT2DPFSSC",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 15,
            "Name": "Aftertreatment 2 DPF Soot Mean Calibration Offset",
            "Offset": 0.0,
            "OperationalHigh": 160.6375,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 5497,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On powerup and on request",
            "Units": "%"
        },
        5498: {
            "Acronym": "AT2DPFSSC",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 31,
            "Name": "Aftertreatment 2 DPF Soot Standard Deviation Calibration Offset",
            "Offset": 0.0,
            "OperationalHigh": 160.6375,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 5498,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On powerup and on request",
            "Units": "%"
        },
        5499: {
            "Acronym": "IVAC",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Intake Valve Actuation System Oil Pressure 1 Solenoid Control",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5499,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        550: {
            "Acronym": "RC",
            "DataRange": "0 to 8,031.875 rpm",
            "EndBit": 103,
            "Name": "Retarder Speed At Point 4 (Retarder Configuration)",
            "Offset": 0.0,
            "OperationalHigh": 8031.875,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 19,
            "Resolution": 0.125,
            "SPN": 550,
            "SPNLength": 16,
            "StartBit": 88,
            "TransmissionRate": "Every 5 s and on change of torque/speed points of more than 10% since last transmission but no faster than every 500 ms",
            "Units": "rpm"
        },
        5500: {
            "Acronym": "IVAC",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Intake Valve Actuation System Oil Pressure 2 Solenoid Control",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5500,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        5501: {
            "Acronym": "DPFC2",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 15,
            "Name": "Aftertreatment 1 DPF Intake Temperature Set Point",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 5501,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "deg F"
        },
        5502: {
            "Acronym": "DPFC2",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 31,
            "Name": "Relative Unburned Fuel Mass from Engine",
            "Offset": 0.0,
            "OperationalHigh": 160.6375,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 5502,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        5503: {
            "Acronym": "DPFC2",
            "DataRange": "0 to 3212.75 g/m",
            "EndBit": 47,
            "Name": "Aftertreatment 1 Fuel Mass Rate",
            "Offset": 0.0,
            "OperationalHigh": 3212.75,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05,
            "SPN": 5503,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "g/m"
        },
        5504: {
            "Acronym": "DPFC1",
            "DataRange": "0 to 3",
            "EndBit": 57,
            "Name": "Hydrocarbon Doser Purging Enable",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5504,
            "SPNLength": 2,
            "StartBit": 56,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010:  1 s and on change",
            "Units": "bit"
        },
        5505: {
            "Acronym": "HCDI1",
            "DataRange": "0 to 3212.75 g/m",
            "EndBit": 15,
            "Name": "Requested Fuel Mass Rate",
            "Offset": 0.0,
            "OperationalHigh": 3212.75,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05,
            "SPN": 5505,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "g/m"
        },
        5506: {
            "Acronym": "HCDI1",
            "DataRange": "0 to 7",
            "EndBit": 18,
            "Name": "HC Doser Status",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5506,
            "SPNLength": 3,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5507: {
            "Acronym": "HCDI1",
            "DataRange": "0 to 3",
            "EndBit": 20,
            "Name": "HC Doser Injecting Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5507,
            "SPNLength": 2,
            "StartBit": 19,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5508: {
            "Acronym": "HCDI1",
            "DataRange": "0 to 7",
            "EndBit": 23,
            "Name": "Diagnostic Status of HC Doser",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5508,
            "SPNLength": 3,
            "StartBit": 21,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5509: {
            "Acronym": "HCDI1",
            "DataRange": "0 to 3",
            "EndBit": 25,
            "Name": "Hydrocarbon Doser Purging Needed",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5509,
            "SPNLength": 2,
            "StartBit": 24,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        551: {
            "Acronym": "RC",
            "DataRange": "-125 to 125 %",
            "EndBit": 39,
            "Name": "Percent Torque At Idle, Point 1 (Retarder Configuration)",
            "Offset": -125.0,
            "OperationalHigh": 0.0,
            "OperationalLow": -125.0,
            "OperationalRange": "-125 to 0%",
            "PGNLength": 19,
            "Resolution": 1,
            "SPN": 551,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "Every 5 s and on change of torque/speed points of more than 10% since last transmission but no faster than every 500 ms",
            "Units": "%"
        },
        5510: {
            "Acronym": "VLS1",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Lubrication Cycle Mode Lock Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5510,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        5511: {
            "Acronym": "VLS1",
            "DataRange": "0 to 7",
            "EndBit": 4,
            "Name": "Lubrication Cycle Set Command",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5511,
            "SPNLength": 3,
            "StartBit": 2,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        5512: {
            "Acronym": "VLS1",
            "DataRange": "0 to 7",
            "EndBit": 7,
            "Name": "Lubrication Operation Mode Set Command",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5512,
            "SPNLength": 3,
            "StartBit": 5,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        5513: {
            "Acronym": "VLS1",
            "DataRange": "0 to 3",
            "EndBit": 9,
            "Name": "Lubrication Pump Enable Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5513,
            "SPNLength": 2,
            "StartBit": 8,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        5514: {
            "Acronym": "VLS1",
            "DataRange": "0 to 3",
            "EndBit": 11,
            "Name": "Lubrication Timing Cycle Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5514,
            "SPNLength": 2,
            "StartBit": 10,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        5515: {
            "Acronym": "VLS2",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Lubrication Cycle Mode Lock Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5515,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        5516: {
            "Acronym": "VLS2",
            "DataRange": "0 to 7",
            "EndBit": 4,
            "Name": "Lubrication Cycle Mode Status",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5516,
            "SPNLength": 3,
            "StartBit": 2,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        5517: {
            "Acronym": "VLS2",
            "DataRange": "0 to 7",
            "EndBit": 7,
            "Name": "Lubrication Operation Mode Status",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5517,
            "SPNLength": 3,
            "StartBit": 5,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        5518: {
            "Acronym": "VLS2",
            "DataRange": "0 to 3",
            "EndBit": 9,
            "Name": "Lubrication Pump Enable Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5518,
            "SPNLength": 2,
            "StartBit": 8,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        5519: {
            "Acronym": "VLS2",
            "DataRange": "0 to 3",
            "EndBit": 11,
            "Name": "Lubrication Reservoir Level",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5519,
            "SPNLength": 2,
            "StartBit": 10,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        552: {
            "Acronym": "RC",
            "DataRange": "-125 to 125 %",
            "EndBit": 63,
            "Name": "Percent Torque At Maximum Speed, Point 2 (Retarder Configuration)",
            "Offset": -125.0,
            "OperationalHigh": 0.0,
            "OperationalLow": -125.0,
            "OperationalRange": "-125 to 0%",
            "PGNLength": 19,
            "Resolution": 1,
            "SPN": 552,
            "SPNLength": 8,
            "StartBit": 56,
            "TransmissionRate": "Every 5 s and on change of torque/speed points of more than 10% since last transmission but no faster than every 500 ms",
            "Units": "%"
        },
        5520: {
            "Acronym": "VLS2",
            "DataRange": "0 to 3",
            "EndBit": 13,
            "Name": "Lubrication Timing Cycle Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5520,
            "SPNLength": 2,
            "StartBit": 12,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        553: {
            "Acronym": "RC",
            "DataRange": "-125 to 125 %",
            "EndBit": 87,
            "Name": "Percent Torque At Point 3 (Retarder Configuration)",
            "Offset": -125.0,
            "OperationalHigh": 0.0,
            "OperationalLow": -125.0,
            "OperationalRange": "-125 to 0%",
            "PGNLength": 19,
            "Resolution": 1,
            "SPN": 553,
            "SPNLength": 8,
            "StartBit": 80,
            "TransmissionRate": "Every 5 s and on change of torque/speed points of more than 10% since last transmission but no faster than every 500 ms",
            "Units": "%"
        },
        5532: {
            "Acronym": "CL",
            "DataRange": "0 to 100 %",
            "EndBit": 15,
            "Name": "Switch Backlight Illumination Brightness Percent",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 5532,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "Every 5 s and on change of state but no faster than every 100 ms",
            "Units": "%"
        },
        5533: {
            "Acronym": "HOP",
            "DataRange": "0 to 1003.984375 Cp",
            "EndBit": 15,
            "Name": "Hydraulic Oil Dynamic Viscosity",
            "Offset": 0.0,
            "OperationalHigh": 1003.984375,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.015625,
            "SPN": 5533,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "30 seconds",
            "Units": "Cp"
        },
        5534: {
            "Acronym": "HOP",
            "DataRange": "0 to 1.961 g/cc",
            "EndBit": 31,
            "Name": "Hydraulic Oil Density",
            "Offset": 0.0,
            "OperationalHigh": 1.961,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 3.052e-05,
            "SPN": 5534,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "30 seconds",
            "Units": "g/cc"
        },
        5535: {
            "Acronym": "HOP",
            "DataRange": "0 to 7.844",
            "EndBit": 47,
            "Name": "Hydraulic Oil Relative Dielectricity (high resolution)",
            "Offset": 0.0,
            "OperationalHigh": 7.844,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5535,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "30 seconds",
            "Units": ""
        },
        5536: {
            "Acronym": "HOP",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 63,
            "Name": "Hydraulic Oil Temperature (High Resolution)",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 5536,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "30 seconds",
            "Units": "deg F"
        },
        5537: {
            "Acronym": "EFP",
            "DataRange": "0 to 1003.984375 Cp",
            "EndBit": 15,
            "Name": "Engine Fuel Dynamic Viscosity",
            "Offset": 0.0,
            "OperationalHigh": 1003.984375,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.015625,
            "SPN": 5537,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "30 seconds",
            "Units": "Cp"
        },
        5538: {
            "Acronym": "EFP",
            "DataRange": "0 to 1.961 g/cc",
            "EndBit": 31,
            "Name": "Engine Fuel Density",
            "Offset": 0.0,
            "OperationalHigh": 1.961,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 3.052e-05,
            "SPN": 5538,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "30 seconds",
            "Units": "g/cc"
        },
        5539: {
            "Acronym": "EFP",
            "DataRange": "0 to 7.844",
            "EndBit": 47,
            "Name": "Engine Fuel Dielectricity (high resolution)",
            "Offset": 0.0,
            "OperationalHigh": 7.844,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5539,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "30 seconds",
            "Units": ""
        },
        554: {
            "Acronym": "RC",
            "DataRange": "-125 to 125 %",
            "EndBit": 111,
            "Name": "Percent Torque At Point 4 (Retarder Configuration)",
            "Offset": -125.0,
            "OperationalHigh": 0.0,
            "OperationalLow": -125.0,
            "OperationalRange": "-125 to 0%",
            "PGNLength": 19,
            "Resolution": 1,
            "SPN": 554,
            "SPNLength": 8,
            "StartBit": 104,
            "TransmissionRate": "Every 5 s and on change of torque/speed points of more than 10% since last transmission but no faster than every 500 ms",
            "Units": "%"
        },
        5540: {
            "Acronym": "EFP",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 63,
            "Name": "Engine Fuel Temperature (High Resolution)",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 5540,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "30 seconds",
            "Units": "deg F"
        },
        5541: {
            "Acronym": "EBCC",
            "DataRange": "0.0 to 72.518869 psi",
            "EndBit": 7,
            "Name": "Engine Turbocharger 1 Turbine Outlet Pressure",
            "Offset": 0.0,
            "OperationalHigh": 72.518869,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.290075476,
            "SPN": 5541,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        5542: {
            "Acronym": "EBCC",
            "DataRange": "0.0 to 72.518869 psi",
            "EndBit": 15,
            "Name": "Desired Engine Turbocharger 1 Turbine Outlet Pressure",
            "Offset": 0.0,
            "OperationalHigh": 72.518869,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.290075476,
            "SPN": 5542,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        5543: {
            "Acronym": "EBCC",
            "DataRange": "0 to 100 %",
            "EndBit": 23,
            "Name": "Engine Exhaust Brake Actuator Command",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 5543,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        5544: {
            "Acronym": "EBCC",
            "DataRange": "0.0 to 72.518869 psi",
            "EndBit": 31,
            "Name": "Engine Turbocharger 2 Turbine Outlet Pressure",
            "Offset": 0.0,
            "OperationalHigh": 72.518869,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.290075476,
            "SPN": 5544,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        5545: {
            "Acronym": "EBCC",
            "DataRange": "0.0 to 72.518869 psi",
            "EndBit": 39,
            "Name": "Desired Engine Turbocharger 2 Turbine Outlet Pressure",
            "Offset": 0.0,
            "OperationalHigh": 72.518869,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.290075476,
            "SPN": 5545,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "psi"
        },
        5546: {
            "Acronym": "ECCVC",
            "DataRange": "0 to 100 %",
            "EndBit": 31,
            "Name": "Engine Auxiliary Cooler Supply Valve 1 Actuator Command",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 5546,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "1 sec",
            "Units": "%"
        },
        5547: {
            "Acronym": "ECCVC",
            "DataRange": "0 to 100 %",
            "EndBit": 39,
            "Name": "Engine Auxiliary Cooler Supply Valve 2 Actuator Command",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 5547,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "1 sec",
            "Units": "%"
        },
        5548: {
            "Acronym": "CSA",
            "DataRange": "0 to 100 %",
            "EndBit": 23,
            "Name": "Engine Cold Start Fuel Igniter Command",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 5548,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "As required",
            "Units": "%"
        },
        5549: {
            "Acronym": "CSA",
            "DataRange": "0 to 3",
            "EndBit": 25,
            "Name": "Engine Cold Start Fuel Igniter Relay",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5549,
            "SPNLength": 2,
            "StartBit": 24,
            "TransmissionRate": "As required",
            "Units": "bit"
        },
        555: {
            "Acronym": "RC",
            "DataRange": "-125 to 125 %",
            "EndBit": 151,
            "Name": "Percent Torque At Peak Torque, Point 5 (Retarder Configuration)",
            "Offset": -125.0,
            "OperationalHigh": 0.0,
            "OperationalLow": -125.0,
            "OperationalRange": "-125 to 0%",
            "PGNLength": 19,
            "Resolution": 1,
            "SPN": 555,
            "SPNLength": 8,
            "StartBit": 144,
            "TransmissionRate": "Every 5 s and on change of torque/speed points of more than 10% since last transmission but no faster than every 500 ms",
            "Units": "%"
        },
        5550: {
            "Acronym": "CSA",
            "DataRange": "0 to 3",
            "EndBit": 27,
            "Name": "Engine Cold Start Fuel Igniter Relay Feedback",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5550,
            "SPNLength": 2,
            "StartBit": 26,
            "TransmissionRate": "As required",
            "Units": "bit"
        },
        5551: {
            "Acronym": "MSI2",
            "DataRange": "0 to 3",
            "EndBit": 39,
            "Name": "Magnet Generator Control",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5551,
            "SPNLength": 2,
            "StartBit": 38,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        5552: {
            "Acronym": "SFC",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Supplemental Fan Enable Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5552,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "1 second",
            "Units": "bit"
        },
        5553: {
            "Acronym": "SFC",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Supplemental Fan Direction Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5553,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "1 second",
            "Units": "bit"
        },
        5554: {
            "Acronym": "SFC",
            "DataRange": "0 to 32,127.5 rpm",
            "EndBit": 23,
            "Name": "Supplemental Fan Speed Command",
            "Offset": 0.0,
            "OperationalHigh": 32127.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.5,
            "SPN": 5554,
            "SPNLength": 16,
            "StartBit": 8,
            "TransmissionRate": "1 second",
            "Units": "rpm"
        },
        5555: {
            "Acronym": "SFC",
            "DataRange": "0 to 3",
            "EndBit": 25,
            "Name": "Supplemental Fan Power Hold",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5555,
            "SPNLength": 2,
            "StartBit": 24,
            "TransmissionRate": "1 second",
            "Units": "bit"
        },
        5556: {
            "Acronym": "SFS",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Supplemental Fan Enable Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5556,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "1 second",
            "Units": "bit"
        },
        5557: {
            "Acronym": "SFS",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Supplemental Fan Direction Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5557,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "1 second",
            "Units": "bit"
        },
        5558: {
            "Acronym": "SFS",
            "DataRange": "0 to 15",
            "EndBit": 7,
            "Name": "Supplemental Fan Velocity Status",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5558,
            "SPNLength": 4,
            "StartBit": 4,
            "TransmissionRate": "1 second",
            "Units": "bit"
        },
        5559: {
            "Acronym": "SFS",
            "DataRange": "0 to 7",
            "EndBit": 10,
            "Name": "Supplemental Fan Controller Temperature Status",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5559,
            "SPNLength": 3,
            "StartBit": 8,
            "TransmissionRate": "1 second",
            "Units": "bit"
        },
        556: {
            "Acronym": "RC",
            "DataRange": "0 to 64,255 Nm",
            "EndBit": 143,
            "Name": "Reference Retarder Torque (Retarder Configuration)",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 19,
            "Resolution": 1,
            "SPN": 556,
            "SPNLength": 16,
            "StartBit": 128,
            "TransmissionRate": "Every 5 s and on change of torque/speed points of more than 10% since last transmission but no faster than every 500 ms",
            "Units": "Nm"
        },
        5560: {
            "Acronym": "SFS",
            "DataRange": "0 to 15",
            "EndBit": 14,
            "Name": "Supplemental Fan Drive Status",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5560,
            "SPNLength": 4,
            "StartBit": 11,
            "TransmissionRate": "1 second",
            "Units": "bit"
        },
        5561: {
            "Acronym": "SFS",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 23,
            "Name": "Supplemental Fan Controller ECU Temperature",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 5561,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "1 second",
            "Units": "deg F"
        },
        5562: {
            "Acronym": "SFS",
            "DataRange": "0 to 32,127.5 rpm",
            "EndBit": 39,
            "Name": "Supplemental Fan Speed",
            "Offset": 0.0,
            "OperationalHigh": 32127.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.5,
            "SPN": 5562,
            "SPNLength": 16,
            "StartBit": 24,
            "TransmissionRate": "1 second",
            "Units": "rpm"
        },
        5563: {
            "Acronym": "SFS",
            "DataRange": "-125 to 125 A",
            "EndBit": 47,
            "Name": "Supplemental Fan Current",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5563,
            "SPNLength": 8,
            "StartBit": 40,
            "TransmissionRate": "1 second",
            "Units": "Amps"
        },
        5564: {
            "Acronym": "SFS",
            "DataRange": "0 to 32127.5 W",
            "EndBit": 63,
            "Name": "Supplemental Fan Power",
            "Offset": 0.0,
            "OperationalHigh": 32127.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.5,
            "SPN": 5564,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "1 second",
            "Units": "W"
        },
        5565: {
            "Acronym": "ECC",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Coolant Level Engine Protection Shutdown Override Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5565,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "Every 10 seconds and on change but no faster than 1 second.",
            "Units": "bit"
        },
        5566: {
            "Acronym": "SHUTDN",
            "DataRange": "0 to 3",
            "EndBit": 61,
            "Name": "Coolant Level Engine Protection Shutdown Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5566,
            "SPNLength": 2,
            "StartBit": 60,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        5567: {
            "Acronym": "GC1",
            "DataRange": "-32128 to 32127",
            "EndBit": 47,
            "Name": "Engine Speed Governor Gain Adjust",
            "Offset": -32128.0,
            "OperationalHigh": 32127.0,
            "OperationalLow": -32128.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5567,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "count"
        },
        5568: {
            "Acronym": "GC1",
            "DataRange": "0 to 10%",
            "EndBit": 55,
            "Name": "Engine Speed Governor Droop",
            "Offset": 0.0,
            "OperationalHigh": 10.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.04,
            "SPN": 5568,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        5569: {
            "Acronym": "DPF1S",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 63,
            "Name": "DPF 1 Soot Sensor ECU Internal Temperature",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 5569,
            "SPNLength": 8,
            "StartBit": 56,
            "TransmissionRate": "1 sec",
            "Units": "deg F"
        },
        557: {
            "Acronym": "RC",
            "DataRange": "0 to 250 steps",
            "EndBit": 15,
            "Name": "Retarder Control Method (Retarder Configuration)",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 19,
            "Resolution": 1,
            "SPN": 557,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "Every 5 s and on change of torque/speed points of more than 10% since last transmission but no faster than every 500 ms",
            "Units": "step"
        },
        5570: {
            "Acronym": "DPF2S",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 63,
            "Name": "DPF 2 Soot Sensor ECU Internal Temperature",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 5570,
            "SPNLength": 8,
            "StartBit": 56,
            "TransmissionRate": "1 sec",
            "Units": "deg F"
        },
        5578: {
            "Acronym": "EFL/P12",
            "DataRange": "0.0 to 145.037738 psi",
            "EndBit": 7,
            "Name": "Engine Fuel Delivery Absolute Pressure",
            "Offset": 0.0,
            "OperationalHigh": 145.037738,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.580150952,
            "SPN": 5578,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        5579: {
            "Acronym": "EFL/P12",
            "DataRange": "0.0 to 145.037738 psi",
            "EndBit": 15,
            "Name": "Engine Filtered Fuel Delivery Pressure",
            "Offset": 0.0,
            "OperationalHigh": 145.037738,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.580150952,
            "SPN": 5579,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        558: {
            "Acronym": "EEC2",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Accelerator Pedal 1 Low Idle Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 558,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "50 msec (preferred) or Engine Speed Dependent (if required by application)",
            "Units": "bit"
        },
        5580: {
            "Acronym": "EFL/P12",
            "DataRange": "0.0 to 145.037738 psi",
            "EndBit": 23,
            "Name": "Engine Filtered Fuel Delivery Absolute Pressure",
            "Offset": 0.0,
            "OperationalHigh": 145.037738,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.580150952,
            "SPN": 5580,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        5581: {
            "Acronym": "AMB2",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 47,
            "Name": "Calculated Ambient Air Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 5581,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        5582: {
            "Acronym": "SRASI",
            "DataRange": "-125 to 125 deg",
            "EndBit": 7,
            "Name": "Static Roll Angle",
            "Offset": -125.0,
            "OperationalHigh": 90.0,
            "OperationalLow": -90.0,
            "OperationalRange": "-90 to 90 deg",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5582,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "deg"
        },
        5583: {
            "Acronym": "SRASI",
            "DataRange": "0 to 7",
            "EndBit": 10,
            "Name": "Static Roll Angle Status",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5583,
            "SPNLength": 3,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5584: {
            "Acronym": "EFL/P12",
            "DataRange": "0 to 250 %",
            "EndBit": 31,
            "Name": "Engine Fuel Filter Degradation",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5584,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "500 ms",
            "Units": "%"
        },
        5589: {
            "Acronym": "AT1AC3",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 15,
            "Name": "Aftertreatment 1 Secondary Air Absolute Pressure",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 5589,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        559: {
            "Acronym": "EEC2",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Accelerator Pedal Kickdown Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 559,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "50 msec (preferred) or Engine Speed Dependent (if required by application)",
            "Units": "bit"
        },
        5590: {
            "Acronym": "AT2AC3",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 15,
            "Name": "Aftertreatment 2 Secondary Air Absolute Pressure",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 5590,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        560: {
            "Acronym": "ETC1",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Transmission Driveline Engaged",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 560,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "10 ms",
            "Units": "bit"
        },
        5603: {
            "Acronym": "CCVS2",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Cruise Control Disable Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5603,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "Every 100 ms and on change but no faster than 20 ms",
            "Units": "bit"
        },
        5604: {
            "Acronym": "CCVS2",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Cruise Control Resume Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5604,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "Every 100 ms and on change but no faster than 20 ms",
            "Units": "bit"
        },
        5605: {
            "Acronym": "CCVS2",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Cruise Control Pause Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5605,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "Every 100 ms and on change but no faster than 20 ms",
            "Units": "bit"
        },
        5606: {
            "Acronym": "CCVS3",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Adaptive Cruise Control Readiness Status",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5606,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "Every 1 sec and on change but no faster than 100 ms",
            "Units": "bit"
        },
        5607: {
            "Acronym": "CCVS3",
            "DataRange": "0 to 7",
            "EndBit": 4,
            "Name": "Cruise Control System Command State",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5607,
            "SPNLength": 3,
            "StartBit": 2,
            "TransmissionRate": "Every 1 sec and on change but no faster than 100 ms",
            "Units": "bit"
        },
        5608: {
            "Acronym": "CCVS3",
            "DataRange": "0 to 255",
            "EndBit": 15,
            "Name": "Source Address of Controlling Device for Disabling Cruise Control",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5608,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "Every 1 sec and on change but no faster than 100 ms",
            "Units": "SA"
        },
        5609: {
            "Acronym": "CCVS3",
            "DataRange": "0 to 255",
            "EndBit": 23,
            "Name": "Source Address of Controlling Device for Pausing Cruise Control",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5609,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "Every 1 sec and on change but no faster than 100 ms",
            "Units": "SA"
        },
        561: {
            "Acronym": "EBC1",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "ASR Engine Control Active",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 561,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        562: {
            "Acronym": "EBC1",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "ASR Brake Control Active",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 562,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5624: {
            "Acronym": "VDC1",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Trailer-VDC Active",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5624,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "100ms",
            "Units": "bit"
        },
        5625: {
            "Acronym": "ETWAI",
            "DataRange": "0 to 100 %",
            "EndBit": 23,
            "Name": "Engine Exhaust Back Pressure Regulator Position",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 5625,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        5626: {
            "Acronym": "ETWAI",
            "DataRange": "0 to 31",
            "EndBit": 28,
            "Name": "Engine Exhaust Back Pressure Regulator Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 5626,
            "SPNLength": 5,
            "StartBit": 24,
            "TransmissionRate": "100 ms",
            "Units": "binary"
        },
        5627: {
            "Acronym": "ETWAI",
            "DataRange": "0 to 7",
            "EndBit": 31,
            "Name": "Engine Exhaust Back Pressure Regulator Temperature Status",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5627,
            "SPNLength": 3,
            "StartBit": 29,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5628: {
            "Acronym": "ETWAI",
            "DataRange": "0 to 15",
            "EndBit": 35,
            "Name": "Engine Exhaust Back Pressure Regulator Control Operation Status",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5628,
            "SPNLength": 4,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5629: {
            "Acronym": "DPFC1",
            "DataRange": "0 to 3",
            "EndBit": 59,
            "Name": "DPF Active Regeneration Inhibited Due to Low Exhaust Gas Pressure",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5629,
            "SPNLength": 2,
            "StartBit": 58,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010:  1 s and on change",
            "Units": "bit"
        },
        563: {
            "Acronym": "EBC1",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Anti-Lock Braking (ABS) Active",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 563,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5630: {
            "Acronym": "CM1",
            "DataRange": "0 to 3",
            "EndBit": 33,
            "Name": "Park Brake Command",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5630,
            "SPNLength": 2,
            "StartBit": 32,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        5631: {
            "Acronym": "EFL/P3",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 63,
            "Name": "Engine Throttle Valve Differential Pressure",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 5631,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        5632: {
            "Acronym": "AT1DPF1S3",
            "DataRange": "0 to 6425.5 MHz",
            "EndBit": 15,
            "Name": "Aftertreatment 1 DPF 1 Soot Sensor Analysis Request Frequency 1",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 5632,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "MHz"
        },
        5633: {
            "Acronym": "AT1DPF1S3",
            "DataRange": "0 to 6425.5 MHz",
            "EndBit": 31,
            "Name": "Aftertreatment 1 DPF 1 Soot Sensor Analysis Request Frequency 2",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 5633,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "1 s",
            "Units": "MHz"
        },
        5634: {
            "Acronym": "AT1DPF1S3",
            "DataRange": "0 to 6425.5 MHz",
            "EndBit": 47,
            "Name": "Aftertreatment 1 DPF 1 Soot Sensor Analysis Request Frequency 3",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 5634,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "1 s",
            "Units": "MHz"
        },
        5635: {
            "Acronym": "AT1DPF1S3",
            "DataRange": "0 to 6425.5 MHz",
            "EndBit": 63,
            "Name": "Aftertreatment 1 DPF 1 Soot Sensor Analysis Request Frequency 4",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 5635,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "1 s",
            "Units": "MHz"
        },
        5636: {
            "Acronym": "AT1DPF1S4",
            "DataRange": "0 to 6425.5 MHz",
            "EndBit": 15,
            "Name": "Aftertreatment 1 DPF 1 Soot Sensor Analysis Frequency 1",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 5636,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "MHz"
        },
        5637: {
            "Acronym": "AT1DPF1S4",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 31,
            "Name": "Aftertreatment 1 DPF 1 Soot Sensor Frequency 1 Soot Signal",
            "Offset": 0.0,
            "OperationalHigh": 160.6375,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 5637,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "1 s",
            "Units": "%"
        },
        5638: {
            "Acronym": "AT1DPF1S4",
            "DataRange": "0 to 6425.5 MHz",
            "EndBit": 47,
            "Name": "Aftertreatment 1 DPF 1 Soot Sensor Analysis Frequency 2",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 5638,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "1 s",
            "Units": "MHz"
        },
        5639: {
            "Acronym": "AT1DPF1S4",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 63,
            "Name": "Aftertreatment 1 DPF 1 Soot Sensor Frequency 2 Soot Signal",
            "Offset": 0.0,
            "OperationalHigh": 160.6375,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 5639,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "1 s",
            "Units": "%"
        },
        564: {
            "Acronym": "EAC1",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Differential Lock State - Central",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 564,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        5640: {
            "Acronym": "AT1DPF1S5",
            "DataRange": "0 to 6425.5 MHz",
            "EndBit": 15,
            "Name": "Aftertreatment 1 DPF 1 Soot Sensor Analysis Frequency 3",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 5640,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "MHz"
        },
        5641: {
            "Acronym": "AT1DPF1S5",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 31,
            "Name": "Aftertreatment 1 DPF 1 Soot Sensor Frequency 3 Soot Signal",
            "Offset": 0.0,
            "OperationalHigh": 160.6375,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 5641,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "1 s",
            "Units": "%"
        },
        5642: {
            "Acronym": "AT1DPF1S5",
            "DataRange": "0 to 6425.5 MHz",
            "EndBit": 47,
            "Name": "Aftertreatment 1 DPF 1 Soot Sensor Analysis Frequency 4",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 5642,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "1 s",
            "Units": "MHz"
        },
        5643: {
            "Acronym": "AT1DPF1S5",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 63,
            "Name": "Aftertreatment 1 DPF 1 Soot Sensor Frequency 4 Soot Signal",
            "Offset": 0.0,
            "OperationalHigh": 160.6375,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 5643,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "1 s",
            "Units": "%"
        },
        5644: {
            "Acronym": "AT1DPF2S3",
            "DataRange": "0 to 6425.5 MHz",
            "EndBit": 15,
            "Name": "Aftertreatment 1 DPF 2 Soot Sensor Analysis Request Frequency 1",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 5644,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "MHz"
        },
        5645: {
            "Acronym": "AT1DPF2S3",
            "DataRange": "0 to 6425.5 MHz",
            "EndBit": 31,
            "Name": "Aftertreatment 1 DPF 2 Soot Sensor Analysis Request Frequency 2",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 5645,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "1 s",
            "Units": "MHz"
        },
        5646: {
            "Acronym": "AT1DPF2S3",
            "DataRange": "0 to 6425.5 MHz",
            "EndBit": 47,
            "Name": "Aftertreatment 1 DPF 2 Soot Sensor Analysis Request Frequency 3",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 5646,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "1 s",
            "Units": "MHz"
        },
        5647: {
            "Acronym": "AT1DPF2S3",
            "DataRange": "0 to 6425.5 MHz",
            "EndBit": 63,
            "Name": "Aftertreatment 1 DPF 2 Soot Sensor Analysis Request Frequency 4",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 5647,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "1 s",
            "Units": "MHz"
        },
        5648: {
            "Acronym": "AT1DPF2S4",
            "DataRange": "0 to 6425.5 MHz",
            "EndBit": 15,
            "Name": "Aftertreatment 1 DPF 2 Soot Sensor Analysis Frequency 1",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 5648,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "MHz"
        },
        5649: {
            "Acronym": "AT1DPF2S4",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 31,
            "Name": "Aftertreatment 1 DPF 2 Soot Sensor Frequency 1 Soot Signal",
            "Offset": 0.0,
            "OperationalHigh": 160.6375,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 5649,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "1 s",
            "Units": "%"
        },
        565: {
            "Acronym": "EAC1",
            "DataRange": "0 to 3",
            "EndBit": 19,
            "Name": "Differential Lock State - Central Front",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 565,
            "SPNLength": 2,
            "StartBit": 18,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        5650: {
            "Acronym": "AT1DPF2S4",
            "DataRange": "0 to 6425.5 MHz",
            "EndBit": 47,
            "Name": "Aftertreatment 1 DPF 2 Soot Sensor Analysis Frequency 2",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 5650,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "1 s",
            "Units": "MHz"
        },
        5651: {
            "Acronym": "AT1DPF2S4",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 63,
            "Name": "Aftertreatment 1 DPF 2 Soot Sensor Frequency 2 Soot Signal",
            "Offset": 0.0,
            "OperationalHigh": 160.6375,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 5651,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "1 s",
            "Units": "%"
        },
        5652: {
            "Acronym": "AT1DPF2S5",
            "DataRange": "0 to 6425.5 MHz",
            "EndBit": 15,
            "Name": "Aftertreatment 1 DPF 2 Soot Sensor Analysis Frequency 3",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 5652,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "MHz"
        },
        5653: {
            "Acronym": "AT1DPF2S5",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 31,
            "Name": "Aftertreatment 1 DPF 2 Soot Sensor Frequency 3 Soot Signal",
            "Offset": 0.0,
            "OperationalHigh": 160.6375,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 5653,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "1 s",
            "Units": "%"
        },
        5654: {
            "Acronym": "AT1DPF2S5",
            "DataRange": "0 to 6425.5 MHz",
            "EndBit": 47,
            "Name": "Aftertreatment 1 DPF 2 Soot Sensor Analysis Frequency 4",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 5654,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "1 s",
            "Units": "MHz"
        },
        5655: {
            "Acronym": "AT1DPF2S5",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 63,
            "Name": "Aftertreatment 1 DPF 2 Soot Sensor Frequency 4 Soot Signal",
            "Offset": 0.0,
            "OperationalHigh": 160.6375,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 5655,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "1 s",
            "Units": "%"
        },
        5657: {
            "Acronym": "NM",
            "DataRange": "0 to 255",
            "EndBit": 7,
            "Name": "NAME Checksum / Error Code",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5657,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        5658: {
            "Acronym": "NM",
            "DataRange": "0 to 1",
            "EndBit": 8,
            "Name": "Manufacturer Code Qualifier Flag",
            "Offset": 0.0,
            "OperationalHigh": 1.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5658,
            "SPNLength": 1,
            "StartBit": 8,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        5659: {
            "Acronym": "NM",
            "DataRange": "0 to 1",
            "EndBit": 9,
            "Name": "ECU Instance Qualifier Flag",
            "Offset": 0.0,
            "OperationalHigh": 1.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5659,
            "SPNLength": 1,
            "StartBit": 9,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        566: {
            "Acronym": "EAC1",
            "DataRange": "0 to 3",
            "EndBit": 21,
            "Name": "Differential Lock State - Central Rear",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 566,
            "SPNLength": 2,
            "StartBit": 20,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        5660: {
            "Acronym": "NM",
            "DataRange": "0 to 1",
            "EndBit": 10,
            "Name": "Function Instance Qualifier Flag",
            "Offset": 0.0,
            "OperationalHigh": 1.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5660,
            "SPNLength": 1,
            "StartBit": 10,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        5661: {
            "Acronym": "NM",
            "DataRange": "0 to 1",
            "EndBit": 11,
            "Name": "Function Qualifier Flag",
            "Offset": 0.0,
            "OperationalHigh": 1.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5661,
            "SPNLength": 1,
            "StartBit": 11,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        5662: {
            "Acronym": "NM",
            "DataRange": "0 to 1",
            "EndBit": 12,
            "Name": "Vehicle System Qualifier Flag",
            "Offset": 0.0,
            "OperationalHigh": 1.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5662,
            "SPNLength": 1,
            "StartBit": 12,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        5663: {
            "Acronym": "NM",
            "DataRange": "0 to 1",
            "EndBit": 13,
            "Name": "Vehicle System Instance Qualifier Flag",
            "Offset": 0.0,
            "OperationalHigh": 1.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5663,
            "SPNLength": 1,
            "StartBit": 13,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        5664: {
            "Acronym": "NM",
            "DataRange": "0 to 1",
            "EndBit": 14,
            "Name": "Industry Group Qualifier Flag",
            "Offset": 0.0,
            "OperationalHigh": 1.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5664,
            "SPNLength": 1,
            "StartBit": 14,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        5665: {
            "Acronym": "NM",
            "DataRange": "0 to 1",
            "EndBit": 15,
            "Name": "Arbitrary Address Capable Qualifier Flag",
            "Offset": 0.0,
            "OperationalHigh": 1.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5665,
            "SPNLength": 1,
            "StartBit": 15,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        5666: {
            "Acronym": "NM",
            "DataRange": "0 to 15",
            "EndBit": 19,
            "Name": "NM Control Mode Indicator",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5666,
            "SPNLength": 4,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        5667: {
            "Acronym": "NM",
            "DataRange": "0 to 2047",
            "EndBit": 31,
            "Name": "Commanded Manufacturer Code",
            "Offset": 0.0,
            "OperationalHigh": 2047.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5667,
            "SPNLength": 11,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        5668: {
            "Acronym": "NM",
            "DataRange": "0 to 7",
            "EndBit": 34,
            "Name": "Commanded ECU Instance",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5668,
            "SPNLength": 3,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        5669: {
            "Acronym": "NM",
            "DataRange": "0 to 31",
            "EndBit": 39,
            "Name": "Commanded Function Instance",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5669,
            "SPNLength": 5,
            "StartBit": 35,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        567: {
            "Acronym": "EAC1",
            "DataRange": "0 to 3",
            "EndBit": 9,
            "Name": "Differential Lock State - Front Axle 1",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 567,
            "SPNLength": 2,
            "StartBit": 8,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        5670: {
            "Acronym": "NM",
            "DataRange": "0 to 255",
            "EndBit": 47,
            "Name": "Commanded Function",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5670,
            "SPNLength": 8,
            "StartBit": 40,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        5671: {
            "Acronym": "NM",
            "DataRange": "0 to 127",
            "EndBit": 55,
            "Name": "Commanded Vehicle System",
            "Offset": 0.0,
            "OperationalHigh": 127.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5671,
            "SPNLength": 7,
            "StartBit": 49,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        5672: {
            "Acronym": "NM",
            "DataRange": "0 to 15",
            "EndBit": 59,
            "Name": "Commanded Vehicle System Instance",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5672,
            "SPNLength": 4,
            "StartBit": 56,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        5673: {
            "Acronym": "NM",
            "DataRange": "0 to 7",
            "EndBit": 62,
            "Name": "Commanded Industry Group",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5673,
            "SPNLength": 3,
            "StartBit": 60,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        5674: {
            "Acronym": "NM",
            "DataRange": "0 to 1",
            "EndBit": 63,
            "Name": "Commanded Arbitrary Address Capable",
            "Offset": 0.0,
            "OperationalHigh": 1.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5674,
            "SPNLength": 1,
            "StartBit": 63,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        5675: {
            "Acronym": "OI",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Operator Shift Prompt",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5675,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "10 s or on change but no more often than 1s",
            "Units": "bit"
        },
        5676: {
            "Acronym": "AEBS1",
            "DataRange": "0 to 15",
            "EndBit": 3,
            "Name": "Advanced emergency braking system state",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5676,
            "SPNLength": 4,
            "StartBit": 0,
            "TransmissionRate": "Every 50ms and on change of \"AEBS system state\" or change of \"Collision warning status\" but no faster than every 10 ms",
            "Units": "bit"
        },
        5677: {
            "Acronym": "AEBS1",
            "DataRange": "0 to 15",
            "EndBit": 7,
            "Name": "Collision warning status",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5677,
            "SPNLength": 4,
            "StartBit": 4,
            "TransmissionRate": "Every 50ms and on change of \"AEBS system state\" or change of \"Collision warning status\" but no faster than every 10 ms",
            "Units": "bit"
        },
        5678: {
            "Acronym": "AEBS1",
            "DataRange": "0 to 7",
            "EndBit": 10,
            "Name": "Relevant object detected for advanced emergency braking system (AEBS)",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5678,
            "SPNLength": 3,
            "StartBit": 8,
            "TransmissionRate": "Every 50ms and on change of \"AEBS system state\" or change of \"Collision warning status\" but no faster than every 10 ms",
            "Units": "bit"
        },
        5679: {
            "Acronym": "AEBS1",
            "DataRange": "0 to 7",
            "EndBit": 13,
            "Name": "Bend off probability of relevant object",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5679,
            "SPNLength": 3,
            "StartBit": 11,
            "TransmissionRate": "Every 50ms and on change of \"AEBS system state\" or change of \"Collision warning status\" but no faster than every 10 ms",
            "Units": "bit"
        },
        568: {
            "Acronym": "EAC1",
            "DataRange": "0 to 3",
            "EndBit": 11,
            "Name": "Differential Lock State - Front Axle 2",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 568,
            "SPNLength": 2,
            "StartBit": 10,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        5680: {
            "Acronym": "AEBS1",
            "DataRange": "0 to 12.5 s",
            "EndBit": 23,
            "Name": "Time to collision with relevant object",
            "Offset": 0.0,
            "OperationalHigh": 12.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05,
            "SPN": 5680,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "Every 50ms and on change of \"AEBS system state\" or change of \"Collision warning status\" but no faster than every 10 ms",
            "Units": "Seconds"
        },
        5681: {
            "Acronym": "AEBS2",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Driver activation demand for Advanced Emergency Braking System",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5681,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "Every 50ms and if new SPN \"Driver activation demand for Advanced Emergency Braking System\" has changed but no faster than every 10 ms",
            "Units": "bit"
        },
        5682: {
            "Acronym": "AEBS2",
            "DataRange": "0 to 15",
            "EndBit": 59,
            "Name": "AEBS2 message Counter",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5682,
            "SPNLength": 4,
            "StartBit": 56,
            "TransmissionRate": "Every 50ms and if new SPN \"Driver activation demand for Advanced Emergency Braking System\" has changed but no faster than every 10 ms",
            "Units": "count"
        },
        5683: {
            "Acronym": "AEBS2",
            "DataRange": "0 to 15",
            "EndBit": 63,
            "Name": "AEBS2 message checksum",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5683,
            "SPNLength": 4,
            "StartBit": 60,
            "TransmissionRate": "Every 50ms and if new SPN \"Driver activation demand for Advanced Emergency Braking System\" has changed but no faster than every 10 ms",
            "Units": "count"
        },
        5684: {
            "Acronym": "GFD",
            "DataRange": "0 to 3212.75 kg/h",
            "EndBit": 31,
            "Name": "Reference Engine Gas Mass Flow Rate",
            "Offset": 0.0,
            "OperationalHigh": 3212.75,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05,
            "SPN": 5684,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "kg/h"
        },
        5685: {
            "Acronym": "AMB2",
            "DataRange": "0.0 to 931.939985519 psi",
            "EndBit": 63,
            "Name": "Barometric Absolute Pressure (High Resolution)",
            "Offset": 0.0,
            "OperationalHigh": 931.939985519,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0145037738,
            "SPN": 5685,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "1 s",
            "Units": "psi"
        },
        569: {
            "Acronym": "EAC1",
            "DataRange": "0 to 3",
            "EndBit": 13,
            "Name": "Differential Lock State - Rear Axle 1",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 569,
            "SPNLength": 2,
            "StartBit": 12,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        570: {
            "Acronym": "EAC1",
            "DataRange": "0 to 3",
            "EndBit": 15,
            "Name": "Differential Lock State - Rear Axle 2",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 570,
            "SPNLength": 2,
            "StartBit": 14,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        5706: {
            "Acronym": "AT1P1I",
            "DataRange": "0 to 100 %",
            "EndBit": 7,
            "Name": "Aftertreatment 1 Diesel Exhaust Fluid Pump Heater",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 5706,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "500 ms",
            "Units": "%"
        },
        5707: {
            "Acronym": "A1SCRDSR2",
            "DataRange": "0 to 100 %",
            "EndBit": 39,
            "Name": "Aftertreatment 1 Diesel Exhaust Fluid Pump Heater Command",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 5707,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "500 ms",
            "Units": "%"
        },
        5708: {
            "Acronym": "EFL/P10",
            "DataRange": "0.0 to 72.518869 psi",
            "EndBit": 63,
            "Name": "Engine Coolant Pressure 2",
            "Offset": 0.0,
            "OperationalHigh": 72.518869,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.290075476,
            "SPN": 5708,
            "SPNLength": 8,
            "StartBit": 56,
            "TransmissionRate": "500 ms",
            "Units": "psi"
        },
        5709: {
            "Acronym": "NSSR",
            "DataRange": "0 to 15",
            "EndBit": 3,
            "Name": "NOx Sensor ATO1 Self-diagnosis Trigger",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5709,
            "SPNLength": 4,
            "StartBit": 0,
            "TransmissionRate": "As required",
            "Units": "bit"
        },
        571: {
            "Acronym": "ERC1",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Retarder Enable - Brake Assist Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 571,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5710: {
            "Acronym": "NSSR",
            "DataRange": "0 to 15",
            "EndBit": 7,
            "Name": "NOx Sensor ATI1 Self-diagnosis Trigger",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5710,
            "SPNLength": 4,
            "StartBit": 4,
            "TransmissionRate": "As required",
            "Units": "bit"
        },
        5711: {
            "Acronym": "NSSR",
            "DataRange": "0 to 15",
            "EndBit": 11,
            "Name": "NOx Sensor ATO2 Self-diagnosis Trigger",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5711,
            "SPNLength": 4,
            "StartBit": 8,
            "TransmissionRate": "As required",
            "Units": "bit"
        },
        5712: {
            "Acronym": "NSSR",
            "DataRange": "0 to 15",
            "EndBit": 15,
            "Name": "NOx Sensor ATI2 Self-diagnosis Trigger",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5712,
            "SPNLength": 4,
            "StartBit": 12,
            "TransmissionRate": "As required",
            "Units": "bit"
        },
        5713: {
            "Acronym": "AT1IG1",
            "DataRange": "0 to 7",
            "EndBit": 55,
            "Name": "NOx Sensor ATO1 Self-diagnosis Status",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5713,
            "SPNLength": 3,
            "StartBit": 53,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        5714: {
            "Acronym": "AT1OG1",
            "DataRange": "0 to 7",
            "EndBit": 55,
            "Name": "NOx Sensor ATI1 Self-diagnosis Status",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5714,
            "SPNLength": 3,
            "StartBit": 53,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        5715: {
            "Acronym": "AT2IG1",
            "DataRange": "0 to 7",
            "EndBit": 55,
            "Name": "NOx Sensor ATO2 Self-diagnosis Status",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5715,
            "SPNLength": 3,
            "StartBit": 53,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        5716: {
            "Acronym": "AT2OG1",
            "DataRange": "0 to 7",
            "EndBit": 55,
            "Name": "NOx Sensor ATI2 Self-diagnosis Status",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5716,
            "SPNLength": 3,
            "StartBit": 53,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        5717: {
            "Acronym": "AT1OGC2",
            "DataRange": "0 to 250",
            "EndBit": 39,
            "Name": "NOx Sensor ATO1 Self-diagnosis Final Result",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5717,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "Ratio"
        },
        5718: {
            "Acronym": "AT1IGC2",
            "DataRange": "0 to 250",
            "EndBit": 39,
            "Name": "NOx Sensor ATI1 Self-diagnosis Final Result",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5718,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "Ratio"
        },
        5719: {
            "Acronym": "AT2OGC2",
            "DataRange": "0 to 250",
            "EndBit": 39,
            "Name": "NOx Sensor ATO2 Self-diagnosis Final Result",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5719,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "Ratio"
        },
        572: {
            "Acronym": "ERC1",
            "DataRange": "0 to 3",
            "EndBit": 7,
            "Name": "Retarder Enable - Shift Assist Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 572,
            "SPNLength": 2,
            "StartBit": 6,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5720: {
            "Acronym": "AT2IGC2",
            "DataRange": "0 to 250",
            "EndBit": 39,
            "Name": "NOx Sensor ATI2 Self-diagnosis Final Result",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5720,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "Ratio"
        },
        5721: {
            "Acronym": "AT1OGC1",
            "DataRange": "0 to 64255 h",
            "EndBit": 55,
            "Name": "NOx Sensor ATO1 Operation Hours Counter",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5721,
            "SPNLength": 16,
            "StartBit": 40,
            "TransmissionRate": "On start-up, and every second until the dewpoint signal state = 1 (SPN 3238) has been received by the transmitter",
            "Units": "hours"
        },
        5722: {
            "Acronym": "AT1IGC1",
            "DataRange": "0 to 64255 h",
            "EndBit": 55,
            "Name": "NOx Sensor ATI1 Operation Hours Counter",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5722,
            "SPNLength": 16,
            "StartBit": 40,
            "TransmissionRate": "On start-up, and every second until the dewpoint signal state = 1 (SPN 3237) has been received by the transmitter",
            "Units": "hours"
        },
        5723: {
            "Acronym": "AT2OGC1",
            "DataRange": "0 to 64255 h",
            "EndBit": 55,
            "Name": "NOx Sensor ATO2 Operation Hours Counter",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5723,
            "SPNLength": 16,
            "StartBit": 40,
            "TransmissionRate": "On start-up, and every second until the dewpoint signal state = 1 (SPN 3240) has been received by the transmitter",
            "Units": "hours"
        },
        5724: {
            "Acronym": "AT2IGC1",
            "DataRange": "0 to 64255 h",
            "EndBit": 55,
            "Name": "NOx Sensor ATI2 Operation Hours Counter",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5724,
            "SPNLength": 16,
            "StartBit": 40,
            "TransmissionRate": "On start-up, and every second until the dewpoint signal state = 1 (SPN 3239) has been received by the transmitter",
            "Units": "hours"
        },
        573: {
            "Acronym": "ETC1",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Transmission Torque Converter Lockup Engaged",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 573,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "10 ms",
            "Units": "bit"
        },
        574: {
            "Acronym": "ETC1",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Transmission Shift In Process",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 574,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "10 ms",
            "Units": "bit"
        },
        575: {
            "Acronym": "EBC1",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "ABS Off-road Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 575,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5750: {
            "Acronym": "WS",
            "DataRange": "0 to 64,255",
            "EndBit": 15,
            "Name": "Wireless Rolling Message Counter",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5750,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "50 ms",
            "Units": "count"
        },
        5751: {
            "Acronym": "WS",
            "DataRange": "0 to 15",
            "EndBit": 23,
            "Name": "Transmitter Command State",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5751,
            "SPNLength": 4,
            "StartBit": 20,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        5752: {
            "Acronym": "WS",
            "DataRange": "0 to 16449635",
            "EndBit": 47,
            "Name": "Transmitter ID",
            "Offset": 0.0,
            "OperationalHigh": 16449635.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5752,
            "SPNLength": 24,
            "StartBit": 24,
            "TransmissionRate": "50 ms",
            "Units": "ID"
        },
        5753: {
            "Acronym": "WS",
            "DataRange": "-125 to 0 dBm",
            "EndBit": 55,
            "Name": "Wireless Signal Strength",
            "Offset": -125.0,
            "OperationalHigh": 0.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.5,
            "SPN": 5753,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "50 ms",
            "Units": "dBm"
        },
        5754: {
            "Acronym": "RBR",
            "DataRange": "0 to 64,255",
            "EndBit": 15,
            "Name": "Unique Beacon Message Identifier Number",
            "Offset": 0.0,
            "OperationalHigh": 64255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5754,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "count"
        },
        5755: {
            "Acronym": "RBR",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Radio Beacon Direction",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5755,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        5756: {
            "Acronym": "RBR",
            "DataRange": "0 to 16449635",
            "EndBit": 47,
            "Name": "Transmitter ID 2",
            "Offset": 0.0,
            "OperationalHigh": 16449635.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5756,
            "SPNLength": 24,
            "StartBit": 24,
            "TransmissionRate": "On request",
            "Units": "ID"
        },
        5757: {
            "Acronym": "RBR",
            "DataRange": "-125 to 0 dBm",
            "EndBit": 55,
            "Name": "Wireless Signal Strength 2",
            "Offset": -125.0,
            "OperationalHigh": 0.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.5,
            "SPN": 5757,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "On request",
            "Units": "dBm"
        },
        5758: {
            "Acronym": "EEC12",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Aftertreatment 1 Intake Gas Sensor Power Supply",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5758,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        5759: {
            "Acronym": "EEC12",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Aftertreatment 1 Outlet Gas Sensor Power Supply",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5759,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        576: {
            "Acronym": "EBC1",
            "DataRange": "0 to 3",
            "EndBit": 19,
            "Name": "ASR Off-road Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 576,
            "SPNLength": 2,
            "StartBit": 18,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5760: {
            "Acronym": "EEC12",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Aftertreatment 2 Intake Gas Sensor Power Supply",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5760,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        5761: {
            "Acronym": "EEC12",
            "DataRange": "0 to 3",
            "EndBit": 7,
            "Name": "Aftertreatment 2 Outlet Gas Sensor Power Supply",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5761,
            "SPNLength": 2,
            "StartBit": 6,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        5762: {
            "Acronym": "TC1",
            "DataRange": "0 to 3",
            "EndBit": 39,
            "Name": "Transmission load reduction inhibit request",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5762,
            "SPNLength": 2,
            "StartBit": 38,
            "TransmissionRate": "When active; 50 ms to transmission and axles",
            "Units": "bit"
        },
        5763: {
            "Acronym": "EEGR1A",
            "DataRange": "0 to 31",
            "EndBit": 4,
            "Name": "Engine Exhaust Gas Recirculation 1 Actuator 1 Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 5763,
            "SPNLength": 5,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "binary"
        },
        5764: {
            "Acronym": "EEGR1A",
            "DataRange": "0 to 7",
            "EndBit": 7,
            "Name": "Engine Exhaust Gas Recirculation 1 Actuator 1 Temperature Status",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5764,
            "SPNLength": 3,
            "StartBit": 5,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5765: {
            "Acronym": "EEGR1A",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 15,
            "Name": "Engine Exhaust Gas Recirculation 1 Actuator 1 Temperature",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 5765,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "deg F"
        },
        5766: {
            "Acronym": "EEGR1A",
            "DataRange": "0 to 100 %",
            "EndBit": 23,
            "Name": "Engine Exhaust Gas Recirculation 1 Actuator 1 Desired Position",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 5766,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        5767: {
            "Acronym": "EEGR1A",
            "DataRange": "0 to 31",
            "EndBit": 28,
            "Name": "Engine Exhaust Gas Recirculation 1 Actuator 2 Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 5767,
            "SPNLength": 5,
            "StartBit": 24,
            "TransmissionRate": "100 ms",
            "Units": "binary"
        },
        5768: {
            "Acronym": "EEGR1A",
            "DataRange": "0 to 7",
            "EndBit": 31,
            "Name": "Engine Exhaust Gas Recirculation 1 Actuator 2 Temperature Status",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5768,
            "SPNLength": 3,
            "StartBit": 29,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5769: {
            "Acronym": "EEGR1A",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 39,
            "Name": "Engine Exhaust Gas Recirculation 1 Actuator 2 Temperature",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 5769,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "deg F"
        },
        577: {
            "Acronym": "EBC1",
            "DataRange": "0 to 3",
            "EndBit": 21,
            "Name": "ASR \"Hill Holder\" Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 577,
            "SPNLength": 2,
            "StartBit": 20,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5770: {
            "Acronym": "EEGR1A",
            "DataRange": "0 to 100 %",
            "EndBit": 47,
            "Name": "Engine Exhaust Gas Recirculation 1 Actuator 2 Desired Position",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 5770,
            "SPNLength": 8,
            "StartBit": 40,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        5771: {
            "Acronym": "EEGR1A",
            "DataRange": "0 to 15",
            "EndBit": 51,
            "Name": "Engine Exhaust Gas Recirculation 1 Actuator 1 Operation Status",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5771,
            "SPNLength": 4,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5772: {
            "Acronym": "EEGR1A",
            "DataRange": "0 to 15",
            "EndBit": 55,
            "Name": "Engine Exhaust Gas Recirculation 1 Actuator 2 Operation Status",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5772,
            "SPNLength": 4,
            "StartBit": 52,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5773: {
            "Acronym": "EEGR1A",
            "DataRange": "0 to 31",
            "EndBit": 4,
            "Name": "Engine Exhaust Gas Recirculation 2 Actuator 1 Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 5773,
            "SPNLength": 5,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "binary"
        },
        5774: {
            "Acronym": "EEGR1A",
            "DataRange": "0 to 7",
            "EndBit": 7,
            "Name": "Engine Exhaust Gas Recirculation 2 Actuator 1 Temperature Status",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5774,
            "SPNLength": 3,
            "StartBit": 5,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5775: {
            "Acronym": "EEGR1A",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 15,
            "Name": "Engine Exhaust Gas Recirculation 2 Actuator 1 Temperature",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 5775,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "deg F"
        },
        5776: {
            "Acronym": "EEGR1A",
            "DataRange": "0 to 100 %",
            "EndBit": 23,
            "Name": "Engine Exhaust Gas Recirculation 2 Actuator 1 Desired Position",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 5776,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        5777: {
            "Acronym": "EEGR1A",
            "DataRange": "0 to 31",
            "EndBit": 28,
            "Name": "Engine Exhaust Gas Recirculation 2 Actuator 2 Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": -3,
            "SPN": 5777,
            "SPNLength": 5,
            "StartBit": 24,
            "TransmissionRate": "100 ms",
            "Units": "binary"
        },
        5778: {
            "Acronym": "EEGR1A",
            "DataRange": "0 to 7",
            "EndBit": 31,
            "Name": "Engine Exhaust Gas Recirculation 2 Actuator 2 Temperature Status",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5778,
            "SPNLength": 3,
            "StartBit": 29,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5779: {
            "Acronym": "EEGR1A",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 39,
            "Name": "Engine Exhaust Gas Recirculation 2 Actuator 2 Temperature",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 5779,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "deg F"
        },
        578: {
            "Acronym": "AI",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 31,
            "Name": "Drive Axle Temperature",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 578,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        5780: {
            "Acronym": "EEGR1A",
            "DataRange": "0 to 100 %",
            "EndBit": 47,
            "Name": "Engine Exhaust Gas Recirculation 2 Actuator 2 Desired Position",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 5780,
            "SPNLength": 8,
            "StartBit": 40,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        5781: {
            "Acronym": "EEGR1A",
            "DataRange": "0 to 15",
            "EndBit": 51,
            "Name": "Engine Exhaust Gas Recirculation 2 Actuator 1 Operation Status",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5781,
            "SPNLength": 4,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5782: {
            "Acronym": "EEGR1A",
            "DataRange": "0 to 15",
            "EndBit": 55,
            "Name": "Engine Exhaust Gas Recirculation 2 Actuator 2 Operation Status",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5782,
            "SPNLength": 4,
            "StartBit": 52,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5783: {
            "Acronym": "EF/TVI2",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 7,
            "Name": "Engine Throttle Valve 1 Temperature",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 5783,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "deg F"
        },
        5784: {
            "Acronym": "EF/TVI2",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 15,
            "Name": "Engine Throttle Valve 2 Temperature",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 5784,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "deg F"
        },
        5785: {
            "Acronym": "EF/TVI2",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 23,
            "Name": "Engine Fuel Valve 1 Temperature",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 5785,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "deg F"
        },
        5786: {
            "Acronym": "EF/TVI2",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 31,
            "Name": "Engine Fuel Valve 2 Temperature",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 5786,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "100 ms",
            "Units": "deg F"
        },
        5787: {
            "Acronym": "EF/TVI2",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 39,
            "Name": "Engine Turbocharger Wastegate Actuator 2 Temperature",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 5787,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "deg F"
        },
        5788: {
            "Acronym": "ETWAI",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 47,
            "Name": "Engine Turbocharger Wastegate Actuator 1 Temperature",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 5788,
            "SPNLength": 8,
            "StartBit": 40,
            "TransmissionRate": "100 ms",
            "Units": "deg F"
        },
        5789: {
            "Acronym": "ETWAI",
            "DataRange": "0 to 100 %",
            "EndBit": 55,
            "Name": "Engine Exhaust Back Pressure Actuator 1 Desired Position",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 5789,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        579: {
            "Acronym": "AI",
            "DataRange": "0.0 to 145.037738 psi",
            "EndBit": 23,
            "Name": "Drive Axle Lift Air Pressure",
            "Offset": 0.0,
            "OperationalHigh": 145.037738,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.580150952,
            "SPN": 579,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "1 s",
            "Units": "psi"
        },
        5790: {
            "Acronym": "ETWAI",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 63,
            "Name": "Engine Exhaust Back Pressure Actuator 1 Temperature",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 5790,
            "SPNLength": 8,
            "StartBit": 56,
            "TransmissionRate": "100 ms",
            "Units": "deg F"
        },
        5791: {
            "Acronym": "ETCBI",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 39,
            "Name": "Engine Turbocharger Compressor Bypass Actuator 1 Temperature",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 5791,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "deg F"
        },
        5792: {
            "Acronym": "ETCBI",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 47,
            "Name": "Engine Turbocharger Compressor Bypass Actuator 2 Temperature",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 5792,
            "SPNLength": 8,
            "StartBit": 40,
            "TransmissionRate": "100 ms",
            "Units": "deg F"
        },
        5793: {
            "Acronym": "ESR",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Desired Engine Fueling State",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5793,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "Every 1 sec and on change but no faster than 100 ms",
            "Units": "bit"
        },
        5794: {
            "Acronym": "EEC13",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Feedback Engine Fueling State",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5794,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "Every 1 sec and on change but no faster than 100 ms",
            "Units": "bit"
        },
        5795: {
            "Acronym": "EEC13",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Engine Fueling Inhibit Allowed",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5795,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "Every 1 sec and on change but no faster than 100 ms",
            "Units": "bit"
        },
        580: {
            "Acronym": "VDS",
            "DataRange": "-2500 to 5531.875 m",
            "EndBit": 63,
            "Name": "Altitude",
            "Offset": -2500.0,
            "OperationalHigh": 5531.875,
            "OperationalLow": -2500.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.125,
            "SPN": 580,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "On request",
            "Units": "m"
        },
        5800: {
            "Acronym": "AV00EF",
            "DataRange": "0 to 63",
            "EndBit": 37,
            "Name": "Auxiliary valve 0 exit/reason code",
            "Offset": 0.0,
            "OperationalHigh": 63.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5800,
            "SPNLength": 6,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5801: {
            "Acronym": "AV01EF",
            "DataRange": "0 to 63",
            "EndBit": 37,
            "Name": "Auxiliary valve 1 exit/reason code",
            "Offset": 0.0,
            "OperationalHigh": 63.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5801,
            "SPNLength": 6,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5802: {
            "Acronym": "AV02EF",
            "DataRange": "0 to 63",
            "EndBit": 37,
            "Name": "Auxiliary valve 2 exit/reason code",
            "Offset": 0.0,
            "OperationalHigh": 63.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5802,
            "SPNLength": 6,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5803: {
            "Acronym": "AV03EF",
            "DataRange": "0 to 63",
            "EndBit": 37,
            "Name": "Auxiliary valve 3 exit/reason code",
            "Offset": 0.0,
            "OperationalHigh": 63.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5803,
            "SPNLength": 6,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5804: {
            "Acronym": "AV04EF",
            "DataRange": "0 to 63",
            "EndBit": 37,
            "Name": "Auxiliary valve 4 exit/reason code",
            "Offset": 0.0,
            "OperationalHigh": 63.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5804,
            "SPNLength": 6,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5805: {
            "Acronym": "AV05EF",
            "DataRange": "0 to 63",
            "EndBit": 37,
            "Name": "Auxiliary valve 5 exit/reason code",
            "Offset": 0.0,
            "OperationalHigh": 63.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5805,
            "SPNLength": 6,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5806: {
            "Acronym": "AV06EF",
            "DataRange": "0 to 63",
            "EndBit": 37,
            "Name": "Auxiliary valve 6 exit/reason code",
            "Offset": 0.0,
            "OperationalHigh": 63.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5806,
            "SPNLength": 6,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5807: {
            "Acronym": "AV07EF",
            "DataRange": "0 to 63",
            "EndBit": 37,
            "Name": "Auxiliary valve 7 exit/reason code",
            "Offset": 0.0,
            "OperationalHigh": 63.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5807,
            "SPNLength": 6,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5808: {
            "Acronym": "AV08EF",
            "DataRange": "0 to 63",
            "EndBit": 37,
            "Name": "Auxiliary valve 8 exit/reason code",
            "Offset": 0.0,
            "OperationalHigh": 63.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5808,
            "SPNLength": 6,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5809: {
            "Acronym": "AV09EF",
            "DataRange": "0 to 63",
            "EndBit": 37,
            "Name": "Auxiliary valve 9 exit/reason code",
            "Offset": 0.0,
            "OperationalHigh": 63.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5809,
            "SPNLength": 6,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        581: {
            "Acronym": "TCFG",
            "DataRange": "0 to 64.255",
            "EndBit": 31,
            "Name": "Transmission Gear Ratio",
            "Offset": 0.0,
            "OperationalHigh": 64.255,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.001,
            "SPN": 581,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "Ratio"
        },
        5810: {
            "Acronym": "AV10EF",
            "DataRange": "0 to 63",
            "EndBit": 37,
            "Name": "Auxiliary valve 10 exit/reason code",
            "Offset": 0.0,
            "OperationalHigh": 63.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5810,
            "SPNLength": 6,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5811: {
            "Acronym": "AV11EF",
            "DataRange": "0 to 63",
            "EndBit": 37,
            "Name": "Auxiliary valve 11 exit/reason code",
            "Offset": 0.0,
            "OperationalHigh": 63.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5811,
            "SPNLength": 6,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5812: {
            "Acronym": "AV12EF",
            "DataRange": "0 to 63",
            "EndBit": 37,
            "Name": "Auxiliary valve 12 exit/reason code",
            "Offset": 0.0,
            "OperationalHigh": 63.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5812,
            "SPNLength": 6,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5813: {
            "Acronym": "AV13EF",
            "DataRange": "0 to 63",
            "EndBit": 37,
            "Name": "Auxiliary valve 13 exit/reason code",
            "Offset": 0.0,
            "OperationalHigh": 63.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5813,
            "SPNLength": 6,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5814: {
            "Acronym": "AV14EF",
            "DataRange": "0 to 63",
            "EndBit": 37,
            "Name": "Auxiliary valve 14 exit/reason code",
            "Offset": 0.0,
            "OperationalHigh": 63.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5814,
            "SPNLength": 6,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5815: {
            "Acronym": "AV15EF",
            "DataRange": "0 to 63",
            "EndBit": 37,
            "Name": "Auxiliary valve 15 exit/reason code",
            "Offset": 0.0,
            "OperationalHigh": 63.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5815,
            "SPNLength": 6,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5816: {
            "Acronym": "FPTO",
            "DataRange": "0 to 31",
            "EndBit": 44,
            "Name": "Front hitch exit/reason code",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5816,
            "SPNLength": 5,
            "StartBit": 40,
            "TransmissionRate": "100ms when engage, otherwise on request",
            "Units": "bit"
        },
        5817: {
            "Acronym": "FPTO",
            "DataRange": "0 to 31",
            "EndBit": 52,
            "Name": "Front PTO exit/reason code",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5817,
            "SPNLength": 5,
            "StartBit": 48,
            "TransmissionRate": "100ms when engage, otherwise on request",
            "Units": "bit"
        },
        5818: {
            "Acronym": "MSSC",
            "DataRange": "0 to 31",
            "EndBit": 52,
            "Name": "Machine selected speed exit/reason code",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5818,
            "SPNLength": 5,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        5819: {
            "Acronym": "RHS",
            "DataRange": "0 to 31",
            "EndBit": 44,
            "Name": "Rear hitch exit/reason code",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5819,
            "SPNLength": 5,
            "StartBit": 40,
            "TransmissionRate": "100ms",
            "Units": "bit"
        },
        582: {
            "Acronym": "VW",
            "DataRange": "0 to 32,127.5 kg",
            "EndBit": 23,
            "Name": "Axle Weight",
            "Offset": 0.0,
            "OperationalHigh": 32127.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.5,
            "SPN": 582,
            "SPNLength": 16,
            "StartBit": 8,
            "TransmissionRate": "On request",
            "Units": "kg"
        },
        5820: {
            "Acronym": "RPTO",
            "DataRange": "0 to 31",
            "EndBit": 52,
            "Name": "Rear PTO exit/reason code",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5820,
            "SPNLength": 5,
            "StartBit": 48,
            "TransmissionRate": "100ms when engage, otherwise on request",
            "Units": "bit"
        },
        5821: {
            "Acronym": "WAND",
            "DataRange": "-64 to 64.51 deg",
            "EndBit": 15,
            "Name": "Wand Angle",
            "Offset": -64.0,
            "OperationalHigh": 64.51,
            "OperationalLow": -64.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.002,
            "SPN": 5821,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "20 ms",
            "Units": "deg"
        },
        5822: {
            "Acronym": "WAND",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Wand Sensor Figure of Merit",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5822,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "20 ms",
            "Units": "bit"
        },
        5823: {
            "Acronym": "LDISP",
            "DataRange": "0 to 6,425.5 mm (0 to 6.4255 m)",
            "EndBit": 15,
            "Name": "Measured Linear Displacement",
            "Offset": 0.0,
            "OperationalHigh": 6425.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 5823,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "20 ms",
            "Units": "m"
        },
        5824: {
            "Acronym": "LDISP",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Linear Displacement Sensor Sensor Figure of Merit",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5824,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "20 ms",
            "Units": "bit"
        },
        5825: {
            "Acronym": "OI",
            "DataRange": "0 to 7",
            "EndBit": 10,
            "Name": "Driver Warning System Indicator Status",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5825,
            "SPNLength": 3,
            "StartBit": 8,
            "TransmissionRate": "10 s or on change but no more often than 1s",
            "Units": "bit"
        },
        5826: {
            "Acronym": "OI",
            "DataRange": "0 to 7",
            "EndBit": 13,
            "Name": "Emission Control System Operator Inducement Severity",
            "Offset": 0.0,
            "OperationalHigh": 7.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5826,
            "SPNLength": 3,
            "StartBit": 11,
            "TransmissionRate": "10 s or on change but no more often than 1s",
            "Units": "bit"
        },
        5827: {
            "Acronym": "AT1HI",
            "DataRange": "0 to 526,385,151.9 km",
            "EndBit": 319,
            "Name": "Aftertreatment 1 Average Distance Between Active DPF Regenerations",
            "Offset": 0.0,
            "OperationalHigh": 526385151.9,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.125,
            "SPN": 5827,
            "SPNLength": 32,
            "StartBit": 288,
            "TransmissionRate": "On request",
            "Units": "m"
        },
        5828: {
            "Acronym": "AT2HI",
            "DataRange": "0 to 526,385,151.9 km",
            "EndBit": 319,
            "Name": "Aftertreatment 2 Average Distance Between Active DPF Regenerations",
            "Offset": 0.0,
            "OperationalHigh": 526385151.9,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 0.125,
            "SPN": 5828,
            "SPNLength": 32,
            "StartBit": 288,
            "TransmissionRate": "On request",
            "Units": "m"
        },
        5829: {
            "Acronym": "EEC14",
            "DataRange": "-125 to 132.02%",
            "EndBit": 15,
            "Name": "Engine Exhaust Gas Recirculation 1 Valve 1 Position Error",
            "Offset": -125.0,
            "OperationalHigh": 132.02,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.004,
            "SPN": 5829,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On Request",
            "Units": "%"
        },
        583: {
            "Acronym": "VDS",
            "DataRange": "-200 to 301.99 deg",
            "EndBit": 47,
            "Name": "Pitch",
            "Offset": -200.0,
            "OperationalHigh": 301.992,
            "OperationalLow": -200.0,
            "OperationalRange": "-200 to +301.992 deg (ASCENT)",
            "PGNLength": 8,
            "Resolution": 0.0078125,
            "SPN": 583,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "deg"
        },
        5830: {
            "Acronym": "EEC14",
            "DataRange": "-125 to 132.02%",
            "EndBit": 31,
            "Name": "Engine Exhaust Gas Recirculation 1 Valve 2 Position Error",
            "Offset": -125.0,
            "OperationalHigh": 132.02,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.004,
            "SPN": 5830,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On Request",
            "Units": "%"
        },
        5831: {
            "Acronym": "EEC11",
            "DataRange": "-125 to 132.02%",
            "EndBit": 47,
            "Name": "Engine Exhaust Gas Recirculation 2 Valve 1 Position Error",
            "Offset": -125.0,
            "OperationalHigh": 132.02,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.004,
            "SPN": 5831,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        5832: {
            "Acronym": "EEC11",
            "DataRange": "-125 to 132.02%",
            "EndBit": 63,
            "Name": "Engine Exhaust Gas Recirculation 2 Valve 2 Position Error",
            "Offset": -125.0,
            "OperationalHigh": 132.02,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.004,
            "SPN": 5832,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        5833: {
            "Acronym": "EEC14",
            "DataRange": "0 to 321.275 g/s",
            "EndBit": 47,
            "Name": "Engine Fuel Mass Flow Rate",
            "Offset": 0.0,
            "OperationalHigh": 321.275,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.005,
            "SPN": 5833,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "On Request",
            "Units": "g/s"
        },
        5834: {
            "Acronym": "DPFC2",
            "DataRange": "0 to 3212.75 g/m",
            "EndBit": 63,
            "Name": "Aftertreatment 2 Fuel Mass Rate",
            "Offset": 0.0,
            "OperationalHigh": 3212.75,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05,
            "SPN": 5834,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "g/m"
        },
        5835: {
            "Acronym": "EPSI",
            "DataRange": "0 to 803.1875 mg/m^3",
            "EndBit": 15,
            "Name": "Aftertreatment 1 Particulate Sensor",
            "Offset": 0.0,
            "OperationalHigh": 803.18753,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0125,
            "SPN": 5835,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "mg/m^3"
        },
        5836: {
            "Acronym": "EPSI",
            "DataRange": "0 to 803.1875 mg/m^3",
            "EndBit": 31,
            "Name": "Aftertreatment 2 Particulate Sensor",
            "Offset": 0.0,
            "OperationalHigh": 803.18753,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0125,
            "SPN": 5836,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "mg/m^3"
        },
        5837: {
            "Acronym": "EEC14",
            "DataRange": "0 to 255",
            "EndBit": 55,
            "Name": "Fuel Type",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5837,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "On Request",
            "Units": "bit"
        },
        584: {
            "Acronym": "VP",
            "DataRange": "-210 to 211.1008122 deg",
            "EndBit": 31,
            "Name": "Latitude",
            "Offset": -210.0,
            "OperationalHigh": 211.108122,
            "OperationalLow": -210.0,
            "OperationalRange": "-210 to +211.108122 deg (NORTH)",
            "PGNLength": 8,
            "Resolution": 10,
            "SPN": 584,
            "SPNLength": 32,
            "StartBit": 0,
            "TransmissionRate": "5 s",
            "Units": "deg"
        },
        5843: {
            "Acronym": "DM57",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Engine OBD Warm-up Sequence",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 0,
            "Resolution": 1,
            "SPN": 5843,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        5844: {
            "Acronym": "DM56",
            "DataRange": "0 to 255 per byte",
            "EndBit": 63,
            "Name": "Model Year",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 0,
            "Resolution": -2,
            "SPN": 5844,
            "SPNLength": 64,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "ASCII"
        },
        5845: {
            "Acronym": "DM56",
            "DataRange": "0 to 255 per byte",
            "EndBit": 1663,
            "Name": "Certification Engine Family Name",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 0,
            "Resolution": -2,
            "SPN": 5845,
            "SPNLength": 1600,
            "StartBit": 64,
            "TransmissionRate": "On request",
            "Units": "ASCII"
        },
        5848: {
            "Acronym": "A1SCRAI2",
            "DataRange": "0 to 19276.5 g/h",
            "EndBit": 15,
            "Name": "Aftertreatment 1 SCR Intermediate NH3",
            "Offset": 0.0,
            "OperationalHigh": 19276.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.3,
            "SPN": 5848,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "50 ms",
            "Units": "g/h"
        },
        5849: {
            "Acronym": "A1SCRAI2",
            "DataRange": "0 to 31",
            "EndBit": 20,
            "Name": "Aftertreatment 1 SCR Intermediate NH3 Sensor Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5849,
            "SPNLength": 5,
            "StartBit": 16,
            "TransmissionRate": "50 ms",
            "Units": "count"
        },
        585: {
            "Acronym": "VP",
            "DataRange": "-210 to 211.1008122 deg",
            "EndBit": 63,
            "Name": "Longitude",
            "Offset": -210.0,
            "OperationalHigh": 211.108122,
            "OperationalLow": -210.0,
            "OperationalRange": "-210 to +211.108122 deg (EAST)",
            "PGNLength": 8,
            "Resolution": 10,
            "SPN": 585,
            "SPNLength": 32,
            "StartBit": 32,
            "TransmissionRate": "5 s",
            "Units": "deg"
        },
        5850: {
            "Acronym": "A1SCRAI2",
            "DataRange": "0 to 3",
            "EndBit": 25,
            "Name": "Aftertreatment 1 SCR Intermediate NH3 Reading Stable",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5850,
            "SPNLength": 2,
            "StartBit": 24,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        5851: {
            "Acronym": "A1SCRAI2",
            "DataRange": "0 to 3",
            "EndBit": 27,
            "Name": "Aftertreatment 1 SCR Intermediate NH3 Gas Sensor Power In Range",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5851,
            "SPNLength": 2,
            "StartBit": 26,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        5852: {
            "Acronym": "A1SCRAI2",
            "DataRange": "0 to 3",
            "EndBit": 29,
            "Name": "Aftertreatment 1 SCR Intermediate NH3 Gas Sensor at Temperature",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5852,
            "SPNLength": 2,
            "StartBit": 28,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        5853: {
            "Acronym": "A1SCRAI2",
            "DataRange": "0 to 31",
            "EndBit": 36,
            "Name": "Aftertreatment 1 SCR Intermediate NH3 Gas Sensor Heater Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5853,
            "SPNLength": 5,
            "StartBit": 32,
            "TransmissionRate": "50 ms",
            "Units": "count"
        },
        5854: {
            "Acronym": "A1SCRAI2",
            "DataRange": "0 to 3",
            "EndBit": 41,
            "Name": "Aftertreatment 1 SCR Intermediate NH3 Gas Sensor Heater Control",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5854,
            "SPNLength": 2,
            "StartBit": 40,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        5855: {
            "Acronym": "A2SCRAI2",
            "DataRange": "0 to 6425.5 m^3/h",
            "EndBit": 15,
            "Name": "Aftertreatment 2 SCR Intermediate NH3",
            "Offset": 0.0,
            "OperationalHigh": 6425.53,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.1,
            "SPN": 5855,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "50 ms",
            "Units": "m^3/h"
        },
        5856: {
            "Acronym": "A2SCRAI2",
            "DataRange": "0 to 31",
            "EndBit": 20,
            "Name": "Aftertreatment 2 SCR Intermediate NH3 Sensor Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5856,
            "SPNLength": 5,
            "StartBit": 16,
            "TransmissionRate": "50 ms",
            "Units": "count"
        },
        5857: {
            "Acronym": "A2SCRAI2",
            "DataRange": "0 to 3",
            "EndBit": 25,
            "Name": "Aftertreatment 2 SCR Intermediate NH3 Reading Stable",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5857,
            "SPNLength": 2,
            "StartBit": 24,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        5858: {
            "Acronym": "A2SCRAI2",
            "DataRange": "0 to 3",
            "EndBit": 27,
            "Name": "Aftertreatment 2 SCR Intermediate NH3 Gas Sensor Power In Range",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5858,
            "SPNLength": 2,
            "StartBit": 26,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        5859: {
            "Acronym": "A2SCRAI2",
            "DataRange": "0 to 1",
            "EndBit": 28,
            "Name": "Aftertreatment 2 SCR Intermediate NH3 Gas Sensor at Temperature",
            "Offset": 0.0,
            "OperationalHigh": 1.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5859,
            "SPNLength": 1,
            "StartBit": 28,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        586: {
            "Acronym": "CI",
            "DataRange": "0 to 255 per byte",
            "EndBit": -1,
            "Name": "Make",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": -2,
            "SPN": 586,
            "SPNLength": 40,
            "StartBit": -1,
            "TransmissionRate": "On request",
            "Units": "ASCII"
        },
        5860: {
            "Acronym": "A2SCRAI2",
            "DataRange": "0 to 31",
            "EndBit": 36,
            "Name": "Aftertreatment 2 SCR Intermediate NH3 Gas Sensor Heater Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5860,
            "SPNLength": 5,
            "StartBit": 32,
            "TransmissionRate": "50 ms",
            "Units": "count"
        },
        5861: {
            "Acronym": "A2SCRAI2",
            "DataRange": "0 to 3",
            "EndBit": 41,
            "Name": "Aftertreatment 2 SCR Intermediate NH3 Gas Sensor Heater Control",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5861,
            "SPNLength": 2,
            "StartBit": 40,
            "TransmissionRate": "50 ms",
            "Units": "bit"
        },
        5862: {
            "Acronym": "A1SCREGT2",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 15,
            "Name": "Aftertreatment 1 SCR Intermediate Gas Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 5862,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "500 ms",
            "Units": "deg F"
        },
        5863: {
            "Acronym": "A1SCREGT2",
            "DataRange": "0 to 31",
            "EndBit": 20,
            "Name": "Aftertreatment 1 SCR Intermediate Gas Temperature Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5863,
            "SPNLength": 5,
            "StartBit": 16,
            "TransmissionRate": "500 ms",
            "Units": "count"
        },
        5864: {
            "Acronym": "A2SCREGT2",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 15,
            "Name": "Aftertreatment 2 SCR Intermediate Gas Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 5864,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "500 ms",
            "Units": "deg F"
        },
        5865: {
            "Acronym": "A2SCREGT2",
            "DataRange": "0 to 31",
            "EndBit": 20,
            "Name": "Aftertreatment 2 SCR Intermediate Gas Temperature Preliminary FMI",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5865,
            "SPNLength": 5,
            "StartBit": 16,
            "TransmissionRate": "500 ms",
            "Units": "count"
        },
        5866: {
            "Acronym": "EEC13",
            "DataRange": "0 to 15",
            "EndBit": 7,
            "Name": "Engine Fueling Inhibit Prevented Reason",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5866,
            "SPNLength": 4,
            "StartBit": 4,
            "TransmissionRate": "Every 1 sec and on change but no faster than 100 ms",
            "Units": "bit"
        },
        5867: {
            "Acronym": "GFI4",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 15,
            "Name": "Supply Gas Methane Percentage",
            "Offset": 0.0,
            "OperationalHigh": 160.6375,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "8 bytes",
            "Resolution": 0.0025,
            "SPN": 5867,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "200 ms",
            "Units": "%"
        },
        587: {
            "Acronym": "CI",
            "DataRange": "0 to 255 per byte",
            "EndBit": -2,
            "Name": "Model",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": -2,
            "SPN": 587,
            "SPNLength": 1600,
            "StartBit": -1,
            "TransmissionRate": "On request",
            "Units": "ASCII"
        },
        5872: {
            "Acronym": "HSS",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Hybrid System Warning Indicator",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5872,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "Every 1 sec and on change but no faster than 100 ms",
            "Units": "bit"
        },
        5873: {
            "Acronym": "HSS",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Hybrid System Overheat Indicator",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5873,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "Every 1 sec and on change but no faster than 100 ms",
            "Units": "bit"
        },
        5874: {
            "Acronym": "HSS",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Hybrid System Stop Indicator",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5874,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "Every 1 sec and on change but no faster than 100 ms",
            "Units": "bit"
        },
        588: {
            "Acronym": "CI",
            "DataRange": "0 to 255 per byte",
            "EndBit": -3,
            "Name": "Serial Number",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": -2,
            "SPN": 588,
            "SPNLength": 1600,
            "StartBit": -1,
            "TransmissionRate": "On request",
            "Units": "ASCII"
        },
        589: {
            "Acronym": "AS",
            "DataRange": "0 to 32,127.5 rpm",
            "EndBit": 15,
            "Name": "Alternator Speed",
            "Offset": 0.0,
            "OperationalHigh": 32127.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.5,
            "SPN": 589,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "1 sec",
            "Units": "rpm"
        },
        59: {
            "Acronym": "ETC3",
            "DataRange": "0 to 100 %",
            "EndBit": 7,
            "Name": "Transmission Shift Finger Gear Position",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 59,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        590: {
            "Acronym": "SHUTDN",
            "DataRange": "0 to 3",
            "EndBit": 7,
            "Name": "Engine Idle Shutdown Timer State",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 590,
            "SPNLength": 2,
            "StartBit": 6,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        591: {
            "Acronym": "SHUTDN",
            "DataRange": "0 to 3",
            "EndBit": 15,
            "Name": "Engine Idle Shutdown Timer Function",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 591,
            "SPNLength": 2,
            "StartBit": 14,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        5913: {
            "Acronym": "TRF2",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 47,
            "Name": "Transmission Oil Temperature 2",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 5913,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        5914: {
            "Acronym": "TRF2",
            "DataRange": "-62.5 to 62.5 L",
            "EndBit": 55,
            "Name": "Transmission Oil Level 2 High / Low",
            "Offset": -62.5,
            "OperationalHigh": 62.5,
            "OperationalLow": -62.5,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.5,
            "SPN": 5914,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "1 s",
            "Units": "liters"
        },
        5915: {
            "Acronym": "TRF2",
            "DataRange": "0 to 15",
            "EndBit": 59,
            "Name": "Transmission Oil Level 2 Countdown Timer",
            "Offset": 0.0,
            "OperationalHigh": 13.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 13 minutes",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5915,
            "SPNLength": 4,
            "StartBit": 56,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        5916: {
            "Acronym": "TRF2",
            "DataRange": "0 to 15",
            "EndBit": 63,
            "Name": "Transmission Oil Level 2 Measurement Status",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 5916,
            "SPNLength": 4,
            "StartBit": 60,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        592: {
            "Acronym": "SHUTDN",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Engine Idle Shutdown Timer Override",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 592,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        593: {
            "Acronym": "SHUTDN",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Engine Idle Shutdown has Shutdown Engine",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 593,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        594: {
            "Acronym": "SHUTDN",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Engine Idle Shutdown Driver Alert Mode",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 594,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        595: {
            "Acronym": "CCVS1",
            "DataRange": "0 to 3",
            "EndBit": 25,
            "Name": "Cruise Control Active",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 595,
            "SPNLength": 2,
            "StartBit": 24,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        596: {
            "Acronym": "CCVS1",
            "DataRange": "0 to 3",
            "EndBit": 27,
            "Name": "Cruise Control Enable Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 596,
            "SPNLength": 2,
            "StartBit": 26,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        597: {
            "Acronym": "CCVS1",
            "DataRange": "0 to 3",
            "EndBit": 29,
            "Name": "Brake Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 597,
            "SPNLength": 2,
            "StartBit": 28,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        598: {
            "Acronym": "CCVS1",
            "DataRange": "0 to 3",
            "EndBit": 31,
            "Name": "Clutch Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 598,
            "SPNLength": 2,
            "StartBit": 30,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        599: {
            "Acronym": "CCVS1",
            "DataRange": "0 to 3",
            "EndBit": 33,
            "Name": "Cruise Control Set Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 599,
            "SPNLength": 2,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        60: {
            "Acronym": "ETC3",
            "DataRange": "0 to 100 %",
            "EndBit": 15,
            "Name": "Transmission Shift Finger Rail Position",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 60,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "On request",
            "Units": "%"
        },
        600: {
            "Acronym": "CCVS1",
            "DataRange": "0 to 3",
            "EndBit": 35,
            "Name": "Cruise Control Coast (Decelerate) Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 600,
            "SPNLength": 2,
            "StartBit": 34,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        601: {
            "Acronym": "CCVS1",
            "DataRange": "0 to 3",
            "EndBit": 37,
            "Name": "Cruise Control Resume Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 601,
            "SPNLength": 2,
            "StartBit": 36,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        602: {
            "Acronym": "CCVS1",
            "DataRange": "0 to 3",
            "EndBit": 39,
            "Name": "Cruise Control Accelerate Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 602,
            "SPNLength": 2,
            "StartBit": 38,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        604: {
            "Acronym": "ETC5",
            "DataRange": "0 to 3",
            "EndBit": 11,
            "Name": "Transmission Neutral Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 604,
            "SPNLength": 2,
            "StartBit": 10,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        605: {
            "Acronym": "SHUTDN",
            "DataRange": "0 to 3",
            "EndBit": 21,
            "Name": "Refrigerant High Pressure Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 605,
            "SPNLength": 2,
            "StartBit": 20,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        606: {
            "Acronym": "ETC1",
            "DataRange": "0 to 3",
            "EndBit": 33,
            "Name": "Engine Momentary Overspeed Enable",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 606,
            "SPNLength": 2,
            "StartBit": 32,
            "TransmissionRate": "10 ms",
            "Units": "bit"
        },
        607: {
            "Acronym": "ETC1",
            "DataRange": "0 to 3",
            "EndBit": 35,
            "Name": "Progressive Shift Disable",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 607,
            "SPNLength": 2,
            "StartBit": 34,
            "TransmissionRate": "10 ms",
            "Units": "bit"
        },
        608: {
            "Acronym": "DM13",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "J1587 (previously SAE J1708 (J1587) Data Link)",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 608,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "Sent when ever a Stop or Start broadcast event is necessary. To maintain the modified state of the vehicle network(s) the commanding device must send the Hold Signal once every 5 seconds.",
            "Units": "bit"
        },
        619: {
            "Acronym": "B",
            "DataRange": "0 to 3",
            "EndBit": 25,
            "Name": "Parking Brake Actuator",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 619,
            "SPNLength": 2,
            "StartBit": 24,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        622: {
            "Acronym": "DM13",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "J1922 (previously SAE J1922 Data Link)",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 622,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "Sent when ever a Stop or Start broadcast event is necessary. To maintain the modified state of the vehicle network(s) the commanding device must send the Hold Signal once every 5 seconds.",
            "Units": "bit"
        },
        623: {
            "Acronym": "DM12",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Red Stop Lamp",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 623,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        624: {
            "Acronym": "DM12",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Amber Warning Lamp",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 624,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        626: {
            "Acronym": "CSA",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Engine Start Enable Device 1",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 626,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "As required",
            "Units": "bit"
        },
        632: {
            "Acronym": "EOI",
            "DataRange": "0 to 3",
            "EndBit": 27,
            "Name": "Engine Fuel Shutoff 1 Control",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 632,
            "SPNLength": 2,
            "StartBit": 26,
            "TransmissionRate": "250 ms",
            "Units": "bit"
        },
        633: {
            "Acronym": "TFAC",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 47,
            "Name": "Engine Fuel Actuator 1 Control Command",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 100%",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 633,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "50 msec (preferred) or Engine Speed Dependent (if required by application)",
            "Units": "%"
        },
        639: {
            "Acronym": "DM13",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "J1939 Network #1, Primary Vehicle Network ( previously SAE J1939 Data Link)",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 639,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "Sent when ever a Stop or Start broadcast event is necessary. To maintain the modified state of the vehicle network(s) the commanding device must send the Hold Signal once every 5 seconds.",
            "Units": "bit"
        },
        641: {
            "Acronym": "EEC6",
            "DataRange": "0 to 100 %",
            "EndBit": 23,
            "Name": "Engine Variable Geometry Turbocharger Actuator #1",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 641,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "100 msec (preferred) or Engine Speed Dependent (if required by application)",
            "Units": "%"
        },
        649: {
            "Acronym": "ETWAC",
            "DataRange": "0 to 160.6375 %",
            "EndBit": 47,
            "Name": "Engine Exhaust Back Pressure Regulator Control Command",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 100%",
            "PGNLength": 8,
            "Resolution": 0.0025,
            "SPN": 649,
            "SPNLength": 16,
            "StartBit": 32,
            "TransmissionRate": "Engine speed dependent",
            "Units": "%"
        },
        681: {
            "Acronym": "TC1",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Transmission Gear Shift Inhibit Request",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 681,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "When active; 50 ms to transmission and axles",
            "Units": "bit"
        },
        682: {
            "Acronym": "TC1",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Transmission Torque Converter Lockup Disable Request",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 682,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "When active; 50 ms to transmission and axles",
            "Units": "bit"
        },
        683: {
            "Acronym": "TC1",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Disengage Driveline Request",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 683,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "When active; 50 ms to transmission and axles",
            "Units": "bit"
        },
        684: {
            "Acronym": "TC1",
            "DataRange": "0 to 100 %",
            "EndBit": 15,
            "Name": "Requested Percent Clutch Slip",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 684,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "When active; 50 ms to transmission and axles",
            "Units": "%"
        },
        685: {
            "Acronym": "TC1",
            "DataRange": "0 to 3",
            "EndBit": 25,
            "Name": "Disengage Differential Lock Request - Front Axle 1",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 685,
            "SPNLength": 2,
            "StartBit": 24,
            "TransmissionRate": "When active; 50 ms to transmission and axles",
            "Units": "bit"
        },
        686: {
            "Acronym": "TC1",
            "DataRange": "0 to 3",
            "EndBit": 27,
            "Name": "Disengage Differential Lock Request - Front Axle 2",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 686,
            "SPNLength": 2,
            "StartBit": 26,
            "TransmissionRate": "When active; 50 ms to transmission and axles",
            "Units": "bit"
        },
        687: {
            "Acronym": "TC1",
            "DataRange": "0 to 3",
            "EndBit": 29,
            "Name": "Disengage Differential Lock Request - Rear Axle 1",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 687,
            "SPNLength": 2,
            "StartBit": 28,
            "TransmissionRate": "When active; 50 ms to transmission and axles",
            "Units": "bit"
        },
        688: {
            "Acronym": "TC1",
            "DataRange": "0 to 3",
            "EndBit": 31,
            "Name": "Disengage Differential Lock Request - Rear Axle 2",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 688,
            "SPNLength": 2,
            "StartBit": 30,
            "TransmissionRate": "When active; 50 ms to transmission and axles",
            "Units": "bit"
        },
        689: {
            "Acronym": "TC1",
            "DataRange": "0 to 3",
            "EndBit": 33,
            "Name": "Disengage Differential Lock Request - Central",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 689,
            "SPNLength": 2,
            "StartBit": 32,
            "TransmissionRate": "When active; 50 ms to transmission and axles",
            "Units": "bit"
        },
        69: {
            "Acronym": "CCVS1",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Two Speed Axle Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 69,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        690: {
            "Acronym": "TC1",
            "DataRange": "0 to 3",
            "EndBit": 35,
            "Name": "Disengage Differential Lock Request - Central Front",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 690,
            "SPNLength": 2,
            "StartBit": 34,
            "TransmissionRate": "When active; 50 ms to transmission and axles",
            "Units": "bit"
        },
        691: {
            "Acronym": "TC1",
            "DataRange": "0 to 3",
            "EndBit": 37,
            "Name": "Disengage Differential Lock Request - Central Rear",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 691,
            "SPNLength": 2,
            "StartBit": 36,
            "TransmissionRate": "When active; 50 ms to transmission and axles",
            "Units": "bit"
        },
        695: {
            "Acronym": "TSC1",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Engine Override Control Mode",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 695,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "To engine: Control Purpose dependent or 10 ms\nTo retarder: 50 ms",
            "Units": "bit"
        },
        696: {
            "Acronym": "TSC1",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Engine Requested Speed Control Conditions",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 696,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "To engine: Control Purpose dependent or 10 ms\nTo retarder: 50 ms",
            "Units": "bit"
        },
        70: {
            "Acronym": "CCVS1",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Parking Brake Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 70,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        701: {
            "Acronym": "AUXIO1",
            "DataRange": "0 to 3",
            "EndBit": 7,
            "Name": "Auxiliary I/O #01",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 701,
            "SPNLength": 2,
            "StartBit": 6,
            "TransmissionRate": "manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        702: {
            "Acronym": "AUXIO1",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Auxiliary I/O #02",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 702,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        703: {
            "Acronym": "AUXIO1",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Auxiliary I/O #03",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 703,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        704: {
            "Acronym": "AUXIO1",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Auxiliary I/O #04",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 704,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        705: {
            "Acronym": "AUXIO1",
            "DataRange": "0 to 3",
            "EndBit": 15,
            "Name": "Auxiliary I/O #05",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 705,
            "SPNLength": 2,
            "StartBit": 14,
            "TransmissionRate": "manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        706: {
            "Acronym": "AUXIO1",
            "DataRange": "0 to 3",
            "EndBit": 13,
            "Name": "Auxiliary I/O #06",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 706,
            "SPNLength": 2,
            "StartBit": 12,
            "TransmissionRate": "manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        707: {
            "Acronym": "AUXIO1",
            "DataRange": "0 to 3",
            "EndBit": 11,
            "Name": "Auxiliary I/O #07",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 707,
            "SPNLength": 2,
            "StartBit": 10,
            "TransmissionRate": "manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        708: {
            "Acronym": "AUXIO1",
            "DataRange": "0 to 3",
            "EndBit": 9,
            "Name": "Auxiliary I/O #08",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 708,
            "SPNLength": 2,
            "StartBit": 8,
            "TransmissionRate": "manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        709: {
            "Acronym": "AUXIO1",
            "DataRange": "0 to 3",
            "EndBit": 23,
            "Name": "Auxiliary I/O #09",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 709,
            "SPNLength": 2,
            "StartBit": 22,
            "TransmissionRate": "manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        710: {
            "Acronym": "AUXIO1",
            "DataRange": "0 to 3",
            "EndBit": 21,
            "Name": "Auxiliary I/O #10",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 710,
            "SPNLength": 2,
            "StartBit": 20,
            "TransmissionRate": "manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        711: {
            "Acronym": "AUXIO1",
            "DataRange": "0 to 3",
            "EndBit": 19,
            "Name": "Auxiliary I/O #11",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 711,
            "SPNLength": 2,
            "StartBit": 18,
            "TransmissionRate": "manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        712: {
            "Acronym": "AUXIO1",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Auxiliary I/O #12",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 712,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        713: {
            "Acronym": "AUXIO1",
            "DataRange": "0 to 3",
            "EndBit": 31,
            "Name": "Auxiliary I/O #13",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 713,
            "SPNLength": 2,
            "StartBit": 30,
            "TransmissionRate": "manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        714: {
            "Acronym": "AUXIO1",
            "DataRange": "0 to 3",
            "EndBit": 29,
            "Name": "Auxiliary I/O #14",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 714,
            "SPNLength": 2,
            "StartBit": 28,
            "TransmissionRate": "manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        715: {
            "Acronym": "AUXIO1",
            "DataRange": "0 to 3",
            "EndBit": 27,
            "Name": "Auxiliary I/O #15",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 715,
            "SPNLength": 2,
            "StartBit": 26,
            "TransmissionRate": "manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        716: {
            "Acronym": "AUXIO1",
            "DataRange": "0 to 3",
            "EndBit": 25,
            "Name": "Auxiliary I/O #16",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 716,
            "SPNLength": 2,
            "StartBit": 24,
            "TransmissionRate": "manufacturer defined, not faster than 100 ms",
            "Units": "bit"
        },
        72: {
            "Acronym": "A1",
            "DataRange": "0 to 100 %",
            "EndBit": 7,
            "Name": "Engine Blower Bypass Valve Position",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 72,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "500 ms",
            "Units": "%"
        },
        723: {
            "Acronym": "ESSI",
            "DataRange": "0 to 32,127.5 rpm",
            "EndBit": 31,
            "Name": "Engine Speed 2",
            "Offset": 0.0,
            "OperationalHigh": 32127.5,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.5,
            "SPN": 723,
            "SPNLength": 16,
            "StartBit": 16,
            "TransmissionRate": "Transmitted only after requested.  After request, broadcast rate is engine speed dependent.  Update stopped after key switch cycle.",
            "Units": "rpm"
        },
        73: {
            "Acronym": "AWPP",
            "DataRange": "0.0 to 580.150952 psi",
            "EndBit": 7,
            "Name": "Auxiliary Pump Pressure",
            "Offset": 0.0,
            "OperationalHigh": 580.150952,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 2.320603808,
            "SPN": 73,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "psi"
        },
        74: {
            "Acronym": "CCSS",
            "DataRange": "0.0 to 155.34275 mph",
            "EndBit": 7,
            "Name": "Maximum Vehicle Speed Limit",
            "Offset": 0.0,
            "OperationalHigh": 155.34275,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.621371,
            "SPN": 74,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "mph"
        },
        740: {
            "Acronym": "ETC3",
            "DataRange": "0 to 3",
            "EndBit": 43,
            "Name": "Transmission Lockup Clutch Actuator",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 740,
            "SPNLength": 2,
            "StartBit": 42,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        748: {
            "Acronym": "ERC2",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Transmission Output Retarder",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 748,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "Every 1 s and on change of state but no faster than every 100 ms.  Grandfathered definition for systems that implemented this message prior to July, 2010: 1 s when active; or on change of state",
            "Units": "bit"
        },
        75: {
            "Acronym": "AI",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 7,
            "Name": "Steering Axle Temperature",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 75,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        767: {
            "Acronym": "ETC5",
            "DataRange": "0 to 3",
            "EndBit": 9,
            "Name": "Transmission Reverse Direction Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 767,
            "SPNLength": 2,
            "StartBit": 8,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        768: {
            "Acronym": "ETC3",
            "DataRange": "0 to 3",
            "EndBit": 33,
            "Name": "Transmission Range High Actuator",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 768,
            "SPNLength": 2,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        769: {
            "Acronym": "ETC3",
            "DataRange": "0 to 3",
            "EndBit": 35,
            "Name": "Transmission Range Low Actuator",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 769,
            "SPNLength": 2,
            "StartBit": 34,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        770: {
            "Acronym": "ETC3",
            "DataRange": "0 to 3",
            "EndBit": 37,
            "Name": "Transmission Splitter Direct Actuator",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 770,
            "SPNLength": 2,
            "StartBit": 36,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        771: {
            "Acronym": "ETC3",
            "DataRange": "0 to 3",
            "EndBit": 39,
            "Name": "Transmission Splitter Indirect Actuator",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 771,
            "SPNLength": 2,
            "StartBit": 38,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        772: {
            "Acronym": "ETC3",
            "DataRange": "0 to 3",
            "EndBit": 25,
            "Name": "Transmission Shift Finger Rail Actuator 1",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 772,
            "SPNLength": 2,
            "StartBit": 24,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        773: {
            "Acronym": "ETC3",
            "DataRange": "0 to 3",
            "EndBit": 27,
            "Name": "Transmission Shift Finger Gear Actuator 1",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 773,
            "SPNLength": 2,
            "StartBit": 26,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        778: {
            "Acronym": "ETC5",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Transmission High Range Sense Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 778,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        779: {
            "Acronym": "ETC5",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Transmission Low Range Sense Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 779,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        780: {
            "Acronym": "ETC3",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "Transmission Shift Finger Neutral Indicator",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 780,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        781: {
            "Acronym": "ETC3",
            "DataRange": "0 to 3",
            "EndBit": 19,
            "Name": "Transmission Shift Finger Engagement Indicator",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 781,
            "SPNLength": 2,
            "StartBit": 18,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        782: {
            "Acronym": "ETC3",
            "DataRange": "0 to 3",
            "EndBit": 21,
            "Name": "Transmission Shift Finger Center Rail Indicator",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 782,
            "SPNLength": 2,
            "StartBit": 20,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        783: {
            "Acronym": "ETC3",
            "DataRange": "0 to 3",
            "EndBit": 29,
            "Name": "Transmission Shift Finger Rail Actuator 2",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 783,
            "SPNLength": 2,
            "StartBit": 28,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        784: {
            "Acronym": "ETC3",
            "DataRange": "0 to 3",
            "EndBit": 31,
            "Name": "Transmission Shift Finger Gear Actuator 2",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 784,
            "SPNLength": 2,
            "StartBit": 30,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        786: {
            "Acronym": "ETC3",
            "DataRange": "0 to 3",
            "EndBit": 45,
            "Name": "Transmission Defuel Actuator",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 786,
            "SPNLength": 2,
            "StartBit": 44,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        787: {
            "Acronym": "ETC3",
            "DataRange": "0 to 3",
            "EndBit": 47,
            "Name": "Transmission Inertia Brake Actuator",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 787,
            "SPNLength": 2,
            "StartBit": 46,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        788: {
            "Acronym": "ETC3",
            "DataRange": "0 to 3",
            "EndBit": 41,
            "Name": "Transmission Clutch Actuator",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 788,
            "SPNLength": 2,
            "StartBit": 40,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        79: {
            "Acronym": "AMB",
            "DataRange": "-459.40000000000003 to 3154.94375 deg F",
            "EndBit": 63,
            "Name": "Road Surface Temperature",
            "Offset": -459.40000000000003,
            "OperationalHigh": 3154.94375,
            "OperationalLow": -459.40000000000003,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.05625,
            "SPN": 79,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "1 s",
            "Units": "deg F"
        },
        80: {
            "Acronym": "DD",
            "DataRange": "0 to 100 %",
            "EndBit": 7,
            "Name": "Washer Fluid Level",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 80,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "1s",
            "Units": "%"
        },
        81: {
            "Acronym": "IC1",
            "DataRange": "0.0 to 18.12971725 psi",
            "EndBit": 7,
            "Name": "Engine DPF Intake Pressure",
            "Offset": 0.0,
            "OperationalHigh": 18.12971725,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.072518869,
            "SPN": 81,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "0.5 s",
            "Units": "psi"
        },
        82: {
            "Acronym": "AIR2",
            "DataRange": "0.0 to 145.037738 psi",
            "EndBit": 7,
            "Name": "Engine Air Start Pressure",
            "Offset": 0.0,
            "OperationalHigh": 145.037738,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.580150952,
            "SPN": 82,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "psi"
        },
        84: {
            "Acronym": "CCVS1",
            "DataRange": "0.0 to 155.961635516 mph",
            "EndBit": 23,
            "Name": "Wheel-Based Vehicle Speed",
            "Offset": 0.0,
            "OperationalHigh": 155.961635516,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.00242723046875,
            "SPN": 84,
            "SPNLength": 16,
            "StartBit": 8,
            "TransmissionRate": "100 ms",
            "Units": "mph"
        },
        86: {
            "Acronym": "CCVS1",
            "DataRange": "0.0 to 155.34275 mph",
            "EndBit": 47,
            "Name": "Cruise Control Set Speed",
            "Offset": 0.0,
            "OperationalHigh": 155.34275,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.621371,
            "SPN": 86,
            "SPNLength": 8,
            "StartBit": 40,
            "TransmissionRate": "100 ms",
            "Units": "mph"
        },
        87: {
            "Acronym": "CCSS",
            "DataRange": "0.0 to 155.34275 mph",
            "EndBit": 15,
            "Name": "Cruise Control High Set Limit Speed",
            "Offset": 0.0,
            "OperationalHigh": 155.34275,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.621371,
            "SPN": 87,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "On request",
            "Units": "mph"
        },
        875: {
            "Acronym": "SHUTDN",
            "DataRange": "0 to 3",
            "EndBit": 19,
            "Name": "Refrigerant Low Pressure Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 875,
            "SPNLength": 2,
            "StartBit": 18,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        88: {
            "Acronym": "CCSS",
            "DataRange": "0.0 to 155.34275 mph",
            "EndBit": 23,
            "Name": "Cruise Control Low Set Limit Speed",
            "Offset": 0.0,
            "OperationalHigh": 155.34275,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.621371,
            "SPN": 88,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "mph"
        },
        897: {
            "Acronym": "TSC1",
            "DataRange": "0 to 3",
            "EndBit": 5,
            "Name": "Override Control Mode Priority",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 897,
            "SPNLength": 2,
            "StartBit": 4,
            "TransmissionRate": "To engine: Control Purpose dependent or 10 ms\nTo retarder: 50 ms",
            "Units": "bit"
        },
        898: {
            "Acronym": "TSC1",
            "DataRange": "0 to 8,031.875 rpm",
            "EndBit": 23,
            "Name": "Engine Requested Speed/Speed Limit",
            "Offset": 0.0,
            "OperationalHigh": 8031.875,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.125,
            "SPN": 898,
            "SPNLength": 16,
            "StartBit": 8,
            "TransmissionRate": "To engine: Control Purpose dependent or 10 ms\nTo retarder: 50 ms",
            "Units": "rpm"
        },
        899: {
            "Acronym": "EEC1",
            "DataRange": "0 to 15",
            "EndBit": 3,
            "Name": "Engine Torque Mode",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 899,
            "SPNLength": 4,
            "StartBit": 0,
            "TransmissionRate": "engine speed dependent",
            "Units": "bit"
        },
        90: {
            "Acronym": "PTO",
            "DataRange": "-40.0 to 410.0 deg F",
            "EndBit": 7,
            "Name": "Power Takeoff Oil Temperature",
            "Offset": -40.0,
            "OperationalHigh": 410.0,
            "OperationalLow": -40.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1.8,
            "SPN": 90,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "deg F"
        },
        900: {
            "Acronym": "ERC1",
            "DataRange": "0 to 15",
            "EndBit": 3,
            "Name": "Retarder Torque Mode",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 900,
            "SPNLength": 4,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        901: {
            "Acronym": "RC",
            "DataRange": "0 to 15",
            "EndBit": 3,
            "Name": "Retarder Type",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 19,
            "Resolution": 1,
            "SPN": 901,
            "SPNLength": 4,
            "StartBit": 0,
            "TransmissionRate": "Every 5 s and on change of torque/speed points of more than 10% since last transmission but no faster than every 500 ms",
            "Units": "bit"
        },
        902: {
            "Acronym": "RC",
            "DataRange": "0 to 15",
            "EndBit": 7,
            "Name": "Retarder Location",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 19,
            "Resolution": 1,
            "SPN": 902,
            "SPNLength": 4,
            "StartBit": 4,
            "TransmissionRate": "Every 5 s and on change of torque/speed points of more than 10% since last transmission but no faster than every 500 ms",
            "Units": "bit"
        },
        903: {
            "Acronym": "ETC5",
            "DataRange": "0 to 3",
            "EndBit": 13,
            "Name": "Transmission Forward Direction Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 903,
            "SPNLength": 2,
            "StartBit": 12,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        904: {
            "Acronym": "EBC2",
            "DataRange": "0.0 to 155.961635516 mph",
            "EndBit": 15,
            "Name": "Front Axle Speed",
            "Offset": 0.0,
            "OperationalHigh": 155.961635516,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.00242723046875,
            "SPN": 904,
            "SPNLength": 16,
            "StartBit": 0,
            "TransmissionRate": "100 ms",
            "Units": "mph"
        },
        905: {
            "Acronym": "EBC2",
            "DataRange": "-4.8544609375 to 4.8544609375 mph",
            "EndBit": 23,
            "Name": "Relative Speed; Front Axle, Left Wheel",
            "Offset": -4.46222049375,
            "OperationalHigh": 4.8544609375,
            "OperationalLow": -4.8544609375,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0388356875,
            "SPN": 905,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "100 ms",
            "Units": "mph"
        },
        906: {
            "Acronym": "EBC2",
            "DataRange": "-4.8544609375 to 4.8544609375 mph",
            "EndBit": 31,
            "Name": "Relative Speed; Front Axle, Right Wheel",
            "Offset": -4.46222049375,
            "OperationalHigh": 4.8544609375,
            "OperationalLow": -4.8544609375,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0388356875,
            "SPN": 906,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "100 ms",
            "Units": "mph"
        },
        907: {
            "Acronym": "EBC2",
            "DataRange": "-4.8544609375 to 4.8544609375 mph",
            "EndBit": 39,
            "Name": "Relative Speed; Rear Axle #1, Left Wheel",
            "Offset": -4.46222049375,
            "OperationalHigh": 4.8544609375,
            "OperationalLow": -4.8544609375,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0388356875,
            "SPN": 907,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "mph"
        },
        908: {
            "Acronym": "EBC2",
            "DataRange": "-4.8544609375 to 4.8544609375 mph",
            "EndBit": 47,
            "Name": "Relative Speed; Rear Axle #1, Right Wheel",
            "Offset": -4.46222049375,
            "OperationalHigh": 4.8544609375,
            "OperationalLow": -4.8544609375,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0388356875,
            "SPN": 908,
            "SPNLength": 8,
            "StartBit": 40,
            "TransmissionRate": "100 ms",
            "Units": "mph"
        },
        909: {
            "Acronym": "EBC2",
            "DataRange": "-4.8544609375 to 4.8544609375 mph",
            "EndBit": 55,
            "Name": "Relative Speed; Rear Axle #2, Left Wheel",
            "Offset": -4.46222049375,
            "OperationalHigh": 4.8544609375,
            "OperationalLow": -4.8544609375,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0388356875,
            "SPN": 909,
            "SPNLength": 8,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "mph"
        },
        91: {
            "Acronym": "EEC2",
            "DataRange": "0 to 100 %",
            "EndBit": 15,
            "Name": "Accelerator Pedal Position 1",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 91,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "50 msec (preferred) or Engine Speed Dependent (if required by application)",
            "Units": "%"
        },
        910: {
            "Acronym": "EBC2",
            "DataRange": "-4.8544609375 to 4.8544609375 mph",
            "EndBit": 63,
            "Name": "Relative Speed; Rear Axle #2, Right Wheel",
            "Offset": -4.46222049375,
            "OperationalHigh": 4.8544609375,
            "OperationalLow": -4.8544609375,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.0388356875,
            "SPN": 910,
            "SPNLength": 8,
            "StartBit": 56,
            "TransmissionRate": "100 ms",
            "Units": "mph"
        },
        911: {
            "Acronym": "SERV",
            "DataRange": "0 to 250 ID",
            "EndBit": 7,
            "Name": "Service Component Identification",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "8 bytes or Variable",
            "Resolution": 1,
            "SPN": 911,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "ID"
        },
        912: {
            "Acronym": "SERV",
            "DataRange": "0 to 250 ID",
            "EndBit": 31,
            "Name": "Service Component Identification",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "8 bytes or Variable",
            "Resolution": 1,
            "SPN": 912,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "On request",
            "Units": "ID"
        },
        913: {
            "Acronym": "SERV",
            "DataRange": "0 to 250 ID",
            "EndBit": 47,
            "Name": "Service Component Identification",
            "Offset": 0.0,
            "OperationalHigh": 250.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "8 bytes or Variable",
            "Resolution": 1,
            "SPN": 913,
            "SPNLength": 8,
            "StartBit": 40,
            "TransmissionRate": "On request",
            "Units": "ID"
        },
        914: {
            "Acronym": "SERV",
            "DataRange": "-160,635 to 160,640 km",
            "EndBit": 23,
            "Name": "Service Distance",
            "Offset": -160635000.0,
            "OperationalHigh": 160640.0,
            "OperationalLow": -160635.0,
            "OperationalRange": "",
            "PGNLength": "8 bytes or Variable",
            "Resolution": 5,
            "SPN": 914,
            "SPNLength": 16,
            "StartBit": 8,
            "TransmissionRate": "On request",
            "Units": "m"
        },
        915: {
            "Acronym": "SERV",
            "DataRange": "-125 to 125 weeks",
            "EndBit": 39,
            "Name": "Service Delay/Calendar Time Based",
            "Offset": -125.0,
            "OperationalHigh": 125.0,
            "OperationalLow": -125.0,
            "OperationalRange": "",
            "PGNLength": "8 bytes or Variable",
            "Resolution": 1,
            "SPN": 915,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "Weeks"
        },
        916: {
            "Acronym": "SERV",
            "DataRange": "-32,127 to 32,128 hr",
            "EndBit": 63,
            "Name": "Service Delay/Operational Time Based",
            "Offset": -115657200.0,
            "OperationalHigh": 32128.0,
            "OperationalLow": -32127.0,
            "OperationalRange": "",
            "PGNLength": "8 bytes or Variable",
            "Resolution": 1,
            "SPN": 916,
            "SPNLength": 16,
            "StartBit": 48,
            "TransmissionRate": "On request",
            "Units": "hours"
        },
        917: {
            "Acronym": "VDHR",
            "DataRange": "0 to 21055406000 meters",
            "EndBit": 31,
            "Name": "High Resolution Total Vehicle Distance",
            "Offset": 0.0,
            "OperationalHigh": 21055406000.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 5,
            "SPN": 917,
            "SPNLength": 32,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "meters"
        },
        918: {
            "Acronym": "VDHR",
            "DataRange": "0 to 21055406000 meters",
            "EndBit": 63,
            "Name": "High Resolution Trip Distance",
            "Offset": 0.0,
            "OperationalHigh": 21055406000.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 5,
            "SPN": 918,
            "SPNLength": 32,
            "StartBit": 32,
            "TransmissionRate": "1 s",
            "Units": "meters"
        },
        92: {
            "Acronym": "EEC2",
            "DataRange": "0 to 250 %",
            "EndBit": 23,
            "Name": "Engine Percent Load At Current Speed",
            "Offset": 0.0,
            "OperationalHigh": 125.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 125%",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 92,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "50 msec (preferred) or Engine Speed Dependent (if required by application)",
            "Units": "%"
        },
        927: {
            "Acronym": "EAC1",
            "DataRange": "0 to 255",
            "EndBit": 7,
            "Name": "Location",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 927,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "500 ms",
            "Units": "bit"
        },
        928: {
            "Acronym": "VW",
            "DataRange": "0 to 255",
            "EndBit": 7,
            "Name": "Axle Location",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 928,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        929: {
            "Acronym": "TIRE",
            "DataRange": "0 to 255",
            "EndBit": 7,
            "Name": "Tire Location",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 929,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "10 s",
            "Units": "bit"
        },
        930: {
            "Acronym": "AI",
            "DataRange": "0 to 255",
            "EndBit": 15,
            "Name": "Drive Axle Location",
            "Offset": 0.0,
            "OperationalHigh": 255.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 930,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        94: {
            "Acronym": "EFL/P1",
            "DataRange": "0.0 to 145.037738 psi",
            "EndBit": 7,
            "Name": "Engine Fuel Delivery Pressure",
            "Offset": 0.0,
            "OperationalHigh": 145.037738,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.580150952,
            "SPN": 94,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "0.5 s",
            "Units": "psi"
        },
        95: {
            "Acronym": "DD",
            "DataRange": "0.0 to 72.518869 psi",
            "EndBit": 23,
            "Name": "Engine Fuel Filter Differential Pressure",
            "Offset": 0.0,
            "OperationalHigh": 72.518869,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.290075476,
            "SPN": 95,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "1s",
            "Units": "psi"
        },
        957: {
            "Acronym": "TCFG",
            "DataRange": "0 to 250",
            "EndBit": 15,
            "Name": "Number of Forward Gear Ratios",
            "Offset": 0.0,
            "OperationalHigh": 125.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 125 gear ratios",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 957,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "On request",
            "Units": "gear value"
        },
        958: {
            "Acronym": "TCFG",
            "DataRange": "0 to 250",
            "EndBit": 7,
            "Name": "Number of Reverse Gear Ratios",
            "Offset": 0.0,
            "OperationalHigh": 125.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 125 gear ratios",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 958,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "gear value"
        },
        959: {
            "Acronym": "TD",
            "DataRange": "0 to 62.5 s",
            "EndBit": 7,
            "Name": "Seconds",
            "Offset": 0.0,
            "OperationalHigh": 59.75,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 59.75 sec",
            "PGNLength": 8,
            "Resolution": 0.25,
            "SPN": 959,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "Seconds"
        },
        96: {
            "Acronym": "DD",
            "DataRange": "0 to 100 %",
            "EndBit": 15,
            "Name": "Fuel Level 1",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 96,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "1s",
            "Units": "%"
        },
        960: {
            "Acronym": "TD",
            "DataRange": "0 to 250 mins",
            "EndBit": 15,
            "Name": "Minutes",
            "Offset": 0.0,
            "OperationalHigh": 59.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 59 min",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 960,
            "SPNLength": 8,
            "StartBit": 8,
            "TransmissionRate": "On request",
            "Units": "minutes"
        },
        961: {
            "Acronym": "TD",
            "DataRange": "0 to 250 hr",
            "EndBit": 23,
            "Name": "Hours",
            "Offset": 0.0,
            "OperationalHigh": 23.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 23 hr",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 961,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "On request",
            "Units": "hours"
        },
        962: {
            "Acronym": "TD",
            "DataRange": "0 to 62.5 days",
            "EndBit": 39,
            "Name": "Day",
            "Offset": 0.0,
            "OperationalHigh": 31.75,
            "OperationalLow": 0.25,
            "OperationalRange": "0.25 to 31.75 day",
            "PGNLength": 8,
            "Resolution": 0.25,
            "SPN": 962,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "Days"
        },
        963: {
            "Acronym": "TD",
            "DataRange": "0 to 250 months",
            "EndBit": 31,
            "Name": "Month",
            "Offset": 0.0,
            "OperationalHigh": 12.0,
            "OperationalLow": 1.0,
            "OperationalRange": "1 to 12 month",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 963,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "On request",
            "Units": "Months"
        },
        964: {
            "Acronym": "TD",
            "DataRange": "1985 to 2235 years",
            "EndBit": 47,
            "Name": "Year",
            "Offset": 1985.0,
            "OperationalHigh": 2235.0,
            "OperationalLow": 1985.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 964,
            "SPNLength": 8,
            "StartBit": 40,
            "TransmissionRate": "On request",
            "Units": "Years"
        },
        965: {
            "Acronym": "SOFT",
            "DataRange": "0 to 250 steps",
            "EndBit": 7,
            "Name": "Number of Software Identification Fields",
            "Offset": 0.0,
            "OperationalHigh": 125.0,
            "OperationalLow": 0.0,
            "OperationalRange": "0 to 125",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 965,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "step"
        },
        966: {
            "Acronym": "CCVS1",
            "DataRange": "0 to 3",
            "EndBit": 61,
            "Name": "Engine Test Mode Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 966,
            "SPNLength": 2,
            "StartBit": 60,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        967: {
            "Acronym": "CCVS1",
            "DataRange": "0 to 3",
            "EndBit": 59,
            "Name": "Engine Idle Decrement Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 967,
            "SPNLength": 2,
            "StartBit": 58,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        968: {
            "Acronym": "CCVS1",
            "DataRange": "0 to 3",
            "EndBit": 57,
            "Name": "Engine Idle Increment Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 968,
            "SPNLength": 2,
            "StartBit": 56,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        969: {
            "Acronym": "EBC1",
            "DataRange": "0 to 3",
            "EndBit": 31,
            "Name": "Remote Accelerator Enable Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 969,
            "SPNLength": 2,
            "StartBit": 30,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        97: {
            "Acronym": "OI",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Water In Fuel Indicator",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 97,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "10 s or on change but no more often than 1s",
            "Units": "bit"
        },
        970: {
            "Acronym": "EBC1",
            "DataRange": "0 to 3",
            "EndBit": 29,
            "Name": "Engine Auxiliary Shutdown Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 970,
            "SPNLength": 2,
            "StartBit": 28,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        971: {
            "Acronym": "EBC1",
            "DataRange": "0 to 3",
            "EndBit": 27,
            "Name": "Engine Derate Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 971,
            "SPNLength": 2,
            "StartBit": 26,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        972: {
            "Acronym": "EBC1",
            "DataRange": "0 to 3",
            "EndBit": 25,
            "Name": "Accelerator Interlock Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 972,
            "SPNLength": 2,
            "StartBit": 24,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        973: {
            "Acronym": "EBC1",
            "DataRange": "0 to 100 %",
            "EndBit": 39,
            "Name": "Engine Retarder Selection",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 973,
            "SPNLength": 8,
            "StartBit": 32,
            "TransmissionRate": "100 ms",
            "Units": "%"
        },
        974: {
            "Acronym": "EEC2",
            "DataRange": "0 to 100 %",
            "EndBit": 31,
            "Name": "Remote Accelerator Pedal Position",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 974,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "50 msec (preferred) or Engine Speed Dependent (if required by application)",
            "Units": "%"
        },
        975: {
            "Acronym": "FD1",
            "DataRange": "0 to 100 %",
            "EndBit": 7,
            "Name": "Estimated Percent Fan Speed",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 975,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "%"
        },
        976: {
            "Acronym": "CCVS1",
            "DataRange": "0 to 31",
            "EndBit": 52,
            "Name": "PTO Governor State",
            "Offset": 0.0,
            "OperationalHigh": 31.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 976,
            "SPNLength": 5,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        977: {
            "Acronym": "FD1",
            "DataRange": "0 to 15",
            "EndBit": 11,
            "Name": "Fan Drive State",
            "Offset": 0.0,
            "OperationalHigh": 15.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 977,
            "SPNLength": 4,
            "StartBit": 8,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        978: {
            "Acronym": "PTO",
            "DataRange": "0 to 3",
            "EndBit": 45,
            "Name": "Engine Remote PTO Governor Variable Speed Control Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 978,
            "SPNLength": 2,
            "StartBit": 44,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        979: {
            "Acronym": "PTO",
            "DataRange": "0 to 3",
            "EndBit": 43,
            "Name": "Engine Remote PTO Governor Preprogrammed Speed Control Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 979,
            "SPNLength": 2,
            "StartBit": 42,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        98: {
            "Acronym": "EFL/P1",
            "DataRange": "0 to 100 %",
            "EndBit": 23,
            "Name": "Engine Oil Level",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 98,
            "SPNLength": 8,
            "StartBit": 16,
            "TransmissionRate": "0.5 s",
            "Units": "%"
        },
        980: {
            "Acronym": "PTO",
            "DataRange": "0 to 3",
            "EndBit": 41,
            "Name": "Engine PTO Governor Enable Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 980,
            "SPNLength": 2,
            "StartBit": 40,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        981: {
            "Acronym": "PTO",
            "DataRange": "0 to 3",
            "EndBit": 55,
            "Name": "Engine PTO Governor Accelerate Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 981,
            "SPNLength": 2,
            "StartBit": 54,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        982: {
            "Acronym": "PTO",
            "DataRange": "0 to 3",
            "EndBit": 53,
            "Name": "Engine PTO Governor Resume Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 982,
            "SPNLength": 2,
            "StartBit": 52,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        983: {
            "Acronym": "PTO",
            "DataRange": "0 to 3",
            "EndBit": 51,
            "Name": "Engine PTO Governor Coast/Decelerate Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 983,
            "SPNLength": 2,
            "StartBit": 50,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        984: {
            "Acronym": "PTO",
            "DataRange": "0 to 3",
            "EndBit": 49,
            "Name": "Engine PTO Governor Set Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 984,
            "SPNLength": 2,
            "StartBit": 48,
            "TransmissionRate": "100 ms",
            "Units": "bit"
        },
        985: {
            "Acronym": "SHUTDN",
            "DataRange": "0 to 3",
            "EndBit": 17,
            "Name": "A/C High Pressure Fan Switch",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 985,
            "SPNLength": 2,
            "StartBit": 16,
            "TransmissionRate": "1 s",
            "Units": "bit"
        },
        986: {
            "Acronym": "CM1",
            "DataRange": "0 to 100 %",
            "EndBit": 7,
            "Name": "Requested Percent Fan Speed",
            "Offset": 0.0,
            "OperationalHigh": 100.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.4,
            "SPN": 986,
            "SPNLength": 8,
            "StartBit": 0,
            "TransmissionRate": "1 s",
            "Units": "%"
        },
        987: {
            "Acronym": "DM12",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Protect Lamp",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": "Variable",
            "Resolution": 1,
            "SPN": 987,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "bit"
        },
        988: {
            "Acronym": "RESET",
            "DataRange": "0 to 3",
            "EndBit": 1,
            "Name": "Trip Group 1",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 988,
            "SPNLength": 2,
            "StartBit": 0,
            "TransmissionRate": "When needed",
            "Units": "bit"
        },
        989: {
            "Acronym": "RESET",
            "DataRange": "0 to 3",
            "EndBit": 3,
            "Name": "Trip Group 2 - Proprietary",
            "Offset": 0.0,
            "OperationalHigh": 3.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 1,
            "SPN": 989,
            "SPNLength": 2,
            "StartBit": 2,
            "TransmissionRate": "When needed",
            "Units": "bit"
        },
        99: {
            "Acronym": "DD",
            "DataRange": "0.0 to 18.12971725 psi",
            "EndBit": 31,
            "Name": "Engine Oil Filter Differential Pressure",
            "Offset": 0.0,
            "OperationalHigh": 18.12971725,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 8,
            "Resolution": 0.072518869,
            "SPN": 99,
            "SPNLength": 8,
            "StartBit": 24,
            "TransmissionRate": "1s",
            "Units": "psi"
        },
        990: {
            "Acronym": "CBI",
            "DataRange": "0 to 526,385,151.9 km",
            "EndBit": 31,
            "Name": "Total Compression Brake Distance",
            "Offset": 0.0,
            "OperationalHigh": 526385151.9,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 16,
            "Resolution": 0.125,
            "SPN": 990,
            "SPNLength": 32,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "m"
        },
        991: {
            "Acronym": "CBI",
            "DataRange": "0.0 to 327080468.2212549 mile",
            "EndBit": 63,
            "Name": "Trip Compression Brake Distance",
            "Offset": 0.0,
            "OperationalHigh": 327080468.2212549,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 16,
            "Resolution": 0.077671375,
            "SPN": 991,
            "SPNLength": 32,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "miles"
        },
        992: {
            "Acronym": "CBI",
            "DataRange": "0.0 to 327080468.2212549 mile",
            "EndBit": 95,
            "Name": "Trip Service Brake Distance",
            "Offset": 0.0,
            "OperationalHigh": 327080468.2212549,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 16,
            "Resolution": 0.077671375,
            "SPN": 992,
            "SPNLength": 32,
            "StartBit": 64,
            "TransmissionRate": "On request",
            "Units": "miles"
        },
        993: {
            "Acronym": "CBI",
            "DataRange": "0 to 4,227,858,431 appl",
            "EndBit": 127,
            "Name": "Trip Service Brake Applications",
            "Offset": 0.0,
            "OperationalHigh": 4227858431.0,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 16,
            "Resolution": 1,
            "SPN": 993,
            "SPNLength": 32,
            "StartBit": 96,
            "TransmissionRate": "On request",
            "Units": "brake appl"
        },
        994: {
            "Acronym": "TFI",
            "DataRange": "0 to 210,554,060.75 hr",
            "EndBit": 31,
            "Name": "Trip Fan On Time",
            "Offset": 0.0,
            "OperationalHigh": 210554060.75,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 16,
            "Resolution": 0.05,
            "SPN": 994,
            "SPNLength": 32,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "hours"
        },
        995: {
            "Acronym": "TFI",
            "DataRange": "0 to 210,554,060.75 hr",
            "EndBit": 63,
            "Name": "Trip Fan On Time Due to the Engine System",
            "Offset": 0.0,
            "OperationalHigh": 210554060.75,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 16,
            "Resolution": 0.05,
            "SPN": 995,
            "SPNLength": 32,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "hours"
        },
        996: {
            "Acronym": "TFI",
            "DataRange": "0 to 210,554,060.75 hr",
            "EndBit": 95,
            "Name": "Trip Fan On Time Due to a Manual Switch",
            "Offset": 0.0,
            "OperationalHigh": 210554060.75,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 16,
            "Resolution": 0.05,
            "SPN": 996,
            "SPNLength": 32,
            "StartBit": 64,
            "TransmissionRate": "On request",
            "Units": "hours"
        },
        997: {
            "Acronym": "TFI",
            "DataRange": "0 to 210,554,060.75 hr",
            "EndBit": 127,
            "Name": "Trip Fan On Time Due to the A/C System",
            "Offset": 0.0,
            "OperationalHigh": 210554060.75,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 16,
            "Resolution": 0.05,
            "SPN": 997,
            "SPNLength": 32,
            "StartBit": 96,
            "TransmissionRate": "On request",
            "Units": "hours"
        },
        998: {
            "Acronym": "TDI",
            "DataRange": "0.0 to 327080468.2212549 mile",
            "EndBit": 31,
            "Name": "Trip Distance on VSL",
            "Offset": 0.0,
            "OperationalHigh": 327080468.2212549,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 12,
            "Resolution": 0.077671375,
            "SPN": 998,
            "SPNLength": 32,
            "StartBit": 0,
            "TransmissionRate": "On request",
            "Units": "miles"
        },
        999: {
            "Acronym": "TDI",
            "DataRange": "0.0 to 327080468.2212549 mile",
            "EndBit": 63,
            "Name": "Trip Gear Down Distance",
            "Offset": 0.0,
            "OperationalHigh": 327080468.2212549,
            "OperationalLow": 0.0,
            "OperationalRange": "",
            "PGNLength": 12,
            "Resolution": 0.077671375,
            "SPN": 999,
            "SPNLength": 32,
            "StartBit": 32,
            "TransmissionRate": "On request",
            "Units": "miles"
        }
    }

mfg_lookup = {
    0: "Reserved ",
    1: "Allied Signal Inc. Elyria, OH USA ",
    2: "Allison Transmission, GMC Indianapolis, IN USA ",
    3: "Ametek, US Gauge Division Sellersville, PA USA ",
    4: "Ametek-Dixson Grand Junction, CO USA ",
    5: "AMP Inc. Harrisburg, PA USA ",
    6: "Berifors Electronics AB Stockholm, Sweden ",
    7: "Case Corp. Burr Ridge, IL USA ",
    8: "Caterpillar Inc. Peoria, IL USA ",
    9: "Chrysler Corp. Auburn Hills, MI USA ",
    10: "Cummins Inc (formerly Cummins Engine Co) Columbus, IN USA ",
    11: "Dearborn Group Inc. Indianapolis, IN & Farmington Hills, MI USA ",
    12: "Deere & Company, Precision Farming East Moline, IL USA ",
    13: "Delco Electronics Kokomo, IN USA ",
    14: "Detroit Diesel Corporation Detroit, MI USA ",
    15: "Dickey-john Corp. Auburn, IL USA ",
    16: "Eaton Corp Southfield, MI USA ",
    17: "Eaton Corp, Corp Res & Dev Milwaukee, WI USA ",
    18: "Eaton Corp, Transmission Div. Kalamazoo, MI USA ",
    19: "Eaton Corp. Trucking Info Services Clemmons, NC USA ",
    20: "Eaton Ltd Worsley, England ",
    21: "Echlin Inc., Midland Brake Inc. Kansas City, MO USA ",
    22: "Ford Motor Co., Electronic Concepts & Systems Dearborn, MI USA ",
    23: "Ford Motor Co., Heavy Truck Dearborn, MI USA ",
    24: "Ford Motor Co., Vehicle Controls Dearborn, MI USA ",
    25: "Daimler Trucks North America LLC (formerly Freightliner LLC) Portland, OR USA ",
    26: "General Motors Corp, Service Technology Grp Romulus, MI USA ",
    27: "GMC Troy, MI USA ",
    28: "Grote Ind. Inc. Madison, IN USA ",
    29: "Hino Motors Ltd. Tokyo, Japan ",
    30: "Isuzu Motors Ltd Kawasaki, Japan ",
    31: "J Pollak Corp Boston, MA USA ",
    32: "Jacobs Vehicle Systems Bloomfield, CT USA ",
    33: "John Deere Waterloo, IA USA ",
    34: "Kelsey Hayes Co. Livonia, MI USA ",
    35: "Kenworth Truck Co. Kirkland, WA USA ",
    36: "Lucas Ind. Solihull WMidlnd, England ",
    37: "Mack Trucks Inc. Hagerstown, MD USA ",
    38: "Micro Processor Systems Inc. Sterling Hts, MI USA ",
    39: "Microfirm Inc. Stillwater, OK USA ",
    40: "Motorola AIEG Inc. Northbrook, IL USA ",
    41: "Motorola Inc. Schaumburg, IL USA ",
    42: "International Truck and Engine Corporation - Engine Electronics (formerly Navistar Intl Trans Co., Engine Electronics) Warrenville, IL USA ",
    43: "International Truck and Engine Corporation - Vehicle Warrenville, IL USA Electronics (formerly Navistar Intl Trans Corp.) ",
    44: "Nippondenso Co. Ltd. Kariya Aichi, Japan ",
    45: "PACCAR Mount Vernon, WA USA ",
    46: "Parasoft Computing Solutions Winston Salem, NC USA ",
    47: "Phillips Semiconductor Sunnyvale, CA USA ",
    48: "Pollak Alphabet El Paso, TX USA ",
    49: "RE America Inc. Cleveland, OH USA ",
    50: "Robert Bosch Corp Broadview, IL USA ",
    51: "Robert Bosch GmbH Stuttgart, Germany ",
    52: "Meritor Automotive, Inc. (formerly Rockwell Automotive) Troy, MI USA ",
    53: "Continental Automotive Systems US, Inc. (formerly Rockwell Land Transportation) Cedar Rapids, IA USA ",
    54: "Meritor Wabco Troy, MI USA ",
    55: "Ryder System Inc. Miami, FL USA ",
    56: "SAIC San Diego, CA USA ",
    57: "Sauer-Danfoss Co (formerly Sauer Sundstrand Co.) Minneapolis, MN USA ",
    58: "SPX Corporation, OTC Div Owatonna, MN USA ",
    59: "VES Inc. Rock Hill, SC USA ",
    60: "Volvo Trucks North America Inc. Greensboro, NC USA ",
    61: "Volvo Truck Corp. Gothenburg, Sweden ",
    62: "Wabco Hanover, Germany ",
    63: "ZF Industries Inc. Vernon Hills, IL USA ",
    64: "unused (formerly SpectraPrecision Laserplane, then SpectraPhysics) ",
    65: "MAN Nutzfahrzeuge AG Munich, Germany ",
    66: "John Deere Construction Equipment Division Dubuque, IA USA ",
    67: "John Deere Coffeyville Works (formerly Funk Manufacturing Company) Coffeyville, KS USA ",
    68: "Scania Södertälje, Sweden ",
    69: "Trimble Navigation ",
    70: "Flex-coil Limited Saskatoon, SK Canada ",
    71: "Vansco Electronics Ltd. Winnipeg, MAN Canada ",
    72: "Sisu Corporation ESPOO, Finland ",
    73: "LeTourneau, Inc. Longview, TX USA ",
    74: "Eaton Axle-Brake Division Kalamazoo, MI USA ",
    75: "Deere & Co, Agricultural Division ",
    76: "unused (formerly Deere & Co, Construction Division) ",
    77: "Deere Power Systems Group ",
    78: "Frank W. Murphy Manufacturing, Inc Tulsa, OK USA ",
    79: "Daimler Benz AG - Engine Division (PBM) Stuttgart, Germany ",
    80: "Twin Disc, Inc. Racine, WI USA ",
    81: "Fire Research Corp. Nesconset, NY USA ",
    82: "Bobcat/Ingersoll-Rand (formerly Melroe/Ingersoll-Rand) Fargo, ND USA ",
    83: "Eaton VORAD Technologies San Diego, CA USA ",
    84: "New Holland UK Limited Basildon, Essex, UK ",
    85: "Kohler Co Kohler, WI USA ",
    86: "C. E. Niehoff & Company Evanston, IL USA ",
    87: "J.C. Bamford Excavators Ltd (JCB) Rocester, Staffordshire, UK ",
    88: "Satloc Precision GPS Scottsdale, AZ USA ",
    89: "Kverneland Group, Electronics Division Nieuw-Vennep, Netherlands ",
    90: "Knorr-Bremse SfN GmbH Munich, Germany ",
    91: "BSG Bodensee Steuergeraete GmbH Immenstaad, Germany ",
    92: "Ag-Chem Equipment Co., Inc. Minnetonka, MN USA ",
    93: "Perkins Engines Company Ltd. Peterborough, UK ",
    94: "CNH Global N.V. Racine, WI USA ",
    95: "Pacific Insight Electronics Corp. Nelson, BC Canada ",
    96: "Mech@tronic IT GmbH Hetzenhausen, Germany ",
    97: "Ag Leader Technology, Inc. Ames, IA USA ",
    98: "Mueller-Elektronik GmbH & Co Salzkotten, Germany ",
    99: "International Transmissions Ltd (ITL) Wrexham, North Wales, UK ",
    100: "VDO Technik AG Rüthi, Switzerland ",
    101: "Sensoria San Diego, CA USA ",
    102: "AGCO GmbH & Co. Marktoberdorf, Germany ",
    103: "Agrocom GmbH & Co. Agrarsystem KG Bielefeld, Germany ",
    104: "Claas Selbstfahrende Erntemaschinen GmbH Harsewinkel, Germany ",
    105: "Kiepe Elektrik GmbH & Co. KG Duesseldorf, Germany ",
    106: "BAE Systems Controls, Inc. Johnson City, NY USA ",
    107: "Grimme Landmaschinen GmbH & Co. KG Damme, Germany ",
    108: "WTK-Elecktronik GmbH Neustadt, Germany ",
    109: "TeeJet Technologies Denmark (formerly LH TechnologiesApS) Aabybro, Denmark ",
    110: "EPIQ Sensor-Nite Fenton, MI USA ",
    111: "Maschinenfabrik Bernhard Krone GmbH Spelle, Germany ",
    112: "MECALAC Annecy le Vieux, France ",
    113: "Stress-Tek, Inc. Kent, WA USA ",
    114: "EControls, Inc. San Antonio, TX USA ",
    115: "NACCO Materials Handling Group, Inc. Portland, OR USA ",
    116: "BEELINE Technologies Brisbane, QLD Australia ",
    117: "HUSCO International Waukesha, WI USA ",
    118: "Intron GmbH Schwaebisch Hall, Germany ",
    119: "IntegriNautics Menlo Park, CA USA ",
    120: "RDS Technology Ltd Minchinhampton, Stroud, UK ",
    121: "HED (Hydro Electronic Devices, Inc.) Hartford, WI USA ",
    122: "FG Wilson (Engineering) Limited Larne, County Antrim, UK ",
    123: "Basler Electric Highland, IL USA ",
    124: "Hydac Electronic Saarbruecken, Germany ",
    125: "Nevada Automotive Test Center Carson City, NV USA ",
    126: "Driver Tech Salt Lake City, UT USA ",
    127: "Holland USA Holland, MI USA ",
    128: "Gerhard Duecker GmbH & Co. KG Stadtlohn, Germany ",
    129: "OMNEX Control Systems Inc. Port Coquitlam, BC, Canada ",
    130: "Nido - Universal Machines B.V. Holten, Netherlands ",
    131: "ITT Industries Eden Prarie, MN USA ",
    132: "Mulag-Fahrzeugwerk Oppenau, Germany ",
    133: "Bucher Schoerling GmbH Hannover, Germany ",
    134: "Iris Technology Ltd Lancaster, Lancs, UK ",
    135: "Airmar Technology Corporation Milford, NH USA ",
    136: "Komatsu Ltd Hiratsuka, Kanagawa, Japan ",
    137: "Maretron Phoenix AZ USA ",
    138: "Georg Fritzmeier GmbH & Co. KG Grosshelfendorf, Germany ",
    139: "Caterpillar Trimble Control Technologies (CTCT), LLC Dayton, OH USA ",
    140: "Lowrance Electronics, Inc. Tulsa, OK USA ",
    141: "Thales Navigation Ltd. Surrey, UK ",
    142: "TRW Automotive (Commercial Steering Systems) Lafayette, IN USA ",
    143: "W. Gmeiner GmbH & Co. Kummersbruck, Germany ",
    144: "Mercury Marine Fond du Lac, WI USA ",
    145: "MurCal Controls Palmdale, CA USA ",
    146: "Maxima Technologies Lancaster, PA USA ",
    147: "Nautibus electronic GmbH Quern, Germany ",
    148: "Blue Water Data, Inc. Salem, NJ USA ",
    149: "Holset Charleston, SC USA ",
    150: "Fleetguard Nashville, TN USA ",
    151: "Raven Industries Inc. - Flow Controls Division Sioux Falls, SD USA ",
    152: "elobau GmbH & Co. KG Leutkirch, Germany ",
    153: "Woodward, Industrial Controls Division Fort Collins, CO USA ",
    154: "Westerbeke Corporation Taunton, MA USA ",
    155: "Vetronix Corporation Santa Barbara, CA ",
    156: "ITT Industries - Cannon Shakopee, MN USA ",
    157: "ISSPRO Inc. Portland, OR USA ",
    158: "Firestone Industrial Products Company Carmel, IN USA ",
    159: "NTech Industries Inc Ukiah, CA USA ",
    160: "Nido Holten, Netherlands ",
    161: "Offshore Systems (UK) Ltd New Milton, Hampshire, UK ",
    162: "Axiomatic Technologies Mississauga, ON Canada ",
    163: "BRP Inc. Waukegan, IL USA ",
    164: "MTU Friedrichshafen GmbH (formerly DaimlerChrysler Friedrichshafen,Germany Off-Highway)",
    165: "CPAC Systems AB Vastra Frolunda, Sweden ",
    166: "Phoenix International Fargo, ND USA ",
    167: "JLG Industries Inc McConnellsburg, PA USA ",
    168: "Xantrex Burnaby, BC Canada ",
    169: "Marlin Technologies Inc. Horicon, WI USA ",
    170: "Computronics Corporation Ltd. Bentley, WA Australia ",
    171: "Wachendorff Elektronik GmbH & Co. KG Geisenheim, Germany ",
    172: "Yanmar Marine USA Adairsville, GA USA ",
    173: "Ryeso, Inc. Palmdale, CA USA ",
    174: "AB Volvo Penta Goteburg, Sweden ",
    175: "Veris Technologies, Inc. Salina, KS USA ",
    176: "Moritz Aerospace Dublin, PA USA ",
    177: "Diagnostic Systems Associates Kalamazoo, MI USA ",
    178: "Siemens VDO Automotive AG Regensburg, Germany ",
    179: "TeeJet Technologies Springfield (formerly Midwest Springfield, IL USA Technologies Illinois, LLC) ",
    180: "Smart Power Systems Reed City, MI  USA ",
    181: "Coretronics, Inc. Eagle, ID USA ",
    182: "Vehicle Systems Engineering B.V. Veenendaal, Netherlands ",
    183: "KDS Controls, Inc Troy, MI USA ",
    184: "EIA Electronics Aartselaar, Belguim ",
    185: "Beede Electrical Instrument Company Penacock, NH USA ",
    186: "Altronic, Inc Girard, OH USA ",
    187: "Air-Weigh Eugene, OR USA ",
    188: "EMP Corp Escanaba, MI USA ",
    189: "QUALCOMM San Diego, CA USA ",
    190: "Hella KGaA Hueck & Co Lippstadt, Germany ",
    191: "XATA Corporation Burnsville, MN USA ",
    192: "Floscan Seattle, WA USA ",
    193: "Jeppesen Marine Portland, OR USA ",
    194: "TriMark Corporation New Hampton, IA USA ",
    195: "General Engine Products Livonia, MI USA ",
    196: "LEMKEN GmbH & Co KG Alpen, Germany ",
    197: "Mechron Power Systems Ottawa, ON Canada ",
    198: "Mystic Valley Communications Mystic, CT USA ",
    199: "Actia Corp Elkhart, IN USA ",
    200: "MGM Brakes Charlotte, NC USA ",
    201: "Disenos y Tecnologia S.A. Barcelona, Spain ",
    202: "Curtis Instruments, Inc Mount Kisco, NY USA ",
    203: "MILtronik GmbH Hilden, Germany ",
    204: "The Morey Corporation Woodridge, IL USA ",
    205: "SmarTire Systems Inc Richmond, BC Canada ",
    206: "port GmbH Halle, Germany ",
    207: "Otto Engineering Carpentersville, IL USA ",
    208: "Drew Technologies, Inc Whitmore Lake, MI USA ",
    209: "Bell Equip. Co. SA (PTY) LTD Richards Bay, South Africa ",
    210: "Iteris, Inc. Anaheim, CA USA ",
    211: "DNA Group Raleigh, NC USA ",
    212: "Sure Power Industries, Inc Tualatin, OR USA ",
    213: "CNH Belgium N.V. Zedelgem, Belgium ",
    214: "MC elettronica Srl Fiesso Umbertiano, Rovigo, Italy ",
    215: "Aetna Engineering/Fireboy-Xintex Grand Rapids, MI USA ",
    216: "Paneltronics Inc. Hialeah Gardens, FL USA ",
    217: "RM Michaelides Software & Elektronik GmbH Fulda, Germany ",
    218: "Gits Manufacturing Company Creston, IA USA ",
    219: "Cat OEM Solutions Mossville, IL USA ",
    220: "Beede Electrical Instrument Company, Inc Penacook, NH USA ",
    221: "SiE Kempten/Allgaeu, Germany ",
    222: "Generac Power Systems, Inc. Waukesha, WI, USA ",
    223: "Vaueo Retarder Co., Ltd. Shajin, Baoan, Shen Zhen, China ",
    224: "EMMI Network, S.L. Gandia, Spain ",
    225: "SKF Gothenburg, Sweden ",
    226: "Monaco Coach Corporation Bend, OR USA ",
    227: "Lykketronic A/S Loegstoer, Denmark ",
    228: "ZF Marine Electronics Mukilteo, WA USA ",
    229: "Garmin International Inc. Olathe, KS USA ",
    230: "Saucon Technologies Bethlehem, PA USA ",
    231: "Topcon Positioning Systems, Inc. Livermore, CA USA ",
    232: "TSD Integrated Controls Minneapolis, MN USA ",
    233: "Yacht Monitoring Solutions Inc. Corpus Christi, TX USA ",
    234: "Mondial electronic GmbH Eggelsberg, Austria ",
    235: "SailorMade Marine Telemetry - Tetra Technology Ltd. Florianopolis, Brazil ",
    236: "Norac Systems International Inc Saskatoon, SK, Canada ",
    237: "Agtron Enterprises Inc Saskatoon, SK, Canada ",
    238: "ZF Friedrichshafen AG Friedrichshafen, Germany ",
    239: "May & Scofield Ltd Basingstoke, Hants, UK ",
    240: "Vanair Mfg New Buffalo, MI USA ",
    241: "Schneider Automation SAS Sophia Antipolis, France ",
    242: "Kokusandenki Co. Ltd Numazu, Shizuoka, Japan ",
    243: "eRide, Inc. San Francisco, CA USA ",
    244: "Techno-Matic Aars, Denmark ",
    245: "Capstan Ag Systems, Inc. Topeka, KS USA ",
    246: "Class 1, Inc Ocala, FL USA ",
    247: "ePULSE Troy, MI USA ",
    248: "Cooper Standard Automotive Active Systems Group Mitchell, ON Canada ",
    249: "Schaltbau GmbH Munich, Germany ",
    250: "Kuhn Group Saverne, France ",
    251: "German Agricultural Society, Test Center (DLG) Gross-Ultstadt, Germany ",
    252: "Sensor-Technik Wiedemann GmbH Kaufbeuren, Germany ",
    253: "Mobile Control Systems Brussels, Belgium ",
    254: "GE Sensing St. Marys, PA USA ",
    255: "MEAS France Toulouse, France ",
    256: "Tyco Electronics AMP Osan-Si, Kyunggi-Do, Korea ",
    257: "Honda Motor Co., Ltd Wako-city, Saitama, Japan ",
    258: "ARAG Rubiera, Italy ",
    259: "Jetter AG Ludwigsburg, Germany ",
    260: "Reichhardt GmbH Steuerungstechnik Hungen, Germany ",
    261: "Red Dot Corporation Seattle, WA USA ",
    262: "HydraForce, Inc Lincolnshire, IL USA ",
    263: "IMMI Westfield, IN USA ",
    264: "Autolync Clayton, Victoria, Australia ",
    265: "MTS Sensor Technologie GmbH Luedenscheid, Germany ",
    266: "International Thermal Research Ltd Richmond, BC Canada ",
    267: "Red Lion Controls, Inc York, PA USA ",
    268: "Accurate Technologies Wixom, MI USA ",
    269: "Saft America Inc., Space & Defense Division Cockeysville, MD USA ",
    270: "Tennant Minneapolis, MN USA ",
    271: "Cole Hersee South Boston, MA USA ",
    272: "Gross Mechanical Laboratories, Inc Hanover, MD USA ",
    273: "Active Research Limited Poole, Dorset, UK ",
    274: "LTW Technology Co., LTD Shinjuang City, Taipei Hsien, Taiwan, R.O.C. ",
    275: "Navico Egersund AS Egersund, Norway ",
    276: "Aqua-Hot Heating Systems Fort Lupton, CO USA ",
    277: "LHP Telematics Columbus, IN USA ",
    278: "Takata Electronics Farmington Hills, MI USA ",
    279: "Geometris LP Houston, TX USA ",
    280: "Leica Geosystems Pty Ltd Brisbane, Queensland Australia ",
    281: "Precision Governors LLC Rockford, IL USA ",
    282: "Medallion Instrumentation Systems Spring Lake, MI USA ",
    283: "CWF Hamilton & Co Ltd Christchurch, New Zealand ",
    284: "Mobile Tech Chile Concepcion, Bio Bio, Chile ",
    285: "Sea Recovery Corp Rancho Dominguez, CA USA ",
    286: "Coelmo srl Acerra, Italy ",
    287: "NTech Industries, Inc Ukiah, CA USA ",
    288: "Mitsubishi FUSO Truck & Bus Corp. Kawasaki, Kanagawa, Japan ",
    289: "Watlow St. Louis, MO USA ",
    290: "Kuebler GmbH VS-Schwenningen, Germany ",
    291: "Groeneveld Transport Efficiency BV Gorinchem, Netherlands ",
    292: "IKUSI - Angel Iglesias S.A. San Sebastian, Guipuzcoa, Spain ",
    293: "Spyder Controls Corp. Lacombe, AB Canada ",
    294: "Grayhill Inc. LaGrange, IL USA ",
    295: "BEP Marine Albany, North Shore City, New Zealand ",
    296: "micro dynamics GmbH Cologne, Germany ",
    297: "Zonar Systems Inc Seattle, WA USA ",
    298: "Holley Performance Bowling Green, KY USA ",
    299: "Rauch Sinzheim, Germany ",
    300: "Systron Donner Automotive Concord, CA USA ",
    301: "Parker Hannifin (UK) Ltd FDE group Thetford, Norfolk, UK ",
    302: "Nissin Kogyo Co., LTD Ueda-city, Nagano prefecture, Japan ",
    303: "CTS Corporation Elkhart, IN USA ",
    304: "EmpirBus AB Uddevalla, Sweden ",
    305: "NovAtel Inc Calgary, AB Canada ",
    306: "Sleipner Motor AB Fredrikstad, Norway ",
    307: "MAS Technologies Harleysville, PA USA ",
    308: "Cyntrx De Pere, WI USA ",
    309: "Krauss-Maffei Wegmann GmbH & Co. KG Munich, Germany ",
    310: "TECNORD srl Modena, Italy ",
    311: "Patrick Power Products Elkridge, MD USA ",
    312: "Lectronix, Inc. Lansing, MI USA ",
    313: "Ilmor Engineering, Inc. Plymouth, MI USA ",
    314: "CSM GmbH Filderstadt, Germany ",
    315: "Icom Incorporated Osaka, Japan ",
    316: "ITT Flow Control Santa Ana, CA USA ",
    317: "Navtronics Bvba Geel, Belgium ",
    318: "SAT Plan Beauvais Tillé, Oise, France ",
    319: "Cadec Global Manchester, NH USA ",
    320: "Miedema Landbouwwerktuigenfabriek B.V. Winsum, Netherlands ",
    321: "Ultra Electronics, Electrics Cheltenham, UK ",
    322: "MICHENKA s.r.o. Koprivnice, Czech Republic ",
    323: "Mobileye Vision Technologies Ltd. Jerusalem, Israel ",
    324: "Snap-on Diagnostics Lincolnshire, IL USA ",
    325: "ASM Automation Sensorik Messtechnik GmbH Moosinning, Germany ",
    326: "Akron Brass Company Wooster, OH USA ",
    327: "Sonceboz SA Sonceboz, Switzerland ",
    328: "Qwerty Electronik AB Bagarmossen, Sweden ",
    329: "Deif A/S Skive, Denmark ",
    330: "Kidde Aerospace and Defense Wilson, NC USA ",
    331: "Horton, Inc. Roseville, MN USA ",
    332: "HWH Corporation Moscow, IA USA ",
    333: "Hadley Products Corporation Grandville, MI USA ",
    334: "Takata-Petri AG Aschaffenburg, Germany ",
    335: "Evo Electric Ltd Woking, Surrey, UK ",
    336: "APE s.r.o. Bradlec, Czech Republic ",
    337: "Carraro S.p.A. Campodarsego, Italy ",
    338: "GRAF-SYTECO Tuningen, Germany ",
    339: "Competence Center ISOBUS e.V. Wallenhorst, Germany ",
    340: "Continental AG Hannover, Germany ",
    341: "Boning GmbH & Co Ganderkesee, Germany ",
    342: "THOMAS MAGNETE GmbH Herdorf, Germany ",
    343: "Baumer Group Villingen-Schwenningen, Germany ",
    344: "Parvus Corporation Salt Lake City, UT USA ",
    345: "Korean Maritime University Busan, Korea ",
    346: "Control Solutions Aurora, IL USA ",
    347: "Honeywell Morristown, NJ USA ",
    348: "Amazonen Werke H. Dreyer Hasbergen, Germany ",
    349: "Suonentieto Kuopio, Finland ",
    350: "Noris Marine Systems GmbH & Co KG Nuremberg, Germany ",
    351: "Thrane & Thrane Gentofte, Denmark ",
    352: "SAME DEUTZ-FAHR Treviglio, Italy ",
    353: "Hegemon Electronics, Inc. Sterling Heights, MI USA ",
    354: "Junkkari OY Yliharma, Finland ",
    355: "Mastervolt International B.V. Amsterdam, Netherlands ",
    356: "Fischer Panda Generators, Inc Oakland Park, FL USA ",
    357: "Hardi International A/S Taastrup, Denmark ",
    358: "Victron Energy BV Almere-Haven, Netherlands ",
    359: "Ludwig Bergmann GmbH Goldenstedt, Germany ",
    360: "HJS Fahrzeugtechnik GmbH & Co KG Menden, Germany ",
    361: "InMach Ulm, Germany ",
    362: "Alois Poettinger Maschinenfabrik GmbH Grieskirchen, Austria ",
    363: "BEI Duncan Goleta, CA USA ",
    364: "OEM Controls Inc. Shelton, CT USA ",
    365: "Digi-Star, LLC Fort Atkinson, WI USA ",
    366: "Viewnyx Corp. Kanata, ON Canada ",
    367: "Fliegl Agrartechnik Toeging, Germany ",
    368: "HANSENHOF_electronic Reifland, Germany ",
    369: "Power Torque Engineering Ltd Coventry, UK ",
    370: "Rolls-Royce Marine AS Aalesund, Norway ",
    371: "Heinzmann GmbH & Co. KG Schoenau, Germany ",
    372: "Delphi Auburn Hills, MI USA ",
    373: "Electronic Design Inc. Sheboygan, WI USA ",
    374: "Northern Lights, Inc. Seattle, WA USA ",
    375: "Williams Controls Inc. Portland, OR USA ",
    376: "Quake Global San Diego, CA USA ",
    377: "ifm electronic gmbh Essen, Germany ",
    378: " Glendinning Marine Products Conway, SC USA ",
    379: " Yamabiko Corporation Iwate, Japan ",
    380: " Suntech International Ltd. Seoul, Korea ",
    1850: "Teleflex Limerick, PA USA ",
    1851: "RayMarine Portsmouth, Hampshire, UK ",
    1852: "Navionics Wareham, MA USA ",
    1853: "Japan Radio Co Seattle, WA USA ",
    1854: "Northstar Technologies Acton, MA USA ",
    1855: "Furuno USA Camas, WA USA ",
    1856: "Trimble Sunnyvale, CA USA ",
    1857: "Simrad Egersund, Norway ",
    1858: "Litton Charlottesville, VA USA ",
    1859: "Kvasar AB Kinnahult, Sweden ",
    1860: "MMP Fircrest, WA USA ",
    1861: "Vector Cantech Novi, MI USA ",
    1862: "Sanshin Shizuoka, Japan ",
    1863: "Thomas G. Faria Co. Uncasville, CT USA ",
}
